package cz.cuni.jagrlib;

import java.util.*;

/**
 * Simple {@link Channel} implementation. <tt>HashSet</tt>s are used for left- and
 * right-side connections.
 *
 * @version 0.25 $Rev: 397 $ $Date: 2007-01-27 18:43:03 +0100 (so, 27 I 2007) $ $Author: pepca $
 * @since   0.01
 * @see     <a href="../../../../cz/cuni/jagrlib/BasicChannel.java">BasicChannel.java</a>
 * @todo    Proxy-interfaces (more than one {@link Piece} called from one {@link Channel}).
 */
public class BasicChannel implements Channel
{
  /**
   * Fully-qualified interface name of the left side (called from the right side).
   */
  protected String leftInterfaceName;

  /**
   * Interface object for the left-side (called by right-side {@link Piece}[s]).
   */
  protected Object leftInterface;

  /**
   * Set of {@link Plug}s connected to the left side.
   */
  protected HashSet< Plug > leftSet;

  /**
   * Fully-qualified interface name of the right side (called from the left side).
   */
  protected String rightInterfaceName;

  /**
   * Interface object for the right-side (called by left-side {@link Piece}[s]).
   */
  protected Object rightInterface;

  /**
   * Set of {@link Plug}s connected to the right side.
   */
  protected HashSet< Plug > rightSet;

  /**
   * Initializing constructor.
   *
   * @param  _leftInterfaceName Fully-qualified interface name of the left side (called
   *                            from the right side).
   * @param  _rightInterfaceName Fully-qualified interface name of the right side (called
   *                             from the left side).
   * @see    #setInterfaceNames
   */
  public BasicChannel ( String _leftInterfaceName, String _rightInterfaceName )
  {
    leftSet  = new HashSet< Plug >();
    rightSet = new HashSet< Plug >();
    setInterfaceNames( _leftInterfaceName, _rightInterfaceName );
  }

  /**
   * (Re-)initialize both interface names.
   *
   * @param  _leftInterfaceName Fully-qualified interface name of the left side (called
   *                            from the right side).
   * @param  _rightInterfaceName Fully-qualified interface name of the right side (called
   *                             from the left side).
   * @see    #getLeftInterfaceName
   * @see    #getRightInterfaceName
   */
  public void setInterfaceNames ( String _leftInterfaceName, String _rightInterfaceName )
  {
    leftInterfaceName  = _leftInterfaceName;
    rightInterfaceName = _rightInterfaceName;
  }

  /**
   * Returns an interface name of the left Channel-side.
   *
   * @return Fully-qualified interface name of the left side (called from the right side).
   * @see    #setInterfaceNames
   * @see    #getRightInterfaceName
   */
  public String getLeftInterfaceName ()
  {
    return leftInterfaceName;
  }

  /**
   * Returns an interface name of the right Channel-side.
   *
   * @return Fully-qualified interface name of the right side (called from the left side).
   * @see    #setInterfaceNames
   * @see    #getLeftInterfaceName
   */
  public String getRightInterfaceName ()
  {
    return rightInterfaceName;
  }

  /**
   * Returns an interface name of the same Channel-side.
   *
   * @param  proPlug The {@link Plug} connected to the same side.
   * @return Fully-qualified interface name of the given Channel-side.
   * @see    #getInterfaceNameAnti
   */
  public String getInterfaceNamePro ( Plug proPlug )
  {
    if ( leftSet.contains( proPlug ) )
      return leftInterfaceName;

    if ( rightSet.contains( proPlug ) )
      return rightInterfaceName;

    return null;
  }

  /**
   * Returns an interface name of the opposite Channel-side.
   *
   * @param  antiPlug The {@link Plug} connected to the opposite side.
   * @return Fully-qualified interface name of the given Channel-side.
   * @see    #getInterfaceNamePro
   */
  public String getInterfaceNameAnti ( Plug antiPlug )
  {
    if ( leftSet.contains( antiPlug ) )
      return rightInterfaceName;

    if ( rightSet.contains( antiPlug ) )
      return leftInterfaceName;

    return null;
  }

  /**
   * Sets new interface object for the left Channel-side.
   *
   * @param  newLeftInterface New interface object (= object called from the opposite side).
   * @return The old interface object or {@code null}.
   * @see    #setRightInterface
   * @see    #setInterface
   */
  public Object setLeftInterface ( Object newLeftInterface )
  {
    Object old    = leftInterface;
    leftInterface = newLeftInterface;
    return old;
  }

  /**
   * Sets new interface object for the right Channel-side.
   *
   * @param  newRightInterface New interface object (= object called from the opposite side).
   * @return The old interface object or {@code null}.
   * @see    #setLeftInterface
   * @see    #setInterface
   */
  public Object setRightInterface ( Object newRightInterface )
  {
    Object old     = rightInterface;
    rightInterface = newRightInterface;
    return old;
  }

  /**
   * Sets new interface object for the given Channel-side. The side is defined by the
   * {@link Plug} connected to it.
   *
   * @param  newInterface New interface object (= object called from the opposite side).
   * @param  called The {@link Plug} defining the side.
   * @return The old interface object or {@code null}.
   * @see    #setLeftInterface
   * @see    #setRightInterface
   */
  public Object setInterface ( Object newInterface, Plug called )
  {
    if ( isLeft( called ) )
      return setLeftInterface( newInterface );

    return setRightInterface( newInterface );
  }

  /**
   * Returns the interface object of the left Channel-side.
   *
   * @return The interface object or {@code null}.
   * @see    #setLeftInterface
   */
  public Object getLeftInterface ()
  {
    return leftInterface;
  }

  /**
   * Returns the interface object of the right Channel-side.
   *
   * @return The interface object or {@code null}.
   * @see    #setRightInterface
   */
  public Object getRightInterface ()
  {
    return rightInterface;
  }

  /**
   * Returns the interface object of the given Channel-side. The side is defined by the
   * {@link Plug} which calls it.
   *
   * @param  caller The {@link Plug} defining the opposite (caller) side.
   * @return The interface object or {@code null}.
   * @see    #setInterface
   */
  public Object getInterface ( Plug caller )
  {
    if ( isLeft( caller ) )
      return rightInterface;

    if ( isRight( caller ) )
      return leftInterface;

    return null;
  }

  /**
   * Connects a new {@link Plug} to the left Channel-side.
   * Updates the {@link Plug}'s {@link Piece}, too.
   *
   * @param  newPlug The {@link Plug} to be connected.
   * @see    #connectRight
   */
  public void connectLeft ( Plug newPlug )
  {
    newPlug.disconnectChannel();
    newPlug.connectChannel( this );
    leftSet.add( newPlug );
  }

  /**
   * Connects a new {@link Plug} to the right Channel-side.
   *
   * @param  newPlug The {@link Plug} to be connected.
   * @see    #connectLeft
   */
  public void connectRight ( Plug newPlug )
  {
    newPlug.disconnectChannel();
    newPlug.connectChannel( this );
    rightSet.add( newPlug );
  }

  /**
   * Connects a new {@link Plug} to the same Channel-side.
   *
   * @param  newPlug The {@link Plug} to be connected.
   * @param  proPlug The {@link Plug} connected to the same side.
   * @see    #connectAnti
   */
  public void connectPro ( Plug newPlug, Plug proPlug )
  {
    if ( isLeft( proPlug ) )
      connectLeft( newPlug );
    else
      connectRight( newPlug );
  }

  /**
   * Connects a new {@link Plug} to the opposite Channel-side.
   *
   * @param  newPlug The {@link Plug} to be connected.
   * @param  antiPlug The {@link Plug} connected to the opposite side.
   * @see    #connectPro
   */
  public void connectAnti ( Plug newPlug, Plug antiPlug )
  {
    if ( isLeft( antiPlug ) )
      connectRight( newPlug );
    else
      connectLeft( newPlug );
  }

  /**
   * Disconnects the old {@link Plug} from the {@link Channel}.
   *
   * @param  oldPlug The {@link Plug} to be disconnected.
   * @see    #disconnectAll
   */
  public void disconnect ( Plug oldPlug )
  {
    if ( leftSet.remove(  oldPlug ) ||
         rightSet.remove( oldPlug ) )
      oldPlug.disconnectChannel();
  }

  /**
   * Disconnects all {@link Plug}s from the {@link Channel}.
   *
   * @see    #disconnect
   */
  public void disconnectAll ()
  {
    int i;

      // left side:
    Plug[] remove = getLeft();
    if ( remove != null )
      for ( i = 0; i < remove.length; i++ )
        remove[i].disconnectChannel();
    leftSet.clear();

      // right side:
    remove = getRight();
    if ( remove != null )
      for ( i = 0; i < remove.length; i++ )
        remove[i].disconnectChannel();
    rightSet.clear();
  }

  /**
   * Returns the left Channel-side count.
   *
   * @return The left-side count (number of {@link Plug}s connected to that side).
   * @see    #getRightCount
   */
  public int getLeftCount ()
  {
    return leftSet.size();
  }

  /**
   * Returns the right Channel-side count.
   *
   * @return The right-side count (number of {@link Plug}s connected to that side).
   * @see    #getLeftCount
   */
  public int getRightCount ()
  {
    return rightSet.size();
  }

  /**
   * Is the given {@link Plug} connected to the left Channel-side?
   *
   * @param  oldPlug The checked {@link Plug}.
   * @return <tt>true</tt> if the given {@link Plug} is connected to the left side.
   * @see    #isRight
   */
  public boolean isLeft ( Plug oldPlug )
  {
    return leftSet.contains( oldPlug );
  }

  /**
   * Is the given {@link Plug} connected to the right Channel-side?
   *
   * @param  oldPlug The checked {@link Plug}.
   * @return <tt>true</tt> if the given {@link Plug} is connected to the right side.
   * @see    #isLeft
   */
  public boolean isRight ( Plug oldPlug )
  {
    return rightSet.contains( oldPlug );
  }

  /**
   * Returns the array of left-side objects ({@link Plug}s).
   *
   * @return The array of {@link Plug}s connected to the left side.
   */
  public Plug[] getLeft ()
  {
    if ( leftSet.size() == 0 ) return null;

    Plug[] init = new Plug[ leftSet.size() ];
    return (Plug[])leftSet.toArray( init );
  }

  /**
   * Returns the array of right-side objects ({@link Plug}s).
   *
   * @return The array of {@link Plug}s connected to the right side.
   */
  public Plug[] getRight ()
  {
    if ( rightSet.size() == 0 ) return null;

    Plug[] init = new Plug[ rightSet.size() ];
    return (Plug[])rightSet.toArray( init );
  }

  /**
   * Returns the array of same-side objects ({@link Plug}s).
   *
   * @param  proPlug The {@link Plug} connected to the same side.
   * @return The array of {@link Plug}s.
   */
  public Plug[] getPro ( Plug proPlug )
  {
    if ( isLeft( proPlug )  ) return getLeft();

    if ( isRight( proPlug ) ) return getRight();

    return null;
  }

  /**
   * Returns the array of opposite-side objects ({@link Plug}s).
   *
   * @param  antiPlug The {@link Plug} connected to the opposite side.
   * @return The array of {@link Plug}s.
   */
  public Plug[] getAnti ( Plug antiPlug )
  {
    if ( isLeft( antiPlug )  ) return getRight();

    if ( isRight( antiPlug ) ) return getLeft();

    return null;
  }

  /**
   * Prepares the {@link Channel} to regular operation.
   *
   * @return <tt>true</tt> if everything is OK ({@link Channel} is prepared
   *         for operation).
   * @todo   Proxy-interfaces (more than one {@link Piece} called from one {@link Channel}).
   */
  public boolean prepare ()
  {
    if ( getLeftInterfaceName() != null )
    {                                     // the left interface object should be there
      if ( getLeftCount() != 1 ) return false;
                                          // proxy-interfaces are not implemented yet
      Plug[] plug = getLeft();
      setLeftInterface( plug[0].getInputInterface() );
    }

    if ( getRightInterfaceName() != null )
    {                                     // the right interface object should be there
      if ( getRightCount() != 1 ) return false;
                                          // proxy-interfaces are not implemented yet
      Plug[] plug = getRight();
      setRightInterface( plug[0].getInputInterface() );
    }

    return true;
  }

}
