package cz.cuni.jagrlib;

/**
 * Basic {@link Plug} implementation - can be used for <b>internal</b> {@link Plug}s -
 * inside executed <tt>Composition</tt>.
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.04
 * @see     <a href="../../../../cz/cuni/jagrlib/BasicPlug.java">BasicPlug.java</a>
 * @see     Plug
 * @see     Piece
 * @see     Channel
 */
public class BasicPlug implements Plug
{
  /**
   * Is that {@link Plug} mandatory?
   */
  protected boolean mandatory;

  /**
   * Can that {@link Plug} be used in multi-channels?
   *
   * @see    Channel
   */
  protected boolean multi;

  /**
   * Is that {@link Plug} cloneable (=generic)?
   *
   * @see    #key
   */
  protected boolean cloneable;

  /**
   * Fully qualified input <tt>interface</tt> name (implemented <tt>interface</tt>).
   */
  protected String inputInterface;

  /**
   * Fully qualified output <tt>interface</tt> name (required <tt>interface</tt>).
   */
  protected String outputInterface;

  /**
   * Associated {@link Template}.
   *
   * @see    #key
   */
  protected Template template;

  /**
   * Key string used in associated {@link Template}.
   *
   * @see    #template
   */
  protected String key;

  /**
   * Assign constructor - initializes a new <tt>BasicPlug</tt> object.
   *
   * @param  _mandatory Is that plug mandatory (has to be connected)?
   * @param  _multi Can that plug be used in multi-channels?
   * @param  _cloneable Is that plug "cloneable" (=generic)? If yes,
   *                    ordinal numbers (starting from 0) will be appended to its name.
   * @param  _inputInterface Fully qualified identifier of input (implemented)
   *                         <tt>interface</tt>.
   * @param  _outputInterface Fully qualified identifier of output (required)
   *                          <tt>interface</tt>.
   * @param  _template {@link Template} this plug is to be associated with.
   * @param  _key Key used in associated {@link Template}.
   */
  public BasicPlug ( boolean _mandatory, boolean _multi, boolean _cloneable,
                     String _inputInterface, String _outputInterface,
                     Template _template, String _key )
  {
    template        = _template;
    key             = _key;
    mandatory       = _mandatory;
    multi           = _multi;
    cloneable       = _cloneable;
    inputInterface  = _inputInterface;
    outputInterface = _outputInterface;
  }

  /**
   * Has this {@link Plug} the <b>input</b> capability (can be called from outside)?
   * @return <tt>true</tt> if this {@link Plug} can be called from outside.
   *
   * @see    #getInputInterfaceName
   * @see    #getInputInterface
   */
  public boolean isInput ()
  {
    return( inputInterface != null );
  }

  /**
   * Has this {@link Plug} the <b>output</b> capability (can call other interfaces)?
   *
   * @return <tt>true</tt> if this {@link Plug} can call other interfaces.
   * @see    #getOutputInterfaceName
   */
  public boolean isOutput ()
  {
    return( outputInterface != null );
  }

  /**
   * Is this {@link Plug} mandatory?
   *
   * @return <tt>true</tt> if this {@link Plug} is <b>mandatory</b> (has to be
   *         connected).
   */
  public boolean isMandatory ()
  {
    return mandatory;
  }

  /**
   * Can this {@link Plug} be used in <b>multi-channels</b>?
   *
   * @return <tt>true</tt> if this {@link Plug} can be connected to <b>multi-channel</b>.
   */
  public boolean isMulti ()
  {
    return multi;
  }

  /**
   * Can this {@link Plug} be <b>cloned</b>? For filters with variable number of (input)
   * channels.
   * <br>Sibling channels should have access <b>key</b>s with common prefix -
   * integer numbers starting from <b>0</b> will be appended.
   *
   * @return <tt>true</tt> if this {@link Plug} can be <b>cloned</b>.
   * @see    #getKey
   */
  public boolean isCloneable ()
  {
    return cloneable;
  }

  /**
   * Returns the fully-qualified name of an internal (implemented) interface - to be called from
   * outside.
   *
   * @return Fully-qualified name of the interface or <tt>null</tt> if it is not used.
   * @see    #isInput
   */
  public String getInputInterfaceName ()
  {
    return inputInterface;
  }

  /**
   * Returns the fully-qualified name of an external (required) interface - to be called from
   * this {@link Piece}.
   *
   * @return Fully-qualified name of the interface or <tt>null</tt> if it is not used.
   * @see    #isOutput
   */
  public String getOutputInterfaceName ()
  {
    return outputInterface;
  }

  /**
   * Returns the internal (implemented) interface object - to be called from outside.
   *
   * @return The internal (implemented) interface object or <tt>null</tt>.
   */
  public Object getInputInterface ()
  {
    if ( getTemplate() == null ) return null;

    return ((Piece)getTemplate()).myInterface( this );
  }

  /**
   * Gets the associated {@link Template}.
   * For topological purposes only.
   *
   * @return The associated {@link Template} or <tt>null</tt>.
   * @see    #getKey
   */
  public Template getTemplate ()
  {
    return template;
  }

  /**
   * Gets the associated {@link Piece}.
   * For topological purposes only.
   *
   * @return The associated {@link Piece} or <tt>null</tt>.
   * @see    #getKey
   */
  public Piece getPiece ()
  {
    return (Piece)template;
  }

  /**
   * Gets the access {@link #key} from associated {@link Template} to this {@link Plug}.
   * For topological purposes only.
   *
   * @return The access <b>key</b> or <tt>null</tt>.
   * @see    #getTemplate
   */
  public String getKey ()
  {
    return key;
  }

  /**
   * Compatibility test: Am I compatible with the given parameters?
   *
   * @param  iInterface Required input interface (or <tt>null</tt>).
   * @param  oInterface Required output interface (or <tt>null</tt>).
   * @param  cl Is the <b>cloneability</b> required?.
   * @param  mul Is the <b>multi-channel</b> property required?.
   * @return <tt>true</tt> if I'm able to implement the given interface.
   */
  public boolean isCompatible ( String iInterface, String oInterface,
                                boolean cl, boolean mul )
  {
    if ( cl && !isCloneable() ) return false;

    if ( mul && !isMulti() ) return false;

    if ( (iInterface == null) !=
         (getInputInterfaceName() == null) ) return false;

    if ( iInterface != null &&
         !iInterface.equals( getInputInterfaceName() ) ) return false;

    if ( (oInterface == null) !=
         (getOutputInterfaceName() == null) ) return false;

    return( oInterface == null ||
            oInterface.equals( getOutputInterfaceName() ) );
  }

  /**
   * Compatibility test: Am I compatible with the given {@link Plug}?
   *
   * @param  pl The checked {@link Plug}.
   * @return <tt>true</tt> if I'm comptible with the given {@link Plug}.
   */
  public boolean isCompatible ( Plug pl )
  {
    if ( (pl.getInputInterfaceName() == null) !=
         (getInputInterfaceName() == null) ) return false;

    if ( getInputInterfaceName() != null &&
         !pl.getInputInterfaceName().equals( getInputInterfaceName() ) ) return false;

    if ( (pl.getOutputInterfaceName() == null) !=
         (getOutputInterfaceName() == null) ) return false;

    return( getOutputInterfaceName() == null ||
            pl.getOutputInterfaceName().equals( getOutputInterfaceName() ) );
  }

  /**
   * Compatibility test: Am I compatible with the opposite {@link Plug}?
   *
   * @param  opposite The opposite {@link Plug}.
   * @return <tt>true</tt> if I'm compatible with the opposite {@link Plug}.
   */
  public boolean isOpposite ( Plug opposite )
  {
    String myIn   = getInputInterfaceName();
    String myOut  = getOutputInterfaceName();
    String oppIn  = opposite.getInputInterfaceName();
    String oppOut = opposite.getOutputInterfaceName();
    boolean compatible;

    if ( (oppIn == null) != (myOut == null) ) return false;
    if ( (oppOut == null) != (myIn == null) ) return false;

    if ( myIn != null )
    {
      try
      {
        compatible = Class.forName( oppOut ).isAssignableFrom( Class.forName( myIn ) );
      }
      catch ( ClassNotFoundException e )
      {
        compatible = oppOut.equals( myIn );
      }
      if ( !compatible ) return false;
    }

    if ( myOut != null )
    {
      try
      {
        compatible = Class.forName( myOut ).isAssignableFrom( Class.forName( oppIn ) );
      }
      catch ( ClassNotFoundException e )
      {
        compatible = myOut.equals( oppIn );
      }
      if ( !compatible ) return false;
    }

    return true;
  }

  /**
   * Connects the {@link Channel} in the associated {@link Piece}.
   *
   * @param  channel The {@link Channel} to be connected to this {@link Plug}.
   */
  public void connectChannel ( Channel channel )
  {
    if ( getPiece() == null ||
         getKey()   == null ) return;

    getPiece().channels.put( getKey(), channel );
  }

  /**
   * Disconnects the associated channel.
   *
   * @return <tt>true</tt> if there was a {@link Channel} connected.
   */
  public boolean disconnectChannel ()
  {
    if ( getPiece() == null ||
         getKey()   == null ||
         !getPiece().channels.containsKey( getKey() ) ) return false;

    getPiece().channels.remove( getKey() );
    return true;
  }

}
