package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;

/**
 * Sphere as simple {@link BoundingVolume bounding-volume}.
 *
 * @version 0.25 $Rev: 391 $ $Date: 2007-01-20 03:28:39 +0100 (so, 20 I 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/BoundingSphere.java">BoundingSphere.java</a>
 */
public class BoundingSphere implements BoundingVolume
{
  //==========================================================================
  //  Data:

  /** Center of the sphere in cartesian coordinates. */
  protected double[] center;

  /** Radius of the sphere. */
  protected double r;

  /** Square of the radius. */
  protected double rr;

  //==========================================================================
  //  Construction:

  public BoundingSphere ( double[] center, double radius )
  {
    this.center = center;
    r           = radius;
    rr          = radius * radius;
  }

  //==========================================================================
  //  Interface BoundingVolume:

  /**
   * Ray vs. bounding-volume intersection test.
   *
   * @param  P0 Ray origin (<tt>double[3]</tt> or <tt>double[4]</tt>).
   * @param  P1 Direction vector of the ray (<tt>double[3]</tt> or <tt>double[4]</tt>).
   * @return <tt>true</tt> if the ray intersects the bounding-volume.
   * @see    #intersection
   */
  public boolean intersectsRay ( double[] P0, double[] P1 )
  {
    int p0len, p1len;
      // valid input parameters?
    if ( P0 == null || (p0len = P0.length) < 3 ||
         P1 == null || (p1len = P1.length) < 3 ) return false;

      // ray origin:
    double ox = P0[0];
    double oy = P0[1];
    double oz = P0[2];
    if ( p0len > 3 )
    {                                     // to cartesian
      double ow = P0[3];
      if ( ow < -Geometry.EPSILON || ow > Geometry.EPSILON )
      {
        ox /= ow;
        oy /= ow;
        oz /= ow;
      }
    }

      // ray direction vector (of arbitrary size):
    double dx = P1[0];
    double dy = P1[1];
    double dz = P1[2];
    if ( p1len > 3 )
    {                                     // to cartesian
      double dw = P1[3];
      if ( dw < -Geometry.EPSILON && dw > Geometry.EPSILON )
      {
        dx /= dw;
        dy /= dw;
        dz /= dw;
      }
    }

      // center[] versus [ox,oy,oz] + t * [dx,dy,dz]:
    double dd  = dx * dx + dy * dy + dz * dz;
    if ( dd < Geometry.EPSILON ) return false;

    double t = (dx * (center[0] - ox) +
                dy * (center[1] - oy) +
                dz * (center[2] - oz)) / dd;

    dx = ox + t * dx - center[0];
    dy = oy + t * dy - center[1];
    dz = oz + t * dz - center[2];

    return( dx * dx + dy * dy + dz * dz <= rr );
  }

  /**
   * Ray vs. bounding-volume intersection test with quantitative result.
   *
   * @param  P0 Ray origin (<tt>double[3]</tt> or <tt>double[4]</tt>).
   * @param  P1 Direction vector of the ray (<tt>double[3]</tt> or <tt>double[4]</tt>).
   * @return <tt>t</tt>-parameter {@code >= 0.0} if the ray intersects the bounding-volume,
   *         number {@code < 0.0} if no intersection exists.
   * @see    #intersectsRay
   */
  public double intersection ( double[] P0, double[] P1 )
  {
    int p0len, p1len;
      // valid input parameters?
    if ( P0 == null || (p0len = P0.length) < 3 ||
         P1 == null || (p1len = P1.length) < 3 ) return -1.0;

      // ray origin:
    double ox = P0[0];
    double oy = P0[1];
    double oz = P0[2];
    if ( p0len > 3 )
    {                                     // to cartesian
      double ow = P0[3];
      if ( ow < -Geometry.EPSILON || ow > Geometry.EPSILON )
      {
        ox /= ow;
        oy /= ow;
        oz /= ow;
      }
    }

      // ray direction vector (of arbitrary size):
    double dx = P1[0];
    double dy = P1[1];
    double dz = P1[2];
    if ( p1len > 3 )
    {                                     // to cartesian
      double dw = P1[3];
      if ( dw < -Geometry.EPSILON && dw > Geometry.EPSILON )
      {
        dx /= dw;
        dy /= dw;
        dz /= dw;
      }
    }

      // center[] versus [ox,oy,oz] + t * [dx,dy,dz]:
    double dd  = dx * dx + dy * dy + dz * dz;
    if ( dd < Geometry.EPSILON ) return -1.0;

    double t = (dx * (center[0] - ox) +
                dy * (center[1] - oy) +
                dz * (center[2] - oz)) / dd;

    dx = ox + t * dx - center[0];
    dy = oy + t * dy - center[1];
    dz = oz + t * dz - center[2];

    double dist2 = dx * dx + dy * dy + dz * dz;

    if ( dist2 >= rr ) return -1.0;

    double dt = Math.sqrt( (rr - dist2) / dd );

    if ( t - dt >= 0.0 ) return t - dt;

    if ( t + dt >= 0.0 ) return 0.0;

    return -1.0;
  }

  /**
   * Rectangular box (AABB) vs. bounding-volume intersection test.
   *
   * @param  o Coordinate of minimal cell-vertex (<tt>double[3]</tt>).
   * @param  size Box size (<tt>double[3]</tt>).
   * @return <tt>true</tt> if intersection of the cell and the bounding-volume is non-empty.
   * @see    #isInside
   */
  public boolean intersectsCell ( double[] o, double[] size )
  {
    if ( o    == null || o.length    != 3 ||
         size == null || size.length != 3 ) return false;

    return( Geometry.pointBox3D( center, o, size ) < r );
  }

  /**
   * Point vs. bounding-volume test ("inside-test").
   *
   * @param  point Coordinates of the 3D point (<tt>double[3]</tt> or
   *               <tt>double[4]</tt>).
   * @return <tt>true</tt> if the given point lies inside the bounding-volume.
   * @see    #intersectsCell
   */
  public boolean isInside ( double[] point )
  {
    if ( point == null ||
         point.length < 3 ) return false;

    double dx, dy, dz;
    if ( point.length > 3 )
    {
      if ( point[3] > -Geometry.EPSILON && point[3] < Geometry.EPSILON ) return false;

      double mul = 1.0 / point[3];
      dx = mul * point[0] - center[0];
      dy = mul * point[1] - center[1];
      dz = mul * point[2] - center[2];
    }
    else
    {
      dx = point[0] - center[0];
      dy = point[1] - center[1];
      dz = point[2] - center[2];
    }

    return( dx * dx + dy * dy + dz * dz <= rr );
  }

}
