package cz.cuni.jagrlib;

import java.util.*;

import cz.cuni.jagrlib.iface.*;

/**
 * Common code for CSG nodes.
 * Contains support for scene hierarchy and attribute machinery.
 *
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.11
 * @see     <a href="../../../../cz/cuni/jagrlib/CSGCommonNode.java">CSGCommonNode.java</a>
 */
public abstract class CSGCommonNode extends DefaultSceneNode
{
  //==========================================================================
  //  Data:

  /** Attribute dictionary (hashing table of [key,value] pairs). */
  protected Map< String, Object > attr;

  /** Handle of node's parent (or {@code null} for root-node). */
  protected SceneNode parent;

  /** Order in the parent node. */
  protected int order;

  /** Root-property. Is this node root of object (component) hierarchy? */
  protected boolean root;

  //==========================================================================
  //  Construction:

  /** Default initializing constructor. */
  protected CSGCommonNode ()
  {
    attr   = null;
    parent = null;
    order  = -1;
    root   = false;
  }

  //==========================================================================
  //  Interface SceneNode:

  /**
   * Returns the given attribute (defined in this {@link SceneNode} or inherited from
   * some parent node).
   * <p>Returned object can be an <b>original</b> data object and must not be altered
   * outside this {@link SceneNode} instance.
   *
   * @param  key Key-string.
   * @return Attribute object or {@code null}. Must not be altered!
   * @see    #getAttributes
   * @see    #setAttribute
   * @see    #selfAttribute
   */
  @Override
  public Object getAttribute ( String key )
  {
    Object value = null;
    if ( attr != null )
      value = attr.get( key );

    if ( value != null )
      return value;

    if ( parent != null )
      return parent.getAttribute( key );

    return null;
  }

  /**
   * Is the given attribute defined in the node itself or is it inherited?
   *
   * @param  key Key-string.
   * @return {@code true} if the node defines the given attribute.
   * @see    #getAttribute
   * @see    #setAttribute
   * @see    #getAttributes
   */
  @Override
  public boolean selfAttribute ( String key )
  {
    if ( attr == null ) return false;

    return attr.containsKey( key );
  }

  /**
   * Defines the given attribute (key-value pair) in the node.
   *
   * @param  key Key-string.
   * @param  value Attribute value (arbitrary type).
   * @return The old attribute value (defined in the node itself&excl;) or
   *         {@code null}.
   * @see    #getAttribute
   * @see    #selfAttribute
   * @see    #getAttributes
   */
  @Override
  public Object setAttribute ( String key, Object value )
  {
    if ( attr == null )
      attr = new HashMap< String, Object >();

    return attr.put( key, value );
  }

  /**
   * Returns the whole inherence-sequence for the given attribute.
   *
   * @param  key Key-string.
   * @return List of attribute values: {@code result[0]} .. definition in the node itself,
   *         {@code result[1]} .. definition in parent node,
   *         {@code result[2]} .. definition in "grand-parent" node, etc.
   * @see    #getAttribute
   * @see    #setAttribute
   * @see    #selfAttribute
   */
  @Override
  public List< Object > getAttributes ( String key )
  {
    List< Object > list = new LinkedList< Object >();

    SceneNode node = this;
    do
    {
      if ( node.selfAttribute( key ) )
        list.add( node.getAttribute( key ) );

      node = node.getParent();
    }
    while ( node != null );

    return list;
  }

  /**
   * Returns the parent of this node.
   *
   * @return The parend node or {@code null} for scene-root.
   * @see    #setParent
   * @see    #getOrder
   * @see    #children
   * @see    #getChild
   */
  @Override
  public SceneNode getParent ()
  {
    return parent;
  }

  /**
   * Sets handle to the parent node.
   *
   * @param  par Parent node (or {@code null}).
   * @see    #getParent
   */
  @Override
  public void setParent ( SceneNode par )
  {
    parent = par;
  }

  /**
   * Returns order of this {@link SceneNode} in its parent.
   *
   * @return Order of this node as a child (or {@code -1} for scene-root).
   * @see    #setOrder
   * @see    #getParent
   */
  @Override
  public int getOrder ()
  {
    return order;
  }

  /**
   * Sets order of this {@link SceneNode} in its parent.
   *
   * @param  ord The new order.
   * @see    #getOrder
   */
  @Override
  public void setOrder ( int ord )
  {
    order = ord;
  }

  /**
   * Returns {@code true} if this node is root of object (component) hierarchy.
   *
   * @see    #setObjectRoot
   */
  @Override
  public boolean isObjectRoot ()
  {
    return root;
  }

  /**
   * Sets this node as root for the object (component).
   *
   * @param  root Should this node be treated as root?
   * @see    #isObjectRoot
   */
  @Override
  public void setObjectRoot ( boolean root )
  {
    this.root = root;
  }

  /**
   * Asserts validity of the given intersection-attributes.
   * Resolves:
   * {@link MicroFacet#VALID_TO_WORLD VALID_TO_WORLD},
   * {@link MicroFacet#VALID_TO_OBJECT VALID_TO_OBJECT},
   * {@link MicroFacet#VALID_TEXTURES VALID_TEXTURES},
   * {@link MicroFacet#VALID_BRDF VALID_BRDF},
   * {@link MicroFacet#VALID_MATERIAL VALID_MATERIAL},
   * {@link MicroFacet#VALID_COLOR VALID_COLOR}.
   *
   * @param  mf The checked intersection.
   * @param  tmp Non-mandatory object holding state-information for deferred computations.
   * @param  attributes Attribute mask (required attributes - see
   *                    {@code MicroFacet.VALID_*}).
   * @return Attributes which were not computed.
   */
  @Override
  public int assertAttributes ( MicroFacet mf, Object tmp, int attributes )
  {
      // attribute dependencies: have been resolved before
    if ( mf == null || (attributes &= ~mf.valid) == 0 ) return attributes;

      // node -> world transform:
    if ( (attributes & MicroFacet.VALID_TO_WORLD) != 0 )
    {
      mf.toWorld     = new TrMatrix( 4, 3 );
      SceneNode node = this;
      int i;
      while ( node != null && (i = node.getOrder()) >= 0 )
      {
        SceneNode parent = node.getParent();
        mf.toWorld.append( parent.getChildMatrix( i ) );
        node = parent;
      }

      mf.valid |= MicroFacet.VALID_TO_WORLD;
      if ( (attributes -= MicroFacet.VALID_TO_WORLD) == 0 ) return 0;
    }

      // node -> object (component) transform:
    if ( (attributes & MicroFacet.VALID_TO_OBJECT) != 0 )
    {
      mf.toObject    = new TrMatrix( 4, 3 );
      SceneNode node = this;
      int i;
      while ( node != null && (i = node.getOrder()) >= 0 && !node.isObjectRoot() )
      {
        SceneNode parent = node.getParent();
        mf.toObject.append( parent.getChildMatrix( i ) );
        node = parent;
      }

      mf.valid |= MicroFacet.VALID_TO_OBJECT;
      if ( (attributes -= MicroFacet.VALID_TO_OBJECT) == 0 ) return 0;
    }

      // list of applicable textures:
    if ( (attributes & MicroFacet.VALID_TEXTURES) != 0 )
    {
      List< Object > tx = getAttributes( ATTR_TEXTURE );
      mf.textures       = new LinkedList< Texture >();

      int    noTextures;
      Object tex;
      if ( tx != null && (noTextures = tx.size()) > 0 )
      {
        ListIterator< Object > li = tx.listIterator( noTextures );
        do
        {
          tex = li.previous();
          if ( tex instanceof Texture )
            mf.textures.add( (Texture)tex );
        }
        while ( li.hasPrevious() );
      }

      mf.valid |= MicroFacet.VALID_TEXTURES;
      if ( (attributes -= MicroFacet.VALID_TEXTURES) == 0 ) return 0;
    }

      // surface color: 1st try
    if ( (attributes & MicroFacet.VALID_COLOR) != 0 )
    {
        // object color has higher priority over BRDF-data (material) color:
      mf.color = (double[])getAttribute( ATTR_COLOR );
      if ( mf.color == null )
      {
        attributes |= (MicroFacet.VALID_BSSRDF) & ~mf.valid;
      }
      else
      {
        mf.color  = Formula.cloneArray( mf.color );
        mf.valid |= MicroFacet.VALID_COLOR;
        if ( (attributes -= MicroFacet.VALID_COLOR) == 0 ) return 0;
      }
    }

      // bssrdf:
    if ( (attributes & MicroFacet.VALID_BSSRDF) != 0 )
    {
      mf.bssrdf = (BSSRDF)getAttribute( ATTR_BSSRDF );
      if ( mf.bssrdf != null )
      {
        mf.valid |= MicroFacet.VALID_BSSRDF;
        if ( (attributes -= MicroFacet.VALID_BSSRDF) == 0 ) return 0;
      }
    }

      // surface color: 2nd try
    if ( (attributes & MicroFacet.VALID_COLOR)  != 0 &&
         (mf.valid   & MicroFacet.VALID_BSSRDF) != 0 )
    {
        // BRDF-data (material) color has lower priority:
      mf.color = mf.bssrdf.getOrigColor( null );
      if ( mf.color != null )
      {
        mf.valid |= MicroFacet.VALID_COLOR;
        if ( (attributes -= MicroFacet.VALID_COLOR) == 0 ) return 0;
      }
    }

    return attributes;
  }

}