package cz.cuni.jagrlib;

import java.util.*;

/**
 * Inner node of CSG scene representation.
 *
 * @version 0.26 $Rev: 467 $ $Date: 2007-05-26 09:00:40 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.11
 * @see     <a href="../../../../cz/cuni/jagrlib/CSGNode.java">CSGNode.java</a>
 * @see     CSGLeaf
 */
public class CSGNode extends CSGCommonNode
{
  //==========================================================================
  //  Child-reference class:

  protected static class Child
  {
    /** Reference to the child node. */
    public SceneNode node;

    /** Transform matrix from child-object space to parent space. */
    public TrMatrix m;

    /** Transform matrix from parent space to child-object space (for rays, ..). */
    public TrMatrix mInv;

    public Child ( SceneNode _node, TrMatrix _m, TrMatrix _mInv )
    {
      node = _node;
      m    = _m;
      mInv = _mInv;
    }

    public TrMatrix getMatrix ()
    {
      if ( m == null )
      {
        if ( mInv == null ) return null;

        m = new TrMatrix( mInv );
        m.invert( (double[][])null );
      }

      return m;
    }

    public TrMatrix getInvMatrix ()
    {
      if ( mInv == null )
      {
        if ( m == null ) return null;

        mInv = new TrMatrix( m );
        mInv.invert( (double[][])null );
      }

      return mInv;
    }

  }

  //==========================================================================
  //  Data:

  /**
   * Set operation code.
   *
   * @see    BinaryOperation
   */
  public int op;

  /** Array of child-references. */
  protected ArrayList< Child > ch;

  /** Initializing constructor. */
  public CSGNode ( int _op )
  {
    op = _op;
    ch = new ArrayList< Child >();
  }

  /**
   * Returns number of child nodes.
   *
   * @return Number of child nodes ({@code 0} for leaf node).
   * @see    #getChild
   * @see    #getChildMatrix
   * @see    #getChildMatrixInv
   */
  @Override
  public final int children ()
  {
    return ch.size();
  }

  /**
   * Returns the given child node.
   *
   * @param  i Order of the requested node (from {@code 0} to {@code children()-1}).
   * @return The requested node or {@code null}.
   * @see    #children
   * @see    #getChildMatrix
   * @see    #getChildMatrixInv
   */
  @Override
  public SceneNode getChild ( int i )
  {
    if ( i < 0 || i >= ch.size() ) return null;
    return ch.get( i ).node;
  }

  /**
   * Returns transformation matrix for the given child node.
   *
   * @param  i Order of the child node.
   * @return Transformation matrix from the child's coordinate system to this node's one.
   *         {@code null} stands for <b>identity</b>.
   * @see    #children
   * @see    #getChild
   * @see    #getChildMatrixInv
   */
  @Override
  public TrMatrix getChildMatrix ( int i )
  {
    if ( i < 0 || i >= ch.size() ) return null;

    final Child chi = ch.get( i );
    return chi.getMatrix();
  }

  /**
   * Returns inverse transformation matrix for the given child node.
   *
   * @param  i Order of the child node.
   * @return Transformation matrix from this node's coordinate system to child's one.
   *         {@code null} stands for <b>identity</b>.
   * @see    #children
   * @see    #getChild
   * @see    #getChildMatrix
   */
  @Override
  public TrMatrix getChildMatrixInv ( int i )
  {
    if ( i < 0 || i >= ch.size() ) return null;

    final Child chi = ch.get( i );
    return chi.getInvMatrix();
  }

  /**
   * Inserts a new child node.
   *
   * @param  child The new child node.
   * @param  m Transformation matrix for the new child node (transformation from child's
   *           coordinate system to this node's one). Can be {@code null}, even
   *           for non-trivial {@code mInv}.
   * @param  mInv Inverse transformation matrix for the new child node (transformation from
   *              this node's coordinate system to child's one). Can be {@code null}, even
   *              for non-trivial {@code m}.
   * @return Order of the inserted child node (or {@code -1} in case of failure).
   * @see    #removeChild
   * @see    #children
   */
  @Override
  public int insertChild ( SceneNode child, TrMatrix m, TrMatrix mInv )
  {
    final Child chi = new Child( child, m, mInv );
    final int n = ch.size();
    ch.add( chi );

    child.setParent( this );
    child.setOrder( n );

    return( n + 1 );
  }

  /**
   * Removes the given child node.
   *
   * @param  i Order of the child node to be removed.
   * @return Number of child nodes after the remove-operation.
   * @see    #insertChild
   * @see    #children
   */
  @Override
  public int removeChild ( int i )
  {
    final int n = ch.size();
    if ( n < 1 ) return 0;
    if ( i < 0 || i >= n ) return n;

    final Child chi = ch.get( i );
    chi.node.setParent( null );
    chi.node.setOrder( -1 );

    ch.remove( i );
    return( n - 1 );
  }

  /**
   * Computes all intersections of the given ray with one child scene. W/o any checks.
   *
   * @param  P0 Ray origin ({@code double[3]}).
   * @param  P1 Direction vector of the ray ({@code double[3]}).
   * @param  i Order of the child node.
   * @return List of all intersections with the given child-scene.
   */
  protected List< MicroFacet > intersection ( double[] P0, double[] P1, int i )
  {
    final Child chi = ch.get( i );

      // child transform:
    final TrMatrix inv = chi.getInvMatrix();
    if ( inv != null )
    {
      double newP[] = new double[ 3 ];
      inv.transformPoint3D( P0, newP );
      P0   = newP;
      newP = new double[ 3 ];
      inv.transformVector3D( P1, newP );
      P1   = newP;
    }

      // intersections:
    return chi.node.intersection( P0, P1 );
  }

  /**
   * Computes all intersections of the given ray with sub-scene surface.
   *
   * @param  P0 Ray origin ({@code double[3]} or {@code double[4]}).
   * @param  P1 Direction vector of the ray ({@code double[3]} or {@code double[4]}).
   * @return List of all intersections with the sub-scene.
   */
  @Override
  public List< MicroFacet > intersection ( double[] P0, double[] P1 )
  {
    if ( P0 == null || P0.length < 3 ||
         P1 == null || P1.length < 3 ||
         ch.size() == 0 ) return null;

    double p0[] = Geometry.toCartesian3D( P0, null );
    double p1[] = new double[ 3 ];
    System.arraycopy( P1, 0, p1, 0, 3 );

    List< MicroFacet > result = intersection( p0, p1, 0 );
    for ( int i = 1; i < ch.size(); i++ )
    {
      List< MicroFacet > arg = intersection( p0, p1, i );
      result = MicroFacet.booleanOp( result, arg, op );
    }

    return result;
  }

}
