package cz.cuni.jagrlib;

/**
 * {@link Channel} connects two or more {@link Piece}s - the <b>left</b> {@link Piece}(s)
 * can call the given interface of the <b>right</b> {@link Piece}(s) and vice versa.
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.01
 * @see     <a href="../../../../cz/cuni/jagrlib/Channel.java">Channel.java</a>
 */
public interface Channel
{
  /**
   * (Re-)initialize both interface names.
   *
   * @param  _leftInterfaceName Fully-qualified interface name of the left side (called
   *                            from the right side).
   * @param  _rightInterfaceName Fully-qualified interface name of the right side (called
   *                             from the left side).
   * @see    #getLeftInterfaceName
   * @see    #getRightInterfaceName
   */
  void setInterfaceNames ( String _leftInterfaceName, String _rightInterfaceName );

  /**
   * Returns an interface name of the left Channel-side.
   *
   * @return Fully-qualified interface name of the left side (called from the right side).
   * @see    #setInterfaceNames
   * @see    #getRightInterfaceName
   */
  String getLeftInterfaceName ();

  /**
   * Returns an interface name of the right Channel-side.
   *
   * @return Fully-qualified interface name of the right side (called from the left side).
   * @see    #setInterfaceNames
   * @see    #getLeftInterfaceName
   */
  String getRightInterfaceName ();

  /**
   * Returns an interface name of the same Channel-side.
   *
   * @param  proPlug The {@link Plug} connected to the same side.
   * @return Fully-qualified interface name of the given Channel-side.
   * @see    #getInterfaceNameAnti
   */
  String getInterfaceNamePro ( Plug proPlug );

  /**
   * Returns an interface name of the opposite Channel-side.
   *
   * @param  antiPlug The {@link Plug} connected to the opposite side.
   * @return Fully-qualified interface name of the given Channel-side.
   * @see    #getInterfaceNamePro
   */
  String getInterfaceNameAnti ( Plug antiPlug );

  /**
   * Sets new interface object for the left Channel-side.
   *
   * @param  newLeftInterface New interface object (= object called from the opposite side).
   * @return The old interface object or <tt>null</tt>.
   * @see    #setRightInterface
   * @see    #setInterface
   */
  Object setLeftInterface ( Object newLeftInterface );

  /**
   * Sets new interface object for the right Channel-side.
   *
   * @param  newRightInterface New interface object (= object called from the opposite side).
   * @return The old interface object or <tt>null</tt>.
   * @see    #setLeftInterface
   * @see    #setInterface
   */
  Object setRightInterface ( Object newRightInterface );

  /**
   * Sets new interface object for the given Channel-side. The side is defined by the
   * {@link Piece} connected to it.
   *
   * @param  newInterface New interface object (= object called from the opposite side).
   * @param  called The {@link Plug} defining the side.
   * @return The old interface object or <tt>null</tt>.
   * @see    #setLeftInterface
   * @see    #setRightInterface
   */
  Object setInterface ( Object newInterface, Plug called );

  /**
   * Returns the interface object of the left Channel-side.
   *
   * @return The interface object or <tt>null</tt>.
   * @see    #setLeftInterface
   */
  Object getLeftInterface ();

  /**
   * Returns the interface object of the right Channel-side.
   *
   * @return The interface object or <tt>null</tt>.
   * @see    #setRightInterface
   */
  Object getRightInterface ();

  /**
   * Returns the interface object of the given Channel-side. The side is defined by the
   * {@link Plug} which calls it.
   *
   * @param  caller The {@link Plug} defining the opposite (caller) side.
   * @return The interface object or <tt>null</tt>.
   * @see    #setInterface
   */
  Object getInterface ( Plug caller );

  /**
   * Connects a new {@link Plug} to the left Channel-side.
   * Updates the {@link Plug}'s {@link Piece}, too.
   *
   * @param  newPlug The {@link Plug} to be connected.
   * @see    #connectRight
   */
  void connectLeft ( Plug newPlug );

  /**
   * Connects a new {@link Plug} to the right Channel-side.
   * Updates the {@link Plug}'s {@link Piece}, too.
   *
   * @param  newPlug The {@link Plug} to be connected.
   * @see    #connectLeft
   */
  void connectRight ( Plug newPlug );

  /**
   * Connects a new {@link Plug} to the same Channel-side.
   * Updates the {@link Plug}'s {@link Piece}, too.
   *
   * @param  newPlug The {@link Plug} to be connected.
   * @param  proPlug The {@link Plug} connected to the same side.
   * @see    #connectAnti
   */
  void connectPro ( Plug newPlug, Plug proPlug );

  /**
   * Connects a new {@link Plug} to the opposite Channel-side.
   * Updates the {@link Plug}'s {@link Piece}, too.
   *
   * @param  newPlug The {@link Plug} to be connected.
   * @param  antiPlug The {@link Plug} connected to the opposite side.
   * @see    #connectPro
   */
  void connectAnti ( Plug newPlug, Plug antiPlug );

  /**
   * Disconnects the old {@link Plug} from the {@link Channel}.
   * Updates the {@link Plug}'s {@link Piece}, too.
   *
   * @param  oldPlug The {@link Plug} to be disconnected.
   * @see    #disconnectAll
   */
  void disconnect ( Plug oldPlug );

  /**
   * Disconnects all {@link Plug}s from the {@link Channel}.
   * Updates the {@link Plug}'s {@link Piece}, too.
   *
   * @see    #disconnect
   */
  void disconnectAll ();

  /**
   * Returns the left Channel-side count.
   *
   * @return The left-side count (number of {@link Plug}s connected to that side).
   * @see    #getRightCount
   */
  int getLeftCount ();

  /**
   * Returns the right Channel-side count.
   *
   * @return The right-side count (number of {@link Plug}s connected to that side).
   * @see    #getLeftCount
   */
  int getRightCount ();

  /**
   * Is the given {@link Plug} connected to the left Channel-side?
   *
   * @param  oldPlug The checked {@link Plug}.
   * @return <tt>true</tt> if the given {@link Plug} is connected to the left side.
   * @see    #isRight
   */
  boolean isLeft ( Plug oldPlug );

  /**
   * Is the given {@link Plug} connected to the right Channel-side?
   *
   * @param  oldPlug The checked {@link Plug}.
   * @return <tt>true</tt> if the given {@link Plug} is connected to the right side.
   * @see    #isLeft
   */
  boolean isRight ( Plug oldPlug );

  /**
   * Returns the array of left-side objects ({@link Plug}s).
   *
   * @return The array of {@link Plug}s connected to the left side.
   */
  Plug[] getLeft ();

  /**
   * Returns the array of right-side objects ({@link Plug}s).
   *
   * @return The array of {@link Plug}s connected to the right side.
   */
  Plug[] getRight ();

  /**
   * Returns the array of same-side objects ({@link Plug}s).
   *
   * @param  proPlug The {@link Plug} connected to the same side.
   * @return The array of {@link Plug}s.
   */
  Plug[] getPro ( Plug proPlug );

  /**
   * Returns the array of opposite-side objects ({@link Plug}s).
   *
   * @param  antiPlug The {@link Plug} connected to the opposite side.
   * @return The array of {@link Plug}s.
   */
  Plug[] getAnti ( Plug antiPlug );

  /**
   * Prepares the {@link Channel} to regular operation.
   *
   * @return <tt>true</tt> if everything is OK ({@link Channel} is prepared
   *         for operation).
   */
  boolean prepare ();

}
