package cz.cuni.jagrlib;

/**
 * Complex number represented as {@code [double,double]} tuple.
 *
 * @version 0.26 $Rev: 456 $ $Date: 2007-05-19 01:03:36 +0200 (so, 19 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/Complex.java">Complex.java</a>
 * @todo    Goniometry, cyclometry
 */
public class Complex
{
  //==========================================================================
  //  Data:

  /** Real part. */
  protected double re;

  /** Imaginary part. */
  protected double im;

  //==========================================================================
  //  Construction:

  public Complex ()
  {
    re = im = 0.0;
  }

  public Complex ( double r )
  {
    re = r;
    im = 0.0;
  }

  public Complex ( double re, double im )
  {
    this.re = re;
    this.im = im;
  }

  public Complex ( Complex c )
  {
    re = c.re;
    im = c.im;
  }

  public final void set ( double re, double im )
  {
    this.re = re;
    this.im = im;
  }

  public final void set ( Complex c )
  {
    re = c.re;
    im = c.im;
  }

  public final void setPolar ( double r, double theta )
  {
    re = r * Math.cos( theta );
    im = r * Math.sin( theta );
  }

  public final static Complex polar ( double r, double theta )
  {
    return new Complex( r * Math.cos( theta ),
                        r * Math.sin( theta ) );
  }

  //==========================================================================
  //  Simple operators / operations:

  public final boolean isInfinite ()
  {
    return Double.isInfinite( re ) ||
           Double.isInfinite( im );
  }

  public final boolean isNaN ()
  {
    return Double.isNaN( re ) ||
           Double.isNaN( im );
  }

  public final boolean isEqual ( Complex c )
  {
    return c.re == re &&
           c.im == im;
  }

  public final boolean isEqualEps ( Complex c )
  {
    return Math.abs( c.re - re ) < Geometry.EPSILON &&
           Math.abs( c.im - im ) < Geometry.EPSILON;
  }

  public final double real ()
  {
    return re;
  }

  public final double imag ()
  {
    return im;
  }

  //==========================================================================
  //  Simple arithmetics:

  /** Argument (polar angle in counterclockwise direction). */
  public final double arg ()
  {
    return Math.atan2( im, re );
  }

  /** Absolute value. */
  public final double abs ()
  {
    return Math.hypot( re, im );
  }

  /** Norm (absolute value squared). */
  public final double norm ()
  {
    return re * re + im * im;
  }

  /** In-place addition. */
  public final void add ( Complex c )
  {
    re += c.re;
    im += c.im;
  }

  /** External addition. */
  public final static Complex add ( Complex a, Complex b )
  {
    return new Complex( a.re + b.re, a.im + b.im );
  }

  /** In-place subtraction. */
  public final void sub ( Complex c )
  {
    re -= c.re;
    im -= c.im;
  }

  /** External subtraction. */
  public final static Complex sub ( Complex a, Complex b )
  {
    return new Complex( a.re - b.re, a.im - b.im );
  }

  /** Conjugates the complex number. */
  public final void setConjugate ()
  {
    im = -im;
  }

  /** Returns conjugate complex number. */
  public final Complex conjugate ()
  {
    return new Complex( re, -im );
  }

  /** In-place multiplication. */
  public final void mul ( Complex c )
  {
    final double oldRe  = re;
    final double oldCRe = c.re;
    re = oldRe * c.re - im * c.im;
    im = oldRe * c.im + im * oldCRe;
  }

  /** External multiplication. */
  public final static Complex mul ( Complex a, Complex b )
  {
    return new Complex( a.re * b.re - a.im * b.im,
                        a.re * b.im + a.im * b.re );
  }

  /** Inverts the complex number. */
  public final void invert ()
  {
    if ( Math.abs( re ) >= Math.abs( im ) )
    {
      final double imOverRe = im / re;
      re  = 1.0 / (re + im * imOverRe);
      im  = -re * imOverRe;
    }
    else
    {
      final double reOverIm = re / im;
      im  = -1.0 / (re * reOverIm + im);
      re  = -im * reOverIm;
    }
  }

  /** Returns inverse number. */
  public final Complex inverse ()
  {
    Complex res = new Complex( this );
    res.invert();
    return res;
  }

  /** Scales the complex number. */
  public final void setScale ( double k )
  {
    re *= k;
    im *= k;
  }

  /** Returns scaled complex number. */
  public final Complex scale ( double k )
  {
    return new Complex( k * re, k * im );
  }

  /** In-place division. */
  public final void div ( Complex c )
  {
    final double oldRe = re;
    final double oldIm = im;
    double rat, sc;

    if ( Math.abs( c.re ) >= Math.abs( c.im ) )
    {
      rat = c.im / c.re;
      sc  = 1.0 / (c.re + c.im * rat);
      re  = sc * (oldRe + oldIm * rat);
      im  = sc * (oldIm - oldRe * rat);
    }
    else
    {
      rat = c.re / c.im;
      sc  = 1.0 / (c.re * rat + c.im);
      re  = sc * (oldRe * rat + oldIm);
      im  = sc * (oldIm * rat - oldRe);
    }
  }

  /** External division. */
  public final static Complex div ( Complex a, Complex b )
  {
    Complex res = new Complex( a );
    res.div( b );
    return res;
  }

  //==========================================================================
  //  More complex operations:

  /** In-place square root. */
  public final void setSqrt ()
  {
    final double mag = abs();

    if ( mag > 0.0 )
    {
      if ( re > 0.0 )
      {
        re  = Math.sqrt( 0.5 * (mag + re) );
        im *= 0.5 / re;
      }
      else
      {
        double newIm = Math.sqrt( 0.5 * (mag - re) );

        if ( im < 0.0 )
          newIm = -newIm;

        re = 0.5 * im / newIm;
        im = newIm;
      }
    }
    else
      re =
      im = 0.0;
  }

  /** External square root. */
  public final Complex sqrt ()
  {
    Complex res = new Complex( this );
    res.setSqrt();
    return res;
  }

  /** In-place exponentiation (e ^ this). */
  public final void setExp ()
  {
    final double sc = Math.exp( re );
    re = sc * Math.cos( im );
    im = sc * Math.sin( im );
  }

  /** External exponentiation (returns e ^ this). */
  public final Complex exp ()
  {
    Complex res = new Complex( this );
    res.setExp();
    return res;
  }

  /** In-place natural logarithm. */
  public final void setLog ()
  {
    final double mag = abs();
    im = arg();
    re = Math.log( mag );
  }

  /** External natural logarithm. */
  public final Complex log ()
  {
    return new Complex( Math.log( abs() ), arg() );
  }

  /** In-place power with integer exponent. */
  public final void setPower ( int exponent )
  {
    if ( exponent == 0 )
    {
      re = 1.0;
      im = 0.0;
      return;
    }

    final boolean inv = exponent < 0;
    if ( inv ) exponent = -exponent;

    double resRe, resIm;
    if ( (exponent & 1) > 0 )
    {
      resRe = re;
      resIm = im;
    }
    else
    {
      resRe = 1.0;
      resIm = 0.0;
    }

    double tmp;
    while ( (exponent >>= 1) > 0 )
    {
      mul( this );
      if ( (exponent & 1) > 0 )
      {
        tmp   = resRe;
        resRe = resRe * re - resIm * im;
        resIm = tmp   * im + resIm * re;
      }
    }

    re = resRe;
    im = resIm;
    if ( inv )
      invert();
  }

  /** External power with integer exponent. */
  public final Complex power ( int exponent )
  {
    Complex res = new Complex( this );
    res.setPower( exponent );
    return res;
  }

  /** In-place power with real exponent. */
  public final void setPower ( double exponent )
  {
    final double mag   = Math.exp( exponent * Math.log( abs() ) );
    final double theta = exponent * arg();

    re = mag * Math.cos( theta );
    im = mag * Math.sin( theta );
  }

  /** External power with real exponent. */
  public final Complex power ( double exponent )
  {
    Complex res = new Complex( this );
    res.setPower( exponent );
    return res;
  }

  /** In-place power with complex exponent. */
  public final void setPower ( Complex exponent )
  {
    final double rr    = Math.log( abs() );
    final double ii    = arg();

    final double mag   = Math.exp( rr * exponent.re - ii * exponent.im );
    final double theta =           rr * exponent.im + ii * exponent.re;

    re = mag * Math.cos( theta );
    im = mag * Math.sin( theta );
  }

  /** External power with complex exponent. */
  public final Complex power ( Complex exponent )
  {
    Complex res = new Complex( this );
    res.setPower( exponent );
    return res;
  }

}
