package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.BlockQuantizer;

/**
 * Default implementation of {@link BlockQuantizer} interface.
 *
 * @version 0.26 $Rev: 456 $ $Date: 2007-05-19 01:03:36 +0200 (so, 19 V 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../cz/cuni/jagrlib/DefaultBlockQuantizer.java">DefaultBlockQuantizer.java</a>
 */
public abstract class DefaultBlockQuantizer extends Piece implements BlockQuantizer
{
  //==========================================================================
  //  Data:

  /** Temporary array for intermediate integer values. */
  protected int[]    tmpI;

  /** Temporary array for intermediate double values. */
  protected double[] tmpD;

  /** Quantizer variant (context). */
  protected int variant   = 0;

  //==========================================================================
  //  Interface BlockQuantizer:

  /**
   * Set actual quantizer variant (context).
   *
   * @param  variant Variant number (from {@code 0}).
   * @return Old variant number.
   * @see    #setBlockSize
   * @see    #setQuality
   * @see    #setData
   */
  public int setVariant ( int variant )
  {
    int old      = this.variant;
    this.variant = variant;
    return old;
  }

  /**
   * Sets actual block size.
   * Can be used both for 1D and 2D quantization.

   * @param  width Horizontal block size (1D block size).
   * @param  height Vertical block size ({@code 1} for 1D).
   * @see    #setVariant
   * @see    #setQuality
   * @see    #setData
   */
  public void setBlockSize ( int width, int height )
  {
  }

  /**
   * Sets actual quality coefficient.
   * Value range for the quality: {@code 0.0f} to {@code 1.0f}.
   *
   * @param  quality New quality coefficient.
   * @see    #setVariant
   * @see    #setBlockSize
   * @see    #setData
   */
  public void setQuality ( float quality )
  {
  }

  /**
   * Sets quantization table for the current variant.
   *
   * @param  qtable New quantization table.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @see    #setVariant
   * @see    #setQuality
   * @see    #setBlockSize
   */
  public void setData ( int[] qtable )
  {
  }

  /**
   * Quantization (floating-point form).
   *
   * @param  coef Source data (coefficients). Size = array size.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @param  quant Pre-allocated result array or {@code null} if
   *         it's size needs to be determined.
   * @return Result array size (even if the quantization was not performed).
   * @see    #quantize(int[],int[])
   * @see    #dequantize(int[],double[])
   */
  public int quantize ( double[] coef, int[] quant )
  {
    if ( coef  == null ||
         quant == null )
      return quantize( (int[])null, null );

    int len = Math.min( coef.length, quant.length );
    if ( tmpI == null || tmpI.length < len )
      tmpI = new int[ len ];

    for ( int i = 0; i < len; i++ )
      tmpI[i] = Formula.round( coef[i] );

    return quantize( tmpI, quant );
  }

  /**
   * Quantization (integer form).
   *
   * @param  coef Source data (coefficients). Size = array size.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @param  quant Pre-allocated result array or {@code null} if
   *         it's size needs to be determined.
   * @return Result array size (even if the quantization was not performed).
   * @see    #quantize(double[],int[])
   * @see    #dequantize(int[],int[])
   */
  public int quantize ( int[] coef, int[] quant )
  {
    if ( coef  == null ||
         quant == null )
      return quantize( (double[])null, null );

    int len = Math.min( coef.length, quant.length );
    if ( tmpD == null || tmpD.length < len )
      tmpD = new double[ len ];

    for ( int i = 0; i < len; i++ )
      tmpD[i] = coef[i];

    return quantize( tmpD, quant );
  }

  /**
   * Dequantization = reconstruction (floating-point form).
   *
   * @param  quant Source data (quantized coefficients). Size = array size.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @param  coef Pre-allocated coefficient array or {@code null} if
   *         it's size needs to be determined.
   * @return Coefficient array size (even if the dequantization was not performed).
   * @see    #dequantize(int[],int[])
   * @see    #quantize(double[],int[])
   */
  public int dequantize ( int[] quant, double[] coef )
  {
    if ( coef  == null ||
         quant == null )
      return dequantize( null, (int[])null );

    int len = Math.min( coef.length, quant.length );
    if ( tmpI == null || tmpI.length < len )
      tmpI = new int[ len ];

    len = dequantize( quant, tmpI );

    for ( int i = 0; i < len; i++ )
      coef[i] = tmpI[i];

    return len;
  }

  /**
   * Dequantization = reconstruction (integer form).
   *
   * @param  quant Source data (quantized coefficients). Size = array size.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @param  coef Pre-allocated coefficient array or {@code null} if
   *         it's size needs to be determined.
   * @return Coefficient array size (even if the dequantization was not performed).
   * @see    #dequantize(int[],double[])
   * @see    #quantize(int[],int[])
   */
  public int dequantize ( int[] quant, int[] coef )
  {
    if ( coef  == null ||
         quant == null )
      return dequantize( null, (double[])null );

    int len = Math.min( coef.length, quant.length );
    if ( tmpD == null || tmpD.length < len )
      tmpD = new double[ len ];

    len = dequantize( quant, tmpD );

    for ( int i = 0; i < len; i++ )
      coef[i] = Formula.round( tmpD[i] );

    return len;
  }

  //==========================================================================
  //  Interface PersistentParamsBinary:

  /**
   * Pack instance parameters into binary format.
   *
   * @return Data array to be stored into binary stream. Can be {@code null}.
   *         The array can be shared by all method calls.
   * @see    #loadData
   */
  public byte[] storeData ()
  {
    return null;
  }

  /**
   * Loads instance parameters from (previously stored) binary data.
   *
   * @param  data Binary data to be interpreted. Can be {@code null}.
   * @return {@code true} if successful.
   * @see    #storeData
   */
  public boolean loadData ( byte[] data )
  {
    return( data == null );
  }

  //==========================================================================
  //  Module registration:

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "BlockQuantizer";

  /** Object category. */
  protected final static String CATEGORY = C_2D + ".quantization";

}
