package cz.cuni.jagrlib;

import java.io.*;

import cz.cuni.jagrlib.iface.*;

/**
 * File format default implementation for Brep-based 3D scene data.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @see     <a href="../../../../cz/cuni/jagrlib/DefaultBrepFileFormat.java">DefaultBrepFileFormat.java</a>
 */
public abstract class DefaultBrepFileFormat extends DefaultFileFormat
{
  //==========================================================================
  //  Routines to override:

  /**
   * Common load code.
   *
   * @param  stream Opened input bit-stream.
   * @param  b Checked B-rep database object.
   * @throws IOException
   */
  protected void commonLoad ( BitStream stream, Brep b )
  throws IOException
  {
    throw new JaGrLibException( new NotImplementedException( "Loading of this Brep format is not implemented!" ) );
  }

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  b Checked B-rep database object.
   * @throws IOException
   */
  protected void commonSave ( BitStream stream, Brep b )
  throws IOException
  {
    throw new JaGrLibException( new NotImplementedException( "Saving of this Brep format is not implemented!" ) );
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /**
   * Retrieves data from the given file (stream).
   * Can be used for "file opening" in virtual-mapped mode.
   *
   * @param  fileName File-name (stream-name) to read from (can be {@code null}).
   * @param  detail Further implementation-specific information ("r" or "rw" for mapped mode).
   * @return {@code true} if the operation was successful.
   * @see    #loadFile(BitStream,Object)
   * @see    #saveFile(String,String)
   * @see    #closeFile
   */
  @Override
  public boolean loadFile ( String fileName, String detail )
  throws IOException
  {
    Brep      b      = (Brep)     getInterface( PL_DATA,   IFACE+"Brep" );
    BitStream stream = (BitStream)getInterface( PL_STREAM, IFACE+"BitStream" );

    if ( b      == null ||
         stream == null )
      throw new JaGrLibException( "loadFile: vital modules are not connected!" );

    boolean result = false;
    try
    {
      stream.open( false, fileName, detail );
      commonLoad( stream, b );
      result = true;
    }
    finally
    {
      stream.close();
    }

    return result;
  }

  /**
   * Retrieves data from the given {@link BitStream}. This metod is optional.
   * <p>The stream should be opened before - only the {@code reOpen(false,null)} call
   * can be used.
   *
   * @param  stream Input bit-stream (opened - call {@code reOpen(false,null)} first).
   * @param  support Implementation-specific data (e.g. second {@link BitStream} for
   *                 file-formats with {@link #DOUBLE_STREAM} property).
   * @return {@code true} if the operation was successful.
   * @see    #loadFile(String,String)
   * @see    #DOUBLE_STREAM
   */
  @Override
  public boolean loadFile ( BitStream stream, Object support )
  throws IOException
  {
    if ( stream == null ) return false;

    final Brep b = (Brep)getInterface( PL_DATA, IFACE+"Brep" );
    if ( b == null )
      throw new JaGrLibException( "loadFile: vital modules are not connected!" );

    commonLoad( stream, b );
    return true;
  }

  /**
   * Saves data to the given file (stream).
   *
   * @param  fileName File-name (stream-name) to write to (can be {@code null}).
   * @param  detail Further implementation-specific information.
   * @return {@code true} if the operation was successful.
   * @see    #saveFile(BitStream,Object)
   * @see    #loadFile(String,String)
   */
  @Override
  public boolean saveFile ( String fileName, String detail )
  throws IOException
  {
    final Brep      b      = (Brep)     getInterface( PL_DATA,   IFACE+"Brep" );
    final BitStream stream = (BitStream)getInterface( PL_STREAM, IFACE+"BitStream" );

    if ( b      == null ||
         stream == null )
      throw new JaGrLibException( "saveFile: vital modules are not connected!" );

    boolean result = false;
    try
    {
      stream.open( true, fileName, detail );
      commonSave( stream, b );
      result = true;
    }
    finally
    {
      stream.close();
    }

    return result;
  }

  /**
   * Saves data to the given {@link BitStream}. This metod is optional.
   * <p>The stream should be opened and empty - the {@code reOpen(true,null)} call
   * is not required.
   *
   * @param  stream Output bit-stream (opened and empty).
   * @param  support Implementation-specific data (e.g. second {@link BitStream} for
   *                 file-formats with {@link #DOUBLE_STREAM} property).
   * @return {@code true} if the operation was successful.
   * @see    #saveFile(String,String)
   * @see    #DOUBLE_STREAM
   */
  @Override
  public boolean saveFile ( BitStream stream, Object support )
  throws IOException
  {
    if ( stream == null ) return false;

    final Brep b = (Brep)getInterface( PL_DATA, IFACE+"Brep" );
    if ( b == null )
      throw new JaGrLibException( "saveFile: vital modules are not connected!" );

    commonSave( stream, b );
    return true;
  }

  //==========================================================================
  //  Module registration:

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "DataFileFormatToBrepAndBitStream";

  /** Object category. */
  protected final static String CATEGORY = C_IO + '.' + C_3D + '.' + C_BREP;

}
