package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;

/**
 * Default implementation of {@link DiscreteTransform} interface.
 * 2D transforms are based on 1D ones (separable extension).
 *
 * @version 0.25 $Rev: 270 $ $Date: 2006-05-24 02:50:03 +0200 (st, 24 V 2006) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/DefaultDiscreteTransform.java">DefaultDiscreteTransform.java</a>
 */
public abstract class DefaultDiscreteTransform extends Piece implements DiscreteTransform, PersistentParamsBinary
{
  //==========================================================================
  //  Data:

  /** Temporary array for intermediate integer values. */
  private int[]    tmpI;

  /** Temporary array for intermediate double values. */
  private double[] tmpD;

  /** Temporary array for intermediate integer values. */
  private int[]    tmp2I;

  /** Temporary array for intermediate double values. */
  private double[] tmp2D;

  //==========================================================================
  //  Interface DiscreteTransform:

  /**
   * Get extreme domain (source) ranges.
   *
   * @param  bounds Pre-allocated array to hold extreme ranges:
   *         <br><tt>bounds[0]</tt> .. minimum size in 1st dim,
   *         <br><tt>bounds[1]</tt> .. maximum size in 1st dim,
   *         <br><tt>bounds[2]</tt> .. minimum size in 2nd dim,
   *         <br><tt>bounds[3]</tt> .. maximum size in 2nd dim, etc.
   * @return <tt>true</tt> if arbitrary lengths are allowed (even non-powers-ow-two).
   */
  public boolean getSizeBounds ( int[] bounds )
  {
    if ( bounds != null )
    {
      int i;
      for ( i = 0; i + 1 < bounds.length; )
      {
        bounds[ i++ ] =  1;
        bounds[ i++ ] = 32;
      }
    }
    return false;
  }

  /**
   * Forward 1D transform (integer form).
   * Can be used if both source and target data are in integer format.
   *
   * <p>Converts argument into <tt>double[]</tt> array and performs {@link #transform1D(double[],int,int,double[])}.
   * Result <tt>double[]</tt> array is rounded back into original result array.
   *
   * @param  sample Source array (containing sample data).
   * @param  start Starting index of the transformed segment.
   * @param  len Source segment length.
   * @param  coef Pre-allocated array to hold transformed data (coefficients).
   * @return Number of coefficients (<tt>0</tt> in case of error).
   * @see    #inverse1D(int[],int,int[],int)
   * @see    #transform1D(double[],int,int,double[])
   */
  public int transform1D ( int[] sample, int start, int len, int[] coef )
  {
    if ( len    < 1 ||
         sample == null ||
         sample.length < start + len ||
         coef   == null ||
         coef.length < len )
      return 0;

    if ( tmp2D == null || tmp2D.length < len + len )
      tmp2D = new double[ len + len ];

    int i;
    for ( i = 0; i < len; i++, start++ )
      tmp2D[i+len] = sample[start];

    if ( transform1D( tmp2D, len, len, tmp2D ) != len ) return 0;

    for ( i = 0; i < len; i++ )
      coef[i] = Formula.round( tmp2D[i] );

    return len;
  }

  /**
   * Forward 1D transform (floating-point form).
   * Has to be used if source and/or target data are floating-point!
   *
   * <p>Converts argument into <tt>int[]</tt> array and performs {@link #transform1D(int[],int,int,int[])}.
   *
   * @param  sample Source array (containing sample data).
   * @param  start Starting index of the transformed segment.
   * @param  len Source segment length.
   * @param  coef Pre-allocated array to hold transformed data (coefficients).
   * @return Number of coefficients (<tt>0</tt> in case of error).
   * @see    #inverse1D(double[],int,double[],int)
   * @see    #transform1D(int[],int,int,int[])
   */
  public int transform1D ( double[] sample, int start, int len, double[] coef )
  {
    if ( len    < 1 ||
         sample == null ||
         sample.length < start + len ||
         coef   == null ||
         coef.length < len )
      return 0;

    if ( tmp2I == null || tmp2I.length < len + len )
      tmp2I = new int[ len + len ];

    int i;
    for ( i = 0; i < len; i++, start++ )
      tmp2I[i+len] = Formula.round( sample[start] );

    if ( transform1D( tmp2I, len, len, tmp2I ) != len ) return 0;

    for ( i = 0; i < len; i++ )
      coef[i] = tmp2I[i];

    return len;
  }

  /**
   * Inverse 1D transform (integer form).
   * Inverse mapping to {@link #transform1D(int[],int,int,int[])}.
   *
   * <p>Converts argument into <tt>double[]</tt> array and performs {@link #inverse1D(double[],int,double[],int)}.
   * Result <tt>double[]</tt> array is rounded back into original result array.
   *
   * @param  coef Source array (holding coefficients).
   * @param  len Number of coefficients.
   * @param  sample Pre-allocated array to hold reconstructed sample data.
   * @param  start Starting index of the reconstructed sample data.
   * @return Sample data length (<tt>0</tt> in case of error).
   * @see    #transform1D(int[],int,int,int[])
   * @see    #inverse1D(double[],int,double[],int)
   */
  public int inverse1D ( int[] coef, int len, int[] sample, int start )
  {
    if ( len    < 1 ||
         sample == null ||
         sample.length < start + len ||
         coef   == null ||
         coef.length < len )
      return 0;

    if ( tmp2D == null || tmp2D.length < len + len )
      tmp2D = new double[ len + len ];

    int i;
    for ( i = 0; i < len; i++ )
      tmp2D[i] = coef[i];

    if ( inverse1D( tmp2D, len, tmp2D, len ) != len ) return 0;

    for ( i = 0; i < len; i++, start++ )
      sample[start] = Formula.round( tmp2D[i+len] );

    return len;
  }

  /**
   * Inverse 1D transform (floating-point form).
   * Inverse mapping to {@link #transform1D(double[],int,int,double[])}.
   *
   * <p>Converts argument into <tt>int[]</tt> array and performs {@link #inverse1D(int[],int,int[],int)}.
   *
   * @param  coef Source array (holding coefficients).
   * @param  len Number of coefficients.
   * @param  sample Pre-allocated array to hold reconstructed sample data.
   * @param  start Starting index of the reconstructed sample data.
   * @return Sample data length (<tt>0</tt> in case of error).
   * @see    #transform1D(double[],int,int,double[])
   * @see    #inverse1D(int[],int,int[],int)
   */
  public int inverse1D ( double[] coef, int len, double[] sample, int start )
  {
    if ( len    < 1 ||
         sample == null ||
         sample.length < start + len ||
         coef   == null ||
         coef.length < len )
      return 0;

    if ( tmp2I == null || tmp2I.length < len + len )
      tmp2I = new int[ len + len ];

    int i;
    for ( i = 0; i < len; i++ )
      tmp2I[i] = Formula.round( coef[i] );

    if ( inverse1D( tmp2I, len, tmp2I, len ) != len ) return 0;

    for ( i = 0; i < len; i++, start++ )
      sample[start] = tmp2I[i+len];

    return len;
  }

  /**
   * Forward 2D transform (integer form).
   * Can be used if both source and target data are in integer format.
   *
   * @param  sample Source array (containing sample data).
   * @param  start Starting index of the transformed data.
   * @param  width Source rectangle's width (faster index).
   * @param  height Source rectangle's height (slower index).
   * @param  stride Source matrix row length.
   * @param  coef Pre-allocated array to hold transformed data (coefficients).
   * @return Number of coefficients (<tt>0</tt> in case of error).
   * @see    #inverse2D(int[],int,int[],int,int,int)
   * @see    #transform2D(double[],int,int,int,int,double[])
   */
  public int transform2D ( int[] sample, int start, int width, int height, int stride, int[] coef )
  {
    int len = width * height;
    if ( sample == null || start + width + (height - 1) * stride > sample.length ||
         coef   == null || len > coef.length )
      return 0;

    if ( height == 1 )
      return transform1D( sample, start, width, coef );

      // array dimensions and indices are OK
    int i, j, ptr1, ptr2;

      // I - horizontal 1D transform, transpose intermediate data into tmp array:
    if ( tmpI == null || tmpI.length < len )
      tmpI = new int[ len ];

    ptr1 = start;
    for ( i = ptr2 = 0; i < height; i++, ptr1 += stride )
    {                                     // transform one sample row into coef row
      if ( transform1D( sample, ptr1, width, coef ) != width ) return 0;

      for ( j = 0, ptr2 = i; j < width; j++, ptr2 += height )
        tmpI[ptr2] = coef[j];
    }

      // II - transform data columns, transpose them into result array:
    for ( i = ptr1 = 0; i < width; i++, ptr1 += height )
    {                                     // transform one column
      if ( transform1D( tmpI, ptr1, height, tmpI ) != height ) return 0;

      for ( j = 0, ptr2 = i; j < height; j++, ptr2 += width )
        coef[ptr2] = tmpI[j];
    }

    return len;
  }

  /**
   * Forward 2D transform (floating-point form).
   * Has to be used if source and/or target data are floating-point!
   *
   * @param  sample Source array (containing sample data).
   * @param  start Starting index of the transformed data.
   * @param  width Source rectangle's width (faster index).
   * @param  height Source rectangle's height (slower index).
   * @param  stride Source matrix row length.
   * @param  coef Pre-allocated array to hold transformed data (coefficients).
   * @return Number of coefficients (<tt>0</tt> in case of error).
   * @see    #inverse2D(double[],int,double[],int,int,int)
   * @see    #transform2D(int[],int,int,int,int,int[])
   */
  public int transform2D ( double[] sample, int start, int width, int height, int stride, double[] coef )
  {
    int len = width * height;
    if ( sample == null || start + width + (height - 1) * stride > sample.length ||
         coef   == null || len > coef.length )
      return 0;

    if ( height == 1 )
      return transform1D( sample, start, width, coef );

      // array dimensions and indices are OK
    int i, j, ptr1, ptr2;

      // I - horizontal 1D transform, transpose intermediate data into tmp array:
    if ( tmpD == null || tmpD.length < len )
      tmpD = new double[ len ];

    ptr1 = start;
    ptr2 = 0;
    for ( i = 0; i < height; i++, ptr1 += stride )
    {                                     // transform one sample row into coef row
      if ( transform1D( sample, ptr1, width, coef ) != width ) return 0;

      ptr2 = i;
      for ( j = 0; j < width; j++, ptr2 += height )
        tmpD[ptr2] = coef[j];
    }

      // II - transform data columns, transpose them into result array:
    ptr1 = 0;
    for ( i = 0; i < width; i++, ptr1 += height )
    {                                     // transform one column
      if ( transform1D( tmpD, ptr1, height, tmpD ) != height ) return 0;

      ptr2 = i;
      for ( j = 0; j < height; j++, ptr2 += width )
        coef[ptr2] = tmpD[j];
    }

    return len;
  }

  /**
   * Inverse 2D transform (integer form).
   * Inverse mapping to {@link #transform2D(int[],int,int,int,int,int[])}.
   *
   * @param  coef Source array (holding coefficients).
   * @param  len Number of coefficients.
   * @param  sample Pre-allocated array to hold reconstructed sample data.
   * @param  start Starting index of the reconstructed rectangle.
   * @param  width Reconstructed rectangle's width (faster index).
   * @param  stride Reconstructed data matrix row length.
   * @return Sample data height (<tt>0</tt> in case of error).
   * @see    #transform2D(int[],int,int,int,int,int[])
   * @see    #inverse2D(double[],int,double[],int,int,int)
   */
  public int inverse2D ( int[] coef, int len, int[] sample, int start, int width, int stride )
  {
    int height = len / width;
    if ( sample == null || start + width + (height - 1) * stride > sample.length ||
         coef   == null || len > coef.length )
     return 0;

    if ( height == 1 )
    {
      inverse1D( coef, width, sample, start );
      return 1;
    }

     // array dimensions and indices are OK
   int i, j, ptr1, ptr2, ptr3;

     // I - vertical inverse 1D transform, intermediate results are stored in sample array:
   int beg = Math.max( width, height );
   if ( tmpI == null || tmpI.length < len + beg )
     tmpI = new int[ len + beg ];

   ptr3 = beg;
   for ( j = 0; j < width; j++, ptr3 += height ) // transform one source column
   {
     ptr1 = j;
     for ( i = 0; i < height; i++, ptr1 += width )
       tmpI[i] = coef[ptr1];

     if ( inverse1D( tmpI, height, tmpI, ptr3 ) != height ) return 0;
   }

     // II - horizontal inverse 1D transforms, results are stored in sample array:
   ptr1 = beg;
   ptr3 = start;
   for ( j = 0; j++ < height; ptr1++, ptr3 += stride ) // one result row
   {
     ptr2 = ptr1;
     for ( i = 0; i < width; i++, ptr2 += height )
       tmpI[i] = tmpI[ptr2];

     if ( inverse1D( tmpI, width, sample, ptr3 ) != width ) return 0;
   }

    return height;
  }

  /**
   * Inverse 2D transform (floating-point form).
   * Inverse mapping to {@link #transform2D(double[],int,int,int,int,double[])}.
   *
   * @param  coef Source array (holding coefficients).
   * @param  len Number of coefficients.
   * @param  sample Pre-allocated array to hold reconstructed sample data.
   * @param  start Starting index of the reconstructed rectangle.
   * @param  width Reconstructed rectangle's width (faster index).
   * @param  stride Reconstructed data matrix row length.
   * @return Sample data height (<tt>0</tt> in case of error).
   * @see    #transform2D(double[],int,int,int,int,double[])
   * @see    #inverse2D(int[],int,int[],int,int,int)
   */
  public int inverse2D ( double[] coef, int len, double[] sample, int start, int width, int stride )
  {
    int height = len / width;
    if ( sample == null || start + width + (height - 1) * stride > sample.length ||
         coef   == null || len > coef.length )
     return 0;

    if ( height == 1 )
    {
      inverse1D( coef, width, sample, start );
      return 1;
    }

     // array dimensions and indices are OK
   int i, j, ptr1, ptr2, ptr3;

     // I - vertical inverse 1D transform, intermediate results are stored in sample array:
   int beg = Math.max( width, height );
   if ( tmpD == null || tmpD.length < len + beg )
     tmpD = new double[ len + beg ];

   ptr3 = beg;
   for ( j = 0; j < width; j++, ptr3 += height ) // transform one source column
   {
     ptr1 = j;
     for ( i = 0; i < height; i++, ptr1 += width )
       tmpD[i] = coef[ptr1];

     if ( inverse1D( tmpD, height, tmpD, ptr3 ) != height ) return 0;
   }

     // II - horizontal inverse 1D transforms, results are stored in sample array:
   ptr1 = beg;
   ptr3 = start;
   for ( j = 0; j++ < height; ptr1++, ptr3 += stride ) // one result row
   {
     ptr2 = ptr1;
     for ( i = 0; i < width; i++, ptr2 += height )
       tmpD[i] = tmpD[ptr2];

     if ( inverse1D( tmpD, width, sample, ptr3 ) != width ) return 0;
   }

    return height;
  }

  //==========================================================================
  //  Interface PersistentParamsBinary:

  /**
   * Pack instance parameters into binary format.
   *
   * @return Data array to be stored into binary stream. Can be <tt>null</tt>.
   *         The array can be shared by all method calls.
   * @see    #loadData
   */
  public byte[] storeData ()
  {
    return null;
  }

  /**
   * Loads instance parameters from (previously stored) binary data.
   *
   * @param  data Binary data to be interpreted. Can be <tt>null</tt>.
   * @return <tt>true</tt> if successful.
   * @see    #storeData
   */
  public boolean loadData ( byte[] data )
  {
    return( data == null );
  }

  //==========================================================================
  //  Module registration:

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "DiscreteTransform";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_TRANSFORM;

}
