package cz.cuni.jagrlib;

import java.io.*;

import cz.cuni.jagrlib.iface.*;

/**
 * File format default implementation.
 *
 * @version 0.26 $Rev: 551 $ $Date: 2007-10-10 19:58:19 +0200 (st, 10 X 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../cz/cuni/jagrlib/DefaultFileFormat.java">DefaultFileFormat.java</a>
 */
public abstract class DefaultFileFormat extends Piece implements DataFileFormat
{
  //==========================================================================
  //  Support routines:

  /**
   * Write the given string followed by system newline character.
   */
  public static void writeLine ( OutputStream os, String s )
  throws IOException
  {
    if ( s != null )                      // write nonempty string first..
      os.write( s.getBytes() );

      // finish the line:
    os.write( Formula.LINE_SEPARATOR_BYTES );
  }

  /**
   * Read next nonempty line.
   * Returns {@code null} if no line is available.
   */
  public static String readLine ( InputStream is )
  throws IOException
  {
    int ch;

      // skip initial newlines:
    while ( (ch = is.read()) >= 0 && (ch == '\n' || ch == '\r') ) ;

    if ( ch < 0 ) return null;

    final StringBuffer sb = new StringBuffer();
    do
    {
      sb.append( (char)ch );
      ch = is.read();
    }
    while ( ch >= 0 && ch != '\n' && ch != '\r' );

    return new String( sb );
  }

  /**
   * Read character token delimited by an white-space.
   * Skips initial white-spaces.
   */
  public static String readToken ( InputStream is )
  throws IOException
  {
    int ch;
    do
      ch = is.read();
    while ( ch >= 0 && Character.isWhitespace( ch ) );

    if ( ch < 0 ) return null;

    StringBuffer sb = new StringBuffer();
    while ( ch >= 0 && !Character.isWhitespace( ch ) )
    {
      sb.append( (char)ch );
      ch = is.read();
    }

    return new String( sb );
  }

  /**
   * Read one binary float number.
   *
   * @param  is Stream to read from.
   * @param  littleEndian {@code true} for little endian.
   */
  public static float readFloat ( InputStream is, boolean littleEndian )
  throws IOException
  {
    int buf = is.read();
    if ( littleEndian )
    {
      buf |= (is.read() <<  8);
      buf |= (is.read() << 16);
      buf |= (is.read() << 24);
    }
    else
    {
      buf = (buf << 8) | is.read();
      buf = (buf << 8) | is.read();
      buf = (buf << 8) | is.read();
    }

    return Float.intBitsToFloat( buf );
  }

  /**
   * Writes binary float number in little endian format.
   */
  public static void writeFloat ( OutputStream os, float fl )
  throws IOException
  {
    int buf = Float.floatToIntBits( fl );
    os.write( buf & 0xff );
    buf >>>= 8;
    os.write( buf & 0xff );
    buf >>>= 8;
    os.write( buf & 0xff );
    buf >>>= 8;
    os.write( buf & 0xff );
  }

  /** Reads 16-bit unsigned integer from the given bit-stream. */
  public final static int readUnsigned16 ( BitStream s, boolean littleEndian )
  throws IOException
  {
    long r = s.read( 16 );
    if ( !littleEndian )
      r = Formula.swapBytes( r, 16 );
    return (int)r;
  }

  /** Writes 16-bit unsigned integer to the given bit-stream. */
  public final static void writeUnsigned16 ( BitStream s, int data, boolean littleEndian )
  throws IOException
  {
    long d = littleEndian ? data : Formula.swapBytes( data, 16 );
    s.write( d, 16 );
  }

  /** Reads 32-bit unsigned integer from the given bit-stream. */
  public final static long readUnsigned32 ( BitStream s, boolean littleEndian )
  throws IOException
  {
    long r = s.read( 32 );
    if ( !littleEndian )
      r = Formula.swapBytes( r, 32 );
    return r;
  }

  /** Writes 32-bit unsigned integer to the given bit-stream. */
  public final static void writeUnsigned32 ( BitStream s, long data, boolean littleEndian )
  throws IOException
  {
    if ( !littleEndian )
      data = Formula.swapBytes( data, 32 );
    s.write( data, 32 );
  }

  /** Reads 16-bit unsigned integer from the given codec stream (big endian). */
  public final static int readUnsigned16 ( EntropyCodec codec )
  throws IOException
  {
    int result = (int)codec.getBits( 8 ) << 8;
    return( result + (int)codec.getBits( 8 ) );
  }

  /** Stores the given 16-bit unsigned integer in big-endian order. */
  public final static void writeUnsigned16 ( EntropyCodec codec, int data )
  throws IOException
  {
    codec.putBits( (data>>8) & 0xff, 8 );
    codec.putBits(  data     & 0xff, 8 );
  }

  /**
   * Reads and sets binary parameters to the given module instance.
   * @return {@code true} if object was successfully set up.
   */
  public final static boolean readParameters ( EntropyCodec codec, Object instance )
  throws IOException
  {
      // I. read hash number based on class-name:
    int hash     = Formula.hashClassName( instance ) & 0xffff;
    int readHash = readUnsigned16( codec );
    if ( hash != readHash ) return false;

      // II. read binary parameters (if object has PersistentParamsBinary type):
    byte[] params    = null;
    int    paramsLen = readUnsigned16( codec );

    if ( paramsLen > 0 )
    {
      params = new byte[ paramsLen ];
      for ( int i = 0; i < paramsLen; i++ )
        params[i] = (byte)codec.getBits( 8 );
    }

    if ( instance instanceof PersistentParamsBinary )
       return( ((PersistentParamsBinary)instance).loadData( params ) );

    return true;
  }

  /** Writes binary parameters of the given module instance. */
  public final static void writeParameters ( EntropyCodec codec, Object instance )
  throws IOException
  {
      // I. write hash number based on class-name:
    int hash = Formula.hashClassName( instance ) & 0xffff;
    writeUnsigned16( codec, hash );

      // II. write binary parameters (if object has PersistentParamsBinary type):
    byte[] params;
    int    paramsLen;
    if ( instance instanceof PersistentParamsBinary &&
         (params = ((PersistentParamsBinary)instance).storeData()) != null &&
         (paramsLen = params.length) > 0 )
    {
      writeUnsigned16( codec, paramsLen );
      for ( int i = 0; i < paramsLen; i++ )
        codec.putBits( params[i], 8 );
    }
    else
      writeUnsigned16( codec, 0 );
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  public abstract int headerLength ();

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type ({@code 0.0} .. no way,
   *         {@code 0.5} .. maybe, {@code 0.9} .. almost sure, {@code 1.0} ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  public abstract double match ( byte[] header, String fileName );

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.png"]) or {@code null}.
   * @see    #match
   */
  public abstract String[] fileNameMasks ();

  /**
   * Retrieves data from the given file (stream).
   * Can be used for "file opening" in virtual-mapped mode.
   *
   * @param  fileName File-name (stream-name) to read from (can be {@code null}).
   * @param  detail Further implementation-specific information ("r" or "rw" for mapped mode).
   * @return {@code true} if the operation was successful.
   * @see    #loadFile(BitStream,Object)
   * @see    #saveFile(String,String)
   * @see    #closeFile
   */
  public abstract boolean loadFile ( String fileName, String detail )
  throws IOException;

  /**
   * Retrieves data from the given {@link BitStream}. This metod is optional.
   * <p>The stream should be opened before - only the {@code reOpen(false,null)} call
   * can be used.
   *
   * @param  stream Input bit-stream (opened - call {@code reOpen(false,null)} first).
   * @param  support Implementation-specific data (e.g. second {@link BitStream} for
   *                 file-formats with {@link #DOUBLE_STREAM} property).
   * @return {@code true} if the operation was successful.
   * @see    #loadFile(String,String)
   * @see    #DOUBLE_STREAM
   */
  public abstract boolean loadFile ( BitStream stream, Object support )
  throws IOException;

  /**
   * Saves data to the given file (stream).
   *
   * @param  fileName File-name (stream-name) to write to (can be {@code null}).
   * @param  detail Further implementation-specific information.
   * @return {@code true} if the operation was successful.
   * @see    #saveFile(BitStream,Object)
   * @see    #loadFile(String,String)
   */
  public abstract boolean saveFile ( String fileName, String detail )
  throws IOException;

  /**
   * Saves data to the given {@link BitStream}. This metod is optional.
   * <p>The stream should be opened and empty - the {@code reOpen(true,null)} call
   * is not required.
   *
   * @param  stream Output bit-stream (opened and empty).
   * @param  support Implementation-specific data (e.g. second {@link BitStream} for
   *                 file-formats with {@link #DOUBLE_STREAM} property).
   * @return {@code true} if the operation was successful.
   * @see    #saveFile(String,String)
   * @see    #DOUBLE_STREAM
   */
  public abstract boolean saveFile ( BitStream stream, Object support )
  throws IOException;

  /**
   * Closes the file (used in file-mapping mode).
   *
   * @see    #loadFile(String,String)
   */
  public void closeFile ()
  throws IOException
  {
  }

  /**
   * Module initialization - called before any {@link Worker#run}.
   * Individual phases (starting at {@code 0}) will be called until maximum
   * requested phase number is reached.
   *
   * @param  phase Current phase number.
   * @return Number of phases needed to initialize this module.
   */
  @Override
  public int init ( int phase )
  {
    BitStream stream;

    if ( isConnected( PL_INPUT ) ||
         !isConnected( PL_STREAM ) ||
         (stream = (BitStream)getInterface( PL_STREAM, IFACE+"BitStream" )) == null ||
         !stream.canRead() )
      return 0;

    if ( phase == 0 )
      try
      {
        loadFile( (String)null, null );
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error reading data file!", e );
      }

    return 1;
  }
}
