package cz.cuni.jagrlib;

import java.util.*;
import java.util.List;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import cz.cuni.jagrlib.iface.*;

/**
 * Default implementation of {@link GraphicsViewer}.
 *
 * @version 0.26 $Rev: 451 $ $Date: 2007-05-09 12:33:35 +0200 (st, 09 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/DefaultGraphicsViewer.java">DefaultGraphicsViewer.java</a>
 */
public abstract class DefaultGraphicsViewer extends Piece implements GraphicsViewer
{
  //==========================================================================
  //  Data:

  /** Force automatic start of preview window. */
  protected boolean autoStart = false;

  /** Minimal X coordinate of actual invalid region. */
  protected int x1 = Integer.MAX_VALUE;

  /** Maximal X coordinate (outer bound) of actual invalid region. */
  protected int x2 = Integer.MIN_VALUE;

  /** Minimal Y coordinate of actual invalid region. */
  protected int y1 = Integer.MAX_VALUE;

  /** Maximal Y coordinate (outer bound) of actual invalid region. */
  protected int y2 = Integer.MIN_VALUE;

  /** Frame title (in order that {@link JFrame#setTitle(java.lang.String)} works every time). */
  protected String frameTitle = "Graphics";

  /** Swing frame window to show graphic output in. */
  protected JFrame frame;

  /** Custom component - drawing target. */
  protected JComponent canvas;

  /**
   * Breaks the <b>repaint loop</b> from outside the {@link #repaintLoop} call.
   *
   * @see    #repaintLoop
   * @see    #stopRepaintLoop
   */
  protected boolean continueRepaintLoop;

  /** Has the user-break occurred? */
  protected boolean userBreak;

  /**
   * Semaphores that will be signaled after the viewer window is closed.
   *
   * @see    Semaphore
   * @see    #addSemaphore
   * @see    #removeSemaphore
   */
  protected List< Semaphore > semaphores;

  /**
   * Set of actual input listeners.
   *
   * @see    #addInputListener
   * @see    #removeInputListener
   */
  protected List< InputListener > listeners;

  /** InputListener connected through "direct" plug. */
  protected InputListener listener;

  /** Set of HUD primitives. */
  protected IntMap< HudPrimitive > hudSet;

  /** Next free HUD id (or id from which a future search will start). */
  protected int nextFreeId = 0;

  //==========================================================================
  //  HUD drawing:

  /** Protocol for HUD drawable primitives. */
  protected interface HudPrimitive
  {
    /** Draws this primitive using the given graphics. */
    void draw ( Graphics g );
  }

  /** Line for HUD drawing. */
  protected static class HudLine implements HudPrimitive
  {
    /** X coordinate of starting point. */
    protected int x1;

    /** Y coordinate of starting point. */
    protected int y1;

    /** X coordinate of ending point. */
    protected int x2;

    /** Y coordinate of ending point. */
    protected int y2;

    public HudLine ( int xa, int ya, int xb, int yb )
    {
      x1 = xa; y1 = ya;
      x2 = xb; y2 = yb;
    }

    /** Draws this primitive using the given graphics. */
    public void draw ( Graphics g )
    {
      g.drawLine( x1, y1, x2, y2 );
    }
  }

  //==========================================================================
  //  Support classes:

  /**
   * Inner class responsible for window-close event handling (by user interaction).
   *
   * @version 0.25 $Rev: 451 $ $Date: 2007-05-09 12:33:35 +0200 (st, 09 V 2007) $ $Author: pepca $
   * @since   0.13
   */
  protected class LocalWindowListener extends WindowAdapter
  {
    /** Close-window event handler. */
    @Override
    public void windowClosing ( WindowEvent e )
    {
      synchronized ( DefaultGraphicsViewer.this )
      {
        userBreak = true;
        continueRepaintLoop = false;
        DefaultGraphicsViewer.this.notify();
      }

      if ( listener != null )
        listener.windowClose( DefaultGraphicsViewer.this );

      if ( listeners != null )
        for ( InputListener il : listeners )
          il.windowClose( DefaultGraphicsViewer.this );

        // signal the semaphores:
      if ( semaphores != null )           // signal all the semaphores
        for ( Semaphore s : semaphores )
          s.semSignal( 1 );

        // close the window manually:
      e.getWindow().dispose();
    }
  }

  protected class LocalMouseListener extends MouseAdapter
  {
    @Override
    public void mousePressed ( MouseEvent e )
    {
      if ( listener  == null &&
           listeners == null ) return;

      e.consume();
      if ( listener != null )
        listener.mouseButton( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), e.getButton(), true, 0 );
      if ( listeners == null ) return;

      //LogFile.debug( "Press(" + e.getButton() + "): " + e.getWhen() );
      for ( InputListener il : listeners )
        il.mouseButton( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), e.getButton(), true, 0 );
    }

    @Override
    public void mouseReleased ( MouseEvent e )
    {
      if ( listener  == null &&
           listeners == null ) return;

      e.consume();
      if ( listener != null )
        listener.mouseButton( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), e.getButton(), false, 0 );
      if ( listeners == null ) return;

      //LogFile.debug( "Release(" + e.getButton() + "): " + e.getWhen() );
      for ( InputListener il : listeners )
        il.mouseButton( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), e.getButton(), false, 0 );
    }

    @Override
    public void mouseEntered ( MouseEvent e )
    {
      if ( listener != null )
        listener.mouseMove( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), 0 );
      if ( listeners == null ) return;

      for ( InputListener il : listeners )
        il.mouseMove( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), 0 );
    }
  }

  protected class LocalMouseMotionListener implements MouseMotionListener
  {
    public void mouseDragged ( MouseEvent e )
    {
      if ( listener  == null &&
           listeners == null ) return;

      e.consume();
      if ( listener != null )
        listener.mouseMove( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), 0 );
      if ( listeners == null ) return;

      for ( InputListener il : listeners )
        il.mouseMove( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), 0 );
    }

    public void mouseMoved ( MouseEvent e )
    {
      if ( listener == null && listeners == null ) return;

      e.consume();
      if ( listener != null )
        listener.mouseMove( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), 0 );
      if ( listeners == null ) return;

      for ( InputListener il : listeners )
        il.mouseMove( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), 0 );
    }
  }

  protected class LocalMouseWheelListener implements MouseWheelListener
  {
    public void mouseWheelMoved ( MouseWheelEvent e )
    {
      if ( listener == null && listeners == null ) return;

      e.consume();
      if ( listener != null )
        listener.mouseWheel( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), e.getButton(), e.getWheelRotation(), 0 );
      if ( listeners == null ) return;

      for ( InputListener il : listeners )
        il.mouseWheel( DefaultGraphicsViewer.this, e.getWhen(), e.getX(), e.getY(), e.getButton(), e.getWheelRotation(), 0 );
    }
  }

  protected class LocalKeyListener extends KeyAdapter
  {
    @Override
    public void keyPressed ( KeyEvent e )
    {
      if ( listener == null && listeners == null ) return;

      e.consume();
      if ( listener != null )
        listener.keyboard( DefaultGraphicsViewer.this, e.getWhen(), true, e.getKeyChar(), e.getKeyCode(), e.getModifiersEx() );
      if ( listeners == null ) return;

      for ( InputListener il : listeners )
        il.keyboard( DefaultGraphicsViewer.this, e.getWhen(), true, e.getKeyChar(), e.getKeyCode(), e.getModifiersEx() );
    }

    @Override
    public void keyReleased ( KeyEvent e )
    {
      if ( listener == null && listeners == null ) return;

      e.consume();
      if ( listener != null )
        listener.keyboard( DefaultGraphicsViewer.this, e.getWhen(), false, e.getKeyChar(), e.getKeyCode(), e.getModifiersEx() );
      if ( listeners == null ) return;

      for ( InputListener il : listeners )
        il.keyboard( DefaultGraphicsViewer.this, e.getWhen(), false, e.getKeyChar(), e.getKeyCode(), e.getModifiersEx() );
    }
  }

  protected class LocalComponentListener extends ComponentAdapter
  {
    @Override
    public void componentResized ( ComponentEvent e )
    {
      if ( listener == null && listeners == null ) return;

      Component c = e.getComponent();
      if ( listener != null )
        listener.setWindowSize( DefaultGraphicsViewer.this, c.getWidth(), c.getHeight() );
      if ( listeners == null ) return;

      for ( InputListener il : listeners )
        il.setWindowSize( DefaultGraphicsViewer.this, c.getWidth(), c.getHeight() );
    }
  }

  //==========================================================================
  //  Support routines:

  /**
   * Does the repaint job itself (according the <b>invalid region</b>).
   *
   * @see    #repaintLoop
   */
  protected void doRepaint ()
  {
    if ( canvas == null ||
         x2     <= x1   ||
         y2     <= y1 ) return;

    canvas.repaint( canvas.getX() + x1, canvas.getY() + y1, x2 - x1, y2 - y1 );
    x1 = y1 = Integer.MAX_VALUE;
    x2 = y2 = Integer.MIN_VALUE;
  }

  /** Returns drawing component or {@code null} if failed. */
  protected abstract JComponent getCanvas ();

  //==========================================================================
  //  Interface GraphicsViewer:

  /**
   * Start (potentially) infinite <b>repaint loop</b>. It can be terminated either by a
   * {@link #stopRepaintLoop} call or by external user interaction (user break, window
   * close, ..).
   *
   * @return {@code true} if the loop was terminated by a {@link #stopRepaintLoop} call.
   * @see    #stopRepaintLoop
   */
  public synchronized boolean repaintLoop ()
  {
    if ( frame == null )                  // looping for the first time
    {
      if ( (canvas = getCanvas()) == null )
        return false;

      frame    = new JFrame( frameTitle );
      listener = (InputListener)getInterface( PL_DIRECT, IFACE+"InputListener" );
      frame.addWindowListener( new LocalWindowListener() );
      frame.getContentPane().add( canvas, BorderLayout.CENTER );

        // GUI listeners:
      frame.addKeyListener(          new LocalKeyListener() );
      canvas.addMouseListener(       new LocalMouseListener() );
      canvas.addMouseMotionListener( new LocalMouseMotionListener() );
      canvas.addMouseWheelListener(  new LocalMouseWheelListener() );
      canvas.addComponentListener(   new LocalComponentListener() );

        // expose the frame:
      frame.pack();
      frame.setVisible( true );
    }

    continueRepaintLoop = true;           // looping state variables
    userBreak           = false;
    canvas.repaint();

    do                                    // repaint-loop itself
    {
      try
      {
        wait();
      }
      catch ( InterruptedException e )
      {
        continueRepaintLoop = false;
        userBreak           = true;
      }
      if ( !continueRepaintLoop ) break;

      doRepaint();

    } while ( continueRepaintLoop );

    return !userBreak;
  }

  /**
   * Stops the running <b>repaint loop</b>.
   *
   * @return {@code true} if the loop was successfully stopped.
   * @see    #repaintLoop
   */
  public synchronized boolean stopRepaintLoop ()
  {
    if ( frame == null || !continueRepaintLoop ) return false;

    continueRepaintLoop = false;
    notify();
    return true;
  }

  /**
   * Destroys the preview window.
   * Stops repaint-loop eventually.
   */
  public synchronized void destroy ()
  {
    if ( frame != null )
    {
      if ( continueRepaintLoop )
      {
        continueRepaintLoop = false;
        notify();
      }
      frame.setVisible( false );
      frame = null;
    }
  }

  /**
   * Causes the instant <b>repainting</b> of the whole graphic output (regardless of the actual
   * <b>invalid region</b>). {@link Trigger#fire} call has the same effect..
   *
   * @see    #invalidate
   * @see    #repaint
   * @see    Trigger#fire
   */
  public synchronized void repaintAll ()
  {
    if ( frame == null || !continueRepaintLoop ) return;

    x1 = 0; x2 = canvas.getWidth();
    y1 = 0; y2 = canvas.getHeight();
    notify();
  }

  /**
   * Invalidates the given rectangular region of the graphics output. Invalid regions will be
   * repainted during the next {@link #repaint} call...
   *
   * @param  xMin X coordinate of the upper left corner.
   * @param  xMax X coordinate of the lower right corner (outside pixel).
   * @param  yMin Y coordinate of the upper left corner.
   * @param  yMax Y coordinate of the upper left corner (outside pixel).
   * @see    #repaint
   */
  public synchronized void invalidate ( int xMin, int xMax, int yMin, int yMax )
  {
    if ( xMin < x1 ) x1 = xMin;
    if ( xMax > x2 ) x2 = xMax;
    if ( yMin < y1 ) y1 = yMin;
    if ( yMax > y2 ) y2 = yMax;
  }

  /**
   * Causes the instant <b>repainting</b> of actual <b>invalid region</b> of the graphic output.
   *
   * @see    #repaintAll
   * @see    #invalidate
   */
  public synchronized void repaint ()
  {
    if ( frame == null || !continueRepaintLoop ) return;

    notify();
  }

  /**
   * Adds a new <b>semaphore</b> - object which wants to be signaled after the viewer window
   * is closed (by user interaction etc.).
   *
   * @param  s A new semaphore instance.
   * @see    #removeSemaphore
   */
  public synchronized void addSemaphore ( Semaphore s )
  {
    if ( s == null ) return;

    if ( semaphores == null )             // the first semaphore
      semaphores = new LinkedList< Semaphore >();

    semaphores.add( s );
  }

  /**
   * Removes one of associated <b>semaphores</b>.
   *
   * @param  s Semaphore to be removed.
   * @see    #addSemaphore
   */
  public synchronized void removeSemaphore ( Semaphore s )
  {
    if ( s == null || semaphores == null ) return;

    semaphores.remove( s );
    if ( semaphores.size() == 0 )
      semaphores = null;
  }

  /** Adds input listener object. */
  public void addInputListener ( InputListener list )
  {
    if ( list == null ) return;

    if ( listeners == null )              // the fist listener
      listeners = new LinkedList< InputListener >();
    listeners.add( list );
  }

  /** Removes (previously added) input listener object. */
  public void removeInputListener ( InputListener list )
  {
    if ( list == null || listeners == null ) return;

    listeners.remove( list );
    if ( listeners.size() == 0 )
      listeners = null;
  }

  /** Removes all HUD primitives. */
  public void hudClear ()
  {
    if ( hudSet != null )
      hudSet.clear();
    nextFreeId = 0;
  }

  /** Returns id of any free HUD slot. */
  public int hudGetFreeId ()
  {
    return nextFreeId;
  }

  /**
   * Replaces the given HUD primitive slot with line from {@code [x1,y1]} to {@code [x2,y2]}.
   */
  public void hudSetLine ( int id, int x1, int y1, int x2, int y2 )
  {
    if ( hudSet == null )
      hudSet = new IntMap<HudPrimitive>();

    HudPrimitive old = hudSet.put( id, new HudLine( x1, y1, x2, y2 ) );
    if ( old != null ) return;            // replace

    if ( nextFreeId == id )
      while ( hudSet.containsKey( ++nextFreeId ) ) ;
  }

  /** Removes existing HUD primitive. */
  public void hudRemove ( int id )
  {
    if ( hudSet == null ) return;

    HudPrimitive old = hudSet.remove( id );
    if ( old == null ) return;

    if ( id < nextFreeId ) nextFreeId = id;
  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Triggers the (undetermined) action.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public synchronized boolean fire ( int type )
  {
    repaintAll();
    return true;
  }

  /**
   * Module initialization - called before any {@link Worker#run}.
   * Individual phases (starting at {@code 0}) will be called until maximum
   * requested phase number is reached.
   *
   * @param  phase Current phase number.
   * @return Number of phases needed to initialize this module.
   */
  @Override
  public int init ( int phase )
  {
    if ( isConnected( PL_INPUT ) &&
         !autoStart )
      return 0;

    if ( phase == 1 )
    {
      PreviewThread runIt = new PreviewThread( this );
      runIt.safeStart();
    }

    return 2;
  }

  //==========================================================================
  //  Module registration:

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_VIEWER;

}
