package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;

/**
 * Default implementation of {@link ImageFunction}.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/DefaultImageFunction.java">DefaultImageFunction.java</a>
 */
public abstract class DefaultImageFunction extends Piece implements ImageFunction
{
  //==========================================================================
  //  Data:

  /** Background color. */
  protected double[] background = { 0.0, 0.0, 0.0 };

  /** Actual time. */
  protected double time = 0.0;

  /** Image function sampling bounds - horizontal minimum. */
  protected double xMin = 0.0;

  /** Image function sampling bounds - horizontal maximum. */
  protected double xMax = 1.0;

  /** Image function sampling bounds - vertical minimum. */
  protected double yMin = 0.0;

  /** Image function sampling bounds - vertical maximum. */
  protected double yMax = 1.0;

  //--------------------------------------------------------------------------
  //  Support routines:

  //==========================================================================
  //  Interface TimeDependent:

  /**
   * Retrieve the valid bounds for time variable.
   *
   * @param  pre Pre-allocated {@code double[2]} array to hold the result (can be
   *             {@code null}).
   * @return {@code result[0]} .. minimum time value, {@code result[1]} .. maximum
   *         time value.
   * @see    #getTime
   * @see    #setTime
   */
  public double[] getTimeInterval ( double[] pre )
  {
    if ( pre == null ||
         pre.length < 2 )
      pre = new double[ 2 ];

    pre[0] = 0.0;
    pre[1] = 1.0;

    return pre;
  }

  /**
   * Returns the actual time value for the object (used in animations).
   *
   * @return Actual time (in seconds).
   * @see    #setTime
   */
  public double getTime ()
  {
    return time;
  }

  /**
   * Sets the actual time value for the object (used in animations).
   *
   * @param  time New scene time (in seconds).
   * @see    #getTime
   * @see    #getTimeInterval
   */
  public void setTime ( double time )
  {
    this.time = Formula.clamp( time, 0.0, 1.0 );
  }

  //==========================================================================
  //  Interface ImageFunction:

  /**
   * Sets bounds (scaling factors) for the image function (e.g. 2D mapping).
   *
   * @param  Xmin Minimum X coordinate.
   * @param  Xmax Maximum X coordinate.
   * @param  Ymin Minimum Y coordinate.
   * @param  Ymax Maximum Y coordinate.
   * @see    #getBounds
   */
  public void setBounds ( double Xmin, double Xmax, double Ymin, double Ymax )
  {
    xMin = Xmin;
    xMax = Xmax;
    if ( xMax <= xMin ) xMax = xMin + 1.0;

    yMin = Ymin;
    yMax = Ymax;
    if ( yMax <= yMin ) yMax = yMin + 1.0;
  }

  /**
   * Returns actual bounds for {@code [x,y]} (2D coordinates).
   *
   * @param  pre Pre-allocated (non-mandatory) {@code double[4]} array to hold results:
   *             {@code pre[0]} .. X<sub>min</sub>, {@code pre[1]} .. X<sub>max</sub>,
   *             {@code pre[2]} .. Y<sub>min</sub>, {@code pre[3]} .. Y<sub>max</sub>.
   * @return Array with all 4 return values.
   * @see    #setBounds
   * @see    TimeDependent#getTimeInterval
   */
  public double[] getBounds ( double[] pre )
  {
    if ( pre == null ||
         pre.length < 4 )
      pre = new double[ 4 ];

    pre[0] = xMin;
    pre[1] = xMax;
    pre[2] = yMin;
    pre[3] = yMax;

    return pre;
  }

  /**
   * Computes one image sample.
   *
   * @param  x X coordinate of the sample.
   * @param  y Y coordinate of the sample.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (sample signature) used for adaptive subsampling.
   * @see    #getSample(double,double,int,int,double[])
   * @see    #getBounds
   */
  public long getSample ( double x, double y, double[] color )
  {
    return getSample( x, y, 0, 1, color );
  }

  /**
   * Computes one image sample. Passes parameters for hidden sampling.
   *
   * @param  x X coordinate of the sample.
   * @param  y Y coordinate of the sample.
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (sample signature) used for adaptive subsampling.
   * @see    #getSample(double,double,double[])
   * @see    #getBounds
   */
  public long getSample ( double x, double y, int order, int total, double[] color )
  {
    return getSample( x, y, color );
  }

}
