package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.ImageSynthesizer;

/**
 * Default implementation of {@link ImageSynthesizer} interface.
 *
 * @version 0.26 $Rev: 435 $ $Date: 2007-05-02 16:02:04 +0200 (st, 02 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/DefaultImageSynthesizer.java">DefaultImageSynthesizer.java</a>
 */
public abstract class DefaultImageSynthesizer extends Piece implements ImageSynthesizer
{
  //==========================================================================
  //  Data:

  /** Actual image width in pixels. */
  protected int imageWidth = 0;

  /** Actual image height in pixels. */
  protected int imageHeight = 0;

  /** Actual time (for {@code RealFunction}). */
  protected double actTime = 0.0;

  //==========================================================================
  //  Interface TimeDependent:

  /**
   * Retrieve the valid bounds for time variable.
   *
   * @param  pre Pre-allocated {@code double[2]} array to hold the result (can be
   *             {@code null}).
   * @return {@code result[0]} .. minimum time value, {@code result[1]} .. maximum
   *         time value.
   * @see    #getTime
   * @see    #setTime
   */
  public double[] getTimeInterval ( double[] pre )
  {
    if ( pre == null || pre.length < 2 )
      pre = new double[ 2 ];

    pre[0] = 0.0;
    pre[1] = 1.0;
    return pre;
  }

  /**
   * Returns the actual time value for the object (used in animations).
   *
   * @return Actual time (in seconds).
   * @see    #setTime
   */
  public double getTime ()
  {
    return actTime;
  }

  /**
   * Sets the actual time value for the object (used in animations).
   *
   * @param  time New scene time (in seconds).
   * @see    #getTime
   * @see    #getTimeInterval
   */
  public void setTime ( double time )
  {
    actTime = time;
  }

  //==========================================================================
  //  Interface ImageSynthesizer:

  /**
   * Sets resolution of an output raster image.
   *
   * @param  width Horizontal image size in pixels.
   * @param  height Vertical image size in pixels.
   */
  public void setBounds ( int width, int height )
  {
    imageWidth  = width;
    imageHeight = height;
  }

  /**
   * Render the given pixel of the image.
   *
   * @param  x X coordinate of the pixel.
   * @param  y Y coordinate of the pixel.
   * @param  pre Pre-allocated {@code double[3]} array to hold results (can be
   *             {@code null}).
   * @return Color of the pixel (in <b>RGB</b> color space).
   * @see    #renderRectangle
   * @see    #renderAll
   */
  public double[] renderPixel ( int x, int y, double[] pre )
  {
    if ( pre == null ||
         pre.length < 3 )
      pre = new double[ 3 ];

    pre[0] =
    pre[1] =
    pre[2] = 0.0;

    return pre;
  }

  /**
   * Render the whole image.
   * Output is written to some connected raster device (not defined in this interface).
   *
   * @see    #renderPixel
   * @see    #renderRectangle
   */
  public void renderAll ()
  {
    renderRectangle( 0, 0, imageWidth, imageHeight );
  }

}
