package cz.cuni.jagrlib;

import java.util.*;
import java.awt.Color;

import cz.cuni.jagrlib.iface.Property;

/**
 * Default implementation of {@link Property} interface.
 * Implements several handy helper functions.
 *
 * @version 0.26 $Rev: 456 $ $Date: 2007-05-19 01:03:36 +0200 (so, 19 V 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../cz/cuni/jagrlib/DefaultProperty.java">DefaultProperty.java</a>
 */
public abstract class DefaultProperty implements Property
{
  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    return null;
  }

  /**
   * Commits all previous changes to the internal object state.
   * Non-mandatory.
   */
  public void commit ()
  {
  }

  /**
   * Sets all properties from the given map.
   *
   * @param  pairs Dictionary of [key,value] pairs.
   */
  public void set ( Map<String,Object> pairs )
  {
    if ( pairs == null ) return;
    Iterator<String> it = pairs.keySet().iterator();
    while ( it.hasNext() )
    {
      String key = it.next();
      if ( key != null ) set( key, pairs.get( key ) );
    }
  }

  //==========================================================================
  //  Support:

  /**
   * Retrieves integer property from general object instance.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @return New value of the property.
   */
  public static int intProperty ( Object value, int def )
  {
    if ( value instanceof Integer )
      return ((Integer)value).intValue();
    if ( value instanceof Long )
      return (int)((Long)value).longValue();
    if ( value instanceof String )
    {
      try
      {
        return Integer.parseInt( (String)value );
      }
      catch ( NumberFormatException e )
      {
      }
    }
    return def;
  }

  /**
   * Retrieves long property from general object instance.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @return New value of the property.
   */
  public static long longProperty ( Object value, long def )
  {
    if ( value instanceof Integer )
      return ((Integer)value).intValue();
    if ( value instanceof Long )
      return ((Long)value).longValue();
    if ( value instanceof String )
    {
      try
      {
        return Long.parseLong( (String)value );
      }
      catch ( NumberFormatException e )
      {
      }
    }
    return def;
  }

  /**
   * Retrieves enum property from general object instance.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @return New value of the property.
   */
  public static <T extends Enum<T> > T enumProperty ( Object value, T def )
  {
    int iType       = intProperty( value, def.ordinal() );
    @SuppressWarnings("unchecked")
    Class<T> cl     = (Class<T>)def.getClass();
    T[] values      = cl.getEnumConstants();

    if ( values == null ||
         iType < 0 || iType >= values.length )
      return def;

    return values[ iType ];
  }

  /**
   * Retrieves string property from general object instance.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @return New value of the property.
   */
  public static String stringProperty ( Object value, String def )
  {
    if ( value == null )
      return def;

    if ( value instanceof String )
      return (String)value;

    return value.toString();
  }

  /**
   * Retrieves integer property from general object instance, does check value bounds.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @param  minVal Minimal allowed value.
   * @param  maxVal Maximal allowed value.
   * @return New value of the property.
   */
  public static int intProperty ( Object value, int def, int minVal, int maxVal )
  {
    int val = def;
    if ( value instanceof Integer )
      val = ((Integer)value).intValue();
    else
    if ( value instanceof Long )
      val = (int)((Long)value).longValue();
    else
    if ( value instanceof String )
    {
      try
      {
        val = Integer.parseInt( (String)value );
      }
      catch ( NumberFormatException e )
      {
      }
    }
    return Formula.clamp( val, minVal, maxVal );
  }

  /**
   * Retrieves long property from general object instance, does check value bounds.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @param  minVal Minimal allowed value.
   * @param  maxVal Maximal allowed value.
   * @return New value of the property.
   */
  public static long longProperty ( Object value, long def, long minVal, long maxVal )
  {
    long val = def;
    if ( value instanceof Integer )
      val = ((Integer)value).intValue();
    else
    if ( value instanceof Long )
      val = ((Long)value).longValue();
    else
    if ( value instanceof String )
    {
      try
      {
        val = Long.parseLong( (String)value );
      }
      catch ( NumberFormatException e )
      {
      }
    }
    return Formula.clamp( val, minVal, maxVal );
  }

  /**
   * Retrieves double property from general object instance.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @return New value of the property.
   */
  public static double doubleProperty ( Object value, double def )
  {
    if ( value instanceof Double )
      return ((Double)value).doubleValue();
    if ( value instanceof Float )
      return ((Float)value).floatValue();
    if ( value instanceof Integer )
      return ((Integer)value).intValue();
    if ( value instanceof Long )
      return ((Long)value).longValue();
    if ( value instanceof String )
    {
      try
      {
        return Double.parseDouble( (String)value );
      }
      catch ( NumberFormatException e )
      {
      }
    }
    return def;
  }

  /**
   * Retrieves double property from general object instance, does check value bounds.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @param  minVal Minimal allowed value.
   * @param  maxVal Maximal allowed value.
   * @return New value of the property.
   */
  public static double doubleProperty ( Object value, double def, double minVal, double maxVal )
  {
    double val = def;
    if ( value instanceof Double )
      val = ((Double)value).doubleValue();
    else
    if ( value instanceof Float )
      val = ((Float)value).floatValue();
    else
    if ( value instanceof Integer )
      val = ((Integer)value).intValue();
    else
    if ( value instanceof Long )
      val = (int)((Long)value).longValue();
    else
    if ( value instanceof String )
    {
      try
      {
        val = Double.parseDouble( (String)value );
      }
      catch ( NumberFormatException e )
      {
      }
    }
    return Formula.clamp( val, minVal, maxVal );
  }

  /**
   * Retrieves float property from general object instance.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @return New value of the property.
   */
  public static float floatProperty ( Object value, float def )
  {
    if ( value instanceof Float )
      return ((Float)value).floatValue();
    if ( value instanceof Double )
      return (float)((Double)value).doubleValue();
    if ( value instanceof Integer )
      return ((Integer)value).intValue();
    if ( value instanceof Long )
      return ((Long)value).longValue();
    if ( value instanceof String )
    {
      try
      {
        return Float.parseFloat( (String)value );
      }
      catch ( NumberFormatException e )
      {
      }
    }
    return def;
  }

  /**
   * Retrieves float property from general object instance, does check value bounds.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @param  minVal Minimal allowed value.
   * @param  maxVal Maximal allowed value.
   * @return New value of the property.
   */
  public static float floatProperty ( Object value, float def, float minVal, float maxVal )
  {
    float val = def;
    if ( value instanceof Float )
      val = ((Float)value).floatValue();
    else
    if ( value instanceof Double )
      val = (float)((Double)value).doubleValue();
    else
    if ( value instanceof Integer )
      val = ((Integer)value).intValue();
    else
    if ( value instanceof Long )
      val = (int)((Long)value).longValue();
    else
    if ( value instanceof String )
    {
      try
      {
        val = Float.parseFloat( (String)value );
      }
      catch ( NumberFormatException e )
      {
      }
    }
    return Formula.clamp( val, minVal, maxVal );
  }

  /**
   * Retrieves boolean property from general object instance.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @return New value of the property.
   */
  public static boolean booleanProperty ( Object value, boolean def )
  {
    if ( value instanceof Boolean )
      return ((Boolean)value).booleanValue();
    if ( value instanceof Integer )
      return ((Integer)value).intValue() != 0;
    if ( value instanceof Long )
      return ((Long)value).longValue() != 0L;
    if ( value instanceof String )
      return Boolean.parseBoolean( (String)value );
    return def;
  }

  /**
   * Retrieves {@link java.awt.Color} property from general object instance.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @return New value of the property.
   */
  public static Color colorProperty ( Object value, Color def )
  {
    if ( value instanceof Color )
      return (Color)value;
    if ( value instanceof String )
      return Formula.parseColor( (String)value );
    return def;
  }

  /**
   * Retrieves {@code double[]} property from general object instance.
   *
   * @param  value Object instance containing the value.
   * @param  def Default property value.
   * @return New value of the property.
   */
  public static double[] colorDoubleProperty ( Object value, double[] def )
  {
    if ( value instanceof double[] )
      return (double[])value;
    if ( value instanceof Color )
    {
      Color c = (Color)value;
      return new double[] { c.getRed()   / 255.0,
                            c.getGreen() / 255.0,
                            c.getBlue()  / 255.0
                          };
    }
    if ( value instanceof String )
      return Formula.parseColorDouble( (String)value );

    return def;
  }

  /**
   * Reads integer parameter (property) from the given instance.
   *
   * @param  object Object instance to be asked.
   * @param  name Property identifier.
   * @param  def Default property value.
   */
  public static int intProperty ( Property object, String name, int def )
  {
    Object value;
    if ( object == null ||
         (value = object.get( name )) == null ) return def;
    return intProperty( value, def );
  }

  /**
   * Reads long parameter (property) from the given instance.
   *
   * @param  object Object instance to be asked.
   * @param  name Property identifier.
   * @param  def Default property value.
   */
  public static long longProperty ( Property object, String name, long def )
  {
    Object value;
    if ( object == null ||
         (value = object.get( name )) == null ) return def;
    return longProperty( value, def );
  }

  /**
   * Reads double parameter (property) from the given instance.
   *
   * @param  object Object instance to be asked.
   * @param  name Property identifier.
   * @param  def Default property value.
   */
  public static double doubleProperty ( Property object, String name, double def )
  {
    Object value;
    if ( object == null ||
         (value = object.get( name )) == null ) return def;
    return doubleProperty( value, def );
  }

  /**
   * Reads float parameter (property) from the given instance.
   *
   * @param  object Object instance to be asked.
   * @param  name Property identifier.
   * @param  def Default property value.
   */
  public static float floatProperty ( Property object, String name, float def )
  {
    Object value;
    if ( object == null ||
         (value = object.get( name )) == null ) return def;
    return floatProperty( value, def );
  }

  /**
   * Reads boolean parameter (property) from the given instance.
   *
   * @param  object Object instance to be asked.
   * @param  name Property identifier.
   * @param  def Default property value.
   */
  public static boolean booleanProperty ( Property object, String name, boolean def )
  {
    Object value;
    if ( object == null ||
         (value = object.get( name )) == null ) return def;
    return booleanProperty( value, def );
  }

}
