package cz.cuni.jagrlib;

import java.io.*;
import java.awt.image.*;
import javax.imageio.ImageIO;

import cz.cuni.jagrlib.iface.*;

/**
 * File format default implementation for raster image data.
 *
 * @version 0.26 $Rev: 479 $ $Date: 2007-06-04 01:59:43 +0200 (po, 04 VI 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/DefaultRasterFileFormat.java">DefaultRasterFileFormat.java</a>
 */
public abstract class DefaultRasterFileFormat extends DefaultFileFormat
{
  //==========================================================================
  //  Routines to override:

  /**
   * Common load code.
   *
   * @param  stream Opened input bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  protected void commonLoad ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    InputStream is = stream.getInputStream();
    if ( is == null )
      throw new FileNotFoundException();

      // decode graphics stream:
    BufferedImage im = ImageIO.read( is );
    if ( im != null ) g.setBufferedImage( im );
  }

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  protected void commonSave ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    throw new JaGrLibException( new NotImplementedException( "Saving of this raster format is not implemented!" ) );
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /**
   * Retrieves data from the given file (stream).
   * Can be used for "file opening" in virtual-mapped mode.
   *
   * @param  fileName File-name (stream-name) to read from (can be {@code null}).
   * @param  detail Further implementation-specific information ("r" or "rw" for mapped mode).
   * @return {@code true} if the operation was successful.
   * @see    #loadFile(BitStream,Object)
   * @see    #saveFile(String,String)
   * @see    #closeFile
   */
  @Override
  public boolean loadFile ( String fileName, String detail )
  throws IOException
  {
    RasterGraphics g      = (RasterGraphics)getInterface( PL_RASTER, IFACE+"RasterGraphics" );
    BitStream      stream = (BitStream)     getInterface( PL_STREAM, IFACE+"BitStream" );
    if ( g      == null ||
         stream == null )
      throw new JaGrLibException( "loadFile: vital modules are not connected!" );

    boolean result = false;
    try
    {
      stream.open( false, fileName, detail );
      commonLoad( stream, g );
      result = true;
    }
    finally
    {
      stream.close();
    }

    return result;
  }

  /**
   * Retrieves data from the given {@link BitStream}. This metod is optional.
   * <p>The stream should be opened before - only the {@code reOpen(false,null)} call
   * can be used.
   *
   * @param  stream Input bit-stream (opened - call {@code reOpen(false,null)} first).
   * @param  support Implementation-specific data (e.g. second {@link BitStream} for
   *                 file-formats with {@link #DOUBLE_STREAM} property).
   * @return {@code true} if the operation was successful.
   * @see    #loadFile(String,String)
   * @see    #DOUBLE_STREAM
   */
  @Override
  public boolean loadFile ( BitStream stream, Object support )
  throws IOException
  {
    if ( stream == null ) return false;

    RasterGraphics g = (RasterGraphics)getInterface( PL_RASTER, IFACE+"RasterGraphics" );
    if ( g == null )
      throw new JaGrLibException( "loadFile: vital modules are not connected!" );

    commonLoad( stream, g );
    return true;
  }

  /**
   * Saves data to the given file (stream).
   *
   * @param  fileName File-name (stream-name) to write to (can be {@code null}).
   * @param  detail Further implementation-specific information.
   * @return {@code true} if the operation was successful.
   * @see    #saveFile(BitStream,Object)
   * @see    #loadFile(String,String)
   */
  @Override
  public boolean saveFile ( String fileName, String detail )
  throws IOException
  {
    RasterGraphics g      = (RasterGraphics)getInterface( PL_RASTER, IFACE+"RasterGraphics" );
    BitStream      stream = (BitStream)     getInterface( PL_STREAM, IFACE+"BitStream" );
    if ( g      == null ||
         stream == null )
      throw new JaGrLibException( "saveFile: vital modules are not connected!" );

    boolean result = false;
    try
    {
      stream.open( true, fileName, detail );
      commonSave( stream, g );
      result = true;
    }
    finally
    {
      stream.close();
    }

    return result;
  }

  /**
   * Saves data to the given {@link BitStream}. This metod is optional.
   * <p>The stream should be opened and empty - the {@code reOpen(true,null)} call
   * is not required.
   *
   * @param  stream Output bit-stream (opened and empty).
   * @param  support Implementation-specific data (e.g. second {@link BitStream} for
   *                 file-formats with {@link #DOUBLE_STREAM} property).
   * @return {@code true} if the operation was successful.
   * @see    #saveFile(String,String)
   * @see    #DOUBLE_STREAM
   */
  @Override
  public boolean saveFile ( BitStream stream, Object support )
  throws IOException
  {
    if ( stream == null ) return false;

    RasterGraphics g = (RasterGraphics)getInterface( PL_RASTER, IFACE+"RasterGraphics" );
    if ( g == null )
      throw new JaGrLibException( "saveFile: vital modules are not connected!" );

    commonSave( stream, g );
    return true;
  }

  //==========================================================================
  //  Module registration:

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "DataFileFormatToRasterGraphicsAndBitStream";

  /** Object category. */
  protected final static String CATEGORY = C_IO + '.' + C_2D + '.' + C_RASTER;

}
