package cz.cuni.jagrlib;

import java.util.*;
import java.awt.image.*;

import cz.cuni.jagrlib.iface.*;

/**
 * Default implementation of {@link RasterGraphics} interface.
 *
 * @version 0.26 $Rev: 462 $ $Date: 2007-05-24 10:40:53 +0200 (čt, 24 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/DefaultRasterGraphics.java">DefaultRasterGraphics.java</a>
 */
public abstract class DefaultRasterGraphics extends Piece implements RasterGraphics
{
  //==========================================================================
  //  Data:

  /** Default background color (for {@code getPixel()} methods). */
  protected final static int[] BACKGROUND = { 0, 0, 0, 255 };

  /** Default background color (for {@code getPixel()} methods). */
  protected final static double[] BACKGROUND_D = { 0.0, 0.0, 0.0, 1.0 };

  protected final static double INV_255 = 1.0 / 255.0;

  /** The current bitmap width. */
  protected int width;

  /** The current bitmap height. */
  protected int height;

  /** The current bitmap mode (one of {@code RasterGraphics.MODE_*} constants). */
  protected int mode;

  /** Number of bands (for {@link #MODE_BAND MODE_BAND}). */
  protected int bands;

  //--------------------------------------------------------------------------
  //  State variables:

  /** Actual binary set operation. */
  protected int binaryOp;

  /** Actual alpha-operation. */
  protected int alphaOp;

  /** Actual drawing alpha (between {@code 0.0} and {@code 1.0}). */
  protected double alpha;

  /** Actual value-transfer function. */
  protected ValueTransferFunction transferFunction;

  //==========================================================================
  //  Support:

  /** Draws the given line using current color. */
  public static final void drawLine ( RasterGraphics out, double x1d, double y1d, double x2d, double y2d )
  {
    int x1 = Formula.round( x1d );
    int y1 = Formula.round( y1d );
    int x2 = Formula.round( x2d );
    int y2 = Formula.round( y2d );

    int D, ax, ay, sx, sy;

    sx = x2 - x1;
    ax = Math.abs( sx ) << 1;
    if ( sx < 0 ) sx = -1;
    else
      if ( sx > 0 ) sx = 1;

    sy = y2 - y1;
    ay = Math.abs( sy ) << 1;
    if ( sy < 0 ) sy = -1;
    else
      if ( sy > 0 ) sy = 1;

    if ( ax > ay )                        // x coordinate is dominant
    {
      D = ay - (ax >> 1);                 // initial D
      ax = ay - ax;                       // ay = increment0; ax = increment1
      while ( x1 != x2 )
      {
        out.setPixel( x1, y1 );
        if ( D >= 0 )                     // lift up the Y coordinate
        {
          y1 += sy;
          D += ax;
        }
        else
          D += ay;
        x1 += sx;
      }
    }

    else                                  // y coordinate is dominant
    {
      D = ax - (ay >> 1);                 // initial D
      ay = ax - ay;                       // ax = increment0; ay = increment1
      while ( y1 != y2 )
      {
        out.setPixel( x1, y1 );
        if ( D >= 0 )                     // lift up the X coordinate
        {
          x1 += sx;
          D += ay;
        }
        else
          D += ax;
        y1 += sy;
      }
    }

    out.setPixel( x1, y1 );
  }

  /** Draws the given line using current color. */
  public static final void drawFrame ( RasterGraphics out, double x1d, double y1d, double x2d, double y2d )
  {
    final int x1 = Formula.round( Math.min( x1d, x2d ) );
          int y1 = Formula.round( Math.min( y1d, y2d ) );
    final int x2 = Formula.round( Math.max( x1d, x2d ) );
    final int y2 = Formula.round( Math.max( y1d, y2d ) );

    out.setHLine( x1, x2+1, y1 );
    while ( y1 < y2 )
    {
      out.setPixel( x1, y1 );
      out.setPixel( x2, y1++ );
    }
    out.setHLine( x1, x2+1, y2 );
  }

  //==========================================================================
  //  Interface RasterGraphics:

  /**
   * {@code RasterImage} (re-)initialization.
   *
   * @param  _width Width of a new bitmap.
   * @param  _height Height of a new bitmap.
   * @param  _mode Pixel value type ({@code RasterGraphics.MODE_*} constant).
   * @param  _bands Number of bands (for {@link #MODE_BAND MODE_BAND}).
   * @see    #setColormapType
   */
  public void init ( int _width, int _height, int _mode, int _bands )
  {
    width            = _width;
    height           = _height;
    mode             = _mode;
    if ( _bands == 0 )
      switch ( _mode )
      {
        case MODE_GRAY:
          _bands = 1;
          break;

        case MODE_RGB:
          _bands = 3;
          break;

        case MODE_RGBA:
          _bands = 4;
          break;

        default:
          _bands = 3;
      }
    bands            = _bands;
    alpha            = 1.0;
    binaryOp         = BinaryOperation.BIN_OP_SET;
    alphaOp          = AlphaOperation.ALPHA_B_OVER_A;
    transferFunction = null;
  }

  /**
   * {@code RasterImage} re-initialization (keeps current bitmap size, mode,
   * and colormap attributes, too).
   */
  public void init ()
  {
    binaryOp         = BinaryOperation.BIN_OP_SET;
    alphaOp          = AlphaOperation.ALPHA_B_OVER_A;
    transferFunction = null;
    alpha            = 1.0;
  }

  /**
   * Width of the bitmap.
   *
   * @return Current bitmap width in pixels.
   */
  public int getWidth ()
  {
    return width;
  }

  /**
   * Height of the bitmap.
   *
   * @return Current bitmap height in pixels.
   */
  public int getHeight ()
  {
    return height;
  }

  /**
   * Sample (color) model of the bitmap.
   *
   * @return Current bitmap mode ({@code RasterGraphics.MODE_*} constants).
   */
  public int getMode ()
  {
    return mode;
  }

  /**
   * Number of color bands of the bitmap.
   *
   * @return Number of color bands.
   */
  public int getBands ()
  {
    return bands;
  }

  /**
   * Maximal gray value.
   *
   * @return Maximal acceptable gray value.
   */
  public int getMaxGray ()
  {
    return 255;
  }

  /**
   * Maximal ordinal number of colormap entry.
   * <p>Default: colormap mode is not implemented!
   *
   * @return Maximal colormap's entry number.
   */
  public int getMaxColormap ()
  {
    return -1;
  }

  /**
   * Maximal RGB[A] value. Also used in multi-band mode.
   *
   * @return Maximal value for R,G,B,A and other bands.
   */
  public int getMaxRGB ()
  {
    return 255;
  }

  /**
   * Copies raster graphics to a {@link BufferedImage} object.
   * <p>Default: {@link BufferedImage} is not used..
   *
   * @return {@link BufferedImage} object.
   * @see    #setBufferedImage
   */
  public BufferedImage getBufferedImage ()
  {
    return null;
  }

  /**
   * Initializes raster graphics from a {@link BufferedImage} object.
   * <p>Default: {@link BufferedImage} is not used..
   *
   * @param  bi Raster data.
   * @see    #getBufferedImage
   */
  public void setBufferedImage ( BufferedImage bi )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Colormap mode (valid only for {@link #MODE_COLORMAP MODE_COLORMAP} bitmap mode).
   *
   * @return Colormap mode (one of ({@code RasterGraphics.MODE_*} constants).
   * @see    RasterGraphics
   */
  public int getColormapType ()
  {
    return MODE_RGB;
  }

  /**
   * Sets the colormap mode (valid only for {@link #MODE_COLORMAP MODE_COLORMAP} bitmap mode).
   * <p>Default: colormap mode is not implemented!
   *
   * @param  _mode Colormap mode (one of ({@code RasterGraphics.MODE_*} constants).
   * @param  _max Maximal index of colormap entry.
   * @see    RasterGraphics
   */
  public void setColormapType ( int _mode, int _max )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Returns the colormap segment (for color modes: "{@code getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND}").
   * <p>Default: colormap mode is not implemented!
   *
   * @param  i0 The first colormap index.
   * @param  length Length of colormap segment.
   * @param  _colors Returned color levels (each band: from 0.0 to 1.0).
   * @see    RasterGraphics
   */
  public void getColormap ( int i0, int length, double[][] _colors )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Returns the colormap segment (for color modes: "{@code getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND}").
   * <p>Default: colormap mode is not implemented!
   *
   * @param  i0 The first colormap index.
   * @param  length Length of colormap segment.
   * @param  _colors Returned color levels (each band: from 0 to #getMaxRGB()).
   */
  public void getColormap ( int i0, int length, int[][] _colors )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the colormap segment (for color modes: "{@code getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND}").
   * Colormap size &amp; mode has to be set before using {@link #setColormapType}.
   * <p>Default: colormap mode is not implemented!
   *
   * @param  _colors Color levels to be set (each band: from {@code 0.0} to {@code 1.0}).
   */
  public void setColormap ( double[][] _colors )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the colormap segment (for color modes: "{@code getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND}").
   * Colormap size &amp; mode has to be set before using #setColormapType.
   * <p>Default: colormap mode is not implemented!
   *
   * @param  _colors Color levels to be set (each band: from {@code 0} to {@link #getMaxRGB}).
   */
  public void setColormap ( int[][] _colors )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the binary set operation (used in all {@code set*()} methods).
   *
   * @param  _op Binary set operation ({@code BinaryOperation.BIN_OP_*} constant).
   * @return The old binary set operation.
   * @see    BinaryOperation Binary operations
   * @see    RasterGraphics
   */
  public int setOperation ( int _op )
  {
    int old  = binaryOp;
    binaryOp = BinaryOperation.stationaryOp( _op );
    return old;
  }

  /**
   * Sets the alpha-channel operation (used in all {@code set*()} methods).
   *
   * @param  _op Alpha-channel operation ({@code AlphaOperation.ALPHA_*} constant).
   * @return The old alpha-channel operation.
   * @see    AlphaOperation Alpha-channel operations
   */
  public int setAlphaOperation ( int _op )
  {
    int old = alphaOp;
    alphaOp = _op;
    return old;
  }

  /**
   * Sets the value-transfer function (used in all {@code set*()} methods).
   *
   * @param  _transfer New value-transfer function or {@code null}.
   * @return The old Value-transfer function or {@code null}.
   * @see    cz.cuni.jagrlib.iface.ValueTransferFunction Value-transfor function prototype
   * @see    RasterGraphics
   */
  public ValueTransferFunction setTransferFunction ( ValueTransferFunction _transfer )
  {
    return transferFunction;
  }

  /**
   * Sets the current drawing color (for {@link #MODE_GRAY MODE_GRAY} or
   * {@link #MODE_COLORMAP MODE_COLORMAP}).
   *
   * @param  _color Current drawing color (0 to {@link #getMaxGray} or
   *         {@link #getMaxColormap}).
   * @see    RasterGraphics
   */
  public void setColor ( int _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the current drawing color (for {@link #MODE_GRAY MODE_GRAY}).
   *
   * @param  _color Current drawing color ({@code 0.0} to {@code 1.0}).
   * @see    RasterGraphics
   */
  public void setColor ( double _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the current drawing color (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  _color Current drawing color (channel values between
   *                {@code 0} and {@link #getMaxRGB}).
   * @see    RasterGraphics
   */
  public void setColor ( int[] _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the current drawing color (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  _color Current drawing color (channel values between
   *                {@code 0.0} and {@code 1.0}).
   * @see    RasterGraphics
   */
  public void setColor ( double[] _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the current drawing alpha (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_GRAY MODE_GRAY}).
   *
   * @param  _alpha Current drawing alpha (values between
   *                {@code 0.0} and {@code 1.0}).
   */
  public void setAlpha ( double _alpha )
  {
    alpha = _alpha;
  }

  /**
   * Sets the given pixel's value.
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @see    #setColor(int)
   * @see    #setColor(double)
   * @see    #setColor(int[])
   * @see    #setColor(double[])
   * @see    RasterGraphics
   */
  public void setPixel ( int x, int y )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Returns the pixel's value (for {@link #MODE_GRAY MODE_GRAY} or
   * {@link #MODE_COLORMAP MODE_COLORMAP}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value ({@code 0} for off-canvas access).
   * @see    RasterGraphics
   */
  public int getPixel ( int x, int y )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Returns the pixel's value (for {@link #MODE_GRAY MODE_GRAY}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value ({@code 0.0} for off-canvas access).
   * @see    RasterGraphics
   */
  public double getPixelDouble ( int x, int y )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Returns the pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value.
   * @see    RasterGraphics
   */
  public void getPixel ( int x, int y, int[] _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Returns the pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value.
   * @see    RasterGraphics
   */
  public void getPixel ( int x, int y, double[] _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Returns the pixel's RGB[A] values (for all graphic modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value (if the array's length is at least 4, alpha will be returned).
   */
  public void getRGB ( int x, int y, int[] _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Returns the pixel's RGB[A] values (for all graphic modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value (if the array's length is at least 4, alpha will be returned).
   */
  public void getRGB ( int x, int y, double[] _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Returns the pixel's gray value (for all modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value ({@code 0} for off-canvas access).
   */
  public int getGray ( int x, int y )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Returns the pixel's gray value (for all modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value ({@code 0.0} for off-canvas access).
   */
  public double getGrayDouble ( int x, int y )
  {
    return getGray( x, y ) * INV_255;
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_GRAY MODE_GRAY} or
   * {@link #MODE_COLORMAP MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0} and {@link #getMaxGray} or
   *                {@link #getMaxColormap}).
   */
  public void putPixel ( int x, int y, int _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_GRAY MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0.0} and {@code 1.0}).
   */
  public void putPixel ( int x, int y, double _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0} and {@link #getMaxRGB}).
   */
  public void putPixel ( int x, int y, int[] _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0.0} and {@code 1.0}).
   */
  public void putPixel ( int x, int y, double[] _color )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y].
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @see    #setColor(int)
   * @see    #setColor(double)
   * @see    #setColor(int[])
   * @see    #setColor(double[])
   * @see    RasterGraphics
   */
  public void setHLine ( int x1, int x2, int y )
  {
    throw new JaGrLibException( new NotImplementedException() );
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_GRAY MODE_GRAY} or {@link #MODE_COLORMAP MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color ({@code 0} to {@link #getMaxGray} or
   *         {@link #getMaxColormap}).
   * @see    RasterGraphics
   */
  public void setHLine ( int x1, int x2, int y, int _color )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 )     x1 = 0;
    if ( x2 > width ) x2 = width;

    while ( x1 < x2 )
      putPixel( x1++, y, _color );
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_GRAY MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color ({@code 0.0} to {@code 1.0}).
   * @see    RasterGraphics
   */
  public void setHLine ( int x1, int x2, int y, double _color )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 )     x1 = 0;
    if ( x2 > width ) x2 = width;

    while ( x1 < x2 )
      putPixel( x1++, y, _color );
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA} or
   * and {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color (channel values between {@code 0} and {@link #getMaxRGB}).
   * @see    RasterGraphics
   */
  public void setHLine ( int x1, int x2, int y, int[] _color )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 )     x1 = 0;
    if ( x2 > width ) x2 = width;

    while ( x1 < x2 )
      putPixel( x1++, y, _color );
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA} or
   * {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color (channel values between {@code 0.0} and {@code 1.0}).
   * @see    RasterGraphics
   */
  public void setHLine ( int x1, int x2, int y, double[] _color )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 )     x1 = 0;
    if ( x2 > width ) x2 = width;

    while ( x1 < x2 )
      putPixel( x1++, y, _color );
  }

  /**
   * Sets the horizontal line's values (for {@link #MODE_GRAY MODE_GRAY} or
   * {@link #MODE_COLORMAP MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between {@code 0} and {@link #getMaxGray} or
   *                 {@link #getMaxColormap}).
   * @see    RasterGraphics
   */
  public void putHLine ( int x1, int x2, int y, int[] _colors )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    int offset = 0;                       // pixel index in _colors[]
    if ( x1 < 0 )
    {
      offset = -x1;
      x1 = 0;
    }
    if ( x2 > width ) x2 = width;

    while ( x1 < x2 )
      putPixel( x1++, y, _colors[offset++] );
  }

  /**
   * Sets the horizontal line's values (for {@link #MODE_GRAY MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between {@code 0.0} and {@code 1.0}).
   * @see    RasterGraphics
   */
  public void putHLine ( int x1, int x2, int y, double[] _colors )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    int offset = 0;                       // pixel index in _colors[]
    if ( x1 < 0 )
    {
      offset = -x1;
      x1 = 0;
    }
    if ( x2 > width ) x2 = width;

    while ( x1 < x2 )
      putPixel( x1++, y, _colors[offset++] );
  }

  /**
   * Sets the horizontal line's values (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between {@code 0} and {@link #getMaxGray}).
   * @see    RasterGraphics
   */
  public void putHLine ( int x1, int x2, int y, int[][] _colors )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    int offset = 0;                       // pixel index in _colors[]
    if ( x1 < 0 )
    {
      offset = -x1;
      x1 = 0;
    }
    if ( x2 > width ) x2 = width;

      // write the data:
    while ( x1 < x2 )
      putPixel( x1++, y, _colors[offset++] );
  }

  /**
   * Sets the horizontal line's values (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between {@code 0.0} and {@code 1.0}).
   * @see    RasterGraphics
   */
  public void putHLine ( int x1, int x2, int y, double[][] _colors )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    int offset = 0;                       // pixel index in _colors[]
    if ( x1 < 0 )
    {
      offset = -x1;
      x1 = 0;
    }
    if ( x2 > width ) x2 = width;

      // write the data:
    while ( x1 < x2 )
      putPixel( x1++, y, _colors[offset++] );
  }

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_GRAY MODE_GRAY} or {@link #MODE_COLORMAP MODE_COLORMAP}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors ({@code 0} to {@link #getMaxGray} or
   *                 {@link #getMaxColormap}, {@code 0} for off-canvas access).
   * @see    RasterGraphics
   */
  public void getHLine ( int x1, int x2, int y, int[] _colors )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 || y < 0 || y >= height )
    {
      Arrays.fill( _colors, 0, x2 - x1, 0 );
      return;
    }

    int offset = 0;
    while ( x1 < x2 )
      _colors[offset++] = getPixel( x1, y );
  }

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_GRAY MODE_GRAY}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors ({@code 0.0} to {@code 1.0},
   *                 {@code 0.0} for off-canvas access).
   * @see    RasterGraphics
   */
  public void getHLine ( int x1, int x2, int y, double[] _colors )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 || y < 0 || y >= height )
    {
      Arrays.fill( _colors, 0, x2 - x1, 0.0 );
      return;
    }

    int offset = 0;
    while ( x1 < x2 )
      _colors[offset++] = getPixelDouble( x1, y );
  }

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors (channel values between {@code 0} and
   *                 {@link #getMaxRGB}, {@code 0} for off-canvas access).
   * @see    RasterGraphics
   */
  public void getHLine ( int x1, int x2, int y, int[][] _colors )
  {
    if ( bands < 2 )
      throw new JaGrLibException( "getHLine (...,int[][]) needs RGB, RGBA or BAND image",
                                   new Throwable( "getHLine (...,int[][])" ) );

    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 || y < 0 || y >= height )
    {
      for ( int i = 0; i < x2 - x1; i++ )
        Arrays.fill( _colors[i], 0 );
      return;
    }

    int offset = 0;
    while ( x1 < x2 )
    {
      if ( x1 > width || x1 < 0 )         // off canvas
        Arrays.fill( _colors[offset], 0 );
      else
        getPixel( x1++, y, _colors[offset] );
      offset++;
    }
  }

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors (channel values between {@code 0.0} and
   *                 {@code 1.0}, {@code 0.0} for off-canvas access).
   * @see    RasterGraphics
   */
  public void getHLine ( int x1, int x2, int y, double[][] _colors )
  {
    if ( bands < 2 )
      throw new JaGrLibException( "getHLine (...,double[][]) needs RGB, RGBA or BAND image",
                                  new Throwable( "getHLine (...,double[][])" ) );

    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 || y < 0 || y >= height )
    {
      for ( int i = 0; i < x2 - x1; i++ )
        Arrays.fill( _colors[i], 0.0 );
      return;
    }

    int offset = 0;
    while ( x1 < x2 )
    {
      if ( x1 > width || x1 < 0 )         // off canvas
        Arrays.fill( _colors[offset], 0.0 );
      else
        getPixel( x1++, y, _colors[offset] );
      offset++;
    }
  }

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1].
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @see    RasterGraphics
   */
  public void setRectangle ( int x1, int y1, int x2, int y2 )
  {
    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1  = x2;
      x2  = tmp;
    }
    if ( x1 >= width || x2 <= 0 ) return;

    if ( y1 > y2 )
    {
      tmp = y1;
      y1  = y2;
      y2  = tmp;
    }
    if ( y1 >= height || y2 <= 0 ) return;

    while ( y1 < y2 )
      setHLine( x1, x2, y1++ );
  }

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1] (for
   * {@link #MODE_GRAY MODE_GRAY} or {@link #MODE_COLORMAP MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _color Drawing color ({@code 0} to {@link #getMaxGray} or
   *         {@link #getMaxColormap}).
   * @see    RasterGraphics
   */
  public void setRectangle ( int x1, int y1, int x2, int y2, int _color )
  {
    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1  = x2;
      x2  = tmp;
    }
    if ( x1 >= width || x2 <= 0 ) return;

    if ( y1 > y2 )
    {
      tmp = y1;
      y1  = y2;
      y2  = tmp;
    }
    if ( y1 >= height || y2 <= 0 ) return;

    while ( y1 < y2 )
      setHLine( x1, x2, y1++, _color );
  }

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1] (for
   * {@link #MODE_GRAY MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _color Drawing color ({@code 0.0} to {@code 1.0}).
   * @see    RasterGraphics
   */
  public void setRectangle ( int x1, int y1, int x2, int y2, double _color )
  {
    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1  = x2;
      x2  = tmp;
    }
    if ( x1 >= width || x2 <= 0 ) return;

    if ( y1 > y2 )
    {
      tmp = y1;
      y1  = y2;
      y2  = tmp;
    }
    if ( y1 >= height || y2 <= 0 ) return;

    while ( y1 < y2 )
      setHLine( x1, x2, y1++, _color );
  }

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1] (for
   * {@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _color Drawing color ({@code 0} to {@link #getMaxRGB}).
   * @see    RasterGraphics
   */
  public void setRectangle ( int x1, int y1, int x2, int y2, int[] _color )
  {
    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1  = x2;
      x2  = tmp;
    }
    if ( x1 >= width || x2 <= 0 ) return;

    if ( y1 > y2 )
    {
      tmp = y1;
      y1  = y2;
      y2  = tmp;
    }
    if ( y1 >= height || y2 <= 0 ) return;

    while ( y1 < y2 )
      setHLine( x1, x2, y1++, _color );
  }

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1] (for
   * {@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _color Drawing color ({@code 0.0} to {@code 1.0}).
   * @see    RasterGraphics
   */
  public void setRectangle ( int x1, int y1, int x2, int y2, double[] _color )
  {
    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1  = x2;
      x2  = tmp;
    }
    if ( x1 >= width || x2 <= 0 ) return;

    if ( y1 > y2 )
    {
      tmp = y1;
      y1  = y2;
      y2  = tmp;
    }
    if ( y1 >= height || y2 <= 0 ) return;

    while ( y1 < y2 )
      setHLine( x1, x2, y1++, _color );
  }

  //==========================================================================
  //  Interface Property:

  protected int setWidth     = 0;

  protected int setHeight    = 0;

  protected int setColorMode = MODE_RGB;

  protected int setBands     = 3;

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // WIDTH:
    if ( key.compareTo(WIDTH) == 0 )
    {
      setWidth = intProperty( value, setWidth );
      return;
    }
      // HEIGHT:
    if ( key.compareTo(HEIGHT) == 0 )
    {
      setHeight = intProperty( value, setHeight );
      return;
    }
      // COLORMODE:
    if ( key.compareTo(COLORMODE) == 0 )
    {
      setColorMode = intProperty( value, setColorMode );
      return;
    }
      // BANDS:
    if ( key.compareTo(BANDS) == 0 )
    {
      setBands = intProperty( value, setBands );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(WIDTH) == 0 )
      return width;
    if ( key.compareTo(HEIGHT) == 0 )
      return height;
    if ( key.compareTo(COLORMODE) == 0 )
      return mode;
    if ( key.compareTo(BANDS) == 0 )
      return bands;

    return null;
  }

  /**
   * Commits all previous changes to the internal object state.
   * Non-mandatory.
   */
  @Override
  public void commit ()
  {
    init( setWidth, setHeight, setColorMode, setBands );
  }

  //==========================================================================
  //  Module registration:

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "RasterGraphics";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DATA + '.' + C_RASTER;

}
