package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.RealFunction;

/**
 * Default implementation of {@link RealFunction} interface.
 *
 * @version 0.25 $Rev: 390 $ $Date: 2007-01-09 17:05:56 +0100 (út, 09 I 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/DefaultRealFunction.java">DefaultRealFunction.java</a>
 */
public abstract class DefaultRealFunction extends Piece implements RealFunction
{
  /** Constant used for numerical derivatives. */
  public final static double EPSILON = 1.0e-6;

  /**
   * Compute the mapping: {@code double -> double}.
   */
  public double f ( double x )
  {
    return f( x, 0.0 );
  }

  /**
   * Compute the mapping: {@code int -> double}.
   */
  public double f ( int x )
  {
    return f( (double)x );
  }

  /**
   * Compute the mapping: {@code double[2] -> double}.
   */
  public double f ( double x, double y )
  {
    return f( x, y, 0.0 );
  }

  /**
   * Compute the mapping: {@code int[2] -> double}.
   */
  public double f ( int x, int y )
  {
    return f( (double)x, (double)y );
  }

  /**
   * Compute the mapping: {@code double[3] -> double}.
   */
  public double f ( double x, double y, double z )
  {
    double[] arg = { x, y, z };
    return f( arg );
  }

  /**
   * Compute the mapping: {@code int[3] -> double}.
   */
  public double f ( int x, int y, int z )
  {
    return f( (double)x, (double)y, (double)z );
  }

  /**
   * Compute the mapping: {@code double[] -> double}.
   */
  public double f ( double[] arg )
  {
    return 0.0;
  }

  /**
   * Compute the mapping: {@code int[] -> double}.
   */
  public double f ( int[] arg )
  {
    return 0.0;
  }

  /**
   * Compute derivative: <b>df(x)/dx</b>. Non-mandatory.
   *
   * @param  x Function argument.
   * @return Function derivative or {@code 0.0} if not implemented.
   */
  public double df ( double x )
  {
    return( (f( x + EPSILON ) - f( x - EPSILON )) / (EPSILON + EPSILON) );
  }

  /**
   * Compute partial derivatives: <b>df(x,y)/dx</b> and <b>df(x,y)/dy</b>. Non-mandatory.
   *
   * @param  x X coordinate (1st argument).
   * @param  y Y coordinate (2nd argument).
   * @param  d Array to hold results ({@code d[0]} = df(x,y)/dx,
   *           {@code d[1]} = df(x,y)/dy).
   * @return {@code true} if this service is implemented.
   */
  public boolean df ( double x, double y, double d[] )
  {
    if ( d == null || d.length < 2 )
      return false;

    d[0] = (f( x + EPSILON, y ) - f( x - EPSILON, y )) / (EPSILON + EPSILON);
    d[1] = (f( x, y + EPSILON ) - f( x, y - EPSILON )) / (EPSILON + EPSILON);
    return true;
  }

  /**
   * Compute partial derivatives: <b>df(x,y,z)/dx</b>, <b>df(x,y,z)/dy</b> and <b>df(x,y,z)/dz</b>.
   * Non-mandatory.
   *
   * @param  x X coordinate (1st argument).
   * @param  y Y coordinate (2nd argument).
   * @param  z Z coordinate (3rd argument).
   * @param  d Array to hold results ({@code d[0]} = df(x,y,z)/dx,
   *           {@code d[1]} = df(x,y,z)/dy, {@code d[2]} = df(x,y,z)/dz).
   * @return {@code true} if this service is implemented.
   */
  public boolean df ( double x, double y, double z, double d[] )
  {
    if ( d == null || d.length < 3 )
      return false;

    d[0] = (f( x + EPSILON, y, z ) - f( x - EPSILON, y, z )) / (EPSILON + EPSILON);
    d[1] = (f( x, y + EPSILON, z ) - f( x, y - EPSILON, z )) / (EPSILON + EPSILON);
    d[2] = (f( x, y, z + EPSILON ) - f( x, y, z - EPSILON )) / (EPSILON + EPSILON);
    return true;
  }

  /**
   * Compute partial derivatives: <b>df(x1,..)/dx1</b>, ...
   * Non-mandatory.
   *
   * @param  arg Coordinate array.
   * @param  d Array to hold results ({@code d[0]} = df(x1,...)/dx1, ...).
   * @return {@code true} if this service is implemented.
   */
  public boolean df ( double[] arg, double d[] )
  {
    return false;
  }

  /**
   * Returns domain of the mapping.
   *
   * @param  dom Domain array ({@code dom[0]} = Xmin, {@code dom[1]} = Xmax, etc.).
   * @return {@code null} if this service is not implemented.
   */
  public double[] getDomain ( double[] dom )
  {
    return null;
  }

  /**
   * Returns domain of the mapping.
   *
   * @param  dom Domain array ({@code dom[0]} = Xmin, {@code dom[1]} = Xmax, etc.).
   * @return {@code null} if this service is not implemented.
   */
  public int[] getDomain ( int[] dom )
  {
    return null;
  }

}
