package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.TrMatrix;

/**
 * Default implementation of {@link Render3D}.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/DefaultRender3D.java">DefaultRender3D.java</a>
 */
public abstract class DefaultRender3D extends Piece implements Render3D, Trigger
{
  //==========================================================================
  //  Data:

  /** Actual rendering style. */
  protected int renderStyle = RENDER_STYLE_LINE;

  /** Actual model-view matrix ({@code null} .. no transform). */
  protected TrMatrix modelView = null;

  /** Dirty-flag for the model-view matrix. */
  protected boolean mvChanged = true;

  /** Actual projection matrix ({@code null} .. no transform). */
  protected TrMatrix projection = null;

  /** Dirty-flag for the projection matrix. */
  protected boolean prChanged = true;

  /** Viewport origin: x-coordinate. */
  protected int vpX0 = 0;

  /** Viewport origin: y-coordinate. */
  protected int vpY0 = 0;

  /** Viewport width in pixels. */
  protected int vpWidth = 800;

  /** Viewport height in pixels. */
  protected int vpHeight = 600;

  /** Dirty-flag for the viewport data. */
  protected boolean vpChanged = true;

  /**
   * Compound transform matrix (model-view + projection + viewport,
   * {@code modelView} .. no transform).
   */
  protected TrMatrix compound = null;

  /** Actual depth-test comparison function. */
  protected int depthComparison = DEPTH_GEQUAL;

  //==========================================================================
  //  Data:

  /** Sets type of depth-test comparison. */
  public void setDepthComparison ( int comp )
  {
    depthComparison = comp;
  }

  /**
   * Sets actual rendering style.
   *
   * @param style New rendering style.
   * @return Old rendering style.
   * @see #RENDER_STYLE_LINE
   * @see #RENDER_STYLE_FLAT
   */
  public int setRenderStyle ( int style )
  {
    int oldStyle = renderStyle;
    renderStyle  = Formula.clamp( style, RENDER_STYLE_LINE, RENDER_STYLE_PHONG );
    return oldStyle;
  }

  /**
   * Sets actual viewport size.
   * The same semantics as {@code glSetViewport()}.
   *
   * @param  x0 Viewport origin (x-coord).
   * @param  y0 Viewport origin (y-coord).
   * @param  width Viewport width in pixels.
   * @param  height Viewport height in pixels.
   */
  public void setViewport ( int x0, int y0, int width, int height )
  {
    vpChanged = (x0     != vpX0 ||
                 y0     != vpY0 ||
                 width  != vpWidth ||
                 height != vpHeight);
    if ( vpChanged )
    {
      vpX0     = x0;
      vpY0     = y0;
      vpWidth  = width;
      vpHeight = height;
    }
  }

  /**
   * Sets actual model-view matrix.
   *
   * @param mv New model-view matrix.
   */
  public void setModelView ( TrMatrix mv )
  {
    modelView = mv;
    mvChanged = true;
  }

  /**
   * Sets actual projection matrix.
   *
   * @param proj New projection matrix.
   */
  public void setProjection ( TrMatrix proj )
  {
    projection = proj;
    prChanged  = true;
  }

  /**
   * Sets actual projection parameters (non-mandatory).
   *
   * @param view Viewing vector (from camera to the scene).
   * @param up Up vector (should always point up on the screen).
   * @param lookAt Point in the middle of the view.
   * @param zoom Zoom factor (camera - lookAt distance, element size on the screen, etc.).
   * @return {@code true} if the method is implemented.
   */
  public boolean setProjection ( double[] view, double[] up, double[] lookAt, double zoom )
  {
    return false;
  }

  /** Asserts validity of the {@link #compound} matrix. */
  public void checkMatrix ()
  {
    if ( mvChanged ||
         prChanged ||
         vpChanged )
    {
      if ( compound == null )
        compound = new TrMatrix( 4, 4 );
      else
        compound.identity( 4, 4 );

      if ( modelView != null )
        compound.append( modelView );

      if ( projection != null )
        compound.append( projection );

      compound.viewport( vpX0, vpY0, vpWidth, vpHeight );

      mvChanged =
      prChanged =
      vpChanged = false;
    }
  }

  /**
   * 3D rendering using actual parameters.
   *
   * @see #setModelView(TrMatrix)
   * @see #setProjection(TrMatrix)
   * @see #setRenderStyle(int)
   * @see #render(TrMatrix,TrMatrix,int)
   */
  public void render ()
  {
    render( modelView, projection, renderStyle );
  }

  /**
   * 3D rendering using the given projection matrix.
   *
   * @param mv Local model-view matrix.
   * @param proj Local projection matrix.
   * @param style Local rendering style.
   * @see #render()
   */
  public void render ( TrMatrix mv, TrMatrix proj, int style )
  {
    render();
  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Redraws the image immediately.
   *
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    render();
    return true;
  }

}
