package cz.cuni.jagrlib;

import java.util.*;

import cz.cuni.jagrlib.iface.*;

/**
 * Default implementation of {@link ValueTransferFunction} interface.
 *
 * @version 0.25 $Rev: 294 $ $Date: 2006-07-06 00:50:34 +0200 (čt, 06 VII 2006) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/DefaultValueTransferFunction.java">DefaultValueTransferFunction.java</a>
 */
public abstract class DefaultValueTransferFunction extends Piece implements ValueTransferFunction, PersistentParamsBinary
{
  //==========================================================================
  //  Interface ValueTransferFunction:

  /**
   * Scalar value-transfer function.
   *
   * @param  value Input value (range from <tt>0</tt> to <tt>max</tt>).
   * @param  max Maximal input/output value.
   * @return Output value (range from <tt>0</tt> to <tt>max</tt>).
   */
  public int transfer ( int value, int max )
  {
    return( (int)( transfer( value / (double)max ) + 0.5 ) );
  }

  /**
   * Scalar inverse value-transfer function.
   *
   * @param  value Input value (range from <tt>0</tt> to <tt>max</tt>).
   * @param  max Maximal input/output value.
   * @return Output value (range from <tt>0</tt> to <tt>max</tt>).
   */
  public int transferInv ( int value, int max )
  {
    return( (int)( transferInv( value / (double)max ) + 0.5 ) );
  }

  /**
   * Scalar value-transfer function.
   *
   * @param  value Input value (range from <tt>0.0</tt> to <tt>1.0</tt>).
   * @return Output value (range from <tt>0.0</tt> to <tt>1.0</tt>).
   */
  public double transfer ( double value )
  {
    return value;
  }

  /**
   * Scalar inverse value-transfer function.
   *
   * @param  value Input value (range from <tt>0.0</tt> to <tt>1.0</tt>).
   * @return Output value (range from <tt>0.0</tt> to <tt>1.0</tt>).
   */
  public double transferInv ( double value )
  {
    return value;
  }

  /**
   * Gr[A] or RGB[A] value-transfer function.
   * Can be used for higher dimensions too.
   *
   * @param  input Input value (scalar range from <tt>0</tt> to <tt>max</tt>).
   * @param  output Output value (scalar range from <tt>0</tt> to <tt>max</tt>).
   *                Can be the same as <tt>input</tt> (in-place operation).
   * @param  max Maximal input/output value.
   */
  public void transfer ( int[] input, int[] output, int max )
  {
    if ( input  == null ||
         output == null ) return;

    if ( input != output )
      System.arraycopy( input, 0, output, 0, Math.min( input.length, output.length ) );
  }

  /**
   * Gr[A] or RGB[A] inverse value-transfer function.
   * Can be used for higher dimensions too.
   *
   * @param  input Input value (scalar range from <tt>0</tt> to <tt>max</tt>).
   * @param  output Output value (scalar range from <tt>0</tt> to <tt>max</tt>).
   *                Can be the same as <tt>input</tt> (in-place operation).
   * @param  max Maximal input/output value.
   */
  public void transferInv ( int[] input, int[] output, int max )
  {
    if ( input  == null ||
         output == null ) return;

    if ( input != output )
      System.arraycopy( input, 0, output, 0, Math.min( input.length, output.length ) );
  }

  /**
   * Gr[A] or RGB[A] value-transfer function.
   * Can be used for higher dimensions too.
   *
   * @param  input Input value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   * @param  output Output value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   *                Can be the same as <tt>input</tt> (in-place operation).
   */
  public void transfer ( double[] input, double[] output )
  {
    if ( input  == null ||
         output == null ) return;

    if ( input != output )
      System.arraycopy( input, 0, output, 0, Math.min( input.length, output.length ) );
  }

  /**
   * Gr[A] or RGB[A] inverse value-transfer function.
   * Can be used for higher dimensions too.
   *
   * @param  input Input value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   * @param  output Output value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   *                Can be the same as <tt>input</tt> (in-place operation).
   */
  public void transferInv ( double[] input, double[] output )
  {
    if ( input  == null ||
         output == null ) return;

    if ( input != output )
      System.arraycopy( input, 0, output, 0, Math.min( input.length, output.length ) );
  }

  /**
   * Gr[A] or RGB[A] -&gt; index (classification) function.
   * Prototype implementation: RGB to gray conversion.
   *
   * @param  input Input value (scalar range from <tt>0</tt> to <tt>max</tt>).
   * @param  max Maximal input value.
   * @return Output index value (color index).
   */
  public int transfer ( int[] input, int max )
  {
    if ( input == null ||
         input.length < 3 ) return 0;

      // assuming input data in format: RGB[A]
    return Formula.rgbToGray( input[0], input[1], input[2] );
  }

  /**
   * Index -&gt; Gr[A] or RGB[A] (reconstruction) function.
   * Prototype implementation: gray to RGB conversion.
   *
   * @param  input Index value (color index).
   * @param  output Output value (scalar range from <tt>0</tt> to <tt>max</tt>).
   * @param  max Maximal output value.
   */
  public void transferInv ( int input, int[] output, int max )
  {
    if ( output == null ) return;

    Arrays.fill( output, Math.min( input, max ) );
  }

  /**
   * Gr[A] or RGB[A] -&gt; index (classification) function.
   * Prototype implementation: RGB to gray conversion.
   *
   * @param  input Input value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   * @return Output index value (color index).
   */
  public int transfer ( double[] input )
  {
    if ( input == null ||
         input.length < 3 ) return 0;

      // assuming input data in format: RGB[A]
    return( (int)( Formula.rgbToGray( input[0], input[1], input[2] ) * 255.0 + 0.5 ) );
  }

  /**
   * Index -&gt; Gr[A] or RGB[A] (reconstruction) function.
   * Prototype implementation: gray to RGB conversion.
   *
   * @param  input Index value (color index).
   * @param  output Output value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   */
  public void transferInv ( int input, double[] output )
  {
    if ( output == null ) return;

    Arrays.fill( output, Math.min( input, 255.0 ) );
  }

  //==========================================================================
  //  Interface PersistentParamsBinary:

  /**
   * Pack instance parameters into binary format.
   *
   * @return Data array to be stored into binary stream. Can be <tt>null</tt>.
   *         The array can be shared by all method calls.
   * @see    #loadData
   */
  public byte[] storeData ()
  {
    return null;
  }

  /**
   * Loads instance parameters from (previously stored) binary data.
   *
   * @param  data Binary data to be interpreted. Can be <tt>null</tt>.
   * @return <tt>true</tt> if successful.
   * @see    #storeData
   */
  public boolean loadData ( byte[] data )
  {
    return( data == null );
  }

}
