package cz.cuni.jagrlib;

import java.awt.Color;
import java.text.NumberFormat;
import java.text.ParsePosition;
import java.util.*;
import java.io.*;
import org.nfunk.jep.JEP;

/**
 * Mathematical support (excluding geometric stuff - see Geometry class).
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../cz/cuni/jagrlib/Formula.java">Formula.java</a>
 */
public class Formula
{
  /** Line separator string. */
  public static final String LINE_SEPARATOR = System.getProperty( "line.separator" );

  /** Line separator as byte array. */
  public static final byte[] LINE_SEPARATOR_BYTES = LINE_SEPARATOR.getBytes();

  //==========================================================================
  //  Arithmetics, algebra:

  /**
   * Computes double-precision number powered to integer exponent.
   *
   * @param  a Base.
   * @param  e Exponent (all integer values are handled correctly).
   * @return {@code a} to the power of {@code e}.
   */
  public final static double power ( double a, int e )
  {
    if ( e == 0 ) return 1.0;
    if ( e < 0 )
    {
      e = -e;
      a = 1.0 / a;
    }
    double acc = (e & 1) != 0 ? a : 1.0;
    while ( (e >>= 1) != 0 )
    {
      a *= a;
      if ( (e & 1) != 0 ) acc *= a;
    }
    return acc;
  }

  /**
   * Computes diagonal of the rectangle block.
   */
  public final static double diagonal ( double x, double y, double z )
  {
    return Math.sqrt( x * x + y * y + z * z );
  }

  /**
   * Computes floating-point remainder which is periodic and positive
   * unlike the {@link Math#IEEEremainder}.
   */
  public final static double remainder ( double a, double b )
  {
    if ( b == 0.0 ) return 0.0;

    final double rem = Math.IEEEremainder( a, b );
    return( rem >= 0.0 ? rem : rem + b );
  }

  /** Computes greates common divisor of two unsigned integers. */
  public final static int gcd ( int u, int v )
  {
    if ( u == 0 ) return v;
    if ( v == 0 ) return u;

    int k = 0;                            // binary exponent of the result

    while ( (u & 1) == 0 &&
            (v & 1) == 0 )                // while both u and v are even
    {
      u >>>= 1;
      v >>>= 1;
      k++;
    }
      // At this point either u or v (or both) is odd
    do
    {
      if ( (u & 1) == 0 )
        u >>>= 1;
      else
      if ( (v & 1) == 0 )
        v >>>= 1;
      else
      if ( u >= v )
        u = (u - v) >>> 1;
      else
        v = (v - u) >>> 1;
    } while ( u > 0 );

    return v << k;
  }

  /** Computes number of bits to represent the given number of possibilities (log_2{n} rounded up). */
  public final static int bits ( long n )
  {
    if ( n <= 1L ) return 0;

    int  b     = 1;
    long bBits = 2L;
    while ( bBits < n )
    {
      bBits += bBits;
      b++;
    }

    return b;
  }

  /** Clamps the source value to the given range. */
  public final static int clamp ( int source, int min, int max )
  {
    return( (source <= min) ? min : ((source >= max) ? max : source) );
  }

  /** Clamps the source value to the given range. */
  public final static long clamp ( long source, long min, long max )
  {
    return( (source <= min) ? min : ((source >= max) ? max : source) );
  }

  /** Clamps the source value to the given range. */
  public final static float clamp ( float source, float min, float max )
  {
    return( (source <= min) ? min : ((source >= max) ? max : source) );
  }

  /** Clamps the source value to the given range. */
  public final static double clamp ( double source, double min, double max )
  {
    return( (source <= min) ? min : ((source >= max) ? max : source) );
  }

  /** Return index of array item with maximum absolute value. */
  public final static int argMax ( double[] array )
  {
    int besti = 0;
    int len;
    double best, act;

    if ( array != null && (len = array.length) > 1 )
    {
      best = Math.abs( array[0] );
      for ( int i = 1; i < len; i++ )
        if ( (act = Math.abs( array[i] )) > best )
        {
          besti = i;
          best  = act;
        }
    }
    return besti;
  }

  /**
   * Rounds double number to the nearest integer.
   * <p>For {@code double -> long} rounding, use {@link Math#round(double)},
   * <br>for {@code float -> int} rounding, use {@link Math#round(float)}.
   */
  public final static int round ( double a )
  {
    return (int)Math.floor( a + 0.5 );
  }

  /**
   * Returns Variable-length code for the given integer number.
   * <p>{@code 0} .. 0, {@code 1} .. 1, {@code -1} .. 2,
   * {@code 2} .. 3, {@code -2} .. 4, etc.
   */
  public final static long vlc ( int a )
  {
    long sign = (a <= 0) ? 0L : -1L;
    a         = Math.abs( a );
    return( (long)a + a + sign );
  }

  /**
   * Returns original signed number from its Variable-length code.
   * <p>{@code 0} .. 0, {@code 1} .. 1, {@code 2} .. -1,
   * {@code 3} .. 2, {@code 4} .. -2, etc.
   */
  public final static int vlcInv ( long code )
  {
    int sign = (code & 1) > 0L ? 1 : -1;
    int abs  = (int)( (code + 1L) >> 1 );
    return( sign * abs );
  }

  /**
   * Returns integer percent value.
   */
  public final static int percent ( long part, long total )
  {
    if ( total <= 0L ) return 0;
    return (int)( (part * 100L + (total >> 1)) / total );
  }

  /** Returns the given value rounded up to the given granularity. */
  public final static int granularity ( int value, int granul )
  {
    value = (value + granul - 1) / granul;
    return( value * granul );
  }

  /**
   * Function for C[++] compatibility:
   * composes floating-point number from the mantissa and binary exponent.
   *
   * @see    #frexp(double,int[])
   */
  public final static double ldexp ( double mant, int exp )
  {
    if ( exp == 0 ) return mant;

    return( mant * power( 2.0, exp ) );
  }

  /**
   * Function for C[++] compatibility:
   * composes floating-point number from the mantissa and binary exponent.
   *
   * @see    #frexp(float,int[])
   */
  public final static float ldexp ( float mant, int exp )
  {
    if ( exp == 0 ) return mant;

    return( (float)( mant * power( 2.0, exp ) ) );
  }

  /**
   * Function for C[++] compatibility:
   * computes normalized mantissa ({@code 0.5} to {@code 1.0}) and binary exponent
   * from the given floating-point number.
   *
   * @see    #ldexp(double,int)
   */
  public final static double frexp ( double x, int[] exp )
  {
    int     e   = 0;
    boolean neg = (x < 0.0);
            x   = Math.abs( x );

    if ( x < 0.5 )
    {
      do
        e--;
      while ( (x += x) < 0.5 );

      if ( exp != null ) exp[0] = e;

      return( neg ? -x : x );
    }

    do
      e++;
    while ( (x *= 0.5) >= 1.0 );

    if ( exp != null ) exp[0] = e;

    return( neg ? -x : x );
  }

  /**
   * Function for C[++] compatibility:
   * computes normalized mantissa ({@code 0.5} to {@code 1.0}) and binary exponent
   * from the given floating-point number.
   *
   * @see    #ldexp(float,int)
   */
  public final static float frexp ( float x, int[] exp )
  {
    int     e   = 0;
    boolean neg = (x < 0.0);
            x   = Math.abs( x );

    if ( x < 0.5f )
    {
      do
        e--;
      while ( (x += x) < 0.5f );

      if ( exp != null ) exp[0] = e;

      return( neg ? -x : x );
    }

    do
      e++;
    while ( (x *= 0.5f) >= 1.0f );

    if ( exp != null ) exp[0] = e;

    return( neg ? -x : x );
  }

  /**
   * Converts a 16-bit unsigned integer into two bytes (used in serialization,
   * {@link cz.cuni.jagrlib.iface.PersistentParamsBinary PersistentParamsBinary}, ..).
   *
   * @param  array Target of the conversion.
   * @param  index Start index.
   * @param  s Number to convert.
   * @return Index pointing after used space.
   * @see    #loadUnsigned16
   */
  public final static int storeUnsigned16 ( byte[] array, int index, int s )
  {
    array[index++] = (byte)(  s        & 0xff );
    array[index++] = (byte)( (s >>> 8) & 0xff );
    return index;
  }

  /**
   * Reads a 16-bit unsigned integer from byte-array (used in serialization,
   * {@link cz.cuni.jagrlib.iface.PersistentParamsBinary PersistentParamsBinary}, ..).
   *
   * @param  array Source array.
   * @param  index Start index.
   * @return Result read from the array.
   * @see    #storeUnsigned16
   */
  public final static int loadUnsigned16 ( byte[] array, int index )
  {
    int s =  ((int)array[index++]) & 0xff;
    return( s + ((((int)array[index]) & 0xff) << 8) );
  }

  /**
   * Converts a float number into four bytes (used in serialization,
   * {@link cz.cuni.jagrlib.iface.PersistentParamsBinary PersistentParamsBinary}, ..).
   *
   * @param  array Target of the conversion.
   * @param  index Start index.
   * @param  f Floating number to convert.
   * @return Index pointing after used space.
   * @see    #loadFloat
   */
  public final static int storeFloat ( byte[] array, int index, float f )
  {
    int bits       = Float.floatToIntBits( f );
    array[index++] = (byte)(  bits         & 0xff );
    array[index++] = (byte)( (bits >>>= 8) & 0xff );
    array[index++] = (byte)( (bits >>>= 8) & 0xff );
    array[index++] = (byte)( (bits >>>= 8) & 0xff );
    return index;
  }

  /**
   * Reads a float-number from byte-array (used in serialization,
   * {@link cz.cuni.jagrlib.iface.PersistentParamsBinary PersistentParamsBinary}, ..).
   *
   * @param  array Source array.
   * @param  index Start index.
   * @return Result read from the array.
   * @see    #storeFloat
   */
  public final static float loadFloat ( byte[] array, int index )
  {
    final int bits =  (((int)array[index  ]) & 0xff)        +
                     ((((int)array[index+1]) & 0xff) <<  8) +
                     ((((int)array[index+2]) & 0xff) << 16) +
                     ((((int)array[index+3]) & 0xff) << 24);
    return Float.intBitsToFloat( bits );
  }

  /** Converts integer number into binary million (Mega) units. */
  public final static double toMega ( long n )
  {
    return( n / (1024.0 * 1024.0) );
  }

  /** freq[] array can contain zeroes.. */
  public static long permutationsWithRepetition ( int[] freq )
  {
    int len;
    if ( freq == null ||
         (len = freq.length) == 0 )
      return 0L;

      // 1st phase: not taking repetitions into account
    int i;
    long r = 1L;
    long m = 1L;
    for ( i = 0; i < len; i++ )
      if ( freq[i] > 0 )
        r *= m++;

    if ( m == 1L ) return 0L;

      // 2nd phase: repetitions
    long j;
    for ( i = 0; i < len; i++ )
      for ( j = 2L; j <= freq[i]; j++ )
        r = (r * m++) / j;

    return r;
  }

  /** Swaps the given number (containing multiple of 8 bits). */
  public static long swapBytes ( long a, int bits )
  {
    long result = 0L;
    int b;
    for ( b = 0; b < bits; b += 8 )
    {
      result = (result << 8) | (a & 0xff);
      a   >>>= 8;
    }
    return result;
  }

  /**
   * Initializes expression to be repeatedly evaluated.
   *
   * @param  expr String representation of the expression.
   * @return Expression to be evaluated afterwards.
   * @see    JEP#addVariable(String,double)
   * @see    JEP#getValue()
   */
  public static JEP initExpression ( String expr )
  {
    JEP e = new JEP();
    e.addStandardFunctions();
    e.addStandardConstants();
    e.setImplicitMul( true );
    e.setAllowUndeclared( true );
    e.parseExpression( expr );
    return e;
  }

  //==========================================================================
  //  Strings:

  /** Parse sequence of (comma, semicolon)-separated quoted strings. */
  public static String[] parseStringList ( String str )
  {
    int ptr = 0;
    int len = str.length();
    while ( ptr < len && Character.isWhitespace( str.charAt( ptr ) ) ) ptr++;
    if ( ptr >= len )
      return new String[ 0 ];

    final char quote = str.charAt( ptr ); // the first non-space character defines 'quotation'
    List< String > list = new ArrayList< String >();
    StringBuffer sb = new StringBuffer();
    char ch;

    do
    {                                     // parse one array item
      sb.setLength( 0 );
      while ( ++ptr < len &&
              (ch = str.charAt( ptr )) != quote )
      {
        if ( ch == '\\' )
        {
          if ( ++ptr >= len )
            break;
          ch = str.charAt( ptr );
        }
        sb.append( ch );
      }
      list.add( sb.toString() );

        // skip over (arbitrary) separator
      while ( ptr < len && Character.isWhitespace( str.charAt( ptr ) ) ) ptr++;
      if ( ptr < len && str.charAt( ptr ) != quote ) ptr++;
      while ( ptr < len && str.charAt( ptr ) != quote ) ptr++;
    }
    while ( ptr < len );

    return (String[])list.toArray();
  }

  /**
   * Checks format string for numeric type class (integer / floating-point).
   * Finds {@link java.lang.String#format(String,Object...) String.format()}-like sequence and
   * returns whether it needs integer numeric value.
   */
  public static boolean isIntFormat ( String fmt )
  {
    int pos = 0;
    int len = fmt.length();
      // 1. looking for the first single '%'
    do
    {
      pos = fmt.indexOf( '%', pos );
      if ( pos < 0 ) return false;
    }
    while ( pos < len - 1 &&
            fmt.charAt( pos + 1 ) == '%' );

      // 2. looking for the first following letter
    do
      pos++;
    while ( pos < len &&
            !Character.isLetter( fmt.charAt( pos ) ) );
    if ( pos >= len ) return false;

    switch ( Character.toLowerCase( fmt.charAt( pos ) ) )
    {
      case 'h':                           // hex
      case 'd':                           // decimal
      case 'o':                           // octal
      case 'x':                           // hex
        return true;
    }

    return false;
  }

  /** Converts number of seconds to readable (HH:MM:SS) format. */
  public static String timeHMS ( double time )
  {
    StringBuffer sb = new StringBuffer();
    boolean leading = false;
      // hours:
    if ( time >= 3600.0 )
    {
      final int h = (int)Math.floor( time ) / 3600;
      sb.append( h )
        .append( ':' );
      time -= h * 3600.0;
      leading = true;
    }
      // minutes:
    if ( leading ||
         time >= 60.0 )
    {
      final int m = (int)Math.floor( time ) / 60;
      if ( m < 10 && leading )
        sb.append( '0' );
      sb.append( m )
        .append( ':' );
      time -= m * 60.0;
      leading = true;
    }
      // seconds:
    final int s = (int)Math.floor( time );
    if ( s < 10 && leading )
      sb.append( '0' );
    sb.append( s );

    return sb.toString();
  }

  //==========================================================================
  //  Colors:

  public final static NumberFormat fmtInt = NumberFormat.getInstance( Locale.US );
  {
    fmtInt.setParseIntegerOnly( true );
  }

  public final static NumberFormat fmtDouble = NumberFormat.getInstance( Locale.US );

  /** Red importance weight. */
  public final static int RED_WEIGHT   = 4897;

  /** Green importance weight. */
  public final static int GREEN_WEIGHT = 9611;

  /** Blue importance weight. */
  public final static int BLUE_WEIGHT  = 1876;

  /** Importance weight shift. */
  public final static int WEIGHT_SHIFT = 14;

  /** Importance weight total. */
  public final static int WEIGHT_TOTAL = 16384;

  /** Importance weight total inverse. */
  public final static double WEIGHT_INVERSE = 1.0 / 16384;

  /** RGB -> gray value convertor. Keeps original data amplitude.. */
  public final static int rgbToGray ( int r, int g, int b )
  {
    return( ( r * RED_WEIGHT +
              g * GREEN_WEIGHT +
              b * BLUE_WEIGHT ) >>> WEIGHT_SHIFT );
  }

  /** RGB -> gray value convertor. Keeps original data amplitude.. */
  public final static double rgbToGray ( double r, double g, double b )
  {
    return( ( r * RED_WEIGHT +
              g * GREEN_WEIGHT +
              b * BLUE_WEIGHT ) * WEIGHT_INVERSE );
  }

  /**
   * Returns String representation of the given {@link Color}.
   *
   * @see    #parseColor
   */
  public final static String colorToString ( Color c )
  {
    return "[" + c.getRed() + ';' + c.getGreen() + ';' + c.getBlue() + ']';
  }

  /**
   * Parse textual color representation.
   * Input format: "[Red;Green;Blue]"
   *
   * @param  s Input string in "[Red;Green;Blue]" format.
   * @return Color object.
   */
  public final static Color parseColor ( String s )
  {
    if ( s == null ||
         s.length() < 2 )
      return new Color( 0, 0, 0 );

    if ( s.indexOf( '.' ) >= 0 )
    {
      double[] c = parseColorDouble( s );
      return new Color( (float)c[0], (float)c[1], (float)c[2] );
    }

    ParsePosition pos = new ParsePosition( 1 );

    int red   = fmtInt.parse( s, pos ).intValue();
    pos.setIndex( pos.getIndex() + 1 );
    int green = fmtInt.parse( s, pos ).intValue();
    pos.setIndex( pos.getIndex() + 1 );
    int blue  = fmtInt.parse( s, pos ).intValue();

    return new Color( red, green, blue );
  }

  /**
   * Parse textual color representation (components are from range {@code [ 0.0, 1.0 ]}).
   * Input format: "[Red;Green;Blue]"
   *
   * @param  s Input string in "[Red;Green;Blue]" format.
   * @return Color object.
   */
  public final static double[] parseColorDouble ( String s )
  {
    ParsePosition pos = new ParsePosition( 1 );

    double red   = fmtDouble.parse( s, pos ).doubleValue();
    pos.setIndex( pos.getIndex() + 1 );
    double green = fmtDouble.parse( s, pos ).doubleValue();
    pos.setIndex( pos.getIndex() + 1 );
    double blue  = fmtDouble.parse( s, pos ).doubleValue();

    if ( red   <= 1.0 &&
         green <= 1.0 &&
         blue  <= 1.0 )
      return new double[] { red, green, blue };

    return new double[] { red   / 255.0,
                          green / 255.0,
                          blue  / 255.0 };
  }

  /**
   * Computes color from the visualization color-ramp.
   *
   * @param  x Ramp position from {@code 0.0} to {@code 1.0}.
   * @param  color Pre-allocated array to hold result.
   * @return RGB[A] color.
   */
  public final static int[] colorRamp ( double x, int[] color )
  {
    if ( color == null || color.length < 3 )
      color = new int[ 3 ];
    else
      if ( color.length > 3 )
        color[3] = 255;

    x = clamp( x, 0.0, 1.0 );

      // I:   [  0,  0,  0] to [  0,  0,255]
    if ( x <= 0.2 )
    {
      color[0] =
      color[1] = 0;
      color[2] = (int)( x * 1275.0 + 0.5 );
    }
    else
      // II:  [  0,  0,255] to [  0,255,128]
    if ( (x -= 0.2) <= 0.2 )
    {
      color[0] = 0;
      color[1] = (int)( x * 1275.0 + 0.5 );
      color[2] = (int)( 255.5 - x * 635.0 );
    }
    else
      // III: [  0,255,128] to [128,255,  0]
    if ( (x -= 0.2) <= 0.2 )
    {
      color[0] = (int)( x * 640.0 + 0.5 );
      color[1] = 255;
      color[2] = (int)( 128.5 - x * 640.0 );
    }
    else
      // IV:  [128,255,  0] to [255,255,  0]
    if ( (x -= 0.2) <= 0.2 )
    {
      color[0] = (int)( 128.5 + x * 635.0 );
      color[1] = 255;
      color[2] = 0;
    }
    else
      // V:   [255,255,  0] to [255,  0,  0]
    {
      color[0] = 255;
      color[1] = (int)( 255.5 - (x - 0.2) * 1275.0 );
      color[2] = 0;
    }
    return color;
  }

  /**
   * Computes {@code double[3]} color from the visualization color-ramp.
   *
   * @param  x Ramp position from {@code 0.0} to {@code 1.0}.
   * @param  color Pre-allocated array to hold result.
   * @return RGB[A] color.
   */
  public final static double[] colorRamp ( double x, double[] color )
  {
    if ( color == null || color.length < 3 )
      color = new double[ 3 ];
    else
      if ( color.length > 3 )
        color[3] = 1.0;

    x = clamp( x, 0.0, 1.0 );

      // I:   [ 0.0, 0.0, 0.0 ] to [ 0.0, 0.0, 1.0 ]
    if ( x <= 0.2 )
    {
      color[0] =
      color[1] = 0.0;
      color[2] = 5.0 * x;
    }
    else
      // II:  [ 0.0, 0.0, 1.0 ] to [ 0.0, 1.0, 0.5 ]
    if ( (x -= 0.2) <= 0.2 )
    {
      color[0] = 0.0;
      color[1] = 5.0 * x;
      color[2] = 1.0 - 2.5 * x;
    }
    else
      // III: [ 0.0, 1.0, 0.5 ] to [ 0.5, 1.0, 0.0 ]
    if ( (x -= 0.2) <= 0.2 )
    {
      color[0] = 2.5 * x;
      color[1] = 1.0;
      color[2] = 0.5 - 2.5 * x;
    }
    else
      // IV:  [ 0.5, 1.0, 0.0 ] to [ 1.0, 1.0, 0.0 ]
    if ( (x -= 0.2) <= 0.2 )
    {
      color[0] = 0.5 + 2.5 * x;
      color[1] = 1.0;
      color[2] = 0.0;
    }
    else
      // V:   [ 1.0, 1.0, 0.0 ] to [ 1.0, 0.0, 0.0 ]
    {
      color[0] = 1.0;
      color[1] = 1.0 - 5.0 * (x - 0.2);
      color[2] = 0.0;
    }
    return color;
  }

  /**
   * Computes {@code float[3]} color from the visualization color-ramp.
   *
   * @param  x Ramp position from {@code 0.0f} to {@code 1.0f}.
   * @param  color Pre-allocated array to hold result.
   * @return RGB[A] color.
   */
  public final static float[] colorRamp ( double x, float[] color )
  {
    if ( color == null || color.length < 3 )
      color = new float[ 3 ];
    else
      if ( color.length > 3 )
        color[3] = 1.0f;

    float xf = (float)clamp( x, 0.0, 1.0 );

      // I:   [ 0.0, 0.0, 0.0 ] to [ 0.0, 0.0, 1.0 ]
    if ( xf <= 0.2 )
    {
      color[0] =
      color[1] = 0.0f;
      color[2] = 5.0f * xf;
    }
    else
      // II:  [ 0.0, 0.0, 1.0 ] to [ 0.0, 1.0, 0.5 ]
    if ( (xf -= 0.2) <= 0.2 )
    {
      color[0] = 0.0f;
      color[1] = 5.0f * xf;
      color[2] = 1.0f - 2.5f * xf;
    }
    else
      // III: [ 0.0, 1.0, 0.5 ] to [ 0.5, 1.0, 0.0 ]
    if ( (xf -= 0.2) <= 0.2 )
    {
      color[0] = 2.5f * xf;
      color[1] = 1.0f;
      color[2] = 0.5f - 2.5f * xf;
    }
    else
      // IV:  [ 0.5, 1.0, 0.0 ] to [ 1.0, 1.0, 0.0 ]
    if ( (xf -= 0.2) <= 0.2 )
    {
      color[0] = 0.5f + 2.5f * xf;
      color[1] = 1.0f;
      color[2] = 0.0f;
    }
    else
      // V:   [ 1.0, 1.0, 0.0 ] to [ 1.0, 0.0, 0.0 ]
    {
      color[0] = 1.0f;
      color[1] = 1.0f - 5.0f * (xf - 0.2f);
      color[2] = 0.0f;
    }
    return color;
  }

  /** Conversion from Radiance's RGBe 32-bit format into HDR floating-point RGB format. */
  public final static void RGBeToRGB ( byte[] rgbe, int startRgbe, double[] rgb, int startRgb )
  {
    if ( rgbe == null || startRgbe + 4 > rgbe.length ||
         rgb  == null || startRgb  + 3 > rgb.length ) return;

    if ( rgbe[startRgbe+3] == 0 )
      rgb[startRgb  ] =
      rgb[startRgb+1] =
      rgb[startRgb+2] = 0.0;
    else
    {
      double f = power( 2.0, ((int)rgbe[startRgbe+3] & 0xff) - 136 );
      rgb[startRgb  ] = (((int)rgbe[startRgbe  ] & 0xff) + 0.5) * f;
      rgb[startRgb+1] = (((int)rgbe[startRgbe+1] & 0xff) + 0.5) * f;
      rgb[startRgb+2] = (((int)rgbe[startRgbe+2] & 0xff) + 0.5) * f;
    }
  }

  protected static int[] exp = new int[ 1 ];

  /** Conversion from HDR floating-point RGB format to Radiance's RGBe 32-bit format. */
  public final static void RGBToRGBe ( byte[] rgbe, int startRgbe, double R, double G, double B )
  {
    if ( rgbe == null || startRgbe + 4 > rgbe.length ) return;

    double m = Math.max( R, G );
    if ( B > m ) m = B;

    if ( m < 1.e-32 )
      rgbe[startRgbe  ] =
      rgbe[startRgbe+1] =
      rgbe[startRgbe+2] =
      rgbe[startRgbe+3] = 0;
    else
    {
      m = frexp( m, exp ) * 255.9999 / m;

      rgbe[startRgbe  ] = (byte)( R * m );
      rgbe[startRgbe+1] = (byte)( G * m );
      rgbe[startRgbe+2] = (byte)( B * m );
      rgbe[startRgbe+3] = (byte)( exp[0] + 128 );
    }
  }

  /**
   * RGB to HSV (HSB) conversion.
   * RGB values are arbitrary (nonnegative),
   * {@code 0.0 <= H < 360.0}, {@code 0.0 <= S <= 1.0 }, V has the same
   * magnitude as RGB.
   */
  public final static double[] rgbToHsv ( double r, double g, double b, double[] hsv )
  {
    if ( hsv        == null ||
         hsv.length < 3 )
      hsv = new double[ 3 ];

    double max    = Math.max( r, Math.max( g, b ) );
    double min    = Math.min( r, Math.min( g, b ) );
    double maxMin = max - min;

    hsv[2] = max;
    hsv[1] = Geometry.isZero( max ) ? 0.0 : maxMin / max;

    if ( Geometry.isZero( hsv[1] ) )
      hsv[0] = 0.0;
    else
    {
      double a  = 1.0 / maxMin;
      double r1 = (max - r) * a;
      double g1 = (max - g) * a;
      double b1 = (max - b) * a;

      if ( max == r )
        hsv[0] = (min == g) ? 5.0 + b1 : 1.0 - g1;
      else
        if ( max == g )
          hsv[0] = (min == b) ? 1.0 + r1 : 3.0 - b1;
        else
          hsv[0] = (min == r) ? 3.0 + g1 : 5.0 - r1;

      hsv[0] *= 60.0;
    }

    return hsv;
  }

  /**
   * HSV (HSB) to RGB conversion.
   * RGB values are arbitrary (nonnegative),
   * {@code 0.0 <= H < 360.0}, {@code 0.0 <= S <= 1.0 }, V has the same
   * magnitude as RGB result.
   */
  public final static double[] hsvToRgb ( double h, double s, double v, double[] rgb )
  {
    if ( rgb        == null ||
         rgb.length < 3 )
      rgb = new double[ 3 ];

    h = remainder( h, 360.0 );
    h    /= 60.0;
    int i = (int)Math.floor( h );
    h    -= i;

    double p1 = v * (1.0 - s);
    double p2 = v * (1.0 - s * h);
    double p3 = v * (1.0 - s * (1.0 - h));

    switch ( i )
    {
      case 0:
        rgb[0] = v;
        rgb[1] = p3;
        rgb[2] = p1;
        break;

      case 1:
        rgb[0] = p2;
        rgb[1] = v;
        rgb[2] = p1;
        break;

      case 2:
        rgb[0] = p1;
        rgb[1] = v;
        rgb[2] = p3;
        break;

      case 3:
        rgb[0] = p1;
        rgb[1] = p2;
        rgb[2] = v;
        break;

      case 4:
        rgb[0] = p3;
        rgb[1] = p1;
        rgb[2] = v;
        break;

      case 5:
        rgb[0] = v;
        rgb[1] = p1;
        rgb[2] = p2;
        break;

      default:
        rgb[0] =
        rgb[1] =
        rgb[2] = 0.0;
    }

    return rgb;
  }

  //==========================================================================
  //  Prime numbers:

  /** Prime numbers - approximately geometric sequence. */
  protected final static int[] primes =
  {
             5,
            11,
            23,
            53,
           113,
           251,
           509,
          1021,
          2039,
          4051,
          8111,
         16223,
         32467,
         64937,
        129887,
        259781,
        519577,
       1039169,
       2078339,
       4156709,
       8313433,
      16626941,
      33253889,
      66507787,
     133015579,
     266031169,
     532062341,
    1064124683,
    2128249369
  };

  protected final static int PRIMES_MAX = primes.length - 1;

  /**
   * Returns prime number from the predefined sequence.
   *
   * @param  from Lower bound (inclusive) for the prime.
   * @return Prime number greater or equal to {@code from}.
   */
  public final static int findPrime ( int from )
  {
    int minI = 0;
    int minP = primes[minI];
    int maxI = PRIMES_MAX;
    int maxP = primes[maxI];

    if ( from <= minP ) return minP;
    if ( from >= maxP ) return maxP;

      // loop invariant: minP < from <= maxP:
    while ( minI + 1 < maxI )
    {
      int midI = (minI + maxI) >> 1;
      int midP = primes[midI];
      if ( midP < from )
      {
        minI = midI;
        minP = midP;
      }
      else
      {
        maxI = midI;
        maxP = midP;
      }
    }

    return maxP;
  }

  /** First prime numbers for hashing steps. */
  public final static int[] smallPrimes =
  {
       1,
       2,
       3,
       5,
       7,
      11,
      13,
      17,
      19,
      23,
      29,
      31,
      37,
      41,
      43,
      47
  };

  /**
   * Number of prime numbers in {@link #smallPrimes} array.
   * Has to be power of two!
   */
  public final static int FIRST_PRIMES = 16;

  /** Returns small prime number based on the given key (hash). */
  public final static int getSmallPrime ( int key )
  {
    return( smallPrimes[ key & (FIRST_PRIMES - 1) ] );
  }

  //==========================================================================
  //  System:

  /** Clone the given array of floats. */
  public final static float[] cloneArray ( float[] arr )
  {
    if ( arr == null ) return null;
    int len = arr.length;
    float[] res = new float[ len ];
    if ( len > 0 )
      System.arraycopy( arr, 0, res, 0, len );
    return res;
  }

  /** Clone the given array of doubles. */
  public final static double[] cloneArray ( double[] arr )
  {
    if ( arr == null ) return null;
    int len = arr.length;
    double[] res = new double[ len ];
    if ( len > 0 )
      System.arraycopy( arr, 0, res, 0, len );
    return res;
  }

  /** Clone the given array of integers. */
  public final static int[] cloneArray ( int[] arr )
  {
    if ( arr == null ) return null;
    int len = arr.length;
    int[] res = new int[ len ];
    if ( len > 0 )
      System.arraycopy( arr, 0, res, 0, len );
    return res;
  }

  /** Converts {@code int[]} vector to readable form. */
  public final static String vectorToString ( int[] v )
  {
    if ( v == null ) return "null";

    StringBuffer sb = new StringBuffer( "[" );
    for ( int i : v )
      sb.append( String.format( " %d", i ) );
    sb.append( " ]" );
    return sb.toString();
  }

  /** Converts {@code double[]} vector to readable form. */
  public final static String vectorToString ( double[] v )
  {
    if ( v == null ) return "null";

    StringBuffer sb = new StringBuffer( "[" );
    for ( double d : v )
      sb.append( String.format( Locale.US, " %.3f", d ) );
    sb.append( " ]" );
    return sb.toString();
  }

  /** Hash-function instance (accelerator). */
  public static HashFunction hashFunction = null;

  /** Returns hash-number computed from class name of the given instance. */
  public final static int hashClassName ( Object object )
  {
    if ( object == null ) return 0;

    if ( hashFunction == null )
      hashFunction = new HashFunction();
    String className = object.getClass().getName();
    return( (int)hashFunction.hash( className ) );
  }

  /** Returns path relative to current directory. */
  public static String relativePath ( String path )
  {
    String userPath    = System.getProperty( "user.dir" );
    final File userDir = new File( userPath );
    final File inPath  = new File( path );
    try
    {
      userPath = userDir.getCanonicalPath();
      path     = inPath.getCanonicalPath();
    }
    catch ( IOException e )
    {}

    if ( path.indexOf( userPath ) == 0 )
    {
      int pref = userPath.length();
      if ( pref > 0 &&
           pref < path.length() &&
           path.charAt( pref ) == File.separatorChar ) pref++;
      path = path.substring( pref );
      if ( File.separatorChar != '/' )
        path = path.replace( File.separatorChar, '/' );
    }

    return path;
  }

  //==========================================================================
  //  [XML] persistence:

  /** Converts arbitrary object instance to String representation (for persistance). */
  public static final String stringValue ( Object value )
  {
    if ( value == null ) return "";

    if ( value instanceof Color )
      return colorToString( (Color)value );

    return value.toString();
  }
}
