package cz.cuni.jagrlib;

import java.util.*;

/**
 * Simple geometric routines.
 *
 * @version 0.26 $Rev: 819 $ $Date: 2008-11-09 11:39:57 +0100 (ne, 09 XI 2008) $ $Author: Tomas $
 * @since   0.12
 * @see     <a href="../../../../cz/cuni/jagrlib/Geometry.java">Geometry.java</a>
 */
public class Geometry
{
  /** The smallest absolute value that can be inverted. */
  public final static double EPSILON  = Double.longBitsToDouble( 8L );

  /** The smallest absolute value that can be inverted. */
  public final static float EPSILON_F = Float.intBitsToFloat( 8 );

  /**
   * Upper bound for 3D rounding errors (two 3D points closer than {@code SCENE_EPSILON}
   * are assumed to be equal).
   */
  public final static double SCENE_EPSILON = 1.e-4;

  /** Compares the given number with zero (using {@link #EPSILON} tolerance). */
  public final static boolean isZero ( double x )
  {
    return x > -EPSILON && x < EPSILON;
  }

  /** Compares the given number with zero (using {@link #EPSILON_F} tolerance). */
  public final static boolean isZero ( float x )
  {
    return x > -EPSILON_F && x < EPSILON_F;
  }

  /**
   * Computes Euclidean distance of two 3D points.
   *
   * @param  a The 1st point.
   * @param  b The 2nd point.
   * @return Euclidean distance of the two points (or {@link Double#POSITIVE_INFINITY} in case of error).
   */
  public final static double pointPoint3D ( double[] a, double[] b )
  {
    int alen, blen;
    if ( a == null || (alen = a.length) < 3 ||
         b == null || (blen = b.length) < 3 ) return Double.POSITIVE_INFINITY;

    double ax, ay, az, w;
    if ( blen == 3 || (w = b[3]) == 1.0 || w > -EPSILON && w < EPSILON )
    {
      ax = b[0];
      ay = b[1];
      az = b[2];
    }
    else
    {
      w  = 1.0 / w;
      ax = w * b[0];
      ay = w * b[1];
      az = w * b[2];
    }

    if ( alen == 3 || (w = a[3]) == 1.0 || w > -EPSILON && w < EPSILON )
    {
      ax -= a[0];
      ay -= a[1];
      az -= a[2];
    }
    else
    {
      w   = 1.0 / w;
      ax -= w * a[0];
      ay -= w * a[1];
      az -= w * a[2];
    }

    return Math.sqrt( ax * ax + ay * ay + az * az );
  }

  /**
   * Computes 2D distance between point {@code p} and line segment {@code ab}.
   *
   * @param  p Point coordinates.
   * @param  a Segment's starting point.
   * @param  b Segment's end point.
   * @return Distance or {@link Double#POSITIVE_INFINITY} in case of error.
   */
  public final static double pointSegment2D ( double[] p, double[] a, double[] b )
  {
    int plen, alen, blen;
    if ( p == null || (plen = p.length) < 2 ||
         a == null || (alen = a.length) < 2 ||
         b == null || (blen = b.length) < 2 ) return Double.POSITIVE_INFINITY;

    double px, py, w;
    if ( plen == 2 || (w = p[2]) == 1.0 || w > -EPSILON && w < EPSILON )
    {
      px = p[0];
      py = p[1];
    }
    else
    {
      w  = 1.0 / w;
      px = w * p[0];
      py = w * p[1];
    }

    double ax, ay;
    if ( alen == 2 || (w = a[2]) == 1.0 || w > -EPSILON && w < EPSILON )
    {
      ax = a[0];
      ay = a[1];
    }
    else
    {
      w  = 1.0 / w;
      ax = w * a[0];
      ay = w * a[1];
    }

    double bx, by;
    if ( blen == 2 || (w = b[2]) == 1.0 || w > -EPSILON && w < EPSILON )
    {
      bx = b[0];
      by = b[1];
    }
    else
    {
      w  = 1.0 / w;
      bx = w * b[0];
      by = w * b[1];
    }

      // [px,py] versus [ax,ay] - [bx,by]:
    return pointSegment2D( px, py, ax, ay, bx, by );
  }

  /**
   * Computes 2D distance between point {@code [ px, py ]} and line segment
   * {@code [ ax, ay ] [ bx, by ]}.
   */
  public final static double pointSegment2D ( double px, double py, double ax, double ay, double bx, double by )
  {
    double dx = bx - ax;
    double dy = by - ay;

      // [px,py] versus [ax,ay] + t * [dx,dy]:
    double dd  = dx * dx + dy * dy;
    if ( dd > EPSILON )
    {
      double dap = dx * (px - ax) + dy * (py - ay);

      boolean posDD;
      if ( (posDD = (dd >= 0.0)) == (dap >= 0.0) ) // t > 0.0
        if ( posDD == (dap >= dd) )                // t >= 1.0
        {
          ax = bx;
          ay = by;
        }
        else                                       // 0.0 < t < 1.0
        {
          double t = dap / dd;
          ax      += t * dx;
          ay      += t * dy;
        }
    }

      // compute distance [px,py] from [ax,ay]:
    return Math.hypot( ax - px,  ay - py );
  }

  /**
   * Computes 3D distance between point {@code p} and the line segment {@code ab}.
   *
   * @param  p Point coordinates.
   * @param  a Segment's starting point.
   * @param  b Segment's end point.
   * @return Distance or {@link Double#POSITIVE_INFINITY} in case of error.
   */
  public final static double pointSegment3D ( double[] p, double[] a, double[] b )
  {
    int plen, alen, blen;
    if ( p == null || (plen = p.length) < 3 ||
         a == null || (alen = a.length) < 3 ||
         b == null || (blen = b.length) < 3 ) return Double.POSITIVE_INFINITY;

    double px, py, pz, w;
    if ( plen == 3 || (w = p[3]) == 1.0 || w > -EPSILON && w < EPSILON )
    {
      px = p[0];
      py = p[1];
      pz = p[2];
    }
    else
    {
      w  = 1.0 / w;
      px = w * p[0];
      py = w * p[1];
      pz = w * p[2];
    }

    double ax, ay, az;
    if ( alen == 3 || (w = a[3]) == 1.0 || w > -EPSILON && w < EPSILON )
    {
      ax = a[0];
      ay = a[1];
      az = a[2];
    }
    else
    {
      w  = 1.0 / w;
      ax = w * a[0];
      ay = w * a[1];
      az = w * a[2];
    }

    double dx, dy, dz;
    if ( blen == 3 || (w = b[3]) == 1.0 || w > -EPSILON && w < EPSILON )
    {
      dx = b[0] - ax;
      dy = b[1] - ay;
      dz = b[2] - az;
    }
    else
    {
      w  = 1.0 / w;
      dx = w * b[0] - ax;
      dy = w * b[1] - ay;
      dz = w * b[2] - az;
    }

      // [px,py,pz] versus [ax,ay,az] + t * [dx,dy,dz]:
    double dd  = dx * dx + dy * dy + dz * dz;
    if ( dd > EPSILON )
    {
      double dap = dx * (px - ax) + dy * (py - ay) + dz * (pz - az);

      boolean posDD;
      if ( (posDD = (dd >= 0.0)) == (dap >= 0.0) ) // t > 0.0
        if ( posDD == (dap >= dd) )                // t >= 1.0
        {
          ax += dx;
          ay += dy;
          az += dz;
        }
        else                                       // 0.0 < t < 1.0
        {
          double t = dap / dd;
          ax      += t * dx;
          ay      += t * dy;
          az      += t * dz;
        }
    }

      // compute distance [px,py,pz] from [ax,ay,az]:
    ax -= px;
    ay -= py;
    az -= pz;
    return Math.sqrt( ax * ax + ay * ay + az * az );
  }

  /**
   * Computes 2D distance between point {@code p} and the rectangular box {@code ul, ul+size}.
   *
   * @param  p Point coordinates.
   * @param  ul Upper left rectangle corner.
   * @param  size Box size.
   * @return Distance or {@code -1.0} if the point lies inside the rectangle.
   */
  public final static double pointBox2D ( double[] p, double[] ul, double[] size )
  {
    double tmpx = ul[0] - p[0];
    double tmpy = ul[1] - p[1];

    if ( tmpx >= 0.0 )                    // left halfplane
    {
      if ( tmpy > 0.0 )
        return Math.hypot( tmpx, tmpy );

      if ( tmpy + size[1] < 0.0 )
        return Math.hypot( tmpx, tmpy + size[1] );

      return tmpx;
    }

    tmpx += size[0];

    if ( tmpx <= 0.0 )                    // right halfplane
    {
      if ( tmpy > 0.0 )
        return Math.hypot( tmpx, tmpy );

      if ( tmpy + size[1] < 0.0 )
        return Math.hypot( tmpx, tmpy + size[1] );

      return -tmpx;
    }

      // the central band:
    if ( tmpy >= 0.0 )
      return tmpy;

    tmpy += size[1];
    if ( tmpy <= 0.0 )
      return -tmpy;

    return -1.0;
  }

  /**
   * Computes 3D distance between point {@code p} and the rectangular box {@code ul, ul+size}.
   *
   * @param  p Point coordinates.
   * @param  ul Upper left box corner.
   * @param  size Box size.
   * @return Distance or {@code -1.0} if the point lies inside the box.
   */
  public final static double pointBox3D ( double[] p, double[] ul, double[] size )
  {
    double tmpx = ul[0] - p[0];
    double tmpy = ul[1] - p[1];
    double tmpz = ul[2] - p[2];

    if ( tmpz >= 0.0 )
    {
      if ( tmpx >= 0.0 )
      {
        if ( tmpy > 0.0 )
          return Formula.diagonal( tmpx, tmpy, tmpz );

        if ( tmpy + size[1] < 0.0 )
          return Formula.diagonal( tmpx, tmpy + size[1], tmpz );

        return Math.hypot( tmpx, tmpz );
      }

      tmpx += size[0];

      if ( tmpx <= 0.0 )
      {
        if ( tmpy > 0.0 )
          return Formula.diagonal( tmpx, tmpy, tmpz );

        if ( tmpy + size[1] < 0.0 )
          return Formula.diagonal( tmpx, tmpy + size[1], tmpz );

        return Math.hypot( tmpx, tmpz );
      }

      if ( tmpy >= 0.0 )
        return Math.hypot( tmpy, tmpz );

      tmpy += size[1];
      if ( tmpy <= 0.0 )
        return Math.hypot( tmpy, tmpz );

      return tmpz;
    }

    tmpz += size[2];

    if ( tmpz <= 0.0 )
    {
      if ( tmpx >= 0.0 )
      {
        if ( tmpy > 0.0 )
          return Formula.diagonal( tmpx, tmpy, tmpz );

        if ( tmpy + size[1] < 0.0 )
          return Formula.diagonal( tmpx, tmpy + size[1], tmpz );

        return Math.hypot( tmpx, tmpz );
      }

      tmpx += size[0];

      if ( tmpx <= 0.0 )
      {
        if ( tmpy > 0.0 )
          return Formula.diagonal( tmpx, tmpy, tmpz );

        if ( tmpy + size[1] < 0.0 )
          return Formula.diagonal( tmpx, tmpy + size[1], tmpz );

        return Math.hypot( tmpx, tmpz );
      }

      if ( tmpy >= 0.0 )
        return Math.hypot( tmpy, tmpz );

      tmpy += size[1];
      if ( tmpy <= 0.0 )
        return Math.hypot( tmpy, tmpz );

      return -tmpz;
    }

    if ( tmpx >= 0.0 )
    {
      if ( tmpy > 0.0 )
        return Math.hypot( tmpx, tmpy );

      if ( tmpy + size[1] < 0.0 )
        return Math.hypot( tmpx, tmpy + size[1] );

      return tmpx;
    }

    tmpx += size[0];

    if ( tmpx <= 0.0 )
    {
      if ( tmpy > 0.0 )
        return Math.hypot( tmpx, tmpy );

      if ( tmpy + size[1] < 0.0 )
        return Math.hypot( tmpx, tmpy + size[1] );

      return -tmpx;
    }

    if ( tmpy >= 0.0 )
      return tmpy;

    tmpy += size[1];
    if ( tmpy <= 0.0 )
      return -tmpy;

    return -1.0;
  }

  /**
   * Computes intersection of the given ray with the given line segment.
   * Even negative intersections are computed (ray-parameter {@code t < 0.0}).
   *
   * Returns Parametric coordinate on the ray if succeeded, {@link Double#NEGATIVE_INFINITY} otherwise.
   */
  public final static double raySegment2D ( double[] origin, double[] direction, double[] a, double[] b )
  {
    return raySegment2D( origin[0],    origin[1],
                         direction[0], direction[1],
                         a[0],         a[1],
                         b[0],         b[1] );
  }

  /**
   * Computes intersection of the given ray with the given line segment.
   * Even negative intersections are computed (ray-parameter {@code t < 0.0}).
   *
   * Returns Parametric coordinate on the ray if succeeded, {@link Double#NEGATIVE_INFINITY} otherwise.
   */
  public final static double raySegment2D ( double ox, double oy, double dx, double dy,
                                            double ax, double ay, double bx, double by )
  {
    final double nx  = ay - by;
    final double ny  = bx - ax;
    final double den = nx * dx + ny * dy;
    if ( isZero( den ) ) return Double.NEGATIVE_INFINITY;

    final double t   = (nx * (ax - ox) + ny * (ay - oy)) / den;
    double resol;

    if ( Math.abs( ny ) > Math.abs( nx ) )
    {                                     // use X coordinate
      resol = ox + t * dx;
      if ( resol < Math.min( ax, bx ) ||
           resol > Math.max( ax, bx ) ) return Double.NEGATIVE_INFINITY;
    }
    else
    {                                     // use Y coordinate
      resol = oy + t * dy;
      if ( resol < Math.min( ay, by ) ||
           resol > Math.max( ay, by ) ) return Double.NEGATIVE_INFINITY;
    }

    return t;
  }

  /**
   * Computes intersection of the ray with the given AABB (works in either dimension).
   *
   * Returns {@code true} if intersection exists (in such case the pre-allocated result array will be
   *         filled with [ t<sub>min</sub>. t<sub>max</sub> ]).
   */
  public final static boolean rayBox ( double[] origin, double[] direction, double[] ul, double[] size, double[] result )
  {
    final int dim = origin.length;
    double tMin = Double.NEGATIVE_INFINITY;
    double tMax = Double.POSITIVE_INFINITY;
    double t1, t2, mul;

    for ( int d = 0; d < dim; d++ )
      if ( isZero( direction[d] ) )
      {
        if ( origin[d] < ul[d] ||
             origin[d] > ul[d] + size[d] )
          return false;
      }
      else
      {
        mul = 1.0 / direction[d];
        t1  = (ul[d] - origin[d]) * mul;
        t2  = t1 + size[d] * mul;

        if ( mul > 0.0 )
        {
          if ( t1 > tMin ) tMin = t1;
          if ( t2 < tMax ) tMax = t2;
        }
        else
        {
          if ( t2 > tMin ) tMin = t2;
          if ( t1 < tMax ) tMax = t1;
        }

        if ( tMax < 0.0 ||
             tMin > tMax ) return false;
      }

    if ( result != null )
    {
      result[0] = tMin;
      result[1] = tMax;
    }

    return true;
  }

  /**
   * Ray-triangle intersection test in 3D.
   * According to Tomas Moller, <a href="http://www.acm.org/jgt/">JGT</a>.
   * <br />{@code origin + t * direction = (1 - u - v) * a + u * b + v * c}.
   *
   * @param  uv Pre-allocated array to hold barycentric coordinates.
   * @return Parametric coordinate on the ray if succeeded, {@link Double#NEGATIVE_INFINITY} otherwise.
   */
  public final static double rayTriangle3D ( double[] origin, double[] direction,
                                             double[] a, double[] b, double[] c,
                                             double[] uv )
  {
    final double[] e1 = { b[0] - a[0],
                          b[1] - a[1],
                          b[2] - a[2] };

    final double[] e2 = { c[0] - a[0],
                          c[1] - a[1],
                          c[2] - a[2] };

    final double[] pvec = vectorProduct( direction, e2, null );

    final double det = e1[0] * pvec[0] +
                       e1[1] * pvec[1] +
                       e1[2] * pvec[2];

    if ( isZero( det ) ) return Double.NEGATIVE_INFINITY;
    final double detInv = 1.0 / det;

    final double[] tvec = { origin[0] - a[0],
                            origin[1] - a[1],
                            origin[2] - a[2] };

    final double u = detInv * ( tvec[0] * pvec[0] +
                                tvec[1] * pvec[1] +
                                tvec[2] * pvec[2] );

    if ( u < 0.0 || u > 1.0 ) return Double.NEGATIVE_INFINITY;

    final double[] qvec = vectorProduct( tvec, e1, null );

    final double v = detInv * ( direction[0] * qvec[0] +
                                direction[1] * qvec[1] +
                                direction[2] * qvec[2] );

    if ( v < 0.0 || u + v > 1.0 ) return Double.NEGATIVE_INFINITY;

    if ( uv != null &&
         uv.length >= 2 )
    {
      uv[0] = u;
      uv[1] = v;
    }

    return( detInv * ( e2[0] * qvec[0] +
                       e2[1] * qvec[1] +
                       e2[2] * qvec[2] ) );
  }

  /**
   * Tests whether the given point lies in the given triangle.
   * In positive case: computes barycentric coordinates of the point:
   * <p>{@code p = (1 - u - v) * a + u * b + v * c}.
   *
   * @param  p 2D point.
   * @param  a Vertex A of the triangle.
   * @param  b Vertex B of the triangle.
   * @param  c Vertex C of the triangle.
   * @param  uv Pre-allocated array to hold barycentric coordinates.
   * @return Barycentric coordinate-array in positive case, {@code null} otherwise.
   */
  public final static double[] pointInTriangle2D ( double[] p, double[] a, double[] b, double[] c, double[] uv )
  {
    if ( uv == null || uv.length < 2 )
      uv = new double[ 2 ];
    else
      uv[0] = uv[1] = 0.0;

    double e0x = p[0] - a[0];
    double e0y = p[1] - a[1];
    double e1x = b[0] - a[0];
    double e1y = b[1] - a[1];
    double e2x = c[0] - a[0];
    double e2y = c[1] - a[1];

    if ( e1x > -EPSILON && e1x < EPSILON )
    {
      if ( e2x > -EPSILON && e2x < EPSILON )
        return null;

      uv[0] = e0x / e2x;

      if ( uv[0] < 0.0 || uv[0] > 1.0 )
        return null;

      if ( e1y > -EPSILON && e1y < EPSILON )
        return null;

      uv[1] = (e0y - e2y * uv[0]) / e1y;

      if ( uv[1] < 0 )
        return null;
    }
    else
    {
      double d = e2y * e1x - e2x * e1y;

      if ( d > -EPSILON && d < EPSILON )
        return null;

      uv[1] = (e0y * e1x - e0x * e1y) / d;

      if ( uv[1] < 0.0 || uv[1] > 1.0 )
        return null;

      uv[0] = (e0x - e2x * uv[1]) / e1x;

      if ( uv[0] < 0 )
        return null;
    }

    if ( uv[0] + uv[1] > 1.0 )
      return null;

    return uv;
  }

  /**
   * Computes a vector from its endpoints.
   *
   * @param  a Vector's starting point.
   * @param  b Vector's end point.
   * @param  result Optional pre-allocated result array ({@code double[3]} or
   *                {@code double[4]}).
   * @return Vector coordinates or {@code null} for bad parameters, {@code double[3]} for
   *         {@code result == null}.
   */
  public final static double[] vector ( double[] a, double[] b, double[] result )
  {
    int alen, blen;
    if ( a == null || (alen = a.length) < 3 ||
         b == null || (blen = b.length) < 3 ) return null;

    double ax, ay, az, w;
    if ( blen == 3 || (w = b[3]) == 1.0 || w > -EPSILON && w < EPSILON )
    {
      ax = b[0];
      ay = b[1];
      az = b[2];
    }
    else
    {
      w  = 1.0 / w;
      ax = w * b[0];
      ay = w * b[1];
      az = w * b[2];
    }
    if ( alen == 3 || (w = a[3]) == 1.0 || w > -EPSILON && w < EPSILON )
    {
      ax -= a[0];
      ay -= a[1];
      az -= a[2];
    }
    else
    {
      ax -= w * a[0];
      ay -= w * a[1];
      az -= w * a[2];
    }
    if ( result == null || result.length < 3 )
      return new double[] { ax, ay, az, 0.0 };

    result[0] = ax;
    result[1] = ay;
    result[2] = az;
    if ( result.length > 3 ) result[3] = 0.0;

    return result;
  }

  /**
   * Computes area of the given triangle.
   * Triangle is defined by its vertices (cartesian .. [x,y,z]).
   */
  public final static double triangleArea ( double[][]v )
  {
    if ( v == null ) return 0.0;

    final double bx = v[1][0] - v[0][0];
    final double by = v[1][1] - v[0][1];
    final double bz = v[1][2] - v[0][2];

    final double cx = v[2][0] - v[0][0];
    final double cy = v[2][1] - v[0][1];
    final double cz = v[2][2] - v[0][2];

      // area = 0.5 * sqrt( b.b * c.c - (b.c)^2 )
    final double bc = bx * cx + by * cy + bz * cz;

    return Math.sqrt( 0.25 * ( (bx * bx + by * by + bz * bz) * (cx * cx + cy * cy + cz * cz) - bc * bc ) );
  }

  /**
   * Computes normal vector of the given triangle.
   * Triangle is defined by its vertices (cartesian .. [x,y,z]).
   */
  public final static double[] triangleNormal ( double[][]v, double[] normal )
  {
    if ( v == null ) return null;

    final double bx = v[1][0] - v[0][0];
    final double by = v[1][1] - v[0][1];
    final double bz = v[1][2] - v[0][2];

    final double cx = v[2][0] - v[0][0];
    final double cy = v[2][1] - v[0][1];
    final double cz = v[2][2] - v[0][2];

    if ( normal        == null ||
         normal.length < 3 )
      normal = new double[ 3 ];

    normal[0] = by * cz - bz * cy;
    normal[1] = bz * cx - bx * cz;
    normal[2] = bx * cy - by * cx;

    return normalize3D( normal, normal );
  }

  /**
   * Computes dot product of two 3D vectors.
   *
   * @param  a The first argument.
   * @param  b The second argument.
   * @return Dot product.
   */
  public final static double dotProduct3D ( double[] a, double[] b )
  {
    if ( a == null || a.length < 3 ||
         b == null || b.length < 3 ) return 0.0;

    return( a[0] * b[0] + a[1] * b[1] + a[2] * b[2] );
  }

  /**
   * Computes vector product of two (either cartesian or homogeneous) 3D vectors.
   *
   * @param  a The first argument.
   * @param  b The second argument.
   * @param  result Optional pre-allocated result array ({@code double[3]} or
   *                {@code double[4]}).
   * @return Vector product: {@code null} for bad parameters, {@code double[3]} for
   *         {@code result == null}.
   */
  public final static double[] vectorProduct ( double[] a, double[] b, double[] result )
  {
    if ( a == null || a.length < 3 ||
         b == null || b.length < 3 ) return null;

    double r0 = a[1] * b[2] - a[2] * b[1];
    double r1 = a[2] * b[0] - a[0] * b[2];
    double r2 = a[0] * b[1] - a[1] * b[0];

    if ( result == null || result.length < 3 )
      return new double[] { r0, r1, r2, 0.0 };

    result[0] = r0;
    result[1] = r1;
    result[2] = r2;
    if ( result.length > 3 )
      result[3] = 0.0;

    return result;
  }

  /**
   * Finds two vectors perpendicular to the given vector (so that their vector product will
   * give the original vector).
   *
   * @param  p Inout vector ({@code double[3]} or {@code double[4]}).
   * @param  p1 The 1st output vector (pre-allocated array of size {@code 3} or
   *            {@code 4}).
   * @param  p2 The 2nd output vector (pre-allocated array of size {@code 3} or
   *            {@code 4}).
   * @see    #vectorProduct
   * @see    #getAxes
   */
  public final static void getPerpendiculars ( double[] p, double[] p1, double[] p2 )
  {
    if ( p  == null || p.length  < 3 ||
         p1 == null || p1.length < 3 ||
         p2 == null || p2.length < 3 ) return;

    if ( p[0] < -EPSILON || p[0] > EPSILON )
    {
      if ( p[0] > 0.0 )
      {
        p1[0] = -p[1];
        p1[1] =  p[0];
      }
      else
      {
        p1[0] =  p[1];
        p1[1] = -p[0];
      }

      p1[2] =  0.0;
      p2[0] = -p[2];
      p2[1] =  0.0;
      p2[2] =  p[0];
    }

    else
    if ( p[1] < -EPSILON || p[1] > EPSILON )
    {
      if ( p[1] > 0.0 )
      {
        p1[0] = -p[1];
        p1[1] =  p[0];
      }
      else
      {
        p1[0] =  p[1];
        p1[1] = -p[0];
      }

      p1[2] =  0.0;
      p2[0] =  0.0;
      p2[1] = -p[2];
      p2[2] =  p[1];
    }

    else
    {
      if ( p[2] > 0.0 )
      {
        p1[0] = -p[2];
        p1[2] =  p[0];
      }
      else
      {
        p1[0] =  p[2];
        p1[2] = -p[0];
      }

      p1[1] =  0.0;
      p2[0] =  0.0;
      p2[1] = -p[2];
      p2[2] =  p[1];
    }

    if ( p1.length > 3 ) p1[3] = 0.0;
    if ( p2.length > 3 ) p2[3] = 0.0;
  }

  /**
   * Finds two other axes to the given vector, their vector product will
   * give the original vector, all three vectors will be perpendicular to each other.
   *
   * @param  p Inout vector ({@code double[3]} or {@code double[4]}).
   * @param  p1 The 1st output vector (pre-allocated array of size {@code 3} or
   *            {@code 4}).
   * @param  p2 The 2nd output vector (pre-allocated array of size {@code 3} or
   *            {@code 4}).
   * @see    #vectorProduct
   * @see    #getPerpendiculars
   */
  public final static void getAxes ( double[] p, double[] p1, double[] p2 )
  {
    if ( p  == null || p.length  < 3 ||
         p1 == null || p1.length < 3 ||
         p2 == null || p2.length < 3 ) return;

    final double ax = Math.abs( p[0] );
    final double ay = Math.abs( p[1] );
    final double az = Math.abs( p[2] );

    if ( ax >= az &&
         ay >= az )                       // ax, ay are dominant
    {
      p1[0] = -p[1];
      p1[1] =  p[0];
      p1[2] =  0.0;
    }

    else
    if ( ax >= ay &&
         az >= ay )                       // ax, az are dominant
    {
      p1[0] = -p[2];
      p1[1] =  0.0;
      p1[2] =  p[0];
    }

    else                                  // ay, az are dominant
    {
      p1[0] =  0.0;
      p1[1] = -p[2];
      p1[2] =  p[1];
    }

    if ( p1.length > 3 ) p1[3] = 0.0;
    if ( p2.length > 3 ) p2[3] = 0.0;

    vectorProduct( p, p1, p2 );
  }

  /**
   * Computes direction vector of perfect specular-reflected ray.
   *
   * @param  normal Normal vector (points to the outside of the solid). {@code double[3]}
   *                or {@code double[4]}.
   * @param  in Input vector (points from the intersection point to the light source).
   *            {@code double[3]} or {@code double[4]}.
   * @param  out Pre-allocated array to hold result - points from the intersection point.
   *             {@code double[3]} or {@code double[4]}, can be {@code null}.
   * @return Direction vector which points from the intersection point.
   * @see    #specularRefraction
   */
  public final static double[] specularReflection ( double[] normal, double[] in, double[] out )
  {
    if ( normal == null || normal.length < 3 ||
         in     == null || in.length     < 3 ) return null;

    if ( out == null ) out = new double[ 4 ];

      // input vector:
    final double inx = in[0];
    final double iny = in[1];
    final double inz = in[2];

      // normal vector:
    final double nx = normal[0];
    final double ny = normal[1];
    final double nz = normal[2];

    double k = nx * inx + ny * iny + nz * inz;
    k += k;
    out[0] = k * nx - inx;
    out[1] = k * ny - iny;
    out[2] = k * nz - inz;
    if ( out.length > 3 ) out[3] = 0.0;

    return out;
  }

  /**
   * Computes direction vector of perfect specular-refracted ray.
   *
   * @param  normal Normal vector ({@code double[3]} or {@code double[4]}).
   * @param  n Relative index of refraction ({@code v(normal) / v(-normal)}).
   * @param  in Input vector (points from the intersection point to the light source).
   *            {@code double[3]} or {@code double[4]}.
   * @param  out Pre-allocated array to hold result - points from the intersection point.
   *             {@code double[3]} or {@code double[4]}, can be {@code null}.
   * @return Direction vector which points from the intersection point or {@code null}
   *         in case of <b>total reflection</b>.
   * @see    #specularReflection
   */
  public final static double[] specularRefraction ( double[] normal, double n,
                                                    double[] in, double[] out )
  {
    if ( normal == null || normal.length < 3 ||
         in     == null || in.length     < 3 ) return null;

      // input vector:
    double inx = in[0];
    double iny = in[1];
    double inz = in[2];
    double d   = inx * inx + iny * iny + inz * inz;
    if ( d > Geometry.EPSILON && (d < 1.0 - Geometry.EPSILON || d > 1.0 + Geometry.EPSILON) )
    {
      d    = 1.0 / Math.sqrt( d );
      inx *= d;
      iny *= d;
      inz *= d;
    }

      // normal vector:
    double nx = normal[0];
    double ny = normal[1];
    double nz = normal[2];
    d         = nx * nx + ny * ny + nz * nz;
    if ( d > Geometry.EPSILON && (d < 1.0 - Geometry.EPSILON || d > 1.0 + Geometry.EPSILON) )
    {
      d   = 1.0 / Math.sqrt( d );
      nx *= d;
      ny *= d;
      nz *= d;
    }

    d = inx * nx + iny * ny + inz * nz;   // scalar product (N*L)
    if ( d < 0.0 )                        // (N*L) should be > 0.0 (N and L in the same half-space)
    {
      d  = -d;
      nx = -nx;
      ny = -ny;
      nz = -nz;
    }
    else
      n  = 1.0 / n;

    double cos2 = 1.0 - n * n * (1.0 - d * d);
    if ( cos2 <= 0.0 ) return null;       // total reflection

    if ( out == null ) out = new double[ 4 ];

    d      = n * d - Math.sqrt( cos2 );
    out[0] = d * nx - n * inx;
    out[1] = d * ny - n * iny;
    out[2] = d * nz - n * inz;
    if ( out.length > 3 ) out[3] = 0.0;

    return out;
  }

  /**
   * Normalizes the given 3D vector (Cartesian or homogeneous) <b>in place</b>.
   *
   * @param  v Vector to be normalized ({@code double[3]} or {@code double[4]}).
   *           {@code v[3]} will be ignored (not changed).
   * @return Size (Euclidean norm) of the original vector.
   */
  public final static double normalize3D ( double[] v )
  {
    if ( v == null || v.length < 3 ) return 0.0;

    double size = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
    if ( size < EPSILON ) return 0.0;

    size = Math.sqrt( size );
    double mul = 1.0 / size;
    v[0] *= mul;
    v[1] *= mul;
    v[2] *= mul;
    if ( v.length > 3 )
      v[3] = 0.0;

    return size;
  }

  /**
   * Normalizes the given 3D vector (Cartesian or homogeneous) and <b>assigns</b> it.
   *
   * @param  from Source coordinate vector ({@code v[3]} /if present/ will be ignored).
   * @param  to Pre-allocated destination coordinate vector ({@code double[3]} or
   *            {@code double[4]}).
   * @return Normalized vector ({@code double[3]} for {@code from == null}) or
   *         {@code null} in case of error.
   */
  public final static double[] normalize3D ( double[] from, double[] to )
  {
    if ( from == null || from.length < 3 ) return null;

    double size = from[0] * from[0] + from[1] * from[1] + from[2] * from[2];
    if ( size < EPSILON ) return null;

    size = 1.0 / Math.sqrt( size );
    if ( to == null || to.length < 3 )
      return new double[] { size * from[0], size * from[1], size * from[2] };

    to[0] = size * from[0];
    to[1] = size * from[1];
    to[2] = size * from[2];
    if ( to.length > 3 ) to[3] = 0.0;

    return to;
  }

  /**
   * Converts the given coordinate vector to Cartesian 3D coordinates.
   *
   * @param  from Source coordinate vector ({@code double[3]}  or {@code double[4]}).
   * @param  to Pre-allocated destination coordinate vector ({@code double[3]} or
   *            {@code null}).
   * @return Cartesian 3D coordinate vector or {@code null} in case of error.
   * @see    #toCartesian
   */
  public final static double[] toCartesian3D ( double[] from, double[] to )
  {
    if ( from == null || from.length < 3 ) return null;

    if ( from.length == 3 || from[3] > -EPSILON && from[3] < EPSILON )
    {
      if ( to == null || to.length < 3 )
        return new double[] { from[0], from[1], from[2] };

      System.arraycopy( from, 0, to, 0, 3 );
      if ( to.length > 3 )
        to[3] = 1.0;
      return to;
    }

    double mul = 1.0 / from[3];

    if ( to == null || to.length < 3 )
      return new double[] { mul * from[0], mul * from[1], mul * from[2] };

    to[0] = mul * from[0];
    to[1] = mul * from[1];
    to[2] = mul * from[2];
    if ( to.length > 3 )
      to[3] = 1.0;

    return to;
  }

  /**
   * Converts the given coordinate vector to Cartesian 3D coordinates.
   *
   * @param  x X coordinate.
   * @param  y Y coordinate.
   * @param  z Z coordinate.
   * @param  w W coordinate.
   * @param  to Pre-allocated destination coordinate vector ({@code double[3]} or
   *            {@code null}).
   * @return Cartesian 3D coordinate vector.
   * @see    #toCartesian3D
   */
  public final static double[] toCartesian ( double x, double y, double z, double w, double[] to )
  {
    if ( w > -EPSILON && w < EPSILON )
    {
      if ( to == null || to.length < 3 )
        return new double[] { x, y, z };

      to[0] = x;
      to[1] = y;
      to[2] = z;
      if ( to.length > 3 )
        to[3] = 1.0;
      return to;
    }

    double mul = 1.0 / w;

    if ( to == null || to.length < 3 )
      return new double[] { mul * x, mul * y, mul * z };

    to[0] = mul * x;
    to[1] = mul * y;
    to[2] = mul * z;
    if ( to.length > 3 )
      to[3] = 1.0;

    return to;
  }

  /**
   * Converts spherical coordinate vector to Cartesian 3D coordinates.
   *
   * @param  sinTheta Sinus of angle theta.
   * @param  cosTheta Cosinus of angle theta.
   * @param  phi      Angle phi.
   * @param  result   Pre-allocated destination coordinate vector
   *                  ({@code double[3]} or {@code null}).
   * @return Cartesian 3D coordinate vector.
   * @see    #toCartesian3D
   */
  public final static double[] toCartesian ( double sinTheta,
                                             double cosTheta,
                                             double phi,
                                             double[] result )
  {
    if ( result == null )
      result = new double[ 3 ];

    result[0] = sinTheta * Math.cos(phi);
    result[1] = cosTheta * Math.sin(phi);
    result[2] = cosTheta;
    
    return result;
  }
  
  /** Converts {@code double[]} vector to readable form. */
  public final static String vectorToString ( double[] v )
  {
    if ( v == null ) return "null";

    StringBuffer sb = new StringBuffer( "[" );
    for ( double d : v )
      sb.append( String.format( Locale.US, " %9.4f", d ) );
    sb.append( " ]" );
    return sb.toString();
  }

}
