package cz.cuni.jagrlib;

/**
 * General purpose hash function algorithms library.
 * Original author: Arash Partow (2002), http://www.partow.net/
 * Adopted by: Josef Pelikan (2006).
 * Copyright notice: Common Public License (http://www.opensource.org/licenses/cpl.php).
 *
 * @version 0.25 $Rev: 350 $ $Date: 2006-11-09 00:34:24 +0100 (čt, 09 XI 2006) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../cz/cuni/jagrlib/HashFunctions.java">HashFunctions.java</a>
 */
public final class HashFunction
{
  //==========================================================================
  //  Constants:

  final static int  BITS_IN_UNSIGNED_INT = 32;
  final static int  THREE_QUARTERS       = (BITS_IN_UNSIGNED_INT * 3) / 4;
  final static int  ONE_EIGHTH           = BITS_IN_UNSIGNED_INT / 8;
  final static long BITS_31              = 0x7FFFFFFFL;
  final static long BITS_32              = 0xFFFFFFFFL;
  final static long BITS_16              = 0xFFFFL;
  final static int  BIT_31               = 0x80000000;
  final static int  BIT_30               = 0x40000000;

  final static int  STATIC_ARRAY         = 16;

  //==========================================================================
  //  Instance data:

  /** Implemented hash-functions. */
  public static enum HashType
  {
    /** LCG-based. */
    HASH_LCG,
    /** Robert Sedgwick. */
    HASH_RS,
    /** Justin Sobel. */
    HASH_JS,
    /** Peter J. Weinberger. */
    HASH_PJW,
    /** ELF (UNIX). */
    HASH_ELF,
    /** Brian Kernighan, Dennis Ritchie. */
    HASH_BKDR,
    /** SDBM database. */
    HASH_SDBM,
    /** Daniel J. Bernstein. */
    HASH_DJB,
    /** Donald E. Knuth. */
    HASH_DEK,
    /** Arash Partow. */
    HASH_AP,
  }

  final static int MAX_TYPE = HashType.HASH_AP.ordinal() + 1;

  /** Current hash-function type. */
  protected HashType type = HashType.HASH_AP;

  /**
   * Hash-function parameter.
   * While used by BKDR, it should be: 31 131 1313 13131 131313, ...
   */
  protected long param = 131L;

  /** Permutation ordinal number (for coordinate hashing). */
  protected int permutation = 0;

  /** Index to actually processed array. */
  protected int index;

  /** Points after the data segment to be processed. */
  protected int endIndex;

  /** Hash data: String. */
  protected String string;

  protected byte[]  byteArray;
  protected short[] shortArray;
  protected char[]  charArray;
  protected int[]   intArray;
  protected long[]  longArray;

  protected short[] staticArray = new short[ STATIC_ARRAY ];

  //==========================================================================
  //  Support:

  /** Initialize staticArray distribution. Use distribute() methods afterwards. */
  protected final void initDistribution ()
  {
    staticArray[0] = 25111;
    staticArray[1] = 30013;
    index          = 0;
    endIndex       = 2;
  }

  /** Distributes integer value (32-bit) into two staticArray items. */
  protected final void distribute ( int x )
  {
    int buf0 = 0;
    int buf1 = 0;
    for ( int b = 0; b++ < 16; x <<= 2 )
    {
      buf0 += buf0 + ( ( (x & BIT_31) != 0 ) ? 1 : 0 );
      buf1 += buf1 + ( ( (x & BIT_30) != 0 ) ? 1 : 0 );
    }
    staticArray[ endIndex++ ] = (short)buf0;
    staticArray[ endIndex++ ] = (short)buf1;
  }

  /** Distributes long value (64-bit) into four staticArray items. */
  protected final void distribute ( long x )
  {
    distribute( (int)( x & BITS_32 ) );
    distribute( (int)( x >>> 32 ) );
  }

  /** Finish staticArray distribution. Prepares staticArray[] for hash() methods. */
  protected final void finishDistribution ()
  {
    staticArray[ endIndex++ ] = 8101;
    staticArray[ endIndex++ ] = 19031;
    index                     = 0;
    if ( permutation > 0 ) permutate();
  }

  /** Returns the next item (char, number). */
  protected final int next ()
  {
    if ( index >= endIndex ) return 0;

    if ( string != null )
      return string.charAt( index++ );

    if ( byteArray != null )
      return byteArray[ index++ ];

    if ( shortArray != null )
      return shortArray[ index++ ];

    if ( charArray != null )
      return charArray[ index++ ];

    if ( intArray != null )
      return intArray[ index++ ];

    if ( longArray != null )
      return (int)( longArray[ index++ ] );

    return staticArray[ index++ ];
  }

  /** Internal hashing function (switches between the implementations). */
  protected final long hash ()
  {
    switch ( type )
    {
      case HASH_LCG:
        return LCGHash();

      case HASH_RS:
        return RSHash();

      case HASH_JS:
        return JSHash();

      case HASH_PJW:
        return PJWHash();

      case HASH_ELF:
        return ELFHash();

      case HASH_BKDR:
        return BKDRHash();

      case HASH_SDBM:
        return SDBMHash();

      case HASH_DJB:
        return DJBHash();

      case HASH_DEK:
        return DEKHash();

      case HASH_AP:
      default:
        return APHash();
    }
  }

  /** Permutates the staticArray[ 0 .. endIndex-1 ] by permutation order "permutation". */
  protected final void permutate ()
  {
    // !!! TODO: not implemented yet !!!
  }

  //==========================================================================
  //  Public methods:

  public HashFunction ()
  {
  }

  public HashFunction ( HashType type )
  {
    this.type = type;
  }

  public HashFunction ( HashType type, long parameter )
  {
    this.type = type;
    param     = parameter;
  }

  public final void setType ( HashType type )
  {
    this.type = type;
  }

  public final long setParam ( long parameter )
  {
    long old = param;
    param    = parameter;
    return old;
  }

  public final int setPerm ( int perm )
  {
    int old     = permutation;
    permutation = perm;
    return old;
  }

  /**
   * Sets hash-function variation.
   *
   * @param  var Variation ordinal number (0, 1, ...).
   */
  public final void setVariation ( int var )
  {
    switch ( var % MAX_TYPE )
    {
      case 0:
        type = HashType.HASH_LCG;
        break;
      case 1:
        type = HashType.HASH_RS;
        break;
      case 2:
        type = HashType.HASH_JS;
        break;
      case 3:
        type = HashType.HASH_PJW;
        break;
      case 4:
        type = HashType.HASH_ELF;
        break;
      case 5:
        type = HashType.HASH_BKDR;
        break;
      case 6:
        type = HashType.HASH_SDBM;
        break;
      case 7:
        type = HashType.HASH_DJB;
        break;
      case 8:
        type = HashType.HASH_DEK;
        break;
      default:
        type = HashType.HASH_AP;
    }
    permutation = var / MAX_TYPE;
  }

  /** Hash-function returns values: 0 to maxHash()-1. */
  public final long maxHash ()
  {
    return BITS_31 + 1;
  }

  /** Hashing the whole string. */
  public final long hash ( String s )
  {
    string   = s;
    index    = 0;
    endIndex = (s == null) ? 0 : s.length();

    long result = hash();

    string   = null;

    return result;
  }

  /** Hashing the substring. */
  public final long hash ( String s, int from, int to )
  {
    string   = s;
    index    = (from < 0) ? 0 : from;
    endIndex = (s == null) ? 0 : s.length();
    if ( to < endIndex ) endIndex = to;

    long result = hash();

    string   = null;

    return result;
  }

  /** Hashing the whole byte-array. */
  public final long hash ( byte[] a )
  {
    byteArray = a;
    index     = 0;
    endIndex  = (a == null) ? 0 : a.length;

    long result = hash();

    byteArray = null;

    return result;
  }

  /** Hashing the segment from byte-array. */
  public final long hash ( byte[] a, int from, int to )
  {
    byteArray = a;
    index     = (from < 0) ? 0 : from;
    endIndex  = (a == null) ? 0 : a.length;
    if ( to < endIndex ) endIndex = to;

    long result = hash();

    byteArray = null;

    return result;
  }

  /** Hashing the whole short-array. */
  public final long hash ( short[] a )
  {
    shortArray = a;
    index      = 0;
    endIndex   = (a == null) ? 0 : a.length;

    long result = hash();

    shortArray = null;

    return result;
  }

  /** Hashing the segment from short-array. */
  public final long hash ( short[] a, int from, int to )
  {
    shortArray = a;
    index      = (from < 0) ? 0 : from;
    endIndex   = (a == null) ? 0 : a.length;
    if ( to < endIndex ) endIndex = to;

    long result = hash();

    shortArray = null;

    return result;
  }

  /** Hashing the whole char-array. */
  public final long hash ( char[] a )
  {
    charArray = a;
    index     = 0;
    endIndex  = (a == null) ? 0 : a.length;

    long result = hash();

    charArray = null;

    return result;
  }

  /** Hashing the segment from char-array. */
  public final long hash ( char[] a, int from, int to )
  {
    charArray = a;
    index     = (from < 0) ? 0 : from;
    endIndex  = (a == null) ? 0 : a.length;
    if ( to < endIndex ) endIndex = to;

    long result = hash();

    charArray = null;

    return result;
  }

  /** Hashing the whole int-array. */
  public final long hash ( int[] a )
  {
    intArray = a;
    index    = 0;
    endIndex = (a == null) ? 0 : a.length;

    long result = hash();

    intArray = null;

    return result;
  }

  /** Hashing the segment from int-array. */
  public final long hash ( int[] a, int from, int to )
  {
    intArray = a;
    index    = (from < 0) ? 0 : from;
    endIndex = (a == null) ? 0 : a.length;
    if ( to < endIndex ) endIndex = to;

    long result = hash();

    intArray = null;

    return result;
  }

  /** Hashing the whole long-array. */
  public final long hash ( long[] a )
  {
    longArray = a;
    index     = 0;
    endIndex  = (a == null) ? 0 : a.length;

    long result = hash();

    longArray = null;

    return result;
  }

  /** Hashing the segment from long-array. */
  public final long hash ( long[] a, int from, int to )
  {
    longArray = a;
    index     = (from < 0) ? 0 : from;
    endIndex  = (a == null) ? 0 : a.length;
    if ( to < endIndex ) endIndex = to;

    long result = hash();

    longArray = null;

    return result;
  }

  /** Hashing integer coordinate tuple. */
  public final long hash ( int x, int y )
  {
    initDistribution();
    distribute( x );
    distribute( y );
    finishDistribution();

    return hash();
  }

  /** Hashing integer coordinate triple. */
  public final long hash ( int x, int y, int z )
  {
    initDistribution();
    distribute( x );
    distribute( y );
    distribute( z );
    finishDistribution();

    return hash();
  }

  /** Hashing float coordinate tuple. */
  public final long hash ( float x, float y )
  {
    initDistribution();
    distribute( Float.floatToIntBits( x ) );
    distribute( Float.floatToIntBits( y ) );
    finishDistribution();

    return hash();
  }

  /** Hashing float coordinate triple. */
  public final long hash ( float x, float y, float z )
  {
    initDistribution();
    distribute( Float.floatToIntBits( x ) );
    distribute( Float.floatToIntBits( y ) );
    distribute( Float.floatToIntBits( z ) );
    finishDistribution();

    return hash();
  }

  /** Hashing double coordinate tuple. */
  public final long hash ( double x, double y )
  {
    initDistribution();
    distribute( Double.doubleToLongBits( x ) );
    distribute( Double.doubleToLongBits( y ) );
    finishDistribution();

    return hash();
  }

  /** Hashing double coordinate triple. */
  public final long hash ( double x, double y, double z )
  {
    initDistribution();
    distribute( Double.doubleToLongBits( x ) );
    distribute( Double.doubleToLongBits( y ) );
    distribute( Double.doubleToLongBits( z ) );
    finishDistribution();

    return hash();
  }

  //==========================================================================
  //  Individual hashing algorithms:

  /**
   * LCG-based hash function (LCG is from "Numerical Recipes in C").
   */
  protected final long LCGHash ()
  {
    long hash = 0L;

    while ( index < endIndex )
      hash = (hash * 1664525L + 1013904223L + next()) & BITS_32;

    return hash;
  }

  /**
   * A simple hash function from Robert Sedgwicks Algorithms in C book.
   * I've added some simple optimizations to the algorithm in order
   * to speed up its hashing process.
   */
  protected final long RSHash ()
  {
    int  b    = 378551;
    int  a    = 63689;
    long hash = 0L;

    while ( index < endIndex )
    {
      hash = hash * a + next();
      a   *= b;
    }

    return( hash & BITS_31 );
  }

  /**
   * A bitwise hash function written by Justin Sobel.
   */
  protected final long JSHash ()
  {
    long hash = 1315423911L;

    while ( index < endIndex )
      hash ^= ((hash << 5) + next() + (hash >> 2));

    return( hash & BITS_31 );
 }

 /**
  * This hash algorithm is based on work by Peter J. Weinberger
  * of AT&T Bell Labs.
  */
  protected final long PJWHash ()
  {
    final long HIGH_BITS = BITS_32 << (BITS_IN_UNSIGNED_INT - ONE_EIGHTH);
    long hash = 0L;
    long test = 0L;

    while ( index < endIndex )
    {
      hash = (hash << ONE_EIGHTH) + next();

      if ( (test = hash & HIGH_BITS) != 0 )
        hash = ( (hash ^ (test >> THREE_QUARTERS)) & ~HIGH_BITS );
    }

    return( hash & BITS_31 );
  }

  /**
   * Similar to the PJW Hash function, but tweaked for 32-bit processors.
   * Its the hash function widely used on most UNIX systems.
   */
  protected final long ELFHash ()
  {
    long hash = 0L;
    long x    = 0L;

    while ( index < endIndex )
    {
      hash = (hash << 4) + next();

      if ( (x = hash & 0xF0000000L) != 0 )
      {
        hash ^= (x >> 24);
        hash &= ~x;
      }
    }

    return( hash & BITS_31 );
  }

  /**
   * This hash function comes from Brian Kernighan and Dennis Ritchie's book
   * "The C Programming Language". It is a simple hash function using a strange
   * set of possible seeds which all constitute a pattern of 31....31...31 etc,
   * it seems to be very similar to the DJB hash function.
   */
  protected final long BKDRHash ()
  {
    //long param = 131L;                    // 31 131 1313 13131 131313 etc..
    long hash = 0L;

    while ( index < endIndex )
      hash = (hash * param) + next();

    return( hash & BITS_31 );
  }

  /**
   * This is the algorithm of choice which is used in the open source SDBM project.
   * The hash function seems to have a good over-all distribution for many different
   * data sets. It seems to work well in situations where there is a high variance
   * in the MSBs of the elements in a data set.
   */
  protected final long SDBMHash ()
  {
    int hash = 0;

    while ( index < endIndex )
      hash = next() + (hash << 6) + (hash << 16) - hash;

    return( hash & BITS_31 );
  }

  /**
   * An algorithm produced by Professor Daniel J. Bernstein and shown first
   * to the world on the usenet newsgroup comp.lang.c.
   * It is one of the most efficient hash functions ever published.
   */
  protected final long DJBHash ()
  {
    int hash = 5381;

    while ( index < endIndex )
      hash = ((hash << 5) + hash) + next();

    return( hash & BITS_31 );
  }

  /**
   * An algorithm proposed by Donald E. Knuth in The Art Of Computer Programming Volume 3,
   * under the topic of sorting and search chapter 6.4.
   */
  protected final long DEKHash ()
  {
    int hash = endIndex - index;

    while ( index < endIndex )
      hash = ((hash << 5) ^ (hash >> 27)) ^ next();

    return( hash & BITS_31 );
  }

  /**
   * An algorithm produced by original author of this library: Arash Partow.
   * See http://www.partow.net/programming/hashfunctions/
   *
   * Arash writes:
   * I took ideas from all of the above hash functions making a hybrid rotative
   * and additive hash function algorithm based around four primes 3,5,7 and 11.
   * There isn't any real mathematical analysis explaining why one should use
   * this hash function instead of the others described above other than the fact
   * that I tired to resemble the design as close as possible to a simple LFSR.
   *
   * An empirical result which demonstrated the distributive abilities of the hash
   * algorithm was obtained using a hash-table with 100003 buckets, hashing
   * The Project Gutenberg Etext of Webster's Unabridged Dictionary, the longest
   * encountered chain length was 7, the average chain length was 2, the number
   * of empty buckets was 4579.
   */
  protected final long APHash ()
  {
    int hash = 0;

    while ( index < endIndex )
      if ( (index & 1) == 0 )
        hash ^= ((hash << 7) ^ next() ^ (hash >> 3));
      else
        hash ^= (~((hash << 11) ^ next() ^ (hash >> 5)));

    return( hash & BITS_31 );
  }

}
