package cz.cuni.jagrlib;

import java.util.*;
import java.io.*;

/**
 * This object maps integer keys to Object values.
 * Each key can map to at most one Object instance, {@code null} values
 * are allowed.
 *
 * <p>Efficient alternative to {@code Map<K,V>} container classes.
 * Method signatures are as similar as possible (for convenience),
 * the class now implements abstract {@link Collection} interface.
 *
 * @author  PE
 * @version 0.26 $Rev: 755 $ $Date: 2008-05-18 23:24:28 +0200 (ne, 18 V 2008) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/IntMap.java">IntMap.java</a>
 */
public class IntMap< T > extends AbstractCollection< T > implements Cloneable, Serializable
{
  private static final long serialVersionUID = 1L;

  //==========================================================================
  //  Data:

  /** Special value used in {@link #keyArray} for empty (unused) slots. */
  public final int NULL_KEY       = Integer.MIN_VALUE;

  /** Special value used in {@link #keyArray} for slots with removed keys. */
  protected final int REMOVED_KEY = NULL_KEY + 1;

  /**
   * Array containing key values in slots where {@link #dataArray} contains
   * useful values.
   * <br>{@link #NULL_KEY} means empty (unused) slot.
   * <br>{@link #REMOVED_KEY} replaces each removed key.
   */
  transient int[] keyArray;

  /**
   * Array containing mapped values - {@code null} means .
   * <br>{@link #NULL_KEY} means empty (unused) slot.
   * <br>{@link #REMOVED_KEY} replaces each removed key.
   */
  transient T[] dataArray;

  /** Accelerator (equals to {@code keyArray.length} and {@code dataArray.length}). */
  transient int arraySize;

  /** Number of used mappings. */
  transient int mapUsed;

  /**
   * Number of used kay-array items.
   * Includes regular mappings plus previously removed items.
   * if {@code arrayDirty &gt;= overload}, array should be rehashed.
   *
   * @see     #overload
   * @see     #rehash
   */
  transient int arrayDirty;

  /** Overload bound. */
  protected float maxLoad = 0.6f;

  /**
   * Actual overload factor (maxLoad * arraySize).
   * If it is reached, map needs rehashing (in case of many remove operations)
   * or expansion of data array.
   *
   * @see     #arrayDirty
   */
  transient int overload;

  //public static long findCounter      = 0L;
  //public static long collisionCounter = 0L;

  //==========================================================================
  //  Support:

  /**
   * Rehashes the map arrays and prepares them for the specified capacity.
   *
   * @param  newSize New map capacity.
   */
  @SuppressWarnings("unchecked")
  protected void rehash ( int newSize )
  {
      // backup old data:
    int[] oldKeyArray  = keyArray;
    T[]   oldDataArray = dataArray;
    int   oldSize      = arraySize;

      // re-initialize the arrays:
    arraySize  = Formula.findPrime( (int)Math.ceil( Math.max( mapUsed, newSize ) / maxLoad ) );
    overload   = Math.min( (int)Math.ceil( arraySize * maxLoad ), arraySize - 1 );
    //LogFile.log( "Rehash: used = " + mapUsed + ", new capacity = " + newSize + ", array = " + arraySize + ", overload = " + overload );
    keyArray   = new int[ arraySize ];
    Arrays.fill( keyArray, NULL_KEY );
    dataArray  = (T[]) new Object[ arraySize ];
    mapUsed    =
    arrayDirty = 0;

      // rehash the existing tuples:
    int i, key;
    for ( i = 0; i < oldSize; i++ )
      if ( (key = oldKeyArray[i]) > REMOVED_KEY )
        put( key, oldDataArray[i] );
  }

  /**
   * Finds slot with the given key value.
   * Returns either used slot (with valid mapping) or unused slot (key/value pair
   * is not present in the map).
   *
   * @param  key Key to be looked for.
   * @return Index into the array: where the given key resides or where
   *         it should be inserted. {@code -1} for invalid key.
   */
  protected final int find ( int key )
  {
    if ( arraySize == 0 ||
         key == NULL_KEY ||
         key == REMOVED_KEY ) return -1;

       // 1st try:
    //findCounter++;
    int i = ( (key * 179 + 1009) & 0x7fffffff ) % arraySize;
    int k = keyArray[i];
    if ( k == key || k == NULL_KEY ) return i;

      // collision:
    //LogFile.log( "Collision: size = " + arraySize + ", key = " + key + ", with = " + k +
    //             ", index = " + i + ", used = " + mapUsed + ", dirty = " + arrayDirty );
    int first = -1;                       // first free slot for insert
    if ( k == REMOVED_KEY ) first = i;

      // determine try-step:
    int step = Formula.getSmallPrime( key * 101 );
    if ( step == arraySize )
      step = 1;
      // (step mod arraySize) won't be zero because step & arraySize are different primes

      // subsequent tries:
    do
    {
      //collisionCounter++;
      i += step;
      if ( i >= arraySize ) i %= arraySize;
      k = keyArray[i];
      if ( k == key ) return i;           // found!

      if ( k == NULL_KEY )                // not found
        return( first < 0 ? i : first );

      if ( k == REMOVED_KEY && first < 0 )
        first = i;                        // position for the new key
    }
    while ( true );
      // arrayDirty < arraySize && prime(arraySize) => the loop is not infinite
  }

  //==========================================================================
  //  Constructors:

  /** Dafault parameters (small initial capacity, default load factor = {@code 0.6}). */
  public IntMap ()
  {
    this( 3 );
  }

  /** Explicit map capacity, default load factor ({@code 0.6}). */
  public IntMap ( int capacity )
  {
    this( capacity, 0.6f );
  }

  /** Explicit map capacity and load-factor. */
  public IntMap ( int capacity, float loadFactor )
  {
    maxLoad = loadFactor;
    if ( capacity > 0 )
      rehash( capacity );
  }

  /**
   * The same mappings as the specified {@code Map}.
   * The {@code IntMap} is created with default load factor (0.6)
   * and an initial capacity sufficient to hold the mappings
   * in the specified {@code Map}.
   *
   * @param   m the map whose mappings are to be placed in this map.
   */
  public IntMap ( Map< ? extends Integer, ? extends T > m )
  {
    maxLoad = 0.6f;
    putAll( m );
  }

  //==========================================================================
  //  Access methods:

  /**
   * Returns the number of key-value mappings in this map.
   * {@code null} values are counted in.
   *
   * @return The number of key-value mappings in this map.
   */
  @Override
  public final int size ()
  {
    return mapUsed;
  }

  /**
   * Is the map empty?
   *
   * @return {@code true} if this map contains no key-value mappings.
   */
  @Override
  public final boolean isEmpty ()
  {
    return mapUsed == 0;
  }

  /**
   * "Contains" predicate for keys
   *
   * @param  key Key whose presence in this map is to be tested.
   * @return {@code true} if this map contains a mapping (even {@code null})
   *         for the specified key.
   */
  public final boolean containsKey ( int key )
  {
    int i = find( key );
    return( i >= 0 && keyArray[i] == key );
  }

  /**
   * "Contains" predicate for values.
   * More formally, returns {@code true} if and only if this map
   * contains at least one mapping to a value {@code v} such that
   * {@code (value==null ? v==null : value.equals(v))}.
   *
   * @param  value Value whose presence in this map is to be tested.
   * @return {@code true} if this map maps one or more keys to the
   *         specified value.
   */
  public boolean containsValue ( T value )
  {
    for ( int i = 0; i < arraySize; i++ )
      if ( keyArray[i] > REMOVED_KEY &&
           (value == null ? dataArray[i] == null : value.equals( dataArray[i] ) ) )
        return true;

    return false;
  }

  /**
   * Returns the value to which this map maps the specified key.
   * Cannot distinguish between empty mapping (i.e. no
   * {@code put( key, x )} was called) and mapping with null
   * value ({@code put( key, null )} was called).
   * The {@code containsKey} predicate can be used to
   * distinguish these two cases.
   *
   * @param  key Key whose associated value is to be returned.
   * @return The value to which this map maps the specified key, or
   *         {@code null} if the map contains no mapping for this key.
   * @see #containsKey(int)
   */
  public T get ( int key )
  {
    int i = find( key );
    return( (i >= 0 && keyArray[i] == key) ? dataArray[i] : null );
  }

  /**
   * Associates the specified value with the specified key in this map
   * If the map previously contained a mapping for this key, the old
   * value is replaced by the specified value (a returned back).
   *
   * @param  key Key with which the specified value is to be associated.
   * @param  value Value to be associated with the specified key.
   * @return Previous value associated with specified key, or {@code null}
   *         if there was no mapping for key.  A {@code null} return can
   *         also indicate that the map previously associated {@code null}
   *         with the specified key.
   */
  public T put ( int key, T value )
  {
    if ( arraySize == 0 )                 // empty map
      rehash( 1 );

    int i = find( key );
    if ( i < 0 ) return null;             // operation failed

    int oldK = keyArray[i];
    if ( oldK == key )                    // replace existing mapping
    {
      T oldV       = dataArray[i];
      dataArray[i] = value;
      //LogFile.log( "Replace: key = " + key + ", at = " + i );
      return oldV;
    }

      // new mapping:
    mapUsed++;
    keyArray[i]  = key;
    dataArray[i] = value;
    //LogFile.log( "Insert: key = " + key + ", at = " + i + ", old = " + oldK );

    if ( oldK == NULL_KEY &&              // empty space is consumed
         ++arrayDirty > overload )
      rehash( mapUsed + 1 );

    return null;
  }

  /**
   * Removes the mapping for this key from this map if it is present
   *
   * <p>Returns the value to which the map previously associated the key, or
   * {@code null} if the map contained no mapping for this key.  (A
   * {@code null} return can also indicate that the map previously
   * associated {@code null} with the specified key).
   * The map will not contain a mapping for the specified key once
   * the call returns.
   *
   * @param  key Key whose mapping is to be removed from the map.
   * @return Previous value associated with specified key, or {@code null}
   *         if there was no mapping for key.
   */
  public T remove ( int key )
  {
    int i = find( key );
    if ( i < 0 ) return null;             // operation failed

    if ( keyArray[i] == key )             // real removal
    {
      mapUsed--;
      keyArray[i]  = REMOVED_KEY;
      T oldV       = dataArray[i];
      dataArray[i] = null;
      return oldV;
    }

    return null;
  }

  /**
   * Copies all of the mappings from the specified map to this map.
   *
   * @param  m Mappings to be stored in this map.
   * @throws ClassCastException if the class of a key or value in the
   *         specified map prevents it from being stored in this map.
   */
  public void putAll ( Map< ? extends Integer, ? extends T > m )
  {
    if ( m == null ) return;

    rehash( mapUsed + m.size() );         // assuming that no new items are present in the current map..

    for ( Map.Entry< ? extends Integer, ? extends T > e : m.entrySet() )
      put( e.getKey().intValue(), e.getValue() );
  }

  /**
   * Removes all mappings from this map.
   */
  @Override
  public void clear ()
  {
    arraySize  =
    mapUsed    =
    arrayDirty = 0;
    keyArray   = null;
    dataArray  = null;
  }

  /**
   * Rehashes the data structure to be more compact (and fast - if
   * {@link #remove} operations were performed before).
   */
  public void compact ()
  {
    if ( arrayDirty > mapUsed )
      rehash( mapUsed );
  }

  /**
   * Returns a set view of the keys contained in this map.
   * The set IS NOT backed by the map, it reflects exact IntMap state
   * in the moment of method invocation.
   *
   * @return A set view of the keys contained in this map.
   */
  public Set< Integer > keySet ()
  {
    Set< Integer > set = new HashSet< Integer >( mapUsed );

    if ( mapUsed > 0 )
      for ( int key : keyArray )
        if ( key > REMOVED_KEY )
          set.add( key );

    return set;
  }

  /**
   * Returns a collection view of the values contained in this map.
   * The collection IS NOT backed by the map, it reflects exact IntMap state
   * in the moment of method invocation.
   *
   * @return A collection view of the values contained in this map.
   */
  public Collection< T > values ()
  {
    Collection< T > vals = new ArrayList< T >( mapUsed );

    for ( int i = 0; i < arraySize; i++ )
      if ( keyArray[i] > REMOVED_KEY )
        vals.add( dataArray[i] );

    return vals;
  }

  /**
   * Iterates through all map tuples.
   * Iterator handles only mappings existing in the moment this method is called.
   * Consecutive adds/removes are not handled.
   *
   * @see     IntTuple
   * @see     #iterator
   */
  public final Iterator< IntTuple< T > > tupleIterator ()
  {
    return new TupleIterator();
  }

  /**
   * Iterates through all target values.
   * Iterator handles only mappings existing in the moment this method is called.
   * Consecutive adds/removes are not handled.
   *
   * @see     #tupleIterator
   */
  @Override
  public final Iterator< T > iterator ()
  {
    return new TargetIterator();
  }

  /** Common iterator ancestor. */
  protected class CommonIterator extends IntTuple< T >
  {
    /** Copy of the map keys. */
    protected int[] keys;

    /** Last used position in {@link #keys} array. Should be incremented in {@code next()}. */
    protected int i = -1;

    /** Next key to be used in {@code next()}. {@link #findNext} sets it. */
    protected int nextKey;

    /** Next value to be used in {@code next()}. {@link #findNext} sets it. */
    protected T nextValue;

    /** Is {@link #nextKey} variable valid? */
    protected boolean nextValid = false;

    /** Finds the next map key ({@link #nextKey}) and returns {@code true} if successful. */
    protected boolean findNext ()
    {
      if ( nextValid )
        return( nextKey != NULL_KEY );

      nextValid = true;

      while ( i + 1 < keys.length )
      {
        nextKey = keys[ ++i ];
        int pos = find( nextKey );
        if ( pos >= 0 && keyArray[pos] == nextKey )
        {
          nextValue = (T)dataArray[pos];
          return true;
        }
      }

      nextKey   = NULL_KEY;
      nextValue = null;
      return false;
    }

    /**
     * Backs up map keys for this iterator.
     * No subsequent map changes will be reflected!
     */
    protected CommonIterator ()
    {
      if ( mapUsed == 0 )
      {
        nextKey   = NULL_KEY;
        nextValue = null;
        nextValid = true;
        return;
      }

      keys = new int[ mapUsed ];

      int k = 0;
      for ( int key : keyArray )
        if ( key > REMOVED_KEY )
          keys[ k++ ] = key;
    }

    /** Part of the {@link Iterator} interface. */
    public boolean hasNext ()
    {
      return findNext();
    }

    /** Part of the {@link Iterator} interface. */
    public void remove ()
    {
      IntMap.this.remove( key );
      key   = NULL_KEY;
      value = null;
    }

  }

  /** Internal class handling tuple-iteration. */
  protected class TupleIterator extends CommonIterator implements Iterator< IntTuple< T > >
  {
    /** Returned tuple will be re-used by the next {@link #next} call! */
    public IntTuple< T > next ()
    {
      if ( findNext() )
      {
        key       = nextKey;
        value     = nextValue;
        nextValid = false;
        return this;
      }

      key   = NULL_KEY;
      value = null;
      return this;
    }
  }

  /** Internal class handling target object iteration. */
  protected class TargetIterator extends CommonIterator implements Iterator< T >
  {
    public T next ()
    {
      if ( !findNext() ) return null;

      key       = nextKey;
      nextValid = false;
      return nextValue;
    }
  }

  /**
   * Returns a shallow copy of this {@code IntMap} instance:
   * the values themselves are not cloned.
   *
   * @return A shallow copy of this map.
   */
  public Object clone ()
  {
    IntMap< T > result = new IntMap< T >( mapUsed, maxLoad );

    for ( Iterator< IntTuple< T > > it = tupleIterator();
          it.hasNext(); )
    {
      IntTuple< T > t = it.next();
      result.put( t.key, t.value );
    }

    return result;
  }

  /**
   * Save the state of the {@code IntMap} instance to a stream (i.e.,
   * serialize it).
   *
   * @serialData The size of IntMap is emitted (int), then for each
   *             key-value mapping represented by the IntMap:
   *             {@code key} and {@code value} are emitted (in the order
   *             that they are returned by {@link #tupleIterator}).
   */
  protected void writeObject ( ObjectOutputStream s )
  throws IOException
  {
      // write out maxLoad and any hidden stuff..
    s.defaultWriteObject();

      // write out map size:
    s.writeInt( mapUsed );

      // write out map entries:
    for ( Iterator< IntTuple< T > > it = tupleIterator();
          it.hasNext(); )
    {
      IntTuple< T > t = it.next();
      s.writeInt( t.key );
      s.writeObject( t.value );
    }
  }

  /**
   * Reconstitute the {@code IntMap} instance from a stream
   * (i.e. deserialize it).
   */
  @SuppressWarnings("unchecked")
  protected void readObject ( ObjectInputStream s )
  throws IOException, ClassNotFoundException
  {
    clear();

      // read in maxLoad and any hidden stuff..
    s.defaultReadObject();

      // read map size:
    int size = s.readInt();

    rehash( size );

      // read map entries:
    for ( int i = 0; i++ < size; )
    {
      int key = s.readInt();
      put( key, (T)s.readObject() );
    }
  }

}
