package cz.cuni.jagrlib;

import java.util.*;
import java.io.*;

/**
 * LRU-caching integer map.
 * Stores only limited number of [int,value] pairs.
 *
 * @author  PE
 * @version 0.26 $Rev: 755 $ $Date: 2008-05-18 23:24:28 +0200 (ne, 18 V 2008) $ $Author: pepca $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/LRUCache.java">LRUCache.java</a>
 */
public class LRUCache< T > extends IntMap< T >
{
  //==========================================================================
  //  Support data:

  static protected class LRUItem
  {
    /** Key of this cache item. */
    public int key;

    /** The older cache item. */
    public LRUItem older;

    /** The younger cache item. */
    public LRUItem younger;
  }

  /** Special value used for LRU-list head. */
  protected final int HEAD_KEY = REMOVED_KEY + 1;

  /** Head of the LRU-list. */
  protected LRUItem head;

  /** LRU-list. Double-linked list with head. */
  IntMap< LRUItem > lru;
  {
    head         = new LRUItem();
    head.key     = HEAD_KEY;
    head.older   =
    head.younger = head;

    lru = new IntMap< LRUItem >();
    lru.put( HEAD_KEY, head );
  }

  /** Marks the given key as most recently used. */
  protected final void access ( int key )
  {
    LRUItem l = lru.get( key );
    if ( l == null ) return;

    l.younger.older    = l.older;
    l.older.younger    = l.younger;
    l.younger          = head;
    l.older            = head.older;
    head.older.younger = l;
    head.older         = l;
  }

  /** Cache capacity in items. */
  protected int cacheCapacity = 12;

  /** Linked list of recycled items (older is used). */
  protected LRUItem recycled = null;

  //==========================================================================
  //  Constructors:

  /** Dafault parameters (small initial capacity, default load factor = {@code 0.6}). */
  public LRUCache ()
  {
    this( 3 );
  }

  /** Explicit map capacity, default load factor ({@code 0.6}). */
  public LRUCache ( int capacity )
  {
    this( capacity, 0.6f );
  }

  /** Explicit map capacity and load-factor. */
  public LRUCache ( int capacity, float loadFactor )
  {
    maxLoad = loadFactor;
    if ( capacity > 0 )
    {
      rehash( capacity );
      lru.rehash( capacity );
    }
  }

  /**
   * The same mappings as the specified {@code Map}.
   * The {@code IntMap} is created with default load factor (0.6)
   * and an initial capacity sufficient to hold the mappings
   * in the specified {@code Map}.
   *
   * @param   m the map whose mappings are to be placed in this map.
   */
  public LRUCache ( Map< ? extends Integer, ? extends T > m )
  {
    maxLoad = 0.6f;
    putAll( m );
  }

  /** Sets a new cache capacity. */
  public void setCacheCapacity ( int cacheCap )
  {
    if ( cacheCap == cacheCapacity ||
         cacheCap < 2 ) return;

      // shrink the cache..
    while ( size() > cacheCap )
      remove( head.younger.key );

    cacheCapacity = cacheCap;
  }

  //==========================================================================
  //  Access methods:

  /**
   * Returns the value to which this map maps the specified key.
   * Cannot distinguish between empty mapping (i.e. no
   * {@code put( key, x )} was called) and mapping with null
   * value ({@code put( key, null )} was called).
   * The {@code containsKey} predicate can be used to
   * distinguish these two cases.
   *
   * @param  key Key whose associated value is to be returned.
   * @return The value to which this map maps the specified key, or
   *         {@code null} if the map contains no mapping for this key.
   * @see #containsKey(int)
   */
  @Override
  public T get ( int key )
  {
    final int i = find( key );
    if ( i < 0 || keyArray[i] != key )
      return null;

    access( key );
    return dataArray[i];
  }

  /**
   * Associates the specified value with the specified key in this map
   * If the map previously contained a mapping for this key, the old
   * value is replaced by the specified value (a returned back).
   *
   * @param  key Key with which the specified value is to be associated.
   * @param  value Value to be associated with the specified key.
   * @return Previous value associated with specified key, or {@code null}
   *         if there was no mapping for key.  A {@code null} return can
   *         also indicate that the map previously associated {@code null}
   *         with the specified key.
   */
  @Override
  public T put ( int key, T value )
  {
    if ( get( key ) == null )             // else => replace (the item has already been LRU-ed)
    {
      if ( size() + 1 > cacheCapacity )
        remove( head.younger.key );

        // new LRU-list item:
      LRUItem l = recycled;
      if ( l == null )
        l = new LRUItem();
      else
        recycled = l.older;
      l.key              = key;
      l.younger          = head;
      l.older            = head.older;
      head.older.younger = l;
      head.older         = l;
      lru.put( key, l );
    }

    return super.put( key, value );
  }

  /**
   * Removes the mapping for this key from this map if it is present
   *
   * <p>Returns the value to which the map previously associated the key, or
   * {@code null} if the map contained no mapping for this key.  (A
   * {@code null} return can also indicate that the map previously
   * associated {@code null} with the specified key).
   * The map will not contain a mapping for the specified key once
   * the call returns.
   *
   * @param  key Key whose mapping is to be removed from the map.
   * @return Previous value associated with specified key, or {@code null}
   *         if there was no mapping for key.
   */
  @Override
  public T remove ( int key )
  {
    T old = super.remove( key );
    if ( old != null )
    {
      LRUItem l = lru.get( key );
      l.younger.older    = l.older;
      l.older.younger    = l.younger;
      l.younger          = null;
      l.older            = recycled;
      recycled           = l;
    }

    return old;
  }

  /**
   * Removes all mappings from this map.
   */
  @Override
  public final void clear ()
  {
    super.clear();
    lru.clear();

    head.older   =
    head.younger = head;
    recycled     = null;
  }

  /**
   * Rehashes the data structure to be more compact (and fast - if
   * {@link #remove} operations were performed before).
   */
  @Override
  public void compact ()
  {
    super.compact();
    lru.compact();
  }

  /**
   * Returns a shallow copy of this {@code IntMap} instance:
   * the values themselves are not cloned.
   *
   * @return A shallow copy of this map.
   */
  @Override
  public Object clone ()
  {
    LRUCache< T > result = new LRUCache< T >( mapUsed, maxLoad );
    result.setCacheCapacity( cacheCapacity );

    for ( Iterator< IntTuple< T > > it = tupleIterator();
          it.hasNext(); )
    {
      IntTuple< T > t = it.next();
      result.put( t.key, t.value );
    }

    return result;
  }

}
