package cz.cuni.jagrlib;

import java.io.*;
import java.util.*;
import java.text.*;
import java.lang.management.*;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;

/**
 * Log-file for debuging messages, warning and errors dumps.
 *
 * @version 0.26 $Rev: 748 $ $Date: 2008-05-18 03:29:50 +0200 (ne, 18 V 2008) $ $Author: pepca $
 * @see     <a href="../../../../cz/cuni/jagrlib/LogFile.java">LogFile.java</a>
 */
public class LogFile
{
  //==========================================================================
  //  Data:

  /** Print verbose debug-logs. */
  public static boolean debugging = false;

  /** Trace mode (used externally). */
  public static boolean tracing = false;

  /** Log-file object. */
  protected static PrintStream logFile;

  /** Is log-file still empty? */
  protected static boolean empty = true;

  /** Log-file stream was not opened before? */
  protected static boolean first = true;

  /** Log-file name. */
  public static String logFileName = "./log.txt";

  /** Time of program startup. */
  public static Date startDate = Calendar.getInstance().getTime();

  /** Property set for "run-time probe". Is reloaded before each composition run. */
  protected static Properties probe = new Properties();

  /** Default probe file name (in {@link Properties} format). */
  public static final String DEFAULT_PROBE = "probe.txt";

  public static final String SEPARATOR1 = "-------------------------------------------------";
  public static final String SEPARATOR2 = "+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+";
  public static final String SEPARATOR3 = "=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=";

  //==========================================================================
  //  Logging:

  /**
   * Checks if the log-file was already initialized.
   * If not, writes the log-file prolog.
   *
   * @return  True if log-file is ready to use.
   */
  protected static boolean checkProlog ()
  {
    if ( empty )
    {
      if ( logFile == null )
      {
        try
        {
          logFile = new PrintStream( new FileOutputStream( logFileName, true ), true );
        }
        catch ( Exception ex )
        {
          logFile = null;
          ex.printStackTrace();
          return false;
        }
      }

      if ( first )
      {
        logFile.println( "------------------------------------------------------------" );
        String date = DateFormat.getDateTimeInstance( DateFormat.LONG, DateFormat.FULL ).format( startDate );
        logFile.println( "    0.000: Start - " + date + ", ver: " + Version.niceVersion() );

          // operating system, JVM:
        RuntimeMXBean         rt = ManagementFactory.getRuntimeMXBean();
        OperatingSystemMXBean os = ManagementFactory.getOperatingSystemMXBean();
        logFile.println( "    0.000: Arch: " + os.getArch() + ", " + os.getName() +
                         " (" + os.getVersion() + "), proc: " + os.getAvailableProcessors() );
        logFile.println( "    0.000: " + rt.getVmName() + ", " + rt.getVmVendor() +
                         " (" + rt.getVmVersion() + ')' );
        first = false;
      }

      empty = false;
    }
    return( logFile != null );
  }

  /** Close underlying log-file stream. */
  public static void close ()
  {
    if ( logFile != null )
    {
      logFile.close();
      logFile = null;
    }
    empty = true;
  }

  /**
   * Writes exception dump (with stack trace).
   *
   * @param  ex Throwable holding stack-trace.
   */
  public static void exception ( Throwable ex )
  {
    if ( checkProlog() )
    {
      printTime();
      ex.printStackTrace( logFile );
    }
    else
      ex.printStackTrace();
  }

  /**
   * Writes exception dump (with stack trace).
   *
   * @param  s Additional message.
   * @param  ex Throwable holding stack-trace.
   */
  public static void exception ( String s, Throwable ex )
  {
    log( s );
    if ( checkProlog() )
      ex.printStackTrace( logFile );
    else
      ex.printStackTrace();
  }

  public static void log ( String s )
  {
    if ( checkProlog() )
    {
      printTime();
      logFile.println( s );
    }
  }

  public static void warning ( String s )
  {
    log( "Warning: " + s );
  }

  public static void error ( String s )
  {
    log( "Error: " + s );
  }

  public static void debug ( String s )
  {
    if ( debugging ) log( s );
  }

  public static void stackTrace ( String s )
  {
    try
    {
      throw new Throwable();
    }
    catch ( Throwable ex )
    {
      exception( s, ex );
    }
  }

  /**
   * Prints time relative to program startup (in millisecond-resolution).
   */
  protected static void printTime ()
  {
    if ( logFile == null ) return;

    Date now = Calendar.getInstance().getTime();
    logFile.print( String.format( Locale.US, "%9.3f: ", (1.e-3 * (now.getTime() - startDate.getTime())) ) );
  }

  //==========================================================================
  //  Memory statistics:

  protected static long lastCollectionCount = 0L;
  protected static long lastCollectionTime  = 0L;

  /** Reset cummulative memory statistics (garbage collector statistics). */
  public static void resetMemoryStatistics ()
  {
    List< GarbageCollectorMXBean > gcbs = ManagementFactory.getGarbageCollectorMXBeans();
    long collectionCount = 0L;
    long collectionTime  = 0L;

    for ( GarbageCollectorMXBean gcb : gcbs )
    {
      collectionCount += gcb.getCollectionCount();
      collectionTime  += gcb.getCollectionTime();
    }

    lastCollectionCount = collectionCount;
    lastCollectionTime  = collectionTime;
  }

  /**
   * Prints memory statistics: heap statistics &amp; garbage-collection statistics.
   */
  public static void memoryStatistics ()
  {
    MemoryMXBean mxb = ManagementFactory.getMemoryMXBean();
    MemoryUsage  mu  = mxb.getHeapMemoryUsage();

    List< GarbageCollectorMXBean > gcbs = ManagementFactory.getGarbageCollectorMXBeans();
    long collectionCount = 0L;
    long collectionTime  = 0L;

    for ( GarbageCollectorMXBean gcb : gcbs )
    {
      collectionCount += gcb.getCollectionCount();
      collectionTime  += gcb.getCollectionTime();
    }

    log( String.format( Locale.US, "Memory: init = %.2fM, used = %.2fM, commited = %.2fM, max = %.1fM, gc# = %d, gcTime = %.3fs",
                        Formula.toMega( mu.getInit() ),
                        Formula.toMega( mu.getUsed() ),
                        Formula.toMega( mu.getCommitted() ),
                        Formula.toMega( mu.getMax() ),
                        (collectionCount - lastCollectionCount),
                        (collectionTime  - lastCollectionTime) * 1.e-3 ) );

    lastCollectionCount = collectionCount;
    lastCollectionTime  = collectionTime;
  }

  //==========================================================================
  //  Probe:

  /** Reads the given "probe" into property set, returns <tt>true</tt> if succeeded. */
  public static boolean readProbe ( String fileName )
  {
    try
    {
      if ( fileName == null ||
           fileName.length() == 0 )
        fileName = DEFAULT_PROBE;

      InputStream is = new FileInputStream( fileName );
      probe.load( is );

      debug( "Probe: " + fileName + " read OK" );
    }
    catch ( IOException e )
    {
      probe.clear();
      return false;
    }

    return true;
  }

  /** Sets new value of the general ({@link String}) property. */
  public static Object setProperty ( String key, String value )
  {
    return probe.setProperty( key, value );
  }

  /** Gets value of the general ({@link String}) property. Returns <tt>null</tt> if not found. */
  public static String stringProperty ( String key )
  {
    return probe.getProperty( key );
  }

  /** Gets value of the given integer property. Returns <tt>def</tt> if not found. */
  public static int intProperty ( String key, int def )
  {
    String value = probe.getProperty( key );
    if ( value == null ) return def;

    try
    {
      return Integer.parseInt( value );
    }
    catch ( NumberFormatException e )
    {
    }

    return def;
  }

  /** Gets value of the given floating-point property. Returns <tt>def</tt> if not found. */
  public static double doubleProperty ( String key, double def )
  {
    String value = probe.getProperty( key );
    if ( value == null ) return def;

    try
    {
      return Double.parseDouble( value );
    }
    catch ( NumberFormatException e )
    {
    }

    return def;
  }

  /** Gets value of the given boolean property. Returns <tt>def</tt> if not found. */
  public static boolean booleanProperty ( String key, boolean def )
  {
    String value = probe.getProperty( key );
    return( (value == null) ? def : Boolean.parseBoolean( value ) );
  }

  //==========================================================================
  //  Log images:

  /** Image id that will be used next time. */
  protected static int nextId = 0;

  /** Collection of log-images. */
  protected static IntMap< LogImage > images = new IntMap< LogImage >();

  /** Create a new log-image. Returns its id. */
  public static int createImage ( int width, int height, int mode, String fileName )
  {
    LogImage li      = new LogImage();
    li.format        = new PNGFileFormat();
    li.image         = new RasterImage();
    li.image.init( width, height, mode, 0 );
    JavaBitStream bs = new JavaBitStream();
    if ( fileName.indexOf( '%' ) >= 0 )
      fileName = String.format( fileName, nextId );
    bs.set( JavaBitStream.STREAM_NAME, fileName );
    try
    {
      li.format.connect( Piece.PL_RASTER, li.image, Piece.PL_INPUT );
      li.format.connect( Piece.PL_STREAM, bs,       Piece.PL_INPUT );
    }
    catch ( BadInterfaceException e )
    {
    }

    images.put( nextId, li );
    return nextId++;
  }

  /** Retrieves {@link RasterGraphics} object from log-image is. */
  public static RasterGraphics getImage ( int id )
  {
    LogImage li = images.get( id );
    if ( li == null ) return null;

    return li.image;
  }

  /** Save the given log-image. */
  public static void saveImage ( int id )
  {
    LogImage li = images.get( id );
    if ( li == null ) return;

    try
    {
      li.format.saveFile( (String)null, null );
    }
    catch ( Exception e )
    {
      error( "Error writting log-image #" + id + '!' );
    }
  }

  /** Closes the log-image (invalidates its id). */
  public static void closeImage ( int id )
  {
    images.remove( id );
  }

}

/** Support class - data container for log-image. */
class LogImage
{
  /** File-format instance. */
  PNGFileFormat format;

  /** Memory bitmap. */
  RasterImage image;

}
