package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;

/**
 * Test program 00: Bresenham's algorithm for line segments with GUI preview.
 *
 * @version 0.25 $Rev: 336 $ $Date: 2006-10-29 23:06:54 +0100 (ne, 29 X 2006) $ $Author: pepca $
 * @since   0.03
 * @see     <a href="../../../../cz/cuni/jagrlib/Main00.java">Main00.java</a>
 */
public class Main00
{
  /** The file-name of output PNG image (or <tt>null</tt> for no save). */
  final static String PNGFileName = /* null */ "lines.png";

  /**
   * Working thread for Bresenham lines.
   *
   * @version 0.25 $Rev: 336 $ $Date: 2006-10-29 23:06:54 +0100 (ne, 29 X 2006) $ $Author: pepca $
   * @since   0.13
   */
  private static class Worker extends Thread
  {
    /** Line-drawing algorithm. */
    private LineRender line;

    /** Pen used for filling of line images. */
    private SolidColorPen pen;

    /** Optional image preview object. */
    private GraphicsViewer preview;

    /** Number of lines to be drawn. */
    private int repeat;

    /**
     * Initializing constructor - assigns <tt>LineRender</tt>,
     * <tt>SolidColorPen</tt> and optionally <tt>GraphicsViewer</tt>.
     *
     * @param  _line Line-drawing algorithm.
     * @param  _pen Pen for line-filling.
     * @param  _preview Optional preview object.
     * @param  rep Number of lines to be drawn.
     */
    public Worker ( LineRender _line, SolidColorPen _pen, GraphicsViewer _preview, int rep )
    {
      super( "Line-drawing worker" );
      line    = _line;
      pen     = _pen;
      preview = _preview;
      repeat  = rep;
      try
      {
        setPriority( Thread.MIN_PRIORITY );
      }
      catch ( Exception e )
      {
        LogFile.error( "Cannot set priority of the working thread!" );
      }
    }

    /**
     * Executive routine of the working thread - computes output picture in many phases.
     */
    public void run ()
    {
      if ( line == null || pen == null ) return;
      int[] color = { 255, 255, 255, 255 };

        // draw random lines:
      for ( int i = 0; i++ < repeat; )
      {
          // draw random line:
        line.drawLine( Math.random() * 511.0, Math.random() * 511.0,
                       Math.random() * 511.0, Math.random() * 511.0 );

          // and stroke it eventually:
        if ( (i & 15) == 0 )
        {
            // random color:
          color[0] = 54 + (int)(Math.random() * 200.0);
          color[1] = 54 + (int)(Math.random() * 200.0);
          color[2] = 54 + (int)(Math.random() * 200.0);
          pen.setColor( color );
          pen.stroke();
        }

          // refresh the output window:
        if ( preview != null && (i & 127) == 0 )
        {
          preview.repaintAll();
          preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - " + i + " lines" );
        }
      }

        // finish repainting:
      if ( preview != null )
      {
        preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - FINISHED" );
        preview.repaintAll();
        preview.stopRepaintLoop();
      }
    }
  }

  /** Main routine - draws random lines. */
  public static void main ( String[] args )
  {
      // create Piece objects:
    LineBresenham      line    = new LineBresenham();
    XTransitionList    bitmask = new XTransitionList();
    SolidColorPen      pen     = new SolidColorPen();
    RasterImage        image   = new RasterImage();
      // preview:
    RasterImagePreview preview = new RasterImagePreview();
      // image save:
    PNGFileFormat      bak     = null;
    JavaBitStream      output  = null;
    if ( PNGFileName != null )
    {
      bak    = new PNGFileFormat();
      output = new JavaBitStream();
      output.set( JavaBitStream.STREAM_NAME,PNGFileName );
    }

      // connect them by channels:
    try
    {
      line.connect(    Piece.PL_OUTPUT, bitmask, Piece.PL_INPUT );
      bitmask.connect( Piece.PL_INPUT,  pen,     Piece.PL_BITMASK );
      pen.connect(     Piece.PL_OUTPUT, image,   Piece.PL_INPUT );
      preview.connect( Piece.PL_OUTPUT, image,   Piece.PL_INPUT );
      if ( bak != null )
      {
        bak.connect(   Piece.PL_RASTER, image,   Piece.PL_INPUT );
        bak.connect(   Piece.PL_STREAM, output,  Piece.PL_INPUT );
      }
    }
    catch ( BadInterfaceException e )
    {
      LogFile.error( "Bad Interface in connecting-time!" );
      LogFile.exception( e );
      return;
    }

      // set Pieces' parameters:
    image.init( 512, 512, RasterGraphics.MODE_RGB, 0 ); // init the RasterImage
    bitmask.init( 512, 512 );                           // init the BitMask
    //bitmask.setOperation( BinaryOperation.BIN_OP_XOR );

      // prepare the "worker" thread:
    Worker worker = new Worker( line, pen, preview, 1024 );

      // run the "worker" thread by hand:
    try
    {
      worker.start();
    }
    catch ( IllegalThreadStateException e )
    {
      LogFile.error( "Cannot start the working thread!" );
      LogFile.exception( e );
      return;
    }

      // preview the result instantly:
    Semaphore sem = new Semaphore( 0 );
    preview.addSemaphore( sem );
    System.err.println( preview.repaintLoop() ? "Image finished OK!" : "User break!" );

      // save the result image in PNG format:
    if ( bak != null )
    {
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( Exception e )
      {
        LogFile.error( "Error writting \"" + PNGFileName + "\" image!" );
      }
    }

      // ... and finally quit the program (kill the working thread eventually):
    sem.semWait( 60000L );                // show the result window at most one minute
    System.exit( 0 );
  }

}
