package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;

/**
 * Test program 01: Alpha-channel test patterns with GUI preview.
 *
 * @version 0.24 $Rev: 336 $ $Date: 2006-10-29 23:06:54 +0100 (ne, 29 X 2006) $ $Author: pepca $
 * @since   0.07
 * @see     <a href="../../../../cz/cuni/jagrlib/Main01.java">Main01.java</a>
 */
public class Main01
{
  /** The file-name of output PNG image (or <tt>null</tt> for no save). */
  final static String PNGFileName = /* null */ "alpha.png";

  /**
   * Working thread for alpha-channel test patterns.
   *
   * @version 0.24 $Rev: 336 $ $Date: 2006-10-29 23:06:54 +0100 (ne, 29 X 2006) $ $Author: pepca $
   * @since   0.13
   */
  private static class Worker extends Thread
  {
    /** Test pattern generator. */
    private AlphaGenerator gen;

    /** Brush used for filling of test patterns. */
    private SolidColorBrush brush;

    /** Optional image preview object. */
    private GraphicsViewer preview;

    /**
     * Initializing constructor - assigns <tt>AlphaGenerator</tt>,
     * <tt>SolidColorBrush</tt> and optionally <tt>GraphicsViewer</tt>.
     *
     * @param  _gen Test pattern generator.
     * @param  _brush Brush used in pattern-fill.
     * @param  _preview Optional preview object.
     */
    public Worker ( AlphaGenerator _gen, SolidColorBrush _brush, GraphicsViewer _preview )
    {
      super( "Alpha-blending worker" );
      gen     = _gen;
      brush   = _brush;
      preview = _preview;
      try
      {
        setPriority( Thread.MIN_PRIORITY );
      }
      catch ( Exception e )
      {
        LogFile.error( "Cannot set priority of the working thread!" );
      }
    }

    /**
     * Executive routine of the working thread - computes output picture in 3 phases.
     */
    public void run ()
    {
      if ( gen == null || brush == null ) return;
      int[] color = { 255, 255, 255, 255 };

        // generate the test pattern (the 1st stage):
      gen.fire( AlphaGenerator.PATTERN_WAVES );
        // random color:
      color[0] = 54 + (int)(Math.random() * 200.0);
      color[1] = 54 + (int)(Math.random() * 200.0);
      color[2] = 54 + (int)(Math.random() * 200.0);
      brush.setColor( color );
      try
      {
        brush.fill();
      }
      catch ( JaGrLibException e )
      {}
      if ( preview != null )
      {
        preview.repaintAll();
        preview.set( GraphicsViewer.WINDOW_TITLE, "After the first stage" );
        try
        {
          sleep( 4000L );
        }
        catch ( InterruptedException e )
        {}
      }

        // the 2nd stage:
      gen.set( AlphaGenerator.PERIOD, 0 );
      gen.fire( AlphaGenerator.PATTERN_TRANS_LEFT );
        // random color:
      color[0] = 54 + (int)(Math.random() * 200.0);
      color[1] = 54 + (int)(Math.random() * 200.0);
      color[2] = 54 + (int)(Math.random() * 200.0);
      brush.setColor( color );
      try
      {
        brush.fill();
      }
      catch ( JaGrLibException e )
      {}
      if ( preview != null )
      {
        preview.repaintAll();
        preview.set( GraphicsViewer.WINDOW_TITLE, "After the second stage" );
        try
        {
          sleep( 4000L );
        }
        catch ( InterruptedException e )
        {}
      }

        // the 3rd stage:
      gen.fire( AlphaGenerator.PATTERN_TRANS_DOWN );
        // random color:
      color[0] = 54 + (int)(Math.random() * 200.0);
      color[1] = 54 + (int)(Math.random() * 200.0);
      color[2] = 54 + (int)(Math.random() * 200.0);
      brush.setColor( color );
      try
      {
        brush.fill();
      }
      catch ( JaGrLibException e )
      {}
      if ( preview != null )
      {
        preview.set( GraphicsViewer.WINDOW_TITLE, "FINISHED" );
        preview.repaintAll();
        preview.stopRepaintLoop();
      }
    }
  }

  /** Main routine - draws test patterns using alpha-channel. */
  public static void main ( String[] args )
  {
      // create Piece objects:
    AlphaGenerator     gen       = new AlphaGenerator();
    AlphaMatrix        alphamask = new AlphaMatrix();
    SolidColorBrush    brush     = new SolidColorBrush();
    RasterImage        image     = new RasterImage();
      // preview:
    RasterImagePreview preview   = new RasterImagePreview();
      // image save:
    PNGFileFormat      bak       = null;
    JavaBitStream      output    = null;
    if ( PNGFileName != null )
    {
      bak    = new PNGFileFormat();
      output = new JavaBitStream();
      output.set( JavaBitStream.STREAM_NAME, PNGFileName );
    }

      // connect them by channels:
    try
    {
      gen.connect(       Piece.PL_OUTPUT, alphamask, Piece.PL_INPUT );
      alphamask.connect( Piece.PL_INPUT,  brush,     Piece.PL_ALPHAMASK );
      brush.connect(     Piece.PL_OUTPUT, image,     Piece.PL_INPUT );
      preview.connect(   Piece.PL_OUTPUT, image,     Piece.PL_INPUT );
      if ( bak != null )
      {
        bak.connect(     Piece.PL_RASTER, image,     Piece.PL_INPUT );
        bak.connect(     Piece.PL_STREAM, output,    Piece.PL_INPUT );
      }
    }
    catch ( BadInterfaceException e )
    {
      LogFile.error( "Bad Interface in connecting-time!" );
      LogFile.exception( e );
      return;
    }

      // set Pieces' parameters:
    image.init( 512, 512, RasterGraphics.MODE_RGBA, 0 ); // init the RasterImage
    alphamask.init( 512, 512 );                          // init the BitMask
    gen.set( AlphaGenerator.WIDTH,  512 );
    gen.set( AlphaGenerator.HEIGHT, 512 );

      // prepare the "worker" thread:
    Worker worker = new Worker( gen, brush, preview );

      // run the "worker" thread by hand:
    try
    {
      worker.start();
    }
    catch ( IllegalThreadStateException e )
    {
      LogFile.error( "Cannot start the working thread!" );
      LogFile.exception( e );
      return;
    }

      // preview the result instantly:
    Semaphore sem = new Semaphore( 0 );
    preview.addSemaphore( sem );
    System.err.println( preview.repaintLoop() ? "Image finished OK!" : "User break!" );

      // save the result image in PNG format:
    if ( bak != null )
    {
      try
      {
        bak.saveFile( (String)null, null );  // file-name was specified in constructor
      }
      catch ( Exception e )
      {
        LogFile.error( "Error writting \"" + PNGFileName + "\" image!" );
      }
    }

      // ... and finally quit the program (kill the working thread eventually):
    sem.semWait( 60000L );                // show the result window at most one minute
    System.exit( 0 );
  }

}
