package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;

/**
 * Test program 02: Ray-casting of static CSG scene with GUI preview.
 *
 * @version 0.26 $Rev: 840 $ $Date: 2009-01-10 11:44:24 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.12
 * @see     <a href="../../../../cz/cuni/jagrlib/Main02.java">Main02.java</a>
 */
public class Main02
{
  /** The file-name of output JPEG image (or {@code null} for no save). */
  final static String JPEGFileName = /* null */ "sCasting02.jpg";

  /** Image width in pixels. */
  final static int WIDTH  = 400;

  /** Image height in pixels. */
  final static int HEIGHT = 300;

  /**
   * Working thread for ray-casting.
   *
   * @version 0.25 $Rev: 840 $ $Date: 2009-01-10 11:44:24 +0100 (so, 10 I 2009) $ $Author: Tomas $
   * @since   0.13
   */
  private static class Worker extends Thread
  {
    /** Associated image synthesizer (ray-casting in this case). */
    private ImageSynthesizer synt;

    /** Optional image preview object. */
    private GraphicsViewer preview;

    /**
     * Initializing constructor - assigns {@code ImageSynthesizer} and optionally
     * {@code GraphicsViewer}.
     *
     * @param  _synt Image synthesizer which is capable of computing images.
     * @param  _preview Optional preview object.
     */
    public Worker ( ImageSynthesizer _synt, GraphicsViewer _preview )
    {
      super( "Ray-casting worker" );
      synt    = _synt;
      preview = _preview;
      try
      {
        setPriority( Thread.MIN_PRIORITY );
      }
      catch ( Exception e )
      {
        LogFile.error( "Cannot set priority of the working thread!" );
      }
    }

    /**
     * Executive routine of the working thread - computes output picture using Ray-casting.
     */
    public void run ()
    {
      if ( synt == null ) return;

        // render the 1st part:
      synt.renderRectangle( 0, 0, WIDTH / 2, HEIGHT / 2 );
      if ( preview != null )
      {
        preview.repaintAll();
        preview.set( GraphicsViewer.WINDOW_TITLE, "After the first stage" );
      }

        // render the 2nd part:
      synt.renderRectangle( 0, HEIGHT / 2, WIDTH / 2, HEIGHT );
      if ( preview != null )
      {
        preview.repaintAll();
        preview.set( GraphicsViewer.WINDOW_TITLE, "After the second stage" );
      }

        // finish rendering:
      synt.renderRectangle( WIDTH / 2, 0, WIDTH, HEIGHT );
      if ( preview != null )
      {
        preview.set( GraphicsViewer.WINDOW_TITLE, "FINISHED" );
        preview.repaintAll();
        preview.stopRepaintLoop();
      }
    }
  }

  /** Main routine - draws ray-casted picture. */
  public static void main ( String[] args )
  {
      // 1. CSG scene and its content:
    StaticCSGScene csg = new StaticCSGScene();

      // light sources:
    StaticPointLightSource light = new StaticPointLightSource();
    light.setParams( 0, new double[] { 10.0, -5.0,  4.0 },
                        new double[] { 30.0, 30.0, 30.0 },
                        new double[] {  0.5,  1.0,  0.1 } );
    AmbientLightSource ambient = new AmbientLightSource();
    ambient.setParams( new double[] { 1.0, 1.0, 1.0 } );
    csg.setLights( new LightSource[] { light, ambient } );

      // camera:
    StaticCamera camera = new StaticCamera();
    camera.setParams( new double[] { 0.0, 0.5, 4.0 },
                      new double[] { 0.0, 0.0, -1.0 },
                      new double[] { 0.0, -1.0, 0.0 },
                      60.0 );
    csg.setCamera( camera );

      // root scene node:
    CSGNode root = new CSGNode( BinaryOperation.BIN_OP_OR );
    csg.setRoot( root );

      // four leaf nodes: spheres
    Sphere sph1 = new Sphere();
    CSGLeaf leaf1 = new CSGLeaf( sph1 );
    Sphere sph2 = new Sphere();
    CSGLeaf leaf2 = new CSGLeaf( sph2 );
    Sphere sph3 = new Sphere();
    CSGLeaf leaf3 = new CSGLeaf( sph3 );
    Sphere sph4 = new Sphere();
    CSGLeaf leaf4 = new CSGLeaf( sph4 );
    root.insertChild( leaf1, null, null );
    TrMatrix m = new TrMatrix( 4, 3 );
    TrMatrix mi = new TrMatrix( 4, 3 );
    m.scaleAxes3D( null, new double[] { 1.1, 0.5, 0.8 }, mi );
    m.translate3D( 0.2, 0.7, 0.4, mi );
    root.insertChild( leaf2, m, mi );
    m = new TrMatrix( 4, 3 );
    mi = new TrMatrix( 4, 3 );
    m.scaleAxes3D( null, new double[] { 0.3, 1.2, 0.3 }, mi );
    m.translate3D( -0.7, 0.6, 1.2, mi );
    root.insertChild( leaf3, m, mi );
    m = new TrMatrix( 4, 3 );
    mi = new TrMatrix( 4, 3 );
    m.scaleAxes3D( null, new double[] { 2.0, 0.08, 0.08 }, mi );
    m.translate3D( 0.0, 0.25, 1.46, mi );
    root.insertChild( leaf4, m, mi );

/*
      // scene attributes: I. shading model
    PhongModel phong = new PhongModel();
    root.setAttribute( SceneNode.ATTR_LIGHT_MODEL, phong );
*/
      //                   II. global material
    PhongModel material1 = new PhongModel();
    root.setAttribute( SceneNode.ATTR_BSSRDF, material1  );

      //                   III. special material for the 2nd solid
    PhongModel material2 = new PhongModel();
    material2.set( "Color", new double[] { 0.1, 0.1, 0.8 } );
    material2.set( "Kd",    0.2 );        // mostly specular material
    material2.set( "Ks",    0.8 );
    material2.set( "H",     32 );
    leaf2.setAttribute( SceneNode.ATTR_BSSRDF, material2  );

      //                   IV. special material for the 3rd solid
    PhongModel material3 = new PhongModel();
    material3.set( "Color", new double[] { 1.0, 0.2, 0.1 } );
    material3.set( "Kd",    0.8 );        // mostly diffuse material
    material3.set( "Ks",    0.2 );
    material3.set( "H",     10 );
    leaf3.setAttribute( SceneNode.ATTR_BSSRDF, material3  );

      //                   V. special material for the 4th solid
    PhongModel material4 = new PhongModel();
    material4.set( "Color", new double[] { 0.2, 1.0, 0.1 } );
    material4.set( "Kd",    0.4 );
    material4.set( "Ks",    0.6 );
    material4.set( "H",     16 );
    leaf4.setAttribute( SceneNode.ATTR_BSSRDF, material4  );

      // 2. Ray-casting object (image function):
    RayCasting casting = new RayCasting();

      // 3. output raster image:
    RasterImage image = new RasterImage();

      // 4. image synthesizer:
    SimpleSynthesizer synt = new SimpleSynthesizer();

      // 5. raster image viewer:
    RasterImagePreview preview = new RasterImagePreview();

      // 6. image output filter:
    JFIFFileFormat bak = null;

      // 7. output stream:
    JavaBitStream output = null;
    if ( JPEGFileName != null )
    {
      bak    = new JFIFFileFormat();
      output = new JavaBitStream();
      output.set( JavaBitStream.STREAM_NAME, JPEGFileName );
    }

      // connect them by channels:
    try
    {
      light.connect(   Piece.PL_OUTPUT,       csg,     Piece.PL_INTERSECTION );
      casting.connect( Piece.PL_OUTPUT,       csg,     Piece.PL_INPUT );
      casting.connect( Piece.PL_INTERSECTION, csg,     Piece.PL_INTERSECTION );
      synt.connect(    Piece.PL_FUNCTION,     casting, Piece.PL_INPUT );
      synt.connect(    Piece.PL_OUTPUT,       image,   Piece.PL_INPUT );
      synt.connect(    Piece.PL_TRIGGER,      preview, Piece.PL_TRIGGER );
      preview.connect( Piece.PL_OUTPUT,       image,   Piece.PL_INPUT );
      if ( bak != null )
      {
        bak.connect(   Piece.PL_RASTER,       image,   Piece.PL_INPUT );
        bak.connect(   Piece.PL_STREAM,       output,  Piece.PL_INPUT );
      }
    }
    catch ( BadInterfaceException e )
    {
      LogFile.error( "Bad Interface in connecting-time!" );
      LogFile.exception( e );
      return;
    }

      // set Pieces' parameters:
    synt.setBounds( WIDTH, HEIGHT );
    image.init( WIDTH, HEIGHT, RasterGraphics.MODE_RGB, 0 ); // init the RasterImage

      // prepare the "worker" thread:
    Worker worker = new Worker( synt, preview );

      // run the "worker" thread by hand:
    try
    {
      worker.start();
    }
    catch ( IllegalThreadStateException e )
    {
      LogFile.error( "Cannot start the working thread!" );
      LogFile.exception( e );
      return;
    }

      // timer thread:
    RedrawTimer timer = new RedrawTimer();
    timer.startTimer( preview, 0L );

      // preview the result instantly:
    Semaphore sem = new Semaphore( 0 );
    preview.addSemaphore( sem );
    System.err.println( preview.repaintLoop() ? "Image finished OK!" : "User break!" );
    timer.stopTimer();

      // save the result image in JPEG format:
    if ( bak != null )
    {
      bak.set( JFIFFileFormat.QUALITY, 95.0f );
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( Exception e )
      {
        LogFile.error( "Error writting \"" + JPEGFileName + "\" image!" );
      }
    }

      // ... and finally quit the program (kill the working thread eventually):
    sem.semWait( 60000L );                // show the result window at most one minute
    System.exit( 0 );
  }

}