package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;

/**
 * Test program 05: Ray-casting of static CSG scene with GUI preview (cubic 3D fractal).
 *
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.24
 * @see     <a href="../../../../cz/cuni/jagrlib/Main05.java">Main05.java</a>
 */
public class Main05
{
  /** The file-name of output JPEG image (or {@code null} for no save). */
  final static String JPEGFileName = /* null */ "sCasting05.jpg";

  /** Image width in pixels. */
  final static int WIDTH  = 400;

  /** Image height in pixels. */
  final static int HEIGHT = 300;

  private static final double SC = 0.3333;

  private static final double DE = 1.0 + SC; // 1.0 + SC;

  private static final int DEPTH = 3;

  /**
   * Working thread for ray-casting.
   *
   * @version 0.25 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
   * @since   0.13
   */
  private static class Worker extends Thread
  {
    /** Associated image synthesizer (ray-casting in this case). */
    private ImageSynthesizer synt;

    /** Optional image preview object. */
    private GraphicsViewer preview;

    /**
     * Initializing constructor - assigns {@code ImageSynthesizer} and optionally
     * {@code GraphicsViewer}.
     *
     * @param  _synt Image synthesizer which is capable of computing images.
     * @param  _preview Optional preview object.
     */
    public Worker ( ImageSynthesizer _synt, GraphicsViewer _preview )
    {
      super( "Ray-casting worker" );
      synt    = _synt;
      preview = _preview;
      try
      {
        setPriority( Thread.MIN_PRIORITY );
      }
      catch ( Exception e )
      {
        LogFile.error( "Cannot set priority of the working thread!" );
      }
    }

    /**
     * Executive routine of the working thread - computes output picture using Ray-casting.
     */
    public void run ()
    {
      if ( synt == null ) return;

        // render the 1st part:
      synt.renderRectangle( 0, 0, WIDTH / 2, HEIGHT / 2 );
      if ( preview != null )
      {
        preview.repaintAll();
        preview.set( GraphicsViewer.WINDOW_TITLE, "After the first stage" );
      }

        // render the 2nd part:
      synt.renderRectangle( 0, HEIGHT / 2, WIDTH / 2, HEIGHT );
      if ( preview != null )
      {
        preview.repaintAll();
        preview.set( GraphicsViewer.WINDOW_TITLE, "After the second stage" );
      }

        // finish rendering:
      synt.renderRectangle( WIDTH / 2, 0, WIDTH, HEIGHT );
      if ( preview != null )
      {
        preview.repaintAll();
        preview.set( GraphicsViewer.WINDOW_TITLE, "FINISHED" );
        preview.stopRepaintLoop();
      }
    }
  }

  private static Property mat[];

  private static int cubeCounter = 0;

  private static CSGCommonNode node ( int exclude, int depth )
  {
    if ( depth <= 0 )
    {
      //Solid cu = new Sphere();
      Solid cu = new Cube();
      cubeCounter++;
      CSGLeaf leaf = new CSGLeaf( cu );
      leaf.setAttribute( SceneNode.ATTR_BSSRDF, mat[0] );
      return leaf;
    }
    else
    {
      CSGNode inner = new CSGNode( BinaryOperation.BIN_OP_OR );
      TrMatrix m  = new TrMatrix( 4, 3 );
      TrMatrix mi = new TrMatrix( 4, 3 );
        // big cube:
      //Solid cu = new Sphere();
      Solid cu = new Cube();
      cubeCounter++;
      CSGLeaf leaf = new CSGLeaf( cu );
      inner.insertChild( leaf, null, null );
      leaf.setAttribute( SceneNode.ATTR_BSSRDF, mat[depth % 4] );
        // children:
      double[] sc = { SC, SC, SC };

      if ( exclude != 7 )
      {
          // x++, y++, z++
        CSGCommonNode ch = node( 0, depth - 1 );
        m  = new TrMatrix( 4, 3 );
        mi = new TrMatrix( 4, 3 );
        m.scaleAxes3D( null, sc, mi );
        m.translate3D( DE, DE, DE, mi );
        inner.insertChild( ch, m, mi );
      }

      if ( exclude != 6 )
      {
          // x--, y++, z++
        CSGCommonNode ch = node( 1, depth - 1 );
        m  = new TrMatrix( 4, 3 );
        mi = new TrMatrix( 4, 3 );
        m.scaleAxes3D( null, sc, mi );
        m.translate3D( -DE, DE, DE, mi );
        inner.insertChild( ch, m, mi );
      }

      if ( exclude != 5 )
      {
          // x++, y--, z++
        CSGCommonNode ch = node( 2, depth - 1 );
        m  = new TrMatrix( 4, 3 );
        mi = new TrMatrix( 4, 3 );
        m.scaleAxes3D( null, sc, mi );
        m.translate3D( DE, -DE, DE, mi );
        inner.insertChild( ch, m, mi );
      }

      if ( exclude != 4 )
      {
          // x--, y--, z++
        CSGCommonNode ch = node( 3, depth - 1 );
        m  = new TrMatrix( 4, 3 );
        mi = new TrMatrix( 4, 3 );
        m.scaleAxes3D( null, sc, mi );
        m.translate3D( -DE, -DE, DE, mi );
        inner.insertChild( ch, m, mi );
      }

      if ( exclude != 3 )
      {
          // x++, y++, z--
        CSGCommonNode ch = node( 4, depth - 1 );
        m  = new TrMatrix( 4, 3 );
        mi = new TrMatrix( 4, 3 );
        m.scaleAxes3D( null, sc, mi );
        m.translate3D( DE, DE, -DE, mi );
        inner.insertChild( ch, m, mi );
      }

      if ( exclude != 2 )
      {
          // x--, y++, z--
        CSGCommonNode ch = node( 5, depth - 1 );
        m  = new TrMatrix( 4, 3 );
        mi = new TrMatrix( 4, 3 );
        m.scaleAxes3D( null, sc, mi );
        m.translate3D( -DE, DE, -DE, mi );
        inner.insertChild( ch, m, mi );
      }

      if ( exclude != 1 )
      {
          // x++, y--, z--
        CSGCommonNode ch = node( 6, depth - 1 );
        m  = new TrMatrix( 4, 3 );
        mi = new TrMatrix( 4, 3 );
        m.scaleAxes3D( null, sc, mi );
        m.translate3D( DE, -DE, -DE, mi );
        inner.insertChild( ch, m, mi );
      }

      if ( exclude != 0 )
      {
          // x--, y--, z--
        CSGCommonNode ch = node( 7, depth - 1 );
        m  = new TrMatrix( 4, 3 );
        mi = new TrMatrix( 4, 3 );
        m.scaleAxes3D( null, sc, mi );
        m.translate3D( -DE, -DE, -DE, mi );
        inner.insertChild( ch, m, mi );
      }

      return inner;
    }
  }

  /** Main routine - draws ray-casted picture. */
  public static void main ( String[] args )
  {
      // 1. CSG scene and its content:
    StaticCSGScene csg = new StaticCSGScene();

      // light sources:
    StaticPointLightSource light = new StaticPointLightSource();
    light.setParams( 0, new double[] { 10.0, -5.0,  4.0 },
                        new double[] { 30.0, 30.0, 30.0 },
                        new double[] {  0.5,  1.0,  0.1 } );

    AmbientLightSource ambient = new AmbientLightSource();
    ambient.setParams( new double[] { 1.0, 1.0, 1.0 } );
    csg.setLights( new LightSource[] { light, ambient } );

      // camera:
    StaticCamera camera = new StaticCamera();
    camera.setParams( new double[] {  1.3, -2.0,  8.0 },
                      new double[] { -0.2,  0.3, -1.0 },
                      new double[] {  0.0, -1.0,  0.0 },
                      50.0 );
    csg.setCamera( camera );

      // scene attributes: I. shading model
    PhongModel phong = new PhongModel();

      // II. global material
    mat = new PhongModel[4];
    mat[0] = new PhongModel();
    mat[0].set( "Color", new double[] { 1.0, 0.9, 0.0 } );

      // III. special material for the 2nd solid
    mat[1] = new PhongModel();
    mat[1].set( "Color", new double[] { 0.0, 0.1, 0.7 } );
    mat[1].set( "Kd",    0.2 );           // mostly specular material
    mat[1].set( "Ks",    0.8 );
    mat[1].set( "H",     32 );

      // IV. special material for the 3rd solid
    mat[2] = new PhongModel();
    mat[2].set( "Color", new double[] { 1.0, 0.2, 0.1 } );
    mat[2].set( "Kd",    0.8 );           // mostly diffuse material
    mat[2].set( "Ks",    0.2 );
    mat[2].set( "H",     10 );

      // V. special material for the 4th solid
    mat[3] = new PhongModel();
    mat[3].set( "Color", new double[] { 0.2, 1.0, 0.1 } );
    mat[3].set( "Kd",    0.4 );
    mat[3].set( "Ks",    0.6 );
    mat[3].set( "H",     16 );

      // root scene node:
    CSGCommonNode root = node( -1, DEPTH );
    System.out.println( "Cubes = " + cubeCounter );
    root.setAttribute( SceneNode.ATTR_BSSRDF, phong  );
    csg.setRoot( root );

      // 2. Ray-casting object (image function):
    RayCasting casting = new RayCasting();

      // 3. output raster image:
    RasterImage image = new RasterImage();

      // 4. image synthesizer:
    JitteringSynthesizer synt = new JitteringSynthesizer();

      // 5. raster image viewer:
    RasterImagePreview preview = new RasterImagePreview();

      // 6. image output filter:
    JFIFFileFormat bak = null;

      // 7. output stream:
    JavaBitStream output = null;
    if ( JPEGFileName != null )
    {
      bak    = new JFIFFileFormat();
      output = new JavaBitStream();
      output.set( JavaBitStream.STREAM_NAME, JPEGFileName );
    }

      // connect them by channels:
    try
    {
      light.connect(   Piece.PL_OUTPUT,       csg,     Piece.PL_INTERSECTION );
      casting.connect( Piece.PL_OUTPUT,       csg,     Piece.PL_INPUT );
      casting.connect( Piece.PL_INTERSECTION, csg,     Piece.PL_INTERSECTION );
      synt.connect(    Piece.PL_FUNCTION,     casting, Piece.PL_INPUT );
      synt.connect(    Piece.PL_OUTPUT,       image,   Piece.PL_INPUT );
      synt.connect(    Piece.PL_TRIGGER,      preview, Piece.PL_TRIGGER );
      preview.connect( Piece.PL_OUTPUT,       image,   Piece.PL_INPUT );
      if ( bak != null )
      {
        bak.connect(   Piece.PL_RASTER,       image,   Piece.PL_INPUT );
        bak.connect(   Piece.PL_STREAM,       output,  Piece.PL_INPUT );
      }
    }
    catch ( BadInterfaceException e )
    {
      LogFile.error( "Bad Interface in connecting-time!" );
      LogFile.exception( e );
      return;
    }

      // set Pieces' parameters:
    synt.setBounds( WIDTH, HEIGHT );
    image.init( WIDTH, HEIGHT, RasterGraphics.MODE_RGB, 0 ); // init the RasterImage

      // prepare the "worker" thread:
    Worker worker = new Worker( synt, preview );

      // run the "worker" thread by hand:
    try
    {
      worker.start();
    }
    catch ( IllegalThreadStateException e )
    {
      LogFile.error( "Cannot start the working thread!" );
      LogFile.exception( e );
      return;
    }

      // timer thread:
    RedrawTimer timer = new RedrawTimer();
    timer.startTimer( preview, 0L );

      // preview the result instantly:
    Semaphore sem = new Semaphore( 0 );
    preview.addSemaphore( sem );
    System.err.println( preview.repaintLoop() ? "Image finished OK!" : "User break!" );
    timer.stopTimer();

      // save the result image in JPEG format:
    if ( bak != null )
    {
      bak.set( JFIFFileFormat.QUALITY, 95.0f );
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( Exception e )
      {
        LogFile.error( "Error writting \"" + JPEGFileName + "\" image!" );
      }
    }

      // ... and finally quit the program (kill the working thread eventually):
    sem.semWait( 60000L );                // show the result window at most one minute
    System.exit( 0 );
  }

}