package cz.cuni.jagrlib;

import java.io.IOException;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import cz.cuni.jagrlib.eval.EvalPlugin;

/**
 * Main program MainEval: automatic module evaluation.
 *
 * @version 0.26 $Rev: 439 $ $Date: 2007-05-05 12:29:50 +0200 (so, 05 V 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../cz/cuni/jagrlib/MainEval.java">MainEval.java</a>
 */
public class MainEval
{
  /** Default XML script. */
  protected static final String DEFAULT_SCRIPT = "eval/script.xml";

    // XML tags:
  protected static final String PLUGIN    = "plugin";
  protected static final String INSTANCE  = "instance";
  protected static final String PROCESSED = "processed";

  /** Number of parsed test instances. */
  protected static int parsed;

  /** Number of successfully finished test instances. */
  protected static int number;

  /** Object for result-collecting (processed mode). */
  protected static Object results = null;

  protected static class Handler extends DefaultHandler
  {
    /** Individual test parameters. */
    protected Map< String, String > param;

    /** Global test parameters. */
    protected Map< String, String > env;

    /** Plugin to be used in evaluations. */
    protected EvalPlugin plugin;

    /** Stream to be passed to test instance.. */
    protected PrintStream out;

    private enum ParseState
    {
      /** Outside root 'test' tag. */
      NONE,
      /** Inside root 'test' tag. */
      TEST,
      /** Inside 'plugin' tag. */
      PLUGIN,
      /** Inside global parameter tag. */
      PARAM,
      /** Inside 'instance' tag. */
      INSTANCE,
      /** inside individual parameter tag. */
      INSTANCE_PARAM
    }

    /** Actual parsing state. */
    protected ParseState state;

    /** Actual parameter name (used both for global and lical ones). */
    protected String paramName;

    /** Actual parameter value (accumulated?). */
    protected String paramValue;

    public Handler ( PrintStream out, Map<String,String> env )
    {
      this.out = out;
      param    = new HashMap< String, String >();
      this.env = (env == null) ? new HashMap< String, String >() : env;
      state    = ParseState.NONE;
    }

    @Override
    public void startElement ( String uri, String localName, String qName, Attributes atts )
    throws SAXException
    {
      switch ( state )
      {
        case NONE:
          state = ParseState.TEST;        // needn't check tag name..
          return;

        case TEST:
          if ( PLUGIN.equals(qName) )
          {
            state      = ParseState.PLUGIN;
            paramValue = new String();
            return;
          }
          if ( INSTANCE.equals(qName) )
          {
            state = ParseState.INSTANCE;
            param.clear();
            return;
          }
            // otherwise I'll treat this tag as global (environment) parameter:
          state      = ParseState.PARAM;
          paramName  = qName;
          paramValue = new String();
          return;

        case INSTANCE:
            // the only possibility here is to start an individual parameter:
          state      = ParseState.INSTANCE_PARAM;
          paramName  = qName;
          paramValue = new String();
          return;

        default:
            // PLUGIN, PARAM, INSTANCE_PARAM - error
          throw new SAXException( "Tag <" + qName + "> found in invalid context" );
      }
    }

    @Override
    public void characters ( char[] ch, int start, int length )
    throws SAXException
    {
      String value = new String( ch, start, length );
      switch ( state )
      {
        case PLUGIN:
        case PARAM:
        case INSTANCE_PARAM:
          paramValue = paramValue.concat( value );
          return;

        default:
          // NONE, TEST, INSTANCE - ignore
      }
    }

    @Override
    public void endElement ( String uri, String localName, String qName )
    throws SAXException
    {
      switch ( state )
      {
        case TEST:                        // end of all tests
          state = ParseState.NONE;        // needn't check tag name..
          if ( env.get( PROCESSED ) != null &&
               results != null )
          {
            plugin.interpretResults( env, results, out );
          }
          return;

        case PLUGIN:
        {
          state = ParseState.TEST;
            // paramValue contains plugin class name..
          if ( paramValue.length() == 0 )
          {
            out.println( "Plugin class name is empty!" );
            return;
          }
          String pluginClass = Template.JAGRLIB2 + paramValue;
          try
          {
            plugin = (EvalPlugin)(Class.forName( pluginClass ).newInstance());
            if ( env.get( PROCESSED ) != null )
              results = plugin.createResults();
          }
          catch ( Exception e )
          {
            out.println( "Cannot instantiate the " + pluginClass + " class!" );
          }
          return;
        }

        case PARAM:
            // paramValue contains global parameter value (can be empty)..
          if ( env.get( paramName ) == null )   // re-definition is not allowed!
            env.put( paramName, paramValue );
          state = ParseState.TEST;
          return;

        case INSTANCE_PARAM:
            // paramValue contains individual parameter value (can be empty)..
          param.put( paramName, paramValue );
          state = ParseState.INSTANCE;
          return;

        case INSTANCE:                    // run the given test instance
          parsed++;
          if ( plugin == null )
          {
            out.println( "Invalid plugin class, cannot run the test instance!" );
          }
          else                            // can run the instance..
          {
            String shortName = param.get( EvalPlugin.CLASS_NAME );
            if ( shortName != null )
              LogFile.debug( "Running '" + shortName + '\'' );

            if ( env.get( PROCESSED ) != null &&
                 results != null )
            {
              if ( plugin.runTest( env, param, results ) )
                number++;
            }
            else
              if ( plugin.runTest( env, param, out ) )
                number++;
          }
          state = ParseState.TEST;
          return;

        default:
            // NONE - error
          throw new SAXException( "Tag </" + qName + "> found in invalid context" );
      }
    }
  }

  /** Main routine for Eval. */
  public static void main ( String[] args )
  {
    LogFile.debugging = false;            // load LogFile class to set application start time correctly..

      // XML script:
    String xmlFile = DEFAULT_SCRIPT;

    Map<String,String> env = new HashMap<String,String>();

      // command-line parsing:
    for ( int i = 0; i < args.length; i++ )
    {
      if ( args[i].equalsIgnoreCase( "-l" ) )
      {                                   // debug-logs
        LogFile.debugging = true;
        continue;
      }
      if ( args[i].equalsIgnoreCase( "-p" ) )
      {                                   // processed output
        env.put( PROCESSED, "" );
        continue;
      }
      if ( args[i].startsWith( "-D" ) )
      {                                   // define global (environment) symbol
        int pos = args[i].indexOf( '=' );
        String symbol, value;
        if ( pos < 0 )
        {                                 // -Dsymbol
          symbol = args[i].substring( 2 );
          value  = "";
        }
        else
        {                                 // -Dsymbol=value
          symbol = args[i].substring( 2, pos );
          value  = args[i].substring( pos + 1 );
        }
        env.put( symbol, value );
        continue;
      }
        // otherwise this should be XML script file-name..
      xmlFile = args[i];
    }

    if ( LogFile.debugging )
      LogFile.readProbe( null );

      // create XML parser:
    SAXParser parser;
    try
    {
      parser = SAXParserFactory.newInstance().newSAXParser();
    }
    catch ( ParserConfigurationException e )
    {
      LogFile.exception( "SAX parser configuration error", e );
      return;
    }
    catch ( SAXException e )
    {
      LogFile.exception( e );
      return;
    }

      // parse the script file:
    Handler h = new Handler( System.out, env );
    number = parsed = 0;
    try
    {
      parser.parse( xmlFile, h );
      System.out.println( "Total " + number + " instances finished" );
      if ( parsed > number )
        System.out.println( " " + (parsed - number) + " instances failed!" );
    }
    catch ( SAXException e )
    {
      LogFile.exception( "Error parsing XML script '" + xmlFile + "'!", e );
      return;
    }
    catch ( IOException e )
    {
      LogFile.exception( "Error reading the '" + xmlFile + "' file!", e );
      return;
    }
  }

}
