package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;
import cz.cuni.jagrlib.testing.*;
import cz.cuni.jagrlib.obscure.BrepSolidDirectory;

/**
 * Test program for form-factors.
 *
 * @author  PE, Tomas Svab
 * @version 0.26 $Rev: 749 $ $Date: 2008-05-18 11:06:15 +0200 (ne, 18 V 2008) $ $Author: pepca $
 * @see     <a href="../../../../cz/cuni/jagrlib/MainFormFactors.java">MainFormFactors.java</a>
 */
public class MainFormFactors
{
  /** Computing modes. */
  static enum Mode
  {
    /** Compute form-factors individually (patch-to-patch). */
    ENTRY,
    /** Compute form-factors by rows. */
    ROW,
    /** Compute form-factors by columns. */
    COL,
  };

  /** Actual computing mode. */
  static Mode mode = Mode.ENTRY;

  /** Total number of rays for one source patch. */
  static int rays = 100;

  /** Tesselation level (1 .. 12 patches, 2 .. 48 patches). */
  static int tess = 1;

  public static void main ( String[] args )
  {
      // command-line arguments:
    for ( int argIndex = 0; argIndex < args.length; argIndex++ )

      if ( args[argIndex].equals( "-v" ) )  // verbose logging
        LogFile.debugging = true;

      else
      if ( args[argIndex].equals( "-m" ) )  // computing mode: 0 .. ENTRY, 1 .. ROW, 2 .. COL
      {
        if ( argIndex+1 < args.length )
        {
          final int m = Integer.parseInt( args[++argIndex] );
          final Mode[] values = mode.getDeclaringClass().getEnumConstants();
          mode = values[ Formula.clamp( m, 0, values.length ) ];
        }
      }

      else
      if ( args[argIndex].equals( "-r" ) )  // number of rays
      {
        if ( argIndex+1 < args.length )
          rays = Integer.parseInt( args[++argIndex] );
      }

      else
      if ( args[argIndex].equals( "-t" ) )  // tessellation (>= 1)
      {
        if ( argIndex+1 < args.length )
          tess = Integer.parseInt( args[++argIndex] );
      }

      else
        System.out.println( "Unknown argument '" + args[argIndex] + "' will be ignored!" );

    LogFile.log( "Tessellation: " + tess + ", rays: " + rays + ", mode: " + mode.toString() );
    LogFile.log( LogFile.SEPARATOR1 );

    final MonteCarloFF        ff      = new MonteCarloFF();
    final VEFDS               brep    = new VEFDS();
    final CornellBox          cornell = new CornellBox();
    final BrepSolidDirectory  solid   = new BrepSolidDirectory();
    final UniformGrid         grid    = new UniformGrid();

      // 1. connect them by channels:
    try
    {
      cornell.connect( Piece.PL_OUTPUT, brep,  Piece.PL_INPUT );
      ff.connect(      Piece.PL_DATA,   brep,  Piece.PL_INPUT );
      ff.connect(      Piece.PL_OUTPUT, solid, Piece.PL_INPUT );
      solid.connect(   Piece.PL_OUTPUT, brep,  Piece.PL_INPUT );
      if ( null != grid )
        solid.connect( Piece.PL_DIRECT, grid,  Piece.PL_INPUT );
    }
    catch ( BadInterfaceException e )
    {
      LogFile.error( "Bad interface in connecting-time!" );
      LogFile.exception( e );
      return;
    }

      // 2. setup the modules:
    ff.set( ff.NUMBER_RAYS, rays );
    cornell.set( cornell.NUMBER_ROWS_BOX, tess );
    cornell.set( cornell.NUMBER_COLS_BOX, tess );
    cornell.set( cornell.INTERIOR, 0 );
    cornell.fire( 0 );

    try
    {
      grid.connect( Piece.PL_DATA, brep, Piece.PL_INPUT );
    }
    catch ( BadInterfaceException e )
    {
      LogFile.error( "Bad interface in connecting-time!" );
      LogFile.exception( e );
      return;
    }

    grid.set( UniformGrid.DIM,           3 );
    grid.set( UniformGrid.MAX_GRID_SIZE, 64 );
    grid.buildDirectory( Brep.FACE, 0 );

    BrepSolidDirectory mesh = new BrepSolidDirectory( brep, grid, -1 );
    mesh.set( BrepSolidDirectory.SMOOTH, true );
    mesh.set( BrepSolidDirectory.SHELL,  true );

      // 3. compute F-F:
    int i, j;
    int elements = ff.totalElements();
    double[][] formFactor = new double[ elements ][ elements ];
    switch ( mode )
    {
      case ENTRY:
        for ( i = 0; i < elements; i++ )
          for ( j = 0; j < elements; j++ )
            formFactor[i][j] = ff.formFactor( i, j );
        break;

      case ROW:
        for ( i = 0; i < elements; i++ )
          ff.formFactorsFrom( i, null, formFactor[i] );
        break;

      case COL:
      {
        double[] column = new double[ elements ];
        for ( i = 0; i < elements; i++ )
        {
          ff.formFactorsTo( i, null, column );
          for ( j = 0; j < elements; j++ )
            formFactor[j][i] = column[j];
        }
        break;
      }
    }

      //  4. log results:
    double[] area = new double[ elements ];
    for ( i = 0; i < elements; i++ )
      area[i] = ff.elementArea( i );
    LogFile.log( "element areas: " + Formula.vectorToString( area ) );

    for ( i = 0; i < elements; i++ )
      LogFile.log( "FF[" + String.format( "%3d", i ) + "]: " + Formula.vectorToString( formFactor[i] ) );

      // FF rows should sum to <= 1..
    double sumFF = 0.0;
    for ( i = 0; i < elements; i++ )
    {
      for ( double f : formFactor[i] )
        sumFF += f;

      LogFile.log( String.format( java.util.Locale.US, "Sum( FF[%3d] ): %.6f", i, sumFF ) );
      sumFF = 0.0;
    }
  }

}
