package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;
import cz.cuni.jagrlib.testing.*;

/**
 * Test program for 3D noise functions.
 *
 * @version 0.26 $Rev: 840 $ $Date: 2009-01-10 11:44:24 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/MainNoise.java">MainNoise.java</a>
 */
public class MainNoise
{
  /** The file-name of output JPEG image (or {@code null} for no save). */
  final static String JPEGFileName = "sNoise.jpg";

  /** Image width in pixels. */
  final static int WIDTH  = 800;

  /** Image height in pixels. */
  final static int HEIGHT = 500;

  /** Number of spheres. */
  final static int OBJECTS = 15;

  /**
   * Working thread for ray-tracing.
   *
   * @version 0.25 $Rev: 840 $ $Date: 2009-01-10 11:44:24 +0100 (so, 10 I 2009) $ $Author: Tomas $
   * @since   0.25
   */
  private static class Worker extends Thread
  {
    /** Associated image synthesizer (ray-tracing). */
    private ImageSynthesizer synt;

    /** Optional image preview object. */
    private GraphicsViewer preview;

    /**
     * Initializing constructor - assigns {@code ImageSynthesizer} and optionally
     * {@code GraphicsViewer}.
     *
     * @param  _synt Image synthesizer which is capable to compute images.
     * @param  _preview Optional preview object.
     */
    public Worker ( ImageSynthesizer _synt, GraphicsViewer _preview )
    {
      super( "Ray-tracing worker" );
      synt    = _synt;
      preview = _preview;
      try
      {
        setPriority( Thread.MIN_PRIORITY );
      }
      catch ( Exception e )
      {
        LogFile.error( "Cannot set priority of the working thread!" );
      }
    }

    /**
     * Executive routine of the working thread - computes output picture using Ray-tracing.
     */
    public void run ()
    {
      if ( synt == null ) return;

        // render the 1st part:
      synt.renderRectangle( 0, 0, WIDTH / 2, HEIGHT / 2 );
      if ( preview != null )
      {
        preview.repaintAll();
        preview.set( GraphicsViewer.WINDOW_TITLE, "After the first stage" );
      }

        // render the 2nd part:
      synt.renderRectangle( 0, HEIGHT / 2, WIDTH / 2, HEIGHT );
      if ( preview != null )
      {
        preview.repaintAll();
        preview.set( GraphicsViewer.WINDOW_TITLE, "After the second stage" );
      }

        // finish rendering:
      synt.renderRectangle( WIDTH / 2, 0, WIDTH, HEIGHT );
      if ( preview != null )
      {
        preview.set( GraphicsViewer.WINDOW_TITLE, "FINISHED" );
        preview.repaintAll();
        preview.stopRepaintLoop();
      }
    }
  }

  /** Main routine - draws ray-traced picture. */
  public static void main ( String[] args )
  {
      // 1. CSG scene and its content:
    StaticCSGScene csg = new StaticCSGScene();

      // light sources:
    StaticPointLightSource light = new StaticPointLightSource();
    light.setParams( 0, new double[] { -1.0,  -0.25,  4.0 },
                        new double[] {  3.0,   3.0,   3.0 },
                        new double[] {  0.5,   1.0,   0.1 } );
    AmbientLightSource ambient = new AmbientLightSource();
    ambient.setParams( new double[] { 0.5, 0.5, 0.5 } );
    csg.setLights( new LightSource[] { light, ambient } );

      // camera:
    StaticCamera camera = new StaticCamera();
    camera.setParams( new double[] { 0.2,   0.35,  4.5 },
                      new double[] { 0.0,   0.0,  -1.0 },
                      new double[] { 0.0,  -1.0,   0.0 },
                      60.0 );
    csg.setCamera( camera );

      // root scene node:
    CSGNode root = new CSGNode( BinaryOperation.BIN_OP_OR );
    csg.setRoot( root );

      // scene attributes: I. shading model
    PhongModel phong = new PhongModel();
    root.setAttribute( SceneNode.ATTR_BSSRDF, phong  );

      // leaf nodes: spheres
    Sphere[]      sph  = new Sphere[ OBJECTS ];
    CSGLeaf[]     leaf = new CSGLeaf[ OBJECTS ];
    SpotTexture[] tex  = new SpotTexture[ OBJECTS ];

    for ( int i = 0; i < OBJECTS; i++ )
    {
      sph[i]      = new Sphere();
      leaf[i]     = new CSGLeaf( sph[i] );
      TrMatrix m  = new TrMatrix( 4, 3 );
      TrMatrix mi = new TrMatrix( 4, 3 );
      m.scaleAxes3D( null, new double[] { 0.29, 0.29, 0.29 }, mi );
      m.translate3D( 0.6 * (i%5) - 1.0, 0.6 * (i/5) - 0.25, 1.5, mi );
      root.insertChild( leaf[i], m, mi );

      PhongModel material = new PhongModel();
      material.set( "Color", new double[] { 1.0, 1.0, 0.8 } );
      material.set( "Kd",    0.9 );
      material.set( "Ks",    0.1 );
      material.set( "H",     2 );
      leaf[i].setAttribute( SceneNode.ATTR_BSSRDF, material  );

      tex[i] = new SpotTexture();
      tex[i].set( Texture.COLOR, new double[] { 0.2, 0.05, 0.0 } );

      leaf[i].setAttribute( SceneNode.ATTR_TEXTURE, tex[i] );
    }

      // noise objects:
    Piece[] noise = new Piece[ OBJECTS ];

    /**/
    for ( int i = 0; i < OBJECTS; i++ )
    {
      noise[ i] = new SimpleNoise();
      noise[ i].set( RealFunction.AMPLITUDE, 2.5 );
      noise[ i].set( SimpleNoise.VARIATION, i );
    }
    /**/
    /*
    noise[ 0] = new NoiseVaclavKlecanda();
    noise[ 1] = new NoiseIvanGregor();
    noise[ 1].set( RealFunction.AMPLITUDE, 2.0 );
    noise[ 2] = new NoiseDanielBalas();
    noise[ 3] = new NoiseDusanPsotny();
    noise[ 3].set( RealFunction.AMPLITUDE, 2.0 );
    noise[ 3].set( RealFunction.VARIATION, 9 );
    noise[ 4] = new NoiseJanKadlec();
    noise[ 4].set( RealFunction.AMPLITUDE, 2.0 );

    noise[ 5] = new NoiseLarysaAharkava();
    noise[ 5].set( RealFunction.VARIATION, 9 );
    noise[ 6] = new NoiseLiborOlsak();
    noise[ 6].set( RealFunction.AMPLITUDE, 2.0 );
    noise[ 7] = new NoiseMartinKaldy();
    noise[ 7].set( RealFunction.VARIATION, 9 );
    noise[ 8] = new NoiseMichalBida();
    noise[ 8].set( RealFunction.VARIATION, 9 );
    noise[ 9] = new NoiseMichalKrkavec();
    noise[ 9].set( RealFunction.AMPLITUDE, 2.0 );

    noise[10] = new NoisePavelCeska();
    noise[10].set( RealFunction.VARIATION, 9 );
    noise[11] = new NoisePavelCizek();
    noise[11].set( RealFunction.AMPLITUDE, 2.0 );
    noise[12] = new NoisePeterLibic();
    noise[12].set( RealFunction.AMPLITUDE, 2.0 );
    noise[13] = new NoiseRudolfKadlec();
    noise[14] = new NoiseVojtechMencl();
    noise[14].set( RealFunction.FREQUENCY, 1 );
    noise[14].set( RealFunction.AMPLITUDE, 2.0 );
    noise[14].set( RealFunction.VARIATION, 9 );
    */

      // 2. Ray-tracing object (image function):
    RayTracing tracing = new RayTracing();

      // 3. output raster image:
    RasterImage image = new RasterImage();

      // 4. image synthesizer:
    JitteringSynthesizer synt = new JitteringSynthesizer();
    synt.set( ImageSynthesizer.SUPERSAMPLING, 4 );

      // 5. raster image viewer:
    RasterImagePreview preview = new RasterImagePreview();

      // 6. image output filter:
    JFIFFileFormat bak = null;

      // 7. output stream:
    JavaBitStream output = null;
    if ( JPEGFileName != null )
    {
      bak    = new JFIFFileFormat();
      output = new JavaBitStream();
      output.set( JavaBitStream.STREAM_NAME, JPEGFileName );
    }

      // connect them by channels:
    try
    {
      light.connect(   Piece.PL_OUTPUT,       csg,      Piece.PL_INTERSECTION );
      tracing.connect( Piece.PL_OUTPUT,       csg,      Piece.PL_INPUT );
      tracing.connect( Piece.PL_INTERSECTION, csg,      Piece.PL_INTERSECTION );
      synt.connect(    Piece.PL_FUNCTION,     tracing,  Piece.PL_INPUT );
      synt.connect(    Piece.PL_OUTPUT,       image,    Piece.PL_INPUT );
      synt.connect(    Piece.PL_TRIGGER,      preview,  Piece.PL_TRIGGER );
      preview.connect( Piece.PL_OUTPUT,       image,    Piece.PL_INPUT );

      for ( int i = 0; i < OBJECTS; i++ )
        if ( noise[i] != null )
          tex[i].connect( Piece.PL_OUTPUT,    noise[i], Piece.PL_INPUT );

      if ( bak != null )
      {
        bak.connect(   Piece.PL_RASTER,       image,    Piece.PL_INPUT );
        bak.connect(   Piece.PL_STREAM,       output,   Piece.PL_INPUT );
      }
    }
    catch ( BadInterfaceException e )
    {
      LogFile.error( "Bad Interface in connecting-time!" );
      LogFile.exception( e );
      return;
    }

      // set Pieces' parameters:
    synt.setBounds( WIDTH, HEIGHT );
    image.init( WIDTH, HEIGHT, RasterGraphics.MODE_RGB, 0 ); // init the RasterImage

      // prepare the "worker" thread:
    Worker worker = new Worker( synt, preview );

      // run the "worker" thread by hand:
    try
    {
      worker.start();
    }
    catch ( IllegalThreadStateException e )
    {
      LogFile.error( "Cannot start the working thread!" );
      LogFile.exception( e );
      return;
    }

      // timer thread:
    RedrawTimer timer = new RedrawTimer();
    timer.startTimer( preview, 0L );

      // preview the result instantly:
    Semaphore sem = new Semaphore( 0 );
    preview.addSemaphore( sem );
    System.err.println( preview.repaintLoop() ? "Image finished OK!" : "User break!" );
    timer.stopTimer();

      // save the result image in JPEG format:
    if ( bak != null )
    {
      bak.set( JFIFFileFormat.QUALITY, 95.0f );
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( Exception e )
      {
        LogFile.error( "Error writting \"" + JPEGFileName + "\" image!" );
      }
    }

      // ... and finally quit the program (kill the working thread eventually):
    sem.semWait( 60000L );                // show the result window at most one minute
    System.exit( 0 );
  }

}