package cz.cuni.jagrlib;

import java.util.Locale;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;

/**
 * Test program for SLE solvers.
 *
 * @version 0.26 $Rev: 754 $ $Date: 2008-05-18 18:08:40 +0200 (ne, 18 V 2008) $ $Author: pepca $
 * @see     <a href="../../../../cz/cuni/jagrlib/MainSLE.java">MainSLE.java</a>
 */
public class MainSLE
{
  /** SLE-data module. */
  final static SimpleSLEDataStore data = new SimpleSLEDataStore();

  /** Actual SLE-solver module. */
  static SLESolver solver = null;

  /** SLE-solver method (default: Jacobi). */
  static int method = 0;

  /** Method identifier in readable format. */
  static String readableMethod = "";

  /** Number of methods in the {@link SLERelaxation} module.. */
  final static int NUM_RELAXATION = SLERelaxation.Methods.class.getEnumConstants().length;

  /** Random generator (deterministic SLE data setup). */
  final static RandomJames rnd = new RandomJames();

  /** SLE size. */
  static int n = 12;

  /** Required solution precision. */
  static double epsilon = 0.00001;

  /** Refine-batch size. */
  static int batch = 100;

  /** Overshooting/overrelaxation term. */
  static double omega = 1.0;

  /** Actual accuracy (returned from {@link SLESolver#refine}). */
  static double accuracy;

  /** Elapsed time in milliseconds. */
  static long millis;

  //==========================================================================
  //  Support:

  /** Setup the composition. */
  public static void setupWeb ()
  {
    if ( method >= NUM_RELAXATION + 2 )
      method = 0;

    if ( method < NUM_RELAXATION )
    {
      solver = new SLERelaxation();
      solver.set( SLERelaxation.USED_METHOD, method );
      readableMethod = SLERelaxation.Methods.class.getEnumConstants()[ method ].toString();
    }
    else
    if ( method < NUM_RELAXATION + 1 )
    {
      solver = new SLEProgressive();
      readableMethod = "Progressive";
    }
    else
    {
      solver = new SuperShootGather();
      readableMethod = "SuperShootGather";
    }

    solver.set( SLESolver.OMEGA, omega );
    solver.setEpsilon( epsilon );

    try
    {
      ((Piece)solver).connect( Piece.PL_OUTPUT, data, Piece.PL_INPUT );
    }
    catch ( BadInterfaceException e )
    {
      LogFile.error( "Bad interface in connecting-time!" );
      LogFile.exception( e );
      return;
    }
  }

  /** Initialize SLE data. */
  public static void initSLE ()
  {
    data.init( n, n );
    int i, j;
    for ( i = 0; i < n; i++ )
      for ( j = 0; j < n; j++ )
        data.setCoeff( i, j, (i == j) ? (0.5 + rnd.uniformNumber()) * n : rnd.uniformNumber() - 0.5 );
    for ( i = 0; i < n; i++ )
      data.setRight( i, rnd.uniformNumber() * n );
  }

  /** Log statistics/results. */
  public static void log ()
  {
    LogFile.log( "Method:     " + solver.getClass().getName() + ':' + readableMethod );
    LogFile.log( "SLE size:   " + n );
    LogFile.log( "Epsilon:    " + epsilon );
    LogFile.log( "Omega:      " + omega );
    LogFile.log( "Iterations: " + solver.get( SLESolver.STEPS ) );
    LogFile.log( "Batch size: " + batch );
    LogFile.log( String.format( Locale.US, "Accuracy:   %.9f", accuracy ) );
    LogFile.log( String.format( Locale.US, "Time:       %.2f s", 0.001*millis ) );

    LogFile.debugging = true;
    if ( LogFile.debugging )
    {
      int i, j;

        // Check the results:
      final double[] solution = data.getSolution( null );
      LogFile.log( "Solution: " + Formula.vectorToString( solution ) );
      double[] row = null;
      StringBuffer sb = new StringBuffer();

      for ( i = 0; i < n; i++ )           // check the i-th equation
      {
        row = data.getRow( i, row );
        double res = -data.getRight( i );
        for ( j = 0; j < n; j++ )
          res += row[j] * solution[j];

        sb.append( String.format( Locale.US, "R[%4d] = %9.6f, ", i, res ) );
        if ( (i+1) % 5 == 0 )
        {
          LogFile.log( sb.toString() );
          sb.setLength( 0 );
        }
      }
      if ( sb.length() > 0 )
        LogFile.log( sb.toString() );

        // Sum(E):
      double sum = 0.0;
      for ( i = 0; i < solution.length; i++ )
        sum += data.getRight( i );
      LogFile.log( "Sum(E): " + sum );

        // Sum(B):
      sum = 0.0;
      for ( double s : solution )
        sum += s;
      LogFile.log( "Sum(Solution): " + sum );
    }
  }

  //==========================================================================
  //  Main:

  /** Main routine - initializes SLE and iterates solver on it.. */
  public static void main ( String[] args )
  {
      // command-line arguments:
    for ( int argIndex = 0; argIndex < args.length; argIndex++ )

      if ( args[argIndex].equals( "-v" ) )  // verbose logging
        LogFile.debugging = true;

      else
      if ( args[argIndex].equals( "-m" ) )  // SLE solver method
      {
        if ( argIndex+1 < args.length )
          method = Integer.parseInt( args[++argIndex] );
      }

      else
      if ( args[argIndex].equals( "-n" ) )  // SLE size (number of variables)
      {
        if ( argIndex+1 < args.length )
          n = Integer.parseInt( args[++argIndex] );
      }

      else
      if ( args[argIndex].equals( "-e" ) )  // accuracy (epsilon)
      {
        if ( argIndex+1 < args.length )
          epsilon = Double.parseDouble( args[++argIndex] );
      }

      else
      if ( args[argIndex].equals( "-o" ) )  // omega (over-relaxation factor)
      {
        if ( argIndex+1 < args.length )
          omega = Double.parseDouble( args[++argIndex] );
      }

      else
        System.out.println( "Unknown argument '" + args[argIndex] + "' will be ignored!" );

      // initialization:
    setupWeb();
    initSLE();
    solver.reset();

      // solving:
    accuracy   = Double.POSITIVE_INFINITY;
    millis     = System.currentTimeMillis();
    do
    {
      accuracy    = solver.refine( batch, true );
      System.out.print( "*" );
    }
    while ( accuracy >= epsilon );
    millis = System.currentTimeMillis() - millis;

    log();
  }
}
