package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;
import cz.cuni.jagrlib.testing.*;
import cz.cuni.jagrlib.obscure.*;

/**
 * Test program for sampling-tests: draws three test-images using the given ImageSynthesizer.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/MainSampling.java">MainSampling.java</a>
 */
public class MainSampling
{
  /** The file-name of output PNG images (or <tt>null</tt> for no save). */
  final static String PNGFileName1 = "Checker0x256.png";
  final static String PNGFileName2 = "Checker45x256.png";
  final static String PNGFileName3 = "CosR3x256.png";

  /** Image width in pixels. */
  final static int WIDTH  = 600;

  /** Image height in pixels. */
  final static int HEIGHT = 450;

  private static String windowTitle ( Object obj, String state )
  {
    return( obj.getClass().getName() + state );
  }

  /**
   * Working thread for ray-tracing.
   *
   * @version 0.25 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
   * @since   0.25
   */
  private static class Worker extends Thread
  {
    /** Associated image synthesizer. */
    private ImageSynthesizer synt;

    /** Optional image preview object. */
    private GraphicsViewer preview;

    /**
     * Initializing constructor - assigns <tt>ImageSynthesizer</tt> and optionally
     * <tt>GraphicsViewer</tt>.
     *
     * @param  _synt Image synthesizer which is capable of computing images.
     * @param  _preview Optional preview object.
     */
    public Worker ( ImageSynthesizer _synt, GraphicsViewer _preview )
    {
      super( "Sampling worker" );
      synt    = _synt;
      preview = _preview;
      try
      {
        setPriority( Thread.MIN_PRIORITY );
      }
      catch ( Exception e )
      {
        LogFile.error( "Cannot set priority of the working thread!" );
      }
    }

    /**
     * Executive routine of the working thread - computes output picture using Ray-tracing.
     */
    public void run ()
    {
      if ( synt == null ) return;

        // complete rendering:
      synt.renderRectangle( 0, 0, WIDTH, HEIGHT );
      if ( preview != null )
      {
        preview.set( GraphicsViewer.WINDOW_TITLE, windowTitle( synt, " - FINISHED" ) );
        preview.repaintAll();
        preview.stopRepaintLoop();
      }
    }
  }

  /** Main routine - draws test picture. */
  public static void main ( String[] args )
  {
    Main.commandLineParams( args );

      // 1. Test function object:
    TestImageFunctions func = new TestImageFunctions();

      // 2. output raster image:
    RasterImage image = new RasterImage();

      // 3. image synthesizer:
    //Piece synt = new JitteringSynthesizer();
    Piece synt = new AdaptiveJitteringSynthesizer();
    //Piece synt = new SynthesizerDanielBalas();
    //Piece synt = new SynthesizerIvanGregor();
    //Piece synt = new SynthesizerJakubMichalko();
    //Piece synt = new SynthesizerJanKolomaznik();
    //Piece synt = new SynthesizerMichalKrkavec();
    //Piece synt = new SynthesizerPavelCeska();
    //Piece synt = new SynthesizerPavelCizek();
    //Piece synt = new SynthesizerPavelSzalbot();
    //Piece synt = new SynthesizerPetrKoval1();
    //Piece synt = new SynthesizerPetrKoval2();

    synt.set( ImageSynthesizer.SUPERSAMPLING, 256 );
    synt.set( ImageSynthesizer.GAMMA, 1.0 );
    ImageSynthesizer s = (ImageSynthesizer)synt;

      // 4. raster image viewer:
    RasterImagePreview preview = new RasterImagePreview();

      // 5. image output filter:
    PNGFileFormat bak = null;

      // 6. output stream:
    JavaBitStream output = null;
    if ( PNGFileName1 != null )
    {
      bak    = new PNGFileFormat();
      output = new JavaBitStream();
    }

      // connect them by channels:
    try
    {
      synt.connect(    Piece.PL_FUNCTION, func,    Piece.PL_INPUT );
      synt.connect(    Piece.PL_OUTPUT,   image,   Piece.PL_INPUT );
      synt.connect(    Piece.PL_TRIGGER,  preview, Piece.PL_TRIGGER );
      preview.connect( Piece.PL_OUTPUT,   image,   Piece.PL_INPUT );
      if ( bak != null )
      {
        bak.connect(   Piece.PL_RASTER,   image,   Piece.PL_INPUT );
        bak.connect(   Piece.PL_STREAM,   output,  Piece.PL_INPUT );
      }
    }
    catch ( BadInterfaceException e )
    {
      LogFile.error( "Bad Interface in connecting-time!" );
      LogFile.exception( e );
      return;
    }

    Worker worker;
    RedrawTimer timer = new RedrawTimer();
    Semaphore sem     = new Semaphore( 0 );
    preview.addSemaphore( sem );
    boolean ok;

    //*** Checker 0 deg *****************************************************

    func.set( TestImageFunctions.FUNCTION, 0 );
    func.set( TestImageFunctions.FREQUENCY, 1000.0 );
    output.set( JavaBitStream.STREAM_NAME, PNGFileName1 );

      // set Pieces' parameters:
    s.setBounds( WIDTH, HEIGHT );
    image.init( WIDTH, HEIGHT, RasterGraphics.MODE_RGB, 0 ); // init the RasterImage
    preview.set( GraphicsViewer.WINDOW_TITLE, windowTitle( s, " - Checker 0 deg" ) );

      // prepare the "worker" thread:
    worker = new Worker( s, preview );

      // run the "worker" thread by hand:
    try
    {
      worker.start();
    }
    catch ( IllegalThreadStateException e )
    {
      LogFile.error( "Cannot start the working thread!" );
      LogFile.exception( e );
      return;
    }

      // timer thread:
    timer.startTimer( preview, 0L );

      // preview the result instantly:
    ok = preview.repaintLoop();
    System.err.println( ok ? "Image finished OK!" : "User break!" );
    timer.stopTimer();

      // save the result image in PNG format:
    if ( bak != null )
    {
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( Exception e )
      {
        LogFile.error( "Error writting \"" + PNGFileName1 + "\" image!" );
      }
    }

    sem.semWait( 10000L );                // show the result window at most 10 sec
    preview.destroy();

    //*** Checker 45 deg ****************************************************

    func.set( TestImageFunctions.FUNCTION, 1 );
    output.set( JavaBitStream.STREAM_NAME, PNGFileName2 );

      // set Pieces' parameters:
    s.setBounds( WIDTH, HEIGHT );
    image.init( WIDTH, HEIGHT, RasterGraphics.MODE_RGB, 0 ); // init the RasterImage
    preview.set( GraphicsViewer.WINDOW_TITLE, windowTitle( s, " - Checker 45 deg" ) );

      // prepare the "worker" thread:
    worker = new Worker( s, preview );

      // run the "worker" thread by hand:
    try
    {
      worker.start();
    }
    catch ( IllegalThreadStateException e )
    {
      LogFile.error( "Cannot start the working thread!" );
      LogFile.exception( e );
      return;
    }

      // timer thread:
    timer = new RedrawTimer();
    timer.startTimer( preview, 0L );

      // preview the result instantly:
    ok = preview.repaintLoop();
    System.err.println( ok ? "Image finished OK!" : "User break!" );
    timer.stopTimer();

      // save the result image in PNG format:
    if ( bak != null )
    {
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( Exception e )
      {
        LogFile.error( "Error writting \"" + PNGFileName2 + "\" image!" );
      }
    }

    sem.semWait( 10000L );                // show the result window at most 10 sec
    preview.destroy();

    //*** Cos(r^3) **********************************************************

    func.set( TestImageFunctions.FUNCTION, 3 );
    func.set( TestImageFunctions.FREQUENCY, 10.0 );
    output.set( JavaBitStream.STREAM_NAME, PNGFileName3 );

      // set Pieces' parameters:
    s.setBounds( WIDTH, HEIGHT );
    image.init( WIDTH, HEIGHT, RasterGraphics.MODE_RGB, 0 ); // init the RasterImage
    preview.set( GraphicsViewer.WINDOW_TITLE, windowTitle( s, " - Cos(r^3)" ) );

      // prepare the "worker" thread:
    worker = new Worker( s, preview );

      // run the "worker" thread by hand:
    try
    {
      worker.start();
    }
    catch ( IllegalThreadStateException e )
    {
      LogFile.error( "Cannot start the working thread!" );
      LogFile.exception( e );
      return;
    }

      // timer thread:
    timer = new RedrawTimer();
    timer.startTimer( preview, 0L );

      // preview the result instantly:
    ok = preview.repaintLoop();
    System.err.println( ok ? "Image finished OK!" : "User break!" );
    timer.stopTimer();

      // save the result image in PNG format:
    if ( bak != null )
    {
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( Exception e )
      {
        LogFile.error( "Error writting \"" + PNGFileName3 + "\" image!" );
      }
    }

      // ... and finally quit the program (kill the working thread eventually):
    sem.semWait( 10000L );                // show the result window at most 10 sec

    //***********************************************************************

    System.exit( 0 );
  }

}
