package cz.cuni.jagrlib;

import java.util.*;
import cz.cuni.jagrlib.iface.*;

/**
 * Micro-facet = microscopic piece of solid surface. This data container is used for
 * ray-surface intersections.
 *
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.11
 * @see     <a href="../../../../cz/cuni/jagrlib/MicroFacet.java">MicroFacet.java</a>
 * @see     Solid
 * @see     BSSRDF
 */
public class MicroFacet implements Comparable< MicroFacet >
{
  //==========================================================================
  //  Data:

  /** Are 3D world coordinates ({@link #coord}) valid? */
  public final static int VALID_COORD       = 0x0001;

  /** Are 3D local coordinates ({@link #locCoord}) valid? */
  public final static int VALID_LOC_COORD   = 0x0002;

  /** Are 3D object coordinates ({@link #objCoord}) valid? */
  public final static int VALID_OBJ_COORD   = 0x0004;

  /** Are 2D texture coordinates ({@link #texCoord}) valid? */
  public final static int VALID_TEX_COORD   = 0x0008;

  /** Is normal vector ({@link #normal}) valid? */
  public final static int VALID_NORMAL      = 0x0010;

  /** Are surface tangent vectors ({@link #tu}, {@link #tv}) valid? */
  public final static int VALID_TANGENT     = 0x0020;

  /** Is leaf -&gt; world transform ({@link #toWorld}) valid? */
  public final static int VALID_TO_WORLD    = 0x0040;

  /** Is leaf -&gt; object (component) transform ({@link #toObject}) valid? */
  public final static int VALID_TO_OBJECT   = 0x0080;

  /** Is surface color ({@link #color}) valid? */
  public final static int VALID_COLOR       = 0x0100;

  /** Is BRDF object ({@link #brdf}) valid? */
  public final static int VALID_BSSRDF      = 0x0200;

  /** Are BRDF parameters ({@link #material}) valid? */
//  public final static int VALID_MATERIAL  = 0x0400;

  /** Is list of textures ({@link #textures}) valid? */
  public final static int VALID_TEXTURES    = 0x0800;

  /** Bit-mask declaring validity of individual data items ({@code VALID_*} constants). */
  public int valid;

  /**
   * Parameter of the intersection point (intermediate CSG results are of both signs,
   * definitive results should be positive).
   */
  public double t;

  /** {@code true} if the ray enters solid interior (transition from air to solid material). */
  public boolean enter;

  /** {@code true} if the ray enters solid geometry from outside. */
  public boolean front;

  /** 3D world coordinates of the intersection point. */
  public double[] coord;

  /** 3D local coordinates of the intersection point. */
  public double[] locCoord;

  /** 3D object coordinates of the intersection point. */
  public double[] objCoord;

  /** 2D (texture) coordinates of the intersection point (relative to the surface). */
  public double[] texCoord;

  /** Normal vector in the point of intersection (points outside the object = to the air). */
  public double[] normal;

  /** Surface tangent vector in direction of {@code U}. */
  public double[] tu;

  /** Surface tangent vector in direction of {@code V}. */
  public double[] tv;

  /** Id of intersected face (for {@link Brep} intersections). */
  public int face;

  /** Transformation matrix from local coordinates to world coordinates. */
  public TrMatrix toWorld;

  /** Transformation matrix from local coordinates to object (component) coordinates. */
  public TrMatrix toObject;

  /** Surface color. This array is copy of an original color and might be altered! */
  public double[] color;

  /**
   * Optical model (BRDF).
   *
   * @see    #material
   */
  public BSSRDF bssrdf;

  /** Surface which a ray has intersection with. */
  public IntersectionAttributes geometry;

  /** Temporary data used for additional geometric computation. */
  public Object geometryTmp;

  /** Associated scene node (solid interior). */
  public IntersectionAttributes sceneNode;

  /** List of textures which can be applicated sequentially. */
  public List< Texture > textures;

  /** Epsilon constant used in geometric-equality tests (local-space coordinates). */
  public double epsilon = Geometry.SCENE_EPSILON;

  /** {@link #epsilon} squared. */
  public double epsilon2 = Geometry.SCENE_EPSILON * Geometry.SCENE_EPSILON;

  //==========================================================================
  //  Support:

  /** Sets local epsilon used for close-priximity tests. Can be solid-dependent. */
  public void setEpsilon ( double eps )
  {
    epsilon  = eps;
    epsilon2 = eps * eps;
  }

  /** Sets local epsilons used for close-priximity tests. Can be solid-dependent. */
  public void setEpsilon ( double eps, double epsSquared )
  {
    epsilon  = eps;
    epsilon2 = epsSquared;
  }

  /**
   * Computes normal vector from valid (and world-transformed) tangent vectors.
   */
  protected void normalFromTangent ()
  {
    normal = Geometry.vectorProduct( tu, tv, null );
    Geometry.normalize3D( normal );
    if ( enter != front )
    {
      normal[0] = -normal[0];
      normal[1] = -normal[1];
      normal[2] = -normal[2];
    }
    valid |= VALID_NORMAL;
  }

  /**
   * Computes all requested micro-facet's parameters.
   *
   * @param  attributes Bit-mask of the requested attributes (see {@code VALID_*} constants).
   * @return Attributes which were not computed.
   * @see    Solid
   */
  public int assertAttributes ( int attributes )
  {
    if ( (attributes &= ~valid) == 0 )
      return 0;                           // there is nothing to do..

      // attribute dependencies:
    if ( (attributes & VALID_NORMAL) != 0 )
      attributes |= VALID_TANGENT;

    if ( (attributes & (VALID_TANGENT | VALID_COORD)) != 0 )
      attributes |= VALID_TO_WORLD;

    if ( (attributes & (VALID_COORD | VALID_OBJ_COORD)) != 0 )
      attributes |= VALID_LOC_COORD;

    if ( (attributes & VALID_OBJ_COORD) != 0 )
      attributes |= VALID_TO_OBJECT;

    attributes &= ~valid;

    if ( sceneNode != null )              // will SceneNode help?
    {
      attributes = sceneNode.assertAttributes( this, null, attributes );
      if ( attributes == 0 ) return 0;
    }

    if ( geometry != null )               // will surface geometry help?
    {
      attributes = geometry.assertAttributes( this, geometryTmp, attributes );
      if ( attributes == 0 ) return 0;
    }

      // intersection coordinates (world system):
    if ( (attributes & VALID_COORD) != 0 )
    {
      if ( toWorld == null )
        coord = locCoord;
      else
      {
        coord = new double[ 3 ];
        toWorld.transformPoint3D( locCoord, coord );
      }
      valid |= VALID_COORD;
      if ( (attributes -= VALID_COORD) == 0 ) return 0;
    }

      // intersection coordinates (object system):
    if ( (attributes & VALID_OBJ_COORD) != 0 )
    {
      if ( toObject == null )
        objCoord = locCoord;
      else
      {
        objCoord = new double[ 3 ];
        toObject.transformPoint3D( locCoord, objCoord );
      }
      valid |= VALID_OBJ_COORD;
      if ( (attributes -= VALID_OBJ_COORD) == 0 ) return 0;
    }

      // compute normal vector from tangent vectors:
    if ( (attributes & VALID_NORMAL)  != 0 &&
         (valid      & VALID_TANGENT) != 0 )
    {
      normalFromTangent();
      if ( (attributes &= ~valid) == 0 ) return 0;
    }

    return attributes;                    // I will left this to someone else..
  }

  /**
   * Merge two lists of {@link MicroFacet}s using the given set operation.
   *
   * @param  a Left operand.
   * @param  b Right operand.
   * @param  op Set operation code ({@code BinaryOperation.BIN_*}).
   * @see    BinaryOperation
   */
  public static List< MicroFacet > booleanOp ( List< MicroFacet > a, List< MicroFacet > b, int op )
  {
    if ( op == BinaryOperation.BIN_OP_ZERO ) return null;

    List< MicroFacet > result = new LinkedList< MicroFacet >();

      // boundary conditions for "a":
    ListIterator< MicroFacet > aIt = (a == null) ? null : a.listIterator();
    MicroFacet aFacet = (aIt != null && aIt.hasNext()) ? aIt.next() : null;
    double  at  = (aFacet != null) ? aFacet.t : Double.POSITIVE_INFINITY;
    boolean aIn = (aFacet != null) && !aFacet.enter; // inside the "a" set?

      // boundary conditions for "b":
    ListIterator< MicroFacet > bIt = (b == null) ? null : b.listIterator();
    MicroFacet bFacet = (bIt != null && bIt.hasNext()) ? bIt.next() : null;
    double  bt  = (bFacet != null) ? bFacet.t : Double.POSITIVE_INFINITY;
    boolean bIn = (bFacet != null) && !bFacet.enter; // inside the "b" set?
    boolean resIn = BinaryOperation.boolOp( op, false, false );

      // merge-loop:
    MicroFacet act;
    while ( aFacet != null || bFacet != null ) // process the nearest intersection
    {
      boolean nearestA = (at < bt);
      if ( nearestA )
        aIn = !aIn;
      else
        bIn = !bIn;

      boolean newIn = BinaryOperation.boolOp( op, aIn, bIn );
      if ( newIn != resIn )               // surface intersection
      {
        result.add( act = (nearestA ? aFacet : bFacet) );
        act.enter =
        resIn     = newIn;
      }

      if ( nearestA )                     // advance in "a" list
      {
        aFacet = (aIt != null && aIt.hasNext()) ? aIt.next() : null;
        at = (aFacet != null) ? aFacet.t : Double.POSITIVE_INFINITY;
      }
      else                                // advance in "b" list:
      {
        bFacet = (bIt != null && bIt.hasNext()) ? bIt.next() : null;
        bt = (bFacet != null) ? bFacet.t : Double.POSITIVE_INFINITY;
      }
    }

    return result;
  }

  /**
   * Compares the given {@code t0} against the stored intersection.
   * Uses local-space {@link #epsilon} tolerance and provided direction vector.
   *
   * @return {@code true} if this intersection is nearer than {@code t0}.
   */
  public final boolean near ( double t0, double[] dir )
  {
    if ( t >= t0 ) return false;

    final double size2 = dir[0] * dir[0] +
                         dir[1] * dir[1] +
                         dir[2] * dir[2];
    t0 -= t;
    return( t0 * t0 * size2 > epsilon2 );
  }

  /**
   * Compares the given {@code t0} against the stored intersection.
   * Uses local-space {@link #epsilon} tolerance and provided direction vector squared length.
   *
   * @return {@code true} if this intersection is nearer than {@code t0}.
   */
  public final boolean near ( double t0, double dirLenSquare )
  {
    if ( t >= t0 ) return false;

    t0 -= t;
    return( t0 * t0 * dirLenSquare > epsilon2 );
  }

  /**
   * Compares the given {@code t0} against the stored intersection.
   * Uses local-space {@link #epsilon} tolerance and provided direction vector.
   *
   * @return {@code true} if this intersection is far than {@code t0}.
   */
  public final boolean far ( double t0, double[] dir )
  {
    if ( t <= t0 ) return false;

    final double size2 = dir[0] * dir[0] +
                         dir[1] * dir[1] +
                         dir[2] * dir[2];
    t0 = t - t0;
    return( t0 * t0 * size2 > epsilon2 );
  }

  /**
   * Compares the given {@code t0} against the stored intersection.
   * Uses local-space {@link #epsilon} tolerance and provided direction vector squared length.
   *
   * @return {@code true} if this intersection is far than {@code t0}.
   */
  public final boolean far ( double t0, double dirLenSquare )
  {
    if ( t <= t0 ) return false;

    t0 = t - t0;
    return( t0 * t0 * dirLenSquare > epsilon2 );
  }

  /** Natural ordering: ascending [ t ]. */
  public int compareTo ( MicroFacet mf )
  {
    if ( mf == null )
      throw new NullPointerException();

    if ( this == mf )
      return 0;

    if ( t != mf.t )
      return( t < mf.t ? -1 : 1 );

    return( hashCode() < mf.hashCode() ? -1 : 1 );
  }
}