package cz.cuni.jagrlib;

import java.util.*;
import java.lang.reflect.*;

import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.iface.ProgressListener;

/**
 * Basic piece for buildnig computer graphics algorithms (compositions).
 *
 * <p>Instances of {@link Piece} object act as <b>nodes</b> in graph of execution while
 * instances of {@link Channel} object are connecting them as graph <b>edges</b>.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.01
 * @see     <a href="../../../../cz/cuni/jagrlib/Piece.java">Piece.java</a>
 * @see     Template
 */
public abstract class Piece extends DefaultProperty implements Template, Breakable
{
  //==========================================================================
  //  Data:

  /** External break did occur? */
  protected volatile boolean userBreak = false;

  /** Progress listener object. */
  protected ProgressListener pl = null;

  /**
   * Map of {@link Plug}s (potential {@link Channel}s).
   *
   * @see    Plug
   */
  protected Map< String, Plug > plugs;

  /**
   * Map of actually connected {@link Channel}s (connected {@link Plug}s).
   * <br />Used in composition-execution mode.
   *
   * @see    Channel
   */
  protected Map< String, Channel > channels;

  /** XML-persistent data about this module instance. */
  protected InfoModule info;

  //==========================================================================
  //  Construction:

  /**
   * Default constructor: creates an empty instance (it should be assigned later).
   * <br />Uses template[0] as default (should be changed for multi-template modules?).
   */
  protected Piece ()
  {
    plugs    = new HashMap< String, Plug >();
    channels = new HashMap< String, Channel >();
    info     = new InfoModule( getReg( 0 ) );
    setTemplateDynamic( this, ALL_PLUGS );
  }

  public InfoModule getInfo ()
  {
    return info;
  }

  //==========================================================================
  //  Interface Template:

  /**
   * Connects a new {@link Plug}. General routine.
   *
   * @param  key New plug's key.
   * @param  _mandatory Is that plug mandatory (has to be connected)?
   * @param  _multi Can that plug be used in multi-channels?
   * @param  _cloneable Is that plug "cloneable" (=generic)? If yes,
   *                    ordinal numbers (starting from 0) will be appended to its name.
   * @param  _inputInterface Fully qualified identifier of input (implemented)
   *                         {@code interface}.
   * @param  _outputInterface Fully qualified identifier of output (required)
   *                          {@code interface}.
   */
  public void newPlug ( String key, boolean _mandatory, boolean _multi, boolean _cloneable,
                        String _inputInterface, String _outputInterface )
  {
    Plug plug = new BasicPlug( _mandatory, _multi, _cloneable,
                               _inputInterface, _outputInterface, this, key );
    plugs.put( key, plug );
  }

  /**
   * Connects a new input {@link Plug}. Common-use routine: creates <b>mandatory</b>,
   * <b>non-cloneable</b>, <b>multi</b> plug.
   *
   * @param  key New plug's key.
   * @param  _inputInterface Fully qualified identifier of input (implemented)
   *                         {@code interface}.
   */
  public void newInputPlug ( String key, String _inputInterface )
  {
    newPlug( key, false, true, false, _inputInterface, null );
  }

  /**
   * Connects a new output {@link Plug}. Common-use routine: creates <b>mandatory</b>,
   * <b>non-cloneable</b>, <b>multi</b> plug.
   *
   * @param  key New plug's key.
   * @param  _outputInterface Fully qualified identifier of output (required)
   *                          {@code interface}.
   */
  public void newOutputPlug ( String key, String _outputInterface )
  {
    newPlug( key, true, true, false, null, _outputInterface );
  }

  /**
   * Connects a new optional output {@link Plug}. Common-use routine: creates <b>optional</b>,
   * <b>non-cloneable</b>, <b>multi</b> plug.
   *
   * @param  key New plug's key.
   * @param  _outputInterface Fully qualified identifier of output (required)
   *                          {@code interface}.
   */
  public void newOptOutputPlug ( String key, String _outputInterface )
  {
    newPlug( key, false, true, false, null, _outputInterface );
  }

  /**
   * Sets registration strings for the module.
   */
  public void setRegStrings ( String _name, String _template, String _category, String _description )
  {
    info.reg.name        = _name;
    info.reg.template    = _template;
    info.reg.category    = _category;
    info.reg.description = _description;
  }

  //==========================================================================
  //  Interface Breakable:

  /**
   * Tries to stop the thread/computation..
   * The process should yield itself.
   * For {@link Runnable} implementators: {@link Thread#interrupt} will be called afterwards.
   */
  public void stop ()
  {
    userBreak = true;
  }

  /**
   * Sets progress state listener.
   */
  public void setProgressListener ( ProgressListener pl )
  {
    this.pl = pl;
  }

  //==========================================================================
  //  Plug manipulation:

  /**
   * Find a plug from its name.
   */
  public Plug getPlug ( String key )
  {
    return plugs.get( key );
  }

  /**
   * Finds {@link Plug} with the given parameters.
   *
   * @param  inputInterface Fully qualified input interface name (or {@code null}).
   * @param  outputInterface Fully qualified output interface name (or {@code null}).
   * @param  cloneable Should the {@link Plug} be <b>cloneable</b>?
   * @param  multi Should the {@link Plug} have the <b>multi-channel</b> property?
   * @return The compatible {@link Plug} or {@code null}.
   */
  public Plug findPlug ( String inputInterface, String outputInterface,
                         boolean cloneable, boolean multi )
  {
    for ( Plug plug : plugs.values() )
      if ( plug.isCompatible( inputInterface, outputInterface, cloneable, multi ) )
        return plug;

    return null;
  }

  /**
   * Finds {@link Plug} compatible with the given one. Checks {@link Plug} interfaces only,
   * neither names nor input/output flags are relevant.
   *
   * @param  pl Plug template (interfaces are important here .. names, order, input/output
   *            are irrelevant).
   * @return The compatible {@link Plug} or {@code null}.
   */
  public Plug findPlug ( Plug pl )
  {
    return findPlug( pl.getInputInterfaceName(),
                     pl.getOutputInterfaceName(),
                     pl.isCloneable(),
                     pl.isMulti() );
  }

  /**
   * Compatibility test: can I implement the given <u>template</u>?
   *
   * @param  temp Object which plays a role of an template.
   * @return Returns {@code true} if I'm compatible with the {@code temp} object.
   */
  public boolean isCompatible ( Piece temp )
  {
    HashSet< Plug > used = new HashSet< Plug >(); // already used Plugs of mine
    for ( Plug plug : temp.plugs.values() ) // find one Plug compatible with "temp"
    {
      if ( !plug.isMandatory() ) continue;  // doesn't matter

      Plug myPlug = findPlug( plug );
      if ( myPlug == null ||
           used.contains( myPlug ) ) return false;

      if ( !myPlug.isCloneable() ) used.add( myPlug );
    }

    return true;
  }

  /**
   * Compatibility test: can I implement the given <u>template</u>?
   *
   * @param  temp Object which plays a role of an template.
   * @param  plugMap Mapping from {@code temp}'s plugs into my ones ({@code null}
   *                 for identity).
   * @return Returns {@code true} if I'm compatible with the {@code temp} object.
   */
  public boolean isCompatible ( Piece temp, Map< String, String > plugMap )
  {
    for ( Plug plug : temp.plugs.values() ) // check one Plug compatible with "temp"
    {
      if ( !plug.isMandatory() ) continue;  // doesn't matter

      String key = (plugMap == null) ? plug.getKey() : plugMap.get( plug.getKey() );
      Plug myPlug = getPlug( key );
      if ( myPlug == null ) return false;
    }

    return true;
  }

  /**
   * Returns interface of this {@link Template} object (callable from outside using
   * the given {@link Plug}).
   *
   * @param  _plug The required {@link Plug}.
   * @return Interface object to be called through the given {@link Plug}.
   */
  public Object myInterface ( Plug _plug )
  {
    return this;
  }

  /** Tests whether the given plug is being used (connected). */
  public boolean isConnected ( String key )
  {
    return( channels.get( key ) != null );
  }

  /**
   * Returns executable interface for the given channel.
   *
   * @param  key String key (identical to {@link Plug}'s one).
   * @return Executable interface {@code Object}.
   */
  public Object getInterface ( String key )
  {
    Plug pl = plugs.get( key );
    Channel ch = channels.get( key );
    if ( ch == null ) return null;

    return ch.getInterface( pl );
  }

  /**
   * Returns executable interface for the given channel. Checks for the
   * interface type.
   *
   * @param  key String key (identical to {@link Plug}'s one).
   * @param  type Fully qualified {@code interface} type.
   * @throws BadInterfaceException
   * @return Executable interface {@code Object}.
   */
  public Object getInterface ( String key, String type )
  {
    Plug pl = plugs.get( key );
    Channel ch = channels.get( key );
    if ( ch == null ) return null;
    Object obj = ch.getInterface( pl );
    try
    {
      if ( !Class.forName( type ).isAssignableFrom( obj.getClass() ) )
        throw new JaGrLibException( new BadInterfaceException() );
    }
    catch ( ClassNotFoundException e )
    {
      throw new JaGrLibException( new BadInterfaceException( "Invalid interface name" ) );
    }

    return obj;
  }

  /**
   * Connects the {@link Piece} to the other one through the given {@link Plug}.
   * Checks the {@link Plug} compatibility (throws the {@link BadInterfaceException}).
   *
   * @param  key Local key.
   * @param  other The other {@link Piece}.
   * @param  otherKey The other {@link Piece}'s key.
   * @throws BadInterfaceException in case the {@link Plug}s are not compatible.
   */
  public void connect ( String key, Piece other, String otherKey )
  throws BadInterfaceException
  {
    Plug myPlug = plugs.get( key );
    Channel myChannel = channels.get( key );
    Plug otherPlug = other.plugs.get( otherKey );
    Channel otherChannel = other.channels.get( otherKey );

    if ( myPlug == null || otherPlug == null )
      throw new BadInterfaceException( "Piece.connect: cannot find plugs: (" + className() + ',' + key + ',' + (myPlug != null) +
                                       ") - (" + other.className() + ',' + otherKey + ',' + (otherPlug != null) + ')' );
    if ( !myPlug.isOpposite( otherPlug ) )
      throw new BadInterfaceException();
    int i;

    if ( myChannel != null )

      if ( otherChannel != null )           // both channels already exist => destroy the other's one:
      {
          // 1. other's opposite side:
        Plug[] move = otherChannel.getAnti( otherPlug );
        if ( move != null )                 // anything to move
          for ( i = 0; i < move.length; i++ )
          {
            otherChannel.disconnect( move[i] );
            myChannel.connectPro( move[i], myPlug );
          }
          // 2. other's side:
        move = otherChannel.getPro( otherPlug );
        if ( move != null )                 // anything to move
          for ( i = 0; i < move.length; i++ )
          {
            otherChannel.disconnect( move[i] );
            myChannel.connectAnti(   move[i], myPlug );
          }
      }
      else                                  // my channel already exists => connect "other" to it
        myChannel.connectAnti( otherPlug, myPlug );

    else
    {

      if ( otherChannel != null )           // other's channel already exists => connect myself to it
        otherChannel.connectAnti( myPlug, otherPlug );

      else                                  // no channels exist => create the new one
      {
        otherChannel = new BasicChannel( myPlug.getInputInterfaceName(),
                                         myPlug.getOutputInterfaceName() );
        otherChannel.connectLeft(  myPlug );
        otherChannel.connectRight( otherPlug );
      }

      myChannel = otherChannel;
    }

    myChannel.prepare();
  }

  /**
   * Module initialization - called before any {@link Runnable#run Worker.run()}.
   * Individual phases (starting at {@code 0}) will be called until maximum
   * requested phase number is reached.
   *
   * @param  phase Current phase number.
   * @return Number of phases needed to initialize this module.
   */
  public int init ( int phase )
  {
    return 0;
  }

  //==========================================================================
  //  Module registration:

  /**
   * Writes error message (via {@link LogFile}) including the class's name.
   */
  protected void logError ( String message )
  {
    LogFile.error( className() + ": " + message );
  }

  /**
   * Writes warning message (via {@link LogFile}) including the class's name.
   */
  protected void logWarning ( String message )
  {
    LogFile.warning( className() + ": " + message );
  }

  /** Returns number of registration templates. */
  public int noTemplates ()
  {
    Class< ? > c = getClass();
    try
    {
      Field  f   = c.getField( "reg" );
      Object reg = f.get( null );
      if ( !reg.getClass().isArray() ) return 1;
      RegPiece[] r = (RegPiece[])reg;
      return r.length;
    }
    catch ( Exception e )
    {
      LogFile.exception( e );
      return 0;
    }
  }

  /**
   * Retrieves registration object ({@link RegPiece}) from a statically
   * declared class member ("RegPiece reg" or "RegPiece[] reg").
   *
   * @param  ord Ordinal number of template to use..
   * @see    #noTemplates()
   */
  public RegPiece getReg ( int ord )
  {
    Class< ? > c = getClass();
    try
    {
      Field f = c.getField( "reg" );
      Object reg = f.get( null );
      if ( !reg.getClass().isArray() ) return (RegPiece)reg;
      RegPiece[] r = (RegPiece[])reg;
      if ( ord >= 0 &&
           ord < r.length ) return r[ord];
    }
    catch ( Exception e )
    {
      LogFile.exception( e );
    }
    return null;
  }

  /**
   * Invokes statically declared method "setTemplate(Template,int)" for
   * the given Piece instance.
   */
  public int setTemplateDynamic ( Template t, int ord )
  {
    Class<?> c = getClass();
    try
    {
      Class<?> par[] = { TEMPLATE_TYPE, Integer.TYPE };
      Method m = c.getMethod( "setTemplate", par );
      if ( m == null )
      {
        logError( "setTemplateDynamic() - cannot find setTemplate(Template,int)" );
        return 0;
      }
      if ( !Integer.TYPE.isAssignableFrom( m.getReturnType() ) )
      {
        logError( "setTemplateDynamic() - result type error" );
        return 0;
      }
      Object p[] = { t, ord };
      Integer result = (Integer)m.invoke( null, p );
      return result.intValue();
    }
    catch ( Exception e )
    {
      LogFile.exception( e );
      return 0;
    }
  }

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg = new RegPiece();

  //==========================================================================
  //  Property registration (dummy):

  /**
   * Starts definition of single property.
   *
   * @param  name String identifier (unique inside one module class).
   * @param  type Value type - full Java class identifier.
   * @param  descr Brief textual description (for human eyes only).
   * @param  visual Should be this property represented visually in GUI?
   * @see    #propEnd
   */
  public void propBegin ( String name, String type, String descr, boolean visual )
  {
  }

  /**
   * Default value (non-mandatory).
   */
  public void propDefault ( Object value )
  {
  }

  /**
   * Minimum and maximum allowed values (for comparable types only, non-mandatory).
   */
  public void propBounds ( Object min, Object max )
  {
  }

  /**
   * Sets visual manipulator for GUI editing of the property.
   * For predefined manipulators.
   */
  public void propManipulator ( int manipulatorType )
  {
  }

  /**
   * Sets visual manipulator for GUI editing of the property.
   * For user-defined manipulators.
   *
   * @param  className Class name (either fully-qualified or local in {@code package cz.cuni.jagrlib.gui}).
   */
  public void propManipulator ( String className )
  {
  }

  /**
   * Adds next enum label (for list-box GUI).
   *
   * @param  label Textual label to appear in the listbox.
   * @param  value Associated property value (real-value).
   * @param  descr Optional item's description (for human eyes only).
   */
  public void propEnum ( String label, Object value, String descr )
  {
  }

  /**
   * Finishes definition of the property.
   * Every property-definition segment started by {@link #propBegin} must
   * be properly closed by {@link #propEnd}.
   *
   * @see    #propBegin
   */
  public void propEnd ()
  {
  }

  //==========================================================================
  //  Debugging:

  protected final String className ()
  {
    Class<?> c = getClass();
    return c.getName();
  }

  /**
   * Static instance of Class object for Template.
   */
  public static final Class<?> TEMPLATE_TYPE = getCl( "Template" );

  static private Class<?> getCl ( String className )
  {
    try
    {
      return Class.forName( JAGRLIB2 + className );
    }
    catch ( ClassNotFoundException e )
    {
      return null;
    }
  }

}
