package cz.cuni.jagrlib;

/**
 * <tt>Plug</tt> is virtual one-sided (local) interface between {@link Piece}'s inside
 * and its outside - it defines possibility of (and conditions for) connecting {@link Channel}s
 * to the {@link Piece} object.
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.01
 * @see     <a href="../../../../cz/cuni/jagrlib/Plug.java">Plug.java</a>
 * @see     Piece
 * @see     Channel
 * @see     BasicPlug
 */
public interface Plug
{
  /**
   * Has this <tt>Plug</tt> the <b>input</b> capability (can be called from outside)?
   *
   * @return <tt>true</tt> if this <tt>Plug</tt> can be called from outside.
   * @see    #getInputInterfaceName
   * @see    #getInputInterface
   */
  boolean isInput ();

  /**
   * Has this <tt>Plug</tt> the <b>output</b> capability (can call other interfaces)?
   *
   * @return <tt>true</tt> if this <tt>Plug</tt> can call other interfaces.
   * @see    #getOutputInterfaceName
   */
  boolean isOutput ();

  /**
   * Is this <tt>Plug</tt> mandatory?
   *
   * @return <tt>true</tt> if this <tt>Plug</tt> is <b>mandatory</b> (has to be
   *         connected).
   */
  boolean isMandatory ();

  /**
   * Can this <tt>Plug</tt> be used in <b>multi-channels</b>?
   *
   * @return <tt>true</tt> if this <tt>Plug</tt> can be connected to <b>multi-channel</b>.
   */
  boolean isMulti ();

  /**
   * Can this <tt>Plug</tt> be <b>cloned</b>? For filters with variable number of (input)
   * channels.
   * <br>Sibling channels should have access <b>key</b>s with common prefix -
   * integer numbers starting from <b>0</b> will be appended.
   *
   * @return <tt>true</tt> if this <tt>Plug</tt> can be <b>cloned</b>.
   * @see    #getKey
   */
  boolean isCloneable ();

  /**
   * Returns the fully-qualified name of an internal (implemented) interface - to be called from
   * outside.
   *
   * @return Fully-qualified name of the interface or <tt>null</tt> if it is not used.
   * @see    #isInput
   */
  String getInputInterfaceName ();

  /**
   * Returns the fully-qualified name of an external (required) interface - to be called from
   * this {@link Piece}.
   *
   * @return Fully-qualified name of the interface or <tt>null</tt> if it is not used.
   * @see    #isOutput
   */
  String getOutputInterfaceName ();

  /**
   * Returns the internal (implemented) interface object - to be called from outside.
   *
   * @return The internal (implemented) interface object or <tt>null</tt>.
   */
  Object getInputInterface ();

  /**
   * Gets the associated {@link Template}. For topological purposes only.
   *
   * @return The associated {@link Template} or <tt>null</tt>.
   * @see    #getKey
   * @see    #getPiece
   */
  Template getTemplate ();

  /**
   * Gets the associated {@link Piece}. For topological purposes only.
   *
   * @return The associated {@link Piece} or <tt>null</tt>.
   * @see    #getKey
   * @see    #getTemplate
   */
  Piece getPiece ();

  /**
   * Gets the access <b>key</b> from associated {@link Template} to this <tt>Plug</tt>.
   * For topological purposes only.
   *
   * @return The access <b>key</b> or <tt>null</tt>.
   * @see    #getTemplate
   */
  String getKey ();

  /**
   * Compatibility test: Am I compatible with the given parameters?
   *
   * @param  iInterface Required input interface (or <tt>null</tt>).
   * @param  oInterface Required output interface (or <tt>null</tt>).
   * @param  cl Is the <b>cloneability</b> required?.
   * @param  mul Is the <b>multi-channel</b> property required?.
   * @return <tt>true</tt> if I'm able to implement the given interface.
   */
  boolean isCompatible ( String iInterface, String oInterface, boolean cl, boolean mul );

  /**
   * Compatibility test: Am I compatible with the given <tt>Plug</tt>?
   *
   * @param  pl The checked <tt>Plug</tt>.
   * @return <tt>true</tt> if I'm comptible with the given <tt>Plug</tt>.
   */
  boolean isCompatible ( Plug pl );

  /**
   * Compatibility test: Am I compatible with the opposite <tt>Plug</tt>?
   *
   * @param  opposite The opposite <tt>Plug</tt>.
   * @return <tt>true</tt> if I'm compatible with the opposite <tt>Plug</tt>.
   */
  boolean isOpposite ( Plug opposite );

  /**
   * Connects the {@link Channel} in the associated {@link Piece}.
   *
   * @param  channel The {@link Channel} to be connected to this <tt>Plug</tt>.
   */
  void connectChannel ( Channel channel );

  /**
   * Disconnects the associated channel.
   *
   * @return <tt>true</tt> if there was a {@link Channel} connected.
   */
  boolean disconnectChannel ();

}
