package cz.cuni.jagrlib;

import java.util.*;

/**
 * Random number generator by F. James adopted by Phil Linttell, James F. Hickling & Josef Pelikan.
 *
 * <p>This random number generator originally appeared in "Toward a Universal
 * Random Number Generator" by George Marsaglia and Arif Zaman.
 * Florida State University Report: FSU-SCRI-87-50 (1987)
 *
 * <p>It was later modified by F. James and published in "A Review of Pseudo-random Number Generators"
 *
 * <p>Converted from FORTRAN to C by Phil Linttell, James F. Hickling
 * Management Consultants Ltd, Aug. 14, 1989.
 *
 * <p>THIS IS THE BEST KNOWN RANDOM NUMBER GENERATOR AVAILABLE. (However, a newly discovered technique
 * can yield a period of 10^600. But that is still in the development stage.)
 *
 * <p>It passes ALL of the tests for random number generators and has a period of 2^144, is completely
 * portable (gives bit identical results on all machines with at least 24-bit mantissas in the
 * floating point representation).
 *
 * <p>The algorithm is a combination of a Fibonacci sequence (with lags of 97 and 33, and operation
 * "subtraction plus one, modulo one") and an "arithmetic sequence" (using subtraction).
 *
 * <p>On a Vax 11/780, this random number generator can produce a number in 13 microseconds.
 *
 * @author  PE
 * @version 0.25 $Rev: 800 $ $Date: 2008-09-13 12:51:04 +0200 (so, 13 IX 2008) $ $Author: Tomas $
 * @since   0.24
 * @see     <a href="../../../../cz/cuni/jagrlib/RandomJames.java">RandomJames.java</a>
 */
public class RandomJames
{
  //==========================================================================
  //  Data:

  protected double[] u;

  protected double c, cd, cm;

  protected int i97, j97;

  protected double prepMean, prepVar, norMul, norAdd;

  protected int prepRep;

  protected int[] perm;

  protected int permPtr;

  protected int permSize;

  //==========================================================================
  //  Support:

  /** Pseudo-normal distribution parameter setup. */
  protected void checkNormal ( double mean, double variance, int rep )
  {
    if ( rep != prepRep || variance != prepVar || mean != prepMean )
    {
      prepRep  = rep;
      prepMean = mean;
      prepVar  = variance;
      norMul   = variance * Math.sqrt( 12.0 / rep );
      norAdd   = mean - rep * 0.5 * norMul;
    }
  }

  //==========================================================================
  //  Public members:

  /** Success flag. */
  public boolean ok;

  /** Initializing constructor. */
  public RandomJames ( int ij, int kl )
  {
    u = new double[ 97 ];
    reset( ij, kl );
    prepMean =  0.0;
    prepVar  = -1.0;
    prepRep  =  0;
  }

  /** Default constructor ( ij = 1802, kl = 9373 ). */
  public RandomJames ()
  {
    u = new double[ 97 ];
    reset( 1802, 9373 );
    prepMean =  0.0;
    prepVar  = -1.0;
    prepRep  =  0;
  }

  /**
   * Deterministic restart of a sequence.
   *
   * @param  ij Random seed #1 [0,31328]
   * @param  kl Random seed #2 [0,30081]
   */
  public void reset ( int ij, int kl )
  {
    double s, t;
    int i, j, k, l, m;
    int ii, jj;

    ij = Formula.clamp( ij, 0, 31328 );
    kl = Formula.clamp( kl, 0, 30081 );
    ok = true;

    i = ((ij/177) % 177) + 2;
    j =  (ij      % 177) + 2;
    k = ((kl/169) % 178) + 1;
    l =  (kl      % 169);

    for ( ii = 0; ii < 97; ii++ )
    {
      s = 0.0;
      t = 0.5;
      for ( jj = 0; jj++ < 24; )
      {
        m = ( (i*j % 179) * k ) % 179;
        i = j;
        j = k;
        k = m;
        l = (53*l + 1) % 169;
        if ( l*m % 64 >= 32 ) s += t;
        t *= 0.5;
      }
      u[ii] = s;
    }

    c  =   362436.0 / 16777216.0;
    cd =  7654321.0 / 16777216.0;
    cm = 16777213.0 / 16777216.0;

    i97 = 96;
    j97 = 32;
  }

  /** Undeterministic restart of a sequence (uses system time). */
  public void randomize ()
  {
    Calendar c = Calendar.getInstance();
    c.setTime( new Date() );
    int s = c.get( Calendar.SECOND );
    int m = c.get( Calendar.MINUTE );
    int h = c.get( Calendar.HOUR );
    int d = c.get( Calendar.DAY_OF_YEAR ) - 1;

    double maxs_sig   = 60.0 + 60.0/60.0 + 24.0/60.0/60.0 + 366.0/24.0/60.0/60.0;
    double maxs_insig = 60.0 + 60.0*60.0 + 24.0*60.0*60.0 + 366.0*24.0*60.0*60.0;

    double s_sig      = s + m/60.0 + h/60.0/60.0 + d/24.0/60.0/60.0;
    double s_insig    = s + m*60.0 + h*60.0*60.0 + d*24.0*60.0*60.0;

    int s1 = (int)(s_sig   / maxs_sig   * 31328.0);
    int s2 = (int)(s_insig / maxs_insig * 30081.0);

    reset( s1, s2 );
  }

  /** Validation test of a generator algorithm. */
  public boolean validate ()
  {
    reset( 1802, 9373 );
    if ( !ok ) return false;

    double[] temp = new double[ 1000 ];
    int i;
    for ( i = 0; i++ < 20; )
    {
      uniformNumbers( temp );
      if ( !ok ) return false;
    }

    uniformNumbers( temp, 0, 6 );
    for ( i = 0; i < 6; i++ )
      temp[i] = temp[i] * 4096.0 * 4096.0 + 0.5;

    return( (int)temp[0] ==  6533892 &&
            (int)temp[1] == 14220222 &&
            (int)temp[2] ==  7275067 &&
            (int)temp[3] ==  6172232 &&
            (int)temp[4] ==  8354498 &&
            (int)temp[5] == 10633180 );
  }

  /** Uniform random number generator from [0,1]. */
  public double uniformNumber ()
  {
    if ( !ok ) return 0.0;

    double uni = u[i97] - u[j97];
    if ( uni < 0.0 ) uni += 1.0;
    u[i97] = uni;
    if ( --i97 < 0 ) i97 = 96;
    if ( --j97 < 0 ) j97 = 96;
    if ( (c   -= cd) < 0.0 ) c   += cm;
    if ( (uni -= c)  < 0.0 ) uni += 1.0;

    return uni;
  }

  /** Vector of uniform random numbers (from [0,1]). */
  public void uniformNumbers ( double[] vec )
  {
    uniformNumbers( vec, 0, vec.length );
  }

  /** Vector of uniform random numbers (from [0,1]). */
  public void uniformNumbers ( double[] vec, int from, int number )
  {
    if ( vec == null || from >= vec.length ) return;

    number += from;
    if ( number > vec.length ) number = vec.length;
    while ( from < number )
      vec[from++] = uniformNumber();
  }

  /** Pseudo-normal (Gaussian) distribution generator. */
  public double normalNumber ( double mean, double variance, int rep )
  {
    checkNormal( mean, variance, rep );
    int i;
    double sum = 0.0;
    for ( i = 0; i++ < rep; ) sum += uniformNumber();
    return( sum * norMul + norAdd );
  }

  /** Pseudo-normal (Gaussian) distribution generator. */
  public void normalNumbers ( double mean, double variance, int rep, double[] vec )
  {
    normalNumbers( mean, variance, rep, vec, 0, vec.length );
  }

  /** Pseudo-normal (Gaussian) distribution generator. */
  public void normalNumbers ( double mean, double variance, int rep,
                              double[] vec, int from, int number )
  {
    if ( vec == null || from >= vec.length ) return;

    number += from;
    if ( number > vec.length ) number = vec.length;
    checkNormal( mean, variance, rep );
    int i;
    double sum;
    while ( from < number )
    {
      sum = 0.0;
      for ( i = 0; i++ < rep; ) sum += uniformNumber();
      vec[from++] =  sum * norMul + norAdd;
    }
  }

  /**
   * Random permutation setup.
   *
   * @param  size Permutation size.
   * @return The 1st item ({@code 0} to {@code size - 1}).
   * @see    #permutationNext
   */
  public int permutationFirst ( int size )
  {
    if ( size < 1 )
    {
      permSize = 0;
      return 0;
    }

    permSize = size;
    if ( perm == null || perm.length < size )
      perm = new int[ size ];

    int i;
    for ( i = 0; i < size - 1; i++ )
      perm[i] = i + 1;
    perm[i] = 0;
    permPtr = 0;

    return permutationNext();
  }

  /**
   * Computes next permutation item.
   *
   * @return Next random item or {@code -1} at the end.
   * @see    #permutationFirst
   */
  public int permutationNext ()
  {
    if ( permSize <= 0 )
      return -1;

    int steps = (int)( uniformNumber() * permSize );
    while ( steps-- > 0 )
      permPtr = perm[permPtr];

    int result    = perm[permPtr];
    perm[permPtr] = perm[result];
    permSize--;

    return result;
  }

  /** Generates random integer number from the given range. */
  public final int randomInteger ( int min, int max )
  {
    if ( min >= max ) return min;

    return( min + (int)((max + 1 - min) * uniformNumber()) );
  }

  /** Generates random double number from the given range. */
  public final double randomDouble ( double min, double max )
  {
    if ( min >= max ) return min;

    return( min + (max - min) * uniformNumber() );
  }

  /** Generates random float number from the given range. */
  public final float randomFloat ( float min, float max )
  {
    if ( min >= max ) return min;

    return( min + (float)( (max - min) * uniformNumber() ) );
  }

  /**
   * Generates random point in the given triangle.
   * Works in either dimension (2D, 3D).
   *
   * @param  a Triangle vertex A.
   * @param  b Triangle vertex B.
   * @param  c Triangle vertex C.
   * @param  result Pre-allocated result array (can be {@code null}).
   * @return Random point or {@code null} in case of failure.
   */
  public final double[] randomPointFromTriangle ( double[] a, double[] b, double[] c, double[] result )
  {
    if ( a == null ||
         b == null ||
         c == null ) return null;

    int len = a.length;
    if ( result == null || result.length < len )
      result = new double[ len ];

    double u = uniformNumber();
    double v = uniformNumber();
    if ( u + v > 1.0 )
    {
      u = 1.0 - u;
      v = 1.0 - v;
    }

    for ( int i = 0; i < len; i++ )
      result[i] = a[i] + u * (b[i] - a[i]) + v * (c[i] - a[i]);

    return result;
  }

  /**
   * Uniformly generates random unit vector from the given latitude-defined strip.
   * <br />To generate arbitrary direction, use {@code uniformDirection( -1.0, 1.0, result )},
   * for direction from the upper hemisphere, use {@code uniformDirection( 0.0, 1.0, result )}.
   *
   * @param  minLat Sine of the minimum latitude.
   * @param  maxLat Sine of the maximum latitude.
   * @param  result Pre-allocated result array (can be {@code null}).
   */
  public final double[] uniformDirection ( double minLat, double maxLat, double[] result )
  {
    if ( result == null || result.length < 3 )
      result = new double[ 4 ];

    if ( minLat < -1.0 ) minLat = -1.0;
    if ( maxLat >  1.0 ) maxLat =  1.0;
    final double sinLat    = randomDouble( minLat, maxLat );
    final double cosLat    = Math.sqrt( 1.0 - sinLat * sinLat );
    final double longitude = uniformNumber() * 2.0 * Math.PI;

    result[0] = Math.cos( longitude ) * cosLat;
    result[1] = Math.sin( longitude ) * cosLat;
    result[2] = sinLat;
    if ( result.length > 3 )
      result[3] = 0.0;

    return result;
  }

}