package cz.cuni.jagrlib;

/**
 * Simple static pseudo-random generators.
 * Mostly LCG (Linear Congruential Generators).
 *
 * @version 0.25 $Rev: 288 $ $Date: 2006-06-20 07:54:40 +0200 (út, 20 VI 2006) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../cz/cuni/jagrlib/RandomStatic.java">RandomStatic.java</a>
 */
public class RandomStatic
{
  protected final static long BITS_32  = 0xffffffffL;
  protected final static long BITS_31  = 0x7fffffffL;

  /**
   * LCG pseudo-random generator from "Numeric Recipes in C".
   *
   * @param  v Random seed.
   * @return The next pseudo-random number in the sequence.
   * @see    #numericRecipesMax
   */
  public final static long numericRecipes ( long v )
  {
    return( (v * 1664525L + 1013904223L) & BITS_32 );
  }

  /** <tt>max + 1</tt> for: "Numeric Recipes in C". */
  public final static long numericRecipesMax ()
  {
    return BITS_32 + 1L;
  }

  /**
   * LCG pseudo-random generator: IBM randu.
   * Bad distribution in 3D space!
   *
   * @param  v Random seed.
   * @return The next pseudo-random number in the sequence.
   * @see    #ibmRanduMax
   */
  public final static long ibmRandu ( long v )
  {
    return( (v * 65539L) & BITS_31 );
  }

  /** <tt>max + 1</tt> for: IBM randu. */
  public final static long ibmRanduMax ()
  {
    return BITS_31 + 1L;
  }

  /**
   * LCG pseudo-random generator: minimal standard LCG.
   *
   * Proposed in Stephen K. Park and Keith W. Miller:
   *  Random Number Generators: Good Ones Are Hard To Find,
   *  Communications of the ACM, 31(10):1192-1201, 1988.
   *
   * @param  v Random seed.
   * @return The next pseudo-random number in the sequence.
   * @see    #parkMillerMax
   */
  public final static long parkMiller ( long v )
  {
    return( (v * 16807L) % BITS_31 );
  }

  /** <tt>max + 1</tt> for minimal standard LCG. */
  public final static long parkMillerMax ()
  {
    return BITS_31;
  }

  /**
   * LCG pseudo-random generator from Maple.
   *
   * @param  v Random seed.
   * @return The next pseudo-random number in the sequence.
   * @see    #mapleMax
   */
  public final static long maple ( long v )
  {
    return( (v * 427419669081L) % 999999999989L );
  }

  /** <tt>max + 1</tt> for Maple. */
  public final static long mapleMax ()
  {
    return 999999999989L;
  }

}
