package cz.cuni.jagrlib;

import cz.cuni.jagrlib.iface.*;

/**
 * Timer object capable of periodic redrawing the attached {@link GraphicsViewer} window.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../cz/cuni/jagrlib/RedrawTimer.java">RedrawTimer.java</a>
 */
public class RedrawTimer extends Thread
{
  /** Default redraw-period in milliseconds. */
  public final static long DEFAULT_PERIOD = 20000L;

  /** Preview window or {@code null} if timer is not active. */
  protected GraphicsViewer preview;

  /** Title to be set to the preview window ({@code null} means "no change"). */
  protected String title = null;

  /** "Continue" flag for activity-period of the timer. */
  protected boolean cont;

  /** Actual timer period in milliseconds. */
  protected long period;

  /** Sets thread name for debugging identification. */
  public RedrawTimer ()
  {
    super( "JaGrLib - redraw timer" );
  }

  /**
   * Executive method of the timer.
   */
  @Override
  public void run ()
  {
    while ( cont && period > 0L )
    {
      try
      {
        sleep( period );
      }
      catch ( InterruptedException e )
      {
        interrupted();
      }
      synchronized ( this )
      {
        if ( title != null )
        {
          preview.set( GraphicsViewer.WINDOW_TITLE, title );
          title = null;
        }
      }
      preview.repaintAll();
    }
    preview = null;
  }

  /**
   * Start redraw timer.
   *
   * @param  preview Preview window which will be redrawn perodically.
   * @param  period Redraw period in milliseconds.
   * @see    #stopTimer
   */
  public void startTimer ( GraphicsViewer preview, long period )
  {
    if ( period < 0L )
      period = DEFAULT_PERIOD;

    if ( preview == null ) return;

    if ( this.preview != null ) stopTimer();

    this.preview = preview;
    this.period  = period;
    cont         = true;
    try
    {
      setPriority( Thread.MIN_PRIORITY );
    }
    catch ( Exception e )
    {
      LogFile.exception( "Cannot set priority of timer thread! " + e.getMessage(), e );
    }
    try
    {
      start();
    }
    catch ( IllegalThreadStateException e )
    {
      LogFile.exception( "Cannot start timer thread! " + e.getMessage(), e );
    }
  }

  /**
   * Stop redraw timer.
   */
  public void stopTimer ()
  {
    cont = false;
    if ( preview != null )
      interrupt();
  }

  /** Set window title. */
  public synchronized void setTitle ( String tit )
  {
    title = tit;
  }
}
