package cz.cuni.jagrlib;

import java.util.List;

import cz.cuni.jagrlib.iface.*;

/**
 * Abstract node of (hierarchical) 3D scene representation.
 *
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.11
 * @see     <a href="../../../../cz/cuni/jagrlib/SceneNode.java">SceneNode.java</a>
 * @see     CSGNode
 * @see     CSGLeaf
 */
public interface SceneNode extends IntersectionAttributes
{
  /** Key for color attribute ({@code double[]}). */
  String ATTR_COLOR    = "color";

  /** Key for texture attribute ({@link Texture} or List&lt;{@link Texture}&gt;). */
  String ATTR_TEXTURE  = "texture";

  /** Key for light-model attribute ({@link BSSRDF}). */
  String ATTR_BSSRDF   = "BSSRDF";

  /**
   * Returns the given attribute (defined in this {@link SceneNode} or inherited from
   * some parent node).
   * <p>Returned object can be an <b>original</b> data object and must not be altered
   * outside this {@link SceneNode} instance.
   *
   * @param  key Key-string.
   * @return Attribute object or {@code null}. Must not be altered!
   * @see    #getAttributes
   * @see    #setAttribute
   * @see    #selfAttribute
   */
  Object getAttribute ( String key );

  /**
   * Is the given attribute defined in the {@link SceneNode} itself or is it inherited?
   *
   * @param  key Key-string.
   * @return {@code true} if the node defines the given attribute.
   * @see    #getAttribute
   * @see    #setAttribute
   * @see    #getAttributes
   */
  boolean selfAttribute ( String key );

  /**
   * Defines the given attribute (key-value pair) in the {@link SceneNode}.
   *
   * @param  key Key-string.
   * @param  value Attribute value (arbitrary type).
   * @return The old attribute value (defined in the {@link SceneNode} itself&excl;) or
   *         {@code null}.
   * @see    #getAttribute
   * @see    #selfAttribute
   * @see    #getAttributes
   */
  Object setAttribute ( String key, Object value );

  /**
   * Returns the whole inherence-sequence for the given attribute.
   *
   * @param  key Key-string.
   * @return List of attribute values: {@code result[0]} .. definition in the node itself,
   *         {@code result[1]} .. definition in parent node,
   *         {@code result[2]} .. definition in "grand-parent" node, etc.
   * @see    #getAttribute
   * @see    #setAttribute
   * @see    #selfAttribute
   */
  List< Object > getAttributes ( String key );

  /**
   * Returns number of child nodes.
   *
   * @return Number of child nodes ({@code 0} for leaf {@link SceneNode}).
   * @see    #getChild
   * @see    #getChildMatrix
   */
  int children ();

  /**
   * Returns the given child node.
   *
   * @param  i Order of the requested node (from {@code 0} to {@link #children()}-1).
   * @return The requested {@link SceneNode} or {@code null}.
   * @see    #children
   * @see    #getChildMatrix
   * @see    #getChildMatrixInv
   */
  SceneNode getChild ( int i );

  /**
   * Returns transformation matrix for the given child node.
   *
   * @param  i Order of the child node.
   * @return Transformation matrix from the child's coordinate system to this node's one.
   *         {@code null} stands for <b>identity</b>.
   * @see    #children
   * @see    #getChild
   * @see    #getChildMatrixInv
   */
  TrMatrix getChildMatrix ( int i );

  /**
   * Returns inverse transformation matrix for the given child node.
   *
   * @param  i Order of the child node.
   * @return Transformation matrix from this node's coordinate system to child's one.
   *         {@code null} stands for <b>identity</b>.
   * @see    #children
   * @see    #getChild
   * @see    #getChildMatrix
   */
  TrMatrix getChildMatrixInv ( int i );

  /**
   * Inserts a new child node.
   *
   * @param  child The new child {@link SceneNode}.
   * @param  m Transformation matrix for the new child node (transformation from child's
   *           coordinate system to this node's one). Can be {@code null}, even
   *           for non-trivial {@code mInv}.
   * @param  mInv Inverse transformation matrix for the new child node (transformation from
   *              this node's coordinate system to child's one). Can be {@code null}, even
   *              for non-trivial {@code m}.
   * @return Order of the inserted child node (or {@code -1} in case of failure).
   * @see    #removeChild
   * @see    #children
   */
  int insertChild ( SceneNode child, TrMatrix m, TrMatrix mInv );

  /**
   * Removes the given child node.
   *
   * @param  i Order of the child node to be removed.
   * @return Number of child nodes after the remove-operation.
   * @see    #insertChild
   * @see    #children
   */
  int removeChild ( int i );

  /**
   * Returns the parent of this {@link SceneNode}.
   *
   * @return The parend node or {@code null} for scene-root.
   * @see    #setParent
   * @see    #getOrder
   * @see    #children
   * @see    #getChild
   */
  SceneNode getParent ();

  /**
   * Sets handle to the parent node.
   *
   * @param  par Parent node (or {@code null}).
   * @see    #getParent
   */
  void setParent ( SceneNode par );

  /**
   * Returns order of this {@link SceneNode} in its parent.
   *
   * @return Order of this node as a child (or {@code -1} for scene-root).
   * @see    #setOrder
   * @see    #getParent
   */
  int getOrder ();

  /**
   * Sets order of this {@link SceneNode} in its parent.
   *
   * @param  ord The new order.
   * @see    #getOrder
   */
  void setOrder ( int ord );

  /**
   * Returns {@code true} if this node is root of object (component) hierarchy.
   *
   * @see    #setObjectRoot
   */
  boolean isObjectRoot ();

  /**
   * Sets this node as root for the object (component).
   *
   * @param  root Should this node be treated as root?
   * @see    #isObjectRoot
   */
  void setObjectRoot ( boolean root );

  /**
   * Sets serial number of the current test.
   * Used in intersection testing.
   */
  int getSerial ();

  /**
   * Returns serial number assigned to the node.
   */
  void setSerial ( int serial );

  /**
   * Computes all intersections of the given ray with this sub-scene.
   *
   * @param  P0 Ray origin ({@code double[3]} or {@code double[4]}).
   * @param  P1 Direction vector of the ray ({@code double[3]} or {@code double[4]}).
   * @return List of all intersections with the sub-scene.
   */
  List< MicroFacet > intersection ( double[] P0, double[] P1 );

}