package cz.cuni.jagrlib;

/**
 * General semaphore for inter-thread synchronization.
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.13
 * @see     <a href="../../../../cz/cuni/jagrlib/Semaphore.java">Semaphore.java</a>
 */
public class Semaphore
{
  /** Number of waiting clients (&lt;0) or number of free signals (&gt;0). */
  protected int counter;

  /**
   * Initializing constructor.
   *
   * @param  limit Initial number of free signals (should be &gt;= 0)
   */
  public Semaphore ( int limit )
  {
    counter = (limit >= 0) ? limit : 0;
  }

  /**
   * Waits for one signal. Blocks the current thread for (potentially) infinite time.
   *
   * @return <tt>true</tt> if the current thread becomes a signal (<tt>false</tt> if
   *         it is interrupted by another thread).
   * @see    #semWait(long)
   */
  public synchronized boolean semWait ()
  {
    if ( --counter < 0 )
    {
      try
      {
        wait();                           // block the thread
      }
      catch ( InterruptedException e )
      {
        return false;
      }
    }

    return true;
  }

  /**
   * Waits for one signal for the restricted amount of time.
   *
   * @param  millis Number of milliseconds to wait (<tt>0L</tt> =&gt; wait forever).
   * @return <tt>true</tt> if the current thread becomes a signal or the given time
   *         expires (<tt>false</tt> if it is interrupted by another thread).
   * @see    #semWait()
   */
  public synchronized boolean semWait ( long millis )
  {
    if ( --counter < 0 )
    {
      try
      {
        wait( millis );                   // block the thread
      }
      catch ( InterruptedException e )
      {
        return false;
      }
    }

    return true;
  }

  /**
   * Signals the number of times. Resumes at most <tt>number</tt> of waiting threads.
   *
   * @param  number Number of signals to be generated.
   * @see    #semSignalAll
   */
  public synchronized void semSignal ( int number )
  {
    if ( number <= 0 ) return;

    do
    {
      if ( ++counter <= 0 )               // there are some waiting threads
        notify();                         // resume one waiting thread
    } while ( --number > 0 );
  }

  /**
   * Resume all waiting threads (at most <tt>limit</tt> ones).
   *
   * @param  limit Maximum number of waiting threads to be resumed (for &lt;= 0 resumes exactly
   *               all currently waiting threads).
   * @see    #semSignal
   */
  public synchronized void semSignalAll ( int limit )
  {
    if ( counter >= 0 ) return;

    if ( limit <= 0 ||
         limit > -counter ) limit = -counter;

    do
    {
      if ( ++counter <= 0 )               // there are some waiting threads
        notify();                         // resume one waiting thread
      else                                // no waiting threads left
      {
        counter = 0;
        return;
      }
    } while ( --limit > 0 );
  }

  /**
   * Returns number of signals available.
   *
   * @return Number of signals available in current time.
   */
  public int getSignalCount ()
  {
    return counter;
  }

}
