package cz.cuni.jagrlib;

/**
 * 2D and 3D transformation matrix - uses homogenous coordinates. Coordinate vectors are treated
 * as row-vectors - matrix multiplications are associated from the left (multiplication from
 * the right side).
 *
 * <p>Matrix size can be 3x2 or 3x3 (for 2D transformations), 4x3 or 4x4 (for 3D transformations).
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.09
 * @see     <a href="../../../../cz/cuni/jagrlib/TrMatrix.java">TrMatrix.java</a>
 * @see     Template
 */
public class TrMatrix
{
  /**
   * The matrix itself.
   *
   * @see    #height
   * @see    #width
   */
  public double m[][];

  /**
   * Matrix height - number of rows (3 for 2D transformation, 4 for 3D transformation).
   *
   * @see    #width
   */
  public int height;

  /**
   * Matrix width - number of columns (2 to 3 for 2D transformation, 3 to 4 for 3D transformation).
   *
   * @see    #height
   */
  public int width;

  /**
   * Default constructor: creates identical 3D transformation (4x4 matrix size).
   */
  public TrMatrix ()
  {
    identity( 4, 4 );
  }

  /**
   * Identity constructor: creates identical transformation of the given size (3x2, 3x3, 4x3
   * or 4x4).
   *
   * @param  h Matrix height (number of rows). Should be 3 for 2D transformations and 4 for
   *           3D ones.
   * @param  w Matrix width (number of columns). Should be 2 or 3 for 2D transformations,
   *           3 or 4 for 3D ones.
   */
  public TrMatrix ( int h, int w )
  {
    identity( h, w );
  }

  /**
   * Copy constructor.
   *
   * @param  s Source matrix.
   */
  public TrMatrix ( TrMatrix s )
  {
    if ( s != null )
      set( s );
    else
      identity( 4, 4 );
  }

  /**
   * Constructs matrix from row vectors.
   */
  public TrMatrix ( double[] row0, double[] row1, double[] row2, double[] row3 )
  {
    if ( row0 == null || row0.length < 2 ||
         row1 == null || row1.length < 2 ||
         row2 == null || row2.length < 2 )
    {
      identity( 4, 4 );
      return;
    }
    width  = row0.length;
    height = (row3 == null) ? 3 : 4;
    if ( height == 3 )
      m = new double[][] { row0, row1, row2 };
    else
      m = new double[][] { row0, row1, row2, row3 };
  }

  /**
   * Logs the matrix using {@link LogFile}.
   */
  public void log ( String name )
  {
    LogFile.log( ((name == null) ? "Matrix" : name) + " [ " + height + ", " + width + " ]:" );

    for ( double[] v : m )
      LogFile.log( Geometry.vectorToString( v ) );
  }

  /**
   * Initializes the matrix using the given size.
   * Matrix coefficients are not defined.
   *
   * @param  h Matrix height (number of rows). Should be 3 for 2D transformations and 4 for
   *           3D ones.
   * @param  w Matrix width (number of columns). Should be 2 or 3 for 2D transformations,
   *           3 or 4 for 3D ones.
   */
  public void init ( int h, int w )
  {
    height = Formula.clamp( h, 3, 4 );
    width  = Formula.clamp( w, height - 1, height );
    if ( m      == null ||
         height != m.length ||
         width  != m[0].length )
      m = new double[ height ][ width ];
  }

  /**
   * Creates identical matrix of the given size.
   *
   * @param  h Matrix height (number of rows). Should be 3 for 2D transformations and 4 for
   *           3D ones.
   * @param  w Matrix width (number of columns). Should be 2 or 3 for 2D transformations,
   *           3 or 4 for 3D ones.
   */
  public void identity ( int h, int w )
  {
    init( h, w );

    int i, j;
    for ( i = 0; i < height; i++ )
      for ( j = 0; j < width; j++ )
        m[i][j] = (i == j) ? 1.0 : 0.0;
  }

  /**
   * Transposes the given matrix.
   *
   * @see    #invert(TrMatrix)
   */
  public void transpose ()
  {
    int i, j;
    for ( i = 0; i < width; i++ )
       for ( j = 0; j < i; j++ )
       {
         double tmp = m[i][j];
         m[i][j]    = m[j][i];
         m[j][i]    = tmp;
       }

    if ( i < height )
      for ( j = 0; j < width; j++ )
        m[i][j] = (i == j) ? 1.0 : 0.0;
  }

  /**
   * Append the given matrix to the actual transform.
   *
   * @param  mul Matrix to be appended (concatenated) to the actual transform.
   * @see    #prepend
   */
  public void append ( TrMatrix mul )
  {
    if ( mul        == null ||
         mul.height != height ) return;

    int    w       = Math.max( width, mul.width );
    double res[][] = new double[ height ][ w ];
    int i, j, k;
    double a;

    for ( i = 0; i < height; i++ )        // one target row
    {
      for ( j = 0; j < mul.width; j++ )   // one target column
      {
        a = 0.0;
        for ( k = 0; k < width; k++ )
          a += m[i][k] * mul.m[k][j];

        if ( i == height - 1 && k < height )
          a += mul.m[k][j];

        res[i][j] = a;
      }

      if ( j < width )
        res[i][j] = m[i][j];
    }

      // finalizing..
    width = w;
    m     = res;
  }

  /**
   * Prepend the given matrix to the actual transform.
   *
   * @param  mul Matrix to be prepended to the actual transform.
   * @see    #append
   */
  public void prepend ( TrMatrix mul )
  {
    if ( mul        == null ||
         mul.height != height ) return;

    int w = Math.max( width, mul.width );
    double res[][] = new double[ height ][ w ];
    int i, j, k;
    double a;

    for ( i = 0; i < height; i++ )        // one target row
    {
      for ( j = 0; j < width; j++ )       // one target column
      {
        a = 0.0;
        for ( k = 0; k < mul.width; k++ )
          a += mul.m[i][k] * m[k][j];

        if ( i == height - 1 && k < height )
          a += m[k][j];

        res[i][j] = a;
      }

      if ( j < mul.width )
        res[i][j] = mul.m[i][j];
    }

      // finalizing..
    width = w;
    m = res;
  }

  /**
   * Inverts arbitrary transformation matrix.
   *
   * @param  src Source transformation matrix (can be {@code m} or {@code null}).
   * @return {@code false} if the source matrix is singular.
   * @see    #invert(TrMatrix)
   */
  public boolean invert ( double[][] src )
  {
    if ( src    == null ) src = m;

    if ( src    == null || src.length    < 2 ||
         src[0] == null || src[0].length < 2 ) return false;

    int size = Math.max( src.length, src[0].length );
    double n[][] = new double[ size ][ size ];
    int i, j, k;

      // init the working double[size][size] matrix:
    for ( i = 0; i < src.length; i++ )
    {
      System.arraycopy( src[i], 0, n[i], 0, src[i].length );
      for ( j = src[i].length; j < size; j++ )
        n[i][j] = (i == j) ? 1.0 : 0.0;
    }

    while ( i < size )
    {
      for ( j = 0; j < size; j++ )
        n[i][j] = (i == j) ? 1.0 : 0.0;
      i++;
    }

      // result matrix:
    double tmp[][] = new double[ size ][ size ];
    for ( i = 0; i < size; i++ )          // init: identity
      for ( j = 0; j < size; j++ )
        tmp[i][j] = (i == j) ? 1.0 : 0.0;

    double maxAbs, absH;
    int maxi;
    for ( j = 0; j < size; j++ )          // elimination in the j-th column
    {
      maxAbs = Math.abs( n[j][j] );
      maxi = j;
      for ( i = j + 1; i < size; i++ )
      {
        absH = Math.abs( n[i][j] );
        if ( absH > maxAbs )
        {
          maxAbs = absH;
          maxi = i;
        }
      }

        // maxi-th row has the greatest absolute value in the j-th column
      if ( maxAbs < Geometry.EPSILON ) return false;  // singular matrix

      if ( maxi != j )
      {
        double t[] = tmp[maxi];
        tmp[maxi]  = tmp[j];
        tmp[j]     = t;
        t          = n[maxi];
        n[maxi]    = n[j];
        n[j]       = t;
      }

      double pivot = 1.0 / n[j][j];
      for ( k = j + 1; k < size; k++ )
        n[j][k] *= pivot;

      for ( k = 0; k < size; k++ )
        tmp[j][k] *= pivot;

      for ( i = 0; i < size; i++ )        // elimination of the i-th row
        if ( i != j )
        {
          pivot = n[i][j];
          for ( k = j + 1; k < size; k++ )
            n[i][k] -= n[j][k] * pivot;

          for ( k = 0; k < size; k++ )
            tmp[i][k] -= tmp[j][k] * pivot;
        }
    }

      // finalizing..
    m      = tmp;
    width  =
    height = size;
    return true;
  }

  /**
   * Matrix assignment.
   *
   * @param  s Source transformation matrix (cannot be {@code null}).
   */
  public void set ( TrMatrix s )
  {
    if ( s == null ) return;

    width  = s.width;
    height = s.height;
    m = new double[ height ][ width ];
    int i;

    for ( i = 0; i < height; i++ )
      System.arraycopy( s.m[i], 0, m[i], 0, width );
  }

  /**
   * Matrix assignment with inversion.
   *
   * @param  s Source transformation matrix (can be {@code null}).
   * @return {@code false} if the source matrix is singular.
   * @see    #invert(double[][])
   * @see    #transpose
   * @see    #set
   */
  public boolean invert ( TrMatrix s )
  {
    return invert( (s == null) ? m : s.m );
  }

  /**
   * Transforms the given matrix to be used on 3D normal vectors.
   *
   * @param  s Source transformation matrix (can be {@code null}).
   * @return {@code false} if the source matrix is singular.
   */
  public boolean setNormalTransform3D ( TrMatrix s )
  {
    if ( s != null &&
         s != this )
      set( s );
    transpose();
    return invert( m );
  }

  /**
   * 2D linear transformation - processes both Cartesian and homogeneous coordinate
   * vectors.
   *
   * @param  src Source coordinate vector ([x,y] or [x,y,w]).
   * @param  dest Destination coordinate vector ([x,y] or [x,y,w]). Can be the same as
   *              {@code src}.
   */
  public void transformPoint2D ( double[] src, double[] dest )
  {
    double x, y, w;
    if ( src.length > 2 )                 // source [x,y,w]
    {
      x = src[0] * m[0][0] + src[1] * m[1][0] + src[2] * m[2][0];
      y = src[0] * m[0][1] + src[1] * m[1][1] + src[2] * m[2][1];
      if ( width > 2 )                    // 3x3 matrix
        w = src[0] * m[0][2] + src[1] * m[1][2] + src[2] * m[2][2];
      else                                // 3x2 matrix
        w = src[2];
    }
    else                                  // source [x,y]
    {
      x = src[0] * m[0][0] + src[1] * m[1][0] + m[2][0];
      y = src[0] * m[0][1] + src[1] * m[1][1] + m[2][1];
      if ( width > 2 )                    // 3x3 matrix
        w = src[0] * m[0][2] + src[1] * m[1][2] + m[2][2];
      else                                // 3x2 matrix
        w = 1.0;
    }

    if ( dest.length > 2 )                // dest [x,y,w]
    {
      dest[0] = x;
      dest[1] = y;
      dest[2] = w;
    }
    else                                  // dest [x,y]
      if ( w == 1.0 ||
           w > -Geometry.EPSILON && w < Geometry.EPSILON )
      {
        dest[0] = x;
        dest[1] = y;
      }
      else
      {
        dest[0] = x / w;
        dest[1] = y / w;
      }
  }

  /**
   * 2D linear transformation - processes both Cartesian and homogeneous coordinate
   * vectors.
   *
   * @param  src Source coordinate vector ([x,y] or [x,y,w]).
   * @param  dest Destination coordinate vector ([x,y] or [x,y,w]). Can be the same as
   *              {@code src}.
   */
  public void transformPoint2D ( float[] src, float[] dest )
  {
    double x, y, w;
    if ( src.length > 2 )                 // source [x,y,w]
    {
      x = src[0] * m[0][0] + src[1] * m[1][0] + src[2] * m[2][0];
      y = src[0] * m[0][1] + src[1] * m[1][1] + src[2] * m[2][1];
      if ( width > 2 )                    // 3x3 matrix
        w = src[0] * m[0][2] + src[1] * m[1][2] + src[2] * m[2][2];
      else                                // 3x2 matrix
        w = src[2];
    }
    else                                  // source [x,y]
    {
      x = src[0] * m[0][0] + src[1] * m[1][0] + m[2][0];
      y = src[0] * m[0][1] + src[1] * m[1][1] + m[2][1];
      if ( width > 2 )                    // 3x3 matrix
        w = src[0] * m[0][2] + src[1] * m[1][2] + m[2][2];
      else                                // 3x2 matrix
        w = 1.0;
    }

    if ( dest.length > 2 )                // dest [x,y,w]
    {
      dest[0] = (float)x;
      dest[1] = (float)y;
      dest[2] = (float)w;
    }
    else                                  // dest [x,y]
      if ( w == 1.0 ||
           w > -Geometry.EPSILON && w < Geometry.EPSILON )
      {
        dest[0] = (float)x;
        dest[1] = (float)y;
      }
      else
      {
        dest[0] = (float)( x / w );
        dest[1] = (float)( y / w );
      }
  }

  /**
   * 2D linear transformation - processes both Cartesian and homogeneous coordinate
   * vectors. Source and destination coordinate vectors reside inside bigger
   * {@code double[]} arrays.
   *
   * @param  src Array holding source coordinate vector.
   * @param  srcI Origin of source coordinate vector.
   * @param  srcLen Length of source coordinate vector.
   * @param  dest Array holding destination coordinate vector. Can be the same as
   *              {@code src} (including {@code srcI == destI}).
   * @param  destI Origin of destination coordinate vector.
   * @param  destLen Length of destination coordinate vector.
   */
  public void transformPoint2D ( double[] src,  int srcI,  int srcLen,
                                 double[] dest, int destI, int destLen )
  {
    double x, y, w;
    if ( srcLen > 2 )                     // source [x,y,w]
    {
      x = src[srcI] * m[0][0] + src[srcI+1] * m[1][0] + src[srcI+2] * m[2][0];
      y = src[srcI] * m[0][1] + src[srcI+1] * m[1][1] + src[srcI+2] * m[2][1];
      if ( width > 2 )                    // 3x3 matrix
        w = src[srcI] * m[0][2] + src[srcI+1] * m[1][2] + src[srcI+2] * m[2][2];
      else                                // 3x2 matrix
        w = src[srcI+2];
    }
    else                                  // source [x,y]
    {
      x = src[srcI] * m[0][0] + src[srcI+1] * m[1][0] + m[2][0];
      y = src[srcI] * m[0][1] + src[srcI+1] * m[1][1] + m[2][1];
      if ( width > 2 )                    // 3x3 matrix
        w = src[srcI] * m[0][2] + src[srcI+1] * m[1][2] + m[2][2];
      else                                // 3x2 matrix
        w = 1.0;
    }

    if ( destLen > 2 )                    // dest [x,y,w]
    {
      dest[destI]   = x;
      dest[destI+1] = y;
      dest[destI+2] = w;
    }
    else                                  // dest [x,y]
      if ( w == 1.0 ||
           w > -Geometry.EPSILON && w < Geometry.EPSILON )
      {
        dest[destI]   = x;
        dest[destI+1] = y;
      }
      else
      {
        dest[destI]   = x / w;
        dest[destI+1] = y / w;
      }
  }

  /**
   * 2D linear transformation - processes both Cartesian and homogeneous coordinate
   * vectors. Source/destination coordinate vectors reside inside bigger
   * {@code double[]}/{@code float[]} arrays.
   *
   * @param  src Array holding source coordinate vector.
   * @param  srcI Origin of source coordinate vector.
   * @param  srcLen Length of source coordinate vector.
   * @param  dest Array holding destination coordinate vector. Can be the same as
   *              {@code src} (including {@code srcI == destI}).
   * @param  destI Origin of destination coordinate vector.
   * @param  destLen Length of destination coordinate vector.
   */
  public void transformPoint2D ( float[] src,  int srcI,  int srcLen,
                                 float[] dest, int destI, int destLen )
  {
    double x, y, w;
    if ( srcLen > 2 )                     // source [x,y,w]
    {
      x = src[srcI] * m[0][0] + src[srcI+1] * m[1][0] + src[srcI+2] * m[2][0];
      y = src[srcI] * m[0][1] + src[srcI+1] * m[1][1] + src[srcI+2] * m[2][1];
      if ( width > 2 )                    // 3x3 matrix
        w = src[srcI] * m[0][2] + src[srcI+1] * m[1][2] + src[srcI+2] * m[2][2];
      else                                // 3x2 matrix
        w = src[srcI+2];
    }
    else                                  // source [x,y]
    {
      x = src[srcI] * m[0][0] + src[srcI+1] * m[1][0] + m[2][0];
      y = src[srcI] * m[0][1] + src[srcI+1] * m[1][1] + m[2][1];
      if ( width > 2 )                    // 3x3 matrix
        w = src[srcI] * m[0][2] + src[srcI+1] * m[1][2] + m[2][2];
      else                                // 3x2 matrix
        w = 1.0;
    }

    if ( destLen > 2 )                    // dest [x,y,w]
    {
      dest[destI]   = (float)x;
      dest[destI+1] = (float)y;
      dest[destI+2] = (float)w;
    }
    else                                  // dest [x,y]
      if ( w == 1.0 ||
           w > -Geometry.EPSILON && w < Geometry.EPSILON )
      {
        dest[destI]   = (float)x;
        dest[destI+1] = (float)y;
      }
      else
      {
        dest[destI]   = (float)( x / w );
        dest[destI+1] = (float)( y / w );
      }
  }

  /**
   * 3D linear transformation - processes both Cartesian and homogeneous coordinate
   * vectors.
   *
   * @param  src Source coordinate vector ([x,y,z] or [x,y,z,w]).
   * @param  dest Destination coordinate vector ([x,y,z] or [x,y,z,w]). Can be the same as
   *              {@code src}.
   */
  public void transformPoint3D ( double[] src, double[] dest )
  {
    if ( height < 4 ) return;

    double x, y, z, w;
    if ( src.length > 3 )                 // source [x,y,z,w]
    {
      x = src[0] * m[0][0] + src[1] * m[1][0] + src[2] * m[2][0] + src[3] * m[3][0];
      y = src[0] * m[0][1] + src[1] * m[1][1] + src[2] * m[2][1] + src[3] * m[3][1];
      z = src[0] * m[0][2] + src[1] * m[1][2] + src[2] * m[2][2] + src[3] * m[3][2];
      if ( width > 3 )                    // 4x4 matrix
        w = src[0] * m[0][3] + src[1] * m[1][3] + src[2] * m[2][3] + src[3] * m[3][3];
      else                                // 4x3 matrix
        w = src[3];
    }
    else                                  // source [x,y,z]
    {
      x = src[0] * m[0][0] + src[1] * m[1][0] + src[2] * m[2][0] + m[3][0];
      y = src[0] * m[0][1] + src[1] * m[1][1] + src[2] * m[2][1] + m[3][1];
      z = src[0] * m[0][2] + src[1] * m[1][2] + src[2] * m[2][2] + m[3][2];
      if ( width > 3 )                    // 4x4 matrix
        w = src[0] * m[0][3] + src[1] * m[1][3] + src[2] * m[2][3] + m[3][3];
      else                                // 4x3 matrix
        w = 1.0;
    }

    if ( dest.length > 3 )                // dest [x,y,z,w]
    {
      dest[0] = x;
      dest[1] = y;
      dest[2] = z;
      dest[3] = w;
    }
    else                                  // dest [x,y,z]
      if ( w == 1.0 ||
           w > -Geometry.EPSILON && w < Geometry.EPSILON )
      {
        dest[0] = x;
        dest[1] = y;
        if ( dest.length == 3 )
          dest[2] = z;
      }
      else
      {
        dest[0] = x / w;
        dest[1] = y / w;
        if ( dest.length == 3 )
          dest[2] = z / w;
      }
  }

  /**
   * 3D linear transformation - processes both Cartesian and homogeneous coordinate
   * vectors.
   *
   * @param  src Source coordinate vector ([x,y,z] or [x,y,z,w]).
   * @param  dest Destination coordinate vector ([x,y,z] or [x,y,z,w]). Can be the same as
   *              {@code src}.
   */
  public void transformPoint3D ( float[] src, float[] dest )
  {
    if ( height < 4 ) return;

    double x, y, z, w;
    if ( src.length > 3 )                 // source [x,y,z,w]
    {
      x = src[0] * m[0][0] + src[1] * m[1][0] + src[2] * m[2][0] + src[3] * m[3][0];
      y = src[0] * m[0][1] + src[1] * m[1][1] + src[2] * m[2][1] + src[3] * m[3][1];
      z = src[0] * m[0][2] + src[1] * m[1][2] + src[2] * m[2][2] + src[3] * m[3][2];
      if ( width > 3 )                    // 4x4 matrix
        w = src[0] * m[0][3] + src[1] * m[1][3] + src[2] * m[2][3] + src[3] * m[3][3];
      else                                // 4x3 matrix
        w = src[3];
    }
    else                                  // source [x,y,z]
    {
      x = src[0] * m[0][0] + src[1] * m[1][0] + src[2] * m[2][0] + m[3][0];
      y = src[0] * m[0][1] + src[1] * m[1][1] + src[2] * m[2][1] + m[3][1];
      z = src[0] * m[0][2] + src[1] * m[1][2] + src[2] * m[2][2] + m[3][2];
      if ( width > 3 )                    // 4x4 matrix
        w = src[0] * m[0][3] + src[1] * m[1][3] + src[2] * m[2][3] + m[3][3];
      else                                // 4x3 matrix
        w = 1.0;
    }

    if ( dest.length > 3 )                // dest [x,y,z,w]
    {
      dest[0] = (float)x;
      dest[1] = (float)y;
      dest[2] = (float)z;
      dest[3] = (float)w;
    }
    else                                  // dest [x,y,z]
      if ( w == 1.0 ||
           w > -Geometry.EPSILON && w < Geometry.EPSILON )
      {
        dest[0] = (float)x;
        dest[1] = (float)y;
        if ( dest.length == 3 )
          dest[2] = (float)z;
      }
      else
      {
        dest[0] = (float)( x / w );
        dest[1] = (float)( y / w );
        if ( dest.length == 3 )
          dest[2] = (float)( z / w );
      }
  }

  /**
   * 3D linear transformation - processes both Cartesian and homogeneous coordinate
   * vectors. Source and destination coordinate vectors reside inside bigger
   * {@code double[]} arrays.
   *
   * @param  src Array holding source coordinate vector.
   * @param  srcI Origin of source coordinate vector.
   * @param  srcLen Length of source coordinate vector.
   * @param  dest Array holding destination coordinate vector. Can be the same as
   *              {@code src} (including {@code srcI == destI}).
   * @param  destI Origin of destination coordinate vector.
   * @param  destLen Length of destination coordinate vector.
   */
  public void transformPoint3D ( double[] src,  int srcI,  int srcLen,
                                 double[] dest, int destI, int destLen )
  {
    if ( height < 4 ) return;

    double x, y, z, w;
    if ( srcLen > 3 )                     // source [x,y,z,w]
    {
      x = src[srcI] * m[0][0] + src[srcI+1] * m[1][0] + src[srcI+2] * m[2][0] + src[srcI+3] * m[3][0];
      y = src[srcI] * m[0][1] + src[srcI+1] * m[1][1] + src[srcI+2] * m[2][1] + src[srcI+3] * m[3][1];
      z = src[srcI] * m[0][2] + src[srcI+1] * m[1][2] + src[srcI+2] * m[2][2] + src[srcI+3] * m[3][2];
      if ( width > 3 )                    // 4x4 matrix
        w = src[srcI] * m[0][3] + src[srcI+1] * m[1][3] + src[srcI+2] * m[2][3] + src[srcI+3] * m[3][3];
      else                                // 4x3 matrix
        w = src[srcI+3];
    }
    else                                  // source [x,y,z]
    {
      x = src[srcI] * m[0][0] + src[srcI+1] * m[1][0] + src[srcI+2] * m[2][0] + m[3][0];
      y = src[srcI] * m[0][1] + src[srcI+1] * m[1][1] + src[srcI+2] * m[2][1] + m[3][1];
      z = src[srcI] * m[0][2] + src[srcI+1] * m[1][2] + src[srcI+2] * m[2][2] + m[3][2];
      if ( width > 3 )                    // 4x4 matrix
        w = src[srcI] * m[0][3] + src[srcI+1] * m[1][3] + src[srcI+2] * m[2][3] + m[3][3];
      else                                // 4x3 matrix
        w = 1.0;
    }

    if ( destLen > 3 )                    // dest [x,y,z,w]
    {
      dest[destI]   = x;
      dest[destI+1] = y;
      dest[destI+2] = z;
      dest[destI+3] = w;
    }
    else                                  // dest [x,y,z]
      if ( w == 1.0 ||
           w > -Geometry.EPSILON && w < Geometry.EPSILON )
      {
        dest[destI]   = x;
        dest[destI+1] = y;
        if ( destLen == 3 )
          dest[destI+2] = z;
      }
      else
      {
        dest[destI]   = x / w;
        dest[destI+1] = y / w;
        if ( destLen == 3 )
          dest[destI+2] = z / w;
      }
  }

  /**
   * 3D linear transformation - processes both Cartesian and homogeneous coordinate
   * vectors. Source and destination coordinate vectors reside inside bigger
   * {@code float[]} arrays.
   *
   * @param  src Array holding source coordinate vector.
   * @param  srcI Origin of source coordinate vector.
   * @param  srcLen Length of source coordinate vector.
   * @param  dest Array holding destination coordinate vector. Can be the same as
   *              {@code src} (including {@code srcI == destI}).
   * @param  destI Origin of destination coordinate vector.
   * @param  destLen Length of destination coordinate vector.
   */
  public void transformPoint3D ( float[] src,  int srcI,  int srcLen,
                                 float[] dest, int destI, int destLen )
  {
    if ( height < 4 ) return;

    double x, y, z, w;
    if ( srcLen > 3 )                     // source [x,y,z,w]
    {
      x = src[srcI] * m[0][0] + src[srcI+1] * m[1][0] + src[srcI+2] * m[2][0] + src[srcI+3] * m[3][0];
      y = src[srcI] * m[0][1] + src[srcI+1] * m[1][1] + src[srcI+2] * m[2][1] + src[srcI+3] * m[3][1];
      z = src[srcI] * m[0][2] + src[srcI+1] * m[1][2] + src[srcI+2] * m[2][2] + src[srcI+3] * m[3][2];
      if ( width > 3 )                    // 4x4 matrix
        w = src[srcI] * m[0][3] + src[srcI+1] * m[1][3] + src[srcI+2] * m[2][3] + src[srcI+3] * m[3][3];
      else                                // 4x3 matrix
        w = src[srcI+3];
    }
    else                                  // source [x,y,z]
    {
      x = src[srcI] * m[0][0] + src[srcI+1] * m[1][0] + src[srcI+2] * m[2][0] + m[3][0];
      y = src[srcI] * m[0][1] + src[srcI+1] * m[1][1] + src[srcI+2] * m[2][1] + m[3][1];
      z = src[srcI] * m[0][2] + src[srcI+1] * m[1][2] + src[srcI+2] * m[2][2] + m[3][2];
      if ( width > 3 )                    // 4x4 matrix
        w = src[srcI] * m[0][3] + src[srcI+1] * m[1][3] + src[srcI+2] * m[2][3] + m[3][3];
      else                                // 4x3 matrix
        w = 1.0;
    }

    if ( destLen > 3 )                    // dest [x,y,z,w]
    {
      dest[destI]   = (float)x;
      dest[destI+1] = (float)y;
      dest[destI+2] = (float)z;
      dest[destI+3] = (float)w;
    }
    else                                  // dest [x,y,z]
      if ( w == 1.0 ||
           w > -Geometry.EPSILON && w < Geometry.EPSILON )
      {
        dest[destI]   = (float)x;
        dest[destI+1] = (float)y;
        if ( destLen == 3 )
          dest[destI+2] = (float)z;
      }
      else
      {
        dest[destI]   = (float)(x / w);
        dest[destI+1] = (float)(y / w);
        if ( destLen == 3 )
          dest[destI+2] = (float)(z / w);
      }
  }

  /**
   * 3D homomorphism - processes both Cartesian and homogeneous coordinate vectors.
   *
   * @param  src Source coordinate vector ([x,y,z] or [x,y,z,w]).
   * @param  dest Destination coordinate vector ([x,y,z] or [x,y,z,w]). Can be the same as
   *              {@code src}.
   * @see    #transformPoint3D(double[],double[])
   */
  public void transformVector3D ( double[] src, double[] dest )
  {
    if ( height < 3 ) return;

    final double x = src[0] * m[0][0] + src[1] * m[1][0] + src[2] * m[2][0];
    final double y = src[0] * m[0][1] + src[1] * m[1][1] + src[2] * m[2][1];
    final double z = src[0] * m[0][2] + src[1] * m[1][2] + src[2] * m[2][2];
    dest[0] = x;
    dest[1] = y;
    dest[2] = z;
    if ( dest.length > 3 )                // dest [x,y,z,w]
      dest[3] = 0.0;
  }

  /**
   * Appends the given translation to the actual transform matrix. Optionally updates the
   * inverse matrix, too.
   *
   * @param  vector Translation vector ({@code double[3]} or {@code double[4]}).
   * @param  inv Optional inverse matrix.
   */
  public void translate3D ( double[] vector, TrMatrix inv )
  {
    if ( vector == null ||
         vector.length < 3 ) return;

    TrMatrix tmp = new TrMatrix( 4, 3 );
    if ( vector.length == 3 ||
         vector[3] > -Geometry.EPSILON && vector[3] < Geometry.EPSILON )
      System.arraycopy( vector, 0, tmp.m[3], 0, 3 );

    else
    {
      tmp.m[3][0] = vector[0] / vector[3];
      tmp.m[3][1] = vector[1] / vector[3];
      tmp.m[3][2] = vector[2] / vector[3];
    }

    append( tmp );
    if ( inv != null )
    {
      tmp.m[3][0] = -tmp.m[3][0];
      tmp.m[3][1] = -tmp.m[3][1];
      tmp.m[3][2] = -tmp.m[3][2];
      inv.prepend( tmp );
    }
  }

  /**
   * Appends the given translation to the actual transform matrix. Optionally updates the
   * inverse matrix, too.
   *
   * @param  tx X-component of the translation vector.
   * @param  ty Y-component of the translation vector.
   * @param  tz Z-component of the translation vector.
   * @param  inv Optional inverse matrix.
   */
  public void translate3D ( double tx, double ty, double tz, TrMatrix inv )
  {
    TrMatrix tmp = new TrMatrix( 4, 3 );
    tmp.m[3][0] = tx;
    tmp.m[3][1] = ty;
    tmp.m[3][2] = tz;

    append( tmp );
    if ( inv != null )
    {
      tmp.m[3][0] = -tx;
      tmp.m[3][1] = -ty;
      tmp.m[3][2] = -tz;
      inv.prepend( tmp );
    }
  }

  /**
   * Appends the given rotation to the actual transform matrix. Optionally updates the
   * inverse matrix, too.
   *
   * @param  axis Axis index ({@code 0} .. X, {@code 1} .. Y, {@code 2} .. Z).
   * @param  sina Sine of the rotation angle.
   * @param  cosa Cosine of the rotation angle.
   * @param  inv Optional inverse matrix.
   * @see    #rotateAxis3D(int,double,TrMatrix)
   * @see    #rotate3D(double[],double[],double,TrMatrix)
   */
  public void rotateAxis3D ( int axis, double sina, double cosa, TrMatrix inv )
  {
    if ( axis < 0 ||
         axis > 2 ) return;

    TrMatrix tmp = new TrMatrix( 4, 3 );

    switch ( axis )
    {
      case 0:                             // X-axis rotation
        tmp.m[1][1] =  cosa; tmp.m[1][2] =  sina;
        tmp.m[2][1] = -sina; tmp.m[2][2] =  cosa;
        break;

      case 1:                             // Y-axis rotation
        tmp.m[0][0] =  cosa; tmp.m[0][2] = -sina;
        tmp.m[2][0] =  sina; tmp.m[2][2] =  cosa;
        break;

      case 2:                             // Z-axis rotation
        tmp.m[0][0] =  cosa; tmp.m[0][1] =  sina;
        tmp.m[1][0] = -sina; tmp.m[1][1] =  cosa;
        break;
    }

    append( tmp );
    if ( inv != null )
    {
      tmp.transpose();
      inv.prepend( tmp );
    }
  }

  /**
   * Appends the given rotation to the actual transform matrix. Optionally updates the
   * inverse matrix, too.
   *
   * @param  axis Axis index ({@code 0} .. X, {@code 1} .. Y, {@code 2} .. Z).
   * @param  angle Rotation angle in radians.
   * @param  inv Optional inverse matrix.
   * @see    #rotateAxis3D(int,double,double,TrMatrix)
   * @see    #rotate3D(double[],double[],double,TrMatrix)
   */
  public void rotateAxis3D ( int axis, double angle, TrMatrix inv )
  {
    rotateAxis3D( axis, Math.sin( angle ), Math.cos( angle ), inv );
  }

  /**
   * Append general rotation to the actual transform matrix. Optionally updates
   * the inverse matrix, too.
   *
   * @param  origin Origin of the axis.
   * @param  direction Directional vector of the axis.
   * @param  angle Rotation angle in radians.
   * @param  inv Optional inverse matrix.
   * @see    #rotateAxis3D(int,double,TrMatrix)
   * @see    #rotateAxis3D(int,double,double,TrMatrix)
   */
  public void rotate3D ( double[] origin, double[] direction, double angle, TrMatrix inv )
  {
    if ( origin    == null || origin.length    < 3 ||
         direction == null || direction.length < 3 ) return;

    TrMatrix m  = new TrMatrix( 4, 3 );
    TrMatrix mi = new TrMatrix( 4, 3 );
    m.moveToZAxis( origin, direction, mi );
    append( m );
    if ( inv != null )
      inv.prepend( mi );

    rotateAxis3D( 2, angle, inv );
    append( mi );
    if ( inv != null )
      inv.prepend( m );
  }

  /**
   * Append the "move-to-z-axis" transform to the actual transform matrix. Optionally
   * updates the inverse matrix, too.
   *
   * @param  origin Origin of the vector.
   * @param  direction Directional vector.
   * @param  inv Optional inverse matrix.
   */
  public void moveToZAxis ( double[] origin, double[] direction, TrMatrix inv )
  {
    if ( origin    == null || origin.length    < 3 ||
         direction == null || direction.length < 3 ) return;

    origin = Geometry.toCartesian3D( origin, null );
      // I. translate origin to [0,0,0]
    translate3D( -origin[0], -origin[1], -origin[2], inv );
    direction = Geometry.normalize3D( direction, null );
    double d = Math.hypot( direction[0], direction[2] );
      // II. y-rotate into yz-plane
    if ( d > Geometry.EPSILON )
      rotateAxis3D( 1, -direction[0]/d, direction[2]/d, inv );
      // III. x-rotate into z-axis
    rotateAxis3D( 0, direction[1], d, inv );
  }

  /**
   * Append axes scale transform to the actual transform matrix. Optionally updates
   * the inverse matrix, too.
   *
   * @param  origin Scale origin or {@code null} for [0,0,0].
   * @param  sX Scale coefficient for X-axis.
   * @param  sY Scale coefficient for Y-axis.
   * @param  sZ Scale coefficient for Z-axis.
   * @param  inv Optional inverse matrix.
   * @see    #scale3D(double[],double,TrMatrix)
   * @see    #scale3D(double[],double[],double,double,TrMatrix)
   * @see    #scaleAxes3D(double[],double[],TrMatrix)
   */
  public void scaleAxes3D ( double[] origin, double sX, double sY, double sZ, TrMatrix inv )
  {
    if ( origin != null )
    {
      if ( origin.length < 3 ) return;
      origin = Geometry.toCartesian3D( origin, null );
        // translate origin to [0,0,0]
      translate3D( -origin[0], -origin[1], -origin[2], inv );
    }

    TrMatrix tmp = new TrMatrix( 4, 3 );
    tmp.m[0][0] = sX;                     // nonuniform scale
    tmp.m[1][1] = sY;
    tmp.m[2][2] = sZ;
    append( tmp );
    if ( inv != null )
    {
      tmp.m[0][0] = 1.0 / sX;
      tmp.m[1][1] = 1.0 / sY;
      tmp.m[2][2] = 1.0 / sZ;
      inv.prepend( tmp );
    }

    if ( origin != null )
    {
        // translate [0,0,0] back to origin
      translate3D( origin[0], origin[1], origin[2], inv );
    }
  }

  /**
   * Append axes scale transform to the actual transform matrix. Optionally updates
   * the inverse matrix, too.
   *
   * @param  origin Scale origin or {@code null} for [0,0,0].
   * @param  scale Non-uniform scale vector.
   * @param  inv Optional inverse matrix.
   * @see    #scale3D(double[],double,TrMatrix)
   * @see    #scale3D(double[],double[],double,double,TrMatrix)
   * @see    #scaleAxes3D(double[],double,double,double,TrMatrix)
   */
  public void scaleAxes3D ( double[] origin, double[] scale, TrMatrix inv )
  {
    if ( scale == null || scale.length  < 3 ) return;

    scaleAxes3D( origin, scale[0], scale[1], scale[2], inv );
  }

  /**
   * Append uniform scale transform to the actual transform matrix. Optionally updates
   * the inverse matrix, too.
   *
   * @param  origin Scale origin or {@code null} for [0,0,0].
   * @param  scale Uniform scale vector.
   * @param  inv Optional inverse matrix.
   * @see    #scaleAxes3D(double[],double[],TrMatrix)
   * @see    #scaleAxes3D(double[],double,double,double,TrMatrix)
   * @see    #scale3D(double[],double[],double,double,TrMatrix)
   */
  public void scale3D ( double[] origin, double scale, TrMatrix inv )
  {
    scaleAxes3D( origin, scale, scale, scale, inv );
  }

  /**
   * Append general scale to the actual transform matrix. Optionally updates
   * the inverse matrix, too.
   *
   * @param  origin Origin of scale.
   * @param  direction Directional vector to be scaled.
   * @param  dirScale Scale factor for the given direction.
   * @param  perScale Scale factor for perpendicular directions.
   * @param  inv Optional inverse matrix.
   * @see    #scaleAxes3D(double[],double[],TrMatrix)
   * @see    #scale3D(double[],double,TrMatrix)
   */
  public void scale3D ( double[] origin, double[] direction,
                        double dirScale, double perScale, TrMatrix inv )
  {
    if ( origin    == null || origin.length    < 3 ||
         direction == null || direction.length < 3 ) return;

    TrMatrix m  = new TrMatrix( 4, 3 );
    TrMatrix mi = new TrMatrix( 4, 3 );
    m.moveToZAxis( origin, direction, mi );
    append( m );
    if ( inv != null )
      inv.prepend( mi );

    TrMatrix tmp = new TrMatrix( 4, 3 );
    tmp.m[0][0] = perScale;
    tmp.m[1][1] = perScale;
    tmp.m[2][2] = dirScale;
    append( tmp );
    if ( inv != null )
    {
      tmp.m[0][0] = 1.0 / perScale;
      tmp.m[1][1] = 1.0 / perScale;
      tmp.m[2][2] = 1.0 / dirScale;
      inv.prepend( tmp );
    }

    append( mi );
    if ( inv != null )
      inv.prepend( m );
  }

  /**
   * Returns the actual matrix in readable format.
   */
  public String toString ()
  {
    StringBuffer sb = new StringBuffer();
    sb.append( "[" + this.width + ',' + this.height + "]:(" );
    for ( int y = 0; y < height; y++ )
    {
      if ( y != 0 ) sb.append( ',' );
      sb.append( '(' );
      for ( int x = 0; x < width; x++ )
      {
        if ( x != 0 ) sb.append( ',' );
        sb.append( m[y][x] );
      }
      sb.append( ')' );
    }

    return sb.toString();
  }

  /**
   * Converts the matrix into 1D array of doubles (with optional OpenGL order).
   */
  public double[] toArray ( int wid, int hei, boolean rowMajor, double[] result )
  {
    if ( wid <= 0 ) wid = width;
    if ( hei <= 0 ) hei = height;

    if ( result == null ||
         result.length < wid * hei )
      result = new double[ wid * hei ];

    int ptr = 0;
    int col, row;
    if ( rowMajor )
      for ( row = 0; row < hei; row++ )
        for ( col = 0; col < wid; col++ )
          result[ ptr++ ] = (col < width && row < height) ? m[row][col] :
                            (col == row ? 1.0 : 0.0);
    else
      for ( col = 0; col < wid; col++ )
        for ( row = 0; row < hei; row++ )
          result[ ptr++ ] = (col < width && row < height) ? m[row][col] :
                            (col == row ? 1.0 : 0.0);

    return result;
  }

  /**
   * Appends the given translation to the actual transform matrix. Optionally updates the
   * inverse matrix, too.
   *
   * @param  vector Translation vector ({@code double[2]} or {@code double[3]}).
   * @param  inv Optional inverse matrix.
   */
  public void translate2D ( double[] vector, TrMatrix inv )
  {
    if ( vector == null ||
         vector.length < 2 ) return;

    TrMatrix tmp = new TrMatrix( 3, 2 );
    if ( vector.length == 2 ||
         vector[2] > -Geometry.EPSILON && vector[2] < Geometry.EPSILON )
      System.arraycopy( vector, 0, tmp.m[2], 0, 2 );

    else
    {
      tmp.m[2][0] = vector[0] / vector[2];
      tmp.m[2][1] = vector[1] / vector[2];
    }

    append( tmp );
    if ( inv != null )
    {
      tmp.m[2][0] = -tmp.m[2][0];
      tmp.m[2][1] = -tmp.m[2][1];
      inv.prepend( tmp );
    }
  }

  /**
   * Appends the given rotation to the actual transform matrix. Optionally updates the
   * inverse matrix, too.
   *
   * @param  sina Sine of the rotation angle.
   * @param  cosa Cosine of the rotation angle.
   * @param  inv Optional inverse matrix.
   */
  public void rotateOrigin2D ( double sina, double cosa, TrMatrix inv )
  {
    TrMatrix tmp = new TrMatrix( 3, 2 );

    tmp.m[0][0] =  cosa; tmp.m[0][1] = -sina;
    tmp.m[1][0] =  sina; tmp.m[1][1] =  cosa;

    append( tmp );
    if ( inv != null )
    {
      tmp.transpose();
      inv.prepend( tmp );
    }
  }

  /**
   * Computes viewing transformation matrix. Similar to {@code glLookAt()}.
   * Accumulates the viewing transform into the current matrix (uses {@link #append}).
   *
   * @param  eye Eye position in the world space.
   * @param  center Point at which the camera points.
   * @param  up The up-vector (will be transformed to vertical direction in the projection plane).
   */
  public void lookAt ( double[] eye, double[] center, double[] up )
  {
    if ( eye    == null || eye.length    < 3 ||
         center == null || center.length < 3 ) return;

    if ( up     == null || up.length     < 3 )
      up = new double[] { 0.0, 0.0, 1.0 };

    double[] z = Geometry.vector( eye, center, new double[ 4 ] );
    Geometry.normalize3D( z );
    double[] x = Geometry.vectorProduct( up, z, new double[ 4 ] );
    double[] y = Geometry.vectorProduct( z,  x, new double[ 4 ] );
    Geometry.normalize3D( x );
    Geometry.normalize3D( y );
    z[0] = -z[0];
    z[1] = -z[1];
    z[2] = -z[2];
    TrMatrix tmp = new TrMatrix( x, y, z, new double[] { 0.0, 0.0, 0.0, 1.0 } );
    tmp.transpose();
    translate3D( -eye[0], -eye[1], -eye[2], null );
    append( tmp );
  }

  /**
   * Computes perspective projection matrix. Similar to {@code glFrustum()}.
   * Accumulates the viewing transform into the current matrix (uses {@link #append}).
   * <br />Accepts {@code far == Double.MAX_VALUE}.
   */
  public void frustum ( double left, double right, double bottom, double top, double near, double far )
  {
    TrMatrix tmp = new TrMatrix( 4, 4 );

      // check frustum parameters:
    if ( left   >= right )
      right = left + 1.0;
    if ( bottom >= top )
      top   = bottom + 1.0;
    if ( near <= Geometry.EPSILON )
      near  = 0.001;
    if ( far  <= near )
      far   = near * 100.0;

    final double n2 = near + near;
    final double rl = right - left;
    final double tb = top - bottom;

    tmp.m[0][0] = n2 / rl;
    tmp.m[1][1] = n2 / tb;
    tmp.m[2][0] = (right + left) / rl;
    tmp.m[2][1] = (top + bottom) / tb;
      // [2][2] depends on finiteness of far
    tmp.m[2][3] = -1.0;
      // [3][2] depends on finiteness of far
    tmp.m[3][3] = 0.0;

    if ( far < Double.MAX_VALUE )
    {
      final double nf = near - far;
      tmp.m[2][2] = (near + far) / nf;
      tmp.m[3][2] = n2 * far / nf;
    }
    else
    {
      tmp.m[2][2] = -1.0;
      tmp.m[3][2] = -n2;
    }

    append( tmp );
  }

  /**
   * Computes perspective projection matrix. Similar to {@code gluPerspective()}.
   * Accumulates the viewing transform into the current matrix (uses {@link #append}).
   * <br />Accepts {@code far == Double.MAX_VALUE}.
   *
   * @param  fov Horizontal field-of-view angle in degrees.
   * @param  aspect Aspect ratio ({@code width / height} of the screen device).
   * @param  near Near Z-coordinate (must be positive).
   * @param  far Far Z-coordinate (must be greater than {@code near}, can be infinite
   *         {@code == Double.MAX_VALUE}).
   */
  public void perspective ( double fov, double aspect, double near, double far )
  {
    double xmin, xmax, ymin, ymax;

    xmax = near * Math.tan( 0.5 * Math.toRadians( fov ) );
    xmin = -xmax;
    ymin = xmin / aspect;
    ymax = xmax / aspect;

    frustum( xmin, xmax, ymin, ymax, near, far );
  }

  /**
   * Computes perspective projection matrix. Accurate {@code gluPerspective()} code.
   * Accumulates the viewing transform into the current matrix (uses {@link #append}).
   * <br />Accepts {@code far = Double.MAX_VALUE}.
   *
   * @param  fov Horizontal field-of-view angle in degrees.
   * @param  aspect Aspect ratio ({@code width / height} of the screen device).
   * @param  near Near Z-coordinate (must be positive).
   * @param  far Far Z-coordinate (must be greater than {@code near}, can be infinite
   *         {@code == Double.MAX_VALUE}).
   */
  public void perspectiveGL ( double fov, double aspect, double near, double far )
  {
      // check & fix the parameters:
    if ( fov <= Geometry.EPSILON )
      fov = 1.0;
    if ( near <= Geometry.EPSILON )
      near  = 0.001;
    if ( far  <= near )
      far   = near * 100.0;

    TrMatrix tmp = new TrMatrix( 4, 4 );
    final double f = 1.0 / Math.tan( 0.5 * Math.toRadians( fov ) );
    final double n2 = near + near;
    tmp.m[0][0] = f / aspect;
    tmp.m[1][1] = f;
    tmp.m[2][3] = -1.0;
    tmp.m[3][3] = 0.0;
    if ( far < Double.MAX_VALUE )
    {
      final double nf = near - far;
      tmp.m[2][2] = (near + far) / nf;
      tmp.m[3][2] = n2 * far / nf;
    }
    else
    {
      tmp.m[2][2] = -1.0;
      tmp.m[3][2] = -n2;
    }

    append( tmp );
  }

  /**
   * Computes orthographic projection matrix. Similar to {@code glOrtho()}.
   * Accumulates the viewing transform into the current matrix (uses {@link #append}).
   * <br />Accepts {@code far == Double.MAX_VALUE}.
   */
  public void ortho ( double left, double right, double bottom, double top, double near, double far )
  {
    TrMatrix tmp = new TrMatrix( 4, 4 );

      // check parameters:
    if ( left   >= right )
      right = left + 1.0;
    if ( bottom >= top )
      top   = bottom + 1.0;
    if ( near <= Geometry.EPSILON )
      near  = 0.001;
    if ( far  <= near )
      far   = near * 100.0;

    final double rl = right - left;
    final double tb = top - bottom;
    final double fn = far - near;

    tmp.m[0][0] =  2.0 / rl;
    tmp.m[1][1] =  2.0 / tb;
    tmp.m[2][2] = -2.0 / fn;
    tmp.m[3][0] = (-right - left) / rl;
    tmp.m[3][1] = (-top - bottom) / tb;
    if ( far < Double.MAX_VALUE )
      tmp.m[3][2] = (far + near) / fn;
    else
      tmp.m[3][2] = 1.0;

    append( tmp );
  }

  /**
   * Computes final scaling 2D transform (viewport transform) mapping NDS
   * (Normalized Device Space) into the given screen coordinate range.
   * Similar to {@code glViewport()}.
   * Accumulates the scaling transform into the current matrix (uses {@link #append}).
   *
   * <p>Z-coordinate (depth) is transformed into {@code [ 0.0, 1.0 ]} range,
   * {@code 0.0} corresponds to the far clipping plane.
   */
  public void viewport ( double x0, double y0, double width, double height )
  {
    translate3D( 1.0, 1.0, 1.0, null );
    scaleAxes3D( null, 0.5 * width, -0.5 * height, 0.5, null );
    translate3D( x0, y0 + height, 0.0, null );
  }

  /** Self-test. */
  public static void main ( String[] args )
  {
    TrMatrix m = new TrMatrix( 4, 4 );
    m.m[0][1] =  0.8;
    m.m[1][2] = -0.6;
    m.log( "Shear " );
    m.translate3D( 1.0, 2.0, 3.0, null );
    m.rotateAxis3D( 1, Math.sin( 1.2 ), Math.cos( 1.2 ), null );
    m.rotateAxis3D( 0, Math.sin( 0.4 ), Math.cos( 0.4 ), null );
    m.log( "Shear/translate/rotate^2 " );
    TrMatrix n = new TrMatrix( m );
    n.invert( (TrMatrix)null );
    n.log( "Inverted matrix " );
    n.append( m );
    n.log( "Unit " );
    n.set( m );
    n.setNormalTransform3D( null );
    n.log( "Vector-transform " );
    double[] a = { 1.0, -2.0, 2.5, 1.0 };
    double[] b = { 1.6,  1.0, 1.1, 1.0 };
    double[] v = Geometry.vector( a, b, null );
    LogFile.log( "A = " + Geometry.vectorToString( a ) );
    LogFile.log( "B = " + Geometry.vectorToString( b ) );
    LogFile.log( "V = " + Geometry.vectorToString( v ) );
    m.transformPoint3D( a, a );
    LogFile.log( "T(A) = " + Geometry.vectorToString( a ) );
    m.transformPoint3D( b, b );
    LogFile.log( "T(B) = " + Geometry.vectorToString( b ) );
    n.transformVector3D( v, v );
    LogFile.log( "T(V) = " + Geometry.vectorToString( v ) );
    v = Geometry.vector( a, b, v );
    LogFile.log( "T(B)-T(A) = " + Geometry.vectorToString( v ) );
    a[0] = 1.0;
    a[1] =
    a[2] =
    a[3] = 0.0;
    b[1] = 1.0;
    b[0] =
    b[2] =
    b[3] = 0.0;
    Geometry.vectorProduct( a, b, v );
    LogFile.log( "X x Y = " + Geometry.vectorToString( v ) );
    a[2] = 1.0;
    a[0] =
    a[1] =
    a[3] = 0.0;
    Geometry.vectorProduct( b, a, v );
    LogFile.log( "Y x Z = " + Geometry.vectorToString( v ) );
    b[0] = 1.0;
    b[1] =
    b[2] =
    b[3] = 0.0;
    Geometry.vectorProduct( a, b, v );
    LogFile.log( "Z x X = " + Geometry.vectorToString( v ) );
  }

}
