package cz.cuni.jagrlib;

import java.util.*;
import java.awt.geom.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.BadInterfaceException;

/**
 * Concrete vector commands - individual commands are declared as inner classes of
 * {@link VCommand}. The {@link VCommand} can be used as container for
 * a linear sequence of {@link VectorCommand} objects (subroutine = group of commands).
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.02
 * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
 * @see     cz.cuni.jagrlib.iface.VectorGraphics interface VectorGraphics
 */
public class VCommand extends VectorCommand
{
  private static final long serialVersionUID = 1L;

  /**
   * List of sub-commands.
   */
  protected List< VectorCommand > commands;

  /**
   * Transforms the command using the given {@link AffineTransform}.
   *
   * @param t  Affine transformation (3x2 matrix of doubles) or {@code null}.
   */
  public void transform ( AffineTransform t )
  {
    if ( t == null || commands == null ) return;

    Iterator< VectorCommand > it = commands.iterator();
    while ( it.hasNext() )
      it.next().transform( t );
  }

  /**
   * Executes the command on the given {@link VectorGraphics} interface.
   *
   * @param  out Output vector interface.
   * @throws BadInterfaceException
   * @see    VectorGraphics
   */
  public void execute ( VectorGraphics out )
  throws BadInterfaceException
  {
    if ( out == null || commands == null ) return;

    Iterator< VectorCommand > it = commands.iterator();
    while ( it.hasNext() )
      it.next().execute( out );
  }

  /**
   * Starts a new <b>path</b> (an old path is destroyed).
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.PushPath
   * @see     VCommand.PopPath
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#newPath
   */
  public class NewPath extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    public String toString ()
    {
      return "NewPath";
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.newPath();
    }

  }

  /**
   * Moves current point to the given coordinate.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.LineTo
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#moveTo
   */
  public class MoveTo extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * X coordinate.
     *
     * @serial
     */
    public double x;

    /**
     * Y coordinate.
     *
     * @serial
     */
    public double y;

    /**
     * Initializing constructor.
     *
     * @param  _x X coordinate.
     * @param  _y Y coordinate.
     */
    public MoveTo ( double _x, double _y )
    {
      x = _x;
      y = _y;
    }

    /**
     * Transforms the command using the given {@link AffineTransform}.
     *
     * @param  t Affine transformation (3x2 matrix of doubles) or {@code null}.
     */
    public void transform ( AffineTransform t )
    {
      if ( t == null ) return;

      Point2D.Double p = new Point2D.Double( x, y );
      t.transform( p, p );
      x = p.x;
      y = p.y;
    }

    public String toString ()
    {
      return "MoveTo(" + x + ',' + y + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.moveTo( x, y );
    }

  }

  /**
   * Draw line from tho current point to the given coordinate.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.MoveTo
   * @see     VCommand.Stroke
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#lineTo
   */
  public class LineTo extends MoveTo
  {
    private static final long serialVersionUID = 1L;

    /**
     * Initializing constructor.
     *
     * @param  _x X coordinate.
     * @param  _y Y coordinate.
     */
    public LineTo ( double _x, double _y )
    {
      super( _x, _y );
    }

    public String toString ()
    {
      return "LineTo(" + x + ',' + y + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.lineTo( x, y );
    }

  }

  /**
   * Draw a circular arc (counterclockwise) and connects it to the current point.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.RArcTo
   * @see     VCommand.MoveTo
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#arcTo
   */
  public class ArcTo extends MoveTo
  {
    private static final long serialVersionUID = 1L;

    /**
     * Arc radius.
     *
     * @serial
     */
    public double r;

    /**
     * Starting angle in degrees.
     *
     * @serial
     */
    public double alpha;

    /**
     * Finishing angle in degrees.
     *
     * @serial
     */
    public double beta;

    /**
     * Initializing constructor.
     *
     * @param  _x X coordinate of arc center.
     * @param  _y Y coordinate of arc center.
     * @param  _r Arc radius.
     * @param  _alpha Starting angle in degrees.
     * @param  _beta Finishing angle in degrees.
     */
    public ArcTo ( double _x, double _y, double _r, double _alpha, double _beta )
    {
      super( _x, _y );
      r     = _r;
      alpha = _alpha;
      beta  = _beta;
    }

    /**
     * Transforms the command using the given {@link AffineTransform}.
     *
     * @param  t Affine transformation (3x2 matrix of doubles) or <tt>null</tt>.
     */
    public void transform ( AffineTransform t )
    {
      if ( t == null ) return;

      super.transform( t );             // arc center
      // !!! not yet !!!
                                        // bounding arc angles
    }

    public String toString ()
    {
      return "ArcTo(" + x + ',' + y + ',' + r + ',' + alpha + ',' + beta + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.arcTo( x, y, r, alpha, beta );
    }

  }

  /**
   * Draw a circular arc (clockwise) and connects it to the current point.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.ArcTo
   * @see     VCommand.MoveTo
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#rArcTo
   */
  public class RArcTo extends ArcTo
  {
    private static final long serialVersionUID = 1L;

    /**
     * Initializing constructor.
     *
     * @param  _x X coordinate of arc center.
     * @param  _y Y coordinate of arc center.
     * @param  _r Arc radius.
     * @param  _alpha Starting angle in degrees.
     * @param  _beta Finishing angle in degrees.
     */
    public RArcTo ( double _x, double _y, double _r, double _alpha, double _beta )
    {
      super( _x, _y, _r, _alpha, _beta );
    }

    public String toString ()
    {
      return "RArcTo(" + x + ',' + y + ',' + r + ',' + alpha + ',' + beta + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.rArcTo( x, y, r, alpha, beta );
    }

  }

  /**
   * Draw a cubic Bezier curve from the current point.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.MoveTo
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#cubicCurveTo
   */
  public class CubicCurveTo extends MoveTo
  {
    private static final long serialVersionUID = 1L;

    /**
     * X coordinate of the third control point (X<sub>2</sub>).
     *
     * @serial
     */
    public double x2;

    /**
     * Y coordinate of the third control point (Y<sub>2</sub>).
     *
     * @serial
     */
    public double y2;

    /**
     * X coordinate of the fourth control point = curve end (X<sub>3</sub>).
     *
     * @serial
     */
    public double x3;

    /**
     * Y coordinate of the fourth control point = curve end (Y<sub>3</sub>).
     *
     * @serial
     */
    public double y3;

    /**
     * Initializing constructor.
     *
     * @param  _x1 X coordinate of the 2nd control point.
     * @param  _y1 Y coordinate of the 2nd control point.
     * @param  _x2 X coordinate of the 3rd control point.
     * @param  _y2 Y coordinate of the 3rd control point.
     * @param  _x3 X coordinate of the 4th control point.
     * @param  _y3 Y coordinate of the 4th control point.
     */
    public CubicCurveTo ( double _x1, double _y1, double _x2, double _y2, double _x3, double _y3 )
    {
      super( _x1, _y1 );
      x2 = _x2;
      y2 = _y2;
      x3 = _x3;
      y3 = _y3;
    }

    public String toString ()
    {
      return "CubicCurveTo(" + x + ',' + y + ',' + x2 + ',' + y2 + ',' + x3 + ',' + y3 + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.cubicCurveTo( x, y, x2, y2, x3, y3 );
    }

  }

  /**
   * Closes the current <b>path</b>.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.NewPath
   * @see     VCommand.Stroke
   * @see     VCommand.Fill
   * @see     VCommand.Clip
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#closePath
   */
  public class ClosePath extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    public String toString ()
    {
      return "ClosePath";
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.closePath();
    }

  }

  /**
   * Strokes the current <b>path</b> (the current path is destroed - use
   * {@link VCommand.PushPath} and {@link VCommand.PopPath}
   * to restore it).
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.PushPath
   * @see     VCommand.PopPath
   * @see     VCommand.Fill
   * @see     VCommand.Clip
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#stroke
   */
  public class Stroke extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    public String toString ()
    {
      return "Stroke";
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.stroke();
    }

  }

  /**
   * Fills interior of the current <b>path</b> (the current path is destroed - use
   * {@link VCommand.PushPath} and {@link VCommand.PopPath}
   * to restore it). In case of non-continuous <b>path</b> several line segments will
   * be added...
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.PushPath
   * @see     VCommand.PopPath
   * @see     VCommand.Stroke
   * @see     VCommand.Clip
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#fill
   */
  public class Fill extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    public String toString ()
    {
      return "Fill";
    }

    /**
     * Executes the command on the given {@link VectorGraphics}<tt> interface</tt>.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.fill();
    }

  }

  /**
   * Sets the current <b>path</b> as the <b>clipping region</b> (the current path is destroed - use
   * {@link VCommand.PushPath} and {@link VCommand.PopPath}
   * to restore it). In case of non-continuous <b>path</b> several line segments will
   * be added...
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.PushPath
   * @see     VCommand.PopPath
   * @see     VCommand.Stroke
   * @see     VCommand.Fill
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#clip
   */
  public class Clip extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    public String toString ()
    {
      return "Clip";
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.clip();
    }

  }

  /**
   * Pushes the current <b>path</b> on the <b>path stack</b>. Later it can be restored
   * using {@link VCommand.PopPath} command.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.PopPath
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#pushPath
   */
  public class PushPath extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    public String toString ()
    {
      return "PushPath";
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.pushPath();
    }

  }

  /**
   * Pops one item from the <b>path stack</b> and assigns the current <b>path</b>.
   * The old current path is destroyed.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.PushPath
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#popPath
   */
  public class PopPath extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    public String toString ()
    {
      return "PopPath";
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.popPath();
    }

  }

  /**
   * Sets the actual drawing color (optionally with alpha-component).
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.Stroke
   * @see     VCommand.Fill
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#setColor
   */
  public class SetColor extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * Color components Gr[A] or RGB[A] (each component lies between
     * {@code 0.0} and {@code 1.0}).
     *
     * @serial
     */
    public double[] color;

    /**
     * Initializing constructor.
     *
     * @param  _color Color (Gr[A] or RGB[A]).
     */
    public SetColor ( double[] _color )
    {
      color = _color;
    }

    public String toString ()
    {
      StringBuffer result = new StringBuffer( "SetColor(" );
      result.append( color[0] );
      int i;
      for ( i = 1; i < color.length; i++ )
        result.append( ',' )
              .append( color[i] );

      return result.toString();
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.setColor( color );
    }

  }

  /**
   * Sets the actual stroke width.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.Stroke
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#setLineWidth
   */
  public class SetLineWidth extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * Stroke width.
     *
     * @serial
     */
    public double width;

    /**
     * Initializing constructor.
     *
     * @param  _width Stroke width.
     */
    public SetLineWidth ( double _width )
    {
      width = _width;
    }

    public String toString ()
    {
      return "SetLineWidth(" + width + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.setLineWidth( width );
    }

  }

  /**
   * Sets the actual stroke style.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.Stroke
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#setLineStyle
   */
  public class SetLineStyle extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * Stroke style (uses {@code VectorCommand.LINE_*} constants).
     *
     * @serial
     * @see    VectorCommand#LINE_DUMMY
     */
    public int style;

    /**
     * Initializing constructor.
     *
     * @param  _style Stroke style (use {@code VectorCommand.LINE_*} constant).
     */
    public SetLineStyle ( int _style )
    {
      style = _style;
    }

    public String toString ()
    {
      return "SetLineStyle(" + style + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.setLineStyle( style );
    }

  }

  /**
   * Sets the actual fill style.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.Fill
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#setFillStyle
   */
  public class SetFillStyle extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * Fill style (uses {@code VectorCommand.FILL_*} constants).
     *
     * @serial
     * @see    VectorCommand#FILL_DUMMY
     */
    public int style;

    /**
     * Initializing constructor.
     *
     * @param  _style Fill style (use {@code VectorCommand.FILL_*} constant).
     */
    public SetFillStyle ( int _style )
    {
      style = _style;
    }

    public String toString ()
    {
      return "SetFillStyle(" + style + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.setFillStyle( style );
    }

  }

  /**
   * Sets the actual stroke callback routine.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.Stroke
   * @see     VCommand.SetLineStyle
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#setLineCallback
   */
  public class SetLineCallback extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * Stroke callback (object implementing {@link StrokeCallback} interface).
     *
     * @serial
     * @see    cz.cuni.jagrlib.iface.StrokeCallback
     */
    public StrokeCallback callback;

    /**
     * Initializing constructor.
     *
     * @param  _callback Stroke callback (implementing {@link StrokeCallback} interface).
     */
    public SetLineCallback ( StrokeCallback _callback )
    {
      callback = _callback;
    }

    public String toString ()
    {
      return "SetLineCallback(" + callback.toString() + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.setLineCallback( callback );
    }

  }

  /**
   * Sets the actual fill callback routine.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.Fill
   * @see     VCommand.SetFillStyle
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#setFillCallback
   */
  public class SetFillCallback extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * Fill callback (object implementing {@link FillCallback} interface).
     *
     * @serial
     * @see    cz.cuni.jagrlib.iface.FillCallback
     */
    public FillCallback callback;

    /**
     * Initializing constructor.
     *
     * @param  _callback Fill callback (implementing {@link FillCallback} interface).
     */
    public SetFillCallback ( FillCallback _callback )
    {
      callback = _callback;
    }

    public String toString ()
    {
      return "SetFillCallback(" + callback.toString() + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.setFillCallback( callback );
    }

  }

  /**
   * Creates the new font instance (from: family, style, size).
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.FontOrientation
   * @see     VCommand.GetFont
   * @see     VCommand.PutFont
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#createFont
   * @see     cz.cuni.jagrlib.iface.TextRender
   */
  public class CreateFont extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * Font family name.
     *
     * @serial
     */
    public String name;

    /**
     * Font style flags (bitwise OR of {@code TextRender.FONT_*} constants).
     *
     * @serial
     * @see    cz.cuni.jagrlib.iface.TextRender#FONT_BOLD
     */
    public int style;

    /**
     * Font size (distance between two regularly spaced text lines).
     *
     * @serial
     */
    public double size;

    /**
     * Initializing constructor.
     *
     * @param  _name Font family name.
     * @param  _style Font style flags (bitwise OR of {@code TextRender.FONT_*} constants).
     * @param  _size Font size (distance between two regularly spaced text lines).
     * @see    cz.cuni.jagrlib.iface.TextRender
     * @see    cz.cuni.jagrlib.iface.TextRender#FONT_BOLD
     */
    public CreateFont ( String _name, int _style, double _size )
    {
      name = _name;
      style = _style;
      size = _size;
    }

    public String toString ()
    {
      return "CreateFont(\"" + name + "\"," + style + ',' + size + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.createFont( name, style, size );
    }

  }

  /**
   * Changes the current font's orientation.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.CreateFont
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#fontOrientation
   * @see     cz.cuni.jagrlib.iface.TextRender
   */
  public class FontOrientation extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * Font angle (in degrees, counterclockwise).
     *
     * @serial
     */
    public double angle;

    /**
     * Initializing constructor.
     *
     * @param  _angle Font angle (in degrees, counterclockwise).
     * @see    cz.cuni.jagrlib.iface.TextRender
     */
    public FontOrientation ( double _angle )
    {
      angle = _angle;
    }

    public String toString ()
    {
      return "FontOrientation(" + angle + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.fontOrientation( angle );
    }

  }

  /**
   * Assigns the actual font (<b>slot A</b>) from the given <b>slot</b>.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.CreateFont
   * @see     VCommand.PutFont
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#getFont
   * @see     cz.cuni.jagrlib.iface.TextRender
   */
  public class GetFont extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * Font slot (from {@code 0} to {@link TextRender#MAX_SLOT}{@code -1}).
     *
     * @serial
     * @see    cz.cuni.jagrlib.iface.TextRender#MAX_SLOT
     */
    public int slot;

    /**
     * Initializing constructor.
     *
     * @param  _slot Font slot (from {@code 0} to {@link TextRender#MAX_SLOT}{@code -1}).
     * @see    cz.cuni.jagrlib.iface.TextRender#MAX_SLOT
     */
    public GetFont ( int _slot )
    {
      slot = _slot;
    }

    public String toString ()
    {
      return "GetFont(" + slot + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.getFont( slot );
    }

  }

  /**
   * Puts the actual font (<b>slot A</b>) into the given <b>slot</b>.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.CreateFont
   * @see     VCommand.GetFont
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#putFont
   * @see     cz.cuni.jagrlib.iface.TextRender
   */
  public class PutFont extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * Font slot (from {@code 0} to {@link TextRender#MAX_SLOT}{@code -1}).
     *
     * @serial
     * @see    cz.cuni.jagrlib.iface.TextRender#MAX_SLOT
     */
    public int slot;

    /**
     * Initializing constructor.
     *
     * @param  _slot Font slot (from {@code 0} to {@link TextRender#MAX_SLOT}{@code -1}).
     * @see    cz.cuni.jagrlib.iface.TextRender#MAX_SLOT
     */
    public PutFont ( int _slot )
    {
      slot = _slot;
    }

    public String toString ()
    {
      return "PutFont(" + slot + ')';
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.putFont( slot );
    }

  }

  /**
   * Generates path of the given text string and <u>appends</u> it to the current path.
   * Uses the <b>current point</b> as starting reference point (<b>current point</b>
   * itself will be updated by this command).
   *
   * <p>Does nothing if the actual font has neither {@link TextRender#PROP_STROKE} nor
   * {@link TextRender#PROP_FILL} capability.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.CreateFont
   * @see     VCommand.FontOrientation
   * @see     VCommand.GetFont
   * @see     VCommand.MoveTo
   * @see     VCommand.DrawText
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#textPath
   * @see     cz.cuni.jagrlib.iface.TextRender
   */
  public class TextPath extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * String to be drawn.
     *
     * @serial
     */
    public String text;

    /**
     * Initializing constructor.
     *
     * @param  _text String to be drawn.
     */
    public TextPath ( String _text )
    {
      text = _text;
    }

    public String toString ()
    {
      return "TextPath(\"" + text + "\")";
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.textPath( text );
    }

  }

  /**
   * Draws the given text using <b>current font</b> and <b>current color</b>.
   * Uses the <b>current point</b> as starting reference point (<b>current point</b>
   * itself will be updated by this command).
   *
   * <p>Works with both vector and bitmap fonts.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../cz/cuni/jagrlib/VCommand.java">VCommand.java</a>
   * @see     VCommand.CreateFont
   * @see     VCommand.FontOrientation
   * @see     VCommand.GetFont
   * @see     VCommand.MoveTo
   * @see     VCommand.SetColor
   * @see     VCommand.TextPath
   * @see     cz.cuni.jagrlib.iface.VectorGraphics#drawText
   * @see     cz.cuni.jagrlib.iface.TextRender
   */
  public class DrawText extends VectorCommand
  {
    private static final long serialVersionUID = 1L;

    /**
     * String to be drawn.
     *
     * @serial
     */
    public String text;

    /**
     * Initializing constructor.
     *
     * @param  _text String to be drawn.
     */
    public DrawText ( String _text )
    {
      text = _text;
    }

    public String toString ()
    {
      return "DrawText(\"" + text + "\")";
    }

    /**
     * Executes the command on the given {@link VectorGraphics} interface.
     *
     * @param  out Output vector interface.
     * @throws BadInterfaceException
     * @see    VectorGraphics
     */
    public void execute ( VectorGraphics out )
    throws BadInterfaceException
    {
      out.drawText( text, null );
    }

  }

}
