package cz.cuni.jagrlib.eval;

import java.util.*;
import java.io.PrintStream;
import java.io.File;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;

/**
 * Evaluation plugin for MainEval: exercise 01, simple picture generation.
 *
 * @version 0.26 $Rev: 439 $ $Date: 2007-05-05 12:29:50 +0200 (so, 05 V 2007) $ $Author: pepca $
 * @since   0.24
 * @see     MainEval
 * @see     <a href="../../../../../cz/cuni/jagrlib/eval/Ex2005_01_a.java">Ex2005_01_a.java</a>
 */
public class Ex2005_01_a extends DefaultPlugin
{
  //==========================================================================
  //  XML params:

  /** Number of items to draw. */
  protected final static String REPEAT = "repeat";
  /** Random seed (for RandomJames). */
  protected final static String SEED1  = "seed1";
  protected final static String SEED2  = "seed2";
  /** Use integer coordinates? */
  protected final static String INT    = "int";

  /**
   * Runs one instance of the test (simple variant, w/o global result storage).
   */
  @Override
  public boolean runTest ( Map< String, String > env, Map< String, String > param, PrintStream out )
  {
      // class name:
    String shortName = param.get( CLASS_NAME );
    if ( shortName == null )
    {
      out.println( "Cannot find the '" + CLASS_NAME + "' test argument!" );
      return false;
    }
    String className = Template.JAGRLIB2 + shortName;

      // picture file-name:
    String outputFileName = param.get( OUTPUT_FILE_NAME );
    if ( outputFileName == null )
      outputFileName = "output.png";
    String tmp = env.get( BASE );
    if ( tmp != null )
    {
      File baseDir = new File( tmp );
      baseDir.mkdirs();
      outputFileName = tmp.concat( outputFileName );
    }

      // repeat:
    int repeat = integerParam( env, param, REPEAT, 200 );

      // random seeds:
    int seed1 = integerParam( env, param, SEED1,  0 );
    int seed2 = integerParam( env, param, SEED2, 12 );

      // use integer coordinates?
    boolean integer = integerParam( env, param, INT, 0 ) != 0;

      // create Piece objects:
    Piece line;
    try
    {
      line = (Piece)(Class.forName( className ).newInstance());
    }
    catch ( Exception e )
    {
      out.println( "Cannot instantiate the '" + shortName + "' class!" );
      return false;
    }
    if ( !(line instanceof LineRenderAnti) )
    {
      out.println( "The '" + shortName + "' class doesn't implement LineRenderAnti interface!" );
      return false;
    }
    AlphaMatrix     alpha  = new AlphaMatrix();
    SolidColorBrush brush  = new SolidColorBrush();
    RasterImage     image  = new RasterImage();
    PNGFileFormat   png    = new PNGFileFormat();
    JavaBitStream   output = new JavaBitStream();
    output.set( JavaBitStream.STREAM_NAME, outputFileName );

      // connect them by channels:
    try
    {
      line.connect(  Piece.PL_OUTPUT,    alpha,  Piece.PL_INPUT );
      brush.connect( Piece.PL_ALPHAMASK, alpha,  Piece.PL_INPUT );
      brush.connect( Piece.PL_OUTPUT,    image,  Piece.PL_INPUT );
      png.connect(   Piece.PL_RASTER,    image,  Piece.PL_INPUT );
      png.connect(   Piece.PL_STREAM,    output, Piece.PL_INPUT );
    }
    catch ( BadInterfaceException e )
    {
      out.println( "Error connecting the '" + shortName + "' module!" );
      LogFile.exception( "Bad Interface in connecting-time!", e );
      return false;
    }

      // set Pieces' parameters:
    image.init( 512, 512, RasterGraphics.MODE_RGBA, 0 );
    alpha.init( 512, 512 );

    int[] color = new int[4];

      // random generator:
    RandomJames rnd = new RandomJames( seed1, seed2 );
    LineRenderAnti l = (LineRenderAnti)line;

      // time measurement:
    long startTime = System.currentTimeMillis();

      // draw random lines:
    try
    {
      for ( int i = 0; i++ < repeat; )
      {
          // random line:
        double x1 = rnd.uniformNumber() * 511.0;
        double y1 = rnd.uniformNumber() * 511.0;
        double x2 = rnd.uniformNumber() * 511.0;
        double y2 = rnd.uniformNumber() * 511.0;
        if ( integer )
        {
          x1 = Math.round( x1 );
          y1 = Math.round( y1 );
          x2 = Math.round( x2 );
          y2 = Math.round( y2 );
        }
        l.drawLine( x1, y1, x2, y2, rnd.randomDouble( 0.5, 3.5 ) );
          // stroke it:
        if ( (i & 15) == 0 )
        {
            // set random color:
          color[0] = rnd.randomInteger( 54, 255 );
          color[1] = rnd.randomInteger( 54, 255 );
          color[2] = rnd.randomInteger( 54, 255 );
          brush.setColor( color );
          brush.fill();
        }
      }
    }
    catch ( Exception e )
    {
      LogFile.exception( e );
    }

      // set random color:
    color[0] = rnd.randomInteger( 54, 255 );
    color[1] = rnd.randomInteger( 54, 255 );
    color[2] = rnd.randomInteger( 54, 255 );
    brush.setColor( color );
    brush.fill();

      // stop time:
    long stopTime = System.currentTimeMillis();

      // save the result:
    try
    {
      png.saveFile( (String)null, null );
    }
    catch ( Exception e )
    {
      LogFile.error( "Error writting '" + outputFileName + "' image!" );
    }

    out.println( String.format( Locale.US, "OK (%5.3fs) %s", 1.e-3 * (stopTime - startTime), shortName ) );
    return true;
  }

}
