package cz.cuni.jagrlib.eval;

import java.util.*;
import java.io.PrintStream;
import java.io.File;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;

/**
 * Evaluation plugin for MainEval: exercise 02, lines with non-integer coordinates, w/o antialiasing.
 *
 * @version 0.26 $Rev: 439 $ $Date: 2007-05-05 12:29:50 +0200 (so, 05 V 2007) $ $Author: pepca $
 * @since   0.24
 * @see     MainEval
 * @see     <a href="../../../../../cz/cuni/jagrlib/eval/Ex2005_02_a.java">Ex2005_02_a.java</a>
 */
public class Ex2005_02_a extends DefaultPlugin
{
  //==========================================================================
  //  XML params:

  /** Number of items to draw. */
  protected final static String REPEAT = "repeat";
  /** Random seed (for RandomJames). */
  protected final static String SEED1  = "seed1";
  protected final static String SEED2  = "seed2";
  /** Circle center. */
  protected final static String CX     = "cx";
  protected final static String CY     = "cy";
  /** Circle radius. */
  protected final static String RAD    = "rad";
  /** Allow lines going exactly through half-points? */
  protected final static String HALF   = "half";

  /**
   * Runs one instance of the test (simple variant, w/o global result storage).
   */
  @Override
  public boolean runTest ( Map< String, String > env, Map< String, String > param, PrintStream out )
  {
      // class name:
    String shortName = param.get( CLASS_NAME );
    if ( shortName == null )
    {
      out.println( "Cannot find the '" + CLASS_NAME + "' test argument!" );
      return false;
    }
    String className = Template.JAGRLIB2 + shortName;

      // picture file-name:
    String outputFileName = param.get( OUTPUT_FILE_NAME );
    if ( outputFileName == null )
      outputFileName = "output.png";
    String tmp = env.get( BASE );
    if ( tmp != null )
    {
      File baseDir = new File( tmp );
      baseDir.mkdirs();
      outputFileName = tmp.concat( outputFileName );
    }

      // repeat:
    int repeat = integerParam( env, param, REPEAT, 200 );

      // random seeds:
    int seed1 = integerParam( env, param, SEED1,  0 );
    int seed2 = integerParam( env, param, SEED2, 12 );

      // allow half-pixels?
    boolean half = integerParam( env, param, HALF, 0 ) != 0;

      // clipping circle:
    double cx   = doubleParam( env, param, CX,  256.0 );
    double cy   = doubleParam( env, param, CY,  256.0 );
    double rad2 = doubleParam( env, param, RAD, 250.0 );
    rad2 *= rad2;

      // create Piece objects:
    Piece line;
    try
    {
      line = (Piece)(Class.forName( className ).newInstance());
    }
    catch ( Exception e )
    {
      out.println( "Cannot instantiate the '" + shortName + "' class!" );
      return false;
    }
    if ( !(line instanceof LineRender) )
    {
      out.println( "The '" + shortName + "' class doesn't implement LineRender interface!" );
      return true;
    }
    XTransitionList mask   = new XTransitionList();
    LineBresenham   exact  = new LineBresenham();
    SolidColorPen   pen    = new SolidColorPen();
    RasterImage     image  = new RasterImage();
    PNGFileFormat   png    = new PNGFileFormat();
    JavaBitStream   output = new JavaBitStream();
    output.set( JavaBitStream.STREAM_NAME, outputFileName );

      // connect them by channels:
    try
    {
      line.connect(  Piece.PL_OUTPUT,  mask,   Piece.PL_INPUT );
      exact.connect( Piece.PL_OUTPUT,  mask,   Piece.PL_INPUT );
      pen.connect(   Piece.PL_BITMASK, mask,   Piece.PL_INPUT );
      pen.connect(   Piece.PL_OUTPUT,  image,  Piece.PL_INPUT );
      png.connect(   Piece.PL_RASTER,  image,  Piece.PL_INPUT );
      png.connect(   Piece.PL_STREAM,  output, Piece.PL_INPUT );
    }
    catch ( BadInterfaceException e )
    {
      out.println( "Error connecting the '" + shortName + "' module!" );
      LogFile.exception( "Bad Interface in connecting-time!", e );
      return false;
    }

      // set Pieces' parameters:
    image.init( 512, 512, RasterGraphics.MODE_RGB, 0 );
    mask.init( 512, 512 );

    int[] color1 = { 255, 255, 192 };
    int[] color2 = {   0,   0,  96 };

      // random generator:
    RandomJames rnd = new RandomJames( seed1, seed2 );
    LineRender l    = (LineRender)line;

      // time measurement:
    long startTime = System.currentTimeMillis();

      // draw random lines and clipped ones:
    try
    {
      for ( int i = 0; i++ < repeat; )
      {
          // random line coordinates:
        int x1, y1, x2, y2;
        int dx, dy;
        do
        {
          x1 = (int)( rnd.uniformNumber() * 512.0 );
          y1 = (int)( rnd.uniformNumber() * 512.0 );
          x2 = (int)( rnd.uniformNumber() * 512.0 );
          y2 = (int)( rnd.uniformNumber() * 512.0 );
          if ( half ) break;
          dx = Math.abs( x2 - x1 );
          dy = Math.abs( y2 - y1 );
          if ( dx > dy ) dy += dy;
          else
          if ( dy > dx ) dx += dx;
        } while ( Formula.gcd( dx, dy ) > 1 );

          // original (exact) line:
        exact.drawLine( x1, y1, x2, y2 );
        pen.setColor( color1 );
        pen.stroke();

          // clip it:
        double xx1 = x1;
        double yy1 = y1;
        double xx2 = x2;
        double yy2 = y2;
        double sx = xx2 - xx1;
        double sy = yy2 - yy1;
          // (xx1 + t*sx - cx)^2 + (yy1 + t*sy - cy)^2 - rad2 = 0
          // t^2 * (sx^2 + sy^2) + t * (2*sx*(xx1-cx) + 2*sy*(yy1-cy)) + (xx1-cx)^2 + (yy1-cy)^2 - rad2 = 0
        double xcx = xx1 - cx;
        double ycy = yy1 - cy;
        double a = sx * sx + sy * sy;
        double b = 2.0 * (sx * xcx + sy * ycy);
        double c = xcx * xcx + ycy * ycy - rad2;
        double d = b * b - 4.0 * a * c;
        if ( a > Geometry.SCENE_EPSILON && d > 0.0 )
        {
          d = Math.sqrt( d );
          a = 0.5 / a;
          double t = (-b + d) * a;
          if ( t < 1.0 )                  // move ending point
          {
            if ( t <= 0.0 ) continue;
            xx2 = xx1 + t * sx;
            yy2 = yy1 + t * sy;
          }
          t = (-b - d) * a;
          if ( t > 0.0 )                  // move starting point
          {
            if ( t >= 1.0 ) continue;
            xx1 += t * sx;
            yy1 += t * sy;
          }

            // .. and stroke it using overwrite-color:
          l.drawLine( xx1, yy1, xx2, yy2 );
          pen.setColor( color2 );
          pen.stroke();
        }
      }
    }
    catch ( Exception e )
    {
      LogFile.exception( e );
    }

      // stop time:
    long stopTime = System.currentTimeMillis();

      // save the result:
    try
    {
      png.saveFile( (String)null, null );
    }
    catch ( Exception e )
    {
      LogFile.error( "Error writting '" + outputFileName + "' image!" );
    }

    out.println( String.format( Locale.US, "OK (%5.3fs) %s", 1.e-3 * (stopTime - startTime), shortName ) );
    return true;
  }

}
