package cz.cuni.jagrlib.eval;

import java.util.*;
import java.io.PrintStream;
import java.io.File;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;

/**
 * Evaluation plugin for MainEval: exercise 43, general ellipses.
 *
 * @version 0.26 $Rev: 439 $ $Date: 2007-05-05 12:29:50 +0200 (so, 05 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     MainEval
 * @see     <a href="../../../../../cz/cuni/jagrlib/eval/Ex2006_43_a.java">Ex2006_43_a.java</a>
 */
public class Ex2006_43_a extends DefaultPlugin
{
  //==========================================================================
  //  XML params:

  /** Number of items to draw. */
  protected final static String REPEAT = "repeat";

  /** Random seed (for RandomJames). */
  protected final static String SEED1  = "seed1";
  protected final static String SEED2  = "seed2";

  /** Image resolution. */
  protected final static String RESOLUTION = "resolution";

  /** Rim center. */
  protected final static String CX     = "cx";
  protected final static String CY     = "cy";

  /** Rim radius. */
  protected final static String RAD    = "rad";

  /**
   * Runs one instance of the test (simple variant, w/o global result storage).
   */
  @Override
  public boolean runTest ( Map< String, String > env, Map< String, String > param, PrintStream out )
  {
      // class name:
    String shortName = param.get( CLASS_NAME );
    if ( shortName == null )
    {
      out.println( "Cannot find the '" + CLASS_NAME + "' test argument!" );
      return false;
    }
    String className = Template.JAGRLIB2 + shortName;

      // picture file-name:
    String outputFileName = param.get( OUTPUT_FILE_NAME );
    if ( outputFileName == null )
      outputFileName = "output.png";
    String tmp = env.get( BASE );
    if ( tmp != null )
    {
      File baseDir = new File( tmp );
      baseDir.mkdirs();
      outputFileName = tmp.concat( outputFileName );
    }

      // repeat:
    int repeat     = integerParam( env, param, REPEAT, 100 );

      // random seeds:
    int seed1      = integerParam( env, param, SEED1,  0 );
    int seed2      = integerParam( env, param, SEED2, 12 );

      // image resolution in pixels:
    int resolution = integerParam( env, param, RESOLUTION, 512 );

      // rim circle:
    double cx      = doubleParam( env, param, CX,  resolution * 0.5 );
    double cy      = doubleParam( env, param, CY,  resolution * 0.5 );
    double rad     = doubleParam( env, param, RAD, resolution * 0.3 );

      // create Piece objects:
    Piece ellipse;
    try
    {
      ellipse = (Piece)(Class.forName( className ).newInstance());
    }
    catch ( Exception e )
    {
      out.println( "Cannot instantiate the '" + shortName + "' class!" );
      return false;
    }
    if ( !(ellipse instanceof EllipseRender) )
    {
      out.println( "The '" + shortName + "' class doesn't implement EllipseRender interface!" );
      return true;
    }

    XTransitionList mask   = new XTransitionList();
    SolidColorPen   pen    = new SolidColorPen();
    RasterImage     image  = new RasterImage();
    PNGFileFormat   png    = new PNGFileFormat();
    JavaBitStream   output = new JavaBitStream();
    output.set( JavaBitStream.STREAM_NAME, outputFileName );

      // connect them by channels:
    try
    {
      ellipse.connect( Piece.PL_OUTPUT,  mask,   Piece.PL_INPUT );
      pen.connect(     Piece.PL_BITMASK, mask,   Piece.PL_INPUT );
      pen.connect(     Piece.PL_OUTPUT,  image,  Piece.PL_INPUT );
      png.connect(     Piece.PL_RASTER,  image,  Piece.PL_INPUT );
      png.connect(     Piece.PL_STREAM,  output, Piece.PL_INPUT );
    }
    catch ( BadInterfaceException e )
    {
      out.println( "Error connecting the '" + shortName + "' module!" );
      LogFile.exception( "Bad Interface in connecting-time!", e );
      return false;
    }

      // set Pieces' parameters:
    image.init( resolution, resolution, RasterGraphics.MODE_RGB, 0 );
    mask.init(  resolution, resolution );

    double[] color1 = { 1.0, 1.0, 1.0 };

      // random generator:
    RandomJames   rnd = new RandomJames( seed1, seed2 );
    EllipseRender el  = (EllipseRender)ellipse;

      // time measurement:
    long startTime = System.currentTimeMillis();

      // draw rim of ellipses:
    try
    {
      double alpha  = 0.0;
      double dalpha = Math.PI * 2.0 / repeat;
      double a      = rad * 0.5;
      double da     = 0.0;
      double b      = a * 0.01;
      double db     = a * 2.0 / repeat;

      for ( int i = 0; i++ < repeat; alpha += dalpha, a += da, b += db )
      {
        double x0 = cx + rad * Math.cos( alpha );
        double y0 = cy - rad * Math.sin( alpha );

        el.drawEllipse( x0, y0, a, b, alpha );
        color1[0] = rnd.randomDouble( 0.1, 1.0 );
        color1[1] = rnd.randomDouble( 0.1, 1.0 );
        color1[2] = rnd.randomDouble( 0.1, 1.0 );
        pen.setColor( color1 );
        pen.stroke();
      }
    }
    catch ( Exception e )
    {
      LogFile.exception( e );
    }

      // stop time:
    long stopTime = System.currentTimeMillis();

      // save the result:
    try
    {
      png.saveFile( (String)null, null );
    }
    catch ( Exception e )
    {
      LogFile.error( "Error writting '" + outputFileName + "' image!" );
    }

    out.println( String.format( Locale.US, "OK (%5.3fs) %s", 1.e-3 * (stopTime - startTime), shortName ) );
    return true;
  }

}
