package cz.cuni.jagrlib.eval;

import java.util.*;
import java.io.PrintStream;
import java.io.File;
import java.awt.geom.Point2D;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.*;

/**
 * Evaluation plugin for {@link MainEval}: exercises 45-46, polygon fill with or w/o anti-aliasing.
 *
 * @version 0.26 $Rev: 439 $ $Date: 2007-05-05 12:29:50 +0200 (so, 05 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     MainEval
 * @see     <a href="../../../../../cz/cuni/jagrlib/eval/Ex2006_45_a.java">Ex2006_45_a.java</a>
 */
public class Ex2006_45_a extends DefaultPlugin
{
  //==========================================================================
  //  XML params:

  /** Use anti-aliasing? */
  protected final static String ANTI_ALIAS = "antiAlias";

  /** Plug name. */
  protected final static String OUT_PLUG   = "outPlug";

  /** Random seed (for RandomJames). */
  protected final static String SEED1      = "seed1";
  protected final static String SEED2      = "seed2";

  /** Image resolution. */
  protected final static String RESOLUTION = "resolution";

  /** I - big star parameters. */
  protected final static String STAR_RAD1  = "starRad1";
  protected final static String STAR_RAD2  = "starRad2";
  protected final static String STAR_EXP   = "starExp";
  protected final static String STAR_SIZE  = "starSize";

  /** II - big fan parameters. */
  protected final static String FAN_MIN    = "fanMin";
  protected final static String FAN_MAX    = "fanMax";
  protected final static String FAN_SIZE   = "fanSize";

  /** III - arrows parameters. */
  protected final static String ARR_RADIUS = "arrRadius";
  protected final static String ARR_SIZE   = "arrSize";

  /** IV - random polygons parameters. */
  protected final static String RND_RADIUS = "rndRadius";
  protected final static String RND_REPEAT = "rndRepeat";

  /**
   * Runs one instance of the test (simple variant, w/o global result storage).
   */
  @Override
  public boolean runTest ( Map< String, String > env, Map< String, String > param, PrintStream out )
  {
      // class name:
    String shortName = param.get( CLASS_NAME );
    if ( shortName == null )
    {
      out.println( "Cannot find the '" + CLASS_NAME + "' test argument!" );
      return false;
    }
    String className = Template.JAGRLIB2 + shortName;

      // picture file-name:
    String outputFileName = param.get( OUTPUT_FILE_NAME );
    if ( outputFileName == null )
      outputFileName = "output.png";
    String tmp = env.get( BASE );
    if ( tmp != null )
    {
      File baseDir = new File( tmp );
      baseDir.mkdirs();
      outputFileName = tmp.concat( outputFileName );
    }

      // antialiasing flag:
    final boolean antiAlias = booleanParam( env, param, ANTI_ALIAS, false );

      // output plug's name:
    final String plugName   = stringParam(  env, param, OUT_PLUG, Piece.PL_OUTPUT );

      // random seeds:
    final int seed1         = integerParam( env, param, SEED1,  0 );
    final int seed2         = integerParam( env, param, SEED2, 12 );

      // image resolution in pixels:
    final int resolution    = integerParam( env, param, RESOLUTION, 600 );

      // I - big star:
    final double starRad1   = doubleParam(  env, param, STAR_RAD1,   80 );
    final double starRad2   = doubleParam(  env, param, STAR_RAD2,  180 );
    final double starExp    = doubleParam(  env, param, STAR_EXP,   3.0 );
    final int    starSize   = integerParam( env, param, STAR_SIZE,  100 );

      // II - big fan:
    final double fanMin     = doubleParam(  env, param, FAN_MIN,     50 );
    final double fanMax     = doubleParam(  env, param, FAN_MAX,    170 );
    final int    fanSize    = integerParam( env, param, FAN_SIZE,    80 );

      // III - arrows:
    final double arrRadius  = doubleParam(  env, param, ARR_RADIUS,  30 );
    final int    arrSize    = integerParam( env, param, ARR_SIZE,   150 );

      // IV - random polygons:
    final double rndRadius  = doubleParam(  env, param, RND_RADIUS,  20 );
    final int    rndRepeat  = integerParam( env, param, RND_REPEAT, 100 );

    final int maxPoly       = Math.max( Math.max( starSize, 5 ), fanSize * 3 ) + 2;

      // create Piece objects:
    Piece polygon;
    try
    {
      polygon = (Piece)(Class.forName( className ).newInstance());
    }
    catch ( Exception e )
    {
      out.println( "Cannot instantiate the '" + shortName + "' class!" );
      return false;
    }
    if ( !(polygon instanceof PolygonFillRender) )
    {
      out.println( "The '" + shortName + "' class doesn't implement PolygonFillRender interface!" );
      return true;
    }

    final XTransitionList mask   = new XTransitionList();
    final AlphaMatrix     alpha  = new AlphaMatrix();
    final SolidColorBrush brush  = new SolidColorBrush();
    final RasterImage     image  = new RasterImage();
    final PNGFileFormat   png    = new PNGFileFormat();
    final JavaBitStream   output = new JavaBitStream();
    output.set( JavaBitStream.STREAM_NAME, outputFileName );

      // connect them by channels:
    try
    {
      if ( antiAlias )
      {
        polygon.connect( plugName,           alpha, Piece.PL_INPUT );
        brush.connect(   Piece.PL_ALPHAMASK, alpha, Piece.PL_INPUT );
      }
      else
      {
        polygon.connect( plugName,         mask, Piece.PL_INPUT );
        brush.connect(   Piece.PL_BITMASK, mask, Piece.PL_INPUT );
      }
      brush.connect(   Piece.PL_OUTPUT,  image,  Piece.PL_INPUT );
      png.connect(     Piece.PL_RASTER,  image,  Piece.PL_INPUT );
      png.connect(     Piece.PL_STREAM,  output, Piece.PL_INPUT );
    }
    catch ( BadInterfaceException e )
    {
      out.println( "Error connecting the '" + shortName + "' module!" );
      LogFile.exception( "Bad Interface in connecting-time!", e );
      return false;
    }

      // set Pieces' parameters:
    if ( antiAlias )
    {
      image.init( resolution, resolution, RasterGraphics.MODE_RGBA, 0 );
      alpha.init( resolution, resolution );
    }
    else
    {
      image.init( resolution, resolution, RasterGraphics.MODE_RGB, 0 );
      mask.init(  resolution, resolution );
    }
    polygon.set( PolygonRender.POLY_INSIDE, PolygonRender.POLY_INSIDE_PARITY );

    double[] color = { 1.0, 1.0, 0.8, 1.0 };

      // random generator:
    final RandomJames rnd  = new RandomJames( seed1, seed2 );

      // support variables:
    final PolygonFillRender poly = (PolygonFillRender)polygon;
    final double half = resolution * 0.5;
    Point2D.Double[] p = new Point2D.Double[ maxPoly ];

      // time measurement:
    long startTime = System.currentTimeMillis();

    try
    {
      int i, j;
      double x0, y0, angle, dAngle;
      angle  = 0.0;
      dAngle = 1.0 / starSize;

        // I - big star:
      x0 = 0.5 * half + 3.0;
      y0 = 0.5 * half + 3.0;
      for ( i = 0; i < starSize; i++, angle += dAngle )
      {
        double alph = Math.pow( angle, starExp ) * 2.0 * Math.PI;
        double rad  = ((i & 1) == 0) ? starRad2 : starRad1;
        p[i]  = new Point2D.Double( x0 + rad * Math.sin( alph ),
                                    y0 - rad * Math.cos( alph ) );
      }
      p[i] = null;

      poly.fillPolygon( p );
      brush.setColor( color );
      brush.fill();

        // II - big fan:
      x0 = 1.5 * half - 3.0;
      y0 = 0.5 * half + 3.0;
      for ( i = j = 0; j++ < fanSize; )
      {
        double alph = rnd.uniformNumber() * 2.0 * Math.PI;
        double rad  = rnd.randomDouble( fanMin, fanMax );
        double fan  = rnd.randomDouble( 0.01, 0.11 );
        p[i++] = new Point2D.Double( x0, y0 );
        p[i++] = new Point2D.Double( x0 + rad * Math.sin( alph ),
                                     y0 - rad * Math.cos( alph ) );
        alph += fan;
        p[i++] = new Point2D.Double( x0 + rad * Math.sin( alph ),
                                     y0 - rad * Math.cos( alph ) );
      }
      p[i] = null;

      poly.fillPolygon( p );
      color[0] = 0.8;
      color[1] =
      color[2] = 1.0;
      brush.setColor( color );
      brush.fill();

        // III - arrows:
      double origX = 0.5 * arrRadius;
      double origY = half + 0.5 * arrRadius;
      double ampl  = half - arrRadius;
      double x, y;

      for ( i = 0; i++ < arrSize; )
      {
        x0    = rnd.randomDouble( origX, origX + ampl );
        y0    = rnd.randomDouble( origY, origY + ampl );
        angle = rnd.uniformNumber() * 2.0 * Math.PI;
        double sina = Math.sin( angle );
        double cosa = Math.cos( angle );

        double tailX = arrRadius * rnd.randomDouble( 0.2, 1.0 );
        double tailY = arrRadius * 0.2 * rnd.uniformNumber();
        double tipY  = arrRadius * rnd.randomDouble( 0.2, 0.7 );
        double tip2Y = tipY * rnd.randomDouble( 0.2, 0.5 );
        double spike = arrRadius * rnd.randomDouble( 0.4, 1.0 );
        double tip2X = spike * 0.2 * rnd.uniformNumber();

        x    = -tailX;
        y    =  tailY;
        p[0] = new Point2D.Double( x0 + x * cosa + y * sina,
                                   y0 - x * sina + y * cosa );
        y    = -tailY;
        p[1] = new Point2D.Double( x0 + x * cosa + y * sina,
                                   y0 - x * sina + y * cosa );
        x    =  tip2X;
        y    = -tip2Y;
        p[2] = new Point2D.Double( x0 + x * cosa + y * sina,
                                   y0 - x * sina + y * cosa );
        x    =  0.0;
        y    = -tipY;
        p[3] = new Point2D.Double( x0 + x * cosa + y * sina,
                                   y0 - x * sina + y * cosa );
        x    =  spike;
        y    =  0.0;
        p[4] = new Point2D.Double( x0 + x * cosa + y * sina,
                                   y0 - x * sina + y * cosa );
        x    =  0.0;
        y    =  tipY;
        p[5] = new Point2D.Double( x0 + x * cosa + y * sina,
                                   y0 - x * sina + y * cosa );
        x    =  tip2X;
        y    =  tip2Y;
        p[6] = new Point2D.Double( x0 + x * cosa + y * sina,
                                   y0 - x * sina + y * cosa );
        p[7] = null;

        poly.fillPolygon( p );
        color[0] = rnd.randomDouble( 0.1, 1.0 );
        color[1] = rnd.randomDouble( 0.1, 1.0 );
        color[2] = rnd.randomDouble( 0.1, 1.0 );
        brush.setColor( color );
        brush.fill();
      }

        // IV - random polygons:
      int n;
      origX =
      origY = half + 0.5 * rndRadius;
      ampl  = half - rndRadius;

      for ( i = 0; i++ < rndRepeat; )
      {
        x0 = rnd.randomDouble( origX, origX + ampl );
        y0 = rnd.randomDouble( origY, origY + ampl );
        n  = Formula.round( 2.6 + rnd.uniformNumber() * 2.8 );
        for ( j = 0; j < n; j++ )
        {
          angle = rnd.uniformNumber() * 2.0 * Math.PI;
          p[j]  = new Point2D.Double( x0 + rndRadius * Math.sin( angle ),
                                      y0 + rndRadius * Math.cos( angle ) );
        }
        if ( rnd.uniformNumber() < 0.5 )
          p[j++] = p[0];
        p[j] = null;

        poly.fillPolygon( p );
        color[0] = rnd.randomDouble( 0.1, 1.0 );
        color[1] = rnd.randomDouble( 0.1, 1.0 );
        color[2] = rnd.randomDouble( 0.1, 1.0 );
        brush.setColor( color );
        brush.fill();
      }
    }
    catch ( Exception e )
    {
      LogFile.exception( e );
    }

      // stop time:
    long stopTime = System.currentTimeMillis();

      // save the result:
    try
    {
      png.saveFile( (String)null, null );
    }
    catch ( Exception e )
    {
      LogFile.error( "Error writting '" + outputFileName + "' image!" );
    }

    out.println( String.format( Locale.US, "OK (%5.3fs) %s", 1.e-3 * (stopTime - startTime), shortName ) );
    return true;
  }

}
