package cz.cuni.jagrlib.gui;

import java.util.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.LogFile;

/**
 * Abstract class that cares about visual representation of composition.
 *
 * @author  Jan Kostlivy, PE
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/gui/CompositionGUI.java">CompositionGUI.java</a>
 */
abstract public class CompositionGUI extends JPanel implements Undo
{
  //==========================================================================
  //  Data:

  /** Info about Composition. */
  public InfoGroup info = new InfoGroup();

  /** Numerical order for modules and groups. */
  protected UniqueId idModules = new UniqueId();

  /** Numerical order for channels. */
  protected UniqueId idChannel = new UniqueId();

  /** Active plug for new channel. */
  protected InfoPlug plugFrom = null;

  /** Active plug for new channel. */
  protected InfoPlug plugTo = null;

  /** Zoom window. */
  protected ZoomWindow zoom;

  /** Undo/Redo object. */
  public ActionsUndo undoReg = null;
  final static public int ACTION_MODULE_REMOVE  = 1;
  final static public int ACTION_CHANNEL_REMOVE = 2;

  /** Flag for deleting modules. */
  public boolean canDelete = false;

  /** Count of selected modules. */
  protected int countSelected = 0;

  //==========================================================================
  //  Methods:

  /**
   * Replaces module with new one.
   *
   * @param  _info Information about old module.
   */
  abstract public void replaceModule ( InfoModule _info );

  /**
   * Adds a new module.
   *
   * @param  _info Information about module.
   */
  abstract public void newModule ( InfoModule _info );

  /**
   * Adds a new channel.
   *
   * @param  _info Information about channel.
   */
  abstract public void newChannel ( InfoChannel _info );

  /**
   * Creates new temporary channel.
   *
   * @return new temporary channel.
   */
  abstract public InfoChannel newChannelTemp ();

  /**
   * Adds a new group.
   */
  abstract public void newGroup ( InfoGroup _info );

  /** Reads composition and creates all it's components. */
  public void readInfo ()
  {
      // delete old
    removeAll();
    if ( undoReg != null )
      undoReg.clear();
    plugFrom = null;
    plugTo   = null;

      // create new
    for ( InfoModule im : info.modules.values() )
      if ( im instanceof InfoGroup )
        newGroup( (InfoGroup)im );
      else
        newModule( im );

    for ( InfoChannel ic : info.channels.values() )
      newChannel( ic );

    idModules.setLastId( info.modules.getMaxKey() );
    idChannel.setLastId( info.channels.getMaxKey() );
  }

  /**
   * [Un]selects all modules.
   */
  public void selectAll ( boolean _sel )
  {
    for ( InfoModule im : info.modules.values() )
      im.infoGUI.selected = _sel;
    repaintParent();
  }

  /**
   * Changes selection-state of module or group.
   *
   * @param  id Module key.
   * @param  sel New selection state.
   */
  public void setSelectModule ( Object id, boolean sel )
  {
    for ( InfoModule im : info.modules.values() )
      if ( im.key.equals( id ) )
      {
        im.infoGUI.selected = sel;
        if ( sel )
          moveToFront( im.infoGUI.gui );
        else
          moveToBack( im.infoGUI.gui );

        repaintParent();
        return;
      }
  }

  /**
   * Sets active plug for a new channel.
   *
   * @param  plug Active plug.
   */
  public void setIDModule ( InfoPlug plug )
  {
    if ( plug.input )
      setIDModuleTo( plug );
    else
      setIDModuleFrom( plug );

    if ( plugFrom != null &&
         plugTo   != null )
      newChannel( null );

    repaintParent();
  }

  /**
   * Sets starting plug for a new channel.
   *
   * @param  plug Active plug.
   */
  protected void setIDModuleFrom ( InfoPlug plug )
  {
    plugFrom = plug;
  }

  /**
   * Sets ending plug for a new channel.
   *
   * @param  plug Active plug.
   */
  protected void setIDModuleTo ( InfoPlug plug )
  {
    plugTo = plug;
  }

  /**
   * Get channel on [x,y].
   *
   * @param  x X-coordinate.
   * @param  y Y-coordinate.
   * @return Channel key or {@code null} if not found.
   */
  protected Integer getChannelAtPoint ( int x, int y )
  {
    for ( Integer key : info.channels.keySet() )
      if ( info.channels.get( key ).infoGUI.gui.contains( x, y ) )
        return key;

    return null;
  }

  /**
   * Creates a new group.
   */
  public void setGroup ()
  {
    throw new UnsupportedOperationException( "Method setGroup() not implemented yet." );
  }

  /**
   * Dismisses the existing group.
   */
  public void unsetGroup ( InfoGroup ig )
  {
    throw new UnsupportedOperationException( "Method unsetGroup() not implemented yet." );
  }

  /**
   * Dismisses all selected groups.
   */
  public void ungroup ()
  {
    for ( InfoModule im : info.modules.values() )
      if ( im.infoGUI.selected &&
           im instanceof InfoGroup )
        unsetGroup( (InfoGroup)im );

    repaintParent();
  }

  /**
   * Connects the given module with others..
   */
  abstract public void connectWithOther ( InfoModule infoM );

  /** Deletes the module. */
  abstract public void delete ();

  /**
   * Deletes the given module.
   *
   * @param  key Key of module or group.
   */
  public void removeModule ( Integer key )
  {
    if ( !canDelete && !info.plugs.isEmpty() )
    {
      JOptionPane.showMessageDialog( this, "You cannot delete modules form group this way" +
                                     "\n(in this version)." +
                                     "\nIt can be done in Tools / View Registry / Edit Data." );
      return;
    }

    InfoModule im = info.modules.get( key );
    remove( im.infoGUI.gui );
    removeChannels( key );
    info.modules.remove( key );
    if ( undoReg != null )
      undoReg.registry( new ActionsUndoReg( this, ACTION_MODULE_REMOVE, im, null ) );

    repaintParent();
  }

  /**
   * Deletes channels connected to the given module.
   *
   * @param  key Module key.
   */
  public void removeChannels ( Integer key )
  {
    Set< Integer > keySet = new HashSet< Integer >( info.channels.keySet() );
    for ( Integer chKey : keySet )
    {
      InfoChannel ic = info.channels.get( chKey );
      if ( key.equals( ic.plugFrom.owner.key ) ||
           key.equals( ic.plugTo.owner.key ) )
        removeChannel( chKey );
    }
  }

  /**
   * Deletes the given channel.
   *
   * @param  key Channel key.
   */
  public void removeChannel ( Integer key )
  {
    InfoChannel rm = info.channels.get( key );
    rm.infoGUI.gui.removeChannel();
    info.channels.remove( rm.key );
    if ( undoReg != null )
      undoReg.registry( new ActionsUndoReg( this, ACTION_CHANNEL_REMOVE, rm, null ) );
  }

  /**
   * Moves all selected items.
   *
   * @param  id Key of item.
   * @param  nx Relative x coordinate.
   * @param  ny Relative y coordinate.
   */
  abstract public void moveSelected ( Integer id, int nx, int ny );

  /**
   * Repaints the composition panel.
   */
  @Override
  public void paintComponent ( Graphics g )
  {
    super.paintComponent( g );
    paintItem( g );
  }

  /**
   * Paints links between modules-plugs composition-plugs.
   */
  public void paintLinks ( Graphics g )
  {
  }

  /**
   * Paint bonus graphics on panel.
   */
  abstract public void paintItem ( Graphics g );

  /**
   * Component initialization.
   */
  private void jbInit () throws Exception
  {
    this.setBackground( Color.white );

    this.addMouseMotionListener( new MouseMotionAdapter ()
    {
      public void mouseMoved ( MouseEvent e )
      {
        this_mouseMoved( e );
      }
    });

    this.setLayout( null );
  }

  /** Sets bounds of Working space. */
  protected void setCompositionSize ()
  {
    Container c = getParent();
    if ( c == null ) return;

    int ph = c.getHeight();
    int pw = c.getWidth();

    for ( InfoModule im : info.modules.values() )
    {
      Dimension pd = im.infoGUI.gui.getMaxBounds();
      if ( ph < pd.height ) ph = pd.height;
      if ( pw < pd.width )  pw = pd.width;
    }

    if ( getHeight() != ph ||
         getWidth()  != pw )
    {
      setPreferredSize( new Dimension( pw, ph ) );
      if ( zoom != null )
        zoom.setComposition( this );
    }
  }

  /** Default constructor. */
  public CompositionGUI ()
  {
    try
    {
      jbInit();
    }
    catch ( Exception e )
    {
      LogFile.exception( e );
    }
  }

  /**
   * Undo last action.
   *
   * @param  action Action to un-done.
   */
  public void undo ( ActionsUndoReg action )
  {
    if ( undoReg == null ) return;

    undoReg.dont = true;  // avoids re-registration of this action

    switch ( action.what )
    {
      case ACTION_MODULE_REMOVE:
      {
        InfoModule im = (InfoModule)action.lastValue;
        newModule( im );
        info.modules.put( im.key, im );
        break;
      }

      case ACTION_CHANNEL_REMOVE:
      {
        InfoChannel ic = (InfoChannel)action.lastValue;
        newChannel( ic );
        info.channels.put( ic.key, ic );
        break;
      }
    }

    repaintParent();
    undoReg.dont = false;
  }

  /**
   * Redoes last undoed action.
   *
   * @param  action Action to re-done.
   */
  public void redo ( ActionsUndoReg action )
  {
    if ( undoReg == null ) return;

    undoReg.dont = true;  // avoids re-registration of this action

    switch ( action.what )
    {
      case ACTION_MODULE_REMOVE:
      {
        InfoModule im = (InfoModule)action.lastValue;
        removeModule( im.key );
        break;
      }

      case ACTION_CHANNEL_REMOVE:
      {
        InfoChannel ic = (InfoChannel)action.lastValue;
        removeChannel( ic.key );
        break;
      }
    }

    repaintParent();
    undoReg.dont = false;
  }

  //==========================================================================
  //  Events:

  /**
   * Mouse move .. sets ToolTips over channels.
   */
  void this_mouseMoved ( MouseEvent e )
  {
    setToolTipText( null );
    Integer ci = getChannelAtPoint( e.getX(), e.getY() );
    if ( ci != null )
    {
      setCursor( Cursor.getPredefinedCursor( Cursor.HAND_CURSOR ) );
      InfoChannel ic = info.channels.get( ci );
      setToolTipText( ic.nameInterface +
                      '\n' + ic.plugFrom.getModuleOwner().reg.name +
                      " --> " + ic.plugTo.getModuleOwner().reg.name );
    }
    else
      setCursor( Cursor.getPredefinedCursor( Cursor.DEFAULT_CURSOR ) );
  }

  /** Repaint the parent composition. */
  public void repaintParent ()
  {
    ((InfoGroup)info.getTopLevelOwner()).infoCompositionGUI.gui.repaint();
  }

  /**
   * Moves module front.
   */
  protected void moveToFront ( ModuleGUI module )
  {
    remove( module );
    add( module, 0 );
  }

  /**
   * Moves module back.
   */
  protected void moveToBack ( ModuleGUI module )
  {
    remove( module );
    add( module, info.modules.size() - 1 );
  }

}
