package cz.cuni.jagrlib.gui;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.net.URL;
import javax.swing.ImageIcon;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Class storing global Skel preferences.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/gui/Options.java">Options.java</a>
 */
public class Options
{
  /** Default resource directory. */
  protected static final String PATH_RESOURCES = "./res/";

  /** Actual resource directory. */
  public static String pathResources           = PATH_RESOURCES;

  /** Default data directory. */
  protected static final String PATH_DATA      = "./data/";

  /** Actual data directory. */
  public static String pathData                = PATH_DATA;

  /** Default resource directory. */
  public static final String FILE_NAME_REG     = "regdata.xml";

  /** Default path to registry database. */
  protected static final String PATH_REG       = "./reg/" + FILE_NAME_REG;

  /** Actual path to registry database. */
  public static String pathReg                 = PATH_REG;

  /** Default source path. */
  protected static final String PATH_SOURCE    = "./";

  /** Actual source path. */
  public static String pathSource              = PATH_SOURCE;

  /** Default external viewer/editor. */
  protected static final String PATH_EDITOR    = "notepad.exe";

  /** Actual external viewer/editor. */
  public static String pathEditor              = PATH_EDITOR;

  /** Options file name - user preferences. */
  public static String optionsFileName         = "options.txt";

  /** Options file name - system preferences. */
  public static String sysOptionsFileName      = "sys-options.txt";

  /** Default file name of system preferences. */
  protected static final String PATH_SYS_OPTIONS = "sys-options-orig.txt";

  /** Options file name - system preferences (original values). */
  public static String sysOptionsOrigFileName  = PATH_SYS_OPTIONS;

  /** User options in "key->value" format. */
  protected static Properties pref;

  /** System options in "key->value" format. */
  protected static Properties system;

  /** Keys which are updated in 'system' rather than in 'pref'. */
  protected static HashSet< String > systemKeys;

  /**
   * Loads preference files from the disk and [re-]initializes working values.
   */
  public static void init ()
  {
    if ( system == null )
    {
      system     = new Properties();
      pref       = new Properties( system ); // system preferences plays role of default value for user pref.

        // Which keys will be stored in 'system' dictionary:
      systemKeys = new HashSet< String >( 8 );
      systemKeys.add( "PathResources" );
      systemKeys.add( "PathData" );
      systemKeys.add( "PathReg" );
      systemKeys.add( "PathSource" );
      systemKeys.add( "PathEditor" );
    }
    try
    {
      load();
    }
    catch ( IOException ex )
    {
      LogFile.exception( ex );
    }
    interpret();
  }

  /**
   * Stores both preference files from dictionaries to disk files.
   * Doesn't fill dictionaries from working state variables.
   *
   * @see    #reflect()
   */
  public static void save ()
  throws IOException
  {
    system.store( new FileOutputStream( sysOptionsFileName ),
                  " This file holds system preferences of the Skel program and was generated automatically, see " + Version.JAGRLIB_URL );
    pref.store( new FileOutputStream( optionsFileName ),
                " This file holds user preferences of the Skel program and was generated automatically, see " + Version.JAGRLIB_URL );
  }

  /**
   * Loads both preference files from disk to the dictionaries.
   * Doesn't parse (interpret) the read values.
   *
   * @see   #interpret()
   */
  public static void load ()
  throws IOException
  {
    File file = new File( sysOptionsFileName );
    if ( !file.exists() )
      file = new File( sysOptionsOrigFileName );

    if ( file.exists() )                  // load system properties from file
      system.load( new FileInputStream( file ) );
    else                                  // try to load system properties from URL resource
    {
      InputStream urlIs = Options.class.getResourceAsStream( PATH_SYS_OPTIONS );
      if ( urlIs != null )
        system.load( urlIs );
    }

    file = new File( optionsFileName );
    if ( file.exists() )
      pref.load( new FileInputStream( file ) );
  }

  /**
   * Parses color description and returns valid Color instance.
   * If the given string cannot be parsed, returns #ff000000 (black).
   */
  public static Color getColor ( String key )
  {
    String str = pref.getProperty( key );
    if ( str == null )
      LogFile.warning( "color value '" + key + "' not found in preferences" );
    else
    {
      try
      {
        long result = Long.decode( str ).longValue();
        return new Color( (int)result, (result & 0xff000000) > 0 );
      }
      catch ( NumberFormatException e )
      {
        LogFile.warning( "cannot parse color value '" + key + "' (" + str + ')' );
      }
    }
    return Color.BLACK;
  }

  /**
   * Parses boolean parameter. Returns 'false' in case of failure.
   */
  public static boolean getBoolean ( String key )
  {
    String str = pref.getProperty( key );
    if ( str == null )
    {
      LogFile.warning( "boolean value '" + key + "' not found in preferences" );
      return false;
    }
    return Boolean.parseBoolean( str );
  }

  /**
   * Parses integer parameter and checks its bounds.
   * Returns 'deflt' in case of failure.
   */
  public static int getInteger ( String key, int min, int max, int deflt )
  {
    String str = pref.getProperty( key );
    if ( str == null )
      LogFile.warning( "integer value '" + key + "' not found in preferences" );
    else
    {
      try
      {
        int result = Integer.decode( str ).intValue();
        if ( result < min || result > max )
        {
          LogFile.warning( "integer value '" + key + "' out of bounds [" + min + ',' + max + "], it will be clipped" );
          result = Formula.clamp( result, min, max );
        }
        return result;
      }
      catch ( NumberFormatException e )
      {
        LogFile.warning( "cannot parse integer value '" + key + "' (" + str + ')' );
      }
    }
    return deflt;
  }

  /** Parses integer parameter w/o bounds checking. */
  public static int getInteger ( String key )
  {
    return getInteger( key, Integer.MIN_VALUE, Integer.MAX_VALUE, 0 );
  }

  /**
   * Parses string parameter. Returns 'deflt' in case of failure.
   */
  public static String getString ( String key, String deflt )
  {
    String str = pref.getProperty( key );
    if ( str == null )
    {
      LogFile.warning( "string value '" + key + "' not found in preferences" );
      return deflt;
    }
    return str;
  }

  /**
   * Parses string parameter. Returns {@code null} in case of failure.
   */
  public static String getString ( String key )
  {
    return getString( key, null );
  }

  /**
   * Interprets preference dictionaries.
   * Fills all working state variables from the actual dictionaries.
   */
  public static void interpret ()
  {
    InfoPlugGUI.colorConnect          = getColor(   "ColorPlugConnect" );
    InfoPlugGUI.colorMandatory        = getColor(   "ColorPlugMandatory" );
    InfoPlugGUI.colorNonMandatory     = getColor(   "ColorPlugNonMandatory" );
    InfoPlugGUI.colorActive           = getColor(   "ColorPlugActive" );
    InfoPlugGUI.colorOpposite         = getColor(   "ColorPlugOpposite" );
    InfoChannelGUI.colorNormal        = getColor(   "ColorChannelNormal" );
    InfoChannelGUI.colorActiveModule  = getColor(   "ColorChannelActive" );
    InfoChannelGUI.colorTemp          = getColor(   "ColorChannelTemp" );
    InfoModuleGUI.colorSelect         = getColor(   "ColorModuleSelect" );
    InfoGroupGUI.colorName            = getColor(   "ColorGroupName" );
    InfoModuleGUI.colorName           = getColor(   "ColorModuleName" );
    ZoomBasic.colorModule             = getColor(   "ColorZoomModule" );
    ZoomBasic.colorMandatory          = getColor(   "ColorZoomMandatory" );
    ZoomBasic.colorNonMandatory       = getColor(   "ColorZoomNonMandatory" );
    InfoGroup.autoConnect             = getBoolean( "AutoConnect" );
    ModuleGUI.setModuleHeight(          getInteger( "ModuleHeight",    0, 10000, 0 ) );
    ModuleGUI.setModuleWidth(           getInteger( "ModuleWidth",     0, 10000, 0 ) );
    ParamGUI.paramLineHeight          = getInteger( "ParameterHeight", 5,  1000, ParamGUI.PARAM_LINE_HEIGHT );
    pathResources                     = getString(  "PathResources", PATH_RESOURCES );
    pathData                          = getString(  "PathData",      PATH_DATA );
    pathReg                           = getString(  "PathReg",       PATH_REG );
    pathSource                        = getString(  "PathSource",    PATH_SOURCE );
    pathEditor                        = getString(  "PathEditor",    PATH_EDITOR );
  }

  /**
   * Fills both dictionaries with the actual working values.
   * Should be called before the {@link #save()} operation.
   */
  public static void reflect ()
  {
    set( "ColorPlugConnect",      InfoPlugGUI.colorConnect );
    set( "ColorPlugMandatory",    InfoPlugGUI.colorMandatory );
    set( "ColorPlugNonMandatory", InfoPlugGUI.colorNonMandatory );
    set( "ColorPlugActive",       InfoPlugGUI.colorActive );
    set( "ColorPlugOpposite",     InfoPlugGUI.colorOpposite );
    set( "ColorChannelNormal",    InfoChannelGUI.colorNormal );
    set( "ColorChannelActive",    InfoChannelGUI.colorActiveModule );
    set( "ColorChannelTemp",      InfoChannelGUI.colorTemp );
    set( "ColorModuleSelect",     InfoModuleGUI.colorSelect );
    set( "ColorGroupName",        InfoGroupGUI.colorName );
    set( "ColorModuleName",       InfoModuleGUI.colorName );
    set( "ColorZoomModule",       ZoomBasic.colorModule );
    set( "ColorZoomMandatory",    ZoomBasic.colorMandatory );
    set( "ColorZoomNonMandatory", ZoomBasic.colorNonMandatory );
    set( "AutoConnect",           InfoGroup.autoConnect );
    set( "ModuleHeight",          ModuleGUI.getModuleHeight() );
    set( "ModuleWidth",           ModuleGUI.getModuleWidth() );
    set( "ParameterHeight",       ParamGUI.paramLineHeight );
    set( "PathResources",         pathResources );
    set( "PathData",              pathData );
    set( "PathReg",               pathReg );
    set( "PathSource",            pathSource );
    set( "PathEditor",            pathEditor );
  }

  /**
   * Converts Color to HexString (#AARRGGBB).
   *
   * @param c Color to be converted (const).
   * @return String representation of the color.
   */
  public static String colorToHexString ( Color c )
  {
    StringBuffer buf = new StringBuffer( Integer.toHexString( c.getRGB() ) );
    while ( buf.length() < 9 )    // 0AARRGGBB
      buf.insert( 0, '0' );

    buf.insert( 1, 'x' );
    return buf.toString();
  }

  /**
   * Sets the string parameter.
   * {@link #interpret()} should be called to reflect all changes in working state variables.
   */
  public static void set ( String key, String value )
  {
    if ( systemKeys.contains( key ) )
    {
      system.setProperty( key, value );
      pref.remove( key );
    }
    else
      pref.setProperty( key, value );
  }

  /**
   * Sets the color parameter.
   * {@link #interpret()} should be called to reflect all changes in working state variables.
   */
  public static void set ( String key, Color c )
  {
    set( key, colorToHexString( c ) );
  }

  /**
   * Sets the boolean parameter.
   * {@link #interpret()} should be called to reflect all changes in working state variables.
   */
  public static void set ( String key, boolean b )
  {
    set( key, Boolean.toString( b ) );
  }

  /**
   * Sets the integer parameter.
   * {@link #interpret()} should be called to reflect all changes in working state variables.
   */
  public static void set ( String key, int i )
  {
    set( key, Integer.toString( i ) );
  }

  /**
   * Retrieves icon object from resource (JAR) or disk file.
   */
  public static ImageIcon getIcon ( String fileName )
  {
    URL iconURL = Options.class.getResource( fileName );
    if ( iconURL != null )
      return new ImageIcon( iconURL );

    return new ImageIcon( Options.pathResources + fileName );
  }

  /**
   * Retrieves resource stream object from URL resource (JAR).
   */
  public static InputStream getResource ( String fileName )
  {
    return Options.class.getResourceAsStream( fileName );
  }

}
