package cz.cuni.jagrlib.gui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.beans.PropertyVetoException;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.Worker;
import cz.cuni.jagrlib.reg.*;

/**
 * Basic internal frame for composition.
 *
 * @author  PE, Jan Kostlivy
 * @version 0.26 $Rev: 454 $ $Date: 2007-05-15 03:35:14 +0200 (út, 15 V 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/gui/WorkSpaceGUIBasic.java">WorkSpaceGUIBasic.java</a>
 */
public class WorkSpaceGUIBasic extends WorkSpaceGUI
{
  private static final long serialVersionUID = 1L;

  /** Zoom window. */
  protected ZoomWindow zoom;

  /** Undo/Redo object. */
  public ActionsUndo undo = new ActionsUndo();

  /** In closing process? */
  protected boolean closing = false;

  // GUI components:
  private JPanel contentPane;
  private JScrollPane jScrollPaneSpace    = new JScrollPane();
  private CompositionGUI panelComposition = InstanceGUI.newCompositionInstance();

  //==========================================================================
  //  WorkSpaceGUI methods:

  /**
   * Tunnel for commands from Main Window to the Composition.
   * <br />Commands:
   * <br />{@code CMD_LOAD} - load composition.
   * <br />{@code CMD_SAVE} - save composition. params = String(WorkSpaceGUI.SAVE, WorkSpaceGUI.SAVEAS)
   * <br />{@code CMD_ADDMODULE} - add module(s) to composition. Module's position is left top corner.
   * <br />{@code CMD_DELETEMODULE} - delete module(s) from composition.
   * <br />{@code CMD_GROUP} - creating new group.
   * <br />{@code CMD_UNGROUP} - dismissing group.
   * <br />{@code CMD_RUN} - run compostition.
   * <br />{@code CMD_UNDO} - undo.
   * <br />{@code CMD_REDO} - redo.
   * <br />{@code CMD_ZOOMWINDOW} - sets Zoom window to composition. params = JInternalFrame(ZoomWindow)
   * <br />{@code CMD_SELECTALL} - select all modules in composition.
   *
   * @param command command from main window.
   * @param params params for that command.
   * @return Result codes:
   *   {@code 0} - OK or no returning value
   *   {@code -1} - Unknown command
   */
  @Override
  public int pipeToComposition ( int command, Object params )
  {
    switch ( command )
    {
      case CMD_LOAD:
        load();
        return 0;

      case CMD_SAVE:
        save( (String)params );
        return 0;

      case CMD_ADDMODULE:
        panelComposition.newModule( null );
        return 0;

      case CMD_DELETEMODULE:
        panelComposition.delete();
        return 0;

      case CMD_GROUP:
        panelComposition.setGroup();
        return 0;

      case CMD_UNGROUP:
        panelComposition.ungroup();
        return 0;

      case CMD_RUN:
        run();
        return 0;

      case CMD_UNDO:
        undo.undo();
        return 0;

      case CMD_REDO:
        undo.redo();
        return 0;

      case CMD_ZOOMWINDOW:
        setZoomWindow( (JInternalFrame)params );
        return 0;

      case CMD_SELECTALL:
        panelComposition.selectAll( true );
        return 0;
    }

    return -1;
  }

  public CompositionGUI getCompositionGUI ()
  {
    return panelComposition;
  }

  /** Loading the composition. */
  @Override
  public void load ()
  {
    try
    {
        // file-name to load:
      if ( Main.loadFileName != null &&
           Main.loadFileName.length() > 0 )
        panelComposition.info.fileName = Main.loadFileName;

      Main.loadFileName = null;

      RegDatabaseBasic.getInputOutput().load( panelComposition.info );

      if ( panelComposition.info.fileName != null &&
           panelComposition.info.fileName.length() > 0 )
      {
        panelComposition.readInfo();
        setTitle( panelComposition.info.fileName );
      }
    }
    catch ( Exception e )
    {
      LogFile.exception( e );
      StringBuffer msg = new StringBuffer( "Error loading composition: " );
      msg.append( e.toString() );
      if ( e.getCause() != null )
      {
        msg.append( "\nCaused by: " );
        msg.append( e.getCause().toString() );
      }
      JOptionPane.showMessageDialog( this, msg.toString(), "Error", JOptionPane.ERROR_MESSAGE );
    }
  }

  /**
   * Saving the composition.
   *
   * @param type Type of saving (SAVE or SAVEAS).
   */
  @Override
  public void save ( String type )
  {
    if ( panelComposition.info == null )
      return;

    try
    {
      RegDatabaseBasic.getInputOutput().save( type, panelComposition.info );
      setTitle( panelComposition.info.fileName );
      undo.clear();
    }
    catch ( Exception e )
    {
      StringBuffer msg = new StringBuffer( "Error saving composition: " );
      msg.append( e.toString() );
      if ( e.getCause() != null )
      {
        msg.append( "\nCaused by: " );
        msg.append( e.getCause().toString() );
      }
      JOptionPane.showMessageDialog( this, msg.toString(), "Error", JOptionPane.ERROR_MESSAGE );
    }
  }

  /** Running the composition. */
  @Override
  public void run ()
  {
      // composition to run..
    final InfoGroup composition = panelComposition.info;

    if ( (composition.getState() & InfoState.NOTCONNECTED_HARD) > 0 )
    {
      JOptionPane.showMessageDialog( this, "At least one mandatory plug is not connected!", "Error", JOptionPane.ERROR_MESSAGE );
      return;
    }

      // create Piece instances:
    Composition.createPieces( composition.modules );

      // connect them by channels:
    Composition.connectPieces( composition.channels );

      // init Piece instances:
    Composition.initPieces( composition.modules );

      // run init() phases of all modules:
    int maxPhase = 1;
    int phase;
    for ( phase = 0; phase < maxPhase; phase++ )
      for ( Integer key : composition.modules.keySet() )
      {
        int max = composition.modules.get( key ).instancePiece.init( phase );
        if ( max > maxPhase )
          maxPhase = max;
      }

      // select one or more worker modules to run:
    for ( Integer key : composition.modules.keySet() )
    {
      final Piece ipp = composition.modules.get( key ).instancePiece;

      if ( ipp instanceof Worker &&
           !Piece.booleanProperty( ipp, Worker.REJECT, false ) )
      {
        Thread runIt = new Thread( "JaGrLib - Composition worker" )
        {
          @Override
          public void run ()
          {
            MainWindow.mainWindow.runningInc( this, ipp, composition );
            try
            {
              ((Worker)ipp).run();
            }
            catch ( Exception e )
            {
              LogFile.exception( e );
              StringBuffer msg = new StringBuffer( "Exception: " );
              msg.append( e.toString() );
              StackTraceElement[] st = e.getStackTrace();
              if ( st != null && st.length > 0 )
              {
                msg.append( '\n' );
                msg.append( st[0].toString() );
                msg.append( " ..." );
              }
              if ( e instanceof JaGrLibException )
              {
                msg.append( "\nCaused by: " );
                msg.append( e.getCause().toString() );
                st = e.getCause().getStackTrace();
                if ( st != null && st.length > 0 )
                {
                  msg.append( '\n' );
                  msg.append( st[0].toString() );
                  msg.append( " ..." );
                }
              }
              JOptionPane.showMessageDialog( MainWindow.mainWindow, msg.toString(), "Exception", JOptionPane.ERROR_MESSAGE );
            }
            finally
            {
              MainWindow.mainWindow.runningDec( this );
            }
          }
        };

        try
        {
          runIt.setPriority( Thread.MIN_PRIORITY );
        }
        catch ( Exception e )
        {
          LogFile.warning( "Cannot set priority of worker thread!" );
          LogFile.exception( e );
        }

        try
        {
          runIt.start();
        }
        catch ( IllegalThreadStateException e )
        {
          LogFile.error( "Cannot start worker thread!" );
          LogFile.exception( e );
        }
      }
    }
  }

  /**
   * Set zoom window to composition.
   *
   * @param _zoom zoom window.
   */
  public void setZoomWindow ( JInternalFrame _zoom )
  {
    zoom                  = (ZoomWindow)_zoom;
    panelComposition.zoom = zoom;
  }

  //==========================================================================
  //  Methods:

  /**
   * Component initialization.
   */
  private void jbInit ()
  throws Exception
  {
    contentPane = (JPanel) this.getContentPane();
    contentPane.setLayout( new BorderLayout() );
    setDefaultCloseOperation( WindowConstants.DO_NOTHING_ON_CLOSE );
    addInternalFrameListener( new InternalFrameAdapter ()
    {
      @Override
      public void internalFrameActivated ( InternalFrameEvent e )
      {
        this_internalFrameActivated( e );
      }

      @Override
      public void internalFrameClosing ( InternalFrameEvent e )
      {
        this_internalFrameClosing( e );
      }
    });

    contentPane.add( jScrollPaneSpace, BorderLayout.CENTER );
    jScrollPaneSpace.getViewport().add( panelComposition, BorderLayout.CENTER );
    panelComposition.undoReg = undo;
  }

  /** Default constructor. */
  public WorkSpaceGUIBasic ()
  {
    try
    {
      jbInit();
    }
    catch ( Exception e )
    {
      LogFile.exception( e );
    }
  }

  //==========================================================================
  //  Events:

  /**
   * Activating this frame.
   * Change zoom window preview.
   */
  protected void this_internalFrameActivated ( InternalFrameEvent e )
  {
    zoom.setComposition( panelComposition );
  }

  /**
   * Closing this frame.
   * Change zoom window preview.
   */
  protected void this_internalFrameClosing ( InternalFrameEvent e )
  {
    if ( !closing )
    {
      if ( undo.changed() )
      {
        int answer = JOptionPane.showConfirmDialog( this, "Composition '" +
                                                    panelComposition.info.fileName +
                                                    "' was changed. Do you want to save it?" );
        if ( answer == JOptionPane.CANCEL_OPTION )
          return;

        if ( answer == JOptionPane.YES_OPTION )
          save( InputOutput.SAVE );
      }

      closing = true;

      try
      {
        setClosed( true );
      }
      catch ( PropertyVetoException veto )
      {
      }
    }

    zoom.setComposition( null );
  }

  /**
   * Resize the composition.
   */
  @Override
  protected void this_componentResized ( ComponentEvent e )
  {
    panelComposition.setCompositionSize();
  }

}
