package cz.cuni.jagrlib.gui.comp;

import java.io.File;
import java.util.Hashtable;
import java.util.Enumeration;
import javax.swing.filechooser.*;

/**
 * Filter for files for JFileChooser.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../../cz/cuni/jagrlib/gui/comp/CommonFileFilter.java">CommonFileFilter.java</a>
 */
public class CommonFileFilter extends FileFilter
{
  /** Unknown file types. */
  //private static String TYPE_UNKNOWN = "Type Unknown";

  /** Hidden file. */
  //private static String HIDDEN_FILE = "Hidden File";

  private Hashtable< String, FileFilter > filters = null;
  private String description = null;
  private String fullDescription = null;
  private boolean useExtensionsInDescription = true;

  /**
   * Creates a file filter. If no filters are added, then all
   * files are accepted.
   *
   * @see #addExtension
   */
  public CommonFileFilter ()
  {
    this.filters = new Hashtable< String, FileFilter >();
  }

  /**
   * Creates a file filter that accepts files with the given extension.
   *   @param  extension .
   * Example: new ExampleFileFilter("xml");
   *
   * @see #addExtension
   */
  public CommonFileFilter ( String extension )
  {
    this( extension, null );
  }

  /**
   * Creates a file filter that accepts the given file type.
   *   @param  extension .
   *   @param  description .
   * Example: new ExampleFileFilter("xml", "XML Files");
   *
   * Note that the "." before the extension is not needed. If
   * provided, it will be ignored.
   *
   * @see #addExtension
   */
  public CommonFileFilter ( String extension, String description )
  {
    this();
    if ( extension   != null ) addExtension( extension );
    if ( description != null ) setDescription( description );
  }

  /**
   * Creates a file filter from the given string array.
   *   @param  filters .
   * Example: new ExampleFileFilter(String {"xml", "html"});
   *
   * Note that the "." before the extension is not needed and
   * will be ignored.
   *
   * @see #addExtension
   */
  public CommonFileFilter ( String[] filters )
  {
    this( filters, null );
  }

  /**
   * Creates a file filter from the given string array and description.
   *   @param  filters .
   *   @param  description .
   * Example: new ExampleFileFilter(String {"xml", "html"}, "Xml and Html Files");
   *
   * Note that the "." before the extension is not needed and will be ignored.
   *
   * @see #addExtension
   */
  public CommonFileFilter ( String[] filters, String description )
  {
    this();
    for ( int i = 0; i < filters.length; i++ )
    {
        // add filters one by one
      addExtension( filters[i] );
    }
    if ( description != null ) setDescription( description );
  }

  /**
   * Return true if this file should be shown in the directory pane,
   * false if it shouldn't.
   *
   * Files that begin with "." are ignored.
   *
   *   @param  f .
   *   @return .
   * @see #getExtension
   * @see FileFilter#accept
   */
  public boolean accept ( File f )
  {
    if ( f != null )
    {
      if ( f.isDirectory() )
      {
        return true;
      }
      String extension = getExtension( f );
      if ( extension != null && filters.get( getExtension( f ) ) != null )
      {
        return true;
      }
    }
    return false;
  }

  /**
   * Return the extension portion of the file's name .
   *
   *   @param  f .
   *   @return .
   * @see #getExtension
   * @see FileFilter#accept
   */
  public String getExtension ( File f )
  {
    if ( f != null )
    {
      String filename = f.getName();
      int i = filename.lastIndexOf( '.' );
      if ( i > 0 && i < filename.length() - 1 )
      {
        return filename.substring( i+1 ).toLowerCase();
      }
    }
    return null;
  }

  /**
   * Adds a filetype "dot" extension to filter against.
   *
   * For example: the following code will create a filter that filters
   * out all files except those that end in ".xml" and ".html":
   *
   *   @param  extension .
   *
   *   ExampleFileFilter filter = new ExampleFileFilter();
   *   filter.addExtension("xml");
   *   filter.addExtension("html");
   *
   * Note that the "." before the extension is not needed and will be ignored.
   */
  public void addExtension ( String extension )
  {
    if ( filters == null )
      filters = new Hashtable< String, FileFilter >( 5 );
    filters.put( extension.toLowerCase(), this );
    fullDescription = null;
  }

  /**
   * Returns the human readable description of this filter. For
   * example: "XML and HTML Files (*.xml, *.html)"
   *
   *   @return .
   * @see #setDescription
   * @see #setExtensionListInDescription
   * @see #isExtensionListInDescription
   * @see FileFilter#getDescription
   */
  public String getDescription ()
  {
    if ( fullDescription == null )
    {
      if ( description == null || isExtensionListInDescription() )
      {
        fullDescription = description == null ? "(" : description + " (";
          // build the description from the extension list
        Enumeration< String > extensions = filters.keys();
        if ( extensions != null )
        {
          fullDescription += '.' + (String)extensions.nextElement();
          while ( extensions.hasMoreElements() )
          {
            fullDescription += ", " + (String)extensions.nextElement();
          }
        }
        fullDescription += ')';
      }
      else
      {
        fullDescription = description;
      }
    }
    return fullDescription;
  }

  /**
   * Sets the human readable description of this filter. For
   * example: filter.setDescription("Xml and Html files");
   *
   * @param  description .
   * @see #setDescription
   * @see #setExtensionListInDescription
   * @see #isExtensionListInDescription
   */
  public void setDescription ( String description )
  {
    this.description = description;
    fullDescription = null;
  }

  /**
   * Determines whether the extension list (.xml, .html, etc) should
   * show up in the human readable description.
   *
   * Only relevant if a description was provided in the constructor
   * or using setDescription();
   *
   *   @param  b .
   * @see #getDescription
   * @see #setDescription
   * @see #isExtensionListInDescription
   */
  public void setExtensionListInDescription ( boolean b )
  {
    useExtensionsInDescription = b;
    fullDescription = null;
  }

  /**
   * Returns whether the extension list (.xml, .html, etc) should
   * show up in the human readable description.
   *
   * Only relevant if a description was provided in the constructor
   * or using setDescription();
   *
   *   @return .
   * @see #getDescription
   * @see #setDescription
   * @see #setExtensionListInDescription
   */
  public boolean isExtensionListInDescription ()
  {
    return useExtensionsInDescription;
  }

}
