package cz.cuni.jagrlib.iface;

import cz.cuni.jagrlib.iface.BitMask;

/**
 * Bit-mask interface with alpha-channel capability (representation of pixel set in 2D plane).
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.01
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/AlphaMask.java">AlphaMask.java</a>
 */
public interface AlphaMask extends BitMask
{
  /**
   * Sets the alpha-channel operation (used in all <tt>set*()</tt> methods).
   *
   * @param  _op Alpha-channel operation (<tt>AlphaOperation.ALPHA_*</tt> constant).
   * @return The old alpha-channel operation.
   * @see    cz.cuni.jagrlib.AlphaOperation Alpha-channel operations
   */
  int setAlphaOperation ( int _op );

  /**
   * Sets the current default <tt>alpha</tt> - for inherited methods.
   *
   * @param  alpha Opacity factor (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void setAlpha ( double alpha );

  /**
   * Sets the given pixel's value to <tt>alpha</tt>.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  alpha Opacity factor (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void putPixel ( int x, int y, double alpha );

  /**
   * Returns the pixel's value.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel opacity (<tt>0.0</tt> for off-canvas access).
   */
  double getAlpha ( int x, int y );

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y-1].
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  alpha Opacity factor (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void setHLine ( int x1, int x2, int y, double alpha );

  /**
   * Sets the horizontal line's values.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  alpha Opacity factors (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void putHLine ( int x1, int x2, int y, double[] alpha );

  /**
   * Returns the pixels' values for the given horizontal line.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  alpha Returned values in <tt>double[]</tt> array (<tt>0.0</tt> for
   *               off-canvas access).
   */
  void getHLine ( int x1, int x2, int y, double[] alpha );

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1].
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  alpha Opacity factor (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void setRectangle ( int x1, int y1, int x2, int y2, double alpha );

  /**
   * Sets the rectangle's values.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  alpha Opacity factors (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void putRectangle ( int x1, int y1, int x2, int y2, double[][] alpha );

  /**
   * Returns the pixels' values for the given rectangle.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  alpha Returned values in <tt>double[][]</tt> array (<tt>0.0</tt> for
   *               off-canvas access).
   */
  void getRectangle ( int x1, int y1, int x2, int y2, double[][] alpha );

}
