package cz.cuni.jagrlib.iface;

import cz.cuni.jagrlib.*;

/**
 * Abstract BSSRDF (Bidirectional Surface Scattering Distribution Function).
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 840 $ $Date: 2009-01-10 11:44:24 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/BSSRDF.java">BSSRDF.java</a>
 * @see     LightSource
 */
public interface BSSRDF extends Property
{
  /** Common material property: color. */
  String MATERIAL_COLOR = "Color";

  /** Common material property: ambient coefficient. */
  String MATERIAL_KA    = "Ka";

  /** Common material property: diffuse coefficient. */
  String MATERIAL_KD    = "Kd";

  /** Common material property: specular coefficient. */
  String MATERIAL_KS    = "Ks";

  /** Common material property: specular exponent. */
  String MATERIAL_H     = "H";

  /** Common material property: transparency. */
  String MATERIAL_KT    = "Kt";

  /** Common material property: index of refraction. */
  String MATERIAL_N     = "N";

  /** Component mask: Diffuse component. */
  int COMP_DIFFUSE         = 1;

  /** Component mask: Specular reflection component. */
  int COMP_SPEC_REFLECTION = 2;

  /** Component mask: Specular reflection component. */
  int COMP_SPEC_REFRACTION = 4;

  /** Component mask: Glossy reflection component. */
  int COMP_GLOSSY          = 5;

  /** Component mask: both specular components (used in ray-tracing). */
  int COMP_SPECULAR        = COMP_SPEC_REFLECTION + COMP_SPEC_REFRACTION;


  /** Component mask: all components (used with light sources). */
  int COMP_ALL             = COMP_DIFFUSE + COMP_SPECULAR + COMP_GLOSSY;

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as monochrome).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @return BSSRDF coefficient <b>f(in-&gt;out)</b>.
   * @see    #monoBSSRDF(MicroFacet,double[],double[],int)
   * @see    #colorBSSRDF(double[],double[],double[],int,double[])
   */
  double monoBSSRDF ( boolean worldCoords, double[] in, double[] out, double[] normal,
                      double[] tangent, int mask );

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as monochrome).
   * Computations are done relatively to the given {@link MicroFacet}
   * (point on surface of some solid).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  mask Which light components are to be computed?
   * @return BSSRDF coefficient <b>f(in-&gt;mf-&gt;out)</b>.
   * @see    #monoBSSRDF(double[],double[],double[],int)
   * @see    #colorBSSRDF(MicroFacet,double[],double[],int,double[])
   */
  double monoBSSRDF ( boolean worldCoords, MicroFacet mf, double[] in, double[] out, int mask );

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as band-color).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BSSRDF spectral coefficients.
   * @see    #monoBSSRDF(double[],double[],double[],int)
   * @see    #colorBSSRDF(MicroFacet,double[],double[],int,double[])
   */
  double[] colorBSSRDF ( boolean worldCoords, double[] in, double[] out,
                         double[] normal, double[] tangent, int mask, double[] result );

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as band-color).
   * Computations are done relatively to the given {@link MicroFacet}
   * (point on surface of some solid).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;mf-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BSSRDF spectral coefficients.
   * @see    #monoBSSRDF(MicroFacet,double[],double[],int)
   * @see    #colorBSSRDF(double[],double[],double[],int,double[])
   */
  double[] colorBSSRDF ( boolean worldCoords, MicroFacet mf, double[] in,
                         double[] out, int mask, double[] result );

  /**
   * Retrieves the original color (defining surface color).
   *
   * @param  color Pre-allocated {@code double[C]} array to hold the color
   *               ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *               color system/ or greater /for more complex spectral color sampling/). Can
   *               be {@code null}.
   * @return Array containing copy of original color (can be altered).
   */
  double[] getOrigColor ( double[] color );

  /**
   * Sets the actual color (for the given BSSRDF data).
   * This color will be used in subsequent BSSRDF calculations.
   * The color array won't be altered.
   *
   * @param  color Color definition - {@code double[C]} array
   *                ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *                color system/ or greater /for more complex spectral color sampling/).
   */
  void setColor ( double[] color );

  /**
   * Scatter ray according to the light model.
   *
   * @param  mf   The {@link MicroFacet} (point on surface of some solid).
   * @param  in   Input direction (should look from the hit point)..
   * @param  u    Random uniform number.
   * @param  v    Random uniform number.
   * @param  mask Which light components are to be computed?
   * @param  out  Output direction in world coordinates.
   * @param  brdf Color definition - {@code double[C]} array
   *              ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *              color system/ or greater /for more complex spectral color sampling/).
   * @return Probability density function.
   */
  double scatter ( MicroFacet mf,
                   double[] in,
                   double u,
                   double v,
                   int mask,
                   double[] out,
                   double[] brdf );

  /**
   * Returns the probability density function.
   *
   * @param worldCoords Should be set to {@code true} if vectors in and out
   *                    are in world coordinates and to {@code false}
   *                    if vectors in and out are in local coordinates.
   * @param mf  The {@link MicroFacet} (point on surface of some solid).
   * @param in  Input direction.
   * @param out Output direction.
   */
  double pdf ( boolean worldCoords, MicroFacet mf, double[] in, double[] out );

  /** Returns the type of BRDF. */
  int type ();

  /** Transform local coordinates to world coordinates. */
  double[] localToWorld ( double[] normal,
                          double[] tangent,
                          double[] src,
                          double[] dest );

  /** Transform world coordinates to local coordinates. */
  double[] worldToLocal ( double[] normal,
                          double[] tangent,
                          double[] src,
                          double[] dest );

}