package cz.cuni.jagrlib.iface;

import java.io.*;

/**
 * Abstract sequential bit-input/output.
 *
 * @version 0.26 $Rev: 551 $ $Date: 2007-10-10 19:58:19 +0200 (st, 10 X 2007) $ $Author: pepca $
 * @since   0.04
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/BitStream.java">BitStream.java</a>
 */
public interface BitStream extends Property
{
  /** Stream-name property (String): file-name etc. */
  String STREAM_NAME = "Stream Name";

  /** Default file name. */
  String DEFAULT_STREAM_NAME = "output.bin";

  /**
   * Stream is able to read data.
   * If {@code true}, stream can be opened automatically in the init-phase.
   */
  boolean canRead ();

  /**
   * Stream initialization.
   *
   * @param  output Open for output?
   * @param  name Stream name (either stream identification e.g. file-name). {@code null}
   *              for implicit bit-stream ({@link #reOpen}).
   * @param  detail Further implementation-specific information.
   * @throws IOException
   * @see    #reOpen
   */
  void open ( boolean output, String name, String detail )
  throws IOException;

  /**
   * Stream re-initialization.
   *
   * @param  output Open for output?
   * @param  detail Further implementation-specific information.
   * @throws IOException
   * @see    #open
   */
  void reOpen ( boolean output, String detail )
  throws IOException;

  /**
   * Retrieves the associated output stream (non-mandatory).
   *
   * @return The associated {@link OutputStream} instance or {@code null} if
   *         not implemented.
   * @see    #getInputStream
   */
  OutputStream getOutputStream ();

  /**
   * Retrieves the associated input stream (non-mandatory).
   *
   * @return The associated {@link InputStream} instance or {@code null} if
   *         not implemented.
   * @see    #getOutputStream
   */
  InputStream getInputStream ();

  /**
   * Stream flush.
   *
   * @throws IOException
   * @see    #open
   * @see    #close
   */
  void flush ()
  throws IOException;

  /**
   * Stream close.
   * Eventually flushes all written data and disconnects the stream.
   *
   * @throws IOException
   * @see    #open
   * @see    #reOpen
   */
  void close ()
  throws IOException;

  /**
   * Actual stream position in bits.
   * After the output stream is closed, returns maximal written position.
   *
   * @return Actual (maximal) stream position in bits.
   * @throws IOException
   * @see    #available
   * @see    #seek
   */
  long position ()
  throws IOException;

  /**
   * How many bits are available to read (or skip).
   *
   * @return Number of bits available to read (or skip).
   * @throws IOException
   * @see    #position
   * @see    #seek
   */
  long available ()
  throws IOException;

  /**
   * Sets actual position of the stream in bits.
   * Non mandatory.
   *
   * @param  pos Required stream position in bits.
   * @return New stream position in bits (or {@code -1L} if not implemented).
   * @throws IOException
   * @see    #position
   */
  long seek ( long pos )
  throws IOException;

  /**
   * Writes a bunch of bits into the stream.
   *
   * @param  bits Bits to be written (LSB first).
   * @param  length Number of bits to write (negative lengths are ignored).
   * @throws IOException
   * @see    #writeBytes
   * @see    #read
   */
  void write ( long bits, int length )
  throws IOException;

  /**
   * Writes a bunch of bytes into the stream.
   *
   * @param  data Bytes to be written (each byte is written in "LSB first" order).
   * @param  start First index to write. Must be less than {@code data.length}.
   * @param  length Number of bytes to write ({@code start + length > data.length} is ignored).
   * @throws IOException
   * @see    #write
   * @see    #readBytes
   */
  void writeBytes ( byte[] data, int start, int length )
  throws IOException;

  /**
   * Reads a bunch of bits from the stream.
   *
   * @param  length Number of bits to read (negative lengths are ignored).
   * @return Read bits (LSB first).
   * @throws IOException
   * @see    #readBytes
   * @see    #write
   */
  long read ( int length )
  throws IOException;

  /**
   * Reads a bunch of bytes from the stream.
   *
   * @param  data Array holding the read bytes (each byte is read in "LSB first" order).
   * @param  start First index to read. Must be less than {@code data.length}.
   * @param  length Number of bytes to read ({@code start + length > data.length} is ignored).
   * @throws IOException
   * @see    #read
   * @see    #writeBytes
   */
  void readBytes ( byte[] data, int start, int length )
  throws IOException;

  /**
   * Skips the given number of bits.
   *
   * @param  length Number of bits to skip (negative lengths are ignored).
   * @return Number of bits actually skipped.
   * @throws IOException
   * @see    #read
   * @see    #seek
   */
  long skip ( long length )
  throws IOException;

}
