package cz.cuni.jagrlib.iface;

import java.io.*;

/**
 * Codec for integer block data.
 * In <b>coding mode</b> external {@link EntropyCodec} module is feeded by linearized data.
 * <br>In <b>decoding mode</b> input is read from {@link EntropyCodec} and converted back to integer
 * matrix.
 *
 * @version 0.25 $Rev: 250 $ $Date: 2006-05-14 22:37:12 +0200 (ne, 14 V 2006) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/BlockCodec.java">BlockCodec.java</a>
 */
public interface BlockCodec extends Property, PersistentParamsBinary
{
  /**
   * Prepares the underlying {@link EntropyCodec}.
   * Should be called before any of {@link #encode}s or {@link #decode}s.
   */
  void init ()
  throws IOException;

  /**
   * Set actual codec variant (context).
   * Should set mode of the underlying {@link EntropyCodec} too.
   *
   * @param  variant Variant number (from <tt>0</tt>).
   * @return Old variant number.
   * @see    #setBlockSize
   */
  int setVariant ( int variant );

  /**
   * Sets actual block size.
   * Can be used both for 1D and 2D quantization.

   * @param  width Horizontal block size (1D block size).
   * @param  height Vertical block size (<tt>1</tt> for 1D).
   * @see    #setVariant
   */
  void setBlockSize ( int width, int height )
  throws IOException;

  /**
   * Block coding.
   * Converts matrix of integer data to {@link EntropyCodec} input.
   * Connected {@link EntropyCodec} module will be feeded by them.
   *
   * @param  data Source data to be encoded.
   * @param  predict Pre-allocated array used for predicition passing.
   *         <tt>predict[1]</tt> will be used as source, <tt>predict[0]</tt> is filled with
   *         prediction for the next block[s]. Can be <tt>null</tt>.
   * @see    #decode
   */
  void encode ( int[] data, int[][] predict )
  throws IOException;

  /**
   * Block decoding.
   * Reads input from underlying {@link EntropyCodec} and converts them back into
   * given integer matrix.
   *
   * @param  data Target data array (pre-allocated to sufficient size).
   * @param  predict Pre-allocated array used for predicition passing.
   *         <tt>predict[1]</tt> will be used as source, <tt>predict[0]</tt> is filled with
   *         prediction for the next block[s]. Can be <tt>null</tt>.
   * @return Number of decoded array items (should be <tt>with * height</tt>
   *         if successful).
   * @see    #encode
   */
  int decode ( int[] data, int[][] predict )
  throws IOException;

}
