package cz.cuni.jagrlib.iface;

/**
 * Quantization of block discrete data.
 *
 * @version 0.25 $Rev: 246 $ $Date: 2006-05-11 01:38:39 +0200 (čt, 11 V 2006) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/BlockQuantizer.java">BlockQuantizer.java</a>
 */
public interface BlockQuantizer extends Property, PersistentParamsBinary
{
  /**
   * Set actual quantizer variant (context).
   *
   * @param  variant Variant number (from <tt>0</tt>).
   * @return Old variant number.
   * @see    #setBlockSize
   * @see    #setQuality
   * @see    #setData
   */
  int setVariant ( int variant );

  /**
   * Sets actual block size.
   * Can be used both for 1D and 2D quantization.

   * @param  width Horizontal block size (1D block size).
   * @param  height Vertical block size (<tt>1</tt> for 1D).
   * @see    #setVariant
   * @see    #setQuality
   * @see    #setData
   */
  void setBlockSize ( int width, int height );

  /**
   * Sets actual quality coefficient.
   * Value range for the quality: <tt>0.0f</tt> to <tt>1.0f</tt>.
   *
   * @param  quality New quality coefficient.
   * @see    #setVariant
   * @see    #setBlockSize
   * @see    #setData
   */
  void setQuality ( float quality );

  /**
   * Sets quantization table for the current variant.
   *
   * @param  qtable New quantization table.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @see    #setVariant
   * @see    #setQuality
   * @see    #setBlockSize
   */
  void setData ( int[] qtable );

  /**
   * Quantization (floating-point form).
   *
   * @param  coef Source data (coefficients). Size = array size.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @param  quant Pre-allocated result array or <tt>null</tt> if
   *         it's size needs to be determined.
   * @return Result array size (even if the quantization was not performed).
   * @see    #quantize(int[],int[])
   * @see    #dequantize(int[],double[])
   */
  int quantize ( double[] coef, int[] quant );

  /**
   * Quantization (integer form).
   *
   * @param  coef Source data (coefficients). Size = array size.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @param  quant Pre-allocated result array or <tt>null</tt> if
   *         it's size needs to be determined.
   * @return Result array size (even if the quantization was not performed).
   * @see    #quantize(double[],int[])
   * @see    #dequantize(int[],int[])
   */
  int quantize ( int[] coef, int[] quant );

  /**
   * Dequantization = reconstruction (floating-point form).
   *
   * @param  quant Source data (quantized coefficients). Size = array size.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @param  coef Pre-allocated coefficient array or <tt>null</tt> if
   *         it's size needs to be determined.
   * @return Coefficient array size (even if the dequantization was not performed).
   * @see    #dequantize(int[],int[])
   * @see    #quantize(double[],int[])
   */
  int dequantize ( int[] quant, double[] coef );

  /**
   * Dequantization = reconstruction (integer form).
   *
   * @param  quant Source data (quantized coefficients). Size = array size.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @param  coef Pre-allocated coefficient array or <tt>null</tt> if
   *         it's size needs to be determined.
   * @return Coefficient array size (even if the dequantization was not performed).
   * @see    #dequantize(int[],double[])
   * @see    #quantize(int[],int[])
   */
  int dequantize ( int[] quant, int[] coef );

}
