package cz.cuni.jagrlib.iface;

import java.util.*;

import cz.cuni.jagrlib.*;

/**
 * Boundary representation (B-rep) of 3D scene. Suitable for both compact and topological
 * 3D databases. Restricted mode uses triangle faces only, general mode could contain
 * arbitrary simple polygons as faces.
 *
 * <p>Vertices, edges, faces, divisions and solids (Vs, Es, Fs, Ds, Ss) are represented by their
 * <b>handles</b> (integer indices). Arbitrary <b>attributes</b> can be associated with Vs, Es, Fs,
 * Ds or Ss by the user. Implicit attribute: vertex coordinates. Attributes are accessed by string
 * identifiers or integer shortcuts.
 * </p>
 *
 * <p>More than one <b>context</b> can be used (e.g. for different scene views, rendering modes).
 * Handles (indices) are shared among all contexts, some attributes can be shared, too.
 * </p>
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.09
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/Brep.java">Brep.java</a>
 */
public interface Brep extends Property
{
  /** Error completion code. */
  int ERROR = -1;

  /** Empty (irelevant) object / parameter. */
  int NULL  = -1;

  //==========================================================================
  //  Database entities:

  /** Global {@link Brep} data / attribute association. */
  int SCENE    = 0;

  /** Context data / attribute association. */
  int CONTEXT  = 1;

  /** Vertex data / attribute association. */
  int VERTEX   = 2;

  /** Edge data / attribute association. */
  int EDGE     = 3;

  /** Face data / attribute association. */
  int FACE     = 4;

  /** Division data / attribute association. */
  int DIVISION = 5;

  /** Solid data / attribute association. */
  int SOLID    = 6;

  //==========================================================================
  //  Contexts:

  /** Default context identifier. */
  int CTX_DEFAULT = 0;

  //==========================================================================
  //  Attribute types:

  /** General attribute type ({@link java.lang.Object Object} successor). */
  int ATTR_OBJECT     =  0;

  /** {@code boolean} attribute type. */
  int ATTR_BOOLEAN    =  1;

  /** {@code int} attribute type. */
  int ATTR_INTEGER    =  2;

  /** {@code long} attribute type. */
  int ATTR_LONG       =  3;

  /** {@code float} attribute type. */
  int ATTR_FLOAT      =  4;

  /** {@code double} attribute type. */
  int ATTR_DOUBLE     =  5;

  /** 2D {@code float} coordinates [x,y]. */
  int ATTR_COORD_2F   =  6;

  /** 2D {@code double} coordinates [x,y]. */
  int ATTR_COORD_2D   =  7;

  /** 2D {@code int} coordinates [x,y]. */
  int ATTR_COORD_2I   =  8;

  /** 2D {@code float} homogeneous coordinates [x,y,w]. */
  int ATTR_COORDH_3F  =  9;

  /** 2D {@code double} homogeneous coordinates [x,y,w]. */
  int ATTR_COORDH_3D  = 10;

  /** 3D {@code float} coordinates [x,y,z]. */
  int ATTR_COORD_3F   = 11;

  /** 3D {@code double} coordinates [x,y,z]. */
  int ATTR_COORD_3D   = 12;

  /** 3D {@code int} coordinates [x,y,z]. */
  int ATTR_COORD_3I   = 13;

  /** 3D {@code float} homogeneous coordinates [x,y,z,w]. */
  int ATTR_COORDH_4F  = 14;

  /** 3D {@code double} homogeneous coordinates [x,y,z,w]. */
  int ATTR_COORDH_4D  = 15;

  /** RGB color attribute type (float[3]). */
  int ATTR_RGB_3F     = 16;

  /** RGBA color attribute type (float[4]). */
  int ATTR_RGBA_4F    = 17;

  /** Order number of the last attribute type. */
  int ATTR_LAST       = ATTR_RGBA_4F;

  //==========================================================================
  //  Interface Property:

  String PRIMARY_COORD_TYPE   = "Primary coords";
  String PROJECTED_COORD_TYPE = "Projected coords";
  String WEDGE_MODE           = "Winged edge";

  //==========================================================================
  //  Implicit attributes (coordinates):

  /** String identifier of implicit coordinate attribute with {@code double} components. */
  String COORD      = "Coord";

  /** String identifier of implicit projected coordinate attribute. */
  String COORD_PROJ = "Coord-proj";

  /** String identifier of implicit data-attribute (edge, solid, etc.). */
  String DATA       = "Data";

  //==========================================================================
  //  Other attribute-names:

  /** Color (vertex, face, solid?). */
  String COLOR      = "Color";

  /** Normal vector (vertex, face). */
  String NORMAL     = "Normal";

  /** Texture coordinates (1D, 2D or 3D). */
  String TEXTURE_0  = "Texture0";
  String TEXTURE_1  = "Texture1";
  String TEXTURE_2  = "Texture2";
  String TEXTURE_3  = "Texture3";

  /** Solid id (face). */
  String SOLID_A    = "Solid";

  /** Material id (face, solid?). */
  String MATERIAL_A = "Material";

  //==========================================================================
  //  Winged-edge structure:

  /**
   * Size of regular edge record (can be stored in one big array).
   */
  int EDGE_SIZE = 2;

  /**
   * Size of winged-edge record (all items are integer numbers /handles/
   * and can be stored in one big array).
   */
  int WEDGE_SIZE = 8;

  /** Winged-edge-offset: 1st vertex (V1). */
  int WEDGE_V1 = 0;

  /** Winged-edge-offset: 2nd vertex (V2). */
  int WEDGE_V2 = 1;

  /** Winged-edge-offset: the left face (FL). */
  int WEDGE_FL = 2;

  /** Winged-edge-offset: the right face (FR). */
  int WEDGE_FR = 3;

  /** Winged-edge-offset: predecessor in the left face (LP). */
  int WEDGE_LP = 4;

  /** Winged-edge-offset: successor in the left face (LN). */
  int WEDGE_LN = 5;

  /** Winged-edge-offset: predecessor in the right face (RP). */
  int WEDGE_RP = 6;

  /** Winged-edge-offset: successor in the right face (RN). */
  int WEDGE_RN = 7;

  //==========================================================================
  //  Database setup:

  /**
   * Set default coordinate types.
   *
   * @param  baseType Base type for the {@link #COORD} attribute.
   * @param  baseProjType Base type for the {@link #COORD_PROJ} (projected) attribute.
   */
  void setCoordTypes ( int baseType, int baseProjType );

  /**
   * Set winged-edge mode.
   * Default value: {@code false} (don't use winged edge).
   *
   * @param  wedge Winged-edges will be used?
   * @see    #getEdgeRecord
   * @see    #WEDGE_SIZE
   */
  void setEdgeMode ( boolean wedge );

  /**
   * Database (re-)initialization. Removes all scene contexts and implicit / explicit
   * attributes.
   */
  void init ();

  //==========================================================================
  //  Scene contexts:

  /**
   * Creates a new scene context (can be pre-initialized from one of existing contexts).
   * No attributes (even implicit) are created for the new context (exception: pre-initialized
   * contexts).
   *
   * @param  ctxFrom Context handle for initialization (non-mandatory parameter -
   *                 can be {@link #NULL}).
   * @return Handle to a new context.
   * @see    #removeContext
   * @see    #copyAttribute(int,int)
   * @see    #copyAttribute(int,int,String)
   */
  int createContext ( int ctxFrom );

  /**
   * Removes an existing scene context. {@link #CTX_DEFAULT} cannot be removed.
   *
   * @param  ctx Handle to a context to be destroyed.
   * @see    #createContext
   */
  void removeContext ( int ctx );

  //==========================================================================
  //  Scene attributes:

  /**
   * Creates a new scene attribute.
   *
   * @param  where Entity which a new attribute will be associated with ({@link #VERTEX},
   *               {@link #EDGE}, ..).
   * @param  ctx Handle to a context (can be {@link #NULL} for shared attributes).
   * @param  ident String identifier of the attribute.
   * @param  type Attribute type ({@code ATTR_*} constant).
   * @return Handle to a new attribute.
   * @see    #copyAttribute(int,int,String)
   * @see    #copyAttribute(int,int)
   * @see    #removeAttribute
   * @see    #copyCoordinates(int,int,String)
   * @see    #roundCoordinates(int,int,String,TrMatrix)
   * @see    #transformCoordinates(int,int,String,TrMatrix)
   */
  int createAttribute ( int where, int ctx, String ident, int type );

  /**
   * Copies the existing scene attribute to a new one (preserves attribute association
   * /{@code where}/ and {@code type}).
   *
   * @param  attrFrom Handle to a source attribute.
   * @param  ctx Handle to a context (cannot be {@link #NULL}).
   * @param  ident String identifier of the attribute.
   * @return Handle to a new attribute.
   * @see    #createAttribute
   * @see    #copyAttribute(int,int)
   * @see    #removeAttribute
   * @see    #copyCoordinates(int,int,String)
   * @see    #roundCoordinates(int,int,String,TrMatrix)
   * @see    #transformCoordinates(int,int,String,TrMatrix)
   */
  int copyAttribute ( int attrFrom, int ctx, String ident );

  /**
   * Copies the existing scene attribute to a new one (preserves attribute association
   * /{@code where}/, {@code type} and identifier string /{@code ident}/).
   *
   * @param  attrFrom Handle to a source attribute.
   * @param  ctx Handle to a context (cannot be {@link #NULL}).
   * @return Handle to a new attribute.
   * @see    #createAttribute
   * @see    #copyAttribute(int,int,String)
   * @see    #removeAttribute
   * @see    #copyCoordinates(int,int,String)
   * @see    #roundCoordinates(int,int,String,TrMatrix)
   * @see    #transformCoordinates(int,int,String,TrMatrix)
   */
  int copyAttribute ( int attrFrom, int ctx );

  /**
   * Removes the given scene attribute.
   *
   * @param  attr Handle to an attribute.
   * @see    #copyAttribute(int,int,String)
   * @see    #copyAttribute(int,int)
   * @see    #createAttribute
   */
  void removeAttribute ( int attr );

  /**
   * Finds a handle of the given scene attribute.
   *
   * @param  where Entity which the attribute is associated with ({@link #VERTEX},
   *               {@link #EDGE}, ..).
   * @param  ctx Handle to a context (can be {@link #NULL} for shared attributes).
   * @param  ident String identifier of the attribute.
   * @return Handle to the given attribute.
   */
  int getAttributeId ( int where, int ctx, String ident );

  /**
   * Returns a {@code where} parameter of the given attribute (entity which the attribute
   * is associated with).
   *
   * @param  attr Handle to the attribute.
   * @return The {@code where} parameter.
   */
  int getAttributeWhere ( int attr );

  /**
   * Returns a context of the given attribute (for non-shared attributes only).
   *
   * @param  attr Handle to the attribute.
   * @return Exclusive context of the attribute (or {@link #NULL}).
   */
  int getAttributeContext ( int attr );

  /**
   * Returns a string identifier of the given attribute.
   *
   * @param  attr Handle to the attribute.
   * @return String identifier of the attribute.
   */
  String getAttributeIdent ( int attr );

  /**
   * Returns a {@code type} of the given attribute.
   *
   * @param  attr Handle to the attribute.
   * @return The {@code type} parameter.
   */
  int getAttributeType ( int attr );

  /**
   * Retrieves value of the given attribute for the given entity (general form).
   *
   * @param  where Entity which the attribute is associated with ({@link #VERTEX},
   *               {@link #EDGE}, ..).
   * @param  ctx Handle to a context (can be {@link #NULL} for shared attributes).
   * @param  ident String identifier of the attribute.
   * @param  id Handle to the entity.
   * @param  pre Pre-allocated result object (can be {@code null}).
   * @return The value of the attribute for the given entity.
   */
  Object getAttribute ( int where, int ctx, String ident, int id, Object pre );

  /**
   * Retrieves value of the given attribute for the given entity (general form).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @param  pre Pre-allocated result object (can be {@code null}).
   * @return The value of the attribute for the given entity.
   */
  Object getAttribute ( int attr, int id, Object pre );

  /**
   * Retrieves value of the given coordinate-attribute for the given entity (for
   * {@code float} [coordinate] types).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @param  coord Optional pre-allocated result array.
   * @return Result array ({@code coord} if possible).
   */
  float[] getAttribute ( int attr, int id, float[] coord );

  /**
   * Retrieves value of the given coordinate-attribute for the given entity (for
   * {@code double} [coordinate] types).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @param  coord Optional pre-allocated result array.
   * @return Result array ({@code coord} if possible).
   */
  double[] getAttribute ( int attr, int id, double[] coord );

  /**
   * Retrieves value of the given coordinate-attribute for the given entity (for
   * {@code int} coordinate types).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @param  coord Optional pre-allocated result array.
   * @return Result array ({@code coord} if possible).
   */
  int[] getAttribute ( int attr, int id, int[] coord );

  /**
   * Retrieves value of the given flag-attribute for the given entity (for
   * {@code boolean} attribute type).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @return Boolean attribute value.
   */
  boolean getBoolAttribute ( int attr, int id );

  /**
   * Retrieves value of the given integer-attribute for the given entity (for
   * {@code int} attribute type).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @return Integer attribute value.
   */
  int getIntAttribute ( int attr, int id );

  /**
   * Retrieves value of the given integer-attribute for the given entity (for
   * {@code long} attribute type).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @return Long-integer attribute value.
   */
  long getLongAttribute ( int attr, int id );

  /**
   * Retrieves value of the given float-attribute for the given entity (for
   * {@code float} attribute type).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @return Double attribute value.
   */
  float getFloatAttribute ( int attr, int id );

  /**
   * Retrieves value of the given float-attribute for the given entity (for
   * {@code double} attribute type).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @return Double attribute value.
   */
  double getDoubleAttribute ( int attr, int id );

  /**
   * Sets value of the given attribute of the given entity (general form).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @param  val New attribute value (can be {@code null}).
   */
  void setAttribute ( int attr, int id, Object val );

  //==========================================================================
  //  Implicit attributes (coordinates):

  /**
   * Copies data from coordinate-attribute to another one. Destination attribute is implicitly
   * created.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  ctxTo Handle to a context (can be {@link #NULL} for shared attributes).
   * @param  identTo String identifier of the attribute.
   * @return Handle to a new coordinate-attribute.
   * @see    #createAttribute
   * @see    #roundCoordinates(int,int,String,TrMatrix)
   * @see    #transformCoordinates(int,int,String,TrMatrix)
   */
  int copyCoordinates ( int attrFrom, int ctxTo, String identTo );

  /**
   * Copies data from coordinate-attribute to another one. Destination attribute must exist before,
   * copying can be restricted to the given handle-set.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  attrTo Handle to a destination coordinate attribute.
   * @param  set Entity handle-set (for {@code null} everything will be copied).
   * @see    #createAttribute
   * @see    #roundCoordinates(int,int,TrMatrix,BitSet)
   * @see    #transformCoordinates(int,int,TrMatrix,BitSet)
   */
  void copyCoordinates ( int attrFrom, int attrTo, BitSet set );

  /**
   * Transforms data from coordinate-attribute to another one. Projection-plane transformation
   * (and roundup) is done implicitly, destination attribute is created implicitly, too.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  ctxTo Handle to a context (can be {@link #NULL} for shared attributes).
   * @param  identTo String identifier of the attribute.
   * @param  tr Linear (projection-plane) transformation.
   * @return Handle to a new coordinate-attribute.
   * @see    #createAttribute
   * @see    #copyCoordinates(int,int,String)
   * @see    #transformCoordinates(int,int,String,TrMatrix)
   */
  int roundCoordinates ( int attrFrom, int ctxTo, String identTo, TrMatrix tr );

  /**
   * Transforms data from coordinate-attribute to another one. Projection-plane transformation
   * (and roundup) is done implicitly, destination attribute must exist before.
   * Transformation can be restricted to the given handle-set.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  attrTo Handle to a destination coordinate attribute.
   * @param  tr Linear (projection-plane) transformation.
   * @param  set Entity handle-set (for {@code null} everything will be transformed).
   * @see    #createAttribute
   * @see    #copyCoordinates(int,int,BitSet)
   * @see    #transformCoordinates(int,int,TrMatrix,BitSet)
   */
  void roundCoordinates ( int attrFrom, int attrTo, TrMatrix tr, BitSet set );

  /**
   * Transforms data from coordinate-attribute to another one. Linear transformation
   * is done implicitly, destination attribute is created implicitly, too.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  ctxTo Handle to a context (can be {@link #NULL} for shared attributes).
   * @param  identTo String identifier of the attribute.
   * @param  tr Linear matrix transformation.
   * @return Handle to a new coordinate-attribute.
   * @see    #createAttribute
   * @see    #copyCoordinates(int,int,String)
   * @see    #roundCoordinates(int,int,String,TrMatrix)
   */
  int transformCoordinates ( int attrFrom, int ctxTo, String identTo, TrMatrix tr );

  /**
   * Transforms data from coordinate-attribute to another one. Linear transformation
   * is done implicitly, destination attribute must exist before.
   * Transformation can be restricted to the given handle-set.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  attrTo Handle to a destination coordinate attribute.
   * @param  tr Linear matrix transformation.
   * @param  set Entity handle-set (for {@code null} everything will be transformed).
   * @see    #createAttribute
   * @see    #copyCoordinates(int,int,BitSet)
   * @see    #roundCoordinates(int,int,TrMatrix,BitSet)
   */
  void transformCoordinates ( int attrFrom, int attrTo, TrMatrix tr, BitSet set );

  //==========================================================================
  //  Iterators:

  /**
   * Initializes iterator for all vertices in the scene.
   *
   * @param  pre Pre-allocated {@link BrepIterator} object (ca be {@code null}).
   * @return The initialized iterator data.
   * @see    #vertexInFaceIterator
   * @see    #nextItem
   */
  BrepIterator vertexIterator ( BrepIterator pre );

  /**
   * Initializes iterator for vertices within the given face.
   *
   * @param  face Handle to the face.
   * @param  pre Pre-allocated {@link BrepIterator} object (ca be {@code null}).
   * @return The initialized iterator data.
   * @see    #vertexIterator
   * @see    #nextItem
   */
  BrepIterator vertexInFaceIterator ( int face, BrepIterator pre );

  /**
   * Initializes iterator for all edges in the scene.
   *
   * @param  pre Pre-allocated {@link BrepIterator} object (ca be {@code null}).
   * @return The initialized iterator data.
   * @see    #edgeInFaceIterator
   * @see    #nextItem
   */
  BrepIterator edgeIterator ( BrepIterator pre );

  /**
   * Initializes iterator for edges within the given face.
   *
   * @param  face Handle to the face.
   * @param  pre Pre-allocated {@link BrepIterator} object (ca be {@code null}).
   * @return The initialized iterator data.
   * @see    #edgeIterator
   * @see    #nextItem
   */
  BrepIterator edgeInFaceIterator ( int face, BrepIterator pre );

  /**
   * Initializes iterator for all faces in the scene.
   *
   * @param  ctx Handle to a context.
   * @param  pre Pre-allocated {@link BrepIterator} object (ca be {@code null}).
   * @return The initialized iterator data.
   * @see    #faceInSolidIterator
   * @see    #nextItem
   */
  BrepIterator faceIterator ( int ctx, BrepIterator pre );

  /**
   * Initializes iterator for faces within the given solid.
   *
   * @param  ctx Handle to a context.
   * @param  solid Handle to the face.
   * @param  pre Pre-allocated {@link BrepIterator} object (ca be {@code null}).
   * @return The initialized iterator data.
   * @see    #faceIterator
   * @see    #nextItem
   */
  BrepIterator faceInSolidIterator ( int ctx, int solid, BrepIterator pre );

  /**
   * Initializes iterator for all divisions in the scene.
   *
   * @param  ctx Handle to a context.
   * @param  pre Pre-allocated {@link BrepIterator} object (ca be {@code null}).
   * @return The initialized iterator data.
   * @see    #divisionInSolidIterator
   * @see    #nextItem
   */
  BrepIterator divisionIterator ( int ctx, BrepIterator pre );

  /**
   * Initializes iterator for divisions within the given solid.
   *
   * @param  ctx Handle to a context.
   * @param  solid Handle to the face.
   * @param  pre Pre-allocated {@link BrepIterator} object (ca be {@code null}).
   * @return The initialized iterator data.
   * @see    #divisionIterator
   * @see    #nextItem
   */
  BrepIterator divisionInSolidIterator ( int ctx, int solid, BrepIterator pre );

  /**
   * Initializes iterator for all solids in the scene.
   *
   * @param  pre Pre-allocated {@link BrepIterator} object (ca be {@code null}).
   * @return The initialized iterator data.
   * @see    #nextItem
   */
  BrepIterator solidIterator ( BrepIterator pre );

  /**
   * Returns the next entity-handle.
   *
   * @param  iter Associated iterator data.
   * @return Handle to the next entity or {@link #NULL} at the end of an iteration.
   */
  int nextItem ( BrepIterator iter );

  //==========================================================================
  //  Basic data access (get* methods):

  /**
   * Returns integer coordinates of the given vertex (implicit attribute
   * {@link #COORD_PROJ}) (Cartesian 2D or 3D).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Optional pre-allocated coordinate vector.
   * @return Result array ({@code coord} if possible).
   * @see    #getVertexCoords(int,int,double[])
   * @see    #getVertexCoords(int,int,float[])
   */
  int[] getVertexCoords ( int ctx, int vert, int[] coord );

  /**
   * Returns floating-point coordinates of the given vertex (implicit attribute
   * {@link #COORD}).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Optional pre-allocated coordinate vector (Cartesian or homogeneous).
   * @return Result array ({@code coord} if possible).
   * @see    #getVertexCoords(int,int,int[])
   * @see    #getVertexCoords(int,int,double[])
   */
  float[] getVertexCoords ( int ctx, int vert, float[] coord );

  /**
   * Returns floating-point coordinates of the given vertex (implicit attribute
   * {@link #COORD}).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Optional pre-allocated coordinate vector (Cartesian or homogeneous).
   * @return Result array ({@code coord} if possible).
   * @see    #getVertexCoords(int,int,int[])
   * @see    #getVertexCoords(int,int,float[])
   */
  double[] getVertexCoords ( int ctx, int vert, double[] coord );

  /**
   * Returns vertex handles of the given edge.
   *
   * @param  edge Handle of the edge.
   * @param  vert Optional pre-allocated array to hold two vertex handles.
   * @return Result array ({@code vert} if possible).
   */
  int[] getEdgeVertices ( int edge, int[] vert );

  /**
   * Returns record of the given edge (useful particularly for winged-edges).
   * All edge records can be stored in one huge {@code int[]} array, user of this
   * method receives such array and origin of the particular edge record.
   *
   * @param  edge Handle of the edge.
   * @param  index Pre-allocated {@code int[]} array to receive starting index of the edge-record.
   *               If this array is longer than {@code int[1]}, actual record-size will be returned after
   *               the starting index. If {code index} is {@code null}, R/O edge record is returned.
   * @return Shared (read/write) array possibly holding all edge records.
   */
  int[] getEdgeRecord ( int edge, int[] index );

  /**
   * Returns edge handles of the given face.
   *
   * @param  face Handle of the face.
   * @param  edge Pre-allocated array to hold edge handles (or {@code null} to get the
   *              face size).
   * @return Number of face edges.
   * @see    #getFaceVertices
   */
  int getFaceEdges ( int face, int[] edge );

  /**
   * Returns vertex handles of the given face.
   *
   * @param  face Handle of the face.
   * @param  vert Pre-allocated array to hold vertex handles (or {@code null} to get the
   *              face size).
   * @return Number of face vertices.
   * @see    #getFaceEdges
   */
  int getFaceVertices ( int face, int[] vert );

  /**
   * Returns face handles of the given division.
   *
   * @param  ctx Handle of the context.
   * @param  div Handle of the division.
   * @param  up Pre-allocated array to hold predecessor-face handles (will be terminated by
   *            {@link #NULL}). {@code null} to get number of predecessors.
   * @param  down Pre-allocated array to hold successor-face handles (will be terminated by
   *              {@link #NULL}). {@code null} to get number of successors.
   * @return Number of successor faces (or number of predecessor faces for
   *         {@code up == null}).
   */
  int getDivisionFaces ( int ctx, int div, int[] up, int[] down );

  /**
   * Returns face handles of the given solid.
   *
   * @param  ctx Handle of the context.
   * @param  solid Handle of the solid.
   * @param  face Pre-allocated array to hold face handles (or {@code null} to get the
   *              solid size).
   * @return Number of faces.
   */
  int getSolidFaces ( int ctx, int solid, int[] face );

  //==========================================================================
  //  Global get*() methods (for faster iterations):

  /**
   * Returns all vertex handles in the database.
   *
   * @param  vert Pre-allocated array to hold vertex handles (or {@code null} to get
   *              the array size).
   * @return Array size.
   * @see    #getAllVertexCoords(int,int,int[])
   * @see    #getAllVertexCoords(int,int,double[])
   */
  int getAllVertices ( int[] vert );

  /**
   * Returns all integer vertex-coordinates in the database (Cartesian - 2D or 3D).
   *
   * @param  attr Handle to the attribute (e.g. {@link #COORD_PROJ}).
   * @param  dim Size of coordinate vector ({@code 2} or {@code 3}).
   * @param  coord Pre-allocated array to hold vertex coordinates (or {@code null} to get
   *               the array size).
   * @return Array size (number of vertices).
   * @see    #getAllVertices
   * @see    #getAllVertexCoords(int,int,double[])
   */
  int getAllVertexCoords ( int attr, int dim, int[] coord );

  /**
   * Returns all floating-point vertex-coordinates in the database (Cartesian or homogeneous).
   *
   * @param  attr Handle to the attribute (e.g. {@link #COORD}).
   * @param  dim Size of coordinate vector ({@code 2} to {@code 4}).
   * @param  coord Pre-allocated array to hold vertex coordinates (or {@code null} to get
   *               the array size).
   * @return Array size (number of vertices).
   * @see    #getAllVertices
   * @see    #getAllVertexCoords(int,int,int[])
   */
  int getAllVertexCoords ( int attr, int dim, float[] coord );

  /**
   * Returns all floating-point vertex-coordinates in the database (Cartesian or homogeneous).
   *
   * @param  attr Handle to the attribute (e.g. {@link #COORD}).
   * @param  dim Size of coordinate vector ({@code 2} to {@code 4}).
   * @param  coord Pre-allocated array to hold vertex coordinates (or {@code null} to get
   *               the array size).
   * @return Array size (number of vertices).
   * @see    #getAllVertices
   * @see    #getAllVertexCoords(int,int,int[])
   */
  int getAllVertexCoords ( int attr, int dim, double[] coord );

  /**
   * Returns all edge handles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  edge Pre-allocated array to hold edge handles (or {@code null} to get
   *              the array size).
   * @return Array size.
   * @see    #getAllEdgeVertices
   */
  int getAllEdges ( int ctx, int[] edge );

  /**
   * Returns vertex handles of all edges in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  vert Pre-allocated array to hold vertex handles (or {@code null} to get
   *              the array size).
   * @return Array size (number of edges).
   * @see    #getAllEdges
   */
  int getAllEdgeVertices ( int ctx, int[] vert );

  /**
   * Returns all face handles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  face Pre-allocated array to hold face handles (or {@code null} to get
   *              the array size).
   * @return Array size.
   */
  int getAllFaces ( int ctx, int[] face );

  /**
   * Returns edge handles of all triangles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  edge Pre-allocated array to hold edge handles (or {@code null} to get
   *              the array size). It will hold data only for triangle faces.
   * @return Array size (number of faces).
   * @see    #getAllTriangleVertices
   */
  int getAllTriangleEdges ( int ctx, int[] edge );

  /**
   * Returns vertex handles of all triangles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  vert Pre-allocated array to hold vertex handles (or {@code null} to get
   *              the array size). It will hold data only for triangle faces.
   * @return Array size (number of faces).
   * @see    #getAllTriangleEdges
   */
  int getAllTriangleVertices ( int ctx, int[] vert );

  /**
   * Returns all division handles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  div Pre-allocated array to hold division handles (or {@code null} to get
   *             the array size).
   * @return Array size.
   */
  int getAllDivisions ( int ctx, int[] div );

  /**
   * Returns all solid handles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  solid Pre-allocated array to hold solid handles (or {@code null} to get
   *               the array size).
   * @return Array size.
   */
  int getAllSolids ( int ctx, int[] solid );

  //==========================================================================
  //  Basic data access (create* and set* methods):

  /**
   * Creates a new vertex.
   *
   * @return Handle of the new vertex.
   */
  int createVertex ();

  /**
   * Sets integer coordinates of the given vertex (implicit attribute
   * {@link #COORD_PROJ}) (Cartesian 2D or 3D).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Coordinate vector (2D or 3D).
   * @see    #createVertex
   */
  void setVertexCoords ( int ctx, int vert, int[] coord );

  /**
   * Sets floating-point coordinates of the given vertex (implicit attribute
   * {@link #COORD}) (Cartesian or homogeneous).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Coordinate vector (2D to 4D).
   * @see    #createVertex
   */
  void setVertexCoords ( int ctx, int vert, float[] coord );

  /**
   * Sets floating-point coordinates of the given vertex (implicit attribute
   * {@link #COORD}) (Cartesian or homogeneous).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Coordinate vector (2D to 4D).
   * @see    #createVertex
   */
  void setVertexCoords ( int ctx, int vert, double[] coord );

  /**
   * Creates a new edge.
   *
   * @return Handle of the new edge.
   */
  int createEdge ();

  /**
   * Sets vertex handles of the given edge.
   *
   * @param  edge Handle of the edge.
   * @param  v1 Handle of the 1st vertex.
   * @param  v2 Handle of the 2nd vertex.
   * @see    #createEdge
   */
  void setEdgeVertices ( int edge, int v1, int v2 );

  /**
   * Creates a new face.
   *
   * @return Handle of the new face.
   */
  int createFace ();

  /**
   * Sets edge handles of the given triangle.
   *
   * @param  face Handle of the face.
   * @param  e1 Handle of the 1st edge.
   * @param  e2 Handle of the 2nd edge.
   * @param  e3 Handle of the 3rd edge.
   * @see    #createFace
   */
  void setTriangleEdges ( int face, int e1, int e2, int e3 );

  /**
   * Sets vertex handles of the given triangle.
   *
   * @param  face Handle of the face.
   * @param  v1 Handle of the 1st vertex.
   * @param  v2 Handle of the 2nd vertex.
   * @param  v3 Handle of the 3rd vertex.
   * @see    #createFace
   */
  void setTriangleVertices ( int face, int v1, int v2, int v3 );

  /**
   * Sets vertex handles of the given face.
   *
   * @param  face Handle of the face.
   * @param  vert Handles of the vertices ({@link #NULL} breaks the sequence}).
   * @see    #createFace
   */
  void setFaceVertices ( int face, int[] vert );

  /**
   * Sets edge handles of the given face.
   *
   * @param  face Handle of the face.
   * @param  edge Array of edge handles (can be prematurely terminated by {@link #NULL}).
   * @see    #createFace
   */
  void setFaceEdges ( int face, int[] edge );

  /**
   * Sets face handles of the given solid.
   *
   * @param  ctx Handle of the context.
   * @param  solid Handle of the solid.
   * @param  face Array of face handles (can be prematurely terminated by {@link #NULL}).
   * @see    #createSolid
   */
  void setSolidFaces ( int ctx, int solid, int[] face );

  /**
   * Creates a new division (in the given context).
   *
   * @param  ctx Handle of the context.
   * @return Handle of the new division.
   */
  int createDivision ( int ctx );

  /**
   * Sets face handles of the given division.
   *
   * @param  ctx Handle of the context.
   * @param  div Handle of the division.
   * @param  up Array of predecessor-face handles (can be prematurely terminated by
   *            {@link #NULL}).
   * @param  down Array of successor-face handles (can be prematurely terminated by
   *              {@link #NULL}).
   * @see    #createDivision
   */
  void setDivisionFaces ( int ctx, int div, int[] up, int[] down );

  /**
   * Creates a new solid (in the given context).
   *
   * @param  ctx Handle of the context.
   * @return Handle of the new solid.
   */
  int createSolid ( int ctx );

  /**
   * Resets the given solid.
   *
   * @param  ctx Handle of the context.
   * @param  solid Handle of the solid.
   * @see    #createSolid
   * @see    #insertFaceIntoSolid
   */
  void resetSolid ( int ctx, int solid );

  /**
   * Inserts another face handle into the given solid.
   *
   * @param  ctx Handle of the context.
   * @param  solid Handle of the solid.
   * @param  face Handle of the new face.
   * @see    #resetSolid
   */
  void insertFaceIntoSolid ( int ctx, int solid, int face );

}
