package cz.cuni.jagrlib.iface;

import java.io.*;

import cz.cuni.jagrlib.Breakable;

/**
 * Interface for data persistence algorithms (general load/save filter).
 *
 * @version 0.26 $Rev: 462 $ $Date: 2007-05-24 10:40:53 +0200 (čt, 24 V 2007) $ $Author: pepca $
 * @since   0.13
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/DataFileFormat.java">DataFileFormat.java</a>
 * @see     BitStream
 */
public interface DataFileFormat extends Property, Breakable
{
  /**
   * Double-stream property (Boolean): file-format uses more than one bit-stream - e.g. for
   * header and raw data.
   *
   * @see    #loadFile(BitStream,Object)
   * @see    #saveFile(BitStream,Object)
   */
  String DOUBLE_STREAM = "DoubleStream";

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  int headerLength ();

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type ({@code 0.0} .. no way,
   *         {@code 0.5} .. maybe, {@code 0.9} .. almost sure, {@code 1.0} ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  double match ( byte[] header, String fileName );

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.jpg", "*.jpeg",
   *         "*.jfif"]) or {@code null}.
   * @see    #match
   */
  String[] fileNameMasks ();

  /**
   * Retrieves data from the given file (stream).
   * Can be used for "file opening" in virtual-mapped mode.
   *
   * @param  fileName File-name (stream-name) to read from (can be {@code null}).
   * @param  detail Further implementation-specific information ("r" or "rw" for mapped mode).
   * @return {@code true} if the operation was successful.
   * @see    #loadFile(BitStream,Object)
   * @see    #saveFile(String,String)
   * @see    #closeFile
   */
  boolean loadFile ( String fileName, String detail )
  throws IOException;

  /**
   * Retrieves data from the given {@link BitStream}. This metod is optional.
   * <p>The stream should be opened before - only the {@link BitStream#reOpen} call
   * can be used.
   *
   * @param  stream Input bit-stream (opened - call {@link BitStream#reOpen} first).
   * @param  support Implementation-specific data (e.g. second {@link BitStream} for
   *                 file-formats with {@link #DOUBLE_STREAM} property).
   * @return {@code true} if the operation was successful.
   * @see    #loadFile(String,String)
   * @see    #DOUBLE_STREAM
   */
  boolean loadFile ( BitStream stream, Object support )
  throws IOException;

  /**
   * Saves data to the given file (stream).
   *
   * @param  fileName File-name (stream-name) to write to (can be {@code null}).
   * @param  detail Further implementation-specific information.
   * @return {@code true} if the operation was successful.
   * @see    #saveFile(BitStream,Object)
   * @see    #loadFile(String,String)
   */
  boolean saveFile ( String fileName, String detail )
  throws IOException;

  /**
   * Saves data to the given {@link BitStream}. This metod is optional.
   * <p>The stream should be opened and empty - the {@code reOpen(true,null)} call
   * is not required.
   *
   * @param  stream Output bit-stream (opened and empty).
   * @param  support Implementation-specific data (e.g. second {@link BitStream} for
   *                 file-formats with {@code DOUBLE_STREAM} property).
   * @return {@code true} if the operation was successful.
   * @see    #saveFile(String,String)
   * @see    #DOUBLE_STREAM
   */
  boolean saveFile ( BitStream stream, Object support )
  throws IOException;

  /**
   * Closes the file (used in file-mapping mode).
   *
   * @see    #loadFile(String,String)
   */
  void closeFile ()
  throws IOException;

}
