package cz.cuni.jagrlib.iface;

/**
 * Discrete transform of raster 1D - 2D data.
 * Transforms segments/rectangles of integer/floating-point samples.
 *
 * @version 0.24 $Rev: 240 $ $Date: 2006-05-07 02:01:03 +0200 (ne, 07 V 2006) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/DiscreteTransform.java">DiscreteTransform.java</a>
 */
public interface DiscreteTransform extends Property, PersistentParamsBinary
{
  /**
   * Get extreme domain (source) ranges.
   *
   * @param  bounds Pre-allocated array to hold extreme ranges:
   *         <br><tt>bounds[0]</tt> .. minimum size in 1st dim,
   *         <br><tt>bounds[1]</tt> .. maximum size in 1st dim,
   *         <br><tt>bounds[2]</tt> .. minimum size in 2nd dim,
   *         <br><tt>bounds[3]</tt> .. maximum size in 2nd dim, etc.
   * @return <tt>true</tt> if arbitrary lengths are allowed (even non-powers-ow-two).
   */
  boolean getSizeBounds ( int[] bounds );

  /**
   * Forward 1D transform (integer form).
   * Can be used if both source and target data are in integer format.
   *
   * @param  sample Source array (containing sample data).
   * @param  start Starting index of the transformed segment.
   * @param  len Source segment length.
   * @param  coef Pre-allocated array to hold transformed data (coefficients).
   * @return Number of coefficients (<tt>0</tt> in case of error).
   * @see    #inverse1D(int[],int,int[],int)
   * @see    #transform1D(double[],int,int,double[])
   */
  int transform1D ( int[] sample, int start, int len, int[] coef );

  /**
   * Forward 1D transform (floating-point form).
   * Has to be used if source and/or target data are floating-point!
   *
   * @param  sample Source array (containing sample data).
   * @param  start Starting index of the transformed segment.
   * @param  len Source segment length.
   * @param  coef Pre-allocated array to hold transformed data (coefficients).
   * @return Number of coefficients (<tt>0</tt> in case of error).
   * @see    #inverse1D(double[],int,double[],int)
   * @see    #transform1D(int[],int,int,int[])
   */
  int transform1D ( double[] sample, int start, int len, double[] coef );

  /**
   * Inverse 1D transform (integer form).
   * Inverse mapping to {@link #transform1D(int[],int,int,int[])}.
   *
   * @param  coef Source array (holding coefficients).
   * @param  len Number of coefficients.
   * @param  sample Pre-allocated array to hold reconstructed sample data.
   * @param  start Starting index of the reconstructed sample data.
   * @return Sample data length (<tt>0</tt> in case of error).
   * @see    #transform1D(int[],int,int,int[])
   * @see    #inverse1D(double[],int,double[],int)
   */
  int inverse1D ( int[] coef, int len, int[] sample, int start );

  /**
   * Inverse 1D transform (floating-point form).
   * Inverse mapping to {@link #transform1D(double[],int,int,double[])}.
   *
   * @param  coef Source array (holding coefficients).
   * @param  len Number of coefficients.
   * @param  sample Pre-allocated array to hold reconstructed sample data.
   * @param  start Starting index of the reconstructed sample data.
   * @return Sample data length (<tt>0</tt> in case of error).
   * @see    #transform1D(double[],int,int,double[])
   * @see    #inverse1D(int[],int,int[],int)
   */
  int inverse1D ( double[] coef, int len, double[] sample, int start );

  /**
   * Forward 2D transform (integer form).
   * Can be used if both source and target data are in integer format.
   *
   * @param  sample Source array (containing sample data).
   * @param  start Starting index of the transformed data.
   * @param  width Source rectangle's width (faster index).
   * @param  height Source rectangle's height (slower index).
   * @param  stride Source matrix row length.
   * @param  coef Pre-allocated array to hold transformed data (coefficients).
   * @return Number of coefficients (<tt>0</tt> in case of error).
   * @see    #inverse2D(int[],int,int[],int,int,int)
   * @see    #transform2D(double[],int,int,int,int,double[])
   */
  int transform2D ( int[] sample, int start, int width, int height, int stride, int[] coef );

  /**
   * Forward 2D transform (floating-point form).
   * Has to be used if source and/or target data are floating-point!
   *
   * @param  sample Source array (containing sample data).
   * @param  start Starting index of the transformed data.
   * @param  width Source rectangle's width (faster index).
   * @param  height Source rectangle's height (slower index).
   * @param  stride Source matrix row length.
   * @param  coef Pre-allocated array to hold transformed data (coefficients).
   * @return Number of coefficients (<tt>0</tt> in case of error).
   * @see    #inverse2D(double[],int,double[],int,int,int)
   * @see    #transform2D(int[],int,int,int,int,int[])
   */
  int transform2D ( double[] sample, int start, int width, int height, int stride, double[] coef );

  /**
   * Inverse 2D transform (integer form).
   * Inverse mapping to {@link #transform2D(int[],int,int,int,int,int[])}.
   *
   * @param  coef Source array (holding coefficients).
   * @param  len Number of coefficients.
   * @param  sample Pre-allocated array to hold reconstructed sample data.
   * @param  start Starting index of the reconstructed rectangle.
   * @param  width Reconstructed rectangle's width (faster index).
   * @param  stride Reconstructed data matrix row length.
   * @return Sample data height (<tt>0</tt> in case of error).
   * @see    #transform2D(int[],int,int,int,int,int[])
   * @see    #inverse2D(double[],int,double[],int,int,int)
   */
  int inverse2D ( int[] coef, int len, int[] sample, int start, int width, int stride );

  /**
   * Inverse 2D transform (floating-point form).
   * Inverse mapping to {@link #transform2D(double[],int,int,int,int,double[])}.
   *
   * @param  coef Source array (holding coefficients).
   * @param  len Number of coefficients.
   * @param  sample Pre-allocated array to hold reconstructed sample data.
   * @param  start Starting index of the reconstructed rectangle.
   * @param  width Reconstructed rectangle's width (faster index).
   * @param  stride Reconstructed data matrix row length.
   * @return Sample data height (<tt>0</tt> in case of error).
   * @see    #transform2D(double[],int,int,int,int,double[])
   * @see    #inverse2D(int[],int,int[],int,int,int)
   */
  int inverse2D ( double[] coef, int len, double[] sample, int start, int width, int stride );

}
