package cz.cuni.jagrlib.iface;

import java.io.IOException;

/**
 * Abstract entropy codec interface.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.04
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/EntropyCodec.java">EntropyCodec.java</a>
 */
public interface EntropyCodec extends Property
{
  /**
   * Codec initialization.
   *
   * @param  output Open for output?
   * @param  ctx Initial context.
   * @throws IOException
   * @see    BitStream
   * @see    #setContext
   */
  void open ( boolean output, int ctx )
  throws IOException;

  /**
   * Sets the given context as current.
   *
   * @param  ctx Id of the new context.
   * @return The old context's id.
   * @throws IOException
   */
  int setContext ( int ctx )
  throws IOException;

  /**
   * Retrieves maximal symbol value for the actual context.
   *
   * @return Maximal symbol value.
   */
  int getMaxSymbol ();

  /**
   * Sets maximal symbol value for the actual context
   * and all subsequently created ones.
   * Could be called before {@link #open}!
   *
   * @param max New maximal symbol value.
   */
  void setMaxSymbol ( int max );

  /**
   * Codec flush.
   *
   * @throws IOException
   * @see    #open
   * @see    #close
   */
  void flush ()
  throws IOException;

  /**
   * Codec close.
   *
   * @throws IOException
   * @see    #open
   * @see    #flush
   */
  void close ()
  throws IOException;

  /**
   * Actual codec external position in symbols (total amount in all contexts).
   *
   * @return Actual codec external position in symbols.
   * @throws IOException
   * @see    #compressed
   */
  long position ()
  throws IOException;

  /**
   * Actual codec compressed position in bits (total amount in all contexts).
   *
   * @return Actual codec compressed position in bits.
   * @throws IOException
   * @see    #position
   */
  long compressed ()
  throws IOException;

  /**
   * Are they any input symbols (bits) available?
   *
   * @return Any symbols (bits) are ready to read.
   * @throws IOException
   * @see    #position
   * @see    #compressed
   */
  boolean available ()
  throws IOException;

  /**
   * Encodes the given symbol (in the current context).
   *
   * @param  symbol Symbol to encode.
   * @throws IOException
   * @see    #putBits
   * @see    #get
   */
  void put ( int symbol )
  throws IOException;

  /* *
   * Encodes one subinterval from the given {@link WheelOfFortune}.
   * Doesn't use contexts.
   *
   * @param  wheel Set of intervals (wheel of fortune).
   * @param  s Index of an interval to encode.
   */
  public void put ( WheelOfFortune wheel, int s )
  throws IOException;

  /**
   * Writes the given "raw" bits (by-passes the entropy encoder).
   *
   * @param  bits Bits to be written ("MSB first" order).
   * @param  length Number of bits to write.
   * @throws IOException
   * @see    #put
   * @see    #getBits
   */
  void putBits ( long bits, int length )
  throws IOException;

  /**
   * Decodes one symbol (using the current context).
   *
   * @return Read symbol ({@code -1} if no symbol is available).
   * @throws IOException
   * @see    #getBits
   * @see    #put
   * @see    #available
   */
  int get ()
  throws IOException;

  /* *
   * Decodes one subinterval from the given {@link WheelOfFortune}.
   * Doesn't use contexts.
   *
   * @param  wheel Set of intervals (wheel of fortune).
   * @return Index of the relevant interval.
   */
  public int get ( WheelOfFortune wheel )
  throws IOException;

  /**
   * Reads the given amount of "raw" bits (by-passes the entropy decoder).
   *
   * @param  length Number of bits to read.
   * @return Read bits (or {@code -1} if no bits are available).
   * @throws IOException
   * @see    #get
   * @see    #putBits
   * @see    #available
   */
  long getBits ( int length )
  throws IOException;

}
