package cz.cuni.jagrlib.iface;

/**
 * Abstract histogram interface - {#link WheelOfFortune} with incremental frequency
 * update capability and decimation mechanism.
 *
 * @author  PE, Jan Kadlec <rrrola@gmail.com>
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.04
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/EntropyHistogram.java">EntropyHistogram.java</a>
 */
public interface EntropyHistogram extends WheelOfFortune
{
  /**
   * Histogram (re-)initialization.
   * Destroys all previously collected statistics.
   *
   * @param  _defaultSymbols Default number of symbols for new contexts.
   * @param  _defaultSumLimit Default frequency sum limit for new contexts ({@code 0} for no bounds).
   * @param  _defaultNonZero Keeps frequencies non-zero at all times.
   *         Default frequencies are the lowest possible.
   * @param  _context The initial context.
   * @see    #setDefaultSymbols
   * @see    #setDefaultSumLimit
   * @see    #setDefaultNonZero
   * @see    #decimate
   */
  void init ( int _defaultSymbols, long _defaultSumLimit, boolean _defaultNonZero, int _context );

  /**
   * Sets the new default number of symbols for new contexts.
   *
   * @param  _defaultSymbols New default number of symbols.
   * @see    #init
   * @see    #setSymbols
   * @see    #getSymbols
   */
  void setDefaultSymbols ( int _defaultSymbols );

  /**
   * Sets the new default frequency sum limit for new contexts.
   *
   * @param  _defaultSumLimit New default frequency sum limit ({@code 0} for no bounds).
   * @see    #init
   * @see    #setSumLimit
   * @see    #getSumLimit
   * @see    #total
   */
  void setDefaultSumLimit ( long _defaultSumLimit );

  /**
   * Changes the frequency sum limit for the current context.
   *
   * @param  _sumLimit New frequency sum limit ({@code 0} for no bounds).
   * @see    #init
   * @see    #setDefaultSumLimit
   * @see    #getSumLimit
   * @see    #decimate
   * @see    #total
   */
  void setSumLimit ( long _sumLimit );

  /**
   * Returns the frequency sum limit for the current context.
   *
   * @return Frequency sum limit ({@code 0} for no bounds).
   * @see    #init
   * @see    #setDefaultSumLimit
   * @see    #setSumLimit
   * @see    #total
   */
  long getSumLimit ();

  /**
   * Set non-zero attribute: keep frequencies non-zero at all times?
   * Value for new contexts.
   *
   * @param  _defaultNonZero Non-zero attribute for new contexts.
   * @see    #init
   * @see    #setNonZero
   * @see    #getNonZero
   */
  void setDefaultNonZero ( boolean _defaultNonZero );

  /**
   * Set non-zero attribute for the current context: keep frequencies non-zero at all times?
   *
   * @param  _nonZero Non-zero attribute for the current context.
   * @see    #init
   * @see    #setDefaultNonZero
   * @see    #getNonZero
   * @see    #decimate
   */
  void setNonZero ( boolean _nonZero );

  /**
   * Read non-zero attribute for the current context
   *
   * @return Non-zero attribute of the current context.
   * @see    #init
   * @see    #setDefaultNonZero
   * @see    #setNonZero
   */
  boolean getNonZero ();

  /**
   * Sets the given context as current.
   * For unknown contexts initializes parameters to default values.
   *
   * @param  _newContext Id of the new context.
   * @return Id of the old context.
   * @see    #init
   * @see    #setDefaultSymbols
   * @see    #setDefaultSumLimit
   * @see    #getContext
   */
  int setContext ( int _newContext );

  /**
   * Returns the current context.
   *
   * @return Id of the current context.
   * @see    #setContext
   */
  int getContext ();

  /**
   * Divides each frequency in the current context by 2.
   * Called automatically when the total frequency sum exceeds the limit.
   *
   * @see    #init
   * @see    #setDefaultSumLimit
   * @see    #setSumLimit
   * @see    #getSumLimit
   * @see    #setDefaultNonZero
   * @see    #setNonZero
   * @see    #getNonZero
   * @see    #total
   */
  void decimate ();

  /**
   * Increments frequencies of all symbols in the current context.
   *
   * @param  _freq Array of frequency increments.
   * @see    #inc(int,long)
   * @see    #inc(int)
   * @see    #decimate
   */
  void inc ( long[] _freq );

  /**
   * Increments a single symbol's frequency.
   *
   * @param  _symbol Symbol code.
   * @param  _freq Frequency increment of the symbol.
   * @see    #inc(long[])
   * @see    #inc(int)
   * @see    #decimate
   */
  void inc ( int _symbol, long _freq );

  /**
   * Increments a single symbol's frequency by {@code 1}.
   *
   * @param  _symbol Symbol code.
   * @see    #inc(int,long)
   * @see    #inc(long[])
   * @see    #decimate
   */
  void inc ( int _symbol );

}
