package cz.cuni.jagrlib.iface;

/**
 * Interface for 3D scene editing using <b>Euler operators</b>:
 * <p>Euler equation: <b>V - E + F = 2 * (S - G) + L</b>, where:
 * </p>
 * <ul>
 *  <li><b>V</b> = # of vertices</li>
 *  <li><b>E</b> = # of edges</li>
 *  <li><b>F</b> = # of faces</li>
 *  <li><b>S</b> = # of shells (solids)</li>
 *  <li><b>G</b> = # of holes through solid (Genus)</li>
 *  <li><b>L</b> = # of inner loops (rings) = holes in faces</li>
 * </ul>
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.10
 * @see     Brep
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/EulerOperators.java">EulerOperators.java</a>
 */
public interface EulerOperators extends Property
{
  /**
   * Make vertex, face, solid.
   * Creates a new (empty) shell containing a new vertex and face.
   *
   * @param  vfs Pre-allocated array to hold handles of the newly created vertex,
   *             face and solid.
   * @see    #Kvfs
   */
  void Mvfs ( int[] vfs );

  /**
   * Kill vertex, face, solid.
   * Eliminates the rest of the given shell.
   *
   * @param  v Vertex handle to be deleted.
   * @param  f Face handle to be deleted.
   * @param  s Solid handle to be deleted.
   * @see    #Mvfs
   */
  void Kvfs ( int v, int f, int s );

  /**
   * Make vertex, edge.
   * Add a new vetrex through a new edge | vertex split.
   *
   * @param  ve Pre-allocated array to hold handles of the newly created vertex
   *            and edge.
   * @param  v1 Handle of the 1st (already existing) vertex.
   * @param  fl Handle of the left face for the new edge.
   * @param  fr Handle of the right face for the new edge (should be the same as fl).
   * @see    #Kve
   */
  void Mve ( int[] ve, int v1, int fl, int fr );

  /**
   * Kill vertex, edge.
   * Edge collapse.
   *
   * @param  v Vertex handle to be deleted.
   * @param  e Edge handle to be deleted.
   * @see    #Mve
   */
  void Kve ( int v, int e );

  /**
   * Make edge, face.
   * Face split.
   *
   * @param  ef Pre-allocated array to hold handles of the newly created edge
   *            and face.
   * @param  v1 Handle of the 1st (already existing) vertex.
   * @param  v2 Handle of the 2nd (already existing) vertex.
   * @see    #Kef
   */
  void Mef ( int[] ef, int v1, int v2 );

  /**
   * Kill edge, face.
   * Face join.
   *
   * @param  e Edge handle to be deleted.
   * @param  f Face handle to be deleted.
   * @see    #Mef
   */
  void Kef ( int e, int f );

  /**
   * Make loop, kill edge.
   * Creates a new inner loop instead of discarded virtual edge (loop split).
   *
   * @param  e Handle of the virtual edge to be deleted.
   * @see    #KlMe
   */
  void MlKe ( int e );

  /**
   * Kill loop, make edge.
   * Merges two loops (inner and outer one) using a new virtual edge.
   *
   * @param  e Pre-allocated array to hold handle of the newly created virtual edge.
   * @param  v1 Handle of the 1st (already existing) vertex.
   * @param  v2 Handle of the 2nd (already existing) vertex.
   * @param  f Face handle in which the virtual edge will be created.
   * @see    #MlKe
   */
  void KlMe ( int[] e, int v1, int v2, int f );

  /**
   * Make loop, hole, kill face.
   * Punches a hole through the given (stil unconcerned) face, on the other side one existing
   * face will be deleted. Doesn't affect edges and vertices at all..
   *
   * @param  fp Handle of the face to be punched (a new loop will be created in it).
   * @param  fd Handle of the face to be deleted (its perimeter will turn into a new loop
   *            on the other side).
   * @see    #KlhMf
   */
  void MlhKf ( int fp, int fd );

  /**
   * Kill loop, hole, make face.
   * Closes one existing hole using a new stopper-face.
   * Doesn't affect edges and vertices at all..
   *
   * @param  f Pre-allocated array to hold handle of the newly created stopper-face.
   * @param  fh Handle of the face containing hole (loop) to be closed.
   * @param  l Ordinal number of the loop within the face.
   * @see    #MlhKf
   */
  void KlhMf ( int[] f, int fh, int l );

}
