package cz.cuni.jagrlib.iface;

/**
 * Data storage for "features" - elements used in feature/based warping/morphing.
 * Features can be 1D (points), 2D (arrows), 3D (trangles) and can be placed in 2D space
 * (plane) od 3D (3D warping).
 *
 * @version 0.26 $Rev: 407 $ $Date: 2007-03-25 15:56:09 +0200 (ne, 25 III 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/FeatureStore.java">FeatureStore.java</a>
 */
public interface FeatureStore extends Property
{
  /** Feature type: point (1D feature). */
  final int FEATURE_POINT    = 0;

  /** Feature type: arrow (2D feature). */
  final int FEATURE_ARROW    = 1;

  /** Feature type: triangle (3D feature). */
  final int FEATURE_TRIANGLE = 2;

  /** Void (null) feature handle. */
  final int NULL = -1;

  /**
   * Inserts a new feature of the given type.
   * Every feature exists in defined time interval (<tt>timeFrom - timeTo</tt>).
   *
   * @param  type Feature type ({@link #FEATURE_POINT}, {@link #FEATURE_ARROW} or {@link #FEATURE_TRIANGLE}).
   * @return Feature handle (id). If succeeded, return value should be &gt;= 0.
   * @see    #removeFeature
   * @see    #setTimes
   * @see    #setCoords
   * @see    #setParam
   */
  int insertFeature ( int timeFrom, int timeTo, int type );

  /**
   * Removes the given feature from the database.
   *
   * @param  id Handle of an existing feature.
   */
  void removeFeature ( int id );

  /**
   * Finds the next feature in the database.
   * Iteration should start with id = {@link #NULL}. This function also returns {@link #NULL}
   * if there is no other feature..
   *
   * @param  id Handle of previous feature (or {@link #NULL} if we are restaring an iteration).
   * @param  time Time of interest.
   * @return Handle of a next feature in database (or {@link #NULL} if the previous was the last one).
   */
  int nextFeature ( int id, double time );

  /**
   * Reads feature-type ({@link #FEATURE_POINT}, {@link #FEATURE_ARROW} or {@link #FEATURE_TRIANGLE}).
   */
  int featureType ( int id );

  /**
   * Sets feature life-time.
   *
   * @param  id Handle of a feature.
   * @param  timeFrom Birth-time of the feature.
   * @param  timeTo Death-time of the feature.
   */
  void setTimes ( int id, int timeFrom, int timeTo );

  /** Returns Birth-time of the given feature. */
  int featureStart ( int id );

  /** Returns Time of death of the given feature. */
  int featureEnd ( int id );

  /**
   * Sets feature coordinates of the given feature in the given keyframe-time.
   *
   * @param  id Handle of a feature.
   * @param  time Keyframe-time.
   * @param  coords Coordinate array (e.g. {@code double[4]} for 2D arrow feature:
   *         {@code [ x0, y0, x1, y1 ]}).
   */
  void setCoords ( int id, int time, double[] coords );

  /**
   * Reads feature coordinates of the given feature in the given time.
   *
   * @param  id Handle of a feature.
   * @param  time Time of interest. Needs not be keyframe-time.
   * @param  coords Pre-allocated coordinate array to be filled
   *         ({@code double[4]} for 2D arrow feature: {@code [ x0, y0, x1, y1 ]}). Can be {@code null}.
   * @return Coordinate array or {@code null}.
   */
  double[] getCoords ( int id, double time, double[] coords );

  /**
   * Sets feature's general parameter in the given keyframe-time.
   *
   * @param  id Handle of a feature.
   * @param  time Keyframe-time ({@link #NULL} for time-independent parameter).
   * @param  name Parameter identifier.
   * @param  value New parameter value.
   */
  void setParam ( int id, int time, String name, Object value );

  /**
   * Reads the given integer parametr in the given time.
   *
   * @param  id Handle of a feature.
   * @param  time Time of interest. Needs not be keyframe-time.
   * @param  name Parameter identifier.
   * @return Integer value (interpolated in time if applicable).
   */
  int getParamInt ( int id, double time, String name );

  /**
   * Reads the given floating-point parametr in the given time.
   *
   * @param  id Handle of a feature.
   * @param  time Time of interest. Needs not be keyframe-time.
   * @param  name Parameter identifier.
   * @return Floating-point value (interpolated in time if applicable).
   */
  double getParamDouble ( int id, double time, String name );

  /**
   * Reads the given general parametr in the given time.
   *
   * @param  id Handle of a feature.
   * @param  time Time of interest. Needs not be keyframe-time.
   * @param  name Parameter identifier.
   * @return Parameter value (interpolated in time if applicable).
   */
  Object getParam ( int id, double time, String name );
}
