package cz.cuni.jagrlib.iface;

/**
 * General interface for geometry searching in 2D and 3D (for various types of search-trees).
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.10
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/GeometrySearch.java">GeometrySearch.java</a>
 * @see     Brep
 */
public interface GeometrySearch extends Property
{
  /**
   * Builds the whole data structure (directory) from scratch.
   *
   * @param  entity Type of stored entities ({@link Brep#VERTEX VERTEX}, {@link Brep#EDGE EDGE} or
   *                {@link Brep#FACE FACE} from {@link Brep}).
   * @param  ctx Handle to the used {@link Brep}-context.
   * @return {@code true} if this service is implemented.
   * @see    #resetDirectory
   * @see    #optimizeDirectory
   * @see    Brep#VERTEX
   * @see    Brep#EDGE
   * @see    Brep#SOLID
   */
  boolean buildDirectory ( int entity, int ctx );

  /**
   * (Re-)initializes the data structure (directory).
   *
   * @param  entity Type of stored entities ({@link Brep#VERTEX VERTEX}, {@link Brep#EDGE EDGE} or
   *                {@link Brep#FACE FACE} from {@link Brep}).
   * @param  ctx Handle to the used {@link Brep}-context.
   * @return {@code true} if this service is implemented.
   * @see    #buildDirectory
   * @see    #optimizeDirectory
   * @see    Brep#VERTEX
   * @see    Brep#EDGE
   * @see    Brep#SOLID
   */
  boolean resetDirectory ( int entity, int ctx );

  /**
   * Inserts another object into the directory. Should be called after
   * {@link #resetDirectory}.
   *
   * @param  handle Handle to an inserted object (its <b>entity type</b> and <b>context</b>
   *                were defined in advance).
   * @see    #resetDirectory
   */
  void insertObject ( int handle );

  /**
   * Removes the given object from the directory.
   *
   * @param  handle Handle to a removed object (its <b>entity type</b> and <b>context</b>
   *                were defined in advance).
   * @return {@code true} if this service is implemented and if the object was
   *         successfully removed.
   * @see    #resetDirectory
   */
  boolean removeObject ( int handle );

  /**
   * Optimizes the data structure (directory). Can be used for <b>finishing</b> of the
   * update phase or for explicit data structure <b>optimization</b>. Non-mandatory.
   * (but application has to call it at the end of update phase).
   *
   * @return {@code true} if this service is implemented.
   * @see    #buildDirectory
   * @see    #resetDirectory
   */
  boolean optimizeDirectory ();

  /**
   * Localizes the given point - finds the object which contains it (first of such objects).
   *
   * @param  coord Coordinate vector of the localized point.
   * @return Handle to the result object or {@link Brep#NULL NULL} if not found (or the service is
   *         not implemented).
   * @see    #localize(double[])
   */
  int localize ( int[] coord );

  /**
   * Localizes the given point - finds the object which contains it (first of such objects).
   *
   * @param  coord Coordinate vector of the localized point.
   * @return Handle to the result object or {@link Brep#NULL NULL} if not found (or the service is
   *         not implemented).
   * @see    #localize(int[])
   */
  int localize ( double[] coord );

  /**
   * Initializes a new <b>interval query</b> for integer coordinates.
   *
   * @param  min Coordinate vector of minimum-vertex.
   * @param  max Coordinate vector of maximum-vertex. Non-mandatory ?
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #intervalQuery(double[],double[],BrepIterator)
   * @see    #getProgress
   */
  BrepIterator intervalQuery ( int[] min, int[] max, BrepIterator pre );

  /**
   * Initializes a new <b>interval query</b> for floating-point coordinates.
   *
   * @param  min Coordinate vector of minimum-vertex.
   * @param  max Coordinate vector of maximum-vertex. Non-mandatory ?
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #intervalQuery(int[],int[],BrepIterator)
   * @see    #getProgress
   */
  BrepIterator intervalQuery ( double[] min, double[] max, BrepIterator pre );

  /**
   * Initializes a new <b>ray-casting pass</b>.
   *
   * @param  start Starting point of the ray.
   * @param  direction Direction vector of the ray.
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #getProgress
   */
  BrepIterator rayPass ( double[] start, double[] direction, BrepIterator pre );

  /**
   * Initializes a new <b>center pass</b>. Objects are enumerated in order defined by increasing
   * distance from the given center point.
   *
   * @param  center Center of a pass.
   * @param  radius Maximum radius to be searched.
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #getProgress
   */
  BrepIterator centerPass ( double[] center, double radius, BrepIterator pre );

  /**
   * Initializes a new <b>sweep-plane pass</b>. Objects are enumerated in order defined by
   * the given sweep-plane.
   *
   * @param  plane (Oriented) normal vector of the sweep-plane.
   * @param  origin Starting point of a pass.
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #getProgress
   */
  BrepIterator sweepPass ( double[] plane, double[] origin, BrepIterator pre );

  /**
   * Returns end value (coordinate) of already finished pass-segment.
   * Semantics depends on pass type: safely processed distance from pass/ray origin, ..
   *
   * @param  iter Pass data.
   * @return Safely processed distance from pass origin.
   */
  double finishedSegment ( BrepIterator iter );

  /**
   * Returns progress parameters of the given pass.
   *
   * @param  iter Pass data.
   * @param  count Array to hold two counters: {@code count[0]} .. number of objects
   *               yet passed, {@code count[1]} .. total number of objects
   *               relevant to this pass (non-mandatory). Can be {@code null}.
   * @return Actual pass-distance (or similar metric).
   */
  double getProgress ( BrepIterator iter, int[] count );

  /**
   * Returns (and resets) counters (for statistics, profiling):
   * <br />counters[0] .. number of point-to-point distance calculations.
   * <br />counters[1] .. number of point-to-box distance calculations.
   * <br />counters[2] .. number of log(N) heap operations.
   *
   * @param  counters Array to hold result. If {@code null} functions returns required length.
   * @return Required array length.
   */
  int getStatistics ( long[] counters );

}
