package cz.cuni.jagrlib.iface;

import cz.cuni.jagrlib.*;

/**
 * Viewer for {@link RasterGraphics}, {@link VectorGraphics} etc. Typical implementation
 * uses some GUI object to [pre]view graphic results.
 *
 * @version 0.25 $Rev: 373 $ $Date: 2006-12-17 20:26:08 +0100 (ne, 17 XII 2006) $ $Author: pepca $
 * @since   0.13
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/GraphicsViewer.java">GraphicsViewer.java</a>
 * @see     VectorGraphics
 * @see     RasterGraphics
 */
public interface GraphicsViewer extends Trigger, Property
{
  /** Window-title property (String): title for frame window holding the graphic output. */
  String WINDOW_TITLE = "WindowTitle";

  /** Window-title property (String): title for frame window holding the graphic output. */
  String BLACK_BACKGROUND = "BlackBackground";

  /**
   * Start (potentially) infinite <b>repaint loop</b>. It can be terminated either by a
   * {@link #stopRepaintLoop} call or by external user interaction (user break, window
   * close, ..).
   *
   * @return <tt>true</tt> if the loop was terminated by a {@link #stopRepaintLoop} call.
   * @see    #stopRepaintLoop
   */
  boolean repaintLoop ();

  /**
   * Stops the running <b>repaint loop</b>.
   *
   * @return <tt>true</tt> if the loop was successfully stopped.
   * @see    #repaintLoop
   */
  boolean stopRepaintLoop ();

  /**
   * Destroys the preview window.
   * Stops repaint-loop eventually.
   */
  void destroy ();

  /**
   * Causes the instant <b>repainting</b> of the whole graphic output (regardless of the actual
   * <b>invalid region</b>). {@link Trigger#fire} call has the same effect..
   *
   * @see    #invalidate
   * @see    #repaint
   * @see    Trigger#fire
   */
  void repaintAll ();

  /**
   * Invalidates the given rectangular region of the graphics output. Invalid regions will be
   * repainted during the next {@link #repaint} call...
   *
   * @param  xMin X coordinate of the upper left corner.
   * @param  xMax X coordinate of the lower right corner (outside pixel).
   * @param  yMin Y coordinate of the upper left corner.
   * @param  yMax Y coordinate of the upper left corner (outside pixel).
   * @see    #repaint
   */
  void invalidate ( int xMin, int xMax, int yMin, int yMax );

  /**
   * Causes the instant <b>repainting</b> of actual <b>invalid region</b> of the graphic output.
   *
   * @see    #repaintAll
   * @see    #invalidate
   */
  void repaint ();

  /**
   * Adds a new <b>semaphore</b> - object which wants to be signaled after the viewer window
   * is closed (by user interaction etc.).
   *
   * @param  s A new semaphore instance.
   * @see    #removeSemaphore
   */
  void addSemaphore ( Semaphore s );

  /**
   * Removes one of associated <b>semaphores</b>.
   *
   * @param  s Semaphore to be removed.
   * @see    #addSemaphore
   */
  void removeSemaphore ( Semaphore s );

  /**
   * Adds input listener object.
   */
  void addInputListener ( InputListener list );

  /**
   * Removes (previously added) input listener object.
   */
  void removeInputListener ( InputListener list );

  /** Removes all HUD primitives. */
  void hudClear ();

  /** Returns id of any free HUD slot. */
  int hudGetFreeId ();

  /**
   * Replaces the given HUD primitive slot with line from {@code [x1,y1]} to {@code [x2,y2]}.
   */
  void hudSetLine ( int id, int x1, int y1, int x2, int y2 );

  /** Removes existing HUD primitive. */
  void hudRemove ( int id );

  /**
   * Preview thread for background repainting of attached GraphicsViewer instance.
   *
   * @version 0.25 $Rev: 373 $ $Date: 2006-12-17 20:26:08 +0100 (ne, 17 XII 2006) $ $Author: pepca $
   * @since   0.25
   * @see     <a href="../../../../../cz/cuni/jagrlib/iface/GraphicsViewer.java">GraphicsViewer.java</a>
   */
  static class PreviewThread extends Thread
  {
    /** Attached viewer instance. */
    final private GraphicsViewer pr;

    /** Initializing constructor - sets minimum thread priority as default. */
    public PreviewThread ( GraphicsViewer preview )
    {
      super( "JaGrLib - GraphicsViewer preview" );

      pr = preview;
      try
      {
        setPriority( Thread.MIN_PRIORITY );
      }
      catch ( Exception e )
      {
        LogFile.exception( "Cannot set priority of preview thread! " + e.getMessage(), e );
      }
    }

    /** Preview rendering loop. */
    @Override
    public void run ()
    {
      pr.repaintLoop();
    }

    /** Safe thread start. */
    public void safeStart ()
    {
      try
      {
        start();
      }
      catch ( IllegalThreadStateException e )
      {
        LogFile.exception( "Cannot start preview thread! " + e.getMessage(), e );
      }
    }
  }

}
