package cz.cuni.jagrlib.iface;

/**
 * Abstract image function (R<sup>2</sup> -&gt; R<sup>N</sup> mapping).
 *
 * @version 0.26 $Rev: 434 $ $Date: 2007-05-02 03:57:09 +0200 (st, 02 V 2007) $ $Author: pepca $
 * @since   0.11
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/ImageFunction.java">ImageFunction.java</a>
 * @see     ImageSynthesizer
 */
public interface ImageFunction extends TimeDependent
{
  /** Property name: background-color ({@code double[3]}). */
  String BACKGROUND_COLOR = "BackgroundColor";

  /**
   * Sets bounds (scaling factors) for the image function (e.g. projection-plane mapping).
   *
   * @param  Xmin Minimum X coordinate.
   * @param  Xmax Maximum X coordinate.
   * @param  Ymin Minimum Y coordinate.
   * @param  Ymax Maximum Y coordinate.
   * @see    #getBounds
   */
  void setBounds ( double Xmin, double Xmax, double Ymin, double Ymax );

  /**
   * Returns actual bounds for {@code [ x, y ]} (projection-plane coordinates).
   *
   * @param  pre Pre-allocated (non-mandatory) {@code double[4]} array to hold results:
   *             {@code pre[0]} .. X<sub>min</sub>, {@code pre[1]} .. X<sub>max</sub>,
   *             {@code pre[2]} .. Y<sub>min</sub>, {@code pre[3]} .. Y<sub>max</sub>.
   * @return Array with all 4 return values.
   * @see    #setBounds
   * @see    TimeDependent#getTimeInterval
   */
  double[] getBounds ( double[] pre );

  /**
   * Computes one image sample.
   *
   * @param  x X coordinate of the sample.
   * @param  y Y coordinate of the sample.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (sample signature) used for adaptive subsampling.
   * @see    #getSample(double,double,int,int,double[])
   * @see    #getBounds
   */
  long getSample ( double x, double y, double[] color );

  /**
   * Computes one image sample. Passes parameters for hidden sampling.
   *
   * @param  x X coordinate of the sample.
   * @param  y Y coordinate of the sample.
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (sample signature) used for adaptive subsampling.
   * @see    #getSample(double,double,double[])
   * @see    #getBounds
   */
  long getSample ( double x, double y, int order, int total, double[] color );

}
