package cz.cuni.jagrlib.iface;

import java.util.List;

/**
 * Interface for <b>Quad-tree</b> data structure.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.04
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/QuadTree.java">QuadTree.java</a>
 */
public interface QuadTree extends Property
{
  /**
   * Quad-tree node.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.04
   * @see     <a href="../../../../../cz/cuni/jagrlib/iface/QuadTree.java">QuadTree.java</a>
   */
  class Node implements Cloneable
  {
    /**
     * Node's data ({@code Boolean}, {@code int[]}, {@code double[]}, etc.).
     * <br>{@code null} can have special meaning - "outside" marker for example..
     */
    public Object data;

    /**
     * Successors of the node. {@code null} for leaf node.
     */
    public Node[] succ;

  }

  /**
   * Initializes the Quad-tree. Starts the <b>build-phase</b> of a Quad-tree.
   *
   * @param  _width The region's width.
   * @param  _height The region's height.
   */
  void init ( int _width, int _height );

  /**
   * The Quad-tree region's width.
   */
  int getWidth ();

  /**
   * The Quad-tree region's height.
   */
  int getHeight ();

  /**
   * The current region is non-uniform (used in the <b>build-phase</b>).
   */
  void split ();

  /**
   * The current region is uniform (used in the <b>build-phase</b>).
   *
   * @param  value The value associated with the current region.
   */
  void uniform ( Object value );

  /**
   * Finishes the Quad-tree construction (<b>build-phase</b>).
   * Checks and joins uniform regions.
   */
  void finish ();

  /**
   * Returns the Quad-tree root node.
   *
   * @return The Quad-tree node or {@code null}.
   */
  Node root ();

  /**
   * Returns lineraized Quad-tree data
   * ({@code null} is used as a "descend" symbol).
   *
   * @return The Leaf-node data ordered by a Quad-tree pass (depth-first pass).
   *         {@code null} plays a role of "descend" symbol.
   */
  List< Object > passOrder ();

  /**
   * Quad-tree inspector object.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.04
   * @see     <a href="../../../../../cz/cuni/jagrlib/iface/QuadTree.java">QuadTree.java</a>
   */
  public interface Inspector
  {
    /**
     * (Re-)initialization of a Quad-tree pass.
     */
    public void init ();

    /**
     * Is the current node inner?
     *
     * @return {@code true} if the current node is inner (non-leaf node).
     */
    public boolean isInner ();

    /**
     * Retrieves the current node's data.
     *
     * @return The current node's data (can be {@code null}).
     */
    public Object getData ();

    /**
     * Assigns the current node's data.
     *
     * @param  data The new data for the current node (can be {@code null}).
     * @return The old node's data.
     */
    public Object setData ( Object data );

    /**
     * Returns the current {@link Node}.
     * @return The current node.
     */
    public Node getNode ();

    /** Advances in the same level (horizontally). */
    public void next ();

    /** Descends to the next deepest level of the tree (does nothing for leaf nodes). */
    public void down ();

    /** Ascends to the previous level of the tree (does nothing for the root). */
    public void up ();

    /** Discards all descendants of the current node. */
    public void makeLeafNode ();
  }

  /**
   * The Quad-tree's inspector.
   *
   * @return The <b>inspector</b> for this Quad-tree instance.
   */
  Inspector inspector ();

}
