package cz.cuni.jagrlib.iface;

import java.util.BitSet;

/**
 * Basis functions representing radiosity distribution over a set of {@link Brep} elements.
 *
 * Elements are represented by integer handles (e.g. faces residing in external {@link Brep} storage).
 * Outgoing radiosity distribution on each element is represented by a set of coefficients, neighbour
 * elements might share some coefficients.
 *
 * There are two major objectives defined in this interface:
 * <ol>
 *  <li>Form-factor computation (between any two coefficients in the scene)</li>
 *  <li>Radiosity distribution (approximation based on basis functions and actual data = coefficients)</li>
 * </ol>
 *
 * @author  PE, Tomas "Source" Svab
 * @version 0.26 $Rev: 497 $ $Date: 2007-08-17 01:01:10 +0200 (pá, 17 VIII 2007) $ $Author: pepca $
 * @since   0.26
 */
public interface RadiosityBasisFunctions
{
  //==========================================================================
  //  Radiosity distribution:

  /**
   * Order of the approximation (for information purposes only).
   * Number of coefficients defining radiosity distribution on single element.
   */
  int getOrder ();

  /**
   * Returns coefficient indexes defining the given element.
   *
   * @param  elem Element handle.
   * @param  result Pre-allocated integer array to hold result (can be {@code null}).
   * @return Size of result array.
   */
  int coefficientsFromElement ( int elem, int[] result );

  /**
   * Returns elements which are affected by the given single coefficient.
   *
   * @param  coeff Coefficient index.
   * @param  result Pre-allocated integer array to hold result (can be {@code null}).
   * @return Size of result array.
   */
  int elementsFromCoefficient ( int coeff, int[] result );

  /**
   * Returns set of coefficients defining the given set of elements.
   *
   * @param  elements Set of elements.
   * @param  result Non-mandatory object to be filled.
   * @return Set of coefficient indices.
   */
  BitSet coefficientsFromElements ( BitSet elements, BitSet result );

  /**
   * Returns set of elements which are affected by the given set of coefficients.
   *
   * @param  coefficients Set of coefficients.
   * @param  result Non-mandatory object to be filled.
   * @return Set of element handles.
   */
  BitSet elementsFromCoefficients ( BitSet coefficients, BitSet result );

  /**
   * Computes distribution coefficients for the single point of the element.
   *
   * @param  elem Element handle.
   * @param  coords World-space (?) coordinates of the point lying on the given element.
   * @param  result Non-mandatory pre-allocated array to hold result.
   * @return Distribution multiplicators (coefficient indices match the result
   *         of {@link #coefficientsFromElement}).
   * @todo   Add some other form of distribution ({@link FaceRender}?).
   */
  double[] distribute ( int elem, double[] coords, double[] result );

  //==========================================================================
  //  Form-factors:

  /** Total number of elements in the scene. */
  int totalElements ();

  /** Total number of coefficients. */
  int totalCoefficients ();

  /**
   * Computes "shooting" form-factors (single source coefficient is fixed).
   *
   * @param  coeff Source-coefficient index.
   * @param  ind Array to hold affected (lit) coefficient indices (can be {@code null}).
   * @param  ff Array to hold respective form-factors (can be {@code null}).
   * @return Number of coefficients lit.
   */
  int formFactorsFrom ( int coeff, int[] ind, double[] ff );

  /**
   * Computes "gathering" form-factors (single target coefficient is fixed).
   *
   * @param  coeff Target-coefficient index.
   * @param  ind Array to hold relevant (visible) coefficient indices (can be {@code null}).
   * @param  ff Array to hold respective form-factors (can be {@code null}).
   * @return Number of relevant coefficients.
   */
  int formFactorsTo ( int coeff, int[] ind, double[] ff );

  /** Computes single form-factor (F_{from,to}). */
  double formFactor ( int from, int to );
}
