package cz.cuni.jagrlib.iface;

import java.awt.image.*;

/**
 * Raster graphics interface (representation of pixel bitmap).
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.02
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/RasterGraphics.java">RasterGraphics.java</a>
 */
public interface RasterGraphics extends Property
{
  String WIDTH     = "Width";
  String HEIGHT    = "Height";
  String COLORMODE = "Color Mode";
  String BANDS     = "Bands";

  /**
   * Gray pixel values (<tt>0</tt> to {@link #getMaxGray}).
   */
  int MODE_GRAY     = 0;

  /**
   * Colormapped pixels (<tt>0</tt> to {@link #getMaxColormap}).
   * @see    #getColormapType
   * @see    #setColormapType
   */
  int MODE_COLORMAP = 1;

  /**
   * True-color pixel values (<tt>0</tt> to {@link #getMaxRGB}).
   */
  int MODE_RGB      = 2;

  /**
   * True-color pixel values with alpha-channel (<tt>0</tt> to {@link #getMaxRGB}).
   */
  int MODE_RGBA     = 3;

  /**
   * Higher dimensional pixel values (<tt>0</tt> to {@link #getMaxRGB}).
   */
  int MODE_BAND     = 4;

  /**
   * {@link RasterGraphics} (re-)initialization.
   *
   * @param  _width Width of a new bitmap.
   * @param  _height Height of a new bitmap.
   * @param  _mode Pixel value type (<tt>RasterGraphics.MODE_*</tt> constant).
   * @param  _bands Number of bands (for {@link #MODE_BAND}).
   * @see    #setColormapType
   */
  void init ( int _width, int _height, int _mode, int _bands );

  /**
   * {@link RasterGraphics} re-initialization (keeps current bitmap size and mode).
   */
  void init ();

  /**
   * Width of the bitmap.
   *
   * @return Current bitmap width in pixels.
   */
  int getWidth ();

  /**
   * Height of the bitmap.
   *
   * @return Current bitmap height in pixels.
   */
  int getHeight ();

  /**
   * Sample (color) model of the bitmap.
   *
   * @return Current bitmap mode (<tt>RasterGraphics.MODE_*</tt> constants).
   */
  int getMode ();

  /**
   * Number of color bands of the bitmap.
   *
   * @return Number of color bands.
   */
  int getBands ();

  /**
   * Maximal gray value.
   *
   * @return Maximal acceptable gray value.
   */
  int getMaxGray ();

  /**
   * Maximal ordinal number of colormap entry.
   *
   * @return Maximal colormap's entry number.
   */
  int getMaxColormap ();

  /**
   * Maximal RGB[A] value. Also used in multi-band mode.
   *
   * @return Maximal value for R,G,B,A and other bands.
   */
  int getMaxRGB ();

  /**
   * Copies raster graphics to a {@link BufferedImage} object.
   *
   * @return {@link BufferedImage} object.
   * @see    #setBufferedImage
   */
  BufferedImage getBufferedImage ();

  /**
   * Initializes raster graphics from a {@link BufferedImage} object.
   *
   * @param  bi Raster data.
   * @see    #getBufferedImage
   */
  void setBufferedImage ( BufferedImage bi );

  /**
   * Colormap mode (valid only for {@link #MODE_COLORMAP} bitmap mode).
   *
   * @return Colormap mode (one of (<tt>RasterGraphics.MODE_*</tt> constants).
   */
  int getColormapType ();

  /**
   * Sets the colormap mode (valid only for {@link #MODE_COLORMAP} bitmap mode).
   *
   * @param  _mode Colormap mode (one of (<tt>RasterGraphics.MODE_*</tt> constants).
   * @param  _max Maximal index of colormap entry.
   */
  void setColormapType ( int _mode, int _max );

  /**
   * Returns the colormap segment (for color modes: "<tt>getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND</tt>").
   *
   * @param  i0 The first colormap index.
   * @param  length Length of colormap segment.
   * @param  _colors Returned color levels (each band: from <tt>0.0</tt> to <tt>1.0</tt>).
   */
  void getColormap ( int i0, int length, double[][] _colors );

  /**
   * Returns the colormap segment (for color modes: "<tt>getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND</tt>").
   *
   * @param  i0 The first colormap index.
   * @param  length Length of colormap segment.
   * @param  _colors Returned color levels (each band: from <tt>0</tt> to {@link #getMaxRGB}).
   */
  void getColormap ( int i0, int length, int[][] _colors );

  /**
   * Sets the colormap segment (for color modes: "<tt>getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND</tt>").
   * Colormap size &amp; mode has to be set before using {@link #setColormapType}.
   *
   * @param  _colors Color levels to be set (each band: from <tt>0.0</tt> to <tt>1.0</tt>).
   */
  void setColormap ( double[][] _colors );

  /**
   * Sets the colormap segment (for color modes: "<tt>getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND</tt>").
   * Colormap size &amp; mode has to be set before using {@link #setColormapType}.
   *
   * @param  _colors Color levels to be set (each band: from <tt>0</tt> to {@link #getMaxRGB}).
   */
  void setColormap ( int[][] _colors );

  /**
   * Sets the binary set operation (used in all <tt>set*()</tt> methods).
   *
   * @param  _op Binary set operation (<tt>BinaryOperation.BIN_OP_*</tt> constant).
   * @return The old binary set operation.
   * @see    cz.cuni.jagrlib.BinaryOperation Binary operations
   */
  int setOperation ( int _op );

  /**
   * Sets the alpha-channel operation (used in all <tt>set*()</tt> methods).
   *
   * @param  _op Alpha-channel operation (<tt>AlphaOperation.ALPHA_*</tt> constant).
   * @return The old alpha-channel operation.
   * @see    cz.cuni.jagrlib.AlphaOperation Alpha-channel operations
   */
  int setAlphaOperation ( int _op );

  /**
   * Sets the value-transfer function (used in all <tt>set*()</tt> methods).
   *
   * @param  _transfer New value-transfer function or <tt>null</tt>.
   * @return The old Value-transfer function or <tt>null</tt>.
   * @see    cz.cuni.jagrlib.iface.ValueTransferFunction Value-transfor function prototype
   */
  ValueTransferFunction setTransferFunction ( ValueTransferFunction _transfer );

  /**
   * Sets the current drawing color (for {@link #MODE_GRAY} or
   * {@link #MODE_COLORMAP}).
   *
   * @param  _color Current drawing color (<tt>0</tt> to {@link #getMaxGray} or
   *         {@link #getMaxColormap}).
   */
  void setColor ( int _color );

  /**
   * Sets the current drawing color (for {@link #MODE_GRAY}).
   *
   * @param  _color Current drawing color (<tt>0.0</tt> to <tt>1.0</tt>).
   */
  void setColor ( double _color );

  /**
   * Sets the current drawing color (for {@link #MODE_RGB},
   * {@link #MODE_RGBA} or {@link #MODE_BAND}).
   *
   * @param  _color Current drawing color (channel values between <tt>0</tt> and {@link #getMaxRGB}).
   */
  void setColor ( int[] _color );

  /**
   * Sets the current drawing color (for {@link #MODE_RGB},
   * {@link #MODE_RGBA} or {@link #MODE_BAND}).
   *
   * @param  _color Current drawing color (channel values between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void setColor ( double[] _color );

  /**
   * Sets the current drawing alpha (for {@link #MODE_RGB},
   * {@link #MODE_RGBA} or {@link #MODE_GRAY}).
   *
   * @param  _alpha Current drawing alpha (values between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void setAlpha ( double _alpha );

  /**
   * Sets the given pixel's value.
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @see    #setColor(int)
   * @see    #setColor(double)
   * @see    #setColor(int[])
   * @see    #setColor(double[])
   */
  void setPixel ( int x, int y );

  /**
   * Sets the given pixel's value (for {@link #MODE_GRAY} or
   * {@link #MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between <tt>0</tt> and {@link #getMaxGray} or
   *                {@link #getMaxColormap}).
   */
  void putPixel ( int x, int y, int _color );

  /**
   * Sets the given pixel's value (for {@link #MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void putPixel ( int x, int y, double _color );

  /**
   * Sets the given pixel's value (for {@link #MODE_RGB},
   * {@link #MODE_RGBA} or {@link #MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between <tt>0</tt> and {@link #getMaxRGB}).
   */
  void putPixel ( int x, int y, int[] _color );

  /**
   * Sets the given pixel's value (for {@link #MODE_RGB},
   * {@link #MODE_RGBA} or {@link #MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void putPixel ( int x, int y, double[] _color );

  /**
   * Returns the pixel's value (for {@link #MODE_GRAY} or
   * {@link #MODE_COLORMAP}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value (<tt>0</tt> for off-canvas access).
   */
  int getPixel ( int x, int y );

  /**
   * Returns the pixel's value (for {@link #MODE_GRAY}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value (<tt>0.0</tt> for off-canvas access).
   */
  double getPixelDouble ( int x, int y );

  /**
   * Returns the pixel's value (for {@link #MODE_RGB},
   * {@link #MODE_RGBA} or {@link #MODE_BAND}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value.
   */
  void getPixel ( int x, int y, int[] _color );

  /**
   * Returns the pixel's value (for {@link #MODE_RGB},
   * {@link #MODE_RGBA} or {@link #MODE_BAND}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value.
   */
  void getPixel ( int x, int y, double[] _color );

  /**
   * Returns the pixel's RGB[A] values (for all graphic modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value (if the array's length is at least 4, alpha will be returned).
   */
  void getRGB ( int x, int y, int[] _color );

  /**
   * Returns the pixel's RGB[A] values (for all graphic modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value (if the array's length is at least 4, alpha will be returned).
   */
  void getRGB ( int x, int y, double[] _color );

  /**
   * Returns the pixel's gray value (for all modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value (<tt>0</tt> for off-canvas access).
   */
  int getGray ( int x, int y );

  /**
   * Returns the pixel's gray value (for all modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value (<tt>0.0</tt> for off-canvas access).
   */
  double getGrayDouble ( int x, int y );

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y].
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @see    #setColor(int)
   * @see    #setColor(double)
   * @see    #setColor(int[])
   * @see    #setColor(double[])
   */
  void setHLine ( int x1, int x2, int y );

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_GRAY} or {@link #MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color (<tt>0</tt> to {@link #getMaxGray} or
   *         {@link #getMaxColormap}).
   */
  void setHLine ( int x1, int x2, int y, int _color );

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color (<tt>0.0</tt> to <tt>1.0</tt>).
   */
  void setHLine ( int x1, int x2, int y, double _color );

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_RGB}, {@link #MODE_RGBA} or {@link #MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color (channel values between <tt>0</tt> and {@link #getMaxRGB}).
   */
  void setHLine ( int x1, int x2, int y, int[] _color );

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_RGB}, {@link #MODE_RGBA} or {@link #MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color (channel values between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void setHLine ( int x1, int x2, int y, double[] _color );

  /**
   * Sets the horizontal line's values (for {@link #MODE_GRAY} or
   * {@link #MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between <tt>0</tt> and {@link #getMaxGray} or
   *                 {@link #getMaxColormap}).
   */
  void putHLine ( int x1, int x2, int y, int[] _colors );

  /**
   * Sets the horizontal line's values (for {@link #MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void putHLine ( int x1, int x2, int y, double[] _colors );

  /**
   * Sets the horizontal line's values (for {@link #MODE_RGB},
   * {@link #MODE_RGBA} or {@link #MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between <tt>0</tt> and {@link #getMaxRGB}).
   */
  void putHLine ( int x1, int x2, int y, int[][] _colors );

  /**
   * Sets the horizontal line's values (for {@link #MODE_RGB},
   * {@link #MODE_RGBA} or {@link #MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  void putHLine ( int x1, int x2, int y, double[][] _colors );

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_GRAY} or {@link #MODE_COLORMAP}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors (<tt>0</tt> to {@link #getMaxGray} or
   *                 {@link #getMaxColormap}, <tt>0</tt> for off-canvas access).
   */
  void getHLine ( int x1, int x2, int y, int[] _colors );

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_GRAY}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors (<tt>0.0</tt> to <tt>1.0</tt>, <tt>0.0</tt> for off-canvas access).
   */
  void getHLine ( int x1, int x2, int y, double[] _colors );

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_RGB}, {@link #MODE_RGBA} or {@link #MODE_BAND}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors (channel values between <tt>0</tt> and
   *                 {@link #getMaxRGB}, <tt>0</tt> for off-canvas access).
   */
  void getHLine ( int x1, int x2, int y, int[][] _colors );

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_RGB}, {@link #MODE_RGBA} or {@link #MODE_BAND}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors (channel values between <tt>0.0</tt> and <tt>1.0</tt>, <tt>0.0</tt> for off-canvas access).
   */
  void getHLine ( int x1, int x2, int y, double[][] _colors );

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1].
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   */
  void setRectangle ( int x1, int y1, int x2, int y2 );

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1] (for
   * {@link #MODE_GRAY} or {@link #MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _color Drawing color (<tt>0</tt> to {@link #getMaxGray} or
   *         {@link #getMaxColormap}).
   */
  void setRectangle ( int x1, int y1, int x2, int y2, int _color );

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1] (for
   * {@link #MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _color Drawing color (<tt>0.0</tt> to <tt>1.0</tt>).
   */
  void setRectangle ( int x1, int y1, int x2, int y2, double _color );

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1] (for
   * {@link #MODE_RGB}, {@link #MODE_RGBA} or {@link #MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _color Drawing color (<tt>0</tt> to {@link #getMaxRGB}).
   */
  void setRectangle ( int x1, int y1, int x2, int y2, int[] _color );

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1] (for
   * {@link #MODE_RGB}, {@link #MODE_RGBA} or {@link #MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _color Drawing color (<tt>0.0</tt> to <tt>1.0</tt>).
   */
  void setRectangle ( int x1, int y1, int x2, int y2, double[] _color );

  /**
   * Sets the rectangle's values (for {@link #MODE_GRAY} or
   * {@link #MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _colors Rectangle colors (between <tt>0</tt> and {@link #getMaxGray} or
   *                 {@link #getMaxColormap}).
   */
  //void putRectangle ( int x1, int y1, int x2, int y2, int[][] _colors );

  /**
   * Sets the rectangle's values (for {@link #MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _colors Rectangle colors (between <tt>0</tt> and {@link #getMaxGray}).
   */
  //void putRectangle ( int x1, int y1, int x2, int y2, double[][] _colors );

  /**
   * Sets the rectangle's values (for {@link #MODE_RGB},
   * {@link #MODE_RGBA} or {@link #MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _colors Rectangle colors (between <tt>0</tt> and {@link #getMaxRGB}).
   */
  //void putRectangle ( int x1, int y1, int x2, int y2, int[][][] _colors );

  /**
   * Sets the rectangle's values (for {@link #MODE_RGB},
   * {@link #MODE_RGBA} or {@link #MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  _colors Rectangle colors (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  //void putRectangle ( int x1, int y1, int x2, int y2, double[][][] _colors );

}
