package cz.cuni.jagrlib.iface;

/**
 * Searching in raster images.
 *
 * @version 0.26 $Rev: 458 $ $Date: 2007-05-20 23:45:00 +0200 (ne, 20 V 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/RasterSearch.java">RasterSearch.java</a>
 */
public interface RasterSearch extends Property
{
  /** Error metrics: Mean Absolute Error (average of absolute differences of individual samples). */
  int ERR_MAE  = 0;

  /** Error metrics: Root Mean Squared Error (average of squared differences of individual samples, square root of that average). */
  int ERR_RMSE = 1;

  /** Orientation: original ({@code 0123} from upper left to lower right corner). */
  int ORI_0123 = 0;

  /** Orientation: mirror of {@link #ORI_0123}. */
  int ORI_0213 = 1;

  /** Orientation: 90-degree rotation of {@link #ORI_0123}. */
  int ORI_2031 = 2;

  /** Orientation: mirror of {@link #ORI_2031}. */
  int ORI_1032 = 3;

  /** Orientation: 180-degree rotation of {@link #ORI_0123}. */
  int ORI_3210 = 4;

  /** Orientation: mirror of {@link #ORI_3210}. */
  int ORI_3120 = 5;

  /** Orientation: 270-degree rotation of {@link #ORI_0123}. */
  int ORI_1302 = 6;

  /** Orientation: mirror of {@link #ORI_1302}. */
  int ORI_2301 = 7;

  /** Result code for failed search. */
  int ORI_FAIL = -1;

  /**
   * Transform matrices from destination space {@code [x,y]} to source space {@code [u,v]}.
   * Code to compute source coordinates:
   * <pre>
   * int x, y, size;    // destination coordinates and rectangle size
   * int ori;           // orientation
   * ...
   * int[] tr = ORI_INV[ ori ];
   * int u = x * tr[0] + y * tr[1] + (size-1) * tr[2];
   * int v = x * tr[3] + y * tr[4] + (size-1) * tr[5];
   * </pre>
   */
  int[][] ORI_INV =
  {
    // x   y  abs  x   y  abs
    {  1,  0,  0,  0,  1,  0 },           // ORI_0123
    {  0,  1,  0,  1,  0,  0 },           // ORI_0213
    {  0, -1,  1,  1,  0,  0 },           // ORI_2031
    { -1,  0,  1,  0,  1,  0 },           // ORI_1032
    { -1,  0,  1,  0, -1,  1 },           // ORI_3210
    {  0, -1,  1, -1,  0,  1 },           // ORI_3120
    {  0,  1,  0, -1,  0,  1 },           // ORI_1302
    {  1,  0,  0,  0, -1,  1 },           // ORI_2301
  };

  /** Maximum number of reference images. */
  int MAX_REFERENCE = 4;

  /**
   * Sets reference image.
   * Reference images have to be set before you can search in them..
   *
   * @param  g Read-only raster image (can be {@code null} to reset the slot).
   * @param  id Reference image identifier ({@code 0}, {@code 1}, .. {@link #MAX_REFERENCE}{@code -1}).
   * @param  scale Scale to be used in precomputing.
   * @param  oriMask Mask of orientations used in future search commands (non-mandatory,
   *         use {@code 0} if not known).
   * @param  pixelTransform Will a linear pixel transform be used?
   * @see    #ORI_0123
   */
  void setReferencePicture ( RasterGraphics g, int id, double scale, int oriMask, boolean pixelTransform );

  /**
   * Sets parameters for subsequent searches.
   *
   * @param  oriMask Mask of valid orientations to search ({@code 1 << }{@link #ORI_0123}, ..).
   * @param  bandMask Mask of searched image bands (Red .. 1, Green .. 2, Blue .. 4, ..).
   * @param  metrics One of {{@link #ERR_MAE}, {@link #ERR_RMSE}}.
   * @param  precision Sub-pixel precision of result coordinates (should be power of {@code 1/2}).
   * @param  absMinX Lower bound for X-coordinate of absolute search.
   * @param  absMaxX Upper bound for X-coordinate of absolute search (can be {@code 0} for unbounded search).
   * @param  absMinY Lower bound for Y-coordinate of absolute search.
   * @param  absMaxY Upper bound for Y-coordinate of absolute search (can be {@code 0} for unbounded search).
   * @param  relX X-amplitude of relative search (must be {@code 0} for absolute search).
   * @param  relY Y-amplitude of relative search.
   * @param  minMul Minimum pixel-value multiplier (should be {@code 1} if no multiplication is allowed).
   * @param  maxMul Maximum pixel-value multiplier (should be {@code 1} if no multiplication is allowed).
   * @param  minAdd Minimum additive coefficient (should be {@code 0} if no addition is allowed).
   * @param  maxAdd Maximum additive coefficient (should be {@code 0} if no addition is allowed).
   */
  void setSearchParameters ( int oriMask, int bandMask, int metrics, double precision,
                             int absMinX, int absMaxX, int absMinY, int absMaxY,
                             int relX, int relY,
                             double minMul, double maxMul, double minAdd, double maxAdd );

  /**
   * Searches for the most similar rectangle in one of predefined reference images.
   *
   * @param  f Raster image containing the sought rectangle.
   * @param  x Origin of the sought rectangle.
   * @param  y Origin of the sought rectangle.
   * @param  width Width of the sought rectangle.
   * @param  height Height of the sought rectangle.
   * @param  ref Reference image to be scanned.
   * @param  result Pre-allocated result array {@code double[5]},
   *         will hold coordinates, coefficients and error measure {@code [x,y,mul,add,err]}.
   * @return Orientation of the best fit (or {@link #ORI_FAIL} in case of failure).
   */
  int findBestRectangle ( RasterGraphics f, int x, int y, int width, int height,
                          int ref, double[] result );

}
