package cz.cuni.jagrlib.iface;

/**
 * General mappings to scalar R range.
 *
 * @version 0.25 $Rev: 287 $ $Date: 2006-06-17 23:26:23 +0200 (so, 17 VI 2006) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/RealFunction.java">RealFunction.java</a>
 */
public interface RealFunction extends Property
{
  /** Property name: amplitude. */
  String AMPLITUDE   = "Amplitude";

  /** Property name: global frequency. */
  String FREQUENCY   = "Frequency";

  /** Property name: frequency in the <tt>X</tt> coordinate. */
  String FREQUENCY_X = "Frequency X";

  /** Property name: frequency in the <tt>Y</tt> coordinate. */
  String FREQUENCY_Y = "Frequency Y";

  /** Property name: variation of the function (used in noise functions). */
  String VARIATION   = "Variation";

  /**
   * Compute the mapping: {@code double -> double}.
   */
  double f ( double x );

  /**
   * Compute the mapping: {@code int -> double}.
   */
  double f ( int x );

  /**
   * Compute the mapping: {@code double[2] -> double}.
   */
  double f ( double x, double y );

  /**
   * Compute the mapping: {@code int[2] -> double}.
   */
  double f ( int x, int y );

  /**
   * Compute the mapping: {@code double[3] -> double}.
   */
  double f ( double x, double y, double z );

  /**
   * Compute the mapping: {@code int[3] -> double}.
   */
  double f ( int x, int y, int z );

  /**
   * Compute the mapping: {@code double[] -> double}.
   */
  double f ( double[] arg );

  /**
   * Compute the mapping: {@code int[] -> double}.
   */
  double f ( int[] arg );

  /**
   * Compute erivative: {@code df(x)/dx}. Non-mandatory.
   *
   * @param  x Function argument.
   * @return Function derivative or <tt>0.0</tt> if not implemented.
   */
  double df ( double x );

  /**
   * Compute partial derivatives: {@code df(x,y)/dx} and {@code df(x,y)/dy}. Non-mandatory.
   *
   * @param  x X coordinate (1st argument).
   * @param  y Y coordinate (2nd argument).
   * @param  d Array to hold results (<tt>d[0]</tt> = df(x,y)/dx,
   *           <tt>d[1]</tt> = df(x,y)/dy).
   * @return <tt>true</tt> if this service is implemented.
   */
  boolean df ( double x, double y, double d[] );

  /**
   * Compute partial derivatives: {@code df(x,y,z)/dx}, {@code df(x,y,z)/dy} and {@code df(x,y,z)/dz}.
   * Non-mandatory.
   *
   * @param  x X coordinate (1st argument).
   * @param  y Y coordinate (2nd argument).
   * @param  z Z coordinate (3rd argument).
   * @param  d Array to hold results (<tt>d[0]</tt> = df(x,y,z)/dx,
   *           <tt>d[1]</tt> = df(x,y,z)/dy, <tt>d[2]</tt> = df(x,y,z)/dz).
   * @return <tt>true</tt> if this service is implemented.
   */
  boolean df ( double x, double y, double z, double d[] );

  /**
   * Compute partial derivatives: {@code df(x1,..)/dx1}, ...
   * Non-mandatory.
   *
   * @param  arg Coordinate array.
   * @param  d Array to hold results (<tt>d[0]</tt> = df(x1,...)/dx1, ...).
   * @return <tt>true</tt> if this service is implemented.
   */
  boolean df ( double[] arg, double d[] );

  /**
   * Returns domain of the mapping.
   *
   * @param  dom Domain array (<tt>dom[0]</tt> = Xmin, <tt>dom[1]</tt> = Xmax, etc.).
   * @return <tt>null</tt> if this service is not implemented.
   */
  double[] getDomain ( double[] dom );

  /**
   * Returns domain of the mapping.
   *
   * @param  dom Domain array (<tt>dom[0]</tt> = Xmin, <tt>dom[1]</tt> = Xmax, etc.).
   * @return <tt>null</tt> if this service is not implemented.
   */
  int[] getDomain ( int[] dom );

}
