package cz.cuni.jagrlib.iface;

/**
 * Data object holding system of linear equations (SLE) data.
 *
 * Both input (matrix of coefficients and right-hand-side vector) and
 * output (solution vector) data can be retrieved.
 *
 * Prepared for actual data storage (full or sparse SLE matrix) on-the-fly computation
 * or proxy-access to some other object..
 *
 * @author  PE, Tomas Svab
 * @version 0.26 $Rev: 757 $ $Date: 2008-05-19 01:09:45 +0200 (po, 19 V 2008) $ $Author: pepca $
 * @since   0.26
 * @see     SLEDataStore
 * @see     SLESolver
 */
public interface SLEData extends Property
{
  //==========================================================================
  //  Input SLE data:

  /**
   * Number of equations = number of matrix rows
   * (usually number of unknowns is equal to number of equations).
   */
  int getRows ();

  /**
   * Number of unknowns = number of matrix columns
   * (usually number of unknowns is equal to number of equations).
   */
  int getColumns ();

  /** Reads a single matrix coefficient. */
  double getCoeff ( int row, int col );

  /** Reads the whole matrix row. */
  double[] getRow ( int row, double[] result );

  /** Reads the whole matrix column. */
  double[] getColumn ( int col, double[] result );

  /** Reads the whole matrix. */
  double[][] getMatrix ();

  /** Reads a single element of the right-side vector. */
  double getRight ( int row );

  /**
   * Reads the whole right-side vector.
   *
   * @param  result Pre-allocated array to hold result (can be {@code null}).
   */
  double[] getRight ( double[] result );

  //==========================================================================
  //  Solution vector:

  /** Reads a single unknown (element of solution vector). */
  double getSolution ( int col );

  /** Reads the whole solution vector. */
  double[] getSolution ( double[] result );

  /** Writes value of a single unknown. */
  void setSolution ( int col, double value );

  /** Writes the whole solution vector. */
  void setSolution ( double[] values );

  //==========================================================================
  //  Incremental SLE changes (non-mandatory):

  /** Value used as start- and stop- value for iterators. */
  int NULL = -1;

  /** Returns index of next right-side element which was changed since last {@link #resetChanges} call. */
  int nextRightChanged ( int row );

  /**
   * Returns column-index of next coefficient which was changed since last {@link #resetChanges} call.
   * Passes through the given matrix row..
   */
  int nextCoeffInRowChanged ( int row, int col );

  /**
   * Returns row-index of next coefficient which was changed since last {@link #resetChanges} call.
   * Passes through the given matrix column..
   */
  int nextCoeffInColumnChanged ( int row, int col );

  /** Resets (acknowledges) all SLE changes. */
  void resetChanges ();

  //==========================================================================
  //  Additional information:

  /** Reads a single element weight (radiosity: element area). */
  double getWeight ( int i );

  /**
   * Reads the whole weight vector.
   *
   * @param  values Pre-allocated array to hold result (can be {@code null}).
   */
  double[] getWeight ( double[] values );

  /** Reads a single reflectance (radiosity: element albedo). */
  double getReflectance ( int i );

  /**
   * Reads the whole reflectance vector.
   *
   * @param  reflectance Pre-allocated array to hold result (can be {@code null}).
   */
  double[] getReflectance ( double[] reflectance );

  //==========================================================================
  //  Multi-component systems:

  /** Get actual component number. */
  int actualComponent ();

  /** Set actual component (non-mandatory). */
  void setComponent ( int c );
}
