package cz.cuni.jagrlib.iface;

/**
 * General sampled data file (D-dimensional array of scalars).
 *
 * @version 0.25 $Rev: 300 $ $Date: 2006-07-20 01:03:44 +0200 (čt, 20 VII 2006) $ $Author: pepca $
 * @since   0.10
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/SampleData.java">SampleData.java</a>
 */
public interface SampleData extends Property
{
  /** 8-bit signed integer value. */
  int SAMPLE_BYTE    = 0;

  /** 16-bit signed integer value. */
  int SAMPLE_SHORT   = 1;

  /** 32-bit signed integer value. */
  int SAMPLE_INT     = 2;

  /** 64-bit signed integer value. */
  int SAMPLE_LONG    = 3;

  /** floating-point (single precision) value. */
  int SAMPLE_FLOAT   = 4;

  /** floating-point (double precision) value. */
  int SAMPLE_DOUBLE  = 5;

  /** Complex number (double[2] array - uses double-sized array). */
  int SAMPLE_COMPLEX = 6;

  /** RGB color (packed into one integer). */
  int SAMPLE_RGBA    = 7;

  /** Property name: sample type. */
  String SAMPLE_TYPE = "Sample type";

  /** Property name: frequency (in Hz). */
  String FREQUENCY   = "Frequency";

  /**
   * (Re-)initialize the data array.
   *
   * @param  type Sample type (one of <tt>SAMPLE_*</tt> constants).
   * @param  mapped Should the array contain mapped values?
   * @param  dim Dimensions of an array.
   * @see    #SAMPLE_BYTE
   * @see    #SAMPLE_SHORT
   * @see    #SAMPLE_INT
   * @see    #SAMPLE_LONG
   * @see    #SAMPLE_FLOAT
   * @see    #SAMPLE_DOUBLE
   * @see    #SAMPLE_COMPLEX
   * @see    #SAMPLE_RGBA
   */
  void resetData ( int type, boolean mapped, int[] dim );

  /**
   * Returns actual sample type (<tt>SAMPLE_*</tt> constant).
   *
   * @return Sample type (<tt>SAMPLE_*</tt> constant).
   * @see    #getDataDim
   * @see    #SAMPLE_BYTE
   * @see    #SAMPLE_SHORT
   * @see    #SAMPLE_INT
   * @see    #SAMPLE_LONG
   * @see    #SAMPLE_FLOAT
   * @see    #SAMPLE_DOUBLE
   * @see    #SAMPLE_COMPLEX
   * @see    #SAMPLE_RGBA
   */
  int getSampleType ();

  /**
   * Returns actual array dimensions.
   *
   * @param  dim Pre-allocated array to hold individual dimensions' sizes.
   * @return Dimensionality (number of dimensions).
   * @see    #getSampleType
   */
  int getDataDim ( int[] dim );

  /**
   * Sets the sample geometry (in regular sample matrix).
   *
   * @param geom Sample periods for all dimensions.
   * @see   #getSampleGeometry
   */
  void setSampleGeometry ( double[] geom );

  /**
   * Gets the sample geometry (in regular sample matrix).
   *
   * @param geom Sample periods for all dimensions.
   * @see   #setSampleGeometry
   */
  void getSampleGeometry ( double[] geom );

  /**
   * Sets (partially) the new sample mapping (for mapped arrays only).
   *
   * @param map Mapping array to be set (should be an array of the apropriate type).
   * @param srcFrom First source index (<tt>map[from]</tt>).
   * @param srcTo After the last source index.
   * @param dstFrom First destination index.
   * @see   #getSampleMapping
   */
  void setSampleMapping ( Object map, int srcFrom, int srcTo, int dstFrom );

  /**
   * Retrieves the whole sample mapping (for mapped arrays only).
   *
   * @param map Optional pre-allocated mapping array to be filled.
   * @param srcFrom First index to be retrieved.
   * @param srcTo After the last index to be retrieved.
   * @param bounds Retrieved index bounds (<tt>bounds[0]</tt> = the 1st mapped value
   *               written in map[0], <tt>bounds[1]</tt> = index after the last value).
   * @see   #setSampleMapping
   */
  Object getSampleMapping ( Object map, int srcFrom, int srcTo, int[] bounds );

  /**
   * Returns the original data array.
   *
   * @param  dim Returned dimensionality (array of individual dimension sizes). Can be null.
   * @return Packed 1D sample array (read / write access) or <tt>null</tt> if the service is not
   *         available.
   * @see    #setDataArray
   */
  Object getDataArray ( int[] dim );

  /**
   * Sets data array in the raw (internal) format. Can be used for data fragment
   * or for the whole data array.
   *
   * @param data Packed 1D source data array (of apropriate type - integer has to be used for mapped types).
   * @param srcFrom Source index to begin from.
   * @param dstFrom Start of the region to be set. Can be null (instead of [0,0,0,..]).
   * @param dstTo End of the region to be set. Can be null (instead of dim).
   */
  void setDataArray ( Object data, int srcFrom, int[] dstFrom, int[] dstTo );

  /**
   * Reads the given sample segment (general data format).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result (packed 1D array).
   * @return Number of returned samples.
   */
  int getSamplesGeneral ( int[] from, int[] to, Object data );

  /**
   * Writes the given sample segment (general data format).
   *
   * @param from Starting index.
   * @param to Index after the written data.
   * @param data Data to be written - in 1D array (integer indices for mapped data arrays).
   */
  void setSamplesGeneral ( int[] from, int[] to, Object data );

  /**
   * Reads the given sample segment (integer formats: byte, short, int, RGBA).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result.
   * @return Number of returned samples.
   */
  int getSamples ( int[] from, int[] to, int[] data );

  /**
   * Reads the given sample segment (long format).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result.
   * @return Number of returned samples.
   */
  int getSamples ( int[] from, int[] to, long[] data );

  /**
   * Reads the given sample segment (float formats: float, double, complex).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result (complex uses double-sized array).
   * @return Number of returned samples.
   */
  int getSamples ( int[] from, int[] to, double[] data );

  /**
   * Writes the given sample segment (integer formats: byte, short, int, RGBA).
   *
   * @param from Starting index.
   * @param to Index after written data.
   * @param data Data to be written (1D array).
   */
  void setSamples ( int[] from, int[] to, int[] data );

  /**
   * Writes the given sample segment (long format).
   *
   * @param from Starting index.
   * @param to Index after written data.
   * @param data Data to be written (1D array).
   */
  void setSamples ( int[] from, int[] to, long[] data );

  /**
   * Writes the given sample segment (float formats: float, double, complex).
   *
   * @param from Starting index.
   * @param to Index after written data.
   * @param data Data to be written (1D array). Complex uses double-sized array.
   */
  void setSamples ( int[] from, int[] to, double[] data );

}
