package cz.cuni.jagrlib.iface;

import java.util.*;

import cz.cuni.jagrlib.*;

/**
 * Elementary solid of 3D scene: it is able to compute ray-intersections (including normal vector,
 * and 2D texture coordinate), manages bounding-volume and polyhedral representation.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.11
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/Solid.java">Solid.java</a>
 * @see     BoundingVolume
 * @see     Brep
 * @see     CSGLeaf
 * @see     MicroFacet
 */
public interface Solid extends Property, IntersectionAttributes
{
  /**
   * Creates polyhedral representation of the solid (non-mandatory method).
   *
   * @param  b Global B-rep data structure.
   * @param  smoothness Quality-factor of an approximation ({@code 0.0} for default
   *                    approximation).
   * @return Handle of the created solid (or {@link Brep#NULL NULL} if the method is not
   *         implemented).
   * @see    Brep#NULL
   */
  int createPolyhedr ( Brep b, double smoothness );

  /**
   * Computes intersection of the solid with the given ray.
   *
   * @param  P0 Ray origin ({@code double[3]} or {@code double[4]}).
   * @param  P1 Direction vector of the ray ({@code double[3]} or {@code double[4]}).
   * @return List of intersection points ({@code MicroFacet}s) sorted by {@code t}
   *         (parametric distance from {@code P0}). {@code null} is returned in case
   *         of bad arguments or if no intersection exists..
   * @see    MicroFacet
   * @see    BoundingVolume
   */
  List< MicroFacet > intersection ( double[] P0, double[] P1 );

  /**
   * Returns bounding volume of the solid (non-mandatory method).
   *
   * @param  toWorld Transform from local (solid) coordinates to world coordinates
   *         (where bounding volume will live).
   * @return Bounding volume of the solid or {@code null} if method is not implemented.
   * @see    #intersection
   */
  BoundingVolume getBoundingVolume ( TrMatrix toWorld );

  /**
   * Point vs. solid test ("inside-test").
   *
   * @param  point Coordinates of the 3D point ({@code double[3]} or
   *               {@code double[4]}).
   * @return {@code true} if the given point lies inside the solid.
   */
  boolean isInside ( double[] point );

}
