package cz.cuni.jagrlib.iface;

import java.io.*;

/**
 * Interface for stroke call-back routines.
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.02
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/StrokeCallback.java">StrokeCallback.java</a>
 */
public interface StrokeCallback extends Serializable
{
  /** Use bit-mask modulation. */
  int MASK_BIT        = 0x01;

  /** Use alpha-channel modulation. */
  int MASK_ALPHA      = 0x02;

  /** Use color modulation. */
  int MASK_COLOR      = 0x04;

  /** Use line width modulation. */
  int MASK_WIDTH      = 0x08;

  /** Use parametric (relative) coordinates. */
  int MASK_PARAMETRIC = 0x10;

  /**
   * Tells which features are used by this stroke-modulation function.
   *
   * @return Flags using <a href="#MASK_BIT">MASK_*</a> constants.
   */
  int getFeatures ();

  /**
   * Returns boolean information about the given point on the stroke.
   *
   * @param  i Point's index (from <tt>0</tt> to <tt>length-1</tt>).
   * @param  length Total stroke length.
   * @param  cyclic Is that stroke cyclic?.
   * @return <tt>true</tt> for the point that has to be drawn.
   */
  boolean getBit ( int i, int length, boolean cyclic );

  /**
   * Returns boolean information about the given point on the stroke.
   *
   * @param  t Point's relative coordinate (from <tt>0.0</tt> to <tt>1.0</tt>).
   * @param  cyclic Is that stroke cyclic?.
   * @return <tt>true</tt> for the point that has to be drawn.
   */
  boolean getBit ( double t, boolean cyclic );

  /**
   * Returns alpha-channel information about the given point on the stroke.
   *
   * @param  i Point's index (from <tt>0</tt> to <tt>length-1</tt>).
   * @param  length Total stroke length.
   * @param  cyclic Is that stroke cyclic?.
   * @return Alpha (opacity) to draw the point with (from <tt>0.0</tt> to <tt>1.0</tt>).
   */
  double getAlpha ( int i, int length, boolean cyclic );

  /**
   * Returns alpha-channel information about the given point on the stroke.
   *
   * @param  t Point's relative coordinate (from <tt>0.0</tt> to <tt>1.0</tt>).
   * @param  cyclic Is that stroke cyclic?.
   * @return Alpha (opacity) to draw the point with (from <tt>0.0</tt> to <tt>1.0</tt>).
   */
  double getAlpha ( double t, boolean cyclic );

  /**
   * Returns color information (RGB) about the given point on the stroke.
   *
   * @param  color RGB color of the point.
   * @param  i Point's index (from <tt>0</tt> to <tt>length-1</tt>).
   * @param  length Total stroke length.
   * @param  cyclic Is that stroke cyclic?.
   */
  void getColor ( double[] color, int i, int length, boolean cyclic );

  /**
   * Returns color information (RGB) about the given point on the stroke.
   *
   * @param  color RGB color of the point.
   * @param  t Point's relative coordinate (from <tt>0.0</tt> to <tt>1.0</tt>).
   * @param  cyclic Is that stroke cyclic?.
   */
  void getColor ( double[] color, double t, boolean cyclic );

  /**
   * Returns width of the given point on the stroke.
   *
   * @param  i Point's index (from <tt>0</tt> to <tt>length-1</tt>).
   * @param  length Total stroke length.
   * @param  cyclic Is that stroke cyclic?.
   * @return Width of the point.
   */
  double getWidth ( int i, int length, boolean cyclic );

  /**
   * Returns width of the given point on the stroke.
   *
   * @param  t Point's relative coordinate (from <tt>0.0</tt> to <tt>1.0</tt>).
   * @param  cyclic Is that stroke cyclic?.
   * @return Width of the point.
   */
  double getWidth ( double t, boolean cyclic );

}
