package cz.cuni.jagrlib.iface;

/**
 * Rendering interface for text output.
 *
 * @version 0.25 $Rev: 359 $ $Date: 2006-11-12 22:20:54 +0100 (ne, 12 XI 2006) $ $Author: pepca $
 * @since   0.02
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/TextRender.java">TextRender.java</a>
 */
public interface TextRender extends Render
{
  /** Bold (heavy) font. */
  int FONT_BOLD       = 0x0001;

  /** Italic (slanted) font. */
  int FONT_ITALIC     = 0x0002;

  /** Underlined font. */
  int FONT_UNDERLINE  = 0x0004;

  /** Font with serifs. */
  int FONT_SERIF      = 0x0020;

  /** Manuscript font. */
  int FONT_SCRIPT     = 0x0040;

  /** Decorative font. */
  int FONT_DECORATIVE = 0x0080;

  /** Icon font (wingdings etc.). */
  int FONT_ICON       = 0x0100;

  /** Monospaced (non-proportional) font. */
  int FONT_MONOSPACED = 0x0200;

  /** The asked font was created exactly. */
  int CREATE_EXACT    = 0;

  /** The asked font was approximated. */
  int CREATE_APPROX   = 1;

  /** Font creation process has failed .. no font has been created. */
  int CREATE_FAIL     = -1;

  /**
   * Maximal number of font <b>slots</b> (excluding <b>slot A</b>).
   * Slots are numbered from <tt>0</tt> to <tt>MAX_SLOT-1</tt>.
   */
  int MAX_SLOT        = 8;

  /** Font path can be stroked (stroked font). */
  int PROP_STROKE     = 0x0001;

  /** Font path can be filled (filled font). */
  int PROP_FILL       = 0x0002;

  /** Font is not able to generate paths (bitmap, non-scalable font). */
  int PROP_BITMAP     = 0x0004;

  /** Font is monospaced (non-proportional). */
  int PROP_MONOSPACED = 0x0008;

  /** Property name: antialiasing mode. */
  String ANTI_MODE    = "Antialiasing";

  /** Antialiasing mode value: depends on output module color format. */
  int ANTI_DEPEND     = 0;

  /** Antialiasing mode value: force antialiasing off (even in RGBA output mode). */
  int ANTI_OFF        = 1;

  /** Antialiasing mode value: force antialiasing on (even in RGB/Gray/.. output mode). */
  int ANTI_ON         = 2;

  /**
   * Creates a new font and places it in <b>slot A</b> (actual, working font).
   *
   * @param  name Font name.
   * @param  style Font style - used if exact font match was not successful
   *               (bitwise OR of <tt>TextRender.FONT_*</tt> constants).
   * @param  size Font size (distance between two regularly spaced text lines).
   * @return Result code (one of <tt>TextRender.CREATE_*</tt> constants).
   * @see    #FONT_BOLD
   * @see    #CREATE_EXACT
   * @see    cz.cuni.jagrlib.VCommand.CreateFont
   */
  int createFont ( String name, int style, double size );

  /**
   * Changes the actual font's orientation.
   *
   * @param  angle New font orientation (in degrees, counterclockwise).
   * @return The old orientation.
   * @see    cz.cuni.jagrlib.VCommand.FontOrientation
   */
  double fontOrientation ( double angle );

  /**
   * Assigns the actual font (<b>slot A</b>) from the given <b>slot</b>.
   *
   * @param  slot Slot number (between <tt>0</tt> and <tt>MAX_SLOT-1</tt>).
   * @see    #MAX_SLOT
   * @see    cz.cuni.jagrlib.VCommand.GetFont
   */
  void getFont ( int slot );

  /**
   * Puts the actual font (<b>slot A</b>) into the given <b>slot</b>.
   *
   * @param  slot Slot number (between <tt>0</tt> and <tt>MAX_SLOT-1</tt>).
   * @see    #MAX_SLOT
   * @see    cz.cuni.jagrlib.VCommand.PutFont
   */
  void putFont ( int slot );

  /**
   * Returns properties (flags) of the actual font (<b>slot A</b>).
   *
   * @return Bitwise OR of <tt>PROP_*</tt> constants.
   * @see    #PROP_STROKE
   */
  int fontProperties ();

  /**
   * Generates path of the given text string and <u>appends</u> it to the current path.
   * Uses the <b>current point</b> as starting reference point (<b>current point</b>
   * itself will be updated by this command).
   *
   * <p>Does nothing if the actual font has neither <tt>PROP_STROKE</tt> nor
   * <tt>PROP_FILL</tt> capability.
   *
   * @param  text The string to be drawn.
   * @see    #textExtent
   * @see    #drawText
   * @see    cz.cuni.jagrlib.VCommand.MoveTo
   * @see    VectorGraphics#moveTo
   * @see    cz.cuni.jagrlib.VCommand.TextPath
   */
  void textPath ( String text );

  /**
   * Support class for holding <b>text extents</b>.
   *
   * <p>Every string output can be placed into <b>bounding rectangle</b>.
   *
   * <p><b>Reference point</b> is a point at the base-line
   * which defines start of each text drawing (<b>starting reference point</b>). After drawing
   * is done, consecutive text should start at the <b>finishing reference point</b> of the first
   * output.
   *
   * <p>Entent rectangle is defined in <b>baseline-based coordinate system</b> (text drawing direction
   * is ignored). Other data are in <b>device coordinates</a> ({@link TextExtent#x0},
   * {@link TextExtent#y0}, {@link TextExtent#x1} and {@link TextExtent#y1}).
   *
   * @version 0.25 $Rev: 359 $ $Date: 2006-11-12 22:20:54 +0100 (ne, 12 XI 2006) $ $Author: pepca $
   * @since   0.02
   * @see     <a href="../../../../../cz/cuni/jagrlib/iface/TextRender.java">TextRender.java</a>
   * @see     TextRender#textPath
   * @see     TextRender#drawText
   * @see     TextRender#textExtent
   * @see     TextRender#getKern
   * @see     cz.cuni.jagrlib.VCommand.TextPath
   * @see     cz.cuni.jagrlib.VCommand.DrawText
   */
  public class TextExtent
  {
    /** X coordinate of starting reference point (device coordinates). */
    public double x0;

    /** Y coordinate of starting reference point (device coordinates). */
    public double y0;

    /** X coordinate of finishing reference point (device coordinates). */
    public double x1;

    /** Y coordinate of finishing reference point (device coordinates). */
    public double y1;

    /** X coordinate of upper-left extent corner (baseline coordinates). */
    public double ulx;

    /** Y coordinate of upper-left extent corner (baseline coordinates). */
    public double uly;

    /** X coordinate of lower-right extent corner (baseline coordinates). */
    public double lrx;

    /** Y coordinate of lower-right extent corner (baseline coordinates). */
    public double lry;
  }

  /**
   * Draws the given text. Starting reference point is explicitly specified.
   * If <tt>ext</tt> parameter is not <tt>null</tt>, the routine fills
   * the given <tt>TextExtent</tt> object.
   *
   * <p>Works with both vector and bitmap fonts.
   *
   * @param  x0 X coordinate of the starting reference point.
   * @param  y0 Y coordinate of the starting reference point.
   * @param  text The string to be drawn.
   * @param  ext The <tt>TextExtent</tt> object to be filled (or <tt>null</tt>).
   * @see    TextRender.TextExtent
   * @see    #textExtent
   * @see    cz.cuni.jagrlib.VCommand.DrawText
   */
  void drawText ( double x0, double y0, String text, TextExtent ext );

  /**
   * Computes the <b>extent</b> of the given text string (without any output change).
   *
   * @param  x0 X coordinate of the starting reference point.
   * @param  y0 Y coordinate of the starting reference point.
   * @param  text The string to be drawn.
   * @param  ext The <tt>TextExtent</tt> object to be filled (or <tt>null</tt>).
   * @return The filled <tt>TextExtent</tt> object.
   * @see    TextRender.TextExtent
   * @see    #drawText
   */
  TextExtent textExtent ( double x0, double y0, String text, TextExtent ext );

  /**
   * Returns the <b>kerning</b> (space correction) which should be used between two given
   * characters.
   *
   * @param  c1 First character.
   * @param  c2 Second character.
   * @param  ext The <tt>TextExtent</tt> object to be filled (or <tt>null</tt>).
   * @return The filled <tt>TextExtent</tt> object (only <tt>dx</tt> and <tt>dy</tt>
   *         are used).
   * @see    TextRender.TextExtent
   * @see    #textExtent
   */
  TextExtent getKern ( char c1, char c2, TextExtent ext );

}
