package cz.cuni.jagrlib.iface;

/**
 * Value-transfer function interface: representation of R -&gt; R function (Pixel -&gt; Pixel).
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.02
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/ValueTransferFunction.java">ValueTransferFunction.java</a>
 */
public interface ValueTransferFunction extends Property
{
  /**
   * Scalar value-transfer function.
   *
   * @param  value Input value (range from <tt>0</tt> to <tt>max</tt>).
   * @param  max Maximal input/output value.
   * @return Output value (range from <tt>0</tt> to <tt>max</tt>).
   */
  int transfer ( int value, int max );

  /**
   * Scalar inverse value-transfer function.
   *
   * @param  value Input value (range from <tt>0</tt> to <tt>max</tt>).
   * @param  max Maximal input/output value.
   * @return Output value (range from <tt>0</tt> to <tt>max</tt>).
   */
  int transferInv ( int value, int max );

  /**
   * Scalar value-transfer function.
   *
   * @param  value Input value (range from <tt>0.0</tt> to <tt>1.0</tt>).
   * @return Output value (range from <tt>0.0</tt> to <tt>1.0</tt>).
   */
  double transfer ( double value );

  /**
   * Scalar inverse value-transfer function.
   *
   * @param  value Input value (range from <tt>0.0</tt> to <tt>1.0</tt>).
   * @return Output value (range from <tt>0.0</tt> to <tt>1.0</tt>).
   */
  double transferInv ( double value );

  /**
   * Gr[A] or RGB[A] value-transfer function.
   * Can be used for higher dimensions too.
   *
   * @param  input Input value (scalar range from <tt>0</tt> to <tt>max</tt>).
   * @param  output Output value (scalar range from <tt>0</tt> to <tt>max</tt>).
   *                Can be the same as <tt>input</tt> (in-place operation).
   * @param  max Maximal input/output value.
   */
  void transfer ( int[] input, int[] output, int max );

  /**
   * Gr[A] or RGB[A] inverse value-transfer function.
   * Can be used for higher dimensions too.
   *
   * @param  input Input value (scalar range from <tt>0</tt> to <tt>max</tt>).
   * @param  output Output value (scalar range from <tt>0</tt> to <tt>max</tt>).
   *                Can be the same as <tt>input</tt> (in-place operation).
   * @param  max Maximal input/output value.
   */
  void transferInv ( int[] input, int[] output, int max );

  /**
   * Gr[A] or RGB[A] value-transfer function.
   * Can be used for higher dimensions too.
   *
   * @param  input Input value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   * @param  output Output value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   *                Can be the same as <tt>input</tt> (in-place operation).
   */
  void transfer ( double[] input, double[] output );

  /**
   * Gr[A] or RGB[A] inverse value-transfer function.
   * Can be used for higher dimensions too.
   *
   * @param  input Input value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   * @param  output Output value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   *                Can be the same as <tt>input</tt> (in-place operation).
   */
  void transferInv ( double[] input, double[] output );

  /**
   * Gr[A] or RGB[A] -&gt; index (classification) function.
   * Typical usage: color quantization.
   *
   * @param  input Input value (scalar range from <tt>0</tt> to <tt>max</tt>).
   * @param  max Maximal input value.
   * @return Output index value (color index).
   */
  int transfer ( int[] input, int max );

  /**
   * Index -&gt; Gr[A] or RGB[A] (reconstruction) function.
   * Typical usage: color mapping using a palette.
   *
   * @param  input Index value (color index).
   * @param  output Output value (scalar range from <tt>0</tt> to <tt>max</tt>).
   * @param  max Maximal output value.
   */
  void transferInv ( int input, int[] output, int max );

  /**
   * Gr[A] or RGB[A] -&gt; index (classification) function.
   * Typical usage: color quantization.
   *
   * @param  input Input value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   * @return Output index value (color index).
   */
  int transfer ( double[] input );

  /**
   * Index -&gt; Gr[A] or RGB[A] (reconstruction) function.
   * Typical usage: color mapping using a palette.
   *
   * @param  input Index value (color index).
   * @param  output Output value (scalar range from <tt>0.0</tt> to <tt>1.0</tt>).
   */
  void transferInv ( int input, double[] output );

}
