package cz.cuni.jagrlib.iface;

import java.util.Iterator;

import cz.cuni.jagrlib.*;

/**
 * Vector graphics interface (scalable atomic 2D drawing commands).
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.02
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/VectorGraphics.java">VectorGraphics.java</a>
 * @see     VectorCommand
 * @see     VCommand
 */
public interface VectorGraphics extends Property
{
  //==========================================================================
  // general commands:

  /**
   * Interface (re-)initialization (new vector-command stream will be produced).
   */
  void open ();

  /**
   * Close the current vector-command stream.
   */
  void close ();

  //==========================================================================
  // vector graphics producer:

  /**
   * NewPath command.
   *
   * @see    cz.cuni.jagrlib.VCommand.NewPath
   */
  void newPath ();

  /**
   * MoveTo(x,y) command.
   *
   * @param  x X coordinate.
   * @param  y Y coordinate.
   * @see    cz.cuni.jagrlib.VCommand.MoveTo
   */
  void moveTo ( double x, double y );

  /**
   * LineTo(x,y) command.
   *
   * @param  x X coordinate.
   * @param  y Y coordinate.
   * @see    cz.cuni.jagrlib.VCommand.LineTo
   */
  void lineTo ( double x, double y );

  /**
   * ArcTo(x,y,r,alpha,beta) command.
   *
   * @param  x X coordinate of arc center.
   * @param  y Y coordinate of arc center.
   * @param  r Arc radius.
   * @param  alpha Starting angle in degrees.
   * @param  beta Finishing angle in degrees.
   * @see    cz.cuni.jagrlib.VCommand.ArcTo
   */
  void arcTo ( double x, double y, double r, double alpha, double beta );

  /**
   * RArcTo(x,y,r,alpha,beta) command.
   *
   * @param  x X coordinate of arc center.
   * @param  y Y coordinate of arc center.
   * @param  r Arc radius.
   * @param  alpha Starting angle in degrees.
   * @param  beta Finishing angle in degrees.
   * @see    cz.cuni.jagrlib.VCommand.RArcTo
   */
  void rArcTo ( double x, double y, double r, double alpha, double beta );

  /**
   * CubicCurveTo(x1,y1,x2,y2,x3,y3) command.
   *
   * @param  x1 X coordinate of the 2nd control point.
   * @param  y1 Y coordinate of the 2nd control point.
   * @param  x2 X coordinate of the 3rd control point.
   * @param  y2 Y coordinate of the 3rd control point.
   * @param  x3 X coordinate of the 4th control point.
   * @param  y3 Y coordinate of the 4th control point.
   * @see    cz.cuni.jagrlib.VCommand.CubicCurveTo
   */
  void cubicCurveTo ( double x1, double y1, double x2, double y2, double x3, double y3 );

  /**
   * ClosePath command.
   *
   * @see    cz.cuni.jagrlib.VCommand.ClosePath
   */
  void closePath ();

  /**
   * Stroke command.
   *
   * @see    cz.cuni.jagrlib.VCommand.Stroke
   */
  void stroke ();

  /**
   * Fill command.
   *
   * @see    cz.cuni.jagrlib.VCommand.Fill
   */
  void fill ();

  /**
   * Clip command.
   *
   * @see    cz.cuni.jagrlib.VCommand.Clip
   */
  void clip ();

  /**
   * PushPath command.
   *
   * @see    cz.cuni.jagrlib.VCommand.PushPath
   */
  void pushPath ();

  /**
   * PopPath command.
   *
   * @see    cz.cuni.jagrlib.VCommand.PopPath
   */
  void popPath ();

  /**
   * SetColor command.
   *
   * @param  color New actual color in <b>Gr[A]</b> or <b>RGB[A]</b> format.
   * @see    cz.cuni.jagrlib.VCommand.SetColor
   */
  void setColor ( double[] color );

  /**
   * SetLineWidth command.
   *
   * @param  width New line width.
   * @see    cz.cuni.jagrlib.VCommand.SetLineWidth
   */
  void setLineWidth ( double width );

  /**
   * SetLineStyle command.
   *
   * @param  style New line style: use {@code VectorCommand.LINE_*} constant.
   * @see    cz.cuni.jagrlib.VCommand.SetLineStyle
   * @see    VectorCommand#LINE_SOLID
   */
  void setLineStyle ( int style );

  /**
   * SetFillStyle command.
   *
   * @param  style New fill style: use {@code VectorCommand.FILL_*} constant.
   * @see    cz.cuni.jagrlib.VCommand.SetFillStyle
   * @see    VectorCommand#FILL_SOLID
   */
  void setFillStyle ( int style );

  /**
   * SetLineCallback command.
   *
   * @param  callback Line callback object.
   * @see    cz.cuni.jagrlib.VCommand.SetLineStyle
   * @see    cz.cuni.jagrlib.VCommand.SetLineCallback
   * @see    VectorCommand#LINE_CALLBACK
   */
  void setLineCallback ( StrokeCallback callback );

  /**
   * SetFillCallback command.
   *
   * @param  callback File callback object.
   * @see    cz.cuni.jagrlib.VCommand.SetFillStyle
   * @see    cz.cuni.jagrlib.VCommand.SetFillCallback
   * @see    VectorCommand#FILL_CALLBACK
   */
  void setFillCallback ( FillCallback callback );

  /**
   * Creates a new font and places it in <b>slot A</b> (actual, working font).
   *
   * @param  name Font name.
   * @param  style Font style - used if exact font match was not successful
   *               (bitwise OR of {@code TextRender.FONT_*} constants).
   * @param  size Font size (distance between two regularly spaced text lines).
   * @return Result code (one of {@code TextRender.CREATE_*} constants).
   * @see    TextRender#FONT_BOLD
   * @see    TextRender#CREATE_EXACT
   * @see    cz.cuni.jagrlib.VCommand.CreateFont
   */
  int createFont ( String name, int style, double size );

  /**
   * Changes the actual font's orientation.
   *
   * @param  angle New font orientation (in degrees, counterclockwise).
   * @return The old orientation.
   * @see    cz.cuni.jagrlib.VCommand.FontOrientation
   */
  double fontOrientation ( double angle );

  /**
   * Assigns the actual font (<b>slot A</b>) from the given <b>slot</b>.
   *
   * @param  slot Slot number (between {@code 0} and {@code TextRender.MAX_SLOT-1}).
   * @see    TextRender#MAX_SLOT
   * @see    cz.cuni.jagrlib.VCommand.GetFont
   */
  void getFont ( int slot );

  /**
   * Puts the actual font (<b>slot A</b>) into the given <b>slot</b>.
   *
   * @param  slot Slot number (between {@code 0} and {@code TextRender.MAX_SLOT-1}).
   * @see    TextRender#MAX_SLOT
   * @see    cz.cuni.jagrlib.VCommand.PutFont
   */
  void putFont ( int slot );

  /**
   * Returns properties (flags) of the actual font (<b>slot A</b>).
   *
   * @return Bitwise OR of {@code TextRender.PROP_*} constants.
   * @see    TextRender#PROP_STROKE
   */
  int fontProperties ();

  /**
   * Generates path of the given text string and <u>appends</u> it to the current path.
   * Uses the <b>current point</b> as starting reference point (<b>current point</b>
   * itself will be updated by this command).
   * <p>Does nothing if the actual font has neither {@link TextRender#PROP_STROKE} nor
   * {@link TextRender#PROP_FILL} capability.
   *
   * @param  text The string to be drawn.
   * @see    cz.cuni.jagrlib.VCommand.TextPath
   * @see    #textExtent
   * @see    #drawText
   * @see    cz.cuni.jagrlib.VCommand.MoveTo
   * @see    #moveTo
   */
  void textPath ( String text );

  /**
   * Draws the given text. Starting reference point is taken from the <b>current point</b>.
   * If {@code ext} parameter is not {@code null}, the routine fills
   * the given {@link TextRender.TextExtent} object.
   * <p>Works with both vector and bitmap fonts.
   *
   * @param  text The string to be drawn.
   * @param  ext The {@link TextRender.TextExtent} object to be filled (or {@code null}).
   * @see    TextRender.TextExtent
   * @see    #textExtent
   * @see    cz.cuni.jagrlib.VCommand.DrawText
   */
  void drawText ( String text, TextRender.TextExtent ext );

  /**
   * Computes the <b>extent</b> of the given text string (without any output change).
   * Starting reference point is taken from the <b>current point</b>.
   *
   * @param  text The string to be drawn.
   * @param  ext The {@link TextRender.TextExtent} object to be filled (or {@code null}).
   * @return The filled {@link TextRender.TextExtent} object.
   * @see    TextRender.TextExtent
   * @see    #drawText
   */
  TextRender.TextExtent textExtent ( String text, TextRender.TextExtent ext );

  /**
   * Returns the <b>kerning</b> (space correction) which should be used between two given
   * characters.
   *
   * @param  c1 First character.
   * @param  c2 Second character.
   * @param  ext The {@link TextRender.TextExtent} object to be filled (or {@code null}).
   * @return The filled {@link TextRender.TextExtent} object (only {@code dx} and
   *         {@code dy} are used).
   * @see    TextRender.TextExtent
   * @see    #textExtent
   */
  TextRender.TextExtent getKern ( char c1, char c2, TextRender.TextExtent ext );

  //==========================================================================
  // vector graphics consumer:

  /**
   * Single command processing.
   *
   * @see    #commands(VectorCommand[])
   * @see    #commands(Iterator)
   */
  void command ( VectorCommand c );

  /**
   * Multiple command processing.
   *
   * @see    #command(VectorCommand)
   * @see    #commands(Iterator)
   */
  void commands ( VectorCommand[] c );
  // !!! TODO: what about a collection?

  /**
   * Multiple command processing.
   *
   * @see    #command(VectorCommand)
   * @see    #commands(VectorCommand[])
   */
  void commands ( Iterator< VectorCommand > c );

}
