package cz.cuni.jagrlib.piece;

import java.util.Arrays;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Can add various types of noise to the raster image.
 *
 * @version 0.25 $Rev: 385 $ $Date: 2007-01-07 01:47:21 +0100 (ne, 07 I 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/AdditiveNoise.java">AdditiveNoise.java</a>
 */
public class AdditiveNoise extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Implemented noise types. */
  static enum NoiseType
  {
    /** Gaussian additive noise. */
    GAUSSIAN,
    /** Additive "Pyramid" noise (probab. density = symmetric pyramid function). */
    PYRAMID,
    /** Additive "Box" noise (probab. density = box function). */
    BOX,
    /** "Salt & pepper" noise. */
    SALT_PEPPER,
  }

  /** Actual noise type. */
  protected NoiseType type    = NoiseType.GAUSSIAN;

  /** Distribution variance (radius for non-gaussian distributions). */
  protected double variance   = 0.05;

  /** Probability of minimum-value collapse ("pepper"). */
  protected double pepperProb = 0.02;

  /** Probability of maximum-value collapse ("salt"). */
  protected double saltProb   = 0.02;

  /** Do result clamping? */
  protected boolean clamping  = true;

  /** Set negative value for randomize. */
  protected int seed          = -1;

  /** Number of repetitions for pseudo-gaussian generator (height value means better quality). */
  protected int gaussRep      = 8;

  //==========================================================================
  //  Support:


  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image filtering.
   *
   * @param  t The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int t )
  {
    final RasterGraphics input  = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    final RasterGraphics output = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );

      // output image:
    int width       = input.getWidth();
    int height      = input.getHeight();
    int iMode       = input.getMode();
    int bands       = input.getBands();
    int oMode       = iMode;
    boolean readRGB = false;

    if ( iMode == RasterGraphics.MODE_COLORMAP )
    {
      oMode   = RasterGraphics.MODE_RGB;
      bands   = 3;
      readRGB = true;
    }

    output.init( width, height, oMode, bands );

      // random generator:
    RandomJames rnd;
    if ( seed >= 0 )
      rnd = new RandomJames( seed >>> 15, seed & 0x7fff );
    else
    {
      rnd = new RandomJames();
      rnd.randomize();
    }

    int x, y, b;
    double[] pixel = new double[ bands ];

    for ( y = 0; y < height; y++ )
      for ( x = 0; x < width; x++ )
      {
        if ( readRGB )
          input.getRGB( x, y, pixel );
        else
          input.getPixel( x, y, pixel );

        switch ( type )
        {
          case GAUSSIAN:
            for ( b = 0; b < bands; b++ )
            {
              pixel[b] += rnd.normalNumber( 0.0, variance, gaussRep );
              if ( clamping )
                pixel[b] = Formula.clamp( pixel[b], 0.0, 1.0 );
            }
            break;

          case PYRAMID:
            for ( b = 0; b < bands; b++ )
            {
              pixel[b] += (rnd.uniformNumber() + rnd.uniformNumber() - 1.0) * variance;
              if ( clamping )
                pixel[b] = Formula.clamp( pixel[b], 0.0, 1.0 );
            }
            break;

          case BOX:
            for ( b = 0; b < bands; b++ )
            {
              pixel[b] += (rnd.uniformNumber() - 0.5) * (variance + variance);
              if ( clamping )
                pixel[b] = Formula.clamp( pixel[b], 0.0, 1.0 );
            }
            break;

          case SALT_PEPPER:
          default:
            double r = rnd.uniformNumber();
            if ( r < pepperProb )
              Arrays.fill( pixel, 0.0 );
            else
            if ( r < pepperProb + saltProb )
              Arrays.fill( pixel, 1.0 );
        }

        output.putPixel( x, y, pixel );
      }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String NOISE_TYPE  = "Noise type";
  public final static String VARIANCE    = "Variance";
  public final static String PEPPER_PROB = "Pepper prob";
  public final static String SALT_PROB   = "Salt prob";
  public final static String CLAMPING    = "Clamping";
  public final static String SEED        = "Seed";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // NOISE_TYPE:
    if ( key.compareTo(NOISE_TYPE) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
      // VARIANCE:
    if ( key.compareTo(VARIANCE) == 0 )
    {
      variance = doubleProperty( value, variance, 0.0, 100.0 );
      return;
    }
      // PEPPER_PROB:
    if ( key.compareTo(PEPPER_PROB) == 0 )
    {
      pepperProb = doubleProperty( value, pepperProb, 0.0, 1.0 );
      return;
    }
      // SALT_PROB:
    if ( key.compareTo(SALT_PROB) == 0 )
    {
      saltProb = doubleProperty( value, saltProb, 0.0, 1.0 );
      return;
    }
      // CLAMPING:
    if ( key.compareTo(CLAMPING) == 0 )
    {
      clamping = booleanProperty( value, clamping );
      return;
    }
      // SEED:
    if ( key.compareTo(SEED) == 0 )
    {
      seed = intProperty( value, seed );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(NOISE_TYPE) == 0 )
      return type.ordinal();
    if ( key.compareTo(VARIANCE) == 0 )
      return variance;
    if ( key.compareTo(PEPPER_PROB) == 0 )
      return pepperProb;
    if ( key.compareTo(SALT_PROB) == 0 )
      return saltProb;
    if ( key.compareTo(CLAMPING) == 0 )
      return clamping;
    if ( key.compareTo(SEED) == 0 )
      return seed;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "AdditiveNoiseFilter";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Adds noise to raster image.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"Trigger" );
      t.newOutputPlug( "image1", IFACE+"RasterGraphics" );
      t.newOutputPlug( "image2", IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( NOISE_TYPE, TYPE_INTEGER, "Noise type", true );
        t.propDefault( NoiseType.GAUSSIAN.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Gaussian noise",      NoiseType.GAUSSIAN.ordinal(),    "Additive Gaussian noise" );
        t.propEnum( "Pyramid noise",       NoiseType.PYRAMID.ordinal(),     "Additive noise with pyramid distribution" );
        t.propEnum( "Box noise",           NoiseType.BOX.ordinal(),         "Additive noise with box distribution" );
        t.propEnum( "Salt & Pepper noise", NoiseType.SALT_PEPPER.ordinal(), "Salt & pepper noise" );
      t.propEnd();
      t.propBegin( VARIANCE, TYPE_DOUBLE, "Distribution variance (radius for non-gaussian distributions)", true );
        t.propDefault( 0.05 );
        t.propBounds(  0.0, 100.0 );
      t.propEnd();
      t.propBegin( PEPPER_PROB, TYPE_DOUBLE, "Probability of minimum-value collapse (Pepper)", true );
        t.propDefault( 0.02 );
        t.propBounds(  0.0, 1.0 );
      t.propEnd();
      t.propBegin( SALT_PROB, TYPE_DOUBLE, "Probability of maximum-value collapse (Salt)", true );
        t.propDefault( 0.02 );
        t.propBounds(  0.0, 1.0 );
      t.propEnd();
      t.propBegin( CLAMPING, TYPE_BOOLEAN, "Clamp result values?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( SEED, TYPE_INTEGER, "Random generator seed (-1 for randomize)", true );
        t.propDefault( -1 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
