package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Generator of alpha-channel test patterns.
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.07
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/AlphaGenerator.java">AlphaGenerator.java</a>
 */
public class AlphaGenerator extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Alpha-mask width in pixels. */
  protected int width  = 256;

  /** Alpha-mask height in pixels. */
  protected int height = 256;

  /** Pattern period in pixels. */
  protected int period = 64;

  //==========================================================================
  //  Interface Property:

  /** Alpha-mask width in pixels. */
  public final static String WIDTH  = "Width";

  /** Alpha-mask height in pixels. */
  public final static String HEIGHT = "Height";

  /** Pattern period in pixels. */
  public final static String PERIOD = "Period";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // WIDTH:
    if ( key.compareTo(WIDTH) == 0 )
    {
      width = intProperty( value, width );
      return;
    }
      // HEIGHT:
    if ( key.compareTo(HEIGHT) == 0 )
    {
      height = intProperty( value, height );
      return;
    }
      // PERIOD:
    if ( key.compareTo(PERIOD) == 0 )
    {
      period = intProperty( value, period );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(WIDTH) == 0 )
      return width;
    if ( key.compareTo(HEIGHT) == 0 )
      return height;
    if ( key.compareTo(PERIOD) == 0 )
      return period;
    return null;
  }

  //==========================================================================
  //  Interface Trigger:

  /** Transition from top (opaque) to bottom (transparent). */
  public final static int PATTERN_TRANS_DOWN  = 0;

  /** Transition from bottom (opaque) to up (transparent). */
  public final static int PATTERN_TRANS_UP    = 1;

  /** Transition from left (opaque) to right (transparent). */
  public final static int PATTERN_TRANS_RIGHT = 2;

  /** Transition from right (opaque) to left (transparent). */
  public final static int PATTERN_TRANS_LEFT  = 3;

  /** "sin(x) * cos(y)" pattern. */
  public final static int PATTERN_WAVES       = 4;

  /**
   * Triggers the action.
   *
   * @param  type The action type (<tt>PATTERN_*</tt> constants).
   * @return <tt>true</tt> if the action was successful.
   */
  public boolean fire ( int type )
  {
    switch ( type )
    {
      case PATTERN_TRANS_DOWN:
      {
        AlphaMask out;
        try
        {
          out = (AlphaMask)getInterface( PL_OUTPUT, IFACE+"AlphaMask" );
          int y;
          double op = 1.0;
          double delta = (period > 0) ? -1.0 / (period-1) : -1.0 / (height-1);
          for ( y = 0; y < height; y++ )
          {
            out.setHLine( 0, width, y, op );
            if ( (op += delta) < 0.0 ) op = 1.0;
          }
        }
        catch ( JaGrLibException e )
        {
          return false;
        }
        return true;
      }

      case PATTERN_TRANS_UP:
      {
        AlphaMask out;
        try
        {
          out = (AlphaMask)getInterface( PL_OUTPUT, IFACE+"AlphaMask" );
          int y;
          double op = 1.0;
          double delta = (period > 0) ? -1.0 / (period-1) : -1.0 / (height-1);
          for ( y = height; y-- > 0; )
          {
            out.setHLine( 0, width, y, op );
            if ( (op += delta) < 0.0 ) op = 1.0;
          }
        }
        catch ( JaGrLibException e )
        {
          return false;
        }
        return true;
      }

      case PATTERN_TRANS_RIGHT:
      {
        AlphaMask out;
        try
        {
          out = (AlphaMask)getInterface( PL_OUTPUT, IFACE+"AlphaMask" );
          int x, y;
          double op;
          double delta = (period > 0) ? -1.0 / (period-1) : -1.0 / (width-1);
          for ( y = 0; y < height; y++ )
          {
            op = 1.0;
            for ( x = 0; x < width; x++ )
            {
              out.putPixel( x, y, op );
              if ( (op += delta) < 0.0 ) op = 1.0;
            }
          }
        }
        catch ( JaGrLibException e )
        {
          return false;
        }
        return true;
      }

      case PATTERN_TRANS_LEFT:
      {
        AlphaMask out;
        try
        {
          out = (AlphaMask)getInterface( PL_OUTPUT, IFACE+"AlphaMask" );
          int x, y;
          double op;
          double delta = (period > 0) ? -1.0 / (period-1) : -1.0 / (width-1);
          for ( y = 0; y < height; y++ )
          {
            op = 1.0;
            for ( x = width; x-- > 0; )
            {
              out.putPixel( x, y, op );
              if ( (op += delta) < 0.0 ) op = 1.0;
            }
          }
        }
        catch ( JaGrLibException e )
        {
          return false;
        }
        return true;
      }

      case PATTERN_WAVES:
      {
        AlphaMask out;
        try
        {
          out = (AlphaMask)getInterface( PL_OUTPUT, IFACE+"AlphaMask" );
          int x, y;
          double cosy;
          double per = 2.0 * Math.PI / period;
          for ( y = 0; y < height; y++ )
          {
            cosy = Math.cos( y * per ) * 0.5;
            for ( x = 0; x < width; x++ )
              out.putPixel( x, y, Math.sin( x * per ) * cosy + 0.5 );
          }
        }
        catch ( JaGrLibException e )
        {
          return false;
        }
        return true;
      }
    }
    return false;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "AlphaSampleGenerator";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "PropertyAndTriggerToAlphaMask";

  /** Object description. */
  private final static String DESCRIPTION = "Draws various alpha-channel test patterns.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DRAW + "." + C_SAMPLE + "." + C_ALPHA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,   IFACE+"Property" );
      t.newInputPlug(  PL_TRIGGER, IFACE+"Trigger" );
      t.newOutputPlug( PL_OUTPUT,  IFACE+"AlphaMask" );
        // Property:
      t.propBegin( WIDTH, TYPE_INTEGER, "Bitmap width in pixels", true );
        t.propDefault( "256" );
      t.propEnd();
      t.propBegin( HEIGHT, TYPE_INTEGER, "Bitmap height in pixels", true );
        t.propDefault( "256" );
      t.propEnd();
      t.propBegin( PERIOD, TYPE_INTEGER, "Pattern period in pixels", true );
        t.propDefault( "64" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
