package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Alpha-mask represented by simple rectangular matrix.
 *
 * @version 0.24 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.06
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/AlphaMatrix.java">AlphaMatrix.java</a>
 * @todo    All <tt>binaryOp</tt> operations.
 */
public class AlphaMatrix extends Piece implements AlphaMask
{
  /**
   * Special Enumerator for <tt>AlphaMatrix</tt>.
   *
   * @version 0.23 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
   * @since   0.06
   */
  public class AlphaMatrixEnumerator implements BitMaskEnumerator
  {
    /** Actual X coordinate. */
    protected int x;

    /** Actual Y coordinate. */
    protected int y;

    /** Default constructor (initializes the enumerator). */
    AlphaMatrixEnumerator ()
    {
      reset();
    }

    /**
     * (re-)initialization of the enumerator.
     */
    public void reset ()
    {
      x = y = 0;
    }

    /**
     * Finds the next non-transparent pixel.
     * Updates <tt>x</tt> and <tt>y</tt>.
     * @return <tt>true</tt> if there are no more opaque pixels.
     */
    protected boolean findNext ()
    {
      if ( data == null || y >= yMax - yMin ) return true;

      do
      {
        if ( x >= xMax - xMin )
        {
          if ( ++y >= yMax - yMin ) return true;
          x = 0;
        }
        if ( data[y][x] != ALPHA_ZERO ) return false;
        x++;
      } while ( true );
    }

    /**
     * Returns the next <tt>BitMask</tt> (<tt>AlphaMask</tt>) pixel.
     *
     * @param  pre Pre-allocated <tt>BitMaskEnumerator.Pixel</tt> instance (non-mandatory).
     * @return The next opaque <tt>BitMaskEnumerator.Pixel</tt> or <tt>null</tt>.
     * @see    #nextHLine
     */
    public BitMaskEnumerator.Pixel nextPixel ( BitMaskEnumerator.Pixel pre )
    {
      if ( data == null || findNext() ) return null;

      if ( pre == null )
        pre = new BitMaskEnumerator.Pixel();

      pre.x = x + xMin;
      pre.y = y + yMin;
      pre.alpha = (data[y][x++] - (double)ALPHA_ZERO) / ALPHA_RANGE;

      return pre;
    }

    /**
     * Returns the next <tt>BitMask</tt> (<tt>AlphaMask</tt>) H-line.
     *
     * @param  pre Pre-allocated <tt>BitMaskEnumerator.HLine</tt> instance (non-mandatory).
     * @return The next opaque <tt>BitMaskEnumerator.HLine</tt> or <tt>null</tt>.
     * @see    #nextPixel
     */
    public BitMaskEnumerator.HLine nextHLine ( BitMaskEnumerator.HLine pre )
    {
      if ( data == null || findNext() ) return null;

      if ( pre == null )
        pre = new BitMaskEnumerator.HLine();

      pre.x1    = x + xMin;
      pre.y     = y + yMin;
      byte alp  = data[y][x++];
      pre.alpha = (alp - (double)ALPHA_ZERO) / ALPHA_RANGE;
      while ( x < xMax - xMin && data[y][x] == alp ) x++;
      pre.x2    = x + xMin;

      return pre;
    }

  }

  //==========================================================================
  //  Data:

  /** Minimum alpha-value (total transparency). */
  protected final byte ALPHA_ZERO = -128;

  /** Maximum alpha-value (total opaqueness). */
  protected final byte ALPHA_MAX = 127;

  /** Alpha-value range. */
  protected final int ALPHA_RANGE = (int)ALPHA_MAX - (int)ALPHA_ZERO;

  /** Width of the domain in pixels. */
  protected int setWidth = 0;

  /** Height of the domain in pixels. */
  protected int setHeight = 0;

  /** Minimum x-coordinate. */
  protected int xMin = 0;

  /** Maximum x-coordinate + 1. */
  protected int xMax = 0;

  /** Minimum scanline coordinate. */
  protected int yMin = 0;

  /** Maximum scanline coordinate + 1. */
  protected int yMax = 0;

  /** 2D array of alpha-values. */
  protected byte data[][];

  //--------------------------------------------------------------------------
  //  State variables:

  /** Current alpha-value (for <tt>BitMask</tt>-inherited routines. */
  protected byte currentAlpha = ALPHA_MAX;

  /** Actual binary set operation (not used yet). */
  protected int binaryOp = BinaryOperation.BIN_OP_SET;

  /** Actual alpha-channel operation (not used yet). */
  protected int alphaOp = AlphaOperation.ALPHA_B_OVER_A;

  //--------------------------------------------------------------------------
  //  Support routines:

  /** (Re-)initializes the whole &quot;data&quot; matrix. */
  protected void initMatrix ()
  {
    int width  = xMax - xMin;
    int height = yMax - yMin;
    if ( data == null || data.length < width || data[0].length < height )
      data = new byte[ height ][ width ];

    int x, y;
    for ( y = 0; y < height; y++ )
      for ( x = 0; x < width; x++ )
        data[y][x] = ALPHA_ZERO;
  }

  /**
   * Asserts that the given point [x,y] is included in the set.
   *
   * @param  x X coordinate to be included.
   * @param  y Y coordinate to be included.
   */
  protected void assertXY ( int x, int y )
  {
    if ( x >= xMin && x < xMax && y >= yMin && y < yMax ) return;

    int width  = xMax - xMin;
    int height = yMax - yMin;
    byte nd[][];
    int i, j, ext;
    if ( x < xMin )                       // extend horizontally to the left
    {
      nd  = new byte[ height ][ xMax - x ];
      ext = xMin - x;
      for ( i = 0; i < height; i++ )
      {
        for ( j = 0; j < ext; ) nd[i][j++] = ALPHA_ZERO;
        System.arraycopy( data[i], 0, nd[i], ext, width );
      }
      data  = nd;
      width = xMax - (xMin = x);
    }
    else
      if ( x >= xMax )                    // extend horizontally to the right
      {
        ext = x + 1 - xMin;
        nd  = new byte[ height ][ ext ];
        for ( i = 0; i < height; i++ )
        {
          System.arraycopy( data[i], 0, nd[i], 0, width );
          for ( j = width; j < ext; ) nd[i][j++] = ALPHA_ZERO;
        }
        data  = nd;
        width = ext;
        xMax  = x + 1;
      }

    if ( y < yMin )                       // extend vertically to the up
    {
      nd  = new byte[ yMax - y ][ width ];
      ext = yMin - y;
      for ( i = 0; i < ext; i++ )
        for ( j = 0; j < width; ) nd[i][j++] = ALPHA_ZERO;
      for ( j = 0; i < yMax - y; i++ )
        System.arraycopy( data[j++], 0, nd[i], 0, width );
      data = nd;
      yMin = y;
    }
    else
      if ( y >= yMax )                    // extend vertically to the bottom
      {
        ext = y + 1 - yMin;
        nd  = new byte[ ext ][ width ];
        for ( i = 0; i < height; i++ )
          System.arraycopy( data[i], 0, nd[i], 0, width );
        for ( ; i < ext; i++ )
          for ( j = 0; j < width; ) nd[i][j++] = ALPHA_ZERO;
        data = nd;
        yMax = y + 1;
      }
  }

  /**
   * Realizes the alpha-channel composition.
   * Uses the <b>current alpha-value</b>.
   *
   * @param  A Original (old) alpha-value.
   * @param  B Assigned (new) alpha-value.
   * @return The compound (result) alpha-value.
   * @todo   Other operations than &quot;B OVER A&quot; and &quot;A OVER B&quot; are
   *         not implemented yet.
   */
  protected final byte composeAlpha ( int A, int B )
  {
    A -= ALPHA_ZERO;
    B -= ALPHA_ZERO;
    return( (byte)(A + B - (A * B) / ALPHA_RANGE + ALPHA_ZERO) );
  }

  //==========================================================================
  //  Interface BitMask:

  /**
   * <tt>AlphaMatrix</tt> (re-)initialization.
   *
   * @param  x Width of a new bitmask.
   * @param  y Height of a new bitmask.
   */
  public void init ( int x, int y )
  {
    if ( x < 1 ) x = 1;
    xMin = 0;
    xMax = x;
    if ( y < 1 ) y = 1;
    yMin = 0;
    yMax = y;
    initMatrix();
  }

  /**
   * <tt>AlphaMatrix</tt> re-initialization (keeps current bitmask size - if the size
   * is relevant).
   */
  public void init ()
  {
    if ( xMax <= xMin || yMax <= yMin )
    {
      xMin = yMin = 0;
      xMax = yMax = 128;
    }
    initMatrix();
  }

  /**
   * Sets the binary set operation (used in all <tt>set*()</tt> methods).
   *
   * @param  _op Binary set operation (<tt>BinaryOperation.BIN_OP_*</tt> constant).
   * @return The old binary set operation.
   * @see    cz.cuni.jagrlib.BinaryOperation Binary operations
   * @todo   All <tt>binaryOp</tt> operations.
   */
  public int setOperation ( int _op )
  {
    int old  = binaryOp;
    binaryOp = _op;
    return old;
  }

  /**
   * Sets the alpha-channel operation (used in all <tt>set*()</tt> methods).
   *
   * @param  _op Alpha-channel operation (<tt>AlphaOperation.ALPHA_*</tt> constant).
   * @return The old alpha-channel operation.
   * @see    cz.cuni.jagrlib.AlphaOperation Alpha-channel operations
   */
  public int setAlphaOperation ( int _op )
  {
    int old = alphaOp;
    alphaOp = _op;
    return old;
  }

  /**
   * Sets the given pixel's value to <tt>true</tt> = &quot;painted&quot;.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   */
  public void setPixel ( int x, int y )
  {
    if ( data == null || currentAlpha == ALPHA_ZERO &&
         (x < xMin || x >= xMax || y < yMin || y >= yMax) ) return;

    assertXY( x, y );
    data[y-yMin][x-xMin] = composeAlpha( data[y-yMin][x-xMin], currentAlpha );
  }

  /**
   * Sets the given pixel's value.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  value New pixel's value.
   */
  public void putPixel ( int x, int y, boolean value )
  {
    if ( data == null || !value &&
         (x < xMin || x >= xMax || y < yMin || y >= yMax) ) return;

    assertXY( x, y );
    data[y-yMin][x-xMin] = value ? currentAlpha : ALPHA_ZERO;
  }

  /**
   * Returns the pixel's value.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value (<tt>false</tt> for off-canvas access).
   */
  public boolean getPixel ( int x, int y )
  {
    if ( data == null || x < xMin || x >= xMax || y < yMin || y >= yMax ) return false;

    return( data[y-yMin][x-xMin] != ALPHA_ZERO );
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y].
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   */
  public void setHLine ( int x1, int x2, int y )
  {
    if ( data == null ) return;

    if ( x1 > x2 )
    {
      int tmp = x1;
      x1      = x2;
      x2      = tmp;
    }
    if ( x1 == x2 || currentAlpha == ALPHA_ZERO &&
         (y < yMin || y >= yMax) ) return;

    assertXY( x1,   y );
    assertXY( x2-1, y );
    y  -= yMin;
    x1 -= xMin;
    x2 -= xMin;
    while ( x1 < x2 )
    {
      data[y][x1] = composeAlpha( data[y][x1], currentAlpha );
      x1++;
    }
  }

  /**
   * Sets the horizontal line'a values.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  values New pixels' values.
   */
  public void putHLine ( int x1, int x2, int y, boolean[] values )
  {
    if ( data == null || values == null ) return;

    if ( x1 > x2 )
    {
      int tmp = x1;
      x1      = x2;
      x2      = tmp;
    }
    int offset = 0;
    if ( x1 < xMin )
    {
      offset = xMin - x1;
      x1 = xMin;
    }
    if ( x1 == x2 ) return;

    assertXY( x1,   y );
    assertXY( x2-1, y );
    y  -= yMin;
    x1 -= xMin;
    x2 -= xMin;
    while ( x1 < x2 )
      data[y][x1++] = values[offset++] ? ALPHA_MAX : ALPHA_ZERO;
  }

  /**
   * Returns the pixels' values for the given horizontal line.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  values Returned values in <tt>boolean[]</tt> array (<tt>false</tt> for
   *                off-canvas access).
   */
  public void getHLine ( int x1, int x2, int y, boolean[] values )
  {
    if ( values == null ) return;

    if ( x1 > x2 )
    {
      int tmp = x1;
      x1      = x2;
      x2      = tmp;
    }
    int offset = 0;

      // empty prolog:
    while ( x1 < xMin )
    {
      values[offset] = false;
      x1++;
      if ( ++offset >= values.length ) return;
    }

      // regular data:
    if ( data != null && y >= yMin && y < yMax )
    {
      y -= yMin;
      while ( x1 < x2 && x1 < xMax )
        values[offset++] = (data[y][x1++-xMin] != ALPHA_ZERO);
    }

      // empty epilog:
    while ( offset < values.length )
      values[offset++] = false;
  }

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1].
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   */
  public void setRectangle ( int x1, int y1, int x2, int y2 )
  {
    if ( data == null ) return;

    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1  = x2;
      x2  = tmp;
    }
    if ( y1 > y2 )
    {
      tmp = y1;
      y1  = y2;
      y2  = tmp;
    }
    if ( x1 == x2 || y1 == y2 ) return;

    do
      setHLine( x1, x2, y1++ );
    while ( y1 < y2 );
  }

  /**
   * Sets the rectangle's values.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  values New pixels' values.
   */
  public void putRectangle ( int x1, int y1, int x2, int y2, boolean[][] values )
  {
    if ( data == null || values == null ) return;

    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1  = x2;
      x2  = tmp;
    }
    if ( y1 > y2 )
    {
      tmp = y1;
      y1  = y2;
      y2  = tmp;
    }
    if ( x1 == x2 || y1 == y2 ) return;

    while ( y1 < y2 )
    {
      putHLine( x1, x2, y1, values[y1-yMin] );
      y1++;
    }
  }

  /**
   * Returns the pixels' values for the given rectangle.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  values Returned values in <tt>boolean[][]</tt> array (<tt>false</tt> for
   *                off-canvas access).
   */
  public void getRectangle ( int x1, int y1, int x2, int y2, boolean[][] values )
  {
    if ( values == null ) return;

    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1  = x2;
      x2  = tmp;
    }
    if ( y1 > y2 )
    {
      tmp = y1;
      y1  = y2;
      y2  = tmp;
    }
    if ( x1 == x2 || y1 == y2 ) return;

    int offset = 0;
    int x;

      // empty prolog:
    while ( y1 < yMin )
    {
      for ( x = 0; x < values[offset].length; )
        values[offset][x++] = false;
      y1++;
      if ( ++offset >= values.length ) return;
    }
    if ( y1 >= y2 ) return;

      // regular data:
    while ( y1 < y2 && y1 < yMax )
    {
      getHLine( x1, x2, y1++, values[offset++] );
    }

      // empty epilog:
    while ( offset < values.length )
    {
      for ( x = 0; x < values[offset].length; )
        values[offset][x++] = false;
      offset++;
    }
  }

  /**
   * Bit-block-transfer routine: copies the source rectangle [sx,sy,width,height] to
   * the new location [dx,dy]. Binary set operation is ignored.
   *
   * @param  dx X coordinate of the upper left destination corner.
   * @param  dy Y coordinate of the upper left destination corner.
   * @param  sx X coordinate of the upper left source corner.
   * @param  sy Y coordinate of the upper left source corner.
   * @param  width Rectangle width.
   * @param  height Rectangle height.
   */
  public void bitBlt ( int dx, int dy, int sx, int sy, int width, int height )
  {
    int delta;

      // check horizontal boundaries:
    if ( dx < xMin )
    {
      width -= delta = xMin - dx;
      sx    += delta;
      dx     = xMin;
    }
    if ( sx < xMin )
    {
      width -= delta = xMin - sx;
      dx    += delta;
      sx     = xMin;
    }
    if ( dx + width > xMax )
      width -= dx + width - xMax;
    if ( sx + width > xMax )
      width -= sx + width - xMax;

      // check vertical boundaries:
    if ( dy < yMin )
    {
      height -= delta = yMin - dy;
      sy     += delta;
      dy      = yMin;
    }
    if ( sy < yMin )
    {
      height -= delta = yMin - sy;
      dy     += delta;
      sy      = yMin;
    }
    if ( dy + height > yMax )
      height -= dy + height - yMax;
    if ( sy + height > yMax )
      height -= sy + height - yMax;
    if ( data == null || width <= 0 || height <= 0 ) return;

      // copy the data:
    sx -= xMin;
    sy -= yMin;
    dx -= xMin;
    dy -= yMin;
    if ( dy <= sy )
      while ( height-- > 0 )              // copy one scanline
        System.arraycopy( data[sy++], sx, data[dy++], dx, width );
    else
      while ( height-- > 0 )              // copy one scanline
        System.arraycopy( data[sy+height], sx, data[dy+height], dx, width );
  }

  /**
   * {@link BitMask} enumeration.
   *
   * @return The <tt>BitMaskEnumerator</tt> object instance.
   */
  public BitMaskEnumerator enumerator ()
  {
    return new AlphaMatrixEnumerator();
  }

  //==========================================================================
  //  Interface AlphaMask:

  /**
   * Sets the current default <tt>alpha</tt> - for inherited methods.
   *
   * @param  alpha Opacity factor (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  public void setAlpha ( double alpha )
  {
    if ( alpha <= 0.0 ) currentAlpha = ALPHA_ZERO;
    else
    if ( alpha >= 1.0 ) currentAlpha = ALPHA_MAX;
    else                currentAlpha = (byte)(ALPHA_ZERO + alpha * ALPHA_RANGE);
  }

  /**
   * Sets the given pixel's value to <tt>alpha</tt>.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  alpha Opacity factor (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  public void putPixel ( int x, int y, double alpha )
  {
    if ( data == null ) return;

    if ( alpha < 0.0 ) alpha = 0.0;
    else
    if ( alpha > 1.0 ) alpha = 1.0;

    assertXY( x, y );
    data[y-yMin][x-xMin] = composeAlpha( data[y-yMin][x-xMin],
                                         (int)(ALPHA_ZERO + alpha * ALPHA_RANGE) );
  }

  /**
   * Returns the pixel's value.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel opacity (<tt>0.0</tt> for off-canvas access).
   */
  public double getAlpha ( int x, int y )
  {
    if ( data == null ||
         x < xMin || x >= xMax ||
         y < yMin || y >= yMax ) return 0.0;

    return( (data[y-yMin][x-xMin] - (double)ALPHA_ZERO) / ALPHA_RANGE );
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y].
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  alpha Opacity factor (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  public void setHLine ( int x1, int x2, int y, double alpha )
  {
    if ( data == null ) return;

    if ( x1 > x2 )
    {
      int tmp = x1;
      x1      = x2;
      x2      = tmp;
    }
    if ( x1 == x2 ) return;

    assertXY( x1,   y );
    assertXY( x2-1, y );
    int alp;
    if ( alpha < 0.0 ) alp = ALPHA_ZERO;
    else
    if ( alpha > 1.0 ) alp = ALPHA_MAX;
    else               alp = (int)(ALPHA_ZERO + alpha * ALPHA_RANGE);

    y  -= yMin;
    x1 -= xMin;
    x2 -= xMin;
    while ( x1 < x2 )
    {
      data[y][x1] = composeAlpha( data[y][x1], alp );
      x1++;
    }
  }

  /**
   * Sets the horizontal line's values.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  alpha Opacity factors (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  public void putHLine ( int x1, int x2, int y, double[] alpha )
  {
    if ( data == null || alpha == null ) return;

    if ( x1 > x2 )
    {
      int tmp = x1;
      x1      = x2;
      x2      = tmp;
    }
    if ( x1 == x2 ) return;

    assertXY( x1,   y );
    assertXY( x2-1, y );
    int offset = 0;

    y  -= yMin;
    x1 -= xMin;
    x2 -= xMin;
    while ( x1 < x2 )
    {
      double alp = alpha[offset++];
      if ( alp < 0.0 ) alp = 0.0;
      else
      if ( alp > 1.0 ) alp = 1.0;

      data[y][x1] = composeAlpha( data[y][x1], (int)(ALPHA_ZERO + alp * ALPHA_RANGE) );
      x1++;
    }
  }

  /**
   * Returns the pixels' values for the given horizontal line.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  alpha Returned values in <tt>double[]</tt> array (<tt>0.0</tt> for
   *               off-canvas access).
   */
  public void getHLine ( int x1, int x2, int y, double[] alpha )
  {
    if ( alpha == null ) return;

    int offset = 0;

      // empty prolog:
    while ( x1 < xMin )
    {
      alpha[offset] = 0.0;
      x1++;
      if ( ++offset >= alpha.length ) return;
    }

      // regular data:
    if ( data != null && y >= yMin && y < yMax )
    {
      y -= yMin;
      while ( x1 < x2 && x1 < xMax )
        alpha[offset++] = (data[y][x1++-xMin] - (double)ALPHA_ZERO) / ALPHA_RANGE;
    }

      // empty epilog:
    while ( offset < alpha.length )
      alpha[offset++] = 0.0;
  }

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1].
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  alpha Opacity factor (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  public void setRectangle ( int x1, int y1, int x2, int y2, double alpha )
  {
    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1  = x2;
      x2  = tmp;
    }
    if ( y1 > y2 )
    {
      tmp = y1;
      y1  = y2;
      y2  = tmp;
    }
    if ( x1 == x2 || y1 == y2 ) return;

    do
      setHLine( x1, x2, y1++, alpha );
    while ( y1 < y2 );
  }

  /**
   * Sets the rectangle's values.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  alpha Opacity factors (between <tt>0.0</tt> and <tt>1.0</tt>).
   */
  public void putRectangle ( int x1, int y1, int x2, int y2, double[][] alpha )
  {
    if ( alpha == null ) return;

    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1  = x2;
      x2  = tmp;
    }
    if ( y1 > y2 )
    {
      tmp = y1;
      y1  = y2;
      y2  = tmp;
    }
    if ( x1 == x2 || y1 == y2 ) return;

    tmp = 0;
    while ( y1 < y2 )
      putHLine( x1, x2, y1++, alpha[tmp++] );
  }

  /**
   * Returns the pixels' values for the given rectangle.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  alpha Returned values in <tt>double[][]</tt> array (<tt>0.0</tt> for
   *               off-canvas access).
   */
  public void getRectangle ( int x1, int y1, int x2, int y2, double[][] alpha )
  {
    if ( alpha == null ) return;

    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1  = x2;
      x2  = tmp;
    }
    if ( y1 > y2 )
    {
      tmp = y1;
      y1  = y2;
      y2  = tmp;
    }

    int offset = 0;
    int x;

      // empty prolog:
    while ( y1 < yMin )
    {
      for ( x = 0; x < alpha[offset].length; )
        alpha[offset][x++] = 0.0;
      y1++;
      if ( ++offset >= alpha.length ) return;
    }
    if ( y1 >= y2 ) return;

      // regular data:
    while ( y1 < y2 && y1 < yMax )
    {
      getHLine( x1, x2, y1++, alpha[offset++] );
    }

      // empty epilog:
    while ( offset < alpha.length )
    {
      for ( x = 0; x < alpha[offset].length; )
        alpha[offset][x++] = 0.0;
      offset++;
    }
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // WIDTH:
    if ( key.compareTo(WIDTH) == 0 )
    {
      setWidth = intProperty( value, setWidth );
      return;
    }
      // HEIGHT:
    if ( key.compareTo(HEIGHT) == 0 )
    {
      setHeight = intProperty( value, setHeight );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(WIDTH) == 0 )
      return setWidth;
    if ( key.compareTo(HEIGHT) == 0 )
      return setHeight;
    return null;
  }

  /**
   * Commits all previous changes to the internal object state.
   * Non-mandatory.
   */
  public void commit ()
  {
    init( setWidth, setHeight );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "AlphaMatrix";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "AlphaMask";

  /** Object description. */
  private final static String DESCRIPTION = "2D alpha-mask represented by a matrix.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DATA + "." + C_SET;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"AlphaMask" );
        // Property:
      t.propBegin( WIDTH, TYPE_INTEGER, "Matrix width in pixels", true );
        t.propDefault( "128" );
      t.propEnd();
      t.propBegin( HEIGHT, TYPE_INTEGER, "Matrix height in pixels", true );
        t.propDefault( "128" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
