package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.xml.JGLScript;

/**
 * Ambient light source.
 *
 * @version 0.26 $Rev: 444 $ $Date: 2007-05-05 22:52:07 +0200 (so, 05 V 2007) $ $Author: pepca $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/AmbientLightSource.java">AmbientLightSource.java</a>
 */
public class AmbientLightSource extends DefaultLightSource
{
  //==========================================================================
  //  Data:

  {
    scriptName = "data/ambient.jgl";
  }

  /**
   * Color of the source: {@code 1D} for monochrome source, {@code 3D} for <b>RGB</b>
   * color space, or more for complex spectrum sampled spaces.
   */
  protected double[] lightColor = new double[] { 1.0 };

  //==========================================================================
  //  Support:

  /**
   * Set light source parameters.
   *
   * @param  color Source color/intensity (can be either 1D /for monochrome computing/,
   *               3D /for RGB color system/ or greater /for more complex spectral color
   *               sampling/).
   */
  public void setParams ( double[] color )
  {
      // source color/intensity:
    if ( color == null )
    {
      if ( lightColor == null ||
           lightColor.length != 1 )
        lightColor = new double[ 1 ];

      lightColor[0] = 1.0;
    }
    else
    {
      if ( lightColor == null ||
           lightColor.length != color.length )
        lightColor = new double[ color.length ];

      System.arraycopy( color, 0, lightColor, 0, color.length );
    }
  }

  /** Read data from JGL script. */
  protected void readScript ()
  {
    if ( scriptName == null ) return;

    JGLScript script = new JGLScript();
    Map< String, Object > map = script.interpret( scriptName );
    if ( map != null )
    {
      Object o;

      if ( (o = map.get( KEY_LIGHT ) ) != null &&
           o instanceof double[] )
        lightColor = (double[])o;
    }
  }

  //==========================================================================
  //  Interface LightSource:

  /**
   * Lighting test from the source to the given {@link MicroFacet}.
   *
   * @param  mf The checked {@link MicroFacet} (point on surface of some solid).
   * @return {@code true} if the light source lights on the given point.
   * @see    #getIntensity
   * @see    #getDirection
   */
  public boolean lights ( MicroFacet mf )
  {
    return true;
  }

  /**
   * Returns light intensity coming from the source to the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  color Pre-allocated {@code double[C]} array to hold result - light intensity
   *               coming from the source to the given point ({@code C} can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @return {@code true} if the light source can light on the given point (<b>cannot</b>
   *         replace the {@link #lights} call!).
   * @see    #lights
   * @see    #getDirection
   */
  public boolean getIntensity ( MicroFacet mf, double[] color )
  {
    if ( color.length <= lightColor.length )
      System.arraycopy( lightColor, 0, color, 0, color.length );
    else
      Arrays.fill( color, lightColor[0] );

    return true;
  }

  /**
   * Returns light direction between the source and the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  dir Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *             from the {@link MicroFacet} to the light source.
   * @return {@code false} if the light source is not directional (thus {@code dir}
   *         is irrelevant).
   * @see    #lights
   * @see    #getIntensity
   */
  public boolean getDirection ( MicroFacet mf, double[] dir )
  {
    return false;
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // COLOR:
    if ( key.compareTo(COLOR) == 0 )
    {
      if ( value instanceof double[] )
        setParams( (double[])value );
      return;
    }
      // PARAMS:
    if ( key.compareTo(PARAMS) == 0 )
    {
      scriptName = stringProperty( value, scriptName );
      readScript();
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(COLOR) == 0 )
      return Formula.cloneArray( lightColor );
    if ( key.compareTo(PARAMS) == 0 )
      return scriptName;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Ambient light source";

  /** Object description. */
  private final static String DESCRIPTION = "Ambient light source shining everywhere and " +
                                            "uniformly to all directions.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"LightSource" );
        // Property:
      t.propBegin( COLOR, TYPE_OBJECT, "Light source color (intensity)", false );
      t.propEnd();
      t.propBegin( PARAMS, TYPE_OBJECT, "Light source definition script", true );
        t.propDefault( "data/ambient.jgl" );
        t.propManipulator( "ParamJGLScript" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
