package cz.cuni.jagrlib.piece;

import java.io.IOException;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Arithmetic codec.
 *
 * @author  Jan Kadlec <rrrola@gmail.com>
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/ArithmeticCodec.java">ArithmeticCodec.java</a>
 */
public class ArithmeticCodec extends Piece implements EntropyCodec
{
  /** Codec state: closed/input/output. */
  protected static enum OpenState
  {
    CLOSED,
    READ,
    WRITE
  }

  /** Actual codec state. */
  protected OpenState state = OpenState.CLOSED;

  /** Associated I/O stream. {@code null} if the codec is not opened. */
  protected BitStream stream;

  /** EntropyHistogram to manage contexts. */
  protected EntropyHistogram hist;

  /** Actual position in the original (logical) stream in symbols. */
  protected long position = 0L;

  /** Actual position in the coded stream in bits. */
  protected long rawBits = 0L;

  /** Default alphabet size (for new context creation). */
  protected int defaultMaxSymbol = 255;

  /** Default context Id. */
  protected int defaultCtx = 0;

  /** Increment of frequency for every encoded symbol. */
  protected int increment = 1;

  /** Actual decimation limit. */
  protected int decimationLimit = 32768;

  /** Maximal decimation limit for entropy histograms. */
  protected static final int MAX_DECIMATION_LIMIT = 1 << 24;

  /**
   * Number of bits for maximum value in interval arithmetic.
   * {@link #VAL_MAX} * {@link #decimationLimit} must not overflow!
   */
  protected static final int  VAL_BITS = 39;

  /** One quarter of the maximal interval. */
  protected static final long VAL_1_4  = (1L << (VAL_BITS - 2));

  /** One half of the maximal interval. */
  protected static final long VAL_1_2  = (VAL_1_4 * 2L);

  /** Three quarters of the maximal interval. */
  protected static final long VAL_3_4  = (VAL_1_4 * 3L);

  /** Maximum working arithmetic value. */
  protected static final long VAL_MAX  = (VAL_1_4 * 4L - 1L);

  /** Number of center-doublings (unknown bits to write later). */
  protected long follow = 0L;

  /** Arithmetic encoder/decoder range and value. */
  protected long l, h, v;

  /** Print summary statistics at the end of encoding? */
  protected boolean statistics = false;

  //==========================================================================
  //  Aritmetic Coder implementation - support functions:

  /**
   * Reads a bit from the input stream (returns {@code 0} on error).
   */
  private long getBit ()
  {
    try
    {
      return stream.read( 1 );
    }
    catch ( IOException e )
    {
      return 0L;
    }
  }

  /**
   * Stretches the encoder range to at least 1/4 of it's maximum.
   * Writes a bit for each range doubling.
   */
  private void updateEncoder ()
  throws IOException
  {
    while ( true )
    {
      if ( h < VAL_1_2 )
      {
        stream.write( 0, 1 );
        while ( follow != 0 )
        {
          stream.write( 1, 1 );
          follow--;
        }
      }

      else
        if ( l >= VAL_1_2 )
        {
          stream.write( 1, 1 );
          while ( follow != 0 )
          {
            stream.write( 0, 1 );
            follow--;
          }
          l -= VAL_1_2;
          h -= VAL_1_2;
        }

        else
          if ( l >= VAL_1_4 && h < VAL_3_4 )
          {
            follow++;
            l -= VAL_1_4;
            h -= VAL_1_4;
          }
          else
            break;

      l += l;
      h += h + 1;
      rawBits++;
    }
  }

  /**
   * Stretches the decoder range to at least 1/4 of it's maximum.
   * Reads a bit for each range doubling.
   */
  private void updateDecoder ()
  {
    while ( true )
    {
      if ( h >= VAL_1_2 )
        if ( l >= VAL_1_2 )
        {
          v -= VAL_1_2;
          l -= VAL_1_2;
          h -= VAL_1_2;
        }
        else
          if ( l >= VAL_1_4 && h < VAL_3_4 )
          {
            v -= VAL_1_4;
            l -= VAL_1_4;
            h -= VAL_1_4;
          }
          else
            break;

      v += v + getBit();
      l += l;
      h += h + 1;
      rawBits++;
    }
  }

  //==========================================================================
  //  Interface EntropyCodec:

  public void flush ()
  throws IOException
  {
    if ( state != OpenState.WRITE )
      throw new IOException( "ArithmeticCodec is not opened for writing!" );

    follow++;
    if ( l >= VAL_1_4 )
    {
      stream.write( 1, 1 );
      while ( follow != 0 )
      {
        stream.write( 0, 1 );
        follow--;
      }
    }
    else
    {
      stream.write( 0, 1 );
      while ( follow != 0 )
      {
        stream.write( 1, 1 );
        follow--;
      }
    }
    rawBits++;
  }

  /**
   * Encodes the given symbol (in the current context).
   *
   * @param  s Symbol to encode.
   * @throws IOException
   * @see    #putBits
   * @see    #get
   */
  public void put ( int s )
  throws IOException
  {
    if ( state != OpenState.WRITE )
      throw new IOException( "ArithmeticCodec is not opened for writing!" );

    if ( s < 0 || s >= hist.getSymbols() )
      throw new IOException( "Invalid symbol!" );

    long r = h - l + 1;
    h      = l + (r * hist.left( s + 1 )) / hist.total() - 1;
    l     += (r * hist.left( s )) / hist.total();
    updateEncoder();
    hist.inc( s, increment );
    position++;
  }

  /**
   * Encodes one subinterval from the given {@link WheelOfFortune}.
   * Doesn't use contexts and assigned histograms..
   *
   * @param  wheel Set of intervals (wheel of fortune).
   * @param  s Index of an interval to encode.
   */
  public void put ( WheelOfFortune wheel, int s )
  throws IOException
  {
    if ( state != OpenState.WRITE )
      throw new IOException( "ArithmeticCodec is not opened for writing!" );

    if ( wheel == null ||
         s < 0 ||
         s >= wheel.getSymbols() )
      throw new IOException( "Invalid wheel or symbol!" );

    long r = h - l + 1;
    h      = l + (r * wheel.left( s + 1 )) / wheel.total() - 1;
    l     += (r * wheel.left( s )) / wheel.total();
    updateEncoder();
    position++;
  }

  /**
   * Writes the given "raw" bits (by-passes the entropy encoder).
   *
   * @param  bits Bits to be written ("MSB first" order).
   * @param  length Number of bits to write.
   * @throws IOException
   * @see    #put
   * @see    #getBits
   */
  public void putBits ( long bits, int length )
  throws IOException
  {
    if ( state != OpenState.WRITE )
      throw new IOException( "ArithmeticCodec is not opened for writing!" );

    if ( length != 0 )
    {
      long r = h - l + 1;
      bits   = r * (bits & ((1 << length) - 1));
      h      = l + ((bits + r) >> length) - 1;
      l     += (bits >> length);
      updateEncoder();
    }
  }

  /**
   * Decodes one symbol (using the current context).
   *
   * @return Read symbol ({@code -1} if no symbol is available).
   * @throws IOException
   * @see    #getBits
   * @see    #put
   * @see    #available
   */
  public int get ()
  throws IOException
  {
    if ( state != OpenState.READ )
      throw new IOException( "ArithmeticCodec is not opened for reading!" );

    if ( v < l || v > h )
      throw new IOException( "Arithmetic decoder out of range!" );

    long r = h - l + 1;
    int  s = hist.convert( ((v - l + 1) * hist.total() - 1) / r );
    h      = l + (r * hist.left( s + 1 )) / hist.total() - 1;
    l     += (r * hist.left( s )) / hist.total();
    updateDecoder();
    hist.inc( s, increment );
    position++;
    return s;
  }

  /**
   * Decodes one subinterval from the given {@link WheelOfFortune}.
   * Doesn't use contexts and assigned histograms..
   *
   * @param  wheel Set of intervals (wheel of fortune).
   * @return Index of the relevant interval.
   */
  public int get ( WheelOfFortune wheel )
  throws IOException
  {
    if ( state != OpenState.READ )
      throw new IOException( "ArithmeticCodec is not opened for reading!" );

    if ( wheel == null )
      throw new IOException( "Invalid wheel!" );

    if ( v < l || v > h )
      throw new IOException( "Arithmetic decoder out of range!" );

    long r = h - l + 1;
    int  s = wheel.convert( ((v - l + 1) * wheel.total() - 1) / r );
    h      = l + (r * wheel.left( s + 1 )) / wheel.total() - 1;
    l     += (r * wheel.left( s )) / wheel.total();
    updateDecoder();
    position++;
    return s;
  }

  /**
   * Reads the given amount of "raw" bits (by-passes the entropy decoder).
   *
   * @param  length Number of bits to read.
   * @return Read bits (or {@code -1} if no bits are available).
   * @throws IOException
   * @see    #get
   * @see    #putBits
   * @see    #available
   */
  public long getBits ( int length )
  throws IOException
  {
    if ( state != OpenState.READ )
      throw new IOException( "ArithmeticCodec is not opened for reading!" );

    if ( v < l || v > h )
      throw new IOException( "Arithmetic decoder out of range!" );

    if ( length != 0 )
    {
      long r    = h - l + 1;
      long bits = (((v - l + 1) << length) - 1) / r;
      h         = l + ((r * bits + r) >> length) - 1;
      l        += ((r * bits) >> length);
      updateDecoder();
      return bits;
    }
    return 0;
  }

  public void open ( boolean output, int ctx )
  throws IOException
  {
    if ( state != OpenState.CLOSED )
      throw new IOException( "ArithmeticCodec has already been opened!" );

    position   =
    rawBits    =
    follow     =
    l          =
    v          = 0;
    h          = VAL_MAX;
    defaultCtx = ctx;

    hist       = (EntropyHistogram)getInterface( "histogram", IFACE+"EntropyHistogram" );
    hist.init( defaultMaxSymbol + 1, decimationLimit, true, defaultCtx );
    if ( hist == null )
      throw new IOException( "Couldn't initialize histogram!" );

    stream = (BitStream)getInterface( PL_OUTPUT, IFACE+"BitStream" );
    if ( stream == null )
      throw new IOException( "Couldn't open stream!" );
    stream.open( output, (String)null, null );

    state = output ? OpenState.WRITE : OpenState.READ;
    if ( !output )
      for ( int i = 0; i < VAL_BITS; i++ )
        v += v + getBit();
  }

  public void close ()
  throws IOException
  {
    if ( state == OpenState.CLOSED ) return;

    if ( state == OpenState.WRITE )
    {
      flush();
      if ( statistics )
      {
        Piece hist = (Piece)getInterface( "histogram", IFACE+"EntropyHistogram" );
        if ( hist != null )
          hist.set( STATISTICS, null );
      }
    }

    stream.close();
    stream = null;

    hist.init( defaultMaxSymbol + 1, decimationLimit, true, defaultCtx );
    position =
    rawBits  = 0;
    state    = OpenState.CLOSED;
  }

  public int setContext ( int ctx )
  throws IOException
  {
    if ( state == OpenState.CLOSED )
      throw new IOException( "ArithmeticCodec is closed!" );

    return hist.setContext( ctx );
  }

  public int getMaxSymbol ()
  {
    if ( state == OpenState.CLOSED ) return -1;

    return( hist.getSymbols() - 1 );
  }

  public void setMaxSymbol ( int maxSymbol )
  {
    if ( state != OpenState.CLOSED )
    {
        // set it for this context and the following ones
      hist.setSymbols( maxSymbol + 1 );
      hist.setDefaultSymbols( maxSymbol + 1 );
    }

    defaultMaxSymbol = maxSymbol;
  }

  public long position ()
  throws IOException
  {
    return position;
  }

  public long compressed ()
  throws IOException
  {
    return rawBits;
  }

  public boolean available ()
  throws IOException
  {
    return( state == OpenState.WRITE &&
            stream.available() != 0 );
  }

  //==========================================================================
  //  Interface Property:

  public final static String INCREMENT        = "Increment";
  public final static String DECIMATION_LIMIT = "Decimation Limit";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null ) return;

      // STATISTICS:
    if ( key.compareTo(STATISTICS) == 0 )
    {
      if ( value != null )
        statistics = booleanProperty( value, statistics );
      else
      {
        Piece hist = (Piece)getInterface( "histogram", IFACE+"EntropyHistogram" );
        if ( hist != null )
          hist.set( STATISTICS, value );
      }
      return;
    }

    if ( value == null ) return;

      // INCREMENT:
    if ( key.compareTo(INCREMENT) == 0 )
    {
      increment = intProperty( value, increment );
      return;
    }
      // DECIMATION_LIMIT:
    if ( key.compareTo(DECIMATION_LIMIT) == 0 )
    {
      decimationLimit = intProperty( value, decimationLimit, 1, MAX_DECIMATION_LIMIT );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(INCREMENT) == 0 )
      return increment;
    if ( key.compareTo(DECIMATION_LIMIT) == 0 )
      return decimationLimit;

    return null;
  }

  //==========================================================================
  //  Module registration

  /** Object name. */
  private final static String NAME = "ArithmeticCodec";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "EntropyCodecToBitStream";

  /** Object description. */
  private final static String DESCRIPTION = "Arithmetic codec implementation.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_COMPRESSION + ".arithmetic";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,    IFACE+"EntropyCodec" );
      t.newOutputPlug( PL_OUTPUT,   IFACE+"BitStream" );
      t.newOutputPlug( "histogram", IFACE+"EntropyHistogram" );
        // Property:
      t.propBegin( INCREMENT, TYPE_INTEGER, "Frequency increment for each encoded symbol", true );
        t.propDefault( 1 );
        t.propBounds(  1, 256 );
      t.propEnd();
      t.propBegin( DECIMATION_LIMIT, TYPE_INTEGER, "Decimation limit", true );
        t.propDefault( 32768 );
        t.propBounds(  1, MAX_DECIMATION_LIMIT );
      t.propEnd();
      t.propBegin( STATISTICS, TYPE_BOOLEAN, "Print object statistics", false );
      t.propEnd();
     }
   return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
