package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Audio filter framework - able to process several simple filters successively.
 *
 * @version 0.25 $Rev: 300 $ $Date: 2006-07-20 01:03:44 +0200 (čt, 20 VII 2006) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/AudioFilter.java">AudioFilter.java</a>
 */
public class AudioFilter extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Maximum number of pluggable filters. */
  public final static int MAX_FILTERS = 5;

  /** Size of transform batch in time-slots. */
  public final static int BATCH_SIZE = 1024;

  /** Implicit volume gain (multiplier). */
  private double gain = 1.0;

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts audio data transform.
   *
   * @param  type The action type (whatever it means..).
   * @return <tt>true</tt> if the action was successful.
   */
  public boolean fire ( int type )
  {
    SampleData input  = (SampleData)getInterface( "audio1", IFACE+"SampleData" );
    SampleData output = (SampleData)getInterface( "audio2", IFACE+"SampleData" );

    Filter1D[] filter = new Filter1D[ MAX_FILTERS ];
    int filters = 0;
    do
    {
      String key = "filter" + (filters + 1);
      filter[filters] = (Filter1D)getInterface( key, IFACE+"Filter1D" );
    } while ( filter[filters] != null &&
              ++filters < MAX_FILTERS );

    if ( input  == null ||
         output == null ||
         filters == 0 ) return false;     // I need at least one filter..

    int formats[][] = new int[ filters + 1 ][ 4 ];
                                          // for each "audio line": channels, bytes-per-sample, samples, frequency
    int[]    dims0  = { 0, 1 };
    int[]    dims1  = { 0, 1 };
    double[] period = { 1.0 };

      // source audio format and size:
    input.getDataDim( dims0 );
    input.getSampleGeometry( period );
    int srcBytesPerSample;
    switch( input.getSampleType() )
    {
      case SampleData.SAMPLE_BYTE:
        srcBytesPerSample = 1;
        break;

      case SampleData.SAMPLE_INT:
        srcBytesPerSample = 4;
        break;

      default:
        srcBytesPerSample = 2;
    }
    int i;

      // get intermediate and target formats:
    formats[0][ Filter1D.IND_CHANNELS ]         = dims0[1];
    formats[0][ Filter1D.IND_BYTES_PER_SAMPLE ] = srcBytesPerSample;
    formats[0][ Filter1D.IND_SAMPLES ]          = dims0[0];
    formats[0][ Filter1D.IND_FREQUENCY ]        = Formula.round( 1.0 / period[0] );
    for ( i = 0; i < filters; i++ )
      filter[i].format( formats[i], formats[i+1] );

      // set target audio format and size:
    dims0[0]  = formats[filters][ Filter1D.IND_SAMPLES ];
    dims0[1]  = formats[filters][ Filter1D.IND_CHANNELS ];
    period[0] = 1.0 / formats[filters][ Filter1D.IND_FREQUENCY ];
    int targetType;
    switch ( formats[filters][ Filter1D.IND_BYTES_PER_SAMPLE ] )
    {
      case 1:
        targetType = SampleData.SAMPLE_BYTE;
        break;
      case 4:
        targetType = SampleData.SAMPLE_INT;
        break;
      default:
        targetType = SampleData.SAMPLE_SHORT;
    }
    output.resetData( targetType, false, dims0 );
    output.setSampleGeometry( period );

      // temporary buffer:
    int maxSamples = BATCH_SIZE * formats[0][ Filter1D.IND_CHANNELS ];
    long samples   = maxSamples;
    for ( i = 0; i < filters; i++ )
    {
      samples = (samples * formats[i+1][ Filter1D.IND_CHANNELS ] * formats[i+1][ Filter1D.IND_SAMPLES ]) /
                (formats[i][ Filter1D.IND_CHANNELS ] * formats[i][ Filter1D.IND_SAMPLES ]);
      if ( (int)samples > maxSamples )
        maxSamples = (int)samples;
    }
    double[][] buffer = new double[ 2 ][ maxSamples ];
    int srcIndex = 0;

      // target format support:
    double  mul     = gain;
    byte[]  bTarget = null;
    short[] sTarget = null;
    int[]   iTarget = null;
    switch ( targetType )
    {
      case SampleData.SAMPLE_BYTE:
        bTarget = new byte[ (int)samples ];
        mul    *= Byte.MAX_VALUE;
        break;
      case SampleData.SAMPLE_INT:
        iTarget = new int[ (int)samples ];
        mul    *= Integer.MAX_VALUE;
        break;
      default:
        sTarget = new short[ (int)samples ];
        mul    *= Short.MAX_VALUE;
    }

      // processing loop - I. init
    for ( i = 0; i < filters; i++ )
      filter[i].reset();

      // processing loop - II. batch loop
    int srcDataPointer = 0;
    int tgtDataPointer = 0;
    int srcSamples     = formats[      0][ Filter1D.IND_SAMPLES ];
    int srcChannels    = formats[      0][ Filter1D.IND_CHANNELS ];
    int tgtSamples     = formats[filters][ Filter1D.IND_SAMPLES ];
    int tgtChannels    = formats[filters][ Filter1D.IND_CHANNELS ];

    int toGet, len;
    outer:
    while ( tgtDataPointer < tgtSamples )
    {
      if ( srcDataPointer < srcSamples )
      {
          // retrieve source data:
        toGet           = Math.min( srcSamples - srcDataPointer, BATCH_SIZE );
        dims0[0]        = srcDataPointer;
        dims0[1]        = 0;
        srcDataPointer += toGet;
        dims1[0]        = srcDataPointer;
        dims1[1]        = srcChannels;
        len             = input.getSamples( dims0, dims1, buffer[srcIndex] );
        if ( len < 0 ) break;             // error

          // filter chain:
        for ( i = 0; i < filters; i++ )
        {
          len = filter[i].batch( buffer[srcIndex], len, buffer[1-srcIndex] );
          if ( len < 0 ) break outer;
          srcIndex = 1 - srcIndex;
        }
      }
      else
      {
        len   = maxSamples;
        toGet = len / tgtChannels;
        if ( toGet > tgtSamples - tgtDataPointer )
          toGet = tgtSamples - tgtDataPointer;
        len   = toGet * tgtChannels;
        Arrays.fill( buffer[srcIndex], 0, len, 0.0 );
      }

        // write result data:
      dims0[0]        = tgtDataPointer;
      dims0[1]        = 0;
      tgtDataPointer += len / tgtChannels;
      dims1[0]        = tgtDataPointer;
      dims1[1]        = tgtChannels;
      switch ( targetType )
      {
        case SampleData.SAMPLE_BYTE:
          for ( i = 0; i < len; i++ )
            bTarget[i] = (byte)Formula.clamp( Formula.round( buffer[srcIndex][i] * mul ), -128, 127 );
          output.setDataArray( bTarget, 0, dims0, dims1 );
          break;
        case SampleData.SAMPLE_SHORT:
          for ( i = 0; i < len; i++ )
            sTarget[i] = (short)Formula.clamp( Formula.round( buffer[srcIndex][i] * mul ), Short.MIN_VALUE, Short.MAX_VALUE );
          output.setDataArray( sTarget, 0, dims0, dims1 );
          break;
        case SampleData.SAMPLE_INT:
          for ( i = 0; i < len; i++ )
            iTarget[i] = Formula.clamp( Formula.round( buffer[srcIndex][i] * mul ), Integer.MIN_VALUE, Integer.MAX_VALUE );
          output.setDataArray( iTarget, 0, dims0, dims1 );
          break;
      }
    }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String GAIN  = "Implicit gain";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // GAIN:
    if ( key.compareTo(GAIN) == 0 )
    {
      gain = doubleProperty( value, gain, 0.0, 100.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(GAIN) == 0 )
      return gain;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "AudioFilter";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToFilter1DAndSampleDataAndSampleData";

  /** Object description. */
  private final static String DESCRIPTION = "Audio filter cascade framework.";

  /** Object category. */
  protected final static String CATEGORY = C_1D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"Trigger" );
      t.newOutputPlug(    "audio1",  IFACE+"SampleData" );
      t.newOutputPlug(    "audio2",  IFACE+"SampleData" );
      t.newOptOutputPlug( "filter1", IFACE+"Filter1D" );
      t.newOptOutputPlug( "filter2", IFACE+"Filter1D" );
      t.newOptOutputPlug( "filter3", IFACE+"Filter1D" );
      t.newOptOutputPlug( "filter4", IFACE+"Filter1D" );
      t.newOptOutputPlug( "filter5", IFACE+"Filter1D" );
        // Property:
      t.propBegin( GAIN, TYPE_DOUBLE, "Implicit gain", true );
        t.propDefault( "1.0" );
        t.propBounds( "0.0", "100.0" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
