package cz.cuni.jagrlib.piece;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.sound.sampled.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Audio replay module with GUI window.
 *
 * @version 0.26 $Rev: 472 $ $Date: 2007-05-31 09:33:37 +0200 (čt, 31 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/AudioReplay.java">AudioReplay.java</a>
 */
public class AudioReplay extends Piece implements Trigger
{
  //==========================================================================
  //  Support routines:

  /** Finishes the replay loop. */
  protected void finishReplay ()
  {
    if ( thread == null ) return;

    synchronized ( thread )
    {
      userBreak = true;
      restart   = false;
      thread.notify();
    }
  }

  /** Start/stop replaying. */
  protected void startStopReplay ()
  {
    if ( thread == null ) return;

    synchronized ( thread )
    {
      stopped = !stopped;
      thread.notify();
    }
  }

  /** Move replay pointer relative to the actual position. */
  protected void moveRelative ( int delta )
  {
    if ( thread == null ) return;

    synchronized ( thread )
    {
      newPtr     = Formula.clamp( samplePtr + delta, fromSample, toSample );
      chPosition = true;
      thread.notify();
    }
  }

  //==========================================================================
  //  Support classes:

  /**
   * Custom component for progress-bar drawing.
   */
  class ProgressCanvas extends JComponent
  {
    private static final long serialVersionUID = 1L;

    /** Progress value from {@code 0.0} to {@code 1.0}. */
    double progress = 0.0;

    @Override
    protected void paintComponent ( Graphics g )
    {
      int    width  = getWidth();
      int    height = getHeight();
      double done   = Formula.clamp( progress, 0.0, 1.0 ) * width;
      int    transX = (int)done;
      done          = (done - transX) * height;
      int    transY = (int)done;
        // done:
      g.setColor( getBackground() );
      g.fillRect( 0, 0, transX, height );
      g.drawLine( transX, 0, transX, transY );
        // to do:
      g.setColor( getForeground() );
      if ( transY < height - 1 )
        g.drawLine( transX, transY + 1, transX, height - 1 );
      g.fillRect( transX + 1, 0, width, height );
    }

    @Override
    public Dimension getPreferredSize ()
    {
      return new Dimension( BAR_WIDTH, BAR_HEIGHT );
    }

    @Override
    public Dimension getMinimumSize ()
    {
      return new Dimension ( BAR_WIDTH / 2, BAR_HEIGHT / 2 );
    }

  }

  /**
   * Thread responsible for sound replaying and progress-bar repainting.
   */
  class ReplayThread extends Thread
  {
    /** Sets thread name for debugging identification. */
    public ReplayThread ()
    {
      super( "JaGrLib - AudioReplay" );
    }

    @Override
    public void run ()
    {
      try
      {
        int    channels   = toChannel - fromChannel;
        int    maxSamples = (int)( channels * frequency * FRAME * 3.0 );
               maxSamples = (maxSamples + 3) & -4;
        int    granul     = bytesPerSample * channels;
        int[]  fromIndex  = new int[ 2 ];
        int[]  toIndex    = new int[ 2 ];
        int[]  buffer     = new int[ maxSamples ];
        byte[] byteBuffer = new byte[ maxSamples * bytesPerSample ];
        int    i, j, k;

        sourceDataLine.open( audioFormat );

        do
        {
          sourceDataLine.start();
          stopped         = false;
          int wholeBuffer = sourceDataLine.getBufferSize();
          fromIndex[1]    = fromChannel;
          toIndex[1]      = toChannel;
          samplePtr       = chPosition ? newPtr : fromSample;
          chPosition      = false;

          int ready;                      // number of ready-to-replay bytes
          int toAdd;                      // number of bytes/timeslots/samples to add

          while ( samplePtr < toSample &&
                  !userBreak )
          {
              // retrieve next bunch of data:
            ready = wholeBuffer - sourceDataLine.available();
            toAdd = (int)(bytesPerSecond * FRAME * 3.0) - ready;
                                          // number of bytes to add
            if ( toAdd > 0 )
            {
                // how much data is needed?
              toAdd = (toAdd + granul - 1) / granul;
                                          // number of time-slots to add
              if ( samplePtr + toAdd > toSample ) toAdd = toSample - samplePtr;

                // retrieve data from source audio file:
              if ( toAdd > 0 )
              {
                fromIndex[0] = samplePtr;
                samplePtr   += toAdd;
                toIndex[0]   = samplePtr;
                toAdd       *= channels;  // number of samples (in all channels) to add
                sampleObject.getSamples( fromIndex, toIndex, buffer );

                  // convert the data:
                if ( bytesPerSample > 1 )
                  for ( i = k = 0; i < toAdd; i++ )
                    for ( j = 0; j++ < bytesPerSample; )
                    {
                      byteBuffer[k++] = (byte)( buffer[i] & 0xff );
                      buffer[i] >>>= 8;
                    }

                  // .. and write it to the audio line:
                sourceDataLine.write( byteBuffer, 0, bytesPerSample * toAdd );
              }
            }

            do                            // stop-loop
            {
                // update progress-bar:
              if ( canvas != null )
              {
                canvas.progress = (samplePtr - fromSample + 0.0) / (toSample - fromSample);
                canvas.repaint();
              }

              if ( stopped )
                synchronized ( this )
                {
                  try
                  {
                    wait();
                  }
                  catch ( InterruptedException e )
                  {
                    userBreak = true;
                  }
                }

              if ( chPosition )
              {
                chPosition = false;
                samplePtr  = newPtr;
              }

            } while ( stopped && !userBreak );

            if ( !userBreak )             // regular delay
              synchronized ( this )
              {
                try
                {
                  wait( (long)( FRAME * 1000L ) );
                }
                catch ( InterruptedException e )
                {
                  userBreak = true;
                }
              }

            if ( chPosition )
            {
              chPosition = false;
              samplePtr  = newPtr;
            }
          }

          canvas.progress = 1.0;
          canvas.repaint();

          sourceDataLine.drain();
          sourceDataLine.stop();
          stopped = true;

          if ( restart )
            synchronized ( this )
            {
              try
              {
                wait();                   // replay was finished
              }
              catch ( InterruptedException e )
              {
              }
            }
        }
        while ( restart );

        sourceDataLine.close();
      }
      catch ( Exception e )
      {
        LogFile.exception( e );
      }
    }
  }

  /**
   * Inner class responsible for window-close event handling (by user interaction).
   */
  private class LocalWindowListener extends WindowAdapter
  {
    /** Close-window event handler. */
    @Override
    public void windowClosing ( WindowEvent e )
    {
      finishReplay();
        // close the window manually:
      e.getWindow().dispose();
    }
  }

  /**
   * Inner listener class for replay control (start/stop/rewind/..).
   */
  private class LocalKeyListener extends KeyAdapter
  {
    @Override
    public void keyReleased ( KeyEvent e )
    {
      e.consume();

      if ( e.getKeyChar() == ' ' )
        startStopReplay();
      else
      if ( e.getKeyChar() == 'q' ||
           e.getKeyCode() == KeyEvent.VK_END )
        finishReplay();
      else
      if ( e.getKeyCode() == KeyEvent.VK_HOME )
        moveRelative( -toSample );
      else
      if ( e.getKeyCode() == KeyEvent.VK_PAGE_UP )
        moveRelative( -(toSample - fromSample) / BIG_STEP );
      else
      if ( e.getKeyCode() == KeyEvent.VK_PAGE_DOWN )
        moveRelative(  (toSample - fromSample) / BIG_STEP );
      else
      if ( e.getKeyCode() == KeyEvent.VK_LEFT )
        moveRelative( -(toSample - fromSample) /
                      ((e.getModifiersEx() & KeyEvent.SHIFT_DOWN_MASK) > 0 ? TINY_STEP : SMALL_STEP) );
      else
      if ( e.getKeyCode() == KeyEvent.VK_RIGHT )
        moveRelative(  (toSample - fromSample) /
                      ((e.getModifiersEx() & KeyEvent.SHIFT_DOWN_MASK) > 0 ? TINY_STEP : SMALL_STEP) );
    }
  }

  //==========================================================================
  //  Data:

  /** Width of progress bar canvas (in pixels). */
  protected static final int BAR_WIDTH  = 512;

  /** Height of progress bar canvas (in pixels). */
  protected static final int BAR_HEIGHT = 40;

  /** Replay frame size in seconds. */
  protected static final double FRAME   = 0.1;

  /** Fraction of big step (PageDown, PageUp). */
  protected static final int BIG_STEP   = 30;

  /** Fraction of small step (Left, Right). */
  protected static final int SMALL_STEP = 400;

  /** Fraction of tiny step (Shift+Left, Shift+Right). */
  protected static final int TINY_STEP  = 5000;

  /** Replay frame title. */
  protected String         frameTitle = "Replay";

  /** Replay info string. */
  protected String         info = "";

  /** Swing frame window to show graphic output in. */
  JFrame                   frame;

  /** Custom component - progress-drawing target. */
  ProgressCanvas           canvas;

  /** Restart the whole replay? */
  protected boolean        restart = true;

  /** Has the user-break occurred? */
  protected boolean        userBreak = false;

  /** Stopped state (space bar can restart the replay). */
  protected boolean        stopped = false;

  /** Sample pointer should be changed. */
  protected boolean        chPosition = false;

  /** New value of {@link #samplePtr}. */
  protected int            newPtr;

  /** Replay frequency in Hz (samples per second). */
  protected int            frequency;

  /** Number of bytes per sample ({@code 8} to {@code 32}). */
  protected int            bytesPerSample;

  /** Accelerator for replay-loop. */
  protected int            bytesPerSecond;

  /** Audio format of the replayed data. */
  protected AudioFormat    audioFormat;

  /** Information about the used audio data line. */
  protected DataLine.Info  dataLineInfo;

  /** Audio data line object. */
  protected SourceDataLine sourceDataLine;

  /** Replay-thread instance. */
  protected ReplayThread   thread;

  /** Source data object. */
  protected SampleData     sampleObject;

  /** Source data pointer (in samples). */
  protected int            samplePtr;

  /** Start of replayed segment (in samples). */
  protected int            fromSample = 0;

  /** End of replayed segment (in samples). */
  protected int            toSample = Integer.MAX_VALUE;

  /** First replayed channel ({@code 0} or {@code 1}). */
  protected int            fromChannel = 0;

  /** After the last replayed channel ({@code 1} or {@code 2}). */
  protected int            toChannel = 2;

  //==========================================================================
  //  Interface Trigger:

  /**
   * Triggers the (undetermined) action.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public synchronized boolean fire ( int type )
  {
    sampleObject = (SampleData)getInterface( PL_OUTPUT, IFACE+"SampleData" );
    if ( sampleObject == null ) return false;

    if ( frame == null )                  // looping for the first time
    {
      canvas = new ProgressCanvas();
      frame  = new JFrame( frameTitle );
      frame.addWindowListener( new LocalWindowListener() );
      frame.getContentPane().add( canvas, BorderLayout.CENTER );

        // GUI listeners:
      frame.addKeyListener( new LocalKeyListener() );

        // expose the frame:
      frame.pack();
    }

    frame.setVisible( true );
      // looping state variables:
    userBreak = false;
    restart   = true;
    canvas.repaint();

    if ( thread == null )
    {
      frequency = intProperty( sampleObject.get( SampleData.FREQUENCY ), 48000 );

      switch( sampleObject.getSampleType() )
      {
        case SampleData.SAMPLE_BYTE:
          bytesPerSample = 1;
          break;

        case SampleData.SAMPLE_INT:
          bytesPerSample = 4;
          break;

        default:
          bytesPerSample = 2;
      }

      int[] dims = { 0, 1 };
        // check channel & sample ranges:
      sampleObject.getDataDim( dims );
      if ( fromChannel >= dims[1] ) fromChannel = dims[1] - 1;
      toChannel = Formula.clamp( toChannel, fromChannel + 1, dims[1] );
      if ( fromSample >= dims[0] ) fromSample = dims[0];
      toSample  = Formula.clamp( toSample, fromSample, dims[0] );
      bytesPerSecond = bytesPerSample * (toChannel - fromChannel) * frequency;

      audioFormat = new AudioFormat( frequency,
                                     bytesPerSample << 3,
                                     toChannel - fromChannel,
                                     true,
                                     false );

      dataLineInfo = new DataLine.Info( SourceDataLine.class,
                                        audioFormat,
                                        (int)( bytesPerSecond * FRAME * 4.0 ) );

      try
      {
        sourceDataLine = (SourceDataLine)AudioSystem.getLine( dataLineInfo );
      }
      catch ( LineUnavailableException e )
      {
        return false;
      }

      thread = new ReplayThread();
      thread.setPriority( Thread.MIN_PRIORITY );
    }

    thread.run();

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String INFO_STRING  = "Info";
  public final static String WINDOW_TITLE = GraphicsViewer.WINDOW_TITLE;

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

        // WINDOW_TITLE:
    if ( key.compareTo( WINDOW_TITLE ) == 0 )
    {
      frameTitle = value.toString();
      if ( frame != null )
        frame.setTitle( frameTitle );
      return;
    }
      // INFO_STRING:
    if ( key.compareTo(INFO_STRING) == 0 &&
         value instanceof String )
    {
      info = (String)value;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(WINDOW_TITLE) == 0 )
    {
      if ( frame == null ) return null;
      return frame.getTitle();
    }
    if ( key.compareTo(INFO_STRING) == 0 )
      return info;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Audio replay";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToSampleData";

  /** Object description. */
  private final static String DESCRIPTION = "Audio data is replayed using javax.sound.sampled package.";

  /** Object category. */
  protected final static String CATEGORY = C_1D + '.' + C_VIEWER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"Trigger" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"SampleData" );
        // Property:
      t.propBegin( WINDOW_TITLE, TYPE_STRING, "Window title", true );
        t.propDefault( "Replay" );
      t.propEnd();
      t.propBegin( INFO_STRING, TYPE_STRING, "Data info", true );
        t.propDefault( "" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
