package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * {@link Brep} -&gt; {@link RasterGraphics} converter, prepares Brep-vertices for compression.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/BrepToImage.java">BrepToImage.java</a>
 */
public class BrepToImage extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Empty background. */
  public final static int[] BACKGROUND = {   0,  0,  0, 255 };

  /** Used color base (green & blue components are used for vertex value). */
  public final static int[] USED       = {   1,  0,  0, 255 };

  /** Linear quantizer for amplitude values. */
  protected double quantize = 1.0;

  //==========================================================================
  //  Interface Render3D:

  /**
   * Starts data transform.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    final Brep           data   = (Brep)          getInterface( PL_DATA,  IFACE+"Brep" );
    final RasterGraphics output = (RasterGraphics)getInterface( PL_IMAGE, IFACE+"RasterGraphics" );

    if ( data   == null ||
         output == null ) return false;

    int ctx        = Brep.CTX_DEFAULT;
    int worldId    = data.getAttributeId( Brep.VERTEX, ctx, Brep.COORD );
    double[] world = new double[ 4 ];
    int[]    color = USED;
    BrepIterator it;
    int vertex;

      // 1. determine output image size:
    it = data.vertexIterator( null );
    int width  = 0;
    int height = 0;
    int x, y, z;

    while ( (vertex = it.next()) != Brep.NULL )
    {
      data.getAttribute( worldId, vertex, world );
      x = Formula.round( world[0] );
      y = Formula.round( world[1] );
      if ( x >= width  ) width  = x + 1;
      if ( y >= height ) height = y + 1;
      }

      // 2. init the output image:
    output.init( width, height, RasterGraphics.MODE_RGB, 3 );
    output.setRectangle( 0, 0, width, height, BACKGROUND );

      // 3. convert data:
    it = data.vertexIterator( it );
    while ( (vertex = it.next()) != Brep.NULL )
    {
      data.getAttribute( worldId, vertex, world );
      x = Formula.clamp( Formula.round( world[0] ), 0, width  - 1 );
      y = Formula.clamp( Formula.round( world[1] ), 0, height - 1 );
      z = Formula.clamp( Formula.round( world[2] * quantize ), 0, 255 );
      color[1] =
      color[2] = z;                       // both green and blue components are modulated
      output.putPixel( x, y, color );
    }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String QUANTIZE = "Z-quantization";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // QUANTIZE:
    if ( key.compareTo(QUANTIZE) == 0 )
    {
      quantize = doubleProperty( value, quantize, 0.0, 100.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(QUANTIZE) == 0 )
      return quantize;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "BrepToImage";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToBrepAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "B-rep to RasterGraphics converter (for vertex geometry compression).";

  /** Object category. */
  protected final static String CATEGORY = C_DATA + '.' + C_COMPRESSION + ".transform";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"Trigger" );
      t.newOutputPlug( PL_DATA,   IFACE+"Brep" );
      t.newOutputPlug( PL_IMAGE,  IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( QUANTIZE, TYPE_DOUBLE, "Z quantization coefficient", true );
        t.propDefault( 1.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
