package cz.cuni.jagrlib.piece;

import java.io.*;
import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Bit-stream object with random access and buffer-capability build on
 * top of {@link RandomAccessFile} class.
 *
 * @version 0.26 $Rev: 551 $ $Date: 2007-10-10 19:58:19 +0200 (st, 10 X 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/BufferedBitStream.java">BufferedBitStream.java</a>
 */
public class BufferedBitStream extends Piece implements BitStream
{
  //==========================================================================
  //  Data:

  /** Data stream name (file-name). */
  protected String streamName = DEFAULT_STREAM_NAME;

  /** Current random access file. */
  protected RandomAccessFile file;

  /** File modification was enabled? */
  protected boolean writeEnabled;

  /** Actual stream-position in bits. */
  protected long position = 0L;

  /** Memory buffer. */
  protected byte[] buffer;

  /** Buffer position in the file (in bits). Multiple of 8. */
  protected long bufferOrigin;

  /** Buffer size in bits. */
  protected long bufferSize = 1L << 15;

  /** Used buffer bits. Has to be multiple of 8. */
  protected long bufferUsed;

  /** Position of the first changed bit in the {@link #buffer}. Relative to {@link #bufferOrigin}. */
  protected long changedMin;

  /** Position of the last changed bit in the {@link #buffer}. Relative to {@link #bufferOrigin}. */
  protected long changedMax;

  /** Actual position within the {@link #buffer}. Relative to {@link #bufferOrigin}. */
  protected long bufferPtr;

  //==========================================================================
  //  Support:

  /**
   * Flushes all unwritten data to disk file.
   * Doesn't call flush() operation of an underlying stream.
   */
  protected void flushBuffer ()
  throws IOException
  {
    if ( changedMin > changedMax ) return;

    int minw = (int)(  changedMin      >> 3 );
    int maxw = (int)( (changedMax + 8) >> 3 );
    file.seek( (bufferOrigin >> 3) + minw );
    file.write( buffer, minw, maxw - minw );
    changedMin = bufferSize;
    changedMax = -1L;
  }

  /** Does internal seek in the memory buffer. */
  protected void bufferSeek ( long bufferPos )
  {
    bufferPtr = bufferPos;
    position  = bufferOrigin + bufferPos;
  }

  /** Reads buffer into memory. Old buffer content can be discarded! */
  protected void readBuffer ( long pos )
  throws IOException
  {
    position     = pos;
    pos       >>>= 3;
    file.seek( pos );
    bufferOrigin = pos << 3;
    int read     = file.read( buffer, 0, (int)( bufferSize >> 3 ) );
    if ( read < 0 ) read = 0;
    bufferUsed   = read << 3;
    if ( bufferUsed < bufferSize )
      Arrays.fill( buffer, read, buffer.length, (byte)0 );
    bufferPtr    = position - bufferOrigin;
    changedMin   = bufferSize;
    changedMax   = -1L;
  }

  protected void changeBufferSize ( long newSize )
  throws IOException
  {
    if ( newSize < 128 ) newSize = 128;   // 16 bytes

    if ( buffer  != null &&
         newSize == bufferSize ) return;

    if ( file != null )                   // the stream is open
    {
      if ( writeEnabled )
        flushBuffer();
    }

    bufferSize = newSize;
    buffer     = new byte[ (int)( newSize >> 3 ) ];

    if ( file != null )
      readBuffer( position );
  }

  //==========================================================================
  //  Interface BitStream:

  /**
   * Stream is able to read data.
   * If {@code true}, stream can be opened automatically in the init-phase.
   */
  public boolean canRead ()
  {
    return true;
  }

  /**
   * Stream initialization.
   *
   * @param  output Open for output?
   * @param  name Stream name (either stream identification e.g. file-name). {@code null}
   *              for implicit bit-stream ({@link #reOpen}).
   * @param  detail Further implementation-specific information.
   * @throws IOException
   * @see    #reOpen
   */
  public void open ( boolean output, String name, String detail )
  throws IOException
  {
    if ( file != null ) close();

      // stream name:
    if ( name       != null ) streamName = name;
    if ( streamName == null ) streamName = DEFAULT_STREAM_NAME;

      // stream type:
    if ( detail != null )
      output = detail.length() >= 2 &&
               detail.charAt( 1 ) == 'w';
    else
      detail = output ? "rw" : "r";
    writeEnabled = output;

      // open the file:
    changeBufferSize( bufferSize );
    file = new RandomAccessFile( streamName, detail );
    readBuffer( 0L );
  }

  /**
   * Stream re-initialization.
   *
   * @param  output Open for output?
   * @param  detail Further implementation-specific information.
   * @throws IOException
   * @see    #open
   */
  public void reOpen ( boolean output, String detail )
  throws IOException
  {
    open( output, null, detail );
  }

  /**
   * Retrieves the associated output stream (non-mandatory).
   *
   * @return The associated {@link OutputStream} instance or {@code null} if
   *         not implemented.
   * @see    #getInputStream
   */
  public OutputStream getOutputStream ()
  {
    return null;
  }

  /**
   * Retrieves the associated input stream (non-mandatory).
   *
   * @return The associated {@link InputStream} instance or {@code null} if
   *         not implemented.
   * @see    #getOutputStream
   */
  public InputStream getInputStream ()
  {
    return null;
  }

  /**
   * Stream flush.
   *
   * @throws IOException
   * @see    #open
   * @see    #close
   */
  public void flush ()
  throws IOException
  {
    if ( file != null &&
         writeEnabled )
    {
      flushBuffer();
      file.getFD().sync();
    }
  }

  /**
   * Stream close.
   * Eventually flushes all written data and disconnects the stream.
   *
   * @throws IOException
   * @see    #open
   * @see    #reOpen
   */
  public void close ()
  throws IOException
  {
    if ( file == null ) return;

    if ( writeEnabled )
      flushBuffer();

    position     = (file.length() << 3);
    file.close();
    file         = null;
    bufferOrigin = 0L;
  }

  /**
   * Actual stream position in bits.
   * After the output stream is closed, returns maximal written position.
   *
   * @return Actual (maximal) stream position in bits.
   * @throws IOException
   * @see    #available
   * @see    #seek
   */
  public long position ()
  throws IOException
  {
    return position;
  }

  /**
   * How many bits are available to read (or skip).
   *
   * @return Number of bits available to read (or skip).
   * @throws IOException
   * @see    #position
   * @see    #seek
   */
  public long available ()
  throws IOException
  {
    if ( file != null )
      return( (file.length() << 3) - position );

    return 0L;
  }

  /**
   * Sets actual position of the stream in bits.
   * Non mandatory.
   *
   * @param  pos Required stream position in bits.
   * @return New stream position in bits (or {@code -1L} if not implemented).
   * @throws IOException
   * @see    #position
   */
  public long seek ( long pos )
  throws IOException
  {
    if ( file == null ) return -1L;

    if ( pos >= bufferOrigin &&
         pos <  bufferOrigin + bufferSize )
    {                                     // no need to move buffer origin
      bufferSeek( pos - bufferOrigin );
    }
    else
    {
      if ( writeEnabled )
        flushBuffer();
      readBuffer( pos );
    }

    return position;
  }

  /**
   * Writes a bunch of bits into the stream.
   *
   * @param  bits Bits to be written (LSB first).
   * @param  length Number of bits to write (negative lengths are ignored).
   * @throws IOException
   * @see    #writeBytes
   * @see    #read
   */
  public void write ( long bits, int length )
  throws IOException
  {
    if ( file == null ||
         !writeEnabled ) return;

    if ( length > 64 ) length = 64;

    while ( length > 0 )                  // contribute to one byte
    {
      if ( bufferPtr >= bufferSize )      // the buffer is full
      {
        flushBuffer();
        readBuffer( position );
      }

      int bp = (int)( bufferPtr >> 3 );   // byte index in buffer[]
      int lo = (int)( bufferPtr & 7 );    // low bits to keep in buffer[bp]
      int le = Math.min( length, 8 - lo ); // bits to write there..

      if ( bufferPtr < changedMin )
        changedMin = bufferPtr;
      if ( bufferPtr + le - 1 > changedMax )
        changedMax = bufferPtr + le - 1;

      int leBits = (1 << le) - 1;
      int mask   = 0xff - (leBits << lo);
      buffer[bp] = (byte)( (buffer[bp] & mask) | ((bits & leBits) << lo) );

      bits    >>>= le;
      length    -= le;
      bufferPtr += le;
      position  += le;
    }
  }

  /**
   * Writes a bunch of bytes into the stream.
   *
   * @param  data Bytes to be written (each byte is written in "LSB first" order).
   * @param  start First index to write. Must be less than {@code data.length}.
   * @param  length Number of bytes to write ({@code start + length > data.length} is ignored).
   * @throws IOException
   * @see    #write
   * @see    #readBytes
   */
  public void writeBytes ( byte[] data, int start, int length )
  throws IOException
  {
    if ( file == null || !writeEnabled ||
         length <= 0  || data == null ||
         start >= data.length ) return;

    if ( start + length > data.length ) length = data.length - start;

    if ( (bufferPtr & 7) > 0 )
      do
        write( data[start++], 8 );
      while ( --length > 0 );

    else
      while ( length > 0 )
      {
        if ( bufferPtr >= bufferSize )    // the buffer is full
        {
          flushBuffer();
          readBuffer( position );
        }

        int bp = (int)( bufferPtr >> 3 ); // byte index in buffer[]
        int le = Math.min( length, (int)(bufferSize >> 3) - bp ); // bytes to write..

        if ( bufferPtr < changedMin )
          changedMin = bufferPtr;
        if ( bufferPtr + (le << 3) - 1 > changedMax )
          changedMax = bufferPtr + (le << 3) - 1;

        System.arraycopy( data, start, buffer, bp, le );

        length    -= le;
        le       <<= 3;
        bufferPtr += le;
        position  += le;
      }
  }

  /**
   * Reads a bunch of bits from the stream.
   *
   * @param  length Number of bits to read (negative lengths are ignored).
   * @return Read bits (LSB first).
   * @throws IOException
   * @see    #readBytes
   * @see    #write
   */
  public long read ( int length )
  throws IOException
  {
    if ( file == null ||
         length <= 0 ) return 0L;

    if ( length > 64 ) length = 64;

    long accu = 0L;
    int  used = 0;
    while ( used < length )               // read from one buffer byte
    {
      if ( bufferPtr >= bufferSize )      // the buffer is over
      {
        flushBuffer();
        readBuffer( position );
      }

      if ( bufferPtr >= bufferUsed )
        break;

      int bp = (int)( bufferPtr >> 3 );   // byte index in buffer[]
      int lo = (int)( bufferPtr & 7 );    // low bits to skip over in buffer[bp]
      int le = Math.min( length - used, 8 - lo ); // bits to read from there..

      long n = (buffer[bp] >> lo) & ((1 << le) - 1);
      accu  |= n << used;

      used      += le;
      bufferPtr += le;
      position  += le;
    }

    return accu;
  }

  /**
   * Reads a bunch of bytes from the stream.
   *
   * @param  data Array holding the read bytes (each byte is read in "LSB first" order).
   * @param  start First index to read. Must be less than {@code data.length}.
   * @param  length Number of bytes to read ({@code start + length > data.length} is ignored).
   * @throws IOException
   * @see    #read
   * @see    #writeBytes
   */
  public void readBytes ( byte[] data, int start, int length )
  throws IOException
  {
    if ( file == null || length <= 0 ||
         data == null || start >= data.length ) return;

    if ( start + length > data.length ) length = data.length - start;
    do
      data[start++] = (byte)read( 8 );
    while ( --length > 0 );
  }

  /**
   * Skips the given number of bits.
   *
   * @param  length Number of bits to skip (negative lengths are ignored).
   * @return Number of bits actually skipped.
   * @throws IOException
   * @see    #seek
   */
  public long skip ( long length )
  throws IOException
  {
    if ( length <= 0 ) return 0L;

    long oldPos = position;
    position    = seek( position + length );
    return( position - oldPos );
  }

  //==========================================================================
  //  Interface Property:

  /** Buffer size in bytes. */
  public final static String BUFFER_SIZE = "Buffer size";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // STREAM_NAME:
    if ( key.compareTo(STREAM_NAME) == 0 &&
         value instanceof String )
    {
      streamName = (String)value;
      return;
    }
      // BUFFER_SIZE:
    if ( key.compareTo(BUFFER_SIZE) == 0 )
    {
      try
      {
        changeBufferSize( intProperty( value, (int)(bufferSize >> 3), 16, 1 << 23 ) << 3 );
      }
      catch ( IOException e )
      {
        LogFile.exception( e );
      }
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(STREAM_NAME) == 0 )
      return streamName;
    if ( key.compareTo(BUFFER_SIZE) == 0 )
      return (int)(bufferSize >> 3);

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Random access bit-stream";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "BitStream";

  /** Object description. */
  private final static String DESCRIPTION = "Random access buffered bit-stream using builtin Java class RandomAccessFile.";

  /** Object category. */
  protected final static String CATEGORY = C_IO + '.' + C_BINARY;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT,    IFACE+"BitStream" );
      t.newInputPlug( PL_PROPERTY, IFACE+"Property" );
        // Property:
      t.propBegin( STREAM_NAME, TYPE_STRING, "File name", true );
        t.propDefault( "file-name" );
      t.propEnd();
      t.propBegin( BUFFER_SIZE, TYPE_INTEGER, "Buffer size in bytes", true );
        t.propDefault( 4096 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
