package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Simple colormap store implementation.
 *
 * @author  PE, Jirik
 * @version 0.25 $Rev: 276 $ $Date: 2006-05-27 18:24:56 +0200 (so, 27 V 2006) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/Colormaps.java">Colormaps.java</a>
 */
public class Colormaps extends Piece implements ColormapStore
{
  //==========================================================================
  //  Data:

  /** Current number of Colormaps. */
  protected int storeSize = 0;

  /** Colormaps representation. */
  protected int[][][] store;

  /** Color mode */
  protected int mode = RasterGraphics.MODE_RGB;

  /** Number of bands for {@link RasterGraphics#MODE_BAND}. */
  protected int bands;

  protected static final double TO_DOUBLE = 1.0 / 255.0;

  //==========================================================================
  //  Interface ColormapStore:

  /**
   * Store (re-)initialization.
   *
   * @param  _mode Colormap item value type (<tt>RasterGraphics.MODE_*</tt> constant).
   * @param  _bands Number of bands (for {@link RasterGraphics#MODE_BAND}).
   * @see    RasterGraphics
   * @see    #init()
   */
  public void init ( int _mode, int _bands )
  {
    if ( _mode != RasterGraphics.MODE_RGB )
      _mode = RasterGraphics.MODE_RGB;

    mode      = _mode;
    bands     = 3;
    storeSize = 0;
    store     = new int[ 1 ][][];
  }

  /**
   * Store re-initialization (keeps current colormap mode).
   *
   * @see    #init(int,int)
   */
  public void init ()
  {
    storeSize = 0;
  }

  /**
   * Returns the store size (colormaps are indexed from <tt>0</tt> to
   * {@link #size}<tt>-1</tt>).
   *
   * @return Store size.
   * @see    #setSize
   */
  public int size ()
  {
    return storeSize;
  }

  /**
   * Assigns the new store size (potentially discards some colormaps).
   *
   * @param  newSize The new store size.
   * @return The old store size.
   * @see    #size
   */
  public int setSize ( int newSize )
  {
    int oldSize = storeSize;
    store       = new int[ newSize ][][];
    storeSize   = newSize;
    return oldSize;
  }

  /**
   * Retrieves the i-th colormap.
   *
   * @param  i Colormap index (from <tt>0</tt> to {@link #size}<tt>-1</tt>).
   * @return Array holding the colormap data.
   * @see    #getColormap
   */
  public double[][] getColormapDouble ( int i )
  {
    if ( i < 0 || i >= storeSize )
      return null;

    int colors = store[i].length;
    double[][] doubleColormap = new double[ colors ][];
    int[][] storei = store[i];

    for ( int j = 0; j < colors; j++ )
    {
      doubleColormap[j] = new double[ storei[j].length ];
      for ( int k = 0; k < storei[j].length; k++ )
        doubleColormap[j][k] = storei[j][k] * TO_DOUBLE;
    }

    return doubleColormap;
  }

  /**
   * Retrieves the i-th colormap.
   *
   * @param  i Colormap index (from <tt>0</tt> to {@link #size}<tt>-1</tt>).
   * @return Array holding the colormap data.
   * @see    #getColormapDouble
   */
  public int[][] getColormap ( int i )
  {
    if ( i < 0 || i >= storeSize )
      return null;

    return store[i];
  }

  /**
   * Assigns the i-th colormap.
   *
   * @param  i Colormap index.
   * @param  cmap The new colormap data (or <tt>null</tt>).
   * @return Array holding the old colormap data (or <tt>null</tt>).
   * @see    #setColormap(int,double[][])
   */
  public int[][] setColormap ( int i, int[][] cmap )
  {
    if ( i < 0 || i >= storeSize )
      return null;

    int[][] oldColormap = store[i];
    store[i] = cmap;

    return oldColormap;
  }

  /**
   * Assigns the i-th colormap.
   *
   * @param  i Colormap index.
   * @param  cmap The new colormap data (or <tt>null</tt>).
   * @return Array holding the old colormap data (or <tt>null</tt>).
   * @see    #setColormap(int,int[][])
   */
  public double[][] setColormap ( int i, double[][] cmap )
  {
    if ( i < 0 || i >= storeSize )
      return null;

    double[][] oldColormap = getColormapDouble( i );

    int colors = cmap.length;
    int[][] newColormap = new int[ colors ][];
    for ( int j = 0; j < colors; j++ )
    {
      newColormap[j] = new int[ cmap[j].length ];
      for ( int k = 0; k < cmap[j].length; k++ )
        newColormap[j][k] = Formula.clamp( (int)(cmap[j][k] * 255.0 + 0.5), 0, 255 );
    }

    store[i] = newColormap;
    return oldColormap;
  }

//==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Colormaps";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ColormapStore";

  /** Object description. */
  private final static String DESCRIPTION = "Colormap storage object (uses RGB exclusively).";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DATA + "." + C_RASTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"ColormapStore" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
