package cz.cuni.jagrlib.piece;

import java.io.IOException;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Gray file format (using predictor, VLC and dynamic arithmetic coder).
 *
 * @version 0.25 $Rev: 399 $ $Date: 2007-02-12 23:13:47 +0100 (po, 12 II 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/CompressedFormatGray.java">CompressedFormatGray.java</a>
 */
public class CompressedFormatGray extends DefaultRasterFileFormat
{
  //==========================================================================
  //  Data:

  //--------------------------------------------------------------------------
  //  Support routines:

  /**
   * Common load code.
   *
   * @param  stream Opened input bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonLoad ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    final EntropyCodec codec   = (EntropyCodec)getInterface( PL_OUTPUT,  IFACE+"EntropyCodec" );
    final IntFunction  predict = (IntFunction) getInterface( PL_COMPARE, IFACE+"IntFunction" );

    if ( codec   == null ||
         predict == null ) return;

    codec.setMaxSymbol( 511 );            // -255 .. 255
    codec.open( false, 0 );

      // Gray header:
    int magic  = readUnsigned16( codec );
    if ( magic != MAGIC )
      throw new IOException( "Error in Gray file format" );

    int width  = readUnsigned16( codec );
    int height = readUnsigned16( codec );

      // predictor identifier = hash( class name )
    int hash     = Formula.hashClassName( predict ) & 0xffff;
    int readHash = readUnsigned16( codec );
    if ( hash != readHash )
      throw new IOException( "Error in Gray file format (bad predictor)" );

      // predictor parameters: predictor type & size
    int predictType  = readUnsigned16( codec );
    predict.set( ImagePredictor.PREDICTOR_TYPE, predictType );
    int precinctSize = readUnsigned16( codec );
    predict.set( ImagePredictor.PRECINCT_SIZE, precinctSize );

      // initialize raster image object:
    g.init( width, height, RasterGraphics.MODE_GRAY, 0 );

      // image data, pixels in scanline order (up-to-bottom, left-to-right)
    int i, j, prediction, error;
    for ( j = 0; j < height; j++ )        // one scanline
      for ( i = 0; i < width; i++ )
      {
        prediction = predict.g( i, j, 0 );
        error      = Formula.vlcInv( codec.get() );
        //LogFile.log( String.format( "Decode[%3d]: %3d, %3d, %4d", i, (prediction+error), prediction, error ) );
        g.putPixel( i, j, prediction + error );
      }

    codec.close();
  }

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonSave ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    final EntropyCodec codec   = (EntropyCodec)getInterface( PL_OUTPUT,  IFACE+"EntropyCodec" );
    final IntFunction  predict = (IntFunction) getInterface( PL_COMPARE, IFACE+"IntFunction" );
    if ( codec   == null ||
         predict == null ) return;

    codec.setMaxSymbol( 511 );            // -255 .. 255
    codec.open( true, 0 );

      // Gray header:
    writeUnsigned16( codec, MAGIC );
    int width = g.getWidth();
    writeUnsigned16( codec, width );
    int height = g.getHeight();
    writeUnsigned16( codec, height );

      // predictor identifier = hash( class name )
    int hash = Formula.hashClassName( predict ) & 0xffff;
    writeUnsigned16( codec, hash );

      // predictor parameters: predictor type & size
    int predictType = intProperty( predict, ImagePredictor.PREDICTOR_TYPE, 0 );
    writeUnsigned16( codec, predictType );
    int precinctSize = intProperty( predict, ImagePredictor.PRECINCT_SIZE, 0 );
    writeUnsigned16( codec, precinctSize );

      // image data, pixels in scanline order (up-to-bottom, left-to-right)
    int i, j, prediction;
    int[] pixel = new int[ 4 ];
    for ( j = 0; j < height; j++ )        // one scanline
      for ( i = 0; i < width; i++ )
      {
        prediction = predict.g( i, j, 0 );
        g.getPixel( i, j, pixel );
        //LogFile.log( String.format( "Encode[%3d]: %3d, %3d, %4d", i, pixel[0], prediction, (pixel[0]-prediction) ) );
        codec.put( (int)Formula.vlc( pixel[0] - prediction ) );
      }

    codec.close();
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /** 16-bit magic number (used at the file beginning). */
  public final static int MAGIC = 0x7a87;

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return 2;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type (<tt>0.0</tt> .. no way,
   *         <tt>0.5</tt> .. maybe, <tt>0.9</tt> .. almost sure, <tt>1.0</tt> ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int len = (header == null) ? 0 : header.length;
    if ( len > 2 ) len = 2;
    if ( len < 2 )                        // not enough bytes in header[]
      return 0.0;
    byte[] h =
    {                                     // BW header
      (byte)((MAGIC>>8) & 0xff),
      (byte)( MAGIC     & 0xff)
    };
    int i;
    for ( i = 0; i < len; i++ )
      if ( header[i] != h[i] ) return 0.0;
    return 1.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.png"]) or <tt>null</tt>.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.gray",
    };
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Gray file format";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "DataFileFormatToRasterGraphicsAndIntFunctionAndEntropyCodecAndBitStream";

  /** Object description. */
  private final static String DESCRIPTION = "Gray file format using external prediction and VLC coding.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,   IFACE+"DataFileFormat" );
      t.newOptOutputPlug( PL_RASTER,  IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_COMPARE, IFACE+"IntFunction" );
      t.newOptOutputPlug( PL_STREAM,  IFACE+"BitStream" );
      t.newOptOutputPlug( PL_OUTPUT,  IFACE+"EntropyCodec" );
        // Property:
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
