package cz.cuni.jagrlib.piece;

import java.io.IOException;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Lossy image compression (using discrete image transform, quantization, block coding
 * and dynamic entropy coder).
 *
 * @version 0.25 $Rev: 399 $ $Date: 2007-02-12 23:13:47 +0100 (po, 12 II 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/CompressedFormatTransform.java">CompressedFormatTransform.java</a>
 */
public class CompressedFormatTransform extends DefaultRasterFileFormat
{
  //==========================================================================
  //  Data:

  /** Transform block width. */
  protected int blockWidth  =  8;

  /** Transform block height. */
  protected int blockHeight =  8;

  /** Actual coding quality (affects encoder quantization tables). */
  protected float quality   = 80.0f;

  /** Use inter-block prediction? */
  protected boolean usePrediction = true;

  /** Number of block which is logged to LogFile. <tt>-1</tt> to swith-off the statistics. */
  protected int statBlock = -1;

  //--------------------------------------------------------------------------
  //  Support routines:

  protected void checkBlockSize ( DiscreteTransform transform, int width, int height )
  {
    int[] bounds = new int[ 4 ];
    boolean arb = transform.getSizeBounds( bounds );
    if ( blockWidth  < bounds[0] ) blockWidth  = bounds[0];
    if ( blockWidth  > bounds[1] ) blockWidth  = bounds[1];
    if ( blockHeight < bounds[2] ) blockHeight = bounds[2];
    if ( blockHeight > bounds[3] ) blockHeight = bounds[3];
    if ( arb )
    {
      if ( blockWidth  > width )  blockWidth  = width;
      if ( blockHeight > height ) blockHeight = height;
    }
    else
    {
      int b = 1;
      int m = Math.min( blockWidth, blockHeight );
      while ( b + b <= m ) b += b;
      blockWidth  =
      blockHeight = b;
    }
  }

  protected void logArray ( String str, double[] arr )
  {
    int i, j, k;
    LogFile.debug( "CompressedFormatTransform: " + str );
    for ( j = k = 0; j++ < blockHeight; )
    {
      StringBuffer sb = new StringBuffer();
      for ( i = 0; i++ < blockWidth; k++ )
        sb.append ( String.format( " %7.2f", arr[k] ) );
      LogFile.debug( new String( sb ) );
    }
  }

  protected void logArray ( String str, int[] arr )
  {
    int i, j, k;
    LogFile.debug( "CompressedFormatTransform: " + str );
    for ( j = k = 0; j++ < blockHeight; )
    {
      StringBuffer sb = new StringBuffer();
      for ( i = 0; i++ < blockWidth; k++ )
        sb.append ( String.format( " %7d", arr[k] ) );
      LogFile.debug( new String( sb ) );
    }
  }

  /**
   * Common load code.
   * Binary format &amp; principle: see {@link #commonSave commonSave()}.
   *
   * @param  stream Opened input bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonLoad ( BitStream stream, RasterGraphics g )
  throws IOException
  {
          ValueTransferFunction pixelFunction = (ValueTransferFunction)getInterface( PL_FUNCTION,  IFACE+"ValueTransferFunction" );
    final DiscreteTransform     transform     = (DiscreteTransform)    getInterface( PL_TRANSFORM, IFACE+"DiscreteTransform" );
    final BlockQuantizer        quantizer     = (BlockQuantizer)       getInterface( PL_QUANTIZER, IFACE+"BlockQuantizer" );
    final BlockCodec            blockCodec    = (BlockCodec)           getInterface( PL_CODEC,     IFACE+"BlockCodec" );
    final EntropyCodec          codec         = (EntropyCodec)         getInterface( PL_OUTPUT,    IFACE+"EntropyCodec" );

    if ( codec == null ) return;          // pixelFunction, transform, quantizer and blockCodec are optional!

    codec.open( false, 0 );
    codec.setMaxSymbol( 4095 );

      // Transform header:
    int magic     = readUnsigned16( codec );
    if ( magic != MAGIC )
      throw new IOException( "Error in Transform file format (bad MAGIC)" );

    int width     = readUnsigned16( codec );
    int height    = readUnsigned16( codec );
    int bands     = (int)codec.getBits( 8 );
    usePrediction = codec.getBits( 8 ) > 0;

      // pixelFunction parameters:
    if ( !readParameters( codec, pixelFunction ) )
      throw new IOException( "Error in Transform file format (bad pixel-transfer function)" );

      // transform parameters:
    if ( !readParameters( codec, transform ) )
      throw new IOException( "Error in Transform file format (bad discrete transform)" );

      // block size:
    blockWidth  = readUnsigned16( codec );
    blockHeight = readUnsigned16( codec );

      // quantizer parameters:
    if ( !readParameters( codec, quantizer ) )
      throw new IOException( "Error in Transform file format (bad quantizer)" );

      // blockCodec parameters:
    if ( !readParameters( codec, blockCodec ) )
      throw new IOException( "Error in Transform file format (bad block-codec)" );

      // initialize raster image object:
    switch ( bands )
    {
      case 1:
        g.init( width, height, RasterGraphics.MODE_GRAY, 0 );
        break;
      case 3:
        g.init( width, height, RasterGraphics.MODE_RGB, 0 );
        break;
      case 4:
        g.init( width, height, RasterGraphics.MODE_RGBA, 0 );
        break;
      default:
        g.init( width, height, RasterGraphics.MODE_BAND, bands );
    }

      // image blocks in scanline order (up-to-bottom, left-to-right)
    int      x, y;                        // block coordinates (upper left corner)
    int      blockLen  = blockWidth * blockHeight;
    double[] blockData = new double[ blockLen ];
    double[] coeff     = (transform == null) ? blockData : new double[ blockLen ];
    int[]    qCoeff    = new int[ blockLen ];
    int      i, j, k, block;
    int[]    pixel     = new int[ 4 ];    // RGB[A] pixel
    int[][]  predict   = usePrediction ?  // new prediction, left preduction, upper prediction
                         new int[3][1] : null;
    int[]    tmp;

    if ( statBlock >= 0 )
      LogFile.debug( "Decoding:" );

    for ( y = block = 0; y < height; y += blockHeight )
      for ( x = 0; x < width; x += blockWidth, block++ )
      {                                   // decode one image block
          // block decoding:
        if ( blockCodec == null )
          for ( i = 0; i < blockLen; i++ )
            qCoeff[i] = Formula.vlcInv( codec.get() );
        else
        {
          if ( usePrediction &&
               x == 0 )                   // 1st blick in a row => use prediction from upper neighbour
          {
            tmp        = predict[2];
            predict[2] = predict[1];
            predict[1] = tmp;
          }

          blockCodec.decode( qCoeff, predict );

          if ( usePrediction )
          {
            tmp        = predict[0];
            predict[0] = predict[1];
            predict[1] = tmp;
            if ( x == 0 )                   // save next prediction into predict[2] too
              predict[2] = Formula.cloneArray( predict[1] );
          }
        }

        if ( statBlock == block )
          logArray( "decoded coefficients", qCoeff );

          // block dequantization (reconstruction):
        if ( quantizer == null )
          for ( i = 0; i < blockLen; i++ )
            coeff[i] = qCoeff[i];
        else
          quantizer.dequantize( qCoeff, coeff );

        if ( statBlock == block )
          logArray( "reconstructed coefficients", coeff );

          // inverse discrete image transform:
        if ( transform != null &&
             transform.inverse2D( coeff, blockLen, blockData, 0, blockWidth, blockWidth ) != blockHeight )
          throw new IOException( "Error in Transform file format (bad inverse transform output length)" );

          // pixelFunction is being ignored..

        if ( statBlock == block )
          logArray( "reconstructed data", blockData );

          // restore gray image block:
        for ( j = y, k = 0; j < y + blockHeight; j++ )
          for ( i = x; i < x + blockWidth; i++, k++ )
            if ( i < width && j < height )
            {
              pixel[0] =
              pixel[1] =
              pixel[2] =
              pixel[3] = Formula.clamp( (int)( blockData[k] + 128.5 ), 0, 255 );
              g.putPixel( i, j, pixel );
            }
      }

    codec.close();
  }

  /**
   * Common save code.
   *
   * <p>Binary format &amp; principle:
   * <ol>
   *   <li>{@link ValueTransferFunction Pixel-transfer function} transforms individual pixel values
   *       (example: RGB -&gt; YC<sub>b</sub>C<sub>r</sub>). Each band will be processed individually.
   *   <li>{@link DiscreteTransform Discrete image transform} is appplied to each image block.
   *   <li>Block of coefficients is quantized using specified {@link BlockQuantizer quantizer}.
   *   <li>Quantized block is transformed (and coded) using the {@link BlockCodec block-codec}
   *       (and underlying {@link EntropyCodec entropy codec}).
   * </ol>
   *
   * @param  stream Opened output bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonSave ( BitStream stream, RasterGraphics g )
  throws IOException
  {
          ValueTransferFunction pixelFunction = (ValueTransferFunction)getInterface( PL_FUNCTION,  IFACE+"ValueTransferFunction" );
    final DiscreteTransform     transform     = (DiscreteTransform)    getInterface( PL_TRANSFORM, IFACE+"DiscreteTransform" );
    final BlockQuantizer        quantizer     = (BlockQuantizer)       getInterface( PL_QUANTIZER, IFACE+"BlockQuantizer" );
    final BlockCodec            blockCodec    = (BlockCodec)           getInterface( PL_CODEC,     IFACE+"BlockCodec" );
    final EntropyCodec          codec         = (EntropyCodec)         getInterface( PL_OUTPUT,    IFACE+"EntropyCodec" );

    if ( codec == null ) return;          // pixelFunction, transform, quantizer and blockCodec are optional!

    codec.open( true, 0 );
    codec.setMaxSymbol( 4095 );

      // Transform header:
    writeUnsigned16( codec, MAGIC );
    int width  = g.getWidth();
    writeUnsigned16( codec, width );
    int height = g.getHeight();
    writeUnsigned16( codec, height );
    int bands  = g.getBands();
    codec.putBits( bands & 0xff, 8 );
    codec.putBits( usePrediction ? 1 : 0, 8 );

      // pixelFunction parameters:
    writeParameters( codec, pixelFunction );

      // transform parameters:
    writeParameters( codec, transform );

      // block size:
    if ( transform != null )              // check block size
      checkBlockSize( transform, width, height );
    writeUnsigned16( codec, blockWidth );
    writeUnsigned16( codec, blockHeight );

    if ( quantizer != null )
    {
      quantizer.setBlockSize( blockWidth, blockHeight );
      quantizer.setQuality( quality * 0.01f );
    }

      // quantizer parameters:
    writeParameters( codec, quantizer );

    if ( blockCodec != null )
    {
      blockCodec.setBlockSize( blockWidth, blockHeight );
      blockCodec.init();
    }

      // blockCodec parameters:
    writeParameters( codec, blockCodec );

      // image blocks in scanline order (up-to-bottom, left-to-right)
    int      x, y;                        // block coordinates (upper left corner)
    int      blockLen  = blockWidth * blockHeight;
    double[] blockData = new double[ blockLen ];
    double[] coeff     = (transform == null) ? blockData : new double[ blockLen ];
    int[]    qCoeff    = new int[ blockLen ];
    int[][]  predict   = usePrediction ?  // new prediction, left preduction, upper prediction
                         new int[3][1] : null;

    int      i, j, k, block;
    int[]    tmp;

    if ( quantizer != null &&
         quantizer.quantize( coeff, null ) != blockLen )
      throw new IOException( "Error in Transform file format (bad quantizer output length)" );

    if ( statBlock >= 0 )
      LogFile.debug( "Encoding (" + Math.round( quality ) + "% quality):" );

    for ( y = block = 0; y < height; y += blockHeight )
      for ( x = 0; x < width; x += blockWidth, block++ )
      {                                   // encode one image block
          // compute gray image block:
        for ( j = y, k = 0; j < y + blockHeight; j++ )
          for ( i = x; i < x + blockWidth; i++ )
            blockData[k++] = 255.0 * (g.getGrayDouble( Math.min( i, width - 1 ), Math.min( j, height - 1 ) ) ) - 128.0;

        if ( statBlock == block )
          logArray( "original data", blockData );

          // pixelFunction is being ignored..

          // discrete image transform:
        if ( transform != null &&
             transform.transform2D( blockData, 0, blockWidth, blockHeight, blockWidth, coeff ) != blockLen )
          throw new IOException( "Error in Transform file format (bad transform output length)" );

        if ( statBlock == block )
          logArray( "coefficients", coeff );

          // block quantization:
        if ( quantizer == null )
          for ( i = 0; i < blockLen; i++ )
            qCoeff[i] = Formula.round( coeff[i] );
        else
          quantizer.quantize( coeff, qCoeff );

        if ( statBlock == block )
          logArray( "encoded coefficients", qCoeff );

          // block coding:
        if ( blockCodec == null )
          for ( i = 0; i < blockLen; i++ )
            codec.put( (int)Formula.vlc( qCoeff[i]) );
        else
        {
          if ( predict != null &&
               x == 0 )                   // 1st blick in a row => use prediction from upper neighbour
          {
            tmp        = predict[2];
            predict[2] = predict[1];
            predict[1] = tmp;
          }

          if ( statBlock == block )
            blockCodec.set( BlockCodec.LOGGING, true );

          blockCodec.encode( qCoeff, predict );

          if ( statBlock == block )
          {
            LogFile.log( (String)blockCodec.get( BlockCodec.LOGGING ) );
            blockCodec.set( BlockCodec.LOGGING, false );
          }

          if ( predict != null )
          {
            tmp        = predict[0];
            predict[0] = predict[1];
            predict[1] = tmp;
            if ( x == 0 )                 // save next prediction into predict[2] too
              predict[2] = Formula.cloneArray( predict[1] );
          }
        }
      }

    codec.close();
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /** 16-bit magic number (used at the file beginning). */
  public final static int MAGIC = 0x1b6d;

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return 2;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type (<tt>0.0</tt> .. no way,
   *         <tt>0.5</tt> .. maybe, <tt>0.9</tt> .. almost sure, <tt>1.0</tt> ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int len = (header == null) ? 0 : header.length;
    if ( len > 2 ) len = 2;
    if ( len < 2 )                        // not enough bytes in header[]
      return 0.0;
    byte[] h =
    {                                     // BW header
      (byte)((MAGIC>>8) & 0xff),
      (byte)( MAGIC     & 0xff)
    };
    int i;
    for ( i = 0; i < len; i++ )
      if ( header[i] != h[i] ) return 0.0;
    return 1.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.png"]) or <tt>null</tt>.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.tr",
    };
  }

  //==========================================================================
  //  Interface Property:

  public final static String BLOCK_WIDTH  = "Block width";
  public final static String BLOCK_HEIGHT = "Block height";
  public final static String QUALITY      = "Quality";
  public final static String PREDICTION   = "Use prediction";
  public final static String STAT_BLOCK   = "Logged block#";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // BLOCK_WIDTH:
    if ( key.compareTo(BLOCK_WIDTH) == 0 )
    {
      blockWidth = intProperty( value, blockWidth, 1, 16384 );
      return;
    }
      // BLOCK_HEIGHT:
    if ( key.compareTo(BLOCK_HEIGHT) == 0 )
    {
      blockHeight = intProperty( value, blockHeight, 1, 16384 );
      return;
    }
      // QUALITY:
    if ( key.compareTo(QUALITY) == 0 )
    {
      quality = floatProperty( value, quality, 0.0f, 100.0f );
      return;
    }
      // PREDICTION:
    if ( key.compareTo(PREDICTION) == 0 )
    {
      usePrediction = booleanProperty( value, usePrediction );
      return;
    }
      // STAT_BLOCK:
    if ( key.compareTo(STAT_BLOCK) == 0 )
    {
      statBlock = intProperty( value, statBlock );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(BLOCK_WIDTH) == 0 )
      return blockWidth;
    if ( key.compareTo(BLOCK_HEIGHT) == 0 )
      return blockHeight;
    if ( key.compareTo(QUALITY) == 0 )
      return quality;
    if ( key.compareTo(PREDICTION) == 0 )
      return usePrediction;
    if ( key.compareTo(STAT_BLOCK) == 0 )
      return statBlock;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Transform file format";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "DataFileFormatToRasterGraphicsAndValueTransferFunctionAndDiscreteTransformAnd" +
                                                "BlockQuantizerAndBlockCodecAndBitStreamAndEntropyCodec";

  /** Object description. */
  private final static String DESCRIPTION = "Lossy file format using external image transform, quantizer and block coding.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,     IFACE+"DataFileFormat" );
      t.newOptOutputPlug( PL_RASTER,    IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_FUNCTION,  IFACE+"ValueTransferFunction" );
      t.newOptOutputPlug( PL_TRANSFORM, IFACE+"DiscreteTransform" );
      t.newOptOutputPlug( PL_QUANTIZER, IFACE+"BlockQuantizer" );
      t.newOptOutputPlug( PL_CODEC,     IFACE+"BlockCodec" );
      t.newOptOutputPlug( PL_STREAM,    IFACE+"BitStream" );
      t.newOutputPlug(    PL_OUTPUT,    IFACE+"EntropyCodec" );
        // Property:
      t.propBegin( BLOCK_WIDTH, TYPE_INTEGER, "Transform block width", true );
        t.propDefault( "8" );
        t.propBounds( "1", "16384" );
      t.propEnd();
      t.propBegin( BLOCK_HEIGHT, TYPE_INTEGER, "Transform block height", true );
        t.propDefault( "8" );
        t.propBounds( "1", "16384" );
      t.propEnd();
      t.propBegin( QUALITY, TYPE_FLOAT, "Encoding quality", true );
        t.propDefault( "80.0f" );
        t.propBounds( "0.0f", "100.0f" );
      t.propEnd();
      t.propBegin( PREDICTION, TYPE_BOOLEAN, "Use inter-block prediction", true );
        t.propDefault( "true" );
      t.propEnd();
      t.propBegin( STAT_BLOCK, TYPE_INTEGER, "Logged block number", true );
        t.propDefault( "-1" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
