package cz.cuni.jagrlib.piece;

import java.awt.Color;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Generates classical Cornell Box.
 * <a href="http://www.graphics.cornell.edu/online/box/data.html">Original data from Cornell University</a>.
 *
 * @author  PE, Tomas Svab
 * @version 0.26 $Rev: 753 $ $Date: 2008-05-18 17:25:57 +0200 (ne, 18 V 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/CornellBox.java">CornellBox.java</a>
 */
public class CornellBox extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Accelerator: input B-rep database. */
  protected Brep brep;

  /** Actual {@link Brep} context. */
  protected int ctx = Brep.CTX_DEFAULT;

  /** Actual vertex-color attribute. */
  protected int attrColorV = Brep.NULL;

  /** Actual vertex-normal attribute. */
  protected int attrNormal = Brep.NULL;

  /** Actual face-color attribute. */
  protected int attrColorF = Brep.NULL;

  /** Actual face-emission attribute. */
  protected int attrEmission = Brep.NULL;

  /** Name of face-emission attribute. */
  protected String emissionName = "Emission";

  /** Supported interiors. */
  protected static enum Interior
  {
    NONE,
    SHORT_BLOCK,
    TALL_BLOCK,
    TWO_BLOCKS,
  }

  /** Actually used number of boxes. */
  protected Interior interior = Interior.NONE;

  /** Automatic tessellation? */
  protected boolean automaticTess = false;

  /** Number of rows for the box. */
  protected int cellsUBox = 1;

  /** Number of columns for the box. */
  protected int cellsVBox = 1;

  /** Number of rows for block (short, tall). */
  protected int cellsUBlock = 1;

  /** Number of columns for block (short, tall). */
  protected int cellsVBlock = 1;

  /** Maximal size of triangle in scene (used if {@link #automaticTess} is true). */
  protected double maxSideLength = 100.0;

  /** Color of the left side of the box. */
  protected Color leftSurfaceColor    = new Color( 0, 255, 0 );

  /** Color of the right side of the box. */
  protected Color rightSurfaceColor   = new Color( 255, 0, 0 );

  /** Color of all other sides. */
  protected Color otherSurfacesColor  = new Color( 255, 255, 255 );

  /** Width of light(s). */
  protected double lightWidth = 130.0;

  /** Height of light(s). */
  protected double lightHeight = 105.0;

  /** All colors in {@code float[]} format. */
  protected float[][] color = null;

  //==========================================================================
  //  Support:

  /**
   * Should be called at the beginning of each Cornell-box construction.
   */
  public void prepare ( Brep data, int ctx )
  {
    this.brep    = data;
    this.ctx     = ctx;
    attrColorV   = brep.createAttribute( Brep.VERTEX, Brep.NULL, Brep.COLOR,   Brep.ATTR_RGB_3F );
    attrNormal   = brep.createAttribute( Brep.VERTEX, Brep.NULL, Brep.NORMAL,  Brep.ATTR_COORDH_4D );
    attrColorF   = brep.createAttribute( Brep.FACE,   Brep.NULL, Brep.COLOR,   Brep.ATTR_RGB_3F );
    attrEmission = brep.createAttribute( Brep.FACE,   Brep.NULL, emissionName, Brep.ATTR_RGB_3F );
    color = new float[][]
    {
      {                                         // 0 .. other
        otherSurfacesColor.getRed()   / 255.0f,
        otherSurfacesColor.getGreen() / 255.0f,
        otherSurfacesColor.getBlue()  / 255.0f
      },
      {                                         // 1 .. right wall
        rightSurfaceColor.getRed()    / 255.0f,
        rightSurfaceColor.getGreen()  / 255.0f,
        rightSurfaceColor.getBlue()   / 255.0f
      },
      {                                         // 2 .. left wall
        leftSurfaceColor.getRed()     / 255.0f,
        leftSurfaceColor.getGreen()   / 255.0f,
        leftSurfaceColor.getBlue()    / 255.0f
      },
      {                                         // 3 .. black
        0.0f, 0.0f, 0.0f
      },
      {                                         // 4 .. light emission
        // 18.4f, 15.6f, 8.0f
        20.0f, 20.0f, 20.0f
      },
      {
        0.78f, 0.78f, 0.78f                     // 5 .. light reflectance
      },
    };
  }

  /**
   * Computes number of edges used in splitting the given segment.
   *
   * @see  #maxSideLength
   */
  protected int autoEdges ( double[] a, double[] b )
  {
    return (int)Math.ceil( Geometry.pointPoint3D( a, b ) / maxSideLength );
  }

  /**
   * Creates quadrilateral with the given parameters.
   * Must be called after {@link #prepare}.
   *
   * @param ur     Coordinates of up right vertex.
   * @param ul     Coordinates of up left vertex.
   * @param ll     Coordinates of low left vertex.
   * @param lr     Coordinates of low right vertex.
   * @param horiz  Number of horizontal edges (columns).
   * @param vert   Number of vertical edges (rows).
   * @param vColor Vertex-color.
   * @param fColor Face-color (reflectivity).
   * @param fEmiss Face-emission (nonzero for light sources).
   */
  public void createQuad ( double[] ur, double[] ul, double[] ll, double[] lr,
                           int horiz, int vert,
                           float[] vColor, float[] fColor, float[] fEmiss )
  {
    final double[] left   = Formula.cloneArray( ul );              // left point
    final double[] right  = Formula.cloneArray( ur );              // right point
    final double[] h      = Geometry.vector( left,  right, null ); // horizontal side
    final double[] vl     = Geometry.vector( left,  ll,    null ); // left side
    final double[] vr     = Geometry.vector( right, lr,    null ); // right side
    final double[] normal = Geometry.vectorProduct( h, vl, null );
    Geometry.normalize3D( normal );

    if ( automaticTess )
    {
      horiz = autoEdges( ul, ur );
      vert  = autoEdges( ul, ll );
    }

    final double vertInv  = 1.0 / vert;
    final double horizInv = 1.0 / horiz;
    int i, j, k, face;
    for ( k = 0; k < 3; k++ )
    {
      vl[k] *= vertInv;                                            // left vertical step
      vr[k] *= vertInv;                                            // right vertical step
    }

    int[]          prev  = new int[ horiz + 1 ];                   // array of vertex-handles (previous line)
    int[]          curr  = new int[ horiz + 1 ];                   // array of vertex-handles (current line)
    final double[] coord = { 0.0, 0.0, 0.0, 1.0 };                 // world coordinate of the actual vertex

    for ( i = 0; i <= vert; i++ )                                  // one vertex line (row):
    {
      int[] tmp = prev;                                            // curr can be used for the current horizontal line
      prev      = curr;
      curr      = tmp;

        // setup the actual line:
      Geometry.vector( left, right, h );
      for ( k = 0; k < 3; k++ )
      {
        h[k]    *= horizInv;                                       // horizontal step
        coord[k] = left[k];
      }

      for ( j = 0; j <= horiz; j++ )                               // one vertex in the actual line:
      {
        curr[j] = brep.createVertex();
        brep.setVertexCoords( ctx, curr[j], coord );
        brep.setAttribute( attrColorV, curr[j], vColor );
        brep.setAttribute( attrNormal, curr[j], normal );

        if ( i > 0 && j > 0 )                                      // create triangle tuple:
        {
          brep.setTriangleVertices( face = brep.createFace(), prev[j-1], prev[j  ], curr[j-1] );
          brep.setAttribute( attrColorF,   face, fColor );
          brep.setAttribute( attrEmission, face, fEmiss );
          brep.setTriangleVertices( face = brep.createFace(), curr[j-1], prev[j  ], curr[j  ] );
          brep.setAttribute( attrColorF,   face, fColor );
          brep.setAttribute( attrEmission, face, fEmiss );
        }

        for ( k = 0; k < 3; k++ )
          coord[k] += h[k];
      }

      for ( k = 0; k < 3; k++ )
      {
        left[k]  += vl[k];
        right[k] += vr[k];
      }
    }
  }

  /**
   * Creates Cornell-box external walls.
   */
  protected void box ()
  {
    double[][] vert =
    {
      { 556.0, 548.8,   0.0 },
      { 552.8,   0.0,   0.0 },
      {   0.0,   0.0,   0.0 },
      {   0.0, 548.8,   0.0 },
      { 556.0, 548.8, 559.2 },
      { 549.6,   0.0, 559.2 },
      {   0.0,   0.0, 559.2 },
      {   0.0, 548.8, 559.2 },
    };

    int[][] ind =
    {
      { 5, 6, 2, 1, 0 },  // floor
      { 3, 7, 4, 0, 0 },  // ceiling
      { 4, 7, 6, 5, 0 },  // back wall
      { 7, 3, 2, 6, 1 },  // right wall
      { 0, 4, 5, 1, 2 },  // left wall
    };

    for ( int[] i : ind )
      createQuad( vert[i[0]], vert[i[1]], vert[i[2]], vert[i[3]],
                  cellsUBox, cellsVBox,
                  color[3], color[i[4]], color[3] );
  }

  /**
   * Creates rectangular light (near the ceiling).
   */
  protected void light ()
  {
    double[][] vert =
    {
      { 278.0 + ( lightWidth / 2 ), 548.79, 279.5 - ( lightHeight / 2 ) },
      { 278.0 - ( lightWidth / 2 ), 548.79, 279.5 - ( lightHeight / 2 ) },
      { 278.0 - ( lightWidth / 2 ), 548.79, 279.5 + ( lightHeight / 2 ) },
      { 278.0 + ( lightWidth / 2 ), 548.79, 279.5 + ( lightHeight / 2 ) },
    };

    createQuad( vert[0], vert[1], vert[2], vert[3],
                cellsUBox, cellsVBox,
                color[5], color[5], color[4] );
  }

  /**
   * Creates short block inside the box.
   */
  protected void shortBlock ()
  {
    double[][] vert =
    {
      { 290.0, 165.0, 114.0 },
      { 290.0,   0.0, 114.0 },
      { 130.0,   0.0,  65.0 },
      { 130.0, 165.0,  65.0 },
      { 240.0, 165.0, 272.0 },
      { 240.0,   0.0, 272.0 },
      {  82.0,   0.0, 225.0 },
      {  82.0, 165.0, 225.0 },
    };

    int[][] ind =
    {
      { 0, 4, 7, 3 },  // top face
      { 5, 4, 0, 1 },  // left face
      { 1, 0, 3, 2 },  // front face
      { 2, 3, 7, 6 },  // right face
      { 5, 4, 7, 6 }   // back face
    };

    for ( int[] i : ind )
      createQuad( vert[i[0]], vert[i[1]], vert[i[2]], vert[i[3]],
                  cellsUBlock, cellsVBlock,
                  color[3], color[0], color[3] );
  }

  /**
   * Creates tall block inside the box.
   */
  protected void tallBlock ()
  {
    double[][] vert =
    {
      { 423.0, 330.0, 247.0 },
      { 423.0,   0.0, 247.0 },
      { 265.0,   0.0, 296.0 },
      { 265.0, 330.0, 296.0 },
      { 472.0, 330.0, 406.0 },
      { 472.0,   0.0, 406.0 },
      { 314.0,   0.0, 456.0 },
      { 314.0, 330.0, 456.0 },
    };

    int[][] ind =
    {
      { 7, 3, 0, 4 },
      { 4, 0, 1, 5 },
      { 7, 4, 5, 6 },
      { 3, 7, 6, 2 },
      { 0, 3, 2, 1 }
    };

    for ( int[] i : ind )
      createQuad( vert[i[0]], vert[i[1]], vert[i[2]], vert[i[3]],
                  cellsUBlock, cellsVBlock,
                  color[3], color[0], color[3] );
  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Triggers the (undetermined) action.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  @Override
  public boolean fire ( int type )
  {
    prepare( (Brep)getInterface( PL_OUTPUT, IFACE+"Brep" ), brep.CTX_DEFAULT );

    light();
    box();
    switch ( interior )
    {
      case SHORT_BLOCK:
        shortBlock();
        break;

      case TWO_BLOCKS:
        shortBlock();

      case TALL_BLOCK:
        tallBlock();
        break;

      default:
    }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String AUTOMATIC_TESS       = "Automatic";
  public final static String NUMBER_ROWS_BOX      = "Rows box";
  public final static String NUMBER_COLS_BOX      = "Cols box";
  public final static String NUMBER_ROWS_BLOCK    = "Rows block";
  public final static String NUMBER_COLS_BLOCK    = "Cols block";
  public final static String MAX_SIDE_LENGTH      = "Max side";
  public final static String LEFT_SURFACE_COLOR   = "Left color";
  public final static String RIGHT_SURFACE_COLOR  = "Right color";
  public final static String OTHER_SURFACES_COLOR = "Other color";
  public final static String LIGHT_WIDTH          = "Light width";
  public final static String LIGHT_HEIGHT         = "Light height";
  public final static String INTERIOR             = "Interior";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // AUTOMATIC_TESS:
    if ( key.compareTo(AUTOMATIC_TESS) == 0 )
    {
      automaticTess = booleanProperty( value, automaticTess );
      return;
    }
      // NUMBER_ROWS_BOX:
    if ( key.compareTo(NUMBER_ROWS_BOX) == 0 )
    {
      cellsUBox = intProperty( value, cellsUBox );
      return;
    }
      // NUMBER_COLS_BOX:
    if ( key.compareTo(NUMBER_COLS_BOX) == 0 )
    {
      cellsVBox = intProperty( value, cellsVBox );
      return;
    }
      // NUMBER_ROWS_BLOCK:
    if ( key.compareTo(NUMBER_ROWS_BLOCK) == 0 )
    {
      cellsUBlock = intProperty( value, cellsUBlock );
      return;
    }
      // NUMBER_COLS_BLOCK:
    if ( key.compareTo(NUMBER_COLS_BLOCK) == 0 )
    {
      cellsVBlock = intProperty( value, cellsVBlock );
      return;
    }
      // MAX_SIDE_LENGTH:
    if ( key.compareTo(MAX_SIDE_LENGTH) == 0 )
    {
      maxSideLength = doubleProperty( value, maxSideLength );
      return;
    }
      // LEFT_SURFACE_COLOR:
    if ( key.compareTo(LEFT_SURFACE_COLOR) == 0 )
    {
      leftSurfaceColor = colorProperty( value, leftSurfaceColor );
      return;
    }
      // RIGHT_SURFACE_COLOR:
    if ( key.compareTo(RIGHT_SURFACE_COLOR) == 0 )
    {
      rightSurfaceColor = colorProperty( value, rightSurfaceColor );
      return;
    }
      // OTHER_SURFACES_COLOR:
    if ( key.compareTo(OTHER_SURFACES_COLOR) == 0 )
    {
      otherSurfacesColor = colorProperty( value, otherSurfacesColor );
      return;
    }
      // LIGHT_WIDTH:
    if ( key.compareTo(LIGHT_WIDTH) == 0 )
    {
      lightWidth = doubleProperty( value, lightWidth );
      return;
    }
      // LIGHT_HEIGHT:
    if ( key.compareTo(LIGHT_HEIGHT) == 0 )
    {
      lightHeight = doubleProperty( value, lightHeight );
      return;
    }
      // BOXES:
    if ( key.compareTo(INTERIOR) == 0 )
    {
      interior = enumProperty( value, interior );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(AUTOMATIC_TESS) == 0 )
      return automaticTess;
    if ( key.compareTo(NUMBER_ROWS_BOX) == 0 )
      return cellsUBox;
    if ( key.compareTo(NUMBER_COLS_BOX) == 0 )
      return cellsVBox;
    if ( key.compareTo(NUMBER_ROWS_BLOCK) == 0 )
      return cellsUBlock;
    if ( key.compareTo(NUMBER_COLS_BLOCK) == 0 )
      return cellsVBlock;
    if ( key.compareTo(MAX_SIDE_LENGTH) == 0 )
      return maxSideLength;
    if ( key.compareTo(LEFT_SURFACE_COLOR) == 0 )
      return leftSurfaceColor;
    if ( key.compareTo(RIGHT_SURFACE_COLOR) == 0 )
      return rightSurfaceColor;
    if ( key.compareTo(OTHER_SURFACES_COLOR) == 0 )
      return otherSurfacesColor;
    if ( key.compareTo(LIGHT_WIDTH) == 0 )
      return lightWidth;
    if ( key.compareTo(LIGHT_HEIGHT) == 0 )
      return lightHeight;
    if ( key.compareTo(INTERIOR) == 0 )
      return interior;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Cornell Box generator";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToBrep";

  /** Object description. */
  private final static String DESCRIPTION = "Creates classical Cornell Box.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_DATA + '.' + C_SCENE;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"Trigger" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"Brep" );
        // Property:
      t.propBegin( AUTOMATIC_TESS, TYPE_BOOLEAN, "Do automatic tesselation?", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( NUMBER_ROWS_BOX, TYPE_INTEGER, "Number of rows at box wall", true );
        t.propDefault( 1 );
      t.propEnd();
      t.propBegin( NUMBER_COLS_BOX, TYPE_INTEGER, "Number of columns at box wall", true );
        t.propDefault( 1 );
      t.propEnd();
      t.propBegin( NUMBER_ROWS_BLOCK, TYPE_INTEGER, "Number of rows at block face", true );
        t.propDefault( 1 );
      t.propEnd();
      t.propBegin( NUMBER_COLS_BLOCK, TYPE_INTEGER, "Number of columns at block face", true );
        t.propDefault( 1 );
      t.propEnd();
      t.propBegin( MAX_SIDE_LENGTH, TYPE_DOUBLE, "Maximal triangle-side length", true );
        t.propDefault( 100.0 );
      t.propEnd();
      t.propBegin( LEFT_SURFACE_COLOR, TYPE_OBJECT, "Color of the left wall", true );
        t.propDefault( "[255;0;0]" );
        t.propManipulator( "ParamColor" );
      t.propEnd();
      t.propBegin( RIGHT_SURFACE_COLOR, TYPE_OBJECT, "Color of the right wall", true );
        t.propDefault( "[0;255;0]" );
        t.propManipulator( "ParamColor" );
      t.propEnd();
      t.propBegin( OTHER_SURFACES_COLOR, TYPE_OBJECT, "Color of all other faces", true );
        t.propDefault( "[255;255;255]" );
        t.propManipulator( "ParamColor" );
      t.propEnd();
      t.propBegin( LIGHT_WIDTH, TYPE_DOUBLE, "Light width", true );
        t.propDefault( 130.0 );
      t.propEnd();
      t.propBegin( LIGHT_HEIGHT, TYPE_DOUBLE, "Light height", true );
        t.propDefault( 105.0 );
      t.propEnd();
      t.propBegin( INTERIOR, TYPE_INTEGER, "Interior type", true );
        t.propDefault( Interior.NONE.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Empty",       Interior.NONE.ordinal(),        "Empty interior" );
        t.propEnum( "Short block", Interior.SHORT_BLOCK.ordinal(), "Short block only" );
        t.propEnum( "Tall block",  Interior.TALL_BLOCK.ordinal(),  "Tall block only" );
        t.propEnum( "Two blocks",  Interior.TWO_BLOCKS.ordinal(),  "Both short and tall blocks" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
