package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Cube as elementary 3D solid.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.24
 * @author  Martin Petricek, PE
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/Cube.java">Cube.java</a>
 */
public class Cube extends DefaultSolid
{
  //==========================================================================
  //  Interface Solid:

  /**
   * Creates polyhedral representation of the solid (non-mandatory method).
   *
   * @param  brep Global B-rep data structure.
   * @param  smoothness Quality-factor of an approximation ({@code 0.0} for default
   *                    approximation).
   * @return Handle of the created solid (or {@link Brep#NULL} if the method is not
   *         implemented).
   * @see    Brep#NULL
   */
  public int createPolyhedr ( Brep brep, double smoothness )
  {
    final int context = Brep.CTX_DEFAULT;

    final int solid = brep.createSolid( context );

      // N=near, F=far, T=top, B=bottom, L=left, R=right
    final int faceN = brep.createFace();
    final int faceF = brep.createFace();
    final int faceT = brep.createFace();
    final int faceB = brep.createFace();
    final int faceL = brep.createFace();
    final int faceR = brep.createFace();

    final int edgeNT = brep.createEdge();
    final int edgeNB = brep.createEdge();
    final int edgeNL = brep.createEdge();
    final int edgeNR = brep.createEdge();
    final int edgeFT = brep.createEdge();
    final int edgeFB = brep.createEdge();
    final int edgeFL = brep.createEdge();
    final int edgeFR = brep.createEdge();
    final int edgeLT = brep.createEdge();
    final int edgeLB = brep.createEdge();
    final int edgeRT = brep.createEdge();
    final int edgeRB = brep.createEdge();

    final int vertexNLT = brep.createVertex();
    final int vertexNLB = brep.createVertex();
    final int vertexNRT = brep.createVertex();
    final int vertexNRB = brep.createVertex();
    final int vertexFLT = brep.createVertex();
    final int vertexFLB = brep.createVertex();
    final int vertexFRT = brep.createVertex();
    final int vertexFRB = brep.createVertex();

    brep.setVertexCoords( context, vertexNLT, new double[]{ -1.,  1., -1. } );
    brep.setVertexCoords( context, vertexNLB, new double[]{ -1., -1., -1. } );
    brep.setVertexCoords( context, vertexNRT, new double[]{  1.,  1., -1. } );
    brep.setVertexCoords( context, vertexNRB, new double[]{  1., -1., -1. } );
    brep.setVertexCoords( context, vertexFLT, new double[]{ -1.,  1.,  1. } );
    brep.setVertexCoords( context, vertexFLB, new double[]{ -1., -1.,  1. } );
    brep.setVertexCoords( context, vertexFRT, new double[]{  1.,  1.,  1. } );
    brep.setVertexCoords( context, vertexFRB, new double[]{  1., -1.,  1. } );

    brep.setEdgeVertices( edgeNT, vertexNLT, vertexNRT );
    brep.setEdgeVertices( edgeNB, vertexNLB, vertexNRB );
    brep.setEdgeVertices( edgeNL, vertexNLT, vertexNLB );
    brep.setEdgeVertices( edgeNR, vertexNRT, vertexNRB );
    brep.setEdgeVertices( edgeFT, vertexFLT, vertexFRT );
    brep.setEdgeVertices( edgeFB, vertexFLB, vertexFRB );
    brep.setEdgeVertices( edgeFL, vertexFLT, vertexFLB );
    brep.setEdgeVertices( edgeFR, vertexFRT, vertexFRB );
    brep.setEdgeVertices( edgeLT, vertexNLT, vertexFLT );
    brep.setEdgeVertices( edgeLB, vertexNLB, vertexFLB );
    brep.setEdgeVertices( edgeRT, vertexNRT, vertexFRT );
    brep.setEdgeVertices( edgeRB, vertexNRB, vertexFRB );

    brep.setFaceEdges( faceN, new int[] { edgeNT, edgeNL, edgeNB, edgeNR } );
    brep.setFaceEdges( faceF, new int[] { edgeFT, edgeFL, edgeFB, edgeFR } );
    brep.setFaceEdges( faceL, new int[] { edgeNL, edgeLT, edgeFL, edgeLB } );
    brep.setFaceEdges( faceR, new int[] { edgeNR, edgeRT, edgeFR, edgeRB } );
    brep.setFaceEdges( faceT, new int[] { edgeNT, edgeLT, edgeFT, edgeRT } );
    brep.setFaceEdges( faceB, new int[] { edgeNB, edgeLB, edgeFB, edgeRB } );

    brep.setSolidFaces( context, solid, new int[] { faceL, faceR, faceB, faceT, faceF, faceN } );

    return solid;
  }

  /**
   * Computes intersection of the solid with the given ray.
   *
   * @param  P0 Ray origin ({@code double[3]} or {@code double[4]}).
   * @param  P1 Direction vector of the ray ({@code double[3]} or {@code double[4]}).
   * @return List of intersection points ({@link MicroFacet}s) sorted by {@code t}
   *         (parametric distance from {@code P0}). {@code null} is returned in case
   *         of bad arguments or if no intersection exists..
   * @see    MicroFacet
   * @see    BoundingVolume
   */
  public List< MicroFacet > intersection ( double[] P0, double[] P1 )
  {
    int P0len;
      // valid input parameters?
    if ( P0 == null || (P0len = P0.length) < 3 ||
         P1 == null ||          P1.length  < 3 ) return null;

      // ray origin:
    double ox = P0[0];
    double oy = P0[1];
    double oz = P0[2];
    if ( P0len > 3 )
    {                                     // to cartesian
      double ow = P0[3];
      if ( Geometry.isZero( ow ) )
        return null;

      ow  = 1.0 / ow;
      ox *= ow;
      oy *= ow;
      oz *= ow;
    }

      // ray direction vector (of arbitrary size):
    final double dx = P1[0];
    final double dy = P1[1];
    final double dz = P1[2];

    double tNear = Double.NEGATIVE_INFINITY;
    double tFar  = Double.POSITIVE_INFINITY;
    double t1, t2, t3;

    if ( !Geometry.isZero( dx ) )
    {                                     // non-parallel
       // +-1 = +-r
      t2 = ( 1.0 - ox) / dx;
      t3 = (-1.0 - ox) / dx;
      t1 = Math.min( t2, t3 );
      t2 = Math.max( t2, t3 );

      if ( t1 > tNear ) tNear = t1;
      if ( t2 < tFar  ) tFar  = t2;

      if ( tFar < tNear ) return null;
    }

    if ( !Geometry.isZero( dy ) )
    {                                     // non-parallel
       // +-1 = +-r
      t2 = ( 1.0 - oy) / dy;
      t3 = (-1.0 - oy) / dy;
      t1 = Math.min( t2, t3 );
      t2 = Math.max( t2, t3 );

      if ( t1 > tNear ) tNear = t1;
      if ( t2 < tFar  ) tFar  = t2;

      if ( tFar < tNear ) return null;
    }

    if ( !Geometry.isZero( dz ) )
    {                                     // non-parallel
       // +-1 = +-r
      t2 = ( 1.0 - oz) / dz;
      t3 = (-1.0 - oz) / dz;
      t1 = Math.min( t2, t3 );
      t2 = Math.max( t2, t3 );

      if ( t1 > tNear ) tNear = t1;
      if ( t2 < tFar  ) tFar  = t2;

      if ( tFar < tNear ) return null;
    }

    LinkedList< MicroFacet > list = new LinkedList< MicroFacet >();
    MicroFacet mf;

    mf          = new MicroFacet();
    mf.t        = tNear;
    mf.enter    =
    mf.front    = true;
    mf.locCoord = new double[]
    {
      ox + mf.t * dx,
      oy + mf.t * dy,
      oz + mf.t * dz
    };
    mf.valid    = MicroFacet.VALID_LOC_COORD;
    list.addLast( mf );

    mf          = new MicroFacet();
    mf.t        = tFar;
    mf.enter    =
    mf.front    = false;
    mf.locCoord = new double[]
    {
      ox + mf.t * dx,
      oy + mf.t * dy,
      oz + mf.t * dz
    };
    mf.valid    = MicroFacet.VALID_LOC_COORD;
    list.addLast( mf );

    return list;
  }

  /**
   * Asserts validity of the given intersection-attributes.
   * Resolves: {@link MicroFacet#VALID_TANGENT}, {@link MicroFacet#VALID_TEX_COORD}.
   *
   * @param  mf The checked intersection.
   * @param  tmp Non-mandatory object holding state-information for deferred computations.
   * @param  attributes Attribute mask (required attributes - see
   *                    {@code MicroFacet.VALID_*}).
   * @return Attributes which were not computed.
   */
  @Override
  public int assertAttributes ( MicroFacet mf, Object tmp, int attributes )
  {
    if ( mf == null ) return attributes;

      // tangent surface vectors:
    if ( (attributes & MicroFacet.VALID_TANGENT) != 0 )
    {
      mf.tu = new double[ 3 ];
      mf.tv = new double[ 3 ];
      int coord = Formula.argMax( mf.locCoord );
      double[] o = new double[ 3 ];
      o[coord] = (mf.locCoord[coord] > 0.0) ? 1.0 : -1.0;
      Geometry.getAxes( o, mf.tu, mf.tv );
      if ( mf.toWorld != null )
      {
        mf.toWorld.transformVector3D( mf.tu, mf.tu );
        mf.toWorld.transformVector3D( mf.tv, mf.tv );
      }
      mf.valid |= MicroFacet.VALID_TANGENT;
      if ( (attributes -= MicroFacet.VALID_TANGENT) == 0 ) return 0;
    }

      // 2D texture coordinates (inverse mapping):
    if ( (attributes & MicroFacet.VALID_TEX_COORD) != 0 )
    {
      int coord = Formula.argMax( mf.locCoord );
      mf.texCoord = new double[ 2 ];
      double r = Math.abs( mf.locCoord[coord] ); // cube size
      int texc = 0;
      for ( int i = 0; i < 3; i++ )
      {
        if ( i == coord ) i++;            // skip
        if ( i >= 3 ) break;
        mf.texCoord[texc] = (mf.locCoord[i] + r) / (2*r);
                                          // 0 .. 1
        texc++;
      }
      mf.valid |= MicroFacet.VALID_TEX_COORD;
      if ( (attributes -= MicroFacet.VALID_TEX_COORD) == 0 ) return 0;
    }

    return attributes;
  }

  /**
   * Point vs. solid test (&quot;inside-test&quot;).
   *
   * @param  point Coordinates of the 3D point ({@code double[3]} or
   *               {@code double[4]}).
   * @return {@code true} if the given point lies inside the solid.
   */
  @Override
  public boolean isInside ( double[] point )
  {
    if ( point == null || point.length < 3 ) return false;

    double rr = (point.length > 3) ? (point[3] * point[3]) : 1.0;
    for ( int i = 0; i < 3; i++ )
      if ( Math.abs( point[i] ) > rr ) return false;

    return true;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Cube";

  /** Object description. */
  private final static String DESCRIPTION = "Cube as a 3D solid for CSG.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"Solid" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
