package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Discrete cosine transform (can process 1D or 2D data of arbitrary size).
 *
 * @version 0.25 $Rev: 246 $ $Date: 2006-05-11 01:38:39 +0200 (čt, 11 V 2006) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/DCT.java">DCT.java</a>
 */
public class DCT extends DefaultDiscreteTransform
{
  //==========================================================================
  //  Data:

  /** Transform bases for last computed transform sizes (accelerator). */
  protected double[][] base = new double[ 2 ][];

  /** Last computed transform sizes (accelerator). */
  protected int[]      len  = { 0, 0 };

  /** Last cache to be recycled (<tt>0</tt> or <tt>1</tt>. */
  protected int        last;

  /** Temporary array for in-place transforms. */
  protected double[]   tmpD;

  //==========================================================================
  //  Support:

  /** Computes base matrix for the given size. */
  protected double[] assertBase ( int l )
  {
      // check the cache first:
    if ( l == len[0] ) return base[0];
    if ( l == len[1] ) return base[1];

      // recycle the last-th array:
    int newLen = l * l;
    if ( base[last] == null ||
         base[last].length < newLen )
      base[last] = new double[ newLen ];

    double[] result = base[last];
    len[last]       = l;
    last            = 1 - last;

      // compute transform matrix of size l * l into array result[]:
    double ck = 1.0 / Math.sqrt( l );
    double ci;
    int n, k, i = 0;

      // 1st row: DC
    for ( n = 0; n++ < l; )
      result[i++] = ck;

      // 2nd to l-th rows: AC
    ck = Math.sqrt( 2.0 / l );
    for ( k = 1; k < l; k++ )
    {
      ci = (Math.PI * k) / (l + l);
      for ( n = 0; n < l; n++ )
        result[i++] = ck * Math.cos( (n + n + 1.0) * ci );
    }

    return result;
  }

  //==========================================================================
  //  Interface DiscreteTransform:

  /**
   * Get extreme domain (source) ranges.
   *
   * @param  bounds Pre-allocated array to hold extreme ranges:
   *         <br><tt>bounds[0]</tt> .. minimum size in 1st dim,
   *         <br><tt>bounds[1]</tt> .. maximum size in 1st dim,
   *         <br><tt>bounds[2]</tt> .. minimum size in 2nd dim,
   *         <br><tt>bounds[3]</tt> .. maximum size in 2nd dim, etc.
   * @return <tt>true</tt> if arbitrary lengths are allowed (even non-powers-ow-two).
   */
  public boolean getSizeBounds ( int[] bounds )
  {
    if ( bounds != null )
    {
      int i;
      for ( i = 0; i + 1 < bounds.length; )
      {
        bounds[ i++ ] =     1;
        bounds[ i++ ] = 16384;
      }
    }
    return true;
  }

  /**
   * Forward 1D transform (floating-point form).
   * Has to be used if source and/or target data are floating-point!
   *
   * @param  sample Source array (containing sample data).
   * @param  start Starting index of the transformed segment.
   * @param  len Source segment length.
   * @param  coef Pre-allocated array to hold transformed data (coefficients).
   * @return Number of coefficients (<tt>0</tt> in case of error).
   * @see    #inverse1D(double[],int,double[],int)
   * @see    #transform1D(int[],int,int,int[])
   */
  public int transform1D ( double[] sample, int start, int len, double[] coef )
  {
    if ( sample == null ||
         sample.length < start + len ||
         coef   == null ||
         coef.length < len )
      return 0;

    if ( len <= 1 )
    {
      coef[0] = sample[start];
      return 1;
    }

    if ( sample == coef &&
         start < len )
    {                                     // temporary array must be used
      if ( tmpD == null || tmpD.length < len )
        tmpD = new double[ len ];

      System.arraycopy( sample, start, tmpD, 0, len );
      sample = tmpD;
      start  = 0;
    }

    double[] b = assertBase( len );
    double acc;
    int i, j, k, ptr;
    for ( j = k = 0; j < len; j++ )
    {
      acc = 0.0;
      ptr = start;
      for ( i = 0; i++ < len; )
        acc += b[k++] * sample[ptr++];
      coef[j] = acc;
    }

    return len;
  }

  /**
   * Inverse 1D transform (floating-point form).
   * Inverse mapping to {@link #transform1D(double[],int,int,double[])}.
   *
   * @param  coef Source array (holding coefficients).
   * @param  len Number of coefficients.
   * @param  sample Pre-allocated array to hold reconstructed sample data.
   * @param  start Starting index of the reconstructed sample data.
   * @return Sample data length (<tt>0</tt> in case of error).
   * @see    #transform1D(double[],int,int,double[])
   * @see    #inverse1D(int[],int,int[],int)
   */
  public int inverse1D ( double[] coef, int len, double[] sample, int start )
  {
    if ( sample == null ||
         sample.length < start + len ||
         coef   == null ||
         sample.length < start + len )
      return 0;

    if ( len <= 1 )
    {
      sample[start] = coef[0];
      return 1;
    }

    if ( sample == coef &&
         start < len )
    {                                     // temporary array must be used
      if ( tmpD == null || tmpD.length < len )
        tmpD = new double[ len ];

      System.arraycopy( coef, 0, tmpD, 0, len );
      coef = tmpD;
    }

    double[] b = assertBase( len );
    double acc;
    int i, j, k;
    for ( j = 0; j < len; j++ )
    {
      acc = 0.0;
      for ( i = 0, k = j; i < len; k += len )
        acc += b[k] * coef[i++];
      sample[start++] = acc;
    }

    return len;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Discrete cosine transform";

  /** Object description. */
  private final static String DESCRIPTION = "Discrete cosine transform (naive implementation). " +
                                            "Can process 1D or 2D data of arbitrary size.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"DiscreteTransform" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
