package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Module holding N-dimensional data matrix.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/DataMatrix.java">DataMatrix.java</a>
 */
public class DataMatrix extends Piece implements SampleData
{
  //==========================================================================
  //  Data:

  /** Actual number of sample-array dimensions. */
  protected int dims;

  /** Actual dimension sizes (dim.length == dims). */
  protected int[] dim;

  /** Accelerators: submatrix sizes (accel[i] = Prod_{j=i..N} dim[j]). */
  protected int[] accel;

  /** Total 1D packed array size (equal to {@code accel[0]}). */
  protected int totalSize;

  /** Actual sample type. If {@link #map} {@code != null}, the actual data array has to be {@link #SAMPLE_INT}. */
  protected int sampleType = SAMPLE_INT;

  /** Sample-array geometry (1 double for one dimension = uniform distance between samples). */
  protected double[] geometry;

  /** The data array itself (byte format). */
  protected byte[] dataByte;

  /** The data array itself (byte format). */
  protected short[] dataShort;

  /** The data array itself (byte format). */
  protected int[] dataInt;

  /** The data array itself (byte format). */
  protected long[] dataLong;

  /** The data array itself (byte format). */
  protected float[] dataFloat;

  /** The data array itself (byte format). */
  protected double[] dataDouble;

  /** Data mapping (used together with integer data matrix). */
  protected Object map;

  //==========================================================================
  //  Support:

  /** Returns internal packed 1D array of the apropriate type. */
  protected Object dataArray ()
  {
    if ( map != null ) return dataInt;

    switch ( sampleType )
    {
      case SAMPLE_BYTE:
        return dataByte;

      case SAMPLE_SHORT:
        return dataShort;

      case SAMPLE_INT:
      case SAMPLE_RGBA:
        return dataInt;

      case SAMPLE_LONG:
        return dataLong;

      case SAMPLE_FLOAT:
        return dataFloat;

      case SAMPLE_DOUBLE:
      case SAMPLE_COMPLEX:
        return dataDouble;

      default:
        return null;
    }
  }

  /** Determines array's component type (returns null if there is a problem). */
  protected Class<?> getComponentClass ( Object array )
  {
    Class<?> cl, comp;
      // check data array type:
    if ( array == null ||
       !(cl = array.getClass()).isArray() ) return null;

    if ( !(comp = cl.getComponentType()).isPrimitive() ) return null;

      // check compatibility of the component type:
    if ( map != null )
      return( comp.equals( Integer.TYPE ) ? comp : null );

    switch ( sampleType )
    {
      case SAMPLE_BYTE:
        if ( comp.equals( Byte.TYPE ) )
          return comp;
        break;

      case SAMPLE_SHORT:
        if ( comp.equals( Short.TYPE ) )
          return comp;
        break;

      case SAMPLE_INT:
      case SAMPLE_RGBA:
        if ( comp.equals( Integer.TYPE ) )
          return comp;
        break;

      case SAMPLE_LONG:
        if ( comp.equals( Long.TYPE ) )
          return comp;
        break;

      case SAMPLE_FLOAT:
        if ( comp.equals( Float.TYPE ) )
          return comp;
        break;

      case SAMPLE_DOUBLE:
      case SAMPLE_COMPLEX:
        if ( comp.equals( Double.TYPE ) )
          return comp;
        break;
    }

    return null;
  }

  /** Computes item index of ind[0] .. ind[dims-2] (w/o the last - fastest - index). */
  protected final int countIndex ( int[] ind )
  {
    int dstInd = 0;                       // destination index
    for ( int i = 1; i < dims; i++ )
      dstInd += accel[i] * ind[i-1];
    return dstInd;
  }

  //==========================================================================
  //  Interface SampleData:

  /**
   * (Re-)initialize the data array.
   *
   * @param  type Sample type (one of {@code SAMPLE_*} constants).
   * @param  mapped Should the array contain mapped values?
   * @param  dim Dimensions of an array.
   * @see    #SAMPLE_BYTE
   * @see    #SAMPLE_SHORT
   * @see    #SAMPLE_INT
   * @see    #SAMPLE_LONG
   * @see    #SAMPLE_FLOAT
   * @see    #SAMPLE_DOUBLE
   * @see    #SAMPLE_COMPLEX
   * @see    #SAMPLE_RGBA
   */
  public void resetData ( int type, boolean mapped, int[] dim )
  {
    if ( dim == null || dim.length == 0 ) return;

    dims       = dim.length;
    this.dim   = dim;
    accel      = new int[ dims ];
    totalSize  = (type == SAMPLE_COMPLEX) ? 2 : 1;
    int i;
    for ( i = dims; --i >= 0; )
      accel[i] = totalSize *= dim[i];

    dataByte   = null;
    dataShort  = null;
    dataInt    = null;
    dataFloat  = null;
    dataDouble = null;

    if ( mapped )
      dataInt = new int[ totalSize ];

    switch ( sampleType = type )
    {
      case SAMPLE_BYTE:
        if ( mapped )
          map = new byte[ 2 ];
        else
          dataByte = new byte[ totalSize ];
        break;

      case SAMPLE_SHORT:
        if ( mapped )
          map = new short[ 2 ];
        else
          dataShort = new short[ totalSize ];
        break;

      case SAMPLE_LONG:
        if ( mapped )
          map = new long[ 2 ];
        else
          dataLong = new long[ totalSize ];
        break;

      case SAMPLE_FLOAT:
        if ( mapped )
          map = new float[ 2 ];
        else
          dataFloat = new float[ totalSize ];
        break;

      case SAMPLE_DOUBLE:
      case SAMPLE_COMPLEX:
        if ( mapped )
          map = new double[ (sampleType == SAMPLE_DOUBLE) ? 2 : 4 ];
        else
          dataDouble = new double[ totalSize ];
        break;

      case SAMPLE_INT:
      case SAMPLE_RGBA:
      default:
        if ( mapped )
          map = new int[ 2 ];
        else
          dataInt = new int[ totalSize ];
    }

  }

  /**
   * Returns actual sample type ({@code SAMPLE_*} constant).
   *
   * @return Sample type ({@code SAMPLE_*} constant).
   * @see    #getDataDim
   * @see    #SAMPLE_BYTE
   * @see    #SAMPLE_SHORT
   * @see    #SAMPLE_INT
   * @see    #SAMPLE_LONG
   * @see    #SAMPLE_FLOAT
   * @see    #SAMPLE_DOUBLE
   * @see    #SAMPLE_COMPLEX
   * @see    #SAMPLE_RGBA
   */
  public int getSampleType ()
  {
    return sampleType;
  }

  /**
   * Returns actual array dimensions.
   *
   * @param  dim Pre-allocated array to hold individual dimensions' sizes.
   * @return Dimensionality (number of dimensions).
   * @see    #getSampleType
   */
  public int getDataDim ( int[] dim )
  {
    if ( this.dim        == null ||
         this.dim.length == 0 ) return 0;

    if ( dim != null &&
         dim.length >= dims )
      System.arraycopy( this.dim, 0, dim, 0, dims );

    return dims;
  }

  /**
   * Sets the sample geometry (in regular sample matrix).
   *
   * @param geom Sample periods for all dimensions.
   * @see   #getSampleGeometry
   */
  public void setSampleGeometry ( double[] geom )
  {
    if ( geom        == null ||
         geom.length < dims ) return;

    geometry = geom;
  }

  /**
   * Gets the sample geometry (in regular sample matrix).
   *
   * @param geom Sample periods for all dimensions.
   * @see   #setSampleGeometry
   */
  public void getSampleGeometry ( double[] geom )
  {
    if ( geometry == null ||
         geom == null ||
         geom.length < dims ) return;

    System.arraycopy( geometry, 0, geom, 0, dims );
  }

  /**
   * Sets (partially) the new sample mapping (for mapped arrays only).
   *
   * @param map Mapping array to be set (should be an array of the apropriate type).
   * @param srcFrom First source index ({@code map[from]}).
   * @param srcTo After the last source index.
   * @param dstFrom First destination index.
   * @see   #getSampleMapping
   */
  public void setSampleMapping ( Object map, int srcFrom, int srcTo, int dstFrom )
  {
    // !!! TODO: !!!
  }

  /**
   * Retrieves the whole sample mapping (for mapped arrays only).
   *
   * @param map Optional pre-allocated mapping array to be filled.
   * @param srcFrom First index to be retrieved.
   * @param srcTo After the last index to be retrieved.
   * @param bounds Retrieved index bounds ({@code bounds[0]} = the 1st mapped value
   *               written in map[0], {@code bounds[1]} = index after the last value).
   * @see   #setSampleMapping
   */
  public Object getSampleMapping ( Object map, int srcFrom, int srcTo, int[] bounds )
  {
    // !!! TODO: !!!
    return null;
  }

  /**
   * Returns the original data array.
   *
   * @param  dim Returned dimensionality (array of individual dimension sizes). Can be null.
   * @return Packed 1D sample array (read / write access) or {@code null} if the service is not
   *         available.
   * @see    #setDataArray
   */
  public Object getDataArray ( int[] dim )
  {
    if ( this.dim == null ||
         dim != null && dim.length < dims ) return null;

    if ( dim != null )
      System.arraycopy( this.dim, 0, dim, 0, dims );

    return dataArray();
  }

  /**
   * Sets data array in the raw (internal) format. Can be used for data fragment
   * or for the whole data array.
   *
   * @param data Packed 1D source data array (of apropriate type - integer has to be used for mapped types).
   * @param srcFrom Source index to begin from.
   * @param dstFrom Start of the region to be set. Can be {@code null} (instead of {@code [0,0,0,..]}).
   * @param dstTo End of the region to be set. Can be {@code null} (instead of {@code dim}).
   */
  public void setDataArray ( Object data, int srcFrom, int[] dstFrom, int[] dstTo )
  {
      // check data array type:
    if ( getComponentClass( data ) == null )
      return;

      // target indices:
    int[] ind = new int[ dims ];
    if ( dstFrom != null )
    {
      if ( dstFrom.length < dims ) return;
      System.arraycopy( dstFrom, 0, ind, 0, dims );
    }
    else
      dstFrom = new int[ dims ];

    if ( dstTo == null )
      dstTo = dim;
    else
      if ( dstTo.length < dims ) return;
      // start .. dstFrom, actual .. ind, stop .. dstTo

    int ty = (map == null) ? SAMPLE_INT : sampleType;
    int i;
    int row = dstTo[dims-1] - ind[dims-1]; // source row size
    int lastIndex = ind[dims-1];
    if ( ty == SAMPLE_COMPLEX ) lastIndex <<= 1;

    switch ( ty )
    {
      case SAMPLE_BYTE:
      {
        byte[] src = (byte[])data;
        do
        {
          System.arraycopy( src, srcFrom,
                            dataByte, countIndex( ind ) + lastIndex,
                            row );
          srcFrom += row;

            // destination index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= dstTo[i] )
          {
            ind[i] = dstFrom[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_SHORT:
      {
        short[] src = (short[])data;
        do
        {
          System.arraycopy( src, srcFrom,
                            dataShort, countIndex( ind ) + lastIndex,
                            row );
          srcFrom += row;

            // destination index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= dstTo[i] )
          {
            ind[i] = dstFrom[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_INT:
      case SAMPLE_RGBA:
      {
        int[] src = (int[])data;
        do
        {
          System.arraycopy( src, srcFrom,
                            dataInt, countIndex( ind ) + lastIndex,
                            row );
          srcFrom += row;

            // destination index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= dstTo[i] )
          {
            ind[i] = dstFrom[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_LONG:
      {
        long[] src = (long[])data;
        do
        {
          System.arraycopy( src, srcFrom,
                            dataLong, countIndex( ind ) + lastIndex,
                            row );
          srcFrom += row;

            // destination index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= dstTo[i] )
          {
            ind[i] = dstFrom[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_FLOAT:
      {
        float[] src = (float[])data;
        do
        {
          System.arraycopy( src, srcFrom,
                            dataFloat, countIndex( ind ) + lastIndex,
                            row );
          srcFrom += row;

            // destination index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= dstTo[i] )
          {
            ind[i] = dstFrom[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_DOUBLE:
      {
        double[] src = (double[])data;
        do
        {
          System.arraycopy( src, srcFrom,
                            dataDouble, countIndex( ind ) + lastIndex,
                            row );
          srcFrom += row;

            // destination index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= dstTo[i] )
          {
            ind[i] = dstFrom[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_COMPLEX:
      {
        double[] src = (double[])data;
        row <<= 1;
        do
        {
          System.arraycopy( src, srcFrom,
                            dataDouble, countIndex( ind ) + lastIndex,
                            row );
          srcFrom += row;

            // destination index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= dstTo[i] )
          {
            ind[i] = dstFrom[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }
    }
  }

  /**
   * Reads the given sample segment (general data format).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result (packed 1D array).
   * @return Number of returned samples.
   */
  public int getSamplesGeneral ( int[] from, int[] to, Object data )
  {
      // check data array type:
    if ( getComponentClass( data ) == null )
      return -1;

      // target indices:
    int[] ind = new int[ dims ];
    if ( from != null )
    {
      if ( from.length < dims ) return -1;
      System.arraycopy( from, 0, ind, 0, dims );
    }
    else
      from = new int[ dims ];

    if ( to == null )
      to = dim;
    else
      if ( to.length < dims ) return -1;
      // start .. from, actual .. ind, stop .. to

    int ty = (map == null) ? SAMPLE_INT : sampleType;
    int i, ptr = 0;
    int row = to[dims-1] - ind[dims-1];   // source row size
    int lastIndex = ind[dims-1];
    if ( ty == SAMPLE_COMPLEX ) lastIndex <<= 1;

    switch ( ty )
    {
      case SAMPLE_BYTE:
      {
        byte[] dst = (byte[])data;
        do
        {
          System.arraycopy( dataByte, countIndex( ind ) + lastIndex,
                            dst, ptr,
                            row );
          ptr += row;

            // index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= to[i] )
          {
            ind[i] = from[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_SHORT:
      {
        short[] dst = (short[])data;
        do
        {
          System.arraycopy( dataShort, countIndex( ind ) + lastIndex,
                            dst, ptr,
                            row );
          ptr += row;

            // index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= to[i] )
          {
            ind[i] = from[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_INT:
      case SAMPLE_RGBA:
      {
        int[] dst = (int[])data;
        do
        {
          System.arraycopy( dataInt, countIndex( ind ) + lastIndex,
                            dst, ptr,
                            row );
          ptr += row;

            // index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= to[i] )
          {
            ind[i] = from[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_LONG:
      {
        long[] dst = (long[])data;
        do
        {
          System.arraycopy( dataLong, countIndex( ind ) + lastIndex,
                            dst, ptr,
                            row );
          ptr += row;

            // index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= to[i] )
          {
            ind[i] = from[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_FLOAT:
      {
        float[] dst = (float[])data;
        do
        {
          System.arraycopy( dataFloat, countIndex( ind ) + lastIndex,
                            dst, ptr,
                            row );
          ptr += row;

            // index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= to[i] )
          {
            ind[i] = from[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_DOUBLE:
      {
        double[] dst = (double[])data;
        do
        {
          System.arraycopy( dataDouble, countIndex( ind ) + lastIndex,
                            dst, ptr,
                            row );
          ptr += row;

            // index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= to[i] )
          {
            ind[i] = from[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }

      case SAMPLE_COMPLEX:
      {
        double[] dst = (double[])data;
        row <<= 1;
        do
        {
          System.arraycopy( dataDouble, countIndex( ind ) + lastIndex,
                            dst, ptr,
                            row );
          ptr += row;

            // index increment:
          i = dims - 2;
          while ( i >= 0 && ++ind[i] >= to[i] )
          {
            ind[i] = from[i];
            i--;
          }
        } while ( i >= 0 );
        break;
      }
    }

    return ptr;
  }

  /**
   * Writes the given sample segment (general data format).
   *
   * @param from Starting index.
   * @param to Index after the written data.
   * @param data Data to be written - in 1D array (integer indices for mapped data arrays).
   */
  public void setSamplesGeneral ( int[] from, int[] to, Object data )
  {
    setDataArray( data, 0, from, to );
  }

  /**
   * Reads the given sample segment (integer formats: byte, short, int, RGBA).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result.
   * @return Number of returned samples.
   */
  public int getSamples ( int[] from, int[] to, int[] data )
  {
    if ( dataInt   == null &&
         dataShort == null &&
         dataByte  == null ) return -1;

      // target indices:
    int[] ind = new int[ dims ];
    if ( from != null )
    {
      if ( from.length < dims ) return -1;
      System.arraycopy( from, 0, ind, 0, dims );
    }
    else
      from = new int[dims];

    if ( to == null )
      to = dim;
    else
      if ( to.length < dims ) return -1;
      // start .. from, actual .. ind, stop .. to

    int i, j, datai, ptr = 0;
    int row = to[dims-1] - ind[dims-1];   // source row size
    int lastIndex = ind[dims-1];

    do
    {
      datai = countIndex( ind ) + lastIndex;

        // copy the data:
      if ( dataInt != null )
      {
        System.arraycopy( dataInt, datai,
                          data, ptr,
                          row );
        ptr += row;
      }
      else
      if ( dataShort != null )
        for ( i = datai, j = 0; j++ < row; )
          data[ptr++] = dataShort[i++];
      else
        for ( i = datai, j = 0; j++ < row; )
          data[ptr++] = dataByte[i++];

        // index increment:
      i = dims - 2;
      while ( i >= 0 && ++ind[i] >= to[i] )
      {
        ind[i] = from[i];
        i--;
      }
    } while ( i >= 0 );

    return ptr;
  }

  /**
   * Reads the given sample segment (long format).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result.
   * @return Number of returned samples.
   */
  public int getSamples ( int[] from, int[] to, long[] data )
  {
    if ( dataLong == null ) return -1;

      // target indices:
    int[] ind = new int[ dims ];
    if ( from != null )
    {
      if ( from.length < dims ) return -1;
      System.arraycopy( from, 0, ind, 0, dims );
    }
    else
      from = new int[dims];

    if ( to == null )
      to = dim;
    else
      if ( to.length < dims ) return -1;
      // start .. from, actual .. ind, stop .. to

    int i, ptr = 0;
    int row = to[dims-1] - ind[dims-1];   // source row size
    int lastIndex = ind[dims-1];

    do
    {
      System.arraycopy( dataLong, countIndex( ind ) + lastIndex,
                        data, ptr,
                        row );
      ptr += row;

        // index increment:
      i = dims - 2;
      while ( i >= 0 && ++ind[i] >= to[i] )
      {
        ind[i] = from[i];
        i--;
      }
    } while ( i >= 0 );

    return ptr;
  }

  /**
   * Reads the given sample segment (float formats: float, double, complex).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result (complex uses double-sized array).
   * @return Number of returned samples.
   */
  public int getSamples ( int[] from, int[] to, double[] data )
  {
    if ( dataDouble == null &&
         dataFloat  == null ) return -1;

      // target indices:
    int[] ind = new int[ dims ];
    if ( from != null )
    {
      if ( from.length < dims ) return -1;
      System.arraycopy( from, 0, ind, 0, dims );
    }
    else
      from = new int[ dims ];

    if ( to == null )
      to = dim;
    else
      if ( to.length < dims ) return -1;
      // start .. from, actual .. ind, stop .. to

    int i, j, datai, ptr = 0;
    int row = to[dims-1] - ind[dims-1];   // source row size
    if ( sampleType == SAMPLE_COMPLEX )
      row <<= 1;
    int lastIndex = (sampleType == SAMPLE_COMPLEX) ? (ind[dims-1] << 1) : ind[dims-1];

    do
    {
      datai = countIndex( ind ) + lastIndex;

        // copy the data:
      if ( dataDouble != null )
      {
        System.arraycopy( dataDouble, datai,
                          data, ptr,
                          row );
        ptr += row;
      }
      else
        for ( i = datai, j = 0; j++ < row; )
          data[ptr++] = dataFloat[i++];

        // index increment:
      i = dims - 2;
      while ( i >= 0 && ++ind[i] >= to[i] )
      {
        ind[i] = from[i];
        i--;
      }
    } while ( i >= 0 );

    return ptr;
  }

  /**
   * Writes the given sample segment (integer formats: byte, short, int, RGBA).
   *
   * @param from Starting index.
   * @param to Index after written data.
   * @param data Data to be written (1D array).
   */
  public void setSamples ( int[] from, int[] to, int[] data )
  {
    if ( dataInt   == null &&
         dataShort == null &&
         dataByte  == null ) return;

     // target indices:
   int[] ind = new int[ dims ];
   if ( from != null )
   {
     if ( from.length < dims ) return;
     System.arraycopy( from, 0, ind, 0, dims );
   }
   else
     from = new int[ dims ];

   if ( to == null )
     to = dim;
   else
     if ( to.length < dims ) return;
     // start .. from, actual .. ind, stop .. to

   int i, j, datai, ptr = 0;
   int row = to[dims-1] - ind[dims-1];   // source row size
   int lastIndex = ind[dims-1];

   do
   {
     datai = countIndex( ind ) + lastIndex;

       // copy the data:
     if ( dataInt != null )
     {
       System.arraycopy( data, ptr,
                         dataInt, datai,
                         row );
       ptr += row;
     }
     else
     if ( dataShort != null )
       for ( i = datai, j = 0; j++ < row; )
         dataShort[i++] = (short)data[ptr++];
     else
       for ( i = datai, j = 0; j++ < row; )
         dataByte[i++] = (byte)data[ptr++];

       // index increment:
     i = dims - 2;
     while ( i >= 0 && ++ind[i] >= to[i] )
     {
       ind[i] = from[i];
       i--;
     }
   } while ( i >= 0 );
  }

  /**
   * Writes the given sample segment (long format).
   *
   * @param from Starting index.
   * @param to Index after written data.
   * @param data Data to be written (1D array).
   */
  public void setSamples ( int[] from, int[] to, long[] data )
  {
    if ( dataLong == null ) return;

      // target indices:
    int[] ind = new int[ dims ];
    if ( from != null )
    {
      if ( from.length < dims ) return;
      System.arraycopy( from, 0, ind, 0, dims );
    }
    else
      from = new int[ dims ];

    if ( to == null )
      to = dim;
    else
      if ( to.length < dims ) return;
      // start .. from, actual .. ind, stop .. to

    int i, ptr = 0;
    int row = to[dims-1] - ind[dims-1];   // source row size
    int lastIndex = ind[dims-1];

    do
    {
      System.arraycopy( data, ptr,
                        dataLong, countIndex( ind ) + lastIndex,
                        row );
      ptr += row;

        // index increment:
      i = dims - 2;
      while ( i >= 0 && ++ind[i] >= to[i] )
      {
        ind[i] = from[i];
        i--;
      }
    } while ( i >= 0 );
  }

  /**
   * Writes the given sample segment (float formats: float, double, complex).
   *
   * @param from Starting index.
   * @param to Index after written data.
   * @param data Data to be written (1D array). Complex uses double-sized array.
   */
  public void setSamples ( int[] from, int[] to, double[] data )
  {
    if ( dataDouble == null &&
         dataFloat  == null ) return;

      // target indices:
    int[] ind = new int[ dims ];
    if ( from != null )
    {
      if ( from.length < dims ) return;
      System.arraycopy( from, 0, ind, 0, dims );
    }
    else
      from = new int[ dims ];

    if ( to == null )
      to = dim;
    else
      if ( to.length < dims ) return;
      // start .. from, actual .. ind, stop .. to

    int i, j, datai, ptr = 0;
    int row = to[dims-1] - ind[dims-1];   // source row size
    if ( sampleType == SAMPLE_COMPLEX )
      row <<= 1;
    int lastIndex = (sampleType == SAMPLE_COMPLEX) ? (ind[dims-1] << 1) : ind[dims-1];

    do
    {
      datai = countIndex( ind ) + lastIndex;

        // copy the data:
      if ( dataDouble != null )
      {
        System.arraycopy( data, ptr,
                          dataDouble, datai,
                          row );
        ptr += row;
      }
      else
        for ( i = datai, j = 0; j++ < row; )
          dataFloat[i++] = (float)data[ptr++];

        // index increment:
      i = dims - 2;
      while ( i >= 0 && ++ind[i] >= to[i] )
      {
        ind[i] = from[i];
        i--;
      }
    } while ( i >= 0 );
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // SAMPLE_TYPE:
    if ( key.compareTo(SAMPLE_TYPE) == 0 )
    {
      sampleType = intProperty( value, sampleType, SAMPLE_BYTE, SAMPLE_RGBA );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(SAMPLE_TYPE) == 0 )
      return sampleType;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "DataMatrix";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "SampleData";

  /** Object description. */
  private final static String DESCRIPTION = "Simple data stored in N-dimensional matrix";

  /** Object category. */
  protected final static String CATEGORY = "nd.data";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"SampleData" );
        // Property:
      t.propBegin( SAMPLE_TYPE, TYPE_INTEGER, "Data item type", true );
        t.propDefault( SAMPLE_INT );
        t.propBounds( SAMPLE_BYTE, SAMPLE_RGBA );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Byte",    SAMPLE_BYTE,    "byte (8-bit signed int)" );
        t.propEnum( "Short",   SAMPLE_SHORT,   "short (16-bit signed int)" );
        t.propEnum( "Int",     SAMPLE_INT,     "int (32-bit signed int)" );
        t.propEnum( "Long",    SAMPLE_LONG,    "long (64-bit signed int)" );
        t.propEnum( "Float",   SAMPLE_FLOAT,   "float (32-bit float)" );
        t.propEnum( "Double",  SAMPLE_DOUBLE,  "double (64-bit float)" );
        t.propEnum( "Complex", SAMPLE_COMPLEX, "complex (two 64-bit floats)" );
        t.propEnum( "RGBA",    SAMPLE_RGBA,    "color in RGBA format (4 x 8bits packed in one 32-bit integer)" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
