package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Linear quantizer with optional dead-zone.
 *
 * @version 0.26 $Rev: 472 $ $Date: 2007-05-31 09:33:37 +0200 (čt, 31 V 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/DeadZoneQuantizer.java">DeadZoneQuantizer.java</a>
 */
public class DeadZoneQuantizer extends Piece implements ScalarQuantizer
{
  //==========================================================================
  //  Data:

  /** Quantizer-step size. */
  protected double delta    = 1.0;

  /** Dead-zone size. */
  protected double deadZone = 0.0;

  /** Accelerator: inverse of quantizer-step. */
  protected double coef     = 1.0;

  //==========================================================================
  //  Interface ScalarQuantizer:

  /**
   * Set actual quantizer variant (context).
   *
   * @param  variant Variant number (from {@code 0}).
   * @return Old variant number.
   * @see    #setBounds
   */
  public int setVariant ( int variant )
  {
    return 0;
  }

  /**
   * Sets input and output bounds.
   *
   * @param  inMin Dead-zone size = origin of linear reconstruction line.
   * @param  inMax Maximum value = center of maximum value step (end of linear reconstruction line).
   * @param  outLevels Number of output levels in the positive axis.
   * @see    #setVariant
   */
  public void setBounds ( double inMin, double inMax, int outLevels )
  {
      // inMin -> 0
      // inMax -> outLevels-1
      // delta = (inMax - inMin) / (outLevels - 1)
    deadZone = inMin;
    if ( inMax <= inMin )
      inMax = inMin + 1.0;
    if ( outLevels < 2 )
      outLevels = 2;

    delta = (inMax - inMin) / (outLevels - 1.0);
    coef  = 1.0 / delta;
  }

  /**
   * Quantization (floating-point form).
   *
   * @param  in Input floating-point value.
   * @return Quantized value (or {@code -1} if failed).
   * @see    #quantize(int)
   */
  public int quantize ( double in )
  {
    final boolean sign = in < 0.0;
    if ( sign ) in = -in;
    int level = Formula.round( (in - deadZone) * coef );
    return( sign ? -level : level );
  }

  /**
   * Quantization (integer form).
   *
   * @param  in Input integer value.
   * @return Quantized value (or {@code -1} if failed).
   * @see    #quantize(double)
   */
  public int quantize ( int in )
  {
    return quantize( (double)in );
  }

  /**
   * Dequantization = reconstruction (floating-point form).
   *
   * @param  quant Quantized value.
   * @return Source (reconstructed) value.
   * @see    #dequantizeInt(int)
   * @see    #quantize(double)
   */
  public double dequantizeDouble ( int quant )
  {
    if ( quant == 0 ) return 0.0;

    final boolean sign = quant < 0;
    if ( sign ) quant = -quant;
    double result = deadZone + delta * quant;

    return( sign ? -result : result );
  }

  /**
   * Dequantization = reconstruction (integer form).
   *
   * @param  quant Quantized value.
   * @return Source (reconstructed) value.
   * @see    #dequantizeDouble(int)
   * @see    #quantize(int)
   */
  public int dequantizeInt ( int quant )
  {
    if ( quant == 0 ) return 0;

    final boolean sign = quant < 0;
    if ( sign ) quant = -quant;
    int result = Formula.round( deadZone + delta * quant );

    return( sign ? -result : result );
  }

  //==========================================================================
  //  Interface PersistentParamsBinary:

  /**
   * Pack instance parameters into binary format.
   *
   * @return Data array to be stored into binary stream. Can be {@code null}.
   *         The array can be shared by all method calls.
   * @see    #loadData
   */
  public byte[] storeData ()
  {
    byte[] arr  = new byte[ 8 ];

      // deadZone: 32 bits
    Formula.storeFloat( arr, 0, (float)deadZone );

      // delta: 32 bits
    Formula.storeFloat( arr, 4, (float)delta );

    return arr;
  }

  /**
   * Loads instance parameters from (previously stored) binary data.
   *
   * @param  data Binary data to be interpreted. Can be {@code null}.
   * @return {@code true} if successful.
   * @see    #storeData
   */
  public boolean loadData ( byte[] data )
  {
    if ( data == null ||
         data.length < 8 ) return false;

      // deadZone: 32 bits
    deadZone = Formula.loadFloat( data, 0 );

      // delta: 32 bits
    delta    = Formula.loadFloat( data, 4 );

    coef     = 1.0 / delta;
    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String DEAD_ZONE = "Minimum";
  public final static String STEP      = "Step";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // DEAD_ZONE:
    if ( key.compareTo(DEAD_ZONE) == 0 )
    {
      deadZone = doubleProperty( value, deadZone );
      return;
    }
      // STEP:
    if ( key.compareTo(STEP) == 0 )
    {
      delta = doubleProperty( value, delta );
      coef  = 1.0 / delta;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(DEAD_ZONE) == 0 )
      return deadZone;
    if ( key.compareTo(STEP) == 0 )
      return delta;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Dead-zone quantizer";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ScalarQuantizer";

  /** Object description. */
  private final static String DESCRIPTION = "Linear quantizer with optional dead-zone.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + ".quantization";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"ScalarQuantizer" );
        // Property:
      t.propBegin( DEAD_ZONE, TYPE_DOUBLE, "Dead-zone size (on the positive axis)", true );
        t.propDefault( 0.0 );
      t.propEnd();
      t.propBegin( STEP, TYPE_DOUBLE, "Step size", true );
        t.propDefault( 1.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
