package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * R<sup>2</sup>->R<sup>2</sup> mappings (analytic warping sample transforms).
 *
 * @version 0.26 $Rev: 431 $ $Date: 2007-04-29 22:42:22 +0200 (ne, 29 IV 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/Deformations.java">Deformations.java</a>
 */
public class Deformations extends DefaultFunctionR2ToR2
{
  //==========================================================================
  //  Data:

  /** Implemented deformation types. */
  protected static enum DeformationType
  {
    /** Zoom. */
    DEF_ZOOM,
    /** Rotate. */
    DEF_ROTATE,
    /** Cubic twist deformation. */
    DEF_TWIST_CUBIC,
    /** Round twist deformation. */
    DEF_TWIST_ROUND,
    /** Magnifying-glass deformation. */
    DEF_MAGNI_GLASS,
  }

  /** Actual function type. */
  protected DeformationType type = DeformationType.DEF_ZOOM;

  /** General deformation parameter (zoom factor, angle, twist amplitude..). */
  protected double factor        = 1.0;

  /** Aspect ratio of the source space. */
  protected double aspectRatio   = 1.0;

  /** Support variables (accelerators?). */
  protected double tmp1, tmp2, tmp3, tmp4;

  /** Dirty flag = had to recompute internal accelerators? */
  protected boolean dirty        = true;

  //==========================================================================
  //  Support:

  /** Internal cubic-twist mapping function. */
  protected void cubicTwist ( boolean forward, double x, double y, double[] res )
  {
    if ( dirty )
    {
      tmp1  = factor / 90.0;              // target angle in 90-degree units
      tmp3  = 0.5 * aspectRatio;
      dirty = false;
    }

    double ax    = 2.0 * (x - tmp3) / aspectRatio;
    double ay    = 2.0 * (y - 0.5);
    double absAx = Math.abs( ax );
    double absAy = Math.abs( ay );
    double r     = Math.max( absAx, absAy );

    double angle;                         // actual angle
    if ( r == absAx )                     // 0 || 2
    {
      tmp2 = 0.5 * (1.0 + ay / r);
      if ( ax > 0 )                       // 0
        angle = 1.0 - tmp2;
      else                                // 2
        angle = 2.0 + tmp2;
    }
    else                                  // 1 || 3
    {
      tmp2 = 0.5 * (1.0 + ax / r);
      if ( ay > 0 )                       // 3
        angle = 3.0 + tmp2;
      else                                // 1
        angle = 2.0 - tmp2;
    }

    double dAngle = tmp1 * (1.0 - r);     // angle delta: in 90-degree units, positive = counterclockwise

    angle = 0.25 * (forward ? angle + dAngle : angle - dAngle);
    angle = 4.0 * (angle - Math.floor( angle ));
      // ... angle is now between 0.0 and 4.0

    ax = Math.floor( angle );
    ay = r * (2.0 * (angle - ax) - 1.0);

    switch ( (int)ax )
    {
      case 0:
        res[0] = tmp3 * (1.0 + r);
        res[1] = 0.5 * (1.0 - ay);
        break;

      case 1:
        res[0] = tmp3 * (1.0 - ay);
        res[1] = 0.5 * (1.0 - r);
        break;

      case 2:
        res[0] = tmp3 * (1.0 - r);
        res[1] = 0.5 * (1.0 + ay);
        break;

      case 3:
        res[0] = tmp3 * (1.0 + ay);
        res[1] = 0.5 * (1.0 + r);
        break;

    }
  }

  protected void setAspectRatio ()
  {
    RasterGraphics image = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    if ( image != null )
    {
      aspectRatio = (image.getWidth() - 1.0) / (image.getHeight() - 1.0);
      dirty = true;
    }
  }

  //==========================================================================
  //  Interface FunctionR2ToR2:

  /**
   * Compute the mapping: <b>f(x,y)</b>.
   *
   * @param  x X coordinate (1st argument).
   * @param  y Y coordinate (2nd argument).
   * @param  res Result array ({@code res[0]} = U coordinate,
   *             {@code res[1]} = V coordinate).
   */
  @Override
  public void f ( double x, double y, double[] res )
  {
    if ( res == null || res.length < 2 ) return;

    switch ( type )
    {
      case DEF_ZOOM:
      {
        if ( dirty )
        {
          tmp1  = 0.5 * aspectRatio;
          dirty = false;
        }
        double ax = x - tmp1;
        double ay = y - 0.5;
        res[0]    = tmp1 + factor * ax;
        res[1]    = 0.5  + factor * ay;
        return;
      }

      case DEF_ROTATE:
      {
        if ( dirty )
        {
          tmp1  = Math.cos( Math.toRadians( factor ) );
          tmp2  = Math.sin( Math.toRadians( factor ) );
          tmp3  = 0.5 * aspectRatio;
          dirty = false;
        }
        double ax = x - tmp3;
        double ay = y - 0.5;
        res[0]    = tmp3 + tmp1 * ax + tmp2 * ay;
        res[1]    = 0.5  - tmp2 * ax + tmp1 * ay;
        return;
      }

      case DEF_TWIST_CUBIC:
        cubicTwist( true, x, y, res );
        return;

      case DEF_TWIST_ROUND:
      {
        if ( dirty )
        {
          tmp1  = 0.5 * Math.min( 1.0, aspectRatio );
          tmp3  = 0.5 * aspectRatio;
          dirty = false;
        }
        double ax = (x - tmp3) / tmp1;
        double ay = (y - 0.5) / tmp1;
        double rr = ax * ax + ay * ay;
        if ( rr >= 1.0 )
        {
          res[0] = x;
          res[1] = y;
        }
        else
        {
          rr = Math.sqrt( rr );
          double angle = Math.toRadians( factor ) * (1.0 - rr);
          double cosa  = Math.cos( angle );
          double sina  = Math.sin( angle );
          res[0]       = tmp3 + tmp1 * ( cosa * ax + sina * ay);
          res[1]       = 0.5  + tmp1 * (-sina * ax + cosa * ay);
        }
        return;
      }

      case DEF_MAGNI_GLASS:
      {
        if ( dirty )
        {
          tmp1  = 0.5 * Math.min( 1.0, aspectRatio );
          tmp2  = 0.5 * (factor - 1.0);
          tmp3  = 0.5 * aspectRatio;
          tmp4  = 0.5 * (1.0 / factor - 1.0);
          dirty = false;
        }
        double ax  = (x - tmp3) / tmp1;
        double ay  = (y - 0.5)  / tmp1;
        double rr  = ax * ax + ay * ay;    // radius squared
        double mag = tmp1 * Math.exp( tmp2 * Math.log( rr ) );
        res[0]     = tmp3 + mag * ax;
        res[1]     = 0.5  + mag * ay;
        return;
      }

    }
  }

  /**
   * Compute the inverse mapping: <b>f<sup>-1</sup>(u,v)</b>.
   *
   * @param  u U coordinate (1st argument).
   * @param  v V coordinate (2nd argument).
   * @param  res Result array ({@code res[0]} = X coordinate,
   *             {@code res[1]} = Y coordinate).
   * @return {@code true} if this service is implemented.
   */
  @Override
  public boolean fInv ( double u, double v, double[] res )
  {
    if ( res == null || res.length < 2 ) return false;

    switch ( type )
    {
      case DEF_ZOOM:
      {
        if ( dirty )
        {
          tmp1  = 0.5 * aspectRatio;
          dirty = false;
        }
        double ax = u - tmp1;
        double ay = v - 0.5;
        res[0]    = tmp1 + ax / factor;
        res[1]    = 0.5  + ay / factor;
        return true;
      }

      case DEF_ROTATE:
      {
        if ( dirty )
        {
          tmp1  = Math.cos( Math.toRadians( factor ) );
          tmp2  = Math.sin( Math.toRadians( factor ) );
          tmp3  = 0.5 * aspectRatio;
          dirty = false;
        }
        double ax = u - tmp3;
        double ay = v - 0.5;
        res[0]    = tmp3 + tmp1 * ax - tmp2 * ay;
        res[1]    = 0.5  + tmp2 * ax + tmp1 * ay;
        return true;
      }

      case DEF_TWIST_CUBIC:
        cubicTwist( false, u, v, res );
        return true;

      case DEF_TWIST_ROUND:
      {
        if ( dirty )
        {
          tmp1  = 0.5 * Math.min( 1.0, aspectRatio );
          tmp3  = 0.5 * aspectRatio;
          dirty = false;
        }
        double ax = (u - tmp3) / tmp1;
        double ay = (v - 0.5)  / tmp1;
        double rr = ax * ax + ay * ay;
        if ( rr >= 1.0 )
        {
          res[0] = u;
          res[1] = v;
        }
        else
        {
          rr = Math.sqrt( rr );
          double angle = Math.toRadians( factor ) * (1.0 - rr);
          double cosa  = Math.cos( angle );
          double sina  = Math.sin( angle );
          res[0]       = tmp3 + tmp1 * (cosa * ax - sina * ay);
          res[1]       = 0.5  + tmp1 * (sina * ax + cosa * ay);
        }
        return true;
      }

      case DEF_MAGNI_GLASS:
      {
        if ( dirty )
        {
          tmp1  = 0.5 * Math.min( 1.0, aspectRatio );
          tmp2  = 0.5 * (factor - 1.0);
          tmp3  = 0.5 * aspectRatio;
          tmp4  = 0.5 * (1.0 / factor - 1.0);
          dirty = false;
        }
        double ax  = (u - tmp3) / tmp1;
        double ay  = (v - 0.5)  / tmp1;
        double rr  = ax * ax + ay * ay;    // radius squared
        double mag = tmp1 * Math.exp( tmp4 * Math.log( rr ) );
        res[0]     = tmp3 + mag * ax;
        res[1]     = 0.5  + mag * ay;
        return true;
      }

    }

    return false;
  }

  /**
   * Returns domain of the mapping.
   *
   * @param  dom Domain array ({@code dom[0]} = Xmin, {@code dom[1]} = Xmax,
   *             {@code dom[2]} = Ymin, {@code dom[3]} = Ymax).
   * @return {@code true} if this service is implemented.
   */
  @Override
  public boolean getDomain ( double[] dom )
  {
    if ( dom == null || dom.length < 4 ) return false;

    setAspectRatio();
    dom[0] = dom[2] = 0.0;
    dom[1] = aspectRatio;
    dom[3] = 1.0;

    return true;
  }

  /**
   * Returns range of the mapping.
   *
   * @param  rng Range array ({@code rng[0]} = Umin, {@code rng[1]} = Umax,
   *             {@code rng[2]} = Vmin, {@code rng[3]} = Vmax).
   * @return {@code true} if this service is implemented.
   */
  @Override
  public boolean getRange ( double[] rng )
  {
    if ( rng == null || rng.length < 4 ) return false;

    setAspectRatio();
    rng[0] =
    rng[2] = 0.0;
    rng[1] = aspectRatio;
    rng[3] = 1.0;

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String DEFORMATION  = "Deformation";
  public final static String FACTOR       = "Factor";
  public final static String ASPECT_RATIO = "Aspect ratio";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // DEFORMATION:
    if ( key.compareTo(DEFORMATION) == 0 )
    {
      type  = enumProperty( value, type );
      dirty = true;
      return;
    }
      // FACTOR:
    if ( key.compareTo(FACTOR) == 0 )
    {
      factor = doubleProperty( value, factor );
      dirty  = true;
      return;
    }
      // ASPECT_RATIO:
    if ( key.compareTo(ASPECT_RATIO) == 0 )
    {
      aspectRatio = doubleProperty( value, aspectRatio, 1.e-4, 1.e4 );
      dirty       = true;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(DEFORMATION) == 0 )
      return type.ordinal();
    if ( key.compareTo(FACTOR) == 0 )
      return factor;
    if ( key.compareTo(ASPECT_RATIO) == 0 )
      return aspectRatio;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Deformations";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "FunctionR2ToR2ToRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Examples of deformation mappings in 2D.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_FUNCTION;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"FunctionR2ToR2" );
      t.newOptOutputPlug( PL_OUTPUT, IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( DEFORMATION, TYPE_INTEGER, "Deformation type", true );
        t.propDefault( DeformationType.DEF_ZOOM.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Zoom",        DeformationType.DEF_ZOOM.ordinal(),        "Centered zoom" );
        t.propEnum( "Rotation",    DeformationType.DEF_ROTATE.ordinal(),      "Centered rotation in degrees" );
        t.propEnum( "Cubic twist", DeformationType.DEF_TWIST_CUBIC.ordinal(), "Centered cubic twist in degrees" );
        t.propEnum( "Round twist", DeformationType.DEF_TWIST_ROUND.ordinal(), "Centered round twist in degrees" );
        t.propEnum( "Magni-glass", DeformationType.DEF_MAGNI_GLASS.ordinal(), "Centered magnifying glass" );
      t.propEnd();
      t.propBegin( FACTOR, TYPE_DOUBLE, "Parameter", true );
        t.propDefault( 1.0 );
      t.propEnd();
      t.propBegin( ASPECT_RATIO, TYPE_DOUBLE, "Aspect ratio", true );
        t.propDefault( 1.0 );
        t.propBounds(  1.e-4, 1.e4 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
