package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Image effect - blur.
 *
 * @author  PE, Jan Kolomazn&iacute;k
 * @version 0.26 $Rev: 450 $ $Date: 2007-05-08 19:11:49 +0200 (út, 08 V 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/EffectBlur.java">EffectBlur.java</a>
 */
public class EffectBlur extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Implemented blur types. */
  protected static enum BlurType
  {
    /** Radial blur. */
    BLUR_RADIAL,
    /** Tangential blur. */
    BLUR_TANGENTIAL,
    /** Linear blur (distance from the given line). */
    BLUR_LINEAR,
    /** Linear perpendicular blur (distance from the given line). */
    BLUR_PERPENDICULAR,
    /** Linear constant blur (motion blur). */
    BLUR_CONSTANT,
  }

  /** Actual blur type. */
  protected BlurType type = BlurType.BLUR_CONSTANT;

  /** Blur/line direction angle in degrees. */
  protected double angle = 10.0;

  /**
   * Blur magnitude:
   * {@link BlurType#BLUR_RADIAL}/{@link BlurType#BLUR_LINEAR}/{@link BlurType#BLUR_PERPENDICULAR}: radius/distance extension coefficient in %,
   * {@link BlurType#BLUR_TANGENTIAL}: blur size in degrees,
   * {@link BlurType#BLUR_CONSTANT}: blur size in pixels.
   */
  protected double size = 10.0;

  /** Center of blur (relative X coordinate). */
  protected double centerX = 0.5;

  /** Center of blur (relative Y coordinate). */
  protected double centerY = 0.5;

  /** Accelerator: input picture. */
  protected RasterGraphics input;

  /** Accelerator: input image width. */
  protected int width;

  /** Accelerator: input image height. */
  protected int height;

  /** Support: input pixel value. */
  protected final double[] iPixel = new double[ 3 ];

  /** Support: output pixel value. */
  protected final double[] oPixel = new double[ 3 ];

  //==========================================================================
  //  Support:

  /**
   * Averages input-image colors along the given line.
   * Result color is returned in {@link #oPixel}.
   */
  protected void lineAverage ( int x1, int y1, int x2, int y2 )
  {
    int D, ax, ay, sx, sy, count;

    sx = x2 - x1;
    ax = Math.abs( sx ) << 1;
    if ( sx < 0 ) sx = -1;
    else
      if ( sx > 0 ) sx = 1;

    sy = y2 - y1;
    ay = Math.abs( sy ) << 1;
    if ( sy < 0 ) sy = -1;
    else
      if ( sy > 0 ) sy = 1;

    input.getRGB( x1, y1, oPixel );       // 1st pixel or 0 if outside..
    count = (x1 >= 0 && x1 < width &&
             y1 >= 0 && y1 < height) ? 1 : 0;

    if ( ax > ay )                        // x coordinate is dominant
    {
      D  = ay - (ax >> 1);                // initial D
      ax = ay - ax;                       // ay = increment0; ax = increment1
      while ( x1 != x2 )
      {
        if ( D >= 0 )                     // lift up the Y coordinate
        {
          y1 += sy;
          D  += ax;
        }
        else
          D  += ay;

        x1 += sx;

        if ( x1 >= 0 && x1 < width &&
             y1 >= 0 && y1 < height )
        {
          count++;
          input.getRGB( x1, y1, iPixel );
          oPixel[0] += iPixel[0];
          oPixel[1] += iPixel[1];
          oPixel[2] += iPixel[2];
        }
      }
    }

    else                                  // y coordinate is dominant
    {
      D  = ax - (ay >> 1);                // initial D
      ay = ax - ay;                       // ax = increment0; ay = increment1
      while ( y1 != y2 )
      {
        if ( D >= 0 )                     // lift up the X coordinate
        {
          x1 += sx;
          D  += ay;
        }
        else
          D  += ax;

        y1 += sy;

        if ( x1 >= 0 && x1 < width &&
             y1 >= 0 && y1 < height )
        {
          count++;
          input.getRGB( x1, y1, iPixel );
          oPixel[0] += iPixel[0];
          oPixel[1] += iPixel[1];
          oPixel[2] += iPixel[2];
        }
      }
    }

    if ( count > 1 )
    {
      double mul = 1.0 / count;
      oPixel[0] *= mul;
      oPixel[1] *= mul;
      oPixel[2] *= mul;
    }
  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transform.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
                         input  = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    final RasterGraphics output = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );

    if ( input  == null ||
         output == null ) return false;

      // output image:
    width  = input.getWidth();
    height = input.getHeight();
    output.init( width, height, RasterGraphics.MODE_RGB, 3 );

      // temporary values:
    int x, y, x1, y1, x2, y2;
    final int x0 = Formula.round( centerX * width  );
    final int y0 = Formula.round( centerY * height );

    switch ( this.type )
    {
      case BLUR_RADIAL:                   // distance from center, relative coefficient
      {
        final double coef = 0.005 * size;
        int dxy;

        for( y = 0; y < height; y++ )
        {
          dxy = Formula.round( coef * (y0 - y) );
          y1  = y - dxy;
          y2  = y + dxy;
          for ( x = 0; x < width; x++ )
          {
            dxy = Formula.round( coef * (x0 - x) );
            lineAverage( x - dxy, y1, x + dxy, y2 );
            output.putPixel( x, y, oPixel );
          }
        }

        break;
      }

      case BLUR_TANGENTIAL:               // distance from center, relative coefficient, perpendicular
      {
        final double coef = 0.005 * size;
        int dx;

        for( y = 0; y < height; y++ )
        {
          dx = Formula.round( coef * (y0 - y) );
          for ( x = 0; x < width; x++ )
          {
            x1  = x - dx;
            x2  = x + dx;
            y2  = Formula.round( coef * (x - x0) );
            y1  = y - y2;
            y2 += y;
            lineAverage( x1, y1, x2, y2 );
            output.putPixel( x, y, oPixel );
          }
        }

        break;
      }

      case BLUR_LINEAR:                   // distance from the given line, relative coefficient
      {
        final double coef = 0.005 * size;
        final double mx   =  Math.cos( Math.toRadians( angle ) );
        final double my   = -Math.sin( Math.toRadians( angle ) );
          // sharp line: [ x0, y0 ] + t * [ mx, my ]
        double t;
        int dx, dy;

        for( y = 0; y < height; y++ )
          for ( x = 0; x < width; x++ )
          {
            t  = mx * (x - x0) + my * (y - y0);
            dx = Formula.round( coef * (x0 + t * mx - x) );
            dy = Formula.round( coef * (y0 + t * my - y) );
            lineAverage( x - dx, y - dy, x + dx, y + dy );
            output.putPixel( x, y, oPixel );
          }
        break;
      }

      case BLUR_PERPENDICULAR:            // distance from the given line, relative coefficient, perpendicular
      {
        final double coef = 0.005 * size;
        final double mx   =  Math.cos( Math.toRadians( angle ) );
        final double my   = -Math.sin( Math.toRadians( angle ) );
          // sharp line: [ x0, y0 ] + t * [ mx, my ]
        double t;
        int dx, dy;

        for( y = 0; y < height; y++ )
          for ( x = 0; x < width; x++ )
          {
            t  = mx * (x - x0) + my * (y - y0);
            dx = Formula.round( coef * (x0 + t * mx - x) );
            dy = Formula.round( coef * (y0 + t * my - y) );
            lineAverage( x - dy, y + dx, x + dy, y - dx );
            output.putPixel( x, y, oPixel );
          }
        break;
      }

      case BLUR_CONSTANT:                 // motion blur (the given direction)
      default:
      {
        final int dx =  Formula.round( 0.5 * size * Math.cos( Math.toRadians( angle ) ) );
        final int dy = -Formula.round( 0.5 * size * Math.sin( Math.toRadians( angle ) ) );

        for( y = 0; y < height; y++ )
          for ( x = 0; x < width; x++ )
          {
            lineAverage( x - dx, y - dy, x + dx, y + dy );
            output.putPixel( x, y, oPixel );
          }

        break;
      }
    }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  protected final static String FILTER_TYPE = "Filter type";
  protected final static String MAGNITUDE   = "Magnitude";
  protected final static String ANGLE       = "Angle";
  protected final static String CENTER_X    = "Center X";
  protected final static String CENTER_Y    = "Center Y";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // FILTER_TYPE:
    if ( key.compareTo(FILTER_TYPE) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
      // MAGNITUDE:
    if ( key.compareTo(MAGNITUDE) == 0 )
    {
      size = doubleProperty( value, size, 0.0, 100.0 );
      return;
    }
      // ANGLE:
    if ( key.compareTo(ANGLE) == 0 )
    {
      angle = doubleProperty( value, angle, 0.0, 100.0 );
      return;
    }
      // CENTER_X:
    if ( key.compareTo(CENTER_X) == 0 )
    {
      centerX = doubleProperty( value, centerX, 0.0, 1.0 );
      return;
    }
      // CENTER_Y:
    if ( key.compareTo(CENTER_Y) == 0 )
    {
      centerY = doubleProperty( value, centerY, 0.0, 1.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(FILTER_TYPE) == 0 )
      return type.ordinal();
    if ( key.compareTo(MAGNITUDE) == 0 )
      return size;
    if ( key.compareTo(ANGLE) == 0 )
      return angle;
    if ( key.compareTo(CENTER_X) == 0 )
      return centerX;
    if ( key.compareTo(CENTER_Y) == 0 )
      return centerY;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "EffectBlur";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterImageAndRasterImage";

  /** Object description. */
  private final static String DESCRIPTION = "Image blur: radial, tangential, linear and motion blur.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"Trigger" );
      t.newOutputPlug( "image1", IFACE+"RasterGraphics" );
      t.newOutputPlug( "image2", IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( FILTER_TYPE, TYPE_INTEGER, "Blur type", true );
        t.propDefault( BlurType.BLUR_CONSTANT.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "constant",      BlurType.BLUR_CONSTANT.ordinal(),      "Constant (motion) blur" );
        t.propEnum( "linear",        BlurType.BLUR_LINEAR.ordinal(),        "Linear blur (distance from defined line)" );
        t.propEnum( "perpendicular", BlurType.BLUR_PERPENDICULAR.ordinal(), "Perpendicular blur (distance from defined line)" );
        t.propEnum( "radial",        BlurType.BLUR_RADIAL.ordinal(),        "Radial blur" );
        t.propEnum( "tangential",    BlurType.BLUR_TANGENTIAL.ordinal(),    "Tangential blur" );
      t.propEnd();
      t.propBegin( MAGNITUDE, TYPE_DOUBLE, "Blur magnitude (in pixels/degrees)", true );
        t.propDefault( 10.0 );
        t.propBounds(  0.0, 100.0 );
      t.propEnd();
      t.propBegin( ANGLE, TYPE_DOUBLE, "Angle in degrees", true );
        t.propDefault( 10.0 );
        t.propBounds(  0.0, 100.0 );
      t.propEnd();
      t.propBegin( CENTER_X, TYPE_DOUBLE, "Center X (relative)", true );
        t.propDefault( 0.5 );
        t.propBounds(  0.0, 1.0 );
      t.propEnd();
      t.propBegin( CENTER_Y, TYPE_DOUBLE, "Center Y (relative)", true );
        t.propDefault( 0.5 );
        t.propBounds(  0.0, 1.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
