package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Image effect example - image pixelization.
 *
 * @version 0.26 $Rev: 450 $ $Date: 2007-05-08 19:11:49 +0200 (út, 08 V 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/EffectPixelize.java">EffectPixelize.java</a>
 */
public class EffectPixelize extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Cell size in pixels. */
  protected int cellSize = 8;

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transformation.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    final RasterGraphics input  = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    final RasterGraphics output = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );
    int i, j, cx, cy;

      // output image:
    int width  = input.getWidth();
    int height = input.getHeight();
    output.init( width, height, RasterGraphics.MODE_RGB, 3 );

      // support values:
    final int  cellsX = (width + cellSize - 1) / cellSize;
    final int  cellsY = (height + cellSize - 1) / cellSize;
    int[]      pixel  = new int[ 4 ];        // I/O pixel
    long[]     out    = new long[ 4 ];       // cell summation
    final long area   = cellSize * cellSize; // cell area
    final long areaH  = area >> 1;           // cell area rounding

    int x0, y0;
    for ( cy = y0 = 0; cy < cellsY; cy++, y0 += cellSize )   // one output cell row
      for ( cx = x0 = 0; cx < cellsX; cx++, x0 += cellSize ) // one output cell
      {
        out[0] = out[1] = out[2] = out[3] = 0L;
        for ( j = y0; j < y0 + cellSize; j++ )
          for ( i = x0; i < x0 + cellSize; i++ )
          {
            input.getRGB( i, j, pixel );
            out[0] += pixel[0];
            out[1] += pixel[1];
            out[2] += pixel[2];
            out[3] += pixel[3];
          }
        if ( x0 + cellSize <= width &&
             y0 + cellSize <= height )
        {                                 // the whole cell lies inside the image
          pixel[0] = (int)( (out[0] + areaH) / area );
          pixel[1] = (int)( (out[1] + areaH) / area );
          pixel[2] = (int)( (out[2] + areaH) / area );
          pixel[3] = (int)( (out[3] + areaH) / area );
        }
        else
        {                                 // this cell should be truncated
          long a = (x0 + cellSize > width)  ? width  - x0 : cellSize;
          long b = (y0 + cellSize > height) ? height - y0 : cellSize;
          a *= b;
          b = a >> 1;
          pixel[0] = (int)( (out[0] + b) / a );
          pixel[1] = (int)( (out[1] + b) / a );
          pixel[2] = (int)( (out[2] + b) / a );
          pixel[3] = (int)( (out[3] + b) / a );
        }
        output.setRectangle( x0, y0, x0+cellSize, y0+cellSize, pixel );
      }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String CELL_SIZE = "Cell size";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // CELL_SIZE:
    if ( key.compareTo(CELL_SIZE) == 0 )
    {
      cellSize = intProperty( value, cellSize );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(CELL_SIZE) == 0 )
      return cellSize;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "EffectPixelize";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Image pixelizer, replaces image rectangles by its average colors.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"Trigger" );
      t.newOutputPlug( "image1", IFACE+"RasterGraphics" );
      t.newOutputPlug( "image2", IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( CELL_SIZE, TYPE_INTEGER, "Cell size in pixels", true );
        t.propDefault( 8 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
