package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Ellipse drawing - two different algorithms for untilted (Bresenham) and tilted (Bond) ellipses.
 *
 * @author  Bohdan Kolecek, PE
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/Ellipse.java">Ellipse.java</a>
 */
public class Ellipse extends Piece implements EllipseRender
{
  //==========================================================================
  //  Data:

  /** Accelerator: attached output module. */
  protected BitMask out = null;

  //==========================================================================
  //  Support:

  /** Draws four symmetric ellipse points. */
  protected final void ellipsePoints ( int x0, int y0,
                                       int x,  int y )
  {
    out.setPixel( x0 - x, y0 - y );
    out.setPixel( x0 + x, y0 - y );
    out.setPixel( x0 - x, y0 + y );
    out.setPixel( x0 + x, y0 + y );
  }

  /** Support polynomial. */
  protected final static double ellipseFunction ( double A, double B, double C2,
                                                  double D, double E, double F,
                                                  double x, double y )
  {
    return( x * (A * x + C2 * y + D) +
            y * (B * y + E) + F );
  }

  //==========================================================================
  //  Interface LineRender:

  /**
   * Draws an axis-aligned ellipse with center [x,y] and radii a, b.
   *
   * @param  x0 X coordinate of ellipse center.
   * @param  y0 X coordinate of ellipse center.
   * @param  a Horizontal radius of the ellipse.
   * @param  b Vertical radius of the ellipse.
   */
  public void drawEllipse ( int x0, int y0, int a, int b )
  {
    if ( out == null )
    {
      out = (BitMask)getInterface( PL_OUTPUT, IFACE+"BitMask" );
      if ( out == null ) return;
    }

    final int aa  = a * a;
    final int aa2 = aa + aa;
    final int bb  = b * b;
    final int bb2 = bb + bb;

    int D  = bb - aa * b + (aa >> 2);
    int dx = bb2;
    int dy = aa * (b + b - 1);
    int x  = 0;
    int y  = b;

    ellipsePoints( x0, y0, x, y );

      // part I:
    while ( dx < dy )
    {
      if ( D >= 0 )
      {
        D  += aa - dy;
        dy -= aa2;
        y--;
      }

      D  += dx + bb;
      dx += bb2;
      x++;

      ellipsePoints( x0, y0, x, y );
    }

      // part II:
    D = bb * x * (x + 1) + (bb >> 2) + aa * ((y - 1) * (y - 1) - bb);

    while ( y > 0 )
    {
      if ( D < 0 )
      {
        D  += dx;
        dx += bb2;
        x++;
      }

      D  += aa2 - dy;
      dy -= aa2;
      y--;

      ellipsePoints( x0, y0, x, y );
    }
  }

  /**
   * Draws an axis-aligned ellipse with center [x,y] and radii a, b.
   *
   * @param  x0 X coordinate of ellipse center.
   * @param  y0 X coordinate of ellipse center.
   * @param  a Horizontal radius of the ellipse.
   * @param  b Vertical radius of the ellipse.
   */
  public void drawEllipse ( double x0, double y0, double a, double b )
  {
    drawEllipse( x0, y0, a, b, 0.0 );
  }

  /**
   * Draws a tilted ellipse with center [x,y] and radii a, b.
   *
   * @param  x0 X coordinate of ellipse center.
   * @param  y0 X coordinate of ellipse center.
   * @param  a Primary radius of the ellipse.
   * @param  b Secondary radius of the ellipse.
   * @param  tilt Tilt angle in radians. {@code 0.0} for axis-aligned ellipse.
   *              Positive orientation means counterclockwise.
   */
  public void drawEllipse ( double x0, double y0, double a, double b, double tilt )
  {
    if ( out == null )
    {
      out = (BitMask)getInterface( PL_OUTPUT, IFACE+"BitMask" );
      if ( out == null ) return;
    }

    /*
     * Following code is based on C. Bond's algorithm (http://www.crbond.com/papers/ellipse.pdf).
     *
     * This implementation takes into account real part of ellipse center coordinates.
     * To achieve that ellipse center coordinates translation cannot be deferred until
     * plot time. This makes following implementation different from the one suggested
     * in Bond's paper.
     */

      // Assert a >= b:
    if ( a < b )
    {
      double tmp = a;
      a = b;
      b = tmp;
      tilt += 0.5 * Math.PI;
    }

      // Determine ellipse coefficients:
    final double aa = a * a;
    final double bb = b * b;
    final double c = Math.sqrt( aa - bb );
    final double xc = c * Math.cos( -tilt );
    final double yc = c * Math.sin( -tilt );

    final double A = aa - xc * xc;
    final double B = aa - yc * yc;
    final double C = -xc * yc;
    final double C2 = C + C;
    final double D = -2 * (A * x0 + C * y0);
    final double E = -2 * (B * y0 + C * x0);
    final double F = -aa * bb + A * x0 * x0 + B * y0 * y0 + C2 * x0 * y0;

      // Compute critical points of the ellipse:
    final double rtA = Math.sqrt( A );
    final double rtB = Math.sqrt( B );
    final double rtABC1 = 1.0 / Math.sqrt( A + B - C2 );
    final double rtABC2 = 1.0 / Math.sqrt( A + B + C2 );

    final double xt  = x0 - C / rtA;
    final double yt  = y0 + rtA;
    final double xtr = x0 + (B - C) * rtABC1;
    final double ytr = y0 + (A - C) * rtABC1;

    final double xr  = x0 + rtB;
    final double yr  = y0 - C / rtB;
    final double xbr = x0 + (B + C) * rtABC2;
    final double ybr = y0 - (A + C) * rtABC2;

    final double xb  = x0 + C / rtA;
    final double yb  = y0 - rtA;
    final double xbl = x0 - (B - C) * rtABC1;
    final double ybl = y0 - (A - C) * rtABC1;

    final double xl  = x0 - rtB;
    final double yl  = y0 + C / rtB;
    final double xtl = x0 - (B + C) * rtABC2;
    final double ytl = y0 + (A + C) * rtABC2;

      // Draw ellipse:
    int x, y;

    x = Formula.round( xt );
    y = Formula.round( yt );
    while ( x < xtr )
    {
      out.setPixel( x++, y );
      if ( ellipseFunction( A, B, C2, D, E, F, x, y - 0.5 ) > 0 ) y--;
    }

    x = Formula.round( xtr );
    y = Formula.round( ytr );
    while ( y > yr )
    {
      out.setPixel( x, y-- );
      if ( ellipseFunction( A, B, C2, D, E, F, x + 0.5, y ) <= 0 ) x++;
    }

    x = Formula.round( xr );
    y = Formula.round( yr );
    while ( y > ybr )
    {
      out.setPixel( x, y-- );
      if ( ellipseFunction( A, B, C2, D, E, F, x - 0.5, y ) > 0 ) x--;
    }

    x = Formula.round( xbr );
    y = Formula.round( ybr );
    while ( x > xb )
    {
      out.setPixel( x--, y );
      if ( ellipseFunction( A, B, C2, D, E, F, x, y - 0.5 ) <= 0 ) y--;
    }

    x = Formula.round( xb );
    y = Formula.round( yb );
    while ( x > xbl )
    {
      out.setPixel( x--, y );
      if ( ellipseFunction( A, B, C2, D, E, F, x, y + 0.5 ) > 0 ) y++;
    }

    x = Formula.round( xbl );
    y = Formula.round( ybl );
    while ( y < yl )
    {
      out.setPixel( x, y++ );
      if ( ellipseFunction( A, B, C2, D, E, F, x - 0.5, y ) <= 0 ) x--;
    }

    x = Formula.round( xl );
    y = Formula.round( yl );
    while ( y < ytl )
    {
      out.setPixel( x, y++ );
      if ( ellipseFunction( A, B, C2, D, E, F, x + 0.5, y ) > 0 ) x++;
    }

    x = Formula.round( xtl );
    y = Formula.round( ytl );
    while ( x < xt )
    {
      out.setPixel( x++, y );
      if ( ellipseFunction( A, B, C2, D, E, F, x, y + 0.5 ) <= 0 ) y++;
    }
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Ellipse";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "EllipseRenderToBitMask";

  /** Object description. */
  private final static String DESCRIPTION = "Draws ellipse using integer/float coordinates.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DRAW + '.' + C_ELLIPSE + '.' + C_INTEGER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"EllipseRender" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"BitMask" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
