package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * {@link Brep} face rendering - flat shading.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/FaceRenderFlat.java">FaceRenderFlat.java</a>
 */
public class FaceRenderFlat extends Piece implements FaceRender
{
  //==========================================================================
  //  Data:

  /** Last used context (for attribute caches). */
  protected int lastCtx = Brep.NULL;

  /** Name of face-color attribute. */
  protected String colorName = Brep.COLOR;

  /** Actual face-color attribute. */
  protected int attrColor = Brep.NULL;

  /** Name of projected-coordinate attribute. */
  protected String coordsName = Brep.COORD_PROJ;

  /** Actual projected-coordinate attribute. */
  protected int attrCoords = Brep.NULL;

  /** Accelerator: input B-rep database. */
  protected Brep data;

  /** Accelerator: raster graphics output object. */
  protected RasterGraphics output;

  /** Accelerator: bitmask object (can be {@code null}). */
  protected BitMaskCore mask;

  /** Fill-counter for fill-outside mode. */
  protected int counter;

  /** Last used face id (for repetitive calls of {@link #fillHLine}). */
  protected int lastFace = Brep.NULL;

  /** Color used if no face-color attribute is defined. */
  protected final static double[] DEFAULT_COLOR = { 0.4, 0.9, 1.0 };

  /** Tmp: vertex handles for one face. */
  protected int[] vert  = new int[ 3 ];

  /** Tmp: vertex coordinates for one edge. */
  protected double[][] vc  = new double[ 2 ][ 4 ];

  /** Tmp: face color ({@code float} format). */
  protected float[] colorF = new float[ 4 ];

  /** Tmp: face color ({@code double} format). */
  protected double[] color = new double[ 3 ];

  /** Tmp: input edge set sorted by {@code [ y0 ]}. */
  protected TreeSet< PolyEdge > input  = new TreeSet< PolyEdge >();

  /** Tmp: actual edge list sorted by {@code [ x ]}. */
  protected TreeSet< PolyEdge > actual = new TreeSet< PolyEdge >( eComparator );

  /** Tmp: edge list for the next line. */
  protected TreeSet< PolyEdge > backup = new TreeSet< PolyEdge >( eComparator );

  //==========================================================================
  //  Support:

  /** Prepare acelerators, returns {@code true} if failed. */
  protected boolean accelerators ( int ctx, int face, boolean maskNeeded )
  {
    if ( data == null )                   // Brep database connected for the 1st time..
    {
      data       = (Brep)getInterface( PL_DATA, IFACE+"Brep" );
      attrColor  =
      attrCoords = Brep.NULL;
      if ( data == null ) return true;
    }

    if ( output == null )                 // raster output connected for the 1st time..
    {
      output = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
      if ( output == null ) return true;
    }

    if ( maskNeeded && mask == null )     // bit-mask data object (if needed)
      mask= (BitMaskCore) getInterface( PL_BITMASK, IFACE+"BitMaskCore" );

    if ( lastCtx    != ctx ||
         attrCoords == Brep.NULL )        // context change => get new attribute Id
    {
      attrCoords = data.getAttributeId( Brep.VERTEX, ctx, coordsName );
      if ( attrCoords == Brep.NULL ) return true;
    }

    if ( lastCtx   != ctx ||
         attrColor == Brep.NULL )         // context change => get new attribute Id
    {
      attrColor = data.getAttributeId( Brep.FACE, ctx, colorName );
      if ( attrColor == Brep.NULL )
        output.setColor( DEFAULT_COLOR );

      lastFace  = Brep.NULL;
    }

    if ( lastFace != face )               // set actual face color
    {
      lastFace = face;

      if ( attrColor != Brep.NULL )
      {
        data.getAttribute( attrColor, face, colorF );
        color[0] = colorF[0];
        color[1] = colorF[1];
        color[2] = colorF[2];
        output.setColor( color );
      }
    }

    lastCtx = ctx;
    return false;
  }

  //==========================================================================
  //  Support classes:

  /**
   * Polygon edge.
   */
  protected static class PolyEdge implements Comparable< PolyEdge >
  {
    /** Starting Y-coordinate. */
    public int y0;

    /** X-coordinate of the edge-hline intersection. */
    public double x;

    /** X difference between two adjacent hlines. */
    public double dx;

    /** Number of hlines to draw. */
    public int count;

    public PolyEdge ( double[] p1, double[] p2 )
    {
      double yy0, dy, yCorr;

      if ( p1[1] < p2[1] )                // top-down edge
      {
        dy    = p2[1] - p1[1];
        yy0   = Math.ceil( p1[1] );
        yCorr = yy0 - p1[1];
        y0    = Formula.round( yy0 );
        dx    = (p2[0] - p1[0]) / dy;
        x     = p1[0] + yCorr * dx;
        count = Formula.round( Math.ceil( p2[1] ) - yy0 );
      }

      else                                // bottom-up edge
      {
        dy    = p1[1] - p2[1];
        yy0   = Math.ceil( p2[1] );
        yCorr = yy0 - p2[1];
        y0    = Formula.round( yy0 );
        dx    = (p1[0] - p2[0]) / dy;
        x     = p2[0] + yCorr * dx;
        count = Formula.round( Math.ceil( p1[1] ) - yy0 );
      }
    }

    /** Prepare the edge record for the next line. */
    public void next ()
    {
      x += dx;
    }

    /** Natural ordering: [ y0, x ]. */
    public int compareTo ( PolyEdge o )
    {
      if ( o == null )
        throw new NullPointerException();

      if ( this == o )
        return 0;

      if ( y0 != o.y0 )
        return( y0 < o.y0 ? -1 : 1 );

      if ( x != o.x )
        return( x < o.x ? -1 : 1 );

      return( hashCode() < o.hashCode() ? -1 : 1 );
    }
  }

  /** Edge comparator for actual edge list: [ x ]. */
  protected static class EdgeComparator implements Comparator< PolyEdge >
  {
    public int compare ( PolyEdge a, PolyEdge b )
    {
      if ( a == b )
        return 0;

      if ( a.x != b.x )
        return( a.x < b.x ? -1 : 1 );

      return( a.hashCode() < b.hashCode() ? -1 : 1 );
    }
  }

  /** Single comparator instance. */
  protected final static EdgeComparator eComparator = new EdgeComparator();

  /**
   * Render horizontal segment (H-line) from {@code [ e1.x, y ]} to {@code [ e2.x, y ]}.
   * For inside-fill mode.
   */
  protected void hLineInside ( int y, PolyEdge e1, PolyEdge e2 )
  {
    int x1 = Formula.round( Math.ceil( e1.x ) );
    int x2 = Formula.round( Math.ceil( e2.x ) );

    if ( x1 >= x2 ) return;

    if ( mask == null )
    {
      output.setHLine( x1, x2, y );
      return;
    }

    do
      if ( mask.getPixel( x1, y ) )
        output.setPixel( x1, y );
    while ( ++x1 < x2 );
  }

  /**
   * Render horizontal segment (H-line) from {@code [ e1.x, y ]} to {@code [ e2.x, y ]}.
   * For outside-fill mode.
   */
  protected void hLineOutside ( int y, PolyEdge e1, PolyEdge e2 )
  {
    int x1 = Formula.round( Math.ceil( e1.x ) );
    int x2 = Formula.round( Math.ceil( e2.x ) );

    if ( x1 >= x2 ) return;

    if ( mask == null )
    {
      output.setHLine( x1, x2, y );
      counter += x2 - x1;
      return;
    }

    final int bakX1 = x1;

    do
      if ( !mask.getPixel( x1, y ) )
      {
        output.setPixel( x1, y );
        counter++;
      }
    while ( ++x1 < x2 );

    mask.setHLine( bakX1, x2, y );
  }

  /**
   * Fills the given face, common routine for both inside- and outside- fill modes.
   * Must be called after successfull {@link #accelerators} call!
   *
   * @return Number of drawn pixels (valid in outside-fill mode only).
   */
  protected int commonFill ( int ctx, int face, boolean inside )
  {
    int faceSize;
    int i, i1;

      // 1. initialize all the lists:
    input.clear();
    actual.clear();
    backup.clear();

      // 2. sort input edge records into input list
    while ( (faceSize = data.getFaceVertices( face, vert )) > vert.length )
      vert = new int[ faceSize ];

    data.getAttribute( attrCoords, vert[0], vc[1] );
    Geometry.toCartesian3D( vc[1], vc[1] );

    for ( i = 0; i < faceSize; i++ )
    {
      if ( (i1 = i + 1) >= faceSize ) i1 = 0;

      data.getAttribute( attrCoords, vert[i1], vc[i & 1] );
      Geometry.toCartesian3D( vc[i & 1], vc[i & 1] );

        // include edge vc[0]--vc[1]:
      if ( Math.ceil( vc[0][1] ) == Math.ceil( vc[1][1] ) )
        continue;                       // discarding horizontal edges

      input.add( new PolyEdge( vc[0], vc[1] ) );
    }

      // 3. initialize actual edge set
    if ( input.size() == 0 ) return 0;
    PolyEdge act = input.first();
    input.remove( act );                // !!! TODO(1.6): act = input.pollFirst()

    int y = act.y0;
    actual.add( act );

      // 4. hline loop
    counter = 0;
    do
    {
        // move all relevant edges from input into actual:
      while ( input.size() > 0 &&
              (act = input.first()).y0 == y )
      {                                 // !!! TODO(1.6): act = input.pollFirst()
        input.remove( act );
        actual.add( act );
      }

      PolyEdge old  = null;             // previous edge (kept in current-line state for potential rendering)
      int      rule = 0;

      do
      {
          // 4a. edge contribution to the current hline
        act  = actual.first();
        actual.remove( act );           // !!! TODO(1.6): act = actual.pollFirst()

        if ( rule != 0 )                // draw H-line segment from the previous edge..
          if ( inside )
            hLineInside( y, old, act );
          else
            hLineOutside( y, old, act );

          // 4b. old edge ages for the next hline
        if ( old != null &&
             --old.count > 0 )
        {
          old.next();
          backup.add( old );
        }

        rule = 1 - rule;
        old  = act;
      }
      while ( actual.size() > 0 );

        // the last edge was not transformed yet..
      if ( --old.count > 0 )
      {
        old.next();
        backup.add( old );
      }

        // swap the actual sets:
      TreeSet< PolyEdge > tmp = backup;
      backup = actual;
      actual = tmp;

        // Y-coordinate of the next line
      y++;
    }
    while ( actual.size() > 0 ||
            input.size()  > 0 );

    return counter;
  }

  //==========================================================================
  //  Interface FaceRender:

  /**
   * Fills the given polygonal face. Uses only pixels which are set in the associated
   * {@link BitMaskCore} (fills the whole face if no {@link BitMaskCore} is used).
   *
   * @param  ctx Handle to a context to draw from.
   * @param  face Handle of the face (in the associated {@link Brep} scene).
   * @see    BitMaskCore
   * @see    Brep
   */
  public void fillInside ( int ctx, int face )
  {
    if ( accelerators( ctx, face, true ) ) return;

    commonFill( ctx, face, true );
  }

  /**
   * Fills the horizontal line from the given polygonal face.
   *
   * @param  ctx Handle to a context to draw from.
   * @param  face Handle of the face (in the associated {@link Brep} scene).
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @see    Brep
   */
  public void fillHLine ( int ctx, int face, int x1, int x2, int y )
  {
    if ( accelerators( ctx, face, false ) ) return;

    output.setHLine( x1, x2, y );
  }

  /**
   * Fills the given polygonal face. Uses only pixels which are not set in the associated
   * {@link BitMaskCore}. Sets the drawn pixels afterwards.
   *
   * @param  ctx Handle to a context to draw from.
   * @param  face Handle of the face (in the associated {@link Brep} scene).
   * @return Number of pixels filled.
   * @see    BitMaskCore
   * @see    Brep
   */
  public int fillOutside ( int ctx, int face )
  {
    if ( accelerators( ctx, face, true ) ) return 0;

    return commonFill( ctx, face, false );
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // PROJECTED_COORDS:
    if ( key.compareTo(PROJECTED_COORDS) == 0 )
    {
      coordsName = stringProperty( value, coordsName );
      attrCoords = Brep.NULL;
      return;
    }
      // FACE_COLOR:
    if ( key.compareTo(FACE_COLOR) == 0 )
    {
      colorName = stringProperty( value, colorName );
      attrColor = Brep.NULL;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(PROJECTED_COORDS) == 0 )
      return coordsName;
    if ( key.compareTo(FACE_COLOR) == 0 )
      return colorName;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "FaceRender flat";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "FaceRenderToBrepAndBitMaskCoreAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Flat-shading of B-rep scene.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + '.' + C_BREP;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 1 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,   IFACE+"FaceRender" );
      t.newOutputPlug(    PL_DATA,    IFACE+"Brep" );
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_BITMASK, IFACE+"BitMaskCore" );
        // Property:
      t.propBegin( PROJECTED_COORDS, TYPE_STRING, "Name of projected-coords attribute", true );
        t.propDefault( Brep.COORD_PROJ );
      t.propEnd();
      t.propBegin( FACE_COLOR, TYPE_STRING, "Name of face-color attribute", true );
        t.propDefault( Brep.COLOR );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
