package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Data store for "features" (for feature-based warping/morphing).
 *
 * @version 0.26 $Rev: 407 $ $Date: 2007-03-25 15:56:09 +0200 (ne, 25 III 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/Features.java">Features.java</a>
 */
public class Features extends Piece implements FeatureStore
{
  /**
   * Internal feature representation.
   *
   * @version 0.26 $Rev: 407 $ $Date: 2007-03-25 15:56:09 +0200 (ne, 25 III 2007) $ $Author: pepca $
   * @since   0.24
   * @see     <a href="../../../../../cz/cuni/jagrlib/piece/Features.java">Features.java</a>
   */
  static class Feature
  {
    //==========================================================================
    //  Data:

    /** Attribute name for feature coordinates. */
    final static String COORDS = "Coords";

    /** Unique identification of the feature. */
    int id;

    /**
     * Feature type.
     *
     * @see FeatureStore#FEATURE_POINT
     * @see FeatureStore#FEATURE_ARROW
     * @see FeatureStore#FEATURE_TRIANGLE
     */
    int type;

    /** Birth time of the feature (integer value in seconds). */
    int birthTime;

    /** Death time of the feature (integer value in seconds). */
    int deathTime;

    /**
     * Mapping: Time -&gt; attribute value.
     * Coordinates are treated as regular attribute ({@link #COORDS}).
     */
    ArrayList< Map< String, Object > > attrMap;

    /**
     * Global (time-independent) feature attributes.
     */
    Map< String, Object > attr;

    //==========================================================================
    //  Construction:

    protected void initAttrMap ()
    {
      attrMap = new ArrayList< Map< String, Object > >( deathTime - birthTime + 1 );
      int i = birthTime;
      do
        attrMap.add( new HashMap< String, Object >() );
      while ( ++i <= deathTime );
    }

    Feature ( int birth, int death, int type )
    {
      if ( death <= birth )
        death = birth + 1;

      this.type = type;
      birthTime = birth;
      deathTime = death;
      initAttrMap();
      attr      = new HashMap< String, Object >();
    }

    void setTimes ( int death, int birth )
    {
      if ( death <= birth )
        death = birth + 1;

      if ( birth > deathTime ||
           death < birthTime )
      {                                     // no attribute will survive
        birthTime = birth;
        deathTime = death;
        initAttrMap();
        return;
      }

        // adjust the beginning:
      if ( birthTime < birth )              // discard time-slots from the beginning
        do
        {
          attrMap.remove( 0 );
        } while ( ++birthTime < birth );
      else
        while ( birth < birthTime )         // insert new time-slots at the beginning
        {
          attrMap.add( 0, new HashMap< String, Object >() );
          birthTime--;
        }

        // adjust the end:
      if ( death < deathTime )              // discard trailing time-slots
      {
        do
        {
          attrMap.remove( death + 1 );
        } while ( death < --deathTime );
        return;
      }
      while ( deathTime < death )           // append new time-slots
      {
        attrMap.add( new HashMap< String, Object >() );
        deathTime++;
      }
    }
  }

  //==========================================================================
  //  Data:

  /** Array of all registered features. Access key = feature id. */
  protected ArrayList< Feature > features = new ArrayList< Feature >( 8 );

  //==========================================================================
  //  Support:

  protected final Feature feature ( int id )
  {
    if ( id < 0 || id >= features.size() ) return null;

    return features.get( id );
  }

  protected final Feature feature ( int id, int time )
  {
    Feature f = feature( id );
    if ( f != null &&
         f.birthTime <= time &&
         time <= f.deathTime )
      return f;

    return null;
  }

  protected final Object getAttr ( int id, double time, String name )
  {
    Feature f = feature( id );
    if ( f == null ) return null;

    if ( time < f.birthTime ) time = f.birthTime;
    else
      if ( time > f.deathTime ) time = f.deathTime;

    int    t   = (int)Math.round( time );
    Object val = f.attrMap.get( t - f.birthTime ).get( name );
    if ( val != null ) return val;

    return f.attr.get( name );
  }

  protected final void setAttr ( int id, int time, String name, Object value )
  {
    Feature f = feature( id, time );
    if ( f != null )
      f.attrMap.get( time - f.birthTime ).put( name, value );
    else
    {
      f = feature( id );
      if ( f != null )
        f.attr.put( name, value );
    }
  }

  //==========================================================================
  //  Interface FeatureStore:

  /**
   * Inserts a new feature of the given type.
   * Every feature exists in defined time interval (timeFrom - timeTo).
   *
   * @param  type Feature type ({@link #FEATURE_POINT}, {@link #FEATURE_ARROW} or {@link #FEATURE_TRIANGLE}).
   * @return Feature handle (id). If succeeded, return value should be &gt;= 0.
   * @see    #removeFeature
   * @see    #setTimes
   * @see    #setCoords
   * @see    #setParam
   */
  public int insertFeature ( int timeFrom, int timeTo, int type )
  {
      // insert/recycle a feature slot:
    int id = -1;
    if ( features.size() > 0 )
      id = features.indexOf( null );

    Feature f = new Feature( timeFrom, timeTo, type );

    if ( id < 0 )
    {                                     // insert new feature slot
      id = features.size();
      features.add( f );
    }
    else
      features.set( id, f );              // reuse the old slot..

    f.id = id;
    return id;
  }

  /**
   * Removes the given feature from the database.
   *
   * @param  id Handle of an existing feature.
   */
  public void removeFeature ( int id )
  {
    if ( id < 0 ||
         id >= features.size() ) return;

    features.set( id, null );
  }

  /**
   * Finds the next feature in the database.
   * Iteration should start with id = {@link #NULL}. This function also returns {@link #NULL}
   * if there is no other feature..
   *
   * @param  id Handle of previous feature (or {@link #NULL} if we are restaring an iteration).
   * @param  time Time of interest.
   * @return Handle of a next feature in database (or {@link #NULL} if the previous was the last one).
   */
  public int nextFeature ( int id, double time )
  {
    if ( ++id < 0 ) id = 0;

      // looking for feature starting at index "id" and being active at the given time:
    while  ( id < features.size() )
    {
      Feature act = features.get( id );
      if ( act != null &&
           (double)act.birthTime <= time &&
           (double)act.deathTime >= time )
        return id;

      id++;
    }

    return NULL;
  }

  /**
   * Reads feature-type ({@link #FEATURE_POINT}, {@link #FEATURE_ARROW} or {@link #FEATURE_TRIANGLE}).
   */
  public int featureType ( int id )
  {
    Feature f = feature( id );
    return( f == null ? NULL : f.type );
  }

  /**
   * Sets feature life-time.
   *
   * @param  id Handle of a feature.
   * @param  timeFrom Birth-time of the feature.
   * @param  timeTo Death-time of the feature.
   */
  public void setTimes ( int id, int timeFrom, int timeTo )
  {
    Feature f = feature( id );
    if ( f != null )
      f.setTimes( timeFrom, timeTo );
  }

  /** Returns Birth-time of the given feature. */
  public int featureStart ( int id )
  {
    Feature f = feature( id );
    return( f == null ? NULL : f.birthTime );
  }

  /** Returns Time of death of the given feature. */
  public int featureEnd ( int id )
  {
    Feature f = feature( id );
    return( f == null ? NULL : f.deathTime );
  }

  /**
   * Sets feature coordinates of the given feature in the given keyframe-time.
   *
   * @param  id Handle of a feature.
   * @param  time Keyframe-time.
   * @param  coords Coordinate array (e.g. {@code double[4]} for 2D arrow feature:
   *         {@code [ x0, y0, x1, y1 ]}).
   */
  public void setCoords ( int id, int time, double[] coords )
  {
    if ( coords == null ||
         coords.length < 2 ) return;

    setAttr( id, time, Feature.COORDS, Formula.cloneArray( coords ) );
  }

  /**
   * Reads feature coordinates of the given feature in the given time.
   *
   * @param  id Handle of a feature.
   * @param  time Time of interest. Needs not be keyframe-time.
   * @param  coords Pre-allocated coordinate array to be filled
   *         ({@code double[4]} for 2D arrow feature: {@code [ x0, y0, x1, y1 ]}). Can be {@code null}.
   * @return Coordinate array or {@code null}.
   */
  public double[] getCoords ( int id, double time, double[] coords )
  {
    // !!! TODO: coordinate interpolation for warping !!!
    double[] c = (double[])getAttr( id, time, Feature.COORDS );
    if ( c == null ) return null;

    if ( coords == null || coords.length < c.length )
      return Formula.cloneArray( c );

    System.arraycopy( c, 0, coords, 0, c.length );
    return coords;
  }

  /**
   * Sets feature's general parameter in the given keyframe-time.
   *
   * @param  id Handle of a feature.
   * @param  time Keyframe-time ({@link #NULL} for time-independent parameter).
   * @param  name Parameter identifier.
   * @param  value New parameter value.
   */
  public void setParam ( int id, int time, String name, Object value )
  {
    setAttr( id, time, name, value );
  }

  /**
   * Reads the given integer parametr in the given time.
   *
   * @param  id Handle of a feature.
   * @param  time Time of interest. Needs not be keyframe-time.
   * @param  name Parameter identifier.
   * @return Integer value (interpolated in time if applicable).
   */
  public int getParamInt ( int id, double time, String name )
  {
    // !!! TODO: attribute interpolation for warping !!!
    Object val = getAttr( id, time, name );
    if ( val == null ) return 0;

    if ( val instanceof Integer )
      return ((Integer)val).intValue();

    if ( val instanceof Double )
      return Formula.round( ((Double)val).doubleValue() );

    if ( val instanceof Float )
      return Math.round( ((Float)val).floatValue() );

    return 0;
  }

  /**
   * Reads the given floating-point parametr in the given time.
   *
   * @param  id Handle of a feature.
   * @param  time Time of interest. Needs not be keyframe-time.
   * @param  name Parameter identifier.
   * @return Floating-point value (interpolated in time if applicable).
   */
  public double getParamDouble ( int id, double time, String name )
  {
    // !!! TODO: attribute interpolation for warping !!!
    Object val = getAttr( id, time, name );
    if ( val == null ) return 0.0;

    if ( val instanceof Integer )
      return ((Integer)val).intValue();

    if ( val instanceof Double )
      return ((Double)val).doubleValue();

    if ( val instanceof Float )
      return ((Float)val).floatValue();

    return 0.0;
  }

  /**
   * Reads the given general parametr in the given time.
   *
   * @param  id Handle of a feature.
   * @param  time Time of interest. Needs not be keyframe-time.
   * @param  name Parameter identifier.
   * @return Parameter value (interpolated in time if applicable).
   */
  public Object getParam ( int id, double time, String name )
  {
    return getAttr( id, time, name );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Features";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "FeatureStoreTo";

  /** Object description. */
  private final static String DESCRIPTION = "Data container for 1D, 2D, 3D features. " +
                                            "Usage: feature-based warping/morphing";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DATA + ".features";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"FeatureStore" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
