package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Audio filter - notch filter (very-narrow-band suppression).
 *
 * @version 0.26 $Rev: 431 $ $Date: 2007-04-29 22:42:22 +0200 (ne, 29 IV 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/Filter1DNotch.java">Filter1DNotch.java</a>
 */
public class Filter1DNotch extends Piece implements Filter1D
{
  //==========================================================================
  //  Data:

  /** Suppressed frequency in Hz. */
  protected double frequency  = 50.0;

  /** Unity (quality measure). */
  protected double unity      = 0.99;

  /** Sample rate in Hz. */
  protected double sampleRate = 0.0;

  /** Number of channels to be processed. */
  protected int channels      = 0;

  /** History windows for each channel (input data). */
  protected double[][] xWindow;

  /** History windows for each channel (output data). */
  protected double[][] yWindow;

  /** Index into {@link #xWindow} and {@link #yWindow} arrays. */
  protected int windowPtr;

  /** Coef for x[i-1]. */
  protected double b1;

  /** Coef for y[i-1]. */
  protected double a1;

  /** Coef for y[i-2]. */
  protected double a2;

  //==========================================================================
  //  Support:

  /**
   * Computes next sample.
   *
   * @param  x Input sample (<b>x<sub>i</sub></b>).
   * @param  channel Channel number.
   * @return Filtered sample (<b>y<sub>i</sub></b>).
   */
  protected double nextSample ( double x, int channel )
  {
    int windowPrev = 1 - windowPtr;
    double result  = x + b1 * xWindow[channel][windowPrev] +      xWindow[channel][windowPtr] +
                         a1 * yWindow[channel][windowPrev] + a2 * yWindow[channel][windowPtr];

      // shift both windows:
    xWindow[channel][windowPtr] = x;
    yWindow[channel][windowPtr] = result;

    return result;
  }

  //==========================================================================
  //  Interface Filter1D:

  /**
   * Defines format modifications.
   * Sets input sample format for subsequent {@link #batch} calls.
   *
   * @param  source Source audio format (channels, bytes-per-sample, samples, frequency).
   * @param  target Target audio format (channels, bytes-per-sample, samples, frequency).
   * @return {@code true} if any format changes will be made ({@code false => format identity}).
   * @see    #batch(double[],int,double[])
   * @see    #batch(int[],int,int[])
   */
  public boolean format ( int[] source, int[] target )
  {
    if ( source == null ||
         source.length < 4 ) return false;

    if ( target != null )
      System.arraycopy( source, 0, target, 0, Math.min( source.length, target.length ) );

    sampleRate = source[ Filter1D.IND_FREQUENCY ];
    channels   = source[ Filter1D.IND_CHANNELS ];

    return false;
  }

  /**
   * Starts a new data pass.
   *
   * @see    #batch(double[],int,double[])
   * @see    #batch(int[],int,int[])
   */
  public void reset ()
  {
    if ( channels <= 0 ||
         sampleRate < Geometry.EPSILON ) return;

    int i;
    if ( xWindow == null ||
         xWindow.length < channels )
      xWindow = new double[ channels ][ 2 ];
    else
      for ( i = 0; i < channels; i++ )
        Arrays.fill( xWindow[i], 0.0 );

    if ( yWindow == null ||
         yWindow.length < channels )
      yWindow = new double[ channels ][ 2 ];
    else
      for ( i = 0; i < channels; i++ )
        Arrays.fill( yWindow[i], 0.0 );

    windowPtr = 0;

    double theta = 2.0 * Math.PI * frequency / sampleRate;
    double c     = Math.cos ( theta );

    b1 = -c - c;
    a1 = -unity * b1;
    a2 = -unity * unity;
  }

  /**
   * Processes one sample batch in <tt>double</tt> format.
   * Sample ordering in case of multi-channel data: {@code channel0[0], channel1[0], ..
   * channelN[0], channel0[1], ..}.
   *
   * @param  source Source data array.
   * @param  sourceSamples Number of source time-slots.
   * @param  target Target data array.
   * @return Number of target time-slots.
   * @see    #format
   * @see    #batch(int[],int,int[])
   */
  public int batch ( double[] source, int sourceSamples, double[] target )
  {
    if ( channels <= 0 ) return -1;

    int timeSlots = sourceSamples / channels;
    int i, ch, ptr;

    for ( i = ptr = 0; i < timeSlots; i++ )
    {
      for ( ch = 0; ch < channels; ch++, ptr++ )
        target[ptr] = nextSample( source[ptr], ch );

      windowPtr = 1 - windowPtr;
    }

    return sourceSamples;
  }

  /**
   * Processes one sample batch in <tt>integer</tt> format.
   * Sample ordering in case of multi-channel data: {@code channel0[0], channel1[0], ..
   * channelN[0], channel0[1], ..}.
   *
   * @param  source Source data array.
   * @param  sourceSamples Number of source time-slots.
   * @param  target Target data array.
   * @return Number of target time-slots.
   * @see    #format
   * @see    #batch(double[],int,double[])
   */
  public int batch ( int[] source, int sourceSamples, int[] target )
  {
    return sourceSamples;
  }

  //==========================================================================
  //  Interface Property:

  public final static String FREQUENCY  = "Frequency";
  public final static String UNITY      = "Unity";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // FREQUENCY:
    if ( key.compareTo(FREQUENCY) == 0 )
    {
      frequency = doubleProperty( value, frequency, 0.0, 1.e9 );
      return;
    }
      // UNITY:
    if ( key.compareTo(UNITY) == 0 )
    {
      unity = doubleProperty( value, unity, 0.1, 0.999999 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(FREQUENCY) == 0 )
      return frequency;
    if ( key.compareTo(UNITY) == 0 )
      return unity;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Filter1DNotch";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "Filter1D";

  /** Object description. */
  private final static String DESCRIPTION = "Audio filter module - notch filter.";

  /** Object category. */
  protected final static String CATEGORY = C_1D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"Filter1D" );
        // Property:
      t.propBegin( FREQUENCY, TYPE_DOUBLE, "Notch frequency", true );
        t.propDefault( 50.0 );
        t.propBounds( 0.0, 1.e9 );
      t.propEnd();
      t.propBegin( UNITY, TYPE_DOUBLE, "Unity (quality measure)", true );
        t.propDefault( 0.99 );
        t.propBounds(  0.1, 0.999999 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
