package cz.cuni.jagrlib.piece;

import java.awt.Point;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.LinkedBlockingQueue;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Flood-fill using FIFO (queue) and point access.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/FloodFillPoint.java">FloodFillPoint.java</a>
 */
public class FloodFillPoint extends Piece implements FloodFillRender
{
  //==========================================================================
  //  Data:

  /** Continuity. Should be either 4 or 8. */
  protected int continuity = 4;

  /** Maximal queue size. */
  protected int maxQueue = 0;

  /** Maximal queue. */
  protected Point[] queue;

  /** Remember the maximal queue? */
  protected boolean rememberQueue = false;

  /** Pre-set maximal queue size. */
  protected int setMaxQueue = -1;

  /** Save file-name template. */
  protected String saveTemplate = "eval/save%03d.png";

  /** Interval between two save points (in drawn pixels). */
  protected int saveInterval = 1000;

  //==========================================================================
  //  Interface FloodFillRender:

  /**
   * Fills the continuous 2D area defined by the starting point [x,y],
   * using callback border {@link BitMask}. Computes either <em>4-continuous</em>
   * or <em>8-continuous</em> area.
   *
   * @param  x Seed's X coordinate.
   * @param  y Seed's Y coordinate.
   */
  public void floodFill ( int x, int y )
  {
    final BitMask        out     = (BitMask)       getInterface( PL_OUTPUT,  IFACE+"BitMask" );
    final BitMask        cl      = (BitMask)       getInterface( PL_BITMASK, IFACE+"BitMask" );
    final SolidPen       pen     = (SolidPen)      getInterface( "pen",      IFACE+"SolidPen" );
    final GraphicsViewer preview = (GraphicsViewer)getInterface( PL_RASTER,  IFACE+"GraphicsViewer" );
    final DataFileFormat save    = (DataFileFormat)getInterface( "save",     IFACE+"DataFileFormat" );
    out.init();

    BitMask bak = null;
    if ( preview != null ||
         save    != null )
    {
      bak = new XTransitionList();
      int wid = intProperty( out, BitMask.WIDTH,  512 );
      int hei = intProperty( out, BitMask.HEIGHT, 512 );
      bak.init( wid, hei );
    }
    else
      bak = out;

    Queue< Point > q = new LinkedBlockingQueue< Point >();
    bak.setPixel( x, y );
    out.setPixel( x, y );
    q.add( new Point( x, y ) );
    int actQueue = 1;
    int drawn    = 0;
    int saveFile = 0;
    maxQueue     = 1;

    do
    {
      Point p = q.remove();
      actQueue--;
      x = p.x;
      y = p.y;
      //LogFile.log( String.format( "Pop: %4d,%4d", x, y ) );

      if ( (!bak.getPixel( x-1, y )) && cl.getPixel( x-1, y ) )
      {
        bak.setPixel( x-1, y );
        out.setPixel( x-1, y );
        q.add( new Point( x-1, y ) );
        actQueue++;
        drawn++;
      }

      if ( !bak.getPixel( x+1, y ) && cl.getPixel( x+1, y ) )
      {
        bak.setPixel( x+1, y );
        out.setPixel( x+1, y );
        q.add( new Point( x+1, y ) );
        actQueue++;
        drawn++;
      }

      if ( !bak.getPixel( x, y-1 ) && cl.getPixel( x, y-1 ) )
      {
        bak.setPixel( x, y-1 );
        out.setPixel( x, y-1 );
        q.add( new Point( x, y-1 ) );
        actQueue++;
        drawn++;
      }

      if ( !bak.getPixel( x, y+1 ) && cl.getPixel( x, y+1 ) )
      {
        bak.setPixel( x, y+1 );
        out.setPixel( x, y+1 );
        q.add( new Point( x, y+1 ) );
        actQueue++;
        drawn++;
      }

      if ( continuity > 4 )
      {
        if ( (!bak.getPixel( x-1, y-1 )) && cl.getPixel( x-1, y-1 ) )
        {
          bak.setPixel( x-1, y-1 );
          out.setPixel( x-1, y-1 );
          q.add( new Point( x-1, y-1 ) );
          actQueue++;
          drawn++;
        }

        if ( !bak.getPixel( x+1, y-1 ) && cl.getPixel( x+1, y-1 ) )
        {
          bak.setPixel( x+1, y-1 );
          out.setPixel( x+1, y-1 );
          q.add( new Point( x+1, y-1 ) );
          actQueue++;
          drawn++;
        }

        if ( !bak.getPixel( x-1, y+1 ) && cl.getPixel( x-1, y+1 ) )
        {
          bak.setPixel( x-1, y+1 );
          out.setPixel( x-1, y+1 );
          q.add( new Point( x-1, y+1 ) );
          actQueue++;
          drawn++;
        }

        if ( !bak.getPixel( x+1, y+1 ) && cl.getPixel( x+1, y+1 ) )
        {
          bak.setPixel( x+1, y+1 );
          out.setPixel( x+1, y+1 );
          q.add( new Point( x+1, y+1 ) );
          actQueue++;
          drawn++;
        }
      }

        // maximal queue size:
      if ( actQueue > maxQueue )
      {
        maxQueue = actQueue;
        if ( rememberQueue )
          queue = q.toArray( new Point[ 0 ] );
      }

      if ( (drawn >= saveInterval || actQueue == setMaxQueue) &&
           pen != null &&
           (preview != null || save != null) )
      {
        pen.stroke();
        if ( preview != null )
          preview.repaintAll();
        if ( save != null &&
             saveTemplate.endsWith( ".png" ) )
        {
          try
          {
            save.saveFile( String.format( saveTemplate, saveFile ), null );
          }
          catch ( IOException e )
          {
            LogFile.exception( "Cannot save PNG file (" + saveTemplate + ")!", e );
          }
          saveFile++;
        }
        drawn = 0;
      }

    } while ( actQueue != setMaxQueue &&
              !q.isEmpty() );
  }

  //==========================================================================
  //  Interface Property:

  public final static String MAX_QUEUE     = "MaxQueue";
  public final static String QUEUE         = "Queue";
  public final static String REM_QUEUE     = "RememberQueue";
  public final static String TEMPLATE      = "SaveTemplate";
  public final static String SAVE_INTERVAL = "SaveInterval";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // CONTINUITY:
    if ( key.compareTo(CONTINUITY) == 0 )
    {
      continuity = intProperty( value, continuity ) < 8 ? 4 : 8;
      return;
    }
      // REM_QUEUE:
    if ( key.compareTo(REM_QUEUE) == 0 )
    {
      rememberQueue = booleanProperty( value, rememberQueue );
      return;
    }
      // MAX_QUEUE:
    if ( key.compareTo(MAX_QUEUE) == 0 )
    {
      setMaxQueue = intProperty( value, setMaxQueue );
      return;
    }
      // TEMPLATE:
    if ( key.compareTo(TEMPLATE) == 0 )
    {
      if ( value instanceof String )
        saveTemplate = (String)value;
      return;
    }
      // SAVE_INTERVAL:
    if ( key.compareTo(SAVE_INTERVAL) == 0 )
    {
      saveInterval = intProperty( value, saveInterval, 1, 1000000 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(CONTINUITY) == 0 )
      return continuity;
    if ( key.compareTo(MAX_QUEUE) == 0 )
      return maxQueue;
    if ( key.compareTo(QUEUE) == 0 )
      return queue;
    if ( key.compareTo(REM_QUEUE) == 0 )
      return rememberQueue;
    if ( key.compareTo(TEMPLATE) == 0 )
      return saveTemplate;
    if ( key.compareTo(SAVE_INTERVAL) == 0 )
      return saveInterval;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "FloodFillPoint";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "FloodFillRenderToBitMaskAndBitMask";

  /** Object description. */
  private final static String DESCRIPTION = "Flood-fill point/based algorithm, uses FIFO.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DRAW + '.' + C_FILL + '.' + C_INTEGER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,   IFACE+"FloodFillRender" );
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"BitMask" );
      t.newOutputPlug(    PL_BITMASK, IFACE+"BitMask" );
      t.newOptOutputPlug( "pen",      IFACE+"Pen" );
      t.newOptOutputPlug( PL_RASTER,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( "save",     IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( CONTINUITY, TYPE_INTEGER, "Continuity style", true );
        t.propDefault( 4 );
        t.propBounds( 4, 8 );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "4-continuous", 4 ,"4-continuous fill" );
        t.propEnum( "8-continuous", 8 ,"8-continuous fill" );
      t.propEnd();
      t.propBegin( MAX_QUEUE, TYPE_INTEGER, "Max queue size", true );
      t.propEnd();
      t.propBegin( REM_QUEUE, TYPE_BOOLEAN, "Remember maximal queue?", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( TEMPLATE, TYPE_STRING, "Save file-name template", true );
        t.propDefault( "eval/save%03d.png" );
      t.propEnd();
      t.propBegin( SAVE_INTERVAL, TYPE_INTEGER, "Save interval in pixels", true );
        t.propDefault( 1000 );
        t.propBounds( 1, 1000000 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
