package cz.cuni.jagrlib.piece;

import java.util.*;
import java.io.IOException;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Fractal image compression module. Uses adaptive Quad-tree partitioning.
 *
 * @version 0.26 $Rev: 464 $ $Date: 2007-05-24 15:41:38 +0200 (čt, 24 V 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/FractalCompressionQtree.java">FractalCompressionQtree.java</a>
 */
public class FractalCompressionQtree extends DefaultRasterFileFormat
{
  //==========================================================================
  //  Data:

  /** Decoded image block, acts as Q-tree node as well. */
  protected class ImageBlock implements Comparable< ImageBlock >
  {
    /** Upper-left corner coordinates. */
    public int x0, y0;

    /** Block size in pixels. */
    public int size;

    /** Best source coordinates. */
    public double sourceX, sourceY;

    /** Best source orientation (see {@link RasterSearch#ORI_0123}, ..). */
    public int orientation;

    /** Pixel transform coefficients (contrast, brightness). */
    public double s, o;

    /** Reconstruction error (for adaptive encoder). */
    public double error;

    /**
     * Block order (descending reconstruction error).
     * {@code 0} for the best (=final) nodes.
     */
    public int order;

    /** Descendants (can be empty). */
    public ImageBlock[] desc;

    /** Non-initialized (inner) Q-tree node. */
    public ImageBlock ( int x, int y, int siz )
    {
      x0          = x;
      y0          = y;
      size        = siz;
      orientation = RasterSearch.ORI_FAIL;
      order       = 0;
    }

    /** Initialized (matched) Q-tree node. */
    public ImageBlock ( int x, int y, int siz, RasterGraphics g )
    {
      this( x, y, siz );
      search.setSearchParameters( 1 << RasterSearch.ORI_0123, 1, RasterSearch.ERR_RMSE, 1.0,
                                  1-size, 0, 1-size, 0, 0, 0,
                                  -sMax, sMax, -sMax, sMax + 1.0 );
      match( g );
    }

    /**
     * Inserts the given child sub-tree,
     * creates {@link #desc} array if necessary.
     */
    public void setChild ( int ord, ImageBlock ch )
    {
      if ( desc == null )
        desc = new ImageBlock[ 4 ];

      desc[ord] = ch;
    }

    /**
     * Computes block match in the prepared reference image.
     * Returns {@code true} if succeeded.
     */
    public boolean match ( RasterGraphics g )
    {
      if ( userBreak ) return false;

      orientation = search.findBestRectangle( g, x0, y0, size, size, 0, bingo );
      if ( orientation == RasterSearch.ORI_FAIL ) return false;

      sourceX = bingo[0];
      sourceY = bingo[1];
      s       = bingo[2];
      o       = bingo[3];
      error   = bingo[4];
      heap.add( this );
      return true;
    }

    /**
     * Encoding the sub-tree rooted here. Code of one node will be:
     * <pre>
     * CTX_CTRL:  { CTRL_REC     | recursive descent (w/o block match)
     *              CTRL_NODE    | inner node (contains inferior block match)
     *              CTRL_FAIL    | failed leaf node (dead-end)
     *              &gt;= CTRL_ORI0 } leaf node (1st symbol defines its orientation)
     * [raw-bits: order]           block order number (if 'bestCodeOnly' is not set)
     * CTX_COORD: x, y             reference image coordinates
     * CTX_S:     s                quantized 's' coef (multiplier)
     * CTX_O:     o                quantized 'o' coef (addend)
     * </pre>
     */
    public void encode ()
    throws IOException
    {
      codec.setContext( CTX_CTRL );

      if ( desc != null &&
           (orientation == RasterSearch.ORI_FAIL || bestCodeOnly) )
        codec.put( CTRL_REC );

      else
      if ( desc == null &&
           orientation == RasterSearch.ORI_FAIL )
        codec.put( CTRL_FAIL );

      else
      {
        if ( !bestCodeOnly && desc != null )
          codec.put( CTRL_NODE );

        codec.put( CTRL_ORI0 + orientation );

        if ( !bestCodeOnly )
          codec.putBits( order, blockOrderBits );

        codec.setContext( CTX_COORD );
        codec.put( Formula.round( sourceX ) + maxBlockSize );
        codec.put( Formula.round( sourceY ) + maxBlockSize );

        codec.setContext( CTX_S );
        codec.put( (int)Formula.vlc( qS.quantize( s ) ) );

        codec.setContext( CTX_O );
        codec.put( qO.quantize( o ) );
      }

      if ( desc != null )
        for ( ImageBlock ib : desc )
          if ( ib != null ) ib.encode();
    }

    /**
     * Performs one step of fractal transform (PIFS).
     * Source image was already downsampled by a factor of 2.
     */
    public void transformImage ()
    {
      if ( order >= minNodeOrder ||
           order == 0 && desc == null )
      {
        int x, y, b;
        final int bands = actIm.getBands();
        final double[] pixel = new double[ bands ];

        final int[] tr = RasterSearch.ORI_INV[ orientation ];
        final int x2x  = tr[0];
        final int y2x  = tr[1];
        final int absx = Formula.round( sourceX ) + (size - 1) * tr[2];
        final int x2y  = tr[3];
        final int y2y  = tr[4];
        final int absy = Formula.round( sourceY ) + (size - 1) * tr[5];

        for ( y = 0; y < size; y++ )
        {
          int srcX = absx + y * y2x;
          int srcY = absy + y * y2y;
          for ( x = 0; x < size; x++, srcX += x2x, srcY += x2y )
          {
            prevIm.getPixel( srcX, srcY, pixel );
            for ( b = 0; b < bands; b++ )
              pixel[b] = Formula.clamp( s * pixel[b] + o, 0.0, 1.0 );
            actIm.putPixel( x0 + x, y0 + y, pixel );
          }
        }
      }

      if ( desc == null ) return;

      for ( ImageBlock ib : desc )
        if ( ib != null ) ib.transformImage();
    }

    /**
     * Write decomposition diagram into the {@link FractalCompressionQtree#log} image.
     */
    public void logTree ( int depth )
    {
      if ( log  == null ) return;

      depth++;
      if ( desc == null )
      {
        log.setColor( LOG_COLOR[ depth % LOG_COLOR.length ] );
        DefaultRasterGraphics.drawFrame( log, x0, y0, x0 + size-1, y0 + size-1 );
      }
      else
        for ( ImageBlock ib : desc )
          if ( ib != null ) ib.logTree( depth );
    }

    /** Natural ordering: descending [ error ]. */
    public int compareTo ( ImageBlock o )
    {
      if ( o == null )
        throw new NullPointerException();

      if ( this == o )
        return 0;

      if ( error != o.error )
        return( error > o.error ? -1 : 1 );

      return( hashCode() < o.hashCode() ? -1 : 1 );
    }
  }

  /** Heap of unfinished Q-tree nodes (encoder only). */
  protected TreeSet< ImageBlock > heap;

  /** Maximal block size in pixels. */
  protected int maxBlockSize = 32;

  /** Minimal block size in pixels. */
  protected int minBlockSize = 4;

  /** RMSE error limit for Q-tree adaptive subdivision. */
  protected double errorLimit = 0.04;

  /** Encode the best blocks only (non-redundant code)? */
  protected boolean bestCodeOnly = true;

  /** Log intermediate results? */
  protected boolean logging = true;

  /** Number of quantized values of s. */
  protected int sValues = 32;

  /** Number of quantized values of o. */
  protected int oValues = 128;

  /** Maximal amplitude of s. */
  protected double sMax = 1.1;

  /** Number of bits to encode block order# (for progressive [= redundant] mode only). */
  protected int blockOrderBits = 16;

  /** Entropy-codec context: control, orientation. */
  protected static final int CTX_CTRL  = 0;

  /** Entropy-codec context: coordinates. */
  protected static final int CTX_COORD = 1;

  /** Entropy-codec context: s-coefficient. */
  protected static final int CTX_S     = 2;

  /** Entropy-codec context: o-coefficient. */
  protected static final int CTX_O     = 3;

  /** Control alphabet: recursion (inner node w/o block data). */
  protected static final int CTRL_REC  = 0;

  /** Control alphabet: inner node (with block data). */
  protected static final int CTRL_NODE = 1;

  /** Control alphabet: failed leaf node. */
  protected static final int CTRL_FAIL = 2;

  /** Control alphabet: 1st code assigned for orientation (marking the leaf node). */
  protected static final int CTRL_ORI0 = 3;

  /** Maximal number of transforms steps in image reconstruction. */
  protected static final int MAX_ITERATIONS = 10;

  /** Accelerator: searchable object. */
  protected RasterSearch search;

  /** Accelerator: entropy codec. */
  protected EntropyCodec codec;

  /** Accelerator: [dead-zone] quantizer for 's'. */
  protected ScalarQuantizer qS;

  /** Accelerator: [linear] quantizer for 'o'. */
  protected ScalarQuantizer qO;

  /** Image width in pixels. */
  protected int width;

  /** Image height in pixels. */
  protected int height;

  /** Size of the smallest rectangle containing the whole image. */
  protected int totalSize;

  /** Number of image bands (1 for {@link RasterGraphics#MODE_GRAY}). */
  protected int bands;

  /** Reconstruction: only nodes with {@code order == 0} or {@code order >= minNodeOrder}. */
  protected int minNodeOrder;

  /** Target image for fractal transform (PIFS). */
  protected HDRImage actIm;

  /** Source image for fractal transform (PIFS). */
  protected HDRImage prevIm;

  /** Global array for block-search results. */
  protected double[] bingo = new double[ 5 ];

  //==========================================================================
  //  Logging:

  /** Log-image frame colors. */
  protected final static double[][] LOG_COLOR =
  {
    { 0.9, 0.0, 0.0 },
    { 0.0, 0.7, 0.0 },
    { 0.0, 0.0, 1.0 },
  };

  /** Log-image. */
  protected RasterGraphics log;

  //==========================================================================
  //  Support:

  /**
   * Builds one branch of a Q-tree, returns its root.
   * Blocks with size at most {@link #maxBlockSize} will go to the {@link #heap}.
   */
  protected ImageBlock buildInitialTree ( RasterGraphics g, int x0, int y0, int size )
  {
    if ( userBreak ) return null;

    ImageBlock block = new ImageBlock( x0, y0, size );

    if ( size <= maxBlockSize )           // find match in reference frame
    {
      if ( block.match( g ) )
      {
        LogFile.debug( String.format( Locale.US, "Best[%3d,%3d,%2d]: [%4d,%4d], s=%6.3f, o=%6.1f, err=%8.5f",
                                      block.x0, block.y0, block.size,
                                      Formula.round( block.sourceX * 2.0 ), Formula.round( block.sourceY * 2.0 ),
                                      block.s, (block.o * 255.0), block.error ) );
      }
      return block;
    }

      // block is too big => create descendants
    final int half = size >> 1;

    block.setChild( 0, buildInitialTree( g, x0, y0, half ) );

    if ( x0 + half < width )
      block.setChild( 1, buildInitialTree( g, x0 + half, y0, half ) );

    if ( y0 + half < height )
    {
      block.setChild( 2, buildInitialTree( g, x0, y0 + half, half ) );

      if ( x0 + half < width )
        block.setChild( 3, buildInitialTree( g, x0 + half, y0 + half, half ) );
    }

    return block;
  }

  /**
   * Loads Q-tree from attached {@link EntropyCodec}, returns its root.
   */
  protected ImageBlock loadTree ( int x0, int y0, int size )
  throws IOException
  {
    ImageBlock block = new ImageBlock( x0, y0, size );

    codec.setContext( CTX_CTRL );
    int nodeType = codec.get();

    if ( nodeType == CTRL_FAIL ) return block;

      // inner node (will read its descendants)?
    final boolean inner = nodeType < CTRL_ORI0;

    if ( nodeType != CTRL_REC )           // node data will be read
    {
      if ( nodeType == CTRL_NODE )        // skip node prefix => read node orientation
        nodeType = codec.get();

      block.orientation = nodeType - CTRL_ORI0;
      if ( !bestCodeOnly )
        block.order = (int)codec.getBits( blockOrderBits );

      codec.setContext( CTX_COORD );
      block.sourceX = codec.get() - maxBlockSize;
      block.sourceY = codec.get() - maxBlockSize;

      codec.setContext( CTX_S );
      block.s = qS.dequantizeDouble( Formula.vlcInv( codec.get() ) );

      codec.setContext( CTX_O );
      block.o = qO.dequantizeDouble( codec.get() );

      LogFile.debug( String.format( Locale.US, "Load node [%3d,%3d,%2d], s=%6.3f, o=%6.1f",
                                    block.x0, block.y0, block.size, block.s, (block.o * 255.0) ) );
    }

    if ( inner )                          // read node descendants
    {
      final int half = size >> 1;

      block.setChild( 0, loadTree( x0, y0, half ) );

      if ( x0 + half < width )
        block.setChild( 1, loadTree( x0 + half, y0, half ) );

      if ( y0 + half < height )
      {
        block.setChild( 2, loadTree( x0, y0 + half, half ) );

        if ( x0 + half < width )
          block.setChild( 3, loadTree( x0 + half, y0 + half, half ) );
      }
    }

    return block;
  }

  /**
   * Common load code.
   *
   * @param  stream Opened input bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonLoad ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    codec = (EntropyCodec)   getInterface( PL_OUTPUT,    IFACE+"EntropyCodec" );
    qS    = (ScalarQuantizer)getInterface( "quantizerS", IFACE+"ScalarQuantizer" );
    qO    = (ScalarQuantizer)getInterface( "quantizerO", IFACE+"ScalarQuantizer" );

    if ( codec == null ||
         qS    == null ||
         qO    == null ) return;

    codec.open( false, CTX_CTRL );
    codec.setMaxSymbol( CTRL_ORI0 + 7 );  // Control/orientation alphabet

      // FRAC header:
    int magic  = readUnsigned16( codec );
    if ( magic != MAGIC )
      throw new IOException( "Error in FRAC file format" );

    width        = readUnsigned16( codec );
    height       = readUnsigned16( codec );
    bands        = (int)codec.getBits( 7 );
    bestCodeOnly = codec.getBits( 1 ) != 0L;

    totalSize    = 1 << Formula.bits( Math.max( width, height ) );
    if ( !bestCodeOnly )
      blockOrderBits = (int)codec.getBits( 8 );
    maxBlockSize = readUnsigned16( codec );
    sValues      = readUnsigned16( codec );
    oValues      = readUnsigned16( codec );

      // prepare codec contexts:
    codec.setContext( CTX_COORD );
    codec.setMaxSymbol( ((totalSize + 1) >> 1) + maxBlockSize );

    codec.setContext( CTX_S );
    codec.setMaxSymbol( sValues );

    codec.setContext( CTX_O );
    codec.setMaxSymbol( oValues - 1 );

      // plugin parameters:
    if ( !readParameters( codec, codec ) )
      throw new IOException( "Error in FRAC file format (bad entropy codec)" );
    if ( !readParameters( codec, qS ) )
      throw new IOException( "Error in FRAC file format (bad s-quantizer)" );
    if ( !readParameters( codec, qO ) )
      throw new IOException( "Error in FRAC file format (bad o-quantizer)" );

    ImageBlock root = loadTree( 0, 0, totalSize );
    codec.close();

      // image reconstruction: two temporary floating-point images:
    actIm  = new HDRImage( width, height, RasterGraphics.MODE_GRAY );
    prevIm = new HDRImage( width, height, RasterGraphics.MODE_GRAY );
    int x, y, i;

      // image reconstruction I: random image init
    RandomJames rnd = new RandomJames();
    for ( y = 0; y < height; y++ )
      for ( x = 0; x < width; x++ )
        actIm.putPixel( x, y, rnd.uniformNumber() );

      // reconstruction image-log?:
    if ( logging )
    {
      final int logId = LogFile.createImage( width, height, RasterGraphics.MODE_GRAY, "rec%03d.png" );
      log = LogFile.getImage( logId );
      for ( y = 0; y < height; y++ )
        for ( x = 0; x < width; x++ )
          log.putPixel( x, y, actIm.getGrayDouble( x, y ) );
      LogFile.saveImage( logId );
    }

      // image reconstructin II: transform loop
    minNodeOrder = Integer.MAX_VALUE;
    for ( i = 0; i < MAX_ITERATIONS; i++ )
    {
      HDRImage tmpIm = prevIm;
      prevIm         = actIm;
      actIm          = tmpIm;

        // downsampling source image:
      prevIm.downsample ( 2, true );

        // PIFS transform:
      root.transformImage();

        // reconstruction image-log?:
      if ( logging )
      {
        final int logId = LogFile.createImage( width, height, RasterGraphics.MODE_GRAY, "rec%03d.png" );
        log = LogFile.getImage( logId );
        for ( y = 0; y < height; y++ )
          for ( x = 0; x < width; x++ )
            log.putPixel( x, y, Formula.clamp( actIm.getGrayDouble( x, y ), 0.0, 1.0 ) );
        LogFile.saveImage( logId );
      }
    }

      // initialize raster image object:
    switch ( bands )
    {
      case 1:
        g.init( width, height, RasterGraphics.MODE_GRAY, 0 );
        break;
      case 3:
        g.init( width, height, RasterGraphics.MODE_RGB, 0 );
        break;
      case 4:
        g.init( width, height, RasterGraphics.MODE_RGBA, 0 );
        break;
      default:
        g.init( width, height, RasterGraphics.MODE_BAND, bands );
    }

      // copy the last reconstructed image to the result object:
    for ( y = 0; y < height; y++ )
      for ( x = 0; x < width; x++ )
        g.putPixel( x, y, Formula.clamp( actIm.getGrayDouble( x, y ), 0.0, 1.0 ) );

    actIm  =
    prevIm = null;
  }

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonSave ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    codec  = (EntropyCodec)   getInterface( PL_OUTPUT,    IFACE+"EntropyCodec" );
    search = (RasterSearch)   getInterface( "search",     IFACE+"RasterSearch" );
    qS     = (ScalarQuantizer)getInterface( "quantizerS", IFACE+"ScalarQuantizer" );
    qO     = (ScalarQuantizer)getInterface( "quantizerO", IFACE+"ScalarQuantizer" );

    if ( codec  == null ||
         search == null ||
         qS     == null ||
         qO     == null ) return;

    codec.open( true, CTX_CTRL );
    codec.setMaxSymbol( CTRL_ORI0 + 7 );  // Control/orientation alphabet

      // FRAC header:
    writeUnsigned16( codec, MAGIC );
    width  = g.getWidth();
    writeUnsigned16( codec, width );
    height = g.getHeight();
    writeUnsigned16( codec, height );
    bands  = 1;                           // number of color bands (1 for MODE_GRAY)
    codec.putBits( bands, 7 );
    codec.putBits( bestCodeOnly ? 1L : 0L, 1 ); // only the best blocks will be encoded?

    totalSize = 1 << Formula.bits( Math.max( width, height ) );
    if ( !bestCodeOnly )
    {
      int maxBlocks  = totalSize / minBlockSize;
      maxBlocks     *= maxBlocks;
      blockOrderBits = Formula.bits( maxBlocks ) + 1;
      codec.putBits( blockOrderBits, 8 );
    }
    writeUnsigned16( codec, maxBlockSize );
    writeUnsigned16( codec, sValues );
    writeUnsigned16( codec, oValues );

      // prepare codec contexts:
    codec.setContext( CTX_COORD );
    codec.setMaxSymbol( ((totalSize + 1) >> 1) + maxBlockSize );

    codec.setContext( CTX_S );
    codec.setMaxSymbol( sValues );

    codec.setContext( CTX_O );
    codec.setMaxSymbol( oValues - 1 );

      // set both quantizers:
    qS.setBounds( 0.0, sMax, sValues / 2 );     // positive axis only
    qO.setBounds( -sMax, sMax + 1.0, oValues ); // the whole scale

      // plugin parameters:
    writeParameters( codec, codec );
    writeParameters( codec, qS );
    writeParameters( codec, qO );

      // encoding tree - initial shape:
    heap = new TreeSet< ImageBlock >();
    search.setReferencePicture( g, 0, 0.5, 1 << RasterSearch.ORI_0123, true );
    search.setSearchParameters( 1 << RasterSearch.ORI_0123, 1, RasterSearch.ERR_RMSE, 1.0,
                                1-maxBlockSize, 0, 1-maxBlockSize, 0, 0, 0,
                                -sMax, sMax, -sMax, sMax + 1.0 );
    ImageBlock root = buildInitialTree( g, 0, 0, totalSize );

      // adaptive Q-tree refinement (until all blocks have good enough matches..)
    ImageBlock b;
    int blockOrder = 0;
    while ( !heap.isEmpty() &&
            (b = heap.first()).error > errorLimit )
    {
      if ( userBreak )
      {
        codec.close();
        return;
      }

      heap.remove( b );
      b.order = ++blockOrder;
      LogFile.debug( String.format( Locale.US, "Improving[%3d,%3d,%2d], err: %.5f", b.x0, b.y0, b.size, b.error ) );

      if ( b.size > minBlockSize )        // going to try its descendants
      {
        final int half = b.size >> 1;

        b.setChild( 0, new ImageBlock( b.x0, b.y0, half, g ) );

        if ( b.x0 + half < width )
          b.setChild( 1, new ImageBlock( b.x0 + half, b.y0, half, g ) );

        if ( b.y0 + half < height )
        {
          b.setChild( 2, new ImageBlock( b.x0, b.y0 + half, half, g ) );

          if ( b.x0 + half < width )
            b.setChild( 3, new ImageBlock( b.x0 + half, b.y0 + half, half, g ) );
        }
      }
    }
    heap = null;

      // decomposition image-log?:
    if ( logging )
    {
      final int logId = LogFile.createImage( width, height, RasterGraphics.MODE_RGB, "fract%03d.png" );
      log = LogFile.getImage( logId );
      final double[] pixel = new double[ 3 ];
      for ( int y = 0; y < height; y++ )
        for ( int x = 0; x < width; x++ )
        {
          g.getRGB( x, y, pixel );
          log.putPixel( x, y, pixel );
        }
      root.logTree( 0 );
      LogFile.saveImage( logId );
    }

      // Now Q-tree contains good (or too small) blocks,
      // I'm gonna to encode them:
    root.encode();

    codec.close();
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /** 16-bit magic number (used at the file beginning). */
  public final static int MAGIC = 0xf4ac;

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return 2;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type ({@code 0.0} .. no way,
   *         {@code 0.5} .. maybe, {@code 0.9} .. almost sure, {@code 1.0} ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int len = (header == null) ? 0 : header.length;
    if ( len > 2 ) len = 2;
    if ( len < 2 )                        // not enough bytes in header[]
      return 0.0;
    byte[] h =
    {                                     // RGB header
      (byte)((MAGIC>>8) & 0xff),
      (byte)( MAGIC     & 0xff)
    };
    int i;
    for ( i = 0; i < len; i++ )
      if ( header[i] != h[i] ) return 0.0;
    return 1.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.png"]) or {@code null}.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.frac",
    };
  }

  //==========================================================================
  //  Interface Property:

  public final static String MAX_BLOCK_SIZE = "Max block size";
  public final static String MIN_BLOCK_SIZE = "Min block size";
  public final static String ERROR_LIMIT    = "Error limit";
  public final static String S_VALUES       = "S levels";
  public final static String S_MAX          = "S amplitude";
  public final static String O_VALUES       = "O levels";
  public final static String ONLY_BEST      = "Best code only";
  public final static String LOGGING        = "Log images";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // MAX_BLOCK_SIZE:
    if ( key.compareTo(MAX_BLOCK_SIZE) == 0 )
    {
      maxBlockSize = intProperty( value, maxBlockSize, 8, 1024 );
      return;
    }
      // MIN_BLOCK_SIZE:
    if ( key.compareTo(MIN_BLOCK_SIZE) == 0 )
    {
      minBlockSize = intProperty( value, minBlockSize, 2, 64 );
      return;
    }
      // ERROR_LIMIT:
    if ( key.compareTo(ERROR_LIMIT) == 0 )
    {
      errorLimit = doubleProperty( value, errorLimit, 0.0, 1.0 );
      return;
    }
      // S_VALUES:
    if ( key.compareTo(S_VALUES) == 0 )
    {
      sValues = intProperty( value, sValues, 4, 32768 );
      return;
    }
      // S_MAX:
    if ( key.compareTo(S_MAX) == 0 )
    {
      sMax = doubleProperty( value, sMax, 0.1, 1.5 );
      return;
    }
      // O_VALUES:
    if ( key.compareTo(O_VALUES) == 0 )
    {
      oValues = intProperty( value, oValues, 8, 65536 );
      return;
    }
      // ONLY_BEST:
    if ( key.compareTo(ONLY_BEST) == 0 )
    {
      bestCodeOnly = booleanProperty( value, bestCodeOnly );
      return;
    }
      // LOGGING:
    if ( key.compareTo(LOGGING) == 0 )
    {
      logging = booleanProperty( value, logging );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(MAX_BLOCK_SIZE) == 0 )
      return maxBlockSize;
    if ( key.compareTo(MIN_BLOCK_SIZE) == 0 )
      return minBlockSize;
    if ( key.compareTo(ERROR_LIMIT) == 0 )
      return errorLimit;
    if ( key.compareTo(S_VALUES) == 0 )
      return sValues;
    if ( key.compareTo(S_MAX) == 0 )
      return sMax;
    if ( key.compareTo(O_VALUES) == 0 )
      return oValues;
    if ( key.compareTo(ONLY_BEST) == 0 )
      return bestCodeOnly;
    if ( key.compareTo(LOGGING) == 0 )
        return logging;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Fractal codec";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "DataFileFormatToRasterGraphicsAndRasterSearchAndScalarQuantizer2AndEntropyCodecAndBitStream";

  /** Object description. */
  private final static String DESCRIPTION = "Fractal image compression using adaptive Q-tree partitioning.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,     IFACE+"DataFileFormat" );
      t.newOutputPlug(    PL_RASTER,    IFACE+"RasterGraphics" );
      t.newOptOutputPlug( "search",     IFACE+"RasterSearch" );
      t.newOutputPlug(    "quantizerS", IFACE+"ScalarQuantizer" );
      t.newOutputPlug(    "quantizerO", IFACE+"ScalarQuantizer" );
      t.newOptOutputPlug( PL_STREAM,    IFACE+"BitStream" );
      t.newOptOutputPlug( PL_OUTPUT,    IFACE+"EntropyCodec" );
        // Property:
      t.propBegin( MAX_BLOCK_SIZE, TYPE_INTEGER, "Maximum encoding block size in pixels", true );
        t.propDefault( 32 );
        t.propBounds(  8, 1024 );
      t.propEnd();
      t.propBegin( MIN_BLOCK_SIZE, TYPE_INTEGER, "Minimum encoding block size in pixels", true );
        t.propDefault( 4 );
        t.propBounds(  2, 64 );
      t.propEnd();
      t.propBegin( ERROR_LIMIT, TYPE_DOUBLE, "RMSE error limit for adaptive Q-tree subdivision", true );
        t.propDefault( 0.04 );
        t.propBounds(  0.0, 1.0 );
      t.propEnd();
      t.propBegin( S_VALUES, TYPE_INTEGER, "Number of quantization values for 's' (multiplication)", true );
        t.propDefault( 32 );
        t.propBounds(  4, 32768 );
      t.propEnd();
      t.propBegin( S_MAX, TYPE_DOUBLE, "Maximum amplitude of 's' (multiplication)", true );
        t.propDefault( 1.1 );
        t.propBounds(  0.1, 1.5 );
      t.propEnd();
      t.propBegin( O_VALUES, TYPE_INTEGER, "Number of quantization values for 'o' (additive coef)", true );
        t.propDefault( 128 );
        t.propBounds(  8, 65536 );
      t.propEnd();
      t.propBegin( ONLY_BEST, TYPE_BOOLEAN, "Encode only the best blocks (no redundancy)?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( LOGGING, TYPE_BOOLEAN, "Log intermediate results?", true );
        t.propDefault( true );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
