package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Function samples. R<sup>2</sup>-&gt;R so far.
 *
 * @version 0.26 $Rev: 431 $ $Date: 2007-04-29 22:42:22 +0200 (ne, 29 IV 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/FunctionSamples.java">FunctionSamples.java</a>
 */
public class FunctionSamples extends DefaultRealFunction
{
  //==========================================================================
  //  Data:

  /** Implemented function types. */
  protected static enum FunctionType
  {
    /** Orthogonal waves. */
    FUNC_WAVES_ORTHO,
    /** Radial waves. */
    FUNC_WAVES_RADIAL,
    /** Waves with sinc. */
    FUNC_WAVES_SINC,
  }

  /** Actual function type. */
  protected FunctionType type = FunctionType.FUNC_WAVES_ORTHO;

  /** Actual function amplitude. */
  protected double amplitude  = 1.0;

  /** Actual frequency in X-axis. */
  protected double freqX      = 1.0;

  /** Actual frequency in Y-axis. */
  protected double freqY      = 1.0;

  /** Actual fading coefficient ({@code 0.0} for no fading). */
  protected double fading     = 0.0;

  //==========================================================================
  //  Interface RealFunction:

  /**
   * Compute the mapping: <b>f(x,y)</b>.
   *
   * @param  x X coordinate (1st argument).
   * @param  y Y coordinate (2nd argument).
   */
  @Override
  public double f ( double x, double y )
  {
    x *= freqX;
    y *= freqY;
    double ampl = amplitude;
    double radius = 0.0;
    if ( fading > 0.0 ||
         type == FunctionType.FUNC_WAVES_RADIAL )
    {
      radius = Math.hypot( x, y );
      if ( fading > 0.0 )
        ampl *= Math.exp( -radius * radius * fading );
    }

    switch ( type )
    {
      case FUNC_WAVES_ORTHO:
        return( ampl * (Math.cos( x ) + Math.cos( y )) );

      case FUNC_WAVES_RADIAL:
        return( ampl * Math.cos( radius ) );

      case FUNC_WAVES_SINC:
        radius = Math.abs( x ) + Math.abs( y );
        if ( Math.abs( radius ) < Geometry.SCENE_EPSILON )
          return ampl;
        return( ampl * Math.sin( radius ) / radius );

      default:
        return 0.0;
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String FUNCTION    = "Function";
  public final static String FADING      = "Fading";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // FUNCTION:
    if ( key.compareTo(FUNCTION) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
      // AMPLITUDE:
    if ( key.compareTo(AMPLITUDE) == 0 )
    {
      amplitude = doubleProperty( value, amplitude, 1.e-3, 1.e4 );
      return;
    }
      // FREQUENCY_X:
    if ( key.compareTo(FREQUENCY_X) == 0 )
    {
      freqX = doubleProperty( value, freqX, 0.0, 1.e4 );
      return;
    }
      // FREQUENCY_Y:
    if ( key.compareTo(FREQUENCY_Y) == 0 )
    {
      freqY = doubleProperty( value, freqY, 0.0, 1.e4 );
      return;
    }
      // FADING:
    if ( key.compareTo(FADING) == 0 )
    {
      fading = doubleProperty( value, fading, 0.0, 1.e4 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(FUNCTION) == 0 )
      return type.ordinal();
    if ( key.compareTo(AMPLITUDE) == 0 )
      return amplitude;
    if ( key.compareTo(FREQUENCY_X) == 0 )
      return freqX;
    if ( key.compareTo(FREQUENCY_Y) == 0 )
      return freqY;
    if ( key.compareTo(FADING) == 0 )
      return fading;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "FunctionSamples";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "RealFunction";

  /** Object description. */
  private final static String DESCRIPTION = "Sample functions R^2 -> R.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + "." + C_FUNCTION;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"RealFunction" );
        // Property:
      t.propBegin( FUNCTION, TYPE_INTEGER, "Function type", true );
        t.propDefault( FunctionType.FUNC_WAVES_ORTHO.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Orthogonal waves", FunctionType.FUNC_WAVES_ORTHO.ordinal(),  "cos(x)+cos(y)" );
        t.propEnum( "Radial waves",     FunctionType.FUNC_WAVES_RADIAL.ordinal(), "cos(r)" );
        t.propEnum( "Sinc waves",       FunctionType.FUNC_WAVES_SINC.ordinal(),   "sin(x+y)/(x+y)" );
      t.propEnd();
      t.propBegin( AMPLITUDE, TYPE_DOUBLE, "Amplitude", true );
        t.propDefault( 1.0 );
        t.propBounds(  1.e-3, 1.e4 );
      t.propEnd();
      t.propBegin( FREQUENCY_X, TYPE_DOUBLE, "Frequency X", true );
        t.propDefault( 1.0 );
        t.propBounds(  0.0, 1.e4 );
      t.propEnd();
      t.propBegin( FREQUENCY_Y, TYPE_DOUBLE, "Frequency Y", true );
        t.propDefault( 1.0 );
        t.propBounds(  0.0, 1.e4 );
      t.propEnd();
      t.propBegin( FADING, TYPE_DOUBLE, "Fading", true );
        t.propDefault( 0.0 );
        t.propBounds(  0.0, 1.e4 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
