package cz.cuni.jagrlib.piece;

import java.io.*;
import java.awt.image.*;
import javax.imageio.ImageIO;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * GIF file format (persistence of 2D raster graphics data).
 *
 * @version 0.25 $Rev: 399 $ $Date: 2007-02-12 23:13:47 +0100 (po, 12 II 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/GIFFileFormat.java">GIFFileFormat.java</a>
 */
public class GIFFileFormat extends DefaultRasterFileFormat
{
  //==========================================================================
  //  Data:

  /** Binary file header. */
  protected final static byte[] HEADER =
  {                                       // GIF header
    (byte)'G', (byte)'I', (byte)'F',
    (byte)'8', (byte)'9', (byte)'a'
  };

  //--------------------------------------------------------------------------
  //  Support routines:

  /**
   * Common load code.
   *
   * @param  stream Opened input bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonLoad ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    InputStream gifFile = stream.getInputStream();
    if ( gifFile == null )
      throw new FileNotFoundException();

      // decode GIF file into the raster graphics:
    BufferedImage im = ImageIO.read( gifFile );
    if ( im != null ) g.setBufferedImage( im );
  }

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonSave ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    OutputStream gifFile = stream.getOutputStream();
    if ( gifFile == null )
      throw new FileNotFoundException();

      // encode the raster graphics:
    ImageIO.write( g.getBufferedImage(), "GIF", gifFile );
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null )
      return null;
    return null;
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return HEADER.length;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type (<tt>0.0</tt> .. no way,
   *         <tt>0.5</tt> .. maybe, <tt>0.9</tt> .. almost sure, <tt>1.0</tt> ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int len = (header == null) ? 0 : header.length;
    if ( len > HEADER.length ) len = HEADER.length;
    if ( len < HEADER.length )            // not enough bytes in header[] => look at file-name
    {
      if ( fileName == null || fileName.length() < 5 ) return 0.0;
      int pos = fileName.lastIndexOf( '.' );
      if ( pos < 0 ) return 0.0;
      String suffix = fileName.substring( pos );
      if ( suffix.compareToIgnoreCase( "gif" ) == 0 ) return 0.9;
      return 0.0;
    }

    for ( int i = 0; i < 4; i++ )
      if ( header[i] != HEADER[i] ) return 0.0;
    if ( header[4] != '7' && header[4] != '9' ) return 0.0;
    if ( header[5] != HEADER[5] ) return 0.0;

    return 1.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.gif"]) or <tt>null</tt>.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.gif",
    };
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "GIF file format";

  /** Object description. */
  private final static String DESCRIPTION = "GIF filter, " +
                                            "uses javax.imageio package.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"DataFileFormat" );
      t.newOutputPlug(    PL_RASTER, IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_STREAM, IFACE+"BitStream" );
        // Property:
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
