package cz.cuni.jagrlib.piece;

import javax.media.opengl.GL;
import javax.media.opengl.glu.GLU;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * GL scene rendering of connected {@link Brep}-scene.
 *
 * @author  Stuchl, PE
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/GLBrepRender.java">GLBrepRender.java</a>
 */
public class GLBrepRender extends DefaultRender3D implements GLGraphics
{
  //==========================================================================
  //  Data:

  /** Accelerator: connected Brep scene. */
  protected Brep brep = null;

  /** Angle of perspective view (in degrees). */
  protected float perspectiveFovy = 40.0f;

  /** Near-plane distance. */
  protected float perpectiveZNear = 0.2f;

  /** Far-plane distance. */
  protected float perpectiveZFar = 50.0f;

  /** Background color. */
  protected float[] backgroundColor = { 0.0f, 0.0f, 0.0f, 1.0f };

  /** Current shading model. */
  protected int shadingModel = GL.GL_FLAT;

  /** How many OpenGL lights will be used? ({@code 0} for no GL lighting)  */
  protected int lights = 0;

  /** Constants for "glEnable". */
  protected final static int[] glEnableConstants = { GL.GL_DEPTH_TEST };

  /** Current "glDepthFunc" value. */
  protected int glDepthFunc = GL.GL_LEQUAL;

  /** Hint tuples for "glHint". */
  protected int[][] glHints = { { GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST } };

  /** Clear-depth value. */
  protected float clearDepth = 1.0f;

  //==========================================================================
  //  Painting:

  /**
   * Routine responsible for GL context repainting.
   *
   * @param  gl GL commands access object.
   * @param  glu GLU commands access object.
   */
  public void paintRoutine ( GL gl, GLU glu )
  {
    if ( brep == null )
      brep = (Brep)getInterface( PL_OUTPUT, IFACE+"Brep" );

    initBuffer( gl );
    setupModelView( gl, glu );
    drawBrep( gl );
  }

  /**
   * GL buffer initialization (called at the start of every frame).
   */
  private void initBuffer ( GL gl )
  {
    gl.glClear( GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT );
  }

  /**
   * Model-view matrix setup.
   */
  private void setupModelView ( GL gl, GLU glu )
  {
      // I'm using GL_MODELVIEW mode..
    if ( modelView == null )              // default camera position and orientation
    {
      gl.glLoadIdentity();
      glu.gluLookAt( 0.0f, 8.0f, 0.0f,
                     0.0f, 0.0f, 0.0f,
                     0.0f, 0.0f, 1.0f );
    }
    else
    {
      gl.glLoadMatrixd( modelView.toArray( 4, 4, true, null ), 0 );
    }

      // swap Y <-> Z coordinates for left-handed system..
    gl.glMultMatrixd( new double[] { 1.0, 0.0, 0.0, 0.0,
                                     0.0, 0.0, 1.0, 0.0,
                                     0.0, 1.0, 0.0, 0.0,
                                     0.0, 0.0, 0.0, 1.0 }, 0 );

    /*
    double[] mat = new double[ 16 ];
    gl.glGetDoublev( GL.GL_MODELVIEW_MATRIX, mat, 0 );
    LogFile.log( "ModelView: " + Formula.vectorToString( mat ) );
    */

    mvChanged = false;
  }

  /**
   * Draws Brep scene into the given GL context..
   */
  private void drawBrep ( GL gl )
  {
    int ctx      = Brep.CTX_DEFAULT;
    int fColId   = brep.getAttributeId( Brep.FACE,   ctx, Brep.COLOR );
    int normalId = brep.getAttributeId( Brep.VERTEX, ctx, Brep.NORMAL );
    int vColId   = brep.getAttributeId( Brep.VERTEX, ctx, Brep.COLOR );

    BrepIterator it = brep.faceIterator( ctx, null );

    int[]    vert   = new int[ 3 ];
    float[]  colorF = new float[ 4 ];
    double[] world  = new double[ 4 ];
    double[] normal = new double[ 4 ];

    int faceId, i;
    int faceSize;

    if ( lights > 0 )
      setMaterial( gl );

    while ( (faceId = it.next()) != Brep.NULL && !userBreak )
    {
      while ( (faceSize = brep.getFaceVertices( faceId, vert )) > vert.length )
        vert = new int[ faceSize ];

      if ( shadingModel == GL.GL_FLAT )
        if ( lights == 0 )
        {                                 // flat shading from implicit face colors
          brep.getAttribute( fColId, faceId, colorF );
          gl.glColor4fv( colorF, 0 );
        }
        else
        {                                 // flat shading from vertex normals
          brep.getAttribute( normalId, vert[0], normal );
          for ( i = 1; i < faceSize; i++ )
          {
            brep.getAttribute( normalId, vert[i], world );
            normal[0] += world[0];
            normal[1] += world[1];
            normal[2] += world[2];
          }
          Geometry.normalize3D( normal );
          gl.glNormal3dv( normal, 0 );
        }

      gl.glBegin( GL.GL_POLYGON );
      for ( i = 0; i < faceSize; i++ )    // vertex loop
      {
        if ( shadingModel != GL.GL_FLAT )
          if ( lights == 0 )
          {                               // Gouraud shading using implicit vertex colors
            brep.getAttribute( vColId, vert[i], colorF );
            gl.glColor4fv( colorF, 0 );
          }
          else
          {                               // Gouraud shading using vertex normals
            brep.getAttribute( normalId, vert[i], normal );
            gl.glNormal3dv( normal, 0 );
          }

        brep.getVertexCoords( ctx, vert[i], world );
        gl.glVertex4dv( world, 0 );
      }
      gl.glEnd();
    }
  }

  /**
   * GL-material setup.
   */
  private void setMaterial ( GL gl )
  {
    float[] materialAmbient   = { 0.3f, 0.3f, 0.3f, 1.0f };  // ambient color
    float[] materialDiffuse   = { 0.0f, 1.0f, 0.3f, 1.0f };  // diffuse color
    float[] materialSpecular  = { 0.8f, 0.5f, 0.5f, 1.0f };  // specular color
    float[] materialShininess = { 32.0f };                   // shininess (exponent)

    gl.glMaterialfv( GL.GL_FRONT, GL.GL_AMBIENT,   materialAmbient, 0 );
    gl.glMaterialfv( GL.GL_FRONT, GL.GL_DIFFUSE,   materialDiffuse, 0 );
    gl.glMaterialfv( GL.GL_FRONT, GL.GL_SPECULAR,  materialSpecular, 0 );
    gl.glMaterialfv( GL.GL_FRONT, GL.GL_SHININESS, materialShininess, 0 );
  }

  //==========================================================================
  //  Interface GLGraphics:

  /**
   * Initialization of GL context.
   * It's called in init-time of every GL context.
   *
   * @param  gl GL commands access object.
   * @param  glu GLU commands access object.
   */
  public void initRoutine ( GL gl, GLU glu )
  {
      // Basic GL init:
    gl.glShadeModel( shadingModel );

      // Set background color:
    gl.glClearColor( backgroundColor[0],
                     backgroundColor[1],
                     backgroundColor[2],
                     backgroundColor[3] );

      // glEnable global constants:
    for ( int i : glEnableConstants )
      gl.glEnable( i );
    if ( lights > 0 )
    {
      gl.glEnable( GL.GL_LIGHTING );
      gl.glEnable( GL.GL_LIGHT0 );
    }

      // Depth-buffer setup:
    gl.glClearDepth( clearDepth );
    gl.glDepthFunc(  glDepthFunc );

      // GL-hint tuples:
    for ( int[] i : glHints )
      gl.glHint( i[0], i[1] );

    if ( lights > 0 )
    {
      final float[] lightPosition = { -1.0f, 1.0f, 1.0f, 0.0f };
      gl.glLightfv( GL.GL_LIGHT0, GL.GL_POSITION, lightPosition, 0 );
    }
  }

  /**
   * Reshape routine is called every time the dimensions of GL context are changed.
   * It usually contains projection setup.
   *
   * @param  gl GL commands access object.
   * @param  glu GLU commands access object.
   * @param  x X position of reshaped canvas.
   * @param  y Y position of reshaped canvas.
   * @param  width Width of reshaped canvas.
   * @param  height Height of reshaped canvas.
   */
  public void reshapeRoutine ( GL gl, GLU glu, int x, int y, int width, int height )
  {
    if ( height <= 0 )  // avoid divide by zero
      height = 1;

    final float aspect = (float)width / (float)height;
    gl.glViewport( x, y, width, height );
    vpChanged = false;

      // Projection matrix setup:
    gl.glMatrixMode( GL.GL_PROJECTION );
    if ( projection == null )             // default projection matrix (perspective)
    {
      gl.glLoadIdentity();
      glu.gluPerspective( perspectiveFovy, aspect, perpectiveZNear, perpectiveZFar );
    }
    else
    {
      gl.glLoadMatrixd( projection.toArray( 4, 4, true, null ), 0 );
    }
    prChanged = false;

    /*
    double[] mat = new double[ 16 ];
    gl.glGetDoublev( GL.GL_PROJECTION_MATRIX, mat, 0 );
    LogFile.log( "Projection: " + Formula.vectorToString( mat ) );
    */

      // Model-view matrix setup:
    gl.glMatrixMode( GL.GL_MODELVIEW );
    gl.glLoadIdentity();
  }

  //==========================================================================
  //  Interface Property:

  public final static String USE_LIGHTS     = "GL lights";
  public final static String SHADING_MODEL  = "Shading model";

  /**
   * Sets the given property.
   *
   * @param key Key string.
   * @param value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // USE_LIGHTS:
    if ( key.compareTo(USE_LIGHTS) == 0 )
    {
      lights = intProperty( value, lights, 0, 4 );
      return;
    }
      // SHADING_MODEL:
    if ( key.compareTo(SHADING_MODEL) == 0 )
    {
      shadingModel = intProperty( value, shadingModel );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(USE_LIGHTS) == 0 )
      return lights;
    if ( key.compareTo(SHADING_MODEL) == 0 )
      return shadingModel;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "GLBrepRender";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "GLGraphicsToBrep";

  /** Object description. */
  private final static String DESCRIPTION = "Rendering of Brep scene using OpenGL.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + '.' + C_BREP;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"GLGraphics" );
      t.newInputPlug(  PL_RENDER, IFACE+"Render3D" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"Brep" );
        // Property:
      t.propBegin( USE_LIGHTS, TYPE_INTEGER, "How many GL lights to use?", true );
        t.propDefault( 0 );
        t.propBounds(  0, 4 );
      t.propEnd();
      t.propBegin( SHADING_MODEL, TYPE_INTEGER, "GL shading model (flat, smooth)", true );
        t.propDefault( GL.GL_FLAT );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Flat",    GL.GL_FLAT,   "Flat shading (each face has its constant color)" );
        t.propEnum( "Gouraud", GL.GL_SMOOTH, "Gouraud shading (colors are interpolated from vertices)" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;

  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
