package cz.cuni.jagrlib.piece;

import java.util.*;
import java.util.List;
import java.awt.*;
import java.awt.event.*;

import javax.media.opengl.*;
import javax.media.opengl.glu.GLU;
import javax.swing.*;
//import com.sun.opengl.util.Screenshot;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Window able to display GL graphics.
 *
 * @author  Stuchl
 * @version 0.26 $Rev: 702 $ $Date: 2008-01-26 10:00:15 +0100 (so, 26 I 2008) $ $Author: Tomas $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/GLWindow.java">GLWindow.java</a>
 */
public class GLWindow extends Piece implements GraphicsViewer
{
  //==========================================================================
  //  Data:

  /** Default canvas width in pixels. */
  protected static final int DEFAULT_WIDTH = 800;

  /** Default canvas height in pixels. */
  protected static final int DEFAULT_HEIGHT = 600;

  /** Current width of rendering canvas (in pixels). */
  protected int width  = DEFAULT_WIDTH;

  /** Current height of rendering canvas (in pixels). */
  protected int height = DEFAULT_HEIGHT;

  /** Current state of fulscreen mode. */
  protected boolean fullscreen = false;

  /** Minimal X coordinate of actual invalid region. */
  protected int x1 = Integer.MAX_VALUE;

  /** Maximal X coordinate (outer bound) of actual invalid region. */
  protected int x2 = Integer.MIN_VALUE;

  /** Minimal Y coordinate of actual invalid region. */
  protected int y1 = Integer.MAX_VALUE;

  /** Maximal Y coordinate (outer bound) of actual invalid region. */
  protected int y2 = Integer.MIN_VALUE;

  /** Minimal canvas width. */
  protected static final int CANVAS_WIDTH_MIN = 32;

  /** Maximal canvas width. */
  protected static final int CANVAS_WIDTH_MAX = 4096;

  /** Minimal canvas height. */
  protected static final int CANVAS_HEIGHT_MIN = 32;

  /** Maximal canvas height. */
  protected static final int CANVAS_HEIGHT_MAX = 4096;

  /** Default fullscreen value. */
  protected static final boolean DEFAULT_FULLSCREEN_VALUE = false;

  /** Default window title. */
  protected static final String DEFAULT_FRAME_TITLE = "GL Graphics";

  /** Frame title (in order that setTitle() works every time). */
  protected String frameTitle = DEFAULT_FRAME_TITLE;

  /** Swing frame window to show graphic output in. */
  protected JFrame frame;

  /** GL canvas component - drawing target. */
  protected GLCanvas canvas;

  /** Actual GL graphics module (rendering source). */
  protected GLGraphics gr;

  /**
   * Interthread variable for creating snapshot. Snapshot can be done only in
   * drawing thread. This variable tells drawing thread to make snapshot.
   *
   * @see GLEventListener#display
   */
  protected boolean makeSnapshot = false;

  /**
   * Breaks the <b>repaint loop</b> from outside the {@link #repaintLoop} call.
   *
   * @see #repaintLoop
   * @see #stopRepaintLoop
   */
  protected boolean continueRepaintLoop;

  /** Has the user-break occurred? */
  protected boolean userBreak;

  /**
   * Semaphores that will be signaled after the viewer window is closed.
   *
   * @see Semaphore
   * @see #addSemaphore
   * @see #removeSemaphore
   */
  protected List< Semaphore > semaphores;

  /**
   * Set of actual input listeners.
   *
   * @see #addInputListener
   * @see #removeInputListener
   */
  protected List< InputListener > listeners;

  /** InputListener connected through "direct" plug. */
  protected InputListener listener;

  /** Minimal interval between two FPS-display times. */
  protected static final long FPS_PERIOD = 1000L;

  /** FPS-queue size (maximal number of previous frames used in FPS computation). */
  protected static final int FPS_QUEUE_SIZE = 80;

  /** The FPS-queue data, actual frame is defined by {@link #fpsPtr}. */
  protected final long[] fpsQueue = new long[ FPS_QUEUE_SIZE ];

  /** Index of actual frame in {@link #fpsQueue}. */
  protected int fpsPtr = 0;

  /** The last FPS-display time. */
  protected long fpsLast = System.currentTimeMillis();

  //==========================================================================
  //  Local listener classes:

  /**
   * Inner class responsible for window-close event handling (by user interaction).
   *
   * @version 0.26 $Rev: 702 $ $Date: 2008-01-26 10:00:15 +0100 (so, 26 I 2008) $ $Author: Tomas $
   */
  private class LocalWindowListener extends WindowAdapter
  {
    /** Close-window event handler. */
    @Override
    public void windowClosing ( WindowEvent e )
    {
      synchronized ( GLWindow.this )
      {
        userBreak = true;
        continueRepaintLoop = false;
        GLWindow.this.notify();
      }

      // signal the semaphores:
      if ( semaphores != null ) // signal all the semaphores
      {
        for ( Semaphore currentSemaphore : semaphores )
          if ( currentSemaphore != null )
            currentSemaphore.semSignal( 1 );
      }
      // close the window manually:
      e.getWindow().dispose();
    }
  }

  private class LocalMouseListener extends MouseAdapter
  {
    @Override
    public void mousePressed ( MouseEvent e )
    {
      if ( listener == null && listeners == null )
        return;

      e.consume();

      if ( listener != null )
        listener.mouseButton( this, e.getWhen(), e.getX(), e.getY(), e.getButton(), true, 0 );

      if ( listeners == null )
        return;

      for ( InputListener currentInputListener : listeners )
        currentInputListener.mouseButton( this, e.getWhen(), e.getX(), e.getY(), e.getButton(), true, 0 );
    }

    @Override
    public void mouseReleased ( MouseEvent e )
    {
      if ( listener == null && listeners == null )
        return;

      e.consume();

      if ( listener != null )
        listener.mouseButton( this, e.getWhen(), e.getX(), e.getY(), e.getButton(), false, 0 );

      if ( listeners == null )
        return;

      for ( InputListener currentInputListener : listeners )
        currentInputListener.mouseButton( this, e.getWhen(), e.getX(), e.getY(), e.getButton(), false, 0 );
    }

    @Override
    public void mouseEntered ( MouseEvent e )
    {
      if ( listener != null )
        listener.mouseMove( this, e.getWhen(), e.getX(), e.getY(), 0 );

      if ( listeners == null )
        return;

      for ( InputListener currentInputListener : listeners )
        currentInputListener.mouseMove( this, e.getWhen(), e.getX(), e.getY(), 0 );
    }
  }

  private class LocalMouseMotionListener implements MouseMotionListener
  {
    public void mouseDragged ( MouseEvent e )
    {
      if ( listener == null && listeners == null )
        return;

      e.consume();

      if ( listener != null )
        listener.mouseMove( this, e.getWhen(), e.getX(), e.getY(), 0 );

      if ( listeners == null )
        return;

      for ( InputListener currentInputListener : listeners )
        currentInputListener.mouseMove( this, e.getWhen(), e.getX(), e.getY(), 0 );
    }

    public void mouseMoved ( MouseEvent e )
    {
      if ( listener == null && listeners == null )
        return;

      e.consume();

      if ( listener != null )
        listener.mouseMove( this, e.getWhen(), e.getX(), e.getY(), 0 );

      if ( listeners == null )
        return;

      for ( InputListener currentInputListener : listeners )
        currentInputListener.mouseMove( this, e.getWhen(), e.getX(), e.getY(), 0 );
    }
  }

  private class LocalMouseWheelListener implements MouseWheelListener
  {
    public void mouseWheelMoved ( MouseWheelEvent e )
    {
      if ( listener == null && listeners == null )
        return;

      e.consume();

      if ( listener != null )
        listener.mouseWheel( this, e.getWhen(), e.getX(), e.getY(), e.getButton(), e.getWheelRotation(), 0 );

      if ( listeners == null )
        return;

      for ( InputListener currentInputListener : listeners )
        currentInputListener.mouseWheel( this, e.getWhen(), e.getX(), e.getY(), e.getButton(), e.getWheelRotation(), 0 );
    }
  }

  private class LocalKeyListener extends KeyAdapter
  {
    @Override
    public void keyPressed ( KeyEvent e )
    {
      if ( listener == null && listeners == null )
        return;

      e.consume();

      if ( listener != null )
        listener.keyboard( this, e.getWhen(), true, e.getKeyChar(), e.getKeyCode(), e.getModifiersEx() );

      if ( listeners == null )
        return;

      for ( InputListener currentInputListener : listeners )
        currentInputListener.keyboard( this, e.getWhen(), true, e.getKeyChar(), e.getKeyCode(), e.getModifiersEx() );
    }

    @Override
    public void keyReleased ( KeyEvent e )
    {
      if ( listener == null && listeners == null )
        return;

      e.consume();

      if ( listener != null )
        listener.keyboard( this, e.getWhen(), false, e.getKeyChar(), e.getKeyCode(), e.getModifiersEx() );

      if ( listeners == null )
        return;

      for ( InputListener currentInputListener : listeners )
        currentInputListener.keyboard( this, e.getWhen(), false, e.getKeyChar(), e.getKeyCode(), e.getModifiersEx() );
    }
  }

  private class LocalComponentListener extends ComponentAdapter
  {
    @Override
    public void componentResized ( ComponentEvent e )
    {
      if ( listener == null && listeners == null )
        return;

      Component c = e.getComponent();

      if ( listener != null )
        listener.setWindowSize( this, c.getWidth(), c.getHeight() );

      if ( listeners == null )
        return;

      for ( InputListener currentListener : listeners )
        currentListener.setWindowSize( this, c.getWidth(), c.getHeight() );

      canvas.repaint();
    }
  }

  /**
   * GL listener for handling GL canvas initializing and refreshing.
   *
   * @author  Stuchl
   * @version 0.26 $Rev: 702 $ $Date: 2008-01-26 10:00:15 +0100 (so, 26 I 2008) $ $Author: Tomas $
   */
  private class LocalGLEventListener implements GLEventListener
  {
    /** Current GLGraphics (drawing routines). */
    private final GLGraphics glGraphics;

    public LocalGLEventListener ( GLGraphics glGraphics )
    {
      this.glGraphics = glGraphics;
    }

    /**
     * Called by the drawable immediately after the OpenGL context is
     * initialized. Can be used to perform one-time OpenGL initialization such
     * as setup of lights and display lists. Note that this method may be called
     * more than once if the underlying OpenGL context for the GLAutoDrawable is
     * destroyed and recreated, for example if a GLCanvas is removed from the
     * widget hierarchy and later added again.
     */
    public void init ( GLAutoDrawable drawable )
    {
      LogFile.debug( "JOGL GL init() from GLEventListener implementation." );
      // FOR DEBUGING GL:
      //drawable.setGL( new DebugGL( drawable.getGL() ) );

      final GL gl = drawable.getGL();
      GLU glu = new GLU();

      glGraphics.initRoutine( gl, glu );
    }

    /**
     * Called by the drawable to initiate OpenGL rendering by the client. After
     * all GLEventListeners have been notified of a display event, the drawable
     * will swap its buffers if {@link GLAutoDrawable#setAutoSwapBufferMode
     * setAutoSwapBufferMode} is enabled.
     */
    public void display ( GLAutoDrawable drawable )
    {
      LogFile.debug( "JOGL GL display() from GLEventListener implementation." );
      try
      {
        //fetch GL and GLU instances
        final GL gl = drawable.getGL();
        GLU glu = new GLU();

        glGraphics.paintRoutine( gl, glu ); // paints connected gl commands
        gl.glFlush();
      }
      catch ( GLException e )
      {
        LogFile.exception( e );
      }
    }

    /**
     * Called by the drawable during the first repaint after the component has
     * been resized. The client can update the viewport and view volume of the
     * window appropriately, for example by a call to
     * {@link javax.media.opengl.GL#glViewport}; note that for convenience the
     * component has already called {@code glViewport(x,y,width,height)}
     * when this method is called, so the client may not have to do anything in
     * this method.
     */
    public void reshape ( GLAutoDrawable drawable, int x, int y, int width, int height )
    {
      LogFile.debug( "JOGL GL reshape() from GLEventListener implementation." );
      try
      {
        final GL gl = drawable.getGL();
        final GLU glu = new GLU();

        glGraphics.reshapeRoutine( gl, glu, x, y, width, height );

        //reset transform matrix
        gl.glLoadIdentity();
      }
      catch ( GLException e )
      {
        LogFile.exception( e );
      }
    }

    /**
     * <p>Called by the drawable when the display mode or the display device
     * associated with the GLAutoDrawable has changed. The two boolean
     * parameters indicate the types of change(s) that have occurred.
     * </p>
     * <p>An example of a display <i>mode</i> change is when the bit depth changes
     * (e.g., from 32-bit to 16-bit color) on monitor upon which the
     * GLAutoDrawable is currently being displayed.
     * </p>
     * <p>An example of a display <i>device</i> change is when the user drags the
     * window containing the GLAutoDrawable from one monitor to another in a
     * multiple-monitor setup.
     * </p>
     * <p>The reason that this function handles both types of changes (instead of
     * handling mode and device changes in separate methods) is so that
     * applications have the opportunity to respond to display changes in the
     * most efficient manner. For example, the application may need make fewer
     * adjustments to compensate for a device change if it knows that the mode
     * on the new device is identical the previous mode.
     * </p>
     *
     * <p><b>NOTE: Implementations are not required to implement this method. The
     * Reference Implementation DOES NOT IMPLEMENT this method.</b>
     * </p>
     */
    public void displayChanged ( GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged )
    {
      LogFile.debug( "JOGL GL displayChanged() from GLEventListener implementation." );
      // Not yet implemented in JOGL
    }
  }

  //--------------------------------------------------------------------------
  //  Support routines:

  /**
   * Does the repaint job itself (according the <b>invalid region</b>).
   *
   * @see #repaintLoop
   */
  protected void doRepaint()
  {
    if ( canvas == null || x2 <= x1 || y2 <= y1 ) return;

    canvas.repaint( canvas.getX() + x1, canvas.getY() + y1, x2 - x1, y2 - y1 );
    x1 = y1 = Integer.MAX_VALUE;
    x2 = y2 = Integer.MIN_VALUE;

      // FPS estimation:
    final long now = System.currentTimeMillis();
    fpsQueue[ fpsPtr ] = now;
    if ( ++fpsPtr >= FPS_QUEUE_SIZE ) fpsPtr = 0;

    if ( now > fpsLast + FPS_PERIOD )     // display the FPS-estimate
    {
      fpsLast = now;

      int frames = FPS_QUEUE_SIZE - 1;    // comparing "now" to fpsQueue[ptr]
      int ptr    = fpsPtr;
      final long oldest = now - FPS_PERIOD - FPS_PERIOD;
      while ( frames > 1 &&
              oldest > fpsQueue[ptr] )
      {
        if ( ++ptr >= FPS_QUEUE_SIZE ) ptr = 0;
        frames--;
      }

        // Average FPS:
      int fps = (now == fpsQueue[ptr]) ? 1212 : Formula.round( (1000.0 * frames) / (now - fpsQueue[ptr]) );
      frame.setTitle( String.format( "%s (%d fps)", frameTitle, fps ) );
    }
  }

  //==========================================================================
  //  Interface GraphicsViewer:

  /**
   * Start (potentially) infinite <b>repaint loop</b>. It can be terminated
   * either by a {@code stopRepaintLoop()} call or by external user
   * interaction (user break, window close, ..).
   *
   * @return {@code true} if the loop was terminated by a
   *         {@link #stopRepaintLoop} call.
   * @see    #stopRepaintLoop
   */
  public synchronized boolean repaintLoop ()
  {
    if ( frame == null )                  // looping for the first time
    {
      try
      {
        gr = (GLGraphics)getInterface( PL_OUTPUT, IFACE+"GLGraphics" );
      }
      catch ( JaGrLibException e )
      {
        LogFile.exception( e );
        return false;
      }

      if ( gr == null )
      {
        LogFile.error( "GLGraphics is not connected." );
        return false;
      }

        // Fetch GL canvas
      canvas = new GLCanvas();

        // Create new swing frame
      frame = new JFrame( frameTitle );

        // Get connected InputListener
      listener = (InputListener)getInterface( PL_DIRECT, IFACE+"InputListener" );

        // Add new window listener (closing, resizing) to frame
      frame.addWindowListener( new LocalWindowListener() );

        // Add GL canvas to newly created frame
      frame.getContentPane().add( canvas, BorderLayout.CENTER );

        // Add GL event listener (repaints and draws scene from GLGraphics implementation)
        // !!! FIXME: must be created new instance for each GLDrawable
      canvas.addGLEventListener( new LocalGLEventListener( gr ) );

        // Add mouse * listeners to cacth mouse actions
      canvas.addMouseListener(       new LocalMouseListener() );
      canvas.addMouseMotionListener( new LocalMouseMotionListener() );
      canvas.addMouseWheelListener(  new LocalMouseWheelListener() );

        // Add key listener to cacch key strokes
      LocalKeyListener lkl = new LocalKeyListener();
      frame.addKeyListener( lkl );

        // Add component litener (here is GLCanvas the only component)
      canvas.addComponentListener( new LocalComponentListener() );

        // Set auto swapping buffers of GL context
        // !!! TODO: add this propery to GLWindow property list
      canvas.setAutoSwapBufferMode( true );

        // Expose the frame:
      setSize();
      frame.pack();

        // Show the frame
      frame.setVisible( true );
    }

    continueRepaintLoop = true;           // looping state variables
    userBreak           = false;
    canvas.repaint();

    do                                    // repaint-loop itself
    {
      try
      {
        wait();
      }
      catch ( InterruptedException e )
      {
        continueRepaintLoop = false;
        userBreak = true;
      }

      if ( !continueRepaintLoop ) break;

      doRepaint();

    } while ( continueRepaintLoop );

    return !userBreak;
  }

  /**
   * Destroys the preview window. Stops repaint-loop eventually.
   */
  public synchronized void destroy ()
  {
    if ( frame != null )
    {
      if ( continueRepaintLoop )
      {
        continueRepaintLoop = false;
        notify();
      }
      frame.setVisible( false );
      frame = null;
    }
  }

  /**
   * Stops the running <b>repaint loop</b>.
   *
   * @return {@code true} if the loop was successfully stopped.
   * @see #repaintLoop
   */
  public synchronized boolean stopRepaintLoop ()
  {
    if ( frame == null || !continueRepaintLoop ) return false;

    continueRepaintLoop = false;
    notify();
    return true;
  }

  /**
   * Causes the instant <b>repainting</b> of the whole graphic output
   * (regardless of the actual <b>invalid region</b>).
   * {@code Trigger.fire()} call has the same effect..
   *
   * @see #invalidate
   * @see #repaint
   * @see Trigger#fire
   */
  public synchronized void repaintAll ()
  {
    if ( frame == null || !continueRepaintLoop ) return;

    x1 = 0;
    x2 = canvas.getWidth();
    y1 = 0;
    y2 = canvas.getHeight();
    notify();
  }

  /**
   * Invalidates the given rectangular region of the graphics output. Invalid
   * regions will be repainted during the next {@code repaint()} call...
   *
   * @param xMin X coordinate of the upper left corner.
   * @param xMax X coordinate of the lower right corner (outside pixel).
   * @param yMin Y coordinate of the upper left corner.
   * @param yMax Y coordinate of the upper left corner (outside pixel).
   * @see   #repaint
   */
  public synchronized void invalidate ( int xMin, int xMax, int yMin, int yMax )
  {
    if ( xMin < x1 ) x1 = xMin;
    if ( xMax > x2 ) x2 = xMax;
    if ( yMin < y1 ) y1 = yMin;
    if ( yMax > y2 ) y2 = yMax;
  }

  /**
   * Causes the instant <b>repainting</b> of actual <b>invalid region</b> of
   * the graphic output.
   *
   * @see #repaintAll
   * @see #invalidate
   */
  public synchronized void repaint ()
  {
    if ( frame == null || !continueRepaintLoop ) return;

    notify();
  }

  /**
   * Adds a new <b>semaphore</b> - object which wants to be signaled after the
   * viewer window is closed (by user interaction etc.).
   *
   * @param s
   *          A new semaphore instance.
   * @see #removeSemaphore
   */
  public synchronized void addSemaphore ( Semaphore s )
  {
    if ( s == null ) return;

    if ( semaphores == null )             // the first semaphore
      semaphores = new LinkedList< Semaphore >();

    semaphores.add( s );
  }

  /**
   * Removes one of associated <b>semaphores</b>.
   *
   * @param s Semaphore to be removed.
   * @see   #addSemaphore
   */
  public synchronized void removeSemaphore ( Semaphore s )
  {
    if ( s == null || semaphores == null ) return;

    semaphores.remove( s );
    if ( semaphores.size() == 0 )
      semaphores = null;
  }

  /** Adds input listener object. */
  public void addInputListener ( InputListener list )
  {
    if ( list == null ) return;

    if ( listeners == null )              // the fist listener
      listeners = new LinkedList< InputListener >();

    listeners.add( list );
  }

  /** Removes (previously added) input listener object. */
  public void removeInputListener ( InputListener list )
  {
    if ( list == null || listeners == null ) return;

    listeners.remove( list );
    if ( listeners.size() == 0 )
      listeners = null;
  }

  public void hudClear ()
  {
    throw new UnsupportedOperationException( "Method " + this.getClass().getName() + ".hudClear" );
  }

  public int hudGetFreeId ()
  {
    throw new UnsupportedOperationException( "Method " + this.getClass().getName() + ".hudGetFreeId" );
  }

  public void hudSetLine ( int id, int x1, int y1, int x2, int y2 )
  {
    throw new UnsupportedOperationException( "Method " + this.getClass().getName() + ".hudSetLine" );
  }

  public void hudRemove ( int id )
  {
    throw new UnsupportedOperationException( "Method " + this.getClass().getName() + ".hudRemove" );
  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Triggers the (undetermined) action.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public synchronized boolean fire ( int type )
  {
    repaintAll();
    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String FULLSCREEN    = "fullscreen";
  public final static String CANVAS_WIDTH  = "Canvas width";
  public final static String CANVAS_HEIGHT = "Canvas height";

  /**
   * Sets the given property.
   *
   * @param key Key string.
   * @param value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // WINDOW_TITLE:
    if ( key.compareTo(WINDOW_TITLE) == 0 )
    {
      frameTitle = value.toString();
      if ( frame != null )
        frame.setTitle( frameTitle );
      return;
    }
      // CANVAS_WIDTH:
    if ( key.compareTo(CANVAS_WIDTH) == 0 )
    {
      width = intProperty( value, width, CANVAS_WIDTH_MIN, CANVAS_WIDTH_MAX );
      return;
    }
      // CANVAS_HEIGHT:
    if ( key.compareTo(CANVAS_HEIGHT) == 0 )
    {
      height = intProperty( value, height, CANVAS_HEIGHT_MIN, CANVAS_HEIGHT_MAX );
      return;
    }
      // FULLSCREEN
    if ( key.compareTo(FULLSCREEN) == 0 )
    {
      fullscreen = booleanProperty( value, DEFAULT_FULLSCREEN_VALUE );
      setSize();
      return;
    }
  }

  /**
   * Sets size of frame due to fullscreen variable and width and height
   * variables
   */
  public void setSize ()
  {
    if ( canvas == null || frame == null ) return;

    if ( fullscreen )
    {
      Dimension size = Toolkit.getDefaultToolkit().getScreenSize().getSize();
      if ( !frame.isDisplayable() )
      {
          // Remove window decorations
        frame.setUndecorated( true );
      }
      canvas.setBounds( 0, 0, size.width, size.height );
      frame.setSize( size );
    }
    else
    {
      canvas.setBounds( 0, 0, width, height );
      frame.setSize( width, height );
    }
    repaintAll();
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(WINDOW_TITLE) == 0 )
      return frameTitle;
    if ( key.compareTo(CANVAS_WIDTH) == 0 )
      return width;
    if ( key.compareTo(CANVAS_HEIGHT) == 0 )
      return height;
    if ( key.compareTo(FULLSCREEN) == 0 )
      return fullscreen;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "GL graphics viewer";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "GraphicsViewerToGLGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "GL graphics is rendered in simple Swing window.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_VIEWER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"GraphicsViewer" );
      t.newOutputPlug(    PL_OUTPUT, IFACE+"GLGraphics" );
      t.newOptOutputPlug( PL_DIRECT, IFACE+"InputListener" );
        // Property:
      t.propBegin( CANVAS_WIDTH, TYPE_INTEGER, "Canvas width in pixels", true );
        t.propDefault( DEFAULT_WIDTH );
        t.propBounds( CANVAS_WIDTH_MIN, CANVAS_WIDTH_MAX );
      t.propEnd();
      t.propBegin( CANVAS_HEIGHT, TYPE_INTEGER, "Canvas height in pixels", true );
        t.propDefault( DEFAULT_HEIGHT );
        t.propBounds( CANVAS_HEIGHT_MIN, CANVAS_HEIGHT_MAX );
      t.propEnd();
      t.propBegin( WINDOW_TITLE, TYPE_STRING, "Window title", true );
        t.propDefault( DEFAULT_FRAME_TITLE );
      t.propEnd();
      t.propBegin( FULLSCREEN, TYPE_BOOLEAN, "Fullscreen mode", true );
        t.propDefault( DEFAULT_FULLSCREEN_VALUE );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
