package cz.cuni.jagrlib.piece;

import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * HDR file format used in Radiance (RGBe RLE-encoded format).
 *
 * @version 0.25 $Rev: 399 $ $Date: 2007-02-12 23:13:47 +0100 (po, 12 II 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/HDRFileFormat.java">HDRFileFormat.java</a>
 */
public class HDRFileFormat extends DefaultRasterFileFormat
{
  //==========================================================================
  //  Data:

  /** Binary file header. */
  protected final static byte[] HEADER =
  {                                       // Radiance header
    (byte)'#', (byte)'?', (byte)'R', (byte)'A',
    (byte)'D', (byte)'I', (byte)'A', (byte)'N',
    (byte)'C', (byte)'E'
  };

  /** RGBe scanline, byte order: <tt>[ Red, Green, Blue, Exp ]</tt>. */
  protected static byte[] scanline = null;

  /** Minimum scanline length for encoding. */
  protected final static int MIN_ELEN = 8;

  /** Maximum scanline length for encoding. */
  protected final static int MAX_ELEN = 0x7fff;

  /** Minimum run length. */
  protected final static int MIN_RUN  = 4;

  //--------------------------------------------------------------------------
  //  Support routines:

  /** Asserts minimum length of the {@link #scanline} array (in pixels). */
  protected static void assertScanline ( int width )
  {
    int len = width << 2;
    if ( scanline != null &&
         scanline.length >= len ) return;

    scanline = new byte[ len ];
  }

  /**
   * Reads data from the old binary format into the {@link #scanline} array.
   *
   * @param  start Starting index into the {@link #scanline} array (not pixel index!).
   * @param  len Length to read (in pixels).
   * @param  is Binary stream to read from.
   * @param  unget First byte read before calling this method (or <tt>-1</tt> if none was).
   */
  protected static void oldReadScanline ( int start, int len, InputStream is, int unget )
  throws IOException
  {
    int repShift = 0;                     // LSB/MSB RLE repeat attribute
    int i;
    int readBunch = (unget < 0) ? 4 : 3;
    if ( unget >= 0 )
      scanline[start++] = (byte)unget;

    while ( len-- > 0 )                   // read one quadruple
    {
      for ( i = 0; i++ < readBunch; )
        scanline[start++] = (byte)is.read(); // reading rour bytes

      if ( scanline[start-2] == 1 &&
           scanline[start-3] == 1 &&
           scanline[start-4] == 1 )       // RLE packet
      {
        start -= 4;
        i      = ((int)scanline[start+3] & 0xff) << repShift;
                                          // number of pixels to repeat
        len   -= (i - 1);                 // one item was counted before
        i    <<= 2;                       // number of bytes

        while ( i-- > 0 )
        {
          scanline[start] = scanline[start-4];
          start++;
        }

        repShift += 8;                    // next RLE packet would contain MSB repeat count
      }

      else                                // regular pixel
        repShift = 0;

      readBunch = 4;
    }
  }

  /** Reads data in new RLE format (band-ordered) into the {@link #scanline} array. */
  protected static void readScanline ( int len, InputStream is )
  throws IOException
  {
    assertScanline( len );

    if ( len < MIN_ELEN ||
         len > MAX_ELEN )                 // use old scanline format
    {
      oldReadScanline( 0, len, is, -1 );
      return;
    }

    int b = is.read();
    if ( b != 2 )                         // use old scanline format
    {
      oldReadScanline( 0, len, is, b );
      return;
    }

      // check scanline header - should be [ 2, 2, hi(len), lo(len) ]:
    scanline[0] = (byte)b;
    scanline[1] = (byte)is.read();
    scanline[2] = (byte)is.read();
    scanline[3] = (byte)is.read();

    if ( scanline[1] != 2 ||
         (scanline[2] & 0x80) > 0 )       // header mismatch => use the old format
    {
      oldReadScanline( 4, len - 1, is, -1 );
      return;
    }

    int readLen =  ((int)scanline[3] & 0xff) +
                  (((int)scanline[2] & 0x7f) << 8);
    if ( readLen != len )                 // header lenght mismatch => error
      throw new CharConversionException( "Error in HDR scanline" );

      // read regular scanline data:
    int band;                             // band number (0 for Red, .. 3 for Exp)
    int rep;                              // repeat value (for RLE)
    for ( band = 0; band < 4; band++ )
    {
      b = band;                           // index into scanline array
      int i = 0;
      while ( i < len )                   // read one component
      {
        rep = is.read();
        if ( rep > 128 )                  // RLE packet
        {
          rep     &= 0x7f;
          byte val = (byte)is.read();
          while ( rep-- > 0 )
          {
            scanline[b] = val;
            b          += 4;
            i++;
          }
        }
        else                              // raw packet
          while ( rep-- > 0 )
          {
            scanline[b] = (byte)is.read();
            b          += 4;
            i++;
          }
      }
    }
  }

  protected static void writeScanline ( int len, OutputStream os )
  throws IOException
  {
    if ( len < MIN_ELEN ||
         len > MAX_ELEN )                 // use old scanline format
    {
      os.write( scanline, 0, len << 2 );
      return;
    }

    os.write( 2 );
    os.write( 2 );
    os.write( len >> 8 );
    os.write( len & 0x7f );

      // put 4 components separately:
    len <<= 2;
    int b, band;
    for ( band = 0; band < 4; band++ )
    {
      b = band;                           // index into scanline array
      do                                  // encode at least one packet: [raw] [run]
      {
          // looking for RLE packet starting at 'b'
        int run  = 0;
        int brun = b + 4;
        while ( brun < len )
        {
          if ( scanline[brun] == scanline[brun-4] )
          {
            if ( ++run >= MIN_RUN-1 ) break; // found the run!
          }
          else
            run = 0;

          brun += 4;
        }

        brun -= (run << 2);               // start of the run
        int rawLen = (brun - b) >> 2;

          // now write out raw packets:
        while ( rawLen > 0 )
        {
          int rl = Math.min( 128, rawLen );

            // write raw packet of size 'rl' starting at 'b':
          os.write( rl );
          rawLen -= rl;
          do
          {
            os.write( scanline[b] );
            b += 4;
          }
          while ( --rl > 0 );
        }

        if ( run >= MIN_RUN-1 )           // write one run
        {
          brun = scanline[b];             // value to repeat
          run = 1;
          while ( (b += 4) < len &&
                  scanline[b] == brun &&
                  ++run < 127 ) ;
          if ( run == 127 ) b += 4;

          os.write( run + 128 );
          os.write( brun );
        }
      }
      while ( b < len );
    }
  }

  /**
   * Common load code.
   *
   * @param  stream Opened input bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonLoad ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    InputStream hdrFile = stream.getInputStream();
    if ( hdrFile == null )
      throw new FileNotFoundException();

      // decode HDR file into raster graphics:
      // "#?RADIANCE ... 0x0a, 0x0a {+|-}Y <height> {+|-}X <width>"
    boolean bottomUp    = false;
    boolean leftToRight = true;
    int width           = 0;
    int height          = 0;

    try
    {
      int last;
      int act = hdrFile.read();
      do
      {
        last = act;
        act  = hdrFile.read();
      }
      while ( act >= 0 && (last != 0x0a || act != 0x0a) );

      if ( act < 0 )
        throw new CharConversionException( "Error in HDR header" );

      String axis1 = readToken( hdrFile );
      int    dim1  = Integer.parseInt( readToken( hdrFile ) );
      String axis2 = readToken( hdrFile );
      int    dim2  = Integer.parseInt( readToken( hdrFile ) );

      if ( axis1 == null || axis1.length() != 2 ||
           axis1 == null || axis1.length() != 2 )
        throw new CharConversionException( "Error in HDR header" );

      if ( Character.toLowerCase( axis1.charAt( 1 ) ) == 'x' )
      {
        width       = dim1;
        leftToRight = (axis1.charAt( 0 ) == '+');
      }
      else
      if ( Character.toLowerCase( axis1.charAt( 1 ) ) == 'y' )
      {
        height      = dim1;
        bottomUp    = (axis1.charAt( 0 ) == '+');
      }
      else
        throw new CharConversionException( "Error in HDR header" );

      if ( Character.toLowerCase( axis2.charAt( 1 ) ) == 'x' )
      {
        width       = dim2;
        leftToRight = (axis2.charAt( 0 ) == '+');
      }
      else
      if ( Character.toLowerCase( axis2.charAt( 1 ) ) == 'y' )
      {
        height      = dim2;
        bottomUp    = (axis2.charAt( 0 ) == '+');
      }
      else
        throw new CharConversionException( "Error in HDR header" );

    }
    catch ( NumberFormatException ex )
    {
      throw new CharConversionException( "Error in HDR header" );
    }

    if ( width <= 0 || height <= 0 )
      throw new CharConversionException( "Error in HDR header" );

      // init memory bitmap ..
    g.init( width, height, RasterGraphics.MODE_RGB, 3 );

      // .. and finally read the binary data into it:
    int x, y, dy, i;
    double[] pix = new double[ 3 ];
    if ( bottomUp )
    {
      y  = height - 1;
      dy = -1;
    }
    else
    {
      y  = 0;
      dy = 1;
    }

    while ( y >= 0 && y < height )        // read one scanline
    {
      readScanline( width, hdrFile );

        // convert RGBe pixels into float[3] RGB:
      if ( leftToRight )
        for ( x = i = 0; x < width; x++, i += 4 )
        {
          Formula.RGBeToRGB( scanline, i, pix, 0 );
          g.putPixel( x, y, pix );
        }
      else
        for ( x = width, i = 0; --x >= 0; i += 4 )
        {
          Formula.RGBeToRGB( scanline, i, pix, 0 );
          g.putPixel( x, y, pix );
        }

      y += dy;
    }
  }

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonSave ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    OutputStream hdrFile = stream.getOutputStream();
    if ( hdrFile == null )
      throw new FileNotFoundException();

      // write HDR header:
    int width  = g.getWidth();
    int height = g.getHeight();

      // "#?RADIANCE ... 0x0a, 0x0a -Y <height> +X <width>"
    StringBuffer sb = new StringBuffer();
    sb.append( "#?RADIANCE\nFORMAT=32-bit_rle_rgbe\n\n-Y " );
    sb.append( height ).append( " +X " ).append( width ).append( '\n' );
    int x, y, i;
    for ( i = 0; i < sb.length(); i++ )
      hdrFile.write( sb.codePointAt( i ) );

      // write the bitmap:
    double[] pix = new double[ 3 ];
    assertScanline( width );

    for ( y = 0; y < height; y++ )        // encode one scanline
    {
        // read the scanline & convert it to RGBe format:
      for ( x = i = 0; x < width; x++, i += 4 )
      {
        g.getPixel( x, y, pix );
        Formula.RGBToRGBe( scanline, i, pix[0], pix[1], pix[2] );
      }

      writeScanline( width, hdrFile );
    }
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null )
      return null;
    return null;
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return HEADER.length;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type (<tt>0.0</tt> .. no way,
   *         <tt>0.5</tt> .. maybe, <tt>0.9</tt> .. almost sure, <tt>1.0</tt> ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int len = (header == null) ? 0 : header.length;
    if ( len > HEADER.length ) len = HEADER.length;
    if ( len < HEADER.length )            // not enough bytes in header[] => look at file-name
    {
      if ( fileName == null || fileName.length() < 5 ) return 0.0;
      int pos = fileName.lastIndexOf( '.' );
      if ( pos < 0 ) return 0.0;
      String suffix = fileName.substring( pos );
      if ( suffix.compareToIgnoreCase( "hdr" ) == 0 ) return 0.9;
      if ( suffix.compareToIgnoreCase( "pic" ) == 0 ) return 0.7;
      return 0.0;
    }

    for ( int i = 0; i < HEADER.length; i++ )
      if ( header[i] != HEADER[i] ) return 0.0;

    return 1.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.hdr"]) or <tt>null</tt>.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.hdr",
      "*.pic",
    };
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "HDR file format";

  /** Object description. */
  private final static String DESCRIPTION = "HDR filter (Radiance RGBe, RLE-encoded).";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"DataFileFormat" );
      t.newOutputPlug(    PL_RASTER, IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_STREAM, IFACE+"BitStream" );
        // Property:
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
