package cz.cuni.jagrlib.piece;

import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.color.*;
import com.sun.media.jai.codecimpl.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * HDR raster image stored in memory.
 * <br>Uses Java's {@link BufferedImage} class to hold raster data,
 * one channel is stored as {@code float} number.
 *
 * @version 0.26 $Rev: 464 $ $Date: 2007-05-24 15:41:38 +0200 (čt, 24 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/HDRImage.java">HDRImage.java</a>
 * @see     <a href="http://jsourcery.com/output/sun/jai-codec/1.1/RasterFactory.source.html">RasterFactory</a>
 */
public class HDRImage extends DefaultRasterGraphics
{
  //==========================================================================
  //  Data:

  /** Maximum gray/RGB value used in integer part of the {@link RasterGraphics} API. */
  public final static int MAX_VALUE = (1 << 30) - 1;

  /** Background color for image initialization. */
  public final static double[] INIT_COLOR = { 0.0, 0.0, 0.0, 1.0 };

  /** Default foreground color. */
  public final static float[] FOREGROUND = { 1.0f, 1.0f, 1.0f, 1.0f };

  /** The working {@link BufferedImage} object. */
  protected BufferedImage im;

  /**
   * Raw data array in scanline order.
   * Back-buffer of associated {@link DataBufferFloat} instance (accelerator).
   * data[0] .. R[0,0], data[1] .. G[0,0], data[2] .. B[0,0], data[3] .. A[0,0], ..
   */
  protected float[] data;

  //==========================================================================
  //  State variables:

  /** Actual drawing scalar color ({@link #MODE_GRAY MODE_GRAY}). */
  protected float sColor;

  /** Actual drawing color ({@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA}, {@link #MODE_BAND MODE_BAND}). */
  protected float[] color;

  /** Temporary array used for RGBA access. */
  protected double[] tmpColor;

  //==========================================================================
  //  Construction:

  /**
   * Creates an empty image.
   * {@link #init(int,int,int,int)} should be called to initialize image size/content.
   */
  public HDRImage ()
  {
  }

  /**
   * Initializes the new image.
   *
   * @param  _width Width of a new bitmap.
   * @param  _height Height of a new bitmap.
   * @param  _mode Pixel value type ({@code RasterGraphics.MODE_*} constant).
   */
  public HDRImage ( int _width, int _height, int _mode )
  {
    init( _width, _height, _mode, 0 );
  }

  /**
   * Initializes the new image from an existing image.
   *
   * @param  gr Existing raster image.
   * @param  scale Scale coefficient ({@code 1.0} for no scaling}).
   * @param  _mode Pixel value type ({@code RasterGraphics.MODE_*} constant).
   */
  public HDRImage ( RasterGraphics gr, double scale, int _mode )
  {
    int _width  = 512;
    int _height = 512;
    int _bands  = 0;
    if ( gr != null )                     // determine image size
    {
      int oWidth  = gr.getWidth();
      int oHeight = gr.getHeight();
      if ( _mode == MODE_BAND )
        _bands = gr.getBands();
      _width      = Formula.round( oWidth  * scale );
      _height     = Formula.round( oHeight * scale );
    }

    init( _width, _height, _mode, _bands );
    if ( gr == null ) return;

      // initialize image data:
    final double step      = 1.0 / scale;
    final double stepHalf  = 0.5 * step;
    final int    iStep     = (int)Math.ceil( step );
    final double mStepHalf = 0.5 * step / iStep;
    int intX, intY;
    double inX, inY, fracX, fracY;
    int x, y, ix, ix2, iy, iy0, iy2;
    int count;

    if ( mode == MODE_GRAY )
    {
      double acc;
      if ( iStep < 2 )
        for ( y = 0, inY = stepHalf - 0.5; y < _height; y++, inY += step )
        {
          intY  = (int)Math.floor( inY );
          fracY = inY - intY;
          for ( x = 0, inX = stepHalf - 0.5; x < _width; x++, inX += step )
          {
            intX  = (int)Math.floor( inX );
            fracX = inX - intX;
            acc   = (1.0 - fracY) * ( (1.0 - fracX) * gr.getGrayDouble( intX,     intY     ) +
                                             fracX  * gr.getGrayDouble( intX + 1, intY     ) ) +
                           fracY  * ( (1.0 - fracX) * gr.getGrayDouble( intX,     intY + 1 ) +
                                             fracX  * gr.getGrayDouble( intX + 1, intY + 1 ) );
            putPixel( x, y, acc );
          }
        }
      else
        for ( y = 0, inY = mStepHalf - 0.5; y < _height; y++, inY += step )
        {
          iy0 = Formula.round( inY );
          iy2 = Formula.round( inY + step );
          for ( x = 0, inX = mStepHalf - 0.5; x < _width; x++, inX += step )
          {
            acc   = 0.0;
            count = 0;
            ix2 = Formula.round( inX + step );

            for ( iy = iy0; iy < iy2; iy++ )
              for ( ix = Formula.round( inX ); ix < ix2; ix++ )
              {
                acc += gr.getGrayDouble( ix, iy );
                count++;
              }

            putPixel( x, y, acc / count );
          }
        }
    }

    else
    {
      double[] acc = new double[ bands ];
      // !!! TODO: resampling color (multi-band) images !!!
    }
  }

  //==========================================================================
  //  Image functions:

  /**
   * Downsample the image by the given integer factor.
   * Uses simple arithmetic averaging.
   *
   * @param  border Set border color (to {@link #INIT_COLOR background color})?
   */
  public void downsample ( int factor, boolean border )
  {
    if ( factor < 2 ) return;

    int x, y, b;
    int ox, oy;
    final int newWidth  = (width  + factor - 1) / factor;
    final int newHeight = (height + factor - 1) / factor;
    final double mul    = 1.0 / (factor * factor);
    final int lineInc   = bands * width;

    for ( oy = 0; oy < newHeight; oy++ )
      for ( ox = 0; ox < newWidth; ox++ )
      {
        Arrays.fill( tmpColor, 0.0 );
        int ptr0 = bands * factor * (ox + width * oy);
        final int maxX = Math.min( width  - factor * ox, factor );
        final int maxY = Math.min( height - factor * oy, factor );
        for ( y = 0; y++ < maxY; ptr0 += lineInc )
        {
          int ptr = ptr0;
          for ( x = 0; x++ < maxX; )
            for ( b = 0; b < bands; )
              tmpColor[b++] += data[ptr++];
        }

        ptr0 = bands * (ox + width * oy);
        for ( b = 0; b < bands; )
          data[ptr0++] = (float)( tmpColor[b++] * mul );
      }

    if ( border )
    {
      setRectangle( newWidth, 0,         width, newHeight, INIT_COLOR );
      setRectangle( 0,        newHeight, width, height,    INIT_COLOR );
    }
  }

  //==========================================================================
  //  Interface RasterGraphics:

  /**
   * {@code RasterImage} (re-)initialization.
   *
   * @param  _width Width of a new bitmap.
   * @param  _height Height of a new bitmap.
   * @param  _mode Pixel value type ({@code RasterGraphics.MODE_*} constant).
   * @param  _bands Number of bands (for {@link #MODE_BAND MODE_BAND}).
   * @see    #setColormapType
   */
  @Override
  public void init ( int _width, int _height, int _mode, int _bands )
  {
    super.init( _width, _height, _mode, _bands );

    sColor   = 1.0f;
    color    = FOREGROUND;
    tmpColor = new double[ bands ];

    int[] offsets = new int[ bands ];
    for ( int i = 0; i < bands; i++ )
      offsets[i] = i;

    ICC_Profile     ip = (_mode == MODE_GRAY) ? ICC_Profile.getInstance( ColorSpace.CS_GRAY ) :
                                                ICC_Profile.getInstance( ColorSpace.CS_sRGB );
    ColorSpace      cs = new ICC_ColorSpace( ip );
    ColorModel      cm = RasterFactory.createComponentColorModel( DataBuffer.TYPE_FLOAT, cs, _mode == MODE_RGBA,
                                                                  true, Transparency.TRANSLUCENT );

    com.sun.media.jai.codecimpl.util.DataBufferFloat
                    db = new com.sun.media.jai.codecimpl.util.DataBufferFloat( width * height * bands );
    data               = db.getData();

    WritableRaster  wr = RasterFactory.createInterleavedRaster( db, width, height, width * bands, bands,
                                                                offsets, new Point( 0, 0 ) );

    im                 = new BufferedImage( cm, wr, true, null );

    setRectangle( 0, 0, width, height, INIT_COLOR );
  }

  /**
   * {@link RasterImage} re-initialization (keeps current bitmap size, mode,
   * and colormap attributes, too).
   */
  @Override
  public void init ()
  {
    super.init();

    sColor = 1.0f;
    color  = FOREGROUND;

    setRectangle( 0, 0, width, height, INIT_COLOR );
  }

  /**
   * Maximal gray value.
   *
   * @return Maximal acceptable gray value.
   */
  @Override
  public int getMaxGray ()
  {
    return MAX_VALUE;
  }

  /**
   * Maximal RGB[A] value. Also used in multi-band mode.
   *
   * @return Maximal value for R,G,B,A and other bands.
   */
  @Override
  public int getMaxRGB ()
  {
    return MAX_VALUE;
  }

  /**
   * Copies raster graphics to a {@link BufferedImage} object.
   *
   * @return {@link BufferedImage} object.
   * @see    #setBufferedImage
   */
  @Override
  public BufferedImage getBufferedImage ()
  {
    return im;
  }

  /**
   * Sets the current drawing color (for {@link #MODE_GRAY MODE_GRAY}).
   *
   * @param  _color Current drawing color (0 to {@link #getMaxGray}).
   * @see    RasterGraphics
   */
  @Override
  public void setColor ( int _color )
  {
    sColor = Formula.clamp( _color, 0, MAX_VALUE );

    if ( mode  != MODE_GRAY &&
         color != null )
      color[0] =
      color[1] =
      color[2] = sColor;
  }

  /**
   * Sets the current drawing color (for {@link #MODE_GRAY MODE_GRAY}).
   *
   * @param  _color Current drawing color ({@code 0.0} to {@code 1.0}).
   * @see    RasterGraphics
   */
  @Override
  public void setColor ( double _color )
  {
    sColor = (float)_color;

    if ( mode  != MODE_GRAY &&
         color != null )
      color[0] =
      color[1] =
      color[2] = sColor;
  }

  /**
   * Sets the current drawing color (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  _color Current drawing color (channel values between
   *                {@code 0} and {@link #getMaxRGB}).
   * @see    RasterGraphics
   */
  @Override
  public void setColor ( int[] _color )
  {
    if ( _color == null ) return;

    int len = _color.length;
    if ( color        == null ||
         color.length != len )
      color = new float[ len ];

    for ( int i = 0; i < len; i++ )
      color[i] = _color[i];

    if ( mode == MODE_GRAY )
      sColor = color[0];
  }

  /**
   * Sets the current drawing color (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  _color Current drawing color (channel values between
   *                {@code 0.0} and {@code 1.0}).
   * @see    RasterGraphics
   */
  @Override
  public void setColor ( double[] _color )
  {
    if ( _color == null ) return;

    int len = _color.length;
    if ( color == null || color.length != len )
      color = new float[ len ];

    for ( int i = 0; i < len; i++ )
      color[i] = (float)_color[i];

    if ( mode == MODE_GRAY )
      sColor = color[0];
  }

  /**
   * Sets the given pixel's value.
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @see    #setColor(int)
   * @see    #setColor(double)
   * @see    #setColor(int[])
   * @see    #setColor(double[])
   * @see    RasterGraphics
   */
  @Override
  public void setPixel ( int x, int y )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height ) return;

    int ptr = bands * (x + width * y);
    int i;

    if ( mode == MODE_RGBA )
    {
        // color * alpha + oldColor * (1-alpha):
      data[ptr] = (float)( color[0] * alpha + data[ptr] * (1.0 - alpha) );
      ptr++;
      data[ptr] = (float)( color[1] * alpha + data[ptr] * (1.0 - alpha) );
      ptr++;
      data[ptr] = (float)( color[2] * alpha + data[ptr] * (1.0 - alpha) );
      ptr++;
      data[ptr] = (float)(            alpha + data[ptr] * (1.0 - alpha) );
    }
    else
      for ( i = 0; i < bands; )
        data[ptr++] = color[i++];
  }

  /**
   * Returns the pixel's value (for {@link #MODE_GRAY MODE_GRAY}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value ({@code 0} for off-canvas access).
   * @see    RasterGraphics
   */
  @Override
  public int getPixel ( int x, int y )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height ) return 0;

    int ptr = bands * (x + width * y);
    return Math.round( data[ptr] );
  }

  /**
   * Returns the pixel's value (for {@link #MODE_GRAY MODE_GRAY}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value ({@code 0.0} for off-canvas access).
   * @see    RasterGraphics
   */
  @Override
  public double getPixelDouble ( int x, int y )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height ) return 0.0;

    int ptr = bands * (x + width * y);
    return data[ptr];
  }

  /**
   * Returns the pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value.
   * @see    RasterGraphics
   */
  @Override
  public void getPixel ( int x, int y, int[] _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      System.arraycopy( BACKGROUND, 0, _color, 0, Math.min( BACKGROUND.length, _color.length ) );

    else
    {
      int ptr = bands * (x + width * y);
      for ( int i = 0; i < _color.length; i++ )
        _color[i] = Math.round( data[ptr++] );
    }
  }

  /**
   * Returns the pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value.
   * @see    RasterGraphics
   */
  @Override
  public void getPixel ( int x, int y, double[] _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      System.arraycopy( BACKGROUND_D, 0, _color, 0, Math.min( BACKGROUND_D.length, _color.length ) );

    else
    {
      int ptr = bands * (x + width * y);
      for ( int i = 0; i < _color.length; i++ )
        _color[i] = data[ptr++];
    }
  }

  /**
   * Returns the pixel's RGB[A] values (for all graphic modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value (if the array's length is at least 4, alpha will be returned).
   */
  @Override
  public void getRGB ( int x, int y, int[] _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      System.arraycopy( BACKGROUND, 0, _color, 0, Math.min( BACKGROUND.length, _color.length ) );

    else
    {
      int ptr = bands * (x + width * y);

      switch ( mode )
      {
        case MODE_GRAY:
          _color[0] =
          _color[1] =
          _color[2] = Math.round( data[ptr] );
          break;

        case MODE_RGB:
        case MODE_BAND:
          _color[0] = Math.round( data[ptr++] );
          _color[1] = Math.round( data[ptr++] );
          _color[2] = Math.round( data[ptr  ] );
          break;

        case MODE_RGBA:
          _color[0] = Math.round( data[ptr++] );
          _color[1] = Math.round( data[ptr++] );
          _color[2] = Math.round( data[ptr++] );
          _color[3] = Math.round( data[ptr  ] );
          return;
      }

      if ( _color.length > 3 )
        _color[3] = MAX_VALUE;
    }
  }

  /**
   * Returns the pixel's RGB[A] values (for all graphic modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value (if the array's length is at least 4, alpha will be returned).
   */
  @Override
  public void getRGB ( int x, int y, double[] _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      System.arraycopy( BACKGROUND_D, 0, _color, 0, Math.min( BACKGROUND_D.length, _color.length ) );

    else
    {
      int ptr = bands * (x + width * y);

      switch ( mode )
      {
        case MODE_GRAY:
          _color[0] =
          _color[1] =
          _color[2] = data[ptr];
          break;

        case MODE_RGB:
        case MODE_BAND:
          _color[0] = data[ptr++];
          _color[1] = data[ptr++];
          _color[2] = data[ptr  ];
          break;

        case MODE_RGBA:
          _color[0] = data[ptr++];
          _color[1] = data[ptr++];
          _color[2] = data[ptr++];
          _color[3] = data[ptr  ];
          return;
      }

      if ( _color.length > 3 )
        _color[3] = 1.0;
    }
  }

  /**
   * Returns the pixel's gray value (for all modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value ({@code 0} for off-canvas access).
   */
  @Override
  public int getGray ( int x, int y )
  {
    if ( x >= 0 && x < width &&
         y >= 0 && y < height )
    {
      final int ptr = bands * (x + width * y);

      switch ( mode )
      {
        case MODE_GRAY:
          return Math.round( data[ptr] );

        case MODE_RGB:
        case MODE_RGBA:
        case MODE_BAND:
          return Formula.round( Formula.rgbToGray( (double)data[ptr], (double)data[ptr+1], (double)data[ptr+2] ) );
      }
    }

    return 0;
  }

  /**
   * Returns the pixel's gray value (for all modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value ({@code 0.0} for off-canvas access).
   */
  @Override
  public double getGrayDouble ( int x, int y )
  {
    if ( x >= 0 && x < width &&
         y >= 0 && y < height )
    {
      final int ptr = bands * (x + width * y);

      switch ( mode )
      {
        case MODE_GRAY:
          return data[ptr];

        case MODE_RGB:
        case MODE_RGBA:
        case MODE_BAND:
          return Formula.rgbToGray( data[ptr], data[ptr+1], data[ptr+2] );
      }
    }

    return 0.0;
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_GRAY MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0} and {@link #getMaxGray}).
   */
  @Override
  public void putPixel ( int x, int y, int _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      return;

    int ptr   = bands * (x + width * y);
    data[ptr] = _color;
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_GRAY MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0.0} and {@code 1.0}).
   */
  @Override
  public void putPixel ( int x, int y, double _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      return;

    int ptr   = bands * (x + width * y);
    data[ptr] = (float)_color;
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0} and {@link #getMaxRGB}).
   */
  @Override
  public void putPixel ( int x, int y, int[] _color )
  {
    if ( _color == null || _color.length < bands ||
         x < 0 || x >= width ||
         y < 0 || y >= height )
      return;

    int ptr = bands * (x + width * y);
    for ( int i = 0; i < bands; )
      data[ptr++] = _color[i++];
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0.0} and {@code 1.0}).
   */
  @Override
  public void putPixel ( int x, int y, double[] _color )
  {
    if ( _color == null || _color.length < bands ||
         x < 0 || x >= width ||
         y < 0 || y >= height ) return;

    int ptr = bands * (x + width * y);
    for ( int i = 0; i < bands; )
      data[ptr++] = (float)_color[i++];
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y].
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @see    #setColor(int)
   * @see    #setColor(double)
   * @see    #setColor(int[])
   * @see    #setColor(double[])
   * @see    RasterGraphics
   */
  @Override
  public void setHLine ( int x1, int x2, int y )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 ) x1 = 0;
    if ( x2 > width ) x2 = width;
    int ptr = bands * (x1 + width * y);

    switch ( mode )
    {
      case MODE_GRAY:
        while ( x1++ < x2 )
          data[ptr++] = sColor;
        break;

      case MODE_RGBA:
        while ( x1++ < x2 )
        {
            // color * alpha + tmpColor * (1-alpha):
          data[ptr] = (float)( color[0] * alpha + data[ptr] * (1.0 - alpha) );
          ptr++;
          data[ptr] = (float)( color[1] * alpha + data[ptr] * (1.0 - alpha) );
          ptr++;
          data[ptr] = (float)( color[2] * alpha + data[ptr] * (1.0 - alpha) );
          ptr++;
          data[ptr] = (float)(            alpha + data[ptr] * (1.0 - alpha) );
          ptr++;
        }
        break;

      case MODE_RGB:
      case MODE_BAND:
      {
        int i;
        while ( x1++ < x2 )
          for ( i = 0; i < bands; )
            data[ptr++] = color[i++];
        break;
      }
    }
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "HDRImage";

  /** Object description. */
  private final static String DESCRIPTION = "HDR raster image storage (using BufferedImage).";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( WIDTH, TYPE_INTEGER, "Bitmap width in pixels", true );
        t.propDefault( 512 );
      t.propEnd();
      t.propBegin( HEIGHT, TYPE_INTEGER, "Bitmap height in pixels", true );
        t.propDefault( 512 );
      t.propEnd();
      t.propBegin( COLORMODE, TYPE_INTEGER, "Pixel color mode", true );
        t.propDefault( RasterGraphics.MODE_RGB );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "RGB",  RasterGraphics.MODE_RGB,  "RGB color mode" );
        t.propEnum( "RGBA", RasterGraphics.MODE_RGBA, "RGBA color mode (transparency)" );
        t.propEnum( "Gray", RasterGraphics.MODE_GRAY, "Gray mode" );
      t.propEnd();
      t.propBegin( BANDS, TYPE_INTEGER, "Number of color bands", true );
        t.propDefault( 0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
