package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * HDR to LDR image conversion.
 *
 * @version 0.25 $Rev: 280 $ $Date: 2006-06-01 19:35:42 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/HDRMap.java">HDRMap.java</a>
 */
public class HDRMap extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Exposure coefficient. Multiplies original image data. */
  protected double exposure = 1.0;

  /** Gamma-correction for LDR frame-buffer efficiency. */
  protected double gamma = 2.2;

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transformation.
   *
   * @param  type The action type (whatever it means..).
   * @return <tt>true</tt> if the action was successful.
   */
  public boolean fire ( int type )
  {
    final RasterGraphics input  = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    final RasterGraphics output = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );

    if ( input  == null ||
         output == null ) return false;

      // output image:
    int width  = input.getWidth();
    int height = input.getHeight();

    output.init( width, height, RasterImage.MODE_RGB, 0 );

    double[] pixel = new double[ 3 ];
    double g = 1.0 / gamma;
    int i, j;
    for ( j = 0; j < height; j++ )
      for ( i = 0; i < width; i++ )
      {
        input.getRGB( i, j, pixel );
        pixel[0] = Formula.clamp( Math.pow( pixel[0] * exposure, g ), 0.0, 1.0 );
        pixel[1] = Formula.clamp( Math.pow( pixel[1] * exposure, g ), 0.0, 1.0 );
        pixel[2] = Formula.clamp( Math.pow( pixel[2] * exposure, g ), 0.0, 1.0 );
        output.putPixel( i, j, pixel );
      }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String EXPOSURE = "Exposure";
  public final static String GAMMA    = "Gamma";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // EXPOSURE:
    if ( key.compareTo(EXPOSURE) == 0 )
    {
      exposure = doubleProperty( value, exposure );
      return;
    }
      // GAMMA:
    if ( key.compareTo(GAMMA) == 0 )
    {
      gamma = doubleProperty( value, gamma, 0.2, 5.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(EXPOSURE) == 0 )
      return exposure;
    if ( key.compareTo(GAMMA) == 0 )
      return gamma;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "HDR Map";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Conversion of HDR raster image into LDR.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"Trigger" );
      t.newOutputPlug( "image1", IFACE+"RasterGraphics" );
      t.newOutputPlug( "image2", IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( EXPOSURE, TYPE_DOUBLE, "Exposure coefficient", true );
        t.propDefault( "1.0" );
      t.propEnd();
      t.propBegin( GAMMA, TYPE_DOUBLE, "Gamma correction", true );
        t.propDefault( "2.2" );
        t.propBounds( "0.2", "5.0" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
