package cz.cuni.jagrlib.piece;

import java.io.IOException;
import java.util.*;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Adaptive Huffman codec.
 *
 * @author  Ondrej Starek <stareko@seznam.cz>, PE
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/HuffmanCodec.java">HuffmanCodec.java</a>
 */
public class HuffmanCodec extends Piece implements EntropyCodec
{
  /** Node of the Huffman tree. */
  protected static class HuffNode
  {
    /** Node word. */
    public int character;

    /** Frequency. */
    public long freq;

    /** Tree (vertical) links. */
    public HuffNode left;
    public HuffNode right;
    public HuffNode parent;

    /** Sibling (horizontal) links. */
    public HuffNode next;
    public HuffNode prev;

    public HuffNode ( int character )
    {
      this.character = character;
      freq           = 1L;
      // all other data is implicitly initialized to 0 (null)
    }
  }

  /** Huffman tree (used as {@link EntropyCodec}-context). */
  protected static class HuffTree
  {
    /** Total number of code-words (including {@link #NYT}-escape). */
    private int wordCount;

    /** Not-Yet-Transmitted escape code. */
    private int NYT;

    /** Special value used in inner nodes. */
    private int innode;

    /** Code-word size in bits. */
    private int codeSize;

    /** Array of tree leaves. */
    private HuffNode[] leaves;

    /** Root of the Huffman tree. */
    private HuffNode root;

    public HuffTree ( int maxSymbol )
    {
      init( maxSymbol );
    }

    public void init ( int maxSymbol )
    {
      if ( wordCount != maxSymbol + 2 )
      {
        wordCount = maxSymbol + 2;        // including NYT
        NYT       = maxSymbol + 1;
        innode    = maxSymbol + 2;

          // compute code-word size:
        codeSize  = 1;
        int size  = 2;
        while ( size < NYT )
        {
          size += size;
          codeSize++;
        }

          // array of leaves:
        leaves    = new HuffNode[ wordCount ];
      }
      else
        Arrays.fill( leaves, null );

        // root item:
      root        = new HuffNode( NYT );
      leaves[NYT] = root;
    }

    public final int getMaxSymbol ()
    {
      return( wordCount - 2 );
    }

    private void printTreeInner ( HuffNode node )
    {
      if ( node == null ) return;

      printTreeInner( node.left );
      printTreeInner( node.right );
    }

    public void printTree ()
    {
      printTreeInner( root );
    }

    private void splitNYT ( int character )
    {
      HuffNode n1, n2;

      n1 = new HuffNode( NYT );
      n2 = new HuffNode( character );
      leaves[NYT].character = innode;
      leaves[NYT].freq = 1;

        // siblings:
      n1.next = n2;
      n2.next = leaves[NYT];
      n2.prev = n1;
      leaves[NYT].prev = n2;

        // tree links:
      n1.parent = leaves[NYT];
      n2.parent = leaves[NYT];
      leaves[NYT].left  = n1;
      leaves[NYT].right = n2;

        // update leaf node:
      leaves[NYT] = n1;
      leaves[character] = n2;
    }

    private HuffNode swapLast ( HuffNode node )
    {
      HuffNode n;

        // find the last node with the same frequency..
      n = node;
      while ( n.next != null &&
              n.next != root &&
              n.next.freq == node.freq )
        n = n.next;

        // it's me and my parent => dont swap
      if ( n == node ) return node;

        // swap data:
      int d          = node.character;
      node.character = n.character;
      n.character    = d;
      long f         = node.freq;
      node.freq      = n.freq;
      n.freq         = f;

        // swap children:
      HuffNode h;
      h          = node.left;
      node.left  = n.left;
      n.left     = h;
      h          = node.right;
      node.right = n.right;
      n.right    = h;
      if ( node.character == innode )
      {
        node.left.parent  = node;
        node.right.parent = node;
      }
      if ( n.character == innode )
      {
        n.left.parent  = n;
        n.right.parent = n;
      }

        // update list of leaves:
      if ( node.character != innode )
        leaves[node.character] = node;
      if ( n.character != innode )
        leaves[n.character] = n;

      return n;
    }

    private void rearrangeTree ( HuffNode node )
    {
      while ( node != root )
      {
        node = swapLast( node );
        node.freq++;
        node = node.parent;
      }
      node.freq++;
    }

    private void writeCode ( HuffNode node, BitStream stream )
    throws IOException
    {
      if ( node == root ) return;

      writeCode( node.parent, stream );
      stream.write( (node.parent.left == node) ? 0 : 1, 1 );
    }

    public void writeChar ( BitStream stream, int character )
    throws IOException
    {
      if ( character < 0 ||
           character >= NYT )
      {
        LogFile.error( "HuffmanCodec - symbol (" + character + ") out of range (" + (NYT-1) + ')' );
        character = 0;
      }

        // write the code
      if ( leaves[character] == null )
      {
          // write NYT and character:
        writeCode( leaves[NYT], stream );
        stream.write( character, codeSize );

          // insert char into the tree:
        splitNYT( character );
        rearrangeTree( leaves[character].parent );
      }
      else
      {
        writeCode( leaves[character], stream );
        rearrangeTree( leaves[character] );
      }
    }

    public int getChar ( BitStream stream )
    throws IOException
    {
      HuffNode n = root;
      long bit;
      int  ch;

        // find the leaf node:
      while ( n.character == innode )
      {
        bit = stream.read( 1 );
        n = ((bit & 1) != 0) ? n.right : n.left;
      }

      if ( n.character == NYT )
      {
          // if the leaf is NYT, read the next character:
        ch = (int)stream.read( codeSize );
        splitNYT( ch );
        n  = leaves[ch].parent;
      }
      else
        ch = n.character;

        // update the tree:
      rearrangeTree( n );

      return ch;
    }
  }

  //==========================================================================
  //  Data:

  /** Associated I/O stream. {@code null} if the codec is not opened. */
  protected BitStream stream;

  /** Map of contexts. */
  protected IntMap< HuffTree > contexts;

  /** Id of the current context. */
  protected int currentContext = Integer.MIN_VALUE;

  /** Current context (accelerator). */
  protected HuffTree context;

  /** Max-symbol used only for creating of new contexts! */
  protected int maxSymbol = 255;

  /** Opened for output (encoding)? */
  protected boolean output;

  /** Actual position in original (logical) stream. */
  protected long position = 0;

  //==========================================================================
  //  Support:

  /** Discards all contexts. */
  protected void clearContexts ()
  {
    contexts       = new IntMap< HuffTree >( 1 );
    currentContext = Integer.MIN_VALUE;
    context        = null;
  }

  /** [Re-]initializes context set, creates the given solo context. */
  protected void initContexts ( int ctx )
  {
    clearContexts();
    setContextInner( ctx );
  }

  private int setContextInner ( int ctxId )
  {
    if ( ctxId == currentContext ) return currentContext;

    int oldId = currentContext;

    HuffTree hf = contexts.get( ctxId );
    if ( hf == null )
    {
      hf = new HuffTree( maxSymbol );
      contexts.put( ctxId, hf );
    }
    currentContext = ctxId;
    context        = hf;

    return oldId;
  }

  //==========================================================================
  //  Interface EntropyCodec:

  public HuffmanCodec ()
  {
    clearContexts();
  }

  /**
   * Are they any input symbols (bits) available?
   *
   * @return Any symbols (bits) are ready to read.
   * @throws IOException
   * @see    #position
   * @see    #compressed
   */
  public boolean available ()
  throws IOException
  {
    return ( !output &&
             stream != null &&
             stream.available() > 0 );
  }

  /**
   * Codec close.
   *
   * @throws IOException
   * @see    #open
   * @see    #flush
   */
  public void close ()
  throws IOException
  {
    if ( stream == null )
      return;

    stream.close();
    stream   = null;
    position = 0;
    clearContexts();
  }

  /**
   * Actual codec compressed position in bits (total amount in all contexts).
   *
   * @return Actual codec compressed position in bits.
   * @throws IOException
   * @see    #position
   */
  public long compressed ()
  throws IOException
  {
    // !!! TODO: compute exact binary position
    return stream.position();
  }

  /**
   * Codec flush.
   *
   * @throws IOException
   * @see    #open
   * @see    #close
   */
  public void flush ()
  throws IOException
  {
    if ( output )
      stream.flush();
  }

  /**
   * Decodes one symbol (using the current context).
   *
   * @return Read symbol ({@code -1} if no symbol is available).
   * @throws IOException
   * @see    #getBits
   * @see    #put
   * @see    #available
   */
  public int get ()
  throws IOException
  {
    if ( stream == null ||
         output )
      throw new IOException( "HuffmanCodec is not opened for reading!" );

    position++;
    return context.getChar( stream );
  }

  /**
   * Decodes one subinterval from the given {@link WheelOfFortune}.
   * Uses the actual context!
   *
   * @param  wheel Set of intervals (wheel of fortune).
   * @return Index of the relevant interval.
   */
  public int get ( WheelOfFortune wheel )
  throws IOException
  {
    return get();
  }

  /**
   * Reads the given amount of "raw" bits (by-passes the entropy decoder).
   *
   * @param  length Number of bits to read.
   * @return Read bits (or {@code -1} if no bits are available).
   * @throws IOException
   * @see    #get
   * @see    #putBits
   * @see    #available
   */
  public long getBits ( int length )
  throws IOException
  {
    if ( stream == null ||
         output )
      throw new IOException( "HuffmanCodec is not opened for reading!" );

    return stream.read( length );
  }

  /**
   * Codec initialization.
   *
   * @param  output Open for output?
   * @param  dCtx Initial context.
   * @throws IOException
   * @see    BitStream
   * @see    #setContext
   */
  public void open ( boolean output, int dCtx )
  throws IOException
  {
    stream      = (BitStream)getInterface( PL_OUTPUT, IFACE+"BitStream" );
    this.output = output;
    position    = 0;
    stream.open( output, (String)null, null );
    initContexts( dCtx );
  }

  /**
   * Retrieves maximal symbol value for the actual context.
   *
   * @return Maximal symbol value.
   */
  public int getMaxSymbol ()
  {
    if ( stream == null ) return -1;

    if ( context != null )
      return context.getMaxSymbol();

    return maxSymbol;
  }

  /**
   * Sets maximal symbol value for the actual context.
   *
   * @param maxSym New maximal symbol value.
   */
  public void setMaxSymbol ( int maxSym )
  {
    maxSymbol = maxSym;
    if ( context != null )
      context.init( maxSym );
  }

  /**
   * Actual codec external position in symbols (total amount in all contexts).
   *
   * @return Actual codec external position in symbols.
   * @throws IOException
   * @see    #compressed
   */
  public long position ()
  throws IOException
  {
    return position;
  }

  /**
   * Encodes the given symbol (in the current context).
   *
   * @param  symbol Symbol to encode.
   * @throws IOException
   * @see    #putBits
   * @see    #get
   */
  public void put ( int symbol )
  throws IOException
  {
    if ( stream == null ||
         !output )
      throw new IOException( "HuffmanCodec is not opened for writting!" );

    position++;
    context.writeChar( stream, symbol );
  }

  /**
   * Encodes one subinterval from the given {@link WheelOfFortune}.
   * Uses the actual context!
   *
   * @param  wheel Set of intervals (wheel of fortune).
   * @param  s Index of an interval to encode.
   */
  public void put ( WheelOfFortune wheel, int s )
  throws IOException
  {
    put( s );
  }

  /**
   * Writes the given "raw" bits (by-passes the entropy encoder).
   *
   * @param  bits Bits to be written ("MSB first" order).
   * @param  length Number of bits to write.
   * @throws IOException
   * @see    #put
   * @see    #getBits
   */
  public void putBits ( long bits, int length )
  throws IOException
  {
    stream.write( bits, length );
  }

  /**
   * Sets the given context as current.
   *
   * @param  ctx Id of the new context.
   * @return The old context's id.
   * @throws IOException
   */
  public int setContext ( int ctx )
  throws IOException
  {
    return setContextInner( ctx );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "HuffmanCodec";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "EntropyCodecToBitStream";

  /** Object description. */
  private final static String DESCRIPTION = "Adaptive Huffman codec implementation.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_COMPRESSION + ".AHF";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"EntropyCodec" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"BitStream" );
        // Property:
     }
   return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
