package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Raster image comparison.
 *
 * @version 0.26 $Rev: 477 $ $Date: 2007-06-03 03:19:26 +0200 (ne, 03 VI 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/ImageCompare.java">ImageCompare.java</a>
 */
public class ImageCompare extends DefaultRealFunction
{
  //==========================================================================
  //  Data:

  /** Implemented filter precincts. */
  protected static enum FilterPrecinct
  {
    /** Exact comparison - works w/o precinct. */
    FILTER_NONE,
    /** Cylindrical precinct with the given radius. */
    FILTER_CYLINDER,
    /** Cylindrical precinct with the given radius. */
    FILTER_CONE,
  }

  /** Implemented metric types. */
  protected static enum MetricType
  {
    /** Exact comparison - true/false (1/0). */
    METRIC_BOOLEAN,
    /** Root mean square error. */
    METRIC_RMSE,
    /** Mean absolute error. */
    METRIC_MAE,
    /** Total absolute error. */
    METRIC_AE,
  }

  /** Actual filter precinct. */
  protected FilterPrecinct precinct = FilterPrecinct.FILTER_NONE;

  /** Filter support radius in pixels. */
  protected double radius = 0.0;

  /** Actual metric type. */
  protected MetricType metric = MetricType.METRIC_BOOLEAN;

  /** Psychovisual color-component factors are used? */
  protected boolean psychovisual = false;

  //==========================================================================
  //  Support:

  /** Integer radius of the filter support. */
  protected int iRadius;

  /** Filter matrix (odd size .. 2*iRadius+1). */
  protected int[][] filter;

  /** Total sum of filter values. */
  protected int filterSum;

  /** Accelerator (filterSum / 2). */
  protected int filterSumHalf;

  /** Support pixel array (for {@link #filterRGB} routines). */
  protected int[] pixel = new int[ 3 ];

  /** Support pixel array (for {@link #filterRGB} routines). */
  protected double[] pixeld = new double[ 3 ];

  /** Initialize (prepare) filter operation. */
  protected void initFilter ()
  {
    iRadius = (int)radius;
    if ( iRadius <= 0 ||
         precinct == FilterPrecinct.FILTER_NONE ) return;

    int i, j, r;
    int size = 2 * iRadius + 1;
    if ( filter == null ||
         filter.length != size )
      filter = new int[ size ][ size ];

    filterSum = 0;
    int iRadius2 = iRadius * iRadius;
    double dRadius;

    switch ( precinct )
    {
      case FILTER_CYLINDER:
        for ( i = 0; i < size; i++ )
          for ( j = 0; j < size; j++ )
          {
            r = (i - iRadius) * (i - iRadius) + (j - iRadius) * (j - iRadius);
            filterSum += (filter[i][j] = (r <= iRadius2) ? 1 : 0);
          }
        break;

      case FILTER_CONE:
        for ( i = 0; i < size; i++ )
          for ( j = 0; j < size; j++ )
          {
            r = (i - iRadius) * (i - iRadius) + (j - iRadius) * (j - iRadius);
            dRadius = Math.sqrt( r ) / radius;
            if ( dRadius >= 1.0 )
              filter[i][j] = 0;
            else
              filterSum += (filter[i][j] = (int)((1.0 - dRadius) * 16384.0 + 0.5));
          }
        break;
    }
    filterSumHalf = filterSum >> 1;
  }

  /** Filters the given image pixel using prepared convolution filter.. */
  protected void filterRGB ( RasterGraphics im, int i, int j, int[] pix )
  {
    if ( iRadius <= 0 ) return;
    int x, y, c;
    int r = 0;
    int g = 0;
    int b = 0;

    for ( y = -iRadius; y <= iRadius; y++ )
      for ( x = -iRadius; x <= iRadius; x++ )
        if ( (c = filter[iRadius+y][iRadius+x]) > 0 )
        {
          im.getRGB( i+x, j+y, pixel );
          r += c * pixel[0];
          g += c * pixel[1];
          b += c * pixel[2];
        }

    pix[0] = (r + filterSumHalf) / filterSum;
    pix[1] = (g + filterSumHalf) / filterSum;
    pix[2] = (b + filterSumHalf) / filterSum;
  }

  /** Filters the given image pixel using prepared convolution filter.. */
  protected void filterRGB ( RasterGraphics im, int i, int j, double[] pix )
  {
    if ( iRadius <= 0 ) return;
    int x, y, c;
    double r = 0.0;
    double g = 0.0;
    double b = 0.0;

    for ( y = -iRadius; y <= iRadius; y++ )
      for ( x = -iRadius; x <= iRadius; x++ )
        if ( (c = filter[iRadius+y][iRadius+x]) > 0 )
        {
          im.getRGB( i+x, j+y, pixeld );
          r += c * pixeld[0];
          g += c * pixeld[1];
          b += c * pixeld[2];
        }

    pix[0] = r / filterSum;
    pix[1] = g / filterSum;
    pix[2] = b / filterSum;
  }

  //==========================================================================
  //  Interface RealFunction:

  /** Compute the mapping: {@code int -> double}. */
  @Override
  public double f ( int x )
  {
    RasterGraphics im1 = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    RasterGraphics im2 = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );
    if ( im1 == null || im2 == null )
      return Double.POSITIVE_INFINITY;

      // image1:
    int width  = im1.getWidth();
    int height = im1.getHeight();
    if ( width  != im2.getWidth() ||
         height != im2.getHeight() )
      return Double.POSITIVE_INFINITY;

    int i, j;
    int[] pix1 = new int[ 3 ];
    int[] pix2 = new int[ 3 ];
    int  sum   = 0;
    long suml  = 0L;
    long err;

      // filtering:
    boolean prec = (precinct != FilterPrecinct.FILTER_NONE);
    if ( prec ) initFilter();

    switch ( metric )
    {
      case METRIC_BOOLEAN:
        for ( j = 0; j < height; j++ )
          for ( i = 0; i < width; i++ )
          {
            if ( prec )
            {
              filterRGB( im1, i, j, pix1 );
              filterRGB( im2, i, j, pix2 );
            }
            else
            {
              im1.getRGB( i, j, pix1 );
              im2.getRGB( i, j, pix2 );
            }
            if ( pix1[0] != pix2[0] ||
                 pix1[1] != pix2[1] ||
                 pix1[2] != pix2[2] )
              return Double.POSITIVE_INFINITY;
          }
        return 0.0;

      case METRIC_RMSE:
        if ( psychovisual )
        {
          for ( j = 0; j < height; j++ )
            for ( i = 0; i < width; i++ )
            {
              if ( prec )
              {
                filterRGB( im1, i, j, pix1 );
                filterRGB( im2, i, j, pix2 );
              }
              else
              {
                im1.getRGB( i, j, pix1 );
                im2.getRGB( i, j, pix2 );
              }
              err   = pix1[0] - pix2[0];
              suml += Formula.RED_WEIGHT   * err * err;
              err   = pix1[1] - pix2[1];
              suml += Formula.GREEN_WEIGHT * err * err;
              err   = pix1[2] - pix2[2];
              suml += Formula.BLUE_WEIGHT  * err * err;
            }
          return Math.sqrt( suml / ((double)Formula.WEIGHT_TOTAL * width * height) );
        }
        for ( j = 0; j < height; j++ )
          for ( i = 0; i < width; i++ )
          {
            if ( prec )
            {
              filterRGB( im1, i, j, pix1 );
              filterRGB( im2, i, j, pix2 );
            }
            else
            {
              im1.getRGB( i, j, pix1 );
              im2.getRGB( i, j, pix2 );
            }
            err   = pix1[0] - pix2[0];
            suml += err * err;
            err   = pix1[1] - pix2[1];
            suml += err * err;
            err   = pix1[2] - pix2[2];
            suml += err * err;
          }
        return Math.sqrt( suml / (3.0 * width * height) );

      default:
      //case METRIC_AE:
      //case METRIC_MAE:
        if ( psychovisual )
        {
          for ( j = 0; j < height; j++ )
            for ( i = 0; i < width; i++ )
            {
              if ( prec )
              {
                filterRGB( im1, i, j, pix1 );
                filterRGB( im2, i, j, pix2 );
              }
              else
              {
                im1.getRGB( i, j, pix1 );
                im2.getRGB( i, j, pix2 );
              }
              suml += Formula.RED_WEIGHT   * Math.abs( pix1[0] - pix2[0] );
              suml += Formula.GREEN_WEIGHT * Math.abs( pix1[1] - pix2[1] );
              suml += Formula.BLUE_WEIGHT  * Math.abs( pix1[2] - pix2[2] );
            }
          if ( metric == MetricType.METRIC_AE )
          {
            suml = (suml * 3) >>> Formula.WEIGHT_SHIFT;
            return (double)suml;
          }
          return( suml / ((double)Formula.WEIGHT_TOTAL * width * height) );
        }
        for ( j = 0; j < height; j++ )
          for ( i = 0; i < width; i++ )
          {
            if ( prec )
            {
              filterRGB( im1, i, j, pix1 );
              filterRGB( im2, i, j, pix2 );
            }
            else
            {
              im1.getRGB( i, j, pix1 );
              im2.getRGB( i, j, pix2 );
            }
            sum += Math.abs( pix1[0] - pix2[0] );
            sum += Math.abs( pix1[1] - pix2[1] );
            sum += Math.abs( pix1[2] - pix2[2] );
          }
        if ( metric == MetricType.METRIC_AE )
          return sum;
        return( sum / (width * height * 3.0) );
    }
  }

  /** Compute the mapping: {@code int -> double}. */
  @Override
  public double f ( double x )
  {
    RasterGraphics im1 = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    RasterGraphics im2 = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );
    if ( im1 == null || im2 == null )
      return Double.POSITIVE_INFINITY;

      // image1:
    int width  = im1.getWidth();
    int height = im1.getHeight();
    if ( width  != im2.getWidth() ||
         height != im2.getHeight() )
      return Double.POSITIVE_INFINITY;

    int i, j;
    double[] pix1 = new double[ 3 ];
    double[] pix2 = new double[ 3 ];
    double   sum = 0.0;

      // filtering:
    boolean prec = (precinct != FilterPrecinct.FILTER_NONE);
    if ( prec ) initFilter();

    switch ( metric )
    {
      case METRIC_BOOLEAN:
        return f( 0.0 );

      case METRIC_RMSE:
      {
        double err;
        if ( psychovisual )
        {
          for ( j = 0; j < height; j++ )
            for ( i = 0; i < width; i++ )
            {
              if ( prec )
              {
                filterRGB( im1, i, j, pix1 );
                filterRGB( im2, i, j, pix2 );
              }
              else
              {
                im1.getRGB( i, j, pix1 );
                im2.getRGB( i, j, pix2 );
              }
              err  = pix1[0] - pix2[0];
              sum += Formula.RED_WEIGHT   * err * err;
              err  = pix1[1] - pix2[1];
              sum += Formula.GREEN_WEIGHT * err * err;
              err  = pix1[2] - pix2[2];
              sum += Formula.BLUE_WEIGHT  * err * err;
            }
          return Math.sqrt( sum / ((double)Formula.WEIGHT_TOTAL * width * height) );
        }
        for ( j = 0; j < height; j++ )
          for ( i = 0; i < width; i++ )
          {
            if ( prec )
            {
              filterRGB( im1, i, j, pix1 );
              filterRGB( im2, i, j, pix2 );
            }
            else
            {
              im1.getRGB( i, j, pix1 );
              im2.getRGB( i, j, pix2 );
            }
            err  = pix1[0] - pix2[0];
            sum += err * err;
            err  = pix1[1] - pix2[1];
            sum += err * err;
            err  = pix1[2] - pix2[2];
            sum += err * err;
          }
        return Math.sqrt( sum / (3.0 * width * height) );
      }

      default:
      //case METRIC_AE:
      //case METRIC_MAE:
        if ( psychovisual )
        {
          for ( j = 0; j < height; j++ )
            for ( i = 0; i < width; i++ )
            {
              if ( prec )
              {
                filterRGB( im1, i, j, pix1 );
                filterRGB( im2, i, j, pix2 );
              }
              else
              {
                im1.getRGB( i, j, pix1 );
                im2.getRGB( i, j, pix2 );
              }
              sum += Formula.RED_WEIGHT   * Math.abs( pix1[0] - pix2[0] );
              sum += Formula.GREEN_WEIGHT * Math.abs( pix1[1] - pix2[1] );
              sum += Formula.BLUE_WEIGHT  * Math.abs( pix1[2] - pix2[2] );
            }
          if ( metric == MetricType.METRIC_AE )
            return (sum * 3.0) / (double)Formula.WEIGHT_TOTAL;
          return( sum / ((double)Formula.WEIGHT_TOTAL * width * height) );
        }
        for ( j = 0; j < height; j++ )
          for ( i = 0; i < width; i++ )
          {
            if ( prec )
            {
              filterRGB( im1, i, j, pix1 );
              filterRGB( im2, i, j, pix2 );
            }
            else
            {
              im1.getRGB( i, j, pix1 );
              im2.getRGB( i, j, pix2 );
            }
            sum += Math.abs( pix1[0] - pix2[0] );
            sum += Math.abs( pix1[1] - pix2[1] );
            sum += Math.abs( pix1[2] - pix2[2] );
          }
        if ( metric == MetricType.METRIC_AE )
          return sum;
        return( sum / (3.0 * width * height) );
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String FILTER_PRECINCT = "FilterPrecinct";
  public final static String FILTER_RADIUS   = "FilterRadius";
  public final static String METRIC_TYPE     = "MetricType";
  public final static String PSYCHOVISUAL    = "Psychovisual";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // FILTER_PRECINCT:
    if ( key.compareTo(FILTER_PRECINCT) == 0 )
    {
      precinct = enumProperty( value, precinct );
      return;
    }
      // FILTER_RADIUS:
    if ( key.compareTo(FILTER_RADIUS) == 0 )
    {
      radius = doubleProperty( value, radius, 0.0, Double.MAX_VALUE );
      return;
    }
      // METRIC_TYPE:
    if ( key.compareTo(METRIC_TYPE) == 0 )
    {
      metric = enumProperty( value, metric );
      return;
    }
      // PSYCHOVISUAL:
    if ( key.compareTo(PSYCHOVISUAL) == 0 )
    {
      psychovisual = booleanProperty( value, psychovisual );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(FILTER_PRECINCT) == 0 )
      return precinct.ordinal();
    if ( key.compareTo(FILTER_RADIUS) == 0 )
      return radius;
    if ( key.compareTo(METRIC_TYPE) == 0 )
      return metric.ordinal();
    if ( key.compareTo(PSYCHOVISUAL) == 0 )
      return psychovisual;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "ImageCompare";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "RealFunctionToRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Image comparison.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_RASTER + '.' + C_METRIC;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"RealFunction" );
      t.newOutputPlug( "image1", IFACE+"RasterGraphics" );
      t.newOutputPlug( "image2", IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( FILTER_PRECINCT, TYPE_INTEGER, "Filter precinct type", true );
        t.propDefault( FilterPrecinct.FILTER_NONE.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "None",     FilterPrecinct.FILTER_NONE.ordinal(),     "No filtering" );
        t.propEnum( "Cylinder", FilterPrecinct.FILTER_CYLINDER.ordinal(), "Cylindrical filter" );
        t.propEnum( "Cone",     FilterPrecinct.FILTER_CONE.ordinal(),     "Conical filter" );
      t.propEnd();
      t.propBegin( FILTER_RADIUS, TYPE_DOUBLE, "Filter radius in pixels", true );
        t.propDefault( 0.0 );
        t.propBounds(  0.0, 20.0 );
      t.propEnd();
      t.propBegin( METRIC_TYPE, TYPE_INTEGER, "Metric type", true );
        t.propDefault( MetricType.METRIC_BOOLEAN.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Boolean", MetricType.METRIC_BOOLEAN.ordinal(), "Exact image comparison" );
        t.propEnum( "RMSE",    MetricType.METRIC_RMSE.ordinal(),    "Root Mean Square Error" );
        t.propEnum( "MAE",     MetricType.METRIC_MAE.ordinal(),     "Mean Absolute Error" );
        t.propEnum( "AE",      MetricType.METRIC_AE.ordinal(),      "Absolute Error" );
      t.propEnd();
      t.propBegin( PSYCHOVISUAL, TYPE_BOOLEAN, "Use psychovisual component coefficients", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
