package cz.cuni.jagrlib.piece;

import org.nfunk.jep.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Image filter - general image composition.
 *
 * @version 0.26 $Rev: 450 $ $Date: 2007-05-08 19:11:49 +0200 (út, 08 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/ImageCompute.java">ImageCompute.java</a>
 */
public class ImageCompute extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Maximum number of input raster images. */
  public final static int MAX_INPUT = 4;

  /** Input variable names. */
  public final static String[][] iVars =
  {
    { "r1", "g1", "b1", "a1" },
    { "r2", "g2", "b2", "a2" },
    { "r3", "g3", "b3", "a3" },
    { "r4", "g4", "b4", "a4" }
  };

  /** Output variable names ("r" is used for gray value as well). */
  public final static String[] oVars =
  { "r", "g", "b", "a" };

  /** Input constant names. */
  public final static String[] cVars =
  { "x", "y" };

  /** Floating-point constant x. */
  protected double constX = 0.0;

  /** Floating-point constant y. */
  protected double constY = 0.0;

  /** Use {@code 0.0-1.0} clamping? */
  protected boolean clamp = true;

  /**
   * Sequence of {@link org.nfunk.jep.JEP} expressions (separated by newline character)
   * defining color of output pixel.
   *
   * <p>Input values: {@code r1}, {@code g1}, {@code b1}, {@code a1},
   * {@code r2}, {@code g2}, etc.</p>
   *
   * <p>Output pixel: {@code r}, {@code g}, {@code b}, {@code a}</p>
   */
  protected String expr = "";

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transformation.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    int i, j, k, b;
    final RasterGraphics[] input = new RasterGraphics[ MAX_INPUT ];
    for ( i = 0; i < MAX_INPUT; i++ )
      input[i] = (RasterGraphics)getInterface( "raster" + (i+1), IFACE+"RasterGraphics" );
    final RasterGraphics output = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );

    if ( input[0] == null ||
         output   == null ) return false;

      // output image:
    int width      = input[0].getWidth();
    int height     = input[0].getHeight();
    int sourceMode = input[0].getMode();
    int targetMode = sourceMode;
    if ( sourceMode == RasterGraphics.MODE_COLORMAP )
      targetMode = RasterGraphics.MODE_RGB;

    output.init( width, height, targetMode, 0 );

    int outBands = (targetMode == RasterGraphics.MODE_RGBA) ? 4 : 3;
    double[] pixel = new double[ outBands ];

      // prepare expression sequence:
    String[] ex     = (expr == null) ? new String[ 0 ] : expr.split( "\n" );
    int      exLen  = ex.length;          // number of expression lines
    Node[]   lines  = new Node[ exLen ];  // pre-compiled expression nodes
    JEP      parser = new JEP();
    Object   result;

    parser.addStandardConstants();
    parser.addStandardFunctions();
    parser.setImplicitMul( true );
    parser.setAllowUndeclared( true );
    parser.setAllowAssignment( true );

      // input constants:
    parser.addVariable( cVars[0], constX );
    parser.addVariable( cVars[1], constY );

      // result variables:
    for ( b = 0; b < 3; b++ )
      parser.addVariable( oVars[b], 0.0 );
    parser.addVariable( oVars[3], 1.0 );

    for ( b = 0; b < exLen; b++ )
      try
      {
        lines[b] = parser.parse( ex[b] );
      }
      catch ( ParseException e )
      {
        LogFile.error( "ImageCompute: error parsing expression on line " + (b+1) + ": '" + ex[b] + '\'' );
        return false;
      }

      // compute the target image:
    try
    {
      for ( j = 0; j < height; j++ )
        for ( i = 0; i < width; i++ )
        {
            // read input pixels:
          for ( k = 0; k < MAX_INPUT; k++ )
            if ( input[k] != null )
            {
              input[k].getRGB( i, j, pixel );
              for ( b = 0; b < outBands; b++ )
                parser.addVariable( iVars[k][b], pixel[b] );
            }

            // compute sequence of pre-parsed expressions:
          for ( b = 0; b < exLen; b++ )
            parser.evaluate( lines[b] );

            // compose output pixel:
          if ( clamp )
            for ( b = 0; b < outBands; b++ )
            {
              result = parser.getVarValue( oVars[b] );
              pixel[b] = (result instanceof Double) ? Formula.clamp( ((Double)result).doubleValue(), 0.0, 1.0 ) : 0.0;
            }
          else
            for ( b = 0; b < outBands; b++ )
            {
              result = parser.getVarValue( oVars[b] );
              pixel[b] = (result instanceof Double) ? ((Double)result).doubleValue() : 0.0;
            }

          output.putPixel( i, j, pixel );
        }
    }
    catch ( ParseException e )
    {
      LogFile.error( "ImageCompute: error evaluating expressions" );
      return false;
    }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String CONST_X     = "Const x";
  public final static String CONST_Y     = "Const y";
  public final static String CLAMPING    = "Clamping";
  public final static String EXPRESSIONS = "Expressions";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // CONST_X:
    if ( key.compareTo(CONST_X) == 0 )
    {
      constX = doubleProperty( value, constX );
      return;
    }
      // CONST_Y:
    if ( key.compareTo(CONST_Y) == 0 )
    {
      constY = doubleProperty( value, constY );
      return;
    }
      // CLAMPING:
    if ( key.compareTo(CLAMPING) == 0 )
    {
      clamp = booleanProperty( value, clamp );
      return;
    }
      // EXPRESSIONS:
    if ( key.compareTo(EXPRESSIONS) == 0 )
    {
      expr = stringProperty( value, expr );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(CONST_X) == 0 )
      return constX;
    if ( key.compareTo(CONST_Y) == 0 )
      return constY;
    if ( key.compareTo(CLAMPING) == 0 )
      return clamp;
    if ( key.compareTo(EXPRESSIONS) == 0 )
      return expr;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "ImageCompute";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterGraphics5";

  /** Object description. */
  private final static String DESCRIPTION = "Image computation, uses general expressions.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"Trigger" );
      int i;
      for ( i = 0; i < MAX_INPUT; i++ )
        t.newOptOutputPlug( "raster" + (i+1), IFACE+"RasterGraphics" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( CONST_X, TYPE_DOUBLE, "Constant x", true );
        t.propDefault( 0.0 );
      t.propEnd();
      t.propBegin( CONST_Y, TYPE_DOUBLE, "Constant y", true );
        t.propDefault( 0.0 );
      t.propEnd();
      t.propBegin( CLAMPING, TYPE_BOOLEAN, "Clamp result values to [0,1] interval?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( EXPRESSIONS, TYPE_STRING, "Color expressions, each assignment at separate line", true );
        t.propDefault( "" );
        t.propManipulator( MANIPULATOR_MULTILINE );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
