package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Image cropping &amp; scaling.
 * Preserves pixel format (except for {@link RasterGraphics#MODE_COLORMAP} which is
 * converted into {@link RasterGraphics#MODE_RGB}).
 *
 * @version 0.25 $Rev: 385 $ $Date: 2007-01-07 01:47:21 +0100 (ne, 07 I 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/ImageCrop.java">ImageCrop.java</a>
 */
public class ImageCrop extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** X coordinate of upper left corner of the crop area. */
  protected int x0 = 0;

  /** Y coordinate of upper left corner of the crop area. */
  protected int y0 = 0;

  /** Crop area width in pixels ({@code 0} for input image width). */
  protected int width  = 0;

  /** Crop area height in pixels ({@code 0} for input image height). */
  protected int height = 0;

  /** Target image width in pixels ({@code 0} for crop area width). */
  protected int outWidth  = 0;

  /** Target image height in pixels ({@code 0} for crop area height). */
  protected int outHeight = 0;

  /**
   * Resampling types.
   */
  protected static enum ResamplingType
  {
    /** Rounding (fast but low-quality). */
    RES_ROUND,
    /** Bilinear resampling. */
    RES_BILINEAR,
  }

  /** Actual resampling type. */
  protected ResamplingType resampling = ResamplingType.RES_ROUND;

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transformation.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    final RasterGraphics input  = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    final RasterGraphics output = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );

    if ( input  == null ||
         output == null ) return false;

    int i, j;

      // output image:
    int iWid       = (width     == 0) ? input.getWidth()  : width;
    int iHei       = (height    == 0) ? input.getHeight() : height;
    int oWid       = (outWidth  == 0) ? iWid : outWidth;
    int oHei       = (outHeight == 0) ? iHei : outHeight;
    int sourceMode = input.getMode();
    int targetMode = sourceMode;
    int bands      = input.getBands();
    if ( targetMode == RasterGraphics.MODE_COLORMAP )
    {
      targetMode = RasterGraphics.MODE_RGB;
      bands      = 3;
    }

    output.init( oWid, oHei, targetMode, bands );

    int[] pixel = new int[ bands ];

    if ( iWid == oWid &&
         iHei == oHei )                   // cropping w/o scaling
    {
      if ( sourceMode == RasterGraphics.MODE_COLORMAP )
        for ( j = 0; j < oHei; j++ )
          for ( i = 0; i < oWid; i++ )
          {
            input.getRGB( i + x0, j + y0, pixel );
            output.putPixel( i, j, pixel );
          }
      else
        for ( j = 0; j < oHei; j++ )
          for ( i = 0; i < oWid; i++ )
          {
            input.getPixel( i + x0, j + y0, pixel );
            output.putPixel( i, j, pixel );
          }
    }
    else                                  // cropping & scaling
    {
      double kx, ax;                      // src_x = kx * target_x + ax;
      double ky, ay;                      // src_y = ky * target_y + ay;

      kx = iWid / (double)oWid;
      ax = x0 + 0.5 * (kx - 1.0);
      ky = iHei / (double)oHei;
      ay = y0 + 0.5 * (ky - 1.0);

      double x, y;
      int iy, ix;

      if ( resampling == ResamplingType.RES_ROUND )
      {                                   // simple rounding
        if ( sourceMode == RasterGraphics.MODE_COLORMAP )
          for ( j = 0, y = ay; j < oHei; j++, y += ky )
          {
            iy = Formula.round( y );
            for ( i = 0, x = ax; i < oWid; i++, x += kx )
            {
              input.getRGB( Formula.round( x ), iy, pixel );
              output.putPixel( i, j, pixel );
            }
          }
        else
          for ( j = 0, y = ay; j < oHei; j++, y += ky )
          {
            iy = Formula.round( y );
            for ( i = 0, x = ax; i < oWid; i++, x += kx )
            {
              input.getPixel( Formula.round( x ), iy, pixel );
              output.putPixel( i, j, pixel );
            }
          }
      }
      else                                // Bilinear resampling
      {
        double[] pix00 = new double[ bands ];
        double[] pix10 = new double[ bands ];
        double[] pix01 = new double[ bands ];
        double[] pix11 = new double[ bands ];
        double[] pix   = new double[ bands ];
        double   fx, fy, fx1, fy1;
        int      b;

        if ( sourceMode == RasterGraphics.MODE_COLORMAP )
          for ( j = 0, y = ay; j < oHei; j++, y += ky )
          {
            fy  = Math.floor( y );
            iy  = (int)fy;                // y = iy + fy
            fy  = y - fy;
            fy1 = 1.0 - fy;

            for ( i = 0, x = ax; i < oWid; i++, x += kx )
            {
              fx  = Math.floor( x );
              ix  = (int)fx;              // x = ix + fx
              fx  = x - fx;
              fx1 = 1.0 - fx;

                // four surrounding pixels:
              input.getRGB( ix,     iy,     pix00 );
              input.getRGB( ix + 1, iy,     pix10 );
              input.getRGB( ix,     iy + 1, pix01 );
              input.getRGB( ix + 1, iy + 1, pix11 );

                // bilinear interpolation:
              for ( b = 0; b < bands; b++ )
                pix[b] = fx1 * ( fy1 * pix00[b] + fy * pix01[b] ) +
                         fx  * ( fy1 * pix10[b] + fy * pix11[b] );

              output.putPixel( i, j, pix );
            }
          }
        else
          for ( j = 0, y = ay; j < oHei; j++, y += ky )
          {
            fy  = Math.floor( y );
            iy  = (int)fy;                // y = iy + fy
            fy  = y - fy;
            fy1 = 1.0 - fy;

            for ( i = 0, x = ax; i < oWid; i++, x += kx )
            {
              fx  = Math.floor( x );
              ix  = (int)fx;              // x = ix + fx
              fx  = x - fx;
              fx1 = 1.0 - fx;

                // four surrounding pixels:
              input.getPixel( ix,     iy,     pix00 );
              input.getPixel( ix + 1, iy,     pix10 );
              input.getPixel( ix,     iy + 1, pix01 );
              input.getPixel( ix + 1, iy + 1, pix11 );

                // bilinear interpolation:
              for ( b = 0; b < bands; b++ )
                pix[b] = fx1 * ( fy1 * pix00[b] + fy * pix01[b] ) +
                         fx  * ( fy1 * pix10[b] + fy * pix11[b] );

              output.putPixel( i, j, pix );
            }
          }
      }
    }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String X0         = "x0";
  public final static String Y0         = "y0";
  public final static String WIDTH      = "Width";
  public final static String HEIGHT     = "Height";
  public final static String OUT_WIDTH  = "Out width";
  public final static String OUT_HEIGHT = "Out height";
  public final static String RESAMPLING = "Resampling";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // X0:
    if ( key.compareTo(X0) == 0 )
    {
      x0 = intProperty( value, x0 );
      return;
    }
      // Y0:
    if ( key.compareTo(Y0) == 0 )
    {
      y0 = intProperty( value, y0 );
      return;
    }
      // WIDTH:
    if ( key.compareTo(WIDTH) == 0 )
    {
      width = intProperty( value, width, 0, 1000000 );
      return;
    }
      // HEIGHT:
    if ( key.compareTo(HEIGHT) == 0 )
    {
      height = intProperty( value, height, 0, 1000000 );
      return;
    }
      // OUT_WIDTH:
    if ( key.compareTo(OUT_WIDTH) == 0 )
    {
      outWidth = intProperty( value, outWidth, 0, 1000000 );
      return;
    }
      // OUT_HEIGHT:
    if ( key.compareTo(OUT_HEIGHT) == 0 )
    {
      outHeight = intProperty( value, outHeight, 0, 1000000 );
      return;
    }
      // RESAMPLING:
    if ( key.compareTo(RESAMPLING) == 0 )
    {
      resampling = enumProperty( value, resampling );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(X0) == 0 )
      return x0;
    if ( key.compareTo(Y0) == 0 )
      return y0;
    if ( key.compareTo(WIDTH) == 0 )
      return width;
    if ( key.compareTo(HEIGHT) == 0 )
      return height;
    if ( key.compareTo(OUT_WIDTH) == 0 )
      return outWidth;
    if ( key.compareTo(OUT_HEIGHT) == 0 )
      return outHeight;
    if ( key.compareTo(RESAMPLING) == 0 )
      return resampling.ordinal();
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "ImageCrop";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Image cropping & scaling.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"Trigger" );
      t.newOutputPlug( "image1", IFACE+"RasterGraphics" );
      t.newOutputPlug( "image2", IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( X0, TYPE_INTEGER, "X0", true );
        t.propDefault( 0 );
      t.propEnd();
      t.propBegin( Y0, TYPE_INTEGER, "Y0", true );
        t.propDefault( 0 );
      t.propEnd();
      t.propBegin( WIDTH, TYPE_INTEGER, "Input rectangle width", true );
        t.propDefault( 0 );
        t.propBounds(  0, 1000000 );
      t.propEnd();
      t.propBegin( HEIGHT, TYPE_INTEGER, "Input rectangle height", true );
        t.propDefault( 0 );
        t.propBounds(  0, 1000000 );
      t.propEnd();
      t.propBegin( OUT_WIDTH, TYPE_INTEGER, "Result image width", true );
        t.propDefault( 0 );
        t.propBounds(  0, 1000000 );
      t.propEnd();
      t.propBegin( OUT_HEIGHT, TYPE_INTEGER, "Result image height", true );
        t.propDefault( 0 );
        t.propBounds(  0, 1000000 );
      t.propEnd();
      t.propBegin( RESAMPLING, TYPE_INTEGER, "Image resampling type", true );
        t.propDefault( ResamplingType.RES_ROUND.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Round",    ResamplingType.RES_ROUND.ordinal(),    "Rounding" );
        t.propEnum( "Bilinear", ResamplingType.RES_BILINEAR.ordinal(), "Bilinear resampling" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
