package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * K-D-tree coding of raster images.
 * Suitable also for Delaynay vertex encoding (intermediate data are stored in {@link RasterGraphics}).
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/ImageKDTree.java">ImageKDTree.java</a>
 */
public class ImageKDTree extends Piece implements KDTree
{
  //==========================================================================
  //  Data:

  /** Construction mode. */
  protected int mode = MODE_SPACE_BALANCED;

  /** Use sample color as the 3rd dimension? */
  protected boolean colorDimension = true;

  /** K-D-tree root node. */
  protected IntNode root = null;

  /** Accelerator: attached raster image. */
  protected RasterGraphics image = null;

  //==========================================================================
  //  K-D-tree construction:

  /** Recursively build subordinate nodes.. */
  protected IntNode boxSpaceBalanced3D ( IntNode node, int[] col )
  {
      // determine a dimension with the largest extent:
    int maxSize = node.upper[0] - node.lower[0];
    int maxDim  = 0;

    int size = node.upper[1] - node.lower[1];
    if ( size > maxSize )
    {
      maxSize = size;
      maxDim  = 1;
    }
    size = node.upper[2] - node.lower[2];
    if ( size > maxSize )
    {
      maxSize = size;
      maxDim  = 2;
    }

    if ( maxSize == 0 )                   // leaf node => check if there is a sample here..
    {
      image.getRGB( node.lower[0], node.lower[1], col );
      node.cardinality = 1;
      return( col[0] == 0 ? null : node ); // empty leaf node => throw it away
    }

    node.dim   = maxDim;                  // splitting dimension
    node.split = (node.lower[maxDim] + node.upper[maxDim] + 1) >> 1;

      // create and check the left node:
    node.left  = new IntNode();
    node.left.lower = node.lower;         // is it dangerous? I hope not..
    node.left.upper = Formula.cloneArray( node.upper );
    node.left.upper[maxDim] = node.split - 1;
    node.left = boxSpaceBalanced3D( node.left, col );
    node.cardinality = (node.left != null) ? node.left.cardinality : 0;

      // create and check the left node:
    node.right  = new IntNode();
    node.right.upper = node.upper;        // is it dangerous? I hope not..
    node.right.lower = Formula.cloneArray( node.lower );
    node.right.lower[maxDim] = node.split;
    node.right = boxSpaceBalanced3D( node.right, col );
    if ( node.right != null )
      node.cardinality += node.right.cardinality;

    return( node.cardinality > 0 ? node : null );
  }

  /**
   * Builds space-balanced K-D-tree from the attached raster image.
   * Sample-color is used as the 3rd dimension =&gt; it is splitted the same way spatial dimensions are..
   */
  protected void buildSpaceBalanced3D ()
  {
    final int width  = image.getWidth();
    final int height = image.getHeight();
    int x, y;

      // determine color bounds:
    int minColor = 255;
    int maxColor = 0;
    int[] col = new int[ 3 ];
    int color;

    for ( y = 0; y < height; y++ )
      for ( x = 0; x < width; x++ )
      {
        image.getRGB( x, y, col );
        if ( col[0] > 0 )                 // valid sample
        {
          color = col[1];
          if ( color < minColor ) minColor = color;
          if ( color > maxColor ) maxColor = color;
        }
      }

      // .. encoding [ width x height x <minColor,maxColor> ] box
    root = new IntNode();
    root.lower = new int[] {       0,        0, minColor };
    root.upper = new int[] { width-1, height-1, maxColor };

    boxSpaceBalanced3D( root, col );      // root node should remain intact even if there are no samples..
  }

  //==========================================================================
  //  KDTree utilization:

  /** Recursively utilize the given sub-tree. */
  protected void utilize3D ( IntNode node, int[] col )
  {
    if ( node.left  == null &&
         node.right == null )             // leaf node => output the sample
    {
      col[1] =
      col[2] = node.lower[2];
      image.putPixel( node.lower[0], node.lower[1], col );
      return;
    }

    if ( node.left != null )
      utilize3D( node.left, col );

    if ( node.right != null )
      utilize3D( node.right, col );
  }

  //==========================================================================
  //  Interface KDTree:

  /**
   * Builds the K-D-tree using actual settings.
   * Uses actual construction-settings.
   *
   * @see    #mode
   * @see    #colorDimension
   */
  public void buildTree ()
  {
    image = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    if ( image == null ) return;

    root = null;
    switch ( mode )
    {
      case MODE_SPACE_BALANCED:
        if ( colorDimension )
          buildSpaceBalanced3D();
        break;
    }
  }

  /**
   * Returns actual root node.
   */
  public IntNode getIntRoot ()
  {
    return root;
  }

  /**
   * Tree is assigned from externally prepared data.
   *
   * @param  r The new root node.
   */
  public void setRoot ( IntNode r )
  {
    root = r;
  }

  /**
   * Utilize (use) the given tree. There is possibility to enter the tree externally..
   *
   * @param  r The root node or {@code null} if internal tree should be used.
   */
  public void utilizeTree ( IntNode r )
  {
    if ( r == null ) r = root;
    if ( r == null ) return;

    image = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    if ( image == null ) return;

      // check the output image:
    final int width  = r.upper[0] + 1;
    final int height = r.upper[1] + 1;
    if ( width  != image.getWidth() ||
         height != image.getHeight() )
      image.init( width, height, RasterGraphics.MODE_RGB, 0 );
    image.setRectangle( 0, 0, width, height, BrepToImage.BACKGROUND );

    int[] col = BrepToImage.USED;
    switch ( mode )
    {
      case MODE_SPACE_BALANCED:
        if ( colorDimension )
          utilize3D( root, col );
        break;
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String USE_COLOR = "Color is dimension";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // CONSTRUCTION_MODE:
    if ( key.compareTo(CONSTRUCTION_MODE) == 0 )
    {
      mode = intProperty( value, mode, MODE_SPACE_BALANCED, MODE_BREADTH_BALANCED );
      return;
    }
      // USE_COLOR:
    if ( key.compareTo(USE_COLOR) == 0 )
    {
      colorDimension = booleanProperty( value, colorDimension );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(CONSTRUCTION_MODE) == 0 )
      return mode;
    if ( key.compareTo(USE_COLOR) == 0 )
      return colorDimension;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "ImageKDTree";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "KDTreeToRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "K-D-tree for raster image encoding.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_RASTER + '.' + C_DATA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"KDTree" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( CONSTRUCTION_MODE, TYPE_INTEGER, "Construction mode", true );
        t.propDefault( MODE_SPACE_BALANCED );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Space-balanced",   MODE_SPACE_BALANCED,   "Split plane .. middle of the cell" );
        t.propEnum( "Breadth-balanced", MODE_BREADTH_BALANCED, "Split plane .. median" );
      t.propEnd();
      t.propBegin( USE_COLOR, TYPE_BOOLEAN, "Use color as the 3rd dimension?", true );
        t.propDefault( true );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
