package cz.cuni.jagrlib.piece;

import java.util.Arrays;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Raster image prediction template.
 *
 * @version 0.25 $Rev: 385 $ $Date: 2007-01-07 01:47:21 +0100 (ne, 07 I 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/ImagePredictor.java">ImagePredictor.java</a>
 */
public class ImagePredictor extends DefaultIntFunction
{
  //==========================================================================
  //  Data:

  /**
   * Implemented predictors.
   * Basic schema (X is pixel to be predicted):
   * <pre>
   *   A   B
   *     \ |
   *   C - X
   * </pre>
   */
  protected static enum PredictorType
  {
    /** Void predictor (<b>0</b>). #0 in JPEG. */
    PREDICT_NONE,
    /** Horizontal predictor (<b>C</b>). #1 in JPEG. */
    PREDICT_HORIZONTAL,
    /** Vertical predictor (<b>B</b>). #2 in JPEG. */
    PREDICT_VERTICAL,
    /** Oblique predictor (<b>A</b>). #3 in JPEG. */
    PREDICT_OBLIQUE,
    /** Planar predictor (<b>B + C - A</b>). #4 in JPEG. */
    PREDICT_PLANAR,
    /** Leftmost predictor (<b>C + (B - A) / 2</b>). #5 in JPEG. */
    PREDICT_LEFTMOST,
    /** Upmost predictor (<b>B + (C - A) / 2</b>). #6 in JPEG. */
    PREDICT_UPMOST,
    /** Two nearest neighbours (<b>(B + C) / 2</b>). #7 in JPEG. */
    PREDICT_NEAR,
  }

  /** Actual filter precinct. */
  protected PredictorType predictor = PredictorType.PREDICT_PLANAR;

  /** Precinct size (whatever it means - not used yet). */
  protected int precinctSize = 1;

  /** Maximal value of {@link #precinctSize}. */
  public final static int MAX_PRECINCT = 16;

  //==========================================================================
  //  Support:

  /** Support pixel array (RGBA format). */
  protected int[] pixel = new int[ 4 ];

  /** Reference to the raster image (accelerator). */
  protected RasterGraphics image;

  /** Image width (accelerator). */
  protected int width;

  /** Image height (accelerator). */
  protected int height;

  /** Maximum value of one channel (accelerator). */
  protected int maxValue;

  /** Prepares all accelerators, returns {@code true} if failed. */
  protected final boolean assertImage ()
  {
    if ( image != null ) return false;

    image = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    if ( image == null ) return true;

    width    = image.getWidth();
    height   = image.getHeight();
    maxValue = image.getMaxRGB();

    return false;
  }

  /** Takes image borders into account, fills {@link #pixel} array. */
  protected final void readPixel ( int x, int y )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      Arrays.fill( pixel, 128 );
    else
      image.getPixel( x, y, pixel );
  }

  /** Clamps the given value to range {@code 0} - {@link #maxValue}. */
  protected final int clamp ( int value )
  {
    return Formula.clamp( value, 0, maxValue );
  }

  //==========================================================================
  //  Interface IntFunction:

  /**
   * Compute prediction for (yet not transferred) pixel {@code [ x, y ]}.
   *
   * @param  x Horizontal coordinate (from {@code 0} to {@link RasterGraphics#getWidth}<tt> - 1</tt>).
   * @param  y Vertical coordinate (from {@code 0} to {@link RasterGraphics#getHeight}<tt> - 1</tt>).
   * @param  channel Color-channel number: {@code 0} .. Red, {@code 1} .. Green, {@code 2} .. Blue,
   *                 {@code 3} .. Alpha.
   * @return Predicted channel value.
   */
  @Override
  public int g ( int x, int y, int channel )
  {
    if ( assertImage() ) return 0;

    int tmp;
    switch ( predictor )                  // A   B
    {                                     //   \ |
                                          // C - X

      case PREDICT_HORIZONTAL:            // C
        readPixel( x - 1, y );
        return pixel[ channel ];

      case PREDICT_VERTICAL:              // B
        readPixel( x, y - 1 );
        return pixel[ channel ];

      case PREDICT_OBLIQUE:               // A
        readPixel( x - 1, y - 1 );
        return pixel[ channel ];

      case PREDICT_PLANAR:                // B + C - A
        readPixel( x - 1, y );
        tmp  = pixel[ channel ];
        readPixel( x, y - 1 );
        tmp += pixel[ channel ];
        readPixel( x - 1, y - 1 );
        return clamp( tmp - pixel[ channel ] );

      case PREDICT_LEFTMOST:              // C + (B - A) / 2
        readPixel( x, y - 1 );
        tmp  = pixel[ channel ] + 1;      // for rounding!
        readPixel( x - 1, y - 1 );
        tmp -= pixel[ channel ];
        readPixel( x - 1, y );
        return clamp( pixel[ channel ] + (tmp >> 1) );

      case PREDICT_UPMOST:                // B + (C - A) / 2
        readPixel( x - 1, y );
        tmp  = pixel[ channel ] + 1;      // for rounding!
        readPixel( x - 1, y - 1 );
        tmp -= pixel[ channel ];
        readPixel( x, y - 1 );
        return clamp( pixel[ channel ] + (tmp >> 1) );

      case PREDICT_NEAR:                  // (B + C) / 2
        readPixel( x - 1, y );
        tmp  = pixel[ channel ];
        readPixel( x, y - 1 );
        return( (tmp + pixel[ channel ] + 1) >> 1 );

      case PREDICT_NONE:                  // 0
      default:
        return 0;
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String PREDICTOR_TYPE = "PredictorType";
  public final static String PRECINCT_SIZE  = "PrecinctSize";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // PREDICTOR_TYPE:
    if ( key.compareTo(PREDICTOR_TYPE) == 0 )
    {
      predictor = enumProperty( value, PredictorType.PREDICT_NONE );
      return;
    }
      // PRECINCT_SIZE:
    if ( key.compareTo(PRECINCT_SIZE) == 0 )
    {
      precinctSize = intProperty( value, precinctSize, 0, MAX_PRECINCT );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(PREDICTOR_TYPE) == 0 )
      return predictor.ordinal();
    if ( key.compareTo(PRECINCT_SIZE) == 0 )
      return precinctSize;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "ImagePredictor";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "IntFunctionToRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Simple 2D image predictors.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + ".predictor";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"IntFunction" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( PREDICTOR_TYPE, TYPE_INTEGER, "Image predictor type", true );
        t.propDefault( PredictorType.PREDICT_PLANAR.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "None 0",             PredictorType.PREDICT_NONE.ordinal(),       "Void prediction (0)" );
        t.propEnum( "Horizontal C",       PredictorType.PREDICT_HORIZONTAL.ordinal(), "Horizontal prediction (left neighbour)" );
        t.propEnum( "Vertical B",         PredictorType.PREDICT_VERTICAL.ordinal(),   "Vertical prediction (upper neighbour)" );
        t.propEnum( "Oblique A",          PredictorType.PREDICT_OBLIQUE.ordinal(),    "Oblique prediction (upperLeft neighbour)" );
        t.propEnum( "Planar B+C-A",       PredictorType.PREDICT_PLANAR.ordinal(),     "Planar predictor (upper + left - upperLeft)" );
        t.propEnum( "Leftmost C+(B-A)/2", PredictorType.PREDICT_LEFTMOST.ordinal(),   "Leftmost predictor (left + (upper - upperLeft) / 2)" );
        t.propEnum( "Upmost B+(C-A)/2",   PredictorType.PREDICT_UPMOST.ordinal(),     "Upmost predictor (upper + (left - upperLeft) / 2)" );
        t.propEnum( "Near (B+C)/2",       PredictorType.PREDICT_NEAR.ordinal(),       "Two nearest neighbours (upper + left) / 2" );
      t.propEnd();
      t.propBegin( PRECINCT_SIZE, TYPE_INTEGER, "Precinct size in pixels", true );
        t.propDefault( 1 );
        t.propBounds(  0, MAX_PRECINCT );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
