package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Transition effect between two input raster images.
 *
 * @version 0.25 $Rev: 385 $ $Date: 2007-01-07 01:47:21 +0100 (ne, 07 I 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/ImageTransition.java">ImageTransition.java</a>
 */
public class ImageTransition extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Time variable from {@code < 0.0, 1.0 >}. */
  protected double time = 0.0;

  /**
   * Image field function: R<sup>2</sup> {@code -> <0,1>} mapping.
   */
  protected static enum FieldType
  {
    /** Constant value: 1/2. */
    FIELD_NONE,
    /** External field value. */
    FIELD_EXTERNAL,
    /** Linear ramp, see {@link Direction}. */
    FIELD_RAMP,
    /** Random field. */
    FIELD_RANDOM,
  }

  /** Actual field type. */
  protected FieldType fieldType = FieldType.FIELD_NONE;

  /** Eight main directions for ramps, .. */
  public static enum Direction
  {
    /** None = {@code [ 0, 0 ]}. */
    DIR_NONE,
    /** East = {@code [ 1, 0 ]}. */
    DIR_E,
    /** North-east = {@code [ 1, -1 ]}. */
    DIR_NE,
    /** North = {@code [ 0, -1 ]}. */
    DIR_N,
    /** North-west = {@code [ -1, -1 ]}. */
    DIR_NW,
    /** West = {@code [ -1, 0 ]}. */
    DIR_W,
    /** South-west = {@code [ -1, 1 ]}. */
    DIR_SW,
    /** South = {@code [ 0, 1 ]}. */
    DIR_S,
    /** South-east = {@code [ 1, 1 ]}. */
    DIR_SE
  }

  /** Actual field direction. */
  protected Direction fieldDirection = Direction.DIR_NONE;

  /** Set negative value for randomize. */
  protected int seed = -1;

  /** Time transition type. */
  protected static enum TransitionType
  {
    /** External transition function. */
    TR_EXTERNAL,
    /** Cutoff function (at 1/2). */
    TR_CUTOFF,
    /** Linear interpolation. */
    TR_LINEAR,
    /** Goniometric sigma function. */
    TR_SIGMA,
  }

  /** Actual transition type. */
  protected TransitionType transition = TransitionType.TR_LINEAR;

  /** Transition size. {@code 1.0} .. transition extends over the whole image. */
  protected double size = 1.0;

  //==========================================================================
  //  RealFunctions for image field:

  /**
   * Ramp image field function.
   *
   * Only {@link RealFunction#f(double,double)} is used..
   */
  public static class RampField extends DefaultRealFunction
  {
    protected Direction dir;

    public RampField ( Direction d )
    {
      dir = d;
    }

    /**
     * Compute the mapping: {@code double[2] -> double}.
     */
    public double f ( double x, double y )
    {
      switch ( dir )
      {
        case DIR_E:
          return Formula.clamp( x, 0.0, 1.0 );
        case DIR_NE:
          return Formula.clamp( (1.0 + x - y) * 0.5, 0.0, 1.0 );
        case DIR_N:
          return Formula.clamp( 1.0 - y, 0.0, 1.0 );
        case DIR_NW:
          return Formula.clamp( (2.0 - x - y) * 0.5, 0.0, 1.0 );
        case DIR_W:
          return Formula.clamp( 1.0 - x, 0.0, 1.0 );
        case DIR_SW:
          return Formula.clamp( (1.0 - x + y) * 0.5, 0.0, 1.0 );
        case DIR_S:
          return Formula.clamp( y, 0.0, 1.0 );
        case DIR_SE:
          return Formula.clamp( (x + y) * 0.5, 0.0, 1.0 );
      }
      return 0.5;
    }
  }

  /**
   * Random image field function.
   *
   * Only {@link RealFunction#f(double,double)} is used..
   */
  public static class RandomField extends DefaultRealFunction
  {
    final protected RandomJames rnd;

    public RandomField ( int seed )
    {
      if ( seed >= 0 )
        rnd = new RandomJames( seed >>> 15, seed & 0x7fff );
      else
      {
        rnd = new RandomJames();
        rnd.randomize();
      }
    }

    /**
     * Compute the mapping: {@code double[2] -> double}.
     */
    public double f ( double x, double y )
    {
      return rnd.uniformNumber();
    }
  }

  //==========================================================================
  //  RealFunctions for time transition:

  /**
   * Cutoff transition.
   *
   * Only {@link RealFunction#f(double)} is used..
   */
  public static class CutoffTransition extends DefaultRealFunction
  {
    /**
     * Compute the mapping: {@code double -> double}.
     */
    public double f ( double x )
    {
      return( x < 0.5 ? 0.0 : 1.0 );
    }
  }

  /**
   * Linear transition.
   *
   * Only {@link RealFunction#f(double)} is used..
   */
  public static class LinearTransition extends DefaultRealFunction
  {
    /**
     * Compute the mapping: {@code double -> double}.
     */
    public double f ( double x )
    {
      return Formula.clamp( x, 0.0, 1.0 );
    }
  }

  /**
   * Goniometric sigma function.
   *
   * Only {@link RealFunction#f(double)} is used..
   */
  public static class SigmaTransition extends DefaultRealFunction
  {
    /**
     * Compute the mapping: {@code double -> double}.
     */
    public double f ( double x )
    {
      return( 0.5 + 0.5 * Math.sin( (Formula.clamp( x, 0.0, 1.0 ) - 0.5) * Math.PI ) );
    }
  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transformation.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    final RasterGraphics input1 = (RasterGraphics)getInterface( "input1",     IFACE+"RasterGraphics" );
    final RasterGraphics input2 = (RasterGraphics)getInterface( "input2",     IFACE+"RasterGraphics" );
    final RasterGraphics output = (RasterGraphics)getInterface( PL_OUTPUT,    IFACE+"RasterGraphics" );
          RealFunction   field  = (RealFunction)  getInterface( "field",      IFACE+"RealFunction" );
          RealFunction   trans  = (RealFunction)  getInterface( "transition", IFACE+"RealFunction" );

    if ( input1 == null ||
         output == null ) return false;

      // output image:
    int width  = Math.max( input1.getWidth(), 2 );
    int height = Math.max( input1.getHeight(), 2 );
    int mode   = input1.getMode();
    int bands  = Math.max( input1.getBands(), 3 );
    if ( mode == RasterGraphics.MODE_COLORMAP )
      mode = RasterGraphics.MODE_RGB;

    output.init( width, height, mode, bands );

      // image field instance:
    double actField = 0.5;
    switch ( fieldType )
    {
      case FIELD_NONE:
        field = null;
        break;
      case FIELD_RAMP:
        field = new RampField( fieldDirection );
        break;
      case FIELD_RANDOM:
        field = new RandomField( seed );
        break;
    }

      // time transition instance:
    switch ( transition )
    {
      case TR_CUTOFF:
        trans = new CutoffTransition();
        break;
      case TR_SIGMA:
        trans = new SigmaTransition();
        break;
      case TR_EXTERNAL:
        if ( trans != null ) break;
      case TR_LINEAR:
      default:
        trans = new LinearTransition();
    }

    double[] inp1 = new double[ bands ];
    double[] inp2 = new double[ bands ];
    double[] out  = new double[ bands ];
    int i, j, b;
    double coef, tr;
    double x, y;
    final double dx    = 1.0 / (width  - 1.0);
    final double dy    = 1.0 / (height - 1.0);
    final double size1 = 1.0 / size;

      // compute the target image:
    for ( j = 0, y = 0.0; j < height; j++, y += dy )
      for ( i = 0, x = 0.0; i < width; i++, x += dx )
      {
          // read input pixels:
        input1.getRGB( i, j, inp1 );
        if ( input2 != null )
          input2.getRGB( i, j, inp2 );

        if ( field != null )
          actField = field.f( x, y );

        tr   = Formula.clamp( size1 * (actField + size - time - time * size), 0.0, 1.0 );
        coef = Formula.clamp( trans.f( 1.0 - tr ), 0.0, 1.0 );

        for ( b = 0; b < bands; b++ )
          out[b] = (1.0 - coef) * inp1[b] + coef * inp2[b];

        output.putPixel( i, j, out );
      }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String TIME       = "Time";
  public final static String FIELD_TYPE = "Field type";
  public final static String FIELD_DIR  = "Field direction";
  public final static String SEED       = "Seed";
  public final static String TRANS_TYPE = "Transition type";
  public final static String TRANS_SIZE = "Transition size";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // TIME:
    if ( key.compareTo(TIME) == 0 )
    {
      time = doubleProperty( value, time, 0.0, 1.0 );
      return;
    }
      // FIELD_TYPE:
    if ( key.compareTo(FIELD_TYPE) == 0 )
    {
      fieldType = enumProperty( value, fieldType );
      return;
    }
      // FIELD_DIR:
    if ( key.compareTo(FIELD_DIR) == 0 )
    {
      fieldDirection = enumProperty( value, fieldDirection );
      return;
    }
      // SEED:
    if ( key.compareTo(SEED) == 0 )
    {
      seed = intProperty( value, seed );
      return;
    }
      // TRANS_TYPE:
    if ( key.compareTo(TRANS_TYPE) == 0 )
    {
      transition = enumProperty( value, transition );
      return;
    }
      // TRANS_SIZE:
    if ( key.compareTo(TRANS_SIZE) == 0 )
    {
      size = doubleProperty( value, size, 1.0e-6, 100.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(TIME) == 0 )
      return time;
    if ( key.compareTo(FIELD_TYPE) == 0 )
      return fieldType.ordinal();
    if ( key.compareTo(FIELD_DIR) == 0 )
      return fieldDirection.ordinal();
    if ( key.compareTo(SEED) == 0 )
      return seed;
    if ( key.compareTo(TRANS_TYPE) == 0 )
      return transition.ordinal();
    if ( key.compareTo(TRANS_SIZE) == 0 )
      return size;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "ImageTransition";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterGraphics3AndRealFunction2";

  /** Object description. */
  private final static String DESCRIPTION = "Transition of two raster images.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,     IFACE+"Trigger" );
      t.newOutputPlug(    "input1",     IFACE+"RasterGraphics" );
      t.newOptOutputPlug( "input2",     IFACE+"RasterGraphics" );
      t.newOutputPlug(    PL_OUTPUT,    IFACE+"RasterGraphics" );
      t.newOptOutputPlug( "field",      IFACE+"RealFunction" );
      t.newOptOutputPlug( "transition", IFACE+"RealFunction" );
        // Property:
      t.propBegin( TIME, TYPE_DOUBLE, "Time variable from <0.0,1.0>", true );
        t.propDefault( 0.0 );
        t.propBounds(  0.0, 1.0 );
      t.propEnd();
      t.propBegin( FIELD_TYPE, TYPE_INTEGER, "Image field type", true );
        t.propDefault( FieldType.FIELD_NONE.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "None",        FieldType.FIELD_NONE.ordinal(),     "Empty field" );
        t.propEnum( "External",    FieldType.FIELD_EXTERNAL.ordinal(), "External field (see plug)" );
        t.propEnum( "Linear ramp", FieldType.FIELD_RAMP.ordinal(),     "Linear ramp (see Direction)" );
        t.propEnum( "Random",      FieldType.FIELD_RANDOM.ordinal(),   "Random field" );
      t.propEnd();
      t.propBegin( SEED, TYPE_INTEGER, "Random generator seed (-1 for randomize)", true );
        t.propDefault( 12 );
      t.propEnd();
      t.propBegin( FIELD_DIR, TYPE_INTEGER, "Direction of image field", true );
        t.propDefault( Direction.DIR_NONE.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "None",       Direction.DIR_NONE.ordinal(), "Void direction" );
        t.propEnum( "East",       Direction.DIR_E.ordinal(),    "East" );
        t.propEnum( "North-east", Direction.DIR_NE.ordinal(),   "North-east" );
        t.propEnum( "North",      Direction.DIR_N.ordinal(),    "North" );
        t.propEnum( "North-west", Direction.DIR_NW.ordinal(),   "North-west" );
        t.propEnum( "West",       Direction.DIR_W.ordinal(),    "West" );
        t.propEnum( "South-West", Direction.DIR_SW.ordinal(),   "South-West" );
        t.propEnum( "South",      Direction.DIR_S.ordinal(),    "South" );
        t.propEnum( "South-east", Direction.DIR_SE.ordinal(),   "South-east" );
      t.propEnd();
      t.propBegin( TRANS_TYPE, TYPE_INTEGER, "Time transition type", true );
        t.propDefault( TransitionType.TR_LINEAR.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "External", TransitionType.TR_EXTERNAL.ordinal(), "External transition function (see plug)" );
        t.propEnum( "Cutoff",   TransitionType.TR_CUTOFF.ordinal(),   "Cutoff function at 1/2" );
        t.propEnum( "Linear",   TransitionType.TR_LINEAR.ordinal(),   "Linear transition" );
        t.propEnum( "Sigma",    TransitionType.TR_SIGMA.ordinal(),    "Goniometric sigma transition" );
      t.propEnd();
      t.propBegin( TRANS_SIZE, TYPE_DOUBLE, "Transition size (1.0 .. the whole image)", true );
        t.propDefault( 1.0 );
        t.propBounds(  1.0e-6, 100.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
