package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Entropy histogram implementation: map of cummulative frequency interval trees.
 *
 * @author  Jan Kadlec <rrrola@gmail.com>
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/IntervalArray.java">IntervalArray.java</a>
 */
public class IntervalArray extends Piece implements EntropyHistogram
{
  /**
   * Cummulative frequency interval tree.
   * Implemented as an in-place array:
   * <pre>
   *     total a+..+s
   *      |                                             q+r+s
   *      |                i+j+k+l+m+n+o+p                |
   *      |        e+f+g+h        |        m+n+o+p        |
   *      |    c+d    |    g+h    |    k+l    |    o+p    |     s
   * 19:  |  b  |  d  |  f  |  h  |  j  |  l  |  n  |  p  |  r  |
   * p[]  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18
   * </pre>
   *
   * @author  Jan Kadlec <rrrola@gmail.com>
   * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
   * @since   0.24
   */
  protected class CFTree
  {
    /**
     * Upper bound for number of symbols.
     * Depends on {@link ArithmeticCodec#MAX_DECIMATION_LIMIT}.
     */
    static final int MAX_SYMBOLS = 1 << 22;

    /** Actual number of symbols. */
    int    s;
    /** ceil( log2( s ) ). */
    int    m;
    /** CFtree array, p[0] = total sum */
    long[] p;
    /** Total sum limit. */
    long   l;
    /** Minimum frequency. */
    long   min;

    /** Converts a CFTree to a regular array. */
    private final void unwrap ()
    {
      for ( int i = 1; i < s; i++ )
        p[ i & i-1 ] -= p[i];
    }

    /** Converts a regular array to a CFTree. */
    private void wrap ()
    {
      for ( int i = s-1; i > 0; i-- )
        p[ i & i-1 ] += p[i];
    }

    /** Halves all symbol frequencies. */
    public void decimate ()
    {
      unwrap();

      if ( min == 0 )
        for ( int i = 0; i < s; i++ ) p[i] >>= 1;
      else
        for ( int i = 0; i < s; i++ ) p[i] = (p[i] + 1) >> 1;

      wrap();
    }

    /** Sets frequencies. Checks frequency bounds. */
    public void set ( long[] a )
    {
      if ( a == null )
        Arrays.fill( p, min );
      else
      {
        int len = Math.min( a.length, s );
        for ( int i = 0; i < len; i++ )
          p[i] = Math.max( a[i], min );

        if ( len < s )
          Arrays.fill( p, len, s, min );
      }

      wrap();

      while ( p[0] > l ) decimate();
    }

    /** Increments frequencies. Checks bounds. */
    public void inc ( long[] a )
    {
      unwrap();

      for ( int i = 0; i < s; i++ )
      {
        p[i] += a[i];
        if ( p[i] < min ) p[i] = min;
      }

      wrap();

      while ( p[0] > l ) decimate();
    }

    public CFTree ( int symbols, long sumLimit, boolean nonZero )
    {
      init( symbols, sumLimit, nonZero );
    }

    /** [Re-]initializes the array. */
    public void init ( int symbols, long sumLimit, boolean nonZero )
    {
      s = Formula.clamp( symbols, 1, MAX_SYMBOLS );
      for ( m = 1; m < s; m <<= 1 ) ;

      p   = new long[ s ];
      l   = sumLimit;
      min = (nonZero ? 1 : 0);

      Arrays.fill( p, min );

      if ( nonZero )
        wrap();

      long minl = min * s;
      if ( l < minl ) l = s / m + minl;   // sanity check
    }

    /** Resizes the array. */
    public void newSize ( int n )
    {
      if ( n == s ) return;

      n = Formula.clamp( n, 1, MAX_SYMBOLS );
      long[] q = new long[ n ];

      unwrap();

      if ( n <= s )
        System.arraycopy( p, 0, q, 0, n );
      else
      {
        System.arraycopy( p, 0, q, 0, s );
        Arrays.fill( q, s, n, min );
      }
      s = n; p = q;
      for ( m = 1; m < s; m <<= 1 ) ;

      wrap();

      long minl = min * s;
      if ( l < minl ) l = s / m + minl;   // sanity check

      while ( p[0] > l ) decimate();
    }

    /** Changes frequency sum limit. */
    public void newLimit ( long n )
    {
      if ( n == l ) return;

      l = n;
      long minl = min * s;
      if ( l < minl ) l = s / m + minl;   // sanity check

      while ( p[0] > l ) decimate();
    }

    /** Changes minimum frequency. */
    public void newNonZero ( boolean n )
    {
      long nmin = (n ? 1 : 0);
      if ( min == nmin ) return;

      if ( (min = nmin) == 1 )
      {
        unwrap();

        for ( int i = 0; i < s; i++ )
          if ( p[i] == 0 ) p[i] = 1;

        wrap();

        while ( p[0] > l ) decimate();
      }
    }

    /** Interval length. */
    public long getF ( int i )
    {
      if ( i < 0 || i >= s ) return 0;

      long f = p[i];
      int  j = 1;
      while ( (i & j) == 0 && i + j < s )
      {
        f -= p[i+j];
        j <<= 1;
      }
      return f;
    }

    /** Interval start. */
    public long getCf ( int i )
    {
      if ( i <= 0 || i > s ) return 0;

      long cf = p[0];
      while ( i < s )
      {
        cf -= p[i];
        i = (i | i-1) + 1;
      }
      return cf;
    }

    /** Interval number. */
    public int getI ( long cf )
    {
      if ( cf < 0 || cf >= p[0] ) return -1;

      int  i = 0;
      int  j = m;
      long q = p[0];
      while ( (j >>= 1) != 0 )
      {
        if ( i + j < s && cf >= (q -= p[i+j]) )
        {
          cf -= q;
          q   = p[ i += j ];
        }
      }
      return i;
    }

    /** Change interval length. */
    public void update ( int i, long d )
    {
      while ( i > 0 )
      {
        p[i] += d;
        i    &= i - 1;
      }
      if ( (p[0] += d) > l ) decimate();
    }
  }

  /** Identifier of the actual context. */
  protected int ctxId;

  /** The actual context. */
  protected CFTree ctx;

  /** All open contexts. */
  protected IntMap< CFTree > contexts;

  /** Default number of symbols. */
  protected int defaultSymbols;

  /** Default frequency sum limit. */
  protected long defaultSumLimit;

  /** Default non-zero flag. */
  protected boolean defaultNonZero;

  //============================================
  //  Implementation

  public IntervalArray ()
  {
    contexts = new IntMap< CFTree >( 0 );
  }

  /**
   * Histogram (re-)initialization.
   * Destroys all previously collected statistics.
   *
   * @param  _defaultSymbols Default number of symbols for new contexts.
   * @param  _defaultSumLimit Default frequency sum limit for new contexts ({@code 0} for no bounds).
   * @param  _defaultNonZero Keeps frequencies non-zero at all times.
   *         Default frequencies are the lowest possible.
   * @param  _ctxId The initial context.
   * @see    #setDefaultSymbols
   * @see    #setDefaultSumLimit
   * @see    #setDefaultNonZero
   * @see    #decimate
   */
  public void init ( int _defaultSymbols, long _defaultSumLimit, boolean _defaultNonZero, int _ctxId )
  {
    defaultSymbols  = _defaultSymbols;
    defaultSumLimit = _defaultSumLimit;
    defaultNonZero  = _defaultNonZero;
    ctxId           = _ctxId;
    ctx             = new CFTree( defaultSymbols, defaultSumLimit, defaultNonZero );
    contexts.clear();
    contexts.put( ctxId, ctx );
  }

  /**
   * Sets the new default number of symbols for new contexts.
   *
   * @param  i New default number of symbols.
   * @see    #init
   * @see    #setSymbols
   * @see    #getSymbols
   */
  public void setDefaultSymbols ( int i )
  {
    defaultSymbols = i;
  }

  /**
   * Changes the number of symbols in the current context.
   * Keeps previous frequencies (truncated or padded with the default value).
   *
   * @param  i New number of symbols.
   * @see    #setDefaultSymbols
   * @see    #getSymbols
   * @see    #decimate
   */
  public void setSymbols ( int i )
  {
    ctx.newSize( i );
    ctx.set( null );                      // temporary precaution
  }

  /**
   * Returns the number of symbols in the current context.
   *
   * @return Number of symbols.
   * @see    #setDefaultSymbols
   * @see    #setSymbols
   */
  public int getSymbols ()
  {
    return ctx.s;
  }

  /**
   * Sets the new default frequency sum limit for new contexts.
   *
   * @param  l New default frequency sum limit ({@code 0} for no bounds).
   * @see    #init
   * @see    #setSumLimit
   * @see    #getSumLimit
   * @see    #total
   */
  public void setDefaultSumLimit ( long l )
  {
    defaultSumLimit = l;
  }

  /**
   * Changes the frequency sum limit for the current context.
   *
   * @param  l New frequency sum limit ({@code 0} for no bounds).
   * @see    #init
   * @see    #setDefaultSumLimit
   * @see    #getSumLimit
   * @see    #decimate
   * @see    #total
   */
  public void setSumLimit ( long l )
  {
    ctx.newLimit( l );
  }

  /**
   * Returns the frequency sum limit for the current context.
   *
   * @return Frequency sum limit ({@code 0} for no bounds).
   * @see    #init
   * @see    #setDefaultSumLimit
   * @see    #setSumLimit
   * @see    #total
   */
  public long getSumLimit ()
  {
    return ctx.l;
  }

  /**
   * Set non-zero attribute: keep frequencies non-zero at all times?
   * Value for new contexts.
   *
   * @param  b Non-zero attribute for new contexts.
   * @see    #init
   * @see    #setNonZero
   * @see    #getNonZero
   */
  public void setDefaultNonZero ( boolean b )
  {
    defaultNonZero = b;
  }

  /**
   * Set non-zero attribute for the current context: keep frequencies non-zero at all times?
   *
   * @param  b Non-zero attribute for the current context.
   * @see    #init
   * @see    #setDefaultNonZero
   * @see    #getNonZero
   * @see    #decimate
   */
  public void setNonZero ( boolean b )
  {
    ctx.newNonZero( b );
  }

  /**
   * Read non-zero attribute for the current context
   *
   * @return Non-zero attribute of the current context.
   * @see    #init
   * @see    #setDefaultNonZero
   * @see    #setNonZero
   */
  public boolean getNonZero ()
  {
    return( ctx.min != 0 );
  }

  /**
   * Sets the given context as current.
   * For unknown contexts initializes parameters to default values.
   *
   * @param  _ctxId Identifier of the new context.
   * @return Identifier of the old context.
   * @see    #init
   * @see    #setDefaultSymbols
   * @see    #setDefaultSumLimit
   * @see    #getContext
   */
  public int setContext ( int _ctxId )
  {
    if ( _ctxId == ctxId ) return _ctxId;

    int old = ctxId;
    ctxId   = _ctxId;
    if ( (ctx = contexts.get( ctxId )) == null )
    {
      ctx = new CFTree( defaultSymbols, defaultSumLimit, defaultNonZero );  // create new context
      contexts.put( ctxId, ctx );
    }
    return old;
  }

  /**
   * Returns the current context.
   *
   * @return Id of the current context.
   * @see    #setContext
   */
  public int getContext ()
  {
    return ctxId;
  }

  /**
   * Divides each frequency in the current context by 2.
   * Called automatically when the total frequency sum exceeds the limit.
   *
   * @see    #init
   * @see    #setDefaultSumLimit
   * @see    #setSumLimit
   * @see    #getSumLimit
   * @see    #setDefaultNonZero
   * @see    #setNonZero
   * @see    #getNonZero
   * @see    #total
   */
  public void decimate ()
  {
    ctx.decimate();
  }

  /**
   * Sets frequencies of all symbols in the current context.
   * Can be used for context re-initialization ({@code null} parameter).
   *
   * @param  f Array of new frequencies.
   *           Can be {@code null} for context re-initialization.
   * @see    #set(int,long)
   * @see    #decimate
   */
  public void set ( long[] f )
  {
    ctx.set( f );
  }

  /**
   * Sets a single symbol's frequency.
   *
   * @param  symbol Symbol code.
   * @param  _freq New frequency of the symbol.
   * @see    #set(long[])
   * @see    #decimate
   */
  public void set ( int symbol, long _freq )
  {
    long f = ctx.getF( symbol );
    _freq -= f;
    if ( _freq < -f + ctx.min )
      _freq = -f + ctx.min;
    ctx.update( symbol, _freq );
  }

  /**
   * Increments frequencies of all symbols in the current context.
   *
   * @param  f Array of frequency increments.
   * @see    #inc(int,long)
   * @see    #inc(int)
   * @see    #decimate
   */
  public void inc ( long[] f )
  {
    ctx.inc( f );
  }

  /**
   * Increments a single symbol's frequency.
   *
   * @param  symbol Symbol code.
   * @param  _freq Frequency increment of the symbol.
   * @see    #inc(long[])
   * @see    #inc(int)
   * @see    #decimate
   */
  public void inc ( int symbol, long _freq )
  {
    long f = ctx.getF( symbol );
    if ( _freq < -f + ctx.min )
      _freq = -f + ctx.min;
    ctx.update( symbol, _freq );
  }

  /**
   * Increments a single symbol's frequency by {@code 1}.
   *
   * @param  s Symbol code.
   * @see    #inc(int,long)
   * @see    #inc(long[])
   * @see    #decimate
   */
  public void inc ( int s )
  {
    inc( s, 1 );
  }

  /**
   * Returns the frequency sum of all symbols in the current context.
   * Same as {@code left( getSymbols() )}, but might be more effective.
   *
   * @return The total frequency sum.
   * @see    #left
   * @see    #get
   * @see    #convert
   */
  public long total ()
  {
    return ctx.p[0];
  }

  /**
   * Returns the start of the symbol's cummulative frequency interval
   * (the frequency sum of all preceding symbols).
   *
   * @param  s Symbol code.
   * @return The sum of all preceding symbol frequencies.
   * @see    #total
   * @see    #get
   * @see    #convert
   */
  public long left ( int s )
  {
    return ctx.getCf( s );
  }

  /**
   * Returns the frequency of the given symbol.
   * Same as {@code left( _symbol + 1 ) - left( _symbol )}.
   *
   * @param  s Symbol code.
   * @return The frequency of that symbol.
   * @see    #total
   * @see    #left
   * @see    #convert
   */
  public long get ( int s )
  {
    return ctx.getF( s );
  }

  /**
   * Converts a frequency sum to the corresponding symbol.
   *
   * @param  cf Frequency sum of the unknown symbol.
   * @return The symbol which has {@code _sum} in it's cummulative frequency interval.
   * @see    #total
   * @see    #left
   * @see    #get
   */
  public int convert ( long cf )
  {
    return ctx.getI( cf );
  }

  //==========================================================================
  //  Support:

  protected void printStatistics ()
  {
    //if ( IntMap.findCounter > 0L )
    //  LogFile.log( "IntMap collision statistics: lookups = " + IntMap.findCounter +
    //               ", collisions = " + IntMap.collisionCounter +
    //               " (" + Formula.percent( IntMap.collisionCounter, IntMap.findCounter ) + "%)" );

    LogFile.log( "IntervalArray statistics:" );
    Iterator< IntTuple<CFTree> > it = contexts.tupleIterator();
    while ( it.hasNext() )
    {
      IntTuple<CFTree> tu = it.next();
      LogFile.log( "Context: " + tu.key + ", sym = " + tu.value.s + ", sum = " + tu.value.p[0] +
                   ", p[1] = " + tu.value.p[1] );
    }
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null ) return;

      // STATISTICS:
    if ( key.compareTo(STATISTICS) == 0 )
    {
      if ( LogFile.debugging )
        printStatistics();
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    return null;
  }

  //==========================================================================
  //  Module registration

  /** Object name. */
  private final static String NAME = "IntervalArray";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "EntropyHistogramTo";

  /** Object description. */
  private final static String DESCRIPTION = "Cummulative frequency interval tree.";

  /** Object category. */
  protected final static String CATEGORY = C_DATA + '.' + C_COMPRESSION + ".arithmetic";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_OUTPUT, IFACE+"EntropyHistogram" );
        // Property:
      t.propBegin( STATISTICS, TYPE_BOOLEAN, "Print object statistics", false );
      t.propEnd();
     }
   return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
