package cz.cuni.jagrlib.piece;

import java.util.*;
import java.io.*;
import javax.imageio.*;
import javax.imageio.stream.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * JFIF file format (persistence of 2D raster graphics data).
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.13
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/JFIFFileFormat.java">JFIFFileFormat.java</a>
 */
public class JFIFFileFormat extends DefaultRasterFileFormat
{
  //==========================================================================
  //  Data:

  /** Encoding-quality property (Double or Float): quality for JPEG encoder. */
  public final static String QUALITY = "Quality";

  /**
   * Actual quality for JPEG encoder.
   * Value should be set between {@code 0.0f} and {@code 100.0f}.
   */
  protected float quality = 80.0f;

  //==========================================================================
  //  Support:

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonSave ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    OutputStream os = stream.getOutputStream();
    if ( os == null )
      throw new FileNotFoundException();

    ImageOutputStream ios = null;
    try
    {
      ios = ImageIO.createImageOutputStream( os );
    }
    catch ( IOException e )
    {
      throw new IIOException( "Can't create output stream!", e );
    }

    ImageWriter writer = null;
    ImageTypeSpecifier type = ImageTypeSpecifier.createFromRenderedImage( g.getBufferedImage() );
    Iterator< ImageWriter > it = ImageIO.getImageWriters( type, "JPG" );
    if ( it.hasNext() )
      writer = it.next();

    if ( writer == null )
      throw new IIOException( "Can't find JPEG writer!" );

    writer.setOutput( ios );
    ImageWriteParam par = writer.getDefaultWriteParam();
    par.setCompressionMode( ImageWriteParam.MODE_EXPLICIT );
    par.setCompressionQuality( 0.01f * quality );

    try
    {
      writer.write( null, new IIOImage( g.getBufferedImage(), null, null ), par );
    }
    finally
    {
      writer.dispose();
      ios.flush();
    }
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return 11;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type ({@code 0.0} .. no way,
   *         {@code 0.5} .. maybe, {@code 0.9} .. almost sure, {@code 1.0} ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int len = (header == null) ? 0 : header.length;
    if ( len > 11 ) len = 11;
    if ( len < 4 )                        // not enough bytes in header[] => look at file-name
    {
      if ( fileName == null || fileName.length() < 5 ) return 0.0;
      int pos = fileName.lastIndexOf( '.' );
      if ( pos < 0 ) return 0.0;
      String suffix = fileName.substring( pos );
      if ( suffix.compareToIgnoreCase( "jpg" )  == 0 ||
           suffix.compareToIgnoreCase( "jpeg" ) == 0 ||
           suffix.compareToIgnoreCase( "jfif" ) == 0 ) return 0.9;
      return 0.0;
    }

    byte[] h = new byte[]                 // JFIF header (0xFF, SOI, 0xFF, APP0, ?, ?, "JFIF\0")
    {
      (byte)0xFF, (byte)0xD8, (byte)0xFF, (byte)0xE0,
      0x00, 0x00, 0x4A, 0x46, 0x49, 0x46, 0x00
    };
    int i;
    for ( i = 0; i < 4; i++ )
      if ( header[i] != h[i] ) return 0.0;
    if ( len <= 6 ) return 0.95;
    for ( i = 6; i < len; i++ )
      if ( header[i] != h[i] ) return 0.1;
    return 1.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.jpg", "*.jpeg",
   *         "*.jfif"]) or {@code null}.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.jpg", "*.jpeg", "*.jfif"
    };
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // QUALITY:
    if ( key.compareTo(QUALITY) == 0 )
    {
      quality = floatProperty( value, quality, 0.0f, 100.0f );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(DOUBLE_STREAM) == 0 )
      return false;
    if ( key.compareTo(QUALITY) == 0 )
      return quality;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "JFIF file format";

  /** Object description. */
  private final static String DESCRIPTION = "JFIF (JPEG File Interchange Format) filter, " +
                                            "uses javax.imageio package.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"DataFileFormat" );
      t.newOutputPlug(    PL_RASTER, IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_STREAM, IFACE+"BitStream" );
        // Property:
      t.propBegin( DOUBLE_STREAM, TYPE_BOOLEAN, "Use floating-point data stream?", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( QUALITY, TYPE_FLOAT, "Compression quality in %", true );
        t.propBounds( 0.0f, 100.0f );
        t.propDefault( 80.0f );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
