package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * JPEG-like block quantizer.
 *
 * @version 0.26 $Rev: 472 $ $Date: 2007-05-31 09:33:37 +0200 (čt, 31 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/JPEGQuantizer.java">JPEGQuantizer.java</a>
 */
public class JPEGQuantizer extends DefaultBlockQuantizer
{
  //==========================================================================
  //  Data:

  /** Maximum number of quantization tables. */
  public final static int MAX_TABLES = 8;

  /** Quantization tables. */
  protected int[][][] q = new int[ MAX_TABLES ][][];

  /** Obey strict Baseline JPEG rules (8-bit quantizer)? */
  protected boolean strictBaselineJPEG = false;

  //==========================================================================
  //  Support:

  /** Upper bound for q-value - in strict Baseline JPEG system (8-bit quantizer). */
  protected final static int MAX_QUANT = 255;

  protected final static int OUTSIDE_Y = 128;

  /** JPEG recommendation for Y channel and 50% quality (annex K.1). */
  protected final static int[][] Y50 =
  {
    {   16,  11,  10,  16,  24,  40,  51,  61 },
    {   12,  12,  14,  19,  26,  58,  60,  55 },
    {   14,  13,  16,  24,  40,  57,  69,  56 },
    {   14,  17,  22,  29,  51,  87,  80,  62 },
    {   18,  22,  37,  56,  68, 109, 103,  77 },
    {   24,  35,  55,  64,  81, 104, 113,  92 },
    {   49,  64,  78,  87, 103, 121, 120, 101 },
    {   72,  92,  95,  98, 112, 100, 103,  99 }
  };

  protected final static int OUTSIDE_CB =  99;

  /** JPEG recommendation for Cb, Cr channels and 50% quality (annex K.2). */
  protected final static int[][] CB50 =
  {
    {   17,  18,  24,  47,  99,  99,  99,  99 },
    {   18,  21,  26,  66,  99,  99,  99,  99 },
    {   24,  26,  56,  99,  99,  99,  99,  99 },
    {   47,  66,  99,  99,  99,  99,  99,  99 },
    {   99,  99,  99,  99,  99,  99,  99,  99 },
    {   99,  99,  99,  99,  99,  99,  99,  99 },
    {   99,  99,  99,  99,  99,  99,  99,  99 },
    {   99,  99,  99,  99,  99,  99,  99,  99 }
  };

  protected final static int fiftyPercentY ( int x, int y )
  {
    if ( x >= 8 || y >= 8 )
      return OUTSIDE_Y;

    return Y50[y][x];
  }

  protected final int fiftyPercentCbCr ( int x, int y )
  {
    if ( x >= 8 || y >= 8 )
      return OUTSIDE_CB;

    return CB50[y][x];
  }

  //==========================================================================
  //  Interface BlockQuantizer:

  /**
   * Set actual quantizer variant (context).
   *
   * @param  variant Variant number (from {@code 0}).
   * @return Old variant number.
   * @see    #setBlockSize
   * @see    #setQuality
   * @see    #setData
   */
  @Override
  public int setVariant ( int variant )
  {
    int old      = this.variant;
    this.variant = Formula.clamp( variant, 0, MAX_TABLES-1 );
    return old;
  }

  /**
   * Sets actual block size.
   * Can be used both for 1D and 2D quantization.

   * @param  width Horizontal block size (1D block size).
   * @param  height Vertical block size ({@code 1} for 1D).
   * @see    #setVariant
   * @see    #setQuality
   * @see    #setData
   */
  @Override
  public void setBlockSize ( int width, int height )
  {
    if ( q[variant]           == null ||
         q[variant].length    != height ||
         q[variant][0].length != width )
      q[variant] = new int[ height ][ width ];
  }

  /**
   * Sets actual quality coefficient.
   * Value range for the quality: {@code 0.0f} to {@code 1.0f}.
   *
   * @param  quality New quality coefficient.
   * @see    #setVariant
   * @see    #setBlockSize
   * @see    #setData
   */
  @Override
  public void setQuality ( float quality )
  {
    if ( q[variant] == null ) return;

    int height = q[variant].length;
    int width  = q[variant][0].length;
    int i, j;

      // multiplier value based on quality (IJG recommendations):
    if ( quality < 0.01f )
      quality = 0.01f;
    else
    if ( quality > 1.0f )
      quality = 1.0f;

    if ( quality < 0.5f )
      quality = 0.5f / quality;
    else
      quality = 2.0f - quality - quality;

      // upper bound:
    int max = strictBaselineJPEG ? 255 : 32767;

    boolean y = (variant & 1) == 0;

    for ( j = 0; j < height; j++ )
      for ( i = 0; i < width; i++ )
        q[variant][j][i] = Formula.clamp( Math.round( (y ? fiftyPercentY( i, j ) : fiftyPercentCbCr( i, j )) * quality ), 1, max );
  }

  /**
   * Sets quantization table for the current variant.
   *
   * @param  qtable New quantization table.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @see    #setVariant
   * @see    #setQuality
   * @see    #setBlockSize
   */
  @Override
  public void setData ( int[] qtable )
  {
    int width, height;
    if ( qtable     == null ||
         q[variant] == null ||
         qtable.length < (height = q[variant].length) *
                         (width  = q[variant][0].length) )
      return;

      // upper bound:
    int max = strictBaselineJPEG ? 255 : 32767;

    int i, j, k;
    for ( j = k = 0; j < height; j++ )
      for ( i = 0; i < width; i++ )
        q[variant][j][i] = Formula.clamp( qtable[k++], 1, max );
  }

  /**
   * Quantization (floating-point form).
   *
   * @param  coef Source data (coefficients). Size = array size.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @param  quant Pre-allocated result array or {@code null} if
   *         it's size needs to be determined.
   * @return Result array size (even if the quantization was not performed).
   * @see    #quantize(int[],int[])
   * @see    #dequantize(int[],double[])
   */
  @Override
  public int quantize ( double[] coef, int[] quant )
  {
    if ( q[variant] == null )
      return 0;

    int height = q[variant].length;
    int width  = q[variant][0].length;
    int len    = width * height;

    if ( coef  == null ||
         quant == null )
      return len;

    int i, j, k;
    for ( j = k = 0; j < height; j++ )
      for ( i = 0; i < width; i++, k++ )
        quant[k] = Formula.round( coef[k] / q[variant][j][i] );

    return len;
  }

  /**
   * Dequantization = reconstruction (floating-point form).
   *
   * @param  quant Source data (quantized coefficients). Size = array size.
   *         Should have the correct size and layout (scanline top-first, left-first).
   * @param  coef Pre-allocated coefficient array or {@code null} if
   *         it's size needs to be determined.
   * @return Coefficient array size (even if the dequantization was not performed).
   * @see    #dequantize(int[],int[])
   * @see    #quantize(double[],int[])
   */
  @Override
  public int dequantize ( int[] quant, double[] coef )
  {
    if ( q[variant] == null )
      return 0;

    int height = q[variant].length;
    int width  = q[variant][0].length;
    int len    = width * height;

    if ( coef  == null ||
         quant == null )
      return len;

    int i, j, k;
    for ( j = k = 0; j < height; j++ )
      for ( i = 0; i < width; i++, k++ )
        coef[k] = quant[k] * (double)q[variant][j][i];

    return len;
  }

  //==========================================================================
  //  Interface PersistentParamsBinary:

  /**
   * Pack instance parameters into binary format.
   *
   * @return Data array to be stored into binary stream. Can be {@code null}.
   *         The array can be shared by all method calls.
   * @see    #loadData
   */
  @Override
  public byte[] storeData ()
  {
    int count  = 2;
    int tables = 0;
    int j, x, y;
    int width, height;

    for ( j = 0; j < q.length; j++ )
      if ( q[j] != null )                 // found the j-th table
      {
        tables++;
        count += 3;                       // table header (index, height, width)
        height = q[j].length;
        width  = q[j][0].length;
        count += 2 * height * width;
      }

    byte[] arr = new byte[ count ];
    int    ptr = 0;

    arr[ptr++] = (byte)( tables  & 0xff ); // number of tables
    arr[ptr++] = (byte)( variant & 0xff ); // actual quantizer variant

    for ( j = 0; j < q.length; j++ )
      if ( q[j] != null )
      {
        arr[ptr++] = (byte)( j      & 0xff ); // table index
        height     = q[j].length;
        width      = q[j][0].length;
        arr[ptr++] = (byte)( height & 0xff ); // table height
        arr[ptr++] = (byte)( width  & 0xff ); // table width

          // write the j-th table data:
        for ( y = 0; y < height; y++ )
          for ( x = 0; x < width; x++ )
            ptr = Formula.storeUnsigned16( arr, ptr, q[j][y][x] );
      }

    return arr;
  }

  /**
   * Loads instance parameters from (previously stored) binary data.
   *
   * @param  data Binary data to be interpreted. Can be {@code null}.
   * @return {@code true} if successful.
   * @see    #storeData
   */
  @Override
  public boolean loadData ( byte[] data )
  {
    if ( data == null ) return false;

    int ptr    = 0;
    int tables = data[ptr++];
    if ( tables > MAX_TABLES ) return false;
    variant    = data[ptr++];
    if ( variant >= MAX_TABLES ) return false;
    int i, j, x, y;
    int width, height;

    q = new int[ MAX_TABLES ][][];

    for ( i = 0; i++ < tables; )
    {
      j      = data[ptr++];               // j-th table will be restored..
      if ( j >= MAX_TABLES ) return false;
      height = data[ptr++];
      width  = data[ptr++];
      q[j]   = new int[ height ][ width ];

      for ( y = 0; y < height; y++ )
        for ( x = 0; x < width; x++, ptr += 2 )
          q[j][y][x] = Formula.loadUnsigned16( data, ptr );
    }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String STRICT_JPEG  = "Strict JPEG";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // STRICT_JPEG:
    if ( key.compareTo(STRICT_JPEG) == 0 )
    {
      strictBaselineJPEG = booleanProperty( value, strictBaselineJPEG );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(STRICT_JPEG) == 0 )
      return strictBaselineJPEG;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "JPEG quantizer";

  /** Object description. */
  private final static String DESCRIPTION = "Block quantizer from JPEG1 specification.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"BlockQuantizer" );
      // Property:
      t.propBegin( STRICT_JPEG, TYPE_BOOLEAN, "Strict baseline JPEG quantization (8-bit)", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
