package cz.cuni.jagrlib.piece;

import java.io.*;
import java.util.zip.GZIPInputStream;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Bit-stream object implemented with help of <b>java.io</b> streams ({@link InputStream},
 * {@link OutputStream}).
 *
 * @version 0.26 $Rev: 551 $ $Date: 2007-10-10 19:58:19 +0200 (st, 10 X 2007) $ $Author: pepca $
 * @since   0.13
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/JavaBitStream.java">JavaBitStream.java</a>
 */
public class JavaBitStream extends Piece implements BitStream
{
  //==========================================================================
  //  Data:

  /** Stream is intended for output. */
  protected boolean isOutput = true;

  /** Data stream name (file-name etc.). */
  protected String streamName = DEFAULT_STREAM_NAME;

  /** Current input stream. */
  protected InputStream inputStream;

  /** Current output stream. */
  protected OutputStream outputStream;

  /** Bit buffer. */
  protected long buffer;

  /** Actual buffer length in bits. */
  protected int bufferBits;

  /** Actual stream-position in bits. */
  protected long position;

  /** Temporary array for faster I/O operations. */
  protected byte[] array;

  //==========================================================================
  //  Support:

  /**
   * Writes all complete bytes from {@link #buffer} to the output stream.
   *
   * @throws IOException
   * @see    #flushBuffer
   * @see    #buffer
   */
  protected final void writeBuffer ()
  throws IOException
  {
    if ( (bufferBits & -8) == 0 ) return;

    int len = bufferBits >> 3;          // number of bytes to write
    if ( array == null || array.length != len )
      array = new byte[ len ];

    for ( int i = 0; i < len; )
    {
      array[i++] = (byte)(buffer & 0xff);
      buffer >>>= 8;
      bufferBits -= 8;
    }
    outputStream.write( array );
    position += (len << 3);
  }

  /**
   * Flushes out the complete {@link #buffer}. Incomplete byte is padded with zero bits.
   *
   * @throws IOException
   * @see    #writeBuffer
   * @see    #buffer
   */
  protected void flushBuffer ()
  throws IOException
  {
    if ( outputStream != null )
    {
      bufferBits = (bufferBits + 7) & -8;
      writeBuffer();
    }
  }

  /**
   * Reads maximum amount of input bits into {@link #buffer}.
   * Assumes {@link #bufferBits}{@code == 0}.
   *
   * @throws IOException
   * @see    #writeBuffer
   * @see    #buffer
   */
  protected final void readBuffer ()
  throws IOException
  {
    if ( array == null || array.length != 8 )
      array = new byte[ 8 ];

    int len = inputStream.read( array );
    if ( len <= 0 ) return;
    position += (len << 3);
    for ( int i = 0; i < len; )
    {
      buffer |= (((long)array[i++] & 0xffL) << bufferBits);
      bufferBits += 8;
    }
  }

  //==========================================================================
  //  Interface BitStream:

  /**
   * Stream is able to read data.
   * If {@code true}, stream can be opened automatically in the init-phase.
   */
  public boolean canRead ()
  {
    return !isOutput;
  }

  /**
   * Stream initialization.
   *
   * @param  output Open for output?
   * @param  name Stream name (either stream identification e.g. file-name). {@code null}
   *              for implicit bit-stream ({@link #reOpen}).
   * @param  detail Further implementation-specific information.
   * @throws IOException
   * @see    #reOpen
   */
  public void open ( boolean output, String name, String detail )
  throws IOException
  {
    if ( inputStream  != null ||
         outputStream != null ) close();

      // stream name:
    if ( name != null ) streamName = name;
    if ( streamName == null ) streamName = DEFAULT_STREAM_NAME;

      // stream type:
    isOutput = output;

      // open the stream:
    if ( isOutput )                       // FileOutputStream
      outputStream = new FileOutputStream( streamName );
    else                                  // FileInputStream | GZIPInputStream
    {
      inputStream = new FileInputStream( streamName );

      final int extPos = streamName.lastIndexOf( '.' );
      if ( extPos >= 0 &&
           streamName.substring( extPos ).equalsIgnoreCase( ".gz" ) )
        inputStream = new GZIPInputStream( inputStream );
    }

    buffer      = 0L;
    bufferBits  = 0;
    position    = 0L;
  }

  /**
   * Stream re-initialization.
   *
   * @param  output Open for output?
   * @param  detail Further implementation-specific information.
   * @throws IOException
   * @see    #open
   */
  public void reOpen ( boolean output, String detail )
  throws IOException
  {
    open( output, null, detail );
  }

  /**
   * Retrieves the associated output stream (non-mandatory).
   *
   * @return The associated {@link OutputStream} instance or {@code null} if
   *         not implemented.
   * @see    #getInputStream
   */
  public OutputStream getOutputStream ()
  {
    return outputStream;
  }

  /**
   * Retrieves the associated input stream (non-mandatory).
   *
   * @return The associated {@link InputStream} instance or {@code null} if
   *         not implemented.
   * @see    #getOutputStream
   */
  public InputStream getInputStream ()
  {
    return inputStream;
  }

  /**
   * Stream flush.
   *
   * @throws IOException
   * @see    #open
   * @see    #close
   */
  public void flush ()
  throws IOException
  {
    if ( outputStream != null )
    {
      flushBuffer();
      outputStream.flush();
    }
  }

  /**
   * Stream close.
   * Eventually flushes all written data and disconnects the stream.
   *
   * @throws IOException
   * @see    #open
   * @see    #reOpen
   */
  public void close ()
  throws IOException
  {
    if ( inputStream != null )            // close the associated input stream
    {
      inputStream.close();
      inputStream = null;
    }

    if ( outputStream != null )           // close the associated output stream
    {
      flushBuffer();
      outputStream.close();
      outputStream = null;
    }
  }

  /**
   * Actual stream position in bits.
   * After the output stream is closed, returns maximal written position.
   *
   * @return Actual (maximal) stream position in bits.
   * @throws IOException
   * @see    #available
   * @see    #seek
   */
  public long position ()
  throws IOException
  {
    if ( inputStream != null )            // input stream
      return( position - bufferBits );

    if ( outputStream != null )           // output stream
      return( position + bufferBits );

    return position;
  }

  /**
   * How many bits are available to read (or skip).
   *
   * @return Number of bits available to read (or skip).
   * @throws IOException
   * @see    #position
   * @see    #seek
   */
  public long available ()
  throws IOException
  {
    if ( inputStream != null )            // input stream only
      return( (((long)inputStream.available()) << 3) + bufferBits );

    return 0L;
  }

  /**
   * Sets actual position of the stream in bits.
   * Non mandatory.
   *
   * @param  pos Required stream position in bits.
   * @return New stream position in bits (or {@code -1L} if not implemented).
   * @throws IOException
   * @see    #position
   */
  public long seek ( long pos )
  {
    return -1L;
  }

  /**
   * Writes a bunch of bits into the stream.
   *
   * @param  bits Bits to be written (LSB first).
   * @param  length Number of bits to write (negative lengths are ignored).
   * @throws IOException
   * @see    #writeBytes
   * @see    #read
   */
  public void write ( long bits, int length )
  throws IOException
  {
    if ( outputStream == null || length <= 0 ) return;

    if ( length > 64 ) length = 64;
    if ( bufferBits + length >= 64 )      // buffer will overflow => flush it first
    {
      buffer    |= (bits << bufferBits);
      int first  = 64 - bufferBits;
      bufferBits = 64;
      writeBuffer();
      if ( (length -= first) == 0 ) return;
      bits    >>>= first;
    }

    buffer |= ((bits & ((1L << length) - 1L)) << bufferBits);     // not-overflowed bits
    bufferBits += length;
  }

  /**
   * Writes a bunch of bytes into the stream.
   *
   * @param  data Bytes to be written (each byte is written in "LSB first" order).
   * @param  start First index to write. Must be less than {@code data.length}.
   * @param  length Number of bytes to write ({@code start + length > data.length} is ignored).
   * @throws IOException
   * @see    #write
   * @see    #readBytes
   */
  public void writeBytes ( byte[] data, int start, int length )
  throws IOException
  {
    if ( outputStream == null || length <= 0 ||
         data == null || start >= data.length ) return;

    if ( start + length > data.length ) length = data.length - start;
    do
      write( data[start++], 8 );
    while ( --length > 0 );
  }

  /**
   * Reads a bunch of bits from the stream.
   *
   * @param  length Number of bits to read (negative lengths are ignored).
   * @return Read bits (LSB first).
   * @throws IOException
   * @see    #readBytes
   * @see    #write
   */
  public long read ( int length )
  throws IOException
  {
    if ( inputStream == null || length <= 0 ) return 0L;

    if ( length > 64 ) length = 64;
    long out    = 0L;
    int outBits = 0;
    int first   = length;
    if ( first > bufferBits ) first = bufferBits;
    if ( first > 0 )                      // read bits from buffer first
    {
      out = buffer & ((1L << first) - 1L);
      buffer >>>= first;
      bufferBits -= first;
      if ( (length -= first) == 0 ) return out;
      outBits = first;
    }

      // buffer is empty & I need another "length" bits
    readBuffer();
    out |= ((buffer & ((1L << length) - 1L)) << outBits);
    buffer >>>= length;
    bufferBits -= length;
    return out;
  }

  /**
   * Reads a bunch of bytes from the stream.
   *
   * @param  data Array holding the read bytes (each byte is read in "LSB first" order).
   * @param  start First index to read. Must be less than {@code data.length}.
   * @param  length Number of bytes to read ({@code start + length > data.length} is ignored).
   * @throws IOException
   * @see    #read
   * @see    #writeBytes
   */
  public void readBytes ( byte[] data, int start, int length )
  throws IOException
  {
    if ( inputStream == null || length <= 0 ||
         data == null || start >= data.length ) return;

    if ( start + length > data.length ) length = data.length - start;
    do
      data[start++] = (byte)read( 8 );
    while ( --length > 0 );
  }

  /**
   * Skips the given number of bits.
   *
   * @param  length Number of bits to skip (negative lengths are ignored).
   * @return Number of bits actually skipped.
   * @throws IOException
   * @see    #read
   */
  public long skip ( long length )
  throws IOException
  {
    if ( length <= 0 ) return 0L;

    long all = length;
    int len;
    if ( inputStream != null )            // skip input bits
    {
      len = bufferBits;
      if ( len > length ) len = (int)length;
      if ( len > 0 )                      // skip buffered bits first
      {
        buffer >>>= len;
        bufferBits -= len;
        if ( (length -= len) == 0L ) return all;
      }

      long skipped = inputStream.skip(length >> 3);
      position += (skipped <<= 3);        // number of complete bytes been skipped
      if ( (length -= skipped) == 0 ) return all;

      readBuffer();
      len = (int)length;
      if ( len > bufferBits ) len = bufferBits;
      if ( len > 0 )
      {
        buffer >>>= len;
        bufferBits -= len;
        length -= len;                    // bits left
      }
      return( all - length );
    }
    if ( outputStream != null )           // write "length" zero bits
    {
      do
      {
        len = 64;
        if ( len > length ) len = (int)length;
        write( 0L, len );
      } while ( (length -= len) > 0L );
      return all;
    }

    return 0L;
  }

  //==========================================================================
  //  Interface Property:

  /** Stream-type property (boolean): intended for output. */
  public final static String STREAM_OUTPUT = "Output";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // STREAM_OUTPUT:
    if ( key.compareTo(STREAM_OUTPUT) == 0 )
    {
      isOutput = booleanProperty( value, isOutput );
      return;
    }
      // STREAM_NAME:
    if ( key.compareTo(STREAM_NAME) == 0 )
    {
      streamName = stringProperty( value, streamName );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(STREAM_OUTPUT) == 0 )
      return isOutput;
    if ( key.compareTo(STREAM_NAME) == 0 )
      return streamName;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Bit-streams using Java.io";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "PropertyAndBitStream";

  /** Object description. */
  private final static String DESCRIPTION = "Bit-streams using builtin Java classes InputStream " +
                                            "and OutputStream. Optional Zlib decompression.";

  /** Object category. */
  protected final static String CATEGORY = C_IO + '.' + C_BINARY;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT,    IFACE+"BitStream" );
      t.newInputPlug( PL_PROPERTY, IFACE+"Property" );
        // Property:
      t.propBegin( STREAM_OUTPUT, TYPE_BOOLEAN, "Is intended for output?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( STREAM_NAME, TYPE_STRING, "File name", true );
        t.propDefault( "file-name" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
