package cz.cuni.jagrlib.piece;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Text drawing using {@link java.awt.Graphics} implementation.
 *
 * @version 0.25 $Rev: 385 $ $Date: 2007-01-07 01:47:21 +0100 (ne, 07 I 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/JavaText.java">JavaText.java</a>
 */
public class JavaText extends Piece implements TextRender
{
  //==========================================================================
  //  Data:

  /** Font cache size granularity. */
  protected final static int CACHE_GRANUL = 16;

  /** Multiplier {@code 1 / 255}. */
  protected final static double INV255 = 1.0 / 255.0;

  /** Font rendering surface. */
  protected BufferedImage surface;

  /** Graphics associated with the current surface. */
  protected Graphics2D graphics;

  /** Actual surface width in pixels (granularity {@link #CACHE_GRANUL}). */
  protected int surfaceWid;

  /** Actual surface height in pixels (granularity {@link #CACHE_GRANUL}). */
  protected int surfaceHei;

  {
    surface    = new BufferedImage( CACHE_GRANUL, CACHE_GRANUL, BufferedImage.TYPE_BYTE_GRAY );
    graphics   = (Graphics2D)surface.getGraphics();
    graphics.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON );
    surfaceWid =
    surfaceHei = CACHE_GRANUL;
  }

  /** Actual font orientation. */
  protected double orientation = 0.0;

  /** Accelerator: sin( orientation ). */
  protected double sino = 0.0;

  /** Accelerator: cos( orientation ). */
  protected double coso = 1.0;

  /**
   * Use antialiasing?
   * Actual status, depends either on output module type or on {@link #ANTI_MODE} property.
   */
  protected boolean actAntialiasing = false;

  /** Change anti-aliasing mode before next text rendering? */
  protected boolean chAntialiasing = true;

  /** Antialiasing logic. */
  protected int antialiasing = ANTI_DEPEND;

  /** Font slots. Active font is in {@code fonts[0]}. */
  protected Font[] fonts = new Font[ MAX_SLOT ];

  //==========================================================================
  //  Support:

  /** Asserts enough space for the given text bounds (and clears the surface). */
  protected void prepareSurface ( int width, int height )
  {
    if ( width  > surfaceWid ||
         height > surfaceHei )
    {                                     // re-allocate the surface
      if ( width > surfaceWid )
        surfaceWid = Formula.granularity( width,  CACHE_GRANUL );
      if ( height > surfaceHei )
        surfaceHei = Formula.granularity( height, CACHE_GRANUL );
      surface  = new BufferedImage( surfaceWid, surfaceHei, BufferedImage.TYPE_BYTE_GRAY );
      graphics = (Graphics2D)surface.getGraphics();
      if ( actAntialiasing )
        graphics.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON );
      chAntialiasing = false;
    }
    else                                  // just clear the field
    {
      if ( chAntialiasing )
      {
        chAntialiasing = false;
        graphics.setRenderingHint( RenderingHints.KEY_TEXT_ANTIALIASING,
                                   actAntialiasing ? RenderingHints.VALUE_TEXT_ANTIALIAS_ON :
                                   RenderingHints.VALUE_TEXT_ANTIALIAS_OFF );
      }
      graphics.setColor( Color.BLACK );
      graphics.fillRect( 0, 0, width, height );
    }
  }

  //==========================================================================
  //  Interface TextRender:

  /**
   * Creates a new font and places it in <b>slot A</b> (actual, working font).
   *
   * @param  name Font name.
   * @param  style Font style - used if exact font match was not successful
   *               (bitwise OR of {@code TextRender.FONT_*} constants).
   * @param  size Font size (distance between two regularly spaced text lines).
   * @return Result code (one of {@code TextRender.CREATE_*} constants).
   * @see    #FONT_BOLD
   * @see    #CREATE_EXACT
   * @see    cz.cuni.jagrlib.VCommand.CreateFont
   */
  public int createFont ( String name, int style, double size )
  {
    if ( name == null ) return CREATE_FAIL;

    int javaStyle = 0;
    if ( (style & FONT_BOLD) > 0 )
      javaStyle |= Font.BOLD;
    if ( (style & FONT_ITALIC) > 0 )
      javaStyle |= Font.ITALIC;

    fonts[0] = new Font( name, javaStyle, Formula.round( size ) );

    LogFile.debug( "createFont( " + name + ", " + style + ", " + Formula.round( size ) + " ) = " +
                   fonts[0].getFontName() + ", " + fonts[0].getSize2D() );

    return( fonts[0].getFontName().equals( name ) ? CREATE_EXACT : CREATE_APPROX );
  }

  /**
   * Changes the actual font's orientation.
   *
   * @param  angle New font orientation (in degrees, counterclockwise).
   * @return The old orientation.
   * @see    cz.cuni.jagrlib.VCommand.FontOrientation
   */
  public double fontOrientation ( double angle )
  {
    double old  = Math.toDegrees( orientation );
    orientation = Math.toRadians( angle );
    sino        = Math.sin( orientation );
    coso        = Math.cos( orientation );
    return old;
  }

  /**
   * Assigns the actual font (<b>slot A</b>) from the given <b>slot</b>.
   *
   * @param  slot Slot number (between {@code 0} and {@code MAX_SLOT - 1}).
   * @see    #MAX_SLOT
   * @see    cz.cuni.jagrlib.VCommand.GetFont
   */
  public void getFont ( int slot )
  {
    if ( slot <= 0 ||
         slot > MAX_SLOT ) return;

    if ( fonts[slot] != null )
      fonts[0] = fonts[slot];
  }

  /**
   * Puts the actual font (<b>slot A</b>) into the given <b>slot</b>.
   *
   * @param  slot Slot number (between {@code 0} and {@code MAX_SLOT - 1}).
   * @see    #MAX_SLOT
   * @see    cz.cuni.jagrlib.VCommand.PutFont
   */
  public void putFont ( int slot )
  {
    if ( slot <= 0 ||
         slot > MAX_SLOT ) return;

    if ( fonts[0] != null )
      fonts[slot] = fonts[0];
  }

  /**
   * Returns properties (flags) of the actual font (<b>slot A</b>).
   *
   * @return Bitwise OR of {@code PROP_*} constants.
   * @see    #PROP_STROKE
   */
  public int fontProperties ()
  {
    return PROP_FILL;
  }

  /**
   * Generates path of the given text string and <u>appends</u> it to the current path.
   * Uses the <b>current point</b> as starting reference point (<b>current point</b>
   * itself will be updated by this command).
   *
   * <p>Does nothing if the actual font has neither {@link #PROP_STROKE} nor
   * {@link #PROP_FILL} capability.
   *
   * @param  text The string to be drawn.
   * @see    #textExtent
   * @see    #drawText
   * @see    cz.cuni.jagrlib.VCommand.MoveTo
   * @see    VectorGraphics#moveTo
   * @see    cz.cuni.jagrlib.VCommand.TextPath
   */
  public void textPath ( String text )
  {
  }

  /**
   * Draws the given text. Starting reference point is explicitly specified.
   * If {@code ext} parameter is not {@code null}, the routine fills
   * the given {@link cz.cuni.jagrlib.iface.TextRender.TextExtent TextExtent} object.
   *
   * <p>Works with both vector and bitmap fonts.
   *
   * @param  x0 X coordinate of the starting reference point.
   * @param  y0 Y coordinate of the starting reference point.
   * @param  text The string to be drawn.
   * @param  ext The {@link cz.cuni.jagrlib.iface.TextRender.TextExtent TextExtent} object to be filled (or {@code null}).
   * @see    cz.cuni.jagrlib.iface.TextRender.TextExtent
   * @see    #textExtent
   * @see    cz.cuni.jagrlib.VCommand.DrawText
   */
  public void drawText ( double x0, double y0, String text, TextExtent ext )
  {
    if ( fonts[0] == null ||
         graphics == null ) return;

    final BitMask   bm = (BitMask)  getInterface( PL_OUTPUT,    IFACE+"BitMask" );
    final AlphaMask am = (AlphaMask)getInterface( PL_ALPHAMASK, IFACE+"AlphaMask" );

    if ( bm == null &&
         am == null ) return;

    if ( chAntialiasing )
      switch ( antialiasing )
      {
        case ANTI_DEPEND:
          actAntialiasing = (am != null);
          break;
        case ANTI_OFF:
          actAntialiasing = false;
          break;
        default:
          actAntialiasing = true;
      }

    FontMetrics fm       = graphics.getFontMetrics( fonts[0] );
    Rectangle2D bounds2D = fm.getStringBounds( text, graphics );
    double minX  = bounds2D.getMinX();
    double minY  = bounds2D.getMinY();
    double maxX  = bounds2D.getMaxX();
    double maxY  = bounds2D.getMaxY();
    if ( ext != null )
    {
      ext.x0  = x0;
      ext.y0  = y0;
      double advance = fm.stringWidth( text );
      ext.x1  = x0 + advance * coso;
      ext.y1  = y0 - advance * sino;
      ext.ulx = minX;
      ext.uly = minY;
      ext.lrx = maxX;
      ext.lry = maxY;
    }

      // [ minX, minY ] rotated:
    double minimumX =  coso * minX + sino * minY;
    double minimumY = -sino * minX + coso * minY;
    double maximumX = minimumX;
    double maximumY = minimumY;
      // [ minX, maxY ] rotated:
    double actX =  coso * minX + sino * maxY;
    double actY = -sino * minX + coso * maxY;
    if ( actX < minimumX ) minimumX = actX;
    if ( actX > maximumX ) maximumX = actX;
    if ( actY < minimumY ) minimumY = actY;
    if ( actY > maximumY ) maximumY = actY;
      // [ maxX, maxY ] rotated:
    actX =  coso * maxX + sino * maxY;
    actY = -sino * maxX + coso * maxY;
    if ( actX < minimumX ) minimumX = actX;
    if ( actX > maximumX ) maximumX = actX;
    if ( actY < minimumY ) minimumY = actY;
    if ( actY > maximumY ) maximumY = actY;
      // [ maxX, minY ] rotated:
    actX =  coso * maxX + sino * minY;
    actY = -sino * maxX + coso * minY;
    if ( actX < minimumX ) minimumX = actX;
    if ( actX > maximumX ) maximumX = actX;
    if ( actY < minimumY ) minimumY = actY;
    if ( actY > maximumY ) maximumY = actY;

    int width  = Formula.round( Math.ceil( maximumX ) - Math.floor( minimumX ) ) + 1;
    int height = Formula.round( Math.ceil( maximumY ) - Math.floor( minimumY ) ) + 1;

    //LogFile.log( String.format( "Ori: %5.1f, min:%7.2f,%7.2f, max:%7.2f,%7.2f, wid: %4d, hei: %4d, '%s'",
    //                            Math.toDegrees( orientation ), minimumX, minimumY,
    //                            maximumX, maximumY, width, height, text ) );

    prepareSurface( width, height );
    AffineTransform saveAt = graphics.getTransform();
    graphics.rotate( -orientation );
    graphics.setFont( fonts[0] );
    graphics.setColor( Color.WHITE );
    actX = coso * -minimumX - sino * -minimumY;
    actY = sino * -minimumX + coso * -minimumY;
    graphics.drawString( text, (float)actX, (float)actY );
    graphics.setTransform( saveAt );

    WritableRaster raster = surface.getRaster();

    int i, j, val;
    int xi0 = Formula.round( x0 + minimumX );
    int yi0 = Formula.round( y0 + minimumY );

    if ( am != null )
      for ( j = 0; j < height; j++ )
        for ( i = 0; i < width; i++ )
        {
          val = Formula.clamp( raster.getSample( i, j, 0 ), 0, 255 );

          if ( actAntialiasing )
            am.putPixel( xi0 + i, yi0 + j, val * INV255 );
          else
            if ( val >= 128 )
              am.setPixel( xi0 + i, yi0 + j );
        }
    else
      for ( j = 0; j < height; j++ )
        for ( i = 0; i < width; i++ )
        {
          val = Formula.clamp( raster.getSample( i, j, 0 ), 0, 255 );

          if ( val >= 128 )
            bm.setPixel( xi0 + i, yi0 + j );
        }

      // debugging:
    if ( LogFile.debugging )
    {
      int            imId = LogFile.createImage( width, height, RasterGraphics.MODE_GRAY, "text%04d.png" );
      RasterGraphics gr   = LogFile.getImage( imId );

      for ( j = 0; j < height; j++ )
        for ( i = 0; i < width; i++ )
          gr.putPixel( i, j, Formula.clamp( raster.getSample( i, j, 0 ), 0, 255 ) );

      gr.putPixel( Formula.round( -minimumX ), Formula.round( -minimumY ), 160 );
      LogFile.saveImage( imId );
    }
  }

  /**
   * Computes the <b>extent</b> of the given text string (without any output change).
   *
   * @param  x0 X coordinate of the starting reference point.
   * @param  y0 Y coordinate of the starting reference point.
   * @param  text The string to be drawn.
   * @param  ext The {@link cz.cuni.jagrlib.iface.TextRender.TextExtent TextExtent} object to be filled (or {@code null}).
   * @return The filled {@link cz.cuni.jagrlib.iface.TextRender.TextExtent TextExtent} object.
   * @see    cz.cuni.jagrlib.iface.TextRender.TextExtent
   * @see    #drawText
   */
  public TextExtent textExtent ( double x0, double y0, String text, TextExtent ext )
  {
    if ( fonts[0] == null ||
         graphics == null ) return null;

    if ( ext == null )
      ext = new TextExtent();

    FontMetrics fm       = graphics.getFontMetrics( fonts[0] );
    Rectangle2D bounds2D = fm.getStringBounds( text, graphics );
    ext.x0  = x0;
    ext.y0  = y0;
    double advance = fm.stringWidth( text );
    ext.x1  = x0 + advance * coso;
    ext.y1  = y0 - advance * sino;
    ext.ulx = bounds2D.getMinX();
    ext.uly = bounds2D.getMinY();
    ext.lrx = bounds2D.getMaxX();
    ext.lry = bounds2D.getMaxY();
    //LogFile.log( String.format( "Extent(\"%s\"): %.2f, %.2f, %.2f, %.2f",
    //                            text, ext.ulx, ext.uly, ext.lrx, ext.lry ) );

    return ext;
  }

  /**
   * Returns the <b>kerning</b> (space correction) which should be used between two given
   * characters.
   *
   * @param  c1 First character.
   * @param  c2 Second character.
   * @param  ext The {@link cz.cuni.jagrlib.iface.TextRender.TextExtent TextExtent} object to be filled (or {@code null}).
   * @return The filled {@link cz.cuni.jagrlib.iface.TextRender.TextExtent TextExtent} object (only {@code dx} and {@code dy}
   *         are used).
   * @see    cz.cuni.jagrlib.iface.TextRender.TextExtent
   * @see    #textExtent
   */
  public TextExtent getKern ( char c1, char c2, TextExtent ext )
  {
    return null;
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // ANTI_MODE:
    if ( key.compareTo(ANTI_MODE) == 0 )
    {
      int old = antialiasing;
      antialiasing = intProperty( value, antialiasing, ANTI_DEPEND, ANTI_ON );
      if ( old != antialiasing )
        chAntialiasing = true;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(ANTI_MODE) == 0 )
      return antialiasing;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "JavaText";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TextRenderToBitMaskAndAlphaMask";

  /** Object description. */
  private final static String DESCRIPTION = "Draws text using java.awt.Graphics methods.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DRAW + "." + C_TEXT;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,     IFACE+"TextRender" );
      t.newOptOutputPlug( PL_OUTPUT,    IFACE+"BitMask" );
      t.newOptOutputPlug( PL_ALPHAMASK, IFACE+"AlphaMask" );
        // Property:
      t.propBegin( ANTI_MODE, TYPE_INTEGER, "Antialiasing mode", true );
        t.propDefault( ANTI_DEPEND );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Depend", ANTI_DEPEND, "Depends on output color format" );
        t.propEnum( "Off",    ANTI_OFF,    "Antialiasing off" );
        t.propEnum( "On",     ANTI_ON,     "Antialiasing on" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
