package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Jittering image synthesizer with progressive rendering.
 * Static variant of regular NxN jitering.
 *
 * @version 0.26 $Rev: 472 $ $Date: 2007-05-31 09:33:37 +0200 (čt, 31 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/JitteringSynthesizer.java">JitteringSynthesizer.java</a>
 */
public class JitteringSynthesizer extends Piece implements ImageSynthesizer
{
  //==========================================================================
  //  Data:

  /** Actual image width in pixels. */
  protected int imageWidth = 0;

  /** Actual image height in pixels. */
  protected int imageHeight = 0;

  /** Use gray palette (for RealFunction). */
  protected boolean useGray = false;

  /** Use time variable (for RealFunction). */
  protected boolean useTime = false;

  /** Actual time (for RealFunction). */
  protected double actTime = 0.0;

  /** Horizontal scaling coefficient (for RealFunction). */
  protected double mulX = 1.0;

  /** Horizontal translation (for RealFunction). */
  protected double addX = 0.0;

  /** Vertical scaling coefficient (for RealFunction). */
  protected double mulY = 1.0;

  /** Vertical translation (for RealFunction). */
  protected double addY = 0.0;

  /** Linear supersampling factor. Number of samples per pixel = <b>supersampling<sup>2</sup></b>. */
  protected int supersampling = 2;

  /** Actual number of samples per pixel ({@link #supersampling}<sup>2</sup>). */
  protected int totalSamples  = 4;

  /** Initial image cell size in pixels. */
  protected final static int INIT_CELL = 8;

  /** Actual jittering amplitude ({@code 0.0} to {@code 1.0}). */
  protected double jittering = 1.0;

  /** Global random generator. */
  protected static final RandomJames rnd = new RandomJames();

  protected double[] tmp;

  /**
   * Gamma-correction for frame-buffer efficiency.
   * It can be {@code 0.0} to turn off gamma-correction and color clamping.
   */
  protected double gamma = 2.2;

  //==========================================================================
  //  Support:

  /** Transfers image size from associated {@link RasterGraphics} if necessary. */
  protected void imageBounds ()
  {
    if ( imageWidth > 0 ) return;

      // set image size according to the connected image module
    final RasterGraphics r = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    if ( r == null )
    {
      setBounds( 512, 512 );
      return;
    }

    setBounds( r.getWidth(), r.getHeight() );
  }

  protected void computePixel ( ImageFunction f, int x, int y, double[] pre )
  {
    int len = pre.length;
    if ( tmp == null || tmp.length < len )
      tmp = new double[ len ];
    Arrays.fill( pre, 0.0 );

    int i, j, k, ord;
    final double step      = 1.0 / supersampling;
    final double amplitude = jittering * step;
    final double origin    = 0.5 * (step - amplitude);
    double x0, y0;
    for ( j = ord = 0, y0 = y + origin; j++ < supersampling; y0 += step )
      for ( i = 0, x0 = x + origin; i++ < supersampling; x0 += step )
      {
        f.getSample( x0 + amplitude * rnd.uniformNumber(),
                     y0 + amplitude * rnd.uniformNumber(),
                     ord++, totalSamples, tmp );
        for ( k = 0; k < len; k++ )
          pre[k] += tmp[k];
      }

    final double mul = step / supersampling;

    if ( gamma > 0.01 )
    {                                     // gamma-correction and clamping
      final double g = 1.0 / gamma;
      for ( k = 0; k < len; k++ )
        pre[k] = Formula.clamp( Math.pow( pre[k] * mul, g ), 0.0, 1.0 );
    }
    else                                  // no correction, no clamping (for HDRI)
      for ( k = 0; k < len; k++ )
        pre[k] *= mul;
  }

  protected void computePixel ( RealFunction rf, int x, int y, double[] pre )
  {
    int i, j;
    final double step      = 1.0 / supersampling;
    final double amplitude = jittering * step;
    final double origin    = 0.5 * (step - amplitude);
    double x0, y0;
    double result = 0.0;
    for ( j = 0, y0 = y + origin; j++ < supersampling; y0 += step )
      for ( i = 0, x0 = x + origin; i++ < supersampling; x0 += step )
      {
        double xd = mulX * (x0 + amplitude * rnd.uniformNumber()) + addX;
        double yd = mulY * (y0 + amplitude * rnd.uniformNumber()) + addY;
        result   += Formula.clamp( 0.5 * ((useTime ? rf.f( xd, yd, actTime ) : rf.f( xd, yd )) + 1.0), 0.0, 1.0 ); ;
      }

    result *= step / supersampling;
    if ( useGray )
    {
      pre[0] = pre[1] = pre[2] = result;
      if ( pre.length > 3 )
        pre[3] = 1.0;
    }
    else
      Formula.colorRamp( result, pre );
  }

  //==========================================================================
  //  Interface TimeDependent:

  /**
   * Retrieve the valid bounds for time variable.
   *
   * @param  pre Pre-allocated {@code double[2]} array to hold the result (can be
   *             {@code null}).
   * @return {@code result[0]} .. minimum time value, {@code result[1]} .. maximum
   *         time value.
   * @see    #getTime
   * @see    #setTime
   */
  public double[] getTimeInterval ( double[] pre )
  {
    ImageFunction f = (ImageFunction)getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    if ( f != null )
      return f.getTimeInterval( pre );

      // RealFunction:
    if ( !useTime ) return null;
    RealFunction rf = (RealFunction)getInterface( PL_DIRECT, IFACE+"RealFunction" );
    if ( rf == null ) return null;

    double[] dom = new double[ 6 ];
    dom = rf.getDomain( dom );
    if ( dom == null ) return null;

    if ( pre == null || pre.length < 2 )
      pre = new double[ 2 ];

    pre[0] = dom[4];
    pre[1] = dom[5];
    return pre;
  }

  /**
   * Returns the actual time value for the object (used in animations).
   *
   * @return Actual time (in seconds).
   * @see    #setTime
   */
  public double getTime ()
  {
    final ImageFunction f = (ImageFunction)getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    return( (f == null) ? actTime : f.getTime() );
  }

  /**
   * Sets the actual time value for the object (used in animations).
   *
   * @param  time New scene time (in seconds).
   * @see    #getTime
   * @see    #getTimeInterval
   */
  public void setTime ( double time )
  {
    actTime = time;
    final ImageFunction f = (ImageFunction)getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    if ( f != null )
      f.setTime( time );
  }

  //==========================================================================
  //  Interface ImageSynthesizer:

  /**
   * Sets resolution of an output raster image.
   *
   * @param  width Horizontal image size in pixels.
   * @param  height Vertical image size in pixels.
   */
  public void setBounds ( int width, int height )
  {
    imageWidth  = width;
    imageHeight = height;
    if ( imageWidth <= 0 ) return;

    final ImageFunction f = (ImageFunction)getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    if ( f != null )
    {
      f.setBounds( 0.0, width, 0.0, height );
      return;
    }

      // RealFunction:
    final RealFunction rf = (RealFunction)getInterface( PL_DIRECT, IFACE+"RealFunction" );
    if ( rf == null ) return;

    double[] dom = new double[ 6 ];
    dom = rf.getDomain( dom );
    if ( dom == null )                    // assuming 0.0-1.0 in both directions
    {
      mulX = 1.0 / width;
      mulY = 1.0 / height;
      addX =
      addY = 0.0;
      return;
    }
      // set local scaling transformation:
    mulX = (dom[1] - dom[0]) / width;
    addX =  dom[0];
    mulY = (dom[3] - dom[2]) / height;
    addX =  dom[2];
  }

  /**
   * Render the given pixel of the image.
   *
   * @param  x X coordinate of the pixel.
   * @param  y Y coordinate of the pixel.
   * @param  pre Pre-allocated {@code double[3]} array to hold results (can be
   *             {@code null}).
   * @return Color of the pixel (in <b>RGB</b> color space).
   * @see    #renderRectangle
   * @see    #renderAll
   */
  public double[] renderPixel ( int x, int y, double[] pre )
  {
    if ( pre == null || pre.length < 3 )
      pre = new double[ 3 ];

    imageBounds();

    final ImageFunction f = (ImageFunction)getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    if ( f != null )
    {
      computePixel( f, x, y, pre );
      return pre;
    }

      // RealFunction:
    final RealFunction rf = (RealFunction)getInterface( PL_DIRECT, IFACE+"RealFunction" );
    if ( rf == null ) return null;

    computePixel( rf, x, y, pre );
    return pre;
  }

  /**
   * Render the given rectangle.
   * Output is written to some connected raster device (not defined in this interface).
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @see    #renderPixel
   * @see    #renderAll
   */
  public void renderRectangle ( int x1, int y1, int x2, int y2 )
  {
    final ImageFunction  f    = (ImageFunction) getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    final RealFunction   rf   = (RealFunction)  getInterface( PL_DIRECT,   IFACE+"RealFunction" );
    final RasterGraphics r    = (RasterGraphics)getInterface( PL_OUTPUT,   IFACE+"RasterGraphics" );
    final Trigger        trig = (Trigger)       getInterface( PL_TRIGGER,  IFACE+"Trigger" );

    if ( f == null && rf == null ||
         r == null )
      return;

    imageBounds();

    if ( x1 < 0 ) x1 = 0;
    if ( x2 > imageWidth ) x2 = imageWidth;
    if ( y1 < 0 ) y1 = 0;
    if ( y2 > imageHeight ) y2 = imageHeight;
    if ( x1 >= x2 || y1 >= y2 ) return;   // the rectangle is empty

      // run several phases of image rendering:
    int cell = INIT_CELL;
    int x, y;
    boolean xParity, yParity;
    double[] c = new double[ 3 ];         // pixel color
    final long total = (x2 - x1) * (long)(y2 - y1);
    long counter = 0L;

    all:
    do                                    // do one phase
    {
      for ( y = y1, yParity = false;
            y < y2;                       // one image row
            y += cell, yParity = !yParity )

        for ( x = x1, xParity = false;
              x < x2;                     // one image cell
              x += cell, xParity = !xParity )

          if ( cell == INIT_CELL ||
               xParity || yParity )       // process the cell
          {
            if ( userBreak ) break all;

              // determine sample color ..
            if ( f != null )
              computePixel( f, x, y, c );
            else
              computePixel( rf, x, y, c );

              // .. and render it:
            int xMax = x + cell;
            if ( xMax > x2 ) xMax = x2;
            int yMax = y + cell;
            if ( yMax > y2 ) yMax = y2;
            r.setRectangle( x, y, xMax, yMax, c );

            if ( (++counter & 1023) == 0 &&
                 pl != null )
              pl.progress( counter, total );
          }

        // repaint the picture:
      if ( trig != null )
        trig.fire( cell );
      if ( pl != null )
      {
        pl.progress( counter, total );
        pl.repaint();
      }

    } while ( (cell >>= 1) > 0 );         // do one phase
  }

  /**
   * Render the whole image.
   * Output is written to some connected raster device (not defined in this interface).
   *
   * @see    #renderPixel
   * @see    #renderRectangle
   */
  public void renderAll ()
  {
    renderRectangle( 0, 0, imageWidth, imageHeight );
  }

  //==========================================================================
  //  Interface Property:

  public final static String JITTERING = "Jittering";
  public final static String USE_GRAY  = "Gray palette";
  public final static String USE_TIME  = "Use time variable";

  protected int setWidth  = 0;
  protected int setHeight = 0;

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // WIDTH:
    if ( key.compareTo(WIDTH) == 0 )
    {
      setWidth = intProperty( value, setWidth );
      return;
    }
      // HEIGHT:
    if ( key.compareTo(HEIGHT) == 0 )
    {
      setHeight = intProperty( value, setHeight );
      return;
    }
      // SUPERSAMPLING:
    if ( key.compareTo(SUPERSAMPLING) == 0 )
    {
      totalSamples  = intProperty( value, totalSamples, 1, 10000 );
      supersampling = (int)Math.sqrt( totalSamples );
      totalSamples  = supersampling * supersampling;
      return;
    }
      // JITTERING:
    if ( key.compareTo(JITTERING) == 0 )
    {
      jittering = doubleProperty( value, jittering, 0.0, 1.0 );
      return;
    }
      // GAMMA:
    if ( key.compareTo(GAMMA) == 0 )
    {
      gamma = doubleProperty( value, gamma, 0.0, 5.0 );
      return;
    }
      // USE_GRAY:
    if ( key.compareTo(USE_GRAY) == 0 )
    {
      useGray = booleanProperty( value, useGray );
      return;
    }
      // USE_TIME:
    if ( key.compareTo(USE_TIME) == 0 )
    {
      useTime = booleanProperty( value, useTime );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(WIDTH) == 0 )
      return setWidth;
    if ( key.compareTo(HEIGHT) == 0 )
      return setHeight;
    if ( key.compareTo(SUPERSAMPLING) == 0 )
      return totalSamples;
    if ( key.compareTo(SUPERSAMPLING) == 0 )
      return totalSamples;
    if ( key.compareTo(JITTERING) == 0 )
      return jittering;
    if ( key.compareTo(USE_GRAY) == 0 )
      return useGray;
    if ( key.compareTo(USE_TIME) == 0 )
      return useTime;

    return null;
  }

  /**
   * Commits all previous changes to the internal object state.
   * Non-mandatory.
   */
  @Override
  public void commit ()
  {
    setBounds( setWidth, setHeight );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "JitteringSynthesizer";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ImageSynthesizerToImageFunctionAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Image synthesizer with static jittering.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,    IFACE+"ImageSynthesizer" );
      t.newOptOutputPlug( PL_FUNCTION, IFACE+"ImageFunction" );
      t.newOptOutputPlug( PL_DIRECT,   IFACE+"RealFunction" );
      t.newOutputPlug(    PL_OUTPUT,   IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_TRIGGER,  IFACE+"Trigger" );
        // Property:
      t.propBegin( WIDTH, TYPE_INTEGER, "Image width in pixels", true );
        t.propDefault( 0 );
      t.propEnd();
      t.propBegin( HEIGHT, TYPE_INTEGER, "Image height in pixels", true );
        t.propDefault( 0 );
      t.propEnd();
      t.propBegin( SUPERSAMPLING, TYPE_INTEGER, "Supersampling factor", true );
        t.propDefault( 4 );
      t.propEnd();
      t.propBegin( JITTERING, TYPE_DOUBLE, "Jittering amplitude: 0.0 = regular supersampling, 1.0 = full jittering", true );
        t.propDefault( 1.0 );
        t.propBounds(  0.0, 1.0 );
      t.propEnd();
      t.propBegin( GAMMA, TYPE_DOUBLE, "Gamma-correction exponent", true );
        t.propDefault( 2.2 );
        t.propBounds(  0.0, 5.0 );
      t.propEnd();
      t.propBegin( USE_GRAY, TYPE_BOOLEAN, "Use gray palette in case of scalar input?", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( USE_TIME, TYPE_BOOLEAN, "Use time variable (for RealFunction)?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
