package cz.cuni.jagrlib.piece;

import java.io.IOException;
import java.util.*;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;

/**
 * LZW Trie entry.
 * This is only a supporting data container.
 *
 * @see LZWCodec.LZWTrie
 */
class LZWTrieEntry
{
  /** Actual character. */
  public int character;

  /** Reference to my parent (or {@link LZWCodec.LZWTrie#NIL}). */
  public int parent;

  /** Reference to my son (or {@link LZWCodec.LZWTrie#NIL}). */
  public int son;

  /** Reference to my brother (or {@link LZWCodec.LZWTrie#NIL}). */
  public int brother;
}

/**
 * LZW codec.
 *
 * @author  PE, Tom Poch
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/LZWCodec.java">LZWCodec.java</a>
 */
public class LZWCodec extends Piece implements EntropyCodec
{
  protected enum InsertResult
  {
    INSERT_OK,
    INSERT_FAIL,
    INSERT_FULL
  }

  /**
   * Dictionary implemented by trie-like structure
   * (represented by binary tree - first son is son, second one is brother)
   * each entry holds only one character.
   * Each word in the dictionary is represented by a path from root to its vertex.
   *
   * @author  Tom Poch, PE
   * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
   * @since   0.24
   */
  protected class LZWTrie
  {
    /** Void reference value. */
    public final int NIL = -1;

    /** Maximal symbol value. */
    protected int maxSymbol = -1;

    /**
     * Clear-directory command.
     * Value should be 256 for 8-bit alphabet.
     */
    protected int clearDir;

    /**
     * End-of-Data command.
     * Value should be 257 for 8-bit alphabet.
     */
    protected int endOfData;

    /**
     * First free phrase code.
     * Value should be 258 for 8-bit alphabet.
     */
    protected int firstCode;

    /** Internal vocabulary representation (trie). */
    protected ArrayList< LZWTrieEntry > table;

    /**
     * Actual index for trie traversal.
     * (is equal to phrase code)
     */
    protected int actualPos;

    /** Actual codeword length in bits. */
    protected int codeSize;

    /** Initial code size in bits (for dictionary re-initialization). */
    protected int initCodeSize;

    /** Accelerator: (1 << codeSize) - 1. */
    protected int maxCode;

    /** True if code-table is full and is waiting for expansion. */
    protected boolean fullCode;

    /** End-of-Data status. */
    protected boolean eod;

    /** Holds actual phrase. */
    protected int[] phrase = null;

    /** Position in actual phrase. */
    protected int phrasePos = 0;

    protected int character;

    protected int oldCode = -1;

    /**
     * Initializes the dictionary.
     *
     * @param maxSymbol Default code size - 1.
     */
    LZWTrie ( int maxSymbol )
    {
      table = new ArrayList< LZWTrieEntry >();
      setMaxSymbol( maxSymbol );
    }

    int getMaxSymbol ()
    {
      return maxSymbol;
    }

    void setMaxSymbol ( int maxSymbol )
    {
      if ( this.maxSymbol >= maxSymbol ) return;

      this.maxSymbol = maxSymbol;
      clearDir       = maxSymbol + 1;
      endOfData      = clearDir  + 1;
      firstCode      = endOfData + 1;
      codeSize       = 2;
      while ( (maxCode = (1 << codeSize) - 1) <= firstCode )
        codeSize++;
      initCodeSize = codeSize;
      fullCode     = false;
      table.ensureCapacity( maxCode + 1 );
      init();
    }

    /** At least one character is available. */
    boolean available ()
    {
      return( phrase != null && phrasePos < phrase.length );
    }

    /**
     * Change actual codeword length.
     *
     * @param codeSize New codeword length in bits.
     */
    void setCodeSize ( int codeSize )
    {
      this.codeSize = codeSize;
      maxCode       = (1 << codeSize) - 1;
      fullCode      = false;
      table.ensureCapacity( maxCode + 1 );
    }

    /** Resets trie traversing to root node. */
    void goToRoot ()
    {
      actualPos = NIL;
    }

    /**
     * (Re-)initialize the dictionary (trie).
     * clearDir, endOfData, firstCode have to be defined prior
     * to calling this method.
     */
    void init ()
    {
      LZWTrieEntry e;
      table.clear();
      for ( int i = 0; i < firstCode; )
      {
        e = new LZWTrieEntry();
        e.son = e.parent = NIL;
        e.character = i++;
        e.brother = (i == firstCode) ? NIL : i;
        table.add( e );
      }
      goToRoot();
      eod = false;
    }

    /**
     * Incremental search.
     *
     * @param  character New character to be appended.
     * @return Whether the extended phrase is already present in the dictionary.
     */
    boolean move ( int character )
    {
      LZWTrieEntry tmp;
      int tmpPos;

        // if actual is root, start from the first character
        // (table is initialized by trivial phrases)
      if ( actualPos == NIL )
        tmpPos = 0;

        // start from the first son
      else
      {
        tmpPos = table.get( actualPos ).son;
        if ( tmpPos == NIL ) return false;
      }

        // try all the sons:
      while ( (tmp = table.get(tmpPos)).character != character )
      {
        tmpPos = tmp.brother;
        if ( tmpPos == NIL ) break;
      }

        // if character wasn't found
      if ( tmpPos == NIL ) return false;

        // set the new position
      actualPos = tmpPos;

      return true;
    }

    /**
     * Insert a new leaf entry into the dictionary (trie).
     *
     * @param  character Character to be appended.
     * @return Result code.
     */
    InsertResult insertLeaf ( int character )
    {
      if ( table.size() == maxCode + 1 )  // the code is full
        return InsertResult.INSERT_FULL;

        // create new entry:
      LZWTrieEntry newEntry = new LZWTrieEntry();
      newEntry.parent       = actualPos;
      newEntry.character    = character;
      newEntry.brother      =
      newEntry.son          = NIL;

      LZWTrieEntry act;
      int tmpPos;
      if ( actualPos == NIL )
      {
        tmpPos = 0;
        act    = table.get( 0 );
      }  // this shouldn't happen
      else
      {
        act    = table.get( actualPos );
        tmpPos = act.son;
      }

      if ( tmpPos == NIL )
      {
        table.add( newEntry );
        act.son  = table.size() - 1;
        fullCode = (table.size() == maxCode + 1 && variableCode);
        return InsertResult.INSERT_OK;
      }

      while ( (act = table.get(tmpPos)).brother != NIL )
      {
        if ( act.character == character )
          return InsertResult.INSERT_FAIL;
        tmpPos = act.brother;
      }

      table.add( newEntry );
      act.brother = table.size() - 1;
      fullCode    = (table.size() == maxCode + 1 && variableCode);
      return InsertResult.INSERT_OK;
    }

    int getCode ()
    {
      return actualPos;
    }

    void setActualPos ( int code )
    {
      actualPos = code;
    }

    boolean isCode ( int code )
    {
      return !(code >= table.size());
    }

    /**
     * Returns phrase associated with the given code.
     * When code is valid, changes actual position to the code
     * (we will probably insert a new list there soon)
     */
    int[] getPhrase ( int code )
    {
      if ( code >= table.size() ) return null;
      int depth, tmpPos, i;
      LZWTrieEntry tmp;

        // 1. determine phrase size;
      for ( depth = 0, tmpPos = code;
            (tmp = table.get(tmpPos)).parent != NIL;
            tmpPos = tmp.parent )
        depth++;

      int[] ret = new int[ depth + 1 ];
      i         = depth;
      tmpPos    = code;

        // 2. fill the result array (in reverse order)
      while ( tmpPos != NIL )
      {
        tmp      = table.get( tmpPos );
        tmpPos   = tmp.parent;
        ret[i--] = tmp.character;
      }
      return ret;
    }

    int get ()
    throws IOException
    {
      if ( eod )
      {
        LogFile.log( "LZW decode: Reading after EOD marker at position: " + position );
        return -1;
        //throw new IOException( "Reading after EOD marker at position: " + position );
      }
      if ( phrase == null ||
           phrasePos == phrase.length )
      {
          // dictionary-phrase is not ready => we must get it
        if ( oldCode == NIL )
        {
            // brand new dictionary
          oldCode = (int)stream.read( codeSize );
          //LogFile.debug( "Read code: " + oldCode );
          if ( oldCode == clearDir )
          {
            position--;
            oldCode = NIL;
            init();
            setCodeSize( initCodeSize );
            //LogFile.debug( "LZW decode: Dictionary cleanup at position: " + position );
            return get();
          }
          character = oldCode;
          phrase = null;
          return character;
        }

        int code = (int)stream.read( codeSize );
        if ( code == clearDir )
        {
            // clean dictionary if input code is clearDir
          position--;
          oldCode = NIL;
          //LogFile.debug( "LZW decode: Dictionary cleanup at position: " + position );
          init();
          setCodeSize( initCodeSize );
            // we must return next char from stream
          return get();
        }

        if ( code == endOfData )          // premature end of data stream
        {
          eod = true;
          return -1;
        }

        if ( fullCode )                   // the code is full => this symbol should be one bit wider
          code |= (stream.read( 1 ) << codeSize);
        //LogFile.debug( "Read code: " + code );

        if ( isCode(code) )
            // if the code is in the dictionary, get the phrase associated to it
          phrase = getPhrase( code );
        else
        {
            // the code is not known
            // it must be in the last (unfinished) code phrase - its first char
          int[] tmpPhrase = getPhrase( oldCode );
          phrase = new int[ tmpPhrase.length + 1 ];
          System.arraycopy( tmpPhrase, 0, phrase, 0, tmpPhrase.length );
          phrase[tmpPhrase.length] = character;
        }

          // insert oldcode phrase + first char of the code phrase to the dictionary
        character = phrase[0];
        setActualPos( oldCode );
        if ( insertLeaf(character) == InsertResult.INSERT_FULL &&
             variableCode )
        {
            // the next bit is always 0, so we can skip it (not too nice)
          setCodeSize( codeSize + 1 );
          insertLeaf( character );
          //LogFile.debug( "LZW decode: Code resize (" + codeSize + ") at position: " + position );
        }
        oldCode   = code;
        phrasePos = 0;
      }

        // return next char from actual phrase
      return phrase[ phrasePos++ ];
    }

    void put ( int symbol )
    throws IOException
    {
      if ( symbol < 0 || symbol > maxSymbol )
      {
        LogFile.error( "Invalid input symbol: " + symbol + " (should be between " + 0 + " and " + maxSymbol + ")!" );
        if ( symbol < 0 ) symbol = 0;
        else
        if ( symbol > maxSymbol ) symbol = maxSymbol;
      }
        // if input string is in the dictionary, just move and return
      if ( move( symbol ) ) return;
        // input string isn't in the dictionary
        // write the code of string without last character
      stream.write( getCode(), codeSize );
      //LogFile.debug( "Write code: " + actualPos );
        // insert new item into the dictionary
      if ( insertLeaf(symbol) == InsertResult.INSERT_FULL )
      {
        if ( !variableCode || codeSize == maxCodeSize )
        {
            // if the dictionary is full, clean it
          init();
          stream.write( clearDir, codeSize );
          //LogFile.debug( "Write code: " + clearDir + " (clearDir)" );
          setCodeSize( initCodeSize );
          //LogFile.debug( "LZW encode: Dictionary init at position: " + position );
        }
        else
        {
            // if resizing is enabled, increase output symbol length
          setCodeSize( codeSize + 1 );
          insertLeaf( symbol );
          //LogFile.debug( "LZW encode: Dictionary resize (" + codeSize + ") at position: " + position );
        }
      }
      goToRoot();
        // input string init: only the last character of previous phrase
      move( symbol );
    }

  }

  //==========================================================================
  //  Data:

  /** Associated I/O stream. Null if the codec is not opened. */
  protected BitStream stream;

  /** Opened for output (encoding)? */
  protected boolean output;

  /** Id of the actual context. */
  protected int actualCtx;

  /** Working context (= dictionary). */
  protected LZWTrie ctx;

  /** Use variable-code implementation? */
  protected boolean variableCode = true;

  /** Maximal codeword size in bits. */
  protected int maxCodeSize = 20;

  /** Actual position in original (logical) stream. */
  protected long position = 0;

  /** Actual position in compressed (binary) stream. */
  protected long rawBits = 0;

  /** Default alphabet size (for new context creation). */
  protected int defaultMaxSymbol = 255;

  //==========================================================================
  //  Interface EntropyCodec:

  public LZWCodec ()
  {
    ctx = null;
  }

  public boolean available ()
  throws IOException
  {
    return ( !output && stream != null &&
             (ctx.available() || stream.available() > 0) );
  }

  /** Close stream, clean statistics, write out the EOD code.. */
  public void close ()
  throws IOException
  {
    if ( stream == null ) return;

    if ( output )
    {
      flush();
      stream.write( ctx.endOfData, ctx.codeSize );
      //LogFile.debug( "Write code: " + ctx.endOfData + " (endOfData)" );
    }
    //LogFile.debug( "LZW codec: Closing at position: " + position );
    stream.close();
    position =
    rawBits  = 0;
    stream   = null;
  }

  public long compressed ()
  throws IOException
  {
    // !!! TODO: compute exact binary position
    return stream.position();
  }

  public void flush ()
  throws IOException
  {
    if ( output )
    {
      stream.write( ctx.getCode(), ctx.codeSize );
      //LogFile.debug( "Write code: " + ctx.actualPos );
      ctx.goToRoot();
    }
  }

  /** Get the next character from compressed stream. */
  public int get ()
  throws IOException
  {
    if ( stream == null || output )
      throw new IOException( "LZWCodec is not opened for reading!" );

    position++;
    return ctx.get();
  }

  /**
   * Decodes one subinterval from the given {@link WheelOfFortune}.
   * Uses the actual context!
   *
   * @param  wheel Set of intervals (wheel of fortune).
   * @return Index of the relevant interval.
   */
  public int get ( WheelOfFortune wheel )
  throws IOException
  {
    return get();
  }

  public long getBits ( int length )
  throws IOException
  {
    if ( stream == null || output )
      throw new IOException( "LZWCodec is not opened for reading!" );

    rawBits += length;
    //LogFile.debug( "Read " + length + " bits" );
    return stream.read( length );
  }

  public void open ( boolean output, int dCtx )
  throws IOException
  {
    stream      = (BitStream)getInterface( PL_OUTPUT, IFACE+"BitStream" );
    this.output = output;
    position    =
    rawBits     = 0;
    actualCtx   = dCtx;
    ctx         = new LZWTrie( defaultMaxSymbol );
    stream.open( output, (String)null, null );
  }

  /**
   * Retrieves maximal symbol value for the actual context.
   *
   * @return Maximal symbol value.
   */
  public int getMaxSymbol ()
  {
    if ( stream == null ) return -1;
    return ctx.getMaxSymbol();
  }

  /**
   * Sets maximal symbol value for the actual context.
   *
   * @param maxSymbol New maximal symbol value.
   */
  public void setMaxSymbol ( int maxSymbol )
  {
    if ( stream != null )
      ctx.setMaxSymbol( maxSymbol );
    else
      defaultMaxSymbol = maxSymbol;
  }

  public long position ()
  throws IOException
  {
    return position;
  }

  /** Puts the given symbol into the stream. */
  public void put ( int symbol )
  throws IOException
  {
    if ( stream == null ||
         !output )
      throw new IOException( "LZWCodec is not opened for writting!" );

    position++;
    ctx.put( symbol );
  }

  /**
   * Encodes one subinterval from the given {@link WheelOfFortune}.
   * Uses the actual context!
   *
   * @param  wheel Set of intervals (wheel of fortune).
   * @param  s Index of an interval to encode.
   */
  public void put ( WheelOfFortune wheel, int s )
  throws IOException
  {
    put( s );
  }

  public void putBits ( long bits, int length )
  throws IOException
  {
    stream.write( bits, length );
    //LogFile.debug( "Write " + length + " bits (" + bits + ')' );
    rawBits += length;
  }

  public int setContext ( int ctxId )
  {
    int old   = actualCtx;
    actualCtx = ctxId;
    return old;
  }

  //==========================================================================
  //  Interface Property:

  public final static String VARIABLE_CODE = "VariableCode";
  public final static String MAX_CODE_SIZE = "MaxCodeSize";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // VARIABLE_CODE:
    if ( key.compareTo(VARIABLE_CODE) == 0 )
    {
      variableCode = booleanProperty( value, variableCode );
      return;
    }
      // MAX_CODE_SIZE:
    if ( key.compareTo(MAX_CODE_SIZE) == 0 )
    {
      maxCodeSize = intProperty( value, maxCodeSize );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(VARIABLE_CODE) == 0 )
      return variableCode;
    if ( key.compareTo(MAX_CODE_SIZE) == 0 )
      return maxCodeSize;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "LZWCodec";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "EntropyCodecToBitStream";

  /** Object description. */
  private final static String DESCRIPTION = "LZW codec implementation.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_COMPRESSION + ".LZW";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"EntropyCodec" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"BitStream" );
        // Property:
      t.propBegin( VARIABLE_CODE, TYPE_BOOLEAN, "Growable code size", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( MAX_CODE_SIZE, TYPE_INTEGER, "Maximal code size", true );
        t.propDefault( 20 );
        t.propBounds(  3, 30 );
      t.propEnd();
     }
   return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
