package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Antialiased line with arbitrary width and floating-point coordinates.
 *
 * @author  Karel Poledna (poledk@cmail.cz)
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/LineAntialias.java">LineAntialias.java</a>
 */
public class LineAntialias extends Piece implements LineRenderAnti
{
  //==========================================================================
  //  Data:

  /**
   * Line width in pixels.
   *
   * @see    Render#LINE_WIDTH
   */
  protected double width = 1.0;

  //==========================================================================
  //  Interface LineRenderAnti:

  /**
   * Sets default line width in pixels.
   *
   * @param  wid New line width in pixels.
   * @return The old line width in pixels.
   */
  public double setWidth ( double wid )
  {
    double old = width;
    width = wid;
    return old;
  }

  /**
   * Draws a line from [x1,y1] to [x2,y2].
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  y1 Y coordinate of the starting pixel.
   * @param  x2 X coordinate of the line end.
   * @param  y2 Y coordinate of the line end.
   * @param  wid Line width in pixels.
   */
  public void drawLine ( int x1, int y1, int x2, int y2, double wid )
  {
    drawLine( (double)x1, (double)y1, (double)x2, (double)y2, wid );
  }

  /** Pixel area -&gt; intensity transform. */
  protected static final double intensity ( double value )
  {
    return( 1.0 - (1.0 - value) * (1.0 - value) );
  }

  /**
   * Draws a line from [x1,y1] to [x2,y2].
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  y1 Y coordinate of the starting pixel.
   * @param  x2 X coordinate of the line end.
   * @param  y2 Y coordinate of the line end.
   * @param  wid Line width in pixels.
   */
  public void drawLine ( double x1, double y1, double x2, double y2, double wid )
  {
    final AlphaMask out = (AlphaMask)getInterface( PL_OUTPUT, IFACE+"AlphaMask" );
    if ( out == null ) return;

      // Coordinate shift
      // Input coordinates:   origin lies in the center of the pixel
      // Working coordinates: origin defines upper left pixel corner
    x1 += 0.5; y1 += 0.5;
    x2 += 0.5; y2 += 0.5;

    final double ax = Math.abs( x2 - x1 );
    final double ay = Math.abs( y2 - y1 );

    if ( ax > ay )                        // x-axis is dominant
    {
      final double dy = ay / ax;

      final double triangle = 0.5 * dy;   // area of an [1,dy] triangle
      final double itri = 1 - triangle;   // complementary area..
      final double wy = wid * Math.sqrt( 1.0 + dy * dy ); // line width projected to x-axis
      final double axy2 = 0.5 / dy;

        // assert growing y-coordinate
      if ( y1 > y2 )
      {
        double tmp = x1; x1 = x2; x2 = tmp;
               tmp = y1; y1 = y2; y2 = tmp;
      }

      final int startx;                   // first column
      final int stopx;                    // after the last column
      final int stepx;                    // x-coordinate step
      double y;                           // starting y-coordinate (related to pixel border)

      if ( x2 > x1 )
      {
        startx = (int)(x1 + 0.5);
        stopx  = (int)(x2 + 0.5);
        stepx  = 1;
        y      = y1 - ay * (x1 - startx) / ax;
      }
      else
      {
        startx = (int)(x1 - 0.5);
        stopx  = (int)(x2 - 0.5);
        stepx  = -1;
        y      = y1 - ay * (startx + 1 - x1) / ax;
      }

      y -= 0.5 * wy;                      // y-coordinate of the segment upper bound
      int x = startx;                     // first column

      while ( x != stopx )                // draw loop
      {
        int ty = (int)y;                  // upmost affected row
        double t1, t2;                    // area of two upmost pixels (restricted by the bottom-line)

        double dty = (ty + 1) - y;        // "pixel bottom to line top" distance

        if ( dty >= dy )
        {
          t1 = dty - triangle;
          t2 = 1.0;
        }
        else
        {
          t1 = dty * dty * axy2;
          t2 = itri + dty - t1;
        }

        double by  = y + wy;              // y-coordinate of line segment bottom
        int    ly  = (int)by;             // bottom line row
        double l1, l2;                    // area of two bottom pixels (restricted by the bottom-line)
        double dly = (ly + 1) - by;

        if ( dly >= dy )
        {
          l1 = dly - triangle;
          l2 = 1.0;
        }
        else
        {
          l1 = dly * dly * axy2;
          l2 = itri + dly - l1;
        }

        if ( ty == ly )
        {
            // narrow line (both bounds start in the same pixel)
          out.putPixel( x, ty,     intensity( t1 - l1 ) );
          out.putPixel( x, ty + 1, intensity( t2 - l2 ) );
        }
        else
        if ( ty + 1 == ly )
        {
            // a little bit wider line (bounds lie in two adjacent pixels)
          out.putPixel( x, ty,     intensity( t1 ) );
          out.putPixel( x, ly,     intensity( t2 - l1 ) );
          out.putPixel( x, ly + 1, intensity( 1.0 - l2 ) );
        }
        else
        {
            // wide line (at least one pixel lies between line bounds)
          out.putPixel( x, ty,     intensity( t1 ) );
          out.putPixel( x, ty + 1, intensity( t2 ) );

          for ( int yy = ty + 2; yy < ly; yy++ )
            out.putPixel( x, yy, 1.0 );

          out.putPixel( x, ly,     intensity( 1.0 - l1 ) );
          out.putPixel( x, ly + 1, intensity( 1.0 - l2 ) );
        }

        x += stepx;
        y += dy;
      }

    }

    else                                  // y-axis is dominant
    {                                     // there are no comments in this branch, see the first (x-dominant) branch
      final double dx = ax / ay;          // and swap x<->y mentally

      final double triangle = 0.5 * dx;
      final double itri = 1.0 - triangle;
      final double wx = wid * Math.sqrt( 1.0 + dx * dx );
      final double ayx2 = 0.5 / dx;

      if ( x1 > x2 )
      {
        double tmp = x1; x1 = x2; x2 = tmp;
               tmp = y1; y1 = y2; y2 = tmp;
      }

      final int starty;
      final int stopy;
      final int stepy;
      double x;

      if ( y2 > y1 )
      {
        starty = (int)(y1 + 0.5);
        stopy  = (int)(y2 + 0.5);
        stepy  = 1;
        x      = x1 - ax * (y1 - starty) / ay;
      }
      else
      {
        starty = (int)(y1 - 0.5);
        stopy  = (int)(y2 - 0.5);
        stepy  = -1;
        x      = x1 - ax * (starty + 1 - y1) / ay;
      }

      x -= 0.5 * wx;
      int y = starty;

      while ( y != stopy )
      {
        int tx = (int)x;
        double t1, t2;
        double dtx = (tx + 1) - x;

        if ( dtx >= dx )
        {
          t1 = dtx - triangle;
          t2 = 1.0;
        }
        else
        {
          t1 = dtx * dtx * ayx2;
          t2 = itri + dtx - t1;
        }

        double bx  = x + wx;
        int    lx  = (int)bx;
        double l1, l2;
        double dlx = (lx + 1) - bx;

        if ( dlx >= dx )
        {
          l1 = dlx - triangle;
          l2 = 1.0;
        }
        else
        {
          l1 = dlx * dlx * ayx2;
          l2 = itri + dlx - l1;
        }

        if ( tx == lx )
        {
          out.putPixel( tx,     y, intensity( t1 - l1 ) );
          out.putPixel( tx + 1, y, intensity( t2 - l2 ) );
        }
        else
        if ( tx + 1 == lx )
        {
          out.putPixel( tx,     y, intensity( t1 ) );
          out.putPixel( lx,     y, intensity( t2 - l1 ) );
          out.putPixel( lx + 1, y, intensity( 1.0 - l2 ) );
        }
        else
        {
          out.putPixel( tx,     y, intensity( t1 ) );
          out.putPixel( tx + 1, y, intensity( t2 ) );

          for ( int xx = tx + 2; xx < lx; xx++ )
            out.putPixel( xx, y, 1.0 );

          out.putPixel( lx,     y, intensity( 1.0 - l1 ) );
          out.putPixel( lx + 1, y, intensity( 1.0 - l2 ) );
        }

        y += stepy;
        x += dx;
      }
    }
  }

  /**
   * Draws a line from [x1,y1] to [x2,y2].
   * Uses default line width.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  y1 Y coordinate of the starting pixel.
   * @param  x2 X coordinate of the line end.
   * @param  y2 Y coordinate of the line end.
   */
  public void drawLine ( int x1, int y1, int x2, int y2 )
  {
    drawLine( x1, y1, x2, y2, width );
  }

  /**
   * Draws a line from [x1,y1] to [x2,y2].
   * Uses default line width.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  y1 Y coordinate of the starting pixel.
   * @param  x2 X coordinate of the line end.
   * @param  y2 Y coordinate of the line end.
   */
  public void drawLine ( double x1, double y1, double x2, double y2 )
  {
    drawLine( x1, y1, x2, y2, width );
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // LINE_WIDTH:
    if ( key.compareTo(LINE_WIDTH) == 0 )
    {
      width = doubleProperty( value, width, 0.001, 100.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(LINE_WIDTH) == 0 )
      return width;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "LineAntialias";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "LineRenderAntiToAlphaMask";

  /** Object description. */
  private final static String DESCRIPTION = "Draws anti-aliased line segment with arbitrary width and floating-point coordinates.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DRAW + '.' + C_LINE + '.' + C_FLOAT;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"LineRenderAnti" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"AlphaMask" );
        // Property:
      t.propBegin( LINE_WIDTH, TYPE_DOUBLE, "Line width in pixels", true );
        t.propDefault( 1.0 );
        t.propBounds( 0.001, 100.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
