package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Antialiased lines with arbitrary ending points.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/LineAntialiasSimple.java">LineAntialiasSimple.java</a>
 */
public class LineAntialiasSimple extends Piece implements LineRenderAnti
{
  //==========================================================================
  //  Data:

  /**
   * Line width in pixels.
   *
   * @see    Render#LINE_WIDTH
   */
  protected double width = 1.0;

  //==========================================================================
  //  Support:

  /**
   * Computes area of an parallelogram [0,y]-[1,y]-[1,h1]-[0,h0]
   * intersected with pixel [0,y].
   */
  protected final static double area ( int y, double h0, double h1 )
  {
    double floor0 = Math.floor( h0 );
    double fract0 = h0 - floor0;
    floor0       -= y;
    double floor1 = Math.floor( h1 );
    double fract1 = h1 - floor1;
    floor1       -= y;
      // too high:
    if ( floor0 >= 1.0 && floor1 >= 1.0 )
      return 1.0;

    double b;
      // right point is too high:
    if ( floor1 >= 1.0 )
    {
      b = (1.0 - fract0) / (1.0 - fract0 + fract1 );
      return( 1.0 - b * (1.0 - fract0) * 0.5 );
    }

      // left point is too high:
    if ( floor0 >= 1.0 )
    {
      b = (1.0 - fract1) / (1.0 - fract1 + fract0 );
      return( 1.0 - b * (1.0 - fract1) * 0.5 );
    }

      // the whole p11m lies inside the pixel:
    if ( floor0 >= 0.0 && floor1 >= 0.0 )
      return( 0.5 * (fract0 + fract1) );

      // right point is inside:
    if ( floor1 >= 0.0 )
    {
      b = fract1 / (1.0 - fract0 + fract1);
      return( 0.5 * b * fract1 );
    }

      // left point is inside:
    if ( floor0 >= 0.0 )
    {
      b = fract0 / (1.0 - fract1 + fract0);
      return( 0.5 * b * fract0 );
    }

      // too low:
    return 0.0;
  }

  //==========================================================================
  //  Interface LineRenderAnti:

  /**
   * Sets default line width in pixels.
   *
   * @param  wid New line width in pixels.
   * @return The old line width in pixels.
   */
  public double setWidth ( double wid )
  {
    double old = width;
    width = Formula.clamp( wid, 0.001, 100.0 );
    return old;
  }

  /**
   * Draws a line from [x1,y1] to [x2,y2].
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  y1 Y coordinate of the starting pixel.
   * @param  x2 X coordinate of the line end.
   * @param  y2 Y coordinate of the line end.
   * @param  wid Line width in pixels.
   */
  public void drawLine ( int x1, int y1, int x2, int y2, double wid )
  {
    drawLine( (double)x1, (double)y1, (double)x2, (double)y2, wid );
  }

  /**
   * Draws a line from [x1,y1] to [x2,y2].
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  y1 Y coordinate of the starting pixel.
   * @param  x2 X coordinate of the line end.
   * @param  y2 Y coordinate of the line end.
   */
  public void drawLine ( int x1, int y1, int x2, int y2 )
  {
    drawLine( (double)x1, (double)y1, (double)x2, (double)y2, width );
  }

  /**
   * Draws a line from [x1,y1] to [x2,y2].
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  y1 Y coordinate of the starting pixel.
   * @param  x2 X coordinate of the line end.
   * @param  y2 Y coordinate of the line end.
   * @param  wid Line width in pixels.
   */
  public void drawLine ( double x1, double y1, double x2, double y2, double wid )
  {
    AlphaMask out = (AlphaMask)getInterface(PL_OUTPUT,IFACE+"AlphaMask");
    if ( out == null ) return;

    double ax = Math.abs( x2 - x1 );
    double ay = Math.abs( y2 - y1 );
    int major, major0, major1;
    double tmp, d, up0, up1, lo0, lo1, a;
    int minMax, minMin, minor;

    if ( ax > ay )
    {                                     // X-axis is dominant..
      if ( x1 > x2 )
      {
        tmp = x1; x1 = x2; x2 = tmp;
        tmp = y1; y1 = y2; y2 = tmp;
      }
      d = (y2 - y1) / ax;                 // direction vector = [1,d]
      tmp = Math.sqrt( 1.0 + d * d );     // normal vector = [-d,1], minor coef = tmp
      wid *= 0.5 * tmp;                   // half-width projected to minor axis
      major0 = major = Formula.round( x1 ); // lower bound for major axis
      major1         = Formula.round( x2 ); // upper bound for major axis
      tmp = x1 + 0.5 - major0;            // left extension to major0
      up1 = y1 + 0.5 - tmp * d + wid;     // minor coordinate of left-upper point
      lo1 = y1 + 0.5 - tmp * d - wid;     // minor coordinate of left-lower point
      tmp = 1.0 - tmp;                    // coverage of the first pixel

      do
      {                                   // draw a pixel (at least one)
        up0 = up1; up1 += d;
        lo0 = lo1; lo1 += d;
        minMin = (int)Math.min( lo0, lo1 );
        minMax = (int)Math.max( up0, up1 );
        if ( major == major1 )
          tmp = (x2 + 0.5 - major1);
        else
        if ( major != major0 )
          tmp = 1.0;

        for ( minor = minMin; minor <= minMax; minor++ )
        {
          a = area( minor, up0, up1 ) - area( minor, lo0, lo1 );
          out.putPixel( major, minor, tmp * a );
        }

      } while ( ++major <= major1 );
    }

    else                                  // Y-axis is dominant..
    {
      if ( y1 > y2 )
      {
        tmp = x1; x1 = x2; x2 = tmp;
        tmp = y1; y1 = y2; y2 = tmp;
      }
      d = (x2 - x1) / ay;                 // direction vector = [d,1]
      tmp = Math.sqrt( 1.0 + d * d );     // normal vector = [1,-d], minor coef = tmp
      wid *= 0.5 * tmp;                   // half-width projected to minor axis
      major0 = major = Formula.round( y1 ); // lower bound for major axis
      major1         = Formula.round( y2 ); // upper bound for major axis
      tmp = y1 + 0.5 - major0;            // left extension to major0
      up1 = x1 + 0.5 - tmp * d + wid;     // minor coordinate of left-upper point
      lo1 = x1 + 0.5 - tmp * d - wid;     // minor coordinate of left-lower point
      tmp = 1.0 - tmp;                    // coverage of the first pixel

      do
      {                                   // draw a pixel (at least one)
        up0 = up1; up1 += d;
        lo0 = lo1; lo1 += d;
        minMin = (int)Math.min( lo0, lo1 );
        minMax = (int)Math.max( up0, up1 );
        if ( major == major1 )
          tmp = (y2 + 0.5 - major1);
        else
        if ( major != major0 )
          tmp = 1.0;

        for ( minor = minMin; minor <= minMax; minor++ )
        {
          a = area( minor, up0, up1 ) - area( minor, lo0, lo1 );
          out.putPixel( minor, major, tmp * a );
        }

      } while ( ++major <= major1 );
    }
  }

  /**
   * Draws a line from [x1,y1] to [x2,y2].
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  y1 Y coordinate of the starting pixel.
   * @param  x2 X coordinate of the line end.
   * @param  y2 Y coordinate of the line end.
   */
  public void drawLine ( double x1, double y1, double x2, double y2 )
  {
    drawLine( x1, y1, x2, y2, width );
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // LINE_WIDTH:
    if ( key.compareTo(LINE_WIDTH) == 0 )
    {
      width = doubleProperty( value, width, 0.001, 100.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(LINE_WIDTH) == 0 )
      return width;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "LineAntialiasSimple";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "LineRenderAntiToAlphaMask";

  /** Object description. */
  private final static String DESCRIPTION = "Draws an anti-aliased line segment with arbitrary width and floating-point coordinates.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DRAW + '.' + C_LINE + '.' + C_FLOAT;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION);
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"LineRenderAnti" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"AlphaMask" );
        // Property:
      t.propBegin( LINE_WIDTH, TYPE_DOUBLE, "Line width in pixels", true );
        t.propDefault( 1.0 );
        t.propBounds( 0.001, 100.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
