package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Bresenham's algorithm for line segments.
 *
 * @version 0.25 $Rev: 381 $ $Date: 2007-01-02 15:51:44 +0100 (út, 02 I 2007) $ $Author: pepca $
 * @since   0.01
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/LineBresenham.java">LineBresenham.java</a>
 */
public class LineBresenham extends Piece implements LineRender
{
  //==========================================================================
  //  Data:

  /**
   * Line-join state.
   *
   * @see    Render#LINE_JOIN
   */
  protected int join = LINE_JOIN_OVERLAP;

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // LINE_JOIN:
    if ( key.compareTo(LINE_JOIN) == 0 )
    {
      join = intProperty( value, join );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(LINE_JOIN) == 0 )
      return join;
    return null;
  }

  //==========================================================================
  //  Interface LineRender:

  /**
   * Draws a line from [x1,y1] to [x2,y2].
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  y1 Y coordinate of the starting pixel.
   * @param  x2 X coordinate of the line end.
   * @param  y2 Y coordinate of the line end.
   */
  public void drawLine ( int x1, int y1, int x2, int y2 )
  {
    BitMask out = (BitMask)getInterface( PL_OUTPUT, IFACE+"BitMask" );
    int D, ax, ay, sx, sy;

    sx = x2 - x1;
    ax = Math.abs( sx ) << 1;
    if ( sx < 0 ) sx = -1;
    else
      if ( sx > 0 ) sx = 1;

    sy = y2 - y1;
    ay = Math.abs( sy ) << 1;
    if ( sy < 0 ) sy = -1;
    else
      if ( sy > 0 ) sy = 1;

    if ( ax > ay )                          // x coordinate is dominant
    {
      D = ay - (ax >> 1);                   // initial D
      ax = ay - ax;                         // ay = increment0; ax = increment1
      while ( x1 != x2 )
      {
        out.setPixel( x1, y1 );
        if ( D >= 0 )                       // lift up the Y coordinate
        {
          y1 += sy;
          D += ax;
        }
        else
          D += ay;
        x1 += sx;
      }
    }

    else                                    // y coordinate is dominant
    {
      D = ax - (ay >> 1);                   // initial D
      ay = ax - ay;                         // ax = increment0; ay = increment1
      while ( y1 != y2 )
      {
        out.setPixel( x1, y1 );
        if ( D >= 0 )                       // lift up the X coordinate
        {
          x1 += sx;
          D += ay;
        }
        else
          D += ax;
        y1 += sy;
      }
    }
                                            // the very last pixel (only in non-disjoint mode)
    if ( join != LINE_JOIN_DISJOINT ) out.setPixel( x1, y1 );
  }

  /**
   * Draws a line from [x1,y1] to [x2,y2].
   * Calls {@link #drawLine(int,int,int,int)}
   * directly after rounding the <tt>double</tt> arguments.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  y1 Y coordinate of the starting pixel.
   * @param  x2 X coordinate of the line end.
   * @param  y2 Y coordinate of the line end.
   */
  public void drawLine ( double x1, double y1, double x2, double y2 )
  {
    drawLine( Formula.round( x1 ), Formula.round( y1 ),
              Formula.round( x2 ), Formula.round( y2 ) );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Bresenham";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "LineRenderToBitMask";

  /** Object description. */
  private final static String DESCRIPTION = "Draws a line segment using integer coordinates. " +
                                            "Uses Bresenham's algorithm.";
  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DRAW + "." + C_LINE + "." + C_INTEGER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"LineRender" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"BitMask" );
        // Property:
      t.propBegin( LINE_JOIN, TYPE_INTEGER, "Line-join style", true );
        t.propDefault( LINE_JOIN_DISJOINT );
        t.propBounds(  LINE_JOIN_DISJOINT, LINE_JOIN_OVERLAP );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Disjoint", LINE_JOIN_DISJOINT, "Subsequent line segments are disjoint" );
        t.propEnum( "Overlap",  LINE_JOIN_OVERLAP,  "Subsequent line segments have one common pixel" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
