package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Linear scalar quantizer.
 *
 * @version 0.26 $Rev: 472 $ $Date: 2007-05-31 09:33:37 +0200 (čt, 31 V 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/LinearQuantizer.java">LinearQuantizer.java</a>
 */
public class LinearQuantizer extends Piece implements ScalarQuantizer
{
  //==========================================================================
  //  Data:

  /** Minimal input value. */
  protected double iMin   = 0.0;

  /** Maximal input value. */
  protected double iMax   = 1.0;

  /** Number of output levels. */
  protected int levels    = 256;

  /** Accelerator (quantization multiplier). */
  protected double coef   = 255.0;

  /** Accelerator (reconstruction multiplier). */
  protected double deCoef = 1.0 / 255.0;

  //==========================================================================
  //  Interface ScalarQuantizer:

  /**
   * Set actual quantizer variant (context).
   *
   * @param  variant Variant number (from {@code 0}).
   * @return Old variant number.
   * @see    #setBounds
   */
  public int setVariant ( int variant )
  {
    return 0;
  }

  /**
   * Sets input and output bounds.
   *
   * @param  inMin Input value minimum.
   * @param  inMax Input value maximum.
   * @param  outLevels Number of output levels.
   * @see    #setVariant
   */
  public void setBounds ( double inMin, double inMax, int outLevels )
  {
    iMin   = inMin;
    iMax   = inMax;
    if ( iMax <= iMin )
      iMax = iMin + 1.0;
    levels = outLevels;
    if ( levels < 2 )
      levels = 2;

    coef   = (levels - 1.0) / (iMax - iMin);
    deCoef = 1.0 / coef;
  }

  /**
   * Quantization (floating-point form).
   *
   * @param  in Input floating-point value.
   * @return Quantized value (or {@code -1} if failed).
   * @see    #quantize(int)
   */
  public int quantize ( double in )
  {
    return Formula.clamp( Formula.round( (in - iMin) * coef ), 0, levels-1 );
  }

  /**
   * Quantization (integer form).
   *
   * @param  in Input integer value.
   * @return Quantized value (or {@code -1} if failed).
   * @see    #quantize(double)
   */
  public int quantize ( int in )
  {
    return Formula.clamp( Formula.round( (in - iMin) * coef ), 0, levels-1 );
  }

  /**
   * Dequantization = reconstruction (floating-point form).
   *
   * @param  quant Quantized value.
   * @return Source (reconstructed) value.
   * @see    #dequantizeInt(int)
   * @see    #quantize(double)
   */
  public double dequantizeDouble ( int quant )
  {
    quant = Formula.clamp( quant, 0, levels-1 );
    return( iMin + deCoef * quant );
  }

  /**
   * Dequantization = reconstruction (integer form).
   *
   * @param  quant Quantized value.
   * @return Source (reconstructed) value.
   * @see    #dequantizeDouble(int)
   * @see    #quantize(int)
   */
  public int dequantizeInt ( int quant )
  {
    quant = Formula.clamp( quant, 0, levels-1 );
    return Formula.round( iMin + deCoef * quant );
  }

  //==========================================================================
  //  Interface PersistentParamsBinary:

  /**
   * Pack instance parameters into binary format.
   *
   * @return Data array to be stored into binary stream. Can be {@code null}.
   *         The array can be shared by all method calls.
   * @see    #loadData
   */
  public byte[] storeData ()
  {
    byte[] arr  = new byte[ 10 ];

      // iMin: 32 bits
    Formula.storeFloat( arr, 0, (float)iMin );

      // iMax: 32 bits
    Formula.storeFloat( arr, 4, (float)iMax );

      // levels: 16 bits
    Formula.storeUnsigned16( arr, 8, levels );

    return arr;
  }

  /**
   * Loads instance parameters from (previously stored) binary data.
   *
   * @param  data Binary data to be interpreted. Can be {@code null}.
   * @return {@code true} if successful.
   * @see    #storeData
   */
  public boolean loadData ( byte[] data )
  {
    if ( data == null ||
         data.length < 10 ) return false;

      // iMin: 32 bits
    iMin   = Formula.loadFloat( data, 0 );

      // iMax: 32 bits
    iMax   = Formula.loadFloat( data, 4 );

      // levels: 16 bits
    levels = Formula.loadUnsigned16( data, 8 );

    coef   = (levels - 1.0) / (iMax - iMin);
    deCoef = 1.0 / coef;
    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String MIN_VALUE = "Minimum";
  public final static String MAX_VALUE = "Maximum";
  public final static String LEVELS    = "Levels";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // MIN_VALUE:
    if ( key.compareTo(MIN_VALUE) == 0 )
    {
      iMin = doubleProperty( value, iMin );
      return;
    }
      // MAX_VALUE:
    if ( key.compareTo(MAX_VALUE) == 0 )
    {
      iMax = doubleProperty( value, iMax );
      return;
    }
      // LEVELS:
    if ( key.compareTo(LEVELS) == 0 )
    {
      levels = intProperty( value, levels, 2, Integer.MAX_VALUE );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(MIN_VALUE) == 0 )
      return iMin;
    if ( key.compareTo(MAX_VALUE) == 0 )
      return iMax;
    if ( key.compareTo(LEVELS) == 0 )
      return levels;

    return null;
  }

  /**
   * Commits all previous changes to the internal object state.
   * Non-mandatory.
   */
  @Override
  public void commit ()
  {
    setBounds( iMin, iMax, levels );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Linear quantizer";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ScalarQuantizer";

  /** Object description. */
  private final static String DESCRIPTION = "Linear quantizer.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + ".quantization";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"ScalarQuantizer" );
        // Property:
      t.propBegin( MIN_VALUE, TYPE_DOUBLE, "Minimum input value", true );
        t.propDefault( 0.0 );
      t.propEnd();
      t.propBegin( MAX_VALUE, TYPE_DOUBLE, "Maximum input value", true );
        t.propDefault( 1.0 );
      t.propEnd();
      t.propBegin( LEVELS, TYPE_INTEGER, "Number of quantized levels", true );
        t.propDefault( 256 );
        t.propBounds(  2, Integer.MAX_VALUE );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
