package cz.cuni.jagrlib.piece;

import java.io.IOException;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Mesh compression format using arbitrary pass order, zero-skipping and optional DPCM.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/MeshCompressionRLE.java">MeshCompressionRLE.java</a>
 */
public class MeshCompressionRLE extends DefaultRasterFileFormat
{
  //==========================================================================
  //  Data:

  /** Use DPCM (prediction) in amplitude coding? */
  protected boolean useDPCM = true;

  /** Use robust file-format (more checking but also more bits)? */
  protected boolean robustFileFormat = true;

  /** Maximum number of skipped zeroes ({@code maxSkip} means the zero-gap continues). */
  protected int maxSkip = 255;

  //==========================================================================
  //  Codec:

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonSave ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    final EntropyCodec codec = (EntropyCodec)getInterface( PL_OUTPUT, IFACE+"EntropyCodec" );
    final Order2D      order = (Order2D)     getInterface( "order",   IFACE+"Order2D" );

    if ( codec == null ||
         order == null ) return;

      // codec init (0 .. amplitude):
    codec.open( true, 0 );
    codec.setMaxSymbol( useDPCM ? 510 : 255 );

      // File header:
    writeUnsigned16( codec, MAGIC );
    int width  = g.getWidth();
    writeUnsigned16( codec, width );
    int height = g.getHeight();
    writeUnsigned16( codec, height );

    if ( robustFileFormat )
    {
      codec.putBits( useDPCM ? 1 : 0, 1 );
      int type   = intProperty( order, Order2D.ORDER_TYPE, 0 );
      codec.putBits( type, 7 );
      writeUnsigned16( codec, maxSkip );
      writeParameters( codec, codec );
    }

    codec.setContext( 1 );                // codec init (1 .. zero skip)
    codec.setMaxSymbol( maxSkip );

    int last    = 128;                    // amplitude of the last pixel (if DPCM is on)
    int toSkip  = 0;                      // how many void pixels were skipped recently
    int[] color = new int[ 3 ];           // actual pixel color (red > 0 => pixel is valid)

      // image pass:
    order.setSize( width, height );
    //order.init();
    int[] c = new int[ 2 ];
    while ( order.isNext() )
    {
      c = order.next( c );
      if ( c == null ) break;

      g.getRGB( c[0], c[1], color );
      if ( color[0] == 0 )                // void pixel => skip it
      {
        if ( ++toSkip >= maxSkip )
        {
          codec.put( maxSkip );           // split the skip => encode maxSkip
          toSkip -= maxSkip;
          //LogFile.log( "SkipMax " + maxSkip );
        }
      }
      else                                // valid pixel
      {
          // 1. encode number of skipped void pixes
        codec.put( toSkip );
        //LogFile.log( "Pixel " + toSkip + " [" + c[0] + ',' + c[1] + "] " + color[1] );
        toSkip = 0;
          // 2. encode the amplitude
        codec.setContext( 0 );
        if ( useDPCM )
        {
          codec.put( (int)Formula.vlc( color[1] - last ) );
          last = color[1];
        }
        else
          codec.put( color[1] );
        codec.setContext( 1 );
      }
    }

    if ( toSkip > 0 )
    {
      codec.put( toSkip );
      //LogFile.log( "Final skip " + toSkip );
    }
    codec.close();
  }

  /**
   * Common load code.
   *
   * @param  stream Opened input bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonLoad ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    final EntropyCodec codec = (EntropyCodec)getInterface( PL_OUTPUT, IFACE+"EntropyCodec" );
    final Order2D      order = (Order2D)     getInterface( "order",   IFACE+"Order2D" );

    if ( codec == null ||
         order == null ) return;

    codec.open( false, 0 );               // codec init (0 .. amplitude)

      // Pal header:
    int magic     = readUnsigned16( codec );
    if ( magic != MAGIC )
      throw new IOException( "Error in Mesh file format" );

    int width     = readUnsigned16( codec );
    int height    = readUnsigned16( codec );

    if ( robustFileFormat )
    {
      int type = (int)codec.getBits( 1 ); // use DPCM?
      useDPCM  = type != 0;
      type     = (int)codec.getBits( 7 ); // order type
      order.set( Order2D.ORDER_TYPE, type );
      maxSkip  = readUnsigned16( codec );
      if ( !readParameters( codec, codec ) )
        throw new IOException( "Error in Mesh file format (bad codec)" );
    }

    codec.setMaxSymbol( useDPCM ? 510 : 255 );
    codec.setContext( 1 );                // codec init (1 .. zero skip)
    codec.setMaxSymbol( maxSkip );

      // initialize raster image object:
    g.init( width, height, RasterGraphics.MODE_RGB, 3 );
    g.setRectangle( 0, 0, width, height, BrepToImage.BACKGROUND );
    int[] color = BrepToImage.USED;
    int last    = 128;                    // amplitude of the last pixel (if DPCM is on)

      // image data:
    order.setSize( width, height );
    //order.init();
    int[] c = new int[ 2 ];
    int toSkip, i;

      // decoder loop is controlled by input code:
    while ( order.isNext() )
    {
        // read skip sequence[s]:
      while ( order.isNext() )
      {
        i = toSkip = (int)codec.get();
        //LogFile.log( "Skip " + toSkip );
        while ( i-- > 0 )
          c = order.next( c );
        if ( toSkip < maxSkip ) break;
      }
      if ( !order.isNext() ) break;
      c = order.next( c );
      if ( c == null ) break;

        // pixel [ c[0], c[1] ] is valid..
      codec.setContext( 0 );
      i = (int)codec.get();
      //LogFile.log( "Pixel [" + c[0] + ',' + c[1] + "] " + i );
      codec.setContext( 1 );
      if ( useDPCM )
      {
        color[1] =
        color[2] = last + Formula.vlcInv( i );
        last     = color[1];
      }
      else
        color[1] =
        color[2] = i;

      g.putPixel( c[0], c[1], color );
    }

    codec.close();
  }

  //==========================================================================
  //  Interface Property:

  public final static String USE_DPCM = "Use DPCM";
  public final static String MAX_SKIP = "Max skip";
  public final static String ROBUST   = "Robust file-format";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // USE_DPCM:
    if ( key.compareTo(USE_DPCM) == 0 )
    {
      useDPCM = booleanProperty( value, useDPCM );
      return;
    }
      // MAX_SKIP:
    if ( key.compareTo(MAX_SKIP) == 0 )
    {
      maxSkip = intProperty( value, maxSkip, 1, 65535 );
      return;
    }
      // ROBUST:
    if ( key.compareTo(ROBUST) == 0 )
    {
      robustFileFormat = booleanProperty( value, robustFileFormat );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(USE_DPCM) == 0 )
      return useDPCM;
    if ( key.compareTo(MAX_SKIP) == 0 )
      return maxSkip;
    if ( key.compareTo(ROBUST) == 0 )
      return robustFileFormat;

    return null;
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /** 16-bit magic number (used at the file beginning). */
  public final static int MAGIC = 0x8e01;

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return 2;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type ({@code 0.0} .. no way,
   *         {@code 0.5} .. maybe, {@code 0.9} .. almost sure, {@code 1.0} ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int len = (header == null) ? 0 : header.length;
    if ( len > 2 ) len = 2;
    if ( len < 2 )                        // not enough bytes in header[]
      return 0.0;
    byte[] h =
    {                                     // Pal header
      (byte)((MAGIC>>8) & 0xff),
      (byte)( MAGIC     & 0xff)
    };
    int i;
    for ( i = 0; i < len; i++ )
      if ( header[i] != h[i] ) return 0.0;
    return 1.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.png"]) or {@code null}.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.m01",
    };
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "MeshCompressionRLE";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "DataFileFormatToRasterGraphicsAndEntropyCodecAndBitStreamAndOrder2D";

  /** Object description. */
  private final static String DESCRIPTION = "Mesh compression: arbitrary pass, zero skipping, DPCM.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"DataFileFormat" );
      t.newOptOutputPlug( PL_RASTER, IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_STREAM, IFACE+"BitStream" );
      t.newOptOutputPlug( PL_OUTPUT, IFACE+"EntropyCodec" );
      t.newOptOutputPlug( "order",   IFACE+"Order2D" );
        // Property:
      t.propBegin( USE_DPCM, TYPE_BOOLEAN, "Use DPCM (prediction) in amplitude coding?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( MAX_SKIP, TYPE_INTEGER, "Maximal lenght of skip-packet", true );
        t.propDefault( 255 );
        t.propBounds( 1, 65535 );
      t.propEnd();
      t.propBegin( ROBUST, TYPE_BOOLEAN, "Use robust file-format (a little bit longer)?", true );
        t.propDefault( true );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
