package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.obscure.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * This module provides stochastic Monte Carlo methods (hemisphere sampling,
 * area sampling, area-to-area sampling) for form-factor solving.
 *
 * @author  PE, Tomas Svab
 * @version 0.26 $Rev: 756 $ $Date: 2008-05-19 00:10:44 +0200 (po, 19 V 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/MonteCarloFF.java">MonteCarloFF.java</a>
 * @see     RadiosityBasisFunctions
 * @see     Brep
 * @see     BitSet
 * @see     MicroFacet
 */
public class MonteCarloFF extends Piece implements RadiosityBasisFunctions
{
  //==========================================================================
  //  Caching support:

  /** FF row cache. */
  protected LRUCache< double[] > rowCache;

  /** Full row-cache in case of {@link #cacheSize} {@code >= num}. */
  protected double[][] fullCache;

  /** How many rows will be cached? (0 for none) */
  protected int cacheSize  = 0;

  //==========================================================================
  //  Data:

  /** Supported sampling methods. */
  public static enum Sampling
  {
    /** Uniform distribution on the hemisphere. Independent samples, must be weighted! */
    RANDOM_UNIFORM,
    /** Cosine emitter. Independent samples. */
    RANDOM_COSINE,
  }

  /** Actual sampling method. */
  protected Sampling sampling = Sampling.RANDOM_COSINE;

  /** Accelerator: input B-rep database. */
  protected Brep brep;

  /** Accelerator: intersection computing. */
  protected Solid solid;

  /** Accelerator: scene optimalization. */
  protected GeometrySearch geom;

  /** Actual {@link Brep} context. */
  protected int ctx = Brep.CTX_DEFAULT;

  /** Actual vertex-normal attribute. */
  protected int attrNormal = Brep.NULL;

  /** Random number generator. */
  protected final RandomJames rnd = new RandomJames();

  /** Number of rays, which are shooted from one element. */
  protected int numRays = 100;

  /** Element areas. */
  protected double[] area;

  /** Accelerators are initialized? */
  protected boolean isInit = false;

  //==========================================================================
  //  Radiosity distribution (RadiosityBasisFunctions):

  /**
   * Order of the approximation (for information purposes only).
   * Number of coefficients defining radiosity distribution on single element.
   */
  @Override
  public int getOrder ()
  {
    init();
    return 1;
  }

  /**
   * Returns coefficient indexes defining the given element.
   *
   * @param  elem Element handle.
   * @param  result Pre-allocated integer array to hold result (can be {@code null}).
   * @return Size of result array.
   */
  @Override
  public int coefficientsFromElement ( int elem, int[] result )
  {
    init();
    result[0] = elem;
    return getOrder();
  }

  /**
   * Returns elements which are affected by the given single coefficient.
   *
   * @param  coeff Coefficient index.
   * @param  result Pre-allocated integer array to hold result (can be {@code null}).
   * @return Size of result array.
   */
  @Override
  public int elementsFromCoefficient ( int coeff, int[] result )
  {
    init();
    result[0] = coeff;
    return 1;
  }

  /**
   * Returns set of coefficients defining the given set of elements.
   *
   * @param  elements Set of elements.
   * @param  result Non-mandatory object to be filled.
   * @return Set of coefficient indices.
   */
  @Override
  public BitSet coefficientsFromElements ( BitSet elements, BitSet result )
  {
    init();

    if ( elements.isEmpty() )
      return null;

    result = elements;
    return null;
  }

  /**
   * Returns set of elements which are affected by the given set of coefficients.
   *
   * @param  coefficients Set of coefficients.
   * @param  result Non-mandatory object to be filled.
   * @return Set of element handles.
   */
  @Override
  public BitSet elementsFromCoefficients ( BitSet coefficients, BitSet result )
  {
    init();

    if ( coefficients.isEmpty() )
      return null;

    result = coefficients;
    return result;
  }

  /**
   * Computes distribution coefficients for the single point of the element.
   *
   * @param  elem Element handle.
   * @param  coords World-space (?) coordinates of the point lying on the given element.
   * @param  result Non-mandatory pre-allocated array to hold result.
   * @return Distribution multiplicators (coefficient indices match the result
   *         of {@link #coefficientsFromElement}).
   * @todo   Add some other form of distribution ({@link FaceRender}?).
   */
  @Override
  public double[] distribute ( int elem, double[] coords, double[] result )
  {
    init();
    return null;
  }

  //==========================================================================
  //  Form-factors (interface RadiosityBasisFunctions):

  /**
   * Total number of elements in the scene.
   */
  @Override
  public int totalElements ()
  {
    init();
    return brep.getAllFaces( 0, null );
  }

  /**
   * Total number of coefficients.
   */
  @Override
  public int totalCoefficients ()
  {
    init();
    return brep.getAllFaces( 0, null );
  }

  /**
   * Computes "shooting" form-factors (single source coefficient is fixed).
   *
   * @param  coeff Source-coefficient index.
   * @param  ind Array to hold affected (lit) coefficient indices (can be {@code null}).
   * @param  ff Array to hold respective form-factors (can be {@code null}).
   * @return Number of coefficients lit.
   */
  @Override
  public int formFactorsFrom ( int coeff, int[] ind, double[] ff )
  {
    if ( !init() )
      return 0;

    final int num = totalCoefficients();
    if ( ff == null ||
         ff.length < num )
      return 0;

    int i, j;
    double[] cached;

      // look into the full-cache:
    if ( fullCache        != null &&
         fullCache[coeff] != null )
    {
      System.arraycopy( fullCache[coeff], 0, ff, 0, num );
      return num;
    }
      // search the cache:
    if ( rowCache != null )
    {
      cached = rowCache.get( coeff );
      if ( cached != null )
      {
        System.arraycopy( cached, 0, ff, 0, num );
        return num;
      }
    }

    Arrays.fill( ff, 0, num, 0.0 );
    double[][] orig = new double[ numRays ][ 3 ];
    double[][] dir  = new double[ numRays ][ 3 ];
    double[] normal = new double[ 3 ];
    double[][] v    = null;
    double[] weight = generateRays( coeff, orig, dir );
    double wSum = 0.0;
    if ( weight != null )
      for ( double w : weight )
        wSum += w;
    final double denom = 1.0 / ((weight == null) ? numRays : wSum);

      // compute intersection:
    List< MicroFacet > m;
    int face;

    for ( i = 0; i < numRays; i++ )
    {
      m = solid.intersection( orig[i], dir[i] );
      if ( m == null )   // there is no intersection
        continue;

      face = coeff;
        // skip all t <= 0.0 and self-intersections:
      for ( MicroFacet mf : m )
        if ( mf.face != coeff &&
             mf.t > 0.0 )
        {
          face = mf.face;
          break;
        }

      if ( face == coeff )
        continue;

      v = getVertices( face, v );
      Geometry.triangleNormal( v, normal );
      if ( Geometry.dotProduct3D( dir[i], normal ) >= 0.0 )
        continue;

      ff[ face ] += ((weight == null) ? denom : (denom * weight[i]));
    }

      // cache results:
    if ( fullCache != null )
    {
      fullCache[coeff] = new double[ num ];
      System.arraycopy( ff, 0, fullCache[coeff], 0, num );
      return num;
    }

    if ( rowCache != null )
    {
      cached = new double[ num ];
      System.arraycopy( ff, 0, cached, 0, num );
      rowCache.put( coeff, cached );
    }

    return num;
  }

  /**
   * Computes "gathering" form-factors (single target coefficient is fixed).
   *
   * @param  coeff Target-coefficient index.
   * @param  ind Array to hold relevant (visible) coefficient indices (can be {@code null}).
   * @param  ff Array to hold respective form-factors (can be {@code null}).
   * @return Number of relevant coefficients.
   */
  @Override
  public int formFactorsTo ( int coeff, int[] ind, double[] ff )
  {
    final int num = formFactorsFrom( coeff, ind, ff );
    if ( num == 0 ) return 0;

      // convert "shooting" FFs to "gathering" ones:
    assertAreas();
    for ( int i = 0; i < num; i++ )
      ff[i] *= (area[coeff] / area[i]);

    return num;
  }

  /**
   * Computes single form-factor (F_{from,to}).
   * Works only with triangle faces..
   */
  @Override
  public double formFactor ( int from, int to )
  {
    if ( !init() || from == to )
      return 0.0;

    assertAreas();

      // gets vertices:
    int i = from, j = to;
    double[][] vertexI = getVertices( i, null );
    double[][] vertexJ = getVertices( j, null );

      // gets normals:
    double ni[] = Geometry.triangleNormal( vertexI, null );
    double nj[] = Geometry.triangleNormal( vertexJ, null );

      // early-reject checks:
    double[] dirIJ = new double[ 3 ];
    /* Not yet, Cornell-box triangle orientations chould be checked!
    Geometry.vector( vertexI[0], vertexJ[0], dirIJ );
    if ( Geometry.dotProduct3D( dirIJ, ni ) <= 0.0 ||
         Geometry.dotProduct3D( dirIJ, nj ) >= 0.0 )
    {
      Geometry.vector( vertexI[1], vertexJ[0], dirIJ );
      if ( Geometry.dotProduct3D( dirIJ, ni ) <= 0.0 ||
           Geometry.dotProduct3D( dirIJ, nj ) >= 0.0 )
      {
        Geometry.vector( vertexI[2], vertexJ[0], dirIJ );
        if ( Geometry.dotProduct3D( dirIJ, ni ) <= 0.0 ||
             Geometry.dotProduct3D( dirIJ, nj ) >= 0.0 )
          return 0.0;
      }
    }
    */

      // computes form-factor:
    double ff = 0.0;
    List< MicroFacet > m;
    double[] xi = new double[ 3 ];
    double[] xj = new double[ 3 ];
    final double denomAdd = area[j] / numRays;

    for ( int ray = 0; ray < numRays; ray++ )
    {
        // randomly selects point xi on element i and xj on element j:
      rnd.randomPointFromTriangle( vertexI[0], vertexI[1], vertexI[2], xi );
      rnd.randomPointFromTriangle( vertexJ[0], vertexJ[1], vertexJ[2], xj );

        // computes directions:
      Geometry.vector( xi, xj, dirIJ );

        // checks visibility:
      m = solid.intersection( xi, dirIJ );
      if ( m == null ) continue;

      int face = i;
      for ( MicroFacet mf : m )
        if ( mf.face != i &&
             mf.t > 0.0 )
        {
          face = mf.face;
          break;
        }

        // this sample ray did pass through..
      if ( face == j )
      {
        final double radius    =  Geometry.normalize3D( dirIJ );
        final double cosThetaI =  Geometry.dotProduct3D( dirIJ, ni );
        final double cosThetaJ = -Geometry.dotProduct3D( dirIJ, nj );

        if ( cosThetaI <= 0.0 ||
             cosThetaJ <= 0.0 ) continue;

        ff += cosThetaI * cosThetaJ / ( Math.PI * radius * radius + denomAdd );
      }
    }

    return( ff * denomAdd );
  }

  //==========================================================================
  //  Support:

  /** Initialize accelerators.. */
  protected boolean init ( )
  {
    if ( isInit ) return true;
    isInit = true;

    brep  = (Brep)          getInterface( PL_DATA,   IFACE+"Brep" );
    geom  = (GeometrySearch)getInterface( PL_DIRECT, IFACE+"GeometrySearch" );
    solid = (Solid)         getInterface( PL_OUTPUT, IFACE+"Solid" );

    if ( brep  == null ||
         solid == null )
      return false;

    if ( geom != null )
      geom.buildDirectory( Brep.FACE, ctx );

    attrNormal = brep.getAttributeId( Brep.VERTEX, ctx, Brep.NORMAL );

      // row-cache setup:
    final int num = brep.getAllFaces( 0, null );
    if ( cacheSize >= num )                 // full-cache
      fullCache = new double[ num ][];
    else                                    // associative cache
    {
      rowCache = new LRUCache< double[] >( cacheSize );
      rowCache.setCacheCapacity( cacheSize );
    }

    return true;
  }

  /** Computes area of the given element (triangle is presumed) . */
  public double elementArea ( int i )
  {
    double[][] v = getVertices( i, null );
    return Geometry.triangleArea( v );
  }

  /** Return coords of all vertices of the triangle. */
  protected double[][] getVertices ( int handle, double[][] vertex )
  {
    if ( vertex == null )
      vertex = new double[ 3 ][ 3 ];

    int[] vertexID = new int[ 3 ];
    brep.getFaceVertices( handle, vertexID );

    for ( int i = 0; i < 3; i++ )
      brep.getVertexCoords( ctx, vertexID[i], vertex[i] );

    return vertex;
  }

  /** Computes areas of all elements (if necessary). */
  protected void assertAreas ()
  {
    if ( area != null ) return;

    final int elems = totalElements();
    area = new double[ elems ];
    for ( int i = 0; i < elems; i++ )
      area[i] = elementArea( i );
  }

  //==========================================================================
  //  Compute of FF:

  /**
   * Generates required number of rays for shooting throught the hemisphere.
   *
   * @param coeff      ID of element, which is shooting ray.
   * @param origin     Pre-allocated array to hold sampling points on element
   *                   in world coordinates.
   * @param direction  Pre-allocated array to hold sampling points
   *                   on hemisphere or area in world coordinates.
   */
  protected double[] generateRays ( int coeff, double[][] origin, double[][] direction )
  {
    if ( origin    == null || origin[0].length    < 3 ||
         direction == null || direction[0].length < 3 )
      return null;

      // world-coordinates of the vertices:
    final double[][] v = getVertices( coeff, null );

      // sample planar triangle:
    for ( int i = 0; i < numRays; i++ )
      rnd.randomPointFromTriangle( v[0], v[1], v[2], origin[i] );

      // sampling on the hemisphere or area:
    return randomWeighted( v, direction );
  }

  /**
   * Generate uniform random directions by random generating numbers.
   *
   * @param  v   World-space coordinates of the triangle vertices.
   * @param  dir Pre-allocated array to hold normalized direction of ray in tangent space.
   * @return Weight vector or {@code null} if all rays are equally important.
   */
  protected double[] randomWeighted ( double[][] v, double dir[][] )
  {
    if ( dir           == null    ||
         dir.length    != numRays ||
         dir[0].length < 3 )
      return null;

    int i;
    double[] w = null;

    switch ( sampling )
    {
      case RANDOM_UNIFORM:
        w = new double[ numRays ];
        for ( i = 0; i < numRays; i++ )
        {
          rnd.uniformDirection( 0.0, 1.0, dir[i] );
          w[i] = dir[i][2];
        }
        break;

      case RANDOM_COSINE:
        for ( i = 0; i < numRays; i++ )
        {
          double rad2, x, y;
          do
          {
            x = rnd.uniformNumber() * 2.0 - 1.0;
            y = rnd.uniformNumber() * 2.0 - 1.0;
            rad2 = x * x + y * y;
          }
          while ( rad2 >= 1.0 );
          dir[i][0] = x;
          dir[i][1] = y;
          dir[i][2] = Math.sqrt( 1.0 - rad2 );
        }
        break;
    }

      // I. compute the world -> tangent matrix
    final TrMatrix m = new TrMatrix( 3, 3 );          // rows: tangent, binormal, normal
    Geometry.vector( v[0], v[1], m.m[0] );
    Geometry.normalize3D( m.m[0] );                   // tangent
    Geometry.triangleNormal( v, m.m[2] );             // normal
    Geometry.vectorProduct( m.m[0], m.m[2], m.m[1] ); // binormal = tangent X normal

      // II. transform all the directions
    for ( i = 0; i < numRays; i++ )
      m.transformVector3D( dir[i], dir[i] );

    return w;
  }

  //==========================================================================
  //  Interface Property:

  public final static String SAMPLING    = "Sampling type";
  public final static String NUMBER_RAYS = "Number of rays";
  public final static String CACHE_SIZE  = "Cache size";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // SAMPLING:
    if ( key.compareTo(SAMPLING) == 0 )
    {
      sampling = enumProperty( value, sampling );
      return;
    }
      // NUMBER_OF_RAYS:
    if ( key.compareTo(NUMBER_RAYS) == 0 )
    {
      numRays = intProperty( value, numRays    );
      return;
    }
      // CACHE_SIZE:
    if ( key.compareTo(CACHE_SIZE) == 0 )
    {
      cacheSize = intProperty( value, cacheSize );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(SAMPLING) == 0 )
      return sampling.ordinal();
    if ( key.compareTo(NUMBER_RAYS) == 0 )
      return numRays;
    if ( key.compareTo(CACHE_SIZE) == 0 )
      return cacheSize;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Monte-Carlo Form Factor";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "RadiosityBasisFunctionsToBrepAndSolidAndGeometrySearch";

  /** Object description. */
  private final static String DESCRIPTION = "Monte Carlo methods for form-factors";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + '.' + C_RADIOSITY;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"RadiosityBasisFunctions" );
      t.newOutputPlug(    PL_DATA,   IFACE+"Brep" );
      t.newOutputPlug(    PL_OUTPUT, IFACE+"Solid" );
      t.newOptOutputPlug( PL_DIRECT, IFACE+"GeometrySearch" );
        // Property:
      t.propBegin( SAMPLING, TYPE_INTEGER, "Sampling method", true );
        t.propDefault( Sampling.RANDOM_UNIFORM.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Random uniform", Sampling.RANDOM_UNIFORM.ordinal(), "Independent random samples, uniform distribution" );
        t.propEnum( "Random cosine",  Sampling.RANDOM_COSINE.ordinal(),  "Cosine emitter, independent random samples" );
      t.propEnd();
      t.propBegin( NUMBER_RAYS, TYPE_INTEGER, "Number of shooted rays", true  );
        t.propDefault( 100 );
      t.propEnd();
      t.propBegin( CACHE_SIZE, TYPE_INTEGER, "How many rows/cols can be cached?", true  );
        t.propDefault( 0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
