package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.GraphicsViewer;
import cz.cuni.jagrlib.reg.*;

/**
 * Rectangle image frame defined by interactive mouse input.
 *
 * @version 0.25 $Rev: 406 $ $Date: 2007-03-25 00:06:58 +0100 (ne, 25 III 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/MouseFrame.java">MouseFrame.java</a>
 */
public class MouseFrame extends DefaultInputListener
{
  //==========================================================================
  //  Data:

  /** X coordinate of frame origin (or {@link Integer#MIN_VALUE} if no frame is defined). */
  protected int x0 = Integer.MIN_VALUE;

  /** Y coordinate of frame origin. */
  protected int y0 = 0;

  /** X coordinate of opposite frame corner. */
  protected int x1 = 0;

  /** Y coordinate of opposite frame corner. */
  protected int y1 = 0;

  /** X coordinate of right mouse dragging. */
  protected int xr = 0;

  /** Y coordinate of right mouse dragging. */
  protected int yr = 0;

  /** Left mouse button - stretching. */
  protected boolean stretching = false;

  /** Right mouse button - dragging. */
  protected boolean dragging = false;

  /** Accelerator - viewer window reference. */
  protected GraphicsViewer gw = null;

  /** X coordinate of invalidated point. */
  protected int xi = 0;

  /** Y coordinate of invalidated point. */
  protected int yi = 0;

  /** Need to repaint the whole canvas? */
  protected boolean all = false;

  //==========================================================================
  //  Support:

  /** Sets actual HUD frame into attached viewer window. */
  protected void setFrame ()
  {
    if ( gw == null )
      gw = (GraphicsViewer)getInterface( PL_OUTPUT, IFACE+"GraphicsViewer" );

    if ( gw == null )
      return;

    if ( x0 == Integer.MIN_VALUE )
    {
      gw.hudClear();
      gw.repaintAll();
      all = false;
      return;
    }

    gw.hudSetLine( 0, x0, y0, x1, y0 );
    gw.hudSetLine( 1, x1, y0, x1, y1 );
    gw.hudSetLine( 2, x1, y1, x0, y1 );
    gw.hudSetLine( 3, x0, y1, x0, y0 );

    if ( all )
    {
      gw.repaintAll();
      all = false;
    }
    else
    {
      gw.invalidate( Math.min( xi, Math.min( x0, x1 ) ),
                     Math.max( xi, Math.max( x0, x1 ) ) + 1,
                     Math.min( yi, Math.min( y0, y1 ) ),
                     Math.max( yi, Math.max( y0, y1 ) ) + 1 );
      gw.repaint();
    }
  }

  //==========================================================================
  //  Interface InputListener:

  /**
   * Mouse button was pressed/released.
   *
   * @param sender Originator of the event
   * @param when System time in milliseconds.
   * @param x Horizontal screen position in pixels.
   * @param y Vertical screen position in pixels.
   * @param button Button number ({@link #BUTTON1}, etc.).
   * @param press Was it press or release?
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public void mouseButton ( Object sender, long when, int x, int y, int button, boolean press, int flags )
  {
    if ( button == BUTTON1 )              // left mouse button
    {
      x1 = x;
      y1 = y;

      if ( (stretching = press) )
      {
        x0  = x;
        y0  = y;
        all = true;
      }
      else
        if ( x0 == x1 &&
             y0 == y1 )
        {
          x0       = Integer.MIN_VALUE;
          dragging = false;
        }
    }

    if ( button == BUTTON3 )              // right mouse button
      if ( dragging = (press && x0 != Integer.MIN_VALUE) )
      {
        xr = x;
        yr = y;
      }

    setFrame();
  }

  /**
   * Mouse pointer was moved.
   *
   * @param sender Originator of the event
   * @param when System time in milliseconds.
   * @param x Horizontal screen position in pixels.
   * @param y Vertical screen position in pixels.
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public void mouseMove ( Object sender, long when, int x, int y, int flags )
  {
    if ( dragging )
    {
      int dx = x - xr;
      int dy = y - yr;
      xr     = x;
      yr     = y;

      xi     = (dx > 0) ? Math.min( x0, x1 ) : Math.max( x0, x1 );
      yi     = (dy > 0) ? Math.min( y0, y1 ) : Math.max( y0, y1 );

      x0    += dx;
      x1    += dx;
      y0    += dy;
      y1    += dy;
    }

    if ( stretching )
    {
      xi = x1;
      yi = y1;

      x1 = x;
      y1 = y;
    }

    if ( stretching ||
         dragging )
      setFrame();
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "MouseFrame";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "InputListenerToGraphicsViewer";

  /** Object description. */
  private final static String DESCRIPTION = "Mouse events consumer defines rectangular frame.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + ".interaction";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"InputListener" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"GraphicsViewer" );
        // Property:
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
