package cz.cuni.jagrlib.piece;

import java.util.*;
import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Wavefront OBJ file format (persistence of B-rep 3D data).
 *
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/OBJFileFormat.java">OBJFileFormat.java</a>
 */
public class OBJFileFormat extends DefaultBrepFileFormat
{
  //==========================================================================
  //  Support classes:

  /** Vertex triple [ coords, texture, normal ]. */
  protected static class Vertex
  {
    /** Coordinate handle. */
    public int coord;

    /** Texture-coordinate handle ({@code -1} if not present). */
    public int texture;

    /** Normal vector handle ({@code -1} if not present). */
    public int normal;

    public Vertex ( int coo, int tex, int nor )
    {
      coord   = coo;
      texture = tex;
      normal  = nor;
    }

    @Override
    public final boolean equals ( Object obj )
    {
      if ( obj == null ||
           !(obj instanceof Vertex) ) return false;

      final Vertex v = (Vertex)obj;
      return( v.coord   == coord   &&
              v.texture == texture &&
              v.normal  == normal );
    }

    @Override
    public final int hashCode ()
    {
      return (int)(coord + 10103L * normal + 10000151L * texture);
    }
  }

  //==========================================================================
  //  Data:

  /** Triangulate all read faces? */
  protected boolean triangulate = true;

  /** Change coordinate orientation? */
  protected boolean reorient = true;

  /**
   * Id of solid read from input stream. All faces will go into this solid.
   * If {@link Brep#NULL}, no solid identifiers will be set..
   */
  protected int solid = Brep.NULL;

  //==========================================================================
  //  Support routines:

  /** Parses double number, returns {@code true} if succeeded. */
  protected static boolean parseDouble ( String token, double[] arr, int i )
  {
    try
    {
      arr[i] = Double.parseDouble( token );
    }
    catch ( NumberFormatException ex )
    {
      LogFile.log( "OBJ load: error in number (" + token + ')' );
      return false;
    }
    return true;
  }

  /**
   * Parses double-precision vector starting at {@code token[1]},
   * stores result in the {@code arr} array, returns its length.
   */
  protected static int parseVector ( String[] tokens, int len, double[] arr )
  {
    int i = 1;
    int k = 0;
    while ( i < len &&
            k < arr.length &&
            parseDouble( tokens[i], arr, k ) )
    {
      i++;
      k++;
    }

    len = k;
    while ( k < arr.length )
      arr[k++] = 0.0;

    return len;
  }

  /**
   * Parse one face, parsed vertices are written into the {@code vert} array,
   * their number is returned.
   */
  protected static int parseFace ( String[] tokens, int len, int vertices, int textures, int normals, Vertex[] vert )
  {
    int i;
    int dst = 0;
    for ( i = 1; i < len && (tokens[i].length() == 0 || tokens[i].charAt( 0 ) != '#') ; i++ )
    {                                     // parse one token
      String[] ind = tokens[i].split( "/" ); // "v" ["vt" ["vn"]]
        // vertex coordinates (mandatory)
      int coord = 0;
      try
      {
        coord = Integer.parseInt( ind[0] );
      }
      catch ( NumberFormatException ex )
      {
        LogFile.log( "OBJ load: error in index (" + ind[0] + ')' );
        continue;
      }
      if ( coord == 0 ||
           Math.abs( coord ) > vertices )
      {
        LogFile.log( "OBJ load: bad coord index (" + coord + ')' );
        continue;
      }
      if ( coord > 0 )                    // positive index starting from 1
        coord--;
      else                                // negative (relative) index: -1 means the last vertex
        coord += vertices;
        // coord is valid vertex coordinate index!

      int tex    = -1;
      int normal = -1;
      if ( ind.length > 1 )               // are there another parts?
      {
        if ( ind[1].length() > 0 )        // trying to parse texture coordinates..
        {
          try
          {
            tex = Integer.parseInt( ind[1] );
          }
          catch ( NumberFormatException ex )
          {
            tex = 0;
          }
          if ( tex == 0 ||
               Math.abs( tex ) > textures )
            tex = -1;
          else
            if ( tex > 0 )                // positive index starting from 1
              tex--;
            else                          // negative (relative) index: -1 means the last texture
              tex += textures;
        }
        if ( ind.length > 2 )             // are there another parts?
        {
          if ( ind[2].length() > 0 )      // trying to parse normal..
          {
            try
            {
              normal = Integer.parseInt( ind[2] );
            }
            catch ( NumberFormatException ex )
            {
              normal = 0;
            }
            if ( normal == 0 ||
                 Math.abs( normal ) > normals )
              normal = -1;
            else
              if ( normal > 0 )           // positive index starting from 1
                normal--;
              else                        // negative (relative) index: -1 means the last normal
                normal += normals;
          }
        }
      }

        // the whole triple is valid: [ coord, tex, normal ]
        // .. write it to the output array:
      vert[ dst++ ] = new Vertex( coord, tex, normal );
    }

    return dst;
  }

  /**
   * Common load code.
   *
   * @param  stream Opened input bit-stream.
   * @param  b Checked B-rep database object.
   * @throws IOException
   */
  @Override
  protected void commonLoad ( BitStream stream, Brep b )
  throws IOException
  {
    final InputStream objFile = stream.getInputStream();
    if ( objFile == null )
      throw new FileNotFoundException();

      // B-rep setup:
      // !!! TODO: add a new solid to nonempty B-rep database ???
    b.setCoordTypes( Brep.ATTR_COORDH_4D, Brep.ATTR_COORDH_4D );
    b.init();
    int faceColor    = b.createAttribute( Brep.FACE,   Brep.NULL, Brep.COLOR,     Brep.ATTR_RGB_3F );
    int vertexNormal = b.createAttribute( Brep.VERTEX, Brep.NULL, Brep.NORMAL,    Brep.ATTR_COORDH_4D );
    int vertexTex    = b.createAttribute( Brep.VERTEX, Brep.NULL, Brep.TEXTURE_0, Brep.ATTR_COORD_2D );
    int ctx          = Brep.CTX_DEFAULT;

      // all vertex-coordinates in one array:
    final ArrayList< double[] > vertices  = new ArrayList< double[] >();
      // unused vertices:
    final BitSet unusedVertices           = new BitSet();
      // all texture-coordinates in one array:
    final ArrayList< double[] > texCoords = new ArrayList< double[] >();
      // all normal-vectors in one array:
    final ArrayList< double[] > normals   = new ArrayList< double[] >();
      // actual smooth-group:
    int smoothGroup = 0;
      // all materials in single name-space:
    final HashMap< String, PhongModel > materials = new HashMap< String, PhongModel >();
      // actual material:
    PhongModel material = null;
      // B-rep vertices:
    final HashMap< Vertex, Integer > brepVertices = new HashMap< Vertex, Integer >();

    String line;                          // actual input line
    double[] arr  = new double[ 4 ];      // coordinate array
    Vertex[] vert = new Vertex[ 3 ];      // vertex array (for new vertices - not registered yet)
    int[]    tv   = new int[ 3 ];         // face vertex-handles
    float[]  col  = { 1.0f, 0.6f, 0.2f }; // default face color
    double   tmp;

      // line-loop:
    lineLoop:
    do
    {                                     // process one non-empty input line
      do
      {                                   // find next non-empty & non-comment line
        if ( (line = readLine( objFile )) == null )
          break lineLoop;

        line = line.trim();
      }
      while ( line.length()    == 0 ||
              line.charAt( 0 ) == '#' );

      String[] tokens = line.split( "\\p{javaWhitespace}+" );
      int len = 0;
      while ( len < tokens.length &&
              (tokens[len].length() < 1 ||
               tokens[len].charAt( 0 ) != '#') ) len++;

      if ( len < 2 ) continue;

        // "v" - vertex coordinates: "v x y z [w]"
      if ( tokens[0].equalsIgnoreCase( "v" ) )
      {
        if ( (len = parseVector( tokens, len, arr )) < 1 ) continue;
        if ( len < 4 ) arr[ 3 ] = 1.0;
        if ( reorient )
        {
          tmp    = arr[1];
          arr[1] = arr[2];
          arr[2] = tmp;
        }
        unusedVertices.set( vertices.size() );
        vertices.add( arr );
        arr = new double[ 4 ];
        continue;
      }

        // "vn" - normal vector: "vn x y z"
      if ( tokens[0].equalsIgnoreCase( "vn" ) )
      {
        if ( parseVector( tokens, len, arr ) < 1 ) continue;
        if ( reorient )
        {
          tmp    = arr[1];
          arr[1] = arr[2];
          arr[2] = tmp;
        }
        normals.add( arr );
        arr = new double[ 4 ];
        continue;
      }

        // "vt" - texture coordinates: "vt s [t [u]]"
      if ( tokens[0].equalsIgnoreCase( "vt" ) )
      {
        final int texLen = parseVector( tokens, len, arr );
        if ( texLen < 1 ) continue;
        double[] tx = new double[] { arr[0], arr[1] };
        texCoords.add( tx );
        continue;
      }

        // "f" - face definition: "f {v[/[vt][/vn]]}+"
      if ( tokens[0].equalsIgnoreCase( "f" ) )
      {
        if ( len < 4 )
        {
          LogFile.log( "OBJ load: not enough face vertices (" + (len-1) + ')' );
          continue;
        }

          // trying to parse len-1 vertices:
        if ( len - 1 > vert.length )
          vert = new Vertex[ len - 1 ];
        final int faceLen = parseFace( tokens, len, vertices.size(), texCoords.size(), normals.size(), vert );
        int i;
        if ( faceLen < 3 )
        {
          LogFile.log( "OBJ load: error parsing face" );
          continue;
        }

          // register not-yet-registered vertices:
        for ( i = 0; i < faceLen; i++ )
          if ( brepVertices.get( vert[i] ) == null )
          {
            unusedVertices.clear( vert[i].coord );
            int vertId = b.createVertex();
            b.setVertexCoords( ctx,          vertId, vertices.get( vert[i].coord ) );
            if ( vert[i].normal >= 0 )
              b.setAttribute(  vertexNormal, vertId, normals.get( vert[i].normal ) );
            if ( vert[i].texture >= 0 )
              b.setAttribute(  vertexTex,    vertId, texCoords.get( vert[i].texture ) );
            brepVertices.put( vert[i], vertId );
          }

        int face;
        if ( triangulate )                // break the convex polygon into triangle faces
        {
            // create faceLen-2 triangular faces:
          int iUp = 0;                    // triangle A [ iDn, iUp, iUp+1 ] iUp++
          int iDn = faceLen - 1;          // triangle B [ iDn, iUp, iDn-1 ] iDn--
          do
          {
              // triangle A
            tv[0] = brepVertices.get( vert[iDn]   ).intValue();
            tv[1] = brepVertices.get( vert[iUp++] ).intValue();
            tv[2] = brepVertices.get( vert[iUp]   ).intValue();
            b.setFaceVertices( face = b.createFace(), tv );
            b.setAttribute( faceColor, face, col );
              // triangle B
            if ( iUp < iDn - 1 )
            {
              tv[1] = tv[2];
              tv[2] = brepVertices.get( vert[--iDn] ).intValue();
              b.setFaceVertices( face = b.createFace(), tv );
              b.setAttribute( faceColor, face, col );
            }
          }
          while ( iUp < iDn - 1 );
        }
        else
        {
          if ( faceLen != tv.length )
            tv = new int[ faceLen ];
          for ( i = 0; i < faceLen; i++ )
            tv[i] = brepVertices.get( vert[i] ).intValue();
          b.setFaceVertices( face = b.createFace(), tv );
          b.setAttribute( faceColor, face, col );
        }

        if ( (col[2] += 0.01f) >= 1.0f )  // !!! TODO: face colors by OBJ materials !!!
          col[2] = 0.0f;
      }
    }
    while ( true );

      // register unused vertices..
    int v = -1;
    while ( (v = unusedVertices.nextSetBit( v+1 )) >= 0 )
    {
      int vertId = b.createVertex();
      b.setVertexCoords( ctx, vertId, vertices.get( v ) );
    }
  }

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  b Checked B-rep database object.
   * @throws IOException
   */
  protected void commonSave ( BitStream stream, Brep b )
  throws IOException
  {
    final OutputStream objFile = stream.getOutputStream();
    if ( objFile == null )
      throw new FileNotFoundException();

    int vertexNormal = b.getAttributeId( Brep.VERTEX, Brep.NULL, Brep.NORMAL );
    int vertexTex    = b.getAttributeId( Brep.VERTEX, Brep.NULL, Brep.TEXTURE_0 );
    int ctx          = Brep.CTX_DEFAULT;
    int handle;
    double[] data    = null;

      // vertex-geometry section:
    writeLine( objFile, "# vertex coords:" );
    BrepIterator it = b.vertexIterator( null );
    while ( (handle = it.next()) != Brep.NULL )
    {
      data = b.getVertexCoords( ctx, handle, data );
      writeLine( objFile, String.format( Locale.US, "v %f %f %f", data[0], data[1], data[2] ) );
    }

      // vertex-normal section:
    if ( vertexNormal != Brep.NULL )
    {
      writeLine( objFile, "# vertex normals:" );
      it.reset();
      while ( (handle = it.next()) != Brep.NULL )
      {
        b.getAttribute( vertexNormal, handle, data );
        writeLine( objFile, String.format( Locale.US, "vn %f %f %f", data[0], data[1], data[2] ) );
      }
    }

      // vertex-texture section:
    if ( vertexTex != Brep.NULL )
    {
      writeLine( objFile, "# vertex texture coords:" );
      it.reset();
      while ( (handle = it.next()) != Brep.NULL )
      {
        b.getAttribute( vertexTex, handle, data );
        writeLine( objFile, String.format( Locale.US, "vt %f %f", data[0], data[1] ) );
      }
    }

      // faces:
    int[] vert  = new int[ 3 ];
    int faceSize;
    StringBuffer sb = new StringBuffer( "f" );
    String fmt;
    if ( vertexTex != Brep.NULL )
      if ( vertexNormal != Brep.NULL )
        fmt = " %d/%d/%d";
      else
        fmt = " %d/%d";
    else
      if ( vertexNormal != Brep.NULL )
        fmt = " %d//%3$d";                // to be sure!
      else
        fmt = " %d";

    it = b.faceIterator( ctx, it );
    while ( (handle = it.next()) != Brep.NULL ) // one face
    {
      while ( (faceSize = b.getFaceVertices( handle, vert )) > vert.length )
        vert = new int[ faceSize ];
      if ( faceSize < 3 ) continue;

        // faceSize items in vert[] array:
      sb.setLength( 1 );
      for ( int i = 0; i < faceSize; i++ )
      {
        handle = vert[i] + 1;
        sb.append( String.format( fmt, handle, handle, handle ) ); // geometry, texture, normal vector
      }
      writeLine( objFile, sb.toString() );
    }
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return 0;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type ({@code 0.0} .. no way,
   *         {@code 0.5} .. maybe, {@code 0.9} .. almost sure, {@code 1.0} ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int pos = fileName.lastIndexOf( '.' );
    if ( pos < 0 ) return 0.0;

    String suffix = fileName.substring( pos );
    if ( suffix.compareToIgnoreCase( "obj" ) == 0 ) return 0.9;

    return 0.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.png"]) or {@code null}.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.obj",
    };
  }

  //==========================================================================
  //  Interface Property:

  public final static String TRIANGULATE = "Triangulate";
  public final static String REORIENT    = "Reorient";
  public final static String SOLID       = "Solid id";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

       // TRIANGULATE:
    if ( key.compareTo(TRIANGULATE) == 0 )
    {
      triangulate = booleanProperty( value, triangulate );
      return;
    }
      // REORIENT:
    if ( key.compareTo(REORIENT) == 0 )
    {
      reorient = booleanProperty( value, reorient );
      return;
    }
       // SOLID:
    if ( key.compareTo(SOLID) == 0 )
    {
      solid = intProperty( value, solid, -1, Integer.MAX_VALUE );
      return;
    }
 }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(TRIANGULATE) == 0 )
      return triangulate;
    if ( key.compareTo(REORIENT) == 0 )
      return reorient;
    if ( key.compareTo(SOLID) == 0 )
      return solid;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "OBJ file format";

  /** Object description. */
  private final static String DESCRIPTION = "Wavefront OBJ filter (3D B-rep data).";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"DataFileFormat" );
      t.newOutputPlug(    PL_DATA,   IFACE+"Brep" );
      t.newOptOutputPlug( PL_STREAM, IFACE+"BitStream" );
        // Property:
      t.propBegin( TRIANGULATE, TYPE_BOOLEAN, "Triangulate every read face?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( REORIENT, TYPE_BOOLEAN, "Reorient coordinate system?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( SOLID, TYPE_INTEGER, "Solid identifier or -1", true );
        t.propDefault( -1 );
        t.propBounds(  -1, Integer.MAX_VALUE );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}