package cz.cuni.jagrlib.piece;

import java.util.concurrent.ArrayBlockingQueue;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Rectangle image ordering using various Space-Filling-Curves (SFC).
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/OrderSFC.java">OrderSFC.java</a>
 */
public class OrderSFC extends Piece implements Order2D
{
  //==========================================================================
  //  Data:

  /** Implemented order types. */
  protected static enum OrderType
  {
    /** Morton ordering = Z-order (2<sup>N</sup>). */
    ORDER_MORTON,
    /** Peano-Hilbert curve (2<sup>N</sup>). */
    ORDER_HILBERT,
    /** Peano curve (3<sup>N</sup> x 2*3<sup>N-1</sup>). */
    ORDER_PEANO,
    /** Meandering Peano curve (3<sup>N</sup>). */
    ORDER_M_PEANO,
  }

  /** FIFO item (holding coordinates of one sample.). */
  protected static class QItem
  {
    /** X-coordinate of the sample or {@code -1} if no sample is available. */
    public int x;

    /** Y-coordinate of the sample. */
    public int y;

    public QItem ( int x, int y )
    {
      this.x = x;
      this.y = y;
    }
  }

  /** Actual order type. */
  protected OrderType type = OrderType.ORDER_MORTON;

  /** Actual bitmap width in pixels. */
  protected int width = 0;

  /** Actual bitmap height in pixels. */
  protected int height = 0;

  /** Number of samples yet to deliver. */
  protected int toDeliver = 0;

  /** Producer &lt;-&gt; consumer communication buffer. */
  protected ArrayBlockingQueue< QItem > fifo = new ArrayBlockingQueue< QItem >( 256 );

  /** Continue-flag for producer thread. */
  protected volatile boolean produce = false;

  /** Actual SFC producer thread. */
  protected Thread producer = null;

  //==========================================================================
  //  Support:

  /** Stops producer thread. */
  protected void stopProducer ()
  {
    produce = false;
    if ( producer != null )
    {
      //LogFile.log( "Forced producer stop, id = " + producer.getId() );
      producer.interrupt();
      producer  = null;
      toDeliver = 0;
    }
  }

  //==========================================================================
  //  Morton order producer thread:

  /** Producer thread for the Morton order (Z-order). */
  protected class Morton extends Thread
  {
    public Morton ()
    {
      super( "JaGrLib - Morton producer" );
    }

    /** Recursive SFC generation function. */
    protected boolean morton ( int x, int y, int size )
    {
      if ( size == 1 )
      {
        if ( x < width &&
             y < height )
          try
          {
            fifo.put( new QItem( x, y ) );
          }
          catch ( InterruptedException e )
          {
            return false;
          }
        return produce;
      }

      size >>= 1;
      return( morton( x,        y,        size ) &&
              morton( x + size, y,        size ) &&
              morton( x,        y + size, size ) &&
              morton( x + size, y + size, size ) );
    }

    @Override
    public void run ()
    {
      int size = 1;
      while ( size < width ||
              size < height )
        size += size;

      morton( 0, 0, size );
    }
  }

  //==========================================================================
  //  Peano-Hilbert order producer thread:

  /** Producer thread for the Peano-Hilbert order. */
  protected class Hilbert extends Thread
  {
    public Hilbert ()
    {
      super( "JaGrLib - Hilbert producer" );
    }

    /**
     * Recursive SFC generation function.
     *
     * Adopted from: Greg Breinholt, "Algorithm 781: generating Hilbert's space-filling curve by recursion"
     * ACM Transactions on Mathematical Software (TOMS) archive,
     * Volume 24, Issue 2 (June 1998), pp. 184 - 189.
     *
     * @param  x Initial X-coordinate.
     * @param  y Initial Y-coordinate.
     * @param  lg Curve width (2<sup>m</sup>).
     * @param  i1 Starting point of the unit shape.
     * @param  i2 End point of the unit shape.
     */
    protected boolean hilbert ( int x, int y, int lg, int i1, int i2 )
    {
      if ( lg == 1 )
      {                                   // Unit shape reached
        if ( x < width &&
             y < height )
          try
          {
            fifo.put( new QItem( x, y ) );
          }
          catch ( InterruptedException e )
          {
            return false;
          }
        return produce;
      }

      lg >>= 1;
      return( hilbert( x +     i1*lg, y +     i1*lg, lg, i1,   1-i2 ) &&
              hilbert( x +     i2*lg, y + (1-i2)*lg, lg, i1,   i2 ) &&
              hilbert( x + (1-i1)*lg, y + (1-i1)*lg, lg, i1,   i2 ) &&
              hilbert( x + (1-i2)*lg, y +     i2*lg, lg, 1-i1, i2 ) );
    }

    @Override
    public void run ()
    {
      int size = 1;
      while ( size < width ||
              size < height )
        size += size;

      hilbert( 0, 0, size, 0, 0 );
    }
  }

  //==========================================================================
  //  Peano order producer thread:

  /** Producer thread for the Peano order. */
  protected class Peano extends Thread
  {
    public Peano ()
    {
      super( "JaGrLib - Peano producer" );
    }

    /**
     * Recursive SFC generation function.
     *
     * @param  x Initial X-coordinate.
     * @param  y Initial Y-coordinate.
     * @param  sizeX Rectangle width (3<sup>m</sup>).
     * @param  sizeY Rectangle height (2*3<sup>m</sup>).
     * @param  backX Reverse horizontal direction?
     * @param  backY Reverse vertical direction?
     */
    protected boolean peano ( int x, int y, int sizeX, int sizeY, boolean backX, boolean backY )
    {
      if ( sizeX == 1 )
      {                                   // Unit shape reached
        if ( x < width &&
             y < height )
        {
          if ( backY )
          {
            if ( y + 1 < height )
              try
              {
                fifo.put( new QItem( x, y + 1 ) );
              }
              catch ( InterruptedException e )
              {
                return false;
              }
            if ( !produce ) return false;
          }

          try
          {
            fifo.put( new QItem( x, y ) );
          }
          catch ( InterruptedException e )
          {
            return false;
          }
          if ( !produce ) return false;

          if ( !backY )
          {
            if ( y + 1 < height )
              try
              {
                fifo.put( new QItem( x, y + 1 ) );
              }
              catch ( InterruptedException e )
              {
                return false;
              }
          }
        }
        return produce;
      }

      if ( sizeX > sizeY )                // wide rectangle => horizontal split
      {
        sizeX /= 3;
        if ( backX )
            // right-to-left
          return( peano( x + sizeX + sizeX, y, sizeX, sizeY, true,  backY ) &&
                  peano( x + sizeX,         y, sizeX, sizeY, true, !backY ) &&
                  peano( x,                 y, sizeX, sizeY, true,  backY ) );

          // left-to-right
        return( peano( x,                 y, sizeX, sizeY, false,  backY ) &&
                peano( x + sizeX,         y, sizeX, sizeY, false, !backY ) &&
                peano( x + sizeX + sizeX, y, sizeX, sizeY, false,  backY ) );
      }

                                          // tall rectangle => vertical split
      sizeY /= 3;
      if ( backY )
          // bottom-to-up
        return( peano( x, y + sizeY + sizeY, sizeX, sizeY,  backX, true ) &&
                peano( x, y + sizeY,         sizeX, sizeY, !backX, true ) &&
                peano( x, y,                 sizeX, sizeY,  backX, true ) );

        // top-to-bottom
      return( peano( x, y,                 sizeX, sizeY,  backX, false ) &&
              peano( x, y + sizeY,         sizeX, sizeY, !backX, false ) &&
              peano( x, y + sizeY + sizeY, sizeX, sizeY,  backX, false ) );
    }

    @Override
    public void run ()
    {
      int sizeX = 1;
      int sizeY = 2;
      while ( sizeX < width ||
              sizeY < height )
        if ( sizeX < sizeY )
          sizeX += (sizeX + sizeX);
        else
          sizeY += (sizeY + sizeY);

      peano( 0, 0, sizeX, sizeY, false, false );
    }
  }

  //==========================================================================
  //  Meandering Peano order producer thread:

  /** Producer thread for the Meandering Peano order. */
  protected class MPeano extends Thread
  {
    public MPeano ()
    {
      super( "JaGrLib - m-Peano producer" );
    }

    /**
     * Recursive SFC generation function.
     *
     * For m-Peano curve see: John M. Dennis, "Partitioning with Space-Filling Curves on the Cubed-Sphere"
     * International Parallel and Distributed Processing Symposium (IPDPS;03) Nice, France 2003, p269a
     * (www.cisl.ucar.edu/css/papers/sfc3.pdf)
     *
     * @param  x Initial X-coordinate.
     * @param  y Initial Y-coordinate.
     * @param  size Rectangle size (3<sup>m</sup>).
     * @param  majX Major axis x-component.
     * @param  majY Major axis y-component.
     * @param  minX Minor axis x-component.
     * @param  minY Minor axis y-component.
     */
    protected boolean mPeano ( int x, int y, int size, int majX, int majY, int minX, int minY )
    {
      if ( size == 1 )
      {                                   // Unit shape reached
        if ( x < width &&
             y < height )
          try
          {
            fifo.put( new QItem( x, y ) );
          }
          catch ( InterruptedException e )
          {
            //LogFile.log( "Producer stopped: x = " + x + ", y = " + y + ", id = " + this.getId() );
            return false;
          }
        return produce;
      }

      size /= 3;
      final int minXs = minX * size;
      final int minYs = minY * size;
      final int majXs = majX * size;
      final int majYs = majY * size;
      final int backX = -majX - minX;
      final int backY = -majY - minY;

      return( mPeano( x,                                                 y,                                                 size,  minX,  minY,  majX,  majY ) &&
              mPeano( x + minXs,                                         y + minYs,                                         size,  minX,  minY,  majX,  majY ) &&
              mPeano( x + minXs + minXs,                                 y + minYs + minYs,                                 size,  majX,  majY,  minX,  minY ) &&
              mPeano( x + minXs + minXs + majXs,                         y + minYs + minYs + majYs,                         size,  majX,  majY,  minX,  minY ) &&
              mPeano( x + minXs + minXs + majXs + majXs,                 y + minYs + minYs + majYs + majYs,                 size,  majX,  majY,  minX,  minY ) &&
              mPeano( x + minXs + minXs + majXs + majXs + majXs + backX, y + minYs + minYs + majYs + majYs + majYs + backY, size, -majX, -majY, -minX, -minY ) &&
              mPeano( x + minXs + minXs + majXs + majXs + backX,         y + minYs + minYs + majYs + majYs + backY,         size, -minX, -minY, -majX, -majY ) &&
              mPeano( x + minXs + majXs + majXs + backX,                 y + minYs + majYs + majYs + backY,                 size, -minX, -minY, -majX, -majY ) &&
              mPeano( x + majXs + majXs,                                 y + majYs + majYs,                                 size,  majX,  majY,  minX,  minY ) );
    }

    @Override
    public void run ()
    {
      int size = 1;
      while ( size < width ||
              size < height )
        size += (size + size);

      //LogFile.log( "SFC producer started, size = " + size + ", id = " + this.getId() );
      mPeano( 0, 0, size, 1, 0, 0, 1 );
      //LogFile.log( "SFC producer finished, size = " + size + ", id = " + this.getId() );
    }
  }

  //==========================================================================
  //  Interface Order2D:

  /**
   * Sets the source rectangle size. Initializes the pass algorithm. For both <b>Active</b> and
   * <b>Passive</b> modes.
   *
   * @param  _m Horizontal rectangle size (width).
   * @param  _n Vertical rectangle size (height).
   * @see    #init
   */
  public void setSize ( int _m, int _n )
  {
    stop();

    width  = _m;
    height = _n;

    init();
  }

  /**
   * Re-initializes the pass algorithm. For <b>Passive</b> pass mode.
   *
   * @see    #setSize
   */
  public void init ()
  {
    stopProducer();

    switch( type )
    {
      case ORDER_HILBERT:
        producer = new Hilbert();
        break;

      case ORDER_PEANO:
        producer = new Peano();
        break;

      case ORDER_M_PEANO:
        producer = new MPeano();
        break;

      case ORDER_MORTON:
      default:
        producer = new Morton();
    }

    produce   = true;
    toDeliver = width * height;
    fifo.clear();
    //LogFile.log( "init(), toDeliver = " + toDeliver + ", id = " + producer.getId() );

    try
    {
      producer.start();
    }
    catch ( IllegalThreadStateException e )
    {
      LogFile.exception( "Cannot start producer thread! " + e.getMessage(), e );
    }
  }

  /**
   * Returns the next 2D point. For <b>Passive</b> pass mode.
   *
   * @param  coords Pre-allocated array of size 2 (for [X,Y] coordinates).
   *                Can be {@code null}.
   * @return The next point's coordinates ([X,Y]).
   * @see    #init
   */
  public int[] next ( int[] coords )
  {
    if ( producer == null )
      return null;

    if ( toDeliver <= 0 )
    {
      stopProducer();
      return null;
    }

    if ( coords == null )
      coords = new int[ 2 ];

    QItem it = null;
    try
    {
      it = fifo.take();
    }
    catch ( InterruptedException e )
    {
    }

    if ( it == null )
      return null;

    coords[0] = it.x;
    coords[1] = it.y;
    toDeliver--;

    return coords;
  }

  /**
   * Anything to pass? For <b>Passive</b> pass mode.
   *
   * @return {@code true} if there is at least one pixel to pass.
   * @see    #init
   * @see    #next
   */
  public boolean isNext ()
  {
    return( toDeliver > 0 );
  }

  /**
   * Executes the whole rectangle pass. For <b>Active</b> (="modal") pass mode.
   *
   * @see    #setSize
   */
  public void pass ()
  {
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // ORDER_TYPE:
    if ( key.compareTo(ORDER_TYPE) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(ORDER_TYPE) == 0 )
      return type.ordinal();

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "OrderSFC";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "Order2D";

  /** Object description. */
  private final static String DESCRIPTION = "SFC ordering in 2D.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_RASTER + ".order";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"Order2D" );
        // Property:
      t.propBegin( ORDER_TYPE, TYPE_INTEGER, "Order type", true );
        t.propDefault( OrderType.ORDER_MORTON.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Morton",  OrderType.ORDER_MORTON.ordinal(),  "Morton order = Z-order" );
        t.propEnum( "Hilbert", OrderType.ORDER_HILBERT.ordinal(), "Hilbert order" );
        t.propEnum( "Peano",   OrderType.ORDER_PEANO.ordinal(),   "Peano order" );
        t.propEnum( "m-Peano", OrderType.ORDER_M_PEANO.ordinal(), "Meandering Peano order" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
