package cz.cuni.jagrlib.piece;

import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * PFM (Portable Float Map) file format (persistence of HDR 2D raster graphics data).
 *
 * @version 0.25 $Rev: 399 $ $Date: 2007-02-12 23:13:47 +0100 (po, 12 II 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/PFMFileFormat.java">PFMFileFormat.java</a>
 */
public class PFMFileFormat extends DefaultRasterFileFormat
{
  //==========================================================================
  //  Support routines:

  /**
   * Common load code.
   *
   * @param  stream Opened input bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonLoad ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    InputStream pfmFile = stream.getInputStream();
    if ( pfmFile == null )
      throw new FileNotFoundException();

      // decode PFM file into the raster graphics:
      // "PF <width> <height> <aspect_ratio>\r" or
      // "Pf <width> <height> <aspect_ratio>\r"
    boolean gray  = readToken( pfmFile ).equals( "Pf" );
    int width, height;
    double scale;
    try
    {
      width  = Integer.parseInt( readToken( pfmFile ) );
      height = Integer.parseInt( readToken( pfmFile ) );
      scale  = Double.parseDouble( readToken( pfmFile ) );
    }
    catch ( NumberFormatException ex )
    {
      throw new CharConversionException( "Error in PFM header" );
    }

      // init memory bitmap ..
    int bands      = gray ? 1 : 3;
    boolean endian = scale < 0.0;         // little endian
    g.init( width, height, gray ? RasterGraphics.MODE_GRAY : RasterGraphics.MODE_RGB, bands );

      // .. and finally read the binary data into it:
    int x, y;
    if ( gray )
      for ( y = 0; y < height; y++ )
        for ( x = 0; x < width; x++ )
          g.putPixel( x, y, readFloat( pfmFile, endian ) );
    else
    {
      double[] pix = new double[ 3 ];
      for ( y = 0; y < height; y++ )
        for ( x = 0; x < width; x++ )
        {
          pix[0] = readFloat( pfmFile, endian );
          pix[1] = readFloat( pfmFile, endian );
          pix[2] = readFloat( pfmFile, endian );
          g.putPixel( x, y, pix );
        }
    }
  }

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonSave ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    OutputStream pfmFile = stream.getOutputStream();
    if ( pfmFile == null )
      throw new FileNotFoundException();

      // write PFM header:
    boolean gray = (g.getMode() == RasterGraphics.MODE_GRAY);
    int width    = g.getWidth();
    int height   = g.getHeight();

      // "PF <width> <height> <aspect_ratio>\r" or
      // "Pf <width> <height> <aspect_ratio>\r"
    StringBuffer sb = new StringBuffer();
    sb.append( gray ? "Pf\n" : "PF\n" );
    sb.append( width ).append( " " ).append( height ).append( '\n' );
    sb.append( "-1.0\r" );
    int x, y, i;
    for ( i = 0; i < sb.length(); i++ )
      pfmFile.write( sb.codePointAt( i ) );

      // write the bitmap:
    if ( gray )
      for ( y = 0; y < height; y++ )
        for ( x = 0; x < width; x++ )
          writeFloat( pfmFile, (float)g.getPixelDouble( x, y ) );
    else
    {
      double[] pix = new double[ 3 ];
      for ( y = 0; y < height; y++ )
        for ( x = 0; x < width; x++ )
        {
          g.getPixel( x, y, pix );
          writeFloat( pfmFile, (float)pix[0] );
          writeFloat( pfmFile, (float)pix[1] );
          writeFloat( pfmFile, (float)pix[2] );
        }
    }
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null )
      return null;
    return null;
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return 3;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type (<tt>0.0</tt> .. no way,
   *         <tt>0.5</tt> .. maybe, <tt>0.9</tt> .. almost sure, <tt>1.0</tt> ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int len = (header == null) ? 0 : header.length;
    if ( len > 3 ) len = 3;
    if ( len < 3 )                        // not enough bytes in header[] => look at file-name
    {
      if ( fileName == null || fileName.length() < 5 ) return 0.0;
      int pos = fileName.lastIndexOf( '.' );
      if ( pos < 0 ) return 0.0;
      String suffix = fileName.substring( pos );
      if ( suffix.compareToIgnoreCase( "pfm" ) == 0 ) return 0.9;
      return 0.0;
    }
      // PFM header: "Pf " or "PF "
    if ( header[0] != (byte)'P' ||
         header[1] != (byte)'F' && header[1] != (byte)'f' ||
         !Character.isWhitespace( header[2] ) ) return 0.0;

    return 1.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.png"]) or <tt>null</tt>.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.pfm",
    };
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "PFM file format";

  /** Object description. */
  private final static String DESCRIPTION = "PFM filter (Portable Float Map).";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"DataFileFormat" );
      t.newOutputPlug(    PL_RASTER, IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_STREAM, IFACE+"BitStream" );
        // Property:
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
