package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.testing.LambertianModel;
import cz.cuni.jagrlib.reg.*;

/**
 * Phong shading model (with constant or Schlick K() term).
 *
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/PhongModel.java">PhongModel.java</a>
 */
public class PhongModel extends LambertianModel
{
  //==========================================================================
  //  Data:

  /** Original color of the solid surface. */
  public double[] origColor;

  /** Ambient-light coefficient. */
  public double ka;

  /** Diffuse reflection coefficient. */
  public double kd;

  /** Specular reflection coefficient. */
  public double ks;

  /** Specular reflection exponent. */
  public int h;

  /** Transmission (refraction) coefficient. */
  public double kt;

  /** Relative index of refraction. */
  public double n;

  /** Limit cos-value for total reflection. */
  public double cosTotal;

  /** Schlick blending coefficient ({@code 0.0} .. non-Schlick, {@code 1.0} .. Schlick). */
  public double sch;

  /** Default constructor. */
  public PhongModel ()
  {
    origColor = new double[] { 0.4, 0.25, 0.1 };
    ka        = 0.2;
    kd        = 0.6;
    ks        = 0.3;
    h         = 16;
    kt        = 0.0;
    n         = 1.0;
    cosTotal  = 0.0;
    sch       = 0.0;
  }

  //==========================================================================
  //  Support:

  /**
   * Common computation for monochrome BRDF sample (BRDF is treated as monochrome).
   *
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  d BRDF parameters.
   * @param  mask Which light components are to be computed?
   * @return BRDF coefficient {@code f(in->out)}.
   */
  protected double mono ( double[] in, double[] out, double[] normal, int mask )
  {
    final boolean viewOut = out[0] * normal[0] + out[1] * normal[1] + out[2] * normal[2] > 0.0;

    if ( in == null )                     // ambient term only..
        // dim ambient light if viewer is inside
      return( color[0] * (viewOut ? ka : (ka * kt)) );

      // directional light source:
    final double  cosAlpha = in[0] * normal[0] + in[1] * normal[1] + in[2] * normal[2];
    final boolean lightOut = cosAlpha > 0.0;

      // Schlick's adjustments:
    if ( sch > Geometry.EPSILON )
    {
      final double oneU    = 1.0 - Math.abs( cosAlpha );   //  1-u
      final double oneUsq  = oneU * oneU;                  // (1-u)^2
      final double schlick = sch * oneUsq * oneUsq * oneU; // (1-u)^5 * d.sch
      final double schlickComp = 1.0 - schlick;
      ks += schlick * (kd + kt);
      kd *= schlickComp;
      kt *= schlickComp;
    }

    double[] r = null;
    double coef = 1.0;
    if ( viewOut == lightOut )          // viewer and source are on the same side..
    {
      if ( (mask & COMP_SPEC_REFLECTION) != 0 )
      {
        final double cos2 = cosAlpha + cosAlpha;
        r = new double[]                // reflected vector
        {
          cos2 * normal[0] - in[0],
          cos2 * normal[1] - in[1],
          cos2 * normal[2] - in[2]
        };

        if ( !lightOut &&               // total reflection check
             -cosAlpha <= cosTotal )
          if ( (ks += kt) + kd > 1.0 )
            ks = 1.0 - kd;
      }
    }
    else                                // opposite sides => use specular refraction
    {
      if ( (mask & COMP_SPEC_REFRACTION) != 0 )
        r = Geometry.specularRefraction( normal, n, in, null );
      coef = kt;
    }

    final double diffuse = (mask & COMP_DIFFUSE) != 0 ?
                           coef * kd * Math.abs( cosAlpha ) : 0.0;
    double specular = 0.0;
    if ( r != null )
    {
      final double cosBeta = r[0] * out[0] + r[1] * out[1] + r[2] * out[2];
      if ( cosBeta > 0.0 )
        specular = coef * ks * Formula.power( cosBeta, h );
    }

    return( diffuse * color[0] + specular );
  }

  /**
   * Common computation for color BRDF sample (BRDF is treated as band-color).
   *
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  d BRDF parameters.
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                {@code f(in->out)} ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BRDF spectral coefficients.
   */
  protected double[] color ( double[] in, double[] out, double[] normal, int mask,
                             double[] result )
  {
    final int     bands   = color.length;
    final boolean viewOut = out[0] * normal[0] + out[1] * normal[1] + out[2] * normal[2] > 0.0;
    int i;
    double coef;

    if ( in == null )                     // ambient term only..
    {
      if ( result == null ||
           result.length < bands )
        result = new double[ bands ];

        // dim ambient light if viewer is inside
      coef = viewOut ? ka : (ka * kt);
      for ( i = 0; i < bands; i++ )
        result[i] = coef * color[i];

      return result;
    }

      // directional light source:
    final double  cosAlpha = in[0] * normal[0] + in[1] * normal[1] + in[2] * normal[2];
    final boolean lightOut = cosAlpha > 0.0;

      // Schlick's adjustments:
    if ( sch > Geometry.EPSILON )
    {
      final double oneU    = 1.0 - Math.abs( cosAlpha );     //  1-u
      final double oneUsq  = oneU * oneU;                    // (1-u)^2
      final double schlick = sch * oneUsq * oneUsq * oneU; // (1-u)^5 * d.sch
      final double schlickComp = 1.0 - schlick;
      ks += schlick * (kd + kt);
      kd *= schlickComp;
      kt *= schlickComp;
    }

    double[] r = null;
    coef       = 1.0;
    if ( viewOut == lightOut )            // viewer and source are on the same side..
    {
      if ( (mask & COMP_SPEC_REFLECTION) != 0 )
      {
        final double cos2 = cosAlpha + cosAlpha;
        r = new double[]                    // reflected vector
          {
            cos2 * normal[0] - in[0],
            cos2 * normal[1] - in[1],
            cos2 * normal[2] - in[2]
          };

        if ( !lightOut &&                   // total reflection check
             -cosAlpha <= cosTotal )
          if ( (ks += kt) + kd > 1.0 )
            ks = 1.0 - kd;
      }
    }
    else                                  // opposite sides => use specular refraction
    {
      if ( (mask & COMP_SPEC_REFRACTION) != 0 )
        r = Geometry.specularRefraction( normal, n, in, null );
      coef = kt;
    }

    if ( result == null ||
         result.length < bands )
      result = new double[ bands ];

    final double diffuse = (mask & COMP_DIFFUSE) != 0 ?
                           coef * kd * Math.abs( cosAlpha ) : 0.0;
    double specular = 0.0;
    if ( r != null )
    {
      final double cosBeta = r[0] * out[0] + r[1] * out[1] + r[2] * out[2];
      if ( cosBeta > 0.0 )
        specular = coef * ks * Formula.power( cosBeta, h );
    }

    for ( i = 0; i < bands; i++ )
      result[i] = diffuse * color[i] + specular;

    return result;
  }

  //==========================================================================
  //  Interface BSSRDF:

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as monochrome).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @return BSSRDF coefficient <b>f(in-&gt;out)</b>.
   * @see    #monoBSSRDF(MicroFacet,double[],double[],int)
   * @see    #colorBSSRDF(double[],double[],double[],int,double[])
   */
  @Override
  public double monoBSSRDF ( boolean worldCoords, double[] in, double[] out,
                             double[] normal, double[] tangent, int mask )
  {
    return mono( in, out, normal, mask );
  }

  /**
   * Returns the requested BRDF sample (BRDF is treated as monochrome). Computations are done
   * relatively to the given {@code MicroFacet} (point on surface of some solid).
   *
   * @param  mf The {@code MicroFacet} (point on surface of some solid).
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  mask Which light components are to be computed?
   * @return BRDF coefficient <b>f(in-&gt;mf-&gt;out)</b>.
   * @see    #monoBRDF(double[],double[],double[],Property,int)
   * @see    #colorBRDF(MicroFacet,double[],double[],int,double[])
   * @see    MicroFacet#material
   */
  @Override
  public double monoBSSRDF ( boolean worldCoords, MicroFacet mf, double[] in,
                             double[] out, int mask )
  {
    if ( mf == null ||
         mf.assertAttributes( MicroFacet.VALID_NORMAL ) != 0) return 0.0;

    return mono( in, out, mf.normal, mask );
  }

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as band-color).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BSSRDF spectral coefficients.
   * @see    #monoBSSRDF(double[],double[],double[],int)
   * @see    #colorBSSRDF(MicroFacet,double[],double[],int,double[])
   */
  @Override
  public double[] colorBSSRDF ( boolean worldCoords, double[] in, double[] out,
                                double[] normal, double[] tangent, int mask,
                                double[] result )
  {
    return color( in, out, normal, mask, result );
  }

  /**
   * Returns the requested BRDF sample (BRDF is treated as band-color). Computations are done
   * relatively to the given {@code MicroFacet} (point on surface of some solid).
   *
   * @param  mf The {@code MicroFacet} (point on surface of some solid).
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;mf-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BRDF spectral coefficients.
   * @see    #monoBRDF(MicroFacet,double[],double[],int)
   * @see    #colorBRDF(double[],double[],double[],Property,int,double[])
   * @see    MicroFacet#material
   */
  @Override
  public double[] colorBSSRDF ( boolean worldCoords, MicroFacet mf, double[] in,
                                double[] out, int mask, double[] result )
  {
    if ( mf == null ||
         mf.assertAttributes( MicroFacet.VALID_NORMAL |
                              MicroFacet.VALID_COLOR ) != 0 ) return null;

    setColor( mf.color );
    return color( in, out, mf.normal, mask, result );
  }

  /**
   * Accumulates light contributions from set of light sources through the given
   * {@code MicroFacet} (point on surface of some solid) to the viewing direction.
   *
   * @param  mf The {@code MicroFacet} (point on surface of some solid).
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  lights Set of light sources that can light the point.
   * @param  out Output (viewing) ray direction ({@code double[3]} or
   *             {@code double[4]}).
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;mf-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}, otherwise the
   *                initial value will be preseved.
   * @return Output color coefficients.
   * @see    #colorBRDF(MicroFacet,double[],double[],int,double[])
   * @see    MicroFacet#material
   */
  public double[] lightSum ( MicroFacet mf, int order, int total,
                             LightSource[] lights, double[] out,
                             double[] result )
  {
    if ( lights == null ||
         mf     == null ||
         mf.assertAttributes( MicroFacet.VALID_NORMAL |
                              MicroFacet.VALID_COLOR ) != 0 ) return null;

    setColor( mf.color );
    int bands = color.length;

      // initialize the results[] array:
    if ( result == null || result.length < bands )
      result = new double[ bands ];

      // temporary data:
    double[] lDir  = new double[ 4 ];     // light source direction vector
    double[] lInt  = new double[ bands ]; // light source intensity
    double[] contr = new double[ bands ]; // contribution of an individual ray

      // accumulate light from light sources:
    int i;
    for ( LightSource light : lights )
    {
      if ( light == null ) break;

      if ( !light.getIntensity( mf, order, total, lInt ) )
        continue;

      boolean dirSource = light.getDirection( mf, order, total, lDir );
      if ( color( dirSource ? lDir : null, out, mf.normal, COMP_ALL, contr ) == null )
        continue;

      for ( i = 0; i < bands; i++ )
        result[i] += contr[i] * lInt[i];
    }

    return result;
  }

  /**
   * Retrieves the original color (defining surface color).
   *
   * @param  color Pre-allocated {@code double[C]} array to hold the color
   *               ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *               color system/ or greater /for more complex spectral color sampling/). Can
   *               be {@code null}.
   * @param  data Data object with BRDF state information (optional).
   * @return Array containing copy of original color (can be altered).
   */
  @Override
  public double[] getOrigColor ( double[] color )
  {
    if ( this.origColor == null ) return null;

    int bands = this.origColor.length;
    if ( color == null ||
         color.length < bands )
      color = Formula.cloneArray( this.origColor );
    else
      System.arraycopy( this.origColor, 0, color, 0, bands );

    return color;
  }

  /**
   * Sets the actual color (for the given BRDF data).
   * This color will be used in subsequent BRDF calculations.
   * The color array won't be altered.
   *
   * @param  color Color definition - {@code double[C]} array
   *               ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *               color system/ or greater /for more complex spectral color sampling/).
   * @param  data Data object with BRDF state information.
   */
  @Override
  public void setColor ( double[] color )
  {
    if ( color != null )
      this.color = color;
  }

   /**
   * Scatter ray according to the light model.
   *
   * @param  mf   The {@link MicroFacet} (point on surface of some solid).
   * @param  in   Input direction (should look from the hit point)..
   * @param  u    Random uniform number.
   * @param  v    Random uniform number.
   * @param  mask Which light components are to be computed?
   * @param  out  Output direction in world coordinates.
   * @param  brdf Color definition - {@code double[C]} array
   *              ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *              color system/ or greater /for more complex spectral color sampling/).
   * @return Probability density function.
   */
  @Override
  public double scatter ( MicroFacet mf,
                          double[] in,
                          double u,
                          double v,
                          int mask,
                          double[] out,
                          double[] brdf )
  {
    if ( mf   == null ||
         in   == null ||
         brdf == null ||
         out  == null )
      return 0.0;

//    MonteCarlo.cosineSampleHemisphere( u, v, out );
    return pdf( false, mf, in, out );
  }

  /**
   * Return the probability density function.
   *
   * @param in  Input direction.
   * @param out Output direction.
   */
  @Override
  public double pdf ( boolean worldCoords, MicroFacet mf, double[] in, double[] out )
  {
    return 0.0;
  }

  /** Returns the type of BRDF. */
  @Override
  public int type ()
  {
    if ( kd > 0.0 &&
         ( ks > 0.0 || kt > 0.0 ) )
      return BSSRDF.COMP_ALL;
    else if ( ks > 0.0 ||
              kt > 0.0 )
      return BSSRDF.COMP_SPECULAR;
    else
      return BSSRDF.COMP_DIFFUSE;
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

    if ( key.compareTo( BSSRDF.MATERIAL_COLOR ) == 0 &&
         value instanceof double[] )
    {
      origColor = (double[])value;
      return;
    }

    if ( key.compareTo( BSSRDF.MATERIAL_KA ) == 0 )
    {
      ka = doubleProperty( value, ka );
      return;
    }

    if ( key.compareTo( BSSRDF.MATERIAL_KD ) == 0 )
    {
      kd = doubleProperty( value, kd );
      return;
    }

    if ( key.compareTo( BSSRDF.MATERIAL_KS ) == 0 )
    {
      ks = doubleProperty( value, ks );
      return;
    }

    if ( key.compareTo( BSSRDF.MATERIAL_H ) == 0 )
    {
      h = intProperty( value, h );
      return;
    }

    if ( key.compareTo( BSSRDF.MATERIAL_KT ) == 0 )
    {
      kt = doubleProperty( value, kt );
      return;
    }

    if ( key.compareTo( BSSRDF.MATERIAL_N ) == 0 )
    {
      n = doubleProperty( value, n );
      cosTotal = n > 1.0 ? (Math.sqrt( n * n - 1.0 ) / n) : 0.0;
      return;
    }

    if ( key.compareTo( "Sch" ) == 0 )
    {
      sch = doubleProperty( value, sch, 0.0, 1.0 );
      return;
    }

  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo( BSSRDF.MATERIAL_COLOR ) == 0 ) return origColor;
    if ( key.compareTo( BSSRDF.MATERIAL_KA )    == 0 ) return ka;
    if ( key.compareTo( BSSRDF.MATERIAL_KD )    == 0 ) return kd;
    if ( key.compareTo( BSSRDF.MATERIAL_KS )    == 0 ) return ks;
    if ( key.compareTo( BSSRDF.MATERIAL_H )     == 0 ) return h;
    if ( key.compareTo( BSSRDF.MATERIAL_KT )    == 0 ) return kt;
    if ( key.compareTo( BSSRDF.MATERIAL_N )     == 0 ) return n;
    if ( key.compareTo( "Sch" )                     == 0 ) return sch;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Phong shading model";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "BSSRDF";

  /** Object description. */
  private final static String DESCRIPTION = "Phong shading model with constant K() term.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_LIGHT;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"BSSRDF" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}