package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Plane objects (infinite plane, rectangle, triangle) as elementary 3D solids.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/Plane.java">Plane.java</a>
 */
public class Plane extends DefaultSolid
{
  //==========================================================================
  //  Data:

  /** Lower bound for x coordinate. */
  protected double xMin = Double.NEGATIVE_INFINITY;

  /** Upper bound for x coordinate (or see {@link #triangle}). */
  protected double xMax = Double.POSITIVE_INFINITY;

  /** Lower bound for y coordinate. */
  protected double yMin = Double.NEGATIVE_INFINITY;

  /** Upper bound for y coordinate (or see {@link #triangle}). */
  protected double yMax = Double.POSITIVE_INFINITY;

  /** Use three restrictions instead of four (the 3rd one is: {@code xMax * x + yMax * y < 1.0}). */
  protected boolean triangle = false;

  /** Use canonic u coordinate (=x)? */
  protected boolean canonicU = true;

  /** Use canonic v coordinate (=y)? */
  protected boolean canonicV = true;

  /** Scale coefficient for u coordinate. */
  protected double scaleU = 1.0;

  /** Scale coefficient for v coordinate. */
  protected double scaleV = 1.0;

  //==========================================================================
  //  Construction:

  /** Infinite plane. */
  public Plane ()
  {
  }

  /** Rectangle (four restrictions). */
  public Plane ( double xMi, double xMa, double yMi, double yMa )
  {
    xMin = xMi;
    xMax = xMa;
    yMin = yMi;
    yMax = yMa;

      // texture coordinates:
    if ( !Double.isInfinite( xMin ) &&
         !Double.isInfinite( xMax ) )
    {
      canonicU = false;
      scaleU   = 1.0 / (xMax - xMin);
    }
    if ( !Double.isInfinite( yMin ) &&
         !Double.isInfinite( yMax ) )
    {
      canonicV = false;
      scaleV   = 1.0 / (yMax - yMin);
    }
  }

  /** Triangle (three restrictions). */
  public Plane ( double xMa, double yMa )
  {
    yMin     = xMin = 0.0;
    xMax     = (xMa < Geometry.EPSILON) ? 1.0 : 1.0 / xMa;
    yMax     = (yMa < Geometry.EPSILON) ? 1.0 : 1.0 / yMa;
    triangle = true;
  }

  //==========================================================================
  //  Interface Solid:

  /**
   * Computes intersection of the solid with the given ray.
   *
   * @param  P0 Ray origin ({@code double[3]} or {@code double[4]}).
   * @param  P1 Direction vector of the ray ({@code double[3]} or {@code double[4]}).
   * @return List of intersection points ({@link MicroFacet}s) sorted by {@code t}
   *         (parametric distance from {@code P0}). {@code null} is returned in case
   *         of bad arguments or if no intersection exists..
   * @see    MicroFacet
   * @see    BoundingVolume
   */
  public List< MicroFacet > intersection ( double[] P0, double[] P1 )
  {
    final int P0len;
      // valid input parameters?
    if ( P0 == null || (P0len = P0.length) < 3 ||
         P1 == null ||          P1.length  < 3 ) return null;

      // ray origin:
    double Ox = P0[0];
    double Oy = P0[1];
    double Oz = P0[2];
    if ( P0len > 3 )
    {                                     // to cartesian
      double Ow = P0[3];
      if ( Geometry.isZero( Ow ) )
        return null;

      Ow  = 1.0 / Ow;
      Ox *= Ow;
      Oy *= Ow;
      Oz *= Ow;
    }

    final double Dz = P1[2];

    if ( Geometry.isZero( Dz ) )
      return null;

    final double t = -Oz / Dz;
    final double x = Ox + t * P1[0];
    if ( x < xMin ) return null;
    final double y = Oy + t * P1[1];
    if ( y < yMin ) return null;
    double u, v;

    if ( triangle )
    {
      u = x * xMax;
      v = y * yMax;
      if ( u + v > 1.0 ) return null;
    }
    else
      if ( x > xMax ||
           y > yMax ) return null;
      else
      {
        u = canonicU ? x : (x - xMin) * scaleU;
        v = canonicV ? y : (y - yMin) * scaleV;
      }

      // there will be at most one real intersection..
    LinkedList< MicroFacet > list = new LinkedList< MicroFacet >();
    MicroFacet mf;

    mf          = new MicroFacet();
    mf.t        = t;
    mf.enter    =
    mf.front    = Dz < 0.0;
    mf.locCoord = new double[] { x, y, 0.0 };
    mf.texCoord = new double[] { u, v };
    mf.valid    = MicroFacet.VALID_LOC_COORD |
                  MicroFacet.VALID_TEX_COORD;
    list.addLast( mf );

    return list;
  }

  /**
   * Asserts validity of the given intersection-attributes.
   * Resolves:
   * {@link MicroFacet#VALID_TANGENT VALID_TANGENT},
   * {@link MicroFacet#VALID_TEX_COORD VALID_TEX_COORD}.
   *
   * @param  mf The checked intersection.
   * @param  tmp Non-mandatory object holding state-information for deferred computations.
   * @param  attributes Attribute mask (required attributes - see
   *                    {@code MicroFacet.VALID_*}).
   * @return Attributes which were not computed.
   */
  @Override
  public int assertAttributes ( MicroFacet mf, Object tmp, int attributes )
  {
    if ( mf == null ) return attributes;

      // tangent surface vectors:
    if ( (attributes & MicroFacet.VALID_TANGENT) != 0 )
    {
      mf.tu = new double[] { 1.0, 0.0, 0.0 };
      mf.tv = new double[] { 0.0, 1.0, 0.0 };
      if ( mf.toWorld != null )
      {
        mf.toWorld.transformVector3D( mf.tu, mf.tu );
        mf.toWorld.transformVector3D( mf.tv, mf.tv );
      }

      mf.valid |= MicroFacet.VALID_TANGENT;
      if ( (attributes -= MicroFacet.VALID_TANGENT) == 0 ) return 0;
    }

    return attributes;
  }

  /**
   * Point vs. solid test ("inside-test").
   *
   * @param  point Coordinates of the 3D point ({@code double[3]} or
   *               {@code double[4]}).
   * @return {@code true} if the given point lies inside the solid.
   */
  @Override
  public boolean isInside ( double[] point )
  {
    if ( point == null ||
         point.length < 3 ) return false;

    if ( point.length == 3 )
      return( point[2] <= 0.0 );

    return( (point[0] <= 0.0) != (point[3] <= 0.0) );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Plane";

  /** Object description. */
  private final static String DESCRIPTION = "Plane objects (plane, rectangle, triangle) as 3D solids for CSG.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"Solid" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
