package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Computes illumination of 3D scene using radiosity method.
 *
 * @author  Tomas Svab, PE
 * @version 0.26 $Rev: 753 $ $Date: 2008-05-18 17:25:57 +0200 (ne, 18 V 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/Radiosity.java">Radiosity.java</a>
 */
public class Radiosity extends Piece implements GlobalIllumination
{
  //==========================================================================
  //  Data:

  /** Supported color modes. */
  public static enum Colors
  {
    /** Use all three components (R - red, G - green, B - blue). */
    RGB,
    /** Compute the red component (only red component is used). */
    RED,
    /** Compute the green component (only green component is used). */
    GREEN,
    /** Compute the blue component (only blue component is used). */
    BLUE,
    /** Compute the red component (all components are the same -> result is gray). */
    RED_GRAY,
    /** Compute the green component (all components are the same -> result is gray). */
    GREEN_GRAY,
    /** Compute the blue component (all components are the same -> result is gray). */
    BLUE_GRAY,
  }

  /** Actual used color(s). */
  protected Colors color = Colors.RGB;

  /** Accelerator: input B-rep database. */
  protected Brep brep;

  /** Accelerator: input result from solver. */
  protected SLEData data;

  /** Accelerator: input result from solver. */
  protected SLESolver solver;

  /** Accelerator: input information about basis. */
  protected RadiosityBasisFunctions basis;

  /** Actual {@link Brep} context. */
  protected int ctx = Brep.CTX_DEFAULT;

  /** Actual node-result attribute. */
  protected int attrResult = brep.NULL;

  /** Use Gouraud interpolation? */
  protected boolean gouraud = true;

  /** Required accuracy of solution. */
  protected double epsilon = 0.01;

  /** Refine-batch size. */
  protected int batch = 100;

  /** Log debug information? */
  protected boolean log = false;

  //==========================================================================
  //  Support:

  /** Save one result (attribute) into a {@link Brep} database. */
  protected void saveResult ( int handle, float[] result )
  {
    if ( result        == null ||
         result.length < 3 )
      return;

    for ( int i = 0; i < 3; i++ )
      result[i] = Formula.clamp( result[i], 0.0f, 1.0f );
    brep.setAttribute( attrResult, handle, result );
  }

  /** Save all results into a {@link Brep} database. */
  protected void saveResults ( double[][] solution  )
  {
    if ( solution == null )
      return;

    final int numElem = basis.totalElements();
    final int numVert = brep.getAllVertices( null );
    float[] resultItem = new float[ 3 ];
    int i, j;

    if ( !gouraud )
    {
        // no interpolation!
        // .. just save the solution into attached Brep database:
      for ( i = 0; i < numElem; i++ )
      {
        switch ( color )
        {
          case RGB:
            resultItem[0] = (float)solution[0][i];
            resultItem[1] = (float)solution[1][i];
            resultItem[2] = (float)solution[2][i];
            break;

          case RED_GRAY:
          case GREEN_GRAY:
          case BLUE_GRAY:
            Arrays.fill( resultItem, (float)solution[0][i] );
            break;

          case RED:
            Arrays.fill( resultItem, 0.0f );
            resultItem[0] = (float)solution[0][i];
            break;

          case GREEN:
            Arrays.fill( resultItem, 0.0f );
            resultItem[1] = (float)solution[0][i];
            break;

          case BLUE:
            Arrays.fill( resultItem, 0.0f );
            resultItem[2] = (float)solution[0][i];
            break;
        }
        saveResult( i, resultItem );
      }

      return;
    }

      // propagate illumination into vertices:
    float[][] results = new float[ 3 ][ numVert ];    // previously computed illumination (SLE)
    int[] affect      = new int[ numVert ];           // number of faces affecting individual vertices
    final int components = Colors.RGB.equals( color ) ? 3 : 1;

    for ( int k = 0; k < components; k++ )
    {
      Arrays.fill( affect, 0 );

        // computes values in vertices:
      int[] verticesID = new int[ 3 ];
      for ( int element = 0; element < numElem; element++ )
      {
        brep.getFaceVertices( element, verticesID  );
        for ( int vertex = 0; vertex < 3; vertex++ )
        {
          results[ k ][ verticesID[vertex] ] += (float)solution[k][element];
          affect[ verticesID[vertex] ]++;
        }
      }
    }

      // avarages solution:
    for ( i = 0; i < numVert; i++ )
      for ( j = 0; j < components; j++ )
        results[j][i] /= affect[i];

      // saves solution:
    for ( i = 0; i < numVert; i++ )
    {
      switch ( color )
      {
        case RGB:
          resultItem[0] = (float)results[0][i];
          resultItem[1] = (float)results[1][i];
          resultItem[2] = (float)results[2][i];
          break;

        case RED_GRAY:
        case GREEN_GRAY:
        case BLUE_GRAY:
          Arrays.fill( resultItem, (float)results[0][i] );
          break;

        case RED:
          Arrays.fill( resultItem, 0.0f );
          resultItem[0] = (float)results[0][i];
          break;

        case GREEN:
          Arrays.fill( resultItem, 0.0f );
          resultItem[1] = (float)results[0][i];
          break;

        case BLUE:
          Arrays.fill( resultItem, 0.0f );
          resultItem[2] = (float)results[0][i];
          break;
      }
      saveResult( i, resultItem );
    }
  }

  //==========================================================================
  //  Interface GlobalIllumination:

  /**
   * Sets camera position ({@code null} for no camera = pure global lighting).
   *
   * @param  camera World-space coordinates of the camera.
   * @todo   Importance of the camera (instead of its position).
   */
  @Override
  public void setCamera ( double[] camera )
  {
  }

  /**
   * Computes scene illumination from the scratch.
   * If time/iteration limits are nonzero, tries to fulfill them..
   * This method can be called repeatedly - or in combination with {@link #updateIllumination}.
   *
   * @param  maxIterations If nonzero, defines maximum number of iterations (if applicable).
   * @param  maxTime If nonzero, defines maximum time in milliseconds (if applicable).
   * @return {@code true} if finished.
   */
  @Override
  public boolean illuminate ( int maxIterations, long maxTime )
  {
    if ( brep   == null ) brep   = (Brep)                   getInterface( PL_DATA,      IFACE+"Brep" );
    if ( solver == null ) solver = (SLESolver)              getInterface( PL_OUTPUT,    IFACE+"SLESolver" );
    if ( basis  == null ) basis  = (RadiosityBasisFunctions)getInterface( PL_FUNCTION,  IFACE+"RadiosityBasisFunctions" );
    if ( data   == null ) data   = (SLEData)                getInterface( PL_TRANSFORM, IFACE+"SLEData" );

    if ( gouraud )
      attrResult = brep.getAttributeId( brep.VERTEX, Brep.NULL, brep.COLOR );
    else
      attrResult = brep.getAttributeId( brep.FACE,   Brep.NULL, brep.COLOR );

    final int numElem    = basis.totalElements();
    final int components = Colors.RGB.equals( color ) ? 3 : 1;

    double[][] solution = new double[ components ][ numElem ];   // do we really need this?? :-(

    if ( log )
    {
      LogFile.log( "------------------------------" );
      LogFile.log( "Radiosity global illumination:" );
      LogFile.log( "  SLE size:       " + numElem );
      LogFile.log( "  Epsilon:        " + epsilon );
      LogFile.log( "  Omega:          " + solver.get( SLESolver.OMEGA ) );
      LogFile.log( "  Batch size:     " + batch );
    }

      // solves SLE:
    long time       = 0;
    long millis     = System.currentTimeMillis();
    double accuracy = 0.0;
    int steps       = 0;
    int i;

    for ( i = 0; i < components && !userBreak; i++ )
    {
        // sets the actual component:
      switch ( color )
      {
        case RGB:
          data.setComponent( i );
          break;

        case RED:
        case RED_GRAY:
          data.setComponent( 0 );
          break;

        case GREEN:
        case GREEN_GRAY:
          data.setComponent( 1 );
          break;

        case BLUE:
        case BLUE_GRAY:
          data.setComponent( 2 );
          break;
      }

        // computes solution:
      solver.reset();
      do
      {
        accuracy = solver.refine( batch, true );
        time = System.currentTimeMillis() - millis;
        steps += batch;
      }
      while ( steps < maxIterations &&
              time < maxTime / components &&
              accuracy > epsilon &&
              !userBreak );

      if ( log )
      {
        LogFile.log( "Color component " + i );
        LogFile.log( "  Iterations:     " + solver.get( SLESolver.STEPS ) );
        LogFile.log( String.format( Locale.US, "  Total residual: %.6f", solver.get( SLESolver.TOTAL_RESIDUAL ) ) );
        LogFile.log( String.format( Locale.US, "  Accuracy:       %.6f", accuracy ) );
      }

      steps = 0;
      solution[i] = solver.getSolution( solution[i] );
    }

    if ( log )
      LogFile.log( String.format( Locale.US, "Total time:       %.2f s" + (userBreak ? " USER BREAK!" : ""), 0.001*time ) );

    saveResults( solution );

    return true;
  }

  /**
   * Recomputes illumination according to scene & light-source changes.
   * Scene & light-source changes are expressed in the {@link SLEData} interface.
   * Camera position/importance changes should be considered too.
   *
   * If time/iteration limits are nonzero, tries to fulfill them..
   *
   * @param  maxIterations If nonzero, defines maximum number of iterations (if applicable).
   * @param  maxTime If nonzero, defines maximum time in milliseconds (if applicable).
   * @return {@code true} if finished.
   */
  @Override
  public boolean updateIllumination ( int maxIterations, long maxTime )
  {
    solver.update();

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String COLORS  = "Colors";
  public final static String GOURAUD = "Gouraud";
  public final static String EPSILON = "Epsilon";
  public final static String BATCH   = "Batch";
  public final static String LOG     = "Log";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // COLORS:
    if ( key.compareTo(COLORS) == 0 )
    {
      color = enumProperty( value, color );
      return;
    }
      // GOURAUD:
    if ( key.compareTo(GOURAUD) == 0 )
    {
      gouraud = booleanProperty( value, gouraud );
      return;
    }
      // ACCURACY:
    if ( key.compareTo(EPSILON) == 0 )
    {
      epsilon = doubleProperty( value, epsilon  );
      return;
    }
      // BATCH:
    if ( key.compareTo(BATCH) == 0 )
    {
      batch = intProperty( value, batch );
      return;
    }
      // LOG:
    if ( key.compareTo(LOG) == 0 )
    {
      log = booleanProperty( value, log );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(COLORS) == 0 )
      return color.ordinal();
    if ( key.compareTo(GOURAUD) == 0 )
      return gouraud;
    if ( key.compareTo(EPSILON) == 0 )
      return epsilon;
    if ( key.compareTo(BATCH) == 0 )
      return batch;
    if ( key.compareTo(LOG) == 0 )
      return log;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Radiosity";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "GlobalIlluminationToSLESolverAndRadiosityBasisFunctionsAndBrepAndSLEData";

  /** Object description. */
  private final static String DESCRIPTION = "Computes illumination of B-rep scene using Radiosity method";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_LIGHT;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,     IFACE+"GlobalIllumination" );
      t.newOutputPlug( PL_OUTPUT,    IFACE+"SLESolver" );
      t.newOutputPlug( PL_FUNCTION,  IFACE+"RadiosityBasisFunctions" );
      t.newOutputPlug( PL_DATA,      IFACE+"Brep" );
      t.newOutputPlug( PL_TRANSFORM, IFACE+"SLEData" );

        // Property:
      t.propBegin( COLORS, TYPE_INTEGER, "Used color(s)", true );
        t.propDefault( Colors.RGB.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "RGB",        Colors.RGB.ordinal(),        "Use all three components." );
        t.propEnum( "Red",        Colors.RED.ordinal(),        "Compute the red component (only this component is used)." );
        t.propEnum( "Green",      Colors.GREEN.ordinal(),      "Compute the green component (only this component is used)." );
        t.propEnum( "Blue",       Colors.BLUE.ordinal(),       "Compute the blue component (only this component is used)." );
        t.propEnum( "Red gray",   Colors.RED_GRAY.ordinal(),   "Compute the red component (all components are the same -> result is gray)." );
        t.propEnum( "Green gray", Colors.GREEN_GRAY.ordinal(), "Compute the green component (all components are the same -> result is gray)." );
        t.propEnum( "Blue gray",  Colors.BLUE_GRAY.ordinal(),  "Compute the blue component (all components are the same -> result is gray)." );
      t.propEnd();
      t.propBegin( GOURAUD, TYPE_BOOLEAN, "Use Gouraud interpolation?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( EPSILON, TYPE_DOUBLE, "Required solution accuracy", true  );
        t.propDefault( 0.01 );
      t.propEnd();
      t.propBegin( BATCH, TYPE_INTEGER, "Refine-batch size", true );
        t.propDefault( 100 );
      t.propEnd();
      t.propBegin( LOG, TYPE_BOOLEAN, "Log useful information?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
