package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Provides SLE (system of linear equations) from Brep for some SLE Solver.
 *
 * @author  Tomas Svab, PE
 * @version 0.26 $Rev: 757 $ $Date: 2008-05-19 01:09:45 +0200 (po, 19 V 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/RadiosityEquations.java">RadiosityEquations.java</a>
 * @todo    Support for common polygons (not only triangles).
 * @todo    Support for multi-component color models.
 */
public class RadiosityEquations extends Piece implements SLEData
{
  //==========================================================================
  //  Data:

  /** Accelerator: input {@link Brep} database. */
  protected Brep brep;

  /** Accelerator: input radiosity basis functions. */
  protected RadiosityBasisFunctions basis;

  /** Actual {@link Brep} context. */
  protected int ctx = Brep.CTX_DEFAULT;

  /** Attribute name: face color. */
  protected String colorName = brep.COLOR;

  /** Attribute name: face color. */
  protected String emissionName = "Emission";

  /** Attribute: face color (albedo = ro). */
  protected int attrColor = brep.NULL;

  /** Attribute: face emmision (E). */
  protected int attrEmission = brep.NULL;

  /** Attribute: face radiosity (result = B). */
  protected int attrResult   = brep.NULL;

  /** Was the SLE already initialized? */
  protected boolean isInitialized = false;

  /** Actual component number. */
  protected int component = 0;

  //==========================================================================
  //  Input SLE data:

  /**
   * Number of equations = number of matrix rows
   * (usually number of unknowns is equal to number of equations).
   */
  @Override
  public int getRows ()
  {
    if ( !init() ) return NULL;

    return basis.totalCoefficients();
  }

  /**
   * Number of unknowns = number of matrix columns
   * (usually number of unknowns is equal to number of equations).
   */
  @Override
  public int getColumns ()
  {
    if ( !init() ) return NULL;

    return basis.totalCoefficients();
  }

  /** Reads a single matrix coefficient. */
  @Override
  public double getCoeff ( int row, int col )
  {
    if ( !init() ) return 0.0;

    if ( row < 0 || row >= getRows() ||
         col < 0 || col >= getColumns() ) return 0.0;

    return (row == col ? 1.0 : 0.0) - getReflectance( row ) * basis.formFactor( row, col );
  }

  /** Reads the whole matrix row. */
  @Override
  public double[] getRow ( int row, double[] result )
  {
    if ( !init() ) return null;

    final int rows = getRows();
    if ( row < 0 || row >= rows ) return null;
    if ( result        == null ||
         result.length < rows )
      result = new double[ rows ];

    final double P = -getReflectance( row );
    basis.formFactorsFrom( row, null, result );

    for ( int col = 0; col < rows; col++ )
      result[col] *= P;
    result[row] += 1.0;

    return result;
  }

  /** Reads the whole matrix column. */
  @Override
  public double[] getColumn ( int col, double[] result )
  {
    if ( !init() ) return null;

    final int cols = getColumns();
    if ( col < 0 || col >= cols ) return null;
    if ( result        == null ||
         result.length < cols )
      result = new double[ cols ];

    basis.formFactorsTo( col, null, result );

    for ( int row = 0; row < cols; row++ )
      result[row] *= -getReflectance( row );
    result[col] += 1.0;

    return result;
  }

  /** Reads the whole matrix. */
  @Override
  public double[][] getMatrix ()
  {
    if ( !init() ) return null;

    final int rows = getRows();
    final int cols = getColumns();

    double P;
    final double[][] result = new double[ rows ][ cols ];

    for ( int row = 0; row < rows; row++ )
    {
      basis.formFactorsFrom( row, null, result[row] );
      P = -getReflectance( row );

      for ( int col = 0; col < cols; col++ )
        result[row][col] *= P;
      result[row][row] += 1.0;
    }

    return result;
  }

  /** Reads a single element of the right-side vector. */
  @Override
  public double getRight ( int row )
  {
    if ( !init() ) return 0.0;

    final int rows = getRows();
    if ( row < 0 || row >= rows ) return 0.0;

    final float[] E = new float[ 3 ];
    brep.getAttribute( attrEmission, row, E );

    return (double)E[component];
  }

  /** Reads the whole right-side vector. */
  @Override
  public double[] getRight ( double[] result )
  {
    if ( !init() ) return null;

    final int rows = getRows();
    if ( result        == null ||
         result.length < rows )
      result = new double[ rows ];

    final float[] E = new float[ 3 ];
    for ( int row = 0; row < rows; row++ )
    {
      brep.getAttribute( attrEmission, row, E );
      result[row] = (double)E[component];
    }

    return result;
  }

  //==========================================================================
  //  Solution vector:

  /** Reads a single unknown (element of solution vector). */
  @Override
  public double getSolution ( int col )
  {
    return 0.0;
  }

  /** Reads the whole solution vector. */
  @Override
  public double[] getSolution ( double[] result )
  {
    return null;
  }

  /** Writes value of a single unknown. */
  @Override
  public void setSolution ( int col, double value )
  {
  }

  /** Writes the whole solution vector. */
  @Override
  public void setSolution ( double[] values )
  {
  }

  //==========================================================================
  //  Incremental SLE changes (non-mandatory):

  /** Returns index of next right-side element which was changed since last {@link #resetChanges} call. */
  @Override
  public int nextRightChanged ( int row )
  {
    return NULL;
  }

  /**
   * Returns column-index of next coefficient which was changed since last {@link #resetChanges} call.
   * Passes through the given matrix row..
   */
  @Override
  public int nextCoeffInRowChanged ( int row, int col )
  {
    return NULL;
  }

  /**
   * Returns row-index of next coefficient which was changed since last {@link #resetChanges} call.
   * Passes through the given matrix column..
   */
  @Override
  public int nextCoeffInColumnChanged ( int row, int col )
  {
    return NULL;
  }

  /** Resets (acknowledges) all SLE changes. */
  @Override
  public void resetChanges ()
  {
  }

  //==========================================================================
  //  Additional information:

  /** Reads a single element weight (radiosity: element area). */
  @Override
  public double getWeight ( int i )
  {
    if ( !init() ) return 0.0;

    final int rows = getRows();
    if ( i < 0 || i >= rows ) return 0.0;

    final int[] vertexID = new int[ 3 ];
    brep.getFaceVertices( i, vertexID );
    double[][] v = new double[ 3 ][ 3 ];
    for ( int j = 0; j < 3; j++ )
      brep.getVertexCoords( ctx, vertexID[j], v[j] );

    return Geometry.triangleArea( v );
  }

  /**
   * Reads the whole weight vector.
   *
   * @param  values Pre-allocated array to hold result (can be {@code null}).
   */
  @Override
  public double[] getWeight ( double[] values )
  {
    if ( !init() ) return null;

    final int rows = getRows();
    if ( values        == null ||
         values.length < rows )
      values = new double[ rows ];

    for ( int i = 0; i < rows; i++ )
      values[i] = getWeight( i );

    return values;
  }

  /** Reads a single reflectance (radiosity: element albedo). */
  @Override
  public double getReflectance ( int i )
  {
    if ( !init() ) return 0.0;

    final int rows = getRows();
    if ( i < 0 || i >= rows ) return 0.0;

    final float P[] = new float[ 3 ];
    brep.getAttribute( attrColor, i, P );

    return (double)P[ Formula.clamp( component, 0, 2 ) ];
  }

  /**
   * Reads the whole reflectance vector.
   *
   * @param  reflectance Pre-allocated array to hold result (can be {@code null}).
   */
  @Override
  public double[] getReflectance ( double[] reflectance )
  {
    if ( !init() ) return null;

    final int rows = getRows();
    if ( reflectance        == null ||
         reflectance.length < rows )
      reflectance = new double[ rows ];

    final float[] P = new float[ 3 ];
    for ( int row = 0; row < rows; row++ )
    {
      brep.getAttribute( attrColor, row, P );
      reflectance[row] = (double)P[ Formula.clamp( component, 0, 2 ) ];
    }

    return reflectance;
  }

  /** Get actual component number. */
  @Override
  public int actualComponent ()
  {
    return component;
  }

  /** Set actual component (non-mandatory). */
  @Override
  public void setComponent ( int c )
  {
    component = c;
  }

  //==========================================================================
  //  Support:

  /** Initialize accelerators. */
  protected boolean init ()
  {
    if ( isInitialized ) return true;

    if ( brep  == null ) brep  = (Brep)                   getInterface( PL_DATA,     IFACE+"Brep" );
    if ( basis == null ) basis = (RadiosityBasisFunctions)getInterface( PL_FUNCTION, IFACE+"RadiosityBasisFunctions" );

    if ( brep  == null ||
         basis == null )
      return false;

    attrColor    = brep.getAttributeId( brep.FACE, ctx, colorName );
    attrEmission = brep.getAttributeId( brep.FACE, ctx, emissionName );

    isInitialized = true;
    return true;
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Radiosity Equations";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "SlEDataToBrepAndRadiosityBasisFunctions";

  /** Object description. */
  private final static String DESCRIPTION = "SLE system of radiosity equations";

  /** Object category. */
  protected final static String CATEGORY = C_DATA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,    IFACE+"SLEData" );
      t.newOutputPlug( PL_DATA,     IFACE+"Brep" );
      t.newOutputPlug( PL_FUNCTION, IFACE+"RadiosityBasisFunctions" );
        // Property:
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
