package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Rank filter module.
 *
 * @version 0.25 $Rev: 392 $ $Date: 2007-01-21 17:37:56 +0100 (ne, 21 I 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/RankFilter.java">RankFilter.java</a>
 */
public class RankFilter extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Actual filter type. */
  protected RankType type = RankType.MEDIAN;

  /** Actual window shape. */
  protected WindowShape shape = WindowShape.RECT;

  /** Actual window size (radius). */
  protected int radius = 1;

  /** Actual window offset array ([dx0,dy0], [dx1,dy1], ..). */
  protected int[] window;

  /** Working array of ranks. */
  protected Rank[] ranks;

  //==========================================================================
  //  Support:

  protected void initWindow ()
  {
    if ( radius < 1 ) radius = 1;

    int i, j, k = 0;
    switch ( shape )
    {
      case CROSS:
        window = new int[ 2 * (1 + 4 * radius) ];
        for ( i = -radius; i < 0; i++ )
        {
          window[ k++ ] = 0;
          window[ k++ ] = i;
        }
        for ( i = -radius; i <= radius; i++ )
        {
          window[ k++ ] = i;
          window[ k++ ] = 0;
        }
        for ( i = 1; i <= radius; i++ )
        {
          window[ k++ ] = 0;
          window[ k++ ] = i;
        }
        break;

      case ROUND:
      {
        int[] tmp = new int[ 2 * (radius + radius + 1) * (radius + radius + 1) ];
        int   rr  = radius * radius;

        for ( j = -radius; j <= radius; j++ )
          for ( i = -radius; i <= radius; i++ )
            if ( i * i + j * j <= rr )
            {
              tmp[ k++ ] = i;
              tmp[ k++ ] = j;
            }

        window = new int[ k ];
        System.arraycopy( tmp, 0, window, 0, k );
        break;
      }

      default:                            // rectangle window shape
        window = new int[ 2 * (radius + radius + 1) * (radius + radius + 1) ];
        for ( j = -radius; j <= radius; j++ )
          for ( i = -radius; i <= radius; i++ )
          {
            window[ k++ ] = i;
            window[ k++ ] = j;
          }
    }
  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image filtering.
   *
   * @param  t The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int t )
  {
    RasterGraphics input  = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    RasterGraphics output = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );

      // output image:
    int width  = input.getWidth();
    int height = input.getHeight();
    int iMode  = input.getMode();
    int bands  = input.getBands();
    int oMode  = iMode;
    int maxVal = 255;

    switch ( iMode )
    {
      case RasterGraphics.MODE_GRAY:
        maxVal = input.getMaxGray();
        break;

      case RasterGraphics.MODE_COLORMAP:
        oMode  = RasterGraphics.MODE_RGB;
        bands  = 3;

      default:
        maxVal = input.getMaxRGB();
    }

    output.init( width, height, oMode, bands );

      // support data:
    int i, j, k, b;
    int x, y;
    ranks = new Rank[ bands ];
    for ( b = 0; b < bands; b++ )
      ranks[ b ] = new RankHistogram();

    initWindow();

    switch ( oMode )
    {
      case RasterGraphics.MODE_GRAY:
        for ( j = 0; j < height; j++ )
          for ( i = 0; i < width; i++ )
          {
            ranks[ 0 ].init( maxVal );
            for ( k = 0; k < window.length; )
            {
              x = i + window[ k++ ];
              y = j + window[ k++ ];
              if ( x >= 0 && x < width &&
                   y >= 0 && y < height )
                ranks[ 0 ].add( input.getPixel( x, y ) );
            }
            output.putPixel( i, j, ranks[ 0 ].result( type ) );
          }
        break;

      default:
      {
        int[] pix = new int[ bands ];

        for ( j = 0; j < height; j++ )
          for ( i = 0; i < width; i++ )
          {
            for ( b = 0; b < bands; b++ )
              ranks[ b ].init( maxVal );

            for ( k = 0; k < window.length; )
            {
              x = i + window[ k++ ];
              y = j + window[ k++ ];
              if ( x >= 0 && x < width &&
                   y >= 0 && y < height )
              {
                if ( iMode == RasterGraphics.MODE_COLORMAP )
                  input.getRGB( x, y, pix );
                else
                  input.getPixel( x, y, pix );

                for ( b = 0 ; b < bands; b++ )
                  ranks[ b ].add( pix[ b ] );
              }
            }

            for ( b = 0; b < bands; b++ )
              pix[ b ] = ranks[ b ].result( type );

            output.putPixel( i, j, pix );
          }
      }
    }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String FILTER_TYPE  = "Filter type";
  public final static String WINDOW_SHAPE = "Window shape";
  public final static String RADIUS       = "Radius";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // FILTER_TYPE:
    if ( key.compareTo(FILTER_TYPE) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
      // WINDOW_SHAPE:
    if ( key.compareTo(WINDOW_SHAPE) == 0 )
    {
      shape = enumProperty( value, shape );
      return;
    }
      // RADIUS:
    if ( key.compareTo(RADIUS) == 0 )
    {
      radius = intProperty( value, radius, 1, 20 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(FILTER_TYPE) == 0 )
      return type.ordinal();
    if ( key.compareTo(WINDOW_SHAPE) == 0 )
      return shape.ordinal();
    if ( key.compareTo(RADIUS) == 0 )
      return radius;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "RankImageFilter";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Rank image filter.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"Trigger" );
      t.newOutputPlug( "image1", IFACE+"RasterGraphics" );
      t.newOutputPlug( "image2", IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( FILTER_TYPE, TYPE_INTEGER, "Filter type", true );
        t.propDefault( RankType.MEDIAN.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Min (erosion)",      RankType.MIN.ordinal(),    "Minimum (erosion)" );
        t.propEnum( "Max (dilation)",     RankType.MAX.ordinal(),    "Maximum (dilation)" );
        t.propEnum( "Median",             RankType.MEDIAN.ordinal(), "Median filter" );
        t.propEnum( "Middle (min+max)/2", RankType.MIDDLE.ordinal(), "Middle (min+max)/2" );
      t.propEnd();
      t.propBegin( WINDOW_SHAPE, TYPE_INTEGER, "Window shape", true );
        t.propDefault( WindowShape.RECT.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Rectangular",  WindowShape.RECT.ordinal(),  "Rectangular window" );
        t.propEnum( "Round",        WindowShape.ROUND.ordinal(), "Round window" );
        t.propEnum( "Cross-shaped", WindowShape.CROSS.ordinal(), "Cross-shaped window" );
      t.propEnd();
      t.propBegin( RADIUS, TYPE_INTEGER, "Window radius in pixels", true );
        t.propDefault( 1 );
        t.propBounds(  1, 20 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
