package cz.cuni.jagrlib.piece;

import java.util.*;
import java.awt.image.*;
import java.awt.Rectangle;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Raster image stored in memory.
 * <br>Uses Java's {@link BufferedImage} class to hold raster data.
 *
 * @version 0.25 $Rev: 385 $ $Date: 2007-01-07 01:47:21 +0100 (ne, 07 I 2007) $ $Author: pepca $
 * @since   0.03
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/RasterImage.java">RasterImage.java</a>
 */
public class RasterImage extends DefaultRasterGraphics
{
  //==========================================================================
  //  Data:

  /** Background color for image initialization. */
  public final static int[] INIT_COLOR = { 0, 0, 0, 255 };

  /** Default foreground color. */
  public final static int[] FOREGROUND = { 255, 255, 255, 255 };

  /** The working {@link BufferedImage} object. */
  protected BufferedImage im;

  /**
   * Current colormap-entry mode (either {@link #MODE_RGB MODE_RGB} or {@link #MODE_RGBA MODE_RGBA}).
   */
  protected int colormapMode;

  /** Actual colormap size. */
  protected int colormapSize;

  /** Colormap cache (accelerator). */
  protected int[] cmapCache;

  //--------------------------------------------------------------------------
  //  State variables:

  /** Actual drawing scalar color ({@link #MODE_GRAY MODE_GRAY}, {@link #MODE_COLORMAP MODE_COLORMAP}). */
  protected int sColor;

  /** Actual drawing color ({@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA}, {@link #MODE_BAND MODE_BAND}). */
  protected int[] color;

  /** Temporary array used for RGBA access. */
  protected int[] tmpColor;

  //==========================================================================
  //  Interface RasterGraphics:

  /**
   * {@code RasterImage} (re-)initialization.
   *
   * @param  _width Width of a new bitmap.
   * @param  _height Height of a new bitmap.
   * @param  _mode Pixel value type ({@code RasterGraphics.MODE_*} constant).
   * @param  _bands Number of bands (for {@link #MODE_BAND MODE_BAND}).
   * @see    #setColormapType
   */
  public void init ( int _width, int _height, int _mode, int _bands )
  {
    super.init( _width, _height, _mode, _bands );

    colormapSize = 0;
    colormapMode = MODE_RGB;
    int biType;
    switch ( _mode )
    {
      case MODE_RGB:
        biType = BufferedImage.TYPE_INT_RGB;
        bands = 3;
        break;

      case MODE_RGBA:
        biType = BufferedImage.TYPE_INT_ARGB_PRE;
        bands = 4;
        break;

      case MODE_GRAY:
        biType = BufferedImage.TYPE_BYTE_GRAY;
        bands = 1;
        break;

      case MODE_COLORMAP:
        biType = BufferedImage.TYPE_BYTE_INDEXED;
        colormapSize = 256;
        bands = 1;
        break;

      default:
        // !!! TODO: general MODE_BANDS bitmap !!!
        biType = BufferedImage.TYPE_INT_RGB;
        bands  = 3;
    }

    sColor           = 255;
    color            = FOREGROUND;
    tmpColor         = new int[ bands ];
    im               = new BufferedImage( _width, _height, biType );

    setRectangle( 0, 0, width, height, INIT_COLOR );
  }

  /**
   * {@link RasterImage} re-initialization (keeps current bitmap size, mode,
   * and colormap attributes, too).
   */
  public void init ()
  {
    super.init();

    sColor = 255;
    color  = FOREGROUND;

    setRectangle( 0, 0, width, height, INIT_COLOR );
  }

  /**
   * Maximal ordinal number of colormap entry.
   *
   * @return Maximal colormap's entry number.
   */
  public int getMaxColormap ()
  {
    return colormapSize - 1;
  }

  /**
   * Copies raster graphics to a {@link BufferedImage} object.
   *
   * @return {@link BufferedImage} object.
   * @see    #setBufferedImage
   */
  public BufferedImage getBufferedImage ()
  {
    return im;
  }

  /**
   * Initializes raster graphics from a {@link BufferedImage} object.
   *
   * @param  bi Raster data.
   * @see    #getBufferedImage
   */
  public void setBufferedImage ( BufferedImage bi )
  {
    im           = bi;
    width        = bi.getWidth();
    height       = bi.getHeight();
    colormapSize = 0;
    colormapMode = MODE_RGB;

    switch ( bi.getType() )
    {
      case BufferedImage.TYPE_BYTE_GRAY:
      case BufferedImage.TYPE_BYTE_BINARY:
      case BufferedImage.TYPE_USHORT_GRAY:
        if ( !(bi.getColorModel() instanceof IndexColorModel) )
        {
          mode  = MODE_GRAY;
          bands = 1;
          break;
        }

      case BufferedImage.TYPE_BYTE_INDEXED:
      {
        mode               = MODE_COLORMAP;
        IndexColorModel cm = (IndexColorModel)im.getColorModel();
        colormapSize       = cm.getMapSize();
        colormapMode       = (cm.getNumComponents() == 3) ? MODE_RGB : MODE_RGBA;
        bands              = 1;
        break;
      }

      case BufferedImage.TYPE_INT_RGB:
      case BufferedImage.TYPE_INT_BGR:
      case BufferedImage.TYPE_3BYTE_BGR:
      case BufferedImage.TYPE_USHORT_565_RGB:
      case BufferedImage.TYPE_USHORT_555_RGB:
        mode  = MODE_RGB;
        bands = 3;
        break;

      case BufferedImage.TYPE_INT_ARGB:
      case BufferedImage.TYPE_INT_ARGB_PRE:
      case BufferedImage.TYPE_4BYTE_ABGR:
      case BufferedImage.TYPE_4BYTE_ABGR_PRE:
        mode  = MODE_RGBA;
        bands = 4;
        break;

      default:
        mode  = MODE_BAND;
        bands = bi.getSampleModel().getNumBands();
    }

    tmpColor = new int[ bands ];
  }

  /**
   * Colormap mode (valid only for {@link #MODE_COLORMAP MODE_COLORMAP} bitmap mode).
   *
   * @return Colormap mode (one of ({@code RasterGraphics.MODE_*} constants).
   * @see    RasterGraphics
   */
  public int getColormapType ()
  {
    return colormapMode;
  }

  /**
   * Sets the colormap mode (valid only for {@link #MODE_COLORMAP MODE_COLORMAP} bitmap mode).
   *
   * @param  _mode Colormap mode (one of ({@code RasterGraphics.MODE_*} constants).
   * @param  _max Maximal index of colormap entry.
   * @see    RasterGraphics
   */
  public void setColormapType ( int _mode, int _max )
  {
    if ( mode  != MODE_COLORMAP ||
         _mode != MODE_RGB && _mode != MODE_RGBA )
      throw new IllegalArgumentException();

    colormapMode = _mode;
    colormapSize = _max + 1;
  }

  /**
   * Returns the colormap segment (for color modes: "{@code getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND}").
   *
   * @param  i0 The first colormap index.
   * @param  length Length of colormap segment.
   * @param  _colors Returned color levels (each band: from {@code 0.0} to {@code 1.0}).
   * @see    RasterGraphics
   */
  public void getColormap ( int i0, int length, double[][] _colors )
  {
    if ( mode != MODE_COLORMAP )
      throw new IllegalStateException();

    IndexColorModel cm = (IndexColorModel)im.getColorModel();
    int size = cm.getMapSize();
    if ( i0 >= size ) return;
    int[] cmap = new int[ size ];
    cm.getRGBs( cmap );

      // output array size:
    if ( length > _colors.length )
      length = _colors.length;
    if ( i0 + length > size )
      length = size - i0;

      // convert the data:
    int i = 0;
    int cmapBands = (colormapMode == MODE_RGB) ? 3 : 4;
    while ( length-- > 0 )
    {
      if ( _colors[i] == null || _colors[i].length < cmapBands )
        _colors[i] = new double[ cmapBands ];
      _colors[i][0] = ((cmap[i0] >> 16) & 0xff) * INV_255;
      _colors[i][1] = ((cmap[i0] >>  8) & 0xff) * INV_255;
      _colors[i][2] = ((cmap[i0]      ) & 0xff) * INV_255;
      if ( cmapBands > 3 )
        _colors[i][3] = ((cmap[i0] >> 24) & 0xff) * INV_255;
      i0++; i++;
    }
  }

  /**
   * Returns the colormap segment (for color modes: "{@code getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND}").
   *
   * @param  i0 The first colormap index.
   * @param  length Length of colormap segment.
   * @param  _colors Returned color levels (each band: from {@code 0} to {@link #getMaxRGB()}).
   */
  public void getColormap ( int i0, int length, int[][] _colors )
  {
    if ( mode != MODE_COLORMAP )
      throw new IllegalStateException();

    IndexColorModel cm = (IndexColorModel)im.getColorModel();
    int size = cm.getMapSize();
    if ( i0 >= size ) return;

    int[] cmap = new int[ size ];
    cm.getRGBs( cmap );

      // output array size:
    if ( length > _colors.length )
      length = _colors.length;
    if ( i0 + length > size )
      length = size - i0;

      // convert the data:
    int i = 0;
    int cmapBands = (colormapMode == MODE_RGB) ? 3 : 4;
    while ( length-- > 0 )
    {
      if ( _colors[i] == null || _colors[i].length < cmapBands )
        _colors[i] = new int[ cmapBands ];
      _colors[i][0] = ((cmap[i0] >> 16) & 0xff);
      _colors[i][1] = ((cmap[i0] >>  8) & 0xff);
      _colors[i][2] = ((cmap[i0]      ) & 0xff);
      if ( cmapBands > 3 )
        _colors[i][3] = ((cmap[i0] >> 24) & 0xff);
      i0++; i++;
    }
  }

  /**
   * Sets the colormap segment (for color modes: "{@code getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND}").
   * Colormap size &amp; mode has to be set before using {@link #setColormapType}.
   *
   * @param  _colors Color levels to be set (each band: from {@code 0.0} to {@code 1.0}).
   */
  public void setColormap ( double[][] _colors )
  {
    if ( mode != MODE_COLORMAP )
      throw new IllegalStateException();

    if ( _colors == null || _colors.length < colormapSize )
    {
      LogFile.error( "Bad colormap size!" );
      return;
    }

    int cmapBands = (colormapMode == MODE_RGB) ? 3 : 4;
    if ( _colors[0].length < cmapBands )
    {
      LogFile.error( "Bad colormap format!" );
      return;
    }

      // convert colormap data to IndexColorModel-compatible format:
    byte[] cmap = new byte[ colormapSize * cmapBands ];
    int i = 0;
    int j, b;
    for ( j = 0; j < colormapSize; j++ )
      for ( b = 0; b < cmapBands; b++ )
        cmap[i++] = (byte)( _colors[j][b] * 255.0 - 128.0 );
    IndexColorModel icm = new IndexColorModel( 8, colormapSize, cmap,
                                               0, cmapBands > 3 );

      // .. and finally create new BufferedImage object (old data will be discarded)
    im = new BufferedImage( width, height,
                            BufferedImage.TYPE_BYTE_INDEXED, icm );
  }

  /**
   * Sets the colormap segment (for color modes: "{@code getColormapType()
   * == MODE_RGB || getColormapType() == MODE_RGBA || getColormapType() == MODE_BAND}").
   * Colormap size &amp; mode has to be set before using #setColormapType.
   *
   * @param  _colors Color levels to be set (each band: from {@code 0} to {@link #getMaxRGB}).
   */
  public void setColormap ( int[][] _colors )
  {
    if ( mode != MODE_COLORMAP )
      throw new IllegalStateException();

    if ( _colors == null || _colors.length < colormapSize )
    {
      LogFile.error( "Bad colormap size!" );
      return;
    }

    int cmapBands = (colormapMode == MODE_RGB) ? 3 : 4;
    if ( _colors[0].length < cmapBands )
    {
      LogFile.error( "Bad colormap format!" );
      return;
    }

      // convert colormap data to IndexColorModel-compatible format:
    byte[] cmap = new byte[ colormapSize * cmapBands ];
    int i = 0;
    int j, b;
    for ( j = 0; j < colormapSize; j++ )
      for ( b = 0; b < cmapBands; b++ )
        cmap[i++] = (byte)_colors[j][b];
    IndexColorModel icm = new IndexColorModel( 8, colormapSize, cmap,
                                               0, cmapBands > 3 );

      // .. and finally create new BufferedImage object (old data will be discarded)
    im = new BufferedImage( width, height,
                            BufferedImage.TYPE_BYTE_INDEXED, icm );
  }

  /**
   * Sets the current drawing color (for {@link #MODE_GRAY MODE_GRAY} or
   * {@link #MODE_COLORMAP MODE_COLORMAP}).
   *
   * @param  _color Current drawing color (0 to {@link #getMaxGray} or
   *         {@link #getMaxColormap}).
   * @see    RasterGraphics
   */
  public void setColor ( int _color )
  {
    sColor = Formula.clamp( _color, 0, 255 );

    if ( mode  != MODE_GRAY &&
         mode  != MODE_COLORMAP &&
         color != null )
      color[0] = color[1] = color[2] = sColor;
  }

  /**
   * Sets the current drawing color (for {@link #MODE_GRAY MODE_GRAY}).
   *
   * @param  _color Current drawing color ({@code 0.0} to {@code 1.0}).
   * @see    RasterGraphics
   */
  public void setColor ( double _color )
  {
    sColor = Formula.clamp( (int)( _color * 255.0 + 0.5 ), 0, 255 );

    if ( mode  != MODE_GRAY &&
         mode  != MODE_COLORMAP &&
         color != null )
      color[0] = color[1] = color[2] = sColor;
  }

  /**
   * Sets the current drawing color (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  _color Current drawing color (channel values between
   *                {@code 0} and {@link #getMaxRGB}).
   * @see    RasterGraphics
   */
  public void setColor ( int[] _color )
  {
    if ( _color == null ) return;

    int len = _color.length;
    if ( color        == null ||
         color.length != len )
      color = new int[ len ];

    System.arraycopy( _color, 0, color, 0, len );

    if ( mode == MODE_GRAY ||
         mode == MODE_COLORMAP )
      sColor = color[0];
  }

  /**
   * Sets the current drawing color (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  _color Current drawing color (channel values between
   *                {@code 0.0} and {@code 1.0}).
   * @see    RasterGraphics
   */
  public void setColor ( double[] _color )
  {
    if ( _color == null ) return;

    int len = _color.length;
    if ( color == null || color.length != len )
      color = new int[ len ];

    int i;
    for ( i = 0; i < len; i++ )
      color[i] = Formula.clamp( (int)( _color[i] * 255.0 + 0.5 ), 0, 255 );

    if ( mode == MODE_GRAY ||
         mode == MODE_COLORMAP )
      sColor = color[0];
  }

  /**
   * Sets the given pixel's value.
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @see    #setColor(int)
   * @see    #setColor(double)
   * @see    #setColor(int[])
   * @see    #setColor(double[])
   * @see    RasterGraphics
   */
  public void setPixel ( int x, int y )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height ) return;

    switch ( mode )
    {
      case MODE_GRAY:
      case MODE_COLORMAP:
        im.getRaster().setSample( x, y, 0, sColor );
        break;

      case MODE_RGB:
        im.getRaster().setPixel( x, y, color );
        break;

      case MODE_RGBA:
      {
        WritableRaster wr = im.getRaster();
        wr.getPixel( x, y, tmpColor );    // alpha-premultiplied

          // color * alpha + oldColor * (1-alpha):
        tmpColor[0] = (int)( color[0] * alpha + tmpColor[0] * (1.0 - alpha) + 0.5 );
        tmpColor[1] = (int)( color[1] * alpha + tmpColor[1] * (1.0 - alpha) + 0.5 );
        tmpColor[2] = (int)( color[2] * alpha + tmpColor[2] * (1.0 - alpha) + 0.5 );
        tmpColor[3] = (int)( 255      * alpha + tmpColor[3] * (1.0 - alpha) + 0.5 );

        wr.setPixel( x, y, tmpColor );
        break;
      }
    }
  }

  /**
   * Returns the pixel's value (for {@link #MODE_GRAY MODE_GRAY} or
   * {@link #MODE_COLORMAP MODE_COLORMAP}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value ({@code 0} for off-canvas access).
   * @see    RasterGraphics
   */
  public int getPixel ( int x, int y )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height ) return 0;

    return im.getRaster().getSample( x, y, 0 );
  }

  /**
   * Returns the pixel's value (for {@link #MODE_GRAY MODE_GRAY}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value ({@code 0.0} for off-canvas access).
   * @see    RasterGraphics
   */
  public double getPixelDouble ( int x, int y )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height ) return 0.0;

    return im.getRaster().getSample( x, y, 0 ) * INV_255;
  }

  /**
   * Returns the pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value.
   * @see    RasterGraphics
   */
  public void getPixel ( int x, int y, int[] _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      System.arraycopy( BACKGROUND, 0, _color, 0, Math.min( BACKGROUND.length, _color.length ) );

    else
      im.getRaster().getPixel( x, y, _color );
  }

  /**
   * Returns the pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value.
   * @see    RasterGraphics
   */
  public void getPixel ( int x, int y, double[] _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      System.arraycopy( BACKGROUND_D, 0, _color, 0, Math.min( BACKGROUND_D.length, _color.length ) );

    else
    {
      im.getRaster().getPixel( x, y, tmpColor );
      int i;
      for ( i = 0; i < _color.length; i++ )
        _color[i] = tmpColor[i] * INV_255;
    }
  }

  /** Asserts colormap cache values. */
  protected void assertCmapCache ()
  {
    if ( cmapCache == null )
    {
      IndexColorModel cm = (IndexColorModel)im.getColorModel();
      int size  = cm.getMapSize();
      cmapCache = new int[ size ];
      cm.getRGBs( cmapCache );
    }
  }

  /**
   * Returns the pixel's RGB[A] values (for all graphic modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value (if the array's length is at least 4, alpha will be returned).
   */
  public void getRGB ( int x, int y, int[] _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      System.arraycopy( BACKGROUND, 0, _color, 0, Math.min( BACKGROUND.length, _color.length ) );

    else
      switch ( mode )
      {
        case MODE_GRAY:
          _color[0] = _color[1] = _color[2] = im.getRaster().getSample( x, y, 0 );
          if ( _color.length > 3 )
            _color[3] = 255;
          return;

        case MODE_COLORMAP:
        {
          assertCmapCache();
          int index = im.getRaster().getSample( x, y, 0 );
          if ( index < 0 || index >= cmapCache.length )
            System.arraycopy( BACKGROUND, 0, _color, 0, Math.min( BACKGROUND.length, _color.length ) );

          else
          {
            _color[0] = ((cmapCache[index] >> 16) & 0xff);
            _color[1] = ((cmapCache[index] >>  8) & 0xff);
            _color[2] = ((cmapCache[index]      ) & 0xff);
            if ( _color.length > 3 )
              if ( colormapMode == MODE_RGBA )
                _color[3] = ((cmapCache[index] >> 24) & 0xff);
              else
                _color[3] = 255;
          }

          return;
        }

        case MODE_RGB:
          im.getRaster().getPixel( x, y, _color );
          if ( _color.length > 3 )
            _color[3] = 255;
          return;

        case MODE_RGBA:
          if ( _color.length >= bands )
            im.getRaster().getPixel( x, y, _color );
          else
          {
            im.getRaster().getPixel( x, y, tmpColor );
            _color[0] = tmpColor[0];
            _color[1] = tmpColor[1];
            _color[2] = tmpColor[2];
          }
          return;

        case MODE_BAND:
          if ( _color.length >= bands )
            im.getRaster().getPixel( x, y, _color );
          else
          {
            im.getRaster().getPixel( x, y, tmpColor );
            _color[0] = tmpColor[0];
            _color[1] = tmpColor[1];
            _color[2] = tmpColor[2];
            if ( _color.length > 3 )
              _color[3] = tmpColor[3];
          }
          return;
      }
  }

  /**
   * Returns the pixel's RGB[A] values (for all graphic modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Current pixel RGB[A] value (if the array's length is at least 4, alpha will be returned).
   */
  public void getRGB ( int x, int y, double[] _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      System.arraycopy( BACKGROUND_D, 0, _color, 0, Math.min( BACKGROUND_D.length, _color.length ) );

    else
      switch ( mode )
      {
        case MODE_GRAY:
          _color[0] = _color[1] = _color[2] = im.getRaster().getSample( x, y, 0 ) * INV_255;
          if ( _color.length > 3 )
            _color[3] = 1.0;
          return;

        case MODE_COLORMAP:
        {
          assertCmapCache();
          int index = im.getRaster().getSample( x, y, 0 );
          if ( index < 0 || index >= cmapCache.length )
            System.arraycopy( BACKGROUND_D, 0, _color, 0, Math.min( BACKGROUND_D.length, _color.length ) );

          else
          {
            _color[0] = ((cmapCache[index] >> 16) & 0xff) * INV_255;
            _color[1] = ((cmapCache[index] >>  8) & 0xff) * INV_255;
            _color[2] = ((cmapCache[index]      ) & 0xff) * INV_255;
            if ( _color.length > 3 )
              if ( colormapMode == MODE_RGBA )
                _color[3] = ((cmapCache[index] >> 24) & 0xff) * INV_255;
              else
                _color[3] = 1.0;
          }

          return;
        }

        case MODE_RGB:
          im.getRaster().getPixel( x, y, tmpColor );
          _color[0] = tmpColor[0] * INV_255;
          _color[1] = tmpColor[1] * INV_255;
          _color[2] = tmpColor[2] * INV_255;
          if ( _color.length > 3 )
            _color[3] = 1.0;
          return;

        case MODE_RGBA:
          im.getRaster().getPixel( x, y, tmpColor );
          _color[0] = tmpColor[0] * INV_255;
          _color[1] = tmpColor[1] * INV_255;
          _color[2] = tmpColor[2] * INV_255;
          if ( _color.length > 3 )
            _color[3] = tmpColor[3] * INV_255;
          return;

        case MODE_BAND:
          im.getRaster().getPixel( x, y, tmpColor );
          _color[0] = tmpColor[0] * INV_255;
          _color[1] = tmpColor[1] * INV_255;
          _color[2] = tmpColor[2] * INV_255;
          if ( _color.length > 3 )
            if ( bands < 4 )
              _color[3] = 1.0;
            else
              _color[3] = tmpColor[3] * INV_255;
          return;
      }
  }

  /**
   * Returns the pixel's gray value (for all modes).
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value ({@code 0} for off-canvas access).
   */
  public int getGray ( int x, int y )
  {
    if ( x >= 0 && x < width &&
         y >= 0 && y < height )

      switch ( mode )
      {
        case MODE_GRAY:
          return im.getRaster().getSample( x, y, 0 );

        case MODE_COLORMAP:
        {
          assertCmapCache();
          int index = im.getRaster().getSample( x, y, 0 );
          if ( index < 0 || index >= cmapCache.length )
            return 0;
          return Formula.rgbToGray( ((cmapCache[index] >> 16) & 0xff),
                                    ((cmapCache[index] >>  8) & 0xff),
                                    ((cmapCache[index]      ) & 0xff) );
        }

        case MODE_RGB:
        case MODE_RGBA:
        case MODE_BAND:
          im.getRaster().getPixel( x, y, tmpColor );
          return Formula.rgbToGray( tmpColor[0], tmpColor[1], tmpColor[2] );
      }

    return 0;
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_GRAY MODE_GRAY} or
   * {@link #MODE_COLORMAP MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0} and {@link #getMaxGray} or
   *                {@link #getMaxColormap}).
   */
  public void putPixel ( int x, int y, int _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      return;

    im.getRaster().setSample( x, y, 0, _color );
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_GRAY MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0.0} and {@code 1.0}).
   */
  public void putPixel ( int x, int y, double _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      return;

    int col = (int)( _color * 255.0 + 0.5 );
    im.getRaster().setSample( x, y, 0, col );
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0} and {@link #getMaxRGB}).
   */
  public void putPixel ( int x, int y, int[] _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height )
      return;

    im.getRaster().setPixel( x, y, _color );
  }

  /**
   * Sets the given pixel's value (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  _color Pixel color (between {@code 0.0} and {@code 1.0}).
   */
  public void putPixel ( int x, int y, double[] _color )
  {
    if ( x < 0 || x >= width ||
         y < 0 || y >= height ) return;

    int len = Math.min( _color.length, bands );
    int i;
    for ( i = 0; i < len; i++ )
      tmpColor[i] = (int)( _color[i] * 255.0 + 0.5 );

    im.getRaster().setPixel( x, y, tmpColor );
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y].
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @see    #setColor(int)
   * @see    #setColor(double)
   * @see    #setColor(int[])
   * @see    #setColor(double[])
   * @see    RasterGraphics
   */
  public void setHLine ( int x1, int x2, int y )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 ) x1 = 0;
    if ( x2 > width ) x2 = width;
    WritableRaster wr = im.getRaster();

    switch ( mode )
    {
      case MODE_GRAY:
      case MODE_COLORMAP:
        while ( x1 < x2 )
          wr.setSample( x1++, y, 0, sColor );
        break;

      case MODE_RGBA:
        while ( x1 < x2 )
        {
          wr.getPixel( x1, y, tmpColor ); // alpha-premultiplied

            // color * alpha + tmpColor * (1-alpha):
          tmpColor[0] = (int)( color[0] * alpha + tmpColor[0] * (1.0 - alpha) + 0.5 );
          tmpColor[1] = (int)( color[1] * alpha + tmpColor[1] * (1.0 - alpha) + 0.5 );
          tmpColor[2] = (int)( color[2] * alpha + tmpColor[2] * (1.0 - alpha) + 0.5 );
          tmpColor[3] = (int)( 255      * alpha + tmpColor[3] * (1.0 - alpha) + 0.5 );

          wr.setPixel( x1++, y, tmpColor );
        }
        break;

      case MODE_RGB:
      case MODE_BAND:
        while ( x1 < x2 )
          wr.setPixel( x1++, y, color );
        break;
    }
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_GRAY MODE_GRAY} or {@link #MODE_COLORMAP MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color ({@code 0} to {@link #getMaxGray} or
   *         {@link #getMaxColormap}).
   * @see    RasterGraphics
   */
  public void setHLine ( int x1, int x2, int y, int _color )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 ) x1 = 0;
    if ( x2 > width ) x2 = width;
    WritableRaster wr = im.getRaster();
    while ( x1 < x2 )
      wr.setSample( x1++, y, 0, _color );
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_GRAY MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color ({@code 0.0} to {@code 1.0}).
   * @see    RasterGraphics
   */
  public void setHLine ( int x1, int x2, int y, double _color )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 ) x1 = 0;
    if ( x2 > width ) x2 = width;
    WritableRaster wr = im.getRaster();
    int col = (int)( _color * 255.0 + 0.5 );
    while ( x1 < x2 )
      wr.setSample( x1++, y, 0, col );
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA} or
   * and {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color (channel values between {@code 0} and {@link #getMaxRGB}).
   * @see    RasterGraphics
   */
  public void setHLine ( int x1, int x2, int y, int[] _color )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 ) x1 = 0;
    if ( x2 > width ) x2 = width;
    WritableRaster wr = im.getRaster();

    switch ( mode )
    {
      case MODE_GRAY:
      case MODE_COLORMAP:
        while ( x1 < x2 )
          wr.setSample( x1++, y, 0, _color[0] );
        break;

      case MODE_RGB:
      case MODE_BAND:
        while ( x1 < x2 )
          wr.setPixel( x1++, y, _color );
        break;

      case MODE_RGBA:
      {
        double colR = _color[0] * alpha;
        double colG = _color[1] * alpha;
        double colB = _color[2] * alpha;
        double colA = (_color.length < 4) ? 255.0 * alpha : _color[3] * alpha;
        while ( x1 < x2 )
        {
          wr.getPixel( x1, y, tmpColor ); // alpha-premultiplied

            // color * alpha + tmpColor * (1-alpha):
          tmpColor[0] = (int)( colR + tmpColor[0] * (1.0 - alpha) + 0.5 );
          tmpColor[1] = (int)( colG + tmpColor[1] * (1.0 - alpha) + 0.5 );
          tmpColor[2] = (int)( colB + tmpColor[2] * (1.0 - alpha) + 0.5 );
          tmpColor[3] = (int)( colA + tmpColor[3] * (1.0 - alpha) + 0.5 );

          wr.setPixel( x1++, y, tmpColor );
        }
        break;
      }
    }
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y] (for
   * {@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA} or
   * {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _color Drawing color (channel values between {@code 0.0} and {@code 1.0}).
   * @see    RasterGraphics
   */
  public void setHLine ( int x1, int x2, int y, double[] _color )
  {
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 ) x1 = 0;
    if ( x2 > width ) x2 = width;
    WritableRaster wr = im.getRaster();

    switch ( mode )
    {
      case MODE_GRAY:
      case MODE_COLORMAP:
      {
        int col = (int)( _color[0] * 255.0 + 0.5 );
        while ( x1 < x2 )
          wr.setSample( x1++, y, 0, col );
        break;
      }

      case MODE_RGB:
      case MODE_BAND:
        for ( int i = 0; i < bands; i++ )
          tmpColor[i] = (int)( _color[i] * 255.0 + 0.5 );
        while ( x1 < x2 )
          wr.setPixel( x1++, y, tmpColor );
        break;

      case MODE_RGBA:
      {
        double colR = _color[0] * 255.0 * alpha;
        double colG = _color[1] * 255.0 * alpha;
        double colB = _color[2] * 255.0 * alpha;
        double colA = alpha * ((_color.length < 4) ? 255.0 : _color[3] * 255.0);
        while ( x1 < x2 )
        {
          wr.getPixel( x1, y, tmpColor ); // alpha-premultiplied

            // color * alpha + tmpColor * (1-alpha):
          tmpColor[0] = (int)( colR + tmpColor[0] * (1.0 - alpha) + 0.5 );
          tmpColor[1] = (int)( colG + tmpColor[1] * (1.0 - alpha) + 0.5 );
          tmpColor[2] = (int)( colB + tmpColor[2] * (1.0 - alpha) + 0.5 );
          tmpColor[3] = (int)( colA + tmpColor[3] * (1.0 - alpha) + 0.5 );

          wr.setPixel( x1++, y, tmpColor );
        }
        break;
      }
    }
  }

  /**
   * Sets the horizontal line's values (for {@link #MODE_GRAY MODE_GRAY} or
   * {@link #MODE_COLORMAP MODE_COLORMAP}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between {@code 0} and {@link #getMaxGray} or
   *                 {@link #getMaxColormap}).
   * @see    RasterGraphics
   */
  public void putHLine ( int x1, int x2, int y, int[] _colors )
  {
    int offset = 0;                       // pixel index in _colors[]
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 )
    {
      offset = -x1;
      x1 = 0;
    }
    if ( x2 > width ) x2 = width;

    WritableRaster wr = im.getRaster();
    while ( x1 < x2 )
    {
      wr.setSample( x1++, y, 0, _colors[offset] );
      offset++;
    }
  }

  /**
   * Sets the horizontal line's values (for {@link #MODE_GRAY MODE_GRAY}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between {@code 0.0} and {@code 1.0}).
   * @see    RasterGraphics
   */
  public void putHLine ( int x1, int x2, int y, double[] _colors )
  {
    int offset = 0;                       // pixel index in _colors[]
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 )
    {
      offset = -x1;
      x1 = 0;
    }
    if ( x2 > width ) x2 = width;

    WritableRaster wr = im.getRaster();
    while ( x1 < x2 )
    {
      wr.setSample( x1++, y, 0, (int)( _colors[offset] * 255.0 + 0.5 ) );
      offset++;
    }
  }

  /**
   * Sets the horizontal line's values (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between {@code 0} and {@link #getMaxGray}).
   * @see    RasterGraphics
   */
  public void putHLine ( int x1, int x2, int y, int[][] _colors )
  {
    int offset = 0;                       // pixel index in _colors[]
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 )
    {
      offset = -x1;
      x1 = 0;
    }
    if ( x2 > width ) x2 = width;

    WritableRaster wr = im.getRaster();
      // write the data:
    while ( x1 < x2 )
      wr.setPixel( x1++, y, _colors[offset++] );
  }

  /**
   * Sets the horizontal line's values (for {@link #MODE_RGB MODE_RGB},
   * {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   * <p>Ignores any off-canvas access.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Line colors (between {@code 0.0} and {@code 1.0}).
   * @see    RasterGraphics
   */
  public void putHLine ( int x1, int x2, int y, double[][] _colors )
  {
    int offset = 0;                       // pixel index in _colors[]
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 ||
         y < 0 || y >= height ) return;

    if ( x1 < 0 )
    {
      offset = -x1;
      x1 = 0;
    }
    if ( x2 > width ) x2 = width;

    WritableRaster wr = im.getRaster();

    int i;
    int len = Math.min( _colors[0].length, bands );
    while ( x1 < x2 )
    {
      for ( i = 0; i < len; i++ )
        tmpColor[i] = (int)( _colors[offset][i] * 255.0 + 0.5 );
      wr.setPixel( x1++, y, tmpColor );
      offset++;
    }
  }

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_GRAY MODE_GRAY} or {@link #MODE_COLORMAP MODE_COLORMAP}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors ({@code 0} to {@link #getMaxGray} or
   *                 {@link #getMaxColormap}, {@code 0} for off-canvas access).
   * @see    RasterGraphics
   */
  public void getHLine ( int x1, int x2, int y, int[] _colors )
  {
    int offset = 0;
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 || y < 0 || y >= height )
    {
      Arrays.fill( _colors, 0, x2 - x1, 0 );
      return;
    }

    Raster r = im.getData( new Rectangle( x1, y, x2-x1, 1 ) );
    while ( x1 < x2 )
    {
      int[] col;
      if ( x1 > width || x1 < 0 )         // off canvas
      {
        col = new int[ 1 ];
        col[0] = 0;
      }
      else
      {
        col = r.getPixel( x1++, y, (int[])null );
        if ( col.length > 1 )             // more channels than one? RGB? RGBA? make grey by averaging ...
        {
          for( int i = 0; i < col.length; i++ )
            col[0] += col[i];
          col[0] /= col.length;
        }
      }
      _colors[offset] = col[0];
      offset++;
    }
  }

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_GRAY MODE_GRAY}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors ({@code 0.0} to {@code 1.0},
   *                 {@code 0.0} for off-canvas access).
   * @see    RasterGraphics
   */
  public void getHLine ( int x1, int x2, int y, double[] _colors )
  {
    int offset = 0;
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 || y < 0 || y >= height )
    {
      for ( int i = 0; i < x2 - x1; i++ )
        _colors[i] = 0.0;                 // fill by zeroes
      return;
    }

    Raster r = im.getData( new Rectangle( x1, y, x2-x1, 1 ) );
    while ( x1 < x2 )
    {
      double[] col;
      if ( x1 > width || x1 < 0 )         // off canvas
      {
        col = new double[ 1 ];
        col[0] = 0;
      }
      else
      {
        col = r.getPixel( x1++, y, (double[])null );
        if ( col.length > 1 )             // more channels than one? RGB? RGBA? make grey by averaging ...
        {
          for ( int i = 0; i < col.length; i++ )
            col[0] += col[i];
          col[0] /= col.length;
        }
      }
      _colors[offset] = col[0] * INV_255;
      offset++;
    }
  }

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors (channel values between {@code 0} and
   *                 {@link #getMaxRGB}, {@code 0} for off-canvas access).
   * @see    RasterGraphics
   */
  public void getHLine ( int x1, int x2, int y, int[][] _colors )
  {
    if ( bands < 3 || bands > 4 )
      throw new JaGrLibException( "getHLine (...,int[][]) needs RGB, RGBA or BAND image",
                                   new Throwable( "getHLine (...,int[][])" ) );

    int offset = 0;
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 || y < 0 || y >= height )
    {
      for ( int i = 0; i < x2 - x1; i++ )
        Arrays.fill( _colors[i], 0 );
      return;
    }

    Raster r = im.getData( new Rectangle( x1, y, x2-x1, 1 ) );
    while ( x1 < x2 )
    {
      int[] col;
      if ( x1 > width || x1 < 0 )         // off canvas
      {
        col = new int[ bands ];
      }
      else
      {
        col = r.getPixel( x1++, y, (int[])null );
      }
      for ( int j = 0; j < col.length; j++ )
        _colors[offset][j] = col[j];
      offset++;
    }
  }

  /**
   * Returns the values for the given horizontal line (for
   * {@link #MODE_RGB MODE_RGB}, {@link #MODE_RGBA MODE_RGBA} or {@link #MODE_BAND MODE_BAND}).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  _colors Returned colors (channel values between {@code 0.0} and
   *                 {@code 1.0}, {@code 0.0} for off-canvas access).
   * @see    RasterGraphics
   */
  public void getHLine ( int x1, int x2, int y, double[][] _colors )
  {
    if ( bands < 3 || bands > 4 )
      throw new JaGrLibException( "getHLine (...,double[][]) needs RGB, RGBA or BAND image",
                                  new Throwable( "getHLine (...,double[][])" ) );

    int offset = 0;
    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( x1 >= width || x2 <= 0 || y < 0 || y >= height )
    {
      for ( int i = 0; i < x2 - x1; i++ )
        Arrays.fill( _colors[i], 0.0 );
      return;
    }

    Raster r = im.getData( new Rectangle( x1, y, x2-x1, 1 ) );
    while ( x1 < x2 )
    {
      int[] col;
      if ( x1 > width || x1 < 0 )         // off canvas
        col = new int[ bands ];
      else
        col = r.getPixel( x1++, y, (int[])null );

      for ( int i = 0; i < col.length; i++ )
        _colors[offset][i] = col[i] * INV_255;

      offset++;
    }
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "RasterImage";

  /** Object description. */
  private final static String DESCRIPTION = "2D raster image storage (using standard BufferedImage).";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( WIDTH, TYPE_INTEGER, "Bitmap width in pixels", true );
        t.propDefault( 512 );
      t.propEnd();
      t.propBegin( HEIGHT, TYPE_INTEGER, "Bitmap height in pixels", true );
        t.propDefault( 512 );
      t.propEnd();
      t.propBegin( COLORMODE, TYPE_INTEGER, "Pixel color mode", true );
        t.propDefault( RasterGraphics.MODE_RGB );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "RGB",      RasterGraphics.MODE_RGB,      "RGB color mode" );
        t.propEnum( "RGBA",     RasterGraphics.MODE_RGBA,     "RGBA color mode (transparency)" );
        t.propEnum( "Gray",     RasterGraphics.MODE_GRAY,     "Gray mode" );
        t.propEnum( "Colormap", RasterGraphics.MODE_COLORMAP, "Colormapped mode" );
      t.propEnd();
      t.propBegin( BANDS, TYPE_INTEGER, "Number of color bands", true );
        t.propDefault( 0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
