package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Ray casting of CSG scenes.
 *
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/RayCasting.java">RayCasting.java</a>
 */
public class RayCasting extends DefaultImageFunction
{
  //==========================================================================
  //  Data:

  /** Projection-plane linear transform - horizontal additive term. */
  protected double xA = 0.0;

  /** Projection-plane linear transform - horizontal multiplicator. */
  protected double xK = 1.0;

  /** Projection-plane linear transform - vertical additive term. */
  protected double yA = 0.0;

  /** Projection-plane linear transform - vertical multiplicator. */
  protected double yK = 1.0;

  /** Background color (used when ray won't hit anything in a scene). */
  {
    background = new double[] { 0.05, 0.08, 0.1 };
  }

  /** Accelerator: actual Ray-tracing scene (everything but intersections). */
  protected RTScene scene = null;

  /** Accelerator: object used for intersections. */
  protected Intersectable inter = null;

  /** Accelerator: camera object. */
  protected RayGenerator camera = null;

  /** Hash-multiplier for number of light sources. */
  protected final static long HASH_LIGHT   = 101L;

  /** Hash-multiplier for textures. */
  protected final static long HASH_TEXTURE = 11L;

  //==========================================================================
  //  Computing restrictions:

  /** Compute shading (using the defined {@link BSSRDF}). */
  protected boolean doShading = true;

  /** Compute shadows (if {@code false} all light sources will be considered visible). */
  protected boolean doShadows = true;

  /** Take textures into account. */
  protected boolean doTextures = true;

  /** Mask used in {@link MicroFacet#assertAttributes}. */
  protected int dataMask;

  //==========================================================================
  //  Support:

  protected long hashLights ( LightSource[] lights )
  {
    if ( lights == null ) return 0L;

    long sum = 0L;
    for ( LightSource light : lights )
    {
      sum += sum;
      if ( light != null ) sum += HASH_LIGHT;
    }

    return sum;
  }

  /**
   * Asserts the {@link #scene} accelerator and
   * does first-time algorithm setup.
   *
   * @return {@code true} if failed.
   */
  protected boolean assertScene ()
  {
    if ( scene != null ) return false;

    scene = (RTScene)getInterface( PL_OUTPUT, IFACE+"RTScene" );
    if ( scene == null ) return true;

      // further setup:
    background = scene.getBackground();

    dataMask = MicroFacet.VALID_COLOR;

    if ( doShading )
      dataMask |= MicroFacet.VALID_NORMAL |
                  MicroFacet.VALID_BSSRDF;

    if ( doTextures )
      dataMask |= MicroFacet.VALID_TEXTURES;

    if ( inter == null )
    {
      inter = (Intersectable)getInterface( PL_INTERSECTION, IFACE+"Intersectable" );
      if ( inter == null )
        inter = scene;
    }

    LightSource[] lights = scene.getLights();
    if ( lights != null )
      for ( LightSource ls : lights )
        ls.setIntersectable( doShadows ? inter : null );

    camera = scene.getCamera();

    return( camera == null );
  }

  //==========================================================================
  //  Interface TimeDependent:

  /**
   * Retrieve the valid bounds for time variable.
   *
   * @param  pre Pre-allocated {@code double[2]} array to hold the result (can be
   *             {@code null}).
   * @return {@code result[0]} .. minimum time value, {@code result[1]} .. maximum
   *         time value.
   * @see    #getTime
   * @see    #setTime
   */
  @Override
  public double[] getTimeInterval ( double[] pre )
  {
    if ( assertScene() )
      return null;

    return scene.getTimeInterval( pre );
  }

  /**
   * Returns the actual time value for the object (used in animations).
   *
   * @return Actual time (in seconds).
   * @see    #setTime
   */
  @Override
  public double getTime ()
  {
    if ( assertScene() )
      return 0.0;

    return scene.getTime();
  }

  /**
   * Sets the actual time value for the object (used in animations).
   *
   * @param  time New scene time (in seconds).
   * @see    #getTime
   * @see    #getTimeInterval
   */
  @Override
  public void setTime ( double time )
  {
    if ( !assertScene() )
      scene.setTime( time );
  }

  //==========================================================================
  //  Interface ImageFunction:

  /**
   * Sets bounds (scaling factors) for the image function (e.g. projection-plane mapping).
   *
   * @param  Xmin Minimum X coordinate.
   * @param  Xmax Maximum X coordinate.
   * @param  Ymin Minimum Y coordinate.
   * @param  Ymax Maximum Y coordinate.
   * @see    #getBounds
   */
  @Override
  public void setBounds ( double Xmin, double Xmax, double Ymin, double Ymax )
  {
    if ( assertScene() )
      return;

      // projection-plane bounds:
    xMin = Xmin; xMax = Xmax;
    if ( xMax <= xMin ) xMax = xMin + 1.0;
    yMin = Ymin; yMax = Ymax;
    if ( yMax <= yMin ) yMax = yMin + 1.0;
    camera.setAspectRatio( (xMax - xMin) / (yMax - yMin) );

      // projection-plane linear transform:
    double[] vp = camera.getViewport( null );
    if ( vp == null ) return;

    xK = (vp[1] - vp[0]) / (xMax - xMin);
    xA = vp[0] - xMin * xK;
    yK = (vp[3] - vp[2]) / (yMax - yMin);
    yA = vp[2] - yMin * yK;
  }

  /**
   * Computes one image sample. Passes parameters for hidden sampling.
   *
   * @param  x X coordinate of the sample.
   * @param  y Y coordinate of the sample.
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (sample signature) used for adaptive subsampling.
   * @see    #getSample(double,double,double[])
   * @see    #getBounds
   */
  @Override
  public long getSample ( double x, double y, int order, int total, double[] color )
  {
      // initial color:
    if ( color == null ) return -1L;
    Arrays.fill( color, 0.0 );

    if ( assertScene() )
      return 0L;

      // compute one ray:
    double[] p0 = new double[ 3 ];        // ray coordinates
    double[] p1 = new double[ 3 ];
    if ( !camera.getRay( xA + xK * x, yA + yK * y, order, total, p0, p1 ) )
      return 11L;

      // get the intersection:
    MicroFacet mf = inter.intersection( p0, p1 );

    if ( mf == null )                     // no intersection => use background color
    {
      System.arraycopy( background, 0, color, 0, Math.min( color.length, background.length ) );
      return 1L;
    }

      // compute color of the intersection point:
    if ( mf.assertAttributes( dataMask ) != 0 ) return 0L;

      // hash code:
    long hash = mf.sceneNode.hashCode();  // init: hash code of the intersected solid

      // apply textures:
    if ( doTextures &&
         mf.textures != null )
      for ( Texture tex : mf.textures )   // apply one texture
        hash = hash * HASH_TEXTURE + tex.compute( mf );

    if ( doShading )
    {
        // and finally apply the light model:
      Arrays.fill( color, 0.0 );
      mf.bssrdf.setColor( mf.color );
      LightSource[] lights = scene.getLightsToMicroFacet( mf, order, total );
      p1[0] = -p1[0];
      p1[1] = -p1[1];
      p1[2] = -p1[2];
//      mf.brssdf.lightSum( mf, order, total, lights, p1, color );

      return( hash + hashLights( lights ) );
    }

      // w/o shading:
    System.arraycopy( mf.color, 0, color, 0, Math.min( color.length, mf.color.length ) );
    return hash;
  }

  //==========================================================================
  //  Interface Property:

  public final static String SHADING   = "Shading";
  public final static String SHADOWS   = "Shadows";
  public final static String TEXTURES  = "Textures";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // BACKGROUND_COLOR:
    if ( key.compareTo(BACKGROUND_COLOR) == 0 &&
         value instanceof double[] )
    {
      background = (double[])value;
      return;
    }
      // SHADING:
    if ( key.compareTo(SHADING) == 0 )
    {
      doShading = booleanProperty( value, doShading );
      scene     = null;
      return;
    }
      // SHADOWS:
    if ( key.compareTo(SHADOWS) == 0 )
    {
      doShadows = booleanProperty( value, doShadows );
      scene     = null;
      return;
    }
      // TEXTURES:
    if ( key.compareTo(TEXTURES) == 0 )
    {
      doTextures = booleanProperty( value, doTextures );
      scene      = null;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(BACKGROUND_COLOR) == 0 )
      return background;
    if ( key.compareTo(SHADING) == 0 )
      return doShading;
    if ( key.compareTo(SHADOWS) == 0 )
      return doShadows;
    if ( key.compareTo(TEXTURES) == 0 )
      return doTextures;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Ray-casting";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ImageFunctionToRTSceneAndIntersectable";

  /** Object description. */
  private final static String DESCRIPTION = "Ray-casting of CSG scene.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + "." + C_RENDER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,        IFACE+"ImageFunction" );
      t.newInputPlug(  PL_PROPERTY,     IFACE+"Property" );
      t.newOutputPlug( PL_OUTPUT,       IFACE+"RTScene" );
      t.newOutputPlug( PL_INTERSECTION, IFACE+"Intersectable" );
        // Property:
      t.propBegin( SHADING, TYPE_BOOLEAN, "Use light models at all?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( SHADOWS, TYPE_BOOLEAN, "Use shadow rays (for casted shadows)?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( TEXTURES, TYPE_BOOLEAN, "Compute textures?", true );
        t.propDefault( true );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}