package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Ray tracing of CSG scenes.
 *
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.13
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/RayTracing.java">RayTracing.java</a>
 */
public class RayTracing extends RayCasting
{
  //==========================================================================
  //  Data:

  /** Hash-multiplier for refracted rays. */
  protected final static long HASH_REFRACT  = 13L;

  /** Hash-multiplier for reflected rays. */
  protected final static long HASH_REFLECT  = 17L;

  /** Actual sample order. */
  protected int actOrder;

  /** Actual cardinality of sample set. */
  protected int actTotal;

  //==========================================================================
  //  Computing restrictions:

  /** Use reflected rays (recursion). */
  protected boolean doReflections = true;

  /** Use refracted rays - in case of transparent material (recursion). */
  protected boolean doRefractions = true;

  /**
   * Recursion-termination parameter - maximal recursion depth.
   *
   * @see    #minImportance
   */
  protected int maxLevel = 12;

  /**
   * Recursion-termination parameter - minimal importance value which causes ray reflection
   * and/or refraction.
   *
   * @see    #maxLevel
   */
  protected double minImportance = 0.05;

  //==========================================================================
  //  Support:

  /**
   * Asserts the {@link #scene} accelerator and
   * does first-time algorithm setup.
   *
   * @return {@code true} if failed.
   */
  @Override
  protected boolean assertScene ()
  {
    if ( scene != null ) return false;

    if ( super.assertScene() ) return true;

    if ( doReflections ||
         doRefractions )
      dataMask |= MicroFacet.VALID_NORMAL |
                  MicroFacet.VALID_BSSRDF;

    return false;
  }

  /**
   * Recursive shading function - computes color contribution of the given ray (shot from the
   * origin {@code p0} into direction vector {@code p1}). Recursion is stopped
   * by a hybrid method: {@code importance} and {@code level} are checked.
   *
   * @param  level Actual recursion depth.
   * @param  importance Importance of the actual ray.
   * @param  p0 Ray origin.
   * @param  p1 Direction vector of the ray.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (ray sub-signature) used for adaptive subsampling.
   * @see    #minImportance
   * @see    #maxLevel
   * @see    #scene
   * @see    #inter
   */
  protected long shade ( int level, double importance, double[] p0, double[] p1,
                         double[] color )
  {
      // get the intersection:
    MicroFacet mf = inter.intersection( p0, p1 );

    if ( mf == null )                     // no intersection => use background color
    {
      System.arraycopy( background, 0, color, 0, Math.min( color.length, background.length ) );
      return 1L;
    }

      // compute color of the intersection point:
    if ( mf.assertAttributes( dataMask ) != 0 ) return 0L;

    if ( LogFile.tracing )
    {
      LogFile.log( String.format( Locale.US, "Intersection%d at t = %.4f with '%s' (%s,%s), p1.N: %.3f, mask: %04x",
                                  level, mf.t, mf.geometry.getClass().getName(),
                                  mf.enter ? "enter" : "leave", mf.front ? "front" : "back",
                                  Geometry.dotProduct3D( p1, mf.normal ), dataMask ) );
      if ( mf.assertAttributes( MicroFacet.VALID_COORD ) == 0 )
        LogFile.log( String.format( Locale.US, "  A: [ %.4f, %.4f, %.4f ], p1: [ %.4f, %.4f, %.4f ], N: [ %.4f, %.4f, %.4f ]",
                     mf.coord[0], mf.coord[1], mf.coord[2],
                     p1[0], p1[1], p1[2],
                     mf.normal[0], mf.normal[1], mf.normal[2] ) );
    }

      // hash code:
    long hash = mf.sceneNode.hashCode();  // init: hash code of the intersected solid

      // apply textures:
    if ( doTextures &&
         mf.textures != null )
      for ( Texture tex : mf.textures )   // apply one texture
        hash = hash * HASH_TEXTURE + tex.compute( mf );

    int i;
    Arrays.fill( color, 0.0 );
    p1[0] = -p1[0];
    p1[1] = -p1[1];
    p1[2] = -p1[2];
    LightSource[] lights = null;

    mf.bssrdf.setColor( mf.color );
    if ( doShading )
    {
        // apply all the light sources:
      lights = scene.getLightsToMicroFacet( mf, actOrder, actTotal );
      //mf.bssrdf.lightSum( mf, actOrder, actTotal, lights, p1, color );
      hash += hashLights( lights );
    }
    else
    {
        // apply only ambient light source[s]:
      lights = scene.getLights();
      if ( lights != null )
      {
        i = 0;
        for ( LightSource ls : lights )
          if ( ls != null &&
               ls instanceof AmbientLightSource )
            lights[i++] = ls;
        if ( i < lights.length )
          lights[i] = null;
        //mf.bssrdf.lightSum( mf, actOrder, actTotal, lights, p1, color );
        hash += hashLights( lights );
      }
    }

    if ( LogFile.tracing )
    {
      i = 0;
      while ( lights != null && i < lights.length && lights[i] != null ) i++;
      LogFile.log( String.format( Locale.US, "Base surface color: [ %.3f, %.3f, %.3f ], lights: %d",
                                  color[0], color[1], color[2], i ) );
    }

      // check the recursion depth:
    if ( level++ >= maxLevel ||
         !doReflections && !doRefractions )
      return hash;                        // no further recursion

    double maxK;
    final double comp[] = mf.bssrdf.getOrigColor( null );
                                          // color of a partial light component
    final int bands = comp.length;
    double newImportance;
    double[] r = null;

    if ( doReflections )                  // trying to shoot a reflected ray..
    {
      r = Geometry.specularReflection( mf.normal, p1, null );
      double[] ks = mf.bssrdf.colorBSSRDF( true, mf, p1, r, BSSRDF.COMP_SPECULAR, null );

      if ( ks != null )
      {
        maxK = ks[0];
        for ( i = 1; i < bands; i++ )
          if ( ks[i] > maxK ) maxK = ks[i];

        newImportance = importance * maxK;
        if ( newImportance >= minImportance ) // do compute the reflected ray
        {
          if ( LogFile.tracing )
            LogFile.log( String.format( Locale.US, "Reflected ray from [ %.4f, %.4f, %.4f ] to [ %.4f, %.4f, %.4f ], importance: %.4f",
                                        mf.coord[0], mf.coord[1], mf.coord[2],
                                        r[0], r[1], r[2], newImportance ) );

          hash += HASH_REFLECT * shade( level, newImportance, mf.coord, r, comp );

          if ( LogFile.tracing )
            LogFile.log( String.format( Locale.US, "Reflected ray color: [ %.3f, %.3f, %.3f ], coef: [ %.3f, %.3f, %.3f ]",
                                        comp[0], comp[1], comp[2],
                                        ks[0], ks[1], ks[2] ) );

          for ( i = 0; i < bands; i++ )
            color[i] += ks[i] * comp[i];
        }
      }
    }

    if ( doRefractions )                  // trying to shoot a refracted ray..
    {
      maxK = DefaultProperty.doubleProperty( mf.bssrdf, BSSRDF.MATERIAL_KT, 0.0 );
      newImportance = importance * maxK;
      if ( newImportance < minImportance )
        return hash;

        // refracted ray:
      double n = DefaultProperty.doubleProperty( mf.bssrdf, BSSRDF.MATERIAL_N, 1.0 );

      if ( (r = Geometry.specularRefraction( mf.normal, n, p1, r )) == null )
        return hash;

      if ( LogFile.tracing )
        LogFile.log( String.format( Locale.US, "Refracted ray from [ %.4f, %.4f, %.4f ] to [ %.4f, %.4f, %.4f ], importance: %.4f",
                                    mf.coord[0], mf.coord[1], mf.coord[2],
                                    r[0], r[1], r[2], newImportance ) );

      hash += HASH_REFRACT * shade( level, newImportance, mf.coord, r, comp );

      if ( LogFile.tracing )
        LogFile.log( String.format( Locale.US, "Refracted ray color: [ %.3f, %.3f, %.3f ], coef: %.3f",
                                    comp[0], comp[1], comp[2], maxK ) );

      for ( i = 0; i < bands; i++ )
        color[i] += maxK * comp[i];
    }

    return hash;
  }

  //==========================================================================
  //  Interface TimeDependent:

  //  Was inherited completely

  //==========================================================================
  //  Interface ImageFunction:

  //  setBounds() was inherited
  //  getBounds() was inherited

  /**
   * Computes one image sample. Passes parameters for hidden sampling.
   *
   * @param  x X coordinate of the sample.
   * @param  y Y coordinate of the sample.
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (sample signature) used for adaptive subsampling.
   * @see    #getSample(double,double,double[])
   * @see    #getBounds
   */
  @Override
  public long getSample ( double x, double y, int order, int total, double[] color )
  {
      // initial color:
    if ( color == null ) return -1L;
    Arrays.fill( color, 0.0 );

    if ( assertScene() )
      return 0L;

      // verbose tracing:
    if ( LogFile.debugging &&
         LogFile.intProperty( "y", -1 ) == (int)Math.floor( y ) &&
         LogFile.intProperty( "x", -1 ) == (int)Math.floor( x ) )
    {
      LogFile.tracing = true;
      LogFile.log( "Tracing pixel [ " + (int)Math.floor( x ) + ", " + (int)Math.floor( y ) + " ]:" );
    }

      // compute the primary ray:
    double[] p0 = new double[ 3 ];        // ray coordinates
    double[] p1 = new double[ 3 ];
    if ( !camera.getRay( xA + xK * x, yA + yK * y, order, total, p0, p1 ) )
      return 11L;

    actOrder  = order;
    actTotal  = total;
    long hash = shade( 0, 1.0, p0, p1, color );

    if ( LogFile.tracing )
    {
      LogFile.log( String.format( Locale.US, "Sample color: [ %.3f, %.3f, %.3f ]", color[0], color[1], color[2] ) );
      LogFile.tracing = false;
    }

    return hash;
  }

  //  getSample(double,double,double[]) was inherited

  //==========================================================================
  //  Interface Property:

  public final static String REFLECTIONS    = "Reflections";
  public final static String REFRACTIONS    = "Refractions";
  public final static String MAX_RECURSION  = "Max recursion";
  public final static String MIN_IMPORTANCE = "Min importance";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // REFLECTIONS:
    if ( key.compareTo(REFLECTIONS) == 0 )
    {
      doReflections = booleanProperty( value, doReflections );
      scene         = null;
      return;
    }
      // REFRACTIONS:
    if ( key.compareTo(REFRACTIONS) == 0 )
    {
      doRefractions = booleanProperty( value, doRefractions );
      scene         = null;
      return;
    }
      // MAX_RECURSION:
    if ( key.compareTo(MAX_RECURSION) == 0 )
    {
      maxLevel = intProperty( value, maxLevel );
      return;
    }
      // MIN_IMPORTANCE:
    if ( key.compareTo(MIN_IMPORTANCE) == 0 )
    {
      minImportance = doubleProperty( value, minImportance );
      return;
    }

    super.set( key, value );
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(REFLECTIONS) == 0 )
      return doReflections;
    if ( key.compareTo(REFRACTIONS) == 0 )
      return doRefractions;
    if ( key.compareTo(MAX_RECURSION) == 0 )
      return maxLevel;
    if ( key.compareTo(MIN_IMPORTANCE) == 0 )
      return minImportance;

    return super.get( key );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Ray-tracing";

  /** Object description. */
  private final static String DESCRIPTION = "Ray-tracing of CSG scene.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,        IFACE+"ImageFunction" );
      t.newInputPlug(  PL_PROPERTY,     IFACE+"Property" );
      t.newOutputPlug( PL_OUTPUT,       IFACE+"RTScene" );
      t.newOutputPlug( PL_INTERSECTION, IFACE+"Intersectable" );
        // Property:
      t.propBegin( SHADING, TYPE_BOOLEAN, "Use light models at all?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( SHADOWS, TYPE_BOOLEAN, "Use shadow rays (for casted shadows)?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( TEXTURES, TYPE_BOOLEAN, "Compute textures?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( REFLECTIONS, TYPE_BOOLEAN, "Compute reflected rays?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( REFRACTIONS, TYPE_BOOLEAN, "Compute refracted rays?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( MAX_RECURSION, TYPE_INTEGER, "Maximum recursion level", true );
        t.propDefault( 12 );
      t.propEnd();
      t.propBegin( MIN_IMPORTANCE, TYPE_DOUBLE, "Minimum ray importance (recursion limit)", true );
        t.propDefault( 0.05 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}