package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Depth-sort rendering of {@link Brep} database.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/RenderPainter.java">RenderPainter.java</a>
 */
public class RenderPainter extends DefaultRender3D
{
  //==========================================================================
  //  Data:

  /** Canvas background. */
  protected final static int[] BACKGROUND = {   0,  20,  60, 255 };

  /** How many faces to draw (debugging). Zero means "draw all". */
  protected int facesToDraw = 0;

  //==========================================================================
  //  Interface Render3D:

  /**
   * 3D rendering using the given projection matrix.
   *
   * @param mv Local model-view matrix.
   * @param proj Local projection matrix.
   * @param style Local rendering style.
   * @see #render()
   */
  @Override
  public void render ( TrMatrix mv, TrMatrix proj, int style )
  {
    final Brep           data   = (Brep)          getInterface( PL_DATA,     IFACE+"Brep" );
    final RasterGraphics output = (RasterGraphics)getInterface( PL_OUTPUT,   IFACE+"RasterGraphics" );
          BrepIterator   order  = (BrepIterator)  getInterface( PL_ORDER,    IFACE+"BrepIterator" );
    final FaceRender     render = (FaceRender)    getInterface( PL_FUNCTION, IFACE+"FaceRender" );

    if ( data   == null ||
         output == null ||
         render == null ) return;

    final int width  = output.getWidth();
    final int height = output.getHeight();
    output.setRectangle( 0, 0, width, height, BACKGROUND );

    final int ctx     = Brep.CTX_DEFAULT;
    final int worldId = data.getAttributeId( Brep.VERTEX, ctx, Brep.COORD );
    final int projId  = data.transformCoordinates( worldId, ctx, Brep.COORD_PROJ, proj );

    if ( order == null )
      order = data.faceIterator( ctx, null );
    else
    {
      order.set( BrepIterator.SORT_ORDER,  BrepIterator.SORT_ASCENDING );
      order.set( BrepIterator.CONTEXT,     ctx );
      order.set( BrepIterator.PROJ_COORDS, Brep.COORD_PROJ );
    }

      // to be sure (mandatory data for flat-shading only):
    render.set( FaceRender.PROJECTED_COORDS, Brep.COORD_PROJ );
    render.set( FaceRender.FACE_COLOR,       Brep.COLOR );

      // paint loop:
    int faceId;
    int ord = 0;

    order.reset();
    while ( (faceId = order.next()) != Brep.NULL &&
            (facesToDraw <= 0 || ord++ < facesToDraw) &&
            !userBreak )
      render.fillInside( ctx, faceId );
  }

  //==========================================================================
  //  Interface Property:

  public final static String FACES = "Faces";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // FACES:
    if ( key.compareTo(FACES) == 0 )
    {
      facesToDraw = intProperty( value, facesToDraw );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(FACES) == 0 )
      return facesToDraw;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Render depth-sort";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "Render3DToBrepAndBrepIteratorAndFaceRenderAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Depth-sort B-rep rendering (uses external ordering module).";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + '.' + C_BREP;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,    IFACE+"Render3D" );
      t.newOutputPlug(    PL_DATA,     IFACE+"Brep" );
      t.newOptOutputPlug( PL_ORDER,    IFACE+"BrepIterator" );
      t.newOutputPlug(    PL_FUNCTION, IFACE+"FaceRender" );
      t.newOutputPlug(    PL_OUTPUT,   IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( FACES, TYPE_INTEGER, "How many faces to draw (0 .. all)?", true );
        t.propDefault( 0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
