package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * SLE solver - various relaxation methods.
 * Gauss-Seidel and Southwell iterations (with optional over-relaxation) are implemented so far.
 *
 * @author  Tomas Svab, PE
 * @version 0.26 $Rev: 753 $ $Date: 2008-05-18 17:25:57 +0200 (ne, 18 V 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/SLERelaxation.java">SLERelaxation.java</a>
 */
public class SLERelaxation extends Piece implements SLESolver
{
  //==========================================================================
  //  Data:

  /** Accelerator: input SLEData. */
  protected SLEData data;

  /** Supported methods. */
  public static enum Methods
  {
    /** Jacobi iteration. */
    JACOBI,
    /** Gauss-Seidel iteration. */
    GAUSS_SEIDEL,
    /** Southwell iteration. */
    SOUTHWELL,
  }

  /** Actual sub-method. */
  protected Methods usedMethod = Methods.GAUSS_SEIDEL;

  /** Required solution accuracy (max. norm of residual vector). */
  protected double epsilon = 1.e-6;

  /** Over-relaxation factor. {@code 1.0} for plain relaxation. */
  protected double omega = 1.0;

  /** Index of the next unknown to be relaxed (for ordinal methods). */
  protected int i = 0;

  /** Number of performed iteration steps. */
  protected int steps = 0;

  /** Number of unknowns. */
  protected int N = 0;

  /** Actual vector of unknown variables. */
  protected double B[];

  /** Actual matrix column. */
  protected double K[];

  /** Residual vector. */
  protected double r[];

  /** Second residual vector for Jacobi method (used for new solution generation). */
  protected double r2[];

  /** Was the solver already initialized? */
  protected boolean isInitialized = false;

  //==========================================================================
  //  Interface SLESolver:

  /** Sets solver accuracy (whatever it means). */
  @Override
  public double setEpsilon ( double epsilon )
  {
    double old   = this.epsilon;
    this.epsilon = epsilon;
    return old;
  }

  /**
   * Solve the whole system en bloc.
   *
   * @param  save Solution will be saved into the associated {@link SLEData}.
   */
  @Override
  public void solve ( boolean save )
  {
    solve( Integer.MAX_VALUE, save );
  }

  /**
   * Reset the solver (if applicable).
   */
  @Override
  public void reset ()
  {
    if ( data == null ) data = (SLEData)getInterface( PL_OUTPUT, IFACE+"SLEData" );
    if ( data == null )
      return;

    N = data.getColumns();
    B = new double[ N ];                  // I'm starting from the zero-vector (FIXME)
    r = data.getRight( null );            // .. residual vector is trivial in this case
    if ( usedMethod == Methods.JACOBI )   // allocate 2nd residual vector (for new generation solution)
      r2 = Formula.cloneArray( r );
    else
      r2 = null;
    K = new double[ N ];
    i = 0;
    steps = 0;

    isInitialized = true;
  }

  /**
   * Refine the solution performing several iterations.
   *
   * @param  maxIterations Number of iterations requested.
   * @param  save Solution will be saved into the associated {@link SLEData}.
   * @return Achieved precision (if less than epsilon, the system has converged).
   */
  @Override
  public double refine ( int maxIterations, boolean save )
  {
    return solve( maxIterations, save );
  }

  /**
   * Input data was changed, solver should adapt to it.
   * Usually {@link #solve} or {@link #refine} will be called afterwards..
   */
  @Override
  public void update ()
  {
    reset();
  }

  /** Reads a single unknown (element of solution vector). Non-mandatory. */
  @Override
  public double getSolution ( int col )
  {
    if ( col <  0 ||
         col >= N ||
         B == null ) return 0.0;

    return B[col];
  }

  /** Reads the whole solution vector. Non-mandatory. */
  @Override
  public double[] getSolution ( double[] result )
  {
    if ( result == null ||
         result.length < N )
      result = new double[ N ];

    System.arraycopy( B, 0, result, 0, N );
    return result;
  }

  //==========================================================================
  //  Multi-component systems:

  /** Get actual component number. */
  @Override
  public int actualComponent ()
  {
    return 0;
  }

  /** Set actual component (non-mandatory). */
  @Override
  public void setComponent ( int c )
  {
  }

  //==========================================================================
  //  The solver itself:

  /**
   * Performs the given number of iteration steps.
   *
   * @param  maxIterations Maximal number of iterations.
   * @param  save Solution will be saved into the associated {@link SLEData}.
   * @return Residual norm achieved in the iteration.
   */
  protected double solve ( int maxIterations, boolean save )
  {
    if ( data == null ||
         !isInitialized )
    {
      reset();
      if ( data == null )
        return Double.POSITIVE_INFINITY;
    }

      // solving the SLE:
    int    iter = maxIterations;          // number of iterations in the batch
    double maxResidual;                   // maximal residual value ("max" norm)
    double dB;
    int    j;

    do
    {
      if ( usedMethod == Methods.SOUTHWELL )
      {
        maxResidual = 0.0;
          // pick i with the largest |r[i]| = argmax(|r[i]|):
        for ( j = 0; j < N; j++ )
          if ( maxResidual < Math.abs( r[j] ) )
          {
            maxResidual = Math.abs( r[j] );
            i = j;
          }
      }

        // read one (working) column:
      K = data.getColumn( i, K );

        // solution update:
      if ( K[i] > 0.0 )
        dB = omega * r[i] / K[i];
      else
        dB = 0.0;
      B[i] += dB;

        // residual update:
      maxResidual = 0.0;
      switch ( usedMethod )
      {
        case JACOBI:
            // update the 2nd residual (associated with the new generation solution)
          for ( j = 0; j < N; j++ )
          {
            r2[j] -= dB * K[j];

            if ( maxResidual < Math.abs( r2[j] ) )
              maxResidual = Math.abs( r2[j] );
          }
          break;

        case GAUSS_SEIDEL:
        case SOUTHWELL:
          for ( j = 0; j < N; j++ )
          {
            r[j] -= dB * K[j];

            if ( maxResidual < Math.abs( r[j] ) )
              maxResidual = Math.abs( r[j] );
          }
          break;
      }

      if ( ++i >= N )
      {
        i = 0;
        if ( usedMethod == Methods.JACOBI )
            // the new generation was completed => backup the old residual
          System.arraycopy( r2, 0, r, 0, N );
      }
      steps++;
    }
    while ( maxResidual >= epsilon &&     // Is converged?
            --iter > 0 );

      // saving results..
    if ( save )
      data.setSolution( B );

    return maxResidual;
  }

  //==========================================================================
  //  Support for get:

  protected double getTotalResidual ()
  {
    double total = 0.0;
    for ( int j = 0; j < N; j++ )
      total += Math.abs( r[j] );
    return total;
  }

  //==========================================================================
  //  Interface Property:

  public final static String USED_METHOD = "Method";
  public final static String EPSILON     = "Epsilon";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // USED_METHOD:
    if ( key.compareTo(USED_METHOD) == 0 )
    {
      usedMethod = enumProperty( value, usedMethod );
      return;
    }
      // EPSILON:
    if ( key.compareTo(EPSILON) == 0 )
    {
      epsilon = doubleProperty( value, epsilon );
      return;
    }
      // OMEGA:
    if ( key.compareTo(OMEGA) == 0 )
    {
      omega = doubleProperty( value, omega );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(USED_METHOD) == 0 )
      return usedMethod.ordinal();
    if ( key.compareTo(EPSILON) == 0 )
      return epsilon;
    if ( key.compareTo(OMEGA) == 0 )
      return omega;
    if ( key.compareTo(STEPS) == 0 )
      return steps;
    if ( key.compareTo(TOTAL_RESIDUAL) == 0 )
      return getTotalResidual();

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "SLE-relaxation";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "SLESolverToSLEData";

  /** Object description. */
  private final static String DESCRIPTION = "SLE solver - various relaxation methods (Jacobi, Gauss-Seidel, Southwell)";

  /** Object category. */
  protected final static String CATEGORY = "SLE.solver";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"SLESolver" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"SLEData" );
        // Property:
      t.propBegin( USED_METHOD, TYPE_INTEGER, "Sub-method", true );
        t.propDefault( Methods.GAUSS_SEIDEL.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Jacobi",       Methods.JACOBI.ordinal(),       "Jacobi iteration with zero initial value" );
        t.propEnum( "Gauss-Seidel", Methods.GAUSS_SEIDEL.ordinal(), "Gauss-Seidel iteration with zero initial value" );
        t.propEnum( "Southwell",    Methods.SOUTHWELL.ordinal(),    "Southwell iteration with zero initial value" );
      t.propEnd();
      t.propBegin( EPSILON, TYPE_DOUBLE, "Required solution accuracy (norm of residual vector)", true );
        t.propDefault( 1.e-6 );
      t.propEnd();
      t.propBegin( OMEGA, TYPE_DOUBLE, "Over-relaxation factor, 1.0 for plain relaxation", true );
        t.propDefault( 1.0 );
        t.propBounds( 0.0, 2.0 );
      t.propEnd();
      t.propBegin( STEPS, TYPE_INTEGER, "Number of iteration steps performed so far.", false );
      t.propEnd();
      t.propBegin( TOTAL_RESIDUAL, TYPE_DOUBLE, "Sum of all residual entries = solution accuracy", false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
