package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Noise function - simple implementations.
 * Interpolation/approximation in uniform grid.
 *
 * @version 0.26 $Rev: 431 $ $Date: 2007-04-29 22:42:22 +0200 (ne, 29 IV 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/SimpleNoise.java">SimpleNoise.java</a>
 */
public class SimpleNoise extends DefaultRealFunction
{
  //==========================================================================
  //  Data:

  /** Implemented noise types. */
  protected static enum NoiseType
  {
    /** Approximating noise using cubic B-spline bases. */
    NOISE_BSPLINE,
  }

  /** Actual noise type. */
  protected NoiseType type = NoiseType.NOISE_BSPLINE;

  /** Noise dimensionality. Not used yet (not needed). */
  protected int dim = 2;

  /** Actual noise amplitude. */
  protected double amplitude = 1.0;

  /** Actual noise frequency (number of periods within 0.0 - 1.0 domain). */
  protected int frequency = 8;

  /** Hash-function variation. Should be used in continuous segment: 0, 1, 2, ... */
  protected int variation = 2;

  /** Hash-function instance. */
  protected HashFunction hash;
  {
    hash = new HashFunction();
    hash.setVariation( 2 );
  }

  /** Array holding randomly-generated knot points. */
  protected double[] grid;

  protected final void assertGrid ( int len )
  {
    if ( grid == null ||
         grid.length < len )
      grid = new double[ len ];
  }

  /** X-coordinate increment for better hashing distribution. */
  protected final static int X0 = 10103;

  /** Y-coordinate increment for better hashing distribution. */
  protected final static int Y0 = 12011;

  /** Z-coordinate increment for better hashing distribution. */
  protected final static int Z0 =  9103;

  /** X-coordinate multiplier for better hashing distribution. */
  protected final static int XM = 100153;

  /** Y-coordinate multiplier for better hashing distribution. */
  protected final static int YM = 250013;

  /** Z-coordinate multiplier for better hashing distribution. */
  protected final static int ZM =  50111;

  /** How many LS-bits from pseudo-random generator will be used. */
  protected final static long RND_MASK = 0x3ffffff;

  //==========================================================================
  //  Support:

  /**
   * Cubic B-spline basis function blending.
   * Combines four values from the grid[] array,
   * indices are: i0, i0+di, i0+2*di, i0+3*di.
   */
  protected final double bSpline ( int i0, int di, double t )
  {
    double omt = 1.0 - t;
    double t2  = t * t;
    final double ONE_SIXTH = 1.0 / 6.0;

    double result = grid[i0] * omt * omt * omt;
    i0 += di;
    result += grid[i0] * ( 3.0 * t2 * (t - 2.0) + 4.0 );
    i0 += di;
    result += grid[i0] * ( 3.0 * ( t2 * omt + t ) + 1.0 );
    i0 += di;
    result += grid[i0] * t * t2;

    return( ONE_SIXTH * result );
  }

  //==========================================================================
  //  Interface RealFunction:

  /**
   * Compute the mapping: <b>f(x,y)</b>.
   *
   * @param  x X coordinate (1st argument).
   * @param  y Y coordinate (2nd argument).
   */
  @Override
  public double f ( double x, double y )
  {
    int i, j, k;
    long a;

    x     *= frequency;
    y     *= frequency;
    int xi = (int)Math.floor( x );
    int yi = (int)Math.floor( y );
    x     -= xi;
    y     -= yi;
    xi    += X0;
    yi    += Y0;
    double mul = 2.0 * amplitude / (RND_MASK + 1.0);

    switch ( type )
    {
      case NOISE_BSPLINE:
      {
        assertGrid( 16 );
        for ( k = 0, j = -1; j < 3; j++ )
          for ( i = -1; i < 3; i++ )
          {
            a = hash.hash( (xi + i) * XM, (yi + j) * YM );
            a = RandomStatic.numericRecipes( a );
            grid[ k++ ] = mul * ( a & RND_MASK ) - amplitude;
          }

        for ( i = 0; i < 4; i++ )
          grid[i] = bSpline( i, 4, y );

        return bSpline( 0, 1, x );
      }

      default:
        return 0.0;
    }
  }

  /**
   * Compute the mapping: <b>f(x,y,z)</b>.
   *
   * @param  x X coordinate (1st argument).
   * @param  y Y coordinate (2nd argument).
   * @param  z Z coordinate (3rd argument).
   */
  @Override
  public double f ( double x, double y, double z )
  {
    int i, j, k, l;
    long a;

    x     *= frequency;
    y     *= frequency;
    z     *= frequency;
    int xi = (int)Math.floor( x );
    int yi = (int)Math.floor( y );
    int zi = (int)Math.floor( z );
    x     -= xi;
    y     -= yi;
    z     -= zi;
    xi    += X0;
    yi    += Y0;
    zi    += Z0;
    double mul = 2.0 * amplitude / (RND_MASK + 1.0);

    switch ( type )
    {
      case NOISE_BSPLINE:
      {
        assertGrid( 64 );
        for ( l = 0, k = -1; k < 3; k++ )
          for ( j = -1; j < 3; j++ )
            for ( i = -1; i < 3; i++ )
            {
              a = hash.hash( (xi + i) * XM, (yi + j) * YM, (zi + k) * ZM );
              a = RandomStatic.numericRecipes( a );
              grid[ l++ ] = mul * ( a & RND_MASK ) - amplitude;
            }

        for ( i = 0; i < 16; i++ )
          grid[i] = bSpline( i, 16, z );

        for ( i = 0; i < 4; i++ )
          grid[i] = bSpline( i, 4, y );

        return bSpline( 0, 1, x );
      }

      default:
        return 0.0;
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String NOISE_TYPE = "Noise type";
  public final static String DIMENSION  = "Dimension";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // NOISE_TYPE:
    if ( key.compareTo(NOISE_TYPE) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
      // AMPLITUDE:
    if ( key.compareTo(AMPLITUDE) == 0 )
    {
      amplitude = doubleProperty( value, amplitude, 1.e-4, 1.e4 );
      return;
    }
      // FREQUENCY:
    if ( key.compareTo(FREQUENCY) == 0 )
    {
      frequency = intProperty( value, frequency, 1, 1000 );
      return;
    }
      // VARIATION:
    if ( key.compareTo(VARIATION) == 0 )
    {
      variation = intProperty( value, variation );
      hash.setVariation( variation );
      return;
    }
      // DIMENSION:
    if ( key.compareTo(DIMENSION) == 0 )
    {
      dim = intProperty( value, dim, 1, 4 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(NOISE_TYPE) == 0 )
      return type.ordinal();
    if ( key.compareTo(AMPLITUDE) == 0 )
      return amplitude;
    if ( key.compareTo(FREQUENCY) == 0 )
      return frequency;
    if ( key.compareTo(VARIATION) == 0 )
      return variation;
    if ( key.compareTo(DIMENSION) == 0 )
      return dim;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "SimpleNoiseFunctions";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "RealFunction";

  /** Object description. */
  private final static String DESCRIPTION = "Simple implementations of 1D - 4D noise.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + "." + C_FUNCTION;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"RealFunction" );
        // Property:
      t.propBegin( NOISE_TYPE, TYPE_INTEGER, "Noise type", true );
        t.propDefault( NoiseType.NOISE_BSPLINE.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "B-spline", NoiseType.NOISE_BSPLINE.ordinal(), "Cubic approximation B-spline" );
      t.propEnd();
      t.propBegin( AMPLITUDE, TYPE_DOUBLE, "Amplitude", true );
        t.propDefault( 1.0 );
        t.propBounds(  1.e-4, 1.e4 );
      t.propEnd();
      t.propBegin( FREQUENCY, TYPE_INTEGER, "Frequency", true );
        t.propDefault( 8 );
        t.propBounds(  1, 1000 );
      t.propEnd();
      t.propBegin( VARIATION, TYPE_INTEGER, "Variation", true );
        t.propDefault( 2 );
      t.propEnd();
      t.propBegin( DIMENSION, TYPE_INTEGER, "Dimension", true );
        t.propDefault( 2 );
        t.propBounds(  1, 4 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
