package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Data object holding system of linear equations (SLE) data - actual data storage
 * w/o any smartness.
 *
 * @author  PE
 * @version 0.26 $Rev: 757 $ $Date: 2008-05-19 01:09:45 +0200 (po, 19 V 2008) $ $Author: pepca $
 * @since   0.26
 */
public class SimpleSLEDataStore extends Piece implements SLEDataStore
{
  //==========================================================================
  //  Data:

  /** Number of equations (matrix rows). */
  protected int rows;

  /** Number of unknown variables (matrix columns). */
  protected int columns;

  /** Complete rectangular matrix {@code A} as in {@code Ax=b}. */
  protected double[][] matrix;

  /** Right-hand side {@code b} as in {@code Ax=b}. */
  protected double[] right;

  /** Solution vector {@code x} as in {@code Ax=b}. */
  protected double[] solution;

  /** Solution vector {@code x} as in {@code Ax=b}. */
  protected double[] weight;

  /** Solution vector {@code x} as in {@code Ax=b}. */
  protected double[] reflectance;

  //==========================================================================
  //  Input SLE data:

  /**
   * Number of equations = number of matrix rows
   * (usually number of unknowns is equal to number of equations).
   */
  public int getRows ()
  {
    return rows;
  }

  /**
   * Number of unknowns = number of matrix columns
   * (usually number of unknowns is equal to number of equations).
   */
  public int getColumns ()
  {
    return columns;
  }

  /** Reads a single matrix coefficient. */
  public double getCoeff ( int row, int col )
  {
    if ( matrix == null ||
         row < 0 || row >= rows ||
         col < 0 || col >= columns ) return 0.0;

    return matrix[row][col];
  }

  /** Reads the whole matrix row. */
  public double[] getRow ( int row, double[] result )
  {
    if ( matrix == null ||
         row < 0 || row >= rows ) return null;

    if ( result == null ||
         result.length < columns )
      result = new double[ columns ];

    System.arraycopy( matrix[row], 0, result, 0, columns );

    return result;
  }

  /** Reads the whole matrix column. */
  public double[] getColumn ( int col, double[] result )
  {
    if ( matrix == null ||
         col < 0 || col >= columns ) return null;

    if ( result == null ||
         result.length < rows )
      result = new double[ rows ];

    for ( int i = 0; i < rows; i++ )
      result[i] = matrix[i][col];

    return result;
  }

  /** Reads the whole matrix. */
  public double[][] getMatrix ()
  {
    if ( matrix == null ) return null;

    final double[][] result = new double[ rows ][ columns ];

    for ( int row = 0; row < rows; row++ )
      System.arraycopy( matrix[row], 0, result[row], 0, columns );

    return result;
  }

  /** Reads a single element of the right-side vector. */
  public double getRight ( int row )
  {
    if ( right == null ||
         row < 0 || row >= rows ) return 0.0;

    return right[row];
  }

  /**
   * Reads the whole right-side vector.
   *
   * @param  result Pre-allocated array to hold result (can be {@code null}).
   */
  public double[] getRight ( double[] result )
  {
    if ( right == null ) return null;

    if ( result == null ||
         result.length < rows )
      result = new double[ rows ];

    System.arraycopy( right, 0, result, 0, rows );

    return result;
  }

  //==========================================================================
  //  Solution vector:

  /** Reads a single unknown (element of solution vector). */
  public double getSolution ( int col )
  {
    if ( solution == null ||
         col < 0 || col >= columns ) return 0.0;

    return solution[col];
  }

  /** Reads the whole solution vector. */
  public double[] getSolution ( double[] result )
  {
    if ( solution == null ) return null;

    if ( result == null ||
         result.length < columns )
      result = new double[ columns ];

    System.arraycopy( solution, 0, result, 0, columns );

    return result;
  }

  /** Writes value of a single unknown. */
  public void setSolution ( int col, double value )
  {
    if ( solution == null ||
         col < 0 || col >= columns ) return;

    solution[col] = value;
  }

  /** Writes the whole solution vector. */
  public void setSolution ( double[] values )
  {
    if ( values == null ) return;

    System.arraycopy( values, 0, solution, 0, Math.min( values.length, columns ) );
  }

  //==========================================================================
  //  Storing data:

  /**
   * Initializes the SLE.
   * All values will be set to {@code 0.0}.
   */
  public void init ( int rows, int columns )
  {
    this.rows    = rows;
    this.columns = columns;
    matrix       = new double[ rows ][ columns ];
    right        = new double[ rows ];
    solution     = new double[ columns ];
  }

  /**
   * Sets a single matrix coefficient.
   * Preferred metod in sparse-matrix implementations..
   */
  public void setCoeff ( int row, int col, double value )
  {
    if ( matrix == null ||
         row < 0 || row >= rows ||
         col < 0 || col >= columns ) return;

    matrix[row][col] = value;
  }

  /** Sets the whole matrix row. */
  public void setRow ( int row, double[] values )
  {
    if ( matrix == null ||
         values == null ||
         row < 0 || row >= rows ) return;

    System.arraycopy( values, 0, matrix[row], 0, Math.min( values.length, columns ) );
  }

  /** Sets the whole matrix column. */
  public void setColumn ( int col, double[] values )
  {
    if ( matrix == null ||
         values == null ||
         col < 0 || col >= columns ) return;

    for ( int i = 0; i < rows; i++ )
      matrix[i][col] = (i < values.length) ? values[i] : 0.0;
  }

  /** Sets a single element of the right-side vector. */
  public void setRight ( int row, double value )
  {
    if ( right == null ||
         row < 0 || row >= rows ) return;

    right[row] = value;
  }

  /** Sets the whole right-side vector. */
  public void setRight ( double[] values )
  {
    if ( values == null ) return;

    System.arraycopy( values, 0, right, 0, Math.min( values.length, rows ) );
  }

  //==========================================================================
  //  Incremental SLE changes (non-mandatory):

  /** Returns index of next right-side element which was changed since last {@link #resetChanges} call. */
  public int nextRightChanged ( int row )
  {
    return NULL;
  }

  /**
   * Returns column-index of next coefficient which was changed since last {@link #resetChanges} call.
   * Passes through the given matrix row..
   */
  public int nextCoeffInRowChanged ( int row, int col )
  {
    return NULL;
  }

  /**
   * Returns row-index of next coefficient which was changed since last {@link #resetChanges} call.
   * Passes through the given matrix column..
   */
  public int nextCoeffInColumnChanged ( int row, int col )
  {
    return NULL;
  }

  /** Resets (acknowledges) all SLE changes. */
  public void resetChanges ()
  {
  }

  //==========================================================================
  //  Additional information:

  /** Reads a single weight (radiosity: element area). */
  public double getWeight ( int i )
  {
    // TODO
    return 1.0;
  }

  /** Reads the whole weight vector. */
  public double[] getWeight ( double[] values )
  {
    if ( null == values )
      values = new double[ rows ];

    // TODO
    Arrays.fill( values, 1.0 );
    return values;
  }

  /** Reads a single reflectance (radiosity: element albedo). */
  public double getReflectance ( int i )
  {
    // TODO
    return 0.0;
  }

  /** Reads the whole reflectance vector. */
  public double[] getReflectance ( double[] reflectance )
  {
    // TODO
    return null;
  }

  /** Sets a single weight (radiosity: element area). */
  public void setWeight ( int i, double value )
  {
  }

  /** Sets the whole weight vector en bloc. */
  public void setWeight ( double[] values )
  {
  }

  /** Sets a single reflectance (radiosity: element albedo). */
  public void setReflectance ( int i, double value )
  {
  }

  /** Sets the whole reflectance vector. */
  public void setReflectance ( double[] reflectance )
  {
  }

  /** Get actual component number. */
  public int actualComponent ()
  {
    return 0;
  }

  /** Set actual component (non-mandatory). */
  public void setComponent ( int c )
  {
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "SimpleSLEData";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "SLEDataStore";

  /** Object description. */
  private final static String DESCRIPTION = "SLE storage object (full matrix is stored).";

  /** Object category. */
  protected final static String CATEGORY = C_DATA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"SLEDataStore" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
