package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Simple image synthesizer with progressive rendering.
 *
 * @version 0.26 $Rev: 472 $ $Date: 2007-05-31 09:33:37 +0200 (čt, 31 V 2007) $ $Author: pepca $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/SimpleSynthesizer.java">SimpleSynthesizer.java</a>
 */
public class SimpleSynthesizer extends DefaultImageSynthesizer
{
  //==========================================================================
  //  Data:

  /** Initial image cell size in pixels. */
  protected final static int INIT_CELL = 8;

  /** Use gray palette (for {@link RealFunction}). */
  protected boolean useGray = false;

  /** Use time variable (for {@link RealFunction}). */
  protected boolean useTime = false;

  /** Horizontal scaling coefficient (for {@link RealFunction}). */
  protected double mulX = 1.0;

  /** Horizontal translation (for {@link RealFunction}). */
  protected double addX = 0.0;

  /** Vertical scaling coefficient (for {@link RealFunction}). */
  protected double mulY = 1.0;

  /** Vertical translation (for {@link RealFunction}). */
  protected double addY = 0.0;

  /**
   * Gamma-correction for frame-buffer efficiency.
   * It can be {@code 0.0} to turn off gamma-correction and color clamping.
   */
  protected double gamma = 2.2;

  //==========================================================================
  //  Support:

  /** Transfers image size from associated {@link RasterGraphics} if necessary. */
  protected void imageBounds ()
  {
    if ( imageWidth > 0 ) return;

      // set image size according to the connected image module
    final RasterGraphics r = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    if ( r == null )
    {
      setBounds( 512, 512 );
      return;
    }

    setBounds( r.getWidth(), r.getHeight() );
  }

  //==========================================================================
  //  Interface TimeDependent:

  /**
   * Retrieve the valid bounds for time variable.
   *
   * @param  pre Pre-allocated {@code double[2]} array to hold the result (can be
   *             {@code null}).
   * @return {@code result[0]} .. minimum time value, {@code result[1]} .. maximum
   *         time value.
   * @see    #getTime
   * @see    #setTime
   */
  @Override
  public double[] getTimeInterval ( double[] pre )
  {
    ImageFunction f = (ImageFunction)getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    if ( f != null )
      return f.getTimeInterval( pre );

      // RealFunction:
    if ( !useTime ) return null;

    RealFunction rf = (RealFunction)getInterface( PL_DIRECT, IFACE+"RealFunction" );
    if ( rf == null ) return null;

    double[] dom = new double[ 6 ];
    dom = rf.getDomain( dom );
    if ( dom == null ) return null;

    if ( pre == null || pre.length < 2 )
      pre = new double[ 2 ];

    pre[0] = dom[4];
    pre[1] = dom[5];
    return pre;
  }

  /**
   * Returns the actual time value for the object (used in animations).
   *
   * @return Actual time (in seconds).
   * @see    #setTime
   */
  @Override
  public double getTime ()
  {
    ImageFunction f = (ImageFunction)getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    return( (f == null) ? actTime :  f.getTime() );
  }

  /**
   * Sets the actual time value for the object (used in animations).
   *
   * @param  time New scene time (in seconds).
   * @see    #getTime
   * @see    #getTimeInterval
   */
  @Override
  public void setTime ( double time )
  {
    actTime = time;
    ImageFunction f = (ImageFunction)getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    if ( f != null )
      f.setTime( time );
  }

  //==========================================================================
  //  Interface ImageSynthesizer:

  /**
   * Sets resolution of an output raster image.
   *
   * @param  width Horizontal image size in pixels.
   * @param  height Vertical image size in pixels.
   */
  @Override
  public void setBounds ( int width, int height )
  {
    imageWidth  = width;
    imageHeight = height;
    ImageFunction f = (ImageFunction)getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    if ( f != null )
    {
      f.setBounds( 0.0, width, 0.0, height );
      return;
    }

      // RealFunction:
    RealFunction rf = (RealFunction)getInterface( PL_DIRECT, IFACE+"RealFunction" );
    if ( rf == null ) return;

    double[] dom = new double[ 6 ];
    dom = rf.getDomain( dom );
    if ( dom == null )                    // assuming 0.0-1.0 in both directions
    {
      mulX = 1.0 / width;
      mulY = 1.0 / height;
      addX =
      addY = 0.0;
      return;
    }

      // set local scaling transformation:
    mulX = (dom[1] - dom[0]) / width;
    addX =  dom[0];
    mulY = (dom[3] - dom[2]) / height;
    addX =  dom[2];
  }

  /**
   * Render the given pixel of the image.
   *
   * @param  x X coordinate of the pixel.
   * @param  y Y coordinate of the pixel.
   * @param  pre Pre-allocated {@code double[3]} array to hold results (can be
   *             {@code null}).
   * @return Color of the pixel (in <b>RGB</b> color space).
   * @see    #renderRectangle
   * @see    #renderAll
   */
  @Override
  public double[] renderPixel ( int x, int y, double[] pre )
  {
    if ( pre == null || pre.length < 3 )
      pre = new double[ 3 ];

    imageBounds();

    ImageFunction f = (ImageFunction)getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    if ( f != null )
    {
      f.getSample( x + 0.5, y + 0.5, pre );
      if ( gamma > 0.01 )                 // gamma-correction and clamping
      {
        double g = 1 / gamma;
        pre[0] = Formula.clamp( Math.pow( pre[0], g ), 0.0, 1.0 );
        pre[1] = Formula.clamp( Math.pow( pre[1], g ), 0.0, 1.0 );
        pre[2] = Formula.clamp( Math.pow( pre[2], g ), 0.0, 1.0 );
      }
      return pre;
    }

      // RealFunction:
    RealFunction rf = (RealFunction)getInterface( PL_DIRECT, IFACE+"RealFunction" );
    if ( rf == null ) return null;

    double xd = mulX * (x + 0.5) + addX;
    double yd = mulY * (y + 0.5) + addY;
    double result;

    if ( useTime )
      result = rf.f( xd, yd, actTime );
    else
      result = rf.f( xd, yd );

    result = Formula.clamp( 0.5 * (result + 1.0), 0.0, 1.0 );

    if ( useGray )
    {
      pre[0] = pre[1] = pre[2] = result;
      if ( pre.length > 3 )
        pre[3] = 1.0;
    }
    else
      Formula.colorRamp( result, pre );

    return pre;
  }

  /**
   * Render the given rectangle.
   * Output is written to some connected raster device (not defined in this interface).
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @see    #renderPixel
   * @see    #renderAll
   */
  public void renderRectangle ( int x1, int y1, int x2, int y2 )
  {
    final ImageFunction  f    = (ImageFunction) getInterface( PL_FUNCTION, IFACE+"ImageFunction" );
    final RealFunction   rf   = (RealFunction)  getInterface( PL_DIRECT,   IFACE+"RealFunction" );
    final RasterGraphics r    = (RasterGraphics)getInterface( PL_OUTPUT,   IFACE+"RasterGraphics" );
    final Trigger        trig = (Trigger)       getInterface( PL_TRIGGER,  IFACE+"Trigger" );

    if ( f == null && rf == null ||
         r == null )
      return;

    imageBounds();

    if ( x1 < 0 ) x1 = 0;
    if ( x2 > imageWidth ) x2 = imageWidth;
    if ( y1 < 0 ) y1 = 0;
    if ( y2 > imageHeight ) y2 = imageHeight;
    if ( x1 >= x2 || y1 >= y2 ) return;   // the rectangle is empty

      // run several phases of image rendering:
    int cell = INIT_CELL;
    int x, y;
    boolean xParity, yParity;
    final double[] c = new double[ 3 ];   // pixel color
    double xd, yd, result;
    final double g = (gamma > 0.01) ? 1.0 / gamma : 0.0;
    final long total = (x2 - x1) * (long)(y2 - y1);
    long counter = 0L;

    do                                    // do one phase
    {
      for ( y = y1, yParity = false;
            y < y2;                       // one image row
            y += cell, yParity = !yParity )

        for ( x = x1, xParity = false;
              x < x2;                     // one image cell
              x += cell, xParity = !xParity )

          if ( cell == INIT_CELL ||
               xParity || yParity )       // process the cell
          {
            if ( userBreak ) return;

              // determine sample color ..
            if ( f != null )
            {
              f.getSample( x + 0.5, y + 0.5, c );
              if ( g > 0.0 )              // gamma-correction and clamping
              {
                c[0] = Formula.clamp( Math.pow( c[0], g ), 0.0, 1.0 );
                c[1] = Formula.clamp( Math.pow( c[1], g ), 0.0, 1.0 );
                c[2] = Formula.clamp( Math.pow( c[2], g ), 0.0, 1.0 );
              }
            }

            else
            {
              xd = mulX * (x + 0.5) + addX;
              yd = mulY * (y + 0.5) + addY;

              if ( useTime )
                result = rf.f( xd, yd, actTime );
              else
                result = rf.f( xd, yd );

              result = Formula.clamp ( 0.5 * (result + 1.0), 0.0, 1.0 );

              if ( useGray )
                c[0] = c[1] = c[2] = result;
              else
                Formula.colorRamp( result, c );
            }

              // .. and render it:
            int xMax = x + cell;
            if ( xMax > x2 ) xMax = x2;
            int yMax = y + cell;
            if ( yMax > y2 ) yMax = y2;
            r.setRectangle( x, y, xMax, yMax, c );

            if ( (++counter & 1023) == 0 &&
                 pl != null )
              pl.progress( counter, total );
          }

        // repaint the picture:
      if ( trig != null )
        trig.fire( cell );
      if ( pl != null )
      {
        pl.progress( counter, total );
        pl.repaint();
      }

    }
    while ( (cell >>= 1) > 0 );         // do one phase
  }

  //==========================================================================
  //  Interface Property:

  public final static String USE_GRAY = "Gray palette";
  public final static String USE_TIME = "Use time variable";

  protected int setWidth  = 0;
  protected int setHeight = 0;

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // WIDTH:
    if ( key.compareTo(WIDTH) == 0 )
    {
      setWidth = intProperty( value, setWidth );
      return;
    }
      // HEIGHT:
    if ( key.compareTo(HEIGHT) == 0 )
    {
      setHeight = intProperty( value, setHeight );
      return;
    }
      // GAMMA:
    if ( key.compareTo(GAMMA) == 0 )
    {
      gamma = doubleProperty( value, gamma, 0.0, 5.0 );
      return;
    }
      // USE_GRAY:
    if ( key.compareTo(USE_GRAY) == 0 )
    {
      useGray = booleanProperty( value, useGray );
      return;
    }
      // USE_TIME:
    if ( key.compareTo(USE_TIME) == 0 )
    {
      useTime = booleanProperty( value, useTime );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(WIDTH) == 0 )
      return setWidth;
    if ( key.compareTo(HEIGHT) == 0 )
      return setHeight;
    if ( key.compareTo(GAMMA) == 0 )
      return gamma;
    if ( key.compareTo(USE_GRAY) == 0 )
      return useGray;
    if ( key.compareTo(USE_TIME) == 0 )
      return useTime;

    return null;
  }

  /**
   * Commits all previous changes to the internal object state.
   * Non-mandatory.
   */
  public void commit ()
  {
    setBounds( setWidth, setHeight );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Simple image synthesizer";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ImageSynthesizerToImageFunctionAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Image synthesizer with progressive rendering.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,    IFACE+"ImageSynthesizer" );
      t.newOptOutputPlug( PL_FUNCTION, IFACE+"ImageFunction" );
      t.newOptOutputPlug( PL_DIRECT,   IFACE+"RealFunction" );
      t.newOutputPlug(    PL_OUTPUT,   IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_TRIGGER,  IFACE+"Trigger" );
        // Property:
      t.propBegin( WIDTH, TYPE_INTEGER, "Image width in pixels", true );
        t.propDefault( 320 );
      t.propEnd();
      t.propBegin( HEIGHT, TYPE_INTEGER, "Image height in pixels", true );
        t.propDefault( 240 );
      t.propEnd();
      t.propBegin( GAMMA, TYPE_DOUBLE, "Gamma-correction exponent", true );
        t.propDefault( 2.2 );
        t.propBounds(  0.0, 5.0 );
      t.propEnd();
      t.propBegin( USE_GRAY, TYPE_BOOLEAN, "Use gray palette in case of scalar input?", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( USE_TIME, TYPE_BOOLEAN, "Use time variable (for RealFunction)?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
