package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Solid-color brush.
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.06
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/SolidColorBrush.java">SolidColorBrush.java</a>
 */
public class SolidColorBrush extends Piece implements SolidBrush
{
  //==========================================================================
  //  Interface Brush:

  /**
   * Area (re-)initialization.
   */
  public void init ()
  {
    final RasterGraphics r  = (RasterGraphics)getInterface( PL_OUTPUT,    IFACE+"RasterGraphics" );
    final BitMask        bm = (BitMask)       getInterface( PL_BITMASK,   IFACE+"BitMask" );
    final AlphaMask      am = (AlphaMask)     getInterface( PL_ALPHAMASK, IFACE+"AlphaMask" );

    if ( bm != null )
      bm.init( r.getWidth(), r.getHeight() );

    if ( am != null )
      am.init( r.getWidth(), r.getHeight() );
  }

  /**
   * Sets the binary set operation (used in {@link #fill} method).
   *
   * @param  _op Binary set operation (<tt>BinaryOperation.BIN_OP_*</tt> constant).
   * @return The old binary set operation.
   * @see    RasterGraphics
   * @see    cz.cuni.jagrlib.BinaryOperation Binary operations
   */
  public int setOperation ( int _op )
  {
    try
    {
      RasterGraphics r = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
      return r.setOperation( _op );
    }
    catch ( JaGrLibException e )
    {
      return BinaryOperation.BIN_OP_SET;
    }
  }

  /**
   * Sets the alpha-channel operation (used in {@link #fill} method).
   *
   * @param  _op Alpha-channel operation (<tt>AlphaOperation.ALPHA_*</tt> constant).
   * @return The old alpha-channel operation.
   * @see    cz.cuni.jagrlib.AlphaOperation Alpha-channel operations
   */
  public int setAlphaOperation ( int _op )
  {
    try
    {
      RasterGraphics r = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
      return r.setAlphaOperation( _op );
    }
    catch ( JaGrLibException e )
    {
      return AlphaOperation.ALPHA_B_OVER_A;
    }
  }

  /**
   * Sets the value-transfer function (used in {@link #fill} method).
   *
   * @param  _transfer New value-transfer function or <tt>null</tt>.
   * @return The old Value-transfer function or <tt>null</tt>.
   * @see    RasterGraphics
   * @see    cz.cuni.jagrlib.iface.ValueTransferFunction Value-transfor function prototype
   */
  public ValueTransferFunction setTransferFunction ( ValueTransferFunction _transfer )
  {
    try
    {
      RasterGraphics r = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
      return r.setTransferFunction( _transfer );
    }
    catch ( JaGrLibException e )
    {
      return null;
    }
  }

  /**
   * Fills the defined area. Destroys the current area.
   */
  public void fill ()
  {
    final RasterGraphics r  = (RasterGraphics)getInterface( PL_OUTPUT,  IFACE+"RasterGraphics" );
    final BitMask        bm = (BitMask)       getInterface( PL_BITMASK, IFACE+"BitMask" );

    if ( bm != null )
    {
      BitMaskEnumerator e        = bm.enumerator();
      BitMaskEnumerator.HLine hl = new BitMaskEnumerator.HLine();

      while ( e.nextHLine(hl) != null )
        r.setHLine( hl.x1, hl.x2, hl.y );

      bm.init();
    }

    AlphaMask am = (AlphaMask)getInterface( PL_ALPHAMASK, IFACE+"AlphaMask" );
    if ( am != null )
    {
      BitMaskEnumerator e        = am.enumerator();
      BitMaskEnumerator.HLine hl = new BitMaskEnumerator.HLine();

      while ( e.nextHLine(hl) != null )
      {
        r.setAlpha( hl.alpha );
        r.setHLine( hl.x1, hl.x2, hl.y );
      }

      am.init();
    }
  }

  //==========================================================================
  //  Interface SolidBrush:

  /**
   * Sets the current fill color (for {@link RasterGraphics#MODE_GRAY} or
   * {@link RasterGraphics#MODE_COLORMAP}).
   *
   * @param  _color Current fill color.
   * @see    RasterGraphics
   */
  public void setColor ( int _color )
  {
    RasterGraphics r = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    r.setColor( _color );
  }

  /**
   * Sets the current fill color (for {@link RasterGraphics#MODE_GRAY}).
   *
   * @param  _color Current fill color (<tt>0.0</tt> to <tt>1.0</tt>).
   * @see    RasterGraphics
   */
  public void setColor ( double _color )
  {
    RasterGraphics r = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    r.setColor( _color );
  }

  /**
   * Sets the current fill color (for {@link RasterGraphics#MODE_RGB},
   * {@link RasterGraphics#MODE_RGBA} or {@link RasterGraphics#MODE_BAND}).
   *
   * @param  _color Current fill color.
   * @see    RasterGraphics
   */
  public void setColor ( int[] _color )
  {
    RasterGraphics r = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    r.setColor( _color );
  }

  /**
   * Sets the current fill color (for {@link RasterGraphics#MODE_RGB},
   * {@link RasterGraphics#MODE_RGBA} or {@link RasterGraphics#MODE_BAND}).
   *
   * @param  _color Current fill color (channel values between <tt>0.0</tt> and <tt>1.0</tt>).
   * @see    RasterGraphics
   */
  public void setColor ( double[] _color )
  {
    RasterGraphics r = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    r.setColor( _color );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "SolidColorBrush";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME  = "SolidBrushAndBitMaskToRasterGraphics";

  /** Object template identifier. */
  protected final static String TEMPLATE2_NAME = "SolidBrushAndAlphaMaskToRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Solid-color brush.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DRAW + "." + C_RENDER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 1 )
    {
      t.newInputPlug( PL_INPUT, IFACE+"SolidBrush" );
        // Plugs:
      t.setRegStrings( NAME, (ord <= 0) ? TEMPLATE_NAME : TEMPLATE2_NAME ,CATEGORY, DESCRIPTION );
      if ( ord != 1 ) t.newOutputPlug( PL_BITMASK,   IFACE+"BitMask" );
      if ( ord != 0 ) t.newOutputPlug( PL_ALPHAMASK, IFACE+"AlphaMask" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"RasterGraphics" );
    }
    return 2;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece[] reg;
  static
  {
    reg = new RegPiece[ 2 ];
    reg[0] = new RegPiece();
    setTemplate( reg[0], 0 );
    reg[1] = new RegPiece();
    setTemplate( reg[1], 1 );
  }

}
