package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Sphere as elementary 3D solid.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/Sphere.java">Sphere.java</a>
 */
public class Sphere extends DefaultSolid
{
  //==========================================================================
  //  Interface Solid:

  /**
   * Computes intersection of the solid with the given ray.
   *
   * @param  P0 Ray origin ({@code double[3]} or {@code double[4]}).
   * @param  P1 Direction vector of the ray ({@code double[3]} or {@code double[4]}).
   * @return List of intersection points ({@link MicroFacet}s) sorted by {@code t}
   *         (parametric distance from {@code P0}). {@code null} is returned in case
   *         of bad arguments or if no intersection exists..
   * @see    MicroFacet
   * @see    BoundingVolume
   */
  public List< MicroFacet > intersection ( double[] P0, double[] P1 )
  {
    int P0len;
      // valid input parameters?
    if ( P0 == null || (P0len = P0.length) < 3 ||
         P1 == null ||          P1.length  < 3 ) return null;

      // ray origin:
    double Ox = P0[0];
    double Oy = P0[1];
    double Oz = P0[2];
    if ( P0len > 3 )
    {                                     // to cartesian
      double Ow = P0[3];
      if ( Geometry.isZero( Ow ) )
        return null;

      Ow  = 1.0 / Ow;
      Ox *= Ow;
      Oy *= Ow;
      Oz *= Ow;
    }
      // ray direction vector (of arbitrary size):
    final double Dx = P1[0];
    final double Dy = P1[1];
    final double Dz = P1[2];

    final double OD = Ox * Dx + Oy * Dy + Oz * Dz;
    final double DD = Dx * Dx + Dy * Dy + Dz * Dz;
    final double OO = Ox * Ox + Oy * Oy + Oz * Oz;
    double        d = OD * OD + DD * (1.0 - OO); // discriminant
    if ( d <= 0.0 ) return null;          // no intersections

    d = Math.sqrt( d );

      // there will be two intersections: (-OD - d) / DD, (-OD + d) / DD
    LinkedList< MicroFacet > list = new LinkedList< MicroFacet >();
    MicroFacet mf;

      // first intersection (-OD - d) / DD:
    mf          = new MicroFacet();
    mf.t        = (-OD - d) / DD;
    mf.enter    =
    mf.front    = true;
    mf.locCoord = new double[]
    {
      Ox + mf.t * Dx,
      Oy + mf.t * Dy,
      Oz + mf.t * Dz
    };
    mf.valid    = MicroFacet.VALID_LOC_COORD;
    list.addLast( mf );

      // second intersection (-OD + d) / DD:
    mf          = new MicroFacet();
    mf.t        = (-OD + d) / DD;
    mf.enter    =
    mf.front    = false;
    mf.locCoord = new double[]
    {
      Ox + mf.t * Dx,
      Oy + mf.t * Dy,
      Oz + mf.t * Dz
    };
    mf.valid    = MicroFacet.VALID_LOC_COORD;
    list.addLast( mf );

    return list;
  }

  /**
   * Asserts validity of the given intersection-attributes.
   * Resolves:
   * {@link MicroFacet#VALID_TANGENT VALID_TANGENT},
   * {@link MicroFacet#VALID_TEX_COORD VALID_TEX_COORD}.
   *
   * @param  mf The checked intersection.
   * @param  tmp Non-mandatory object holding state-information for deferred computations.
   * @param  attributes Attribute mask (required attributes - see
   *                    {@code MicroFacet.VALID_*}).
   * @return Attributes which were not computed.
   */
  @Override
  public int assertAttributes ( MicroFacet mf, Object tmp, int attributes )
  {
    if ( mf == null ) return attributes;

      // tangent surface vectors:
    if ( (attributes & MicroFacet.VALID_TANGENT) != 0 )
    {
      mf.tu = new double[ 3 ];
      mf.tv = new double[ 3 ];
      Geometry.getAxes( mf.locCoord, mf.tu, mf.tv );
      if ( mf.toWorld != null )
      {
        mf.toWorld.transformVector3D( mf.tu, mf.tu );
        mf.toWorld.transformVector3D( mf.tv, mf.tv );
      }
      mf.valid |= MicroFacet.VALID_TANGENT;
      if ( (attributes -= MicroFacet.VALID_TANGENT) == 0 ) return 0;
    }

      // 2D texture coordinates (inverse mapping):
    if ( (attributes & MicroFacet.VALID_TEX_COORD) != 0 )
    {
      mf.texCoord = new double[ 2 ];
      double r = Math.hypot( mf.locCoord[0], mf.locCoord[1] );
      mf.texCoord[0] = (r < Geometry.EPSILON) ? 0.0 :
                       (Math.atan2( mf.locCoord[1], mf.locCoord[0] ) / (2.0 * Math.PI) + 0.5);
      mf.texCoord[1] = Math.atan2( r, mf.locCoord[2] ) / Math.PI;
      mf.valid |= MicroFacet.VALID_TEX_COORD;
      if ( (attributes -= MicroFacet.VALID_TEX_COORD) == 0 ) return 0;
    }

    return attributes;
  }

  /**
   * Point vs. solid test ("inside-test").
   *
   * @param  point Coordinates of the 3D point ({@code double[3]} or
   *               {@code double[4]}).
   * @return {@code true} if the given point lies inside the solid.
   */
  @Override
  public boolean isInside ( double[] point )
  {
    if ( point == null ||
         point.length < 3 ) return false;

    double rr = (point.length > 3) ? (point[3] * point[3]) : 1.0;
    return( point[0] * point[0] + point[1] * point[1] + point[2] * point[2] <= rr );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Sphere";

  /** Object description. */
  private final static String DESCRIPTION = "Sphere as a 3D solid for CSG.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"Solid" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
