package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.xml.JGLScript;

/**
 * CSG representation of 3D scenes for Ray-tracing rendering methods.
 * Static scene object (no time-variable is implemented).
 *
 * @version 0.26 $Rev: 840 $ $Date: 2009-01-10 11:44:24 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/StaticCSGScene.java">StaticCSGScene.java</a>
 */
public class StaticCSGScene extends Piece implements RTScene
{
  //==========================================================================
  //  Data:

  /** JGL script key: CSG scene. */
  public final static String KEY_SCENE      = "scene";

  /** JGL script key: camera. */
  public final static String KEY_CAMERA     = "camera";

  /** JGL script key: lights. */
  public final static String KEY_LIGHTS     = "lights";

  /** JGL script key: background. */
  public final static String KEY_BACKGROUND = "background";

  /** JGL script key: Brep scene. */
  public final static String KEY_BREP       = "B-rep";

  /** JGL script key: OBJ import filter scene. */
  public final static String KEY_OBJ        = "import";

  /** Scene definition: JGL script file name. */
  protected String fileName = "data/scene1.jgl";

  /** Does scene object need to be read from JGL script? */
  protected boolean dirty = false;

  /** Root node of the CSG-tree. */
  protected SceneNode root;

  /** Static camera. */
  protected RayGenerator camera;

  /** Static array of light sources in the scene. */
  protected LightSource[] lights;

  /** Actual background color. */
  protected double[] background;

  /** Next assigned object id#. */
  protected int nextId = 0;

  //==========================================================================
  //  Support:

  /** Sets default light sources (one point and one ambient). */
  protected void defaultLights ()
  {
      // light sources:
    StaticPointLightSource light = new StaticPointLightSource();
    light.setParams( 10000,
                     new double[] { 10.0, -5.0,  4.0 },
                     new double[] { 30.0, 30.0, 30.0 },
                     new double[] {  0.5,  1.0,  0.1 } );
    AmbientLightSource ambient = new AmbientLightSource();
    ambient.setParams( new double[] { 1.0, 1.0, 1.0 } );
    lights = new LightSource[] { light, ambient };
  }

  /** Sets default camera. */
  protected void defaultCamera ()
  {
      // camera:
    StaticCamera camera = new StaticCamera();
    this.camera         = camera;
    camera.setParams( new double[] { 0.0,  0.5,  4.0 },
                      new double[] { 0.0,  0.0, -1.0 },
                      new double[] { 0.0, -1.0,  0.0 },
                      60.0 );
  }

  /** Initializes CSG scene to the default state (4 ellipsoids). */
  protected void defaultScene ()
  {
      // root scene node:
    CSGNode root  = new CSGNode( BinaryOperation.BIN_OP_OR );
    this.root     = root;

      // four leaf nodes: spheres
    Sphere  sph1  = new Sphere();
    CSGLeaf leaf1 = new CSGLeaf( sph1 );
    Sphere  sph2  = new Sphere();
    CSGLeaf leaf2 = new CSGLeaf( sph2 );
    Sphere  sph3  = new Sphere();
    CSGLeaf leaf3 = new CSGLeaf( sph3 );
    Sphere  sph4  = new Sphere();
    CSGLeaf leaf4 = new CSGLeaf( sph4 );
    root.insertChild( leaf1, null, null );
    TrMatrix m    = new TrMatrix( 4, 3 );
    TrMatrix mi   = new TrMatrix( 4, 3 );
    m.scaleAxes3D( null, new double[]{ 1.1, 0.5, 0.8 }, mi );
    m.translate3D( 0.2, 0.7, 0.4, mi );
    root.insertChild( leaf2, m, mi );
    m             = new TrMatrix( 4, 3 );
    mi            = new TrMatrix( 4, 3 );
    m.scaleAxes3D( null, new double[]{ 0.3, 1.2, 0.3 }, mi );
    m.translate3D( -0.7, 0.6, 1.2, mi );
    root.insertChild( leaf3, m, mi );
    m             = new TrMatrix( 4, 3 );
    mi            = new TrMatrix( 4, 3 );
    m.scaleAxes3D( null, new double[]{ 2.0, 0.08, 0.08 }, mi );
    m.translate3D( 0.0, 0.25, 1.46, mi );
    root.insertChild( leaf4, m, mi );

      // scene attributes: I. global material
    PhongModel material1 = new PhongModel();
    root.setAttribute( SceneNode.ATTR_BSSRDF, material1  );

      // II. special material for the 2nd solid
    PhongModel material2 = new PhongModel();
    material2.set( "Color", new double[] { 0.1, 0.1, 0.8 } );
    material2.set( "Kd",    0.2 );        // mostly specular material
    material2.set( "Ks",    0.8 );
    material2.set( "H",     32 );
    leaf2.setAttribute( SceneNode.ATTR_BSSRDF, material2  );

      // III. special material for the 3rd solid
    PhongModel material3 = new PhongModel();
    material3.set( "Color", new double[] { 1.0, 0.2, 0.1 } );
    material3.set( "Kd",    0.8 );        // mostly diffuse material
    material3.set( "Ks",    0.2 );
    material3.set( "H",     10 );
    leaf3.setAttribute( SceneNode.ATTR_BSSRDF, material3  );

      // IV. special material for the 4th solid
    PhongModel material4 = new PhongModel();
    material4.set( "Color", new double[] { 0.2, 1.0, 0.1 } );
    material4.set( "Kd",    0.4 );
    material4.set( "Ks",    0.6 );
    material4.set( "H",     16 );
    leaf4.setAttribute( SceneNode.ATTR_BSSRDF, material4  );
  }

  /** Initializes default background color. */
  protected void defaultBackground ()
  {
    background = new double[] { 0.05, 0.08, 0.1 };
  }

  /** All scene components have be usable. */
  protected void assertScene ()
  {
    if ( root       != null &&
         camera     != null &&
         lights     != null &&
         background != null &&
         !dirty ) return;

    nextId = 0;
    dirty  = false;

    if ( fileName != null )
    {
      final Brep           brep     = (Brep)          getInterface( PL_DATA,   IFACE+"Brep" );
      final DataFileFormat import3d = (DataFileFormat)getInterface( PL_IMPORT, IFACE+"DataFileFormat" );

      Map< String, Object > input = new HashMap< String, Object > ();
      if ( brep     != null ) input.put( KEY_BREP, brep );
      if ( import3d != null ) input.put( KEY_OBJ,  import3d );

      JGLScript script = new JGLScript();
      Map< String, Object > map = script.interpret( fileName, input );

      if ( map != null )
      {
        Object o;

        if ( (o = map.get( KEY_SCENE ) ) != null &&
             o instanceof SceneNode )
          root = (SceneNode)o;

        if ( (o = map.get( KEY_LIGHTS ) ) != null &&
             o instanceof LightSource[] )
          lights = (LightSource[])o;

        if ( (o = map.get( KEY_CAMERA ) ) != null &&
             o instanceof RayGenerator )
          camera = (RayGenerator)o;

        if ( (o = map.get( KEY_BACKGROUND ) ) != null &&
             o instanceof double[] )
          background = (double[])o;
      }
    }

    if ( root == null )
      defaultScene();

    if ( lights == null )
      defaultLights();

    if ( camera == null )
      defaultCamera();

    if ( background == null )
      defaultBackground();

      // bind light-sources to the scene:
    if ( lights != null )
      for ( LightSource ls : lights )
        ls.setIntersectable( this );
  }

  //==========================================================================
  //  Interface Intersectable:

  /**
   * Computes the nearest intersection of the given ray with scene surface.
   *
   * @param  P0 Ray origin ({@code double[3]} or {@code double[4]}).
   * @param  P1 Direction vector of the ray ({@code double[3]} or {@code double[4]}).
   * @return The nearest intersection point or {@code null}.
   */
  public MicroFacet intersection ( double[] P0, double[] P1 )
  {
    if ( P0 == null || P0.length < 3 ||
         P1 == null || P1.length < 3 ||
         root == null ) return null;

    final double p1Squared = P1[0] * P1[0] + P1[1] * P1[1] + P1[2] * P1[2];

    List< MicroFacet > inter = root.intersection( P0, P1 );
    if ( inter == null ||
         inter.isEmpty() ) return null;

      // there is anything in the list..
    for ( MicroFacet mf : inter )         // find the 1st positive intersection point
      if ( mf.far( 0.0, p1Squared ) ) return mf;

    return null;
  }

  //==========================================================================
  //  Interface TimeDependent:

  /**
   * Retrieve the valid bounds for time variable.
   *
   * @param  pre Pre-allocated {@code double[2]} array to hold the result (can be
   *             {@code null}).
   * @return {@code result[0]} .. minimum time value, {@code result[1]} .. maximum
   *         time value.
   * @see    #getTime
   * @see    #setTime
   */
  public double[] getTimeInterval ( double[] pre )
  {
    if ( pre == null ||
         pre.length < 2 )
      pre = new double[ 2 ];

    pre[0] =
    pre[1] = 0.0;
    return pre;
  }

  /**
   * Returns the actual time value for the scene (used in animations).
   *
   * @return Actual time (in seconds).
   * @see    #setTime
   */
  public double getTime ()
  {
    return 0.0;
  }

  /**
   * Sets the actual time value for the scene (used in animations).
   *
   * @param  time New scene time (in seconds).
   * @see    #getTime
   * @see    #getTimeInterval
   */
  public void setTime ( double time )
  {
  }

  //==========================================================================
  //  Interface RTScene:

  /**
   * Sets root node (assumes hierarchical scene representation based on {@link SceneNode}).
   */
  public void setRoot ( SceneNode root )
  {
    this.root = root;
  }

  /**
   * Sets the camera object (ray-generator) for the scene.
   */
  public void setCamera ( RayGenerator camera )
  {
    this.camera = camera;
  }

  /**
   * Retrieves the actual camera (ray-generator).
   *
   * @return {@code RayGenerator} object.
   */
  public RayGenerator getCamera ()
  {
    assertScene();

    final RayGenerator cam = (RayGenerator)getInterface( PL_OUTPUT, IFACE+"RayGenerator" );
    if ( cam != null )
      camera = cam;

    return camera;
  }

  /**
   * Sets all the light sources from the given array.
   */
  public void setLights ( LightSource[] l )
  {
    lights = l;
  }

  /**
   * Retrieves all light sources from the scene.
   *
   * @return Array of light sources.
   * @see    #getLightsToMicroFacet
   */
  public LightSource[] getLights ()
  {
    assertScene();

    return lights;
  }

  /**
   * Retrieves light sources which light up the given micro-facet.
   *
   * @param  mf Micro-facet which is to be lighted.
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @return Array of light sources ({@code null} marks end of data).
   * @see    #getLights
   */
  public LightSource[] getLightsToMicroFacet ( MicroFacet mf, int order, int total )
  {
    if ( mf     == null ||
         lights == null ) return null;

    LightSource l[] = new LightSource[ lights.length ];
    int i = 0;
    for ( LightSource light : lights )
      if ( light != null && light.lights( mf, order, total ) )
        l[i++] = light;

    return l;
  }

  /** Retrieves background color (used when a ray doesn't hit anything). */
  public double[] getBackground ()
  {
    assertScene();

    return background;
  }

  //==========================================================================
  //  Interface Property:

  public final static String SCENE = "Scene";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // SCENE:
    if ( key.compareTo(SCENE) == 0 )
    {
      fileName = stringProperty( value, fileName );
      dirty    = true;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(SCENE) == 0 )
      return fileName;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "CSG-scene (static)";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "RTScene";

  /** Object description. */
  private final static String DESCRIPTION = "CSG representation of 3D scene for Ray-tracing " +
                                            "(static implementation).";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_DATA + '.' + C_SCENE;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,        IFACE+"RTScene" );
      t.newInputPlug(     PL_INTERSECTION, IFACE+"Intersectable" );
      t.newInputPlug(     PL_PROPERTY,     IFACE+"Property" );
      t.newOptOutputPlug( PL_OUTPUT,       IFACE+"RayGenerator" );
      t.newOptOutputPlug( PL_DATA,         IFACE+"Brep" );
      t.newOptOutputPlug( PL_IMPORT,       IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( SCENE, TYPE_OBJECT, "Scene definition file-name", true );
        t.propDefault( "data/scene1.jgl" );
        t.propManipulator( "ParamJGLScript" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}