package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.xml.JGLScript;

/**
 * Static perspective camera - ray generator.
 *
 * @version 0.26 $Rev: 444 $ $Date: 2007-05-05 22:52:07 +0200 (so, 05 V 2007) $ $Author: pepca $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/StaticCamera.java">StaticCamera.java</a>
 */
public class StaticCamera extends Piece implements RayGenerator
{
  //==========================================================================
  //  Data:

  /** JGL script key: center. */
  public final static String KEY_CENTER = "center";

  /** JGL script key: front. */
  public final static String KEY_FRONT  = "front";

  /** JGL script key: up. */
  public final static String KEY_UP     = "up";

  /** JGL script key: angle. */
  public final static String KEY_ANGLE  = "angle";

  /** JGL script file-name. */
  protected String scriptName = "data/camera.jgl";

  /** Center of the projection. */
  protected double[] center;

  /** Direction vector (normal vector of the projection plane). */
  protected double[] front;

  /** Vector which points up in the projection plane. */
  protected double[] up;

  /** Vector which points to the right in the projection plane. */
  protected double[] right;

  /** Origin of ray direction vector. */
  protected double[] ray0;

  /** X difference vector for ray direction. */
  protected double[] rayDx;

  /** Y difference vector for ray direction. */
  protected double[] rayDy;

  /** Horizontal projection-plane bounds (from left to right). */
  protected double xMin, xMax;

  /** Vertical projection-plane bounds (from up to bottom). */
  protected double yMin, yMax;

  //==========================================================================
  //  Support:

  /**
   * Set camera parameters.
   *
   * @param  cnt Center of the projection ({@code null} for {@code [0,0,0]}).
   * @param  fr Viewing direction (associated with the center of projection plane).
   * @param  u Up vector (which will point up in the projection plane).
   * @param  angle Horizontal viewing angle in degrees.
   * @return {@code true} if parameters were successfully set.
   */
  public boolean setParams ( double[] cnt, double[] fr, double[] u,
                             double angle )
  {
      // center of the projection:
    if ( cnt == null || cnt.length < 3 )
      center = new double[ 3 ];
    else
      center = Geometry.toCartesian3D( cnt, null );

      // viewing vector:
    if ( fr == null || fr.length < 3 ) return false;
    front = Geometry.normalize3D( fr, null );

      // "up" vector:
    if ( u == null || u.length < 3 ) return false;
    up = Geometry.normalize3D( u, null );
    double cosFrontUp = front[0] * up[0] + front[1] * up[1] + front[2] * up[2];
    if ( cosFrontUp > 1.0 - Geometry.EPSILON ) return false;

    up[0] -= cosFrontUp * front[0];
    up[1] -= cosFrontUp * front[1];
    up[2] -= cosFrontUp * front[2];
    Geometry.normalize3D( up );

      // "right" vector:
    right  = Geometry.vectorProduct( up, front, null );
    Geometry.normalize3D( right );

      // viewing half-angle:
    angle *= Math.PI / 360.0;

      // ray-generator support:
    ray0   = front;
    rayDx  = right;
    rayDy  = new double[] { -up[0], -up[1], -up[2] };

    xMax   = Math.tan( angle );
    xMin   = -xMax;

    setAspectRatio ( 4.0 / 3.0 );

    return true;
  }

  /** Read data from JGL script. */
  protected void readScript ()
  {
    if ( scriptName == null ) return;

    JGLScript script = new JGLScript();
    Map< String, Object > map = script.interpret( scriptName );
    if ( map != null )
    {
      Object o;
      double[] cnt = null;
      double[] fr  = { 0.0,  0.0, -1.0 };
      double[] up  = { 0.0, -1.0,  0.0 };
      double angle = 60.0;

      if ( (o = map.get( KEY_CENTER ) ) != null &&
           o instanceof double[] )
        cnt = (double[])o;

      if ( (o = map.get( KEY_FRONT ) ) != null &&
           o instanceof double[] )
        fr = (double[])o;

      if ( (o = map.get( KEY_UP ) ) != null &&
           o instanceof double[] )
        up = (double[])o;

      if ( (o = map.get( KEY_ANGLE ) ) != null &&
           o instanceof Double )
        angle = ((Double)o).doubleValue();

      setParams( cnt, fr, up, angle );
    }
  }

  //==========================================================================
  //  Interface RayGenerator:

  /**
   * Sets aspect ratio of the target image.
   * Default aspect ratio (if not set) should be <b>4/3</b>.
   *
   * @param  aspect Aspect ratio = width / height. Uses real coordinates.
   */
  public void setAspectRatio ( double aspect )
  {
      // picture bounds:
    yMax = xMax / aspect;
    yMin = xMin / aspect;
  }

  /**
   * Returns boundary values for projection-plane coordinates (ray-coordinates).
   *
   * @param  pre Pre-allocated array to hold results ({@code double[4]} array). Can be
   *             {@code null}.
   * @return X and Y coordinate bounds: {@code result[0]} .. X<sub>min</sub>,
   *         {@code result[1]} .. X<sub>max</sub>,
   *         {@code result[2]} .. Y<sub>min</sub>,
   *         {@code result[3]} .. Y<sub>max</sub>.
   * @see    #getRay
   */
  public double[] getViewport ( double[] pre )
  {
    if ( pre == null || pre.length < 4 )
      pre = new double[ 4 ];

    pre[0] = xMin; pre[1] = xMax;
    pre[2] = yMin; pre[3] = yMax;
    return pre;
  }

  /**
   * Ray-generator method.
   *
   * @param  x X coordinate in the projection plane.
   * @param  y Y coordinate in the projection plane.
   * @param  P0 Pre-allocated {@code double[3]} (or {@code double[4]}) array to hold
   *            coordinates of the <b>ray origin</b>.
   * @param  P1 Pre-allocated {@code double[3]} (or {@code double[4]}) array to hold
   *            <b>direction vector</b> of the ray.
   * @return {@code true} if the ray was successfully generated.
   */
  public boolean getRay ( double x, double y, double[] P0, double[] P1 )
  {
    if ( P0 == null || P0.length < 3 ||
         P1 == null || P1.length < 3 ||
         x < xMin   || x > xMax ||
         y < yMin   || y > yMax ) return false;

      // ray origin:
    System.arraycopy( center, 0, P0, 0, 3 );
    if ( P0.length > 3 ) P0[3] = 1.0;

      // ray direction:
    P1[0] = ray0[0] + x * rayDx[0] + y * rayDy[0];
    P1[1] = ray0[1] + x * rayDx[1] + y * rayDy[1];
    P1[2] = ray0[2] + x * rayDx[2] + y * rayDy[2];
    Geometry.normalize3D( P1 );
    if ( P1.length > 3 )
      P1[3] = 0.0;

    return true;
  }

  /**
   * Ray-generator method. Passes parameters for hidden sampling.
   *
   * @param  x X coordinate in the projection plane.
   * @param  y Y coordinate in the projection plane.
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  P0 Pre-allocated {@code double[3]} (or {@code double[4]}) array to hold
   *            coordinates of the <b>ray origin</b>.
   * @param  P1 Pre-allocated {@code double[3]} (or {@code double[4]}) array to hold
   *            <b>direction vector</b> of the ray.
   * @return {@code true} if the ray was successfully generated.
   */
  public boolean getRay ( double x, double y, int order, int total, double[] P0, double[] P1 )
  {
    return getRay( x, y, P0, P1 );
  }

  //==========================================================================
  //  Interface Property:

  public final static String PARAMS = "Params";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // PARAMS:
    if ( key.compareTo(PARAMS) == 0 )
    {
      scriptName = stringProperty( value, scriptName );
      readScript();
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(PARAMS) == 0 )
      return scriptName;

    return null;
  }

  //==========================================================================
  //  Module initialization:

  {
    setParams( null,
               new double[] { 0.0,  0.0, -1.0 },
               new double[] { 0.0, -1.0,  0.0 },
               60.0 );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Perspective camera (static)";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "RayGenerator";

  /** Object description. */
  private final static String DESCRIPTION = "Camera for perspective projection " +
                                            "(static implementation).";
  /** Object category. */
  protected final static String CATEGORY = C_3D + "." + C_CAMERA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"RayGenerator" );
        // Property:
      t.propBegin( PARAMS, TYPE_OBJECT, "Camera parameters", true );
        t.propDefault( "data/camera.jgl" );
        t.propManipulator( "ParamJGLScript" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
