package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.xml.JGLScript;

/**
 * Static point light source.
 *
 * @version 0.26 $Rev: 858 $ $Date: 2009-02-17 23:48:39 +0100 (út, 17 II 2009) $ $Author: Tomas $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/StaticPointLightSource.java">StaticPointLightSource.java</a>
 */
public class StaticPointLightSource extends DefaultLightSource
{
  //==========================================================================
  //  Data:

  /** JGL script key: position of up right vertex. */
  public final static String PHOTONS      = "photons";

  /** JGL script key: light position. */
  public final static String KEY_POSITION = "position";

  /** JGL script key: dim. */
  public final static String KEY_DIM      = "dim";

  {
    scriptName = "data/light.jgl";
  }

  /** Random number generator. */
  protected final RandomJames rnd = new RandomJames();

  /** Number of shooted photons (for methods based on photon tracing). */
  protected int[] photons = { 0 };

  /** Light source coordinates. */
  protected double[] position = new double[] { 0.0, 0.0, 0.0 };

  /**
   * Color of the source: {@code 1D} for monochrome source, {@code 3D} for <b>RGB</b>
   * color space, or more for complex spectrum sampled spaces.
   */
  protected double[] lightColor = new double[] { 1.0 };

  /**
   * Dimming polynom coefficients: light is dimmed by the factor of
   * {@code dim[0] + dim[1] * D + dim[2] * D * D}.
   * Can be {@code null} for no dimming..
   */
  protected double[] dim = new double[] { 0.5, 1.0, 0.1 };

  //==========================================================================
  //  Support:

  /**
   * Set light source parameters.
   *
   * @param  pos Source position.
   * @param  color Source color/intensity (can be either 1D /for monochrome computing/,
   *               3D /for RGB color system/ or greater /for more complex spectral color
   *               sampling/).
   * @param  dimming Dimming polynom coefficients: light is dimmed by the factor of
   *                 {@code dim[0] + dim[1] * D + dim[2] * D * D}. Can be
   *                 {@code null} for no dimming..
   * @return {@code true} if parameters were successfully set.
   */
  public boolean setParams ( int photons, double[] pos, double[] color,
                             double[] dimming )
  {
      // source position:
    if ( pos == null ||
         pos.length < 3 ) return false;

    this.photons[0] = photons;

    position = Geometry.toCartesian3D( pos, null );

      // source color/intensity:
    if ( color == null )
    {
      if ( lightColor        == null ||
           lightColor.length != 1 )
        lightColor = new double[ 1 ];

      lightColor[0] = 1.0;
    }
    else
    {
      if ( lightColor        == null ||
           lightColor.length != color.length )
        lightColor = new double[ color.length ];

      System.arraycopy( color, 0, lightColor, 0, color.length );
    }

      // dimming polynom:
    if ( dimming == null )
      dim = null;
    else
    {
      if ( dim == null || dim.length != 3 )
        dim = new double[ 3 ];

      for ( int i = 0; i < 3 && i < dimming.length; i++ )
        dim[i] = dimming[i];
    }

    return true;
  }

  /** Read data from JGL script. */
  protected void readScript ()
  {
    if ( scriptName == null ) return;

    JGLScript script = new JGLScript();
    Map< String, Object > map = script.interpret( scriptName );
    if ( map != null )
    {
      Object o;

      if ( (o = map.get( PHOTONS ) ) != null &&
           o instanceof int[] )
        photons = (int[])o;

      if ( (o = map.get( KEY_POSITION ) ) != null &&
           o instanceof double[] )
        position = Geometry.toCartesian3D( (double[])o, null );

      if ( (o = map.get( KEY_LIGHT ) ) != null &&
           o instanceof double[] )
        lightColor = (double[])o;

      if ( (o = map.get( KEY_DIM ) ) != null &&
           o instanceof double[] )
        dim = (double[])o;
    }
  }

  /** For debugging purposes only. */
  @Override
  public String toString ()
  {
    StringBuffer sb = new StringBuffer( getClass().getName() );
    sb.append( "[photons=" )
      .append( Formula.vectorToString( photons ) )
      .append( "[pos=" )
      .append( Formula.vectorToString( position ) )
      .append( ",color=" )
      .append( Formula.vectorToString( lightColor ) )
      .append( ",dim=" )
      .append( Formula.vectorToString( dim ) )
      .append( ']' );
    return sb.toString();
  }

  //==========================================================================
  //  Interface LightSource:

  /**
   * Lighting test from the source to the given {@link MicroFacet}.
   *
   * @param  mf The checked {@link MicroFacet} (point on surface of some solid).
   * @return {@code true} if the light source lights on the given point.
   * @see    #getIntensity
   * @see    #getDirection
   */
  @Override
  public boolean lights ( MicroFacet mf )
  {
    if ( mf == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD |
                              MicroFacet.VALID_NORMAL ) != 0 ) return true;

    final double[] direction = Geometry.vector( mf.coord, position, null );
    if ( mf.normal[0] * direction[0] +
         mf.normal[1] * direction[1] +
         mf.normal[2] * direction[2] <= 0.0 ) return false;

      // check the intersection of a light ray with the scene:
    if ( assertInter() )
      return true;

      // does the sample point see the front side of the MicroFacet?
    MicroFacet c = inter.intersection( mf.coord, direction );

    return( c == null ||
            c.far( 1.0, direction ) );
  }

  /**
   * Returns light intensity coming from the source to the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  color Pre-allocated {@code double[C]} array to hold result - light intensity
   *               coming from the source to the given point ({@code C} can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @return {@code true} if the light source can light on the given point (<b>cannot</b>
   *         replace the {@code lights()} call!).
   * @see    #lights
   * @see    #getDirection
   */
  @Override
  public boolean getIntensity ( MicroFacet mf, double[] color )
  {
    double dimCoef = 1.0;
    if ( position != null &&
         dim      != null &&
         mf       != null &&
         mf.assertAttributes( MicroFacet.VALID_COORD ) == 0 )
    {
      double dist = Geometry.pointPoint3D( mf.coord, position );
      dimCoef /= dim[0] + dist * (dim[1] + dist * dim[2]);
    }

    int i;
    if ( color.length <= lightColor.length )
      for ( i = 0; i < color.length; i++ )
         color[i] = dimCoef * lightColor[i];
    else
      for ( i = 0; i < color.length; i++ )
         color[i] = dimCoef * lightColor[0];

    return true;
  }

  /**
   * Returns light direction between the source and the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  dir Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *             from the {@link MicroFacet} to the light source.
   * @return {@code false} if the light source is not directional (thus {@code dir}
   *         is irrelevant).
   * @see    #lights
   * @see    #getIntensity
   */
  @Override
  public boolean getDirection ( MicroFacet mf, double[] dir )
  {
    if ( dir        == null ||
         dir.length <  3 ||
         position   == null ||
         mf         == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD ) != 0 ) return false;

    Geometry.vector( mf.coord, position, dir );
    Geometry.normalize3D( dir );

    return true;
  }

  /**
   * Sample the light by one shadow ray.
   *
   * @param  mf    The {@link MicroFacet} (point on surface of some solid).
   * @param  u     Uniform random number.
   * @param  v     Uniform random number.
   * @param  color Pre-allocated <tt>double[C]</tt> array to hold result - light intensity
   *               coming from the source to the given point (<tt>C</tt> can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @param  dir   Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *               from the {@link MicroFacet} to the light source.
   * @param  pdf   Probability density function (output variable).
   * @return {@code True} if the light is visible.
   */
  @Override
  public boolean sample ( MicroFacet mf,
                          double u,
                          double v,
                          double[] color,
                          double[] dir,
                          double[] pdf,
                          boolean testVisibility )
  {
    if ( dir          == null ||
         dir.length   <  3    ||
         position     == null ||
         color        == null ||
         color.length <  3    ||
         mf           == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD ) != 0 ) return false;

    Geometry.vector( mf.coord, position, dir );
    Geometry.normalize3D( dir );

      // compute pdf:
    double dist = Geometry.pointPoint3D( position, mf.coord );
    double dimCoef = (dim[0] + dist * (dim[1] + dist * dim[2]));
    pdf[0] = dimCoef;

    System.arraycopy( lightColor, 0, color, 0, color.length );

      // check visibility if it's required:
    if ( testVisibility )
    {
      final double[] direction = Geometry.vector( mf.coord, position, null );
      MicroFacet m = inter.intersection( mf.coord, direction );
      return ( m == null ||
               m.far( 1.0, direction ) );
    }
    return true;
  }

  /**
   * Generates one photon leaving this light source.
   *
   * @param  u1    Uniform random number.
   * @param  v1    Uniform random number.
   * @param  u2    Uniform random number.
   * @param  v2    Uniform random number.
   * @param  pos   Pre-allocated array which holds a point on the light source
   *               (an origin of direction).
   * @param  dir   Output direction in world coordinates. Pre-allocated array.
   * @param  power Power of given light source (if you wanna get power
   *               of this photon, you have to divide this power by number
   *               of photons). Pre-allocated array.
   * @return Weight of given ray (standart weight = 1.0).
   */
  @Override
  public double sample ( double u1,
                         double v1,
                         double u2,
                         double v2,
                         double[] pos,
                         double[] dir,
                         double[] power )
  {
    if ( pos   == null ||
         dir   == null ||
         power == null )
      return 0.0;

    do
    {
      dir[0] = rnd.randomDouble( -1.0f, 1.0f );
      dir[1] = rnd.randomDouble( -1.0f, 1.0f );
      dir[2] = rnd.randomDouble( -1.0f, 1.0f );
    } while ( dir[0]*dir[0] +
              dir[1]*dir[1] +
              dir[2]*dir[2] > 1.0f );

    System.arraycopy( position,   0, pos, 0,   pos.length );
    System.arraycopy( lightColor, 0, power, 0, power.length );

    return 1.0;
  }

  /**
   * Return probability density function.
   *
   * @param point  Point on the surface.
   * @param normal Normal at this point.
   * @param dir    Direction from the point.
   */
  @Override
  public double pdf ( final double[] point, final double[] normal, final double[] dir )
  {
    // It's a waste of time to try to hit the delta light source.
    return 0.0;
  }

  /** Is delta light? */
  public boolean isDeltaLight ()
  {
    return true;
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // PARAMS:
    if ( key.compareTo(PARAMS) == 0 )
    {
      scriptName = stringProperty( value, scriptName );
      readScript();
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(PHOTONS) == 0 )
      return Formula.cloneArray( photons );
    if ( key.compareTo(PARAMS) == 0 )
      return scriptName;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Point light source (static)";

  /** Object description. */
  private final static String DESCRIPTION = "Point light source shining uniformly to all " +
                                            "directions (static implementation).";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"LightSource" );
      t.newOptOutputPlug( PL_OUTPUT, IFACE+"Intersectable" );
        // Property:
      t.propBegin( PARAMS, TYPE_OBJECT, "Light source parameters", true );
        t.propDefault( "data/light.jgl" );
        t.propManipulator( "ParamJGLScript" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}