package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * SLE solver: special solver for radiosity problems - Super-Shoot-Gather.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 754 $ $Date: 2008-05-18 18:08:40 +0200 (ne, 18 V 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/SuperShootGather.java">SuperShootGather.java</a>
 */
public class SuperShootGather extends Piece implements SLESolver
{
  //==========================================================================
  //  Data:

  /** Accelerator: input SLEData. */
  protected SLEData data;

  /** Required accuracy of solution */
  protected double epsilon = 1.e-6;

  /** Use ambient term? */
  protected boolean useAmbient = false;

  /** Number of performed iteration steps. */
  protected int steps = 0;

  /** Number of unknowns. */
  protected int N;

  /** Actual vector of unknown variables. */
  protected double[] B;

  /** Second vector of unknown variables for ambient term. */
  protected double[] B2;

  /** Actual row of matrix of unshoot radiosity. */
  protected double[] dBRow;

  /** Actual column of matrix of unshoot radiosity. */
  protected double[] dBCol;

  /**
   * Radiosity of a patch that results from the unshot radiosity bouncing
   * around our restricted environment.
   */
  protected double[] SG;

  /** Actual matrix of shooted radiosity. */
  protected float[][] nB;

  /** Actual matrix row. */
  protected double[] kRow;

  /** Actual matrix column. */
  protected double[] kCol;

  /** Weights of unknowns (element areas). */
  protected double[] weight;

  /** Reflectance of each element. */
  protected double[] reflectance;

  /** Avarage reflectance. */
  protected double refTotal = 0.0;

  /** Ambient term (total unshot radiosity in the environment). */
  protected double ambient = 0.0;

  /** Was the solver already initialized? */
  protected boolean isInitialized = false;

  //==========================================================================
  //  Interface SLESolver:

  /** Sets solver accuracy (whatever it means). */
  @Override
  public double setEpsilon ( double epsilon )
  {
    double old   = this.epsilon;
    this.epsilon = epsilon;
    return old;
  }

  /**
   * Solve the whole system en bloc.
   *
   * @param  save Solution will be saved into the associated {@link SLEDataStore}.
   */
  @Override
  public void solve ( boolean save )
  {
    solve( Integer.MAX_VALUE, save );
  }

  /** Reset iterative solution algorithm (if applicable). */
  @Override
  public void reset ()
  {
    if ( data == null ) data = (SLEData)getInterface( PL_OUTPUT, IFACE+"SLEData" );
    if ( data == null )
      return;

    N      = data.getColumns();
    B      = data.getRight( null );
    SG     = new double[ N ];
    nB     = new float[ N ][ N ];
    dBRow  = new double[ N ];
    dBCol  = new double[ N ];
    kCol   = new double[ N ];
    kRow   = new double[ N ];
    weight = data.getWeight( null );
    steps  = 0;

    if ( useAmbient )
    {
      B2          = new double[ N ];
      reflectance = data.getReflectance( null );
      computeAvgReflectance();
    }

    isInitialized = true;
  }

  /**
   * Refine the solution performing several iterations.
   *
   * @param  maxIterations Number of iterations requested.
   * @param  save Solution will be saved into the associated {@link SLEDataStore}.
   * @return Achieved precision (if less than epsilon, the system has converged).
   */
  @Override
  public double refine ( int maxIterations, boolean save )
  {
    return solve( maxIterations, save );
  }

  /**
   * Input data was changed, solver should adapt to it.
   * Usually {@link #solve} or {@link #refine} will be called afterwards..
   */
  @Override
  public void update ()
  {
    reset();
  }

  /** Reads a single unknown (element of solution vector). Non-mandatory. */
  @Override
  public double getSolution ( int col )
  {
    if ( col <  0 ||
         col >= N ||
         B == null ) return 0.0;

    return( useAmbient ? B2[col] : B[col] );
  }

  /** Reads the whole solution vector. Non-mandatory. */
  @Override
  public double[] getSolution ( double[] result )
  {
    if ( result == null ||
         result.length < N )
      result = new double[ N ];

    System.arraycopy( useAmbient ? B2 : B, 0, result, 0, N );

    return result;
  }

  //==========================================================================
  //  Multi-component systems:

  /** Get actual component number. */
  @Override
  public int actualComponent ()
  {
    return 0;
  }

  /** Set actual component (non-mandatory). */
  @Override
  public void setComponent ( int c )
  {
  }

  //==========================================================================
  //  Individual relaxation methods:

  /**
   * Refine the solution performing several iterations.
   *
   * @param  maxIterations Number of iterations requested.
   * @param  save Solution will be saved into the associated {@link SLEDataStore}.
   * @return Achieved precision (if less than epsilon, the system has converged).
   */
  protected double solve ( int maxIterations, boolean save )
  {
    if ( data == null ||
         !isInitialized )
    {
      reset();
      if ( data == null )
        return Double.POSITIVE_INFINITY;
    }

    int i = 0, j;

    int iter        = maxIterations;
    double accuracy = 0.0;
    double deltaRad = 0.0;

    do
    {
        // pick i:
      double actSG = 0.0, maxSG = 0.0;
      for ( j = 0; j < N; j++ )
      {
        actSG = SG[j] * weight[j];
        if ( actSG > maxSG )
        {
          maxSG = actSG;
          i = j;
        }
      }

        // read one (working) column and row:
      data.getColumn( i, kCol );
      data.getRow( i, kRow );

        // compute unshoot radiosity:
      for ( j = 0; j < N; j++ )
        dBRow[j] = B[i] - nB[i][j];
      for ( j = 0; j < N; j++ )
        dBCol[j] = B[j] - nB[j][i];

        // do a shot:
      for ( j = 0; j < N; j++ )
      {
        if ( j != i )
          B[j] += dBRow[j] * -kCol[j];

        nB[i][j] = (float)B[i];
      }

        // compute SG:
      double numerator = 0.0;
      for ( j = 0; j < N; j++ )
        if ( j != i )
          numerator -= kRow[j] * dBCol[j];

      double denominator = 1.0;
      for ( j = 0; j < N; j++ )
        if ( j != i )
          denominator -= kRow[j] * kCol[j];
      SG[i] = numerator / denominator;

      B[i] += SG[i];

        // do a gather:
      for ( j = 0; j < N; j++ )
      {
        if ( j != i )
        {
          SG[j] = dBCol[j] - kCol[j] * SG[i];
          B[j] += SG[j] - dBCol[j];;
        }

        nB[j][i] = (float)B[j];
        nB[i][j] = (float)B[i];
      }

        // compute max unshoot radiosity:
      accuracy = 0;
      for ( j = 0; j < N; j++ )
        if ( accuracy < Math.abs( SG[j] ) )
          accuracy = Math.abs( SG[j] );

        // increment variables:
      if ( ++i >= N ) i = 0;
      steps++;
    }
    while ( accuracy > epsilon &&
            --iter > 0 );

      // add ambient term to result:
    if ( useAmbient )
    {
      updateAmbient();
      for ( j = 0; j < N; j++ )
        B2[j] = B[j] + ambient * reflectance[j];
    }

      // saving results..
    if ( save )
      data.setSolution( useAmbient ? B2 : B );

    return accuracy;
  }

  //==========================================================================
  //  Support:

  /** Compute avarage reflectance. Must be called before {@link #updateAmbient}. */
  protected void computeAvgReflectance ( )
  {
    if ( reflectance == null )
      return;

    int i;

    double sumARef = 0.0;
    for ( i = 0; i < N; i++ )
      sumARef += reflectance[i] * weight[i];

    double sumA = 0.0;
    for ( i = 0; i < N; i++ )
      sumA += weight[i];

      // compute total reflectance:
    final double averageRef = 1.0 - sumARef / sumA;
    if ( !Geometry.isZero( averageRef ) )
      refTotal = 1.0 / averageRef;
  }

  /** Update ambient term. Must be called after {@link #updateAmbient}. */
  protected void updateAmbient ()
  {
    if ( weight == null )
      return;

    int i;

      // compute avarage DB:
    double sumADB = 0.0;
    for ( i = 0; i < N; i++ )
      sumADB += SG[i] * weight[i];

    double sumA = 0.0;
    for ( i = 0; i < N; i++ )
      sumA += weight[i];

    ambient = sumADB / sumA * refTotal / N;
  }

  //==========================================================================
  //  Support for get:

  protected double getTotalResidual ()
  {
    double total = 0.0;
    for ( int j = 0; j < N; j++ )
      total += Math.abs( SG[j] );
    return total;
  }

  //==========================================================================
  //  Interface Property:

  public final static String EPSILON = "Epsilon";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // EPSILON:
    if ( key.compareTo(EPSILON) == 0 )
    {
      epsilon = doubleProperty( value, epsilon );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(EPSILON) == 0 )
      return epsilon;
    if ( key.compareTo(STEPS) == 0 )
      return steps;
    if ( key.compareTo(TOTAL_RESIDUAL) == 0 )
      return getTotalResidual();

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "SuperShootGather";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "SLESolverToSLEData";

  /** Object description. */
  private final static String DESCRIPTION = "Special SLE solver for radiosity - Super-Shoot-Gather method";

  /** Object category. */
  protected final static String CATEGORY = "SLE.solver";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"SLESolver" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"SLEData" );
        // Property:
      t.propBegin( EPSILON, TYPE_DOUBLE, "Required solution accuracy", true );
        t.propDefault( 1.e-6 );
      t.propEnd();
      t.propBegin( STEPS, TYPE_INTEGER, "Number of iteration steps performed so far.", false );
      t.propEnd();
      t.propBegin( TOTAL_RESIDUAL, TYPE_DOUBLE, "Sum of all residual entries = solution accuracy", false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
