package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Image functions for sampling tests.
 *
 * @version 0.26 $Rev: 421 $ $Date: 2007-04-21 01:55:47 +0200 (so, 21 IV 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/TestImageFunctions.java">TestImageFunctions.java</a>
 */
public class TestImageFunctions extends DefaultImageFunction
{
  //==========================================================================
  //  Data:

  /** Implemented function types. */
  protected static enum FunctionType
  {
    /** Infinite checker (perpendicular). */
    FUNC_CHECKER_0,
    /** Infinite checker (45 degrees). */
    FUNC_CHECKER_45,
    /** Cos(r<sup>2</sup>). */
    FUNC_COS_R2,
    /** Cos(r<sup>3</sup>). */
    FUNC_COS_R3,
    /** Cos(r<sup>4</sup>). */
    FUNC_COS_R4,
  }

  /** Actual function type. */
  protected FunctionType type = FunctionType.FUNC_CHECKER_0;

  /** Foreground color. */
  protected double[] foreground = { 1.0, 1.0, 1.0 };

  /** Image function space frequency multiplier. */
  protected double frequency = 2000.0;

  /** Horizontal center point. */
  protected double center = 0.5;

  /** Horizontal multiplier. */
  protected double mul = 1.0;

  //--------------------------------------------------------------------------
  //  Support routines:

  //==========================================================================
  //  Interface ImageFunction:

  /**
   * Sets bounds (scaling factors) for the image function (e.g. 2D mapping).
   *
   * @param  Xmin Minimum X coordinate.
   * @param  Xmax Maximum X coordinate.
   * @param  Ymin Minimum Y coordinate.
   * @param  Ymax Maximum Y coordinate.
   * @see    #getBounds
   */
  @Override
  public void setBounds ( double Xmin, double Xmax, double Ymin, double Ymax )
  {
    super.setBounds( Xmin, Xmax, Ymin, Ymax );

    center = 0.5 * (xMin + xMax);
    mul    = frequency / (xMax - xMin);
  }

  /**
   * Computes one image sample. Passes parameters for hidden sampling.
   *
   * @param  x X coordinate of the sample.
   * @param  y Y coordinate of the sample.
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (sample signature) used for adaptive subsampling.
   * @see    #getSample(double,double,double[])
   * @see    #getBounds
   */
  @Override
  public long getSample ( double x, double y, int order, int total, double[] color )
  {
      // background color:
    if ( color == null ) return -1L;

    y -= yMin;
    x -= xMin;
    long ord = 0;

    switch ( type )
    {
      case FUNC_CHECKER_0:
        ord = (y < Geometry.EPSILON) ? 0 :
                Math.round( mul * (x - center) / y ) +
                Math.round( mul * 600.0 / y );
        break;

      case FUNC_CHECKER_45:
      {
        double u = 0.0, v = 0.0;
        if ( y >= Geometry.EPSILON )
        {
          u   = mul * (x - center) / y;
          v   = mul * 600.0 / y;
          ord = Math.round( u + v ) + Math.round( u - v );
        }
        break;
      }

      case FUNC_COS_R2:
        x  *= mul;
        y  *= mul;
        ord = Math.round( x * x + y * y );
        break;

      case FUNC_COS_R3:
        x  *= mul;
        y  *= mul;
        x   = x * x + y * y;
        ord = Math.round( x * Math.sqrt( x ) );
        break;

      case FUNC_COS_R4:
        x  *= mul;
        y  *= mul;
        x   = x * x + y * y;
        ord = Math.round( x * x );
        break;
    }

    double[] c = ((ord & 1L) == 0L) ? background : foreground;
    System.arraycopy( c, 0, color, 0, Math.min( c.length, color.length ) );

    return order;
  }

  //==========================================================================
  //  Interface Property:

  public final static String FUNCTION  = "Function";
  public final static String COLOR     = "Color";
  public final static String FREQUENCY = "Frequency";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // FUNCTION:
    if ( key.compareTo(FUNCTION) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
      // FREQUENCY:
    if ( key.compareTo(FREQUENCY) == 0 )
    {
      frequency = doubleProperty( value, frequency );
      return;
    }
      // BACKGROUND_COLOR:
    if ( key.compareTo(BACKGROUND_COLOR) == 0 &&
         value instanceof double[] )
    {
      background = (double[])value;
      return;
    }
      // COLOR:
    if ( key.compareTo(COLOR) == 0 &&
         value instanceof double[] )
    {
      foreground = (double[])value;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(FUNCTION) == 0 )
      return type.ordinal();
    if ( key.compareTo(FREQUENCY) == 0 )
      return frequency;
    if ( key.compareTo(BACKGROUND_COLOR) == 0 )
      return background;
    if ( key.compareTo(COLOR) == 0 )
      return foreground;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Test image functions";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ImageFunction";

  /** Object description. */
  private final static String DESCRIPTION = "Analytic image functions for sampling tests.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_FUNCTION;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"ImageFunction" );
        // Property:
      t.propBegin( FUNCTION, TYPE_INTEGER, "Function type", true );
        t.propDefault( FunctionType.FUNC_CHECKER_0.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Checker 0 deg",  FunctionType.FUNC_CHECKER_0.ordinal(),  "Infinite checker (perpendicular)" );
        t.propEnum( "Checker 45 deg", FunctionType.FUNC_CHECKER_45.ordinal(), "Infinite checker (45 degrees)" );
        t.propEnum( "Cos(r^2)",       FunctionType.FUNC_COS_R2.ordinal(),     "cos(r^2)" );
        t.propEnum( "Cos(r^3)",       FunctionType.FUNC_COS_R3.ordinal(),     "cos(r^3)" );
        t.propEnum( "Cos(r^4)",       FunctionType.FUNC_COS_R4.ordinal(),     "cos(r^4)" );
      t.propEnd();
      t.propBegin( FREQUENCY, TYPE_DOUBLE, "Frequency multiplier", true );
        t.propDefault( 2000.0 );
      t.propEnd();
      t.propBegin( BACKGROUND_COLOR, TYPE_OBJECT, "Background color", false );
      t.propEnd();
      t.propBegin( COLOR, TYPE_OBJECT, "Foreground color", false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
