package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Test image generator - test picture for (laser) printers.
 *
 * @version 0.26 $Rev: 459 $ $Date: 2007-05-21 00:06:41 +0200 (po, 21 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/TestImageLaser.java">TestImageLaser.java</a>
 */
public class TestImageLaser extends DefaultImageSynthesizer
{
  //==========================================================================
  //  Data:

  /** Actual number of rows/columns. */
  protected int tableSize = 8;

  /** Number of dots per one table cell. */
  protected int dots      = 100;

  /** Accelerator: output image. */
  protected RasterGraphics out;

  /** External module: pen for bitmask-filling. */
  protected SolidColorPen pen;

  /** External module: bit-mask. */
  protected XTransitionList bitMask;

  /** External module: text drawing. */
  protected JavaText text;

  //==========================================================================
  //  Support:

  /**
   * Assert support modules and accelerators.
   *
   * @return {@code true} if succeeded.
   */
  protected boolean assertModules ()
  {
      // raster output:
    out = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    if ( out == null ) return false;

    if ( out.getWidth()  != imageWidth ||
         out.getHeight() != imageHeight )
      out.init( imageWidth, imageHeight, RasterGraphics.MODE_GRAY, 0 );

      // drawing support:
    if ( bitMask == null )
    {
      bitMask = new XTransitionList();
      pen     = new SolidColorPen();
      text    = new JavaText();
      try
      {
        text.connect( Piece.PL_OUTPUT,  bitMask,    Piece.PL_INPUT );
        pen.connect(  Piece.PL_BITMASK, bitMask,    Piece.PL_INPUT );
        pen.connect(  Piece.PL_OUTPUT,  (Piece)out, Piece.PL_INPUT );
      }
      catch ( BadInterfaceException e )
      {
        LogFile.exception( "Error connecting pen, text, bit-mask and raster image!", e );
        return false;
      }
    }

    bitMask.init( imageWidth, imageHeight );

    return true;
  }

  //==========================================================================
  //  Interface ImageSynthesizer:

  /**
   * Render the given rectangle.
   * Output is written to some connected raster device (not defined in this interface).
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @see    #renderPixel
   * @see    #renderAll
   */
  public void renderRectangle ( int x1, int y1, int x2, int y2 )
  {
    if ( !assertModules() ) return;       // failure

    int unit  = Math.min( Math.abs( x2 - x1 ), Math.abs( y2 - y1 ) );
    unit     /= (tableSize + 2);
    int table = unit * tableSize;
    int x0    = (x1 + x2 - table + unit) / 2;
    int y0    = (y1 + y2 - table + unit) / 2;
    int line  = Math.max( 2, unit / 36 );
    int unitH = unit / 2;
    int unitT = unit / 3;

    out.setRectangle( x1, y1, x2, y2, 1.0 ); // white background
    int x, y, i, j, k;
    int wid, hei;
    double uni;

      // randomly placed rectangular dots:
    RandomJames rnd = new RandomJames();
    for ( i = dots * tableSize * tableSize; i-- > 0; )
    {
      if ( userBreak ) return;

      uni  = rnd.uniformNumber();
      uni *= uni;
      x    = Formula.round( table * uni );
      uni  = rnd.uniformNumber();
      uni *= uni;
      y    = Formula.round( table * uni );
      wid  = x / unit + 1;
      if ( x + wid > table ) wid = table - x;
      hei  = y / unit + 1;
      if ( y + hei > table ) hei = table - y;
      x   += x0;
      y   += y0;
      out.setRectangle( x, y, x + wid, y + hei, 0.0 );
    }

      // black frame & titles:
    TextRender.TextExtent ext = new TextRender.TextExtent();
    text.createFont( "Arial", 0, 0.5 * unit );
    String val;
    for ( k = 0, x = x0; k++ < tableSize; x += unit )
    {
      val = String.valueOf( k );
      text.textExtent( 0.0, 0.0, val, ext );
      text.drawText( x + unitH - 0.5 * (ext.ulx + ext.lrx),
                     y0 - unitT,
                     val, null );
    }

    for ( y = y0; y < y0 + line; y++ )
      out.setHLine( x0, x0 + table + line, y, 0.0 );
    for ( i = 1; i <= tableSize && !userBreak; i++ )
    {
      for ( j = line; j++ < unit; y++ )
      {
        for ( k = 0, x = x0; k++ < tableSize; x += unit )
          out.setHLine( x, x + line, y, 0.0 );
        out.setHLine( x0 + table, x0 + table + line, y, 0.0 );
      }
      val = String.valueOf( i );
      text.textExtent( 0.0, 0.0, val, ext );
      text.drawText( x0 - unitT - ext.lrx,
                     y - unitH - 0.5 * (ext.uly + ext.lry),
                     val, null );
      for ( j = 0; j++ < line; y++ )
        out.setHLine( x0, x0 + table + line, y, 0.0 );
    }

    pen.setColor( 0.0 );
    pen.stroke();

  }

  //==========================================================================
  //  Interface Property:

  public final static String TABLE_SIZE = "Table size";
  public final static String DOTS       = "Dots";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // TABLE_SIZE:
    if ( key.compareTo(TABLE_SIZE) == 0 )
    {
      tableSize = intProperty( value, tableSize, 1, 100 );
      return;
    }
      // DOTS:
    if ( key.compareTo(DOTS) == 0 )
    {
      dots = intProperty( value, dots, 1, 10000 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(TABLE_SIZE) == 0 )
      return tableSize;
    if ( key.compareTo(DOTS) == 0 )
      return dots;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "TestLaser";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ImageSynthesizerToRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Test image generator - printer test.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_IMAGE;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"ImageSynthesizer" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( TABLE_SIZE, TYPE_INTEGER, "Number of table rows = columns", true );
        t.propDefault( 8 );
        t.propBounds(  1, 100 );
      t.propEnd();
      t.propBegin( DOTS, TYPE_INTEGER, "Number of dots per table cell", true );
        t.propDefault( 100 );
        t.propBounds(  1, 10000 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
