package cz.cuni.jagrlib.piece;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Image functions for 2D textures.
 *
 * @version 0.26 $Rev: 444 $ $Date: 2007-05-05 22:52:07 +0200 (so, 05 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/TestImageFunctions.java">TestImageFunctions.java</a>
 */
public class TextureFunctions extends DefaultImageFunction
{
  //==========================================================================
  //  Data:

  /** Implemented texture types. */
  protected static enum TextureType
  {
    /** Bi-color checker. */
    CHECKER_BI,
    /** Multi-color checker. */
    CHECKER_COLOR,
    /** Stripes. */
    STRIPES,
    /** Polka dots. */
    POLKA_DOTS,
  }

  /** Actual texture type. */
  protected TextureType type = TextureType.CHECKER_BI;

  /** Foreground color. */
  protected double[] foreground = { 1.0, 1.0, 1.0 };

  /** X-frequency multiplier. */
  protected double freqX = 10.0;

  /** Y-frequency multiplier. */
  protected double freqY = 10.0;

  /** {@code cotg( PI / 6 )} */
  protected static final double SQRT_3 = Math.sqrt( 3.0 );

  protected static final double POLKA_POINT = 0.03;

  //==========================================================================
  //  Interface ImageFunction:

  /**
   * Computes one image sample. Passes parameters for hidden sampling.
   *
   * @param  x X coordinate of the sample.
   * @param  y Y coordinate of the sample.
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (sample signature) used for adaptive subsampling.
   * @see    #getSample(double,double,double[])
   * @see    #getBounds
   */
  @Override
  public long getSample ( double x, double y, int order, int total, double[] color )
  {
      // background color:
    if ( color == null ) return -1L;

    x = Formula.remainder( (x - xMin) / (xMax - xMin), 1.0 );
    y = Formula.remainder( (y - yMin) / (yMax - yMin), 1.0 );

    double[] c = null;
    long result = 0L;

    switch ( type )
    {
      case CHECKER_BI:
      {
        final long xi = (long)( x * freqX );
        final long yi = (long)( y * freqY );
        c = (((xi + yi) & 1L) == 0L) ? background : foreground;
        result = xi + RandomStatic.numericRecipes( yi );
        break;
      }

      case CHECKER_COLOR:
      {
        final long xi = (long)( x * freqX );
        final long yi = (long)( y * freqY );
        final boolean dark = ((xi + yi) & 1L) == 0L;
        long rnd1 = RandomStatic.numericRecipes( 673929 + xi * 16223 );
             rnd1 = RandomStatic.numericRecipes( rnd1 );
             rnd1 = RandomStatic.numericRecipes( rnd1 );
        long rnd2 = RandomStatic.ibmRandu( 196533 + yi * 64937 );
             rnd2 = RandomStatic.ibmRandu( rnd2 );
             rnd2 = RandomStatic.ibmRandu( rnd2 );

        if ( dark )
        {
          color[0] = (rnd1 & 1023L) / 2046.0;
          if ( color.length > 1 )
          {
            color[1] = (rnd2 & 1023L) / 2046.0;
            color[2] = ((rnd1 >> 6) & 1023L) / 2046.0;
          }
        }
        else
        {
          color[0] = 0.5 + (rnd1 & 1023L) / 2046.0;
          if ( color.length > 1 )
          {
            color[1] = 0.5 + (rnd2 & 1023L) / 2046.0;
            color[2] = ((rnd1 >> 6) & 1023L) / 1023.0;
          }
        }

        return( rnd1 ^ rnd2 );
      }

      case STRIPES:
      {
        final long ord = (long)( x * freqX + y * freqY );
        c = ((ord & 1L) == 0L) ? background : foreground;
        result = RandomStatic.numericRecipes( ord );
        break;
      }

      case POLKA_DOTS:
      {
        x *= freqX * SQRT_3;
        y *= freqY;
        result = RandomStatic.numericRecipes( (long)x + 13L * (long)y );
        x -= Math.floor( x );
        y -= Math.floor( y );
        if ( x > 0.5 ) x = 1.0 - x;
        if ( y > 0.5 ) y = 1.0 - y;
        double sqrD = y * y + x * x / 3.0;
        x = 0.5 - x;
        y = 0.5 - y;
        double sqrE = y * y + x * x / 3.0;
        if ( sqrE < sqrD ) sqrD = sqrE;

        c = (sqrD > POLKA_POINT) ? background : foreground;
        break;
      }
    }

    System.arraycopy( c, 0, color, 0, Math.min( c.length, color.length ) );
    return result;
  }

  //==========================================================================
  //  Interface Property:

  public final static String TEXTURE   = "Texture";
  public final static String COLOR     = "Color";
  public final static String FREQ_X    = "Frequency X";
  public final static String FREQ_Y    = "Frequency Y";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // TEXTURE:
    if ( key.compareTo(TEXTURE) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
      // FREQ_X:
    if ( key.compareTo(FREQ_X) == 0 )
    {
      freqX = doubleProperty( value, freqX );
      return;
    }
      // FREQ_Y:
    if ( key.compareTo(FREQ_Y) == 0 )
    {
      freqY = doubleProperty( value, freqY );
      return;
    }
      // BACKGROUND_COLOR:
    if ( key.compareTo(BACKGROUND_COLOR) == 0 )
    {
      background = colorDoubleProperty( value, background );
      return;
    }
      // COLOR:
    if ( key.compareTo(COLOR) == 0 )
    {
      foreground = colorDoubleProperty( value, foreground );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(TEXTURE) == 0 )
      return type.ordinal();
    if ( key.compareTo(FREQ_X) == 0 )
      return freqX;
    if ( key.compareTo(FREQ_Y) == 0 )
      return freqY;
    if ( key.compareTo(BACKGROUND_COLOR) == 0 )
      return background;
    if ( key.compareTo(COLOR) == 0 )
      return foreground;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Texture functions";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ImageFunction";

  /** Object description. */
  private final static String DESCRIPTION = "Analytic 2D texture functions.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_FUNCTION;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"ImageFunction" );
        // Property:
      t.propBegin( TEXTURE, TYPE_INTEGER, "Texture type", true );
        t.propDefault( TextureType.CHECKER_BI.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Checker bi-color",    TextureType.CHECKER_BI.ordinal(),    "Bi-color checker" );
        t.propEnum( "Checker multi-color", TextureType.CHECKER_COLOR.ordinal(), "Multi-color checker" );
        t.propEnum( "Stripes",             TextureType.STRIPES.ordinal(),       "Stripes" );
        t.propEnum( "Polka dots",          TextureType.POLKA_DOTS.ordinal(),    "Polka dots" );
      t.propEnd();
      t.propBegin( FREQ_X, TYPE_DOUBLE, "X-frequency multiplier", true );
        t.propDefault( 10.0 );
      t.propEnd();
      t.propBegin( FREQ_Y, TYPE_DOUBLE, "Y-frequency multiplier", true );
        t.propDefault( 10.0 );
      t.propEnd();
      t.propBegin( BACKGROUND_COLOR, TYPE_OBJECT, "Background color", true );
        t.propDefault( "[0;0;0]" );
        t.propManipulator( "ParamColor" );
      t.propEnd();
      t.propBegin( COLOR, TYPE_OBJECT, "Foreground color", true );
        t.propDefault( "[255;255;255]" );
        t.propManipulator( "ParamColor" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
