package cz.cuni.jagrlib.piece;

import java.lang.reflect.*;
import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Boundary representation (B-rep) of 3D scene - implemented by simple arrays. Restricted
 * mode uses triangle faces only, general mode could contain arbitrary simple polygons as
 * faces.
 *
 * <p>Vertices, edges, faces, divisions and solids (Vs, Es, Fs, Ds, Ss) are represented by
 * their <b>handles</b> (integer indices). Arbitrary <b>attributes</b> can be associated
 * with Vs, Es, Fs, Ds or Ss by the user. Implicit attribute: vertex coordinates.
 * Attributes are accessed by string identifiers or integer shortcuts.
 * </p>
 *
 * <p>More than one <b>context</b> can be used (e.g. for different scene views, rendering
 * modes). Handles (indices) are shared among all contexts, some attributes can be shared,
 * too.
 * </p>
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.11
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/VEFDS.java">VEFDS.java</a>
 */
public class VEFDS extends Piece implements Brep
{
  /**
   * Internal attribute representation.
   *
   * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
   * @since   0.11
   */
  protected class Attribute
  {
    /** Attribute's handle (inverse mapping for {@link #attributes}). */
    public int handle;

    /** Entity which the attribute is associated with. */
    public int where;

    /** Handle of the context ({@link #NULL NULL} for shared attributes). */
    public int context;

    /** Identifier (w/o the context-suffix). */
    public String ident;

    /**
     * Attribute type ({@code ATTR_*}).
     *
     * @see    #ATTR_OBJECT
     */
    public int type;

    /**
     * Size of one data-array item. Allocation: {@code size * itemSize}.
     *
     * @see    #size
     * @see    #data
     * @see    #assertDataSize
     */
    public int itemSize;

    /**
     * Actual size of the data[-array] in items ({@code -1} for scalar attributes).
     * Allocation: {@code size * itemSize}.
     *
     * @see    #itemSize
     * @see    #data
     * @see    #assertDataSize
     */
    public int size;

    /**
     * The data[-array] itself. Allocation: {@code size * itemSize}.
     *
     * @see    #itemSize
     * @see    #size
     * @see    #assertDataSize
     */
    public Object data;

    /**
     * Returns a hash code value for the object.
     *
     * @return A hash code value for this object.
     */
    public int hashCode ()
    {
      return( ident.hashCode() * 13 + context );
    }

    /**
     * Initializing constructor - allocates data object.
     *
     * @param  _where Entity which the attribute is associated with.
     * @param  _context Handle of the context ({@link #NULL NULL} for shared attributes).
     * @param  _ident Identifier (w/o the context-suffix).
     * @param  _type Attribute type.
     * @param  _size Array length ({@code 0} for no allocation, {@code -1} for
     *               scalar value).
     * @see    #VEFDS.Attribute(VEFDS.Attribute,int,String)
     * @see    #assertDataSize
     */
    public Attribute ( int _where, int _context, String _ident, int _type, int _size )
    {
      handle  = NULL;
      where   = _where;
      context = _context;
      ident   = _ident;
      type    = _type;

        // itemSize:
      switch ( type )
      {
        case ATTR_COORD_2D:
        case ATTR_COORD_2F:
        case ATTR_COORD_2I:
        case ATTR_DIVISION:
          itemSize = 2;
          break;

        case ATTR_COORD_3D:
        case ATTR_COORD_3F:
        case ATTR_COORD_3I:
        case ATTR_COORDH_3D:
        case ATTR_RGB_3F:
        case ATTR_TRIANGLE:
          itemSize = 3;
          break;

        case ATTR_COORDH_4D:
        case ATTR_RGBA_4F:
          itemSize = 4;
          break;

        case ATTR_EDGE:
          itemSize = edgeSize;
          break;

        default:
          //case ATTR_OBJECT:
          //case ATTR_BOOLEAN:
          //case ATTR_INTEGER:
          //case ATTR_LONG:
          //case ATTR_FLOAT:
          //case ATTR_DOUBLE:
          //case ATTR_FACE:
          //case ATTR_SOLID:
          itemSize = 1;
      }

        // data[-array] itself:
      data = null;
      size = 0;
      assertDataSize( _size );
    }

    /**
     * Initializing constructor - copies data from the given {@code Attribute} instance.
     *
     * @param  _from Instance data shall be copied from.
     * @param  _context Handle of the context ({@link #NULL NULL} for shared attributes).
     * @param  _ident Identifier (w/o the context-suffix).
     * @see    #VEFDS.Attribute(int,int,String,int,int)
     * @see    #assertDataSize
     */
    public Attribute ( Attribute _from, int _context, String _ident )
    {
      handle   = NULL;
      where    = _from.where;
      context  = _context;
      ident    = _ident;
      type     = _from.type;
      itemSize = _from.itemSize;
      data     = null;
      size     = 0;
        // copy the data:
      deepCopy( _from );
    }

    /**
     * Allocates data[-array] of the given size. Grows the array if needed,
     * {@code _size = 0} will de-allocate the data[-array].
     *
     * @param  _size Required data size ({@code 0} will cause array de-allocation).
     * @see    #size
     * @see    #itemSize
     * @see    #data
     */
    protected void assertDataSize ( int _size )
    {
      if ( _size == size ) return;    // nothing has to be changed

        // empty data[-array]:
      if ( _size == 0 )
      {
        size = 0;
        data = null;
        return;
      }

        // non-empty data[-array] (either scalar or vector):
      Object newData = null;
      int oldSize = size;
      size = _size;
      final int arrSize = size * itemSize;

      switch ( type )
      {
        case ATTR_OBJECT:
          if ( size > 0 )
            newData = new Object[ arrSize ];
          else
            return;
          break;

        case ATTR_BOOLEAN:
          if ( size > 0 )
            newData = new boolean[ arrSize ];
          else
          {
            if ( itemSize == 1 )
              data = false;
            else
              data = new boolean[ itemSize ];
            return;
          }
          break;

        case ATTR_INTEGER:
        case ATTR_COORD_2I:
        case ATTR_COORD_3I:
        case ATTR_EDGE:
        case ATTR_TRIANGLE:
          if ( size > 0 )
            newData = new int[ arrSize ];
          else
          {
            if ( itemSize == 1 )
              data = 0;
            else
              data = new int[ itemSize ];
            return;
          }
          break;

        case ATTR_LONG:
          if ( size > 0 )
            newData = new long[ arrSize ];
          else
          {
            if ( itemSize == 1 )
              data = 0L;
            else
              data = new long[ itemSize ];
            return;
          }
          break;

        case ATTR_DOUBLE:
        case ATTR_COORD_2D:
        case ATTR_COORDH_3D:
        case ATTR_COORD_3D:
        case ATTR_COORDH_4D:
          if ( size > 0 )
            newData = new double[ arrSize ];
          else
          {
            if ( itemSize == 1 )
              data = 0.0;
            else
              data = new double[ itemSize ];
            return;
          }
          break;

        case ATTR_FLOAT:
        case ATTR_COORD_2F:
        case ATTR_COORDH_3F:
        case ATTR_COORD_3F:
        case ATTR_COORDH_4F:
        case ATTR_RGB_3F:
        case ATTR_RGBA_4F:
          if ( size > 0 )
            newData = new float[ arrSize ];
          else
          {
            if ( itemSize == 1 )
              data = 0.0f;
            else
              data = new float[ itemSize ];
            return;
          }
          break;

        case ATTR_FACE:
        case ATTR_SOLID:
        case ATTR_DIVISION:
          if ( size > 0 )
            newData = new int[ arrSize ][];
          else
          {
            data = new int[ itemSize ][];
            return;
          }
          break;
        }

        // preserve the old data:
      if ( oldSize > 0 )
      {
        if ( size < oldSize ) oldSize = size;
        System.arraycopy( data, 0, newData, 0, oldSize * itemSize );
      }
      data = newData;
    }

    /**
     * Deep-copy of attribute data from another {@code Attribute} instance.
     *
     * @param  from The Attribute which data will be copied from.
     */
    public void deepCopy ( Attribute from )
    {
      if ( from == null || from.where != where ||
           from.type != type || from.itemSize != itemSize ) return;

      assertDataSize( from.size );
      if ( size == 0 ) return;

        // copy the data:
      if ( from.data != null && from.data.getClass().isArray() )
      {
        int len = Array.getLength( from.data );
        System.arraycopy( from.data, 0, data, 0, len ); // shallow-copy
        switch ( type )
        {
          case ATTR_FACE:
          case ATTR_SOLID:
          case ATTR_DIVISION:
            int[] oldArr;
            int[] newArr;
            for ( int i = 0; i < len; i++ )
               if ( (oldArr = (int[])Array.get( data, i )) != null )
               {
                   // copy the sub-array
                 int l = oldArr.length;
                 newArr = new int[ l ];
                 Array.set( data, i, newArr );
                 System.arraycopy( oldArr, 0, newArr, 0, l );
               }
             break;
        }
      }
      else
        data = from.data;
    }

    /**
     * Copies coordinate data from the {@code Attribute} object to another one.
     * Performs rounding and/or homo =&gt; Euclid conversions.
     *
     * @param  from Source {@code Attribute} object.
     * @param  set Bit-set of handles to be copied ({@code null} for
     *             "copy all").
     * @see    #roundCoordinates
     * @see    #transformCoordinates
     */
    public void copyCoordinates ( Attribute from, BitSet set )
    {
      int s = (size < from.size) ? size : from.size;
      int i, ptr;
      if ( from.type == type )
      {
        if ( set == null )
        {
          System.arraycopy( from.data, 0, data, 0, s * itemSize );
          return;
        }

        for ( i = ptr = 0; i < s; i++, ptr += itemSize )
          if ( set.get( i ) )
            System.arraycopy( from.data, ptr, data, ptr, itemSize );

        return;
      }

      int      ptrFrom;
      double   w;
      int[]    ia;
      double[] da;
      float    wf;
      float[]  fa;

      switch ( type )
      {
        case ATTR_COORD_2D:
        {                                   // destination: [ double x, double y ]
          double[] to = (double[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                }
              return;

            case ATTR_COORDH_3D:
            case ATTR_COORDH_4D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  w = da[ ptrFrom + from.itemSize - 1 ];
                  if ( Geometry.isZero( w ) ) w = 1.0;
                  to[ptr]   = da[ptrFrom]   / w;
                  to[ptr+1] = da[ptrFrom+1] / w;
                }
              return;

            case ATTR_COORDH_3F:
            case ATTR_COORDH_4F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  wf = fa[ ptrFrom + from.itemSize - 1 ];
                  if ( Geometry.isZero( wf ) ) wf = 1.0f;
                  to[ptr]   = fa[ptrFrom]   / wf;
                  to[ptr+1] = fa[ptrFrom+1] / wf;
                }
              return;

            case ATTR_COORD_2I:
            case ATTR_COORD_3I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                }
              return;

            case ATTR_COORD_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = da[ptrFrom];
                  to[ptr+1] = da[ptrFrom+1];
                }
              return;

            case ATTR_COORD_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                }
              return;
          }
        }

        case ATTR_COORD_2F:
        {                                   // destination: [ float x, float y ]
          float[] to = (float[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = (float)da[ptrFrom];
                  to[ptr+1] = (float)da[ptrFrom+1];
                }
              return;

            case ATTR_COORDH_3D:
            case ATTR_COORDH_4D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  w = da[ ptrFrom + from.itemSize - 1 ];
                  if ( Geometry.isZero( w ) ) w = 1.0;
                  to[ptr]   = (float)(da[ptrFrom]   / w);
                  to[ptr+1] = (float)(da[ptrFrom+1] / w);
                }
              return;

            case ATTR_COORDH_3F:
            case ATTR_COORDH_4F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  wf = fa[ ptrFrom + from.itemSize - 1 ];
                  if ( Geometry.isZero( wf ) ) wf = 1.0f;
                  to[ptr]   = fa[ptrFrom]   / wf;
                  to[ptr+1] = fa[ptrFrom+1] / wf;
                }
              return;

            case ATTR_COORD_2I:
            case ATTR_COORD_3I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                }
              return;

            case ATTR_COORD_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = (float)da[ptrFrom];
                  to[ptr+1] = (float)da[ptrFrom+1];
                }
              return;

            case ATTR_COORD_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                }
              return;
          }
        }

        case ATTR_COORDH_3D:
        {                                   // destination: [ double x, double y, double w ]
          double[] to = (double[])data;
          switch ( from.type )
          {
            case ATTR_COORDH_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = fa[ptrFrom+2];
                }
              return;

            case ATTR_COORDH_4D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = da[ptrFrom];
                  to[ptr+1] = da[ptrFrom+1];
                  to[ptr+2] = da[ptrFrom+3];
                }
              return;

            case ATTR_COORDH_4F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = fa[ptrFrom+3];
                }
              return;

            case ATTR_COORD_2I:
            case ATTR_COORD_3I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                  to[ptr+2] = 1.0;
                }
              return;

            case ATTR_COORD_2D:
            case ATTR_COORD_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = da[ptrFrom];
                  to[ptr+1] = da[ptrFrom+1];
                  to[ptr+2] = 1.0;
                }
              return;

            case ATTR_COORD_2F:
            case ATTR_COORD_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = 1.0;
                }
              return;
          }
        }

        case ATTR_COORDH_3F:
        {                                   // destination: [ float x, float y, float w ]
          float[] to = (float[])data;
          switch ( from.type )
          {
            case ATTR_COORDH_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = (float)da[ptrFrom];
                  to[ptr+1] = (float)da[ptrFrom+1];
                  to[ptr+2] = (float)da[ptrFrom+2];
                }
              return;

            case ATTR_COORDH_4D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = (float)da[ptrFrom];
                  to[ptr+1] = (float)da[ptrFrom+1];
                  to[ptr+2] = (float)da[ptrFrom+3];
                }
              return;

            case ATTR_COORDH_4F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = fa[ptrFrom+3];
                }
              return;

            case ATTR_COORD_2I:
            case ATTR_COORD_3I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                  to[ptr+2] = 1.0f;
                }
              return;

            case ATTR_COORD_2D:
            case ATTR_COORD_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = (float)da[ptrFrom];
                  to[ptr+1] = (float)da[ptrFrom+1];
                  to[ptr+2] = 1.0f;
                }
              return;

            case ATTR_COORD_2F:
            case ATTR_COORD_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = 1.0f;
                }
              return;
          }
        }

        case ATTR_COORD_2I:
        {                                   // destination: [ int x, int y ]
          int[] to = (int[])data;
          switch ( from.type )
          {
            case ATTR_COORDH_3D:
            case ATTR_COORDH_4D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  w = da[ ptrFrom + from.itemSize - 1 ];
                  if ( Geometry.isZero( w ) ) w = 1.0;
                  to[ptr]   = Formula.round( da[ptrFrom]   / w );
                  to[ptr+1] = Formula.round( da[ptrFrom+1] / w );
                }
              return;

            case ATTR_COORDH_3F:
            case ATTR_COORDH_4F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  wf = fa[ ptrFrom + from.itemSize - 1 ];
                  if ( Geometry.isZero( wf ) ) wf = 1.0f;
                  to[ptr]   = Math.round( fa[ptrFrom]   / wf );
                  to[ptr+1] = Math.round( fa[ptrFrom+1] / wf );
                }
              return;

            case ATTR_COORD_3I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                }
              return;

            case ATTR_COORD_2D:
            case ATTR_COORD_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = Formula.round( da[ptrFrom]   );
                  to[ptr+1] = Formula.round( da[ptrFrom+1] );
                }
              return;

            case ATTR_COORD_2F:
            case ATTR_COORD_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += from.itemSize )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = Math.round( fa[ptrFrom]   );
                  to[ptr+1] = Math.round( fa[ptrFrom+1] );
                }
              return;
          }
        }

        case ATTR_COORD_3D:
        {                                   // destination: [ double x, double y, double z ]
          double[] to = (double[])data;
          switch ( from.type )
          {
            case ATTR_COORD_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+1] = fa[ptrFrom+2];
                }
              return;

            case ATTR_COORDH_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  w = da[ ptrFrom + 2 ];
                  if ( Geometry.isZero( w ) ) w = 1.0;
                  to[ptr]   = da[ptrFrom]   / w;
                  to[ptr+1] = da[ptrFrom+1] / w;
                  to[ptr+2] = 0.0;
                }
              return;

            case ATTR_COORDH_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  wf = fa[ ptrFrom + 2 ];
                  if ( Geometry.isZero( wf ) ) wf = 1.0f;
                  to[ptr]   = fa[ptrFrom]   / wf;
                  to[ptr+1] = fa[ptrFrom+1] / wf;
                  to[ptr+2] = 0.0;
                }
              return;

            case ATTR_COORDH_4D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  w = da[ ptrFrom + 3 ];
                  if ( Geometry.isZero( w ) ) w = 1.0;
                  to[ptr]   = da[ptrFrom]   / w;
                  to[ptr+1] = da[ptrFrom+1] / w;
                  to[ptr+2] = da[ptrFrom+2] / w;
                }
              return;

            case ATTR_COORDH_4F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  wf = fa[ ptrFrom + 3 ];
                  if ( Geometry.isZero( wf ) ) wf = 1.0f;
                  to[ptr]   = fa[ptrFrom]   / wf;
                  to[ptr+1] = fa[ptrFrom+1] / wf;
                  to[ptr+2] = fa[ptrFrom+2] / wf;
                }
              return;

            case ATTR_COORD_2I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                  to[ptr+2] = 0.0;
                }
              return;

            case ATTR_COORD_3I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                  to[ptr+1] = ia[ptrFrom+2];
                }
              return;

            case ATTR_COORD_2D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = da[ptrFrom];
                  to[ptr+1] = da[ptrFrom+1];
                  to[ptr+2] = 0.0;
                }
              return;

            case ATTR_COORD_2F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = 0.0;
                }
              return;
          }
        }

        case ATTR_COORD_3F:
        {                                   // destination: [ float x, float y, float z ]
          float[] to = (float[])data;
          switch ( from.type )
          {
            case ATTR_COORD_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = (float)da[ptrFrom];
                  to[ptr+1] = (float)da[ptrFrom+1];
                  to[ptr+1] = (float)da[ptrFrom+2];
                }
              return;

            case ATTR_COORDH_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  w = da[ ptrFrom + 2 ];
                  if ( Geometry.isZero( w ) ) w = 1.0;
                  to[ptr]   = (float)(da[ptrFrom]   / w);
                  to[ptr+1] = (float)(da[ptrFrom+1] / w);
                  to[ptr+2] = 0.0f;
                }
              return;

            case ATTR_COORDH_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  wf = fa[ ptrFrom + 2 ];
                  if ( Geometry.isZero( wf ) ) wf = 1.0f;
                  to[ptr]   = fa[ptrFrom]   / wf;
                  to[ptr+1] = fa[ptrFrom+1] / wf;
                  to[ptr+2] = 0.0f;
                }
              return;

            case ATTR_COORDH_4D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  w = da[ ptrFrom + 3 ];
                  if ( Geometry.isZero( w ) ) w = 1.0;
                  to[ptr]   = (float)(da[ptrFrom]   / w);
                  to[ptr+1] = (float)(da[ptrFrom+1] / w);
                  to[ptr+2] = (float)(da[ptrFrom+2] / w);
                }
              return;

            case ATTR_COORDH_4F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  wf = fa[ ptrFrom + 3 ];
                  if ( Geometry.isZero( wf ) ) wf = 1.0f;
                  to[ptr]   = fa[ptrFrom]   / wf;
                  to[ptr+1] = fa[ptrFrom+1] / wf;
                  to[ptr+2] = fa[ptrFrom+2] / wf;
                }
              return;

            case ATTR_COORD_2I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                  to[ptr+2] = 0.0f;
                }
              return;

            case ATTR_COORD_3I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                  to[ptr+1] = ia[ptrFrom+2];
                }
              return;

            case ATTR_COORD_2D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = (float)da[ptrFrom];
                  to[ptr+1] = (float)da[ptrFrom+1];
                  to[ptr+2] = 0.0f;
                }
              return;

            case ATTR_COORD_2F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = 0.0f;
                }
              return;
          }
        }

        case ATTR_COORDH_4D:
        {                                   // destination: [ double x, double y, douvle, z, double w ]
          double[] to = (double[])data;
          switch ( from.type )
          {
            case ATTR_COORDH_4F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = fa[ptrFrom+2];
                  to[ptr+3] = fa[ptrFrom+3];
                }
              return;

            case ATTR_COORDH_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = da[ptrFrom];
                  to[ptr+1] = da[ptrFrom+1];
                  to[ptr+2] = 0.0;
                  to[ptr+3] = da[ptrFrom+2];
                }
              return;

            case ATTR_COORDH_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = 0.0;
                  to[ptr+3] = fa[ptrFrom+2];
                }
              return;

            case ATTR_COORD_2I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                  to[ptr+2] = 0.0;
                  to[ptr+3] = 1.0;
                }
              return;

            case ATTR_COORD_3I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                  to[ptr+2] = ia[ptrFrom+2];
                  to[ptr+3] = 1.0;
                }
              return;

            case ATTR_COORD_2D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = da[ptrFrom];
                  to[ptr+1] = da[ptrFrom+1];
                  to[ptr+2] = 0.0;
                  to[ptr+3] = 1.0;
                }
              return;

            case ATTR_COORD_2F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = 0.0;
                  to[ptr+3] = 1.0;
                }
              return;

            case ATTR_COORD_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = da[ptrFrom];
                  to[ptr+1] = da[ptrFrom+1];
                  to[ptr+2] = da[ptrFrom+2];
                  to[ptr+3] = 1.0;
                }
              return;

            case ATTR_COORD_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = fa[ptrFrom+2];
                  to[ptr+3] = 1.0;
                }
              return;
          }
        }

        case ATTR_COORDH_4F:
        {                                   // destination: [ float x, float y, float, z, float w ]
          float[] to = (float[])data;
          switch ( from.type )
          {
            case ATTR_COORDH_4D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = (float)da[ptrFrom];
                  to[ptr+1] = (float)da[ptrFrom+1];
                  to[ptr+2] = (float)da[ptrFrom+2];
                  to[ptr+3] = (float)da[ptrFrom+3];
                }
              return;

            case ATTR_COORDH_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = (float)da[ptrFrom];
                  to[ptr+1] = (float)da[ptrFrom+1];
                  to[ptr+2] = 0.0f;
                  to[ptr+3] = (float)da[ptrFrom+2];
                }
              return;

            case ATTR_COORDH_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = 0.0f;
                  to[ptr+3] = fa[ptrFrom+2];
                }
              return;

            case ATTR_COORD_2I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                  to[ptr+2] = 0.0f;
                  to[ptr+3] = 1.0f;
                }
              return;

            case ATTR_COORD_3I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                  to[ptr+2] = ia[ptrFrom+2];
                  to[ptr+3] = 1.0f;
                }
              return;

            case ATTR_COORD_2D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = (float)da[ptrFrom];
                  to[ptr+1] = (float)da[ptrFrom+1];
                  to[ptr+2] = 0.0f;
                  to[ptr+3] = 1.0f;
                }
              return;

            case ATTR_COORD_2F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = 0.0f;
                  to[ptr+3] = 1.0f;
                }
              return;

            case ATTR_COORD_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = (float)da[ptrFrom];
                  to[ptr+1] = (float)da[ptrFrom+1];
                  to[ptr+2] = (float)da[ptrFrom+2];
                  to[ptr+3] = 1.0f;
                }
              return;

            case ATTR_COORD_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = fa[ptrFrom];
                  to[ptr+1] = fa[ptrFrom+1];
                  to[ptr+2] = fa[ptrFrom+2];
                  to[ptr+3] = 1.0f;
                }
              return;
          }
        }

        case ATTR_COORD_3I:
        {                                   // destination: [ int x, int y, int z ]
          int[] to = (int[])data;
          switch ( from.type )
          {
            case ATTR_COORDH_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  w = da[ptrFrom + 2];
                  if ( Geometry.isZero( w ) ) w = 1.0;
                  to[ptr]   = Formula.round( da[ptrFrom]   / w );
                  to[ptr+1] = Formula.round( da[ptrFrom+1] / w );
                  to[ptr+2] = 0;
                }
              return;

            case ATTR_COORDH_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  wf = fa[ ptrFrom + 2 ];
                  if ( Geometry.isZero( wf ) ) wf = 1.0f;
                  to[ptr]   = Math.round( fa[ptrFrom]   / wf );
                  to[ptr+1] = Math.round( fa[ptrFrom+1] / wf );
                  to[ptr+2] = 0;
                }
              return;

            case ATTR_COORDH_4D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  w = da[ptrFrom + 3];
                  if ( Geometry.isZero( w ) ) w = 1.0;
                  to[ptr]   = Formula.round( da[ptrFrom]   / w );
                  to[ptr+1] = Formula.round( da[ptrFrom+1] / w );
                  to[ptr+2] = Formula.round( da[ptrFrom+2] / w );
                }
              return;

            case ATTR_COORDH_4F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  wf = fa[ ptrFrom + 3 ];
                  if ( Geometry.isZero( wf ) ) w = 1.0f;
                  to[ptr]   = Math.round( fa[ptrFrom]   / wf );
                  to[ptr+1] = Math.round( fa[ptrFrom+1] / wf );
                  to[ptr+2] = Math.round( fa[ptrFrom+2] / wf );
                }
              return;

            case ATTR_COORD_2I:
              ia = (int[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = ia[ptrFrom];
                  to[ptr+1] = ia[ptrFrom+1];
                  to[ptr+2] = 0;
                }
              return;

            case ATTR_COORD_2D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = Formula.round( da[ptrFrom]   );
                  to[ptr+1] = Formula.round( da[ptrFrom+1] );
                  to[ptr+2] = 0;
                }
              return;

            case ATTR_COORD_2F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = Math.round( fa[ptrFrom]   );
                  to[ptr+1] = Math.round( fa[ptrFrom+1] );
                  to[ptr+2] = 0;
                }
              return;

            case ATTR_COORD_3D:
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = Formula.round( da[ptrFrom]   );
                  to[ptr+1] = Formula.round( da[ptrFrom+1] );
                  to[ptr+2] = Formula.round( da[ptrFrom+2] );
                }
              return;

            case ATTR_COORD_3F:
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  to[ptr]   = Math.round( fa[ptrFrom]   );
                  to[ptr+1] = Math.round( fa[ptrFrom+1] );
                  to[ptr+2] = Math.round( fa[ptrFrom+2] );
                }
              return;
          }
        }
      }
    }

    /**
     * Rounds coordinate data from the {@code Attribute} object to another one.
     * Performs linear planar transformation, rounding and homo =&gt; Euclid conversions.
     *
     * @param from Source {@code Attribute} object (can be {@code this}).
     * @param tr Transformation matrix (only the planar part is used).
     * @param set Bit-set of handles to be rounded ({@code null} for "round
     *          all").
     * @see #copyCoordinates
     * @see #transformCoordinates
     */
    public void roundCoordinates ( Attribute from, TrMatrix tr, BitSet set )
    {
      // !!! TODO: special method is not written yet !!!
      transformCoordinates( from, tr, set );
    }

    /**
     * Transforms coordinate data from the {@code Attribute} object to another one.
     * Performs linear transformation, rounding and/or homo =&gt; Euclid conversions.
     *
     * @param from Source {@code Attribute} object (can be {@code this}).
     * @param tr Transformation matrix.
     * @param set Bit-set of handles to be transformed ({@code null} for
     *          "transform all").
     * @see #copyCoordinates
     * @see #roundCoordinates
     */
    public void transformCoordinates ( Attribute from, TrMatrix tr, BitSet set )
    {
      if ( tr == null )
      {
        copyCoordinates( from, set );
        return;
      }

      int s = (size < from.size) ? size : from.size;
      int i, ptr, ptrFrom;
      double[] tmp  = new double[ 4 ];
      float[]  tmpf = new float[ 4 ];
      double[] da;
      float[]  fa;

      switch ( type )
      {
        case ATTR_COORD_2D:
        {                                   // destination: [ double x, double y ]
          double[] to = (double[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2D:        // [ d x, d y ] -> [ d x, d y ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 2 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint2D( da, ptr, 2, to, ptr, 2 );
              return;

            case ATTR_COORD_2F:        // [ f x, f y ] -> [ d x, d y ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptr];
                  tmp[1] = fa[ptr+1];
                  tr.transformPoint2D( tmp, 0, 2, to, ptr, 2 );
                }
              return;

            case ATTR_COORD_3D:        // [ d x, d y, d z ] -> [ d x, d y ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptrFrom, 3, tmp, 0, 3 );
                  to[ptr]   = tmp[0];
                  to[ptr+1] = tmp[1];
                }
              return;

            case ATTR_COORD_3F:        // [ f x, f y, f z ] -> [ d x, d y ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptrFrom];
                  tmp[1] = fa[ptrFrom+1];
                  tmp[2] = fa[ptrFrom+2];
                  tr.transformPoint3D( tmp, 0, 3, tmp, 0, 3 );
                  to[ptr]   = tmp[0];
                  to[ptr+1] = tmp[1];
                }
              return;

            case ATTR_COORDH_3D:       // [ d x, d y, d w ] -> [ d x, d y ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint2D( da, ptrFrom, 3, to, ptr, 2 );
              return;

            case ATTR_COORDH_3F:       // [ f x, f y, f w ] -> [ d x, d y ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptrFrom];
                  tmp[1] = fa[ptrFrom+1];
                  tmp[2] = fa[ptrFrom+3];
                  tr.transformPoint2D( tmp, 0, 3, to, ptr, 2 );
                }
              return;

            case ATTR_COORDH_4D:       // [ d x, d y, d z, d w ] -> [ d x, d y ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptrFrom, 4, tmp, 0, 3 );
                  to[ptr]   = tmp[0];
                  to[ptr+1] = tmp[1];
                }
              return;

            case ATTR_COORDH_4F:       // [ f x, f y, f z, f w ] -> [ d x, d y ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptrFrom];
                  tmp[1] = fa[ptrFrom+1];
                  tmp[2] = fa[ptrFrom+2];
                  tmp[3] = fa[ptrFrom+3];
                  tr.transformPoint3D( tmp, 0, 4, tmp, 0, 3 );
                  to[ptr]   = tmp[0];
                  to[ptr+1] = tmp[1];
                }
              return;

            case ATTR_COORD_2I:        // [ i x, i y ] -> [ d x, d y ]
            case ATTR_COORD_3I:        // [ i x, i y, i z ] -> [ d x, d y ]
              // !!! not implemented !!!
              return;
          }
        }

        case ATTR_COORD_2F:
        {                                   // destination: [ float x, float y ]
          float[] to = (float[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2D:        // [ d x, d y ] -> [ f x, f y ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 2 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint2D( da, ptr, 2, tmp, 0, 2 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                }
              return;

            case ATTR_COORD_2F:        // [ f x, f y ] -> [ f x, f y ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 2 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint2D( fa, ptr, 2, to, ptr, 2 );
              return;

            case ATTR_COORD_3D:        // [ d x, d y, d z ] -> [ f x, f y ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptrFrom, 3, tmp, 0, 3 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                }
              return;

            case ATTR_COORD_3F:        // [ f x, f y, f z ] -> [ f x, f y ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint3D( fa, ptrFrom, 3, to, ptr, 2 );
              return;

            case ATTR_COORDH_3D:       // [ d x, d y, d w ] -> [ f x, f y ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint2D( da, ptrFrom, 3, tmp, 0, 2 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                }
              return;

            case ATTR_COORDH_3F:       // [ f x, f y, f w ] -> [ f x, f y ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint2D( fa, ptrFrom, 3, to, ptr, 2 );
              return;

            case ATTR_COORDH_4D:       // [ d x, d y, d z, d w ] -> [ f x, f y ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptrFrom, 4, tmp, 0, 3 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                }
              return;

            case ATTR_COORDH_4F:       // [ f x, f y, f z, f w ] -> [ f x, f y ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( fa, ptrFrom, 4, tmpf, 0, 3 );
                  to[ptr]   = tmpf[0];
                  to[ptr+1] = tmpf[1];
                }
              return;

            case ATTR_COORD_2I:        // [ i x, i y ] -> [ f x, f y ]
            case ATTR_COORD_3I:        // [ i x, i y, i z ] -> [ f x, f y ]
              // !!! not implemented !!!
              return;
          }
        }

        case ATTR_COORD_3D:
        {                                   // destination: [ double x, double y, double z ]
          double[] to = (double[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2D:        // [ d x, d y ] -> [ d x, d y, d z ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = da[ptrFrom];
                  tmp[1] = da[ptrFrom+1];
                  tmp[2] = 0.0;
                  tr.transformPoint3D( tmp, 0, 3, to, ptr, 3 );
                }
              return;

            case ATTR_COORD_2F:        // [ f x, f y ] -> [ d x, d y, d z ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptrFrom];
                  tmp[1] = fa[ptrFrom+1];
                  tmp[2] = 0.0;
                  tr.transformPoint3D( tmp, 0, 3, to, ptr, 3 );
                }
              return;

            case ATTR_COORD_3D:        // [ d x, d y, d z ] -> [ d x, d y, d z ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint3D( da, ptr, 3, to, ptr, 3 );
              return;

            case ATTR_COORD_3F:        // [ f x, f y, f z ] -> [ d x, d y, d z ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptr];
                  tmp[1] = fa[ptr+1];
                  tmp[2] = fa[ptr+2];
                  tr.transformPoint3D( tmp, 0, 3, to, ptr, 3 );
                }
              return;

            case ATTR_COORDH_3D:       // [ d x, d y, d w ] -> [ d x, d y, d z ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = da[ptr];
                  tmp[1] = da[ptr+1];
                  tmp[2] = 0.0;
                  tmp[3] = da[ptr+2];
                  tr.transformPoint3D( tmp, 0, 4, to, ptr, 3 );
                }
              return;

            case ATTR_COORDH_3F:       // [ f x, f y, f w ] -> [ d x, d y, d z ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptr];
                  tmp[1] = fa[ptr+1];
                  tmp[2] = 0.0;
                  tmp[3] = fa[ptr+2];
                  tr.transformPoint3D( tmp, 0, 4, to, ptr, 3 );
                }
              return;

            case ATTR_COORDH_4D:       // [ d x, d y, d z, d w ] -> [ d x, d y, d z ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint3D( da, ptrFrom, 4, to, ptr, 3 );
              return;

            case ATTR_COORDH_4F:       // [ f x, f y, f z, f w ] -> [ d x, d y, d z ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptrFrom];
                  tmp[1] = fa[ptrFrom+1];
                  tmp[2] = fa[ptrFrom+2];
                  tmp[3] = fa[ptrFrom+3];
                  tr.transformPoint3D( tmp, 0, 4, to, ptr, 3 );
                }
              return;

            case ATTR_COORD_2I:        // [ i x, i y ] -> [ d x, d y, d z ]
            case ATTR_COORD_3I:        // [ i x, i y, i z ] -> [ d x, d y, d z ]
              // !!! not implemented !!!
              return;
          }
        }

        case ATTR_COORD_3F:
        {                                   // destination: [ float x, float y, float z ]
          float[] to = (float[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2D:        // [ d x, d y ] -> [ f x, f y, f z ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = da[ptrFrom];
                  tmp[1] = da[ptrFrom+1];
                  tmp[2] = 0.0;
                  tr.transformPoint3D( tmp, 0, 3, tmp, 0, 3 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                  to[ptr+2] = (float)tmp[2];
                }
              return;

            case ATTR_COORD_2F:        // [ f x, f y ] -> [ f x, f y, f z ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmpf[0] = fa[ptrFrom];
                  tmpf[1] = fa[ptrFrom + 1];
                  tmpf[2] = 0.0f;
                  tr.transformPoint3D( tmpf, 0, 3, to, ptr, 3 );
                }
              return;

            case ATTR_COORD_3D:        // [ d x, d y, d z ] -> [ f x, f y, f z ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptr, 3, tmp, 0, 3 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                  to[ptr+2] = (float)tmp[2];
                }
              return;

            case ATTR_COORD_3F:        // [ f x, f y, f z ] -> [ f x, f y, f z ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint3D( fa, ptr, 3, to, ptr, 3 );
              return;

            case ATTR_COORDH_3D:       // [ d x, d y, d w ] -> [ f x, f y, f z ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = da[ptr];
                  tmp[1] = da[ptr+1];
                  tmp[2] = 0.0;
                  tmp[3] = da[ptr+2];
                  tr.transformPoint3D( tmp, 0, 4, tmp, 0, 3 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                  to[ptr+2] = (float)tmp[2];
                }
              return;

            case ATTR_COORDH_3F:       // [ f x, f y, f w ] -> [ f x, f y, f z ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmpf[0] = fa[ptr];
                  tmpf[1] = fa[ptr+1];
                  tmpf[2] = 0.0f;
                  tmpf[3] = fa[ptr+2];
                  tr.transformPoint3D( tmpf, 0, 4, to, ptr, 3 );
                }
              return;

            case ATTR_COORDH_4D:       // [ d x, d y, d z, d w ] -> [ f x, f y, f z ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptrFrom, 4, tmp, 0, 3 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                  to[ptr+2] = (float)tmp[2];
                }
              return;

            case ATTR_COORDH_4F:       // [ f x, f y, f z, f w ] -> [ f x, f y, f z ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint3D( fa, ptrFrom, 4, to, ptr, 3 );
              return;

            case ATTR_COORD_2I:        // [ i x, i y ] -> [ f x, f y, f z ]
            case ATTR_COORD_3I:        // [ i x, i y, i z ] -> [ f x, f y, f z ]
              // !!! not implemented !!!
              return;
          }
        }

        case ATTR_COORDH_3D:
        {                                   // destination: [ double x, double y, double w ]
          double[] to = (double[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2D:        // [ d x, d y ] -> [ d x, d y, d w ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint2D( da, ptrFrom, 2, to, ptr, 3 );
              return;

            case ATTR_COORD_2F:        // [ f x, f y ] -> [ d x, d y, d w ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptrFrom];
                  tmp[1] = fa[ptrFrom+1];
                  tr.transformPoint2D( tmp, 0, 2, to, ptr, 3 );
                }
              return;

            case ATTR_COORD_3D:        // [ d x, d y, d z ] -> [ d x, d y, d w ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptr, 3, tmp, 0, 4 );
                  to[ptr]   = tmp[0];
                  to[ptr+1] = tmp[1];
                  to[ptr+2] = tmp[3];
                }
              return;

            case ATTR_COORD_3F:        // [ f x, f y, f z ] -> [ d x, d y, d w ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptr];
                  tmp[1] = fa[ptr+1];
                  tmp[2] = fa[ptr+2];
                  tr.transformPoint3D( tmp, 0, 3, tmp, 0, 4 );
                  to[ptr]   = tmp[0];
                  to[ptr+1] = tmp[1];
                  to[ptr+2] = tmp[3];
                }
              return;

            case ATTR_COORDH_3D:       // [ d x, d y, d w ] -> [ d x, d y, d w ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint2D( da, ptr, 3, to, ptr, 3 );
              return;

            case ATTR_COORDH_3F:       // [ f x, f y, f w ] -> [ d x, d y, d w ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptr];
                  tmp[1] = fa[ptr+1];
                  tmp[2] = fa[ptr+2];
                  tr.transformPoint2D( tmp, 0, 3, to, ptr, 3 );
                }
              return;

            case ATTR_COORDH_4D:       // [ d x, d y, d z, d w ] -> [ d x, d y, d w ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptrFrom, 4, tmp, 0, 4 );
                  to[ptr]   = tmp[0];
                  to[ptr+1] = tmp[1];
                  to[ptr+2] = tmp[3];
                }
              return;

            case ATTR_COORDH_4F:       // [ f x, f y, f z, f w ] -> [ d x, d y, d w ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( fa, ptrFrom, 4, tmpf, 0, 4 );
                  to[ptr]   = tmpf[0];
                  to[ptr+1] = tmpf[1];
                  to[ptr+2] = tmpf[3];
                }
              return;

            case ATTR_COORD_2I:       // [ i x, i y ] -> [ d x, d y, d w ]
            case ATTR_COORD_3I:       // [ i x, i y, i z ] -> [ d x, d y, d w ]
              // !!! not implemented !!!
              return;
          }
        }

        case ATTR_COORDH_3F:
        {                                   // destination: [ float x, float y, float w ]
          float[] to = (float[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2D:        // [ d x, d y ] -> [ f x, f y, f w ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint2D( da, ptrFrom, 2, tmp, 0, 3 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                  to[ptr+2] = (float)tmp[2];
                }
              return;

            case ATTR_COORD_2F:        // [ f x, f y ] -> [ f x, f y, f w ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint2D( fa, ptrFrom, 2, to, ptr, 3 );
              return;

            case ATTR_COORD_3D:        // [ d x, d y, d z ] -> [ f x, f y, f w ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptr, 3, tmp, 0, 4 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                  to[ptr+2] = (float)tmp[3];
                }
              return;

            case ATTR_COORD_3F:        // [ f x, f y, f z ] -> [ f x, f y, f w ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( fa, ptr, 3, tmpf, 0, 4 );
                  to[ptr]   = tmpf[0];
                  to[ptr+1] = tmpf[1];
                  to[ptr+2] = tmpf[3];
                }
              return;

            case ATTR_COORDH_3D:       // [ d x, d y, d w ] -> [ f x, f y, f w ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint2D( da, ptr, 3, tmp, 0, 3 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                  to[ptr+2] = (float)tmp[2];
                }
              return;

            case ATTR_COORDH_3F:       // [ f x, f y, f w ] -> [ f x, f y, f w ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint2D( fa, ptr, 3, to, ptr, 3 );
              return;

            case ATTR_COORDH_4D:       // [ d x, d y, d z, d w ] -> [ f x, f y, f w ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptrFrom, 4, tmp, 0, 4 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                  to[ptr+2] = (float)tmp[3];
                }
              return;

            case ATTR_COORDH_4F:       // [ f x, f y, f z, f w ] -> [ f x, f y, f w ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( fa, ptrFrom, 4, tmpf, 0, 4 );
                  to[ptr]   = tmpf[0];
                  to[ptr+1] = tmpf[1];
                  to[ptr+2] = tmpf[3];
                }
              return;

            case ATTR_COORD_2I:       // [ i x, i y ] -> [ f x, f y, f w ]
            case ATTR_COORD_3I:       // [ i x, i y, i z ] -> [ f x, f y, f w ]
              // !!! not implemented !!!
              return;
          }
        }

        case ATTR_COORDH_4D:
        {                                   // destination: [ double x, double y, double z, double w ]
          double[] to = (double[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2D:        // [ d x, d y ] -> [ d x, d y, d z, d w ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = da[ptrFrom];
                  tmp[1] = da[ptrFrom+1];
                  tmp[2] = 0.0;
                  tr.transformPoint3D( tmp, 0, 3, to, ptr, 4 );
                }
              return;

            case ATTR_COORD_2F:        // [ f x, f y ] -> [ d x, d y, d z, d w ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptrFrom];
                  tmp[1] = fa[ptrFrom+1];
                  tmp[2] = 0.0;
                  tr.transformPoint3D( tmp, 0, 3, to, ptr, 4 );
                }
              return;

            case ATTR_COORD_3D:        // [ d x, d y, d z ] -> [ d x, d y, d z, d w ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint3D( da, ptrFrom, 3, to, ptr, 4 );
              return;

            case ATTR_COORD_3F:        // [ f x, f y, f z ] -> [ d x, d y, d z, d w ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptrFrom];
                  tmp[1] = fa[ptrFrom+1];
                  tmp[2] = fa[ptrFrom+2];
                  tr.transformPoint3D( tmp, 0, 3, to, ptr, 4 );
                }
              return;

            case ATTR_COORDH_3D:       // [ d x, d y, d w ] -> [ d x, d y, d z, d w ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = da[ptrFrom];
                  tmp[1] = da[ptrFrom+1];
                  tmp[2] = 0.0;
                  tmp[3] = da[ptrFrom+2];
                  tr.transformPoint3D( tmp, 0, 4, to, ptr, 4 );
                }
              return;

            case ATTR_COORDH_3F:       // [ f x, f y, f w ] -> [ d x, d y, d z, d w ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptrFrom];
                  tmp[1] = fa[ptrFrom+1];
                  tmp[2] = 0.0;
                  tmp[3] = fa[ptrFrom+2];
                  tr.transformPoint3D( tmp, 0, 4, to, ptr, 4 );
                }
              return;

            case ATTR_COORDH_4D:       // [ d x, d y, d z, d w ] -> [ d x, d y, d z, d w ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 4 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint3D( da, ptr, 4, to, ptr, 4 );
              return;

            case ATTR_COORDH_4F:       // [ f x, f y, f z, f w ] -> [ d x, d y, d z, d w ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 4 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = fa[ptr];
                  tmp[1] = fa[ptr+1];
                  tmp[2] = fa[ptr+2];
                  tmp[3] = fa[ptr+3];
                  tr.transformPoint3D( tmp, 0, 4, to, ptr, 4 );
                }
              return;

            case ATTR_COORD_2I:        // [ i x, i y ] -> [ d x, d y, d z, d w ]
            case ATTR_COORD_3I:        // [ i x, i y, i z ] -> [ d x, d y, d z, d w ]
              // !!! not implemented !!!
              return;
          }
        }

        case ATTR_COORDH_4F:
        {                                   // destination: [ float x, float y, float z, float w ]
          float[] to = (float[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2D:        // [ d x, d y ] -> [ f x, f y, f z, f w ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = da[ptrFrom];
                  tmp[1] = da[ptrFrom+1];
                  tmp[2] = 0.0;
                  tr.transformPoint3D( tmp, 0, 3, tmp, 0, 4 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                }
              return;

            case ATTR_COORD_2F:        // [ f x, f y ] -> [ f x, f y, f z, f w ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmpf[0] = fa[ptrFrom];
                  tmpf[1] = fa[ptrFrom+1];
                  tmpf[2] = 0.0f;
                  tr.transformPoint3D( tmpf, 0, 3, to, ptr, 4 );
                }
              return;

            case ATTR_COORD_3D:        // [ d x, d y, d z ] -> [ f x, f y, f z, f w ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptrFrom, 3, tmp, 0, 4 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                  to[ptr+2] = (float)tmp[2];
                }
              return;

            case ATTR_COORD_3F:        // [ f x, f y, f z ] -> [ f x, f y, f z, f w ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint3D( fa, ptrFrom, 3, to, ptr, 4 );
              return;

            case ATTR_COORDH_3D:       // [ d x, d y, d w ] -> [ f x, f y, f z, f w ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = da[ptrFrom];
                  tmp[1] = da[ptrFrom+1];
                  tmp[2] = 0.0;
                  tmp[3] = da[ptrFrom+2];
                  tr.transformPoint3D( tmp, 0, 4, tmp, 0, 4 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                  to[ptr+2] = (float)tmp[2];
                  to[ptr+3] = (float)tmp[3];
                }
              return;

            case ATTR_COORDH_3F:       // [ f x, f y, f w ] -> [ f x, f y, f z, f w ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 4, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmpf[0] = fa[ptrFrom];
                  tmpf[1] = fa[ptrFrom+1];
                  tmpf[2] = 0.0f;
                  tmpf[3] = fa[ptrFrom+2];
                  tr.transformPoint3D( tmpf, 0, 4, to, ptr, 4 );
                }
              return;

            case ATTR_COORDH_4D:       // [ d x, d y, d z, d w ] -> [ f x, f y, f z, f w ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptr, 4, tmp, 0, 4 );
                  to[ptr]   = (float)tmp[0];
                  to[ptr+1] = (float)tmp[1];
                  to[ptr+2] = (float)tmp[2];
                  to[ptr+3] = (float)tmp[3];
                }
              return;

            case ATTR_COORDH_4F:       // [ f x, f y, f z, f w ] -> [ f x, f y, f z, f w ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 4 )
                if ( set == null || set.get( i ) )
                  tr.transformPoint3D( fa, ptr, 4, to, ptr, 4 );
              return;

            case ATTR_COORD_2I:        // [ i x, i y ] -> [ f x, f y, f z, f w ]
            case ATTR_COORD_3I:        // [ i x, i y, i z ] -> [ f x, f y, f z, f w ]
              // !!! not implemented !!!
              return;
          }
        }

        case ATTR_COORD_2I:
        {                                   // destination: [ int x, int y ]
          int[] to = (int[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2D:        // [ d x, d y ] -> [ i x, i y ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 2 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint2D( da, ptr, 2, tmp, 0, 2 );
                  to[ptr]     = Formula.round( tmp[0] );
                  to[ptr + 1] = Formula.round( tmp[1] );
                }
              return;

            case ATTR_COORD_2F:        // [ f x, f y ] -> [ i x, i y ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 2 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint2D( fa, ptr, 2, tmpf, 0, 2 );
                  to[ptr]   = Math.round( tmpf[0] );
                  to[ptr+1] = Math.round( tmpf[1] );
                }
              return;

            case ATTR_COORD_3D:        // [ d x, d y, d z ] -> [ i x, i y ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptrFrom, 3, tmp, 0, 3 );
                  to[ptr]   = Formula.round( tmp[0] );
                  to[ptr+1] = Formula.round( tmp[1] );
                }
              return;

            case ATTR_COORD_3F:        // [ f x, f y, f z ] -> [ i x, i y ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( fa, ptrFrom, 3, tmpf, 0, 3 );
                  to[ptr]   = Formula.round( tmpf[0] );
                  to[ptr+1] = Formula.round( tmpf[1] );
                }
              return;

            case ATTR_COORDH_3D:       // [ d x, d y, d w ] -> [ i x, i y ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint2D( da, ptrFrom, 3, tmp, 0, 2 );
                  to[ptr]   = Formula.round( tmp[0] );
                  to[ptr+1] = Formula.round( tmp[1] );
                }
              return;

            case ATTR_COORDH_3F:       // [ f x, f y, f w ] -> [ i x, i y ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint2D( fa, ptrFrom, 3, tmpf, 0, 2 );
                  to[ptr]   = Math.round( tmpf[0] );
                  to[ptr+1] = Math.round( tmpf[1] );
                }
              return;

            case ATTR_COORDH_4D:       // [ d x, d y, d z, d w ] -> [ i x, i y ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptrFrom, 4, tmp, 0, 3 );
                  to[ptr]   = Formula.round( tmp[0] );
                  to[ptr+1] = Formula.round( tmp[1] );
                }
              return;

            case ATTR_COORDH_4F:       // [ f x, f y, f z, f w ] -> [ i x, i y ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 2, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( fa, ptrFrom, 4, tmpf, 0, 3 );
                  to[ptr]   = Formula.round( tmpf[0] );
                  to[ptr+1] = Formula.round( tmpf[1] );
                }
              return;

            case ATTR_COORD_2I:        // [ i x, i y ] -> [ i x, i y ]
            case ATTR_COORD_3I:        // [ i x, i y, i z ] -> [ i x, i y ]
              // !!! not implemented !!!
              return;
          }
        }

        case ATTR_COORD_3I:
        {                                   // destination: [ int x, int y, int z ]
          int[] to = (int[])data;
          switch ( from.type )
          {
            case ATTR_COORD_2D:        // [ d x, d y ] -> [ i x, i y, i z ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = da[ptrFrom];
                  tmp[1] = da[ptrFrom + 1];
                  tmp[2] = 0.0;
                  tr.transformPoint3D( tmp, 0, 3, tmp, 0, 3 );
                  to[ptr]   = Formula.round( tmp[0] );
                  to[ptr+1] = Formula.round( tmp[1] );
                  to[ptr+2] = Formula.round( tmp[2] );
                }
              return;

            case ATTR_COORD_2F:        // [ f x, f y ] -> [ i x, i y, i z ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 2 )
                if ( set == null || set.get( i ) )
                {
                  tmpf[0] = fa[ptrFrom];
                  tmpf[1] = fa[ptrFrom + 1];
                  tmpf[2] = 0.0f;
                  tr.transformPoint3D( tmpf, 0, 3, tmpf, 0, 3 );
                  to[ptr]   = Math.round( tmpf[0] );
                  to[ptr+1] = Math.round( tmpf[1] );
                  to[ptr+2] = Math.round( tmpf[2] );
                }
              return;

            case ATTR_COORD_3D:        // [ d x, d y, d z ] -> [ i x, i y, i z ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptr, 3, tmp, 0, 3 );
                  to[ptr]   = Formula.round( tmp[0] );
                  to[ptr+1] = Formula.round( tmp[1] );
                  to[ptr+2] = Formula.round( tmp[2] );
                }
              return;

            case ATTR_COORD_3F:        // [ f x, f y, f z ] -> [ i x, i y, i z ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( fa, ptr, 3, tmpf, 0, 3 );
                  to[ptr]   = Formula.round( tmpf[0] );
                  to[ptr+1] = Formula.round( tmpf[1] );
                  to[ptr+2] = Formula.round( tmpf[2] );
                }
              return;

            case ATTR_COORDH_3D:       // [ d x, d y, d w ] -> [ i x, i y, i z ]
              da = (double[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmp[0] = da[ptr];
                  tmp[1] = da[ptr+1];
                  tmp[2] = 0.0;
                  tmp[3] = da[ptr+2];
                  tr.transformPoint3D( tmp, 0, 4, tmp, 0, 3 );
                  to[ptr]   = Formula.round( tmp[0] );
                  to[ptr+1] = Formula.round( tmp[1] );
                  to[ptr+2] = Formula.round( tmp[2] );
                }
              return;

            case ATTR_COORDH_3F:       // [ f x, f y, f w ] -> [ i x, i y, i z ]
              fa = (float[])from.data;
              for ( i = ptr = 0; i < s; i++, ptr += 3 )
                if ( set == null || set.get( i ) )
                {
                  tmpf[0] = fa[ptr];
                  tmpf[1] = fa[ptr+1];
                  tmpf[2] = 0.0f;
                  tmpf[3] = fa[ptr+2];
                  tr.transformPoint3D( tmpf, 0, 4, tmpf, 0, 3 );
                  to[ptr]   = Math.round( tmpf[0] );
                  to[ptr+1] = Math.round( tmpf[1] );
                  to[ptr+2] = Math.round( tmpf[2] );
                }
              return;

            case ATTR_COORDH_4D:       // [ d x, d y, d z, d w ] -> [ i x, i y, i z ]
              da = (double[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( da, ptrFrom, 4, tmp, 0, 3 );
                  to[ptr]   = Formula.round( tmp[0] );
                  to[ptr+1] = Formula.round( tmp[1] );
                  to[ptr+2] = Formula.round( tmp[2] );
                }
              return;

            case ATTR_COORDH_4F:       // [ f x, f y, f z, f w ] -> [ i x, i y, i z ]
              fa = (float[])from.data;
              for ( i = ptr = ptrFrom = 0; i < s; i++, ptr += 3, ptrFrom += 4 )
                if ( set == null || set.get( i ) )
                {
                  tr.transformPoint3D( fa, ptrFrom, 4, tmpf, 0, 3 );
                  to[ptr]   = Math.round( tmpf[0] );
                  to[ptr+1] = Math.round( tmpf[1] );
                  to[ptr+2] = Math.round( tmpf[2] );
                }
              return;

            case ATTR_COORD_2I:        // [ i x, i y ] -> [ i x, i y, i z ]
            case ATTR_COORD_3I:        // [ i x, i y, i z ] -> [ i x, i y, i z ]
              // !!! not implemented !!!
              return;
          }
        }
      }
    }
  }

  /**
   * Attribute key - [ entity, context, ident ] compound.
   * Used in attribute lookup (hashing).
   *
   * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
   */
  protected static class AttrKey
  {
    /** Hash-coefficient for {@link #entity}. */
    protected final static int HASH_ENTITY  = 1031;

    /** Hash-coefficient for {@link #context}. */
    protected final static int HASH_CONTEXT = 2003;

    /** Entity identifier ({@link Brep#SCENE} to {@link Brep#SOLID}). */
    public int entity;

    /** Context identifier or {@link Brep#NULL} if shared. */
    public int context;

    /** Attribute identifier. */
    public String name;

    public AttrKey ()
    {
    }

    public AttrKey ( int where, int ctx, String ident )
    {
      entity  = where;
      context = ctx;
      name    = ident;
    }

    @Override
    public final boolean equals ( Object obj )
    {
      if ( obj == null ||
           !(obj instanceof AttrKey) ) return false;

      final AttrKey ak = (AttrKey)obj;
      return( ak.entity  == entity &&
              ak.context == ak.context &&
              ak.name.equals( name ) );
    }

    @Override
    public final int hashCode ()
    {
      int hash = (name == null) ? 13 : name.hashCode();
      return( hash + entity * HASH_ENTITY + context * HASH_CONTEXT );
    }
  }

  /**
   * Support class - attribute map.
   *
   * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
   */
  protected static class MapAttr extends HashMap< AttrKey, Attribute >
  {
    private static final long serialVersionUID = 1L;

    /** Global attribute key (used for fast lookup). */
    protected static final AttrKey key = new AttrKey();

    /** Fast attribute lookup (doesn't need to create new {@link VEFDS.AttrKey} instance). */
    public synchronized final Attribute get ( int where, int ctx, String ident )
    {
      key.entity  = where;
      key.context = ctx;
      key.name    = ident;
      return get( key );
    }

    /** Fast attribute lookup, tries shared attribute as well. */
    public synchronized final Attribute getExt ( int where, int ctx, String ident )
    {
      key.entity  = where;
      key.context = ctx;
      key.name    = ident;
      final Attribute a = get( key );
      if ( a != null || ctx == NULL ) return a;

        // try shared attribute as well..
      key.context = NULL;
      return get( key );
    }

    /** Convenient put method. */
    public final void put ( Attribute a )
    {
      put( new AttrKey( a.where, a.context, a.ident ), a );
    }
  }

  //==========================================================================
  //  Data:

  /** Minimum attribute-array length. */
  protected final static int MIN_ARRAY = 1024;

  /** Edge attribute type (V<sub>1</sub>, V<sub>2</sub> or winged-edge record). */
  final static int ATTR_EDGE     = ATTR_LAST + 1;

  /** Triangle attribute type (E<sub>1</sub>, E<sub>2</sub>, E<sub>3</sub>). */
  final static int ATTR_TRIANGLE = ATTR_EDGE + 1;

  /** Face attribute type (reference to an array: [ E<sub>1</sub>, E<sub>2</sub>, ... E<sub>N</sub> ]). */
  final static int ATTR_FACE     = ATTR_TRIANGLE + 1;

  /** Division attribute type (two references to arrays: [ F<sub>1</sub>, .. F<sub>K</sub> ] and [ F'<sub>1</sub>, .. F'<sub>N</sub> ]). */
  final static int ATTR_DIVISION = ATTR_FACE + 1;

  /** Solid attribute type (reference to an array: [ F<sub>1</sub>, .. F<sub>N</sub> ]). */
  final static int ATTR_SOLID    = ATTR_DIVISION + 1;

  /** Attribute mappings (global, context, vertex, edge, face, division, solid -associated). */
  protected final MapAttr attrMap = new MapAttr();

  /**
   * Number of actually used vertex, edge, face, division and solid handles (granul-allocated).
   * Contexts are 1-allocated.
   */
  protected final int[] used = new int[ SOLID+1 ];

  /** Global attribute map: handle {@code -> Attribute}. attributes-allocated. */
  protected final IntMap< Attribute > attributes = new IntMap< Attribute >( 5 );

  /** Bit-mask of used context IDs. */
  protected final BitSet usedContexts = new BitSet();

  /** Type of primary coordinate attribute {@link #COORD COORD}. */
  protected int primaryType = ATTR_COORDH_4D;

  /** Type of projected coordinate attribute {@link #COORD_PROJ COORD_PROJ}. */
  protected int primaryProjType = ATTR_COORD_3I;

  /** Edge-record size ({@link #WEDGE_SIZE} in winged-edge mode). */
  protected int edgeSize = EDGE_SIZE;

  /** Coordinate types or edge mode were changed since the last {@link #init}. */
  protected boolean dirty = false;

  //==========================================================================
  //  Support routines:

  /**
   * Returns granul-based array length.
   *
   * @param  len Required array length.
   * @return Valid array length.
   */
  protected final static int getGranulLength ( int len )
  {
    int l = MIN_ARRAY;

    while ( l < len )
      l += l;

    return l;
  }

  /**
   * Adjusts attribute-array lengths for the given entity.
   *
   * @param  where The entity code.
   * @param  len New length of an entity-array.
   */
  protected void adjustAttributes ( int where, int len )
  {
    if ( where == SCENE ) return;         // SCENE attributes need not to be resized

    if ( where != CONTEXT )
      len = getGranulLength( len );       // non-CONTEXT attributes are granul-allocated

    for ( Attribute a : attributes )
      if ( a != null && a.where == where )
        a.assertDataSize( len );
  }

  /** Assigns an attribute-handle to the new attribute. */
  protected int insertAttribute ( Attribute a )
  {
      // find new attribute's id:
    int id = 0;
    while ( attributes.get( id ) != null ) id++;

    a.handle = id;
    attributes.put( id, a );

      // attribute hash-map:
    attrMap.put( a );

    return id;
  }

  /**
   * Stores one data item into the given {@code Attribute}.
   *
   * @param  a The attribute which data will be stored to.
   * @param  id Entity handle (array index).
   * @param  val Attribute value.
   */
  protected void store ( Attribute a, int id, Object val )
  {
    if ( a == null ) return;

    if ( a.data != null && a.data.getClass().isArray() )
    {                                     // array attribute => store(id)
      if ( id < 0 || (id+1) * a.itemSize > Array.getLength( a.data ) ) return;

      if ( a.itemSize > 1 )
      {                                   // one data item ... more than one array item
        if ( val == null || !val.getClass().isArray() ||
             Array.getLength( val ) < a.itemSize ) // invalid "val" => create default one

          switch ( a.type )
          {
            case ATTR_OBJECT:
              val = new Object[ a.itemSize ];
              break;

            case ATTR_BOOLEAN:
              val = new boolean[ a.itemSize ];
              break;

            case ATTR_INTEGER:
            case ATTR_COORD_2I:
            case ATTR_COORD_3I:
            case ATTR_EDGE:
            case ATTR_TRIANGLE:
              val = new int[ a.itemSize ];
              break;

            case ATTR_LONG:
              val = new long[ a.itemSize ];
              break;

            case ATTR_DOUBLE:
            case ATTR_COORD_2D:
            case ATTR_COORD_3D:
            case ATTR_COORDH_3D:
            case ATTR_COORDH_4D:
              val = new double[ a.itemSize ];
              break;

            case ATTR_FLOAT:
            case ATTR_COORD_2F:
            case ATTR_COORD_3F:
            case ATTR_COORDH_3F:
            case ATTR_COORDH_4F:
            case ATTR_RGB_3F:
            case ATTR_RGBA_4F:
              val = new float[ a.itemSize ];
              break;

            case ATTR_FACE:
            case ATTR_SOLID:
            case ATTR_DIVISION:
              val = new int[ a.itemSize ][];
              break;
          }

        System.arraycopy( val, 0, a.data, id * a.itemSize, a.itemSize );
        return;
      }

        // a.itemSize == 1:
      if ( val != null && val.getClass().isArray() )
      {
        System.arraycopy( val, 0, a.data, id, 1 );
        return;
      }

      Array.set( a.data, id, val );
      return;
    }

    a.data = val;
  }

  /**
   * Fetches one data item from the given {@code Attribute}.
   *
   * @param  a The attribute which data will be fetched from.
   * @param  id Entity handle (array index).
   * @param  pre Pre-allocated result object (can be {@code null}).
   * @return The value of the attribute for the given entity.
   */
  protected Object fetch ( Attribute a, int id, Object pre )
  {
    if ( a == null ) return null;

    if ( a.data != null && a.data.getClass().isArray() )
    {                                     // array attribute => fetch(id)
      if ( id < 0 || (id+1) * a.itemSize > Array.getLength( a.data ) ) return null;

      if ( a.itemSize > 1 )
      {                                   // one data item ... more than one array item
        if ( pre == null || !pre.getClass().isArray() ||
             Array.getLength( pre ) < a.itemSize )
          switch ( a.type )
          {
            case ATTR_OBJECT:
              pre = new Object[ a.itemSize ];
              break;

            case ATTR_BOOLEAN:
              pre = new boolean[ a.itemSize ];
              break;

            case ATTR_INTEGER:
            case ATTR_COORD_2I:
            case ATTR_COORD_3I:
            case ATTR_EDGE:
            case ATTR_TRIANGLE:
              pre = new int[ a.itemSize ];
              break;

            case ATTR_LONG:
              pre = new long[ a.itemSize ];
              break;

            case ATTR_DOUBLE:
            case ATTR_COORD_2D:
            case ATTR_COORD_3D:
            case ATTR_COORDH_3D:
            case ATTR_COORDH_4D:
              pre = new double[ a.itemSize ];
              break;

            case ATTR_FLOAT:
            case ATTR_COORD_2F:
            case ATTR_COORD_3F:
            case ATTR_COORDH_3F:
            case ATTR_COORDH_4F:
            case ATTR_RGB_3F:
            case ATTR_RGBA_4F:
              pre = new float[ a.itemSize ];
              break;

            case ATTR_FACE:
            case ATTR_SOLID:
            case ATTR_DIVISION:
              pre = new int[ a.itemSize ][];
              break;
          }

        System.arraycopy( a.data, id * a.itemSize, pre, 0, a.itemSize );
        return pre;
      }

        // a.itemSize == 1:
      if ( pre != null && pre.getClass().isArray() )
      {
        System.arraycopy( a.data, id, pre, 0, 1 );
        return pre;
      }

      return Array.get( a.data, id );
    }

    return a.data;
  }

  /**
   * Fetches {@code int[]} value from the given {@code Attribute}.
   *
   * @param  a The attribute which data will be fetched from.
   * @param  id Entity handle (array index).
   * @param  pre Pre-allocated result object (can be {@code null}).
   * @return The value of the attribute for the given entity.
   */
  protected int[] fetch ( Attribute a, int id, int[] pre )
  {
    if ( a      == null ||
         a.data == null ||
         id     <  0    ||
         id     >= a.size ) return null;

    if ( a.type == ATTR_FACE  ||
         a.type == ATTR_SOLID ||
         a.type == ATTR_DIVISION )
    {
      final int[] src = ((int[][])(a.data))[ id ];
      if ( pre == null ||
           pre.length < src.length )
        pre = new int[ src.length ];

      System.arraycopy( src, 0, pre, 0, src.length );
      if ( src.length < pre.length )
        Arrays.fill( pre, src.length, pre.length, NULL );
      return pre;
    }

    if ( a.type != ATTR_INTEGER   && a.type != ATTR_COORD_2I  &&
         a.type != ATTR_COORD_3I  && a.type != ATTR_EDGE      &&
         a.type != ATTR_TRIANGLE ) return null;

    if ( pre == null ||
         pre.length < a.itemSize )
      pre = new int[ a.itemSize ];

    System.arraycopy( a.data, id * a.itemSize, pre, 0, a.itemSize );
    if ( a.itemSize < pre.length )
      Arrays.fill( pre, a.itemSize, pre.length, NULL );
    return pre;
  }

  /**
   * Fetches {@code float[]} value from the given {@code Attribute}.
   *
   * @param  a The attribute which data will be fetched from.
   * @param  id Entity handle (array index).
   * @param  pre Pre-allocated result object (can be {@code null}).
   * @return The value of the attribute for the given entity.
   */
  protected float[] fetch ( Attribute a, int id, float[] pre )
  {
    if ( a      == null ||
         a.data == null ||
         (a.type != ATTR_FLOAT     && a.type != ATTR_COORD_2F  &&
          a.type != ATTR_COORD_3F  && a.type != ATTR_COORDH_3F &&
          a.type != ATTR_COORDH_4F && a.type != ATTR_RGB_3F    &&
          a.type != ATTR_RGBA_4F) ||
         id     <  0    ||
         id     >= a.size ) return null;

    if ( pre == null ||
         pre.length < a.itemSize )
      pre = new float[ a.itemSize ];

    System.arraycopy( a.data, id * a.itemSize, pre, 0, a.itemSize );
    return pre;
  }

  /**
   * Fetches {@code double[]} value from the given {@code Attribute}.
   *
   * @param  a The attribute which data will be fetched from.
   * @param  id Entity handle (array index).
   * @param  pre Pre-allocated result object (can be {@code null}).
   * @return The value of the attribute for the given entity.
   */
  protected double[] fetch ( Attribute a, int id, double[] pre )
  {
    if ( a      == null ||
         a.data == null ||
         (a.type != ATTR_DOUBLE   && a.type != ATTR_COORD_2D  &&
          a.type != ATTR_COORD_3D && a.type != ATTR_COORDH_3D &&
          a.type != ATTR_COORDH_4D) ||
         id     <  0    ||
         id     >= a.size ) return null;

    if ( pre == null ||
         pre.length < a.itemSize )
      pre = new double[ a.itemSize ];

    System.arraycopy( a.data, id * a.itemSize, pre, 0, a.itemSize );
    return pre;
  }

  /** Appends the given integer to the 1-allocated array. */
  public final static int[] arrayAppend ( int[] arr, int item )
  {
    if ( arr == null )
      return new int[] { item };

    int   s      = arr.length;
    int[] result = new int[ s + 1 ];
    System.arraycopy( arr, 0, result, 0, s );
    result[s]    = item;

    return result;
  }

  /** Finds an edge from its vertices. Creates a new one if not found. */
  public int assertEdge ( int v1, int v2, int face )
  {
    final Attribute a = attrMap.getExt( EDGE, NULL, DATA );
    if ( a == null ||
         a.type != ATTR_EDGE ) return NULL;

    int i;
    final int max   = used[EDGE] * edgeSize;
    int[] arr = (int[])(a.data);
    for ( i = 0; i < max; i += edgeSize )
      if ( arr[i] == v1 && arr[i+1] == v2 ||
           arr[i] == v2 && arr[i+1] == v1 )
        return( i / edgeSize );

      // create a new edge:
    final int e = createEdge();
    arr = (int[])(a.data);
    i = e * edgeSize;
    arr[ i ]     = v1;
    arr[ i + 1 ] = v2;
    if ( edgeSize == WEDGE_SIZE )
    {
      arr[ i + WEDGE_FL ] =
      arr[ i + WEDGE_FR ] = face;
      arr[ i + WEDGE_LP ] =
      arr[ i + WEDGE_LN ] =
      arr[ i + WEDGE_RP ] =
      arr[ i + WEDGE_RN ] = e;
    }

    return e;
  }

  //==========================================================================
  //  Interface Brep:

  /**
   * Set default coordinate types.
   *
   * @param  baseType Base type for the {@link #COORD COORD} attribute.
   * @param  baseProjType Base type for the {@link #COORD_PROJ COORD_PROJ} (projected) attribute.
   */
  public void setCoordTypes ( int baseType, int baseProjType )
  {
    dirty           = dirty || primaryType != baseType || primaryProjType != baseProjType;

    primaryType     = baseType;
    primaryProjType = baseProjType;
  }

  /**
   * Set winged-edge mode.
   * Default value: {@code false} (don't use winged edge).
   *
   * @param  wedge Winged-edges will be used?
   * @see    #getEdgeRecord
   * @see    #WEDGE_SIZE
   */
  public void setEdgeMode ( boolean wedge )
  {
    dirty    = dirty || wedge != (edgeSize == WEDGE_SIZE);

    edgeSize = wedge ? WEDGE_SIZE : EDGE_SIZE;
  }

  /**
   * Database (re-)initialization.
   * Removes all scene contexts and implicit / explicit attributes.
   */
  public void init ()
  {
    dirty = false;

      // attribute mappings:
    attributes.clear();
    attrMap.clear();
    Arrays.fill( used, 0 );

      // implicit context:
    usedContexts.clear();
    createContext( NULL );

      // primary coordinate attribute:
    createAttribute( VERTEX,   CTX_DEFAULT, COORD, primaryType );
    if ( edgeSize == WEDGE_SIZE )
      createAttribute( VERTEX, NULL,        DATA,  ATTR_INTEGER );

      // another implicit attributes:
    createAttribute( EDGE,     NULL,        DATA,  ATTR_EDGE );
    createAttribute( FACE,     NULL,        DATA,  (edgeSize == WEDGE_SIZE) ? ATTR_INTEGER : ATTR_FACE );
    createAttribute( DIVISION, NULL,        DATA,  ATTR_DIVISION );
    createAttribute( SOLID,    NULL,        DATA,  ATTR_SOLID );
  }

  //--------------------------------------------------------------------------
  //  Scene contexts:

  /**
   * Creates a new scene context (can be pre-initialized from one of existing contexts).
   * No attributes (even implicit) are created for the new context (exception: pre-initialized
   * contexts).
   *
   * @param  ctxFrom Context handle for initialization (non-mandatory parameter -
   *                 can be {@link #NULL NULL}).
   * @return Handle to a new context.
   * @see    #copyAttribute(int,int,String)
   * @see    #copyAttribute(int,int)
   * @see    #removeContext
   */
  public int createContext ( int ctxFrom )
  {
      // find ID for the new context:
    int id = usedContexts.nextClearBit( CTX_DEFAULT );
    usedContexts.set( id );
    adjustAttributes( CONTEXT, usedContexts.length() );

      // pre-initialization:
    if ( ctxFrom != NULL && usedContexts.get( ctxFrom ) )
      for ( Attribute a : attributes )
        if ( a != null && a.context == ctxFrom ) // copy the attribute
          copyAttribute( a.handle, id );

    return id;
  }

  /**
   * Removes an existing scene context. {@link #CTX_DEFAULT} cannot be removed.
   *
   * @param  ctx Handle to a context to be destroyed.
   * @see    #createContext
   */
  public void removeContext ( int ctx )
  {
      // default (or non-existent) context shouldn't be removed:
    if ( ctx <= CTX_DEFAULT || !usedContexts.get( ctx ) ) return;

      // the context will be removed: remove its attributes first..
    for ( Attribute a : attributes )
      if ( a != null && a.context == ctx )
        removeAttribute( a.handle );

    usedContexts.clear( ctx );
    adjustAttributes( CONTEXT, usedContexts.length() );
  }

  /**
   * Creates a new scene attribute.
   *
   * @param  where Entity which a new attribute will be associated with ({@link #VERTEX VERTEX},
   *               {@link #EDGE EDGE}, ..).
   * @param  ctx Handle to a context (can be {@link #NULL NULL} for shared attributes).
   * @param  ident String identifier of the attribute.
   * @param  type Attribute type ({@code ATTR_*} constant).
   * @return Handle to a new attribute.
   * @see    #copyAttribute(int,int,String)
   * @see    #copyAttribute(int,int)
   * @see    #removeAttribute
   * @see    #copyCoordinates(int,int,String)
   * @see    #roundCoordinates(int,int,String,TrMatrix)
   * @see    #transformCoordinates(int,int,String,TrMatrix)
   */
  public int createAttribute ( int where, int ctx, String ident, int type )
  {
    if ( where < SCENE ||
         where > SOLID ||
         ctx != NULL && !usedContexts.get( ctx ) ) return NULL;

      // check for existing attributes:
    final Attribute a = attrMap.get( where, ctx, ident );
    if ( a != null )
    {
      if ( a.type == type ) return a.handle;
      return NULL;                        // use "removeAttribute()" first!
    }

      // attribute-array length:
    int len = -1;
    switch ( where )
    {
      case CONTEXT:                       // 1-allocated:
        len = usedContexts.length();
        break;

      case VERTEX:
      case EDGE:
      case FACE:
      case DIVISION:
      case SOLID:                         // granul-allocated:
        len = getGranulLength( used[where] );
        break;
    }

      // create a new attribute & connect it:
    return insertAttribute( new Attribute( where, ctx, ident, type, len ) );
  }

  /**
   * Copies the existing scene attribute to a new one (preserves attribute association
   * /{@code where}/ and {@code type}).
   *
   * @param  attrFrom Handle to a source attribute.
   * @param  ctx Handle to a context (cannot be {@link #NULL NULL}).
   * @param  ident String identifier of the attribute.
   * @return Handle to a new attribute.
   * @see    #createAttribute
   * @see    #copyAttribute(int,int)
   * @see    #removeAttribute
   * @see    #copyCoordinates(int,int,String)
   * @see    #roundCoordinates(int,int,String,TrMatrix)
   * @see    #transformCoordinates(int,int,String,TrMatrix)
   */
  public int copyAttribute ( int attrFrom, int ctx, String ident )
  {
      // check the "ctx" parameter:
    if ( ctx == NULL ) return NULL;

    final Attribute aFrom = attributes.get( attrFrom );

    if ( aFrom == null ||
         aFrom.context == ctx && aFrom.ident.equals( ident ) ) return NULL;

       // check for existing attributes:
    Attribute a = attrMap.get( aFrom.where, ctx, ident );
    if ( a != null )
    {
      if ( a.type != aFrom.type ) return NULL;  // use "removeAttribute()" first!

      a.deepCopy( aFrom );
      return a.handle;
    }

      // create a new attribute & connect it:
    return insertAttribute( new Attribute( aFrom, ctx, ident ) );
  }

  /**
   * Copies the existing scene attribute to a new one (preserves attribute association
   * /{@code where}/, {@code type} and identifier string /{@code ident}/).
   *
   * @param  attrFrom Handle to a source attribute.
   * @param  ctx Handle to a context (cannot be {@link #NULL NULL}).
   * @return Handle to a new attribute.
   * @see    #createAttribute
   * @see    #copyAttribute(int,int,String)
   * @see    #removeAttribute
   * @see    #copyCoordinates(int,int,String)
   * @see    #roundCoordinates(int,int,String,TrMatrix)
   * @see    #transformCoordinates(int,int,String,TrMatrix)
   */
  public final int copyAttribute ( int attrFrom, int ctx )
  {
    String ident = getAttributeIdent( attrFrom );
    return copyAttribute( attrFrom, ctx, ident );
  }

  /**
   * Removes the given scene attribute.
   *
   * @param  attr Handle to an attribute.
   * @see    #copyAttribute(int,int,String)
   * @see    #copyAttribute(int,int)
   * @see    #createAttribute
   */
  public void removeAttribute ( int attr )
  {
    Attribute a = attributes.get( attr );
    if ( a == null ) return;

    attributes.remove( a.handle );
    attrMap.remove( a );
  }

  /**
   * Finds a handle of the given scene attribute.
   *
   * @param  where Entity which the attribute is associated with ({@link #VERTEX VERTEX},
   *               {@link #EDGE EDGE}, ..).
   * @param  ctx Handle to a context (can be {@link #NULL NULL} for shared attributes).
   * @param  ident String identifier of the attribute.
   * @return Handle to the given attribute.
   */
  public final int getAttributeId ( int where, int ctx, String ident )
  {
    Attribute a = attrMap.getExt( where, ctx, ident );
    return( (a == null) ? NULL : a.handle );
  }

  /**
   * Return a {@code where} parameter of the given attribute (entity which the attribute
   * is associated with).
   *
   * @param  attr Handle to the attribute.
   * @return The {@code where} parameter.
   */
  public final int getAttributeWhere ( int attr )
  {
    final Attribute a = attributes.get( attr );
    return( (a == null) ? NULL : a.where );
  }

  /**
   * Return a context of the given attribute (for non-shared attributes only).
   *
   * @param  attr Handle to the attribute.
   * @return Exclusive context of the attribute (or {@link #NULL NULL}).
   */
  public final int getAttributeContext ( int attr )
  {
    final Attribute a = attributes.get( attr );
    return( (a == null) ? NULL : a.context );
  }

  /**
   * Returns a string identifier of the given attribute.
   *
   * @param  attr Handle to the attribute.
   * @return String identifier of the attribute ({@code null} if failed).
   */
  public final String getAttributeIdent ( int attr )
  {
    final Attribute a = attributes.get( attr );
    return( (a == null) ? null : a.ident );
  }

  /**
   * Returns a {@code type} of the given attribute.
   *
   * @param  attr Handle to the attribute.
   * @return The {@code type} parameter.
   */
  public final int getAttributeType ( int attr )
  {
    final Attribute a = attributes.get( attr );
    return( (a == null) ? NULL : a.type );
  }

  /**
   * Retrieves value of the given attribute for the given entity (general form).
   *
   * @param  where Entity which the attribute is associated with ({@link #VERTEX VERTEX},
   *               {@link #EDGE EDGE}, ..).
   * @param  ctx Handle to a context (can be {@link #NULL NULL} for shared attributes).
   * @param  ident String identifier of the attribute.
   * @param  id Handle to the entity.
   * @param  pre Pre-allocated result object (can be {@code null}).
   * @return The value of the attribute for the given entity.
   */
  public final Object getAttribute ( int where, int ctx, String ident, int id, Object pre )
  {
    return fetch( attrMap.getExt( where, ctx, ident ), id, pre );
  }

  /**
   * Retrieves value of the given attribute for the given entity (general form).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @param  pre Pre-allocated result object (can be {@code null}).
   * @return The value of the attribute for the given entity.
   */
  public final Object getAttribute ( int attr, int id, Object pre )
  {
    return fetch( attributes.get( attr ), id, pre );
  }

  /**
   * Sets value of the given attribute of the given entity (general form).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @param  val New attribute value (can be {@code null}).
   */
  public final void setAttribute ( int attr, int id, Object val )
  {
    store( attributes.get( attr ), id, val );
  }

  /**
   * Retrieves value of the given coordinate-attribute for the given entity (for
   * {@code float} [coordinate] types).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @param  coord Optional pre-allocated result array.
   * @return Result array ({@code coord} if possible).
   */
  public final float[] getAttribute ( int attr, int id, float[] coord )
  {
    return fetch( attributes.get( attr ), id, coord );
  }

  /**
   * Retrieves value of the given coordinate-attribute for the given entity (for
   * {@code double} coordinate types).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @param  coord Optional pre-allocated result array.
   * @return Result array ({@code coord} if possible).
   */
  public final double[] getAttribute ( int attr, int id, double[] coord )
  {
    return fetch( attributes.get( attr ), id, coord );
  }

  /**
   * Retrieves value of the given coordinate-attribute for the given entity (for
   * {@code int} coordinate types).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @param  coord Optional pre-allocated result array.
   * @return Result array ({@code coord} if possible).
   */
  public int[] getAttribute ( int attr, int id, int[] coord )
  {
    return fetch( attributes.get( attr ), id, coord );
  }

  /**
   * Retrieves value of the given flag-attribute for the given entity (for
   * {@code boolean} attribute type).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @return Boolean attribute value.
   */
  public boolean getBoolAttribute ( int attr, int id )
  {
    Object result = fetch( attributes.get( attr ), id, (Object)null );
    if ( result == null ) return false;

    if ( result.getClass().isArray() && Array.getLength( result ) > 0 )
      result = Array.get( result, 0 );
    if ( result instanceof Boolean )
      return( ((Boolean)result).booleanValue() );

    return false;
  }

  /**
   * Retrieves value of the given integer-attribute for the given entity (for
   * {@code int} attribute type).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @return Integer attribute value.
   */
  public int getIntAttribute ( int attr, int id )
  {
    Object result = fetch( attributes.get( attr ), id, (Object)null );
    if ( result == null ) return NULL;

    if ( result.getClass().isArray() && Array.getLength( result ) > 0 )
      result = Array.get( result, 0 );
    if ( result instanceof Integer )
      return( ((Integer)result).intValue() );
    if ( result instanceof Long )
      return( ((Long)result).intValue() );

    return NULL;
  }

  /**
   * Retrieves value of the given integer-attribute for the given entity (for
   * {@code long} attribute type).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @return Long-integer attribute value.
   */
  public long getLongAttribute ( int attr, int id )
  {
    Object result = fetch( attributes.get( attr ), id, (Object)null );
    if ( result == null ) return NULL;

    if ( result.getClass().isArray() && Array.getLength( result ) > 0 )
      result = Array.get( result, 0 );
    if ( result instanceof Long )
      return( ((Long)result).longValue() );
    if ( result instanceof Integer )
      return( ((Integer)result).longValue() );

    return NULL;
  }

  /**
   * Retrieves value of the given float-attribute for the given entity (for
   * {@code double} attribute type).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @return Double attribute value.
   */
  public double getDoubleAttribute ( int attr, int id )
  {
    Object result = fetch( attributes.get( attr ), id, (Double)null );
    if ( result == null ) return NULL;

    if ( result.getClass().isArray() && Array.getLength( result ) > 0 )
      result = Array.get( result, 0 );
    if ( result instanceof Double )
      return( ((Double)result).doubleValue() );
    if ( result instanceof Float )
      return( ((Float)result).doubleValue() );

    return 0.0;
  }

  /**
   * Retrieves value of the given float-attribute for the given entity (for
   * {@code float} attribute type).
   *
   * @param  attr Handle to the attribute.
   * @param  id Handle to the entity.
   * @return Double attribute value.
   */
  public float getFloatAttribute ( int attr, int id )
  {
    Object result = fetch( attributes.get( attr ), id, (Object)null );
    if ( result == null ) return NULL;

    if ( result.getClass().isArray() && Array.getLength( result ) > 0 )
      result = Array.get( result, 0 );
    if ( result instanceof Float )
      return( ((Float)result).floatValue() );
    if ( result instanceof Double )
      return( ((Double)result).floatValue() );

    return 0.0f;
  }

  //--------------------------------------------------------------------------
  //  implicit attributes (coordinates):

  /**
   * Copies data from coordinate-attribute to another one. Destination attribute is implicitly
   * created.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  ctxTo Handle to a context (can be {@link #NULL NULL} for shared attributes).
   * @param  identTo String identifier of the attribute.
   * @return Handle to a new coordinate-attribute.
   * @see    #createAttribute
   * @see    #roundCoordinates(int,int,String,TrMatrix)
   * @see    #transformCoordinates(int,int,String,TrMatrix)
   */
  public int copyCoordinates ( int attrFrom, int ctxTo, String identTo )
  {
    return copyAttribute( attrFrom, ctxTo, identTo );
  }

  /**
   * Copies data from coordinate-attribute to another one. Destination attribute must exist before,
   * copying can be restricted to the given handle-set.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  attrTo Handle to a destination coordinate attribute.
   * @param  set Entity handle-set (for {@code null} everything will be copied).
   * @see    #createAttribute
   * @see    #roundCoordinates(int,int,TrMatrix,BitSet)
   * @see    #transformCoordinates(int,int,TrMatrix,BitSet)
   */
  public void copyCoordinates ( int attrFrom, int attrTo, BitSet set )
  {
    Attribute from = attributes.get( attrFrom );
    Attribute to   = attributes.get( attrTo );
    if ( from == null ||
         to   == null ||
         from.where != to.where ) return;

    to.copyCoordinates( from, set );
  }

  /**
   * Transforms data from coordinate-attribute to another one. Projection-plane transformation
   * (and roundup) is done implicitly, destination attribute is created implicitly, too.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  ctxTo Handle to a context (can be {@link #NULL NULL} for shared attributes).
   * @param  identTo String identifier of the attribute.
   * @param  tr Linear (projection-plane) transformation.
   * @return Handle to a new coordinate-attribute.
   * @see    #createAttribute
   * @see    #copyCoordinates(int,int,String)
   * @see    #transformCoordinates(int,int,String,TrMatrix)
   */
  public int roundCoordinates ( int attrFrom, int ctxTo, String identTo, TrMatrix tr )
  {
    Attribute from = attributes.get( attrFrom );
    if ( from == null ) return NULL;

    int typeTo;
    switch ( from.type )
    {
      case ATTR_COORD_2D:
      case ATTR_COORD_2F:
      case ATTR_COORDH_3D:
      case ATTR_COORDH_3F:
        typeTo = ATTR_COORD_2I;
        break;

      case ATTR_COORD_3D:
      case ATTR_COORD_3F:
      case ATTR_COORDH_4D:
      case ATTR_COORDH_4F:
        typeTo = ATTR_COORD_3I;
        break;

      default:
        return NULL;
    }

    int attrTo = createAttribute( from.where, ctxTo, identTo, typeTo );
    if ( attrTo == NULL ) return NULL;

    attributes.get( attrTo ).roundCoordinates( from, tr, null );
    return attrTo;
  }

  /**
   * Transforms data from coordinate-attribute to another one. Projection-plane transformation
   * (and roundup) is done implicitly, destination attribute must exist before.
   * Transformation can be restricted to the given handle-set.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  attrTo Handle to a destination coordinate attribute.
   * @param  tr Linear (projection-plane) transformation.
   * @param  set Entity handle-set (for {@code null} everything will be transformed).
   * @see    #createAttribute
   * @see    #copyCoordinates(int,int,BitSet)
   * @see    #transformCoordinates(int,int,TrMatrix,BitSet)
   */
  public void roundCoordinates ( int attrFrom, int attrTo, TrMatrix tr, BitSet set )
  {
    Attribute from = attributes.get( attrFrom );
    Attribute to   = attributes.get( attrTo );
    if ( from == null ||
         to   == null ||
         from.where != to.where ) return;

    to.roundCoordinates( from, tr, set );
  }

  /**
   * Transforms data from coordinate-attribute to another one. Linear transformation
   * is done implicitly, destination attribute is created implicitly, too.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  ctxTo Handle to a context (can be {@link #NULL NULL} for shared attributes).
   * @param  identTo String identifier of the attribute.
   * @param  tr Linear matrix transformation.
   * @return Handle to a new coordinate-attribute.
   * @see    #createAttribute
   * @see    #copyCoordinates(int,int,String)
   * @see    #roundCoordinates(int,int,String,TrMatrix)
   */
  public int transformCoordinates ( int attrFrom, int ctxTo, String identTo, TrMatrix tr )
  {
    Attribute from = attributes.get( attrFrom );
    if ( from == null ) return NULL;

    int attrTo = createAttribute( from.where, ctxTo, identTo, from.type );
    if ( attrTo == NULL ) return NULL;

    attributes.get( attrTo ).transformCoordinates( from, tr, null );
    return attrTo;
  }

  /**
   * Transforms data from coordinate-attribute to another one. Linear transformation
   * is done implicitly, destination attribute must exist before.
   * Transformation can be restricted to the given handle-set.
   *
   * @param  attrFrom Handle to a source coordinate attribute.
   * @param  attrTo Handle to a destination coordinate attribute.
   * @param  tr Linear matrix transformation.
   * @param  set Entity handle-set (for {@code null} everything will be transformed).
   * @see    #createAttribute
   * @see    #copyCoordinates(int,int,BitSet)
   * @see    #roundCoordinates(int,int,TrMatrix,BitSet)
   */
  public void transformCoordinates ( int attrFrom, int attrTo, TrMatrix tr, BitSet set )
  {
    Attribute from = attributes.get( attrFrom );
    Attribute to   = attributes.get( attrTo );
    if ( from == null ||
         to   == null ||
         from.where != to.where ) return;

    to.transformCoordinates( from, tr, set );
  }

  //==========================================================================
  //  Iterators:

  /** Iterator for all vertices in the scene. */
  protected class VertexIterator extends DefaultBrepIteratorOrd
  {
    /**
     * Returns id of the next object.
     *
     * @return {@link Brep#NULL} if there is no next object available.
     */
    public int next ()
    {
      if ( n < used[VERTEX] )
        return n++;

      return NULL;
    }

    /**
     * How many items remain in the set?
     * Non-mandatory - must return {@link Brep#NULL} if not implemented.
     */
    public int left ()
    {
      return( used[VERTEX] - n );
    }
  }

  /**
   * Initializes iterator for all vertices in the scene.
   *
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #vertexInFaceIterator
   * @see    #nextItem
   */
  public BrepIterator vertexIterator ( BrepIterator pre )
  {
    if ( pre != null &&
         pre instanceof VertexIterator )
      pre.reset();
    else
      pre = new VertexIterator();

    return pre;
  }

  /** Iterator for vertices in the given face. */
  protected class VertexInFaceIterator extends DefaultBrepIteratorArr
  {
    /** Face's Id. */
    protected int face;

    public VertexInFaceIterator ( int face )
    {
      init( face );
    }

    /** Can change face's id. */
    public void init ( int face )
    {
      this.face = face;
      arr = null;
      reset();
    }

    /** Reset the iterator. */
    @Override
    public void reset ()
    {
      n = 0;
      if ( arr == null )
      {
          // prepare face-data attribute:
        Attribute a = attrMap.get( FACE, NULL, DATA );
        if ( a == null ) return;
        int[] edges;
        if ( a.type == ATTR_FACE )            // general FACE
          edges = ((int[][])(a.data))[face];
        else                                  // triangle face
        {
          edges = new int[ 3 ];
          System.arraycopy( a.data, face * 3, edges, 0, 3 );
        }

          // edges: array of edges in the given face
        if ( edges == null ||
             edges.length < 2 ) return;

          // prepare edge-data attribute:
        a = attrMap.get( EDGE, NULL, DATA );
        if ( a == null ) return;

          // # of vertices:
        arr = new int[ edges.length ];
          // fill-in vertex handles:
        int prev = edges.length - 1;          // previous edge's handle
        int act  = 0;                         // actual edge's handle
        int i;

        int[] e = (int[])(a.data);            // edge data: vertex handle pairs
        int prev1 = e[ edges[prev] * edgeSize ];
        int prev2 = e[ edges[prev] * edgeSize + 1 ];

        for ( i = 0; i < edges.length; i++ )  // examine one pair of adjacent edges
        {
            // previous edge: [prev1,prev2]
          int actv1 = e[ edges[act] * edgeSize ];
          int actv2 = e[ edges[act] * edgeSize + 1 ]; // actual edge: [actv1,actv2]
          if ( prev1 == actv1 || prev2 == actv1 )
            arr[i] = actv1;
          else
            arr[i] = actv2;
          prev1 = actv1;
          prev2 = actv2;
          prev  = act++;
        }
      }
    }
  }

  /**
   * Initializes iterator for vertices within the given face.
   *
   * @param  face Handle to the face.
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #vertexIterator
   * @see    #nextItem
   */
  public BrepIterator vertexInFaceIterator ( int face, BrepIterator pre )
  {
    if ( pre != null &&
         pre instanceof VertexInFaceIterator )
      ((VertexInFaceIterator)pre).init( face );
    else
      pre = new VertexInFaceIterator( face );

    return pre;
  }

  /** Iterator for all edges in the scene. */
  protected class EdgeIterator extends DefaultBrepIteratorOrd
  {
    /**
     * Returns id of the next object.
     *
     * @return {@link Brep#NULL} if there is no next object available.
     */
    public int next ()
    {
      if ( n < used[EDGE] )
        return n++;

      return NULL;
    }

    /**
     * How many items remain in the set?
     * Non-mandatory - must return {@link Brep#NULL} if not implemented.
     */
    public int left ()
    {
      return( used[EDGE] - n );
    }
  }

  /**
   * Initializes iterator for all edges in the scene.
   *
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #edgeInFaceIterator
   * @see    #nextItem
   */
  public BrepIterator edgeIterator ( BrepIterator pre )
  {
    if ( pre != null &&
         pre instanceof EdgeIterator )
      pre.reset();
    else
      pre = new EdgeIterator();

    return pre;
  }

  /** Iterator for edges in the given face. */
  protected class EdgeInFaceIterator extends DefaultBrepIteratorArr
  {
    /** Face's id. */
    protected int face;

    public EdgeInFaceIterator ( int face )
    {
      init( face );
    }

    public void init ( int face )
    {
      this.face = face;
      arr = null;
      reset();
    }

    /** Reset the iterator. */
    @Override
    public void reset ()
    {
      n = 0;
      if ( arr == null )
      {
          // prepare face-data attribute:
        final Attribute a = attrMap.get( FACE, NULL, DATA );
        if ( a == null ) return;

        if ( a.type == ATTR_FACE )            // general FACE
          arr = ((int[][])(a.data))[face];
        else                                  // triangle face
        {
          arr = new int[ 3 ];
          System.arraycopy( a.data, face * 3, arr, 0, 3 );
        }
      }
    }
  }

  /**
   * Initializes iterator for edges within the given face.
   *
   * @param  face Handle to the face.
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #edgeIterator
   * @see    #nextItem
   */
  public BrepIterator edgeInFaceIterator ( int face, BrepIterator pre )
  {
    if ( pre != null &&
         pre instanceof EdgeInFaceIterator )
      ((EdgeInFaceIterator)pre).init( face );
    else
      pre = new EdgeInFaceIterator( face );

    return pre;
  }

  /** Iterator for all faces in the scene. */
  protected class FaceIterator extends DefaultBrepIteratorOrd
  {
    /**
     * Returns id of the next object.
     *
     * @return {@link Brep#NULL} if there is no next object available.
     */
    public int next ()
    {
      if ( n < used[FACE] )
        return n++;

      return NULL;
    }

    /**
     * How many items remain in the set?
     * Non-mandatory - must return {@link Brep#NULL} if not implemented.
     */
    public int left ()
    {
      return( used[FACE] - n );
    }
  }

  /**
   * Initializes iterator for all faces in the scene.
   *
   * @param  ctx Handle to a context.
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #faceInSolidIterator
   * @see    #nextItem
   */
  public BrepIterator faceIterator ( int ctx, BrepIterator pre )
  {
    if ( pre != null &&
         pre instanceof FaceIterator )
      pre.reset();
    else
      pre = new FaceIterator();

    return pre;
  }

  /** Iterator for faces in the given solid. */
  protected class FaceInSolidIterator extends DefaultBrepIteratorArr
  {
    /** Actual context. */
    protected int ctx;

    /** Solid's id. */
    protected int solid;

    public FaceInSolidIterator ( int ctx, int solid )
    {
      init( ctx, solid );
    }

    /** [Re-]initialize the iterator. */
    public void init ( int ctx, int solid )
    {
      this.ctx   = ctx;
      this.solid = solid;
      arr        = null;
      reset();
    }

    /** Reset the iterator. */
    @Override
    public void reset ()
    {
      n = 0;
      if ( arr == null )
      {
          // prepare solid-data attribute:
        final Attribute a = attrMap.getExt( SOLID, ctx, DATA );
        if ( a == null ) return;

        arr = ((int[][])(a.data))[solid];
      }
    }
  }

  /**
   * Initializes iterator for faces within the given solid.
   *
   * @param  ctx Handle to a context.
   * @param  solid Handle to the face.
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #faceIterator
   * @see    #nextItem
   */
  public BrepIterator faceInSolidIterator ( int ctx, int solid, BrepIterator pre )
  {
    if ( pre != null &&
         pre instanceof EdgeInFaceIterator )
      ((FaceInSolidIterator)pre).init( ctx, solid );
    else
      pre = new FaceInSolidIterator( ctx, solid );

    return pre;
  }

  /** Iterator for all divisions in the scene. */
  protected class DivisionIterator extends DefaultBrepIteratorOrd
  {
    public DivisionIterator ( int ctx )
    {
      init( ctx );
    }

    public void init ( int ctx )
    {
      // ctx is not used yet
      reset();
    }

    /**
     * Returns id of the next object.
     *
     * @return {@link Brep#NULL} if there is no next object available.
     */
    public int next ()
    {
      if ( n < used[DIVISION] )
        return n++;

      return NULL;
    }

    /**
     * How many items remain in the set?
     * Non-mandatory - must return {@link Brep#NULL} if not implemented.
     */
    public int left ()
    {
      return( used[DIVISION] - n );
    }
  }

  /**
   * Initializes iterator for all divisions in the scene.
   *
   * @param  ctx Handle to a context.
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #divisionInSolidIterator
   * @see    #nextItem
   */
  public BrepIterator divisionIterator ( int ctx, BrepIterator pre )
  {
    if ( pre != null &&
         pre instanceof DivisionIterator )
      ((DivisionIterator)pre).init( ctx );
    else
      pre = new DivisionIterator( ctx );

    return pre;
  }

  /**
   * Initializes iterator for divisions within the given solid.
   *
   * @param  ctx Handle to a context.
   * @param  solid Handle to the face.
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #divisionIterator
   * @see    #nextItem
   */
  public BrepIterator divisionInSolidIterator ( int ctx, int solid, BrepIterator pre )
  {
    // !!! TODO: !!!
    return null;
  }

  /** Iterator for all solids in the scene. */
  protected class SolidIterator extends DefaultBrepIteratorOrd
  {
    /**
     * Returns id of the next object.
     *
     * @return {@link Brep#NULL} if there is no next object available.
     */
    public int next ()
    {
      if ( n < used[SOLID] )
        return n++;

      return NULL;
    }

    /**
     * How many items remain in the set?
     * Non-mandatory - must return {@link Brep#NULL} if not implemented.
     */
    public int left ()
    {
      return( used[SOLID] - n );
    }
  }

  /**
   * Initializes iterator for all solids in the scene.
   *
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #nextItem
   */
  public BrepIterator solidIterator ( BrepIterator pre )
  {
    if ( pre != null &&
         pre instanceof SolidIterator )
      pre.reset();
    else
      pre = new SolidIterator();

    return pre;
  }

  /**
   * Returns the next entity-handle.
   *
   * @param  iter Associated iterator data.
   * @return Handle to the next entity or {@link #NULL NULL} at the end of an iteration.
   */
  public int nextItem ( BrepIterator iter )
  {
    if ( iter == null ) return NULL;

    return iter.next();
  }

  //==========================================================================
  //  Basic data access (get* methods):

  /**
   * Returns integer coordinates of the given vertex (implicit attribute
   * {@link #COORD_PROJ COORD_PROJ}) (Cartesian 2D or 3D).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Pre-allocated coordinate vector.
   * @see    #getVertexCoords(int,int,double[])
   * @see    #getVertexCoords(int,int,float[])
   */
  public int[] getVertexCoords ( int ctx, int vert, int[] coord )
  {
    final Attribute a = attrMap.get( VERTEX, ctx, COORD_PROJ );
    if ( a == null ||
         (a.type != ATTR_COORD_2I && a.type != ATTR_COORD_3I) ||
         vert < 0 ||
         vert >= a.size ) return null;

    int s;
    if ( coord == null || coord.length < 2 )
    {
      coord = new int[ s = a.itemSize ];
    }
    else
      s = Math.min( a.itemSize, coord.length );

    System.arraycopy( a.data, vert * a.itemSize, coord, 0, s );
    return coord;
  }

  /**
   * Returns floating-point coordinates of the given vertex (implicit attribute
   * {@link #COORD COORD}).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Optional pre-allocated coordinate vector (Cartesian or homogeneous).
   * @return Result array ({@code coord} if possible).
   * @see    #getVertexCoords(int,int,int[])
   * @see    #getVertexCoords(int,int,float[])
   */
  public double[] getVertexCoords ( int ctx, int vert, double[] coord )
  {
    final Attribute a = attrMap.get( VERTEX, ctx, COORD );
    if ( a == null ||
         (a.type != ATTR_COORD_2D  && a.type != ATTR_COORD_3D &&
          a.type != ATTR_COORDH_3D && a.type != ATTR_COORDH_4D) ||
         vert < 0 ||
         vert >= a.size ) return null;

    int s, tgt;
    if ( coord == null || coord.length < 2 )
    {
      coord = new double[ s = tgt = a.itemSize ];
    }
    else
      s = Math.min( a.itemSize, tgt = coord.length );

    final int ind = vert * a.itemSize;
    System.arraycopy( a.data, ind, coord, 0, s );

    if ( tgt != a.itemSize )
      switch ( a.type )
      {
        case ATTR_COORD_2D:
          if ( tgt > 2 ) coord[2] = 1.0;
          break;

        case ATTR_COORD_3D:
          if ( tgt > 3 ) coord[3] = 1.0;
          break;

        case ATTR_COORDH_3D:
          if ( tgt == 2 )
          {
            double w = ((double[])(a.data))[ ind + 2 ];
            if ( !Geometry.isZero( w ) )
            {
              coord[0] /= w;
              coord[1] /= w;
            }
          }
          break;

        case ATTR_COORDH_4D:
          if ( tgt == 3 )
          {
            double w = ((double[])(a.data))[ ind + 3 ];
            if ( !Geometry.isZero( w ) )
            {
              coord[0] /= w;
              coord[1] /= w;
              coord[2] /= w;
            }
          }
      }

    return coord;
  }

  /**
   * Returns floating-point coordinates of the given vertex (implicit attribute
   * {@link #COORD COORD}).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Optional pre-allocated coordinate vector (Cartesian or homogeneous).
   * @return Result array ({@code coord} if possible).
   * @see    #getVertexCoords(int,int,int[])
   * @see    #getVertexCoords(int,int,double[])
   */
  public float[] getVertexCoords ( int ctx, int vert, float[] coord )
  {
    final Attribute a = attrMap.get( VERTEX, ctx, COORD );
    if ( a == null ||
         (a.type != ATTR_COORD_2F  && a.type != ATTR_COORD_3F &&
          a.type != ATTR_COORDH_3F && a.type != ATTR_COORDH_4F) ||
         vert < 0 ||
         vert >= a.size ) return null;

    int s, tgt;
    if ( coord == null || coord.length < 2 )
    {
      coord = new float[ s = tgt = a.itemSize ];
    }
    else
      s = Math.min( a.itemSize, tgt = coord.length );

    final int ind = vert * a.itemSize;
    System.arraycopy( a.data, ind, coord, 0, s );

    if ( tgt != a.itemSize )
      switch ( a.type )
      {
        case ATTR_COORD_2F:
          if ( tgt > 2 ) coord[2] = 1.0f;
          break;

        case ATTR_COORD_3F:
          if ( tgt > 3 ) coord[3] = 1.0f;
          break;

        case ATTR_COORDH_3F:
          if ( tgt == 2 )
          {
            float w = ((float[])(a.data))[ ind + 2 ];
            if ( !Geometry.isZero( w ) )
            {
              coord[0] /= w;
              coord[1] /= w;
            }
          }
          break;

        case ATTR_COORDH_4F:
          if ( tgt == 3 )
          {
            float w = ((float[])(a.data))[ ind + 3 ];
            if ( !Geometry.isZero( w ) )
            {
              coord[0] /= w;
              coord[1] /= w;
              coord[2] /= w;
            }
          }
      }

    return coord;
  }

  /**
   * Returns vertex handles of the given edge.
   *
   * @param  edge Handle of the edge.
   * @param  vert Optional pre-allocated array to hold two vertex handles.
   * @return Result array ({@code vert} if possible).
   */
  public int[] getEdgeVertices ( int edge, int[] vert )
  {
    final Attribute a = attrMap.get( EDGE, NULL, DATA );
    if ( a == null ||
         a.type != ATTR_EDGE||
         edge < 0 ||
         edge >= a.size ) return null;

    if ( vert == null ||
         vert.length < 2 )
      vert = new int[ 2 ];

    System.arraycopy( a.data, edge * edgeSize, vert, 0, 2 );
    return vert;
  }

  /**
   * Returns record of the given edge (useful particularly for winged-edges).
   * All edge records can be stored in one huge {@code int[]} array, user of this
   * method receives such array and origin of the particular edge record.
   *
   * @param  edge Handle of the edge.
   * @param  index Pre-allocated {@code int[]} array to receive starting index of the edge-record.
   *               If this array is longer than {@code int[1]}, actual record-size will be returned after
   *               the starting index. If {code index} is {@code null}, R/O edge record is returned.
   * @return Shared (read/write) array possibly holding all edge records.
   */
  public int[] getEdgeRecord ( int edge, int[] index )
  {
    final Attribute a = attrMap.get( EDGE, NULL, DATA );
    if ( a == null ||
         a.type != ATTR_EDGE ||
         edge < 0 ||
         edge >= a.size ) return null;

    if ( index == null ||
         index.length < 1 )
    {                                     // return read-only edge record
      index = new int[ edgeSize ];
      System.arraycopy( a.data, edge * edgeSize, index, 0, edgeSize );
      return index;
    }

    index[0] = edge * edgeSize;
    if ( index.length > 1 )
      index[1] = edgeSize;
    return (int[])a.data;
  }

  /**
   * Returns edge handles of the given face.
   *
   * @param  face Handle of the face.
   * @param  edge Pre-allocated array to hold edge handles (or {@code null} to get the
   *              face size).
   * @return Number of face edges.
   * @see    #getFaceVertices
   */
  public int getFaceEdges ( int face, int[] edge )
  {
    final Attribute a = attrMap.get( FACE, NULL, DATA );
    if ( a == null ||
         (a.type != ATTR_FACE && a.type != ATTR_TRIANGLE) ||
         face < 0 ||
         face >= a.size ) return 0;

    if ( a.type == ATTR_FACE )            // general face
    {
      int[] fa = ((int[][])(a.data))[face];
      if ( fa == null ) return 0;
      final int s = fa.length;
      if ( edge != null &&
           edge.length >= s )
        System.arraycopy( fa, 0, edge, 0, s );
      return s;
    }

      // triangle face:
    if ( edge != null &&
         edge.length >=3 )
      System.arraycopy( a.data, face * 3, edge, 0, 3 );
    return 3;
  }

  /**
   * Returns vertex handles of the given face.
   *
   * @param  face Handle of the face.
   * @param  vert Pre-allocated array to hold vertex handles (or {@code null} to get the
   *              face size).
   * @return Number of face vertices.
   * @see    #getFaceEdges
   */
  public int getFaceVertices ( int face, int[] vert )
  {
    final BrepIterator it = vertexInFaceIterator( face, null );
    if ( it == null ) return 0;

    int s = it.left();
    if ( vert != null )
      it.toArray( vert );

    return s;
  }

  /**
   * Returns face handles of the given division.
   *
   * @param  ctx Handle of the context.
   * @param  div Handle of the division.
   * @param  up Pre-allocated array to hold predecessor-face handles (will be terminated by
   *            {@link #NULL NULL}). {@code null} to get number of predecessors.
   * @param  down Pre-allocated array to hold successor-face handles (will be terminated by
   *              {@link #NULL NULL}). {@code null} to get number of successors.
   * @return Number of successor faces (or number of predecessor faces for
   *         {@code up == null}).
   */
  public int getDivisionFaces ( int ctx, int div, int[] up, int[] down )
  {
    // !!! TODO: !!!
    return 0;
  }

  /**
   * Returns face handles of the given solid.
   *
   * @param  ctx Handle of the context.
   * @param  solid Handle of the solid.
   * @param  face Pre-allocated array to hold face handles (or {@code null} to get the
   *              solid size).
   * @return Number of faces.
   */
  public int getSolidFaces ( int ctx, int solid, int[] face )
  {
    final BrepIterator it = faceInSolidIterator( ctx, solid, null );
    if ( it == null ) return 0;

    int s = it.left();
    if ( face != null )
      it.toArray( face );

    return s;
  }

  // global get*() methods (for faster iterations):

  /**
   * Returns all vertex handles in the database.
   *
   * @param  vert Pre-allocated array to hold vertex handles (or {@code null} to get
   *              the array size).
   * @return Array size.
   * @see    #getAllVertexCoords(int,int,int[])
   * @see    #getAllVertexCoords(int,int,double[])
   */
  public int getAllVertices ( int[] vert )
  {
    int s = used[VERTEX];
    if ( vert == null ) return s;

    if ( s > vert.length ) s = vert.length;
    for ( int i = 0; i < s; i++ )
      vert[i] = i;
    return s;
  }

  /**
   * Returns all integer vertex-coordinates in the database (Cartesian - 2D or 3D).
   *
   * @param  attr Handle to the attribute (e.g. {@link #COORD_PROJ COORD_PROJ}).
   * @param  dim Size of coordinate vector ({@code 2} or {@code 3}).
   * @param  coord Pre-allocated array to hold vertex coordinates (or {@code null} to get
   *               the array size).
   * @return Array size (number of vertices).
   * @see    #getAllVertices
   * @see    #getAllVertexCoords(int,int,double[])
   */
  public int getAllVertexCoords ( int attr, int dim, int[] coord )
  {
    // !!! TODO: !!!
    return 0;
  }

  /**
   * Returns all floating-point vertex-coordinates in the database (Cartesian or homogeneous).
   *
   * @param  attr Handle to the attribute (e.g. {@link #COORD COORD}).
   * @param  dim Size of coordinate vector ({@code 2} to {@code 4}).
   * @param  coord Pre-allocated array to hold vertex coordinates (or {@code null} to get
   *               the array size).
   * @return Array size (number of vertices).
   * @see    #getAllVertices
   * @see    #getAllVertexCoords(int,int,int[])
   */
  public int getAllVertexCoords ( int attr, int dim, double[] coord )
  {
    // !!! TODO: !!!
    return 0;
  }

  /**
   * Returns all floating-point vertex-coordinates in the database (Cartesian or homogeneous).
   *
   * @param  attr Handle to the attribute (e.g. {@link #COORD COORD}).
   * @param  dim Size of coordinate vector ({@code 2} to {@code 4}).
   * @param  coord Pre-allocated array to hold vertex coordinates (or {@code null} to get
   *               the array size).
   * @return Array size (number of vertices).
   * @see    #getAllVertices
   * @see    #getAllVertexCoords(int,int,int[])
   */
  public int getAllVertexCoords ( int attr, int dim, float[] coord )
  {
    // !!! TODO: !!!
    return 0;
  }

  /**
   * Returns all edge handles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  edge Pre-allocated array to hold edge handles (or {@code null} to get
   *              the array size).
   * @return Array size.
   * @see    #getAllEdgeVertices
   */
  public int getAllEdges ( int ctx, int[] edge )
  {
    // !!! TODO: !!!
    return 0;
  }

  /**
   * Returns vertex handles of all edges in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  vert Pre-allocated array to hold vertex handles (or {@code null} to get
   *              the array size).
   * @return Array size (number of edges).
   * @see    #getAllEdges
   */
  public int getAllEdgeVertices ( int ctx, int[] vert )
  {
    // !!! TODO: !!!
    return 0;
  }

  /**
   * Returns all face handles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  face Pre-allocated array to hold face handles (or {@code null} to get
   *              the array size).
   * @return Array size.
   */
  public int getAllFaces ( int ctx, int[] face )
  {
    int s = used[FACE];
    if ( face == null ||
         face.length < s ) return s;

    for ( int i = 0; i < s; i++ )
      face[i] = i;
    return s;
  }

  /**
   * Returns edge handles of all triangles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  edge Pre-allocated array to hold edge handles (or {@code null} to get
   *              the array size). It will hold data only for triangle faces.
   * @return Array size (number of faces).
   * @see    #getAllTriangleVertices
   */
  public int getAllTriangleEdges ( int ctx, int[] edge )
  {
    // !!! TODO: !!!
    return 0;
  }

  /**
   * Returns vertex handles of all triangles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  vert Pre-allocated array to hold vertex handles (or {@code null} to get
   *              the array size). It will hold data only for triangle faces.
   * @return Array size (number of faces).
   * @see    #getAllTriangleEdges
   */
  public int getAllTriangleVertices ( int ctx, int[] vert )
  {
    // !!! TODO: !!!
    return 0;
  }

  /**
   * Returns all division handles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  div Pre-allocated array to hold division handles (or {@code null} to get
   *             the array size).
   * @return Array size.
   */
  public int getAllDivisions ( int ctx, int[] div )
  {
    // !!! TODO: !!!
    return 0;
  }

  /**
   * Returns all solid handles in the given context of the database.
   *
   * @param  ctx Handle of the context.
   * @param  solid Pre-allocated array to hold solid handles (or {@code null} to get
   *               the array size).
   * @return Array size.
   */
  public int getAllSolids ( int ctx, int[] solid )
  {
    // !!! TODO: !!!
    return 0;
  }

  // basic data access (create* and set* methods):

  /**
   * Creates a new vertex.
   *
   * @return Handle of the new vertex.
   */
  public int createVertex ()
  {
    int id = used[VERTEX]++;
    adjustAttributes( VERTEX, id+1 );
    return id;
  }

  /**
   * Sets integer coordinates of the given vertex (implicit attribute
   * {@link #COORD_PROJ COORD_PROJ}) (Cartesian 2D or 3D).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Coordinate vector (2D or 3D).
   * @see    #createVertex
   */
  public void setVertexCoords ( int ctx, int vert, int[] coord )
  {
    if ( coord == null ||
         coord.length < 2 ) return;

    final Attribute a = attrMap.getExt( VERTEX, ctx, COORD_PROJ );
    if ( a == null ||
         (a.type != ATTR_COORD_2I && a.type != ATTR_COORD_3I) ||
         vert < 0 ||
         vert >= used[VERTEX] ) return;

    final int s = Math.min( coord.length, a.itemSize );
    final int ind = vert * a.itemSize;
    System.arraycopy( coord, 0, a.data, ind, s );
    if ( s < a.itemSize )
      ((int[])(a.data))[ ind + s ] = 0;
  }

  /**
   * Sets floating-point coordinates of the given vertex (implicit attribute
   * {@link #COORD COORD}) (Cartesian or homogeneous).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Coordinate vector (2D to 4D).
   * @see    #createVertex
   */
  public void setVertexCoords ( int ctx, int vert, double[] coord )
  {
    if ( coord == null ||
         coord.length < 2 ) return;

    final Attribute a = attrMap.getExt( VERTEX, ctx, COORD );
    if ( a == null ||
         (a.type != ATTR_COORD_2D  && a.type != ATTR_COORD_3D &&
          a.type != ATTR_COORDH_3D && a.type != ATTR_COORDH_4D) ||
         vert < 0 ||
         vert >= used[VERTEX] ) return;

    final int s = Math.min( coord.length, a.itemSize );
    final int ind = vert * a.itemSize;
    System.arraycopy( coord, 0, a.data, ind, s );

    if ( s < a.itemSize )
    {
      final double[] arr = (double[])(a.data);
      switch ( a.type )
      {
        case ATTR_COORD_3D:
          arr[ ind + 2 ] = 0.0;
          return;

        case ATTR_COORDH_3D:
          arr[ ind + 2 ] = 1.0;
          return;

        case ATTR_COORDH_4D:
          arr[ ind + 3 ] = 1.0;
          if ( s < 3 )
            arr[ ind + 2 ] = 0.0;
          return;
      }
    }
  }

  /**
   * Sets floating-point coordinates of the given vertex (implicit attribute
   * {@link #COORD COORD}) (Cartesian or homogeneous).
   *
   * @param  ctx Handle of the context.
   * @param  vert Handle of the vertex.
   * @param  coord Coordinate vector (2D to 4D).
   * @see    #createVertex
   */
  public void setVertexCoords ( int ctx, int vert, float[] coord )
  {
    if ( coord == null ||
         coord.length < 2 ) return;
    final Attribute a = attrMap.getExt( VERTEX, ctx, COORD );
    if ( a == null ||
         (a.type != ATTR_COORD_2F  && a.type != ATTR_COORD_3F &&
          a.type != ATTR_COORDH_3F && a.type != ATTR_COORDH_4F) ||
         vert < 0 ||
         vert >= used[VERTEX] ) return;

    final int s = Math.min( coord.length, a.itemSize );
    final int ind = vert * a.itemSize;
    System.arraycopy( coord, 0, a.data, ind, s );

    if ( s < a.itemSize )
    {
      final float[] arr = (float[])(a.data);
      switch ( a.type )
      {
        case ATTR_COORD_3F:
          arr[ ind + 2 ] = 0.0f;
          return;

        case ATTR_COORDH_3F:
          arr[ ind + 2 ] = 1.0f;
          return;

        case ATTR_COORDH_4F:
          arr[ ind + 3 ] = 1.0f;
          if ( s < 3 )
            arr[ ind + 2 ] = 0.0f;
          return;
      }
    }
  }

  /**
   * Creates a new edge.
   *
   * @return Handle of the new edge.
   */
  public int createEdge ()
  {
    int id = used[EDGE]++;
    adjustAttributes( EDGE, id+1 );
    return id;
  }

  /**
   * Sets vertex handles of the given edge.
   *
   * @param  edge Handle of the edge.
   * @param  v1 Handle of the 1st vertex.
   * @param  v2 Handle of the 2nd vertex.
   * @see    #createEdge
   */
  public void setEdgeVertices ( int edge, int v1, int v2 )
  {
    if ( edge < 0 ||
         edge >= used[EDGE] ) return;

    final Attribute a = attrMap.get( EDGE, NULL, DATA );
    if ( a == null ||
         (a.type != ATTR_EDGE) ) return;

    ((int[])(a.data))[ edge * edgeSize     ] = v1;
    ((int[])(a.data))[ edge * edgeSize + 1 ] = v2;
  }

  /**
   * Creates a new face.
   *
   * @return Handle of the new face.
   */
  public int createFace ()
  {
    int id = used[FACE]++;
    adjustAttributes( FACE, id+1 );
    return id;
  }

  /**
   * Sets edge handles of the given triangle.
   *
   * @param  face Handle of the face.
   * @param  e1 Handle of the 1st edge.
   * @param  e2 Handle of the 2nd edge.
   * @param  e3 Handle of the 3rd edge.
   * @see    #createFace
   */
  public void setTriangleEdges ( int face, int e1, int e2, int e3 )
  {
    if ( face < 0 ||
         face >= used[FACE] ) return;

    final Attribute a = attrMap.get( FACE, NULL, DATA );
    if ( a == null ||
         (a.type != ATTR_FACE && a.type != ATTR_TRIANGLE) ) return;

    if ( a.type == ATTR_FACE )            // arbitrary face
    {
      int[] fa = ((int[][])(a.data))[face];
      if ( fa == null || fa.length != 3 )
      {
        fa = new int[ 3 ];
        ((int[][])(a.data))[face] = fa;
      }
      fa[0] = e1;
      fa[1] = e2;
      fa[2] = e3;
    }
    else                                  // triangle
    {
      final int ind = face * 3;
      ((int[])(a.data))[ ind   ] = e1;
      ((int[])(a.data))[ ind+1 ] = e2;
      ((int[])(a.data))[ ind+2 ] = e3;
    }
  }

  /**
   * Sets vertex handles of the given triangle.
   *
   * @param  face Handle of the face.
   * @param  v1 Handle of the 1st vertex.
   * @param  v2 Handle of the 2nd vertex.
   * @param  v3 Handle of the 3rd vertex.
   * @see    #createFace
   */
  public void setTriangleVertices ( int face, int v1, int v2, int v3 )
  {
    if ( face < 0 ||
         face >= used[FACE] ) return;

    int e1 = assertEdge( v1, v2, face );
    int e2 = assertEdge( v2, v3, face );
    int e3 = assertEdge( v3, v1, face );

    setTriangleEdges( face, e1, e2, e3 );
  }

  /**
   * Sets vertex handles of the given face.
   *
   * @param  face Handle of the face.
   * @param  vert Handles of the vertices ({@link #NULL} breaks the sequence}).
   * @see    #createFace
   */
  public void setFaceVertices ( int face, int[] vert )
  {
    if ( face < 0 ||
         face >= used[FACE] ||
         vert == null ||
         vert.length < 3 ) return;

    int len = 0;
    for ( int v : vert )
      if ( v != NULL )
        len++;
      else
        break;

    if ( len < 3 ) return;

    int[] e = new int[ len ];
    int i0, i1;
    for ( i0 = 0; i0 < len; i0++ )
    {
      if ( (i1 = i0 + 1) >= len ) i1 = 0;
      e[i0] = assertEdge( vert[i0], vert[i1], face );
    }

    setFaceEdges( face, e );
  }

  /**
   * Sets edge handles of the given face.
   *
   * @param  face Handle of the face.
   * @param  edge Array of edge handles (can be prematurely terminated by {@link #NULL NULL}).
   * @see    #createFace
   */
  public void setFaceEdges ( int face, int[] edge )
  {
    if ( edge == null ||
         edge.length < 1 ||
         face < 0 ||
         face >= used[FACE] ) return;

    Attribute a = attrMap.get( FACE, NULL, DATA );

    if ( a == null ) return;

    if ( a.type == ATTR_INTEGER )         // winged-edge
    {
      ((int[])(a.data))[face] = edge[0];
      return;
    }

    if ( edge.length < 2 ||
         a.type != ATTR_FACE && (edge.length != 3 || a.type != ATTR_TRIANGLE) ) return;

    if ( a.type == ATTR_FACE )            // arbitrary face
      ((int[][])(a.data))[face] = edge;
    else                                  // triangle
      System.arraycopy( edge, 0, a.data, face * 3, 3 );
  }

  /**
   * Sets face handles of the given solid.
   *
   * @param  ctx Handle of the context.
   * @param  solid Handle of the solid.
   * @param  face Array of face handles (can be prematurely terminated by {@link #NULL NULL}).
   * @see    #createSolid
   */
  public void setSolidFaces ( int ctx, int solid, int[] face )
  {
    if ( face == null ||
         face.length < 2 ||
         solid < 0 ||
         solid >= used[SOLID] ) return;

    Attribute a = attrMap.get( SOLID, ctx, DATA );
    if ( a == null ||
         a.type != ATTR_SOLID ) return;

    ((int[][])(a.data))[solid] = face;
  }

  /**
   * Creates a new division (in the given context).
   *
   * @param  ctx Handle of the context.
   * @return Handle of the new division.
   */
  public int createDivision ( int ctx )
  {
    int id = used[DIVISION]++;
    adjustAttributes( DIVISION, id+1 );
    return id;
  }

  /**
   * Sets face handles of the given division.
   *
   * @param  ctx Handle of the context.
   * @param  div Handle of the division.
   * @param  up Array of predecessor-face handles (can be prematurely terminated by
   *            {@link #NULL NULL}).
   * @param  down Array of successor-face handles (can be prematurely terminated by
   *              {@link #NULL NULL}).
   * @see    #createDivision
   */
  public void setDivisionFaces ( int ctx, int div, int[] up, int[] down )
  {
    // !!! TODO: !!!
  }

  /**
   * Creates a new solid (in the given context).
   *
   * @param  ctx Handle of the context.
   * @return Handle of the new solid.
   */
  public int createSolid ( int ctx )
  {
    if ( ctx != NULL &&
         !usedContexts.get( ctx ) ) return NULL; // invalid context

    if ( getAttributeId( SOLID, ctx, DATA ) == NULL ) // create new attribute array
      createAttribute( SOLID, ctx, DATA, ATTR_SOLID );

    int id = used[SOLID]++;
    adjustAttributes( SOLID, id+1 );
    return id;
  }

  /**
   * Resets the given solid.
   *
   * @param  ctx Handle of the context.
   * @param  solid Handle of the solid.
   * @see    #createSolid
   * @see    #insertFaceIntoSolid
   */
  public void resetSolid ( int ctx, int solid )
  {
    if ( ctx != NULL && !usedContexts.get( ctx ) ||
         solid < 0 ||
         solid >= used[SOLID] ) return;

    final Attribute a = attrMap.getExt( SOLID, ctx, DATA );
    if ( a == null ||
         a.type != ATTR_SOLID ) return;

    ((int[][])(a.data))[solid] = null;
  }

  /**
   * Inserts another face handle into the given solid.
   *
   * @param  ctx Handle of the context.
   * @param  solid Handle of the solid.
   * @param  face Handle of the new face.
   * @see    #resetSolid
   */
  public void insertFaceIntoSolid ( int ctx, int solid, int face )
  {
    if ( ctx != NULL && !usedContexts.get( ctx ) ||
         solid < 0 ||
         solid >= used[SOLID] ||
         face < 0 ||
         face >= used[FACE] ) return;

    final Attribute a = attrMap.getExt( SOLID, ctx, DATA );
    if ( a == null ||
         a.type != ATTR_SOLID ) return;

    ((int[][])(a.data))[solid] = arrayAppend( ((int[][])(a.data))[solid], face );
  }

  //==========================================================================
  //  Module initialization:

  {
    init();
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // PRIMARY_COORD_TYPE:
    if ( key.compareTo(PRIMARY_COORD_TYPE) == 0 )
    {
      int newType = intProperty( value, primaryType, ATTR_COORD_2F, ATTR_COORDH_4D );
      if ( newType != primaryType )
      {
        dirty = true;
        primaryType = newType;
      }
      return;
    }
      // PROJECTED_COORD_TYPE:
    if ( key.compareTo(PROJECTED_COORD_TYPE) == 0 )
    {
      int newType = intProperty( value, primaryProjType, ATTR_COORD_2F, ATTR_COORDH_4D );
      if ( newType != primaryProjType )
      {
        dirty = true;
        primaryProjType = newType;
      }
      return;
    }
      // WEDGE_MODE:
    if ( key.compareTo(WEDGE_MODE) == 0 )
    {
      boolean newMode = booleanProperty( value, edgeSize == WEDGE_SIZE );
      if ( newMode != (edgeSize == WEDGE_SIZE) )
      {
        dirty = true;
        edgeSize = newMode ? WEDGE_SIZE : EDGE_SIZE;
      }
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(PRIMARY_COORD_TYPE) == 0 )
      return primaryType;
    if ( key.compareTo(PROJECTED_COORD_TYPE) == 0 )
      return primaryProjType;
    if ( key.compareTo(WEDGE_MODE) == 0 )
      return (edgeSize == WEDGE_SIZE);

    return null;
  }

  /**
   * Commits all previous changes to the internal object state.
   * Non-mandatory.
   */
  public void commit ()
  {
    if ( dirty )
      init();
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "VEFDS-scene";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "Brep";

  /** Object description. */
  private final static String DESCRIPTION = "Simple Boundary representation of 3D scene - VEFDS lists " +
                                            "(Vertex Edge Face Division Solid).";
  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_DATA + '.' + C_SCENE;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"Brep" );
        // Property:
      t.propBegin( PRIMARY_COORD_TYPE, TYPE_INTEGER, "Primary coordinate type", true );
        t.propDefault( ATTR_COORDH_4D );
        t.propBounds( ATTR_COORD_2F, ATTR_COORDH_4D );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "2f (x,y)",      ATTR_COORD_2F,  "Cartesian float 2D" );
        t.propEnum( "2d (x,y)",      ATTR_COORD_2D,  "Cartesian double 2D" );
        t.propEnum( "2i (x,y)",      ATTR_COORD_2I,  "Cartesian integer 2D" );
        t.propEnum( "h3f (x,y,w)",   ATTR_COORDH_3F, "Homogenous float 2D" );
        t.propEnum( "h3d (x,y,w)",   ATTR_COORDH_3D, "Homogenous double 2D" );
        t.propEnum( "3f (x,y,z)",    ATTR_COORD_3F,  "Cartesian float 3D" );
        t.propEnum( "3d (x,y,z)",    ATTR_COORD_3D,  "Cartesian double 3D" );
        t.propEnum( "3i (x,y,z)",    ATTR_COORD_3I,  "Cartesian integer 3D" );
        t.propEnum( "h4f (x,y,z,w)", ATTR_COORDH_4F, "Homogenous float 3D" );
        t.propEnum( "h4d (x,y,z,w)", ATTR_COORDH_4D, "Homogenous double 3D" );
      t.propEnd();
      t.propBegin( PROJECTED_COORD_TYPE, TYPE_INTEGER, "Projected coordinate type", true );
        t.propDefault( ATTR_COORD_3I );
        t.propBounds( ATTR_COORD_2F, ATTR_COORDH_4D );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "2f (x,y)",      ATTR_COORD_2F,  "Cartesian float 2D" );
        t.propEnum( "2d (x,y)",      ATTR_COORD_2D,  "Cartesian double 2D" );
        t.propEnum( "2i (x,y)",      ATTR_COORD_2I,  "Cartesian integer 2D" );
        t.propEnum( "h3f (x,y,w)",   ATTR_COORDH_3F, "Homogenous float 2D" );
        t.propEnum( "h3d (x,y,w)",   ATTR_COORDH_3D, "Homogenous double 2D" );
        t.propEnum( "3f (x,y,z)",    ATTR_COORD_3F,  "Cartesian float 3D" );
        t.propEnum( "3d (x,y,z)",    ATTR_COORD_3D,  "Cartesian double 3D" );
        t.propEnum( "3i (x,y,z)",    ATTR_COORD_3I,  "Cartesian integer 3D" );
        t.propEnum( "h4f (x,y,z,w)", ATTR_COORDH_4F, "Homogenous float 3D" );
        t.propEnum( "h4d (x,y,z,w)", ATTR_COORDH_4D, "Homogenous double 3D" );
      t.propEnd();
      t.propBegin( WEDGE_MODE, TYPE_BOOLEAN, "Use winged-edge data structure?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
