package cz.cuni.jagrlib.piece;

import java.lang.reflect.*;
import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * WAV sound file filter.
 * Keeps data in disk file, accessed via memory-mapped mode in {@link BitStream}.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/WAVFileFormatMapped.java">WAVFileFormatMapped.java</a>
 */
public class WAVFileFormatMapped extends DefaultFileFormat implements SampleData
{
  //==========================================================================
  //  Data:

  /** Opened bit-stream or {@code null}. */
  protected BitStream stream;

  /** Is the WAV file writable? */
  protected boolean writable;

  /** Actual number of samples (time axis). */
  protected long samples;

  /** Actual number of channels. */
  protected int channels;

  /** Number of bytes in one time-slot. */
  protected int blockAlign;

  /** Actual sample accuracy in bits. */
  protected int bitsPerSample;

  /** Physical sample size in bytes. */
  protected int bytesPerSample;

  /** Sampling frequency in Hz. */
  protected long frequency;

  /** Data offset within the binary file (in bytes). */
  protected long dataOffset;

  /** Was the binary header changed? */
  protected boolean headerChanged;

  /** Actual sample type. */
  protected int sampleType;

  /** Sound duration in seconds. */
  protected double duration;

  /** File properties were changed (for {@link #commit}). */
  protected boolean propChanged;

  //==========================================================================
  //  Construction:

  public WAVFileFormatMapped ()
  {
    samples        = 480000L;
    channels       = 1;
    blockAlign     = 2;
    bitsPerSample  = 16;
    bytesPerSample = 2;
    frequency      = 48000;
    dataOffset     = 44;
    headerChanged  = true;
    sampleType     = SAMPLE_SHORT;
    duration       = 10.0;
    propChanged    = false;
  }

  //==========================================================================
  //  Support:

  protected final static long TAG_RIFF = 0x46464952L;
  protected final static long TAG_WAVE = 0x45564157L;
  protected final static long TAG_FMT  = 0x20746D66L;
  protected final static long TAG_DATA = 0x61746164L;

  /**
   * Reads header of existing WAV file.
   *
   * @return {@code true} if the header was OK.
   */
  protected boolean readHeader ()
  {
    try
    {
      if ( stream == null ||
           stream.seek( 0L ) != 0L ) return false;

      long r = readUnsigned32( stream, true );
      if ( r != TAG_RIFF ) return false;  // "RIFF"

      long left = readUnsigned32( stream, true );

      r = readUnsigned32( stream, true );
      if ( r != TAG_WAVE ) return false;  // "WAVE"

        // Format chunk
      r = readUnsigned32( stream, true );
      if ( r != TAG_FMT ) return false;   // "fmt "

      long len       = readUnsigned32( stream, true );
      int  tmp       = readUnsigned16( stream, true );
      if ( tmp != 1 ) return false;       // PCM = uncompressed

      channels       = readUnsigned16( stream, true );
      if ( channels < 1 || channels > 16 ) return false;

      frequency      = readUnsigned32( stream, true );
      if ( frequency < 1 ) return false;  // frequency

      long byteRate  = readUnsigned32( stream, true );
      blockAlign     = readUnsigned16( stream, true );
      bitsPerSample  = readUnsigned16( stream, true );
      bytesPerSample = (bitsPerSample + 7) >> 3;

        // fmt consistency:
      if ( blockAlign != channels * bytesPerSample ||
           blockAlign * frequency != byteRate )
        return false;

        // extra bits?
      dataOffset = 44;
      left      -= 28;
      if ( (len -= 16) > 0 )
      {
        stream.seek( (36 + len) << 3 );
        left       -= len;
        dataOffset += len;
      }

        // Data chunk:
      r = readUnsigned32( stream, true );
      if ( r != TAG_DATA ) return false;  // "data"

      len     = readUnsigned32( stream, true );
      samples =  len / blockAlign;
      len     = (len + 1) & -2L;          // 16-bit boundary

        // data length consistency:
      if ( (left -= 8) < len ) return false;

      duration = samples / (double)frequency;
    }
    catch ( IOException e )
    {
      return false;
    }

    headerChanged = false;
    return true;
  }

  /**
   * Synchronize disk file header with the actual object state.
   *
   * @return {@code true} if the operation was successfull.
   */
  protected boolean writeHeader ()
  {
    if ( stream == null ) return false;

    if ( !headerChanged ) return true;

    if ( !writable ) return false;

    try
    {
      if ( stream.seek( 0L ) != 0L ) return false;

      writeUnsigned32( stream, TAG_RIFF, true );

      long len = (samples * blockAlign + 1L) & -2L;
                                            // data length rounded up to 16-bit boundary
      writeUnsigned32( stream, dataOffset + len - 8L, true );

      writeUnsigned32( stream, TAG_WAVE, true );
      writeUnsigned32( stream, TAG_FMT, true );
      long extraLen = dataOffset - 44;
      writeUnsigned32( stream, 16 + extraLen, true );

      writeUnsigned16( stream, 1, true );   // PCM coding
      writeUnsigned16( stream, channels, true );
      writeUnsigned32( stream, frequency, true );
      long byteRate = blockAlign * frequency;
      writeUnsigned32( stream, byteRate, true );
      writeUnsigned16( stream, blockAlign, true );
      writeUnsigned16( stream, bitsPerSample, true );

      if ( extraLen > 0L )
      {
        writeUnsigned16( stream, (int)extraLen, true );
        stream.seek( (36 + extraLen) << 3 );
      }

      writeUnsigned32( stream, TAG_DATA, true );
      writeUnsigned32( stream, samples * blockAlign, true );
    }
    catch ( IOException e )
    {
      return false;
    }

    headerChanged = false;
    return true;
  }

  /**
   * Checks array's component type.
   *
   * @return Array size or {@code -1} if array type is not compatible.
   */
  protected int checkComponentClass ( Object array, boolean allowDouble )
  {
    Class<?> cl, comp;

      // check data array type:
    if ( array == null ||
       !(cl = array.getClass()).isArray() ) return -1;

    if ( !(comp = cl.getComponentType()).isPrimitive() ) return -1;

    if ( comp.equals( Byte.TYPE ) ||
         comp.equals( Short.TYPE ) ||
         comp.equals( Integer.TYPE ) )
      return Array.getLength( array );

    if ( allowDouble &&
         comp.equals( Double.TYPE ) )
      return Array.getLength( array );

    return -1;
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return 16;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type ({@code 0.0} .. no way,
   *         {@code 0.5} .. maybe, {@code 0.9} .. almost sure, {@code 1.0} ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int len = (header == null) ? 0 : header.length;
    if ( len > 16 ) len = 16;

    if ( len < 16 )                       // not enough bytes in header[] => look at file-name
    {
      if ( fileName == null || fileName.length() < 5 ) return 0.0;
      int pos = fileName.lastIndexOf( '.' );
      if ( pos < 0 ) return 0.0;

      String suffix = fileName.substring( pos );
      if ( suffix.compareToIgnoreCase( "wav" ) == 0 ) return 0.9;

      return 0.0;
    }

    byte[] h = new byte[]
    {                                     // RIFF/WAV header
      (byte)'R', (byte)'I', (byte)'F', (byte)'F',
      0,         0,         0,         0,
      (byte)'W', (byte)'A', (byte)'V', (byte)'E',
      (byte)'f', (byte)'m', (byte)'t', (byte)' '
    };

    int i;
    for ( i = 0; i < len; i++ )
      if ( h[i] != 0 && header[i] != h[i] ) return 0.0;

    return 1.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.png"]) or {@code null}.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.wav",
    };
  }

  /**
   * Retrieves data from the given file (stream).
   * Can be used for "file opening" in virtual-mapped mode.
   *
   * @param  fileName File-name (stream-name) to read from (can be {@code null}).
   * @param  detail Further implementation-specific information ("r" or "rw" for mapped mode).
   * @return {@code true} if the operation was successful.
   * @see    #loadFile(BitStream,Object)
   * @see    #saveFile(String,String)
   * @see    #closeFile
   */
  @Override
  public boolean loadFile ( String fileName, String detail )
  throws IOException
  {
    if ( stream != null ) stream.close();

    try
    {
      stream = (BitStream)getInterface( PL_STREAM, IFACE+"BitStream" );
      if ( stream == null )
        return false;

      writable = detail != null &&
                 detail.length() > 1 &&
                 detail.charAt( 1 ) == 'w';

      stream.open( writable, fileName, detail );
      if ( readHeader() ) return true;

      return( writable && writeHeader() );
    }
    catch ( JaGrLibException e )
    {
      LogFile.exception( e );
    }

    return false;
  }

  /**
   * Retrieves data from the given {@link BitStream}. This metod is optional.
   * <p>The stream should be opened before - only the {@link BitStream#reOpen} call
   * can be used.
   *
   * @param  stream Input bit-stream (opened - call {@link BitStream#reOpen} first).
   * @param  support Implementation-specific data (e.g. second {@link BitStream} for
   *                 file-formats with {@link #DOUBLE_STREAM} property).
   * @return {@code true} if the operation was successful.
   * @see    #loadFile(String,String)
   * @see    #DOUBLE_STREAM
   */
  @Override
  public boolean loadFile ( BitStream stream, Object support )
  throws IOException
  {
    if ( this.stream != null ) this.stream.close();

    this.stream = stream;

    if ( readHeader() ) return true;

    return( writable && writeHeader() );
  }

  /**
   * Saves data to the given file (stream).
   *
   * @param  fileName File-name (stream-name) to write to (can be {@code null}).
   * @param  detail Further implementation-specific information.
   * @return {@code true} if the operation was successful.
   * @see    #saveFile(BitStream,Object)
   * @see    #loadFile(String,String)
   */
  @Override
  public boolean saveFile ( String fileName, String detail )
  {
    return false;
  }

  /**
   * Saves data to the given {@link BitStream}. This metod is optional.
   * <p>The stream should be opened and empty - the {@code reOpen(true,null)} call
   * is not required.
   *
   * @param  stream Output bit-stream (opened and empty).
   * @param  support Implementation-specific data (e.g. second {@link BitStream} for
   *                 file-formats with {@link #DOUBLE_STREAM} property).
   * @return {@code true} if the operation was successful.
   * @see    #saveFile(String,String)
   * @see    #DOUBLE_STREAM
   */
  @Override
  public boolean saveFile ( BitStream stream, Object support )
  throws IOException
  {
    return false;
  }

  /**
   * Closes the file (used in file-mapping mode).
   *
   * @see    #loadFile(String,String)
   */
  public void closeFile ()
  throws IOException
  {
    if ( stream == null ) return;

    writeHeader();
    stream.close();
    stream = null;
  }

  //==========================================================================
  //  Interface SampleData:

  /**
   * (Re-)initialize the data array.
   *
   * @param  type Sample type (one of {@code SAMPLE_*} constants).
   * @param  mapped Should the array contain mapped values?
   * @param  dim Dimensions of an array.
   * @see    #SAMPLE_BYTE
   * @see    #SAMPLE_SHORT
   * @see    #SAMPLE_INT
   * @see    #SAMPLE_LONG
   * @see    #SAMPLE_FLOAT
   * @see    #SAMPLE_DOUBLE
   * @see    #SAMPLE_COMPLEX
   * @see    #SAMPLE_RGBA
   */
  public void resetData ( int type, boolean mapped, int[] dim )
  {
    if ( !writable ||
         dim == null ||
         dim.length == 0 ||
         dim.length > 2 ) return;

    headerChanged = true;

    channels = (dim.length == 1) ? 1 : dim[1];
    samples  = dim[0];

    switch ( sampleType = type )
    {
      case SAMPLE_BYTE:                   // 8-bit unsigned
        bitsPerSample  =  8;
        bytesPerSample =  1;
        break;

      case SAMPLE_INT:                    // 32-bit signed
        bitsPerSample  = 32;
        bytesPerSample =  4;
        break;

      case SAMPLE_SHORT:                  // 16-bit signed
      default:
        sampleType     = SAMPLE_SHORT;
        bitsPerSample  = 16;
        bytesPerSample =  2;
    }

    blockAlign = bytesPerSample * channels;
    duration   = samples / (double)frequency;
  }

  /**
   * Returns actual sample type ({@code SAMPLE_*} constant).
   *
   * @return Sample type ({@code SAMPLE_*} constant).
   * @see    #getDataDim
   * @see    #SAMPLE_BYTE
   * @see    #SAMPLE_SHORT
   * @see    #SAMPLE_INT
   * @see    #SAMPLE_LONG
   * @see    #SAMPLE_FLOAT
   * @see    #SAMPLE_DOUBLE
   * @see    #SAMPLE_COMPLEX
   * @see    #SAMPLE_RGBA
   */
  public int getSampleType ()
  {
    return sampleType;
  }

  /**
   * Returns actual array dimensions.
   *
   * @param  dim Pre-allocated array to hold individual dimensions' sizes.
   * @return Dimensionality (number of dimensions).
   * @see    #getSampleType
   */
  public int getDataDim ( int[] dim )
  {
    int dims = (channels > 1) ? 2 : 1;

    if ( dim != null &&
         dim.length >= dims )
    {
      dim[0] = (int)samples;
      if ( dims > 1 ) dim[1] = channels;
    }

    return dims;
  }

  /**
   * Sets the sample geometry (in regular sample matrix).
   *
   * @param geom Sample periods for all dimensions.
   * @see   #getSampleGeometry
   */
  public void setSampleGeometry ( double[] geom )
  {
    if ( geom == null ||
         geom.length < 1 ||
         geom[0] < Geometry.EPSILON ) return;

    int newF = Formula.round( 1.0 / geom[0] );
    if ( newF != frequency )
    {
      propChanged = true;
      if ( writable )
        headerChanged = true;
      frequency = newF;
    }
  }

  /**
   * Gets the sample geometry (in regular sample matrix).
   *
   * @param geom Sample periods for all dimensions.
   * @see   #setSampleGeometry
   */
  public void getSampleGeometry ( double[] geom )
  {
    if ( geom == null ||
         geom.length < 1 ) return;

    geom[0] = (frequency == 0) ? 1.0 : 1.0 / frequency;
  }

  /**
   * Sets (partially) the new sample mapping (for mapped arrays only).
   *
   * @param map Mapping array to be set (should be an array of the apropriate type).
   * @param srcFrom First source index ({@code map[from]}).
   * @param srcTo After the last source index.
   * @param dstFrom First destination index.
   * @see   #getSampleMapping
   */
  public void setSampleMapping ( Object map, int srcFrom, int srcTo, int dstFrom )
  {
  }

  /**
   * Retrieves the whole sample mapping (for mapped arrays only).
   *
   * @param map Optional pre-allocated mapping array to be filled.
   * @param srcFrom First index to be retrieved.
   * @param srcTo After the last index to be retrieved.
   * @param bounds Retrieved index bounds ({@code bounds[0]} = the 1st mapped value
   *               written in map[0], {@code bounds[1]} = index after the last value).
   * @see   #setSampleMapping
   */
  public Object getSampleMapping ( Object map, int srcFrom, int srcTo, int[] bounds )
  {
    return null;
  }

  /**
   * Returns the original data array.
   *
   * @param  dim Returned dimensionality (array of individual dimension sizes). Can be null.
   * @return Packed 1D sample array (read / write access) or {@code null} if the service is not
   *         available.
   * @see    #setDataArray
   */
  public Object getDataArray ( int[] dim )
  {
    return null;
  }

  /**
   * Sets data array in the raw (internal) format. Can be used for data fragment
   * or for the whole data array.
   *
   * @param data Packed 1D source data array (of apropriate type).
   * @param srcFrom Source index to begin from.
   * @param dstFrom Start of the region to be set. Can be null (instead of [0,0,0,..]).
   * @param dstTo End of the region to be set. Can be null (instead of dim).
   */
  public void setDataArray ( Object data, int srcFrom, int[] dstFrom, int[] dstTo )
  {
    long sampleFrom  = 0L;
    int  channelFrom = 0;
    if ( dstFrom != null &&
         dstFrom.length >= 1 )
    {
      sampleFrom = dstFrom[0];
      if ( dstFrom.length >= 2 )
        channelFrom = dstFrom[1];
    }

    long sampleTo  = samples;
    int  channelTo = channels;
    if ( dstTo != null &&
         dstTo.length >= 1 )
    {
      sampleTo = dstTo[0];
      if ( dstTo.length >= 2 )
        channelTo = dstTo[1];
    }

      // check data array type:
    int srcLen = checkComponentClass( data, false );
    if ( srcLen < srcFrom + (int)(sampleTo - sampleFrom) * (channelTo - channelFrom) )
      return;

    long sa;
    int  ch;
    long offset0 = dataOffset + channelFrom * bytesPerSample;

    try
    {
      switch ( sampleType )
      {
        case SAMPLE_BYTE:
          for ( sa = sampleFrom; sa < sampleTo; sa++ )
          {
            stream.seek( (offset0 + sa * blockAlign) << 3 );
            for ( ch = channelFrom; ch++ < channelTo; )
              stream.write( Formula.clamp( Array.getInt( data, srcFrom++ ), -128, 127 ) + 128, 8 );
          }
          break;

        case SAMPLE_SHORT:
          for ( sa = sampleFrom; sa < sampleTo; sa++ )
          {
            stream.seek( (offset0 + sa * blockAlign) << 3 );
            for ( ch = channelFrom; ch++ < channelTo; )
              stream.write( Formula.clamp( Array.getInt( data, srcFrom++ ), Short.MIN_VALUE, Short.MAX_VALUE ), 16 );
          }
          break;

        default:
          for ( sa = sampleFrom; sa < sampleTo; sa++ )
          {
            stream.seek( (offset0 + sa * blockAlign) << 3 );
            for ( ch = channelFrom; ch++ < channelTo; )
              stream.write( Array.getInt( data, srcFrom++ ), 32 );
          }
      }
    }
    catch ( IOException e )
    {}
  }

  /**
   * Reads the given sample segment (general data format).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result (packed 1D array).
   * @return Number of returned samples.
   */
  public int getSamplesGeneral ( int[] from, int[] to, Object data )
  {
    long sampleFrom  = 0;
    int  channelFrom = 0;
    if ( from != null &&
         from.length >= 1 )
    {
      sampleFrom = from[0];
      if ( from.length >= 2 )
        channelFrom = from[1];
    }

    long sampleTo  = samples;
    int  channelTo = channels;
    if ( to != null &&
         to.length >= 1 )
    {
      sampleTo = to[0];
      if ( to.length >= 2 )
        channelTo = to[1];
    }

      // check data array type:
    int srcLen = checkComponentClass( data, true );
    if ( srcLen < (int)(sampleTo - sampleFrom) * (channelTo - channelFrom) )
      return -1;

    long sa;
    int  ch, ptr = 0;
    long offset0 = dataOffset + channelFrom * bytesPerSample;

    try
    {
      switch ( sampleType )
      {
        case SAMPLE_BYTE:
          for ( sa = sampleFrom; sa < sampleTo; sa++ )
          {
            stream.seek( (offset0 + sa * blockAlign) << 3 );
            for ( ch = channelFrom; ch++ < channelTo; )
              Array.setByte( data, ptr++, (byte)( stream.read( 8 ) - 128 ) );
          }
          break;

        case SAMPLE_SHORT:
          for ( sa = sampleFrom; sa < sampleTo; sa++ )
          {
            stream.seek( (offset0 + sa * blockAlign) << 3 );
            for ( ch = channelFrom; ch++ < channelTo; )
              Array.setShort( data, ptr++, (short)stream.read( 16 ) );
          }
          break;

        default:
          for ( sa = sampleFrom; sa < sampleTo; sa++ )
          {
            stream.seek( (offset0 + sa * blockAlign) << 3 );
            for ( ch = channelFrom; ch++ < channelTo; )
              Array.setInt( data, ptr++, (int)stream.read( 32 ) );
          }
      }
    }
    catch ( IOException e )
    {}

    return ptr;
  }

  /**
   * Writes the given sample segment (general data format).
   *
   * @param from Starting index.
   * @param to Index after the written data.
   * @param data Data to be written - in 1D array.
   */
  public void setSamplesGeneral ( int[] from, int[] to, Object data )
  {
    setDataArray( data, 0, from, to );
  }

  /**
   * Reads the given sample segment (integer formats: byte, short, int).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result.
   * @return Number of returned samples.
   */
  public int getSamples ( int[] from, int[] to, int[] data )
  {
    return getSamplesGeneral( from, to, data );
  }

  /**
   * Reads the given sample segment (long format).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result.
   * @return Number of returned samples.
   */
  public int getSamples ( int[] from, int[] to, long[] data )
  {
    return getSamplesGeneral( from, to, data );
  }

  /**
   * Reads the given sample segment (float formats: float, double).
   *
   * @param from Starting index.
   * @param to Index after retrieved data.
   * @param data Pre-allocated array to hold the result (complex uses double-sized array).
   * @return Number of returned samples.
   */
  public int getSamples ( int[] from, int[] to, double[] data )
  {
    int len = getSamplesGeneral( from, to, data );

    if ( len > 0 )
    {
      double mul = 1.0;
      switch ( sampleType )
      {
        case SAMPLE_BYTE:
          mul = -1.0 / Byte.MIN_VALUE;
          break;
        case SAMPLE_INT:
          mul = -1.0 / Integer.MIN_VALUE;
          break;
        default:
          mul = -1.0 / Short.MIN_VALUE;
      }

      for ( int i = 0; i < len; )
        data[i++] *= mul;
    }

    return len;
  }

  /**
   * Writes the given sample segment (integer formats: byte, short, int).
   *
   * @param from Starting index.
   * @param to Index after written data.
   * @param data Data to be written (1D array).
   */
  public void setSamples ( int[] from, int[] to, int[] data )
  {
    setDataArray( data, 0, from, to );
  }

  /**
   * Writes the given sample segment (long format).
   *
   * @param from Starting index.
   * @param to Index after written data.
   * @param data Data to be written (1D array).
   */
  public void setSamples ( int[] from, int[] to, long[] data )
  {
    // !!! TODO: not implemented yet !!!
  }

  /**
   * Writes the given sample segment (float formats: float, double).
   *
   * @param from Starting index.
   * @param to Index after written data.
   * @param data Data to be written (1D array). Complex uses double-sized array.
   */
  public void setSamples ( int[] from, int[] to, double[] data )
  {
    // !!! TODO: not implemented yet !!!
  }

  //==========================================================================
  //  Interface Property:

  public final static String CHANNELS  = "Channels";
  public final static String DURATION  = "Duration";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
    int newValue;
      // SAMPLE_TYPE:
    if ( key.compareTo(SAMPLE_TYPE) == 0 )
    {
      newValue = intProperty( value, sampleType, SAMPLE_BYTE, SAMPLE_INT );
      if ( newValue != sampleType )
        propChanged = true;
      sampleType = newValue;
      return;
    }
      // CHANNELS:
    if ( key.compareTo(CHANNELS) == 0 )
    {
      newValue = intProperty( value, channels, 1, 16 );
      if ( newValue != channels )
        propChanged = true;
      channels = newValue;
      return;
    }
      // FREQUENCY:
    if ( key.compareTo(FREQUENCY) == 0 )
    {
      newValue = intProperty( value, (int)frequency );
      if ( newValue != (int)frequency )
        propChanged = true;
      frequency = newValue;
      return;
    }
      // DURATION:
    if ( key.compareTo(DURATION) == 0 )
    {
      double dur = doubleProperty( value, duration, 0.0, 1.e9 );
      if ( dur != duration )
        propChanged = true;
      duration = dur;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(SAMPLE_TYPE) == 0 )
      return sampleType;
    if ( key.compareTo(CHANNELS) == 0 )
      return channels;
    if ( key.compareTo(FREQUENCY) == 0 )
      return (int)frequency;
    if ( key.compareTo(DURATION) == 0 )
      return duration;
    return null;
  }

  /**
   * Commits all previous changes ({@link #set}) to the internal object state.
   * Non-mandatory.
   */
  public void commit ()
  {
    if ( propChanged )
    {
      samples = Math.round( duration * frequency );
      if ( stream != null )
      {
        headerChanged = true;
        writeHeader();
      }
      propChanged = false;
    }
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "WAV file format";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "DataFileFormatAndSampleData";

  /** Object description. */
  private final static String DESCRIPTION = "WAV sound file filter (data are stored in disk file).";

  /** Object category. */
  protected final static String CATEGORY = C_IO + "." + C_1D + "." + C_DATA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"SampleData" );
      t.newInputPlug(  PL_DATA,   IFACE+"DataFileFormat" );
      t.newOutputPlug( PL_STREAM, IFACE+"BitStream" );
        // Property:
      t.propBegin( SAMPLE_TYPE, TYPE_INTEGER, "Data item type", true );
        t.propDefault( SAMPLE_INT );
        t.propBounds( SAMPLE_BYTE, SAMPLE_INT );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Byte",  SAMPLE_BYTE,  "byte (8-bit unsigned int)" );
        t.propEnum( "Short", SAMPLE_SHORT, "short (16-bit signed int)" );
        t.propEnum( "Int",   SAMPLE_INT,   "int (32-bit signed int)" );
      t.propEnd();
      t.propBegin( CHANNELS, TYPE_INTEGER, "Number of channels", true );
        t.propDefault( 1 );
        t.propBounds(  1, 16 );
      t.propEnd();
      t.propBegin( FREQUENCY, TYPE_INTEGER, "Frequency in Hz", true );
        t.propDefault( 48000 );
      t.propEnd();
      t.propBegin( DURATION, TYPE_DOUBLE, "Sound duration in seconds", true );
        t.propDefault( 10.0 );
        t.propBounds(  0.0, 1.e9 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
