package cz.cuni.jagrlib.piece;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Internal X-transition representation.
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.03
 */
class XItem
{
  /** X coordinate of a "transition" (=change). */
  int x;

  /**
   * Initializing constructor.
   * @param  _x Initial X coordinate of the transition.
   */
  XItem ( int _x )
  {
    x = _x;
  }
}

/**
 * This type can be used for array creation..
 * (unlike LinkedList<XItem>)
 */
class XLine extends LinkedList<XItem>
{
  private static final long serialVersionUID = 1L;

  public XItem[] toXArray ()
  {
    return toArray( new XItem[0] );
  }
}

/**
 * Bit-mask represented by "X-transition list" data structure.
 *
 * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.03
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/XTransitionList.java">XTransitionList.java</a>
 */
public class XTransitionList extends Piece implements BitMask
{
  //==========================================================================
  //  Data:

  /**
   * Special Enumerator for {@link XTransitionList}.
   *
   * @version 0.25 $Rev: 281 $ $Date: 2006-06-01 20:55:34 +0200 (čt, 01 VI 2006) $ $Author: pepca $
   * @since   0.03
   */
  public class XTransitionListEnumerator implements BitMaskEnumerator
  {
    /** Actual X coordinate. */
    protected int x;

    /** Finishing X coordinate. */
    protected int xEnd;

    /** Actual Y coordinate. */
    protected int y;

    /** Actual scanline-array (of {@link XItem}). */
    protected XItem[] scan;

    /** <tt>scan</tt> array pointer (). */
    protected int i;

    /** Default constructor (initializes the enumerator). */
    XTransitionListEnumerator ()
    {
      reset();
    }

    /**
     * (re-)initialization of the enumerator.
     */
    public void reset ()
    {
      x = BitMaskEnumerator.MINUS_INFINITY;
      y = yMin;
      scan = null;
      i = 0;
      while ( y < yMax && scanlines[y-yMin].isEmpty() ) y++;
      if ( y >= yMax ) return;              // no pixel is available..

      scan = scanlines[y-yMin].toXArray();
      i = 0;
      x = scan[0].x;
      xEnd = (1 < scan.length) ? scan[1].x : BitMaskEnumerator.PLUS_INFINITY;
    }

    /**
     * Finds the next positive span.
     * @return <tt>true</tt> if any pixels are available.
     */
    protected boolean findNext ()
    {
      if ( x < xEnd )
      {                                     // at least one pixel is available in the current span
        if ( xEnd == BitMaskEnumerator.PLUS_INFINITY ) x = xEnd;
        return true;
      }

      i += 2;
      if ( i >= scan.length )
      {                                     // find the new non-empty scanline
        scan = null;
        do y++;
        while ( y < yMax && scanlines[y-yMin].isEmpty() );
        if ( y >= yMax ) return false;      // no more pixels available..
        scan = scanlines[y-yMin].toArray( new XItem[0] );
        i = 0;
      }
      x = scan[i].x;
      xEnd = (i+1 < scan.length) ? (scan[i+1]).x : BitMaskEnumerator.PLUS_INFINITY;
      return true;
    }

    /**
     * Returns the next {@link BitMask} ({@link AlphaMask}) pixel.
     *
     * @param  pre Pre-allocated <tt>BitMaskEnumerator.Pixel</tt> instance (non-mandatory).
     * @return The next opaque <tt>BitMaskEnumerator.Pixel</tt> or <tt>null</tt>.
     * @see    #nextHLine
     */
    public BitMaskEnumerator.Pixel nextPixel ( BitMaskEnumerator.Pixel pre )
    {
      if ( scan == null ) return null;

      if ( pre == null ) pre = new BitMaskEnumerator.Pixel();
      pre.x     = x++;
      pre.y     = y;
      pre.alpha = 1.0;
      findNext();
      return pre;
    }

    /**
     * Returns the next {@link BitMask} ({@link AlphaMask}) H-line.
     *
     * @param  pre Pre-allocated <tt>BitMaskEnumerator.HLine</tt> instance (non-mandatory).
     * @return The next opaque <tt>BitMaskEnumerator.HLine</tt> or <tt>null</tt>.
     * @see    #nextPixel
     */
    public BitMaskEnumerator.HLine nextHLine ( BitMaskEnumerator.HLine pre )
    {
      if ( scan == null ) return null;

      if ( pre == null ) pre = new BitMaskEnumerator.HLine();
      pre.x1 = x;
      x =
      pre.x2 = xEnd;
      pre.y  = y;
      findNext();
      return pre;
    }

  }

  /** Minimum scanline coordinate. */
  protected int yMin;

  /** Maximum scanline coordinate + 1. */
  protected int yMax;

  /** Array of scanlines (they are holding {@link XItem}s). */
  protected XLine[] scanlines;

  //--------------------------------------------------------------------------
  //  State variables:

  /** Actual binary set operation. */
  protected int binaryOp = BinaryOperation.BIN_OP_OR;

  //--------------------------------------------------------------------------
  //  Support routines:

  /**
   * Asserts that the given scanline is included in the set.
   * @param  y Y coordinate to be included.
   */
  protected void assertY ( int y )
  {
    XLine[] sc;
    int i;
    if ( y < yMin )
    {                                       // extend scanlines[] at the beginning
      sc = new XLine[ yMax - y ];
      System.arraycopy( scanlines, 0, sc, yMin - y, yMax - yMin );
      for ( i = 0; i < yMin-y; )
        sc[i++] = new XLine();
      scanlines = sc;
      yMin = y;
    }
    else
    if ( y >= yMax )
    {                                       // extend scanlines[] at the end
      sc = new XLine[ y + 1 - yMin ];
      System.arraycopy( scanlines, 0, sc, 0, yMax - yMin );
      for ( i = yMax-yMin; i < y+1-yMin; )
        sc[i++] = new XLine();
      scanlines = sc;
      yMax = y + 1;
    }
  }

  /**
   * Sets the given scanline-set (using {@link #binaryOp} operation).
   *
   * @param  l The scanline-set.
   * @param  y Y coordinate of the scanline.
   * @return The new X-transition list.
   */
  protected XLine setList ( XLine l, int y )
  {
    XLine result = new XLine();
    ListIterator<XItem> lIt = l.listIterator();
    int lx = lIt.hasNext() ? lIt.next().x : BitMaskEnumerator.PLUS_INFINITY;
    boolean lIn = false;                    // inside the "l" set?
    ListIterator<XItem> myIt = scanlines[y-yMin].listIterator();
    int myx = myIt.hasNext() ? myIt.next().x : BitMaskEnumerator.PLUS_INFINITY;
    boolean myIn = false;                   // inside the "my" set?
    boolean resIn = BinaryOperation.boolOp( binaryOp, false, false );
    do
    {                                       // process the leftmost XItem
      boolean advL, advMy;
      int x = myx;
      if ( (advL = lx <= x) )
      {                                     // item in "l" is the nearest
        x = lx;
        lIn = !lIn;
      }
      if ( (advMy = myx <= x) )             // item in "my" is the nearest
        myIn = !myIn;
      boolean newRes = BinaryOperation.boolOp( binaryOp, myIn, lIn );
      if ( resIn != newRes )
      {
        result.add( new XItem( x ) );
        resIn = newRes;
      }
      if ( advL )                           // advance in "l" list
        lx = lIt.hasNext() ? lIt.next().x : BitMaskEnumerator.PLUS_INFINITY;
      if ( advMy )                          // advance in "my" list
        myx = myIt.hasNext() ? myIt.next().x : BitMaskEnumerator.PLUS_INFINITY;
    } while ( lx  < BitMaskEnumerator.PLUS_INFINITY ||
              myx < BitMaskEnumerator.PLUS_INFINITY );

    return result;
  }

  /**
   * Sets the given scanline-set (<b>hard-set</b> .. by-passes the <tt>binaryOp</tt>
   * operation) within the given bounds (from <tt>x1</tt> to <tt>x2-1</tt>).
   *
   * @param  l The scanline-set.
   * @param  y Y coordinate of the scanline.
   * @param  x1 X coordinate of the scanline-set begin.
   * @param  x2 X coordinate of the first point after the scanline-set.
   * @return The new X-transition list.
   */
  protected XLine setList ( XLine l, int y, int x1, int x2 )
  {
    XLine result = new XLine();

    ListIterator<XItem> lIt = l.listIterator();
    int lx = lIt.hasNext() ? lIt.next().x : BitMaskEnumerator.PLUS_INFINITY;
    boolean lIn  = false;                   // inside the "l" set?

    ListIterator<XItem> myIt = scanlines[y-yMin].listIterator();
    int myx = myIt.hasNext() ? myIt.next().x : BitMaskEnumerator.PLUS_INFINITY;
    boolean myIn  = false;                  // inside the "my" set?

    boolean waitx1 = true;                  // waiting for "x1"
    boolean waitx2 = true;                  // waiting for "x2"
    boolean resIn  = false;

      // list processing loop:
    while ( true )
    {                                       // process the leftmost XItem
      int x = waitx1 ? x1 : (waitx2 ? x2 : BitMaskEnumerator.PLUS_INFINITY);
                                            // have to wait for both x1 and x2
      if ( lx < x ) x = lx;                 // the nearest of "l". "my" and [x1,x2] segment
      if ( myx < x ) x = myx;
      if ( x >= BitMaskEnumerator.PLUS_INFINITY ) break;

      boolean advL, advMy;
      if ( (advL = lx <= x) )               // item in "l" is the nearest
        lIn = !lIn;
      if ( (advMy = myx <= x) )             // item in "my" is the nearest
        myIn = !myIn;
      if ( x1 == x ) waitx1 = false;
      if ( x2 == x ) waitx2 = false;
      boolean newRes = (waitx1 == waitx2) ? myIn : lIn;
      if ( resIn != newRes )
      {
        result.add( new XItem( x ) );
        resIn = newRes;
      }
      if ( advL )                           // advance in "l" list
        lx = lIt.hasNext() ? lIt.next().x : BitMaskEnumerator.PLUS_INFINITY;
      if ( advMy )                          // advance in "my" list
        myx = myIt.hasNext() ? myIt.next().x : BitMaskEnumerator.PLUS_INFINITY;
    }

    return result;
  }

  /**
   * Finds the given X coordinate in the given <tt>List</tt>. Modifies <tt>ListIterator</tt>
   * such that the <tt>XItem</tt> found (it.x >= x) is the next to be processed.
   *
   * @param  x Boundary X coordinate
   * @param  li The initialized <tt>ListIterator</tt>.
   * @return The <b>inside-status</b> of the set before the next {@link XItem}.
   */
  protected static boolean find ( int x, ListIterator<XItem> li )
  {
    boolean inside = false;

    while ( li.hasNext() )
    {
      XItem it = li.next();
      if ( it.x >= x )
      {
        li.previous();
        return inside;
      }
      inside = !inside;
    }

    return inside;
  }

  //==========================================================================
  //  Interface BitMask:

  /**
   * <tt>XTransitionList</tt> (re-)initialization.
   *
   * @param  x Width of a new bitmask.
   * @param  y Height of a new bitmask.
   */
  public void init ( int x, int y )
  {
    if ( y < 1 ) y = 1;
    yMin = 0;
    yMax = y;
    scanlines = new XLine[ yMax ];
    for ( y = 0; y < yMax; y++ )
      scanlines[y] = new XLine();
  }

  /**
   * <tt>XTransitionList</tt> re-initialization (keeps current bitmask size
   * - if the size is relevant).
   */
  public void init ()
  {
    if ( yMax <= yMin ) yMax = yMin + 1;
    scanlines = new XLine[ yMax - yMin ];
    int y;
    for ( y = 0; y < yMax - yMin; y++ )
      scanlines[y] = new XLine();
  }

  /**
   * Sets the binary set operation (used in all <tt>set*()</tt> methods).
   *
   * @param  _op Binary set operation (<tt>BinaryOperation.BIN_OP_*</tt> constant).
   * @return The old binary set operation.
   * @see    cz.cuni.jagrlib.BinaryOperation Binary operations
   */
  public int setOperation ( int _op )
  {
    int oldOp = binaryOp;
    binaryOp  = BinaryOperation.stationaryOp( _op );
    return oldOp;
  }

  /**
   * Sets the given pixel's value to <tt>true</tt> = "painted".
   * Accepts current <b>binary set operation</b>.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   */
  public void setPixel ( int x, int y )
  {
    assertY( y );
    XLine l = new XLine();
    l.add( new XItem( x ) );
    l.add( new XItem( x+1 ) );
    scanlines[y-yMin] = setList( l, y );
  }

  /**
   * Sets the given pixel's value (ignores current <b>binary set operation</b>).
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @param  value New pixel's value.
   */
  public void putPixel ( int x, int y, boolean value )
  {
    if ( value ) assertY( y );
    if ( y < yMin || y >= yMax ) return;

    XLine l = new XLine();
    if ( value )
      l.add( new XItem( x ) );
    scanlines[y-yMin] = setList( l, y, x, x+1 );
  }

  /**
   * Returns the pixel's value.
   *
   * @param  x X coordinate of a pixel.
   * @param  y Y coordinate of a pixel.
   * @return Pixel value (<tt>false</tt> for off-canvas access).
   */
  public boolean getPixel ( int x, int y )
  {
    if ( y < yMin || y >= yMax ) return false;

    ListIterator<XItem> li = scanlines[y-yMin].listIterator();
    return find( x+1, li );
  }

  /**
   * Sets the horizontal line from [min(x1,x2),y] to [max(x1,x2)-1,y].
   * Accepts current <b>binary set operation</b>.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   */
  public void setHLine ( int x1, int x2, int y )
  {
    if ( x1 >= x2 )
    {
      if ( x1 == x2 ) return;
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }

    assertY( y );
    XLine l = new XLine();
    l.add( new XItem( x1 ) );
    l.add( new XItem( x2 ) );
    scanlines[y-yMin] = setList( l, y );
  }

  /**
   * Sets the horizontal line's values (ignores current <b>binary set operation</b>).
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  values New pixels' values.
   */
  public void putHLine ( int x1, int x2, int y, boolean[] values )
  {
    if ( values == null ) return;

    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }

    if ( x1 + values.length < x2 ) x2 = x1 + values.length;
    if ( x1 == x2 ) return;

    assertY( y );
    XLine l = new XLine();
    int x   = x1;                           // XTR coordinate
    int i   = 0;                            // values[i]
    boolean old = false;                    // the last logical value

    while ( x < x2 )
    {
      if ( values[i] != old )
      {
        l.add( new XItem( x ) );
        old = !old;
      }
      x++;
      i++;
    }

    scanlines[y-yMin] = setList( l, y, x1, x2 );
  }

  /**
   * Returns the pixels' values for the given horizontal line.
   *
   * @param  x1 X coordinate of the starting pixel.
   * @param  x2 X coordinate of the first pixel after the line.
   * @param  y Common Y coordinate of the line.
   * @param  values Returned values in <tt>boolean[]</tt> array (<tt>false</tt> for
   *                off-canvas access).
   */
  public void getHLine ( int x1, int x2, int y, boolean[] values )
  {
    if ( values == null ) return;

    if ( x1 > x2 )
    {
      int tmp = x1;
      x1 = x2;
      x2 = tmp;
    }

    if ( x1 + values.length < x2 ) x2 = x1 + values.length;
    if ( x1 == x2 ) return;

    int i = 0;
    if ( y < yMin || y >= yMax )
    {                                       // outside the actual set
      while ( x1++ < x2 )
        values[i++] = false;
      return;
    }

    ListIterator<XItem> it = scanlines[y-yMin].listIterator();
    int x = it.hasNext() ? it.next().x : BitMaskEnumerator.PLUS_INFINITY;
    boolean in = false;                     // inside the actual set?
    while ( x <= x1 )
    {                                       // skip the prefix
      in = !in;
      x = it.hasNext() ? it.next().x : BitMaskEnumerator.PLUS_INFINITY;
    }

      // x1 < x:
    while ( x <= x2 )
    {                                       // fill one inner segment
      do
        values[i++] = in;
      while ( x1++ < x );
      in = !in;
      x = it.hasNext() ? it.next().x : BitMaskEnumerator.PLUS_INFINITY;
    }

    while ( x1++ < x2 )                     // fill the rest
      values[i++] = in;
  }

  /**
   * Sets the rectangle from [min(x1,x2),min(y1,y2)] to [max(x1,x2)-1,max(y1,y2)-1].
   * Accepts current <b>binary set operation</b>.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   */
  public void setRectangle ( int x1, int y1, int x2, int y2 )
  {
    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( y1 > y2 )
    {
      tmp = y1;
      y1 = y2;
      y2 = tmp;
    }
    if ( x1 == x2 || y1 == y2 ) return;

    assertY( y2 - 1 );
    do
      setHLine( x1, x2, y1++ );
    while ( y1 < y2 );
  }

  /**
   * Sets the rectangle's values (ignores current <b>binary set operation</b>).
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  values New pixels' values.
   */
  public void putRectangle ( int x1, int y1, int x2, int y2, boolean[][] values )
  {
    if ( values == null ) return;

    int tmp;
    if ( x1 > x2 )
    {
      tmp = x1;
      x1 = x2;
      x2 = tmp;
    }
    if ( y1 > y2 )
    {
      tmp = y1;
      y1 = y2;
      y2 = tmp;
    }
    if ( x1 == x2 || y1 == y2 ) return;

    while ( y1 < y2 )
    {
      putHLine( x1, x2, y1, values[ y1 - yMin ] );
      y1++;
    }
  }

  /**
   * Returns the pixels' values for the given rectangle.
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @param  values Returned values in <tt>boolean[][]</tt> array (<tt>false</tt> for
   *                off-canvas access).
   */
  public void getRectangle ( int x1, int y1, int x2, int y2, boolean[][] values )
  {
    if ( values == null || x1 >= x2 ) return;

    int offset = 0;
    int x;
      // empty prolog:

    while ( y1 < yMin )
    {
      for ( x = 0; x < values[offset].length; )
        values[offset][x++] = false;
      y1++;
      if ( ++offset >= values.length ) return;
    }

    if ( y1 >= y2 ) return;

      // regular data:
    while ( y1 < y2 && y1 < yMax )
      getHLine( x1, x2, y1++, values[ offset++ ] );

      // empty epilog:
    while ( offset < values.length )
    {
      for ( x = 0; x < values[offset].length; )
        values[offset][x++] = false;
      offset++;
    }
  }

  /**
   * Bit-block-transfer routine: copies the source rectangle [sx,sy,width,height] to
   * the new location [dx,dy]. Binary set operation is ignored.
   *
   * @param  dx X coordinate of the upper left destination corner.
   * @param  dy Y coordinate of the upper left destination corner.
   * @param  sx X coordinate of the upper left source corner.
   * @param  sy Y coordinate of the upper left source corner.
   * @param  width Rectangle width.
   * @param  height Rectangle height.
   */
  public void bitBlt ( int dx, int dy, int sx, int sy, int width, int height )
  {
    if ( width <= 0 || height <= 0 ) return;

    boolean tmp[][] = new boolean[ height ][ width ];
    getRectangle( sx, sy, sx + width, sy + height, tmp );
    putRectangle( dx, dy, dx + width, dy + height, tmp );
  }

  /**
   * {@link BitMask} enumeration.
   *
   * @return The <tt>BitMaskEnumerator</tt> object instance.
   */
  public BitMaskEnumerator enumerator ()
  {
    return new XTransitionListEnumerator();
  }

  //==========================================================================
  //  Interface Property:

  protected int setWidth  = 0;

  protected int setHeight = 0;

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // WIDTH:
    if ( key.compareTo(WIDTH) == 0 )
    {
      setWidth = intProperty( value, setWidth );
      return;
    }
      // HEIGHT:
    if ( key.compareTo(HEIGHT) == 0 )
    {
      setHeight = intProperty( value, setHeight );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(WIDTH) == 0 )
      return setWidth;
    if ( key.compareTo(HEIGHT) == 0 )
      return setHeight;
    return null;
  }

  /**
   * Commits all previous changes to the internal object state.
   * Non-mandatory.
   */
  public void commit ()
  {
    init( setWidth, setHeight );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "XTransitionList";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "BitMask";

  /** Object description. */
  private final static String DESCRIPTION = "2D bit-mask represented by a \"X-transition " +
                                            "list\" data structure.";
  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DATA + "." + C_SET;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"BitMask" );
        // Property:
      t.propBegin( WIDTH, TYPE_INTEGER, "Bitmask width in pixels", true );
        t.propDefault( "512" );
      t.propEnd();
      t.propBegin( HEIGHT, TYPE_INTEGER, "Bitmask height in pixels", true );
        t.propDefault( "512" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
