package cz.cuni.jagrlib.piece;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.zip.*;

import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;

/**
 * ZLIB codec. Uses built-in Java ZLIB port from {@code java.util.zip} package.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/ZLIBCodec.java">ZLIBCodec.java</a>
 */
public class ZLIBCodec extends Piece implements EntropyCodec
{
  /** OutputStream proxy (for {@link DeflaterOutputStream}). */
  protected class OutStream extends OutputStream
  {
    /** Write one byte to JaGrLib's {@link BitStream}. */
    @Override
    public void write ( int b )
    throws IOException
    {
      if ( stream == null || !output )
        throw new IOException( "ZLIBCodec: cannot write to output stream" );

      stream.write( b, 8 );
      rawBits += 8;
    }

    /** Write whole array of bytes to JaGrLib's {@link BitStream}. */
    @Override
    public void write ( byte[] b )
    throws IOException
    {
      if ( stream == null || !output )
        throw new IOException( "ZLIBCodec: cannot write to output stream" );

      for ( int i = 0; i < b.length; i++ )
        stream.write( b[i], 8 );
      rawBits += (b.length << 3);
    }

    /** Write fragment of byte array to JaGrLib's {@link BitStream}. */
    @Override
    public void write ( byte[] b, int off, int len )
    throws IOException
    {
      if ( stream == null || !output )
        throw new IOException( "ZLIBCodec: cannot write to output stream" );

      if ( len <= 0 ) return;
      rawBits += (len << 3);
      while ( len-- > 0 )
        stream.write( b[off++], 8 );
    }
  }

  /** InputStream proxy (for {@link InflaterInputStream}). */
  protected class InStream extends InputStream
  {
    /** Read one byte from JaGrLib's {@link BitStream}. */
    @Override
    public int read ()
    throws IOException
    {
      if ( stream == null || output )
        throw new IOException( "ZLIBCodec: cannot read from input stream" );

      rawBits += 8;
      return (int)stream.read( 8 );
    }
  }

  //==========================================================================
  //  Data:

  /** Associated I/O stream. {@code null} if the codec is not opened. */
  protected BitStream stream;

  /** Opened for output (encoding)? */
  protected boolean output;

  /** Actual output stream (for {@link DeflaterOutputStream} link). */
  protected OutStream oStream;

  /** Actual input stream (for {@link InflaterInputStream} link). */
  protected InStream iStream;

  /** Id of the actual context. */
  protected int actualCtx;

  /** The actual context (coder). */
  protected DeflaterOutputStream oCtx;

  /** The actual context (decoder). */
  protected InflaterInputStream iCtx;

  /**
   * Compression level.
   * Must be between {@code 0} (fastest) and {@code 9} (best compression).
   *
   * @see Deflater#setLevel(int)
   */
  protected int level = 9;

  /**
   * Deflater compression strategy.
   *
   * @see Deflater#setStrategy(int)
   */
  protected int strategy = Deflater.FILTERED;

  /** Actual position in original (logical) stream. */
  protected long position = 0;

  /** Actual position in compressed (binary) stream. */
  protected long rawBits = 0;

  //==========================================================================
  //  Interface EntropyCodec:

  public ZLIBCodec ()
  {
    iCtx = null;
    oCtx = null;
  }

  public boolean available ()
  throws IOException
  {
    return ( !output && stream != null &&
             (iCtx.available() > 0 || stream.available() > 0) );
  }

  /** Close the stream.. */
  public void close ()
  throws IOException
  {
    if ( stream == null )
      return;

    if ( output )
      oCtx.finish();
    //LogFile.debug( "ZLIB codec: Closing at position: " + position );
    stream.close();
    position =
    rawBits  = 0;
    stream   = null;
  }

  public long compressed ()
  throws IOException
  {
    // !!! TODO: compute exact binary position
    return stream.position();
  }

  public void flush ()
  throws IOException
  {
    if ( output )
      oCtx.flush();
  }

  /** Get next character from compressed stream. */
  public int get ()
  throws IOException
  {
    if ( stream == null || output )
      throw new IOException( "ZLIBCodec is not opened for reading!" );

    position++;
    return iCtx.read();
  }

  /**
   * Decodes one subinterval from the given {@link WheelOfFortune}.
   * Uses the actual context!
   *
   * @param  wheel Set of intervals (wheel of fortune).
   * @return Index of the relevant interval.
   */
  public int get ( WheelOfFortune wheel )
  throws IOException
  {
    return get();
  }

  public long getBits ( int length )
  throws IOException
  {
    if ( stream == null || output )
      throw new IOException( "ZLIBCodec is not opened for reading!" );

    rawBits += length;
    return stream.read( length );
  }

  public void open ( boolean output, int ctx )
  throws IOException
  {
    stream      = (BitStream)getInterface( PL_OUTPUT, IFACE+"BitStream" );
    this.output = output;
    position    =
    rawBits     = 0;
    actualCtx   = ctx;
    stream.open( output, (String)null, null );
    if ( output )
    {
      if ( oStream == null )
        oStream = new OutStream();

      iCtx = null;
      Deflater deflater = new Deflater( level );
      deflater.setStrategy( strategy );
      oCtx = new DeflaterOutputStream( oStream, deflater );
    }
    else
    {
      if ( iStream == null )
        iStream = new InStream();

      iCtx = new InflaterInputStream( iStream );
      oCtx = null;
    }
  }

  /**
   * Retrieves maximal symbol value for the actual context.
   *
   * @return Maximal symbol value ({@code 255} for ZLIBCodec class).
   */
  public int getMaxSymbol ()
  {
    return 255;
  }

  /**
   * Sets maximal symbol value for the actual context.
   *
   * @param maxSymbol New maximal symbol value.
   */
  public void setMaxSymbol ( int maxSymbol )
  {
    if ( maxSymbol != 255 )
      LogFile.debug( "ZLIBCodec: the only supported maxSymbol is 255!" );
  }

  public long position ()
  {
    return position;
  }

  /** Puts the given symbol into the stream. */
  public void put ( int symbol )
  throws IOException
  {
    if ( stream == null || !output )
      throw new IOException( "ZLIBCodec is not opened for writting!" );

    position++;
    oCtx.write( symbol );
  }

  /**
   * Encodes one subinterval from the given {@link WheelOfFortune}.
   * Uses the actual context!
   *
   * @param  wheel Set of intervals (wheel of fortune).
   * @param  s Index of an interval to encode.
   */
  public void put ( WheelOfFortune wheel, int s )
  throws IOException
  {
    put( s );
  }

  public void putBits ( long bits, int length )
  throws IOException
  {
    stream.write( bits, length );
    rawBits += length;
  }

  public int setContext ( int ctx )
  throws IOException
  {
    int old   = actualCtx;
    actualCtx = ctx;
    return old;
  }

  //==========================================================================
  //  Interface Property:

  public final static String COMPRESSION_LEVEL    = "CompressionLevel";
  public final static String COMPRESSION_STRATEGY = "CompressionStrategy";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // COMPRESSION_LEVEL:
    if ( key.compareTo(COMPRESSION_LEVEL) == 0 )
    {
      level = intProperty( value, level, 0, 9 );
      return;
    }
      // COMPRESSION_STRATEGY:
    if ( key.compareTo(COMPRESSION_STRATEGY) == 0 )
    {
      strategy = intProperty( value, strategy );
      if ( strategy != Deflater.DEFAULT_STRATEGY &&
           strategy != Deflater.FILTERED &&
           strategy != Deflater.HUFFMAN_ONLY )
        strategy = Deflater.DEFAULT_STRATEGY;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(COMPRESSION_LEVEL) == 0 )
      return level;
    if ( key.compareTo(COMPRESSION_STRATEGY) == 0 )
      return strategy;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "ZLIBCodec";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "EntropyCodecToBitStream";

  /** Object description. */
  private final static String DESCRIPTION = "ZLIB codec implementation (using built-in Java classes).";

  /** Object category. */
  protected final static String CATEGORY = C_2D + ".compression.ZLIB";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"EntropyCodec" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"BitStream" );
        // Property:
      t.propBegin( COMPRESSION_LEVEL, TYPE_INTEGER, "Compression level (0 .. fastest, 9 .. best)", true );
        t.propDefault( 9 );
        t.propBounds(  0, 9 );
      t.propEnd();
      t.propBegin( COMPRESSION_STRATEGY, TYPE_INTEGER, "ZLIB Compression strategy", true );
        t.propDefault( Deflater.FILTERED );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Default strategy", Deflater.DEFAULT_STRATEGY, "Default strategy" );
        t.propEnum( "Huffman only",     Deflater.HUFFMAN_ONLY,     "Huffman only" );
        t.propEnum( "Filtered",         Deflater.FILTERED,         "Filtered (best for small random values)" );
      t.propEnd();
     }
   return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
