package cz.cuni.jagrlib.piece;

import java.util.*;
import java.io.IOException;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Zig-zag block coding.
 * Used as back-end in block transform compression methods of raster 2D data.
 *
 * @version 0.25 $Rev: 385 $ $Date: 2007-01-07 01:47:21 +0100 (ne, 07 I 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/piece/ZigZagCodec.java">ZigZagCodec.java</a>
 */
public class ZigZagCodec extends Piece implements BlockCodec
{
  //==========================================================================
  //  Data:

  /** Actual codec variant (not used yet). */
  protected int variant = 0;

  /** Actual block width in samples. */
  protected int width   = 8;

  /** Actual block height in samples. */
  protected int height  = 8;

  /** Actual entropy codec (accelerator). */
  protected EntropyCodec codec;

  /**
   * {@link StringBuffer} for logging purposes.
   *
   * @see   #LOGGING
   */
  protected StringBuffer logSb;

  /** Context number for non-zero values encoding. */
  protected final static int CTX_VALUE = 0;

  /** Context number for zero-skipping. */
  protected final static int CTX_SKIP  = 1;

  //==========================================================================
  //  Support:

  /** Performs one step, returns {@code true} if succeeded. */
  protected final boolean step ( int[] pos )
  {
    pos[0] += pos[2];
    pos[1] += pos[3];

    if ( pos[0] >= 0 && pos[0] < width &&
         pos[1] >= 0 && pos[1] < height )
      return true;

      // primarily: over any border
    pos[2] = -pos[2];                     // invert current orientation
    pos[3] = -pos[3];

    if ( pos[0] >= width )                // primarily: over the right border
    {
      pos[0]--;
      return( (pos[1] += 2) < height );
    }

    if ( pos[1] >= height )               // primarily: over the bottom border
    {
      pos[1]--;
      return( (pos[0] += 2) < width );
    }

    if ( pos[0] < 0 )                     // primarily: over the left border
      pos[0] = 0;                         // go back to the border
    else                                  // primarily: over the top border
      pos[1] = 0;                         // go back to the border

    return true;
  }

  /** Performs step of the given size, returns {@code true} if succeeded. */
  protected final boolean step ( int[] pos, int size )
  {
    while ( size-- > 0 )                  // make one step
      if ( !step( pos ) )
        return false;

    return true;
  }

  /** Finds next non-zero number, alters pos, returns {@link #CTX_SKIP}-encoded value. */
  protected final int findNonzero ( int[] data, int[] pos )
  {
    int steps = 1;                        // 1 means "no skips" (in our CTX_SKIP encoding)
    do
    {
      if ( !step( pos ) )
        return 0;

      if ( data[ pos[0] + pos[1] * width ] != 0 )
        return steps;

      steps++;
    }
    while ( true );
  }

  //==========================================================================
  //  Interface BlockCodec:

  /**
   * Prepares the underlying {@link EntropyCodec}.
   * Should be called before any of {@link #encode}s or {@link #decode}s.
   */
  public void init ()
  throws IOException
  {
    codec = (EntropyCodec)getInterface( PL_OUTPUT, IFACE+"EntropyCodec" );
      // prepare context for VLC values:
    codec.setContext( CTX_VALUE );
    codec.setMaxSymbol( 4095 );
      // prepare context for ZERO-skips:
      // 0 .. EOB, 1 .. no skip, 2 .. skip one zero, ..
    codec.setContext( CTX_SKIP );
    codec.setMaxSymbol( width * height );
  }

  /**
   * Set actual codec variant (context).
   * Should set mode of the underlying {@link EntropyCodec} too.
   *
   * @param  variant Variant number (from <tt>0</tt>).
   * @return Old variant number.
   * @see    #setBlockSize
   */
  public int setVariant ( int variant )
  {
    int old      = this.variant;
    this.variant = variant;
    return old;
  }

  /**
   * Sets actual block size.
   * Can be used both for 1D and 2D quantization.

   * @param  width Horizontal block size (1D block size).
   * @param  height Vertical block size (<tt>1</tt> for 1D).
   * @see    #setVariant
   */
  public void setBlockSize ( int width, int height )
  throws IOException
  {
    this.width  = Math.max( width,  1 );
    this.height = Math.max( height, 1 );
      // prepare context for ZERO-skips:
    if ( codec != null )
    {
      codec.setContext( CTX_SKIP );
      codec.setMaxSymbol( this.width * this.height );
    }
  }

  /**
   * Block coding.
   * Converts matrix of integer data to {@link EntropyCodec} input.
   * Connected {@link EntropyCodec} module will be feeded by them.
   *
   * @param  data Source data to be encoded.
   * @param  predict Pre-allocated array used for predicition passing.
   *         <tt>predict[1]</tt> will be used as source, <tt>predict[0]</tt> is filled with
   *         prediction for the next block[s]. Can be {@code null}.
   * @see    #decode
   */
  public void encode ( int[] data, int[][] predict )
  throws IOException
  {
    if ( codec == null ||
         data  == null ||
         data.length < width * height )
      return;

    int[] pos = { 0, 0, 1, -1 };          // { -1, 1, 1, -1 }
    int st, val;

    StringBuffer sb = LogFile.debugging ? logSb : null;
    if ( sb != null )
    {
      sb.setLength( 0 );
      sb.append( "Tmp code: " );
    }

      // prediction processing:
    if ( predict != null && predict.length >= 2 )
    {                                     // predict[0][0] .. prediction for next block's data[0]
      if ( predict[0] == null || predict[0].length < 1 )
        predict[0] = new int[ 1 ];
      predict[0][0] = data[0];
                                          // predict[1][0] .. prediction for data[0]
      if ( predict[1] != null && predict[1].length > 0 )
      {
        data[0] -= predict[1][0];
        if ( sb != null )
          sb.append( "(" + (data[0] + predict[1][0]) + ")->(" + data[0] + ") " );
      }

    }

    codec.setContext( CTX_VALUE );
    val = (int)Formula.vlc( data[ 0 ] );

    if ( sb != null )
      sb.append( data[ 0 ] );

    codec.put( val );

    do
    {
      st = findNonzero( data, pos );

      codec.setContext( CTX_SKIP );
      codec.put( st );

      if ( sb != null )
      {
        if ( st == 0 )
          sb.append( "[EOB]" );
        else
          sb.append( "[" + (st-1) + "]" );
      }

      if ( st == 0 ) return;

      codec.setContext( CTX_VALUE );
      val = (int)Formula.vlc( data[ pos[0] + pos[1] * width ] );

      if ( sb != null )
        sb.append( Formula.vlcInv( val ) );

      codec.put( val );
    }
    while ( true );
  }

  /**
   * Block decoding.
   * Reads input from underlying {@link EntropyCodec} and converts them back into
   * given integer matrix.
   *
   * @param  data Target data array (pre-allocated to sufficient size).
   * @param  predict Pre-allocated array used for predicition passing.
   *         <tt>predict[1]</tt> will be used as source, <tt>predict[0]</tt> is filled with
   *         prediction for the next block[s]. Can be {@code null}.
   * @return Number of decoded array items (should be <tt>with * height</tt>
   *         if successful).
   * @see    #encode
   */
  public int decode ( int[] data, int[][] predict )
  throws IOException
  {
    if ( codec == null ||
         data  == null ||
         data.length < width * height )
      return 0;

    Arrays.fill( data, 0 );
    int[] pos = { 0, 0, 1, -1 };          // { -1, 1, 1, -1 }
    int st, val;

    /*
    StringBuffer sb = LogFile.debugging ? new StringBuffer() : null;
    */
    codec.setContext( CTX_VALUE );
    val = Formula.vlcInv( codec.get() );
    data[ 0 ] = val;

    codec.setContext( CTX_SKIP );
    while ( (st = codec.get()) > 0 )
    {
      if ( !step( pos, st ) ) return 0;
      codec.setContext( CTX_VALUE );
      val = Formula.vlcInv( codec.get() );
      data[ pos[0] + pos[1] * width ] = val;
      codec.setContext( CTX_SKIP );

      /*
      if ( sb != null )
        sb.append( "[" + (st-1) + "]" + val );
      */
    }

      // prediction processing:
    if ( predict != null && predict.length >= 2 )
    {                                     // predict[1][0] .. prediction for data[0]
      if ( predict[1] != null && predict[1].length > 0 )
        data[0] += predict[1][0];
                                          // predict[0][0] .. prediction for next block's data[0]
      if ( predict[0] == null || predict[0].length < 1 )
        predict[0] = new int[ 1 ];
      predict[0][0] = data[0];
    }

    /*
    if ( sb != null )
      LogFile.log( new String( sb.append( "[-1]" ) ) );
    */

    return( width * height );
  }

  //==========================================================================
  //  Interface PersistentParamsBinary:

  /**
   * Pack instance parameters into binary format.
   *
   * @return Data array to be stored into binary stream. Can be {@code null}.
   *         The array can be shared by all method calls.
   * @see    #loadData
   */
  public byte[] storeData ()
  {
    byte[] arr = new byte[ 3 ];

    arr[0] = (byte)( variant & 0xff ); // actual codec variant
    arr[1] = (byte)( width   & 0xff ); // table width
    arr[2] = (byte)( height  & 0xff ); // table height

    return arr;
  }

  /**
   * Loads instance parameters from (previously stored) binary data.
   *
   * @param  data Binary data to be interpreted. Can be {@code null}.
   * @return {@code true} if successful.
   * @see    #storeData
   */
  public boolean loadData ( byte[] data )
  {
    if ( data == null ||
         data.length < 3 ) return false;

    variant = data[0];
    width   = data[1];
    height  = data[2];

    try
    {
      init();
    }
    catch ( IOException e )
    {
      throw new JaGrLibException( "ZigZagCodec: cannot load parameters!", e );
    }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null ) return;
      // LOGGING:
    if ( key.compareTo(LOGGING) == 0 )
    {
      boolean newLogging = booleanProperty( value, logSb != null );
      if ( newLogging != (logSb != null) )
        logSb = newLogging ? new StringBuffer() : null;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
      // LOGGING:
    if ( key.compareTo(LOGGING) == 0 )
    {
      if ( logSb != null )
        return new String( logSb );
      return null;
    }
    return null;
  }

  //==========================================================================
  //  Module registration

  /** Object name. */
  private final static String NAME = "ZigZagCodec";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "BlockCodecToEntropyCodec";

  /** Object description. */
  private final static String DESCRIPTION = "Zig-zag encoding of 2D integer data.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_COMPRESSION + ".coding";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"BlockCodec" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"EntropyCodec" );
        // Property:
      t.propBegin( LOGGING, TYPE_BOOLEAN, "Log temporary code of the block?", false );
        t.propDefault( "false" );
      t.propEnd();
     }
   return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
