package cz.cuni.jagrlib.reg;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;

/**
 * Runnable composition.
 *
 * @author  PE, Jan Kostlivy
 * @version 0.26 $Rev: 454 $ $Date: 2007-05-15 03:35:14 +0200 (út, 15 V 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/reg/Composition.java">Composition.java</a>
 */
public class Composition
{
  /**
   * Create Piece instances.
   *
   * @param modules Map of all relevant modules.
   */
  public static void createPieces ( HashMapModules modules )
  {
    for ( Integer key : modules.keySet() )
    {
      InfoModule im  = modules.get( key );

      if ( im instanceof InfoGroup )
      {
        createPieces(  ((InfoGroup)im).modules );
        connectPieces( ((InfoGroup)im).channels );
        initPieces(    ((InfoGroup)im).modules );
      }
      else
      {
        try
        {
          im.instancePiece = RegDatabaseBasic.newPieceInstance( im );
        }
        catch ( Exception e )
        {
          LogFile.exception( e );
        }
      }
    }
  }

  /**
   * Dispose Piece instances.
   *
   * @param modules Map of all relevant modules.
   */
  public static void disposePieces ( HashMapModules modules )
  {
    for ( Integer key : modules.keySet() )
    {
      InfoModule im  = modules.get( key );

      if ( im instanceof InfoGroup )
        disposePieces( ((InfoGroup)im).modules );
      else
        im.instancePiece = null;
    }
  }

  /**
   * Connect pieces by channels.
   *
   * @param channels Map of all relevant channels.
   */
  public static void connectPieces ( HashMapChannels channels )
  {
    for ( Integer key : channels.keySet() )
    {
      InfoChannel ic  = channels.get( key );
      try
      {
        Piece pieceOut = ic.plugFrom.getModuleOwner().instancePiece;
        Piece pieceIn  = ic.plugTo.getModuleOwner().instancePiece;
        pieceOut.connect( ic.plugFrom.type, pieceIn, ic.plugTo.type );
      }
      catch ( BadInterfaceException e )
      {
        LogFile.exception( "Bad interface in connecting-time!", e );
      }
    }
  }

  /**
   * Inits all Piece instances.
   *
   * @param modules Map of all relevant modules.
   */
  public static void initPieces ( HashMapModules modules )
  {
    for ( Integer key : modules.keySet() )
    {
      InfoModule im = modules.get( key );
      Piece      ip = im.instancePiece;
      if ( ip != null )
      {
        Map< String, Object > p = new HashMap< String, Object >();
        for ( String keyp : im.params.keySet() )
          p.put( im.params.get( keyp ).name,
                 im.params.get( keyp ).value );

        ip.set( p );
        ip.commit();
      }
    }
  }

  /**
   * Runs the given composition.
   *
   * @param composition Composition to run.
   */
  public static void run ( InfoGroup composition )
  {
      // create Piece instances:
    createPieces( composition.modules );

      // connect them by channels:
    connectPieces( composition.channels );

      // init Piece instances:
    initPieces( composition.modules );

      // select one or more worker modules to run:
    for ( Integer key : composition.modules.keySet() )
    {
      final Piece ipp = composition.modules.get( key ).instancePiece;

      if ( ipp instanceof Worker &&
           !Piece.booleanProperty( ipp, Worker.REJECT, false ) )
      {
        Thread runIt = new Thread( "JaGrLib - Composition worker" )
        {
          @Override
          public void run ()
          {
            try
            {
              ((Worker)ipp).run();
            }
            catch ( Exception e )
            {
              LogFile.exception( e );
            }
          }
        };

        try
        {
          runIt.setPriority( Thread.MIN_PRIORITY );
        }
        catch ( Exception e )
        {
          LogFile.warning( "Cannot set priority of worker thread!" );
          LogFile.exception( e );
        }

        try
        {
          runIt.start();
        }
        catch ( IllegalThreadStateException e )
        {
          LogFile.error( "Cannot start worker thread!" );
          LogFile.exception( e );
        }
      }
    }
  }

}
