package cz.cuni.jagrlib.reg;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import cz.cuni.jagrlib.gui.*;

/**
 * Information about concrete group instance (stored in XML file).
 *
 * @author  Jan Kostlivy, PE
 * @version 0.25 $Rev: 381 $ $Date: 2007-01-02 15:51:44 +0100 (út, 02 I 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/reg/InfoGroup.java">InfoGroup.java</a>
 */
public class InfoGroup extends InfoModule
{
  /** File-name of the composition. */
  public String fileName = "";

  /** Information about composition's GUI. */
  public InfoCompositionGUI infoCompositionGUI = null;

  /** Do the group resides in the registry? */
  public boolean registry = true;

  /** Change of this group modifies registry as well? */
  public boolean editRegisty = false;

  /**
   * Set of all information about modules and groups in group.
   * Contains only objects of {@link InfoModule} or it's descendants (like {@link InfoGroup}).
   */
  public HashMapModules modules = new HashMapModules();

  /**
   * Set of all information about channels in the group.
   * Contains only objects of {@link InfoChannel}.
   */
  public HashMapChannels channels = new HashMapChannels();

  /** Are new modules connected to the rest of composition automatically? */
  public static boolean autoConnect = false;

  public final static String TI_GROUP       = "Group";
  public final static String TI_COMPOSITION = "Composition";

  /**
   * Determine if this group has the key or has module with the key.
   *
   * @param _key In demand key.
   * @return {@code true} if group has the key.
   */
  public boolean isOwner ( Integer _key )
  {
    if ( super.isOwner( _key ) )
      return true;

    for ( InfoModule im : modules.values() )
      if ( im.isOwner( _key ) )
        return true;

    return false;
  }

  /**
   * State check - checks connection status..
   *
   * @return One of InfoState.* constants.
   */
  public int getState ()
  {
    int result = 0;
    for ( InfoModule im : modules.values() )
      result |= im.getState();

    return result;
  }

  /**
   * Group initialization.
   * Sets plugs and parameters from original modules.
   */
  public void init ()
  {
    plugs.clear();
    params.clear();
    for ( InfoModule im : modules.values() )
    {
      setPlugs( im );
      setParams( im );
      im.owner = this;
    }
  }

  /**
   * Copies all plugs from the given module.
   *
   * @param im Source module.
   */
  public void setPlugs ( InfoModule im )
  {
    for ( InfoPlug ip : im.plugs.values() )
      if ( ip.infoGUI == null ||
           ip.infoGUI.visible )
      {
        InfoPlug ipg = new InfoPlug( ip );
        ipg.owner = this;
        plugs.put( im.key.toString() + "." + ip.type, ipg );
        newLink( ip, ipg );
      }
  }

  /**
   * Setting parameters from it's module.
   *
   * @param im Information about module.
   */
  public void setParams ( InfoModule im )
  {
    for ( String name : im.params.keySet() )
    {
      InfoParam ip = im.params.get( name );
      if ( ip.infoGUI == null ||
           ip.infoGUI.visible )
        params.put( im.key.toString() + "." + name, ip );
    }
  }

  /**
   * Connecting module's and group's plug.
   *
   * @param ipmodule plug from module.
   * @param ipgroup plug from group.
   * @return new channel - link.
   */
  public InfoChannel newLink ( InfoPlug ipmodule, InfoPlug ipgroup )
  {
    InfoChannel ic   = new InfoChannel();
    ic.infoGUI       = new InfoChannelGUI();
    ic.plugFrom      = ipmodule;
    ic.plugTo        = ipgroup;
    ic.nameInterface = ipmodule.name;
    ic.owner         = this;
    return ic;
  }

  /**
   * Looking for plug in group's channels.
   *
   * @param ip Information about plug.
   * @return {@code true} if group has channel with this plug or {@code false} otherwise.
   */
  public boolean findPlugInChannels ( InfoPlug ip )
  {
    for ( InfoChannel ic : channels.values() )
      if ( ic.plugFrom.equals( ip ) ||
           ic.plugTo.equals( ip ) )
        return true;

    return false;
  }

  /**
   * Gets upper-left corner of group.
   */
  public Dimension getTopLeftPoint ()
  {
    int minX    = Integer.MAX_VALUE;
    int minY    = Integer.MAX_VALUE;

    for ( InfoModule im : modules.values() )
    {
      if ( minX > im.infoGUI.x ) minX = im.infoGUI.x;
      if ( minY > im.infoGUI.y ) minY = im.infoGUI.y;
    }

    return new Dimension( minX, minY );
  }

  /** Default constructor. */
  public InfoGroup ()
  {
    super( new RegPiece() );
    modules.addActionListener( new ActionListener ()
    {
      public void actionPerformed ( ActionEvent e )
      {
        change_actionPerformed( e );
      }
    });
    channels.addActionListener ( new ActionListener ()
    {
      public void actionPerformed ( ActionEvent e )
      {
        change_actionPerformed( e );
      }
    });
  }

  void change_actionPerformed ( ActionEvent e )
  {
    changePerformed( e.getID(), e.getSource() );
  }

  /*-------------------------------------------------*/
  /*------------------- New Group -------------------*/
  /*-------------------------------------------------*/

  /**
   * Creation of a new group.
   *
   * @param superGroup superGroup with selected modules.
   * @return New group or {@code null}.
   */
  public static InfoGroup createNewGroup ( InfoGroup superGroup )
  {
    InfoGroup newGroup = new InfoGroup();
    newGroup.infoGUI   = new InfoGroupGUI();

    newGroup.infoCompositionGUI          = new InfoCompositionGUI();
    newGroup.infoCompositionGUI.gui      = InstanceGUI.newCompositionInstance();
    newGroup.infoCompositionGUI.gui.info = newGroup;

    newGroup.reg.packageName  = "GROUP";
    newGroup.reg.name         = "New group";
    newGroup.reg.className    = newGroup.reg.name;
    newGroup.reg.template     = "Group";

    newGroup.infoGUI.selected = true;
    newGroup.infoGUI.x        = 0;
    newGroup.infoGUI.y        = 0;
    newGroup.infoGUI.height   = ModuleGUI.getModuleHeight();
    newGroup.infoGUI.width    = ModuleGUI.getModuleWidth();

    Vector< ModuleGUI > oldModules = new Vector< ModuleGUI >();

    for ( InfoModule im : superGroup.modules.values() )
      if ( im.infoGUI.selected )
      {
        newGroup.modules.put( im.key, im );
        oldModules.add( im.infoGUI.gui );
      }

      // moving internal channels into the group:
    moveChannelsToGroup( superGroup.channels, newGroup );

    if ( newGroup.modules.size() > 0 )
    {
      newGroup.moveToTopLeft();
      newGroup.initVisibility();

      if ( RegDatabaseBasic.registryGroup( newGroup ) )
      {
        for ( Integer key : newGroup.modules.keySet() )
          superGroup.modules.remove( key );

        for ( ModuleGUI mg : oldModules )
          superGroup.infoCompositionGUI.gui.remove( mg );

        for ( Integer key : newGroup.channels.keySet() )
          superGroup.channels.remove( key );

        newGroup.init();

        for ( InfoChannel ic : superGroup.channels.values() )
        {
          if ( superGroup.modules.get( ic.plugFrom.owner.key ) == null )
            for ( InfoPlug ipp : newGroup.plugs.values() )
              if ( ic.plugFrom.equals( ipp.linkToOrig ) )
                ic.plugFrom = ipp;

          if ( superGroup.modules.get( ic.plugTo.owner.key ) == null )
            for ( InfoPlug ipp : newGroup.plugs.values() )
              if ( ic.plugTo.equals( ipp.linkToOrig ) )
                ic.plugTo = ipp;
        }

        newGroup.infoGUI.height = ModuleGUI.getModuleHeight( newGroup.getVisibleParamLines() );
        return newGroup;
      }

      for ( InfoModule im : newGroup.modules.values() )
        im.infoGUI = (InfoModuleGUI)im.nextInfo.remove( "infoGUI" );

      superGroup.infoCompositionGUI.gui.readInfo();
    }

    return null;
  }

  /**
   * Moves channels (with both plugs in group) to new group.
   *
   * @param superChannels channels from parent group.
   * @param newGroup new group.
   */
  public static void moveChannelsToGroup ( HashMapChannels superChannels, InfoGroup newGroup )
  {
      // moving channels which are entirely inside a group
    for ( Integer key : superChannels.keySet() )
    {
      InfoChannel ic = superChannels.get( key );
      InfoModule imf = newGroup.modules.get( ic.plugFrom.owner.key );
      InfoModule imt = newGroup.modules.get( ic.plugTo.owner.key );

      if ( imf != null &&
           imt != null )
        newGroup.channels.put( key, ic );
    }
  }

  /**
   * Moves all modules to top left corner (relative to their positions).
   */
  public void moveToTopLeft ()
  {
    Dimension tlp = getTopLeftPoint();
    for ( InfoModule im : modules.values() )
    {
      if ( im.nextInfo == null )
        im.nextInfo = new HashMap< String, Object >();

      im.nextInfo.put( "infoGUI", im.infoGUI );
      InfoModuleGUI img = (InfoModuleGUI)im.infoGUI.clone();
      img.selected = false;
      img.x = im.infoGUI.x - tlp.width;
      img.y = im.infoGUI.y - tlp.height;
      im.infoGUI = img;
    }
  }

  /** Sets initial visibility of plugs. */
  public void initVisibility ()
  {
    for ( InfoModule im : modules.values() )
      for ( InfoPlug ip : im.plugs.values() )
        ip.infoGUI.visible = ip.mandatory;
  }

  transient ActionListener actionListener;

  /**
   * Group change - add/delete module/channel.
   *
   * @param id Identification of action.
   * @param _obj Module or channel.
   */
  public void changePerformed ( int id, Object _obj )
  {
    ActionEvent e = new ActionEvent( _obj, id, null );
    processActionEvent( e );
  }

  /**
   * Adds the specified action listener to receive action events from
   * this object. Action events occur when any action is performed.
   * If l is null, no exception is thrown and no action is performed.
   *
   * @param         l the action listener
   * @see           #removeActionListener
   * @see           java.awt.event.ActionListener
   */
  public synchronized void addActionListener ( ActionListener l )
  {
    if ( l != null )
      actionListener = AWTEventMulticaster.add( actionListener, l );
  }

  /**
   * Removes the specified action listener so that it no longer
   * receives action events from this object.
   * If l is null, no exception is thrown and no action is performed.
   *
   * @param           l     the action listener
   * @see             #addActionListener
   * @see             java.awt.event.ActionListener
   */
  public synchronized void removeActionListener ( ActionListener l )
  {
    if ( l != null )
      actionListener = AWTEventMulticaster.remove( actionListener, l );
  }

  /**
   * Processes action events occurring on this object
   * by dispatching them to any registered
   * {@link ActionListener} objects.
   *
   * <p>This method is not called unless action events are
   * enabled for this object. Action events are enabled
   * when one of the following occurs:
   * <ul>
   * <li>An {@link ActionListener} object is registered
   * via {@link #addActionListener}.
   * <li>Action events are enabled via <tt>enableEvents</tt>.
   * </ul>
   * Note that if the event parameter is {@code null}
   * the behavior is unspecified and may result in an
   * exception.
   *
   * @param       e the action event
   * @see         java.awt.event.ActionListener
   * @see         java.awt.Component#enableEvents
   */
  protected void processActionEvent ( ActionEvent e )
  {
    if ( actionListener != null )
      actionListener.actionPerformed( e );
  }

}
