package cz.cuni.jagrlib.reg;

import java.util.*;

import cz.cuni.jagrlib.LogFile;
import cz.cuni.jagrlib.Template;

/**
 * Information about a plug.
 *
 * @author  Jan Kostlivy, PE
 * @version 0.26 $Rev: 431 $ $Date: 2007-04-29 22:42:22 +0200 (ne, 29 IV 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/reg/InfoPlug.java">InfoPlug.java</a>
 */
public class InfoPlug implements Cloneable
{
  /** Name of plug (= interface!). */
  public String name = "";

  /** Type of plug. Unique for one module, used as hash-map key. */
  public String type = "";

  /** Plug type (input/output). */
  public boolean input;

  /** Is the plug mandatory? (only for output plugs) */
  public boolean mandatory = true;

  /** Is the plug is "multi"? */
  public boolean multi = false;

  /** Plug description. */
  public String description = "";

  /** Owner of plug. It can be {@link InfoModule}. */
  public InfoModule owner;

  /** Is the plug connected? */
  private boolean assigned = false;

  /** Connected channels. */
  public HashMapChannels channels;

  /** Information about plug's GUI. */
  public InfoPlugGUI infoGUI = null;

  /** This is space for all other information. */
  public Map< String, Object > nextInfo;

  /** Link to original plug (for groups). */
  public InfoPlug linkToOrig = null;

  public final static String TI_PLUG = "Plug";

  public String inputOutput ()
  {
    return( input ? "Input" : "Output" );
  }

  /**
   * Registry new connected channel.
   */
  public void registryChannel ( InfoChannel _channel )
  {
    if ( channels == null )
      channels = new HashMapChannels();

    setAssigned( true );
    channels.put( _channel.key, _channel );
  }

  /**
   * Unregistry connected channel.
   */
  public void unregistryChannel ( InfoChannel _channel )
  {
    channels.remove( _channel.key );
    if ( channels.isEmpty() )
      setAssigned( false );
  }

  /**
   * Tests whether this plug is connected to another plug.
   *
   * @return {@code true} if plug is connected.
   */
  public boolean isAssigned ()
  {
    if ( linkToOrig != null )
      return linkToOrig.isAssigned();

    return assigned;
  }

  /**
   * Set assigning of plug.
   *
   * @param _assign New value of assigned.
   */
  public void setAssigned ( boolean _assign )
  {
    if ( linkToOrig != null )
      linkToOrig.setAssigned( _assign );

    assigned = _assign;
  }

  /**
   * Get owner of the plug.
   */
  public InfoModule getModuleOwner ()
  {
    if ( linkToOrig != null )
      return linkToOrig.getModuleOwner();

    return owner;
  }

  /**
   * Tests whether the two plugs are opposite.
   *
   * @param _plug1 First plug.
   * @param _plug2 Second plug.
   * @return {@code true} if they are opposite.
   */
  static public boolean isOpposite ( InfoPlug _plug1, InfoPlug _plug2 )
  {
    boolean from1 = _plug1.input;
    if ( from1 == _plug2.input ) return false; // wrong orientation

    if ( from1 )
    {
      InfoPlug tmp = _plug1;
      _plug1       = _plug2;
      _plug2       = tmp;
    }

      // _plug1 .. source (output), _plug2 .. target (input)
    try
    {
      return Class.forName( Template.IFACE + _plug1.name ).isAssignableFrom( Class.forName( Template.IFACE + _plug2.name ) );
    }
    catch ( ClassNotFoundException e )
    {
      LogFile.exception( e );
      return false;
    }
  }

  /**
   * Plug state check.
   *
   * @return  One of {@link InfoState} constants.
   * @see     InfoState
   */
  public int getState ()
  {
    if ( input || isAssigned() )
      return InfoState.OK;

    return( mandatory ? InfoState.NOTCONNECTED_HARD : InfoState.NOTCONNECTED_SOFT );
  }

  /**
   * Tests whether the plug is unconnected or multi-plug:
   * i.e. - if it is posible to connect another channel.
   *
   * @return {@code true} if it is posible to connect another channel.
   */
  public boolean isFree ()
  {
    return( multi || !isAssigned() );
  }

  /** Default constructor. */
  public InfoPlug()
  {}

  /**
   * Creates a new InfoPlug.
   *
   * @param  _name Name of plug.
   * @param  _type Type of plug.
   * @param  _input Plug type (input/output).
   * @param  _mandatory Flag if the plug is mandatory.
   * @param  _multi Flag if the plug is multi.
   * @param  _description Description of plug.
   */
  public InfoPlug ( String _name,
                    String _type,
                    boolean _input,
                    boolean _mandatory,
                    boolean _multi,
                    String _description )
  {
    name        = _name;
    type        = _type;
    input       = _input;
    mandatory   = !_input &&
                  _mandatory;
    multi       = _multi;
    description = _description;
  }

  /**
   * Creates plug for groups with link to original plug.
   *
   * @param ip original plug.
   */
  public InfoPlug ( InfoPlug ip )
  {
    this( ip.name, ip.type, ip.input, ip.mandatory, ip.multi, ip.description );
    linkToOrig = ip;
    owner      = null;
    assigned   = false;
    channels   = ip.channels;
    infoGUI    = null;
    nextInfo   = ip.nextInfo;
  }

  public Object clone ()
  {
    return new InfoPlug( this );
  }

}
