package cz.cuni.jagrlib.reg;

import java.util.*;

import cz.cuni.jagrlib.*;

/**
 * Registration data for Piece.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.23
 * @see     <a href="../../../../../cz/cuni/jagrlib/reg/RegPiece.java">RegPiece.java</a>
 */
public class RegPiece implements Template
{
  /**
   * Fully qualified package name.
   * (e.g.: "cz.cuni.jagrlib.piece").
   */
  public String packageName;

  /**
   * Local class name.
   * (e.g.: "LineBresenham").
   */
  public String className;

  /**
   * Short name of a module.
   * (e.g.: "Bresenham").
   */
  public String name;

  /**
   * Module template.
   * (e.g.: "LineRenderToBitMask").
   */
  public String template;

  /**
   * Module category.
   * (e.g.: "2d.draw.line.integer").
   */
  public String category;

  /**
   * Longer description of a module.
   */
  public String description;

  /** Set of module plugs. */
  public Map< String, InfoPlug >  plugs  = new LinkedHashMap< String, InfoPlug >();

  /** Set of module parameters. */
  public Map< String, InfoParam > params = new LinkedHashMap< String, InfoParam >();

  /** Separates local interface name from fully qualified class name. */
  public final static String normalizeInterfaceName ( String iface )
  {
    int pos = iface.lastIndexOf( '.' );
    if ( pos < 0 ) return iface;

    return iface.substring( pos + 1 );
  }

  /** Returns full clas name of the module. */
  public final String fullClassName ()
  {
    return( packageName + '.' + className );
  }

  /**
   * Connects a new {@link Plug}.
   * General routine.
   *
   * @param  key New plug's key.
   * @param  _mandatory Is that plug mandatory (has to be connected)?
   * @param  _multi Can that plug be used in multi-channels?
   * @param  _cloneable Is that plug "cloneable" (=generic)? If yes,
   *                    ordinal numbers (starting from 0) will be appended to its name.
   * @param  _inputInterface Fully qualified identifier of input (implemented)
   *                         {@code interface}.
   * @param  _outputInterface Fully qualified identifier of output (required)
   *                          {@code interface}.
   */
  public void newPlug ( String key, boolean _mandatory, boolean _multi, boolean _cloneable,
                        String _inputInterface, String _outputInterface )
  {
    if ( _inputInterface == null )
        // output plug
      plugs.put( key, new InfoPlug( normalizeInterfaceName( _outputInterface ), key, false, _mandatory, _multi, "" ) );
    else
    if ( _outputInterface == null )
      plugs.put( key, new InfoPlug( normalizeInterfaceName( _inputInterface ), key, true, _mandatory, _multi, "" ) );
  }

  /**
   * Connects a new input {@link Plug}. Common-use routine: creates <b>optional</b>,
   * <b>non-cloneable</b>, <b>multi</b> plug.
   *
   * @param  key New plug's key.
   * @param  _inputInterface Fully qualified identifier of input (implemented)
   *                         {@code interface}.
   */
  public void newInputPlug ( String key, String _inputInterface )
  {
    plugs.put( key, new InfoPlug( normalizeInterfaceName( _inputInterface ), key, true, false, true, "" ) );
  }

  /**
   * Connects a new output {@link Plug}.
   * Common-use routine: creates <b>mandatory</b>, <b>non-cloneable</b>, <b>multi</b> plug.
   *
   * @param  key New plug's key.
   * @param  _outputInterface Fully qualified identifier of output (required)
   *                          {@code interface}.
   */
  public void newOutputPlug ( String key, String _outputInterface )
  {
    plugs.put( key, new InfoPlug( normalizeInterfaceName( _outputInterface ), key, false, true, true, "" ) );
  }

  /**
   * Connects a new optional output {@link Plug}.
   * Common-use routine: creates <b>optional</b>, <b>non-cloneable</b>, <b>multi</b> plug.
   *
   * @param  key New plug's key.
   * @param  _outputInterface Fully qualified identifier of output (required)
   *                          {@code interface}.
   */
  public void newOptOutputPlug ( String key, String _outputInterface )
  {
    plugs.put( key, new InfoPlug( normalizeInterfaceName( _outputInterface ), key, false, false, true, "" ) );
  }

  /**
   * Sets registration strings for the module.
   */
  public void setRegStrings ( String _name, String _template, String _category, String _description )
  {
    name        = _name;
    template    = _template;
    category    = _category;
    description = _description;
  }

  //==========================================================================
  //  Property registration:

  protected InfoParam param = null;

  protected void logError ( String msg )
  {
    LogFile.error( "registration of " + fullClassName() + '(' + name + ") - " + msg );
  }

  /**
   * Starts definition of single property.
   *
   * @param  name String identifier (unique inside one module class).
   * @param  type Value type - full Java class identifier.
   * @param  descr Brief textual description (for human eyes only).
   * @param  visual Should be this property represented visually in GUI?
   * @see    #propEnd
   */
  public void propBegin ( String name, String type, String descr, boolean visual )
  {
    if ( param != null )
    {
      logError( "Property definition nested: " + name + " inside of " + param.name + " definition" );
      return;
    }

    param = new InfoParam( name, type, descr, visual );
  }

  /**
   * Default value (non-mandatory).
   */
  public void propDefault ( Object value )
  {
    if ( param == null )
    {
      logError( "propDefault() is not called after propBegin()!" );
      return;
    }

    param.defaultValue = value;
  }

  /**
   * Minimum and maximum allowed values (for comparable types only, non-mandatory).
   */
  public void propBounds ( Object min, Object max )
  {
    if ( param == null )
    {
      logError( "propBounds() is not called after propBegin()!" );
      return;
    }

    param.setBounds( min, max );
  }

  /**
   * Sets visual manipulator for GUI editing of the property.
   * For predefined manipulators.
   */
  public void propManipulator ( int manipulatorType )
  {
    if ( param == null )
    {
      logError( "propManipulator(int) is not called after propBegin()!" );
      return;
    }

    param.propManip = manipulatorType;
  }

  /**
   * Sets visual manipulator for GUI editing of the property.
   * For user-defined manipulators.
   *
   * @param  className Class name (either fully-qualified or local in {@code package cz.cuni.jagrlib.gui}).
   */
  public void propManipulator ( String className )
  {
    if ( param == null )
    {
      logError( "propManipulator(String) is not called after propBegin()!" );
      return;
    }

    param.propManip      = MANIPULATOR_CUSTOM;
    param.propManipClass = className;
  }

  /**
   * Adds next enum label (for list-box GUI).
   *
   * @param  label Textual label to appear in the listbox.
   * @param  value Associated property value (real-value).
   * @param  descr Optional item's description (for human eyes only).
   */
  public void propEnum ( String label, Object value, String descr )
  {
    if ( param == null )
    {
      logError( "propEnum() is not called after propBegin()!" );
      return;
    }

    param.setEnum( label, value, descr );
  }

  /**
   * Finishes definition of the property.
   * Every property-definition segment started by {@link #propBegin} must
   * be properly closed by {@link #propEnd}.
   *
   * @see    #propBegin
   */
  public void propEnd ()
  {
    if ( param == null )
    {
      logError( "propEnd() without matching propBegin()!" );
      return;
    }

    param.propEnd();

    params.put( param.name, param );
    param = null;
  }

  /**
   * Static instance of Class object for RegPiece.
   */
  public static final Class<?> TYPE = getCl();

  static private Class<?> getCl ()
  {
    try
    {
      return Class.forName( JAGRLIB2 + "reg.RegPiece" );
    }
    catch ( ClassNotFoundException e )
    {
      return null;
    }
  }

}
