package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Anisotropic microfacet distribution.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/AnisotropicMicrofacet.java">AnisotropicMicrofacet.java</a>
 */
public class AnisotropicMicrofacet extends Piece implements MicrofacetDistribution
{
  /** Specular reflection exponent for x axis. */
  double exponentX;

  /** Specular reflection exponent for y axis. */
  double exponentY;

  /**
   * Sample the given microfacet distribution.
   *
   * @param  mf   The {@link MicroFacet} (point on surface of some solid).
   * @param  in   Input direction (should look from the hit point).
   * @param  u    Random uniform number.
   * @param  v    Random uniform numbers.
   * @param  out  Output direction in world coordinates.
   * @param  brdf Color definition - {@code double[C]} array
   *              ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *              color system/ or greater /for more complex spectral color sampling/).
   * @return Probability density function.
   */
  public double sample ( MicroFacet mf,
                         double[] in,
                         double u,
                         double v,
                         double[] out,
                         double[] brdf )
  {
    double phi, exp = Math.sqrt( ( exponentX + 1.0 ) / ( exponentY + 1.0 ) );
    if ( u < 0.25 )
    {
        // first quadrant:
      u = 4.0 * u;
      phi = Math.atan( exp * Math.tan( 0.5 * Math.PI * u ) );
    }
    else if ( u < 0.5 )
    {
        // second quadrant:
      u = 1.0 - 4.0 * (0.5 - u);
      phi = Math.atan( exp * Math.tan( 0.5 * Math.PI * u ) );
      phi = Math.PI - phi;
    }
    else if ( u < 0.75 )
    {
        // third quadrant:
      u = 4.0 * (u - 0.5);
      phi = Math.atan( exp * Math.tan( 0.5 * Math.PI * u ) );
      phi += Math.PI;
    }
    else
    {
        // fourth quadrant:
      u = 1.0 - 4.0 * (1.0 - u);
      phi = Math.atan( exp * Math.tan( 0.5 * Math.PI * u ) );
      phi = 2.0 * Math.PI - phi;
    }

      // compute half-vector:
    double cosPhi = Math.cos( phi );
    double sinPhi = Math.sin( phi );
    double exponent = 1.0 / ( cosPhi*cosPhi*exponentX + sinPhi*sinPhi*exponentY + 1.0 );
    double cosTheta = Math.pow( u, exponent );
    double sinTheta = Math.sqrt( Math.max( 0.0, 1.0 - cosTheta * cosTheta ) );
    double[] half = new double[ 3 ];
    Geometry.toCartesian( sinTheta, cosTheta, phi, half );
    if ( out[2] * half[2] < 0.0 ) half[2] = half[2];

      // compute out (incident) direction:
    double twoDotIH = 2.0 * Geometry.dotProduct3D( in, half );
    for ( int i = 0; i < out.length; i++ )
      out[i] = -in[i] + twoDotIH * half[i];

    return D( half ) / ( 4.0 * Geometry.dotProduct3D( out, half ) );
  }

  /**
   * Return the probability density function.
   *
   * @param in  Input direction in local coordinates.
   * @param out Output direction in local coordinates.
   */
  public double pdf ( double[] in, double[] out )
  {
    double[] half = new double[ 3 ];
    for ( int i = 0; i < half.length; i++ )
      half[i] = in[i] + out[2];
    Geometry.normalize3D( half );

    return D(half) / ( 4.0 * Geometry.dotProduct3D( in, half ) );
  }

  /**
   * Return a microfacet distribution term.
   *
   * @param half Half-vector in local coordinates.
   */
  public double D ( double[] half )
  {
    double cosTheta = Math.abs( half[2] );
    double exponent = ( exponentX * half[0] * half[0] +
                        exponentY * half[1] * half[1] ) / (1.0 - cosTheta * cosTheta);
    double invTwoPI = 1.0 / (2.0 * Math.PI);
    return Math.sqrt( (exponentX + 2.0) * (exponentY + 2.0) ) *
           invTwoPI * Math.pow( cosTheta, exponent );
  }

  //==========================================================================
  //  Interface Property:

  public final static String EXPONENT_X = "Exponent X";
  public final static String EXPONENT_Y = "Exponent Y";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

    if ( key.compareTo( EXPONENT_X ) == 0 )
    {
      exponentX = doubleProperty( value, exponentX );
      return;
    }
    if ( key.compareTo( EXPONENT_Y ) == 0 )
    {
      exponentY = doubleProperty( value, exponentY );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo( EXPONENT_X ) == 0 ) return exponentX;
    if ( key.compareTo( EXPONENT_Y ) == 0 ) return exponentY;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Anisotropic microfacet distribution";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "BSSRDF";

  /** Object description. */
  private final static String DESCRIPTION = "Anisotropic microfacet distribution.";

  /** Object category. */
  protected final static String CATEGORY = Piece.C_3D + Piece.C_DATA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( Piece.PL_INPUT, Piece.IFACE+"BSSRDF" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}