package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Algorithm for drawing Bezier cubic curve.
 *
 * @version 0.26 $Rev: 431 $ $Date: 2007-04-29 22:42:22 +0200 (ne, 29 IV 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/BezierRegular.java">BezierRegular.java</a>
 */
public class BezierRegular extends Piece implements CurveRender
{
  //==========================================================================
  //  Data:

  /** Number of line segments per curve. */
  protected int number = 50;

  //==========================================================================
  //  Interface Property:

  public final static String NUMBER = "Number";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // NUMBER:
    if ( key.compareTo(NUMBER) == 0 )
    {
      number = intProperty( value, number, 1, 10000 );
      return;
    }
    return;
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(NUMBER) == 0 )
      return number;
    return null;
  }

  //==========================================================================
  //  Interface CurveRender

  private void point ( double[] p, double t, double[] result )
  {
    double nt = 1.0 - t;
    double k0 =      nt * nt * nt;
    double k1 = 3.0 * t * nt * nt;
    double k2 = 3.0 * t *  t * nt;
    double k3 =       t *  t *  t;
    result[0] = p[0] * k0 + p[2] * k1 + p[4] * k2 + p[6] * k3;
    result[1] = p[1] * k0 + p[3] * k1 + p[5] * k2 + p[7] * k3;
  }

  /**
   * Draws curve segment using the given array of control points.
   *
   * @param cp Array of control points.
   * @param orig First array item to be used.
   * @param len Number of control points (can define curve order).
   * @param dim Number of coordinate components (2 for plane, ..).
   */
  public void drawCurve ( double[] cp, int orig, int len, int dim )
  {
    LineRenderAnti lines = (LineRenderAnti)getInterface( PL_OUTPUT, IFACE+"LineRenderAnti" );

    double t = 0.0;
    double dt = 1.0 / (number <= 0 ? 1 : number);
    double[] po = new double[ 2 ];
    double[] pn = new double[ 2 ];

    point( cp, t, po );                   // starting point

    while ( t <= 1.0 )
    {
      point( cp, t += dt, pn );
      lines.drawLine( po[0], po[1], pn[0], pn[1] );
      po[0] = pn[0];
      po[1] = pn[1];
    }
  }

  /**
   * Draws curve segment using the given array of control points.
   *
   * @param cp Array of control points.
   * @param orig First array item to be used.
   * @param len Number of control points (can define curve order).
   * @param dim Number of coordinate components (2 for plane, ..).
   */
  public void drawCurve ( int[] cp, int orig, int len, int dim )
  {
    int dataLen = len * dim;
    double[] dcp = new double[ dataLen ];
    for ( int i = 0; i < dataLen; i++ )
      dcp[i] = (double)cp[orig++];

    drawCurve( dcp, 0, len, dim );
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "BezierRegular";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "CurveRenderToLineRenderAnti";

  /** Object description. */
  private final static String DESCRIPTION = "Draws Bezier curve using regular parametric algorithm.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DRAW + "." + C_CURVE + "." + C_FLOAT;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"CurveRender" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"LineRenderAnti" );
        // Property:
      t.propBegin( NUMBER, TYPE_INTEGER, "Number of line segments", true );
        t.propDefault( 50 );
        t.propBounds(  1, 10000 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
