package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Blinn microfacet distribution.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 841 $ $Date: 2008-09-23 22:28:54 +0200 (út, 23 IX 2008) $ $Author: Tomas $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/BlinnMicrofacet.java">BlinnMicrofacet.java</a>
 */
public class BlinnMicrofacet extends Piece implements MicrofacetDistribution
{
  /** Specular reflection exponent. */
  double exponent;

  /**
   * Sample the given microfacet distribution.
   *
   * @param  mf   The {@link MicroFacet} (point on surface of some solid).
   * @param  in   Input direction (should look from the hit point).
   * @param  u    Random uniform number.
   * @param  v    Random uniform numbers.
   * @param  out  Output direction in world coordinates.
   * @param  brdf Color definition - {@code double[C]} array
   *              ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *              color system/ or greater /for more complex spectral color sampling/).
   * @return Probability density function.
   */
  public double sample ( MicroFacet mf,
                         double[] in,
                         double u,
                         double v,
                         double[] out,
                         double[] brdf )
  {
      // compute half-angle vector:
    double cosTheta = Math.pow( u, 1.0 / ( exponent + 1.0 ) );
    double sinTheta = Math.sqrt( Math.max( 0.0, 1.0 - cosTheta*cosTheta ) );
    double phi      = v * 2.0 * Math.PI;
    double[] half = new double[ 3 ];
    Geometry.toCartesian( sinTheta, cosTheta, phi, half );
    if ( in[2]*half[2] < 0.0 ) half[2] = -half[2];

      // compute incident direction:
    double dotIH2 = 2.0 * Geometry.dotProduct3D( in, half );
    for ( int i = 0; i < out.length; i++ )
      out[i] = in[i] * dotIH2 * half[i];

      // compute pdf:
    double pdf = ( (exponent + 1.0) * Math.pow( cosTheta, exponent ) ) /
                 ( 2.0 * Math.PI * 4.0 * Geometry.dotProduct3D( in, half ) );

    return pdf;
  }

  /**
   * Return the probability density function.
   *
   * @param in  Input direction in local coordinates.
   * @param out Output direction in local coordinates.
   */
  public double pdf ( double[] in, double[] out )
  {
      // compute half-vector:
    double[] half = new double[ 3 ];
    for ( int i = 0; i < half.length; i++ )
      half[i] = in[i] + out[i];
    Geometry.normalize3D( half );

    double cosTheta = half[2];
    return ( ( exponent + 1.0 ) * Math.pow( cosTheta, exponent ) )
           / ( 2.0 * Math.PI * 4.0 * Geometry.dotProduct3D( in, half ) );
  }

  /**
   * Return a microfacet distribution term.
   *
   * @param half Half-vector in local coordinates.
   */
  public double D ( double[] half )
  {
    double cosTheta = half[2];
    double invTwoPI = 1.0 / (2.0 * Math.PI);
    return (exponent + 2.0) * invTwoPI * Math.pow( cosTheta, exponent );
  }

  //==========================================================================
  //  Interface Property:

  public final static String EXPONENT = "Exponent";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

    if ( key.compareTo( EXPONENT ) == 0 )
    {
      exponent = doubleProperty( value, exponent );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo( EXPONENT ) == 0 ) return exponent;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Blinn microfacet distribution";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "BSSRDF";

  /** Object description. */
  private final static String DESCRIPTION = "Blinn microfacet distribution.";

  /** Object category. */
  protected final static String CATEGORY = Piece.C_3D + Piece.C_DATA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( Piece.PL_INPUT, Piece.IFACE+"BSSRDF" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}