package cz.cuni.jagrlib.testing;

import java.awt.Color;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Image filter example - blue-screen masking.
 *
 * @version 0.26 $Rev: 439 $ $Date: 2007-05-05 12:29:50 +0200 (so, 05 V 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/BlueScreen.java">BlueScreen.java</a>
 */
public class BlueScreen extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Central background color. */
  protected Color backgroundCenter     = new Color( 100, 100, 240 );

  /** background classification tolerance. */
  protected double backgroundTolerance = 0.08;

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transformation.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    RasterGraphics input  = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    RasterGraphics output = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );
    int i, j;

      // output image:
    int width  = input.getWidth();
    int height = input.getHeight();
    output.init( width, height, RasterGraphics.MODE_RGBA, 4 );

      // temporary values:
    int tol = (int)( 255 * backgroundTolerance );
    int minR = backgroundCenter.getRed()   - tol;
    int maxR = backgroundCenter.getRed()   + tol;
    int minG = backgroundCenter.getGreen() - tol;
    int maxG = backgroundCenter.getGreen() + tol;
    int minB = backgroundCenter.getBlue()  - tol;
    int maxB = backgroundCenter.getBlue()  + tol;
    int[] pixel = new int[ 4 ];           // I/O pixel

    for ( j = 0; j < height; j++ )
      for ( i = 0; i < width; i++ )       // one pixel
      {
        input.getRGB( i, j, pixel );

        if ( pixel[0] >= minR && pixel[0] <= maxR &&
             pixel[1] >= minG && pixel[1] <= maxG &&
             pixel[2] >= minB && pixel[2] <= maxB )
          pixel[0] = pixel[1] = pixel[2] = pixel[3] = 0;
        else
          pixel[3] = 255;

        output.putPixel( i, j, pixel );
      }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String BACKGROUND_CENTER    = "Background color";
  public final static String BACKGROUND_TOLERANCE = "Tolerance";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // BACKGROUND_CENTER:
    if ( key.compareTo(BACKGROUND_CENTER) == 0 )
    {
      backgroundCenter = colorProperty( value, backgroundCenter );
      return;
    }
      // BACKGROUND_TOLERANCE:
    if ( key.compareTo(BACKGROUND_TOLERANCE) == 0 )
    {
      backgroundTolerance = doubleProperty( value, backgroundTolerance, 0.0, 1.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(BACKGROUND_CENTER) == 0 )
      return backgroundCenter;
    if ( key.compareTo(BACKGROUND_TOLERANCE) == 0 )
      return backgroundTolerance;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "BlueScreen";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Blue-screen masking filter.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"Trigger" );
      t.newOutputPlug( "image1", IFACE+"RasterGraphics" );
      t.newOutputPlug( "image2", IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( BACKGROUND_CENTER, TYPE_OBJECT, "Background color", true );
        t.propDefault( "[100;100;240]" );
        t.propManipulator( "ParamColor" );
      t.propEnd();
      t.propBegin( BACKGROUND_TOLERANCE, TYPE_DOUBLE, "Tolerance", true );
        t.propDefault( "0.08" );
        t.propBounds( "0.0", "1.0" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
