package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * 3D rendering example - wire-frame rendering of {@link Brep} database.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/BrepSolid.java">BrepSolid.java</a>
 */
public class BrepSolid extends DefaultSolid
{
  //==========================================================================
  //  Data:

  /** Explicit B-rep database object. */
  protected Brep brep;

  /** Should the {@link #assertBrep} check the data? */
  protected boolean first = true;

  /**
   * Set of solid ids, can be {@code null},
   * in that case the whole B-rep database is used.
   */
  protected BitSet solids;

  /** Smooth normal-vector interpolation? */
  protected boolean smooth = true;

  /** Accelerator: VERTEX attribute - B-rep coordinates. */
  protected int worldId;

  /** Accelerator: VERTEX attribute - normal vector. */
  protected int normalId;

  /** Accelerator: VERTEX attribute - texture coordinate. */
  protected int textureId;

  /** Shell-mode - all intersections are doubled (=two-sided faces). */
  protected boolean shell = true;

  /**
   * Temporary object to hold intersection data for {@link BrepSolid}.
   *
   * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   */
  protected static class TmpData
  {
    /** Barycentric coordinates in the intersected triangle. */
    double u, v;

    /** Vertex ids. */
    int va, vb, vc;

    /** Normal vector in B-rep coordinates. */
    double[] normal;
  }

  //==========================================================================
  //  Construction:

  /**
   * Implicit (plug) B-rep database.
   */
  public BrepSolid ()
  {
  }

  /**
   * Explicit B-rep database, one (or all) solids.
   *
   * @param  data Explicitly defined B-rep database.
   * @param  solid Either one solid id or {@link Brep#NULL} for the whole database.
   */
  public BrepSolid ( Brep data, int solid )
  {
    brep = data;
    if ( solid >= 0 )
    {
      solids = new BitSet();
      solids.set( solid );
    }
  }

  /**
   * Explicit B-rep database, arbitrary solid set.
   *
   * @param  data Explicitly defined B-rep database.
   * @param  set Arbitrary solid-set (can be {@code null}).
   */
  public BrepSolid ( Brep data, BitSet set )
  {
    brep   = data;
    solids = set;
  }

  /** Inserts another solid id. */
  public void addSolid ( int solid )
  {
    if ( solids == null )
      solids = new BitSet();

    solids.set( solid );
  }

  //==========================================================================
  //  Support:

  /**
   * Asserts the scene object, returns {@code true} if failed.
   */
  protected boolean assertBrep ()
  {
    if ( !first ) return false;

    first = false;
    if ( brep == null )
      brep = (Brep)getInterface( PL_OUTPUT, IFACE+"Brep" );
    if ( brep == null )
      return true;

    worldId   = brep.getAttributeId( Brep.VERTEX, Brep.CTX_DEFAULT, Brep.COORD );
    normalId  = brep.getAttributeId( Brep.VERTEX, Brep.CTX_DEFAULT, Brep.NORMAL );
    textureId = brep.getAttributeId( Brep.VERTEX, Brep.CTX_DEFAULT, Brep.TEXTURE_0 );

    return false;
  }

  //==========================================================================
  //  Interface Solid:

  /**
   * Computes intersection of the solid with the given ray.
   *
   * @param  P0 Ray origin ({@code double[3]} or {@code double[4]}).
   * @param  P1 Direction vector of the ray ({@code double[3]} or {@code double[4]}).
   * @return List of intersection points ({@link MicroFacet}s) sorted by {@code t}
   *         (parametric distance from {@code P0}). {@code null} is returned in case
   *         of bad arguments or if no intersection exists..
   * @see    MicroFacet
   * @see    BoundingVolume
   */
  public List< MicroFacet > intersection ( double[] P0, double[] P1 )
  {
    if ( assertBrep() ) return null;

    BrepIterator it = brep.faceIterator( Brep.CTX_DEFAULT, null );
    int faceId;
    int[] vert  = new int[ 3 ];           // vertex handles
    int faceSize;
    int i;
    double[][] vc = new double[ 3 ][ 4 ]; // vertex local coordinates
    double[]   uv = new double[ 2 ];

    LinkedList< MicroFacet > list = null;
    MicroFacet mf1,  mf2;
    TmpData    tmp1, tmp2;

    while ( (faceId = it.next()) != Brep.NULL )
    {
      while ( (faceSize = brep.getFaceVertices( faceId, vert )) > vert.length )
        vert = new int[ faceSize ];

      for ( i = 0; i < 3; i++ )
      {
        brep.getAttribute( worldId, vert[i], vc[i] );
        Geometry.toCartesian3D( vc[i], vc[i] );
      }

      double t = Geometry.rayTriangle3D( P0, P1, vc[0], vc[1], vc[2], uv );
      if ( Double.isInfinite( t ) ) continue;

        // intersection found:
      if ( list == null )
        list = new LinkedList< MicroFacet >();

      mf1            = new MicroFacet();
      mf1.t          = t;
      if ( shell )
        mf1.t       -= Geometry.SCENE_EPSILON;
      mf1.enter      =
      mf1.front      = true;
      mf1.locCoord   = new double[]
      {
        P0[0] + mf1.t * P1[0],
        P0[1] + mf1.t * P1[1],
        P0[2] + mf1.t * P1[2]
      };
      mf1.face       = faceId;
      mf1.valid      = MicroFacet.VALID_LOC_COORD;
        // tmp data object:
      tmp1           = new TmpData();
      tmp1.u         = uv[0];
      tmp1.v         = uv[1];
      tmp1.va        = vert[0];
      tmp1.vb        = vert[1];
      tmp1.vc        = vert[2];
      if ( smooth && normalId != Brep.NULL )
      {                                   // smooth (interpolated) normal vector
        tmp1.normal = new double[ 4 ];
        brep.getAttribute( normalId, tmp1.va, tmp1.normal );
        final double w = 1.0 - tmp1.u - tmp1.v;
        for ( i = 0; i < 3; i++ )
          tmp1.normal[i] *= w;
        final double[] vec = new double[ 4 ];
        brep.getAttribute( normalId, tmp1.vb, vec );
        for ( i = 0; i < 3; i++ )
          tmp1.normal[i] += tmp1.u * vec[i];
        brep.getAttribute( normalId, tmp1.vc, vec );
        for ( i = 0; i < 3; i++ )
          tmp1.normal[i] += tmp1.v * vec[i];
      }
      else                                // constant normal vector
      {
        for ( i = 0; i < 3; i++ )
        {
          vc[1][i] -= vc[0][i];
          vc[2][i] -= vc[0][i];
        }
        tmp1.normal = Geometry.vectorProduct( vc[1], vc[2], null );
      }
        // tmp.normal now holds normal vector in local (B-rep) coordinates
      if ( P1[0] * tmp1.normal[0] +
           P1[1] * tmp1.normal[1] +
           P1[2] * tmp1.normal[2] > 0.0 )
      {
        if ( shell )
          for ( i = 0; i < 3; i++ )
            tmp1.normal[i] = -tmp1.normal[i];
        else
          mf1.enter =
          mf1.front = false;
      }
      mf1.geometryTmp = tmp1;

      list.add( mf1 );

      if ( !shell ) continue;

      mf2            = new MicroFacet();
      mf2.t          = t;
      mf2.enter      =
      mf2.front      = false;
      mf2.locCoord   = new double[]
      {
        P0[0] + mf2.t * P1[0],
        P0[1] + mf2.t * P1[1],
        P0[2] + mf2.t * P1[2]
      };
      mf2.face       = faceId;
      mf2.valid      = MicroFacet.VALID_LOC_COORD;
        // tmp data object:
      tmp2           = new TmpData();
      tmp2.u         = uv[0];
      tmp2.v         = uv[1];
      tmp2.va        = vert[0];
      tmp2.vb        = vert[1];
      tmp2.vc        = vert[2];
      tmp2.normal    = new double[]
      {
        -tmp1.normal[0],
        -tmp1.normal[1],
        -tmp1.normal[2],
        0.0
      };
      mf2.geometryTmp = tmp2;

      list.add( mf2 );

      if ( t > Geometry.SCENE_EPSILON ) break;
    }

    if ( list != null )
      Collections.sort( list );

    return list;
  }

  /**
   * Asserts validity of the given intersection-attributes.
   * Resolves:
   * {@link MicroFacet#VALID_TANGENT VALID_TANGENT},
   * {@link MicroFacet#VALID_TEX_COORD VALID_TEX_COORD}.
   *
   * @param  mf The checked intersection.
   * @param  tmp Non-mandatory object holding state-information for deferred computations.
   * @param  attributes Attribute mask (required attributes - see
   *                    {@code MicroFacet.VALID_*}).
   * @return Attributes which were not computed.
   */
  @Override
  public int assertAttributes ( MicroFacet mf, Object tmp, int attributes )
  {
    if ( mf  == null ||
         tmp == null ||
         !(tmp instanceof TmpData) ) return attributes;

    TmpData data = (TmpData)tmp;

      // tangent surface vectors:
    if ( (attributes & MicroFacet.VALID_TANGENT) != 0 )
    {
      mf.tu = new double[ 3 ];
      mf.tv = new double[ 3 ];
      if ( data.normal != null )
        Geometry.normalize3D( data.normal );
      else
        data.normal = new double[] { 0.0, 0.0, 1.0 };
      Geometry.getAxes( data.normal, mf.tu, mf.tv );
      if ( mf.toWorld != null )
      {
        mf.toWorld.transformVector3D( mf.tu, mf.tu );
        mf.toWorld.transformVector3D( mf.tv, mf.tv );
      }

      mf.valid |= MicroFacet.VALID_TANGENT;
      if ( (attributes -= MicroFacet.VALID_TANGENT) == 0 ) return 0;
    }

      // 2D texture coordinates (inverse mapping):
    if ( (attributes & MicroFacet.VALID_TEX_COORD) != 0 )
    {
      if ( textureId != Brep.NULL )
      {
          // bilinear interpolation of texture coordinates:
        mf.texCoord = new double[ 2 ];
        brep.getAttribute( textureId, data.va, mf.texCoord );
        final double w = 1.0 - data.u - data.v;
        mf.texCoord[0] *= w;
        mf.texCoord[1] *= w;
        final double[] tex = new double[ 2 ];
        brep.getAttribute( textureId, data.vb, tex );
        mf.texCoord[0] += data.u * tex[0];
        mf.texCoord[1] += data.u * tex[1];
        brep.getAttribute( textureId, data.vc, tex );
        mf.texCoord[0] += data.v * tex[0];
        mf.texCoord[1] += data.v * tex[1];
      }
      else
        mf.texCoord = new double[] { data.u, data.v };

      mf.valid |= MicroFacet.VALID_TEX_COORD;
      if ( (attributes -= MicroFacet.VALID_TEX_COORD) == 0 ) return 0;
    }

    return attributes;
  }

  //==========================================================================
  //  Interface Property:

  public final static String SMOOTH = "Smooth normals";
  public final static String SHELL  = "Shell mode";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // SMOOTH:
    if ( key.compareTo(SMOOTH) == 0 )
    {
      smooth = booleanProperty( value, smooth );
      return;
    }
      // SHELL:
    if ( key.compareTo(SHELL) == 0 )
    {
      shell = booleanProperty( value, shell );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(SMOOTH) == 0 )
      return smooth;
    if ( key.compareTo(SHELL) == 0 )
      return shell;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "BrepSolid";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "SolidToBrep";

  /** Object description. */
  private final static String DESCRIPTION = "B-rep based solid for ray-tracing.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"Solid" );
      t.newOptOutputPlug( PL_OUTPUT, IFACE+"Brep" );
        // Property:
      t.propBegin( SMOOTH, TYPE_BOOLEAN, "Smooth normal vectors?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( SHELL, TYPE_BOOLEAN, "Shell-mode (all faces are two-sided)?", true );
        t.propDefault( true );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
