package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Texture specimen - simple checkerboard (modifies base surface color).
 *
 * @version 0.26 $Rev: 745 $ $Date: 2008-05-17 00:40:51 +0200 (so, 17 V 2008) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/CheckerTexture.java">CheckerTexture.java</a>
 */
public class CheckerTexture extends Piece implements Texture
{
  //==========================================================================
  //  Data:

  /** Texture frequency for U coordinate (2D texture space). */
  protected double fu = 10.0;

  /** Texture frequency for V coordinate (2D texture space). */
  protected double fv = 10.0;

  /** Secondary color. Primary color is inherited from COLOR material attribute, etc. */
  protected double[] color2 = { 0.2, 0.4, 1.0 };

  //==========================================================================
  //  Construction:

  public CheckerTexture ()
  {
  }

  public CheckerTexture ( double fu, double fv, double[] color )
  {
    this.fu = fu;
    this.fv = fv;
    if ( color != null )
      color2 = color;
  }

  //==========================================================================
  //  Interface Texture:

  /**
   * Computes the texture value in the given point (given by a {@link MicroFacet} record).
   *
   * @param  mf {@link MicroFacet} object to modify.
   * @return Hash-value (sample signature) used for adaptive subsampling.
   */
  public long compute ( MicroFacet mf )
  {
    if ( mf.assertAttributes( MicroFacet.VALID_COLOR |
                              MicroFacet.VALID_TEX_COORD ) != 0 )
      return 0L;

    double u = mf.texCoord[0] * fu;
    double v = mf.texCoord[1] * fv;

    long ui = (long)Math.floor( u );
    long vi = (long)Math.floor( v );

    if ( ((ui + vi) & 1) != 0 )
      if ( mf.color.length < 3 )
        mf.color = Formula.cloneArray( color2 );
      else
        System.arraycopy( color2, 0, mf.color, 0, 3 );

    return( ui + RandomStatic.numericRecipes( vi ) );
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // FU:
    if ( key.compareTo(FU) == 0 )
    {
      fu = doubleProperty( value, fu );
      return;
    }
      // FV:
    if ( key.compareTo(FV) == 0 )
    {
      fv = doubleProperty( value, fv );
      return;
    }
      // COLOR:
    if ( key.compareTo(COLOR) == 0 )
    {
      if ( value instanceof double[] )
      {
        double[] v = (double[])value;
        if ( color2 == null || color2.length < 3 )
          color2 = new double[ 3 ];
        System.arraycopy( v, 0, color2, 0, Math.min( color2.length, v.length ) );
      }
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(FU) == 0 )
      return fu;
    if ( key.compareTo(FV) == 0 )
      return fv;
    if ( key.compareTo(COLOR) == 0 )
      return color2;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "CheckerTexture";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "Texture";

  /** Object description. */
  private final static String DESCRIPTION = "Texture specimen -  simple checkerboard texture (modifies base surface color).";

  /** Object category. */
  protected final static String CATEGORY = C_2D + ".texture";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"Texture" );
        // Property:
      t.propBegin( FU, TYPE_DOUBLE, "Frequency U", true );
        t.propDefault( 10.0 );
      t.propEnd();
      t.propBegin( FV, TYPE_DOUBLE, "Frequency V", true );
        t.propDefault( 10.0 );
      t.propEnd();
      t.propBegin( COLOR, TYPE_OBJECT, "Secondary color", false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
