package cz.cuni.jagrlib.testing;

import java.io.IOException;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Pal file format (using simple Entropy codec to encode colormapped raster data).
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/CompressedFormatPal.java">CompressedFormatPal.java</a>
 */
public class CompressedFormatPal extends DefaultRasterFileFormat
{
  //==========================================================================
  //  Data:

  //--------------------------------------------------------------------------
  //  Support routines:

  /**
   * Common load code.
   *
   * @param  stream Opened input bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonLoad ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    EntropyCodec codec = (EntropyCodec)getInterface( PL_OUTPUT, IFACE+"EntropyCodec" );
    if ( codec == null ) return;

    codec.open( false, 0 );

      // Pal header:
    int magic     = readUnsigned16( codec );
    if ( magic != MAGIC )
      throw new IOException( "Error in Pal file format" );

    int width     = readUnsigned16( codec );
    int height    = readUnsigned16( codec );
    int cmapSize  = (int)codec.getBits( 8 ) + 1;
    int cmapBands = (int)codec.getBits( 8 ); // 1 (gray), 3 (RGB) or 4 (RGBA)
    codec.setMaxSymbol( cmapSize - 1 );

    int i, j;
    if ( cmapBands > 1 )
    {
        // colormap data:
      int b;
      int[][] cmap = new int[ cmapSize ][ cmapBands ];
      for ( i = 0; i < cmapSize; i++ )
        for ( b = 0; b < cmapBands; b++ )
          cmap[i][b] = (int)codec.getBits( 8 );

        // initialize raster image object:
      g.init( width, height, RasterGraphics.MODE_COLORMAP, 0 );
      g.setColormapType( (cmapBands == 3) ? RasterGraphics.MODE_RGB : RasterGraphics.MODE_RGBA,
                         cmapSize - 1 );
      g.setColormap( cmap );
    }
    else                                  // gray image => no colormap
      g.init( width, height, RasterGraphics.MODE_GRAY, 0 );

      // image data:
    for ( j = 0; j < height; j++ )        // one scanline
      for ( i = 0; i < width; i++ )
        g.putPixel( i, j, codec.get() );

    codec.close();
  }

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonSave ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    EntropyCodec codec = (EntropyCodec)getInterface( PL_OUTPUT, IFACE+"EntropyCodec" );
    if ( codec == null ) return;

      // analyze the colormap:
    int[][] cmap = null;
    int cmapSize;
    int cmapBands;
    int i, j, b;
    if ( g.getMode() == RasterGraphics.MODE_COLORMAP )
    {
      cmapSize  = g.getMaxColormap() + 1;
      if ( cmapSize > 256 ) cmapSize = 256;
      cmapBands = (g.getColormapType() == RasterGraphics.MODE_RGB) ? 3 : 4;
      cmap      = new int[ cmapSize ][ cmapBands ];
      g.getColormap( 0, cmapSize, cmap );
    }
    else
    {                                     // encode image as gray
      cmapSize  = 256;
      cmapBands = 1;
    }

    codec.open( true, 0 );
    codec.setMaxSymbol( cmapSize - 1 );   // <= 8 bits per pixel

      // Pal header:
    writeUnsigned16( codec, MAGIC );
    int width  = g.getWidth();
    writeUnsigned16( codec, width );
    int height = g.getHeight();
    writeUnsigned16( codec, height );
    codec.putBits( cmapSize - 1, 8 );
    codec.putBits( cmapBands, 8 );

    if ( cmap != null )
    {
        // colormap data:
      for ( i = 0; i < cmapSize; i++ )
        for ( b = 0; b < cmapBands; b++ )
          codec.putBits( cmap[i][b], 8 );

        // image data:
      for ( j = 0; j < height; j++ )      // one scanline
        for ( i = 0; i < width; i++ )
          codec.put( g.getPixel( i, j ) );
    }
    else
        // image data:
      for ( j = 0; j < height; j++ )      // one scanline
        for ( i = 0; i < width; i++ )
          codec.put( g.getGray( i, j ) );

    codec.close();
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

    // !!! none yet !!!
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    // !!! none yet !!!

    return null;
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /** 16-bit magic number (used at the file beginning). */
  public final static int MAGIC = 0x8145;

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return 2;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type ({@code 0.0} .. no way,
   *         {@code 0.5} .. maybe, {@code 0.9} .. almost sure, {@code 1.0} ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int len = (header == null) ? 0 : header.length;
    if ( len > 2 ) len = 2;
    if ( len < 2 )                        // not enough bytes in header[]
      return 0.0;
    byte[] h =
    {                                     // Pal header
      (byte)((MAGIC>>8) & 0xff),
      (byte)( MAGIC     & 0xff)
    };
    int i;
    for ( i = 0; i < len; i++ )
      if ( header[i] != h[i] ) return 0.0;
    return 1.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.png"]) or {@code null}.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.pal",
    };
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Pal file format template";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "DataFileFormatToRasterGraphicsAndEntropyCodecAndBitStream";

  /** Object description. */
  private final static String DESCRIPTION = "Pal filter - trivial implementation.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"DataFileFormat" );
      t.newOptOutputPlug( PL_RASTER, IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_STREAM, IFACE+"BitStream" );
      t.newOptOutputPlug( PL_OUTPUT, IFACE+"EntropyCodec" );
        // Property:
      // !!! none yet !!!
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
