package cz.cuni.jagrlib.testing;

import java.io.IOException;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * RGB file format (using simple Entropy codec to encode raster data).
 *
 * @version 0.25 $Rev: 399 $ $Date: 2007-02-12 23:13:47 +0100 (po, 12 II 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/CompressedFormatRGB.java">CompressedFormatRGB.java</a>
 */
public class CompressedFormatRGB extends DefaultRasterFileFormat
{
  //==========================================================================
  //  Data:

  //==========================================================================
  //  Support:

  /**
   * Common load code.
   *
   * @param  stream Opened input bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonLoad ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    final EntropyCodec codec = (EntropyCodec)getInterface( PL_OUTPUT, IFACE+"EntropyCodec" );

    if ( codec == null ) return;

    codec.open( false, 0 );
    codec.setMaxSymbol( 255 );

      // RGB header:
    int magic  = readUnsigned16( codec );
    if ( magic != MAGIC )
      throw new IOException( "Error in RGB file format" );

    int width  = readUnsigned16( codec );
    int height = readUnsigned16( codec );
    int bands  = (int)codec.getBits( 8 );

      // codec parameters:
    if ( !readParameters( codec, codec ) )
      throw new IOException( "Error in RGB file format (bad entropy codec)" );

      // initialize raster image object:
    switch ( bands )
    {
      case 1:
        g.init( width, height, RasterGraphics.MODE_GRAY, 0 );
        break;
      case 3:
        g.init( width, height, RasterGraphics.MODE_RGB, 0 );
        break;
      case 4:
        g.init( width, height, RasterGraphics.MODE_RGBA, 0 );
        break;
      default:
        g.init( width, height, RasterGraphics.MODE_BAND, bands );
    }

      // image data, band-interlaced:
    int i, j, b;
    int[] pixel = new int[bands];
    for ( j = 0; j < height; j++ )        // one scanline
      for ( b = 0; b < bands; b++ )       // one band
        for ( i = 0; i < width; i++ )
        {
          g.getPixel( i, j, pixel );
          pixel[b] = codec.get();
          g.putPixel( i, j, pixel );
        }

    codec.close();
  }

  /**
   * Common save code.
   *
   * @param  stream Opened output bit-stream.
   * @param  g Checked raster-graphics object.
   * @throws IOException
   */
  @Override
  protected void commonSave ( BitStream stream, RasterGraphics g )
  throws IOException
  {
    final EntropyCodec codec = (EntropyCodec)getInterface( PL_OUTPUT, IFACE+"EntropyCodec" );

    if ( codec == null ) return;

    codec.open( true, 0 );
    codec.setMaxSymbol( 255 );            // 8 bits per band

      // RGB header:
    writeUnsigned16( codec, MAGIC );
    int width  = g.getWidth();
    writeUnsigned16( codec, width );
    int height = g.getHeight();
    writeUnsigned16( codec, height );
    int bands  = g.getBands();
    codec.putBits( bands & 0xff, 8 );

      // codec parameters:
    writeParameters( codec, codec );

      // image data, band-interlaced:
    int i, j, b;
    int[] pixel = new int[bands];
    for ( j = 0; j < height; j++ )        // one scanline
      for ( b = 0; b < bands; b++ )       // one band
        for ( i = 0; i < width; i++ )
        {
          g.getPixel( i, j, pixel );
          codec.put( pixel[b] );
        }

    codec.close();
  }

  //==========================================================================
  //  Interface DataFileFormat:

  /** 16-bit magic number (used at the file beginning). */
  public final static int MAGIC = 0xac37;

  /**
   * Length of file-format header (number of bytes which is necessary for file-format matching).
   *
   * @return Number of bytes from beginning of a file which are enough for file-format
   *         match test.
   * @see    #match
   */
  @Override
  public int headerLength ()
  {
    return 2;
  }

  /**
   * File-format match test. Guesses whether the binary file-header (and file-name) can belong
   * to this file-format type.
   *
   * @param  header Bytes from beginning of the file.
   * @param  fileName Optional file-name string.
   * @return Probability of being my file-format type (<tt>0.0</tt> .. no way,
   *         <tt>0.5</tt> .. maybe, <tt>0.9</tt> .. almost sure, <tt>1.0</tt> ..
   *         absolutely - no need to check another formats).
   * @see    #headerLength
   * @see    #fileNameMasks
   */
  @Override
  public double match ( byte[] header, String fileName )
  {
    int len = (header == null) ? 0 : header.length;
    if ( len > 2 ) len = 2;
    if ( len < 2 )                        // not enough bytes in header[]
      return 0.0;
    byte[] h =
    {                                     // RGB header
      (byte)((MAGIC>>8) & 0xff),
      (byte)( MAGIC     & 0xff)
    };
    int i;
    for ( i = 0; i < len; i++ )
      if ( header[i] != h[i] ) return 0.0;
    return 1.0;
  }

  /**
   * Returns file-name masks associated with the file-format type. Needs not be implemented
   * in systems where "file-name -&gt; file-type" mapping is irrelevant.
   *
   * @return Array of wild-card file-name masks (e.g. ["*.png"]) or <tt>null</tt>.
   * @see    #match
   */
  @Override
  public String[] fileNameMasks ()
  {
    return new String[]
    {
      "*.rgb",
    };
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // !!! not yet !!!
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null )
      return null;
      // !!! not yet !!!
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "RGB file format template";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "DataFileFormatToRasterGraphicsAndEntropyCodecAndBitStream";

  /** Object description. */
  private final static String DESCRIPTION = "RGB filter - trivial implementation.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"DataFileFormat" );
      t.newOptOutputPlug( PL_RASTER, IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_STREAM, IFACE+"BitStream" );
      t.newOptOutputPlug( PL_OUTPUT, IFACE+"EntropyCodec" );
        // Property:
        // !!! none yet !!!
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
