package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Cook-Torrance shading model.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/CookTorranceModel.java">CookTorranceModel.java</a>
 * @todo    Support of more lobes.
 */
public class CookTorranceModel extends LambertianModel
{
  //==========================================================================
  //  Data:

  /** Actual color of the solid surface. */
  public double[] colorD;

  /** Actual color of the solid surface. */
  public double[] colorS;

  /** Fresnel term. */
  public Fresnel fresnel;

  /** Microfacet distribution of some material (represents one lobe). */
  public MicrofacetDistribution distribution;

  /** Default constructor. */
  public CookTorranceModel ()
  {
    colorD = new double[] { 1.0, 1.0, 1.0 };
    colorS = new double[] { 1.0, 1.0, 1.0 };
  }

  protected RandomJames rnd;

  //==========================================================================
  //  Interface BSSRDF:

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as monochrome).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @return BSSRDF coefficient <b>f(in-&gt;out)</b>.
   * @see    #monoBSSRDF(MicroFacet,double[],double[],int)
   * @see    #colorBSSRDF(double[],double[],double[],int,double[])
   */
  @Override
  public double monoBSSRDF ( boolean worldCoords,
                             double[] in,
                             double[] out,
                             double[] normal,
                             double[] tangent,
                             int mask )
  {
    if ( in     == null ||
         out    == null ||
         normal == null )
      return 0.0;

      // ensure that coordinates are local:
    double[] inLoc  = new double[ 3 ];
    double[] outLoc = new double[ 3 ];
    if ( worldCoords )
    {
      worldToLocal( normal, tangent, in,  inLoc  );
      worldToLocal( normal, tangent, out, outLoc );
    }
    else
    {
      System.arraycopy( in,  0, inLoc,  0, in.length  );
      System.arraycopy( out, 0, outLoc, 0, out.length );
    }

      // compute BRDF coefficient (diffuse component):
    double result = 0.0;
    if ( mask == BSSRDF.COMP_DIFFUSE ||
         mask == BSSRDF.COMP_ALL )
    {
      result += colorD[0] / Math.PI;
    }

      // compute BRDF coefficient (specular component):
    if ( mask == BSSRDF.COMP_SPECULAR ||
         mask == BSSRDF.COMP_ALL )
    {
      double cosThetaI = inLoc[2];
      double cosThetaO = outLoc[2];

        // half vector:
      double[] half = new double[ 3 ];
      half[0]       = inLoc[0] + outLoc[0];
      half[1]       = inLoc[1] + outLoc[1];
      half[2]       = inLoc[2] + outLoc[2];
      Geometry.normalize3D( half );

      double cosThetaH = Geometry.dotProduct3D( inLoc, half );
      double G = G( outLoc, inLoc, half );

      result += ( colorS[0] * distribution.D( half ) * G
                  * fresnel.evaluate( cosThetaH, null )[0] ) /
                  ( Math.PI * cosThetaI * cosThetaO );
    }

    return result;
  }

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as band-color).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BSSRDF spectral coefficients.
   * @see    #monoBSSRDF(double[],double[],double[],int)
   * @see    #colorBSSRDF(MicroFacet,double[],double[],int,double[])
   */
  @Override
  public double[] colorBSSRDF ( boolean worldCoords,
                                double[] in,
                                double[] out,
                                double[] normal,
                                double[] tangent,
                                int mask,
                                double[] result )
  {
    if ( in     == null ||
         out    == null ||
         normal == null )
      return null;

    if ( result == null )
      result = new double[ colorD.length ];

      // ensure that coordinates are local:
    double[] inLoc  = new double[ 3 ];
    double[] outLoc = new double[ 3 ];
    if ( worldCoords )
    {
      worldToLocal( normal, tangent, in,  inLoc  );
      worldToLocal( normal, tangent, out, outLoc );
    }
    else
    {
      System.arraycopy( in,  0, inLoc,  0, in.length  );
      System.arraycopy( out, 0, outLoc, 0, out.length );
    }

      // compute BRDF coefficient (diffuse component):
    final double invPI = 1.0 / Math.PI;
    int i;
    if ( mask == BSSRDF.COMP_DIFFUSE ||
         mask == BSSRDF.COMP_ALL )
    {
      for ( i = 0; i < result.length; i++ )
        result[i] += colorD[i] * invPI;
    }

      // compute BRDF coefficient (specular component):
    if ( mask == BSSRDF.COMP_SPECULAR ||
         mask == BSSRDF.COMP_ALL )
    {
      double cosThetaI = inLoc[2];
      double cosThetaO = outLoc[2];

        // compute half-angle vector:
      double[] half = new double[ 3 ];
      half[0]       = inLoc[0] + outLoc[0];
      half[1]       = inLoc[1] + outLoc[1];
      half[2]       = inLoc[2] + outLoc[2];
      Geometry.normalize3D( half );

      double cosThetaH = Geometry.dotProduct3D( inLoc, half );
      double G = G( outLoc, inLoc, half );

      double DGinvPI   = distribution.D( half ) * G * invPI;
      double invThetas = 1.0 / (cosThetaI * cosThetaO);
      double[] fres = new double[ result.length ];
      fresnel.evaluate( cosThetaH, fres );
      for ( i = 0; i < result.length; i++ )
        result[i] += DGinvPI * fres[i] * colorS[i] * invThetas;
    }
    return result;
  }

  /**
   * Retrieves the original color (defining surface color).
   *
   * @param  color Pre-allocated {@code double[C]} array to hold the color
   *               ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *               color system/ or greater /for more complex spectral color sampling/). Can
   *               be {@code null}.
   * @return Array containing copy of original color (can be altered).
   */
  @Override
  public double[] getOrigColor ( double[] color )
  {
    if ( colorD == null ) return null;

    int bands = colorD.length;
    if ( color == null ||
         color.length < bands )
      color = Formula.cloneArray( colorD );
    else
      System.arraycopy( colorD, 0, color, 0, bands );

    return color;
  }

  /**
   * Sets the actual color (for the given BSSRDF data).
   * This color will be used in subsequent BSSRDF calculations.
   * The color array won't be altered.
   *
   * @param  color Color definition - {@code double[C]} array
   *               ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *               color system/ or greater /for more complex spectral color sampling/).
   */
  @Override
  public void setColor ( double[] color )
  {
    if ( color != null )
      this.colorD = color;
  }

  /**
   * Scatter ray according to the light model.
   *
   * @param  mf   The {@link MicroFacet} (point on surface of some solid).
   * @param  in   Input direction (should look from the hit point)..
   * @param  u    Random uniform number.
   * @param  v    Random uniform number.
   * @param  mask Which light components are to be computed?
   * @param  out  Output direction in world coordinates.
   * @param  brdf Color definition - {@code double[C]} array
   *              ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *              color system/ or greater /for more complex spectral color sampling/).
   * @return Probability density function.
   */
  @Override
  public double scatter ( MicroFacet mf,
                          double[] in,
                          double u,
                          double v,
                          int mask,
                          double[] out,
                          double[] brdf )
  {
    if ( mf   == null ||
         in   == null ||
         brdf == null ||
         out  == null )
      return 0.0;

    double[] inLoc = new double[ 3 ];
    worldToLocal( mf.normal, mf.tu, in, inLoc );

      // compute output direction:
    double ksi = rnd.randomDouble( 0.0, 1.0 );
    if ( ksi < 0.5 )
        // sample diffuse term:
      MonteCarlo.cosineSampleHemisphere( u, v, out );
    else
    {
        // sample lobe:
      distribution.sample( mf, inLoc, u, v, out, brdf );
    }

    colorBSSRDF( false, mf, inLoc, out, mask, brdf );
    double pdf = pdf( false, mf, inLoc, out );

    localToWorld( mf.normal, mf.tu, out, out );
    return pdf;
  }

  /**
   * Returns the probability density function.
   *
   * @param in     Input direction.
   * @param normal Normal at the point of intersection.
   * @param out    Output direction.
   */
  @Override
  public double pdf ( boolean worldCoords, MicroFacet mf, double[] in, double[] out )
  {
    if ( in  == null ||
         out == null ||
         mf  == null )
      return 0.0;

      // ensure that coordinates are local:
    double[] inLoc  = new double[ 3 ];
    double[] outLoc = new double[ 3 ];
    if ( worldCoords )
    {
      worldToLocal( mf.normal, mf.tu, in,  inLoc  );
      worldToLocal( mf.normal, mf.tv, out, outLoc );
    }
    else
    {
      System.arraycopy( in,  0, inLoc,  0, in.length  );
      System.arraycopy( out, 0, outLoc, 0, out.length );
    }

    double pdf = 0.0;
    if ( inLoc[2] * outLoc[2] > 0.0 )
      pdf = outLoc[2] / Math.PI;
    pdf += distribution.pdf( in, out );
    return pdf / 2.0;
  }

  /** Returns the type of BRDF. */
  @Override
  public int type ()
  {
    return BSSRDF.COMP_SPECULAR | BSSRDF.COMP_DIFFUSE;
  }

  //==========================================================================
  //  Support methods:

  /**
   * Computes geometric term, which models self-shadowing effects.
   *
   * @param in   Input direction in local coordinates.
   * @param out  Output direction in local coordinates.
   * @param half Half vector.
   */
  protected double G ( double[] in, double[] out, double[] half )
  {
    if ( in   == null ||
         out  == null ||
         half == null )
      return 0.0;

    double dotNI = Math.abs( in[2] );
    double dotNO = Math.abs( out[2] );
    double dotNH = Math.abs( half[2] );
    double dotIH = Math.abs( Geometry.dotProduct3D( in, half ) );
    return Math.min( 1.f,
                     Math.min( ( 2.0 * dotNH * dotNI / dotIH ),
                               ( 2.0 * dotNH * dotNO / dotIH ) ) );
  }

  //==========================================================================
  //  Interface Property:

  public final static String COLOR_D      = "Color diffuse";
  public final static String COLOR_S      = "Color specular";
  public final static String FRESNEL      = "Fresnel";
  public final static String DISTRIBUTION = "Distribution";

/**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

    if ( key.compareTo( COLOR_D ) == 0 &&
         value instanceof double[] )
    {
      colorD = (double[])value;
      return;
    }
    if ( key.compareTo( COLOR_S ) == 0 &&
         value instanceof double[] )
    {
      colorS = (double[])value;
      return;
    }
    if ( key.compareTo( FRESNEL ) == 0 )
    {
      fresnel = (Fresnel)value;
      return;
    }
    if ( key.compareTo( DISTRIBUTION ) == 0 )
    {
      distribution = (MicrofacetDistribution)value;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo( COLOR_D ) == 0 ) return colorD;
    if ( key.compareTo( COLOR_S ) == 0 ) return colorS;
    if ( key.compareTo( FRESNEL ) == 0 ) return fresnel;
    if ( key.compareTo( DISTRIBUTION ) == 0 ) return distribution;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Cook-Torrance shading model";

  /** Object description. */
  private final static String DESCRIPTION = "Cook-Torrance shading model.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"BSSRDF" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}