package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.piece.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Default renderer of CSG scenes.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/DefaultRenderer.java">DefaultRenderer.java</a>
 */
public abstract class DefaultRenderer extends DefaultImageFunction
{
  //==========================================================================
  //  Projection data:

  /** Projection-plane linear transform - horizontal additive term. */
  protected double xA = 0.0;

  /** Projection-plane linear transform - horizontal multiplicator. */
  protected double xK = 1.0;

  /** Projection-plane linear transform - vertical additive term. */
  protected double yA = 0.0;

  /** Projection-plane linear transform - vertical multiplicator. */
  protected double yK = 1.0;

  /** Hash-multiplier for number of light sources. */
  protected final static long HASH_LIGHT   = 101L;

  /** Hash-multiplier for textures. */
  protected final static long HASH_TEXTURE = 11L;

  //==========================================================================
  //  Data:

  /** Sampling in direct illumination computation. */
  protected static enum Sampling
  {
    SAMPLING_LIGHTS,
    SAMPLING_BSDF,
    SAMPLING_BOTH,
  }

  /** Actual way of computation of direct illumination. */
  protected Sampling sampling = Sampling.SAMPLING_BOTH;

  /** Supported reflections. */
  protected static enum Reflections
  {
    ALL,
    DIFFUSE,
    GLOSSY,
    SPECULAR,
  }

  /** Actual used reflections. */
  protected Reflections reflections = Reflections.ALL;

  /** Accelerator: actual Ray-tracing scene (everything but intersections). */
  protected RTScene scene = null;

  /** Accelerator: object used for intersections. */
  protected Intersectable inter = null;

  /** Accelerator: camera object. */
  protected RayGenerator camera = null;

  /** Number of shadow rays. */
  protected int numShadowRays;

  /** Actual sample order. */
  protected int actOrder;

  /** Actual cardinality of sample set. */
  protected int actTotal;

  /** Is it aplied occlusion? */
  protected boolean occlusion = true;

  /** {@code True} if indirect illumination should be used. */
  protected boolean useTextures = true;

  /** Was the renderer initialized? */
  protected boolean isInit = false;

  /** Mask used in {@link MicroFacet#assertAttributes}. */
  protected int dataMask;

  /** Random number generator. */
  protected final RandomJames rnd = new RandomJames();

  //==========================================================================
  //  Interface ImageFunction:

  /**
   * Asserts the {@link #scene} accelerator and
   * does first-time algorithm setup.
   *
   * @return {@code true} if failed.
   */
  protected boolean assertScene ()
  {
    if ( scene != null ) return false;

    scene = (RTScene)getInterface( PL_OUTPUT, IFACE+"RTScene" );
    if ( scene == null ) return true;

      // further setup:
    background = scene.getBackground();

    dataMask = MicroFacet.VALID_COLOR  |
               MicroFacet.VALID_NORMAL |
               MicroFacet.VALID_BSSRDF;

    if ( useTextures )
      dataMask |= MicroFacet.VALID_TEXTURES;

    if ( inter == null )
    {
      inter = (Intersectable)getInterface( PL_INTERSECTION, IFACE+"Intersectable" );
      if ( inter == null )
        inter = scene;
    }

    LightSource[] lights = scene.getLights();
    if ( lights != null )
      for ( LightSource ls : lights )
        ls.setIntersectable( inter );

    camera = scene.getCamera();

    return( camera == null );
  }

  /** Prepare scene for rendering. */
  protected boolean init ()
  {
    return true;
  }

    //==========================================================================
  //  Interface TimeDependent:

  /**
   * Retrieve the valid bounds for time variable.
   *
   * @param  pre Pre-allocated {@code double[2]} array to hold the result (can be
   *             {@code null}).
   * @return {@code result[0]} .. minimum time value, {@code result[1]} .. maximum
   *         time value.
   * @see    #getTime
   * @see    #setTime
   */
  @Override
  public double[] getTimeInterval ( double[] pre )
  {
    if ( assertScene() )
      return null;

    return scene.getTimeInterval( pre );
  }

  /**
   * Returns the actual time value for the object (used in animations).
   *
   * @return Actual time (in seconds).
   * @see    #setTime
   */
  @Override
  public double getTime ()
  {
    if ( assertScene() )
      return 0.0;

    return scene.getTime();
  }

  /**
   * Sets the actual time value for the object (used in animations).
   *
   * @param  time New scene time (in seconds).
   * @see    #getTime
   * @see    #getTimeInterval
   */
  @Override
  public void setTime ( double time )
  {
    if ( !assertScene() )
      scene.setTime( time );
  }

  //==========================================================================
  //  Interface ImageFunction:

    /**
   * Sets bounds (scaling factors) for the image function (e.g. projection-plane mapping).
   *
   * @param  Xmin Minimum X coordinate.
   * @param  Xmax Maximum X coordinate.
   * @param  Ymin Minimum Y coordinate.
   * @param  Ymax Maximum Y coordinate.
   * @see    #getBounds
   */
  @Override
  public void setBounds ( double Xmin, double Xmax, double Ymin, double Ymax )
  {
    if ( assertScene() )
      return;

      // projection-plane bounds:
    xMin = Xmin; xMax = Xmax;
    if ( xMax <= xMin ) xMax = xMin + 1.0;
    yMin = Ymin; yMax = Ymax;
    if ( yMax <= yMin ) yMax = yMin + 1.0;
    camera.setAspectRatio( (xMax - xMin) / (yMax - yMin) );

      // projection-plane linear transform:
    double[] vp = camera.getViewport( null );
    if ( vp == null ) return;

    xK = (vp[1] - vp[0]) / (xMax - xMin);
    xA = vp[0] - xMin * xK;
    yK = (vp[3] - vp[2]) / (yMax - yMin);
    yA = vp[2] - yMin * yK;
  }

  /**
   * Computes one image sample. Passes parameters for hidden sampling.
   *
   * @param  x X coordinate of the sample.
   * @param  y Y coordinate of the sample.
   * @param  order Order of this sample inside the actual pixel.
   * @param  total Total number of samples in the actual pixel.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (sample signature) used for adaptive subsampling.
   * @see    #getSample(double,double,double[])
   * @see    #getBounds
   */
  @Override
  public long getSample ( double x, double y, int order, int total, double[] color )
  {
      // initial color:
    if ( color == null ) return -1L;
    Arrays.fill( color, 0.0 );

    if ( assertScene() )
      return 0L;

    if ( !isInit )
     if ( !init() )
       return 0;

      // verbose tracing:
    if ( LogFile.debugging &&
         LogFile.intProperty( "y", -1 ) == (int)Math.floor( y ) &&
         LogFile.intProperty( "x", -1 ) == (int)Math.floor( x ) )
    {
      LogFile.tracing = true;
      LogFile.log( "Tracing pixel [ " + (int)Math.floor( x ) + ", " + (int)Math.floor( y ) + " ]:" );
    }

      // compute the primary ray:
    double[] p0 = new double[ 3 ];        // ray coordinates
    double[] p1 = new double[ 3 ];
    if ( !camera.getRay( xA + xK * x, yA + yK * y, order, total, p0, p1 ) )
      return 11L;

    actOrder  = order;
    actTotal  = total;
    long hash = shade( 0, 1.0, p0, p1, color );

    if ( LogFile.tracing )
    {
      LogFile.log( String.format( Locale.US, "Sample color: [ %.3f, %.3f, %.3f ]", color[0], color[1], color[2] ) );
      LogFile.tracing = false;
    }

    return hash;
  }

  /**
   * Recursive shading function - computes color contribution of the given ray (shot from the
   * origin {@code p0} into direction vector {@code p1}). Recursion is stopped
   * by a hybrid method: {@code importance} and {@code level} are checked.
   *
   * @param  level Actual recursion depth.
   * @param  importance Importance of the actual ray.
   * @param  p0 Ray origin.
   * @param  p1 Direction vector of the ray.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (ray sub-signature) used for adaptive subsampling.
   * @see    #minImportance
   * @see    #maxLevel
   * @see    #scene
   * @see    #inter
   * @todo   Some better way of computing importance.
   */
  protected long shade ( int level, double importance, double[] p0, double[] p1,
                         double[] color )
  {
    return 0;
  }

  //==========================================================================
  //  Methods for direct illumination:

  /**
   * Compute accure direct illumination.
   *
   * @param  mf    The {@link MicroFacet} (point on surface of some solid).
   * @param  out   Outcoming direction.
   * @param  color Pre-allocated <tt>double[C]</tt> array to hold result - light intensity
   *               coming from the source to the given point (<tt>C</tt> can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @todo   Support of multiple important sampling.
   * @todo   Maybe the number of shadow rays for multiple-importance sampling
   *         of light source and BSDF should be different.
   *         (options - both, lights, bsdf - are not sufficient)
   */
  protected void directIllumination ( MicroFacet mf, double[] out, double[] color )
  {
    LightSource[] light = scene.getLights();

    int i, j;
    double invQuantity = 1.0 / numShadowRays;
    double[] Li = new double[ color.length ];
    for ( j = 0; j < light.length; j++ )
    {
      Arrays.fill( Li, 0.0 );
      for ( i = 0; i < numShadowRays; i++ )
      {
        if ( sampling == Sampling.SAMPLING_BOTH ||
             sampling == Sampling.SAMPLING_LIGHTS )
          sampleLightSource( mf,
                             light[j],
                             rnd.randomDouble( 0.0, 1.0 ),
                             rnd.randomDouble( 0.0, 1.0 ),
                             out,
                             Li );

        if ( sampling == Sampling.SAMPLING_BOTH ||
             sampling == Sampling.SAMPLING_BSDF )
          sampleBSDF( mf,
                      light[j],
                      rnd.randomDouble( 0.0, 1.0 ),
                      rnd.randomDouble( 0.0, 1.0 ),
                      out,
                      Li );
      }
      for ( i = 0; i < color.length; i++ )
        color[i] += invQuantity * Li[i];
    }
  }

  /**
   * Sample light source with multiple importance sampling.
   *
   * @param  mf    The {@link MicroFacet} (point on surface of some solid).
   * @param  light Sampled light source.
   * @param  u     Uniform random number (for light sampling).
   * @param  v     Uniform random number (for light sampling).
   * @param  out   Outcoming direction of light.
   * @param  Ld    Pre-allocated <tt>double[C]</tt> array to hold result - light intensity
   *               coming from the source to the given point (<tt>C</tt> can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @todo   Should be in some "neutral" modul, because other renders
   *         may need it as well (e.g. Metropolis Light Transport).
   */
  protected void sampleLightSource ( MicroFacet mf,
                                     LightSource light,
                                     double u,
                                     double v,
                                     double[] out,
                                     double[] Ld )
  {
    if ( Ld    == null ||
         light == null ||
         out   == null ||
         mf    == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD ) != 0 )
      return;

    double[] pdfLight = new double[ 1 ];         // probability density function of light
    double pdfBRDF    = 0.0;                     // probability density function of BRDF
    double[] in       = new double[ 3 ];         // input direction
    double[] Li       = new double[ Ld.length ]; // input radiance
    double[] brdf     = new double[ Ld.length ]; // brdf of surface

      // set supported types of materials:
    int flags = reflections.ordinal();
    switch ( flags )
    {
      case 0:
        flags = BSSRDF.COMP_ALL;
        break;

      case 1:
        flags = BSSRDF.COMP_DIFFUSE;
        break;

      case 2:
        flags = BSSRDF.COMP_GLOSSY;
        break;

      case 3:
        flags = BSSRDF.COMP_SPECULAR;
        break;
    }

      // sample light:
    boolean visible = light.sample( mf, u, v, Li, in, pdfLight, occlusion );
    if ( visible           &&
         pdfLight[0] > 0.0 &&
         ( Li[0] > 0.0 || Li[1] > 0.0 || Li[2] > 0.0 ) )
    {
        // compute BRDF:
      mf.bssrdf.colorBSSRDF( true, mf, out, in, flags, brdf );
      double cos = Geometry.dotProduct3D( in, mf.normal );

        // compute contribution:
      double invPdf = 1.0 / pdfLight[0];
      if ( light.isDeltaLight() )
        for ( int i = 0; i < Ld.length; i++ )
          Ld[i] += brdf[i] * Li[i] * cos * invPdf;
      else
      {
        pdfBRDF = mf.bssrdf.pdf( true, mf, out, in );
        if ( pdfBRDF > 0.0 )
        {
          double weight = MonteCarlo.balanceHeuristic( 1.0, pdfLight[0], 1.0, pdfBRDF );
          for ( int i = 0; i < Ld.length; i++ )
            Ld[i] += brdf[i] * Li[i] * cos * invPdf * weight;
        }
      }
    }
  }

  /**
   * Sample BSDF with multiple importance sampling.
   *
   * @param  mf    The {@link MicroFacet} (point on surface of some solid).
   * @param  light Sampled light source.
   * @param  u1    Uniform random number (for light sampling).
   * @param  v1    Uniform random number (for light sampling).
   * @param  u2    Uniform random number (for BRDF sampling).
   * @param  v2    Uniform random number (for BRDF sampling).
   * @param  out   Outcoming direction.
   * @param  Ld    Pre-allocated <tt>double[C]</tt> array to hold result - light intensity
   *               coming from the source to the given point (<tt>C</tt> can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @todo   Should be in some "neutral" modul, because other renders
   *         may need it as well (e.g. Metropolis Light Transport).
   */
  protected void sampleBSDF ( MicroFacet mf,
                              LightSource light,
                              double u,
                              double v,
                              double[] out,
                              double[] Ld )
  {
    if ( Ld    == null ||
         light == null ||
         out   == null ||
         mf    == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD ) != 0 )
      return;

    if ( light.isDeltaLight() )
      // It's a waste of time to try to hit the delta light source.
      return;

      // set supported types of materials:
    int flags = reflections.ordinal();
    switch ( flags )
    {
      case 0:
        flags = BSSRDF.COMP_ALL;
        break;

      case 1:
        flags = BSSRDF.COMP_DIFFUSE;
        break;

      case 2:
        flags = BSSRDF.COMP_GLOSSY;
        break;

      case 3:
          // specular material: only one reflected ray
          // (which is cast somewhere else)
        return;
    }

    double[] pdfLight = new double[ 1 ];         // probability density function of light
    double pdfBRDF    = 0.0;                     // probability density function of BRDF
    double[] in       = new double[ 3 ];         // input direction
    double[] brdf     = new double[ Ld.length ]; // brdf of surface

      // variables in and out are swapped, because path-tracing is backward
      // to light transport in real world, so input direction to BSDF sampling
      // is actually output direction in light transport in real world
    out[0] = -out[0];
    out[1] = -out[1];
    out[2] = -out[2];
    pdfBRDF = mf.bssrdf.scatter( mf, out, u, v, flags, in, brdf );

      // sample BRDF:
    if ( pdfBRDF > 0.0 &&
         ( brdf[0] > 0.0 || brdf[1] > 0.0 || brdf[2] > 0.0 ) )
    {
      in[0] = -in[0];
      in[1] = -in[1];
      in[2] = -in[2];
      pdfLight[0] = light.pdf( mf.coord, mf.normal, in );
      in[0] = -in[0];
      in[1] = -in[1];
      in[2] = -in[2];
      if ( pdfLight[0] > 0.0 )
      {
        double weight = MonteCarlo.balanceHeuristic( 1.0, pdfBRDF, 1.0, pdfLight[0] );
        MicroFacet m = inter.intersection( mf.coord, in );
        double[] distance = new double[ 3 ];
        Geometry.vector( m.coord, mf.coord, distance );
        if ( m != null )// &&
//             m.near( 0.9, distance ) )
        {
          double invPdf = 1.0 / pdfBRDF;
          double cos    = Geometry.dotProduct3D( mf.normal, in );
          for ( int i = 0; i < Ld.length; i++ )
            Ld[i] += brdf[i] * mf.color[i] * cos * invPdf * weight;
        }
      }
    }
  }

  /**
   * Handle specular reflection which is specific so it's not handled
   * sample BSDF like the other types of reflection (diffuse and glossy).
   */
  protected void specular ( int level,
                            MicroFacet mf,
                            double[] out,
                            double[] L )
  {
    if ( mf  == null ||
         out == null ||
         L   == null )
      return;

      // possibly end tracing:
    if ( level > 3 )
    {
      //newImportance = importance * cos * brdfAvrg;
      if ( rnd.randomDouble( 0.0, 1.0 ) < 0.5 )
        return;
    }

    double[] in   = new double[ 3 ];
    double[] brdf = new double[ 3 ];
    mf.bssrdf.scatter( mf, out, rnd.randomDouble( 0.0, 1.0 ),
                       rnd.randomDouble( 0.0, 1.0 ), BSSRDF.COMP_SPECULAR,
                       in, brdf );
    double cos = Math.abs( Geometry.dotProduct3D( mf.normal, in ) );

      // add contribution:
    if ( brdf[0] > 0.0 || brdf[1] > 0.0 || brdf[2] > 0.0 )
    {
      double[] Li = new double[ L.length ];
      shade( level, 1.0, mf.coord, in, Li );
      for ( int i = 0; i < Li.length; i++ )
        L[i] += brdf[i] * Li[i] * cos;
    }
  }

}