package cz.cuni.jagrlib.testing;

import java.util.TreeSet;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Depth-sorting example - incorrect behavior for general scenes with large faces.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/DepthSort.java">DepthSort.java</a>
 */
public class DepthSort extends Piece implements BrepIterator
{
  //==========================================================================
  //  Data:

  /** Actual sort order. */
  protected int sortOrder = SORT_ASCENDING;

  /** Actual {@link Brep} context to sort. */
  protected int ctx = Brep.CTX_DEFAULT;

  /** {@link Brep}-attribute name holding projected coordinates. */
  protected String coordsName = Brep.COORD_PROJ;

  /** Accelerator: projected coordinates. */
  protected int attrCoords;

  /** Accelerator: Brep database. */
  protected Brep data;

  /** Tmp: vertex coordinate. */
  protected double[] vc = new double[ 4 ];

  /** Tmp: vertices-in-face iterator. */
  protected BrepIterator vit;

  /** Pre-computed face ordering. */
  protected int[] arr;

  /** Index to {@link #arr} - item to be returned next. */
  protected int n;

  //==========================================================================
  //  Support:

  /** Sorted-set item ({@link Brep} face). */
  protected class Face implements Comparable< Face >
  {
    /** Id of the face. */
    public int id;

    /** Depth of the center-of-gravity. */
    protected double depth;

    public Face ( int face )
    {
      id  = face;
      vit = data.vertexInFaceIterator( face, vit );
      depth = 0.0;
      int count = 0;
      int vertex;
      double[] vc = new double[ 4 ];

      while ( (vertex = vit.next()) != Brep.NULL )
      {
        data.getAttribute( attrCoords, vertex, vc );
        Geometry.toCartesian3D( vc, vc );
        depth += vc[2];
        count++;
      }

      depth /= count;
    }

    /** Natural ordering (depth). */
    public int compareTo ( Face f )
    {
      if ( f == null )
        throw new NullPointerException();

      if ( this == f )
        return 0;

      if ( depth != f.depth )
        return( (sortOrder == SORT_ASCENDING) == (depth < f.depth) ? -1 : 1 );

      return( id < f.id ? -1 : 1 );
    }
  }

  /** Precomputes the ordering, returns {@code true} if failed. */
  protected boolean sort ()
  {
    if ( arr != null ) return false;
    n = 0;

    data = (Brep)getInterface( PL_DATA, IFACE+"Brep" );
    if ( data == null ) return true;

    attrCoords = data.getAttributeId( Brep.VERTEX, ctx, coordsName );
    if ( attrCoords == Brep.NULL ) return true;

    TreeSet< Face > ts = new TreeSet< Face >();
    BrepIterator it = data.faceIterator( ctx, null );
    int face;
    while ( (face = it.next()) != Brep.NULL )
      ts.add( new Face( face ) );

    arr  = new int[ ts.size() ];
    face = 0;
    for ( Face f : ts )
      arr[face++] = f.id;

    return false;
  }

  //==========================================================================
  //  Interface BrepIterator:

  /** Reset the iterator. */
  public void reset ()
  {
    sort();
  }

  /**
   * Returns id of the next object.
   *
   * @return {@link Brep#NULL} if there is no next object available.
   */
  public int next ()
  {
    if ( sort() ) return Brep.NULL;

    if ( n < arr.length )
      return arr[n++];

    return Brep.NULL;
  }

  /**
   * Fills the provided array by all items (pre-allocated array could be {@code null}
   * or too small - in these cases new one will be created).
   * End of the data is marked using {@link Brep#NULL}.
   */
  public int[] toArray ( int[] pre )
  {
    if ( sort() ) return null;

    if ( pre == null ||
         pre.length < arr.length )
      pre = new int[ arr.length ];

    System.arraycopy( arr, 0, pre, 0, arr.length );
    if ( pre.length > arr.length )
      pre[ arr.length ] = Brep.NULL;

    return pre;
  }

  /**
   * How many items remain in the set?
   * Non-mandatory - must return {@link Brep#NULL} if not implemented.
   */
  public int left ()
  {
    if ( sort() ) return Brep.NULL;

    return( arr.length - n );
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // SORT_ORDER:
    if ( key.compareTo(SORT_ORDER) == 0 )
    {
      sortOrder = intProperty( value, sortOrder, SORT_ASCENDING, SORT_DESCENDING );
      arr       = null;
      return;
    }
      // CONTEXT:
    if ( key.compareTo(CONTEXT) == 0 )
    {
      ctx = intProperty( value, ctx );
      arr = null;
      return;
    }
      // PROJ_COORDS:
    if ( key.compareTo(PROJ_COORDS) == 0 )
    {
      coordsName = stringProperty( value, coordsName );
      arr        = null;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(SORT_ORDER) == 0 )
      return sortOrder;
    if ( key.compareTo(CONTEXT) == 0 )
      return ctx;
    if ( key.compareTo(PROJ_COORDS) == 0 )
      return coordsName;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Depth-sorting pilot";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "BrepIteratorToBrep";

  /** Object description. */
  private final static String DESCRIPTION = "Simple depth-sorting order.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + '.' + C_BREP;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"BrepIterator" );
      t.newOutputPlug( PL_DATA,  IFACE+"Brep" );
        // Property:
      t.propBegin( SORT_ORDER, TYPE_INTEGER, "Sorting order", true );
        t.propDefault( SORT_ASCENDING );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Ascending",  SORT_ASCENDING,  "Face with less Z-coordinate goes first" );
        t.propEnum( "Descending", SORT_DESCENDING, "Face with greater Z-coordinate goes first" );
      t.propEnd();
      t.propBegin( CONTEXT, TYPE_INTEGER, "Brep context", false );
        t.propDefault( Brep.CTX_DEFAULT );
      t.propEnd();
      t.propBegin( PROJ_COORDS, TYPE_STRING, "Name of projected-coords attribute", false );
        t.propDefault( Brep.COORD_PROJ );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
