package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.piece.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Direct light map visualizer for JGL scenes.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 859 $ $Date: 2009-02-21 00:14:16 +0100 (so, 21 II 2009) $ $Author: Tomas $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/DirectLightMapVisualizer.java">DirectLightMapVisualizer.java</a>
 * @todo    Hash-value for adaptive jittering.
 */
public class DirectLightMapVisualizer extends DefaultRenderer
{
  //==========================================================================
  //  Data:

  /** {@code True} if caustics should be used. */
  protected boolean useCaustics;

  /** {@code True} if global illumination should be used. */
  protected boolean useGlobal;

  /** Are they used shadow photons? */
  protected boolean shadowPhotons = false;

  /** Light map for global (direct,indirect) photon storing. */
  protected LightMap global;

  /** Caustic map for cautic photon storing. */
  protected LightMap caustic;

  /** Shadow map for shadow photon storing. */
  protected LightMap shadow;

  /** Volume map for volume photon storing. */
  protected LightMap volume;

  /** Light tracer to compute illumination of scene. */
  protected LightTracing tracer;

  //==========================================================================
  //  Support functions:

  /** Prepare scene for rendering. */
  @Override
  protected boolean init ()
  {
    if ( scene == null )
      scene = (RTScene)getInterface(       PL_OUTPUT,       IFACE+"RTScene"       );
    if ( inter == null )
      inter = (Intersectable)getInterface( PL_INTERSECTION, IFACE+"Intersectable" );
    if ( scene == null ||
         inter == null )
      return false;

    if ( global == null )
      global = (LightMap)getInterface(     PL_PHOTON,  IFACE+"LightMap"     );
    if ( tracer == null )
      tracer = (LightTracing)getInterface( PL_TRIGGER, IFACE+"LightTracing" );
    if ( caustic == null )
      caustic = (LightMap)getInterface(    PL_CAUSTIC, IFACE+"LightMap"     );
    if ( shadow  == null )
      shadow = (LightMap)getInterface(     PL_SHADOW,  IFACE+"LightMap"     );

    if ( tracer      != null &&
         reflections != Reflections.SPECULAR )
      tracer.illuminateScene();

    isInit = true;

    return true;
  }

  //==========================================================================
  //  Interface ImageFunction:

  /**
   * Recursive shading function - computes color contribution of the given ray (shot from the
   * origin {@code p0} into direction vector {@code p1}). Recursion is stopped
   * by a hybrid method: {@code importance} and {@code level} are checked.
   *
   * @param  level Actual recursion depth.
   * @param  importance Importance of the actual ray.
   * @param  p0 Ray origin.
   * @param  p1 Direction vector of the ray.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (ray sub-signature) used for adaptive subsampling.
   * @see    #minImportance
   * @see    #maxLevel
   * @see    #scene
   * @see    #inter
   * @todo   Some better way of computing importance.
   */
  @Override
  protected long shade ( int level, double importance, double[] p0, double[] p1,
                         double[] color )
  {
      // get the intersection:
    MicroFacet mf = inter.intersection( p0, p1 );
    if ( mf == null )                     // no intersection => use background color
    {
      System.arraycopy( background, 0, color, 0, Math.min( color.length, background.length ) );
      return 1L;
    }

      // compute and log important attributes of the intersection point:
    if ( mf.assertAttributes( dataMask ) != 0 ) return 0L;
    if ( mf.assertAttributes( MicroFacet.VALID_COORD ) == 0 && LogFile.tracing )
    {
      LogFile.log( String.format( Locale.US, "Intersection%d at t = %.4f with '%s' (%s,%s), p1.N: %.3f, mask: %04x",
                                  level, mf.t, mf.geometry.getClass().getName(),
                                  mf.enter ? "enter" : "leave", mf.front ? "front" : "back",
                                  Geometry.dotProduct3D( p1, mf.normal ), dataMask ) );

      LogFile.log( String.format( Locale.US, "  A: [ %.4f, %.4f, %.4f ], p1: [ %.4f, %.4f, %.4f ], N: [ %.4f, %.4f, %.4f ]",
                                  mf.coord[0], mf.coord[1], mf.coord[2],
                                  p1[0], p1[1], p1[2],
                                  mf.normal[0], mf.normal[1], mf.normal[2] ) );
    }

      // hash code:
    long hash = mf.sceneNode.hashCode();  // init: hash code of the intersected solid

      // apply textures:
    if ( useTextures &&
         mf.textures != null )
      for ( Texture tex : mf.textures )
        hash = hash * HASH_TEXTURE + tex.compute( mf );

    Arrays.fill( color, 0.0 );
    p1[0] = -p1[0];
    p1[1] = -p1[1];
    p1[2] = -p1[2];
    double[] Le   = new double[ color.length ];    // emission
    double[] Lc   = new double[ color.length ];    // caustics
    double[] Ld   = new double[ color.length ];    // global illumination
    double[] brdf = new double[ color.length ];
    mf.bssrdf.getOrigColor( brdf );
    int i;

      // compute direct illumination:
    if ( useGlobal )
    {
        // just estimate density:
      global.irradiance( mf.coord, mf.normal, Ld );
      for ( i = 0; i < color.length; i++ )
        Ld[i] *= brdf[i];

        // handle specular reflections:
      if ( mf.bssrdf.type() == BSSRDF.COMP_SPECULAR )
        specular( level, mf, p1, Ld );
    }

      // compute caustics:
    if ( useCaustics && caustic != null )
    {
      caustic.irradiance( mf.coord, mf.normal, Lc );
      for ( i = 0; i < color.length; i++ )
        Lc[i] *= brdf[i];
    }

    for ( i = 0; i < color.length; i++ )
      color[i] += Le[i] + Ld[i] + Lc[i];

    return hash;
  }

  //==========================================================================
  //  Interface Property:

  public final static String REFLECTIONS = "Reflections";
  public final static String GLOBAL      = "Global";
  public final static String CAUSTICS    = "Caustics";
  public final static String TEXTURES    = "Textures";
  public final static String LOG         = "Log tracing";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // REFLECTIONS:
    if ( key.compareTo(REFLECTIONS) == 0 )
    {
      reflections = enumProperty( value, reflections );
      return;
    }
      // GLOBAL:
    if ( key.compareTo(GLOBAL) == 0 )
    {
      useGlobal = booleanProperty( value, useGlobal );
      return;
    }
      // CAUSTICS:
    if ( key.compareTo(CAUSTICS) == 0 )
    {
      useCaustics = booleanProperty( value, useCaustics );
      return;
    }
      // TEXTURES:
    if ( key.compareTo(TEXTURES) == 0 )
    {
      useTextures = booleanProperty( value, useTextures );
      return;
    }
      // LOG:
    if ( key.compareTo(LOG) == 0 )
    {
      LogFile.tracing = booleanProperty( value, LogFile.tracing );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(REFLECTIONS) == 0 )
      return reflections.ordinal();
    if ( key.compareTo(GLOBAL) == 0 )
      return useGlobal;
    if ( key.compareTo(CAUSTICS) == 0 )
      return useCaustics;
    if ( key.compareTo(TEXTURES) == 0 )
      return useTextures;
    if ( key.compareTo(LOG) == 0 )
      return LogFile.tracing;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Direct light map visualizer";

  /** Object name. */
  private final static String TEMPLATE_NAME = "Direct light map visualizer";

  /** Object description. */
  private final static String DESCRIPTION = "Visualizer of light maps";

  /** Object category. */
  protected final static String CATEGORY = C_3D + "." + C_RENDER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,        IFACE+"ImageFunction" );
      t.newOutputPlug(    PL_OUTPUT,       IFACE+"RTScene" );
      t.newOutputPlug(    PL_INTERSECTION, IFACE+"Intersectable" );
      t.newOptOutputPlug( PL_TRIGGER,      IFACE+"LightTracing" );
      t.newOptOutputPlug( PL_PHOTON,       IFACE+"LightMap" );
      t.newOptOutputPlug( PL_CAUSTIC,      IFACE+"LightMap" );
      t.newOptOutputPlug( PL_SHADOW,       IFACE+"LightMap" );
      t.newOptOutputPlug( PL_VOLUME,       IFACE+"LightMap" );
        // Property:
      t.propBegin( REFLECTIONS, TYPE_INTEGER, "Reflections", true );
        t.propDefault( Reflections.ALL.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "All",      Reflections.ALL.ordinal(),      "All types of materials are used." );
        t.propEnum( "Diffuse",  Reflections.DIFFUSE.ordinal(),  "Only diffuse materials are used." );
        t.propEnum( "Glossy",   Reflections.GLOSSY.ordinal(),   "Only glossy materials are used." );
        t.propEnum( "Specular", Reflections.SPECULAR.ordinal(), "Only specular materials are used." );
      t.propEnd();
      t.propBegin( GLOBAL,   TYPE_BOOLEAN, "Compute global illumination?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( CAUSTICS, TYPE_BOOLEAN, "Compute caustics?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( TEXTURES, TYPE_BOOLEAN, "Compute textures?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( LOG,      TYPE_BOOLEAN, "Log the tracing?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}