package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.xml.JGLScript;

/**
 * Directional light source shining uniformly in one direction.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 858 $ $Date: 2009-02-17 23:48:39 +0100 (út, 17 II 2009) $ $Author: Tomas $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/DirectionalLightSource.java">DirectionalLightSource.java</a>
 */
public class DirectionalLightSource extends DefaultLightSource
{
  //==========================================================================
  //  Data:

  {
    scriptName = "data/directional.jgl";
  }

  /** Number of shooted photons (for methods based on photon tracing). */
  protected int[] photons = { 0 };

  /** Direction of light. */
  protected double[] direction = new double[ 3 ];

  /**
   * Color of the source: {@code 1D} for monochrome source, {@code 3D} for <b>RGB</b>
   * color space, or more for complex spectrum sampled spaces.
   */
  protected double[] lightColor = new double[] { 1.0 };

  //==========================================================================
  //  Support:

  /**
   * Set light source parameters.
   *
   * @param  photons Number of photons to shoot.
   * @param  dir     Direction of the light.
   * @param  color   Source color/intensity (can be either 1D /for monochrome
   *                 computing/, 3D /for RGB color system/ or greater
   *                 /for more complex spectral color sampling/).
   */
  public boolean setParams ( int photons, double[] dir, double[] color )
  {
    if ( dir == null || dir.length < 3 )
      return false;

    this.photons[0] = photons;

      // direction:
    if ( dir == null )
      this.direction = null;
    else
    {
      if ( this.direction == null || this.direction.length != 3 )
        this.direction = new double[ 3 ];

      for ( int i = 0; i < 3 && i < dir.length; i++ )
        this.direction[i] = dir[i];
    }
    Geometry.normalize3D( direction );

      // source color/intensity:
    if ( color == null )
    {
      if ( lightColor == null ||
           lightColor.length != 1 )
        lightColor = new double[ 1 ];

      lightColor[0] = 1.0;
    }
    else
    {
      if ( lightColor == null ||
           lightColor.length != color.length )
        lightColor = new double[ color.length ];

      System.arraycopy( color, 0, lightColor, 0, color.length );
    }

    return true;
  }

  /** Read data from JGL script. */
  protected void readScript ()
  {
    if ( scriptName == null ) return;

    JGLScript script = new JGLScript();
    Map< String, Object > map = script.interpret( scriptName );
    if ( map != null )
    {
      Object o;

      if ( (o = map.get( PHOTONS ) ) != null &&
           o instanceof int[] )
        photons = (int[])o;

      if ( (o = map.get( DIR ) ) != null &&
           o instanceof double[] )
        direction = (double[])o;

      if ( (o = map.get( KEY_LIGHT ) ) != null &&
           o instanceof double[] )
        lightColor = (double[])o;
    }
  }

  /** For debugging purposes only. */
  @Override
  public String toString ()
  {
    StringBuffer sb = new StringBuffer( getClass().getName() );
    sb.append( "[photons=" )
      .append( Formula.vectorToString( photons ) )
      .append( ",direction=" )
      .append( Formula.vectorToString( direction  ) )
      .append( ",color=" )
      .append( Formula.vectorToString( lightColor ) );
    return sb.toString();
  }

  //==========================================================================
  //  Interface LightSource:

  /**
   * Lighting test from the source to the given {@link MicroFacet}.
   *
   * @param  mf The checked {@link MicroFacet} (point on surface of some solid).
   * @return {@code true} if the light source lights on the given point.
   * @see    #getIntensity
   * @see    #getDirection
   */
  public boolean lights ( MicroFacet mf )
  {
    if ( mf == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD |
                              MicroFacet.VALID_NORMAL ) != 0 ) return true;

    if ( mf.normal[0] * direction[0] +
         mf.normal[1] * direction[1] +
         mf.normal[2] * direction[2] <= 0.0 ) return false;

      // check the intersection of a light ray with the scene:
    if ( assertInter() )
      return true;

      // does the sample point see the front side of the MicroFacet?
    MicroFacet c = inter.intersection( mf.coord, direction  );

    return( c == null ||
            c.far( 1.0, direction  ) );
  }

  /**
   * Returns light intensity coming from the source to the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  color Pre-allocated {@code double[C]} array to hold result - light intensity
   *               coming from the source to the given point ({@code C} can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @return {@code true} if the light source can light on the given point (<b>cannot</b>
   *         replace the {@link #lights} call!).
   * @see    #lights
   * @see    #getDirection
   */
  public boolean getIntensity ( MicroFacet mf, double[] color )
  {
    if ( color.length <= lightColor.length )
      System.arraycopy( lightColor, 0, color, 0, color.length );
    else
      Arrays.fill( color, lightColor[0] );

    return true;
  }

  /**
   * Returns light direction between the source and the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  dir Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *             from the {@link MicroFacet} to the light source.
   * @return {@code false} if the light source is not directional (thus {@code dir}
   *         is irrelevant).
   * @see    #lights
   * @see    #getIntensity
   */
  public boolean getDirection ( MicroFacet mf, double[] dir )
  {
    if ( dir        == null || dir.length      < 3 ||
         this.direction   == null || this.direction.length < 3 )
      return false;

    dir = this.direction;
    return true;
  }

   /**
   * Sample the light by one shadow ray.
   *
   * @param  mf    The {@link MicroFacet} (point on surface of some solid).
   * @param  u     Uniform random number.
   * @param  v     Uniform random number.
   * @param  color Pre-allocated <tt>double[C]</tt> array to hold result - light intensity
   *               coming from the source to the given point (<tt>C</tt> can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @param  dir   Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *               from the {@link MicroFacet} to the light source.
   * @param  pdf   Probability density function (output variable).
   * @return {@code True} if the light is visible.
   */
  @Override
  public boolean sample ( MicroFacet mf,
                          double u,
                          double v,
                          double[] color,
                          double[] dir,
                          double[] pdf,
                          boolean testVisibility )
  {
    if ( color == null ||
         dir   == null ||
         pdf   == null ||
         mf    == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD ) != 0 )
      return false;

    dir[0] = -direction[0];
    dir[1] = -direction[1];
    dir[2] = -direction[2];
    pdf[0] = 1.0 / Geometry.dotProduct3D( mf.normal, dir );

    System.arraycopy( lightColor, 0, color, 0, color.length );

      // check visibility if it's required:
    if ( testVisibility )
    {
      MicroFacet m = inter.intersection( mf.coord, dir );
      return m == null;
    }

    return true;
  }

  /**
   * Generates one photon leaving this light source.
   *
   * @param  u1    Uniform random number.
   * @param  v1    Uniform random number.
   * @param  u2    Uniform random number.
   * @param  v2    Uniform random number.
   * @param  pos   Pre-allocated array which holds a point on the light source
   *               (an origin of direction).
   * @param  dir   Output direction in world coordinates. Pre-allocated array.
   * @param  power Power of given light source (if you wanna get power
   *               of this photon, you have to divide this power by number
   *               of photons). Pre-allocated array.
   * @return Weight of given ray (standart weight = 1.0).
   */
  public double sample ( double u1,
                         double v1,
                         double u2,
                         double v2,
                         double[] pos,
                         double[] dir,
                         double[] power )
  {
    if ( pos   == null ||
         dir   == null ||
         power == null )
      return 0.0;

    MonteCarlo.cosineSampleHemisphere( u2, v2, dir );
//    m.transformVector3D( dir, dir );
    // TODO: problem with matrix light->world

    System.arraycopy( lightColor, 0, power, 0, power.length );

    return 1.0;
  }

  /**
   * Return probability density function.
   *
   * @param point  Point on the surface.
   * @param normal Normal at this point.
   * @param dir    Direction from the point to the light.
   */
  @Override
  public double pdf ( final double[] point, final double[] normal, final double[] dir )
  {
    return 0.0;
  }

  /** Is delta light? */
  @Override
  public boolean isDeltaLight ()
  {
    return true;
  }

  //==========================================================================
  //  Interface Property:

  /** JGL script key: position of up right vertex. */
  public final static String PHOTONS = "photons";

  /** JGL script key: position of up right vertex. */
  public final static String DIR     = "direction";

  /** JGL script key: color. */
  public final static String COLOR   = "color";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // PARAMS:
    if ( key.compareTo(PARAMS) == 0 )
    {
      scriptName = stringProperty( value, scriptName );
      readScript();
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(PHOTONS) == 0 )
      return Formula.cloneArray( photons );
    if ( key.compareTo(DIR) == 0 )
      return Formula.cloneArray( direction  );
    if ( key.compareTo(COLOR) == 0 )
      return Formula.cloneArray( lightColor );
    if ( key.compareTo(PARAMS) == 0 )
      return scriptName;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Directional light source";

  /** Object description. */
  private final static String DESCRIPTION = "Directional light source shining uniformly in one direction.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"LightSource" );
        // Property:
      t.propBegin( COLOR, TYPE_OBJECT, "Light source color (intensity)", false );
      t.propEnd();
      t.propBegin( PARAMS, TYPE_OBJECT, "Light source definition script", true );
        t.propDefault( "data/directional.jgl" );
        t.propManipulator( "ParamJGLScript" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}