package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Feature-based warping/morphing: sample R^2->R^2 mapping using feature set stored in FeratureStore.
 *
 * @version 0.26 $Rev: 474 $ $Date: 2007-06-01 15:54:10 +0200 (pá, 01 VI 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/FeatureDeformation.java">FeatureDeformation.java</a>
 */
public class FeatureDeformation extends DefaultFunctionR2ToR2 implements TimeDependent
{
  //==========================================================================
  //  Data:

  /** Aspect ratio of the source space. */
  protected double aspectRatio = 1.0;

  /** Actual time (for warping). */
  protected double actualTime  = 1.0;

  //==========================================================================
  //  Support:

  protected void setAspectRatio ()
  {
    RasterGraphics image = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    if ( image != null )
      aspectRatio = (image.getWidth() - 1.0) / (image.getHeight() - 1.0);
  }

  //==========================================================================
  //  Interface FunctionR2ToR2:

  /**
   * Compute the mapping: <b>f(x,y)</b>.
   *
   * @param  x X coordinate (1st argument).
   * @param  y Y coordinate (2nd argument).
   * @param  res Result array ({@code res[0]} = U coordinate,
   *             {@code res[1]} = V coordinate).
   */
  @Override
  public void f ( double x, double y, double[] res )
  {
    if ( res == null || res.length < 2 ) return;

    // !!! TODO: implementation of forward deformation mapping !!!
    res[0] = x;
    res[1] = y;

      // test implementation - emphasizing features
    FeatureStore f = (FeatureStore)getInterface( PL_DATA, IFACE+"FeatureStore" );
    if ( f == null ) return;

      // boundary times:
    int t1 = (int)Math.ceil( actualTime ); // target time
    int t0 = t1 > 0 ? t1 - 1 : 0;         // source time

      // find the closest source feature:
    int id        = FeatureStore.NULL;
    int minId     = FeatureStore.NULL;
    double minD   = Double.POSITIVE_INFINITY;
    double[] src  = null;
    double[] srcA = new double[ 2 ];
    double[] srcB = new double[ 2 ];
    double[] p    = { x, y };

    do
    {
      id = f.nextFeature( id, t0 );
      if ( id == FeatureStore.NULL ) break;

        // check the feature type:
      if ( f.featureType( id ) != FeatureStore.FEATURE_ARROW ) continue;

      if ( (src = f.getCoords( id, t0, src )) == null ||
           src.length < 4 ) continue;

        // src[0] to src[3] .. source coordinates of the feature
      srcA[0] = src[0];
      srcA[1] = src[1];
      srcB[0] = src[2];
      srcB[1] = src[3];
      double actD = Geometry.pointSegment2D( p, srcA, srcB );
      if ( actD < minD )                  // this feature is closer
      {
        minD  = actD;
        minId = id;
      }

    } while ( true );

      // artifically paint the feature:
    if ( minD < 0.05 )
    {
      res[0] = f.getParamDouble( minId, t0, "a" );
      res[1] = minD * f.getParamDouble( minId, t0, "b" );
    }
  }

  /**
   * Compute the inverse mapping: <b>f<sup>-1</sup>(u,v)</b>.
   *
   * @param  u U coordinate (1st argument).
   * @param  v V coordinate (2nd argument).
   * @param  res Result array ({@code res[0]} = X coordinate,
   *             {@code res[1]} = Y coordinate).
   * @return {@code true} if this service is implemented.
   */
  @Override
  public boolean fInv ( double u, double v, double[] res )
  {
    if ( res == null || res.length < 2 ) return false;

    // !!! TODO: implementation of inverse deformation mapping !!!
    res[0] = u;
    res[1] = v;
    return true;
  }

  /**
   * Returns domain of the mapping.
   *
   * @param  dom Domain array ({@code dom[0]} = Xmin, {@code dom[1]} = Xmax,
   *             {@code dom[2]} = Ymin, {@code dom[3]} = Ymax).
   * @return {@code true} if this service is implemented.
   */
  @Override
  public boolean getDomain ( double[] dom )
  {
    if ( dom == null || dom.length < 4 ) return false;

    setAspectRatio();
    dom[0] =
    dom[2] = 0.0;
    dom[1] = aspectRatio;
    dom[3] = 1.0;

    return true;
  }

  /**
   * Returns range of the mapping.
   *
   * @param  rng Range array ({@code rng[0]} = Umin, {@code rng[1]} = Umax,
   *             {@code rng[2]} = Vmin, {@code rng[3]} = Vmax).
   * @return {@code true} if this service is implemented.
   */
  @Override
  public boolean getRange ( double[] rng )
  {
    if ( rng == null || rng.length < 4 ) return false;

    setAspectRatio();
    rng[0] =
    rng[2] = 0.0;
    rng[1] = aspectRatio;
    rng[3] = 1.0;

    return true;
  }

  //==========================================================================
  //  Interface TimeDependent:

  /**
   * Retrieve the valid bounds for time variable.
   *
   * @param  pre Pre-allocated {@code double[2]} array to hold the result (can be
   *             {@code null}).
   * @return {@code result[0]} .. minimum time value, {@code result[1]} .. maximum
   *         time value.
   * @see    #getTime
   * @see    #setTime
   */
  public double[] getTimeInterval ( double[] pre )
  {
    if ( pre == null )
      pre = new double[ 2 ];

    pre[0] = 0.0;
    pre[1] = 1.0;
    // !!! TODO: get time interval from associated FeatureStore !!!
    return pre;
  }

  /**
   * Returns the actual time value for the object (used in animations).
   *
   * @return Actual time (in seconds).
   * @see    #setTime
   */
  public double getTime ()
  {
    return actualTime;
  }

  /**
   * Sets the actual time value for the object (used in animations).
   *
   * @param  time New scene time (in seconds).
   * @see    #getTime
   * @see    #getTimeInterval
   */
  public void setTime ( double time )
  {
    actualTime = time;
  }

  //==========================================================================
  //  Interface Property:

  public final static String ASPECT_RATIO = "Aspect ratio";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // ASPECT_RATIO:
    if ( key.compareTo(ASPECT_RATIO) == 0 )
    {
      aspectRatio = doubleProperty( value, aspectRatio, 1e-4, 1e4 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(ASPECT_RATIO) == 0 )
      return aspectRatio;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "FeatureDeformation";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "FunctionR2ToR2ToFeatureStoreAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Feature-based warping deformation function.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_FUNCTION;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"FunctionR2ToR2" );
      t.newOutputPlug(    PL_DATA,   IFACE+"FeatureStore" );
      t.newOptOutputPlug( PL_OUTPUT, IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( ASPECT_RATIO, TYPE_DOUBLE, "Aspect ratio", true );
        t.propDefault( 1.0 );
        t.propBounds(  1.e-4, 1.e4 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
