package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Fractal specimen - simple concentrating circles.
 *
 * @version 0.26 $Rev: 453 $ $Date: 2007-05-14 21:50:00 +0200 (po, 14 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/FractalCircles.java">FractalCircles.java</a>
 */
public class FractalCircles extends DefaultImageSynthesizer
{
  //==========================================================================
  //  Data:

  /** Radius exponent. */
  protected double exponent = 2.0;

  /** Fractal frequency multiplier (uniform in both coordinate components). */
  protected double freq = 0.1;

  /** Actual index of external colormap (see {@link ColormapStore}). */
  protected int cmapNo = 0;

  // Inherited variables imageWidth and imageHeight are set to image extent in pixels.
  // See DefaultImageSynthesizer class for details.

  //==========================================================================
  //  Interface ImageSynthesizer:

  /**
   * Render the given pixel of the image.
   *
   * @param  x X coordinate of the pixel.
   * @param  y Y coordinate of the pixel.
   * @param  pre Pre-allocated {@code double[3]} array to hold results (can be
   *             {@code null}).
   * @return Color of the pixel (in <b>RGB</b> color space).
   * @see    #renderRectangle
   * @see    #renderAll
   */
  public double[] renderPixel ( int x, int y, double[] pre )
  {
    if ( pre == null ||
         pre.length < 3 )
      pre = new double[ 3 ];

    double r = freq * freq * (x * x + y * y);
    long ord = Math.round( Math.pow( r, exponent * 0.5 ) );

    final ColormapStore cmap = (ColormapStore)getInterface( PL_COLORMAP, IFACE+"ColormapStore" );

    if ( cmap == null )
    {
      Formula.colorRamp( (ord & 255) / 255.0, pre );
    }
    else
    {
      double[][] cm = cmap.getColormapDouble( cmapNo );
      System.arraycopy( cm[ (int)(ord % cm.length) ], 0, pre, 0, 3 );
    }

    return pre;
  }

  /**
   * Render the given rectangle.
   * Output is written to some connected raster device (not defined in this interface).
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @see    #renderPixel
   * @see    #renderAll
   */
  public void renderRectangle ( int x1, int y1, int x2, int y2 )
  {
      // raster output:
    final RasterGraphics out = (RasterGraphics)getInterface( PL_OUTPUT,   IFACE+"RasterGraphics" );
    if ( out == null ) return;

      // external colormap:
    final ColormapStore cmap = (ColormapStore) getInterface( PL_COLORMAP, IFACE+"ColormapStore" );
    double[][] cm = null;
    int cmSize    = 0;
    if ( cmap != null )
    {
      cm     = cmap.getColormapDouble( cmapNo );
      cmSize = cm.length;
    }

      // temps & accelerators:
    final double freq2 = freq * freq;
    final double exp   = 0.5 * exponent;
    double[] color = new double[ 3 ];

      // rendering loop:
    int x, y;
    for ( y = y1; y < y2; y++ )
    {
      int yy = y * y;
      for ( x = x1; x < x2; x++ )
      {
        if ( userBreak ) return;

        double r = freq2 * (x * x + yy);
        long ord = Math.round( Math.pow( r, exp ) );

        if ( cm == null )
          Formula.colorRamp( (ord & 255) / 255.0, color );
        else
          System.arraycopy( cm[ (int)(ord % cmSize) ], 0, color, 0, 3 );

        out.putPixel( x, y, color );
      }
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String EXPONENT = "Exponent";
  public final static String FREQ     = "Freq";
  public final static String CMAP_NO  = "Colormap#";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // EXPONENT:
    if ( key.compareTo(EXPONENT) == 0 )
    {
      exponent = doubleProperty( value, exponent, 0.1, 20.0 );
      return;
    }
      // FREQ:
    if ( key.compareTo(FREQ) == 0 )
    {
      freq = doubleProperty( value, freq, 0.0, 100.0 );
      return;
    }
      // CMAP_NO:
    if ( key.compareTo(CMAP_NO) == 0 )
    {
      cmapNo = intProperty( value, cmapNo, 0, 100 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(EXPONENT) == 0 )
      return exponent;
    if ( key.compareTo(FREQ) == 0 )
      return freq;
    if ( key.compareTo(CMAP_NO) == 0 )
      return cmapNo;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "FractalCircles";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ImageSynthesizerToRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Fractal specimen - simple concentrating circles.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + ".fractal";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,    IFACE+"ImageSynthesizer" );
      t.newOutputPlug(    PL_OUTPUT,   IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_COLORMAP, IFACE+"ColormapStore" );
        // Property:
      t.propBegin( EXPONENT, TYPE_DOUBLE, "Radius exponent", true );
        t.propDefault( 2.0 );
        t.propBounds(  0.1, 20.0 );
      t.propEnd();
      t.propBegin( FREQ, TYPE_DOUBLE, "Frequency", true );
        t.propDefault( 0.1 );
        t.propBounds(  0.0, 100.0 );
      t.propEnd();
      t.propBegin( CMAP_NO, TYPE_INTEGER, "Colormap number", true );
        t.propDefault( 0 );
        t.propBounds(  0, 100 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
