package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Represents Fresnel equation for conductors.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/FresnelConductor.java">FresnelConductor.java</a>
 */
public class FresnelConductor extends Piece implements Fresnel
{
  /** Index of refraction of the conductor. */
  protected double[] eta;

  /** Absorption coefficient. */
  protected double[] k;

  /** Default constructor. */
  public FresnelConductor ()
  {
    eta = new double[] { 1.0, 1.0, 1.0 };
    k   = new double[] { 0.0, 0.0, 0.0 };
  }

  /**
   * Evaluate given Fresnel equation.
   *
   * @param  cosI Dot product of incoming direction and normal of the surface.
   * @param  refl Pre-allocated field for reflectance (non-mandatory).
   * @return Reflectance at surface for a given direction.
   */
  public double[] evaluate ( double cosI, double[] refl )
  {
    if ( refl == null )
      refl = new double[ 3 ];

    Math.abs( cosI );
    Arrays.fill( refl, 0.0 );

    double tmpN     = 0.0;
    double tmpD     = 0.0;
    double reflPar2 = 0.0;  // square reflectance for parallel polarized light
    double reflPer2 = 0.0;  // square reflectance for perpendicular polarized light
    for ( int i = 0; i < refl.length; i++ )
    {
      tmpD     = eta[i]*eta[i] + k[i]*k[i];
      tmpN     = tmpD * cosI*cosI;
      reflPar2 = (tmpN - (2.f * eta[i] * cosI) + 1) /
      	         (tmpN + (2.f * eta[i] * cosI) + 1);
      reflPer2 = (tmpD - (2.f * eta[i] * cosI) + cosI*cosI) /
 	         (tmpD + (2.f * eta[i] * cosI) + cosI*cosI);
      refl[i]  = (reflPar2 + reflPer2) / 2.f;
    }

    return refl;
  }

  //==========================================================================
  //  Interface Property:

  public final static String ETA = "Eta";
  public final static String K   = "K";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

    if ( key.compareTo( ETA ) == 0 &&
         value instanceof double[] )
    {
      eta = (double[])value;
      return;
    }
    if ( key.compareTo( K ) == 0 &&
         value instanceof double[] )
    {
      k = (double[])value;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo( ETA ) == 0 ) return eta;
    if ( key.compareTo( K ) == 0 ) return k;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Fresnel equation for conductors";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "BSSRDF";

  /** Object description. */
  private final static String DESCRIPTION = "Fresnel equation for conductors.";

  /** Object category. */
  protected final static String CATEGORY = Piece.C_3D + Piece.C_DATA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( Piece.PL_INPUT, Piece.IFACE+"BSSRDF" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}