package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Represents Fresnel equation for dielectric.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/iface/FresnelDielectric.java">FresnelDielectric.java</a>
 */
public class FresnelDielectric extends Piece implements Fresnel
{
  /** Index of refraction for incoming media. */
  double[] etaI;

  /** Index of refraction for transmitted media. */
  double[] etaT;

  /** Default constructor. */
  public FresnelDielectric ()
  {
    etaI = new double[] { 1.0, 1.0, 1.0 };
    etaT = new double[] { 1.0, 1.0, 1.0 };
  }

  /**
   * Evaluate given Fresnel equation.
   *
   * @param  cosI Dot product of incoming direction and normal of the surface.
   * @param  refl Pre-allocated field for reflectance (non-mandatory).
   * @return Reflectance at surface for a given direction.
   */
  public double[] evaluate ( double cosI, double[] refl )
  {
    if ( refl == null )
      refl = new double[ 3 ];

    Arrays.fill( refl, 0.0 );
    cosI = Formula.clamp( cosI, -1.0, 1.0 );

      // determine etas:
    double[] ei = new double[ etaI.length ];
    System.arraycopy( etaI, 0, ei, 0, etaI.length );
    double[] et = new double[ etaT.length ];
    System.arraycopy( etaT, 0, et, 0, etaT.length );
    if ( cosI < 0.0 )
      swap( ei, et );

      // compute transmitance of the ray:
    double sinI = Math.sqrt( Math.max( 0.0, 1.0 - cosI*cosI ) );
    for ( int i = 0; i < refl.length; i++ )
    {
        // handle total internal reflection:
      double sinT = (ei[i] / et[i]) * sinI;
      if ( sinT > 1.0 )
      {
        refl[i] = 1.0;
        continue;
      }

      double cosT    = Math.sqrt( Math.max( 0.0, 1.0 - sinT*sinT ) );
      double cosIAbs = Math.abs( cosI );
      double refPar  = 0.0;      // reflectance for parallel polarized light
      double refPer  = 0.0;      // reflectance for perpendicular polarized light
      refPar = ( (et[i] * cosIAbs) - (ei[i] * cosT) ) /
               ( (et[i] * cosIAbs) + (ei[i] * cosT) );
      refPer = ( (ei[i] * cosIAbs) - (et[i] * cosT) ) /
               ( (ei[i] * cosIAbs) + (et[i] * cosT) );
      refl[i] = refPar*refPar + refPer*refPer / 2.0;
    }

    return refl;
  }

  protected void swap ( double[] ei, double[] et )
  {
    double[] i = new double[ ei.length ];
    System.arraycopy( ei, 0, i,  0,  ei.length );
    System.arraycopy( et, 0, ei, 0,  ei.length );
    System.arraycopy( i,  0, et,  0, ei.length );
  }

  //==========================================================================
  //  Interface Property:

  public final static String ETA_I = "EtaI";
  public final static String ETA_T = "EtaT";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

    if ( key.compareTo( ETA_I ) == 0 &&
         value instanceof double[] )
    {
      etaI = (double[])value;
      return;
    }
    if ( key.compareTo( ETA_T ) == 0 &&
         value instanceof double[] )
    {
      etaT = (double[])value;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo( ETA_I ) == 0 ) return etaI;
    if ( key.compareTo( ETA_T ) == 0 ) return etaT;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Fresnel equation for dielectric";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "BSSRDF";

  /** Object description. */
  private final static String DESCRIPTION = "Fresnel equation for dielectric.";

  /** Object category. */
  protected final static String CATEGORY = Piece.C_3D + Piece.C_DATA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( Piece.PL_INPUT, Piece.IFACE+"BSSRDF" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}