package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * 3D rendering example - graph of R^2 -&gt; R mapping.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/GraphR2ToR.java">GraphR2ToR.java</a>
 */
public class GraphR2ToR extends DefaultRender3D
{
  //==========================================================================
  //  Data:

  /** Canvas background. */
  protected final static int[] BACKGROUND = {   0,  20,  60, 255 };

  /** Foreground color. */
  protected final static int[] FOREGROUND = { 255, 240, 200, 255 };

  /** Number of cells in X direction. */
  protected int cellsX  =    20;

  /** Number of cells in Y direction. */
  protected int cellsY  =    20;

  protected double minX = -5.0;
  protected double maxX =  5.0;
  protected double minY = -5.0;
  protected double maxY =  5.0;

  //==========================================================================
  //  Interface Render3D:

  /**
   * 3D rendering using the given projection matrix.
   *
   * @param mv Local model-view matrix.
   * @param proj Local projection matrix.
   * @param style Local rendering style.
   * @see #render()
   */
  @Override
  public void render ( TrMatrix mv, TrMatrix proj, int style )
  {
    RealFunction   func   = (RealFunction)  getInterface( PL_FUNCTION, IFACE+"RealFunction" );
    RasterGraphics output = (RasterGraphics)getInterface( PL_OUTPUT,   IFACE+"RasterGraphics" );
    int width  = output.getWidth();
    int height = output.getHeight();
    output.setRectangle( 0, 0, width, height, BACKGROUND );
    output.setColor( FOREGROUND );

    double x, y;
    double stepX = (maxX - minX) / cellsX;
    double stepY = (maxY - minY) / cellsY;
    int i, j;
    double[] last0 = new double[3];
    double[] curr0 = new double[3];
    double[] last1 = new double[3];
    double[] curr1 = new double[3];
    double[] tmp;

      // minY border:
    curr0[0] = minX;
    curr0[1] = minY;
    curr0[2] = func.f( minX, minY );
    proj.transformPoint3D( curr0, curr0 );
    x = minX + stepX;
    for ( j = 0; j++ < cellsX; x += stepX )
    {
      tmp = last0; last0 = curr0; curr0 = tmp;
      curr0[0] = x;
      curr0[1] = minY;
      curr0[2] = func.f( x, minY );
      proj.transformPoint3D( curr0, curr0 );
      DefaultRasterGraphics.drawLine( output, last0[0], last0[1], curr0[0], curr0[1] );
    }
      // row loop:
    y = minY;
    for ( i = 0; i++ < cellsY && !userBreak; y += stepY ) // one row
    {
      curr0[0] = minX;
      curr0[1] = y;
      curr0[2] = func.f( minX, y );
      proj.transformPoint3D( curr0, curr0 );
      curr1[0] = minX;
      curr1[1] = y + stepY;
      curr1[2] = func.f( minX, y + stepY );
      proj.transformPoint3D( curr1, curr1 );
      DefaultRasterGraphics.drawLine( output, curr0[0], curr0[1], curr1[0], curr1[1] );
      x = minX + stepX;
      for ( j = 0; j++ < cellsX; x += stepX )
      {
        tmp = last0; last0 = curr0; curr0 = tmp;
        tmp = last1; last1 = curr1; curr1 = tmp;
        curr0[0] = x;
        curr0[1] = y;
        curr0[2] = func.f( x, y );
        proj.transformPoint3D( curr0, curr0 );
        curr1[0] = x;
        curr1[1] = y + stepY;
        curr1[2] = func.f( x, y + stepY );
        proj.transformPoint3D( curr1, curr1 );
        DefaultRasterGraphics.drawLine( output, last1[0], last1[1], curr1[0], curr1[1] );
        DefaultRasterGraphics.drawLine( output, curr0[0], curr0[1], curr1[0], curr1[1] );
      }
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String CELLS_X = "Cells X";
  public final static String CELLS_Y = "Cells Y";
  public final static String MIN_X   = "Min X";
  public final static String MAX_X   = "Max X";
  public final static String MIN_Y   = "Min Y";
  public final static String MAX_Y   = "Max Y";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // RENDER_STYLE:
    if ( key.compareTo(RENDER_STYLE) == 0 )
    {
      renderStyle = intProperty( value, renderStyle, RENDER_STYLE_LINE, RENDER_STYLE_PHONG );
      return;
    }
      // CELLS_X:
    if ( key.compareTo(CELLS_X) == 0 )
    {
      cellsX = intProperty( value, cellsX, 1, 1000 );
      return;
    }
      // CELLS_Y:
    if ( key.compareTo(CELLS_Y) == 0 )
    {
      cellsY = intProperty( value, cellsY, 1, 1000 );
      return;
    }
      // MIN_X:
    if ( key.compareTo(MIN_X) == 0 )
    {
      minX = doubleProperty( value, minX );
      return;
    }
      // MAX_X:
    if ( key.compareTo(MAX_X) == 0 )
    {
      maxX = doubleProperty( value, maxX );
      return;
    }
      // MIN_Y:
    if ( key.compareTo(MIN_Y) == 0 )
    {
      minY = doubleProperty( value, minY );
      return;
    }
      // MAX_Y:
    if ( key.compareTo(MAX_Y) == 0 )
    {
      maxY = doubleProperty( value, maxY );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(RENDER_STYLE) == 0 )
      return renderStyle;
    if ( key.compareTo(CELLS_X) == 0 )
      return cellsX;
    if ( key.compareTo(CELLS_Y) == 0 )
      return cellsY;
    if ( key.compareTo(MIN_X) == 0 )
      return minX;
    if ( key.compareTo(MAX_X) == 0 )
      return maxX;
    if ( key.compareTo(MIN_Y) == 0 )
      return minY;
    if ( key.compareTo(MAX_Y) == 0 )
      return maxY;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Graph R^2->R";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "Render3DToFunctionR2ToRAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Sample R^2->R function renderer.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + ".graph";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,    IFACE+"Render3D" );
      t.newOutputPlug( PL_FUNCTION, IFACE+"RealFunction" );
      t.newOutputPlug( PL_OUTPUT,   IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( RENDER_STYLE, TYPE_INTEGER, "Render style", true );
        t.propDefault( RENDER_STYLE_LINE );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Line drawing",    RENDER_STYLE_LINE,    "" );
        t.propEnum( "Flat shading",    RENDER_STYLE_FLAT,    "" );
        t.propEnum( "Gouraud shading", RENDER_STYLE_GOURAUD, "" );
        t.propEnum( "Phong shading",   RENDER_STYLE_PHONG,   "" );
      t.propEnd();
      t.propBegin( CELLS_X, TYPE_INTEGER, "Cells X", true );
        t.propDefault( 20 );
      t.propEnd();
      t.propBegin( CELLS_Y, TYPE_INTEGER, "Cells Y", true );
        t.propDefault( 20 );
      t.propEnd();
      t.propBegin( MIN_X, TYPE_DOUBLE, "Min X", true );
        t.propDefault( -5.0 );
      t.propEnd();
      t.propBegin( MAX_X, TYPE_DOUBLE, "Max X", true );
        t.propDefault( 5.0 );
      t.propEnd();
      t.propBegin( MIN_Y, TYPE_DOUBLE, "Min Y", true );
        t.propDefault( -5.0 );
      t.propEnd();
      t.propBegin( MAX_Y, TYPE_DOUBLE, "Max Y", true );
        t.propDefault( 5.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
