package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Image filter example - image convert.
 *
 * @version 0.25 $Rev: 385 $ $Date: 2007-01-07 01:47:21 +0100 (ne, 07 I 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/ImageConvert.java">ImageConvert.java</a>
 */
public class ImageConvert extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Requested color mode. See {@code RasterGraphics.MODE_*}. */
  protected int colorMode;

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transformation.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    RasterGraphics input  = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    RasterGraphics output = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );
    int i, j;

      // output image:
    int width      = input.getWidth();
    int height     = input.getHeight();
    int sourceMode = input.getMode();
    int targetMode = colorMode;
    output.init( width, height, targetMode, 0 );

    switch ( targetMode )
    {
      case RasterGraphics.MODE_COLORMAP:
      {
        int[][] oCmap;
        if ( sourceMode == RasterGraphics.MODE_COLORMAP )
        {
            // input colormap:
          int palSize   = input.getMaxColormap() + 1;
          int palMode   = input.getColormapType();
          int cmapBands = (palMode == RasterGraphics.MODE_RGB) ? 3 : 4;
          int[][] iCmap = new int[ palSize ][ cmapBands ];
          input.getColormap( 0, palSize, iCmap );

            // count number of used colors:
          boolean[] used = new boolean[ palSize ];
          for ( j = 0; j < height; j++ )
            for ( i = 0; i < width; i++ )
              used[ input.getPixel( i, j ) ] = true;
          int noUsed = 0;
          for ( i = 0; i < palSize; i++ )
            if ( used[i] ) noUsed++;

            // colormap conversion:
          oCmap      = new int[ noUsed ][];
          int[] perm = new int[ palSize ];
          j = 0;                          // first free item in oCmap
          for ( i = 0; i < palSize; i++ )
            if ( used[i] )
            {
              perm[i] = j;
              oCmap[j++] = iCmap[i];
            }

            // target image initialization:
          output.setColormapType( palMode, noUsed - 1 );
          output.setColormap( oCmap );

            // bitmap conversion:
          for ( j = 0; j < height; j++ )
            for ( i = 0; i < width; i++ )
              output.putPixel( i, j, perm[ input.getPixel( i, j ) ] );

          return true;
        }

        if ( sourceMode == RasterGraphics.MODE_GRAY )
        {
          oCmap = new int[ 256 ][ 3 ];
          for ( i = 0; i < 256; i++ )
            oCmap[i][0] = oCmap[i][1] = oCmap[i][2] = i;

            // target image initialization:
          output.setColormapType( RasterGraphics.MODE_RGB, 255 );
          output.setColormap( oCmap );

            // bitmap conversion:
          for ( j = 0; j < height; j++ )
            for ( i = 0; i < width; i++ )
              output.putPixel( i, j, input.getGray( i, j ) );

          return true;
        }

          // otherwise: use 3-3-2 palette w/o dithering
          // create 3-3-2 palette:
        double R, G, B;
        double dRG = 255.0 / 7;
        double dB  = 255.0 / 3;
        int r, g, b;
        oCmap = new int[ 256 ][ 3 ];

        for ( r = i = 0, R = 0.0; r < 8; r++, R += dRG )
          for ( g = 0, G = 0.0; g < 8; g++, G += dRG )
            for ( b = 0, B = 0.0; b < 4; b++, B += dB )
            {
              oCmap[i][0] = (int)R;
              oCmap[i][1] = (int)G;
              oCmap[i][2] = (int)B;
              i++;
            }
        output.setColormapType( RasterGraphics.MODE_RGB, 255 );
        output.setColormap( oCmap );

          // bitmap conversion:
        int[] pixel = new int[ 3 ];
        for ( j = 0; j < height; j++ )
          for ( i = 0; i < width; i++ )
          {
            input.getRGB( i, j, pixel );
            r = ((pixel[0] + 16) >> 5) & 7;
            g = ((pixel[1] + 16) >> 5) & 7;
            b = ((pixel[2] + 32) >> 6) & 3;
            output.putPixel( i, j, (r << 5) + (g << 2) + b );
          }

        return true;
      }

      case RasterGraphics.MODE_GRAY:
        for ( j = 0; j < height; j++ )
          for ( i = 0; i < width; i++ )
            output.putPixel( i, j, input.getGray( i, j ) );
        return true;

      default:                            // MODE_RGB[A]
      {
        int[] pixel = new int[ targetMode == RasterGraphics.MODE_RGB ? 3 : 4 ];
        for ( j = 0; j < height; j++ )
          for ( i = 0; i < width; i++ )
          {
            input.getRGB( i, j, pixel );
            output.putPixel( i, j, pixel );
          }
      }
    }
    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String COLORMODE = "Color Mode";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // COLORMODE:
    if ( key.compareTo(COLORMODE) == 0 )
    {
      colorMode = intProperty( value, colorMode );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(COLORMODE) == 0 )
      return colorMode;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "ImageConvert";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Image conversion, test module.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"Trigger" );
      t.newOutputPlug( "image1", IFACE+"RasterGraphics" );
      t.newOutputPlug( "image2", IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( COLORMODE, TYPE_INTEGER, "Pixel color mode", true );
        t.propDefault( RasterGraphics.MODE_RGB );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "RGB",      RasterGraphics.MODE_RGB,      "RGB color mode" );
        t.propEnum( "RGBA",     RasterGraphics.MODE_RGBA,     "RGBA color mode (transparency)" );
        t.propEnum( "Gray",     RasterGraphics.MODE_GRAY,     "Gray mode" );
        t.propEnum( "Colormap", RasterGraphics.MODE_COLORMAP, "Colormapped mode" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
