package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Computes 3D-to-2D transformation matrix from interactive mouse/keyboard events.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/InteractiveProjection.java">InteractiveProjection.java</a>
 */
public class InteractiveProjection extends DefaultInputListener
{
  //==========================================================================
  //  Data:

  /** Zoom coefficient for one mouse-wheel "tick" minus 1.0. */
  public static final double WHEEL_ZOOM = 0.03;

  /** Standard {@link #eye}-{@link #center} distance. */
  public static final double EYE_DISTANCE = 8.0;

  /** Actual window width in pixels. */
  protected int width  = 0;

  /** Actual window height in pixels. */
  protected int height = 0;

  /** Actual model-view transform. */
  protected TrMatrix modelView = new TrMatrix( 4, 3 );

  /** Actual projection matrix. */
  protected TrMatrix proj = new TrMatrix( 4, 3 );

  /** View reference point azimuth ({@code 0.0} .. yz plane). */
  protected double azimuth = 0.0;

  /** View reference point elevation angle (positive .. positive z-coordinate). */
  protected double elevation = 0.0;

  /** Actual result zoom factor. */
  protected double zoom = 1.0;

  /** Zoom factor produces by the mouse-wheel. */
  protected double wheelZoom = 1.0;

  /** Mouse-move processing is switched on? */
  protected boolean live = false;

  /** Actual translation component. */
  protected double[] transl = { 0.0, 0.0, 0.0 };

  /** Called for the first time? */
  protected boolean first = true;

  /** Accelerator: notification object #1. */
  protected Trigger trigger1;

  /** Accelerator: notification object #2. */
  protected Trigger trigger2;

  /** Accelerator: rendering object. */
  protected Render3D render;

  /** Point I'm always looking at.. */
  protected double[] center = { 0.0, 0.0, 0.0 };

  /** Actual eye's position (aka VRP). */
  protected double[] eye = { 0.0, EYE_DISTANCE, 0.0 };

  //==========================================================================
  //  Support:

  /** Assembles the projection matrix and triggers the "redraw" action. */
  protected void updateMatrix ()
  {
      // model-view transform:
    modelView.identity( 4, 4 );
    final double dist   = EYE_DISTANCE / wheelZoom;
    final double cosEle = Math.cos( elevation );
    eye[0] = dist * Math.sin( azimuth ) * cosEle;
    eye[1] = dist * Math.cos( azimuth ) * cosEle;
    eye[2] = dist *                       Math.sin( elevation );
    modelView.lookAt( eye, center, null );

      // projection matrix: perspective
    proj.identity( 4, 4 );
    proj.perspective( 40.0, width / (double)height, 0.2, 50.0 );

      // accelerators:
    if ( first )
    {
      first    = false;
      trigger1 = (Trigger) getInterface( "trigger1", IFACE+"Trigger" );
      trigger2 = (Trigger) getInterface( "trigger2", IFACE+"Trigger" );
      render   = (Render3D)getInterface( PL_OUTPUT,  IFACE+"Render3D" );
    }
      // notification:
    if ( render  != null )
    {
      render.setModelView( modelView );
      render.setProjection( proj );
      render.setViewport( 0, 0, width, height );
    }
    if ( trigger1 != null ) trigger1.fire( 0 );
    if ( trigger2 != null ) trigger2.fire( 0 );
  }

  //==========================================================================
  //  Interface InputListener:

  /**
   * Change the window shape.
   *
   * @param width Horizontal window size in pixels.
   * @param height Vertical window size in pixels.
   */
  @Override
  public void setWindowSize ( Object sender, int width, int height )
  {
    this.width  = width;
    this.height = height;
    updateMatrix();
  }

  /**
   * Mouse button was pressed/released.
   *
   * @param when System time in milliseconds.
   * @param x Horizontal screen position in pixels.
   * @param y Vertical screen position in pixels.
   * @param button Button number ({@link #BUTTON1}, etc.).
   * @param press Was it press or release?
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public void mouseButton ( Object sender, long when, int x, int y, int button, boolean press, int flags )
  {
    if ( button != BUTTON1 ||
         press ) return;

    live = !live;
  }

  /**
   * Mouse pointer was moved.
   *
   * @param sender Originator of the event
   * @param when System time in milliseconds.
   * @param x Horizontal screen position in pixels.
   * @param y Vertical screen position in pixels.
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public void mouseMove ( Object sender, long when, int x, int y, int flags )
  {
    if ( !live ) return;

    azimuth   = -((double)x / width  - 0.5) * 2.0 * Math.PI;
    elevation =  ((double)y / height - 0.5) *       Math.PI;
    updateMatrix();
  }

  /**
   * Mouse wheel was rotated.
   *
   * @param sender Originator of the event
   * @param when System time in milliseconds.
   * @param x Horizontal screen position in pixels.
   * @param y Vertical screen position in pixels.
   * @param wheel Wheel number ({@link #BUTTON1}, etc.).
   * @param amount Rotation angle (positive number means down - or towards the user).
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public void mouseWheel ( Object sender, long when, int x, int y, int wheel, int amount, int flags )
  {
    if ( amount == 0 ) return;

    wheelZoom *= 1.0 + WHEEL_ZOOM * amount;
    updateMatrix();
  }

  //==========================================================================
  //  Interface Property:

  public final static String MATRIX = "Matrix";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // MATRIX:
    if ( key.compareTo(MATRIX) == 0 )
    {
      if ( value != null && value instanceof TrMatrix )
        proj = (TrMatrix)value;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(MATRIX) == 0 )
      return proj;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "InteractiveProjection";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "InputListenerToTrigger";

  /** Object description. */
  private final static String DESCRIPTION = "Computes projection matrix from interactive mouse events.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_PROJECTION;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,   IFACE+"InputListener" );
      t.newOptOutputPlug( "trigger1", IFACE+"Trigger" );
      t.newOptOutputPlug( "trigger2", IFACE+"Trigger" );
      t.newOptOutputPlug( PL_OUTPUT,  IFACE+"Render3D" );
        // Property:
      t.propBegin( MATRIX, TYPE_OBJECT, "Projection matrix", false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
