package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Lambertian shading model.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 858 $ $Date: 2009-02-17 23:48:39 +0100 (út, 17 II 2009) $ $Author: Tomas $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/LambertianModel.java">LambertianModel.java</a>
 */
public class LambertianModel extends Piece implements BSSRDF
{
  //==========================================================================
  //  Data:

  /** Actual color of the solid surface. */
  public double[] color;

  /** Default constructor. */
  public LambertianModel ()
  {
    color = new double[] { 1.0, 1.0, 1.0 };
  }

  //==========================================================================
  //  Interface BSSRDF:

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as monochrome).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @return BSSRDF coefficient <b>f(in-&gt;out)</b>.
   * @see    #monoBSSRDF(MicroFacet,double[],double[],int)
   * @see    #colorBSSRDF(double[],double[],double[],int,double[])
   */
  public double monoBSSRDF ( boolean worldCoords,
                             double[] in,
                             double[] out,
                             double[] normal,
                             double[] tangent,
                             int mask )
  {
    if ( normal == null )
      return 0.0;

    if ( mask == BSSRDF.COMP_DIFFUSE ||
         mask == BSSRDF.COMP_ALL )
    {
      return color[0] / Math.PI;
    }
    else
      return 0.0;
  }

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as monochrome).
   * Computations are done relatively to the given {@link MicroFacet}
   * (point on surface of some solid).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  mask Which light components are to be computed?
   * @return BSSRDF coefficient <b>f(in-&gt;mf-&gt;out)</b>.
   * @see    #monoBSSRDF(double[],double[],double[],int)
   * @see    #colorBSSRDF(MicroFacet,double[],double[],int,double[])
   */
  public double monoBSSRDF ( boolean worldCoords,
                             MicroFacet mf,
                             double[] in,
                             double[] out,
                             int mask )
  {
    if ( mf == null ||
         mf.assertAttributes( MicroFacet.VALID_NORMAL ) != 0 )
      return 0.0;

    return monoBSSRDF( worldCoords, in, out, mf.normal, mf.tu, mask );
  }

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as band-color).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BSSRDF spectral coefficients.
   * @see    #monoBSSRDF(double[],double[],double[],int)
   * @see    #colorBSSRDF(MicroFacet,double[],double[],int,double[])
   */
  public double[] colorBSSRDF ( boolean worldCoords,
                                double[] in,
                                double[] out,
                                double[] normal,
                                double[] tangent,
                                int mask,
                                double[] result )
  {
    if ( normal == null )
      return null;

    if ( mask == BSSRDF.COMP_DIFFUSE ||
         mask == BSSRDF.COMP_ALL )
    {
      if ( result == null ||
           result.length < color.length )
        result = new double[ color.length ];

      double invPI = (1.0 / Math.PI);
      for ( int i = 0; i < color.length; i++ )
        result[i] = color[i] * invPI;
      return result;
    }
    else
      return null;
  }

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as band-color).
   * Computations are done relatively to the given {@link MicroFacet}
   * (point on surface of some solid).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;mf-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BSSRDF spectral coefficients.
   * @see    #monoBSSRDF(MicroFacet,double[],double[],int)
   * @see    #colorBSSRDF(double[],double[],double[],int,double[])
   */
  public double[] colorBSSRDF ( boolean worldCoords,
                                MicroFacet mf,
                                double[] in,
                                double[] out,
                                int mask,
                                double[] result )
  {
    if ( mf == null ||
         mf.assertAttributes( MicroFacet.VALID_NORMAL   |
                              MicroFacet.VALID_COLOR ) != 0 )
      return null;

    return colorBSSRDF( worldCoords, in, out, mf.normal, mf.tu, mask, result );
  }

  /**
   * Retrieves the original color (defining surface color).
   *
   * @param  color Pre-allocated {@code double[C]} array to hold the color
   *               ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *               color system/ or greater /for more complex spectral color sampling/). Can
   *               be {@code null}.
   * @return Array containing copy of original color (can be altered).
   */
  public double[] getOrigColor ( double[] color )
  {
    if ( this.color == null ) return null;

    int bands = this.color.length;
    if ( color == null ||
         color.length < bands )
      color = Formula.cloneArray( this.color );
    else
      System.arraycopy( this.color, 0, color, 0, bands );

    return color;
  }

  /**
   * Sets the actual color (for the given BSSRDF data).
   * This color will be used in subsequent BSSRDF calculations.
   * The color array won't be altered.
   *
   * @param  color Color definition - {@code double[C]} array
   *               ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *               color system/ or greater /for more complex spectral color sampling/).
   */
  public void setColor ( double[] color )
  {
    if ( color != null )
      this.color = color;
  }

  /**
   * Scatter ray according to the light model.
   *
   * @param  mf   The {@link MicroFacet} (point on surface of some solid).
   * @param  in   Input direction (should look from the hit point)..
   * @param  u    Random uniform number.
   * @param  v    Random uniform number.
   * @param  mask Which light components are to be computed?
   * @param  out  Output direction in world coordinates.
   * @param  brdf Color definition - {@code double[C]} array
   *              ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *              color system/ or greater /for more complex spectral color sampling/).
   * @return Probability density function.
   */
  public double scatter ( MicroFacet mf,
                          double[] in,
                          double u,
                          double v,
                          int mask,
                          double[] out,
                          double[] brdf  )
  {
    if ( mf   == null ||
         in   == null ||
         brdf == null ||
         out  == null )
      return 0.0;

    double[] inLoc = new double[ 3 ];
    worldToLocal( mf.normal, mf.tu, in, inLoc );

    MonteCarlo.cosineSampleHemisphere( u, v, out );
    colorBSSRDF( false, mf, inLoc, out, mask, brdf );
    double pdf = pdf( false, mf, inLoc, out );

    localToWorld( mf.normal, mf.tu, out, out );
    return pdf;
  }

  /**
   * Returns the probability density function.
   *
   * @param worldCoords Should be set to {@code true} if vectors in and out
   *                    are in world coordinates and to {@code false}
   *                    if vectors in and out are in local coordinates.
   * @param in  Input direction.
   * @param out Output direction.
   */
  public double pdf ( boolean worldCoords, MicroFacet mf, double[] in, double[] out )
  {
    if ( in  == null ||
         out == null )
      return 0.0;

    if ( worldCoords )
    {
      if ( Geometry.dotProduct3D( in,  mf.normal ) *
           Geometry.dotProduct3D( out, mf.normal ) > 0.0 )
        return Math.abs( Geometry.dotProduct3D( out, mf.normal ) ) / Math.PI;
      else
        return 0.0;
    }
    else
    {
      if ( in[2] * out[2] > 0.0 )
        return Math.abs( out[2] ) / Math.PI;
      else
        return 0.0;
    }
  }

  /** Returns the type of BRDF. */
  public int type ()
  {
    return BSSRDF.COMP_DIFFUSE;
  }

  //==========================================================================
  //  Support methods:

  /** Transform local coordinates to world coordinates. */
  public double[] localToWorld ( double[] normal,
                                 double[] tangent,
                                 double[] src,
                                 double[] dest )
  {
    if ( normal  == null ||
         tangent == null ||
         src     == null )
      return null;
    if ( dest == null )
      dest = new double[ 3 ];

    Geometry.normalize3D( tangent );
    double[] binormal = new double[ 3 ];
    Geometry.vectorProduct( tangent, normal, binormal );

    double x = (src[0]*tangent[0]) + (src[1]*binormal[0]) + (src[2]*normal[0]);
    double y = (src[0]*tangent[1]) + (src[1]*binormal[1]) + (src[2]*normal[1]);
    double z = (src[0]*tangent[2]) + (src[1]*binormal[2]) + (src[2]*normal[2]);

    dest[0] = x;
    dest[1] = y;
    dest[2] = z;
    return dest;
  }

  /** Transform world coordinates to local coordinates. */
  public double[] worldToLocal ( double[] normal,
                                 double[] tangent,
                                 double[] src,
                                 double[] dest )
  {
    if ( normal  == null ||
         tangent == null ||
         src     == null )
      return null;
    if ( dest == null )
      dest = new double[ 3 ];

    Geometry.normalize3D( tangent );
    double[] binormal = new double[ 3 ];
    Geometry.vectorProduct( tangent, normal, binormal );

    double x = Geometry.dotProduct3D( src, tangent  );
    double y = Geometry.dotProduct3D( src, binormal );
    double z = Geometry.dotProduct3D( src, normal   );

    dest[0] = x;
    dest[1] = y;
    dest[2] = z;
    return dest;
  }

  //==========================================================================
  //  Interface Property:

  public final static String REFLECTIONS = "Reflections";

/**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

    if ( key.compareTo( BSSRDF.MATERIAL_COLOR ) == 0 &&
         value instanceof double[] )
    {
      color = (double[])value;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo( BSSRDF.MATERIAL_COLOR ) == 0 ) return color;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Lambertian shading model";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "BSSRDF";

  /** Object description. */
  private final static String DESCRIPTION = "Lambertian shading model.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_LIGHT;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"BSSRDF" );
    }
    return 1;
  }

    /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}