package cz.cuni.jagrlib.testing;

import java.awt.Color;
import org.nfunk.jep.JEP;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;
import cz.cuni.jagrlib.piece.*;

/**
 * Raster drawing specimen - Lissajous curve.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/Lissajous.java">Lissajous.java</a>
 */
public class Lissajous extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Y/x ratio. */
  protected String ratio = "1.0";

  /** Phase shift (will be multiplied by PI). */
  protected String phase = "0.5";

  /** Time step (in radians). */
  protected double dt    = 0.01;

  /** Upper time bound (will be multiplied by PI). */
  protected double maxT  = 12.0;

  /** Drawing color. */
  protected Color color  = new Color( 200, 240, 0 );

  /** Draw a curve using line segments? */
  protected boolean polyline = false;

  /** Line width in pixels. */
  protected double lineWidth = 1.0;

  /** Accelerator: output image. */
  protected RasterGraphics output;

  /** External module: line drawing. */
  protected LineAntialias line;

  /** External module: alpha-mask. */
  protected AlphaMatrix alpha;

  /** External module: brush for alpha-filling. */
  protected SolidColorBrush brush;

  /** Image width in pixels (taken from {@link RasterImage}). */
  protected int width;

  /** image height in pixels (taken from {@link RasterImage}). */
  protected int height;

  //==========================================================================
  //  Support:

  /**
   * Assert support modules and accelerators.
   *
   * @return {@code true} if succeeded.
   */
  protected boolean assertModules ()
  {
      // raster output:
    output = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    if ( output == null ) return false;

    width  = output.getWidth();
    height = output.getHeight();
    if ( !polyline ) return true;         // point drawing doesn't need support modules..

      // assert correct pixel mode:
    if ( output.getMode() != RasterGraphics.MODE_RGBA )
      output.init( width, height, RasterGraphics.MODE_RGBA, 0 );

      // drawing support:
    alpha = new AlphaMatrix();
    brush = new SolidColorBrush();
    line  = new LineAntialias();
    try
    {
      line.connect(  Piece.PL_OUTPUT,    alpha,         Piece.PL_INPUT );
      brush.connect( Piece.PL_ALPHAMASK, alpha,         Piece.PL_INPUT );
      brush.connect( Piece.PL_OUTPUT,    (Piece)output, Piece.PL_INPUT );
    }
    catch ( BadInterfaceException e )
    {
      LogFile.exception( "Error connecting modules together: line, brush, alpha-mask, raster image!", e );
      return false;
    }

      // module initialization:
    alpha.init( width, height );

    return true;
  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transformation.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    if ( !assertModules() ) return false; // failure

      // output image init:
    output.setAlpha( 1.0 );
    output.setRectangle( 0, 0, width, height, new int[] { 0, 0, 40, 255 } );

      // y/x ratio defined by an arithmetic expression:
    final JEP vRatio = Formula.initExpression( ratio );
    final double k   = vRatio.getValue();

      // Y phase shift defined by an arithmetic expression:
    final JEP vPhase = Formula.initExpression( phase );
    final double ph  = vPhase.getValue() * Math.PI;

      // loop accelerators:
    final double mt  = maxT * Math.PI + dt;
    final int[] col  = { color.getRed(), color.getGreen(), color.getBlue(), 255 };
    final double cx  = 0.5 * width;       // origin
    final double cy  = 0.5 * height;
    final double ax  = 0.9 * cx;          // amplitudes
    final double ay  = 0.9 * cy;

      // main curve rendering loop:
    double t;
    if ( polyline )                       // poly-line using anti-aliased line drawing module
    {
      brush.setColor( col );
      line.setWidth( lineWidth );

        // curve starting point:
      double x = cx;                      // sin(0) = 0
      double y = cy + ay * Math.sin( ph );

      for ( t = dt; t < mt && !userBreak; t += dt )
      {
        double x0 = x;
        double y0 = y;
        x         = cx + ax * Math.sin( t );
        y         = cy + ay * Math.sin( ph + k * t );

          // draw anti-aliased line:
        line.drawLine( x0, y0, x, y );
        brush.fill();
      }
    }

    else                                  // simple point-drawing
      for ( t = 0.0; t < mt && !userBreak; t += dt )
      {
        int x = Formula.round( cx + ax * Math.sin( t ) );
        int y = Formula.round( cy + ay * Math.sin( ph + k * t ) );
        output.putPixel( x, y, col );
      }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String RATIO    = "Ratio";
  public final static String PHASE    = "Phase shift";
  public final static String STEP     = "Step";
  public final static String MAX_T    = "Max T";
  public final static String COLOR    = "Color";
  public final static String POLYLINE = "Polyline";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // RATIO:
    if ( key.compareTo(RATIO) == 0 )
    {
      ratio = stringProperty( value, ratio );
      return;
    }
      // PHASE:
    if ( key.compareTo(PHASE) == 0 )
    {
      phase = stringProperty( value, phase );
      return;
    }
      // STEP:
    if ( key.compareTo(STEP) == 0 )
    {
      dt = doubleProperty( value, dt, 1.e-5, 1.0 );
      return;
    }
      // MAX_T:
    if ( key.compareTo(MAX_T) == 0 )
    {
      maxT = doubleProperty( value, maxT, 0.0, 1.e10 );
      return;
    }
      // COLOR:
    if ( key.compareTo(COLOR) == 0 )
    {
      color = colorProperty( value, color );
      return;
    }
      // POLYLINE:
    if ( key.compareTo(POLYLINE) == 0 )
    {
      polyline = booleanProperty( value, polyline );
      return;
    }
      // Render.LINE_WIDTH:
    if ( key.compareTo(Render.LINE_WIDTH) == 0 )
    {
      lineWidth = doubleProperty( value, lineWidth, 0.01, 10.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(RATIO) == 0 )
      return ratio;
    if ( key.compareTo(PHASE) == 0 )
      return phase;
    if ( key.compareTo(STEP) == 0 )
      return dt;
    if ( key.compareTo(MAX_T) == 0 )
      return maxT;
    if ( key.compareTo(COLOR) == 0 )
      return color;
    if ( key.compareTo(POLYLINE) == 0 )
      return polyline;
    if ( key.compareTo(Render.LINE_WIDTH) == 0 )
      return lineWidth;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Lissajous";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Image drawing specimen - Lissajous curve.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_IMAGE;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"Trigger" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( RATIO, TYPE_STRING, "Y/X ratio", true );
        t.propDefault( "1.0" );
      t.propEnd();
      t.propBegin( PHASE, TYPE_STRING, "Y phase shift (will be multiplied by PI)", true );
        t.propDefault( "0.5" );
      t.propEnd();
      t.propBegin( STEP, TYPE_DOUBLE, "Time step (in radians)", true );
        t.propDefault( 0.01 );
        t.propBounds( 1.e-5, 1.0 );
      t.propEnd();
      t.propBegin( MAX_T, TYPE_DOUBLE, "Time upper bound (will be multiplied by PI)", true );
        t.propDefault( 12.0 );
        t.propBounds( 0.0, 1.e10 );
      t.propEnd();
      t.propBegin( COLOR, TYPE_OBJECT, "Drawing color", true );
        t.propDefault( "[200;240;0]" );
        t.propManipulator( "ParamColor" );
      t.propEnd();
      t.propBegin( POLYLINE, TYPE_BOOLEAN, "Draw a curve using line segments?", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( Render.LINE_WIDTH, TYPE_DOUBLE, "Line-width in pixels", true );
        t.propDefault( 1.0 );
        t.propBounds( 0.01, 10.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
