package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import java.util.*;

/**
 * Monte Carlo methods.
 *
 * @author  Tomas Svab
 * @version 0.25 $Rev: 821 $ $Date: 2008-11-17 21:35:59 +0100 (po, 17 XI 2008) $ $Author: Tomas $
 * @since   0.24
 * @see     <a href="../../../../cz/cuni/jagrlib/testing/MonteCarlo.java">MonteCarlo.java</a>
 */
public class MonteCarlo
{
  /** Random number generator. */
  protected static final RandomJames rnd = new RandomJames();

  /** Samples hemisphere uniformly. */
  public static double[] uniformSampleHemisphere ( double ksi1,
                                                   double ksi2,
                                                   double[] result )
  {
    if ( result == null )
      result = new double[ 3 ];
    result[2] = ksi1;
    double r = Math.sqrt( Math.max( 0.0, 1.0 - ksi1*ksi1 ) );
    double phi = 2.0 * Math.PI * ksi2;
    result[0] = r * Math.cos( phi );
    result[1] = r * Math.sin( phi );
    return result;
  }

  /** Samples hemisphere according to the cosine-distribution. */
  public static double[] cosineSampleHemisphere ( double ksi1,
                                                  double ksi2,
                                                  double[] result )
  {
    if ( result == null )
      result = new double[ 3 ];
    result = concentricSampleDisk( ksi1, ksi2, result );
    result[2] = Math.sqrt( Math.max( 0.f, 1.f - result[0]*result[0] - result[1]*result[1] ) );
    return result;
  }

  /** Samples concentric disk. */
  public static double[] concentricSampleDisk( double ksi1,
                                               double ksi2,
                                               double[] result  )
  {
    if ( result == null )
      result = new double[ 3 ];

    double r, theta;

      // map uniform random numbers to [-1,1]^2:
    double sx = 2 * ksi1 - 1;
    double sy = 2 * ksi2 - 1;

      // map square to (r,\theta):
    if ( sx == 0.0 && sy == 0.0 )
    {
      result[0] = 0.0;
      result[1] = 0.0;
      return null;
    }
    if ( sx >= -sy )
    {
      if ( sx > sy )
      {
          // handle first region of disk:
        r = sx;
        if ( sy > 0.0 )
	  theta = sy/r;
	else
	  theta = 8.0f + sy/r;
      }
      else
      {
	  // handle second region of disk:
        r = sy;
	theta = 2.0f - sx/r;
      }
    }
    else
    {
      if ( sx <= sy )
      {
          // handle third region of disk:
        r = -sx;
	theta = 4.0f - sy/r;
      }
      else
      {
          // handle fourth region of disk:
	r = -sy;
	theta = 6.0f + sx/r;
      }
    }
    theta *= Math.PI / 4.f;
    result[0] = r*Math.cos( theta );
    result[1] = r*Math.sin( theta );

    return result;
  }

  /** Generates point on the light. */
  public static double[] sampleSquare ( double u,
                                        double v,
                                        final double[] vertUR,
                                        final double[] vertUL,
                                        final double[] vertLL,
                                        final double[] vertLR,
                                        double[] pos )
  {
    if ( pos == null )
      pos = new double[ 3 ];

    if ( rnd.uniformNumber() < 0.5 )
      randomPointFromTriangle( u, v, vertUL, vertUR, vertLR, pos );
    else
      randomPointFromTriangle( u, v, vertUL, vertLR, vertLL, pos );

    return pos;
  }

  /**
   * Generates random point in the given triangle.
   * Works in either dimension (2D, 3D).
   *
   * @param  a Triangle vertex A.
   * @param  b Triangle vertex B.
   * @param  c Triangle vertex C.
   * @param  result Pre-allocated result array (can be {@code null}).
   * @return Random point or {@code null} in case of failure.
   */
  public static final double[] randomPointFromTriangle ( double u,
                                                         double v,
                                                         double[] a,
                                                         double[] b,
                                                         double[] c,
                                                         double[] result )
  {
    if ( a == null ||
         b == null ||
         c == null ) return null;

    int len = a.length;
    if ( result == null || result.length < len )
      result = new double[ len ];

    if ( u + v > 1.0 )
    {
      u = 1.0 - u;
      v = 1.0 - v;
    }

    for ( int i = 0; i < len; i++ )
      result[i] = a[i] + u * (b[i] - a[i]) + v * (c[i] - a[i]);

    return result;
  }

  /**
   * Uniformly samples cone.
   *
   * @param  u           Uniform random number.
   * @param  v           Uniform random number.
   * @param  cosThetaMax Maximal possible cosinus of angle theta.
   * @param  dir         Pre-allocated array to hold result.
   * @return Point on the cone.
   */
  public static double[] uniformSampleCone ( double u,
			 		     double v,
					     double cosThetaMax,
                                             double[] result )
  {
    if ( result == null )
      result = new double[ 3 ];

    double cosTheta = lerp( u, cosThetaMax, 1.f );
    double sinTheta = Math.sqrt( 1.f - cosTheta*cosTheta );
    double phi = v * 2.f * Math.PI;

    result[0] = Math.cos( phi ) * sinTheta;
    result[1] = Math.sin( phi ) * sinTheta;
    result[2] = cosTheta;

    return result;
  }

  /** Are these directions on the same side of hemisphere. */
  public static boolean sameHemisphere ( double[] in, double[] normal, double[] out )
  {
    if ( in     == null ||
         normal == null ||
         out    == null )
      return false;

    return Geometry.dotProduct3D( in, normal ) *
           Geometry.dotProduct3D( out, normal ) < 0.0;
  }

  /** */
  public static double lerp ( double t, double v1, double v2 )
  {
    return (1.f - t) * v1 + t * v2;
  }

  /**
   * Balance heuristic.
   *
   * @param  f    Function f.
   * @param  pdfF Underlying probability density function of f.
   * @param  g    Function g.
   * @param  pdfG Underlying probability density function of g.
   * @return Weight of the f function.
   */
  public static double balanceHeuristic ( double f, double pdfF, double g, double pdfG )
  {
    return (f * pdfF) / (f * pdfF + g * pdfG);
  }

  /**
   * Power heuristic.
   *
   * @param  f    Function f.
   * @param  pdfF Underlying probability density function of f.
   * @param  g    Function g.
   * @param  pdfG Underlying probability density function of g.
   * @return Weight of the f function.
   */
  public static double powerHeuristic ( double f, double pdfF, double g, double pdfG )
  {
    double fp = f * pdfF,
           gp = g * pdfG;
    return (fp * fp) / (fp * fp + gp * gp);
  }
}