package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Order2D example - serpentine image order.
 *
 * @version 0.25 $Rev: 385 $ $Date: 2007-01-07 01:47:21 +0100 (ne, 07 I 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/OrderSerpentine.java">OrderSerpentine.java</a>
 */
public class OrderSerpentine extends Piece implements Order2D
{
  //==========================================================================
  //  Data:

  /** Implemented order types. */
  protected static enum OrderType
  {
    /** Scanline order (from left to right, from top to bottom). */
    ORDER_SCANLINE,
    /** Serpentine order (from top to bottom, first line from left to right). */
    ORDER_SERPENTINE,
  }

  /** Actual order type { scanline | serpentine }. */
  protected OrderType type = OrderType.ORDER_SCANLINE;

  /** Actual bitmap width in pixels. */
  protected int width;

  /** Actual bitmap height in pixels. */
  protected int height;

  /** X-coordinate of the next pixel in the current pass. */
  protected int actX;

  /** Y-coordinate of the next pixel in the current pass (== #height at the end of pass). */
  protected int actY;

  //==========================================================================
  //  Interface Order2D:

  /**
   * Sets the source rectangle size. Initializes the pass algorithm. For both <b>Active</b> and
   * <b>Passive</b> modes.
   *
   * @param  _m Horizontal rectangle size (width).
   * @param  _n Vertical rectangle size (height).
   * @see    #init
   */
  public void setSize ( int _m, int _n )
  {
    width  = _m;
    height = _n;
    init();
  }

  /**
   * Re-initializes the pass algorithm. For <b>Passive</b> pass mode.
   *
   * @see    #setSize
   */
  public void init ()
  {
    actX = actY = 0;
  }

  /**
   * Returns the next 2D point. For <b>Passive</b> pass mode.
   *
   * @param  coords Pre-allocated array of size 2 (for [X,Y] coordinates).
   *                Can be {@code null}.
   * @return The next point's coordinates ([X,Y]).
   * @see    #init
   */
  public int[] next ( int[] coords )
  {
    if ( actY >= height )
      return null;

    if ( coords == null )
      coords = new int[ 2 ];

    coords[0] = actX;
    coords[1] = actY;
    switch ( type )
    {
      case ORDER_SCANLINE:
        if ( ++actX >= width )
        {
          actX = 0;
          actY++;
        }
        break;

      case ORDER_SERPENTINE:
        if ( (actY & 1) == 0 )            // even line => from left
        {
          if ( ++actX >= width )
          {
            actX--;
            actY++;
          }
        }
        else                              // odd line => from right
          if ( --actX < 0 )
          {
            actX++;
            actY++;
          }
        break;
    }

    return coords;
  }

  /**
   * Anything to pass? For <b>Passive</b> pass mode.
   *
   * @return <tt>true</tt> if there is at least one pixel to pass.
   * @see    #init
   * @see    #next
   */
  public boolean isNext ()
  {
    return( actY < height );
  }

  /**
   * Executes the whole rectangle pass. For <b>Active</b> (="modal") pass mode.
   *
   * @see    #setSize
   */
  public void pass ()
  {
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // ORDER_TYPE:
    if ( key.compareTo(ORDER_TYPE) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(ORDER_TYPE) == 0 )
      return type.ordinal();
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "OrderSertpentine";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "Order2D";

  /** Object description. */
  private final static String DESCRIPTION = "Serpentine image order.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + ".order";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"Order2D" );
        // Property:
      t.propBegin( ORDER_TYPE, TYPE_INTEGER, "Order type", true );
        t.propDefault( OrderType.ORDER_SCANLINE.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Scanline",   OrderType.ORDER_SCANLINE.ordinal(),   "Scanline order" );
        t.propEnum( "Serpentine", OrderType.ORDER_SERPENTINE.ordinal(), "Serpentine order" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
