package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Ordered Dithering.
 * Can scale input raster image, converts it to {@link RasterGraphics#MODE_GRAY} and
 * uses only {@code 0.0} and {@code 1.0} values.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/OrderedDither.java">OrderedDither.java</a>
 */
public class OrderedDither extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Target image width in pixels ({@code 0} for original size). */
  protected int width      = 0;

  /** Target image height in pixels ({@code 0} for original size). */
  protected int height     = 0;

  /** Scale factor ({@code 0.0} or {@code 1.0} for original size). */
  protected double scale   = 1.0;

  /** Simulated gamma-exponent. */
  protected double gamma   = 2.0;

  /** Actual dither matrix size (power of two). */
  protected int matrixSize = 2;

  /** Dither matrix. */
  protected double[][] matrix = null;

  //==========================================================================
  //  Support:

  /** Computes dither matrix if necessary. */
  protected void assertMatrix ()
  {
    if ( matrix        != null &&
         matrix.length == matrixSize ) return;

    matrix = new double[ matrixSize ][ matrixSize ];
    int ss = matrixSize * matrixSize;
    int ord, inc, mask;
    int x, y;
    double[] row;
    double granul = 1.0 / (ss + 1.0);

    for ( y = 0; y < matrixSize; y++ )
    {
      row = matrix[y];

      for ( x = 0; x < matrixSize; x++ )
      {
        ord  = 1;                         // pixel order number
        inc  = ss;                        // pixel order increment
        mask = 1;
        while ( (inc >>= 2) > 0 )
        {
          if ( (y & mask) == 0 )
          {
            if ( (x & mask) > 0 )
              ord += 2 * inc;
          }
          else
            ord += ((x & mask) == 0 ? 3 * inc : inc);

          mask += mask;
        }

        row[x] = ord * granul;
      }
    }
  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transform.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    final RasterGraphics input  = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    final RasterGraphics output = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );

    if ( output == null ) return false;

    assertMatrix();
    int i, j;

      // output image:
    int iWid = (input == null) ? 1000 : input.getWidth();
    int iHei = (input == null) ?  250 : input.getHeight();
    int oWid = width;
    int oHei = height;

    if ( oWid == 0 )
      if ( Geometry.isZero( scale ) ||
           Geometry.isZero( scale - 1.0 ) )
        oWid = iWid;
      else
        oWid = Formula.round( iWid * scale );
    if ( oWid < 2 ) oWid = 2;

    if ( oHei == 0 )
      if ( Geometry.isZero( scale ) ||
           Geometry.isZero( scale - 1.0 ) )
        oHei = iHei;
      else
        oHei = Formula.round( iHei * scale );
    if ( oHei < 2 ) oHei = 2;

    output.init( oWid, oHei, RasterGraphics.MODE_GRAY, 0 );

    double pixel;
    int jmod;

    if ( input == null )                  // void input => draw horizontal vignette
    {
      double dGray = -1.0 / (oWid - 1.0);
      for ( j = 0; j < oHei && !userBreak; j++ )
      {
        jmod = j % matrixSize;
        for ( i = 0, pixel = 1.0; i < oWid; i++, pixel += dGray )
          output.putPixel( i, j, (pixel > matrix[jmod][i % matrixSize]) ? 1.0 : 0.0 );
      }

      return true;
    }

    if ( iWid == oWid &&
         iHei == oHei )                   // dithering (no scaling)
      for ( j = 0; j < oHei && !userBreak; j++ )
      {
        jmod = j % matrixSize;
        for ( i = 0; i < oWid; i++ )
        {
          pixel = Math.pow( input.getGrayDouble( i, j ), gamma );
          output.putPixel( i, j, (pixel > matrix[jmod][i % matrixSize]) ? 1.0 : 0.0 );
        }
      }

    else                                  // halftoning (scaling)
    {
      double kx;                          // src_x = kx * target_x
      double ky;                          // src_y = ky * target_y

      kx = (iWid - 1.0) / (oWid - 1.0);
      ky = (iHei - 1.0) / (oHei - 1.0);

      double x, y;
      double pix00, pix10, pix01, pix11;
      int    iy, ix;
      double fx, fy, fx1, fy1;

      for ( j = 0, y = 0.0; j < oHei && !userBreak; j++, y += ky )
      {
        jmod = j % matrixSize;
        fy   = Math.floor( y );
        iy   = (int)fy;                   // y = iy + fy
        fy   = y - fy;
        fy1  = 1.0 - fy;

        for ( i = 0, x = 0.0; i < oWid; i++, x += kx )
        {
          fx  = Math.floor( x );
          ix  = (int)fx;                  // x = ix + fx
          fx  = x - fx;
          fx1 = 1.0 - fx;

            // four surrounding pixels:
          pix00 = Math.pow( input.getGrayDouble( ix,     iy     ), gamma );
          pix10 = Math.pow( input.getGrayDouble( ix + 1, iy     ), gamma );
          pix01 = Math.pow( input.getGrayDouble( ix,     iy + 1 ), gamma );
          pix11 = Math.pow( input.getGrayDouble( ix + 1, iy + 1 ), gamma );

            // bilinear interpolation:
          pixel = fx1 * ( fy1 * pix00 + fy * pix01 ) +
                  fx  * ( fy1 * pix10 + fy * pix11 );

          output.putPixel( i, j, (pixel > matrix[jmod][i % matrixSize]) ? 1.0 : 0.0 );
        }
      }
    }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String WIDTH       = "Width";
  public final static String HEIGHT      = "Height";
  public final static String SCALE       = "Scale";
  public final static String GAMMA       = "Gamma";
  public final static String MATRIX_SIZE = "Matrix size";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // WIDTH:
    if ( key.compareTo(WIDTH) == 0 )
    {
      width = intProperty( value, width, 0, 1000000 );
      return;
    }
      // HEIGHT:
    if ( key.compareTo(HEIGHT) == 0 )
    {
      height = intProperty( value, height, 0, 1000000 );
      return;
    }
      // SCALE:
    if ( key.compareTo(SCALE) == 0 )
    {
      scale = doubleProperty( value, scale, 0.0, 1000.0 );
      return;
    }
      // GAMMA:
    if ( key.compareTo(GAMMA) == 0 )
    {
      gamma = doubleProperty( value, gamma, 0.01, 100.0 );
      return;
    }
      // MATRIX_SIZE:
    if ( key.compareTo(MATRIX_SIZE) == 0 )
    {
      int ms = intProperty( value, matrixSize, 2, 1024 );
      matrixSize = 2;
      while ( matrixSize + matrixSize <= ms )
        matrixSize += matrixSize;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(WIDTH) == 0 )
      return width;
    if ( key.compareTo(HEIGHT) == 0 )
      return height;
    if ( key.compareTo(SCALE) == 0 )
      return scale;
    if ( key.compareTo(GAMMA) == 0 )
      return gamma;
    if ( key.compareTo(MATRIX_SIZE) == 0 )
      return matrixSize;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "OrderedDither";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Ordered Dither.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_RASTER + '.' + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT, IFACE+"Trigger" );
      t.newOptOutputPlug( "image1", IFACE+"RasterGraphics" );
      t.newOutputPlug(    "image2", IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( WIDTH, TYPE_INTEGER, "Output width", true );
        t.propDefault( 0 );
        t.propBounds(  0, 1000000 );
      t.propEnd();
      t.propBegin( HEIGHT, TYPE_INTEGER, "Output height", true );
        t.propDefault( 0 );
        t.propBounds(  0, 1000000 );
      t.propEnd();
      t.propBegin( SCALE, TYPE_DOUBLE, "Image scale factor", true );
        t.propDefault( 1.0 );
        t.propBounds(  0.0, 1000.0 );
      t.propEnd();
      t.propBegin( GAMMA, TYPE_DOUBLE, "Simulated gamma-exponent", true );
        t.propDefault( 2.0 );
        t.propBounds(  0.01, 100.0 );
      t.propEnd();
      t.propBegin( MATRIX_SIZE, TYPE_INTEGER, "Dither matrix size", true );
        t.propDefault( 2 );
        t.propManipulator( MANIPULATOR_COMBO );
        for ( int i = 2; i <= 1024; i += i )
        {
          String val = String.valueOf( i );
          t.propEnum( val, i, val );
        }
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
