package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Oren-Nayar shading model.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/OrenNayarModel.java">OrenNayarModel.java</a>
 */
public class OrenNayarModel extends LambertianModel
{
  //==========================================================================
  //  Data:

  /** Angle. */
  public double sigma = 0.0;

  /** Precomputed value. */
  public double A;

  /** Precomputed value. */
  public double B;

  /** Default constructor. */
  public OrenNayarModel ()
  {
    color = new double[] { 1.0, 1.0, 1.0 };
  }

  //==========================================================================
  //  Interface BSSRDF:

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as monochrome).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @return BSSRDF coefficient <b>f(in-&gt;out)</b>.
   * @see    #monoBSSRDF(MicroFacet,double[],double[],int)
   * @see    #colorBSSRDF(double[],double[],double[],int,double[])
   */
  @Override
  public double monoBSSRDF ( boolean worldCoords,
                             double[] in,
                             double[] out,
                             double[] normal,
                             double[] tangent,
                             int mask )
  {
    if ( normal == null )
      return 0.0;

    if ( !( mask == BSSRDF.COMP_DIFFUSE ||
            mask == BSSRDF.COMP_ALL ) )
      return 0.0;

    double[] inLoc  = new double[ 3 ];
    double[] outLoc = new double[ 3 ];
    if ( worldCoords )
    {
      worldToLocal( normal, tangent, in,  inLoc  );
      worldToLocal( normal, tangent, out, outLoc );
    }
    else
    {
      System.arraycopy( in,  0, inLoc,  0, in.length  );
      System.arraycopy( out, 0, outLoc, 0, out.length );
    }

    double cosThetaI = inLoc[2];
    double cosThetaO = outLoc[2];
    double sinThetaI = Math.sqrt( Math.max( 0.f, 1.f - cosThetaI*cosThetaI ) );
    double sinThetaO = Math.sqrt( Math.max( 0.f, 1.f - cosThetaO*cosThetaO ) );

      // compute cosine term of Oren-Nayar model:
    double maxCos = 0.f;
    if ( sinThetaI > 1e-4 &&
         sinThetaO > 1e-4 )
    {
      double sinPhiI = inLoc[1] / sinThetaI,
             cosPhiI = inLoc[0] / sinThetaI;
      double sinPhiO = inLoc[1] / sinThetaO,
             cosPhiO = inLoc[0] / sinThetaO;
      double dcos = cosPhiI * cosPhiO + sinPhiI * sinPhiO;
      maxCos = Math.max( 0.0, dcos );
    }

      // compute sine and tangent terms of Oren-Nayar model:
    double sinAlpha, tanBeta;
    if ( cosThetaI > cosThetaO )
    {
      sinAlpha = sinThetaO;
      tanBeta  = sinThetaI / cosThetaI;
    }
    else
    {
      sinAlpha = sinThetaI;
      tanBeta =  sinThetaO / cosThetaO;
    }

      // compute result:
    double invPI = 1.0 / Math.PI;
    return color[0] * invPI * (A + B * maxCos * sinAlpha * tanBeta);
  }

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as band-color).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BSSRDF spectral coefficients.
   * @see    #monoBSSRDF(double[],double[],double[],int)
   * @see    #colorBSSRDF(MicroFacet,double[],double[],int,double[])
   */
  @Override
  public double[] colorBSSRDF ( boolean worldCoords,
                                double[] in,
                                double[] out,
                                double[] normal,
                                double[] tangent,
                                int mask,
                                double[] result )
  {
    if ( in      == null ||
         out     == null ||
         normal  == null ||
         tangent == null )
      return null;

    if ( !( mask == BSSRDF.COMP_DIFFUSE ||
            mask == BSSRDF.COMP_ALL ) )
      return new double[] { 0.0, 0.0, 0.0 };

    if ( result == null )
      result = new double[ 3 ];

      // make sure that vectors in and out are in local coordinates:
    double[] inLoc  = new double[ 3 ];
    double[] outLoc = new double[ 3 ];
    if ( worldCoords )
    {
      worldToLocal( normal, tangent, in, inLoc   );
      worldToLocal( normal, tangent, out, outLoc );
    }
    else
    {
      System.arraycopy( in,  0, inLoc,  0, in.length  );
      System.arraycopy( out, 0, outLoc, 0, out.length );
    }

    double cosThetaI = inLoc[2];
    double cosThetaO = outLoc[2];
    double sinThetaI = Math.sqrt( Math.max( 0.f, 1.f - cosThetaI*cosThetaI ) );
    double sinThetaO = Math.sqrt( Math.max( 0.f, 1.f - cosThetaO*cosThetaO ) );

      // compute cosine term of Oren-Nayar model:
    double maxCos = 0.f;
    if ( sinThetaI > 1E-4 &&
         sinThetaO > 1E-4 )
    {
      double sinPhiI = inLoc[1] / sinThetaI,
             cosPhiI = inLoc[0] / sinThetaI;
      double sinPhiO = outLoc[1] / sinThetaO,
             cosPhiO = outLoc[0] / sinThetaO;
      double dcos = cosPhiI * cosPhiO + sinPhiI * sinPhiO;
      maxCos = Math.max( 0.0, dcos );
    }

      // compute sine and tangent terms of Oren-Nayar model:
    double sinAlpha, tanBeta;
    if ( cosThetaI > cosThetaO )
    {
      sinAlpha = sinThetaO;
      tanBeta  = sinThetaI / cosThetaI;
    }
    else
    {
      sinAlpha = sinThetaI;
      tanBeta =  sinThetaO / cosThetaO;
    }

      // compute result:
    double invPI = 1.0 / Math.PI;
    for ( int i = 0; i < result.length; i++ )
      result[i] = color[i] * invPI * (A + B * maxCos * sinAlpha * tanBeta);

    return result;
  }

  //==========================================================================
  //  Interface Property:

  public final static String SIGMA = "Sigma";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

    if ( key.compareTo( BSSRDF.MATERIAL_COLOR ) == 0 &&
         value instanceof double[] )
    {
      color = (double[])value;
      return;
    }

    if ( key.compareTo( SIGMA ) == 0 )
    {
      sigma = doubleProperty( value, sigma );

      double sigmaRad  = (sigma * Math.PI) / 180.0;
      double sigma2    = sigmaRad*sigmaRad;
      A = 1.f - (sigma2 / (2.f * (sigma2 + 0.33f)));
      B = 0.45f * sigma2 / (sigma2 + 0.09f);

      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo( BSSRDF.MATERIAL_COLOR ) == 0 ) return color;
    if ( key.compareTo( SIGMA ) == 0 ) return sigma;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Oren-Nayar shading model";

  /** Object description. */
  private final static String DESCRIPTION = "Oren-Nayar shading model.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"BSSRDF" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}