package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.piece.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Path tracing of JGL scenes.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 860 $ $Date: 2009-02-26 23:25:37 +0100 (čt, 26 II 2009) $ $Author: Tomas $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/PathTracing.java">PathTracing.java</a>
 */
public class PathTracing extends DefaultRenderer
{
  //==========================================================================
  //  Data:

  /** Light map for direct photon storing. */
  protected LightMap global;

  /** Light tracer to compute illumination of scene. */
  protected LightTracing tracer;

  /** Number of traced reflected rays for indirect illumination. */
  protected int numRays;

  /** {@code True} if direct illumination should be used. */
  protected boolean useDirect;

  /** {@code True} if caustics should be used. */
  protected boolean useCaustics;

  /** {@code True} if indirect illumination should be used. */
  protected boolean useIndirect;

  /** Are they used shadow photons? */
  protected boolean shadowPhotons = false;

  //==========================================================================
  //  Support functions:

  /** Prepare scene for rendering. */
  @Override
  protected boolean init ()
  {
    if ( scene == null )
      scene = (RTScene)getInterface(       PL_OUTPUT,       IFACE+"RTScene"       );
    if ( inter == null )
      inter = (Intersectable)getInterface( PL_INTERSECTION, IFACE+"Intersectable" );
    if ( scene == null ||
         inter == null )
      return false;

    if ( global == null )
      global = (LightMap)getInterface(     PL_PHOTON,  IFACE+"LightMap"     );
    if ( tracer == null )
      tracer = (LightTracing)getInterface( PL_TRIGGER, IFACE+"LightTracing" );

    if ( tracer      != null &&
         reflections != Reflections.SPECULAR )
      tracer.illuminateScene();

    isInit = true;

    return true;
  }

  //==========================================================================
  //  Interface ImageFunction:

  /**
   * Recursive shading function - computes color contribution of the given ray (shot from the
   * origin {@code p0} into direction vector {@code p1}). Recursion is stopped
   * by a hybrid method: {@code importance} and {@code level} are checked.
   *
   * @param  level Actual recursion depth.
   * @param  importance Importance of the actual ray.
   * @param  p0 Ray origin.
   * @param  p1 Direction vector of the ray.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (ray sub-signature) used for adaptive subsampling.
   * @see    #minImportance
   * @see    #maxLevel
   * @see    #scene
   * @see    #inter
   * @todo   Some better way of computing importance.
   */
  @Override
  protected long shade ( int level, double importance, double[] p0, double[] p1,
                         double[] color )
  {
      // get the intersection:
    MicroFacet mf = inter.intersection( p0, p1 );
    if ( mf == null )                     // no intersection => use background color
    {
      System.arraycopy( background, 0, color, 0, Math.min( color.length, background.length ) );
      return 1L;
    }

      // compute and log important attributes of the intersection point:
    if ( mf.assertAttributes( dataMask ) != 0 ) return 0L;
    if ( mf.assertAttributes( MicroFacet.VALID_COORD ) == 0 && LogFile.tracing )
    {
      LogFile.log( String.format( Locale.US, "Intersection%d at t = %.4f with '%s' (%s,%s), p1.N: %.3f, mask: %04x",
                                  level, mf.t, mf.geometry.getClass().getName(),
                                  mf.enter ? "enter" : "leave", mf.front ? "front" : "back",
                                  Geometry.dotProduct3D( p1, mf.normal ), dataMask ) );

      LogFile.log( String.format( Locale.US, "  A: [ %.4f, %.4f, %.4f ], p1: [ %.4f, %.4f, %.4f ], N: [ %.4f, %.4f, %.4f ]",
                                  mf.coord[0], mf.coord[1], mf.coord[2],
                                  p1[0], p1[1], p1[2],
                                  mf.normal[0], mf.normal[1], mf.normal[2] ) );
    }

      // hash code:
    long hash = mf.sceneNode.hashCode();  // init: hash code of the intersected solid

      // apply textures:
    if ( useTextures &&
         mf.textures != null )
      for ( Texture tex : mf.textures )
        hash = hash * HASH_TEXTURE + tex.compute( mf );

    Arrays.fill( color, 0.0 );
    p1[0] = -p1[0];
    p1[1] = -p1[1];
    p1[2] = -p1[2];
    double[] Le   = new double[ color.length ];    // emission
    double[] Lc   = new double[ color.length ];    // caustics
    double[] Ld   = new double[ color.length ];    // direct light
    double[] Li   = new double[ color.length ];    // indirect (reflected) light
    double[] brdf = new double[ color.length ];
    double pdf    = 1.0;
    double cos    = 0.0;
    mf.bssrdf.getOrigColor( brdf );

    int i;
    if ( useDirect || ( useIndirect && level > 0 ) )
    {
      directIllumination( mf, p1, Ld );

        // handle specular reflections:
      if ( mf.bssrdf.type() == BSSRDF.COMP_SPECULAR )
        specular( level, mf, p1, Ld );
    }

    if ( useIndirect )
    {
        // compute new importance:
      double newImportance = importance;
      double brdfAvrg      = ( brdf[0] + brdf[1] + brdf[2] ) / 3.0;
      if ( brdfAvrg >= 1.0 ) brdfAvrg = 0.9;
      newImportance = importance * brdfAvrg;

        // sample the hemishere:
      double[] subLi = new double[ color.length ];
      double[] dir = new double[ 3 ];
      pdf = mf.bssrdf.scatter( mf,
                               p1,
                               rnd.randomDouble( 0.0, 1.0 ),
                               rnd.randomDouble( 0.0, 1.0 ),
                               BSSRDF.COMP_ALL,
                               dir,
                               brdf );
      cos = Geometry.dotProduct3D( mf.normal, dir );

        // possibly end tracing:
      if ( rnd.randomDouble( 0.0, 1.0 ) < newImportance && pdf > 0.0 )
      {
        double invImp = 1.0 / newImportance;
        double invPdf = 1.0 / pdf;
        hash += shade( ++level, newImportance, mf.coord, dir, subLi );
        for ( i = 0; i < color.length; i++ )
          Li[i] += brdf[i] * subLi[i] * cos * invImp * invPdf;
      }
    }

    for ( i = 0; i < color.length; i++ )
      color[i] += Le[i] + Ld[i] + Lc[i] + Li[i];

    return hash;
  }

  //==========================================================================
  //  Interface Property:

  public final static String DIRECT      = "Direct";
  public final static String SHADOW_RAYS = "Number of shadow rays";
  public final static String SAMPLING    = "Sampling";
  public final static String INDIRECT    = "Indirect";
  public final static String REFLECTIONS = "Reflections";
  public final static String OCCLUSION   = "Occlusion";
  public final static String TEXTURES    = "Textures";
  public final static String LOG         = "Log tracing";

   /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // SHADOW_RAYS:
    if ( key.compareTo(SHADOW_RAYS) == 0 )
    {
      numShadowRays = intProperty( value, numShadowRays );
      return;
    }
      // SAMPLING:
    if ( key.compareTo(SAMPLING) == 0 )
    {
      sampling = enumProperty( value, sampling );
      return;
    }
      // REFLECTIONS:
    if ( key.compareTo(REFLECTIONS) == 0 )
    {
      reflections = enumProperty( value, reflections );
      return;
    }
      // DIRECT:
    if ( key.compareTo(DIRECT) == 0 )
    {
      useDirect = booleanProperty( value, useDirect );
      return;
    }
      // INDIRECT:
    if ( key.compareTo(INDIRECT) == 0 )
    {
      useIndirect = booleanProperty( value, useIndirect );
      return;
    }
      // OCCLUSION:
    if ( key.compareTo(OCCLUSION) == 0 )
    {
      occlusion = booleanProperty( value, occlusion );
      return;
    }
      // TEXTURES:
    if ( key.compareTo(TEXTURES) == 0 )
    {
      useTextures = booleanProperty( value, useTextures );
      return;
    }
      // LOG:
    if ( key.compareTo(LOG) == 0 )
    {
      LogFile.tracing = booleanProperty( value, LogFile.tracing );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(SHADOW_RAYS) == 0 )
      return numShadowRays;
    if ( key.compareTo(SAMPLING) == 0 )
      return sampling.ordinal();
    if ( key.compareTo(REFLECTIONS) == 0 )
      return reflections.ordinal();
    if ( key.compareTo(DIRECT) == 0 )
      return useDirect;
    if ( key.compareTo(INDIRECT) == 0 )
      return useIndirect;
    if ( key.compareTo(TEXTURES) == 0 )
      return useTextures;
    if ( key.compareTo(LOG) == 0 )
      return LogFile.tracing;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Path tracing";

  /** Object name. */
  private final static String TEMPLATE_NAME = "Path tracing";

  /** Object description. */
  private final static String DESCRIPTION = "Path-tracing of JGL scenes";

  /** Object category. */
  protected final static String CATEGORY = C_3D + "." + C_RENDER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,        IFACE+"ImageFunction" );
      t.newOutputPlug(    PL_OUTPUT,       IFACE+"RTScene" );
      t.newOutputPlug(    PL_INTERSECTION, IFACE+"Intersectable" );
      t.newOptOutputPlug( PL_TRIGGER,      IFACE+"LightTracing" );
      t.newOptOutputPlug( PL_PHOTON,       IFACE+"LightMap" );
        // Property:
      t.propBegin( SHADOW_RAYS,  TYPE_INTEGER, "Number of shadow rays for light integrating hemisphere", true );
        t.propDefault( 10 );
      t.propEnd();
      t.propBegin( SAMPLING, TYPE_INTEGER, "Sampling", true );
        t.propDefault( Sampling.SAMPLING_BOTH.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Sampling both",   Sampling.SAMPLING_BOTH.ordinal(),   "Sample according to lights as well as according to BSDF." );
        t.propEnum( "Sampling BSDF",   Sampling.SAMPLING_BSDF.ordinal(),   "Sample according to BSDF only (problem with delta lights!)." );
        t.propEnum( "Sampling lights", Sampling.SAMPLING_LIGHTS.ordinal(), "Sample according to light only." );
      t.propEnd();
      t.propBegin( REFLECTIONS, TYPE_INTEGER, "Reflections", true );
        t.propDefault( Reflections.ALL.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "All",      Reflections.ALL.ordinal(),      "All types of materials are used." );
        t.propEnum( "Diffuse",  Reflections.DIFFUSE.ordinal(),  "Only diffuse materials are used." );
        t.propEnum( "Glossy",   Reflections.GLOSSY.ordinal(),   "Only glossy materials are used." );
        t.propEnum( "Specular", Reflections.SPECULAR.ordinal(), "Only specular materials are used." );
      t.propEnd();
      t.propBegin( DIRECT,       TYPE_BOOLEAN, "Compute direct?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( INDIRECT,     TYPE_BOOLEAN, "Compute indirect?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( TEXTURES,     TYPE_BOOLEAN, "Compute textures?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( OCCLUSION,   TYPE_BOOLEAN, "Occlusion?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( LOG,          TYPE_BOOLEAN, "Log the tracing?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}