package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Module for photon storing. Primary intended for photon mapping application.
 *
 * @author Tomas Svab
 * @version 0.26 $Rev: 860 $ $Date: 2009-02-26 23:25:37 +0100 (čt, 26 II 2009) $ $Author: Tomas $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/PhotonMap.java">PhotonMap.java</a>
 * @todo    Add some more effective method for photon selecting.
 */
public class PhotonMap extends Piece implements LightMap
{
  //==========================================================================
  //  Support classes:

  /** Single emitted photon. */
  class Photon
  {
    /** Photon position. */
    public double[] pos;

    /** Photon power. */
    public float[] power;

    /** Incoming direction. */
    public char theta, phi;

    /** Splitting plane for kd-tree. */
    public short plane;

    Photon ()
    {
      pos   = new double[ 3 ];
      power = new float[ 3 ];
    }
  };

  /** Contains nearest photons. */
  class NearestPhotons
  {
    /** Maximal number of stored photons. */
    public int max;

    /** Number of found photons. */
    public int found;

    /** Value 1 if the heap has been already build, 0 if hasn't. */
    public int gotHeap;

    /** Position of location of estimate. */
    public double[] pos;

    /** Normal of location of estimate. */
    public double[] normal;

    /** Maximal squared distance between photon and a point of estimate. */
    public double[] dist2;

    /** The list of nearest photons. */
    public Photon[] index;

    NearestPhotons ( double[] pos, double[] normal, double maxDist, int numPhotons )
    {
      if ( pos != null )
        this.pos = pos;
      else
        pos = new double[ 3 ];

      if ( normal != null )
        this.normal = normal;
      else
        normal = new double[ 3 ];

      dist2  = new double[ numPhotons + 1 ];
      index  = new Photon[ numPhotons + 1 ];

      found    = 0;
      gotHeap  = 0;
      dist2[0] = maxDist * maxDist;

      max = numPhotons;
      for ( int i = 0; i < max + 1; i++ )
        index[i] = new Photon();
    }
  };

  //==========================================================================
  //  Data:

  /** Volumes used for finding photons. */
  protected enum Volume
  {
    /** Simple sphere. */
    SPHERE,
    /** More precise than sphere especially in the corners and curved surfaces. */
    DISC,
  };

  /** Actual volume used for finding photons. */
  protected Volume volume = Volume.SPHERE;

  /** Supported filters used in a radiance/irradiance estimation. */
  protected enum Filter
  {
    /** Estimate without filtering. */
    NONE,
    /** Cone filter. */
    CONE,
    /** Gaussian filter. */
    GAUSSIAN,
    /** Simpson's kenel. */
    SIMPSONS_KERNEL,
  };

  /** Actual filter for radiance/irradiance estimate. */
  protected Filter filter = Filter.NONE;

  /** Types of searching of nearest photons. */
  protected enum Type
  {
    /** KD-tree is used -> O(log n). */
    KD_TREE,
    /** Each photon is tested -> O(n). */
    ARRAY,
  };

  /** Actual type of searching nearest photons. */
  protected Type type = Type.KD_TREE;

  /** Maximal number of stored photons. */
  protected int maxPhotons;

  /** Number of stored photons. */
  protected int storedPhotons;

  /** Half of stored photons. */
  protected int halfStoredPhotons;

  /** The list of photons stored in the photon map. */
  protected Photon[] photon;

  /** Sinus of angle theta. */
  protected float[] sintheta;

  /** Cosinus of angle theta. */
  protected float[] costheta;

  /** Sinus of angle phi. */
  protected float[] sinphi;

  /** Cosinus of angle phi. */
  protected float[] cosphi;

  /** Minimals of bounding box. */
  protected double[] bboxMin;

  /** Maximals of bounding box. */
  protected double[] bboxMax;

  /** Number of already scaled photons. */
  protected int prevScale;

  /** Maximal number of photons which can be used for radiosity estimate. */
  protected int estimate;

  /**
   * Maximal radius of the sphere/circle (maximal distance between
   * estimate point and a photon).
   */
  protected double radius;

  /**
   * Maximal error of normal and direction of photon is radiance/irradiance
   * estimate.
   *
   * If it's used sphere, values should be between -1.0 and 0.0.
   * 0.0 means that every photon on the surface which is directed
   * towards the object (dot product of normal and direction is negative)
   * is used. This could cause some visual problems in the corners, that's
   * why should be this coefficient smaller than 0.0.
   *
   * On the other hand, if it's used disc, this coefficient has to be
   * positive. Maximal value should be something about 0.1 (greater value
   * cause same problems as sphere).
   */
  protected double dirError;

  /**
   * Filter constant characterizing the cone filter. Must be equal
   * or greater than one.
   */
  protected double k;

  /** Was the photon map initialized? */
  protected boolean isInit = false;

  /** Was the photon map balanced? */
  protected boolean isBalanced = false;

  /** Log information about estimates? */
  protected boolean logEstimate = false;

  /** Log all photons stored in photon map? */
  protected boolean logPhotons = false;

  //==========================================================================
  //  Constants:

  /** For Gaussian filte. */
  protected final static double ALPHA = 0.918;

  /** For Gaussian filter. */
  protected final static double BETA = 1.953;

  //==========================================================================
  //  Support methods:

  protected void prepare ()
  {
    storedPhotons     = 0;
    halfStoredPhotons = 0;
    photon            = new Photon[ maxPhotons + 1 ];
    photon[0]         = new Photon();
    prevScale         = 1;

    bboxMin = new double[ 3 ];
    bboxMax = new double[ 3 ];
    bboxMin[0] = bboxMin[1] = bboxMin[2] =  1.E8;
    bboxMax[0] = bboxMax[1] = bboxMax[2] = -1.E8;

    double angle = 0.0;
    costheta = new float[ 256 ];
    sintheta = new float[ 256 ];
    cosphi   = new float[ 256 ];
    sinphi   = new float[ 256 ];
    for ( int i = 0; i < 256; i++ )
    {
      angle = (double)i * ( 1.0/256.0 ) * Math.PI;
      costheta[i] = (float)Math.cos( angle );
      sintheta[i] = (float)Math.sin( angle );
      cosphi[i]   = (float)Math.cos( 2.0*angle );
      sinphi[i]   = (float)Math.sin( 2.0*angle );
    }

    isInit = true;
  }

  /**
   * Finds the nearest photons in the photon map under the given conditions (np).
   *
   * @param index Index of starting photon.
   * @param np    NearestPhotons class as a photon-store.
   */
  protected NearestPhotons locatePhotons ( int index, NearestPhotons np )
  {
     if ( photon == null || photon[index] == null )
       return null;

     Photon p = photon[index];

     double dist1 = np.pos[ p.plane ] - p.pos[ p.plane ];
     if ( type == Type.KD_TREE )
     {
       if ( index < halfStoredPhotons )
       {
         if ( dist1 > 0.0 )       // search right plane first
         {
           locatePhotons( 2*index + 1, np );
           if ( dist1 * dist1 < np.dist2[0] )
             locatePhotons( 2*index, np );
         }
         else                     // search left plane first
         {
           locatePhotons( 2*index, np );
           if ( dist1 * dist1 < np.dist2[0] )
             locatePhotons( 2*index + 1, np );
         }
       }
     }
     else
       locatePhotons( ++index, np );

       // compute squared distance between a current photon and np->pos:
     dist1 = p.pos[0] - np.pos[0];
     double dist2 = dist1 * dist1;
     dist1 = p.pos[1] - np.pos[1];
     dist2 += dist1 * dist1;
     dist1 = p.pos[2] - np.pos[2];
     dist2 += dist1 * dist1;

     if ( dist2 < np.dist2[0] )
     {
         // insert photon in the candidate list:
       if ( np.found < np.max )
       {
           // heap is not full; use array
         np.found++;
         np.dist2[ np.found ] = dist2;
         np.index[ np.found ] = p;
       }
       else
       {
         int j, parent;
         if ( np.gotHeap == 0 )
         {
            // build heap:
          double dst2;
          Photon phot;
          int half_found = np.found / 2;
          for ( int k = half_found; k >= 1; k-- )
          {
            parent = k;
            phot = np.index[k];
            dst2 = np.dist2[k];
            while ( parent <= half_found )
            {
              j = parent + parent;
              if ( j < np.found && np.dist2[j] < np.dist2[j + 1] )
                j++;
              if ( dst2 >= np.dist2[j] )
                break;
              np.dist2[parent] = np.dist2[j];
              np.index[parent] = np.index[j];
              parent = j;
            }
            np.dist2[parent] = dst2;
            np.index[parent] = phot;
          }
          np.gotHeap = 1;
        }

          // insert new photon into max heap
          // delete largest element, insert new and reorder the heap
        parent = 1;
        j = 2;
        while ( j <= np.found )
        {
          if ( j < np.found && np.dist2[j] < np.dist2[j + 1] )
            j++;
          if ( dist2 > np.dist2[j] )
            break;
          np.dist2[parent] = np.dist2[j];
          np.index[parent] = np.index[j];
          parent = j;
          j += j;
        }
        np.index[parent] = p;
        np.dist2[parent] = dist2;
        np.dist2[0] = np.dist2[1];
      }
    }

    return np;
  }

  /**
   * Finds the nearest photons in the photon map under the given conditions (np).
   *
   * @param index Index of starting photon.
   * @param np    NearestPhotons class as a photon-store.
   */
  protected NearestPhotons locatePhotons ( NearestPhotons np )
  {
     if ( photon == null )
       return null;

     for ( int i = 0; i < this.storedPhotons; i++ )
     {
       Photon p = photon[i];

         // compute squared distance between a current photon and np->pos:
       double dist1 = p.pos[0] - np.pos[0];
       double dist2 = dist1 * dist1;
       dist1 = p.pos[1] - np.pos[1];
       dist2 += dist1 * dist1;
       dist1 = p.pos[2] - np.pos[2];
       dist2 += dist1 * dist1;

       if ( dist2 < np.dist2[0] )
       {
           // insert photon in the candidate list:
         if ( np.found < np.max )
         {
             // heap is not full; use array
           np.found++;
           np.dist2[ np.found ] = dist2;
           np.index[ np.found ] = p;
         }
         else
         {
           int j, parent;
           if ( np.gotHeap == 0 )
           {
              // build heap:
            double dst2;
            Photon phot;
            int half_found = np.found / 2;
            for ( int k = half_found; k >= 1; k-- )
            {
              parent = k;
              phot = np.index[k];
              dst2 = np.dist2[k];
              while ( parent <= half_found )
              {
                j = parent + parent;
                if ( j < np.found && np.dist2[j] < np.dist2[j + 1] )
                  j++;
                if ( dst2 >= np.dist2[j] )
                  break;
                np.dist2[parent] = np.dist2[j];
                np.index[parent] = np.index[j];
                parent = j;
              }
              np.dist2[parent] = dst2;
              np.index[parent] = phot;
            }
            np.gotHeap = 1;
          }

            // insert new photon into max heap
            // delete largest element, insert new and reorder the heap
          parent = 1;
          j = 2;
          while ( j <= np.found )
          {
            if ( j < np.found && np.dist2[j] < np.dist2[j + 1] )
              j++;
            if ( dist2 > np.dist2[j] )
              break;
            np.dist2[parent] = np.dist2[j];
            np.index[parent] = np.index[j];
            parent = j;
            j += j;
          }
          np.index[parent] = p;
          np.dist2[parent] = dist2;
          np.dist2[0] = np.dist2[1];
        }
      }
    }

    return np;
  }

  /** Returns the direction of a photon. */
  protected double[] photonDir ( Photon p, double[] dir )
  {
    if ( dir == null )
      dir = new double[ 3 ];

    dir[0] = sintheta[ p.theta ] * cosphi[ p.phi ];
    dir[1] = sintheta[ p.theta ] * sinphi[ p.phi ];
    dir[2] = costheta[ p.theta ];

    return dir;
  }

  /**
   * Scales the power of all photons in the once they have been
   * emitted from the light source.
   */
  public void scalePhotonPower ( float scale )
  {
    if ( storedPhotons > 0 )
    {
      for ( int i = prevScale; i <= storedPhotons; i++ )
      {
        photon[i].power[0] *= scale;
        photon[i].power[1] *= scale;
        photon[i].power[2] *= scale;
      }
      prevScale = storedPhotons;
    }
  }

  //==========================================================================
  //  Balancing of KD-tree:

  /**
   * Creates a left-balanced KD-tree from the flat photon array.
   * These functions should be called before the photon map is used
   * for rendering.
   */
  protected void balance ()
  {
    if ( storedPhotons < 1 )
      return;

    Photon[] organiz = new Photon[ storedPhotons + 1 ];
    for ( int i = 0; i <= storedPhotons; i++ )
      organiz[i] = photon[i];

    balanceSegment( 1, 1, storedPhotons, organiz, photon );

    double maxPower = 0.0;
    int id          = 0;
    if ( logPhotons )
      for ( int i = 1; i <= storedPhotons; i++ )
      {
        LogFile.log( i + ") photon: pos=" + Formula.vectorToString( photon[i].pos ) +
                     ", power=[" + photon[i].power[0] + ", " +
                                 + photon[i].power[1] + ", " +
                                 + photon[i].power[2] + "]"  +
                     ", plane="  + photon[i].plane );

        if ( maxPower < photon[i].power[0] )
        {
          maxPower = photon[i].power[0];
          id = i;
        }
        else if ( maxPower < photon[i].power[1] )
        {
          maxPower = photon[i].power[1];
          id = i;
        }
        else if ( maxPower < photon[i].power[2] )
        {
          maxPower = photon[i].power[2];
          id = i;
        }
      }

    LogFile.log( id + ") photon: power=[" + photon[id].power[0] + ", " +
                                          + photon[id].power[1] + ", " +
                                          + photon[id].power[2] + "]" );

    halfStoredPhotons = storedPhotons/2 - 1;
    isBalanced = true;
  }

  /**
   * Splits photon array into two separates pieces around the median, with
   * all photons below the median in the lower half and all photons
   * above the median in the upper half.
   *
   * @param start  Start of photon block in array.
   * @param end    End of photon block in array.
   * @param median Desired median number.
   * @param axis   Axis to split along.
   * @param p      Photons for splitting.
   */
  protected void medianSplit ( final int start,
                               final int end,
                               final int median,
                               final int axis,
                               Photon[] p )
  {
    int left  = start;
    int right = end;

    while ( right > left )
    {
      final double v = p[right].pos[axis];
      int i = left - 1;
      int j = right;
      for ( ;; )
      {
        while ( p[++i].pos[axis] < v )
          ;
        while ( p[--j].pos[axis] > v && j > left )
          ;
        if ( i >= j )
          break;
        swap( p, i, j );
      }

      swap( p, i, right );
      if ( i >= median )
        right = i - 1;
      if ( i <= median )
        left  = i + 1;
    }
  }

  /**
   * Recursively balances KD-tree.
   *
   * @param index
   * @param start   Start of photon block in array.
   * @param end     End of photon block in array.
   * @param organiz Array for organization of photons which is used when
   *                splitting comes up (input variable).
   * @param balance Balanced array of photons (output variable).
   */
  protected void balanceSegment ( final int index,
                                  final int start,
                                  final int end,
                                  Photon[] organiz,
                                  Photon[] balance )
  {
      // compute new median:
    int median = 1;
    while ( ( 4*median ) <= ( end - start + 1 ) )
      median += median;
    if ( ( 3*median ) <= ( end - start + 1 ) )
    {
      median += median;
      median += start - 1;
    } else
      median = end - median + 1;

      // find axis to split along:
    int axis = 2;
    if ( ( bboxMax[0] - bboxMin[0] ) > ( bboxMax[1] - bboxMin[1] ) &&
         ( bboxMax[0] - bboxMin[0] ) > ( bboxMax[2] - bboxMin[2] ) )
      axis = 0;
    else if ( ( bboxMax[1] - bboxMin[1] ) > ( bboxMax[2] - bboxMin[2] ) )
      axis = 1;

      // partition photon block around the median:
    medianSplit( start, end, median, axis, organiz );
    balance[ index ]       = organiz[ median ];
    balance[ index ].plane = (short)axis;

      // recursively balance the left and right block:
    if ( median > start )
    {
        // balance left segment
      if ( start < median - 1 )
      {
        final double tmp = bboxMax[axis];
        bboxMax[axis] = balance[index].pos[axis];
        balanceSegment( 2*index, start, median - 1, organiz, balance );
        bboxMax[axis] = tmp;
      }
      else
      {
        balance[ 2*index ] = organiz[start];
      }
    }

    if ( median < end )
    {
        // balance right segment:
      if ( median + 1 < end )
      {
        final double tmp = bboxMin[axis];
        bboxMin[axis] = balance[index].pos[axis];
        balanceSegment( 2*index + 1, median + 1, end, organiz, balance );
        bboxMin[axis] = tmp;
      }
      else
      {
        balance[ 2*index+1 ] = organiz[end];
      }
    }
  }

  /** Swap two photons (a, b) in the list. */
  protected void swap ( Photon[] ph, int a, int b )
  {
    Photon ph2 = ph[a];
    ph[a]      = ph[b];
    ph[b]      = ph2;
  }

  //==========================================================================
  //  Input methods:

  /** Add single photon. */
  public void addPhoton ( double[] origin, double[] dir, float[] power )
  {
    if ( storedPhotons >= maxPhotons )
      return;

    if ( !isInit )
      prepare();

    storedPhotons++;
    photon[ storedPhotons ] = new Photon();

      // set properties of the photon:
    for ( int i = 0; i < 3; i++ )
    {
      photon[ storedPhotons ].pos[i] = origin[i];

      if ( photon[ storedPhotons ].pos[i] < bboxMin[i] )
        bboxMin[i] = photon[ storedPhotons ].pos[i];
      if ( photon[ storedPhotons ].pos[i] > bboxMax[i] )
        bboxMax[i] = photon[ storedPhotons ].pos[i];

      photon[ storedPhotons ].power[i] = power[i];
    }

      // compute theta:
    int theta = (int)( Math.acos( dir[2] ) * ( 256.0/Math.PI ) );
    if ( theta > 255 )
      photon[ storedPhotons ].theta = (char)255;
    else
      photon[ storedPhotons ].theta = (char)theta;

      // compute phi:
    int phi = (int)( Math.atan2( dir[1], dir[0] ) * ( 256.0/( 2.0*Math.PI ) ) );
    if ( phi > 255 )
      photon[ storedPhotons ].phi = (char)255;
    else if ( phi < 0 )
      photon[ storedPhotons ].phi = (char)(phi+256);
    else
      photon[ storedPhotons ].phi = (char)phi;
  }

  /** Add single ray. */
  public void addRays ( double[] origin, double[] dir )
  {
  }

  /** Add single pencil. */
  public void addPencil ( double[] center, double[][] dir, float[] radiance,
                          double distance )
  {
  }

  /** Add single contribution of energy to the some element. */
  public void addRadiosity ( int handle, float[] amount )
  {
  }

  //==========================================================================
  //  Illumination methods:

  /**
   * Computes irradiance at some point.
   *
   * @param position  Position of point in which is irradiance computing.
   * @param direction
   * @param rad       Pre-allocated array to hold resulting irradiance.
   * @return Resulting irradiance.
   */
  public double[] radiance ( double[] position, double[] dir, double[] rad )
  {
    return null;
  }

  /**
   * Computes irradiance at some point.
   *
   * @param position Position of point in which is irradiance computing.
   * @param irrad    Pre-allocated array to hold resulting irradiance.
   * @return Resulting irradiance.
   */
  public double[] irradiance ( double[] position, double[] normal, double[] irrad )
  {
    if ( position == null ||
         normal   == null )
     return null;

    if ( !isBalanced )
      balance();

    irrad[0] = irrad[1] = irrad[2] = 0.0;

    NearestPhotons np = new NearestPhotons( position, normal, radius, estimate );
    if ( type == Type.KD_TREE )
      locatePhotons( 1, np );
    else
      locatePhotons( np );

    if ( np.found < 10 )
      return irrad;

      // sum irradiance from all photons:
    double weight = 1.0;
    double error  = 1.0;
    double[] pdir = new double[ 3 ];
    for ( int i = 1; i <= np.found; i++ )
    {
      Photon p = np.index[ i ];

        // check direction of the photon:
      photonDir( p, pdir );
      error = pdir[0] * normal[0] +
              pdir[1] * normal[1] +
              pdir[2] * normal[2];
      if ( error >= 0.0 ) continue;
      if ( volume == Volume.DISC )
      {
        Geometry.vector( np.pos, p.pos, pdir );
        error = pdir[0] * normal[0] +
                pdir[1] * normal[1] +
                pdir[2] * normal[2];
        if ( error >= dirError ) continue;
      }

        // apply filter:
      if ( filter != Filter.NONE )
      {
        if ( volume == Volume.SPHERE )
          Geometry.vector( p.pos, np.pos, pdir );
        double distance2 = pdir[0] * pdir[0] +
                           pdir[1] * pdir[1] +
                           pdir[2] * pdir[2];

          // compute weight by given filter:
        if ( filter == Filter.CONE )
          weight = 1.0 - ( Math.sqrt(distance2) / ( k * radius ) );
        else if ( filter == Filter.GAUSSIAN )
        {
          double exponent = distance2 / (2*radius*radius);
          weight = ALPHA * ( 1.0 - ( ( 1.0 - Math.exp(-BETA * exponent) ) /
                                     ( 1.0 - Math.exp(-BETA) ) ) );
        }
        else if ( filter == Filter.SIMPSONS_KERNEL )
        {
          double s = ( 1 - distance2 / (radius*radius) );
          weight = 3.0 * s * s;
        }
      }
      irrad[0] += p.power[0] * weight;
      irrad[1] += p.power[1] * weight;
      irrad[2] += p.power[2] * weight;
    }

      // estimate density:
    float tmp = 1.0f;
    if ( filter == Filter.CONE )
      tmp = (float)( 1.0f / ( ( Math.PI * np.dist2[0] ) * (1.0 - 2.0/(3.0*k)) ) );
    else
      tmp = (float)( 1.0f / ( Math.PI * np.dist2[0] ) );

    irrad[0] *= tmp;
    irrad[1] *= tmp;
    irrad[2] *= tmp;

    if ( logEstimate )
      LogFile.log( "photons: irrad=" + Formula.vectorToString( irrad ) +
                   ", found=" + np.found );

    return irrad;
  }

  /** Computes the radiosity of the single element. */
  public float[] radiosity ( int handle )
  {
    return null;
  }

  //==========================================================================
  //  Getting method:

  /** Returns number of saved photons. */
  public int getNumberPhotons ()
  {
    return storedPhotons;
  }

  /** Returns number of saved rays. */
  public int getNumberRays ()
  {
    return 0;
  }

  /** Returns number of saved pencils. */
  public int getNumberPencils ()
  {
    return 0;
  }

  /** Returns number of saved elements. */
  public int getNumberElements ()
  {
    return 0;
  }

  /** Is light map full? */
  public boolean isFull ()
  {
    return maxPhotons == storedPhotons;
  }

  //==========================================================================
  //  Interface Property:

  public final static String VOLUME       = "Volume";
  public final static String TYPE         = "Search type";
  public final static String FILTER       = "Filter";
  public final static String MAX_PHOTONS  = "Max photons";
  public final static String ESTIMATE     = "Estimate";
  public final static String RADIUS       = "Radius";
  public final static String DIR_ERROR    = "Direction error";
  public final static String K            = "K";
  public final static String LOG_ESTIMATE = "Log estimate";
  public final static String LOG_PHOTONS  = "Log photons";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // VOLUME:
    if ( key.compareTo(VOLUME) == 0 )
    {
      volume = enumProperty( value, volume );
      return;
    }
      // TYPE:
    if ( key.compareTo(TYPE) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
      // FILTER:
    if ( key.compareTo(FILTER) == 0 )
    {
      filter = enumProperty( value, filter );
      return;
    }
      // MAX_PHOTONS:
    if ( key.compareTo(MAX_PHOTONS) == 0 )
    {
      maxPhotons = intProperty( value, maxPhotons );
      return;
    }
      // ESTIMATE:
    if ( key.compareTo(ESTIMATE) == 0 )
    {
      estimate = intProperty( value, estimate );
      return;
    }
      // RADIUS:
    if ( key.compareTo(RADIUS) == 0 )
    {
      radius = doubleProperty( value, radius  );
      return;
    }
      // DIR_ERROR:
    if ( key.compareTo(DIR_ERROR) == 0 )
    {
      dirError = doubleProperty( value, dirError );
      return;
    }
      // K:
    if ( key.compareTo(K) == 0 )
    {
      k = doubleProperty( value, k );
      return;
    }
      // LOG_ESTIMATE:
    if ( key.compareTo(LOG_ESTIMATE) == 0 )
    {
      logEstimate = booleanProperty( value, logEstimate );
      return;
    }
      // LOG_PHOTONS:
    if ( key.compareTo(LOG_PHOTONS) == 0 )
    {
      logPhotons = booleanProperty( value, logPhotons );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(VOLUME) == 0 )
      return volume;
    if ( key.compareTo(TYPE) == 0 )
      return type;
    if ( key.compareTo(FILTER) == 0 )
      return filter;
    if ( key.compareTo(MAX_PHOTONS) == 0 )
      return maxPhotons;
    if ( key.compareTo(ESTIMATE) == 0 )
      return estimate;
    if ( key.compareTo(RADIUS) == 0 )
      return radius;
    if ( key.compareTo(DIR_ERROR) == 0 )
      return dirError;
    if ( key.compareTo(K) == 0 )
      return k;
    if ( key.compareTo(LOG_ESTIMATE) == 0 )
      return logEstimate;
    if ( key.compareTo(LOG_PHOTONS) == 0 )
      return logPhotons;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Photon map";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "LightMap";

  /** Object description. */
  private final static String DESCRIPTION = "Photon map for photon storing";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_DATA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT,  IFACE+"LightMap" );
        // Property:
      t.propBegin( VOLUME, TYPE_INTEGER, "Volumes used for finding photons.", true );
        t.propDefault( Volume.SPHERE.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Disc",   Volume.DISC.ordinal(),   "More precise than sphere especially in the corners and curved surfaces." );
        t.propEnum( "Sphere", Volume.SPHERE.ordinal(), "Simple sphere." );
      t.propEnd();
      t.propBegin( TYPE, TYPE_INTEGER, "Types of searching of nearest photons.", true );
        t.propDefault( Type.KD_TREE.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Array",   Type.ARRAY.ordinal(),   "Each photon is tested -> O(n)." );
        t.propEnum( "KD-tree", Type.KD_TREE.ordinal(), "KD-tree is used -> O(log n)." );
      t.propEnd();
      t.propBegin( FILTER, TYPE_INTEGER, "Supported filters used in a radiance/irradiance estimation.", true );
        t.propDefault( Filter.NONE.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Cone",      Filter.CONE.ordinal(),            "Cone filter" );
        t.propEnum( "Gaussian",  Filter.GAUSSIAN.ordinal(),        "Gaussian filter" );
        t.propEnum( "None",      Filter.NONE.ordinal(),            "Without filtering." );
        t.propEnum( "Simpson's", Filter.SIMPSONS_KERNEL.ordinal(), "Simpson's kernel" );
      t.propEnd();
      t.propBegin( MAX_PHOTONS, TYPE_INTEGER, "Maximal number of stored photons.", true );
        t.propDefault( 100000 );
      t.propEnd();
      t.propBegin( ESTIMATE, TYPE_INTEGER, "Maximal number of photons which can be used for radiosity estimate.", true );
        t.propDefault( 500 );
      t.propEnd();
      t.propBegin( RADIUS, TYPE_DOUBLE, "Maximal radius of the sphere/circle (maximal distance between estimate point and a photon)", true        );
        t.propDefault( 100.0 );
      t.propEnd();
      t.propBegin( DIR_ERROR, TYPE_DOUBLE, "Maximal error of normal and direction of photon is radiance/irradiance estimate.", true );
        t.propDefault( 0.1 );
        t.propBounds( 0.0, 1.0 );
      t.propEnd();
      t.propBegin( K, TYPE_DOUBLE, "Filter constant characterizing the cone filter.", true );
        t.propDefault( 1.1 );
        t.propBounds( 1.0, Double.MAX_VALUE );
      t.propEnd();
      t.propBegin( LOG_ESTIMATE, TYPE_BOOLEAN, "Log information about estimates?", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( LOG_PHOTONS, TYPE_BOOLEAN, "Log all photons stored in photon map?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}