package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.piece.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Illuminates scene by tracing photons through the scene and saving them into
 * the map.
 *
 * @author Tomas Svab
 * @version 0.26 $Rev: 860 $ $Date: 2009-02-26 23:25:37 +0100 (čt, 26 II 2009) $ $Author: Tomas $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/PhotonTracing.java">PhotonTracing.java</a>
 * @todo    Add importance of photon.
 * @todo    Check if it's empty space in the map (if not, end tracing).
 */
public class PhotonTracing extends Piece implements LightTracing
{
  //==========================================================================
  //  Data:

  /** Accelerator: scene. */
  protected RTScene scene;

  /** Accelerator: intersection between scene and ray. */
  protected Intersectable inter;

  /** Light map for photon storing. */
  protected LightMap global;

  /** Caustic map for cautic photon storing. */
  protected LightMap caustic;

  /** Shadow map for shadow photon storing. */
  protected LightMap shadow;

  /** Volume map for volume photon storing. */
  protected LightMap volume;

  /** Random number generator. */
  protected final RandomJames rnd = new RandomJames();

  /** Mask used in {@link MicroFacet#assertAttributes}. */
  protected int dataMask;

  /** Use direct illumination? */
  protected boolean useDirect = true;

  /** Use illumination illumination? */
  protected boolean useIndirect = true;

  /** Use caustics? */
  protected boolean useCaustics = true;

  /** Use shadow photons? */
  protected boolean useShadow = false;

  /**
   * Should be photons saved into the photon map when they are still
   * in light source? In the case of direct visualization may be
   * useful to set value on the {@code true}, otherwise lights ourselfs
   * won't be visible.
   */
  protected boolean savePhotonsInLights = false;

  /** Is the photon map full? */
  protected boolean isMapFull = false;

  /** Was the tracer initialized? */
  protected boolean isInit = false;

  //==========================================================================
  //  Interface LightTracing:

  /**
   * Illuminate scene (transfer light from the light sources to the scene).
   */
  public void illuminateScene ()
  {
    if ( !isInit )
      if ( !init() )
        return;

    LightSource[] ls = scene.getLights();
    double[] temp    = new double[ 3 ];
    float[] power    = new float[ 3 ];
    double[] pos     = new double[ 3 ];
    double[] dir     = new double[ 3 ];

    int photon = 0;
    double pdf = 0.0;
    boolean doneGlobal = false;
    boolean doneCaustic  = false;
    while ( !doneGlobal || !doneCaustic )
    {
        // pick light:
      int i = 0;

      photon++;

        // generate ray:
      pdf = ls[i].sample( rnd.randomDouble( 0.0, 1.0 ),
                          rnd.randomDouble( 0.0, 1.0 ),
                          rnd.randomDouble( 0.0, 1.0 ),
                          rnd.randomDouble( 0.0, 1.0 ),
                          pos,
                          dir,
                          temp );

        // trace photon:
      double invPdf = 1.0 / pdf;
      for ( int k = 0; k < temp.length; k++ )
        power[k] = (float)(temp[k] * invPdf);
      tracePhoton( 0, false, 1.0, power, pos, dir );

      if ( photon >= 10000 )
      {
          // check status (is it done?):
        if ( global == null || global.isFull() || ( !useDirect && !useIndirect ) )
          doneGlobal = true;
        if ( caustic  == null || !useCaustics || caustic.isFull() )
          doneCaustic  = true;
      }
    }
    if ( global != null )
      global.scalePhotonPower( (float)(1.0 / photon) );
    if ( caustic != null )
      caustic.scalePhotonPower(  (float)(1.0 / photon) );
  }

  //==========================================================================
  //  Support functions:

  /** Initializes everything important and builds photon map by photon tracing. */
  public boolean init ()
  {
    if ( !useDirect && !useIndirect && !useCaustics && !useShadow )
      return false;

    if ( scene == null )
      scene = (RTScene)getInterface( PL_OUTPUT, IFACE+"RTScene" );
    if ( inter == null )
      inter = (Intersectable)getInterface( PL_INTERSECTION, IFACE+"Intersectable" );
    if ( scene == null ||
         inter == null )
      return false;

    if ( global == null )
      global  = (LightMap)getInterface( PL_PHOTON, IFACE+"LightMap"  );
    if ( caustic  == null )
      caustic = (LightMap)getInterface( PL_CAUSTIC, IFACE+"LightMap" );
    if ( shadow   == null )
      shadow  = (LightMap)getInterface( PL_SHADOW, IFACE+"LightMap"  );
    if ( volume   == null )
      volume  = (LightMap)getInterface( PL_VOLUME, IFACE+"LightMap"  );

    dataMask |= MicroFacet.VALID_NORMAL |
                MicroFacet.VALID_BSSRDF |
                MicroFacet.VALID_COLOR;

    isInit = true;

    return true;
  }

  /**
   * Trace photon through the scene by recursive calling of this function.
   *
   * @param prevHitSpec True if photon hit specular material in previous intersection.
   */
  public void tracePhoton ( int level,
                            boolean prevHitSpec,
                            double importance,
                            final float[] power,
                            double[] pos,
                            double[] dir )
  {
    if ( power == null ||
         pos   == null ||
         dir   == null )
      return;

    MicroFacet mf = inter.intersection( pos, dir );
    if ( mf == null )
      return;

    if ( mf.assertAttributes( dataMask ) != 0 ) return;
    mf.assertAttributes( MicroFacet.VALID_COORD );

      // save photons:
    int type = mf.bssrdf.type();
    if ( type == BSSRDF.COMP_DIFFUSE ||
         type == BSSRDF.COMP_GLOSSY  ||
         type == BSSRDF.COMP_ALL )
    {
      if ( global != null && !global.isFull() )
      {
        if ( ( useDirect && level == 0 ) || ( useIndirect && level > 0 ) )
          global.addPhoton( mf.coord, dir, power );
      }
      if ( caustic != null && !caustic.isFull() && useCaustics && prevHitSpec )
      {
        caustic.addPhoton( mf.coord, dir, power );
        return;
      }

      prevHitSpec = false;
    }
    else if ( type == BSSRDF.COMP_SPECULAR        ||
              type == BSSRDF.COMP_SPEC_REFLECTION ||
              type == BSSRDF.COMP_SPEC_REFRACTION )
      prevHitSpec = true;

      // possibly end tracing:
    double[] brdf = new double[ power.length ];
    mf.bssrdf.getOrigColor( brdf );
    double brdfAvrg = 0.0;
    for ( int i = 0; i < brdf.length; i++ )
      brdfAvrg += brdf[i];
    brdfAvrg /= brdf.length;
    double newImportance = importance * brdfAvrg;
    if ( rnd.randomDouble( 0.0, 1.0 ) >= newImportance )
      return;

      // scatter ray:
    double[] out  = new double[ 3 ];
    dir[0] = -dir[0];
    dir[1] = -dir[1];
    dir[2] = -dir[2];
    double pdf = mf.bssrdf.scatter( mf,
                                    dir,
                                    rnd.randomDouble( 0.0, 1.0 ),
                                    rnd.randomDouble( 0.0, 1.0 ),
                                    BSSRDF.COMP_ALL,
                                    out,
                                    brdf );
    if ( pdf == 0.0 )
      return;

      // compute new power:
    double invPdf = 1.0 / pdf;
    float[] powerNew = Formula.cloneArray( power );
    double cos = Math.abs( Geometry.dotProduct3D(out, mf.normal) );
    for ( int j = 0; j < powerNew.length; j++ )
      powerNew[j] *= brdf[j] * cos * invPdf;

      // possibly end tracing:
    if ( powerNew[0] == 0.0 &&
         powerNew[1] == 0.0 &&
         powerNew[2] == 0.0 )
      return;

      // trace photon:
    if ( useIndirect || useCaustics )
      tracePhoton( prevHitSpec ? level : ++level, prevHitSpec, newImportance, powerNew, mf.coord, out );
  }

  //==========================================================================
  //  Support methods:

  /**
   * Trace shadow ray and save shadow photon.
   */
  protected void traceShadowPhoton ( double[] pos, double[] dir, float[] power )
  {
      // TODO: probably it will have to be used some different function
      // for intersection, this computes only first intersection
      // (and tracing is slow; moreover doesn't works)
    MicroFacet mf = inter.intersection( pos, dir );
    if ( mf == null )
      return;

    if ( mf.assertAttributes( dataMask ) != 0 ) return;
    mf.assertAttributes( MicroFacet.VALID_COORD );

    shadow.addPhoton( mf.coord, dir, power );

    traceShadowPhoton( mf.coord, dir, power );
  }

  //==========================================================================
  //  Interface Property:

  public final static String SAMPLING     = "Sampling type";
  public final static String DIRECT       = "Direct";
  public final static String INDIRECT     = "Indirect";
  public final static String CAUSTICS     = "Caustics";
  public final static String SHADOW       = "Shadow photons";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // DIRECT:
    if ( key.compareTo(DIRECT) == 0 )
    {
      useDirect = booleanProperty( value, useDirect );
      return;
    }
      // INDIRECT:
    if ( key.compareTo(INDIRECT) == 0 )
    {
      useIndirect = booleanProperty( value, useIndirect );
      return;
    }
      // CAUSTICS:
    if ( key.compareTo(CAUSTICS) == 0 )
    {
      useCaustics = booleanProperty( value, useCaustics );
      return;
    }
      // CAUSTICS:
    if ( key.compareTo(SHADOW) == 0 )
    {
      useShadow = booleanProperty( value, useShadow );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(DIRECT) == 0 )
      return useDirect;
    if ( key.compareTo(INDIRECT) == 0 )
      return useIndirect;
    if ( key.compareTo(CAUSTICS) == 0 )
      return useCaustics;
    if ( key.compareTo(SHADOW) == 0 )
      return useShadow;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Photon tracing";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "";

  /** Object description. */
  private final static String DESCRIPTION = "Photon tracing to compute illumination of scene";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_LIGHT;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,        IFACE+"LightTracing" );
      t.newOutputPlug(    PL_OUTPUT,       IFACE+"RTScene" );
      t.newOutputPlug(    PL_INTERSECTION, IFACE+"Intersectable" );
      t.newOptOutputPlug( PL_PHOTON,       IFACE+"LightMap" );
      t.newOptOutputPlug( PL_CAUSTIC,      IFACE+"LightMap" );
      t.newOptOutputPlug( PL_SHADOW,       IFACE+"LightMap" );
      t.newOptOutputPlug( PL_VOLUME,       IFACE+"LightMap" );
        // Property:
      t.propBegin( DIRECT,   TYPE_BOOLEAN, "Use direct illumination?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( INDIRECT, TYPE_BOOLEAN, "Use indirect illumination?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( CAUSTICS, TYPE_BOOLEAN, "Compute caustics?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( SHADOW,   TYPE_BOOLEAN, "Trace and save shadow photons?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}