package cz.cuni.jagrlib.testing;

import java.awt.Point;
import java.awt.geom.Point2D;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Antialiased thick polyline template.
 *
 * @version 0.25 $Rev: 385 $ $Date: 2007-01-07 01:47:21 +0100 (ne, 07 I 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/PolylineAntialias.java">PolylineAntialias.java</a>
 */
public class PolylineAntialias extends Piece implements PolygonStrokeRender
{
  //==========================================================================
  //  Data:

  /**
   * Line-join style.
   *
   * @see    Render#LINE_JOIN
   * @see    Render#LINE_JOIN_MITER
   * @see    Render#LINE_JOIN_ROUND
   * @see    Render#LINE_JOIN_BEVEL
   */
  protected int joinStyle = LINE_JOIN_ROUND;

  /**
   * Actual miter-limit in pixels.
   * Special value {@code 0.0} converts {@link Render#LINE_JOIN_MITER} to {@link Render#LINE_JOIN_BEVEL}.
   *
   * @see    PolygonStrokeRender#LINE_MITER_LIMIT
   */
  protected double miterLimit = 2.0;

  /**
   * Line width in pixels.
   *
   * @see    Render#LINE_WIDTH
   */
  protected double width = 1.0;

  //==========================================================================
  //  Interface PolygonStrokeRender:

  /**
   * Sets default line width in pixels.
   *
   * @param  wid New line width in pixels.
   * @return The old line width in pixels.
   */
  public double setWidth ( double wid )
  {
    double old = width;
    width      = wid;
    return old;
  }

  /**
   * Sets default miter limit in pixels.
   *
   * @param  miterLimit New miter limit in pixels.
   * @return The old miter limit in pixels.
   */
  public double setMiterLimit ( double miterLimit )
  {
    double old      = this.miterLimit;
    this.miterLimit = miterLimit;
    return old;
  }

  /**
   * Sets default join style.
   *
   * @param  joinStyle New join style.
   * @return The old join style.
   * @see    Render#LINE_JOIN_MITER
   * @see    Render#LINE_JOIN_ROUND
   * @see    Render#LINE_JOIN_BEVEL
   */
  public int setJoinStyle ( int joinStyle )
  {
    int old        = this.joinStyle;
    this.joinStyle = joinStyle;
    return old;
  }

  /**
   * Strokes the given polyline.
   *
   * @param  v Array of polygon vertices (can be closed,
   *           {@code null} can interrupt the sequence).
   */
  public void strokePolyline ( Point[] v )
  {
    if ( v    == null || v.length < 2 ||
         v[0] == null || v[1] == null ) return;

    AlphaMask out = (AlphaMask)getInterface( PL_OUTPUT, IFACE+"AlphaMask" );
    int i, j, dMajor;
    double minor, dMinor;
    double radius = (width - 1.0) * 0.5;
    double center = Math.min( width, 1.0 );
    double edge   = (radius > 0.0) ? radius - Math.floor( radius ) : 0.0;

    for ( i = 0; i < v.length - 1 && v[i+1] != null;
          i++ )                           // one line segment
    {
      int x1 = v[i].x;
      int y1 = v[i].y;
      int x2 = v[i+1].x;
      int y2 = v[i+1].y;
      int dx = x2 - x1;
      int dy = y2 - y1;
      if ( Math.abs( dx ) >= Math.abs( dy ) )
      {                                   // X is the main axis
        dMajor = (x2 > x1) ? 1 : -1;
        minor  = y1 + 0.5;
        dMinor = (dx != 0) ? ((double)dy) / Math.abs( dx ) : 0.0;
        do
        {                                 // draw one line stop
          y1 = (int)minor;
          out.putPixel( x1, y1, center );

          for ( j = 1; j <= radius; j++ )
          {
            out.putPixel( x1, y1-j, 1.0 );
            out.putPixel( x1, y1+j, 1.0 );
          }
          if ( edge > 0.0 )
          {
            out.putPixel( x1, y1-j, edge );
            out.putPixel( x1, y1+j, edge );
          }
          if ( x1 == x2 ) break;

          x1    += dMajor;
          minor += dMinor;
        } while ( x1 != x2 );
      }

      else
      {                                   // Y is the main axis
        dMajor = (y2 > y1) ? 1 : -1;
        minor  = x1 + 0.5;
        dMinor = (dy != 0) ? ((double)dx) / Math.abs( dy ) : 0.0;
        do
        {                                 // draw one line stop
          x1 = (int)minor;
          out.putPixel( x1, y1, center);

          for ( j = 1; j <= radius; j++ )
          {
            out.putPixel( x1-j, y1, 1.0 );
            out.putPixel( x1+j, y1, 1.0 );
          }
          if ( edge > 0.0 )
          {
            out.putPixel( x1-j, y1, edge );
            out.putPixel( x1+j, y1, edge );
          }
          if ( y1 == y2 ) break;

          y1    += dMajor;
          minor += dMinor;
        } while ( y1 != y2 );
      }
    }
  }

  /**
   * Strokes the given polyline.
   *
   * @param  v Array of polygon vertices (can be closed,
   *           {@code null} can interrupt the sequence).
   */
  public void strokePolyline ( Point2D.Double[] v )
  {
    int len;
    if ( v == null ||
         (len = v.length) == 0 ) return;

    int i;
    for ( i = 0; i < len && v[i] != null; i++ ) ;
    len = i;
    Point[] vi = new Point[ len ];

    for ( i = 0; i < len; i++ )
      vi[i] = new Point( Formula.round( v[i].x ), Formula.round( v[i].y ) );

    strokePolyline( vi );
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // LINE_JOIN:
    if ( key.compareTo(LINE_JOIN) == 0 )
    {
      joinStyle = intProperty( value, joinStyle, LINE_JOIN_MITER, LINE_JOIN_BEVEL );
      return;
    }
      // LINE_WIDTH:
    if ( key.compareTo(LINE_WIDTH) == 0 )
    {
      width = doubleProperty( value, width, 0.01, 100.0 );
      return;
    }
      // LINE_MITER_LIMIT:
    if ( key.compareTo(LINE_MITER_LIMIT) == 0 )
    {
      miterLimit = doubleProperty( value, miterLimit, 0.0, 1000.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(LINE_JOIN) == 0 )
      return joinStyle;
    if ( key.compareTo(LINE_WIDTH) == 0 )
      return width;
    if ( key.compareTo(LINE_MITER_LIMIT) == 0 )
      return miterLimit;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "PolylineAntialias";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "PolygonStrokeRenderToAlphaMask";

  /** Object description. */
  private final static String DESCRIPTION = "Draws anti-aliased poly-line.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DRAW + "." + C_LINE + "." + C_INTEGER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"PolygonStrokeRender" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"AlphaMask" );
        // Property:
      t.propBegin( LINE_JOIN, TYPE_INTEGER, "Line-join style", true );
        t.propDefault( LINE_JOIN_ROUND );
        t.propBounds(  LINE_JOIN_MITER, LINE_JOIN_BEVEL );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Miter", LINE_JOIN_MITER, "Sharp (miter) join style" );
        t.propEnum( "Round", LINE_JOIN_ROUND, "Round loin style" );
        t.propEnum( "Bevel", LINE_JOIN_BEVEL, "Bevel (cap) join style, cap height = 0" );
      t.propEnd();
      t.propBegin( LINE_MITER_LIMIT, TYPE_DOUBLE, "Line miter limit (reduces potentially infinite sharp spits)", true );
        t.propDefault( 2.0 );
        t.propBounds(  0.0, 1000.0 );
      t.propEnd();
      t.propBegin( LINE_WIDTH, TYPE_DOUBLE, "Line width", true );
        t.propDefault( 1.0 );
        t.propBounds(  0.01, 100.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
