package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.piece.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Ray tracing of JGL scenes.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 858 $ $Date: 2009-02-17 23:48:39 +0100 (út, 17 II 2009) $ $Author: Tomas $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/RayTracingNew.java">RayTracingNew.java</a>
 * @todo    Hash-value for adaptive jittering.
 */
public class RayTracingNew extends RayCastingNew
{
  //==========================================================================
  //  Interface ImageFunction:

  /**
   * Recursive shading function - computes color contribution of the given ray (shot from the
   * origin {@code p0} into direction vector {@code p1}). Recursion is stopped
   * by a hybrid method: {@code importance} and {@code level} are checked.
   *
   * @param  level Actual recursion depth.
   * @param  importance Importance of the actual ray.
   * @param  p0 Ray origin.
   * @param  p1 Direction vector of the ray.
   * @param  color Pre-allocated array to hold result color vector ({@code double[1]}
   *               for monochrome images, {@code double[3]} for RGB images,
   *               larger arrays can be used for more complex spectral sampling).
   * @return Hash-value (ray sub-signature) used for adaptive subsampling.
   * @see    #minImportance
   * @see    #maxLevel
   * @see    #scene
   * @see    #inter
   * @todo   Some better way of computing importance.
   */
  @Override
  protected long shade ( int level, double importance, double[] p0, double[] p1,
                         double[] color )
  {
      // get the intersection:
    MicroFacet mf = inter.intersection( p0, p1 );
    if ( mf == null )                     // no intersection => use background color
    {
      System.arraycopy( background, 0, color, 0, Math.min( color.length, background.length ) );
      return 1L;
    }

      // compute and log important attributes of the intersection point:
    if ( mf.assertAttributes( dataMask ) != 0 ) return 0L;
    if ( mf.assertAttributes( MicroFacet.VALID_COORD ) == 0 && LogFile.tracing )
    {
      LogFile.log( String.format( Locale.US, "Intersection%d at t = %.4f with '%s' (%s,%s), p1.N: %.3f, mask: %04x",
                                  level, mf.t, mf.geometry.getClass().getName(),
                                  mf.enter ? "enter" : "leave", mf.front ? "front" : "back",
                                  Geometry.dotProduct3D( p1, mf.normal ), dataMask ) );

      LogFile.log( String.format( Locale.US, "  A: [ %.4f, %.4f, %.4f ], p1: [ %.4f, %.4f, %.4f ], N: [ %.4f, %.4f, %.4f ]",
                                  mf.coord[0], mf.coord[1], mf.coord[2],
                                  p1[0], p1[1], p1[2],
                                  mf.normal[0], mf.normal[1], mf.normal[2] ) );
    }

      // hash code:
    long hash = mf.sceneNode.hashCode();  // init: hash code of the intersected solid

      // apply textures:
    if ( useTextures &&
         mf.textures != null )
      for ( Texture tex : mf.textures )
        hash = hash * HASH_TEXTURE + tex.compute( mf );

    Arrays.fill( color, 0.0 );
    p1[0] = -p1[0];
    p1[1] = -p1[1];
    p1[2] = -p1[2];
    double[] brdf = new double[ color.length ];
    mf.bssrdf.getOrigColor( brdf );

    directIllumination( mf, p1, color );
    if ( mf.bssrdf.type() == BSSRDF.COMP_SPECULAR )
      specular( level, mf, p1, color );

    return hash;
  }

  //==========================================================================
  //  Interface Property:

  public final static String REFLECTIONS = "Reflections";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // SHADOW_RAYS:
    if ( key.compareTo(SHADOW_RAYS) == 0 )
    {
      numShadowRays = intProperty( value, numShadowRays );
      return;
    }
      // SAMPLING:
    if ( key.compareTo(SAMPLING) == 0 )
    {
      sampling = enumProperty( value, sampling );
      return;
    }
      // REFLECTIONS:
    if ( key.compareTo(REFLECTIONS) == 0 )
    {
      reflections = enumProperty( value, reflections );
      return;
    }
      // OCCLUSION:
    if ( key.compareTo(OCCLUSION) == 0 )
    {
      occlusion = booleanProperty( value, occlusion );
      return;
    }
      // TEXTURES:
    if ( key.compareTo(TEXTURES) == 0 )
    {
      useTextures = booleanProperty( value, useTextures );
      return;
    }
      // LOG:
    if ( key.compareTo(LOG) == 0 )
    {
      LogFile.tracing = booleanProperty( value, LogFile.tracing );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(SHADOW_RAYS) == 0 )
      return numShadowRays;
    if ( key.compareTo(SAMPLING) == 0 )
      return sampling.ordinal();
    if ( key.compareTo(REFLECTIONS) == 0 )
      return reflections.ordinal();
    if ( key.compareTo(TEXTURES) == 0 )
      return useTextures;
    if ( key.compareTo(LOG) == 0 )
      return LogFile.tracing;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Ray tracing new";

  /** Object name. */
  private final static String TEMPLATE_NAME = "Ray tracing";

  /** Object description. */
  private final static String DESCRIPTION = "Ray-tracing of JGL scenes";

  /** Object category. */
  protected final static String CATEGORY = C_3D + "." + C_RENDER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,        IFACE+"ImageFunction" );
      t.newOutputPlug(    PL_OUTPUT,       IFACE+"RTScene" );
      t.newOutputPlug(    PL_INTERSECTION, IFACE+"Intersectable" );
        // Property:
      t.propBegin( SAMPLING, TYPE_INTEGER, "Sampling", true );
        t.propDefault( Sampling.SAMPLING_BOTH.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Sampling both",   Sampling.SAMPLING_BOTH.ordinal(),   "Sample according to lights as well as according to BSDF." );
        t.propEnum( "Sampling BSDF",   Sampling.SAMPLING_BSDF.ordinal(),   "Sample according to BSDF only (problem with delta lights!)." );
        t.propEnum( "Sampling lights", Sampling.SAMPLING_LIGHTS.ordinal(), "Sample according to light only." );
      t.propEnd();
      t.propBegin( REFLECTIONS, TYPE_INTEGER, "Reflections", true );
        t.propDefault( Reflections.ALL.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "All",      Reflections.ALL.ordinal(),      "All types of materials are used." );
        t.propEnum( "Diffuse",  Reflections.DIFFUSE.ordinal(),  "Only diffuse materials are used." );
        t.propEnum( "Glossy",   Reflections.GLOSSY.ordinal(),   "Only glossy materials are used." );
        t.propEnum( "Specular", Reflections.SPECULAR.ordinal(), "Only specular materials are used." );
      t.propEnd();
      t.propBegin( SHADOW_RAYS, TYPE_INTEGER, "Number of shadow rays for light integrating hemisphere", true );
        t.propDefault( 10 );
      t.propEnd();
      t.propBegin( OCCLUSION,   TYPE_BOOLEAN, "Occlusion?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( TEXTURES,    TYPE_BOOLEAN, "Compute textures?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( LOG,         TYPE_BOOLEAN, "Log the tracing?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}