package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * 3D rendering example - wire-frame rendering of {@link Brep} database.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/RenderWireframe.java">RenderWireframe.java</a>
 */
public class RenderWireframe extends DefaultRender3D
{
  //==========================================================================
  //  Data:

  /** Canvas background. */
  protected final static int[] BACKGROUND = {   0,  20,  60, 255 };

  /** Foreground color. */
  protected final static int[] FOREGROUND = { 255, 240, 200, 255 };

  /** Draw normal vectors (for debugging purposes)? */
  protected boolean normals = false;

  /** Draw vertices only? */
  protected boolean verticesOnly = false;

  //==========================================================================
  //  Interface Render3D:

  /**
   * 3D rendering using the given projection matrix.
   *
   * @param mv Local model-view matrix.
   * @param proj Local projection matrix.
   * @param style Local rendering style.
   * @see #render()
   */
  @Override
  public void render ( TrMatrix mv, TrMatrix proj, int style )
  {
    final Brep           data   = (Brep)          getInterface( PL_DATA,   IFACE+"Brep" );
    final RasterGraphics output = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );

    if ( data   == null ||
         output == null ) return;

      // projection:
    setModelView( mv );
    setProjection( proj );
    checkMatrix();

    int width  = output.getWidth();
    int height = output.getHeight();
    output.setRectangle( 0, 0, width, height, BACKGROUND );
    output.setColor( FOREGROUND );

    int ctx      = Brep.CTX_DEFAULT;
    int worldId  = data.getAttributeId( Brep.VERTEX, ctx, Brep.COORD );
    int projId   = data.transformCoordinates( worldId, ctx, Brep.COORD_PROJ, compound );
    int colorId  = data.getAttributeId( Brep.FACE, ctx, Brep.COLOR );
    int normalId = data.getAttributeId( Brep.VERTEX, ctx, Brep.NORMAL );

    BrepIterator it = data.faceIterator( ctx, null );
    int faceId;
    int[] vert  = new int[ 3 ];
    int faceSize;
    int i, i1;
    double[][] vc  = new double[ 2 ][ 4 ];
    float[] colorF = new float[ 4 ];
    double[] color = new double[ 3 ];
    double[] norm  = new double[ 4 ];
    double[] world = new double[ 4 ];
    int faces      = 0;

    if ( !verticesOnly )                  // wire-frame mode
      while ( (faceId = it.next()) != Brep.NULL &&
              !userBreak )
      {
        faces++;
        while ( (faceSize = data.getFaceVertices( faceId, vert )) > vert.length )
          vert = new int[ faceSize ];

        data.getAttribute( colorId, faceId, colorF );
        color[0] = colorF[0];
        color[1] = colorF[1];
        color[2] = colorF[2];
        output.setColor( color );

        data.getAttribute( projId, vert[0], vc[1] );
        Geometry.toCartesian3D( vc[1], vc[1] );

        for ( i = 0; i < faceSize; i++ )
        {
          if ( (i1 = i + 1) >= faceSize ) i1 = 0;

          data.getAttribute( projId, vert[i1], vc[i & 1] );
          Geometry.toCartesian3D( vc[i & 1], vc[i & 1] );

          if ( normals )
          {
            data.getAttribute( worldId,  vert[i1], world );
            data.getAttribute( normalId, vert[i1], norm );
            norm[0] += world[0];
            norm[1] += world[1];
            norm[2] += world[2];
            norm[3] += world[3];
            compound.transformPoint3D( norm, norm );
            Geometry.toCartesian3D( norm, norm );
            DefaultRasterGraphics.drawLine( output, vc[i & 1][0], vc[i & 1][1], norm[0], norm[1] );
          }

          DefaultRasterGraphics.drawLine( output, vc[0][0], vc[0][1], vc[1][0], vc[1][1] );
        }
      }

    if ( verticesOnly ||                  // vertex mode
         faces == 0 && !userBreak )       // if there were no faces, draw vertices instead..
    {
      it = data.vertexIterator( it );
      while ( (faceId = it.next()) != Brep.NULL &&
              !userBreak )
      {
        faces++;
        data.getAttribute( projId, faceId, vc[0] );
        Geometry.toCartesian3D( vc[0], vc[0] );
        data.getAttribute( worldId, faceId, world );
        Formula.colorRamp( Formula.remainder( world[2] / 255.0, 1.0 ), color );
        output.setColor( color );
        output.putPixel( Formula.round( vc[0][0] ), Formula.round( vc[0][1] ), color );
      }
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String NORMALS  = "Draw normals";
  public final static String VERTICES = "Vertices only";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // RENDER_STYLE:
    if ( key.compareTo(RENDER_STYLE) == 0 )
    {
      renderStyle = intProperty( value, renderStyle, RENDER_STYLE_LINE, RENDER_STYLE_PHONG );
      return;
    }
      // NORMALS:
    if ( key.compareTo(NORMALS) == 0 )
    {
      normals = booleanProperty( value, normals );
      return;
    }
      // VERTICES:
    if ( key.compareTo(VERTICES) == 0 )
    {
      verticesOnly = booleanProperty( value, verticesOnly );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(RENDER_STYLE) == 0 )
      return renderStyle;
    if ( key.compareTo(NORMALS) == 0 )
      return normals;
    if ( key.compareTo(VERTICES) == 0 )
      return verticesOnly;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Render wireframe";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "Render3DToBrepAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Wireframe B-rep rendering.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + '.' + C_BREP;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"Render3D" );
      t.newInputPlug(  PL_DIRECT, IFACE+"Trigger" );
      t.newOutputPlug( PL_DATA,   IFACE+"Brep" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( RENDER_STYLE, TYPE_INTEGER, "Render style", true );
        t.propDefault( RENDER_STYLE_LINE );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Line drawing",    RENDER_STYLE_LINE,    "" );
        t.propEnum( "Flat shading",    RENDER_STYLE_FLAT,    "" );
        t.propEnum( "Gouraud shading", RENDER_STYLE_GOURAUD, "" );
        t.propEnum( "Phong shading",   RENDER_STYLE_PHONG,   "" );
      t.propEnd();
      t.propBegin( NORMALS, TYPE_BOOLEAN, "Draw normal vectors", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( VERTICES, TYPE_BOOLEAN, "Draw vertices only, ignore edges & faces", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
