package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Error-distribution dithering along external space-filling-curve.
 * Scales input raster image, gamma-corrects it, converts it to {@link RasterGraphics#MODE_GRAY} and
 * uses only {@code 0.0} and {@code 1.0} values.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/SFCDither.java">SFCDither.java</a>
 */
public class SFCDither extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Target image width in pixels ({@code 0} for original size). */
  protected int width        = 0;

  /** Target image height in pixels ({@code 0} for original size). */
  protected int height       = 0;

  /** Scale factor ({@code 0.0} or {@code 1.0} for original size). */
  protected double scale     = 1.0;

  /** Simulated gamma-exponent. */
  protected double gamma     = 2.0;

  /** Threshold-perturbation interval ({@code 0.0} for no perturbation). */
  protected double perturb   = 0.0;

  /** Contribution to the closest neighbour. */
  protected double neighbour = 1.0;

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transform.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    final RasterGraphics input  = (RasterGraphics)getInterface( "image1", IFACE+"RasterGraphics" );
    final RasterGraphics output = (RasterGraphics)getInterface( "image2", IFACE+"RasterGraphics" );
    final Order2D        order  = (Order2D)       getInterface( PL_ORDER, IFACE+"Order2D" );

    if ( output == null ||
         order  == null ) return false;

    int i, j;

      // output image:
    int iWid = (input == null) ? 1000 : input.getWidth();
    int iHei = (input == null) ?  250 : input.getHeight();
    int oWid = width;
    int oHei = height;

    if ( oWid == 0 )
      if ( Geometry.isZero( scale ) ||
           Geometry.isZero( scale - 1.0 ) )
        oWid = iWid;
      else
        oWid = Formula.round( iWid * scale );
    if ( oWid < 2 ) oWid = 2;

    if ( oHei == 0 )
      if ( Geometry.isZero( scale ) ||
           Geometry.isZero( scale - 1.0 ) )
        oHei = iHei;
      else
        oHei = Formula.round( iHei * scale );
    if ( oHei < 2 ) oHei = 2;

    output.init( oWid, oHei, RasterGraphics.MODE_GRAY, 0 );
    order.setSize( oWid, oHei );

    final int[] c = new int[ 2 ];         // actual coordinates

      // support variables for input image reading:
    final double dGray = 1.0 / (oWid - 1.0);
    final double kx = (iWid - 1.0) / (oWid - 1.0); // src_x = kx * target_x
    final double ky = (iHei - 1.0) / (oHei - 1.0); // src_y = ky * target_y
    double x, y;
    double pix00, pix10, pix01, pix11;
    int    iy, ix;
    double fx, fy, fx1, fy1;

      // error distribution:
    double pixel, nPixel, nnPixel;
    nPixel = nnPixel = 0.0;
    RandomJames rnd = (perturb > 0.0) ? new RandomJames() : null;
    final double thrMin = 0.5 - perturb * 0.5;
    double threshold = 0.5;

    while ( order.isNext() && !userBreak )
    {
      order.next( c );                    // next pixel to process: [ c[0], c[1] ]
      pixel  = nPixel;
      nPixel = nnPixel;

      if ( input == null )                // void input => horizontal vignette
        pixel += 1.0 - c[0] * dGray;

      else
      if ( iWid == oWid &&
           iHei == oHei )                 // dithering (no scaling)
        pixel += Math.pow( input.getGrayDouble( c[0], c[1] ), gamma );

      else                                // halftoning (scaling)
      {
        x   = c[0] * kx;                  // input-image coordinate
        fx  = Math.floor( x );
        ix  = (int)fx;                    // x = ix + fx
        fx  = x - fx;
        fx1 = 1.0 - fx;

        y   = c[1] * ky;                  // input-image coordinate
        fy  = Math.floor( y );
        iy  = (int)fy;                    // y = iy + fy
        fy  = y - fy;
        fy1 = 1.0 - fy;

          // four surrounding pixels:
        pix00 = Math.pow( input.getGrayDouble( ix,     iy     ), gamma );
        pix10 = Math.pow( input.getGrayDouble( ix + 1, iy     ), gamma );
        pix01 = Math.pow( input.getGrayDouble( ix,     iy + 1 ), gamma );
        pix11 = Math.pow( input.getGrayDouble( ix + 1, iy + 1 ), gamma );

          // bilinear interpolation:
        pixel += fx1 * ( fy1 * pix00 + fy * pix01 ) +
                 fx  * ( fy1 * pix10 + fy * pix11 );
      }

        // use white-noise?
      if ( rnd != null )
        threshold = thrMin + perturb * rnd.uniformNumber();

      x = (pixel > threshold) ? 1.0 : 0.0;
      output.putPixel( c[0], c[1], x );

        // error distribution:
      y       = pixel - x;                // total error to distribute
      nPixel += neighbour * y;
      nnPixel = (1.0 - neighbour) * y;
    }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String WIDTH     = "Width";
  public final static String HEIGHT    = "Height";
  public final static String SCALE     = "Scale";
  public final static String GAMMA     = "Gamma";
  public final static String PERTURB   = "Threshold perturbation";
  public final static String NEIGHBOUR = "Neighbour";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // WIDTH:
    if ( key.compareTo(WIDTH) == 0 )
    {
      width = intProperty( value, width, 0, 1000000 );
      return;
    }
      // HEIGHT:
    if ( key.compareTo(HEIGHT) == 0 )
    {
      height = intProperty( value, height, 0, 1000000 );
      return;
    }
      // SCALE:
    if ( key.compareTo(SCALE) == 0 )
    {
      scale = doubleProperty( value, scale, 0.0, 1000.0 );
      return;
    }
      // GAMMA:
    if ( key.compareTo(GAMMA) == 0 )
    {
      gamma = doubleProperty( value, gamma, 0.01, 100.0 );
      return;
    }
      // PERTURB:
    if ( key.compareTo(PERTURB) == 0 )
    {
      perturb = doubleProperty( value, perturb, 0.0, 1.0 );
      return;
    }
      // NEIGHBOUR:
    if ( key.compareTo(NEIGHBOUR) == 0 )
    {
      neighbour = doubleProperty( value, neighbour, 0.0, 1.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(WIDTH) == 0 )
      return width;
    if ( key.compareTo(HEIGHT) == 0 )
      return height;
    if ( key.compareTo(SCALE) == 0 )
      return scale;
    if ( key.compareTo(GAMMA) == 0 )
      return gamma;
    if ( key.compareTo(PERTURB) == 0 )
      return perturb;
    if ( key.compareTo(NEIGHBOUR) == 0 )
      return neighbour;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "SFC-dither";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToOrder2DAndRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Error-distribution along attached space-filling-curve.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_RASTER + '.' + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT, IFACE+"Trigger" );
      t.newOptOutputPlug( "image1", IFACE+"RasterGraphics" );
      t.newOutputPlug(    "image2", IFACE+"RasterGraphics" );
      t.newOutputPlug(    PL_ORDER, IFACE+"Order2D" );
        // Property:
      t.propBegin( WIDTH, TYPE_INTEGER, "Output width", true );
        t.propDefault( 0 );
        t.propBounds(  0, 1000000 );
      t.propEnd();
      t.propBegin( HEIGHT, TYPE_INTEGER, "Output height", true );
        t.propDefault( 0 );
        t.propBounds(  0, 1000000 );
      t.propEnd();
      t.propBegin( SCALE, TYPE_DOUBLE, "Image scale factor", true );
        t.propDefault( 1.0 );
        t.propBounds(  0.0, 1000.0 );
      t.propEnd();
      t.propBegin( GAMMA, TYPE_DOUBLE, "Simulated gamma-exponent", true );
        t.propDefault( 2.0 );
        t.propBounds(  0.01, 100.0 );
      t.propEnd();
      t.propBegin( PERTURB, TYPE_DOUBLE, "Threshold-perturbation interval", true );
        t.propDefault( 0.0 );
        t.propBounds(  0.0, 1.0 );
      t.propEnd();
      t.propBegin( NEIGHBOUR, TYPE_DOUBLE, "Contribution to the closest neighbour", true );
        t.propDefault( 1.0 );
        t.propBounds(  0.0, 1.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
