package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Solver for multi-level hierarchical radiosity method.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 747 $ $Date: 2008-05-17 15:21:58 +0200 (so, 17 V 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/SLEHierarchical.java">SLEHierarchical.java</a>
 */
public class SLEHierarchical extends Piece implements SLESolver
{
  //==========================================================================
  //  Support classes:

  /** Single node of the tree. */
  class Node
  {
    /** Recieved radiosity which wasn't shooted. */
    public float Bg;

    /** Total emiting radiosity. */
    public float Bs;

    /** Emission. */
    public float E;

    /** Size of area. */
    public float A;

    /** Reflectance. */
    public float r;

    /** List of children. */
    public Node[] children;

    /** First gathering link of node. */
    public Link link;

    /** Handle of the given node (for the access to the B-rep). */
    public int handle;

    /** Initializes all color-based variables of the node. */
    public void init ()
    {
        // get reflectance:
      float[] tempR = new float[ 3 ];
      brep.getAttribute( attrColor, handle, tempR );
      this.r = tempR[ colorMode ];

        // get emission:
      float[] tempE = new float[ 3 ];
      brep.getAttribute( attrEmission, handle, tempE );
      this.E = this.Bs = tempE[ colorMode ];
    }
  }

  /** Single link of the tree. */
  class Link
  {
    /** Shooting node (emmitor). */
    public Node p;

    /** Gathering node (reciever). */
    public Node q;

    /** Form factor from q to p. */
    public float ff;

    /** Next gathering link of node q. */
    public Link next;
  }

  //==========================================================================
  //  Data:

  /** Supported types of multi-level hierarchical radiosity. */
  public static enum AlgorithmTypes
  {
    STATIC,
    ADAPTIVE,
  }

  /** Actual type of multi-level hierarchical radiosity. */
  protected AlgorithmTypes type = AlgorithmTypes.STATIC;

  /** Accelerator: input B-rep database. */
  protected Brep brep;

  /** Accelerator: input radiosity basis functions. */
  protected RadiosityBasisFunctions basis;

  /** Actual {@link Brep} context. */
  protected int ctx = Brep.CTX_DEFAULT;

  /** Actual vertex-normal attribute. */
  protected int attrNormal   = brep.NULL;

  /** Actual face-color attribute. */
  protected int attrColor    = brep.NULL;

  /** Actual face-emmision attribute. */
  protected int attrEmission = brep.NULL;

  /** Required solution accuracy (max. norm of residual vector). */
  protected double epsilon = 1.e-6;

  /** Number of performed iteration steps. */
  protected int steps = 0;

  /** Number of unknowns. */
  protected int N = 0;

  /** List of nodes. */
  protected Node[] nodes;

  /** Maximal area of the element. */
  protected float maxA;

  /** Maximal size of the single form factor. */
  protected float maxF;

  /** Actual color mode (0 = red, 1 = green, 2 = blue). */
  protected int colorMode = 1;

  /** Was the solver already initialized? */
  protected boolean isInitialized = false;

  /** Was the color-based variables already initialized? */
  protected boolean isColorInit = false;

  //==========================================================================
  //  Interface SLESolver:

  /** Sets solver accuracy (whatever it means). */
  @Override
  public double setEpsilon ( double epsilon )
  {
    double old   = this.epsilon;
    this.epsilon = epsilon;
    return old;
  }

  /**
   * Solve the whole system en bloc.
   *
   * @param  save Solution will be saved into the associated {@link SLEData}.
   */
  @Override
  public void solve ( boolean save )
  {
    solve( Integer.MAX_VALUE, save );
  }

  /**
   * Reset the solver (if applicable).
   */
  @Override
  public void reset ()
  {
    if ( !isInitialized )
    {
      if ( brep == null )
        brep = (Brep)getInterface( PL_DATA, IFACE+"Brep" );
      if ( basis == null )
        basis = (RadiosityBasisFunctions)getInterface( PL_FUNCTION, IFACE+"RadiosityBasisFunctions" );
      if ( brep  == null ||
           basis == null )
        return;

      attrColor    = brep.getAttributeId( Brep.FACE,   ctx, Brep.COLOR );
      attrEmission = brep.getAttributeId( Brep.FACE,   ctx, "Emission" );
      attrNormal   = brep.getAttributeId( Brep.VERTEX, ctx, Brep.NORMAL );

      init();
      buildLinks();

      isInitialized = true;
    }

    steps = 0;
  }

  /**
   * Refine the solution performing several iterations.
   *
   * @param  maxIterations Number of iterations requested.
   * @param  save Solution will be saved into the associated {@link SLEData}.
   * @return Achieved precision (if less than epsilon, the system has converged).
   */
  @Override
  public double refine ( int maxIterations, boolean save )
  {
    return solve( maxIterations, save );
  }

  /**
   * Input data was changed, solver should adapt to it.
   * Usually {@link #solve} or {@link #refine} will be called afterwards..
   */
  @Override
  public void update ()
  {
    reset();
  }

  /** Reads a single unknown (element of solution vector). Non-mandatory. */
  @Override
  public double getSolution ( int col )
  {
    if ( col <  0 ||
         col >= N ) return 0.0;

    return nodes[col].Bs;
  }

  /** Reads the whole solution vector. Non-mandatory. */
  @Override
  public double[] getSolution ( double[] result )
  {
    if ( result == null ||
         result.length < N )
      result = new double[ N ];

    for ( int i = 0; i < N; i++ )
      result[i] = nodes[i].Bs;

    return result;
  }

  //==========================================================================
  //  Multi-component systems:

  /** Get actual component number. */
  @Override
  public int actualComponent ()
  {
    return colorMode;
  }

  /** Set actual component (non-mandatory). */
  @Override
  public void setComponent ( int c )
  {
    colorMode = c;

      // reinitialize color-based variables:
    for ( int i = 0; i < N; i++ )
      nodes[i].init();

    isColorInit = true;
  }

  //==========================================================================
  //  The solver itself:

  /**
   * Performs the given number of iteration steps.
   *
   * @param  maxIterations Maximal number of iterations.
   * @param  save Solution will be saved into the associated {@link SLEData}.
   * @return Residual norm achieved in the iteration.
   */
  protected double solve ( int maxIterations, boolean save )
  {
    if ( brep == null ||
         !isInitialized )
    {
      reset();
      if ( brep == null )
        return Double.POSITIVE_INFINITY;
    }

    if ( !isColorInit )
      setComponent( 1 );

    int p;
    int iter = maxIterations;
    double max;
    do
    {
      max = 0.0;

      for ( p = 0; p < N; p++ )
        gatherRad( nodes[p] );
      for ( p = 0; p < N; p++ )
        pushPullRad( nodes[p], 0.0f );

      for ( p = 0; p < N; p++ )
        if ( nodes[p].Bg > max )
          max = nodes[p].Bg;

      steps++;
    }
    while ( --iter > 0 && epsilon < max );

    return max;
  }

  //==========================================================================
  //  Preprocessing:

  /** Initialize nodes. */
  protected void init ()
  {
    N = brep.getAllFaces( ctx, null );
    nodes = new Node[ N ];
    for ( int i = 0; i < N; i++ )
    {
      nodes[i]        = new Node();
      nodes[i].handle = i;
      nodes[i].A      = (float)getWeight( i );
    }
  }

  /** Build initial set of links. */
  protected void buildLinks ()
  {
    for ( int i = 0; i < N; i++ )
      for ( int j = 0; j < N; j++ )
        refine( nodes[i], nodes[j] );
  }

  /** Establish links between p and q. */
  protected void refine ( Node p, Node q )
  {
    if ( oracle( p, q ) )
      link( p, q );
    else                                // linking these nodes isn't fine
    {
      Node node = divide( p, q );       // pick and subdivide one node
      if ( node == q )                  // q was subdivided
        for ( int child = 0; child < 4; child++ )
          refine( p, q.children[ child ] );
      else if ( node == p )             // p was subdivided
        for ( int child = 0; child < 4; child++ )
          refine( p.children[ child ], q );
      else
        link( p, q );                   // not subdividable at all
    }
  }

  /**
   * Returns a decision whether or not to link two nodes based
   * on the approximating error.
   */
  protected boolean oracle ( Node p, Node q )
  {
    if ( p.A < maxA &&
         q.A < maxA )
      return true;

    return ( estimateFormFactors( p, q ) < maxF );
  }

  /**
   * Builds the link between nodes a and b by computing the form factor
   * between the areas represented at this nodes.
   */
  protected void link ( Node p, Node q )
  {
    Link l = new Link();
    l.ff   = (float)basis.formFactor( q.handle, p.handle );
    l.p    = p;
    l.q    = q;
    l.next = q.link;
    q.link = l;
  }

  /**
   * Returns node a or b depending on for which it appears using lower level
   * nodes will reduce the error.
   */
  protected Node divide ( Node p, Node q )
  {
    if ( p.A > q.A )
      return subdivideNode( p );
    else
      return subdivideNode( q );
  }

  /** Subdivides single node to the three children. */
  protected Node subdivideNode ( Node a )
  {
    if ( a == null )
      return null;

      // computes new vertex (center of the face):
    double[][] vertex = new double[ 3 ][ 3 ];
    vertex = getVertices( a.handle, vertex );
    double[] center   = new double[ 3 ];
    for ( int j = 0; j < 3; j++ )
      center[j] = ( vertex[0][j] + vertex[1][j] + vertex[2][j] ) / 3.0;

      // gets handles of vertices:
    int[] handle = new int[ 4 ];
    brep.getFaceVertices( a.handle, handle );
    handle[3] = brep.createVertex();
    int[][] faceHandle =
    {
      { handle[0], handle[1], handle[3] },
      { handle[0], handle[2], handle[3] },
      { handle[1], handle[2], handle[3] },
    };

      // creates children:
    a.children = new Node[ 3 ];
    for ( int i = 0; i < 3; i++ )
    {
        // creates new faces:
      a.children[i] = new Node();
      int face = brep.createFace();
      brep.setFaceVertices( face, faceHandle[i] );

      a.children[i].handle = face;
      a.children[i].A      = (float)getWeight( face );
      a.children[i].init();
    }

    return a;
  }

  //==========================================================================
  //  Form factors estimate:

  /** Estimate unoccluded form factor. */
  protected double estimateFormFactors ( Node p, Node q )
  {
    int i;

      // gets vertices:
    double[][] vertexP = new double[ 3 ][ 3 ];
    double[][] vertexQ = new double[ 3 ][ 3 ];
    getVertices( p.handle, vertexP );
    getVertices( q.handle, vertexQ );

      // computes center of element p:
    double[] centerE = new double[ 3 ];
    for ( i = 0; i < centerE.length; i++ )
      centerE[i] = ( vertexP[ 0 ][i] +
                     vertexP[ 1 ][i] +
                     vertexP[ 2 ][i] ) / centerE.length;

      // computes center of sphere which is around the element q:
    double[] centerS = new double[ 3 ];
    double radius = 0.0;
    sphereFromPoints( vertexQ, centerS, radius );

      // gets normal:
    int[] vertexID = new int[ 3 ];
    brep.getFaceVertices( q.handle, vertexID );
    double[] normal = new double[ 4 ];
    brep.getAttribute( ctx, vertexID[0], normal );

      // computes direction:
    double[] direction = new double[ 3 ];
    Geometry.vector( centerE, centerS, direction );

      // computes omega:
    double A = Math.PI * radius * radius;
    double distance = lengthVector( direction );
    double omega = A / (distance * distance);

    Geometry.normalize3D( direction );
    return ( Geometry.dotProduct3D( normal, direction ) / Math.PI ) * omega;
  }

  /** Creates bounding sphere (center, radius) from the triangle points. */
  protected void sphereFromPoints ( double[][] point,
                                    double[] center,
                                    double radius )
  {
    if ( point  == null ||
         center == null )
      return;

    int i;

      // computes center of sphere:
    double[] min = new double[ 3 ];
    double[] max = new double[ 3 ];
    computeExtremes( point, min, max );
    for ( i = 0; i < center.length; i++ )
      center[i]  = 0.5 * ( max[i] + min[i] );

      // computes radius:
    double[] halfDiag = new double[ 3 ];
    for ( i = 0; i < 3; i++ )
      halfDiag[i] = 0.5 * ( max[i] - min[i] );
    radius = lengthVector( halfDiag );
  }

  /** Computes extremes of collection of vertices. */
  protected void computeExtremes ( double[][] point,
                                   double[] min,
                                   double[] max )
  {
    if ( point == null ||
         min   == null ||
         max   == null )
      return;

    System.arraycopy( point[0], 0, min, 0, point.length );
    System.arraycopy( point[0], 0, max, 0, point.length );
    double[] p = new double[ point[0].length ];
    for ( int i = 1; i < point.length; i++ )
    {
      p = Formula.cloneArray( point[i] );
      for ( int j = 0; j < point[0].length; j++ )
      {
        if ( p[j] < min[j] )
          min[j] = p[j];
        else if ( p[j] > max[j] )
          max[j] = p[j];
      }
    }
  }

  /** Length of given vector. */
  protected double lengthVector ( double[] vector )
  {
    return Math.sqrt( vector[0] * vector[0] +
                      vector[1] * vector[1] +
                      vector[2] * vector[2] );
  }

  /** Return coords of all vertices of the given triangle. */
  protected double[][] getVertices ( int handle, double[][] vertex )
  {
    if ( vertex == null ||
         vertex.length != 3 ||
         vertex[0].length != 3 )
      vertex = new double[ 3 ][ 3 ];

    int[] vertexID = new int[ 3 ];
    brep.getFaceVertices( handle, vertexID );

    for ( int i = 0; i < 3; i++ )
      brep.getVertexCoords( ctx, vertexID[i], vertex[i] );

    return vertex;
  }

  //==========================================================================
  //  Solving:

  /** Gathers energy over each incoming link (converting Bs to Bg). */
  protected void gatherRad ( Node p )
  {
    Link link = p.link;

      // gather energy across all links:
    p.Bg = 0.0f;
    while ( link != null )
    {
      p.Bg += p.r * link.ff * link.p.Bs;
      link = link.next;
    }

    if ( p.children != null )
      for ( int child = 0; child < p.children.length; child++ )
        gatherRad( p.children[ child ] );
  }

  /**
   * Pushes the recieved energy Bg down to the children of each quadtree
   * node, and pulls the results back up the quadtrees by area averaging.
   *
   * @param  p      Node which pushes energy down to the children.
   * @param  BDown  Recieved energy by node a.
   * @return ?
   */
  protected float pushPullRad ( Node p, float BDown )
  {
    float BUp;

    if ( p.children == null )       // p is a leaf
      BUp = p.E + p.Bg + BDown;
    else
    {
      BUp = 0.0f;
      for ( int i = 0; i < p.children.length; i++ )
        BUp += ( p.children[i].A / p.A ) * pushPullRad( p.children[i], p.Bg + BDown );
    }
    p.Bs = BUp;

    return BUp;
  }

  //==========================================================================
  //  Support:

  /** Reads a single weight (size of element of solution vector). */
  protected double getWeight ( int i )
  {
    int[] vertexID = new int[ 3 ];
    brep.getFaceVertices( i, vertexID );

    int j;

    double[][] vertex = new double[ 3 ][ 3 ];
    for ( j = 0; j < 3; j++ )
      brep.getVertexCoords( ctx, vertexID[j], vertex[j] );

    double a = Math.sqrt( ( vertex[0][0] - vertex[1][0] ) * ( vertex[0][0] - vertex[1][0] ) +
                          ( vertex[0][1] - vertex[1][1] ) * ( vertex[0][1] - vertex[1][1] ) +
                          ( vertex[0][2] - vertex[1][2] ) * ( vertex[0][2] - vertex[1][2] ) );
    double b = Math.sqrt( ( vertex[0][0] - vertex[2][0] ) * ( vertex[0][0] - vertex[2][0] ) +
                          ( vertex[0][1] - vertex[2][1] ) * ( vertex[0][1] - vertex[2][1] ) +
                          ( vertex[0][2] - vertex[2][2] ) * ( vertex[0][2] - vertex[2][2] ) );
    double c = Math.sqrt( ( vertex[1][0] - vertex[2][0] ) * ( vertex[1][0] - vertex[2][0] ) +
                          ( vertex[1][1] - vertex[2][1] ) * ( vertex[1][1] - vertex[2][1] ) +
                          ( vertex[1][2] - vertex[2][2] ) * ( vertex[1][2] - vertex[2][2] ) );

    double s = ( a + b + c ) / 2;     // half perimeter

      // area of element by Heron's formula:
    double area = Math.sqrt( s * (s - a) * (s - b) * (s - c) );
    return area;
  }

  /** Return total unshoot radiosity in dependence on actual color mode. */
  protected double getTotalResidual ()
  {
    double totalResidual = 0.0;
    for ( int i = 0; i < N; i++ )
      totalResidual += nodes[i].Bg;
    return totalResidual;
  }

  //==========================================================================
  //  Interface Property:

  public final static String TYPE    = "Type";
  public final static String EPSILON = "Epsilon";
  public final static String MAX_A   = "Min A";
  public final static String MAX_F   = "Min F";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // TYPE:
    if ( key.compareTo(TYPE) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
      // EPSILON:
    if ( key.compareTo(EPSILON) == 0 )
    {
      epsilon = doubleProperty( value, epsilon );
      return;
    }
      // MIN_A:
    if ( key.compareTo(MAX_A) == 0 )
    {
      maxA = floatProperty( value, maxA  );
      return;
    }
      // MIN_F:
    if ( key.compareTo(MAX_F) == 0 )
    {
      maxF = floatProperty( value, maxF  );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(TYPE) == 0 )
      return type.ordinal();
    if ( key.compareTo(EPSILON) == 0 )
      return epsilon;
    if ( key.compareTo(MAX_A) == 0 )
      return maxA;
    if ( key.compareTo(MAX_F) == 0 )
      return maxF;
    if ( key.compareTo(STEPS) == 0 )
      return steps;
    if ( key.compareTo(TOTAL_RESIDUAL) == 0 )
      return getTotalResidual();

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "SLE-hierarchical";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "SLESolverToSLEData";

  /** Object description. */
  private final static String DESCRIPTION = "Solver for multi-level hierarchical radiosity";

  /** Object category. */
  protected final static String CATEGORY = "SLE.solver";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,    IFACE+"SLESolver" );
      t.newOutputPlug( PL_DATA,     IFACE+"Brep" );
      t.newOutputPlug( PL_FUNCTION, IFACE+"RadiosityBasisFunctions" );
        // Property:
      t.propBegin( TYPE, TYPE_INTEGER, "Type", true );
        t.propDefault( AlgorithmTypes.STATIC.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Static",   AlgorithmTypes.STATIC.ordinal(),   "Static" );
        t.propEnum( "Adaptive", AlgorithmTypes.ADAPTIVE.ordinal(), "Adaptive" );
      t.propEnd();
      t.propBegin( EPSILON, TYPE_DOUBLE, "Required solution accuracy (norm of residual vector)", true );
        t.propDefault( 1.e-6 );
      t.propEnd();
      t.propBegin( MAX_A, TYPE_FLOAT, "Maximal area of the element", true  );
        t.propDefault( 1.0 );
      t.propEnd();
      t.propBegin( MAX_F, TYPE_FLOAT, "Maximal size of the single form factor", true  );
        t.propDefault( 1.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
