package cz.cuni.jagrlib.testing;

import java.awt.Color;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Sample colormap generator.
 *
 * @version 0.26 $Rev: 439 $ $Date: 2007-05-05 12:29:50 +0200 (so, 05 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/SampleColormap.java">SampleColormap.java</a>
 */
public class SampleColormap extends Piece implements ColormapStore
{
  //==========================================================================
  //  Data:

  /** Implemented colormap types. */
  protected static enum ColormapType
  {
    /** Gray color ramp. */
    CMAP_GRAY,
    /** Tricolor. */
    CMAP_TRICOLOR,
  }

  /** Actual colormap type. */
  protected ColormapType type  = ColormapType.CMAP_TRICOLOR;

  /** Fictional parameter - Color1. */
  protected Color color1       = Color.RED;

  /** Colormap cache arrays. */
  protected double[][][] store = new double[ 2 ][][];

  //==========================================================================
  //  Support:

  protected void assertColormap ( int i )
  {
    if ( store[i] != null ) return;

    int j;
    switch ( i )
    {
      case 1:                             // variable type
        switch ( type )
        {
          case CMAP_TRICOLOR:             // tricolor
            store[1] = new double[ 192 ][ 3 ];
            final double COEF = 1.0 / 63.0;
            for ( j = 0; j < 64; j++ )
            {
              store[1][j][0] = j * COEF;
              store[1][j][1] =
              store[1][j][2] = 0.0;
            }
            for ( ; j < 128; j++ )
            {
              store[1][j][0] =
              store[1][j][1] = 0.0;
              store[1][j][2] = (j - 64) * COEF;
            }
            for ( ; j < 192; j++ )
            {
              store[1][j][0] =
              store[1][j][1] =
              store[1][j][2] = (j - 128) * COEF;
            }
            return;

          default:                        // gray color ramp
            store[1] = new double[ 256 ][ 3 ];
            for ( j = 0; j < 256; j++ )
              store[1][j][0] =
              store[1][j][1] =
              store[1][j][2] = j / 255.0;
            return;
        }

      default:                            // cold-warm color ramp
        store[0] = new double[ 256 ][ 3 ];
        for ( j = 0; j < 256; j++ )
          Formula.colorRamp( j / 255.0, store[0][j] );
    }
  }

  //==========================================================================
  //  Interface ColormapStore:

  /**
   * Store (re-)initialization.
   *
   * @param  _mode Colormap item value type ({@code RasterGraphics.MODE_*} constant).
   * @param  _bands Number of bands (for {@link RasterGraphics#MODE_BAND}).
   * @see    RasterGraphics
   * @see    #init()
   */
  public void init ( int _mode, int _bands )
  {
    store = new double[ 2 ][][];
  }

  /**
   * Store re-initialization (keeps current colormap mode).
   *
   * @see    #init(int,int)
   */
  public void init ()
  {
  }

  /**
   * Returns the store size (colormaps are indexed from {@code 0} to
   * {@link #size}{@code - 1}).
   *
   * @return Store size.
   * @see    #setSize
   */
  public int size ()
  {
    return 2;
  }

  /**
   * Assigns the new store size (potentially discards some colormaps).
   *
   * @param  newSize The new store size.
   * @return The old store size.
   * @see    #size
   */
  public int setSize ( int newSize )
  {
    return 2;
  }

  /**
   * Retrieves the i-th colormap.
   *
   * @param  i Colormap index (from {@code 0} to {@link #size}{@code - 1}).
   * @return Array holding the colormap data.
   * @see    #getColormap
   */
  public double[][] getColormapDouble ( int i )
  {
    if ( i < 0 || i >= 2 )
      return null;

    assertColormap( i );

    return store[i];
  }

  /**
   * Retrieves the i-th colormap.
   *
   * @param  i Colormap index (from {@code 0} to {@link #size}{@code - 1}).
   * @return Array holding the colormap data.
   * @see    #getColormapDouble
   */
  public int[][] getColormap ( int i )
  {
    if ( i < 0 || i >= 2 )
      return null;

    assertColormap( i );

    final double[][] src  = store[i];
    final int        len  = src.length;
    int[][]          copy = new int[ len ][ 3 ];
    int j, b;
    for ( j = 0; j < len; j++ )
      for ( b = 0; b < 3; b++ )
        copy[j][b] = Formula.round( src[j][b] * 255.0 );

    return copy;
  }

  /**
   * Assigns the i-th colormap.
   *
   * @param  i Colormap index.
   * @param  cmap The new colormap data (or {@code null}).
   * @return Array holding the old colormap data (or {@code null}).
   * @see    #setColormap(int,double[][])
   */
  public int[][] setColormap ( int i, int[][] cmap )
  {
    return null;
  }

  /**
   * Assigns the i-th colormap.
   *
   * @param  i Colormap index.
   * @param  cmap The new colormap data (or {@code null}).
   * @return Array holding the old colormap data (or {@code null}).
   * @see    #setColormap(int,int[][])
   */
  public double[][] setColormap ( int i, double[][] cmap )
  {
    return null;
  }

  //==========================================================================
  //  Interface Property:

  public final static String TYPE   = "Colormap type";
  public final static String COLOR1 = "Color1";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // TYPE:
    if ( key.compareTo(TYPE) == 0 )
    {
      type     = enumProperty( value, type );
      store[1] = null;                    // reset colormap cache
      return;
    }
      // COLOR1:
    if ( key.compareTo(COLOR1) == 0 )
    {
      color1   = colorProperty( value, color1 );
      store[1] = null;                    // reset colormap cache
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(TYPE) == 0 )
      return type.ordinal();
    if ( key.compareTo(COLOR1) == 0 )
      return color1;

    return null;
  }

//==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "SampleColormap";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ColormapStore";

  /** Object description. */
  private final static String DESCRIPTION = "Synthetic colormaps (sample module).";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DATA + "." + C_RASTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"ColormapStore" );
        // Property:
      t.propBegin( TYPE, TYPE_INTEGER, "Colormap[1] type", true );
        t.propDefault( ColormapType.CMAP_TRICOLOR.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Gray ramp", ColormapType.CMAP_GRAY.ordinal(),     "Gray ramp (256 items)" );
        t.propEnum( "Tricolor",  ColormapType.CMAP_TRICOLOR.ordinal(), "Tricolor (192 items)" );
      t.propEnd();
      t.propBegin( COLOR1, TYPE_OBJECT, "Color 1", true );
        t.propDefault( "[255;0;0]" );
        t.propManipulator( "ParamColor" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
