package cz.cuni.jagrlib.testing;

import java.util.Locale;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.piece.HDRImage;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Sample raster-search implementation.
 * Uses only luminance search, one pixel pisition accuracy and RMSE.
 *
 * @version 0.26 $Rev: 474 $ $Date: 2007-06-01 15:54:10 +0200 (pá, 01 VI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/SampleRasterSearch.java">SampleRasterSearch.java</a>
 */
public class SampleRasterSearch extends Piece implements RasterSearch
{
  //==========================================================================
  //  Data:

  /** Use only gray (luminance) component of the images? */
  protected boolean useGray = true;

  /** Actual orientation mask. */
  protected int actOriMask  = 1 << ORI_0123;

  /** Actual band mask. */
  protected int actBandMask = 1;

  /** Actual metrics. */
  protected int actMetrics  = ERR_RMSE;

  /** Actual position accuracy. */
  protected double actPrecision = 1.0;

  /** Actual minimum X coordinate of a search. */
  protected int actMinX = 0;

  /** Actual maximum X coordinate of a search (can be {@code 0} for unbounded search). */
  protected int actMaxX = 0;

  /** Actual minimum Y coordinate of a search. */
  protected int actMinY = 0;

  /** Actual maximum Y coordinate of a search (can be {@code 0} for unbounded search). */
  protected int actMaxY = 0;

  /** Actual maximum X displacement for relative search (should be {@code 0} for absolute search). */
  protected int actRelX = 0;

  /** Actual maximum Y displacement for relative search (can be {@code 0} for absolute search). */
  protected int actRelY = 0;

  /**
   * Actual minimum value of multiplication coefficient {@code s}.
   * Must be set to {@code 1.0} if no pixel transform is allowed).
   */
  protected double actMinS = 1.0;

  /**
   * Actual maximum value of multiplication coefficient {@code s}.
   * Must be set to {@code 1.0} if no pixel transform is allowed).
   */
  protected double actMaxS = 1.0;

  /** Actual minimum value of additive coefficient {@code o}. */
  protected double actMinO = 0.0;

  /** Actual maximum value of additive coefficient {@code o}. */
  protected double actMaxO = 0.0;

  /** Reference pictures. */
  protected HDRImage[] refs = new HDRImage[ MAX_REFERENCE ];

  //==========================================================================
  //  Interface Trigger:

  /**
   * Sets reference image.
   * Reference images have to be set before you can search in them..
   *
   * @param  g Read-only raster image (can be {@code null} to reset the slot).
   * @param  id Reference image identifier ({@code 0}, {@code 1}, .. {@link #MAX_REFERENCE}{@code -1}).
   * @param  scale Scale to be used in precomputing.
   * @param  oriMask Mask of orientations used in future search commands (non-mandatory,
   *         use {@code 0} if not known).
   * @param  pixelTransform Will a linear pixel transform be used?
   * @see    #ORI_0123
   */
  public void setReferencePicture ( RasterGraphics g, int id, double scale, int oriMask, boolean pixelTransform )
  {
    if ( g == null ) return;

    id = Formula.clamp( id, 0, MAX_REFERENCE-1 );
    refs[id] = new HDRImage( g, scale, useGray ? RasterGraphics.MODE_GRAY : g.getMode() );
  }

  /**
   * Sets parameters for subsequent searches.
   *
   * @param  oriMask Mask of valid orientations to search ({@code 1 << }{@link #ORI_0123}, ..).
   * @param  bandMask Mask of searched image bands (Red .. 1, Green .. 2, Blue .. 4, ..).
   * @param  metrics One of {{@link #ERR_MAE}, {@link #ERR_RMSE}}.
   * @param  precision Sub-pixel precision of result coordinates (should be power of {@code 1/2}).
   * @param  absMinX Lower bound for X-coordinate of absolute search.
   * @param  absMaxX Upper bound for X-coordinate of absolute search (can be {@code 0} for unbounded search).
   * @param  absMinY Lower bound for Y-coordinate of absolute search.
   * @param  absMaxY Upper bound for Y-coordinate of absolute search (can be {@code 0} for unbounded search).
   * @param  relX X-amplitude of relative search (must be {@code 0} for absolute search).
   * @param  relY Y-amplitude of relative search.
   * @param  minMul Minimum pixel-value multiplier (should be {@code 1} if no multiplication is allowed).
   * @param  maxMul Maximum pixel-value multiplier (should be {@code 1} if no multiplication is allowed).
   * @param  minAdd Minimum additive coefficient (should be {@code 0} if no addition is allowed).
   * @param  maxAdd Maximum additive coefficient (should be {@code 0} if no addition is allowed).
   */
  public void setSearchParameters ( int oriMask, int bandMask, int metrics, double precision,
                                    int absMinX, int absMaxX, int absMinY, int absMaxY,
                                    int relX, int relY,
                                    double minMul, double maxMul, double minAdd, double maxAdd )
  {
    actOriMask   = oriMask;
    actBandMask  = bandMask;
    actMetrics   = metrics;
    actPrecision = precision;
    actMinX      = absMinX;
    actMaxX      = absMaxX;
    actMinY      = absMinY;
    actMaxY      = absMaxY;
    actRelX      = relX;
    actRelY      = relY;
    actMinS      = minMul;
    actMaxS      = maxMul;
    actMinO      = minAdd;
    actMaxO      = maxAdd;
  }

  /**
   * Searches for the most similar rectangle in one of predefined reference images.
   *
   * @param  f Raster image containing the sought rectangle.
   * @param  x Origin of the sought rectangle.
   * @param  y Origin of the sought rectangle.
   * @param  width Width of the sought rectangle.
   * @param  height Height of the sought rectangle.
   * @param  ref Reference image to be scanned.
   * @param  result Pre-allocated result array {@code double[5]},
   *         will hold coordinates, coefficients and error measure {@code [x,y,mul,add,err]}.
   * @return Orientation of the best fit (or {@link #ORI_FAIL} in case of failure).
   */
  public int findBestRectangle ( RasterGraphics f, int x, int y, int width, int height,
                                 int ref, double[] result )
  {
    ref = Formula.clamp( ref, 0, MAX_REFERENCE-1 );
    final HDRImage r = refs[ref];
    if ( f             == null   ||
         r             == null   ||
         result        == null   ||
         result.length < 5       ||
         actMetrics    != ERR_RMSE )
      return ORI_FAIL;

    final int rWidth  = r.getWidth();
    final int rHeight = r.getHeight();

    if ( rWidth  < width  ||
         rHeight < height )
      return ORI_FAIL;

    int fWidth  = f.getWidth();
    int fHeight = f.getHeight();

    if ( x + width  <= 0      ||
         x          >= fWidth ||
         y + height <= 0      ||
         y          >= fHeight )
    {                                     // the rectangle lies outside of the picture
      result[0] = rWidth;
      result[1] = 0;
      result[2] = 1.0;
      result[3] =
      result[4] = 0.0;
      return ORI_0123;
    }

      // constant terms from f's rectangle:
    double sumB     = 0.0;
    double sumBB    = 0.0;
    double actA, actB;
    final double n  = width * height;
    final double nInv = 1.0 / n;
    int xi, yi;
    for ( yi = y; yi < y + height; yi++ )
      for ( xi = x; xi < x + width; xi++ )
      {
        sumB  += (actB = f.getGrayDouble( xi, yi ));
        sumBB += actB * actB;
      }

      // searching bounds:
    int    searchMinX = actMinX;
    int    searchMaxX = actMaxX;
    if ( searchMaxX == 0 ||
         searchMaxX > rWidth )
      searchMaxX = rWidth;
    int    searchMinY = actMinY;
    int    searchMaxY = actMaxY;
    if ( searchMaxY == 0 ||
         searchMaxY > rHeight - 1 )
      searchMaxY = rHeight - 1;

      // relative search?
    if ( actRelX > 0 )
    {
      if ( searchMinX < x - actRelX )
        searchMinX = x - actRelX;
      if ( searchMaxX > x + actRelX )
        searchMaxX = x + actRelX;
      if ( searchMinY < y - actRelY )
        searchMinY = y - actRelY;
      if ( searchMaxY > y + actRelY )
        searchMaxY = y + actRelY;
    }

      // extreme found so far:
    int    minX   = 0;
    int    minY   = 0;
    int    minOri = ORI_0123;
    double minS   = 1.0;
    double minO   = 0.0;                  // multiplied by n^2
    double minErr = Double.POSITIVE_INFINITY; // multiplied by n^2
    double maxErr = 0.0;

      // search loop variables:
    int    actX, actY;
    double sumA, sumAA, sumAB;
    double actS, actO, actErr, denom;

      // search loop:
    for ( actY = searchMinY; actY <= searchMaxY; actY++ )
      for ( actX = searchMinX; actX <= searchMaxX; actX++ )
      {                                   // test one reference-image position [ actX, actY ]
        sumA  =
        sumAA = 0.0;
          // orientation-invariant computation:
        for ( yi = 0; yi < height; yi++ )
          for ( xi = 0; xi < width; xi++ )
          {
              // reference image:
            actA   = r.getGrayDouble( actX + xi, actY + yi );

            sumA  += actA;
            sumAA += actA * actA;
          }

        sumAB = 0.0;
          // ORI_0123 orientation is fixed here:
        for ( yi = 0; yi < height; yi++ )
          for ( xi = 0; xi < width; xi++ )
          {
              // reference image:
            actA   = r.getGrayDouble( actX + xi, actY + yi );
              // searched rectangle:
            actB   = f.getGrayDouble( x    + xi, y    + yi );

            sumAB += actA * actB;
          }

          // compute actual s, o, err
          // According to Y. Fisher: Fractal Image Compression, SIGGRAPH '92 Course notes
        denom = n * sumAA - sumA * sumA;
        if ( Geometry.isZero( denom ) )
        {
          actS = Formula.clamp( 0.0, actMinS, actMaxS );
          actO = Formula.clamp( sumB * nInv, actMinO, actMaxO );
        }
        else
        {
          actS = Formula.clamp( (n * sumAB - sumA * sumB) / denom, actMinS, actMaxS );
          actO = Formula.clamp( nInv * (sumB - actS * sumA), actMinO, actMaxO );
        }
        actErr = sumBB + actS * (actS * sumAA - 2.0 * (sumAB - actO * sumA)) + actO * (actO * n - sumB - sumB);

        if ( actErr < minErr )            // better match was found..
        {
          minX   = actX;
          minY   = actY;
          minOri = ORI_0123;
          minS   = actS;
          minO   = actO;
          minErr = actErr;
        }
        if ( actErr > maxErr )
          maxErr = actErr;
      }

      // result:
    result[0] = minX;
    result[1] = minY;
    result[2] = minS;
    result[3] = minO;
    result[4] = minErr * nInv;
    //LogFile.debug( String.format( Locale.US, "Err: %.5f, max: %.5f", result[4], (maxErr * nInv) ) );

    return minOri;
  }

  //==========================================================================
  //  Interface Property:

  public final static String USE_GRAY = "Luminance only";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // USE_GRAY:
    if ( key.compareTo(USE_GRAY) == 0 )
    {
      useGray = booleanProperty( value, useGray );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(USE_GRAY) == 0 )
      return useGray;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Sample Raster-search";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "RasterSearch";

  /** Object description. */
  private final static String DESCRIPTION = "Sample implementation of RasterSearch interface.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + ".search";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"RasterSearch" );
        // Property:
      t.propBegin( USE_GRAY, TYPE_BOOLEAN, "Use only gray (luminance) component?", true );
        t.propDefault( true );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
