package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Solid specimen - elementary 3D solid.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/SolidSpecimen.java">SolidSpecimen.java</a>
 */
public class SolidSpecimen extends DefaultSolid
{
  //==========================================================================
  //  Data:

  /**
   * Temporary object to hold intersection data for {@link SolidSpecimen}.
   *
   * @version 0.25 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
   * @since   0.12
   */
  protected static class TmpData
  {
    // !!! TODO: this is only a specimen !!!

    /** Support for tangent vector.. */
    double[] dry;
  }

  //==========================================================================
  //  Interface Solid:

  /**
   * Computes intersection of the solid with the given ray.
   *
   * @param  P0 Ray origin ({@code double[3]} or {@code double[4]}).
   * @param  P1 Direction vector of the ray ({@code double[3]} or {@code double[4]}).
   * @return List of intersection points ({@link MicroFacet}s) sorted by {@code t}
   *         (parametric distance from {@code P0}). {@code null} is returned in case
   *         of bad arguments or if no intersection exists..
   * @see    MicroFacet
   * @see    BoundingVolume
   */
  public List< MicroFacet > intersection ( double[] P0, double[] P1 )
  {
    int P0len, P1len;
    if ( P0 == null || (P0len = P0.length) < 3 ||
         P1 == null || (P1len = P1.length) < 3 ) return null;

    double Ox = P0[0];                  // ray origin:
    double Oy = P0[1];
    double Oz = P0[2];
    if ( P0len > 3 )
    {
      double Ow = P0[3];
      if ( Ow < -Geometry.EPSILON || Ow > Geometry.EPSILON )
      {
        Ox /= Ow;
        Oy /= Ow;
        Oz /= Ow;
      }
    }

    double Dx = P1[0];                  // ray direction vector (of arbitrary size):
    double Dy = P1[1];
    double Dz = P1[2];
    if ( P1len > 3 )
    {
      double Dw = P1[3];
      if ( Dw < -Geometry.EPSILON && Dw > Geometry.EPSILON )
      {
        Dx /= Dw;
        Dy /= Dw;
        Dz /= Dw;
      }
    }

    LinkedList< MicroFacet > list = null;

    // !!! TODO: this is only a specimen !!!
    if ( false )
    {
      if ( list == null )
        list = new LinkedList< MicroFacet >();
        // specimen: insert an intersection
      MicroFacet mf  = new MicroFacet();
      mf.t           = 0.0;
      mf.enter       =                    // == front at this time
      mf.front       = true;              // ray hits front-side of the surface
      TmpData tmp    = new TmpData();
      mf.locCoord    = new double[]
      {
        Ox + mf.t * Dx,
        Oy + mf.t * Dy,
        Oz + mf.t * Dz
      };
      mf.geometryTmp = tmp;
      mf.valid       = MicroFacet.VALID_LOC_COORD;
      list.addLast( mf );
    }

    return list;
  }

  /**
   * Asserts validity of the given intersection-attributes.
   * Resolves: {@link MicroFacet#VALID_TANGENT}, {@link MicroFacet#VALID_TEX_COORD}.
   *
   * @param  mf The checked intersection.
   * @param  tmp Non-mandatory object holding state-information for deferred computations.
   * @param  attributes Attribute mask (required attributes - see
   *                    {@code MicroFacet.VALID_*}).
   * @return Attributes which were not computed.
   */
  @Override
  public int assertAttributes ( MicroFacet mf, Object tmp, int attributes )
  {
    if ( mf  == null ||
         tmp == null ) return attributes;

    TmpData td = (TmpData)tmp;

      // tangent surface vectors:
    if ( (attributes & MicroFacet.VALID_TANGENT) != 0 )
    {
      // !!! TODO: this is only a specimen !!!
      mf.valid |= MicroFacet.VALID_TANGENT;
      if ( (attributes -= MicroFacet.VALID_TANGENT) == 0 ) return 0;
    }

      // 2D texture coordinates (inverse mapping):
    if ( (attributes & MicroFacet.VALID_TEX_COORD) != 0 )
    {
      // !!! TODO: this is only a specimen !!!
      mf.valid |= MicroFacet.VALID_TEX_COORD;
      if ( (attributes -= MicroFacet.VALID_TEX_COORD) == 0 ) return 0;
    }

    return attributes;
  }

  /**
   * Point vs. solid test ("inside-test").
   *
   * @param  point Coordinates of the 3D point ({@code double[3]} or
   *               {@code double[4]}).
   * @return {@code true} if the given point lies inside the solid.
   */
  @Override
  public boolean isInside ( double[] point )
  {
    // this method is not mandatory, if not defined, delete it!
    return false;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "SolidSpecimen";

  /** Object description. */
  private final static String DESCRIPTION = "Solid specimen -  3D solid for CSG.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"Solid" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
