package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Represents perfect specular reflection.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/SpecularReflection.java">SpecularReflection.java</a>
 */
public class SpecularReflection extends LambertianModel
{
  //==========================================================================
  //  Data:

  /** Fresnel term. */
  public Fresnel fresnel;

  /** Default constructor. */
  public SpecularReflection ()
  {
    color = new double[] { 1.0, 1.0, 1.0 };
  }

  //==========================================================================
  //  Interface BSSRDF:

  /**
   * Returns the requested BRSSDF sample (BSSRDF is treated as monochrome).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @return BSSRDF coefficient <b>f(in-&gt;out)</b>.
   * @see    #monoBSSRDF(MicroFacet,double[],double[],int)
   * @see    #colorBSSRDF(double[],double[],double[],int,double[])
   */
  @Override
  public double monoBSSRDF ( boolean worldCoords,
                             double[] in,
                             double[] out,
                             double[] normal,
                             double[] tangent,
                             int mask )
  {
    if ( normal == null ||
         out    == null )
      return 0.0;

    if ( mask == BSSRDF.COMP_SPECULAR ||
         mask == BSSRDF.COMP_ALL )
      return color[0] * fresnel.evaluate( Geometry.dotProduct3D( out, normal ), null )[0];
    else
      return 0.0;
  }

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as band-color).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BSSRDF spectral coefficients.
   * @see    #monoBSSRDF(double[],double[],double[],int)
   * @see    #colorBSSRDF(MicroFacet,double[],double[],int,double[])
   */
  @Override
  public double[] colorBSSRDF ( boolean worldCoords,
                                double[] in,
                                double[] out,
                                double[] normal,
                                double[] tangent,
                                int mask,
                                double[] result )
  {
    if ( normal == null ||
         out    == null )
      return null;

    if ( mask == BSSRDF.COMP_SPECULAR ||
         mask == BSSRDF.COMP_ALL )
    {
      if ( result == null )
        result = new double[ color.length ];

      double[] fres = new double[ color.length ];
      fresnel.evaluate( Geometry.dotProduct3D( out, normal ), fres );
      for ( int i = 0; i < color.length; i++ )
        result[i] = color[i] * fres[i];
    }

    return result;
  }

  /**
   * Scatter ray according to the light model.
   *
   * @param  mf   The {@link MicroFacet} (point on surface of some solid).
   * @param  in   Input direction (should look from the hit point)..
   * @param  u    Random uniform number.
   * @param  v    Random uniform number.
   * @param  mask Which light components are to be computed?
   * @param  out  Output direction in world coordinates.
   * @param  brdf Color definition - {@code double[C]} array
   *              ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *              color system/ or greater /for more complex spectral color sampling/).
   * @return Probability density function.
   */
  @Override
  public double scatter ( MicroFacet mf,
                          double[] in,
                          double u,
                          double v,
                          int mask,
                          double[] out,
                          double[] brdf  )
  {
    if ( mf   == null ||
         in   == null ||
         brdf == null ||
         out  == null )
      return 0.0;

    if ( mask == BSSRDF.COMP_SPECULAR ||
         mask == BSSRDF.COMP_ALL )
    {
      Geometry.specularReflection( mf.normal, in, out );
      colorBSSRDF( false, mf, in, out, mask, brdf );
      return 1.0;
    }
    else
      return 0.0;
  }

  /**
   * Returns the probability density function.
   *
   * @param worldCoords Should be set to {@code true} if vectors in and out
   *                    are in world coordinates and to {@code false}
   *                    if vectors in and out are in local coordinates.
   * @param in  Input direction.
   * @param out Output direction.
   */
  @Override
  public double pdf ( boolean worldCoords, MicroFacet mf, double[] in, double[] out )
  {
    return 0.0;
  }

  /** Returns the type of BRDF. */
  @Override
  public int type ()
  {
    return BSSRDF.COMP_SPECULAR;
  }

  //==========================================================================
  //  Interface Property:

  public final static String FRESNEL = "Fresnel";

/**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

    if ( key.compareTo( BSSRDF.MATERIAL_COLOR ) == 0 &&
         value instanceof double[] )
    {
      color = (double[])value;
      return;
    }
    if ( key.compareTo( FRESNEL ) == 0 )
    {
      fresnel = (Fresnel)value;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo( BSSRDF.MATERIAL_COLOR ) == 0 ) return color;
    if ( key.compareTo( FRESNEL ) == 0 ) return fresnel;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Specular reflection";

  /** Object description. */
  private final static String DESCRIPTION = "Specular reflection";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"BSSRDF" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}