package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Represents perfect specular transmission.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 841 $ $Date: 2009-01-10 14:02:22 +0100 (so, 10 I 2009) $ $Author: Tomas $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/SpecularTransmission.java">SpecularTransmission.java</a>
 */
public class SpecularTransmission extends LambertianModel
{
  //==========================================================================
  //  Data:

  /** Fresnel term. */
  public Fresnel fresnel;

  /** Index of refraction for incoming media. */
  public double etaI;

  /** Index of refraction for transmitted media. */
  public double etaT;

  /** Default constructor. */
  public SpecularTransmission ()
  {
    color = new double[] { 1.0, 1.0, 1.0 };
    etaI = 1.0;
    etaT = 1.0;
  }

  //==========================================================================
  //  Interface BSSRDF:

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as monochrome).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @return BSSRDF coefficient <b>f(in-&gt;out)</b>.
   * @see    #monoBSSRDF(MicroFacet,double[],double[],int)
   * @see    #colorBSSRDF(double[],double[],double[],int,double[])
   */
  @Override
  public double monoBSSRDF ( boolean worldCoords,
                             double[] in,
                             double[] out,
                             double[] normal,
                             double[] tangent,
                             int mask )
  {
    if ( normal == null ||
         out    == null )
      return 0.0;

    if ( mask == BSSRDF.COMP_SPECULAR ||
         mask == BSSRDF.COMP_ALL )
    {
      return color[0] * fresnel.evaluate( Geometry.dotProduct3D( out, normal ), null )[0];
    }
    else
      return 0.0;
  }

  /**
   * Returns the requested BSSRDF sample (BSSRDF is treated as band-color).
   *
   * @param  worldCoords Should be set to {@code true} if vectors in and out
   *                     are in world coordinates and to {@code false}
   *                     if vectors in and out are in local coordinates.
   * @param  in Input ray direction ({@code double[3]} or {@code double[4]}).
   * @param  out Output (viewing) ray direction ({@code double[3]} or {@code double[4]}).
   * @param  normal Normal vector of a surface.
   * @param  tangent Tangent vector of a surface.
   * @param  mask Which light components are to be computed?
   * @param  result Pre-allocated {@code double[C]} array to hold BRDF coefficients
   *                <b>f(in-&gt;out)</b> ({@code C} can be either 1
   *                /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *                complex spectral color sampling/). Can be {@code null}.
   * @return Output BSSRDF spectral coefficients.
   * @see    #monoBSSRDF(double[],double[],double[],int)
   * @see    #colorBSSRDF(MicroFacet,double[],double[],int,double[])
   */
  @Override
  public double[] colorBSSRDF ( boolean worldCoords,
                                double[] in,
                                double[] out,
                                double[] normal,
                                double[] tangent,
                                int mask,
                                double[] result )
  {
    if ( normal == null ||
         out    == null )
      return null;

    if ( mask == BSSRDF.COMP_SPECULAR )
    {
      if ( worldCoords )
      {
        worldToLocal( normal, tangent, in, in );
        worldToLocal( normal, tangent, out, out );
      }

      if ( result == null )
        result = new double[ color.length ];

        // compute new direction:
      double eta  = etaI / etaT;
      double cosT = getCosT( etaI, etaT, in[2] );
      out[0] = -eta * in[0];
      out[1] = -eta * in[1];
      out[2] = cosT;

        // compute brdf:
      double[] fres = new double[ color.length ];
      fresnel.evaluate( in[2], fres );
      for ( int i = 0; i < result.length; i++ )
        result[i] = ( 1.0 - fres[i] ) * color[0] / ( Math.abs(cosT) * eta * eta );

      if ( worldCoords )
      {
        localToWorld( normal, tangent, in, in );
        localToWorld( normal, tangent, out, out );
      }
    }

    return result;
  }

  /**
   * Scatter ray according to the light model.
   *
   * @param  mf   The {@link MicroFacet} (point on surface of some solid).
   * @param  in   Input direction (should look from the hit point)..
   * @param  u    Random uniform number.
   * @param  v    Random uniform number.
   * @param  mask Which light components are to be computed?
   * @param  out  Output direction in world coordinates.
   * @param  brdf Color definition - {@code double[C]} array
   *              ({@code C} can be either 1 /for monochrome computing/, 3 /for RGB
   *              color system/ or greater /for more complex spectral color sampling/).
   * @return Probability density function.
   */
  @Override
  public double scatter ( MicroFacet mf,
                          double[] in,
                          double u,
                          double v,
                          int mask,
                          double[] out,
                          double[] brdf  )
  {
    if ( mf   == null ||
         in   == null ||
         brdf == null ||
         out  == null )
      return 0.0;

    if ( mask == BSSRDF.COMP_SPECULAR ||
         mask == BSSRDF.COMP_ALL )
    {
      double[] inLoc = new double[ 3 ];
      worldToLocal( mf.normal, mf.tu, in, inLoc );

        // determine etas:
      double ei = etaI;
      double et = etaT;
      if ( inLoc[2] < 0.0 )
      {
        ei = etaT;
        et = etaI;
      }

        // handle total internal reflection:
      double eta   = ei / et;
      double sinT2 = eta * eta * ( 1.0 - inLoc[2] * inLoc[2] );
      if ( sinT2 > 1.0 )
        return 0.0;

        // compute new direction:
      double cosT = Math.sqrt( Math.max( 0.0, 1.0 - sinT2 ) );
      if ( inLoc[2] > 0.0 )
        cosT = -cosT;
      out[0] = -eta * inLoc[0];
      out[1] = -eta * inLoc[1];
      out[2] = cosT;

        // compute brdf:
      double[] fres = new double[ color.length ];
      fresnel.evaluate( inLoc[2], fres );
      for ( int i = 0; i < brdf.length; i++ )
      {
        brdf[i] = ( 1.0 - fres[i] ) * color[i] / ( Math.abs(cosT) * eta * eta );
        brdf[i] = Formula.clamp( brdf[i], 0.0, 1.0 );
      }
      localToWorld( mf.normal, mf.tu, out, out );
      return 1.0;
    }
    else
      return 0.0;
  }

  /**
   * Returns the probability density function.
   *
   * @param worldCoords Should be set to {@code true} if vectors in and out
   *                    are in world coordinates and to {@code false}
   *                    if vectors in and out are in local coordinates.
   * @param in  Input direction.
   * @param out Output direction.
   */
  @Override
  public double pdf ( boolean worldCoords, MicroFacet mf, double[] in, double[] out )
  {
    return 0.0;
  }

  /** Returns the type of BRDF. */
  @Override
  public int type ()
  {
    return BSSRDF.COMP_SPECULAR;
  }

  //==========================================================================
  //  Support methods:

  /**
   * Computes cosinus of transmitted ray.
   *
   * @param ei Index of refraction for incoming media.
   * @param et Index of refraction for transmitted media.
   * @param z  Z coordinate of incoming direction (local coordinates!).
   */
  protected double getCosT ( double ei, double et, double z )
  {
    double eta   = ei / et;
    double eta2  = eta * eta;
    double sinI2 = 1.f - (z*z);
    double sinT2 = eta2 * sinI2;
    return Math.sqrt( Math.max( 0.0, 1.0 - sinT2 ) );
  }

  //==========================================================================
  //  Interface Property:

  public final static String FRESNEL = "Fresnel";
  public final static String ETA_I   = "EtaI";
  public final static String ETA_T   = "EtaT";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

    if ( key.compareTo( BSSRDF.MATERIAL_COLOR ) == 0 &&
         value instanceof double[] )
    {
      color = (double[])value;
      return;
    }
    if ( key.compareTo( FRESNEL ) == 0 )
    {
      fresnel = (Fresnel)value;
      return;
    }
    if ( key.compareTo( ETA_I ) == 0 )
    {
      etaI = doubleProperty( value, etaI );
      return;
    }
    if ( key.compareTo( ETA_T ) == 0 )
    {
      etaT = doubleProperty( value, etaT );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo( BSSRDF.MATERIAL_COLOR ) == 0 ) return color;
    if ( key.compareTo( FRESNEL ) == 0 ) return fresnel;
    if ( key.compareTo( ETA_I ) == 0 ) return etaI;
    if ( key.compareTo( ETA_T ) == 0 ) return etaT;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Specular transmission";

  /** Object description. */
  private final static String DESCRIPTION = "Specular transmission";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"BSSRDF" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}