package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Texture specimen - simple 3D spot-like texture (modifies base surface color).
 *
 * @version 0.25 $Rev: 280 $ $Date: 2006-06-01 19:35:42 +0200 (čt, 01 VI 2006) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/SpotTexture.java">SpotTexture.java</a>
 */
public class SpotTexture extends Piece implements Texture
{
  //==========================================================================
  //  Data:

  /** Texture frequency (uniform in all coordinate components). */
  protected double freq = 10.0;

  /** Secondary color. Primary color is inherited from COLOR material attribute, etc. */
  protected double[] color2 = { 0.2, 0.4, 1.0 };

  //==========================================================================
  //  Construction:

  public SpotTexture ()
  {
  }

  public SpotTexture ( double f, double[] color )
  {
    freq = f;
    if ( color != null )
      color2 = color;
  }

  //==========================================================================
  //  Interface Texture:

  /**
   * Computes the texture value in the given point (given by a {@link MicroFacet} record).
   *
   * @param  mf <tt>MicroFacet</tt> object to modify.
   * @return Hash-value (sample signature) used for adaptive subsampling.
   */
  public long compute ( MicroFacet mf )
  {
    if ( mf.assertAttributes( MicroFacet.VALID_COLOR |
                              MicroFacet.VALID_OBJ_COORD ) != 0 )
      return 0L;

    double x = mf.objCoord[0] * freq;
    double y = mf.objCoord[1] * freq;
    double z = mf.objCoord[2] * freq;

    RealFunction tex = (RealFunction)getInterface( PL_OUTPUT, IFACE+"RealFunction" );
    if ( tex == null ) return 0L;

    double t = tex.f( x, y, z );

    if ( mf.color.length < 3 )
      mf.color = new double[ 3 ];

      // blend the original color (mf.color) with my color (color2):
    t = 0.5 * (Formula.clamp( t, -1.0, 1.0 ) + 1.0);
    double ct = 1.0 - t;

    mf.color[0] = ct * mf.color[0] + t * color2[0];
    mf.color[1] = ct * mf.color[1] + t * color2[1];
    mf.color[2] = ct * mf.color[2] + t * color2[2];

    return( Math.round( x ) +
            RandomStatic.parkMiller( Math.round( y ) ) +
            RandomStatic.numericRecipes( Math.round( z ) ) );
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // FREQ:
    if ( key.compareTo(FREQ) == 0 )
    {
      freq = doubleProperty( value, freq );
      return;
    }
      // COLOR:
    if ( key.compareTo(COLOR) == 0 )
    {
      if ( value instanceof double[] )
      {
        double[] v = (double[])value;
        if ( color2 == null || color2.length < 3 )
          color2 = new double[ 3 ];
        System.arraycopy( v, 0, color2, 0, Math.min( color2.length, v.length ) );
      }
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(FREQ) == 0 )
      return freq;
    if ( key.compareTo(COLOR) == 0 )
      return color2;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "SpotTexture";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "Texture";

  /** Object description. */
  private final static String DESCRIPTION = "Texture specimen -  simple 3D spot-like texture (modifies base surface color).";

  /** Object category. */
  protected final static String CATEGORY = C_3D + ".texture";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"Texture" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"RealFunction" );
        // Property:
      t.propBegin( FREQ, TYPE_DOUBLE, "Frequency", true );
        t.propDefault( "10.0" );
      t.propEnd();
      t.propBegin( COLOR, TYPE_OBJECT, "Secondary color", false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
