package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.xml.JGLScript;

/**
 * Square light source shining non-uniformly (amount of power depends
 * on divergence of the photon direction) through the whole hemisphere which
 * lies on front side of the face.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 821 $ $Date: 2008-11-17 21:35:59 +0100 (po, 17 XI 2008) $ $Author: Tomas $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/StaticAreaLightSource.java">StaticAreaLightSource.java</a>
 */
public class StaticAreaLightSource extends DefaultLightSource
{
  //==========================================================================
  //  Data:

  {
    scriptName = "data/square.jgl";
  }

  /** Random number generator. */
  protected final RandomJames rnd = new RandomJames();

  /** Coordinates of up right vertex. */
  protected double[] vertexUR = new double[] { 0.0, 0.0, 0.0 };

  /** Coordinates of up left vertex. */
  protected double[] vertexUL = new double[] { 0.0, 0.0, 0.0 };

  /** Coordinates of low left vertex. */
  protected double[] vertexLL = new double[] { 0.0, 0.0, 0.0 };

  /** Coordinates of low right vertex. */
  protected double[] vertexLR = new double[] { 0.0, 0.0, 0.0 };

  /** Normal of the light. */
  protected double[] normal   = new double[] { 0.0, 0.0, 0.0 };

  /**
   * Color of the source: {@code 1D} for monochrome source, {@code 3D} for <b>RGB</b>
   * color space, or more for complex spectrum sampled spaces.
   */
  protected double[] lightColor = new double[] { 1.0 };

  /**
   * Dimming polynom coefficients: light is dimmed by the factor of
   * {@code dim[0] + dim[1] * D + dim[2] * D * D}.
   * Can be {@code null} for no dimming..
   */
  protected double[] dim = new double[] { 0.5, 1.0, 0.1 };

  /** TBN (tangent, binormal, normal) matrix. */
  protected TrMatrix m;

  //==========================================================================
  //  Support:

  /**
   * Set light source parameters.
   *
   * @param  power    Power of light.
   * @param  urVertex Up right vertex of square.
   * @param  ulVertex Up left vertex of square.
   * @param  llVertex Low left vertex of square.
   * @param  lrVertex Low right vertex of square.
   * @param  color Source color/intensity (can be either 1D /for monochrome computing/,
   *               3D /for RGB color system/ or greater /for more complex spectral color
   *               sampling/).
   * @todo Support of spectral color sampling.
   */
  public boolean setParams ( double[] urVertex,
                             double[] ulVertex,
                             double[] llVertex,
                             double[] lrVertex,
                             double[] color,
                             double[] dimming )
  {
    if ( urVertex == null || urVertex.length < 3 ||
         ulVertex == null || ulVertex.length < 3 ||
         llVertex == null || llVertex.length < 3 ||
         lrVertex == null || lrVertex.length < 3 )
      return false;

      // up right vertex:
    if ( urVertex == null )
      this.vertexUR = null;
    else
    {
      if ( this.vertexUR == null || this.vertexUR.length != 3 )
        this.vertexUR = new double[ 3 ];

      for ( int i = 0; i < 3 && i < urVertex.length; i++ )
        this.vertexUR[i] = urVertex[i];
    }

      // up left vertex:
    if ( ulVertex == null )
      this.vertexUL = null;
    else
    {
      if ( this.vertexUL == null || this.vertexUL.length != 3 )
        this.vertexUL = new double[ 3 ];

      for ( int i = 0; i < 3 && i < ulVertex.length; i++ )
        this.vertexUL[i] = ulVertex[i];
    }

      // low left vertex:
    if ( llVertex == null )
      this.vertexLL = null;
    else
    {
      if ( this.vertexLL == null || this.vertexLL.length != 3 )
        this.vertexLL = new double[ 3 ];

      for ( int i = 0; i < 3 && i < llVertex.length; i++ )
        this.vertexLL[i] = llVertex[i];
    }

      // low right vertex:
    if ( lrVertex == null )
      this.vertexLR = null;
    else
    {
      if ( this.vertexLR == null || this.vertexLR.length != 3 )
        this.vertexLR = new double[ 3 ];

      for ( int i = 0; i < 3 && i < lrVertex.length; i++ )
        this.vertexLR[i] = lrVertex[i];
    }

      // source color/intensity:
    if ( color == null )
    {
      if ( lightColor == null ||
           lightColor.length != 1 )
        lightColor = new double[ 1 ];

      lightColor[0] = 1.0;
    }
    else
    {
      if ( lightColor == null ||
           lightColor.length != color.length )
        lightColor = new double[ color.length ];

      System.arraycopy( color, 0, lightColor, 0, color.length );
    }

      // dimming polynom:
    if ( dimming == null )
      dim = null;
    else
    {
      if ( dim == null || dim.length != 3 )
        dim = new double[ 3 ];

      for ( int i = 0; i < 3 && i < dimming.length; i++ )
        dim[i] = dimming[i];
    }

      // compute normal:
    double[][] vertices = { vertexUL, vertexUR, vertexLR };
    Geometry.triangleNormal( vertices, normal );
    Geometry.normalize3D( normal );

      // initialize TBN matrix:
    double[][] v = { vertexUL, vertexLR, vertexLL };
    m = new TrMatrix( 3, 3 );   // rows: tangent, binormal, normal
    Geometry.triangleNormal( v, m.m[2] );
    Geometry.vector( vertexUR, vertexUL, m.m[0] );
    Geometry.normalize3D( m.m[0] );
    Geometry.vectorProduct( m.m[0], m.m[2], m.m[1] );

    return true;
  }

  /** Read data from JGL script. */
  protected void readScript ()
  {
    if ( scriptName == null ) return;

    JGLScript script = new JGLScript();
    Map< String, Object > map = script.interpret( scriptName );
    if ( map != null )
    {
      Object o;

      if ( (o = map.get( UR_VERTEX ) ) != null &&
           o instanceof double[] )
        vertexUR = (double[])o;

      if ( (o = map.get( UL_VERTEX ) ) != null &&
           o instanceof double[] )
        vertexUL = (double[])o;

      if ( (o = map.get( LL_VERTEX ) ) != null &&
           o instanceof double[] )
        vertexLL = (double[])o;

      if ( (o = map.get( LR_VERTEX ) ) != null &&
           o instanceof double[] )
        vertexLR = (double[])o;

      if ( (o = map.get( KEY_LIGHT ) ) != null &&
           o instanceof double[] )
        lightColor = (double[])o;

      if ( (o = map.get( KEY_DIM ) ) != null &&
           o instanceof double[] )
        dim = (double[])o;

    }
  }

  /** For debugging purposes only. */
  @Override
  public String toString ()
  {
    StringBuffer sb = new StringBuffer( getClass().getName() );
    sb.append( ",urVertex=" )
      .append( Formula.vectorToString( vertexUR ) )
      .append( ",ulVertex=" )
      .append( Formula.vectorToString( vertexUL ) )
      .append( ",llVertex=" )
      .append( Formula.vectorToString( vertexLL ) )
      .append( ",lrVertex=" )
      .append( Formula.vectorToString( vertexLR ) )
      .append( ",lightColor=" )
      .append( Formula.vectorToString( lightColor ) )
      .append( ",dim=" )
      .append( Formula.vectorToString( dim ) )
      .append( ']' );
    return sb.toString();
  }

  //==========================================================================
  //  Interface LightSource:

  /**
   * Lighting test from the source to the given {@link MicroFacet}.
   *
   * @param  mf The checked {@link MicroFacet} (point on surface of some solid).
   * @return {@code true} if the light source lights on the given point.
   * @see    #getIntensity
   * @see    #getDirection
   */
  public boolean lights ( MicroFacet mf )
  {
    if ( mf    == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD ) != 0 )
      return false;

    double[] pos = new double[ 3 ];
    double[] dir = new double[ 3 ];
    MonteCarlo.sampleSquare( rnd.uniformNumber(),
                             rnd.uniformNumber(),
                             vertexUR,
                             vertexUL,
                             vertexLL,
                             vertexLR,
                             pos );
    Geometry.vector( mf.coord, pos, dir );
    Geometry.normalize3D( dir );

      // check the intersection of a light ray with the scene:
    if ( assertInter() )
      return true;

      // does the sample point see the front side of the MicroFacet?
    MicroFacet m = inter.intersection( mf.coord, dir );
    if ( m == null ||
         m.far( 0.9, dir ) )
      return true;

    return false;
  }

  /**
   * Returns light intensity coming from the source to the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  color Pre-allocated {@code double[C]} array to hold result - light intensity
   *               coming from the source to the given point ({@code C} can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @return {@code true} if the light source can light on the given point (<b>cannot</b>
   *         replace the {@link #lights} call!).
   * @see    #lights
   * @see    #getDirection
   */
  public boolean getIntensity ( MicroFacet mf, double[] color )
  {
    if ( color == null ||
         mf    == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD ) != 0 )
      return false;

    double[] pos = new double[ 3 ];
    double[] dir = new double[ 3 ];
    int i = 0, j = 0, k = 0;
/*
      // normalize number of shadow rays:
    int quantitySQRT = (int)Math.sqrt( (double)shadowRays[0] );
    shadowRays[0]    = quantitySQRT * quantitySQRT;
    double[] u = new double[ shadowRays[0] ];
    double[] v = new double[ shadowRays[0] ];

      // generate stratified random numbers:
    double interval  = 1.0 / quantitySQRT;
    for ( i = 0; i < quantitySQRT; i++ )
    {
      for ( j = 0; j < quantitySQRT; j++ )
      {
        u[k] = rnd.randomDouble( i*interval, (i+1)*interval );
        v[k] = rnd.randomDouble( j*interval, (j+1)*interval );
        k++;
      }
    }

    double dist    = 0.0;
    double dimCoef = 1.0;
    double invPI   = 1.0 / Math.PI;
    double pdf     = 0.0;
    for ( i = 0; i < shadowRays[0]; i++ )
    {
      MonteCarlo.sampleSquare( u[i], v[i], vertexUR, vertexUL, vertexLL, vertexLR, pos );
      Geometry.vector( mf.coord, pos, dir );

      dist += Geometry.pointPoint3D( pos, mf.coord );

        // does the sample point see the front side of the MicroFacet?
      MicroFacet c = inter.intersection( mf.coord, dir );

      if ( c == null ||
           c.far( 0.9, dir ) )
      {

        dir[0] = -dir[0];
        dir[1] = -dir[1];
        dir[2] = -dir[2];
        Geometry.normalize3D( dir );
        double cos = Geometry.dotProduct3D( dir, normal );
        pdf = invPI * cos;
        for ( j = 0; j < color.length; j++ )
          color[j] += cos * lightColor[j];
      }
    }

      // compute dim coefficient:
    double invNumRays = 1.0 / (double)shadowRays[0];
    dist *= invNumRays;
    dimCoef /= dim[0] + dist * (dim[1] + dist * dim[2]);

      // compute color:
    for ( i = 0; i < color.length; i++ )
    {
      color[i] *= invNumRays * dimCoef;
      Formula.clamp( color[i], 0.0, 1.0 );
    }
*/
     return true;
  }

  /**
   * Returns light direction between the source and the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  dir Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *             from the {@link MicroFacet} to the light source.
   * @return {@code false} if the light source is not directional (thus {@code dir}
   *         is irrelevant).
   * @see    #lights
   * @see    #getIntensity
   */
  @Override
  public boolean getDirection ( MicroFacet mf, double[] dir )
  {
    if ( dir == null ||
         mf  == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD ) != 0 )
      return false;

    double[] pos = new double[ 3 ];
    MonteCarlo.sampleSquare( rnd.uniformNumber(),
                             rnd.uniformNumber(),
                             vertexUR,
                             vertexUL,
                             vertexLL,
                             vertexLR,
                             pos );

    Geometry.vector( mf.coord, pos, dir );
    Geometry.normalize3D( dir );

    return true;
  }

  /**
   * Sample the light by one shadow ray.
   *
   * @param  mf    The {@link MicroFacet} (point on surface of some solid).
   * @param  u     Uniform random number.
   * @param  v     Uniform random number.
   * @param  color Pre-allocated <tt>double[C]</tt> array to hold result - light intensity
   *               coming from the source to the given point (<tt>C</tt> can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @param  dir   Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *               from the {@link MicroFacet} to the light source.
   * @param  pdf   Probability density function (output variable).
   * @return {@code True} if the light is visible.
   */
  @Override
  public boolean sample ( MicroFacet mf,
                          double u,
                          double v,
                          double[] color,
                          double[] dir,
                          double[] pdf,
                          boolean testVisibility )
  {
    if ( color == null ||
         dir   == null ||
         pdf   == null ||
         mf    == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD ) != 0 )
      return false;

      // sample the light:
    double[] pos = new double[ 3 ];
    MonteCarlo.sampleSquare( u, v, vertexUR, vertexUL, vertexLL, vertexLR, pos );

      // compute radiance leaving the light:
    Geometry.vector( pos, mf.coord, dir );
    Geometry.normalize3D( dir );
    double cos = Geometry.dotProduct3D( dir, normal );
    cos = Formula.clamp( cos, 0.0, 1.0 );
    System.arraycopy( lightColor, 0, color, 0, color.length );

      // compute pdf:
    double dist = Geometry.pointPoint3D( pos, mf.coord );
    double dimCoef = (dim[0] + dist * (dim[1] + dist * dim[2]));
    double area = area();
    if ( area == 0.0 || cos == 0.0 )
    {
      pdf[0] = 0.0;
      return false;
    }
    pdf[0] = dimCoef / (area * cos);

      // change direction (from MicroFacet to light):
    dir[0] = -dir[0];
    dir[1] = -dir[1];
    dir[2] = -dir[2];

      // check visibility if it's required:
    if ( testVisibility )
    {
      final double[] direction = Geometry.vector( mf.coord, pos, null );
      MicroFacet m = inter.intersection( mf.coord, direction );
      return ( m == null ||
               m.far( 1.0, direction ) );
    }
    return true;
  }

  /**
   * Generates one photon leaving this light source.
   *
   * @param  u1    Uniform random number.
   * @param  v1    Uniform random number.
   * @param  u2    Uniform random number.
   * @param  v2    Uniform random number.
   * @param  pos   Pre-allocated array which holds a point on the light source
   *               (an origin of direction).
   * @param  dir   Output direction in world coordinates. Pre-allocated array.
   * @param  power Power of given light source (if you wanna get power
   *               of this photon, you have to divide this power by number
   *               of photons). Pre-allocated array.
   * @return Probability density function of given ray.
   */
  @Override
  public double sample ( double u1,
                         double v1,
                         double u2,
                         double v2,
                         double[] pos,
                         double[] dir,
                         double[] power )
  {
    if ( pos   == null ||
         dir   == null ||
         power == null )
      return 0.0;

    MonteCarlo.sampleSquare( u1, v1, vertexUR, vertexUL, vertexLL, vertexLR, pos );
    MonteCarlo.uniformSampleHemisphere( u2, v2, dir );
    m.transformVector3D( dir, dir );

    System.arraycopy( lightColor, 0, power, 0, power.length );

    // from PI / ( area() * 2.0 * PI * dot(dir, normal) )
    return 1.0 / ( area() * 2.0 * Geometry.dotProduct3D( dir, normal ) );
  }

  /**
   * Return probability density function.
   *
   * @param point  Point on the surface.
   * @param normal Normal at this point.
   * @param dir    Direction from the point to the light.
   */
  @Override
  public double pdf ( final double[] point, final double[] normal, final double[] dir )
  {
    if ( point  == null ||
         normal == null ||
         dir    == null )
      return 0.0;

    MicroFacet mf = inter.intersection( point, dir );
    if ( mf != null &&
         mf.near( 1.0, dir ) )
    {
      mf.assertAttributes( MicroFacet.VALID_COORD );
      double dist = Geometry.pointPoint3D( point, mf.coord );

      double dimCoef = 1.0;
      dimCoef = dim[0] + dist * (dim[1] + dist * dim[2]);
      double cos = Geometry.dotProduct3D( dir, normal );
      cos = Formula.clamp( cos, 0.0, 1.0 );
      if ( cos != 0.0 )
        return dimCoef / (area() * cos);
      else
        return 0.0;
    }
    return 0.0;
  }

  /** Is delta light? */
  @Override
  public boolean isDeltaLight ()
  {
    return false;
  }

  /** Computes area of rectangle. */
  protected double area ()
  {
    return heron( vertexUL, vertexLL, vertexLR ) + heron( vertexUL, vertexLL, vertexUR );
  }

  /** Computes area of triangle by Heron's formula. */
  protected double heron ( double[] A, double[] B, double[] C )
  {
    double a = Math.sqrt( (B[0] - C[0])*(B[0] - C[0]) +
                          (B[1] - C[1])*(B[1] - C[1]) +
                          (B[2] - C[2])*(B[2] - C[2]) );
    double b = Math.sqrt( (A[0] - C[0])*(A[0] - C[0]) +
                          (A[1] - C[1])*(A[1] - C[1]) +
                          (A[2] - C[2])*(A[2] - C[2]) );
    double c = Math.sqrt( (B[0] - A[0])*(B[0] - A[0]) +
                          (B[1] - A[1])*(B[1] - A[1]) +
                          (B[2] - A[2])*(B[2] - A[2]) );
    double s = (a + b + c) / 2;
    return Math.sqrt( s * (s - a) * (s - b) * (s - c) );
  }

  //==========================================================================
  //  Interface Property:

  /** JGL script key: position of up right vertex. */
  public final static String UR_VERTEX  = "ur_vertex";

  /** JGL script key: position of up left vertex. */
  public final static String UL_VERTEX  = "ul_vertex";

  /** JGL script key: position of low left vertex. */
  public final static String LL_VERTEX  = "ll_vertex";

  /** JGL script key: position of low right vertex. */
  public final static String LR_VERTEX  = "lr_vertex";

  /** JGL script key: dim. */
  public final static String KEY_DIM    = "dim";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // PARAMS:
    if ( key.compareTo(PARAMS) == 0 )
    {
      scriptName = stringProperty( value, scriptName );
      readScript();
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(UR_VERTEX) == 0 )
      return Formula.cloneArray( vertexUR  );
    if ( key.compareTo(UL_VERTEX) == 0 )
      return Formula.cloneArray( vertexUL    );
    if ( key.compareTo(LL_VERTEX) == 0 )
      return Formula.cloneArray( vertexLL    );
    if ( key.compareTo(LR_VERTEX) == 0 )
      return Formula.cloneArray( vertexLR    );
    if ( key.compareTo(COLOR) == 0 )
      return Formula.cloneArray( lightColor );
    if ( key.compareTo(PARAMS) == 0 )
      return scriptName;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Static square light source";

  /** Object description. */
  private final static String DESCRIPTION = "Square light source shining through " +
                                            "the whole hemisphere which is on " +
                                            "the front face.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"LightSource" );
        // Property:
      t.propBegin( COLOR, TYPE_OBJECT, "Light source color (intensity)", false );
      t.propEnd();
      t.propBegin( PARAMS, TYPE_OBJECT, "Light source definition script", true );
        t.propDefault( "data/square.jgl" );
        t.propManipulator( "ParamJGLScript" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}