package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.xml.JGLScript;

/**
 * Spherical light shining all around. Every point on the sphere shining
 * through the hemisphere.
 *
 * @version 0.26 $Rev: 786 $ $Date: 2008-06-17 20:21:15 +0200 (út, 17 VI 2008) $ $Author: Tomas $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/StaticSphericalLightSource.java">StaticSphericalLightSource.java</a>
 */
public class StaticSphericalLightSource extends DefaultLightSource
{
  //==========================================================================
  //  Data:

  {
    scriptName = "data/spherical.jgl";
  }

  /** Number of shooted photons (for methods based on photon tracing). */
  protected int[] photons = { 0 };

  /** Center of the sphere. */
  protected double[] center = new double[ 3 ];

  /** Radius of the sphere. */
  protected double[] radius = new double[] { 1.0 };

  /**
   * Color of the source: {@code 1D} for monochrome source, {@code 3D} for <b>RGB</b>
   * color space, or more for complex spectrum sampled spaces.
   */
  protected double[] lightColor = new double[] { 1.0 };

  /**
   * Dimming polynom coefficients: light is dimmed by the factor of
   * {@code dim[0] + dim[1] * D + dim[2] * D * D}.
   * Can be {@code null} for no dimming..
   */
  protected double[] dim = new double[] { 0.5, 1.0, 0.1 };

  //==========================================================================
  //  Support:

  /**
   * Set light source parameters.
   *
   * @param  photons  Number of photons to shoot.
   * @param  power    Power of light.
   * @param  urVertex Up right vertex of square.
   * @param  ulVertex Up left vertex of square.
   * @param  llVertex Low left vertex of square.
   * @param  lrVertex Low right vertex of square.
   * @param  color Source color/intensity (can be either 1D /for monochrome computing/,
   *               3D /for RGB color system/ or greater /for more complex spectral color
   *               sampling/).
   */
  public boolean setParams ( int photons, float[] power, double[] center,
                             double[] radius, double[] color )
  {
    if ( center == null || center.length < 3 ||
         radius == null || radius.length < 3 )
      return false;

    this.photons[0] = photons;

      // center:
    if ( center == null )
      this.center = null;
    else
    {
      if ( this.center == null || this.center.length != 3 )
        this.center = new double[ 3 ];

      for ( int i = 0; i < 3 && i < center.length; i++ )
        this.center[i] = center[i];
    }

      // radius:
    if ( radius == null )
      this.radius = null;
    else
    {
      if ( this.radius == null || this.radius.length != 3 )
        this.radius = new double[ 3 ];

      for ( int i = 0; i < 3 && i < radius.length; i++ )
        this.radius[i] = radius[i];
    }

      // source color/intensity:
    if ( color == null )
    {
      if ( lightColor == null ||
           lightColor.length != 1 )
        lightColor = new double[ 1 ];

      lightColor[0] = 1.0;
    }
    else
    {
      if ( lightColor == null ||
           lightColor.length != color.length )
        lightColor = new double[ color.length ];

      System.arraycopy( color, 0, lightColor, 0, color.length );
    }

    return true;
  }

  /** Read data from JGL script. */
  protected void readScript ()
  {
    if ( scriptName == null ) return;

    JGLScript script = new JGLScript();
    Map< String, Object > map = script.interpret( scriptName );
    if ( map != null )
    {
      Object o;

      if ( (o = map.get( PHOTONS ) ) != null &&
           o instanceof int[] )
        photons = (int[])o;

      if ( (o = map.get( CENTER ) ) != null &&
           o instanceof double[] )
        center = (double[])o;

      if ( (o = map.get( RADIUS ) ) != null &&
           o instanceof double[] )
        radius = (double[])o;

      if ( (o = map.get( KEY_LIGHT ) ) != null &&
           o instanceof double[] )
        lightColor = (double[])o;
    }
  }

  /** For debugging purposes only. */
  @Override
  public String toString ()
  {
    StringBuffer sb = new StringBuffer( getClass().getName() );
    sb.append( "[photons=" )
      .append( Formula.vectorToString( photons ) )
      .append( ",center=" )
      .append( Formula.vectorToString( center ) )
      .append( ",radius=" )
      .append( Formula.vectorToString( radius ) )
      .append( ",color=" )
      .append( Formula.vectorToString( lightColor ) );
    return sb.toString();
  }

  //==========================================================================
  //  Interface LightSource:

  /**
   * Lighting test from the source to the given {@link MicroFacet}.
   *
   * @param  mf The checked {@link MicroFacet} (point on surface of some solid).
   * @return {@code true} if the light source lights on the given point.
   * @see    #getIntensity
   * @see    #getDirection
   * @todo   Sample hemisphere (create direction).
   */
  public boolean lights ( MicroFacet mf )
  {
    if ( mf == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD |
                              MicroFacet.VALID_NORMAL ) != 0 ) return true;

    double[] direction = null; //Geometry.vector( mf.coord, vertex[i], null );
    if ( mf.normal[0] * direction[0] +
         mf.normal[1] * direction[1] +
         mf.normal[2] * direction[2] <= 0.0 ) return false;

      // check the intersection of a light ray with the scene:
    if ( assertInter() )
      return true;

      // does the sample point see the front side of the MicroFacet?
    MicroFacet c = inter.intersection( mf.coord, direction );

    return( c == null ||
            c.far( 1.0, direction ) );
  }

  /**
   * Returns light intensity coming from the source to the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  color Pre-allocated {@code double[C]} array to hold result - light intensity
   *               coming from the source to the given point ({@code C} can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @return {@code true} if the light source can light on the given point (<b>cannot</b>
   *         replace the {@link #lights} call!).
   * @see    #lights
   * @see    #getDirection
   */
  public boolean getIntensity ( MicroFacet mf, double[] color )
  {
    double dimCoef = 1.0;
    double[] vertex = new double[ 3 ]; // TODO: sample hemisphere (origin of direction) or use (center[]-radius*weight)
    if ( dim       != null &&
         mf        != null &&
         mf.assertAttributes( MicroFacet.VALID_COORD ) == 0 )
    {
      double dist = Geometry.pointPoint3D( mf.coord, vertex );
      dimCoef /= dim[0] + dist * (dim[1] + dist * dim[2]);
    }

    int i;
    if ( color.length <= lightColor.length )
      for ( i = 0; i < color.length; i++ )
         color[i] = dimCoef * lightColor[i];
    else
      for ( i = 0; i < color.length; i++ )
         color[i] = dimCoef * lightColor[0];

    return true;
  }

  /**
   * Returns light direction between the source and the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  dir Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *             from the {@link MicroFacet} to the light source.
   * @return {@code false} if the light source is not directional (thus {@code dir}
   *         is irrelevant).
   * @see    #lights
   * @see    #getIntensity
   * @todo   Maybe it should return something (position on the light could
   *         be random or center).
   */
  public boolean getDirection ( MicroFacet mf, double[] dir )
  {
    return false;
  }

  //==========================================================================
  //  Interface Property:

  /** JGL script key: position of up right vertex. */
  public final static String PHOTONS   = "photons";

  /** JGL script key: position of up right vertex. */
  public final static String POWER     = "power";

  /** JGL script key: center of sphere. */
  public final static String CENTER    = "center";

  /** JGL script key: radius of sphere. */
  public final static String RADIUS    = "radius";

  /** JGL script key: color. */
  public final static String COLOR     = "color";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // PARAMS:
    if ( key.compareTo(PARAMS) == 0 )
    {
      scriptName = stringProperty( value, scriptName );
      readScript();
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(PHOTONS) == 0 )
      return Formula.cloneArray( photons );
    if ( key.compareTo(CENTER) == 0 )
      return Formula.cloneArray( center );
    if ( key.compareTo(RADIUS) == 0 )
      return Formula.cloneArray( radius );
    if ( key.compareTo(COLOR) == 0 )
      return Formula.cloneArray( lightColor );
    if ( key.compareTo(PARAMS) == 0 )
      return scriptName;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Static spherical light source";

  /** Object description. */
  private final static String DESCRIPTION = "Square light source shining through " +
                                            "the whole hemisphere which is on " +
                                            "the front face.";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"LightSource" );
        // Property:
      t.propBegin( COLOR, TYPE_OBJECT, "Light source color (intensity)", false );
      t.propEnd();
      t.propBegin( PARAMS, TYPE_OBJECT, "Light source definition script", true );
        t.propDefault( "data/spherical.jgl" );
        t.propManipulator( "ParamJGLScript" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}