package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.xml.JGLScript;

/**
 * Static point light source.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 821 $ $Date: 2008-11-17 21:35:59 +0100 (po, 17 XI 2008) $ $Author: Tomas $
 * @since   0.12
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/StaticSpotLightSource.java">StaticSpotLightSource.java</a>
 */
public class StaticSpotLightSource extends DefaultLightSource
{
  //==========================================================================
  //  Data:

  /** Random number generator. */
  protected final RandomJames rnd = new RandomJames();

  /** Light source coordinates. */
  protected double[] position = new double[] { 0.0, 0.0, 0.0 };

  /** Light source coordinates. */
  protected double[] direction = new double[] { 0.0, 0.0, 0.0 };

  /**
   * Color of the source: {@code 1D} for monochrome source, {@code 3D} for <b>RGB</b>
   * color space, or more for complex spectrum sampled spaces.
   */
  protected double[] lightColor = new double[] { 1.0 };

  /**
   * Angle which defines outer cone. Objects inside this are illuminated, but
   * intensity of this illumination fall off from full intensity to
   * no intensity.
   */
  protected double[] thetaMax = { 0 };

  /** Cosinus of angle totalWidth. */
  protected double cosThetaMax;

  /**
   * Angle which defines inner cone. Objects inside this cone are fully
   * illuminated.
   */
  protected double[] fallOffStart = { 0 };

  /** Cosinus of angle cosFallOffStart. */
  protected double cosFallOffStart;

  /**
   * Dimming polynom coefficients: light is dimmed by the factor of
   * {@code dim[0] + dim[1] * D + dim[2] * D * D}.
   * Can be {@code null} for no dimming..
   */
  protected double[] dim = new double[] { 0.5, 1.0, 0.1 };

  //==========================================================================
  //  Support:

  /**
   * Set light source parameters.
   *
   * @param  pos Source position.
   * @param  color Source color/intensity (can be either 1D /for monochrome computing/,
   *               3D /for RGB color system/ or greater /for more complex spectral color
   *               sampling/).
   * @param  dimming Dimming polynom coefficients: light is dimmed by the factor of
   *                 {@code dim[0] + dim[1] * D + dim[2] * D * D}. Can be
   *                 {@code null} for no dimming..
   * @return {@code true} if parameters were successfully set.
   */
  public boolean setParams ( double[] pos,
                             double[] dir,
                             double[] color,
                             double width,
                             double fallOff,
                             double[] dimming )
  {
      // source position:
    if ( pos == null ||
         pos.length < 3 ) return false;

    position  = Geometry.toCartesian3D( pos, null );
    direction = Geometry.toCartesian3D( dir, null );
    Geometry.normalize3D( direction );

      // source color/intensity:
    if ( color == null )
    {
      if ( lightColor        == null ||
           lightColor.length != 1 )
        lightColor = new double[ 1 ];

      lightColor[0] = 1.0;
    }
    else
    {
      if ( lightColor        == null ||
           lightColor.length != color.length )
        lightColor = new double[ color.length ];

      System.arraycopy( color, 0, lightColor, 0, color.length );
    }

    this.thetaMax[0]   = width;
    this.fallOffStart[0] = fallOff;

    cosThetaMax     = Math.cos( Math.toRadians( thetaMax[0] ) );
    cosFallOffStart = Math.cos( Math.toRadians( fallOffStart[0] ) );

      // dimming polynom:
    if ( dimming == null )
      dim = null;
    else
    {
      if ( dim == null || dim.length != 3 )
        dim = new double[ 3 ];

      for ( int i = 0; i < 3 && i < dimming.length; i++ )
        dim[i] = dimming[i];
    }

    return true;
  }

  /** Read data from JGL script. */
  protected void readScript ()
  {
    if ( scriptName == null ) return;

    JGLScript script = new JGLScript();
    Map< String, Object > map = script.interpret( scriptName );
    if ( map != null )
    {
      Object o;

      if ( (o = map.get( KEY_POSITION ) ) != null &&
           o instanceof double[] )
        position = Geometry.toCartesian3D( (double[])o, null );

      if ( (o = map.get( DIRECTION ) ) != null &&
           o instanceof double[] )
        direction = Geometry.toCartesian3D( (double[])o, null );

      if ( (o = map.get( KEY_LIGHT ) ) != null &&
           o instanceof double[] )
        lightColor = (double[])o;

      if ( (o = map.get( PHOTONS ) ) != null &&
           o instanceof double[] )
        thetaMax = (double[])o;

      if ( (o = map.get( PHOTONS ) ) != null &&
           o instanceof double[] )
        fallOffStart = (double[])o;

      if ( (o = map.get( KEY_DIM ) ) != null &&
           o instanceof double[] )
        dim = (double[])o;
    }
  }

  /** For debugging purposes only. */
  @Override
  public String toString ()
  {
    StringBuffer sb = new StringBuffer( getClass().getName() );
    sb.append( ",pos=" )
      .append( Formula.vectorToString( position ) )
      .append( ",dir=" )
      .append( Formula.vectorToString( direction ) )
      .append( ",color=" )
      .append( Formula.vectorToString( lightColor ) )
      .append( ",total width=" )
      .append( Formula.vectorToString( thetaMax ) )
      .append( ",fall off start=" )
      .append( Formula.vectorToString( fallOffStart ) )
      .append( ",dim=" )
      .append( Formula.vectorToString( dim ) )
      .append( ']' );
    return sb.toString();
  }

  //==========================================================================
  //  Interface LightSource:

  /**
   * Lighting test from the source to the given {@link MicroFacet}.
   *
   * @param  mf The checked {@link MicroFacet} (point on surface of some solid).
   * @return {@code true} if the light source lights on the given point.
   * @see    #getIntensity
   * @see    #getDirection
   */
  @Override
  public boolean lights ( MicroFacet mf )
  {
    if ( mf == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD |
                              MicroFacet.VALID_NORMAL ) != 0 ) return true;

    final double[] direction = Geometry.vector( mf.coord, position, null );
    if ( mf.normal[0] * direction[0] +
         mf.normal[1] * direction[1] +
         mf.normal[2] * direction[2] <= 0.0 ) return false;

      // check the intersection of a light ray with the scene:
    if ( assertInter() )
      return true;

      // does the sample point see the front side of the MicroFacet?
    MicroFacet c = inter.intersection( mf.coord, direction );

    return( c == null ||
            c.far( 1.0, direction ) );
  }

  /**
   * Returns light intensity coming from the source to the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  color Pre-allocated {@code double[C]} array to hold result - light intensity
   *               coming from the source to the given point ({@code C} can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @return {@code true} if the light source can light on the given point (<b>cannot</b>
   *         replace the {@code lights()} call!).
   * @see    #lights
   * @see    #getDirection
   */
  @Override
  public boolean getIntensity ( MicroFacet mf, double[] color )
  {
    double dimCoef = 1.0;
    if ( position != null &&
         dim      != null &&
         mf       != null &&
         mf.assertAttributes( MicroFacet.VALID_COORD ) == 0 )
    {
      double dist = Geometry.pointPoint3D( mf.coord, position );
      dimCoef /= dim[0] + dist * (dim[1] + dist * dim[2]);
    }

    int i;
    if ( color.length <= lightColor.length )
      for ( i = 0; i < color.length; i++ )
         color[i] = dimCoef * lightColor[i];
    else
      for ( i = 0; i < color.length; i++ )
         color[i] = dimCoef * lightColor[0];

    return true;
  }

  /**
   * Returns light direction between the source and the given {@link MicroFacet}.
   *
   * @param  mf The {@link MicroFacet} (point on surface of some solid).
   * @param  dir Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *             from the {@link MicroFacet} to the light source.
   * @return {@code false} if the light source is not directional (thus {@code dir}
   *         is irrelevant).
   * @see    #lights
   * @see    #getIntensity
   */
  @Override
  public boolean getDirection ( MicroFacet mf, double[] dir )
  {
    if ( dir        == null ||
         dir.length <  3 ||
         position   == null ||
         mf         == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD ) != 0 ) return false;

    Geometry.vector( mf.coord, position, dir );
    Geometry.normalize3D( dir );

    return true;
  }

  /**
   * Sample the light by one shadow ray.
   *
   * @param  mf    The {@link MicroFacet} (point on surface of some solid).
   * @param  u     Uniform random number.
   * @param  v     Uniform random number.
   * @param  color Pre-allocated <tt>double[C]</tt> array to hold result - light intensity
   *               coming from the source to the given point (<tt>C</tt> can be either 1
   *               /for monochrome computing/, 3 /for RGB color system/ or greater /for more
   *               complex spectral color sampling/).
   * @param  dir   Pre-allocated array to hold the result - <b>normalised</b> direction vector
   *               from the {@link MicroFacet} to the light source.
   * @param  pdf   Probability density function (output variable).
   * @return {@code True} if the light is visible.
   */
  @Override
  public boolean sample ( MicroFacet mf,
                          double u,
                          double v,
                          double[] color,
                          double[] dir,
                          double[] pdf,
                          boolean testVisibility )
  {
    if ( dir          == null ||
         dir.length   <  3    ||
         position     == null ||
         color        == null ||
         color.length <  3    ||
         mf           == null ||
         mf.assertAttributes( MicroFacet.VALID_COORD ) != 0 ) return false;

    Geometry.vector( position, mf.coord, dir );
    Geometry.normalize3D( dir );

    pdf[0] = 1.0;

    double dist = Geometry.pointPoint3D( position, mf.coord );
    double dimCoef = (dim[0] + dist * (dim[1] + dist * dim[2]));
    double invDim = 1.0 / dimCoef;
    for ( int i = 0; i < color.length; i++ )
      color[i] = lightColor[i] * getFallOff( dir ) * invDim;
    dir[0] = -dir[0];
    dir[1] = -dir[1];
    dir[2] = -dir[2];

      // check visibility if it's required:
    if ( testVisibility )
    {
      final double[] d = Geometry.vector( mf.coord, position, null );
      MicroFacet m = inter.intersection( mf.coord, d );
      return ( m == null ||
               m.far( 1.0, d ) );
    }
    return true;
  }

  /**
   * Generates one photon leaving this light source.
   *
   * @param  u1    Uniform random number.
   * @param  v1    Uniform random number.
   * @param  u2    Uniform random number.
   * @param  v2    Uniform random number.
   * @param  pos   Pre-allocated array which holds a point on the light source
   *               (an origin of direction).
   * @param  dir   Output direction in world coordinates. Pre-allocated array.
   * @param  power Power of given light source (if you wanna get power
   *               of this photon, you have to divide this power by number
   *               of photons). Pre-allocated array.
   * @return Weight of given ray (standart weight = 1.0).
   */
  public double sample ( double u1,
                         double v1,
                         double u2,
                         double v2,
                         double[] pos,
                         double[] dir,
                         double[] power )
  {
    if ( pos   == null ||
         dir   == null ||
         power == null )
      return 0.0;

    MonteCarlo.uniformSampleCone( u1, v1, cosThetaMax, dir );
    // TODO: transform direction to the world coordinates!
    System.arraycopy( position, 0, pos, 0, pos.length );

    for ( int i = 0; i < power.length; i++ )
      power[i] = lightColor[i] * getFallOff( dir );

    return 1.0 / ( 2.0 * Math.PI * (1.0 - cosThetaMax) );
  }

  /**
   * Return probability density function.
   *
   * @param point  Point on the surface.
   * @param normal Normal at this point.
   * @param dir    Direction from the point.
   */
  @Override
  public double pdf ( final double[] point, final double[] normal, final double[] dir )
  {
    // It's a waste of time to try to hit the delta light source.
    return 0.0;
  }

  /** Is delta light? */
  public boolean isDeltaLight ()
  {
    return true;
  }

  //==========================================================================
  //  Support methods:

  /** Returns fall off for given direction. */
  protected double getFallOff ( double[] dir )
  {
    double cosTheta = Geometry.dotProduct3D( direction, dir );
    if ( cosTheta < cosThetaMax )
      return 0.0;
    if ( cosTheta > cosFallOffStart )
      return 1.0;

    double delta = ( cosTheta - cosThetaMax ) / ( cosFallOffStart - cosThetaMax );
    return delta*delta*delta*delta;
  }

  /** Transform from light space to world space. */
  protected void transform ( double[] dir )
  {
    // TBN (tangent, binormal, normal) matrix.
    TrMatrix m = new TrMatrix( 3, 3 );
    m.m[2] = direction;
  }

  //==========================================================================
  //  Interface Property:

  /** JGL script key: position of up right vertex. */
  public final static String PHOTONS        = "photons";

  /** JGL script key: light position. */
  public final static String KEY_POSITION   = "position";

  /** JGL script key: light direction. */
  public final static String DIRECTION      = "direction";

  /** JGL script key: total width. */
  public final static String TOTAL_WIDTH    = "total width";

  /** JGL script key: fall off start. */
  public final static String FALL_OFF_START = "fall off start";

  /** JGL script key: dim. */
  public final static String KEY_DIM        = "dim";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // PARAMS:
    if ( key.compareTo(PARAMS) == 0 )
    {
      scriptName = stringProperty( value, scriptName );
      readScript();
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(DIRECTION) == 0 )
      return Formula.cloneArray( direction );
    if ( key.compareTo(TOTAL_WIDTH) == 0 )
      return Formula.cloneArray( thetaMax  );
    if ( key.compareTo(FALL_OFF_START) == 0 )
      return Formula.cloneArray( fallOffStart );
    if ( key.compareTo(PARAMS) == 0 )
      return scriptName;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Sport light source (static)";

  /** Object description. */
  private final static String DESCRIPTION = "Spot light source";

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,  IFACE+"LightSource" );
      t.newOptOutputPlug( PL_OUTPUT, IFACE+"Intersectable" );
        // Property:
      t.propBegin( PARAMS, TYPE_OBJECT, "Light source parameters", true );
        t.propDefault( "data/light.jgl" );
        t.propManipulator( "ParamJGLScript" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}