package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Creates sample 3D scene ({@link Brep}) for testing of rendering algorithms.
 * <br />So far implemented: graph of R<sup>2</sup> to R function (provided externally via the {@link RealFunction} module).
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/Test3DScene.java">Test3DScene.java</a>
 */
public class Test3DScene extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Number of cells in X direction. */
  protected int cellsX  = 20;

  /** Number of cells in Y direction. */
  protected int cellsY  = 20;

  protected double minX = -5.0;
  protected double maxX =  5.0;
  protected double minY = -5.0;
  protected double maxY =  5.0;

  /** Number of fans to be drawn. */
  protected int fans = 10;

  /** Maximal radius for fan generation. */
  protected double maxRadius = 3.0;

  /** Set negative value for randomize. */
  protected int seed = -1;

  /** Global random generator. Will be initialized in every generation run. */
  protected RandomJames rnd = null;

  /** Non-penetrating faces? */
  protected boolean painter = false;

  /** Generate transparent colors? */
  protected boolean transparent = false;

  /** Color multiplier for vertex colors. */
  protected double colorMul = 0.2;

  /** Fan minimum vertical position. */
  protected static final double FAN_MINZ = 0.0;

  /** Fan maximum vertical position. */
  protected static final double FAN_MAXZ = 8.0;

  /** Relative blade thickness (thickness / width ratio). */
  protected static final double BLADE_THICKNESS = 0.01;

  /** Blade width in world space. */
  protected static final double BLADE_WIDTH = 0.8;

  /** Additional triangle radius in world space. */
  protected static final double ADD_RADIUS = 0.1;

  /** Maximum number of fan double-blades. */
  protected static final int MAX_BLADES = 8;

  //==========================================================================
  //  Support:

  /** Generate random (optionally transparent) color. Uses exactly three random-generator calls. */
  protected void randomColor ( float[] color, boolean tran )
  {
    color[0] = rnd.randomFloat( 0.3f, 1.0f );
    color[1] = rnd.randomFloat( 0.3f, 1.0f );
    color[2] = rnd.randomFloat( 0.3f, 1.0f );
    if ( transparent )
      if ( tran )
      {
        color[3]  = 0.3f * (color[0] + color[1]);
        color[0] *= color[3];
        color[1] *= color[3];
        color[2] *= color[3];
      }
      else
        color[3] = 1.0f;
  }

  /** Permutate the given color vector, sets non-transparent result color. Doesn't call random-generator at all. */
  protected void permuteColor ( float[] color )
  {
    float tmp = color[0];
    color[0]  = color[1];
    color[1]  = color[2];
    color[2]  = tmp;
    if ( transparent )
      color[3] = 1.0f;
  }

  /** Generates one random fan. */
  protected void fan ( Brep data, int ctx, int vertexColor, int faceColor, int vertexNormal, int vertexTex )
  {
    double angle = rnd.uniformNumber() * Math.PI;
    int count    = rnd.randomInteger( 2, MAX_BLADES );
    final double dAngle = Math.PI / count;
    final double offLen = 0.5 * BLADE_WIDTH * BLADE_THICKNESS;
    final double epsLen = offLen * 1.01 / Math.tan( 0.5 * dAngle );
    final double radius = maxRadius * rnd.randomDouble( 0.4, 1.0 );
    double[] center = { rnd.randomDouble( minX,     maxX ),
                        rnd.randomDouble( minY,     maxY ),
                        rnd.randomDouble( FAN_MINZ, FAN_MAXZ ),
                        1.0 };
    double[] axis   = rnd.uniformDirection( -1.0, 1.0, null ); // normalized
    double[] target = { center[0] + BLADE_WIDTH * axis[0],
                        center[1] + BLADE_WIDTH * axis[1],
                        center[2] + BLADE_WIDTH * axis[2],
                        1.0 };
    double[] base1  = new double[ 4 ];
    double[] base2  = new double[ 4 ];
    Geometry.getAxes( axis, base1, base2 );
    double[] blade  = new double[ 4 ];
    double[] blEps  = new double[ 4 ];
    double[] norm   = new double[ 4 ];
    double[] offset = new double[ 4 ];
    double[] act    = { 0.0, 0.0, 0.0, 1.0 };
    double[] add    = { 0.0, 0.0, 0.0, 1.0 };
    int[] vert      = new int[ 4 ];
    float[] color   = new float[ transparent ? 4 : 3 ];
    float[] colorF  = new float[ transparent ? 4 : 3 ];
    double[][] tex  = {
                        { 0.0, 0.0 },
                        { 0.1, 0.0 },
                        { 0.1, 1.0 },
                        { 0.0, 1.0 }
                      };
    int i, face;

    do
    {
      final double sina = Math.sin( angle );
      final double cosa = Math.cos( angle );

      for ( i = 0; i < 3; i++ )
        blade[i] = cosa * base1[i] + sina * base2[i];
      Geometry.normalize3D( blade, blade );

      Geometry.vectorProduct( blade, axis, norm );
      Geometry.normalize3D( norm );

        // norm and blade are normalized:
      for ( i = 0; i < 3; i++ )
      {
        offset[i] = norm[i]  * offLen;
        blEps[i]  = blade[i] * epsLen;
        blade[i] *= radius;
      }

        // I. positive side:
      for ( i = 0; i < 4; i++ )
      {
        vert[i] = data.createVertex();
          // vertex color:
        randomColor( color, true );
        data.setAttribute( vertexColor,  vert[i], color );
          // vertex normal:
        data.setAttribute( vertexNormal, vert[i], norm );
          // texture coordinates:
        data.setAttribute( vertexTex,    vert[i], tex[i] );
      }

        // vertex coordinates:
      for ( i = 0; i < 3; i++ )
        act[i] = target[i] + blade[i] + offset[i];
      data.setVertexCoords( ctx, vert[0], act );

      for ( i = 0; i < 3; i++ )
        act[i] = center[i] + blade[i] + offset[i];
      data.setVertexCoords( ctx, vert[1], act );

      if ( painter || transparent )
      {                                   // two small quadrangles
        for ( i = 0; i < 3; i++ )
          act[i] = center[i] + blEps[i] + offset[i];
        data.setVertexCoords( ctx, vert[2], act );

        for ( i = 0; i < 3; i++ )
          act[i] = target[i] + blEps[i] + offset[i];
        data.setVertexCoords( ctx, vert[3], act );

          // topology - quadrilateral face:
        randomColor( colorF, true );
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );

          // 2nd quadrangle:
        for ( i = 0; i < 4; i++ )
        {
          vert[i] = data.createVertex();
            // vertex color:
          data.setAttribute( vertexColor,  vert[i], color );
            // vertex normal:
          data.setAttribute( vertexNormal, vert[i], norm );
            // texture coordinates:
          data.setAttribute( vertexTex,    vert[i], tex[i] );
        }

        for ( i = 0; i < 3; i++ )
          act[i] = target[i] - blEps[i] + offset[i];
        data.setVertexCoords( ctx, vert[0], act );

        for ( i = 0; i < 3; i++ )
          act[i] = center[i] - blEps[i] + offset[i];
        data.setVertexCoords( ctx, vert[1], act );

        for ( i = 0; i < 3; i++ )
          act[i] = center[i] - blade[i] + offset[i];
        data.setVertexCoords( ctx, vert[2], act );

        for ( i = 0; i < 3; i++ )
          act[i] = target[i] - blade[i] + offset[i];
        data.setVertexCoords( ctx, vert[3], act );

          // topology - quadrilateral face:
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );

          // add four small triangles at the blade thirds
        permuteColor( color );
        permuteColor( colorF );

        for ( i = 0; i < 3; i++ )
        {
          vert[i] = data.createVertex();
            // vertex color:
          data.setAttribute( vertexColor,  vert[i], color );
            // vertex normal:
          data.setAttribute( vertexNormal, vert[i], norm );
            // texture coordinates:
          data.setAttribute( vertexTex,    vert[i], tex[i] );
        }
        vert[3] = Brep.NULL;

        for ( i = 0; i < 3; i++ )
          add[i] = 0.5 * (target[i] + center[i]) + 0.35 * blade[i] + 6.0 * offset[i];
        data.setVertexCoords( ctx, vert[0], add );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] - axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[1], act );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] + axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[2], act );

          // topology - triangle face:
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );

          // 2nd additional triangle:
        for ( i = 0; i < 3; i++ )
        {
          vert[i] = data.createVertex();
            // vertex color:
          data.setAttribute( vertexColor,  vert[i], color );
            // vertex normal:
          data.setAttribute( vertexNormal, vert[i], norm );
            // texture coordinates:
          data.setAttribute( vertexTex,    vert[i], tex[i] );
        }

        for ( i = 0; i < 3; i++ )
          add[i] = 0.5 * (target[i] + center[i]) + 0.75 * blade[i] + 6.0 * offset[i];

        data.setVertexCoords( ctx, vert[0], add );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] - axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[1], act );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] + axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[2], act );

          // topology - triangle face:
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );

          // 3rd additional triangle:
        for ( i = 0; i < 3; i++ )
        {
          vert[i] = data.createVertex();
            // vertex color:
          data.setAttribute( vertexColor,  vert[i], color );
            // vertex normal:
          data.setAttribute( vertexNormal, vert[i], norm );
            // texture coordinates:
          data.setAttribute( vertexTex,    vert[i], tex[i] );
        }

        for ( i = 0; i < 3; i++ )
          add[i] = 0.5 * (target[i] + center[i]) - 0.25 * blade[i] + 6.0 * offset[i];

        data.setVertexCoords( ctx, vert[0], add );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] - axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[1], act );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] + axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[2], act );

          // topology - triangle face:
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );

          // 4th additional triangle:
        for ( i = 0; i < 3; i++ )
        {
          vert[i] = data.createVertex();
            // vertex color:
          data.setAttribute( vertexColor,  vert[i], color );
            // vertex normal:
          data.setAttribute( vertexNormal, vert[i], norm );
            // texture coordinates:
          data.setAttribute( vertexTex,    vert[i], tex[i] );
        }

        for ( i = 0; i < 3; i++ )
          add[i] = 0.5 * (target[i] + center[i]) - 0.65 * blade[i] + 6.0 * offset[i];

        data.setVertexCoords( ctx, vert[0], add );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] - axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[1], act );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] + axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[2], act );

          // topology - triangle face:
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );
      }

      else
      {                                   // one long (penetrating) quadrangle
        for ( i = 0; i < 3; i++ )
          act[i] = center[i] - blade[i] + offset[i];
        data.setVertexCoords( ctx, vert[2], act );

        for ( i = 0; i < 3; i++ )
          act[i] = target[i] - blade[i] + offset[i];
        data.setVertexCoords( ctx, vert[3], act );

          // topology - quadrilateral face:
        randomColor( colorF, true );
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );
      }

        // II. negative side:
      for ( i = 0; i < 3; i++ )
        norm[i] = -norm[i];

      for ( i = 0; i < 4; i++ )
      {
        vert[i] = data.createVertex();
          // vertex color:
        randomColor( color, true );
        data.setAttribute( vertexColor,  vert[i], color );
          // vertex normal:
        data.setAttribute( vertexNormal, vert[i], norm );
          // texture coordinates:
        data.setAttribute( vertexTex,    vert[i], tex[3-i] );
      }

        // vertex coordinates:
      for ( i = 0; i < 3; i++ )
        act[i] = target[i] - blade[i] - offset[i];
      data.setVertexCoords( ctx, vert[0], act );

      for ( i = 0; i < 3; i++ )
        act[i] = center[i] - blade[i] - offset[i];
      data.setVertexCoords( ctx, vert[1], act );

      if ( painter || transparent )
      {                                   // two small quadrangles
        for ( i = 0; i < 3; i++ )
          act[i] = center[i] - blEps[i] - offset[i];
        data.setVertexCoords( ctx, vert[2], act );

        for ( i = 0; i < 3; i++ )
          act[i] = target[i] - blEps[i] - offset[i];
        data.setVertexCoords( ctx, vert[3], act );

          // topology - quadrilateral face:
        randomColor( colorF, true );
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );

          // 2nd quadrangle:
        for ( i = 0; i < 4; i++ )
        {
          vert[i] = data.createVertex();
            // vertex color:
          data.setAttribute( vertexColor,  vert[i], color );
            // vertex normal:
          data.setAttribute( vertexNormal, vert[i], norm );
            // texture coordinates:
          data.setAttribute( vertexTex,    vert[i], tex[3-i] );
        }

        for ( i = 0; i < 3; i++ )
          act[i] = target[i] + blEps[i] - offset[i];
        data.setVertexCoords( ctx, vert[0], act );

        for ( i = 0; i < 3; i++ )
          act[i] = center[i] + blEps[i] - offset[i];
        data.setVertexCoords( ctx, vert[1], act );

        for ( i = 0; i < 3; i++ )
          act[i] = center[i] + blade[i] - offset[i];
        data.setVertexCoords( ctx, vert[2], act );

        for ( i = 0; i < 3; i++ )
          act[i] = target[i] + blade[i] - offset[i];
        data.setVertexCoords( ctx, vert[3], act );

          // topology - quadrilateral face:
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );

          // add four small triangles at the blade thirds
        permuteColor( color );
        permuteColor( colorF );

        for ( i = 0; i < 3; i++ )
        {
          vert[i] = data.createVertex();
            // vertex color:
          data.setAttribute( vertexColor,  vert[i], color );
            // vertex normal:
          data.setAttribute( vertexNormal, vert[i], norm );
            // texture coordinates:
          data.setAttribute( vertexTex,    vert[i], tex[i] );
        }
        vert[3] = Brep.NULL;

        for ( i = 0; i < 3; i++ )
          add[i] = 0.5 * (target[i] + center[i]) + 0.35 * blade[i] - 6.0 * offset[i];

        data.setVertexCoords( ctx, vert[0], add );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] + axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[1], act );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] - axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[2], act );

          // topology - triangle face:
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );

          // 2nd additional triangle:
        for ( i = 0; i < 3; i++ )
        {
          vert[i] = data.createVertex();
            // vertex color:
          data.setAttribute( vertexColor,  vert[i], color );
            // vertex normal:
          data.setAttribute( vertexNormal, vert[i], norm );
            // texture coordinates:
          data.setAttribute( vertexTex,    vert[i], tex[i] );
        }

        for ( i = 0; i < 3; i++ )
          add[i] = 0.5 * (target[i] + center[i]) + 0.75 * blade[i] - 6.0 * offset[i];

        data.setVertexCoords( ctx, vert[0], add );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] + axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[1], act );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] - axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[2], act );

          // topology - triangle face:
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );

          // 3rd additional triangle:
        for ( i = 0; i < 3; i++ )
        {
          vert[i] = data.createVertex();
            // vertex color:
          data.setAttribute( vertexColor,  vert[i], color );
            // vertex normal:
          data.setAttribute( vertexNormal, vert[i], norm );
            // texture coordinates:
          data.setAttribute( vertexTex,    vert[i], tex[i] );
        }

        for ( i = 0; i < 3; i++ )
          add[i] = 0.5 * (target[i] + center[i]) - 0.25 * blade[i] - 6.0 * offset[i];

        data.setVertexCoords( ctx, vert[0], add );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] + axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[1], act );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] - axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[2], act );

          // topology - triangle face:
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );

          // 4th additional triangle:
        for ( i = 0; i < 3; i++ )
        {
          vert[i] = data.createVertex();
            // vertex color:
          data.setAttribute( vertexColor,  vert[i], color );
            // vertex normal:
          data.setAttribute( vertexNormal, vert[i], norm );
            // texture coordinates:
          data.setAttribute( vertexTex,    vert[i], tex[i] );
        }

        for ( i = 0; i < 3; i++ )
          add[i] = 0.5 * (target[i] + center[i]) - 0.65 * blade[i] - 6.0 * offset[i];

        data.setVertexCoords( ctx, vert[0], add );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] + axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[1], act );

        for ( i = 0; i < 3; i++ )
          act[i] = add[i] - axis[i] * ADD_RADIUS - 0.1 * blade[i];
        data.setVertexCoords( ctx, vert[2], act );

          // topology - triangle face:
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );
      }

      else
      {                                   // one long (penetrating) quadrangle
        for ( i = 0; i < 3; i++ )
          act[i] = center[i] + blade[i] - offset[i];
        data.setVertexCoords( ctx, vert[2], act );

        for ( i = 0; i < 3; i++ )
          act[i] = target[i] + blade[i] - offset[i];
        data.setVertexCoords( ctx, vert[3], act );

          // topology - quadrilateral face:
        randomColor( colorF, true );
        data.setFaceVertices( face = data.createFace(), vert );
        data.setAttribute( faceColor, face, colorF );
      }

      angle += dAngle;
    }
    while ( --count > 0 );

  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Prepares 3D scene (attached {@link Brep} object).
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    final RealFunction func = (RealFunction)getInterface( PL_FUNCTION, IFACE+"RealFunction" );
    final Brep         data = (Brep)        getInterface( PL_OUTPUT,   IFACE+"Brep" );

    if ( data == null )
      return false;

      // B-rep related data:
    int      ctx   = Brep.CTX_DEFAULT;
    int[][]  vert  = new int[ cellsY + 1 ][ cellsX + 1 ];

      // geometric data:
    double x, y;
    double stepX   = (maxX - minX) / cellsX;
    double stepY   = (maxY - minY) / cellsY;
    double[] coord = { 0.0, 0.0, 0.0, 1.0 }; // world coordinate of the vertex
    double[] der   = new double[ 2 ];        // partial derivatives
    double[] tx    = { 1.0, 0.0, 0.0, 0.0 }; // tangent vector along X-axis
    double[] ty    = { 0.0, 1.0, 0.0, 0.0 }; // tangent vector along Y-axis
    double[] norm  = new double[ 4 ];        // normal vector
    float[] color  = new float[ transparent ? 4 : 3 ]; // vertex color
    double[] tex   = new double[ 2 ];        // texture coordinates
    final double dU = 1.0 / cellsX;
    final double dV = 1.0 / cellsY;
    int i, j, face;
    double ramp;
    if ( transparent ) color[3] = 1.0f;

      // random generator:
    if ( seed >= 0 )
      rnd = new RandomJames( seed >>> 15, seed & 0x7fff );
    else
    {
      rnd = new RandomJames();
      rnd.randomize();
    }

      // B-rep setup:
    data.setCoordTypes( Brep.ATTR_COORDH_4D, Brep.ATTR_COORDH_4D );
    data.init();
    int vertexColor  = data.createAttribute( Brep.VERTEX, Brep.NULL, Brep.COLOR,     transparent ? Brep.ATTR_RGBA_4F : Brep.ATTR_RGB_3F );
    int faceColor    = data.createAttribute( Brep.FACE,   Brep.NULL, Brep.COLOR,     transparent ? Brep.ATTR_RGBA_4F : Brep.ATTR_RGB_3F );
    int vertexNormal = data.createAttribute( Brep.VERTEX, Brep.NULL, Brep.NORMAL,    Brep.ATTR_COORDH_4D );
    int vertexTex    = data.createAttribute( Brep.VERTEX, Brep.NULL, Brep.TEXTURE_0, Brep.ATTR_COORD_2D );

    if ( func != null )
    {
        // 3D graph geometry:
      for ( y = minY, i = 0, tex[1] = 0.0;
            i <= cellsY;
            i++, y += stepY, tex[1] += dV ) // one row
        for ( x = minX, j = 0, tex[0] = 0.0;
              j <= cellsX;
              j++, x += stepX, tex[0] += dU ) // one column
        {
            // vertex coordinate:
          coord[0]   = x;
          coord[1]   = y;
          coord[2]   = func.f( x, y );
          vert[i][j] = data.createVertex();
          data.setVertexCoords( ctx, vert[i][j], coord );
            // vertex color:
          ramp = Formula.remainder( (coord[2] * colorMul), 2.0 );
          if ( ramp > 1.0 ) ramp = 2.0 - ramp;
          Formula.colorRamp( ramp, color );
          data.setAttribute( vertexColor, vert[i][j], color );
            // vertex normal vector:
          func.df( x, y, der );
          tx[2] = der[0];
          ty[2] = der[1];
          Geometry.vectorProduct( tx, ty, norm );
          Geometry.normalize3D( norm );
          data.setAttribute( vertexNormal, vert[i][j], norm );
          data.setAttribute( vertexTex,    vert[i][j], tex );
        }

        // 3D graph topology (triangle mesh):
      for ( i = 0; i < cellsY; i++ )
        for ( j = 0; j < cellsX; j++ )
        {                                 // one quad
          randomColor( color, false );
          data.setTriangleVertices( face = data.createFace(), vert[i  ][j  ], vert[i  ][j+1], vert[i+1][j  ] );
          data.setAttribute( faceColor, face, color );
          data.setTriangleVertices( face = data.createFace(), vert[i+1][j  ], vert[i  ][j+1], vert[i+1][j+1] );
          data.setAttribute( faceColor, face, color );
        }
    }

      // generate random fans:
    for ( i = 0; i++ < fans; )
      fan( data, ctx, vertexColor, faceColor, vertexNormal, vertexTex );

    return true;
  }

  /**
   * Module initialization - called before any {@link Worker#run}.
   * Individual phases (starting at {@code 0}) will be called until maximum
   * requested phase number is reached.
   *
   * @param  phase Current phase number.
   * @return Number of phases needed to initialize this module.
   */
  @Override
  public int init ( int phase )
  {
    if ( isConnected( PL_INPUT ) )
      return 0;

    if ( phase == 0 )
      fire( 0 );

    return 1;
  }

  //==========================================================================
  //  Interface Property:

  public final static String CELLS_X     = "Cells X";
  public final static String CELLS_Y     = "Cells Y";
  public final static String MIN_X       = "Min X";
  public final static String MAX_X       = "Max X";
  public final static String MIN_Y       = "Min Y";
  public final static String MAX_Y       = "Max Y";
  public final static String COLOR_MUL   = "Color mul";
  public final static String FANS        = "Fans";
  public final static String RADIUS      = "Radius";
  public final static String SEED        = "Seed";
  public final static String PAINTER     = "Non-penetrating";
  public final static String TRANSPARENT = "Transparent";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // CELLS_X:
    if ( key.compareTo(CELLS_X) == 0 )
    {
      cellsX = intProperty( value, cellsX, 1, 1000 );
      return;
    }
      // CELLS_Y:
    if ( key.compareTo(CELLS_Y) == 0 )
    {
      cellsY = intProperty( value, cellsY, 1, 1000 );
      return;
    }
      // MIN_X:
    if ( key.compareTo(MIN_X) == 0 )
    {
      minX = doubleProperty( value, minX );
      return;
    }
      // MAX_X:
    if ( key.compareTo(MAX_X) == 0 )
    {
      maxX = doubleProperty( value, maxX );
      return;
    }
      // MIN_Y:
    if ( key.compareTo(MIN_Y) == 0 )
    {
      minY = doubleProperty( value, minY );
      return;
    }
      // MAX_Y:
    if ( key.compareTo(MAX_Y) == 0 )
    {
      maxY = doubleProperty( value, maxY );
      return;
    }
      // COLOR_MUL:
    if ( key.compareTo(COLOR_MUL) == 0 )
    {
      colorMul = doubleProperty( value, colorMul );
      return;
    }
      // FANS:
    if ( key.compareTo(FANS) == 0 )
    {
      fans = intProperty( value, fans );
      return;
    }
      // RADIUS:
    if ( key.compareTo(RADIUS) == 0 )
    {
      maxRadius = doubleProperty( value, maxRadius );
      return;
    }
      // SEED:
    if ( key.compareTo(SEED) == 0 )
    {
      seed = intProperty( value, seed );
      return;
    }
      // TRANSPARENT:
    if ( key.compareTo(TRANSPARENT) == 0 )
    {
      transparent = booleanProperty( value, transparent );
      return;
    }
      // PAINTER:
    if ( key.compareTo(PAINTER) == 0 )
    {
      painter = booleanProperty( value, painter );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(CELLS_X) == 0 )
      return cellsX;
    if ( key.compareTo(CELLS_Y) == 0 )
      return cellsY;
    if ( key.compareTo(MIN_X) == 0 )
      return minX;
    if ( key.compareTo(MAX_X) == 0 )
      return maxX;
    if ( key.compareTo(MIN_Y) == 0 )
      return minY;
    if ( key.compareTo(MAX_Y) == 0 )
      return maxY;
    if ( key.compareTo(COLOR_MUL) == 0 )
      return colorMul;
    if ( key.compareTo(FANS) == 0 )
      return fans;
    if ( key.compareTo(RADIUS) == 0 )
      return maxRadius;
    if ( key.compareTo(SEED) == 0 )
      return seed;
    if ( key.compareTo(PAINTER) == 0 )
      return painter;
    if ( key.compareTo(TRANSPARENT) == 0 )
      return transparent;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Sample 3D scene";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToFunctionR2ToR";

  /** Object description. */
  private final static String DESCRIPTION = "Creates sample 3D scene (B-rep).";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_DATA + '.' + C_SCENE;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,    IFACE+"Trigger" );
      t.newOptOutputPlug( PL_FUNCTION, IFACE+"RealFunction" );
      t.newOutputPlug(    PL_OUTPUT,   IFACE+"Brep" );
        // Property:
      t.propBegin( CELLS_X, TYPE_INTEGER, "Cells X", true );
        t.propDefault( 20 );
      t.propEnd();
      t.propBegin( CELLS_Y, TYPE_INTEGER, "Cells Y", true );
        t.propDefault( 20 );
      t.propEnd();
      t.propBegin( MIN_X, TYPE_DOUBLE, "Min X", true );
        t.propDefault( -5.0 );
      t.propEnd();
      t.propBegin( MAX_X, TYPE_DOUBLE, "Max X", true );
        t.propDefault( 5.0 );
      t.propEnd();
      t.propBegin( MIN_Y, TYPE_DOUBLE, "Min Y", true );
        t.propDefault( -5.0 );
      t.propEnd();
      t.propBegin( MAX_Y, TYPE_DOUBLE, "Max Y", true );
        t.propDefault( 5.0 );
      t.propEnd();
      t.propBegin( COLOR_MUL, TYPE_DOUBLE, "Color multiplier for vertex colors", true );
        t.propDefault( 0.2 );
      t.propEnd();
      t.propBegin( FANS, TYPE_INTEGER, "Number of random fans", true );
        t.propDefault( 10 );
      t.propEnd();
      t.propBegin( RADIUS, TYPE_DOUBLE, "Max fan radius", true );
        t.propDefault( 3.0 );
      t.propEnd();
      t.propBegin( SEED, TYPE_INTEGER, "Random generator seed (-1 for randomize)", true );
        t.propDefault( -1 );
      t.propEnd();
      t.propBegin( PAINTER, TYPE_BOOLEAN, "Generate non-penetrating faces?", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( TRANSPARENT, TYPE_BOOLEAN, "Generate transparent colors?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
