package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.piece.*;

/**
 * Test image specimen - gamma tester.
 *
 * @version 0.26 $Rev: 453 $ $Date: 2007-05-14 21:50:00 +0200 (po, 14 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/TestGamma.java">TestGamma.java</a>
 */
public class TestGamma extends DefaultImageSynthesizer
{
  //==========================================================================
  //  Data:

  /** Actual output color mode. */
  protected int mode = RasterGraphics.MODE_RGB;

  /** Accelerator: output image. */
  protected RasterGraphics out;

  /** External module: brush for alpha-filling. */
  protected SolidColorBrush brush;

  /** External module: alpha-mask. */
  protected AlphaMatrix alpha;

  /** External module: text drawing. */
  protected JavaText text;

  //==========================================================================
  //  Support:

  /**
   * Assert support modules and accelerators.
   *
   * @return {@code true} if succeeded.
   */
  protected boolean assertModules ()
  {
      // raster output:
    out = (RasterGraphics)getInterface( PL_OUTPUT, IFACE+"RasterGraphics" );
    if ( out == null ) return false;

    if ( out.getWidth()  != imageWidth ||
         out.getHeight() != imageHeight ||
         out.getMode()   != mode )
      out.init( imageWidth, imageHeight, mode, 0 );

      // drawing support:
    alpha = new AlphaMatrix();
    alpha.init( imageWidth, imageHeight );
    brush = new SolidColorBrush();
    text  = new JavaText();
    try
    {
      text.connect(  Piece.PL_ALPHAMASK, alpha,      Piece.PL_INPUT );
      brush.connect( Piece.PL_ALPHAMASK, alpha,      Piece.PL_INPUT );
      brush.connect( Piece.PL_OUTPUT,    (Piece)out, Piece.PL_INPUT );
    }
    catch ( BadInterfaceException e )
    {
      LogFile.exception( "Error connecting brush, text, alpha-mask and raster image!", e );
      return false;
    }

    text.set( TextRender.ANTI_MODE, (mode == RasterGraphics.MODE_RGBA) ?
              TextRender.ANTI_ON : TextRender.ANTI_OFF );

    return true;
  }

  /** Draw one vertical unit (7 vertical columns). */
  protected void verticalUnit ( int x0, int y0, int width, int height,
                                double[] color0, double[] color1 )
  {
    if ( userBreak ) return;

    int i, j, b;
    final int bands = color0.length;

      // column bounds:
    final int hUnit = width / 8;
    final int x1    = x0 + hUnit;
    final int x2    = x1 + hUnit;
    final int x3    = x2 + hUnit;
    final int x7    = x0 + width;
    final int x6    = x7 - hUnit;
    final int x5    = x6 - hUnit;
    final int x4    = x5 - hUnit;

      // vignette:
    double[] c  = Formula.cloneArray( color0 );
    double[] dc = new double[ bands ];
    double mul = 1.0 / (height - 1.0);
    for ( b = 0; b < bands; b++ )
      dc[b] = (color1[b] - color0[b]) * mul;

    for ( i = y0; i < y0 + height && !userBreak; i++ )
    {
      for ( j = x1; j < x6; j++ )
        out.putPixel( j, i, c );
      for ( b = 0; b < bands; b++ )
        c[b] += dc[b];
    }

      // 25% value dithering:
    for ( i = y0; i < y0 + height && !userBreak; i++ )
      if ( (i & 1) == 0 )
        out.setHLine( x0, x1, i, color0 );
      else
        for ( j = x0; j < x1; j++ )
          out.putPixel( j, i, ((j & 1) == ((i >> 1) & 1)) ? color0 : color1 );

      // 50% value dithering (checkerboard):
    for ( i = y0; i < y0 + height && !userBreak; i++ )
      for ( j = x2; j < x3; j++ )
        out.putPixel( j, i, ((j & 1) == (i & 1)) ? color0 : color1 );

      // 50% value dithering (hlines):
    for ( i = y0; i < y0 + height && !userBreak; i++ )
      for ( j = x4; j < x5; j++ )
        out.putPixel( j, i, ((i & 1) == 0) ? color0 : color1 );

      // 75% value dithering:
    for ( i = y0; i < y0 + height && !userBreak; i++ )
      if ( (i & 1) == 0 )
        out.setHLine( x6, x7, i, color1 );
      else
        for ( j = x6; j < x7; j++ )
          out.putPixel( j, i, ((j & 1) == ((i >> 1) & 1)) ? color0 : color1 );

      // scale:
    TextRender.TextExtent ext = new TextRender.TextExtent();
    text.createFont( "Arial", 0, height / 40.0 );
    text.textExtent( 0.0, 0.0, "0.05", ext );
    double x;
    double dy = -0.5 * (ext.uly + ext.lry);
    double y  = y0 + 0.05 * height;

    for ( i = 1; i < 20 && !userBreak; i++, y += 0.05 * height )
    {
      String val = String.format( "%.2f", (0.05 * i) );
      text.textExtent( 0.0, 0.0, val, ext );
      x = x0 + width * 0.5 - 0.5 * (ext.ulx + ext.lrx);
      text.drawText( x, y + dy, val, null );
      brush.setColor( (i < 10) ? color1 : color0 );
      brush.fill();
    }
  }

  //==========================================================================
  //  Interface ImageSynthesizer:

  /**
   * Render the given rectangle.
   * Output is written to some connected raster device (not defined in this interface).
   *
   * @param  x1 X coordinate of the upper left corner.
   * @param  y1 Y coordinate of the upper left corner.
   * @param  x2 X coordinate of the lower right corner (outside pixel).
   * @param  y2 Y coordinate of the upper left corner (outside pixel).
   * @see    #renderPixel
   * @see    #renderAll
   */
  public void renderRectangle ( int x1, int y1, int x2, int y2 )
  {
    if ( !assertModules() ) return;       // failure

    int rWidth  = x2 - x1;
    int rHeight = y2 - y1;

    switch ( mode )
    {
      case RasterGraphics.MODE_GRAY:      // one unit
      {
        double[] c0 = { 0.0 };
        double[] c1 = { 1.0 };
        verticalUnit( x1, y1, rWidth, rHeight, c0, c1 );
        break;
      }

      case RasterGraphics.MODE_RGB:       // four units: R, G, B, W
      {
        int width = rWidth / 4;
        double[] black = { 0.0, 0.0, 0.0 };
        double[] c1    = { 1.0, 0.0, 0.0 };
        verticalUnit( x1,             y1, width, rHeight, black, c1 );
        c1[0] = 0.0;
        c1[1] = 1.0;
        verticalUnit( x1 +     width, y1, width, rHeight, black, c1 );
        c1[1] = 0.0;
        c1[2] = 1.0;
        verticalUnit( x1 + 2 * width, y1, width, rHeight, black, c1 );
        c1[0] = 1.0;
        c1[1] = 1.0;
        verticalUnit( x1 + 3 * width, y1, rWidth - 3 * width, rHeight, black, c1 );
        break;
      }

      case RasterGraphics.MODE_RGBA:      // five units: R, G, B, A, W
      {
        int width = imageWidth / 5;
        double[] black = { 0.0, 0.0, 0.0, 1.0 };
        double[] c1    = { 1.0, 0.0, 0.0, 1.0 };
        verticalUnit( x1,             y1, width, rHeight, black, c1 );
        c1[0] = 0.0;
        c1[1] = 1.0;
        verticalUnit( x1 +     width, y1, width, rHeight, black, c1 );
        c1[1] = 0.0;
        c1[2] = 1.0;
        verticalUnit( x1 + 2 * width, y1, width, rHeight, black, c1 );
        black[3] = 0.0;
        c1[0] = 1.0;
        c1[1] = 1.0;
        verticalUnit( x1 + 3 * width, y1, width, rHeight, black, c1 );
        black[3] = 1.0;
        verticalUnit( x1 + 4 * width, y1, rWidth - 4 * width, rHeight, black, c1 );
        break;
      }
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String COLORMODE = "Color mode";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // COLORMODE:
    if ( key.compareTo(COLORMODE) == 0 )
    {
      mode = intProperty( value, mode, RasterGraphics.MODE_GRAY, RasterGraphics.MODE_RGBA );
      if ( mode == RasterGraphics.MODE_COLORMAP )
        mode = RasterGraphics.MODE_RGB;
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(COLORMODE) == 0 )
      return mode;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "TestGamma";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "ImageSynthesizerToRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Test image specimen - gamma tester.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_IMAGE;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"ImageSynthesizer" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( COLORMODE, TYPE_INTEGER, "Output color mode", true );
        t.propDefault( RasterGraphics.MODE_RGB );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "RGB",  RasterGraphics.MODE_RGB,  "RGB color mode" );
        t.propEnum( "RGBA", RasterGraphics.MODE_RGBA, "RGBA color mode (with transparency)" );
        t.propEnum( "Gray", RasterGraphics.MODE_GRAY, "Gray mode" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
