package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

import cz.cuni.jagrlib.worker.TriangleWarpingWorker;

/**
 * Triangle-mesh warping: sample {@code R^2 -> R^2} mapping using triangle mesh stored in {@link Brep}.
 *
 * @version 0.26 $Rev: 431 $ $Date: 2007-04-29 22:42:22 +0200 (ne, 29 IV 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/TriMeshDeformation.java">TriMeshDeformation.java</a>
 */
public class TriMeshDeformation extends DefaultFunctionR2ToR2
{
  //==========================================================================
  //  Data:

  /** Id of the 2nd context (containing target mesh coordinates). */
  protected int ctx2 = 1;

  protected final int[] vertices = new int[ 3 ];

  protected final double[] p     = new double[ 2 ];
  protected final double[] a     = new double[ 2 ];
  protected final double[] b     = new double[ 2 ];
  protected final double[] c     = new double[ 2 ];
  protected final double[] uv    = new double[ 2 ];

  protected int     lastFace     = -1;
  protected boolean lastForward  = true;

  protected Brep    data;

  //==========================================================================
  //  Support:

  /**
   * Compute domain/range bounds of triangle mesh.
   *
   * @param  ctx Context id of space to be used (source/target space).
   * @param  b Bounds to be filled (xMin, xMax, yMin, yMax).
   * @return True if bounds were computed.
   */
  protected boolean computeBounds ( int ctx, double[] b )
  {
    if ( b == null || b.length < 4 ) return false;

    data = (Brep)getInterface( PL_DATA, IFACE+"Brep" );
    if ( data == null ) return false;

    BrepIterator it = data.vertexIterator( null );
    int v = it.next();
    if ( v == Brep.NULL ) return false;

    double[] c = new double[ 2 ];
    double xMin, xMax, yMin, yMax;
    int count = 1;
    data.getVertexCoords( ctx, v, c );
    xMin = xMax = c[0];
    yMin = yMax = c[1];

    while ( (v = it.next()) != Brep.NULL )
    {
      count++;
      data.getVertexCoords( ctx, v, c );
      if ( c[0] < xMin ) xMin = c[0];
      else
      if ( c[0] > xMax ) xMax = c[0];
      if ( c[1] < yMin ) yMin = c[1];
      else
      if ( c[1] > yMax ) yMax = c[1];
    }

    b[0] = xMin; b[1] = xMax;
    b[2] = yMin; b[3] = yMax;
    return true;
  }

  protected boolean testLastFace ( int ctxFrom, int ctxTo, double[] res )
  {
    if ( lastFace >= 0 && data.getFaceVertices( lastFace, vertices ) == 3 )
    {
      data.getVertexCoords( ctxFrom, vertices[0], a );
      data.getVertexCoords( ctxFrom, vertices[1], b );
      data.getVertexCoords( ctxFrom, vertices[2], c );

      if ( Geometry.pointInTriangle2D( p, a, b, c, uv ) != null )
      {
        data.getVertexCoords( ctxTo, vertices[0], a );
        data.getVertexCoords( ctxTo, vertices[1], b );
        data.getVertexCoords( ctxTo, vertices[2], c );

        final double w = 1.0 - uv[0] - uv[1];
        res[0] = a[0] * w + b[0] * uv[0] + c[0] * uv[1];
        res[1] = a[1] * w + b[1] * uv[0] + c[1] * uv[1];
        return true;
      }
    }

    return false;
  }

  /**
   * Find triangle in which the given point {@code [ x, y ]} lies.
   */
  protected void map ( boolean forward, double x, double y, double[] res )
  {
    if ( res == null || res.length < 2 ) return;

    int ctxFrom, ctxTo;
    if ( forward )
    {
      ctxFrom = Brep.CTX_DEFAULT;
      ctxTo   = ctx2;
    }
    else
    {
      ctxFrom = ctx2;
      ctxTo   = Brep.CTX_DEFAULT;
    }

    if ( data == null )
      data = (Brep)getInterface( PL_DATA, IFACE+"Brep" );

    p[0] = x;
    p[1] = y;

    if ( forward == lastForward )
    {
      if ( testLastFace( ctxFrom, ctxTo, res ) ) return;
    }
    else
      forward = lastForward;

    BrepIterator it = data.faceIterator( Brep.CTX_DEFAULT, null );

    while ( (lastFace = it.next()) != Brep.NULL )
      if ( testLastFace( ctxFrom, ctxTo, res ) ) return;

    res[0] = x;
    res[1] = y;
  }

  //==========================================================================
  //  Interface FunctionR2ToR2:

  /**
   * Compute the mapping: <b>f(x,y)</b>.
   *
   * @param  x X coordinate (1st argument).
   * @param  y Y coordinate (2nd argument).
   * @param  res Result array ({@code res[0]} = U coordinate,
   *             {@code res[1]} = V coordinate).
   */
  @Override
  public void f ( double x, double y, double[] res )
  {
    map( true, x, y, res );
  }

  /**
   * Compute the inverse mapping: <b>f<sup>-1</sup>(u,v)</b>.
   *
   * @param  u U coordinate (1st argument).
   * @param  v V coordinate (2nd argument).
   * @param  res Result array ({@code res[0]} = X coordinate,
   *             {@code res[1]} = Y coordinate).
   * @return {@code true} if this service is implemented.
   */
  @Override
  public boolean fInv ( double u, double v, double[] res )
  {
    map( false, u, v, res );
    return true;
  }

  /**
   * Returns domain of the mapping.
   *
   * @param  dom Domain array ({@code dom[0]} = Xmin, {@code dom[1]} = Xmax,
   *             {@code dom[2]} = Ymin, {@code dom[3]} = Ymax).
   * @return {@code true} if this service is implemented.
   */
  @Override
  public boolean getDomain ( double[] dom )
  {
    return computeBounds( Brep.CTX_DEFAULT, dom );
  }

  /**
   * Returns range of the mapping.
   *
   * @param  rng Range array ({@code rng[0]} = Umin, {@code rng[1]} = Umax,
   *             {@code rng[2]} = Vmin, {@code rng[3]} = Vmax).
   * @return {@code true} if this service is implemented.
   */
  @Override
  public boolean getRange ( double[] rng )
  {
    return computeBounds( ctx2, rng );
  }

  //==========================================================================
  //  Interface Property:

  protected final static String CONTEXT2 = TriangleWarpingWorker.CONTEXT2;

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // CONTEXT2:
    if ( key.compareTo(CONTEXT2) == 0 )
    {
      ctx2 = intProperty( value, ctx2 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(CONTEXT2) == 0 )
      return ctx2;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "TriMeshDeformation";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "FunctionR2ToR2ToBrep";

  /** Object description. */
  private final static String DESCRIPTION = "Triangle-mesh warping deformation function.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_FUNCTION;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"FunctionR2ToR2" );
      t.newOutputPlug( PL_DATA,  IFACE+"Brep" );
        // Property:
      t.propBegin( CONTEXT2, TYPE_INTEGER, "Context 2", false );
        t.propDefault( 1 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
