package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;
import cz.cuni.jagrlib.worker.TriangleWarpingWorker;

/**
 * Interactive triangle-mesh editor specimen.
 *
 * @version 0.25 $Rev: 406 $ $Date: 2007-03-25 00:06:58 +0100 (ne, 25 III 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/TriMeshEditor.java">TriMeshEditor.java</a>
 */
public class TriMeshEditor extends DefaultInputListener implements Trigger
{
  //==========================================================================
  //  Data:

  /** GUI actions/states. */
  protected static enum State
  {
    /** Nothing-to-do state. */
    NONE,
    /** Moving a vertex. */
    MOVING,
    /** Finish vertex moving (repaint?). */
    FINISH,
    /** Loop break (will destroy the window). */
    BREAK,
    /** Loop quit (the window was destroyed before). */
    QUIT,
  }

    // Accelerators & state:

  /** Preview window for input image. */
  protected GraphicsViewer iPreview = null;

  /** Preview window for output image. */
  protected GraphicsViewer oPreview = null;

  /** B-rep database. */
  protected Brep brep = null;

  /** Context2 used in B-rep (output triangle mesh). */
  protected int ctx2;

  /** Number of triangle rows (vertical). */
  protected int rows;

  /** Number of triangle columns (horizontal). */
  protected int columns;

  /** Image width in pixels. */
  protected int width;

  /** Image height in pixels. */
  protected int height;

  /** Actual action/state. */
  protected State state = State.NONE;

  /** New mouse coordinates. */
  protected int nx, ny;

  /** Output preview pane is active? */
  protected boolean outputActive;

  //==========================================================================
  //  Support:

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transformation.
   *
   * @param  type The action type (not used).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
      // initialization:
                         brep     = (Brep)          getInterface( PL_DATA,    IFACE+"Brep" );
                         iPreview = (GraphicsViewer)getInterface( "view1",    IFACE+"GraphicsViewer" );
                         oPreview = (GraphicsViewer)getInterface( "view2",    IFACE+"GraphicsViewer" );
    final RasterGraphics output   = (RasterGraphics)getInterface( PL_OUTPUT,  IFACE+"RasterGraphics" );
    final Trigger        warping  = (Trigger)       getInterface( PL_TRIGGER, IFACE+"Trigger" );

    if ( brep     == null ||
         oPreview == null ||
         output   == null ||
         warping  == null ) return false;

    width  = output.getWidth();
    height = output.getHeight();

    // !!! TODO: initialize HUDs for both preview windows !!!

      // interaction:
    do
    {
        // local copies of state variables (to minimize the critical section):
      State cState;
      int cx, cy;

      synchronized ( this )
      {
        try
        {
          wait();
        }
        catch ( InterruptedException e )
        {
          state = State.BREAK;
        }

          // copy state/action into cState and change the new state..
        cState = state;

          // copy actual coordinates into c*:
        cx = nx;
        cy = ny;
      }

      switch ( cState )
      {
        case MOVING:
          // !!! TODO: in the middle of vertex-moving => change and repaint HUD !!!
          oPreview.repaint();
          break;

        case FINISH:
          // !!! TODO: vertex-moving was finished => repaint result image !!!
          warping.fire( 0 );
          oPreview.repaint();
          break;

        case BREAK:
          if ( iPreview != null )
            iPreview.destroy();
          oPreview.destroy();

        case QUIT:
          return true;
      }
    }
    while ( true );
  }

  /**
   * Mouse button was pressed/released.
   *
   * @param sender Originator of the event
   * @param when System time in milliseconds.
   * @param x Horizontal screen position in pixels.
   * @param y Vertical screen position in pixels.
   * @param button Button number ({@link #BUTTON1}, etc.).
   * @param press Was it press or release?
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public void mouseButton ( Object sender, long when, int x, int y, int button, boolean press, int flags )
  {
    if ( button != BUTTON1 ) return;

    synchronized ( this )
    {
      if ( press )
      {
        outputActive = (sender == oPreview);
        // !!! TODO: find moved vertex (input/output), start moving process.. !!!
        state = State.MOVING;
        notify();
      }
      else
        if ( state == State.MOVING )
        {
          // !!! TODO: finishing vertex-moving.. !!!
          state = State.FINISH;
          notify();
        }
    }
  }

  /**
   * Keyboard key was pressed/released.
   *
   * @param sender Originator of the event
   * @param when System time in milliseconds.
   * @param press Was it press or release?
   * @param ch Character represented by the key.
   * @param keyCode Key code.
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public void keyboard ( Object sender, long when, boolean press, char ch, int keyCode, int flags )
  {
    if ( ch == 'q' )                         // break
    synchronized ( this )
    {
      state = State.BREAK;
      notify();
    }
  }

  /**
   * Mouse pointer was moved.
   *
   * @param sender Originator of the event
   * @param when System time in milliseconds.
   * @param x Horizontal screen position in pixels.
   * @param y Vertical screen position in pixels.
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public synchronized void mouseMove ( Object sender, long when, int x, int y, int flags )
  {
    if ( state == State.MOVING )
    {
      // !!! TODO: change active HUD, invalidate the changed region.. !!!
      notify();
    }
  }

  /**
   * Window was closed (user interaction..).
   *
   * @param sender Originator of the event
   */
  @Override
  public synchronized void windowClose ( Object sender )
  {
    state = State.QUIT;
    notify();
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // CONTEXT2:
    if ( key.compareTo(TriangleWarpingWorker.CONTEXT2) == 0 )
    {
      ctx2 = intProperty( value, ctx2 );
      return;
    }
      // MESH_ROWS:
    if ( key.compareTo(TriangleWarpingWorker.MESH_ROWS) == 0 )
    {
      rows = intProperty( value, rows, 1, 1000 );
      return;
    }
      // MESH_COLUMNS:
    if ( key.compareTo(TriangleWarpingWorker.MESH_COLUMNS) == 0 )
    {
      columns = intProperty( value, columns, 1, 1000 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    if ( key.compareTo(TriangleWarpingWorker.CONTEXT2) == 0 )
      return ctx2;
    if ( key.compareTo(TriangleWarpingWorker.MESH_ROWS) == 0 )
      return rows;
    if ( key.compareTo(TriangleWarpingWorker.MESH_COLUMNS) == 0 )
      return columns;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "TriMeshEditor";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerAndInputListenerToGraphicsViewer2AndBrep";

  /** Object description. */
  private final static String DESCRIPTION = "Triangle mesh editor specimen.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_VECTOR + "." + C_EDITOR;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,   IFACE+"Trigger" );
      t.newInputPlug(     PL_DIRECT,  IFACE+"InputListener" );
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"RasterGraphics" );
      t.newOptOutputPlug( "view1",    IFACE+"GraphicsViewer" );
      t.newOutputPlug(    "view2",    IFACE+"GraphicsViewer" );
      t.newOutputPlug(    PL_DATA,    IFACE+"Brep" );
      t.newOutputPlug(    PL_TRIGGER, IFACE+"Trigger" );
        // Property:
      t.propBegin( TriangleWarpingWorker.CONTEXT2, TYPE_INTEGER, "Context 2", false );
      t.propEnd();
      t.propBegin( TriangleWarpingWorker.MESH_ROWS, TYPE_INTEGER, "Mesh rows", false );
        t.propBounds( 1, 1000 );
      t.propEnd();
      t.propBegin( TriangleWarpingWorker.MESH_COLUMNS, TYPE_INTEGER, "Mesh columns", false );
        t.propBounds( 1, 1000 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
