package cz.cuni.jagrlib.testing;

import java.util.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Uniform grid for fast searching.
 *
 * @version 0.26 $Rev: 495 $ $Date: 2007-07-22 01:15:17 +0200 (ne, 22 VII 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/UniformGrid.java">UniformGrid.java</a>
 */
public class UniformGrid extends DefaultGeometrySearch
{
  //==========================================================================
  //  Support classes:

  /** Support class for grid-pass. */
  protected static class GridHeapItem implements Comparable< GridHeapItem >
  {
    double distance;

    /** Object (vertex, edge) id - if the item holds an object (otherwise it is {@link Brep#NULL NULL}. */
    int id;

    /** Cell-index - if the item holds a cell (otherwise it is {@link Brep#NULL NULL}). */
    int cellIndex;

    /** Initializes neutral heap-item. */
    GridHeapItem ( double dist )
    {
      distance  = dist;
      id        =
      cellIndex = Brep.NULL;
    }

    /** Natural ordering [ distance, {id | cellIndex} ]. */
    public final int compareTo ( GridHeapItem b )
    {
      if ( id        == b.id &&
           cellIndex == b.cellIndex ) return 0;

      if ( distance < b.distance ) return -1;
      if ( distance > b.distance ) return  1;
      if ( id       < b.id ) return -1;
      if ( id       > b.id ) return  1;
      return( cellIndex < b.cellIndex ? -1 : 1 );
    }

    /** Natural ordering [ distance, {id | cellIndex} ]. */
    public final boolean equals ( Object b )
    {
      if ( b == this ) return true;
      if ( b == null ||
           !(b instanceof GridHeapItem) ) return false;

      final GridHeapItem gb = (GridHeapItem)b;
      return( id        == gb.id &&
              cellIndex == gb.cellIndex );
    }
  }

  //==========================================================================
  //  Data:

  /** Tolerance for cell-boundary tests. */
  protected static final double GRID_EPSILON = 1.e-6;

  /** Maximal grid size (in one dimension). */
  protected int maxGrid = 64;

  /** Actual dimensionality of the grid. */
  protected int dim = 2;

  /** Use alternative implementation? */
  protected boolean alt = false;

  /** Actual left boundary. */
  protected double[] leftBound;

  /** Actual right boundary (relative to the {@link #leftBound}). */
  protected double[] rightBound;

  /** Actual cell size. */
  protected double[] cellSize;

  /** Actual number of cells. */
  protected int[] cells;

  /**
   * Accelerator array.
   * {@code accel[0]} .. total array size,
   * {@code accel[1]} .. horizontal size (y++),
   * {@code accel[2]} .. depth layer size (z++), ..
   */
  protected int[] accel;

  /** Array of individual cells. */
  protected int[][] data;

  /** Accelarator - reference to Brep instance. */
  protected Brep geometry;

  /** Actual entity type ({@link Brep#VERTEX VERTEX}, {@link Brep#EDGE EDGE} or {@link Brep#FACE FACE} are supported). */
  protected int entity;

  /** Actual context. */
  protected int ctx;

  /** Cell granularity (the rest of each cell is filled by {@link Brep#NULL NULL}). */
  protected final static int CELL_GRANUL = 16;

  /** Deferred inserted items. */
  protected int[] deferred;

  /** Statistics: point to box distance. */
  protected long pointBoxCounter = 0L;

  /** Statistics: point to point distance. */
  protected long pointPointCounter = 0L;

  /** Statistics: heap operations (insert, remove first). */
  protected long heapCounter = 0L;

  //==========================================================================
  //  Support:

  /**
   * Computes coordinate bounds of the whole data set.
   * leftBound/rightBound are filled with min/max values.
   */
  protected void computeBounds ()
  {
    if ( leftBound == null ||
         leftBound.length != dim )
      leftBound = new double[ dim ];
    if ( rightBound == null ||
         rightBound.length != dim )
      rightBound = new double[ dim ];

    int i, j;
    for ( i = 0; i < dim; i++ )
    {
      leftBound[i]  = Double.POSITIVE_INFINITY;
      rightBound[i] = Double.NEGATIVE_INFINITY;
    }

    BrepIterator it;
    int id;
    double[] act = new double[ Math.max( dim, 4 ) ];

    switch ( entity )
    {
      case Brep.VERTEX:                   // vertices
        it = geometry.vertexIterator( null );

        while ( (id = it.next()) != Brep.NULL )
        {                                 // take one vertex into account
          geometry.getVertexCoords( ctx, id, act );
          for ( i = 0; i < dim; i++ )
          {
            if ( act[i] < leftBound[i]  ) leftBound[i]  = act[i];
            if ( act[i] > rightBound[i] ) rightBound[i] = act[i];
          }
        }
        break;

      case Brep.EDGE:                     // edges
      {
        it = geometry.edgeIterator( null );
        int[] vert = new int[ 2 ];

        while ( (id = it.next()) != Brep.NULL )
        {                                 // take one edge into account
          geometry.getEdgeVertices( id, vert );
          for ( j = 0; j < 2; j++ )
          {
            geometry.getVertexCoords( ctx, vert[j], act );
            for ( i = 0; i < dim; i++ )
            {
              if ( act[i] < leftBound[i]  ) leftBound[i]  = act[i];
              if ( act[i] > rightBound[i] ) rightBound[i] = act[i];
            }
          }
        }
        break;
      }

      case Brep.FACE:                     // faces
      {
        it = geometry.faceIterator( ctx, null );
        int[] vert = new int[ 3 ];
        int len;

        while ( (id = it.next()) != Brep.NULL )
        {                                 // take one face into account
          while ( (len = geometry.getFaceVertices( id, vert )) > vert.length )
            vert = new int[ len ];

          for ( j = 0; j < len; j++ )
          {
            geometry.getVertexCoords( ctx, vert[j], act );
            for ( i = 0; i < dim; i++ )
            {
              if ( act[i] < leftBound[i]  ) leftBound[i]  = act[i];
              if ( act[i] > rightBound[i] ) rightBound[i] = act[i];
            }
          }
        }
        break;
      }
    }
  }

  /**
   * Computes coordinate bounds of the current set plus deferred insertions.
   * leftBound/rightBound are filled with min/max values.
   */
  protected void computeBoundsDeferred ()
  {
    if ( leftBound == null ||
         leftBound.length != dim )
      leftBound = new double[ dim ];
    if ( rightBound == null ||
         rightBound.length != dim )
      rightBound = new double[ dim ];

    int i, j, k, l;
    for ( i = 0; i < dim; i++ )
    {
      leftBound[i]  = Double.POSITIVE_INFINITY;
      rightBound[i] = Double.NEGATIVE_INFINITY;
    }

      // current data:
    double[] act = new double[ Math.max( dim, 4 ) ];
    int id;
    int[] vert   = new int[ 2 ];
    int flen;

    if ( data != null )
      for ( i = 0; i < accel[0]; i++ )
        if ( data[i] != null )
        {
          int len = data[i].length;

          switch ( entity )
          {
            case Brep.VERTEX:             // vertices
              for ( j = 0; j < len && (id = data[i][j]) != Brep.NULL; j++ )
              {
                geometry.getVertexCoords( ctx, id, act );
                for ( k = 0; k < dim; k++ )
                {
                  if ( act[k] < leftBound[k]  ) leftBound[k]  = act[k];
                  if ( act[k] > rightBound[k] ) rightBound[k] = act[k];
                }
              }
              break;

            case Brep.EDGE:               // edges
              for ( j = 0; j < len && (id = data[i][j]) != Brep.NULL; j++ )
              {
                geometry.getEdgeVertices( id, vert );
                for ( l = 0; l < 2; l++ )
                {
                  geometry.getVertexCoords( ctx, vert[l], act );
                  for ( k = 0; k < dim; k++ )
                  {
                    if ( act[k] < leftBound[k]  ) leftBound[k]  = act[k];
                    if ( act[k] > rightBound[k] ) rightBound[k] = act[k];
                  }
                }
              }
              break;

            case Brep.FACE:               // faces
              for ( j = 0; j < len && (id = data[i][j]) != Brep.NULL; j++ )
              {
                while ( (flen = geometry.getFaceVertices( id, vert )) > vert.length )
                  vert = new int[ flen ];

                for ( l = 0; l < flen; l++ )
                {
                  geometry.getVertexCoords( ctx, vert[l], act );
                  for ( k = 0; k < dim; k++ )
                  {
                    if ( act[k] < leftBound[k]  ) leftBound[k]  = act[k];
                    if ( act[k] > rightBound[k] ) rightBound[k] = act[k];
                  }
                }
              }
              break;
          }
        }

      // deferred insertions:
    if ( deferred != null )
      switch ( entity )
      {
        case Brep.VERTEX:                 // vertices
          for ( i = 0; i < deferred.length && (id = deferred[i]) != Brep.NULL; i++ )
          {
            geometry.getVertexCoords( ctx, id, act );
            for ( k = 0; k < dim; k++ )
            {
              if ( act[k] < leftBound[k]  ) leftBound[k]  = act[k];
              if ( act[k] > rightBound[k] ) rightBound[k] = act[k];
            }
          }
          break;

        case Brep.EDGE:                   // edges
          for ( i = 0; i < deferred.length && (id = deferred[i]) != Brep.NULL; i++ )
          {
            geometry.getEdgeVertices( id, vert );
            for ( l = 0; l < 2; l++ )
            {
              geometry.getVertexCoords( ctx, vert[l], act );
              for ( k = 0; k < dim; k++ )
              {
                if ( act[k] < leftBound[k]  ) leftBound[k]  = act[k];
                if ( act[k] > rightBound[k] ) rightBound[k] = act[k];
              }
            }
          }
          break;

        case Brep.FACE:                   // faces
          for ( i = 0; i < deferred.length && (id = deferred[i]) != Brep.NULL; i++ )
          {
            while ( (flen = geometry.getFaceVertices( id, vert )) > vert.length )
              vert = new int[ flen ];

            for ( l = 0; l < flen; l++ )
            {
              geometry.getVertexCoords( ctx, vert[l], act );
              for ( k = 0; k < dim; k++ )
              {
                if ( act[k] < leftBound[k]  ) leftBound[k]  = act[k];
                if ( act[k] > rightBound[k] ) rightBound[k] = act[k];
              }
            }
          }
          break;
      }
  }

  /** Computes array-indexing accelerators. */
  protected final void computeAccelerators ()
  {
      // index = coord[0] + coord[1] * accel[1] + coord[2] * accel[2] ..
      // accel[0] = total array size
    if ( accel == null ||
         accel.length != dim )
      accel = new int[ dim ];

    int tmp = 1;
    int i;
    for ( i = 0; i < dim; i++ )
      accel[ (i+1 < dim) ? i+1 : 0 ] = (tmp *= cells[i]);
  }

  /** Determines grid scale and sets the cells[] array. */
  protected final void scaleGrid ()
  {
    if ( cells == null ||
         cells.length != dim )
      cells = new int[ dim ];

    Arrays.fill( cells, maxGrid );
    computeAccelerators();

      // set cell size:
      // !!! TODO: improve grid-scale strategy !!!
    int i;
    for ( i = 0; i < dim; i++ )
      rightBound[i] -= leftBound[i];

    if ( cellSize == null ||
         cellSize.length != dim )
      cellSize = new double[ dim ];

    for ( i = 0; i < dim; i++ )
      cellSize[i] = rightBound[i] / cells[i];
  }

  /** Inserts a new item into the given cell. */
  protected final void insert ( int index, int id )
  {
    if ( data[index] == null )            // brand new cell..
    {
      data[index] = new int[ CELL_GRANUL ];
      Arrays.fill( data[index], 1, CELL_GRANUL, Brep.NULL );
      data[index][0] = id;
      return;
    }

    int len = data[index].length;
    int i = len - 1;
    while ( data[index][i] == Brep.NULL ) i--;

    if ( i == len - 1 )                   // expand the array
    {
      int newLen  = len + CELL_GRANUL;
      int[] tmp   = data[index];
      data[index] = new int[ newLen ];
      System.arraycopy( tmp, 0, data[index], 0, len );
      Arrays.fill( data[index], len + 1, newLen, Brep.NULL );
    }

    data[index][i+1] = id;
  }

  /** Computes nex integer-coordinate vector, returns {@code true} if succeeded. */
  protected final boolean nextCoord ( int[] ind, int[] mi, int[] ma )
  {
    int d = dim;
    while ( --d >= 0 )
    {
      if ( ++ind[d] <= ma[d] )
        return true;
      ind[d] = mi[d];
    }
    return false;
  }

  /** Inserts a new vertex 'id'. */
  protected final void insertVertex ( int id, double[] coords )
  {
    geometry.getVertexCoords( ctx, id, coords );

      // coords holds vertex coordinates:
    int d, ii;
    ii = Formula.clamp( (int)( (coords[0] - leftBound[0]) / cellSize[0] ), 0, cells[0]-1 );
    for ( d = 1; d < dim; d++ )
      ii += accel[d] * Formula.clamp( (int)( (coords[d] - leftBound[d]) / cellSize[d] ), 0, cells[d]-1 );

    insert( ii, id );
  }

  /** Inserts a new edge 'id'. */
  protected final void insertEdge ( int id, double[] coords, int[] vert, int[][] icoords, int[] ind )
  {
    geometry.getEdgeVertices( id, vert );

    int i, d, ii;
    for ( i = 0; i < 2; i++ )             // process one vertex
    {
      geometry.getVertexCoords( ctx, vert[i], coords );
      for ( d = 0; d < dim; d++ )
        icoords[i][d] = Formula.clamp( (int)( (coords[d] - leftBound[d]) / cellSize[d] ), 0, cells[d]-1 );
    }
    for ( d = 0; d < dim; d++ )
      if ( icoords[0][d] > icoords[1][d] )
      {
        i             = icoords[0][d];
        icoords[0][d] = icoords[1][d];
        icoords[1][d] = i;
      }

    System.arraycopy( icoords[0], 0, ind, 0, dim );
    do
    {
      ii = ind[0];
      for ( d = 1; d < dim; d++ )
        ii += accel[d] * ind[d];

      insert( ii, id );
    }
    while ( nextCoord ( ind, icoords[0], icoords[1] ) );
  }

  /** Inserts a new face 'id'. */
  protected final int[] insertFace ( int id, double[] coords, int[] vert, int[][] icoords, int[] ind )
  {
    int len;
    while ( (len = geometry.getFaceVertices( id, vert )) > vert.length )
      vert = new int[ len ];

    Arrays.fill( icoords[0], Integer.MAX_VALUE );
    Arrays.fill( icoords[1], Integer.MIN_VALUE );
    int i, d, ii;
    for ( i = 0; i < len; i++ )           // process one vertex
    {
      geometry.getVertexCoords( ctx, vert[i], coords );
      for ( d = 0; d < dim; d++ )
      {
        ii = Formula.clamp( (int)( (coords[d] - leftBound[d]) / cellSize[d] ), 0, cells[d]-1 );
        if ( ii < icoords[0][d] ) icoords[0][d] = ii;
        if ( ii > icoords[1][d] ) icoords[1][d] = ii;
      }
    }

    System.arraycopy( icoords[0], 0, ind, 0, dim );
    do
    {
      ii = ind[0];
      for ( d = 1; d < dim; d++ )
        ii += accel[d] * ind[d];

      insert( ii, id );
    }
    while ( nextCoord ( ind, icoords[0], icoords[1] ) );

    return vert;
  }

  //==========================================================================
  //  Iterator classes:

  /**
   * Custom iterator for center pass.
   */
  protected class CenterIterator extends DefaultBrepIterator
  {
    /** Center of the pass. */
    protected double[] center = new double[ dim ];

    /** Maximal radius to be searched. */
    protected double maxRadius;

    /** Working heap. */
    protected SortedSet< GridHeapItem > heap;

    /** Set of already examined objects. */
    protected final BitSet   used = new BitSet();

    /** Temporary array to hold vertex coordinates. */
    protected final double[] act0 = new double[ dim ];
    protected final double[] act1 = new double[ dim ];

    /** Temporary array for edge (face) vertex handles. */
    protected final int[]    vert = new int[ 2 ];

    public void init ( double[] center, double maxR )
    {
      System.arraycopy( center, 0, this.center, 0, dim );
      maxRadius = maxR;
      reset();
    }

    /** Inserts all objects from the given cell into the heap. */
    protected void insertCell ( int index )
    {
      final int[] cell = data[index];
      if ( cell == null ) return;

      int id, i = 0;
      final int len = cell.length;
      double dist;

      if ( entity == Brep.VERTEX )
        while ( i < len && (id = cell[i]) != Brep.NULL )
        {
          pointPointCounter++;
          geometry.getVertexCoords( ctx, id, act0 );
          dist = Math.hypot( act0[0] - center[0], act0[1] - center[1] );
          if ( dist <= maxRadius )
          {
            GridHeapItem hi = new GridHeapItem( dist );
            hi.id = id;
            heapCounter++;
            heap.add( hi );
          }
          i++;
        }
      else
      {
        while ( i < len && (id = cell[i]) != Brep.NULL )
        {
          if ( !used.get( id ) )
          {
            used.set( id );
            geometry.getEdgeVertices( id, vert );
            geometry.getVertexCoords( ctx, vert[0], act0 );
            geometry.getVertexCoords( ctx, vert[1], act1 );
            pointPointCounter++;
            dist = Geometry.pointSegment2D( center[0], center[1], act0[0], act0[1], act1[0], act1[1] );
            if ( dist <= maxRadius )
            {
              GridHeapItem hi = new GridHeapItem( dist );
              hi.id = id;
              heapCounter++;
              heap.add( hi );
            }
          }
          i++;
        }
      }
    }

    /** Initializes the heap (should be called at the beginning of each pass). */
    protected void initHeap ()
    {
      if ( data     == null ||
           geometry == null ) return;

      heap = new TreeSet< GridHeapItem >();
      int x, y, index;
      final double[] xy = { leftBound[0], leftBound[1] };

      for ( y = index = 0; y++ < cells[1]; xy[1] += cellSize[1] )

        for ( x = 0, xy[0] = leftBound[0]; x++ < cells[0]; xy[0] += cellSize[0], index++ )

          if ( data[index] != null )      // nonempty cell:
          {
            pointBoxCounter++;
            double dist = Geometry.pointBox2D( center, xy, cellSize );

            if ( dist <= maxRadius )      // the cell isn't too far
              if ( dist > 0.0 )
              {                           // the whole cell..
                GridHeapItem hi = new GridHeapItem( dist );
                hi.cellIndex = index;
                heapCounter++;
                heap.add( hi );
              }
              else
                insertCell( index );      // individual objects
          }
    }

    /** Reset the iterator. */
    public void reset ()
    {
      heap = null;
      used.clear();
    }

    /** Returns the next object or {@link Brep#NULL NULL}. */
    public final int next ()
    {
      if ( heap == null )
        initHeap();

      GridHeapItem hi;

      while ( !heap.isEmpty() )
      {
        heapCounter++;
        hi = heap.first();
        heap.remove( hi );
        if ( hi.id != Brep.NULL )
          return hi.id;
        insertCell( hi.cellIndex );
      }

      return Brep.NULL;
    }
  }

  /**
   * Custom iterator for ray-segment-pass (new version - 2DDDA).
   */
  protected class RaySegmentIterator2D extends DefaultBrepIterator
  {
    /** Ray origin. */
    protected final double[] origin = new double[ dim ];

    /** Ray direction. */
    protected final double[] direction = new double[ dim ];

    /** Actual cell-coordinate (horizontal). */
    protected int xCell;

    /** Actual cell-coordinate (vertical). */
    protected int yCell;

    /** Actual cell index. */
    protected int iCell;

    /** Integer x-increment. */
    protected int incx;

    /** Integer y-increment. */
    protected int incy;

    /** Index increment (vertical). */
    protected int inciy;

    /** 2DDDA constant (distance between adjacent x-wall intersections). */
    protected double dx;

    /** 2DDDA constant (distance between adjacent y-wall intersections). */
    protected double dy;

    /** 2DDDA variable - distance to the nearest x-wall. */
    protected double lx;

    /** 2DDDA variable - distance to the nearest y-wall. */
    protected double ly;

    /** 2DDDA variable - distance to the nearest wall. */
    protected double minl;

    /** Ray-length searched so far. */
    protected double bound;

    /** Result heap. */
    protected SortedSet< GridHeapItem > result = null;

    /** Set of already examined objects. */
    protected final BitSet   used = new BitSet();

    /** Temporary array to hold vertex coordinates. */
    protected final double[] act0 = new double[ 3 ];
    protected final double[] act1 = new double[ 3 ];

    /** Temporary array for edge vertex handles. */
    protected final int[]    vert = new int[ 3 ];

    /** T-bounds for ray-AABB intersection. */
    protected final double[] t    = new double[ 2 ];

    /** Accelerator: attribute holding edge vertices. */
    protected final int attrEdge   = geometry.getAttributeId( Brep.EDGE,   Brep.NULL, Brep.DATA );

    /** Accelerator: attribute holding vertex coordinates. */
    protected final int attrVertex = geometry.getAttributeId( Brep.VERTEX, ctx,       Brep.COORD );

    /** Initializes all constants depending on the ray geometry. */
    public void init ( double[] orig, double[] dir )
    {
      System.arraycopy( orig, 0, origin,    0, dim );
      System.arraycopy( dir,  0, direction, 0, dim );
      incx  = (int)Math.signum( dir[0] );
      incy  = (int)Math.signum( dir[1] );
      inciy = incy * accel[1];
      dx    = Geometry.isZero( dir[0] ) ? Double.MAX_VALUE : Math.abs( cellSize[0] / dir[0] );
      dy    = Geometry.isZero( dir[1] ) ? Double.MAX_VALUE : Math.abs( cellSize[1] / dir[1] );
      reset();
    }

    /** Inserts all objects from the given cell into the heap. */
    protected void insertCell ( int index )
    {
      final int[] cell = data[index];
      if ( cell == null ) return;

      int id, i = 0;
      final int len = cell.length;
      double dist;

      while ( i < len && (id = cell[i]) != Brep.NULL )
      {
        if ( !used.get( id ) )
        {
          used.set( id );
          geometry.getAttribute( attrEdge,   id,      vert );
          geometry.getAttribute( attrVertex, vert[0], act0 );
          geometry.getAttribute( attrVertex, vert[1], act1 );
          pointPointCounter++;
          dist = Geometry.raySegment2D( origin, direction, act0, act1 );
          if ( dist >= 0.0 )
          {
            GridHeapItem hi = new GridHeapItem( dist );
            hi.id = id;
            heapCounter++;
            result.add( hi );
          }
        }
        i++;
      }
    }

    /**
     * Initializes the pass (should be called at the beginning of each pass).
     *
     * @return {@code true} if the result set is empty.
     */
    protected boolean initPass ()
    {
      if ( data     == null ||
           geometry == null ||
           !Geometry.rayBox( origin, direction, leftBound, rightBound, t ) ) return true;

        // init 2DDDA pass:
      result = new TreeSet< GridHeapItem >();

      if ( t[0] > 0.0 )                   // origin lies outside of the grid hull
      {
        origin[0] += t[0] * direction[0];
        origin[1] += t[0] * direction[1];
      }

      final double x = (origin[0] - leftBound[0]) / cellSize[0];
      final double y = (origin[1] - leftBound[1]) / cellSize[1];
      xCell = Formula.clamp( (int)Math.floor( x ), 0, cells[0]-1 );
      yCell = Formula.clamp( (int)Math.floor( y ), 0, cells[1]-1 );
      iCell = xCell + yCell * accel[1];

        // nearest distance in x:
      lx = x - Math.floor( x );
      if ( incx > 0 )
        lx = 1.0 - lx;
      if ( incx == 0 )
        lx = Double.MAX_VALUE;
      else
        if ( lx < GRID_EPSILON )
          lx = dx;
        else
          lx *= dx;

        // nearest distance in y:
      ly = y - Math.floor( y );
      if ( incy > 0 )
        ly = 1.0 - ly;
      if ( incy == 0 )
        ly = Double.MAX_VALUE;
      else
        if ( ly < GRID_EPSILON )
          ly = dy;
        else
          ly *= dy;

      bound =
      minl  = Math.min( lx, ly );

        // process the initial cell:
      insertCell( iCell );

      return false;
    }

    /** Reset the iterator. */
    public void reset ()
    {
      result = null;
      used.clear();
    }

    /** Returns the next object or {@link Brep#NULL NULL}. */
    public final int next ()
    {
      if ( result == null &&
           initPass() )
        return Brep.NULL;

      GridHeapItem hi = null;
      if ( !result.isEmpty() )
        hi = result.first();

      while ( hi == null ||
              hi.distance > bound )       // no element is ready
      {
          // 2DDDA step:
        if ( lx < ly )                    // x-wall is closer
        {
          ly    -= lx;
          lx     = dx;
          xCell += incx;
          iCell += incx;
        }
        else                              // y-wall is closer
        {
          lx    -= ly;
          ly     = dy;
          yCell += incy;
          iCell += inciy;
        }

        if ( xCell >= 0 && xCell < cells[0] &&
             yCell >= 0 && yCell < cells[1] )
        {
          bound +=
          minl   = Math.min( lx, ly );
          insertCell( iCell );
          if ( !result.isEmpty() )
            hi = result.first();
        }
        else
        {
          if ( hi == null )
            return Brep.NULL;

          bound = Double.POSITIVE_INFINITY;
        }
      }

      if ( hi == null ) return Brep.NULL;

      heapCounter++;
      result.remove( hi );
      return hi.id;
    }
  }

  /**
   * Custom iterator for ray-segment-pass (old version).
   */
  protected class RayIteratorOld extends DefaultBrepIterator
  {
    /** Ray origin. */
    protected final double[] origin = new double[ dim ];

    /** Ray direction. */
    protected final double[] direction = new double[ dim ];

    /** Working heap. */
    protected SortedSet< GridHeapItem > heap;

    /** Set of already examined objects. */
    protected final BitSet   used = new BitSet();

    /** Temporary array to hold vertex coordinates. */
    protected final double[] act0 = new double[ dim ];
    protected final double[] act1 = new double[ dim ];

    /** Temporary array for edge (face) vertex handles. */
    protected final int[]    vert = new int[ 2 ];

    /** Accelerator: attribute holding edge vertices. */
    protected final int attrEdge   = geometry.getAttributeId( Brep.EDGE,   Brep.NULL, Brep.DATA );

    /** Accelerator: attribute holding vertex coordinates. */
    protected final int attrVertex = geometry.getAttributeId( Brep.VERTEX, ctx,       Brep.COORD );

    public void init ( double[] orig, double[] dir )
    {
      System.arraycopy( orig, 0, origin,    0, dim );
      System.arraycopy( dir,  0, direction, 0, dim );
      reset();
    }

    /** Inserts all objects from the given cell into the heap. */
    protected void insertCell ( int index )
    {
      if ( entity == Brep.VERTEX ) return;

      final int[] cell = data[index];
      if ( cell == null ) return;

      int id, i = 0;
      final int len = cell.length;
      double dist;

      while ( i < len && (id = cell[i]) != Brep.NULL )
      {
        if ( !used.get( id ) )
        {
          used.set( id );
          geometry.getAttribute( attrEdge,   id,      vert );
          geometry.getAttribute( attrVertex, vert[0], act0 );
          geometry.getAttribute( attrVertex, vert[1], act1 );
          pointPointCounter++;
          dist = Geometry.raySegment2D( origin, direction, act0, act1 );
          if ( dist >= 0.0 )
          {
            GridHeapItem hi = new GridHeapItem( dist );
            hi.id = id;
            heapCounter++;
            heap.add( hi );
          }
        }
        i++;
      }
    }

    /** Initializes the heap (should be called at the beginning of each pass). */
    protected void initHeap ()
    {
      if ( data     == null ||
           geometry == null ) return;

      heap = new TreeSet< GridHeapItem >();
      int x, y, index;
      final double[] xy = { leftBound[0], leftBound[1] };
      final double[] t  = new double[ 2 ]; // t-bounds for ray-AABB intersection

      for ( y = index = 0; y++ < cells[1]; xy[1] += cellSize[1] )

        for ( x = 0, xy[0] = leftBound[0]; x++ < cells[0]; xy[0] += cellSize[0], index++ )

          if ( data[index] != null )      // nonempty cell:
          {
            pointBoxCounter++;
            if ( !Geometry.rayBox( origin, direction, xy, cellSize, t ) ) continue;

            if ( t[0] <= 0.0 )
              insertCell( index );        // origin lies inside of the cell..

            else                          // the cell will be intersected later
            {
              GridHeapItem hi = new GridHeapItem( t[0] );
              hi.cellIndex = index;
              heapCounter++;
              heap.add( hi );
            }
          }
    }

    /** Reset the iterator. */
    public void reset ()
    {
      heap = null;
      used.clear();
    }

    /** Returns the next object or {@link Brep#NULL NULL}. */
    public final int next ()
    {
      if ( heap == null )
        initHeap();

      GridHeapItem hi;

      while ( !heap.isEmpty() )
      {
        heapCounter++;
        hi = heap.first();
        heap.remove( hi );
        if ( hi.id != Brep.NULL )
          return hi.id;
        insertCell( hi.cellIndex );
      }

      return Brep.NULL;
    }
  }

  /**
   * Custom iterator for ray-triangle-pass (3DDDA).
   */
  protected class RayTriangleIterator3D extends DefaultBrepIterator
  {
    /** Ray origin. */
    protected double[] origin    = new double[ 3 ];

    /** Ray direction (unit vector). */
    protected double[] direction = new double[ 3 ];

    /** Actual cell-coordinate (horizontal). */
    protected int xCell;

    /** Actual cell-coordinate (vertical). */
    protected int yCell;

    /** Actual cell-coordinate (depth). */
    protected int zCell;

    /** Actual cell index. */
    protected int iCell;

    /** Integer x-increment. */
    protected int incx;

    /** Integer y-increment. */
    protected int incy;

    /** Integer z-increment. */
    protected int incz;

    /** Index increment (vertical). */
    protected int inciy;

    /** Index increment (depth). */
    protected int inciz;

    /** 3DDDA constant (distance between adjacent x-wall intersections). */
    protected double dx;

    /** 3DDDA constant (distance between adjacent y-wall intersections). */
    protected double dy;

    /** 3DDDA constant (distance between adjacent z-wall intersections). */
    protected double dz;

    /** 3DDDA variable - distance to the nearest x-wall. */
    protected double lx;

    /** 3DDDA variable - distance to the nearest y-wall. */
    protected double ly;

    /** 3DDDA variable - distance to the nearest z-wall. */
    protected double lz;

    /** 3DDDA variable - distance to the nearest wall. */
    protected double minl;

    /** Ray-length searched so far. */
    protected double bound;

    /** Result heap. */
    protected SortedSet< GridHeapItem > result = null;

    /** Set of already examined objects. */
    protected final BitSet   used = new BitSet();

    /** Temporary array to hold vertex coordinates. */
    protected final double[] act0 = new double[ 4 ];
    protected final double[] act1 = new double[ 4 ];
    protected final double[] act2 = new double[ 4 ];

    /** Temporary array for face vertex handles. */
    protected final int[]    vert = new int[ 3 ];

    /** T-bounds for ray-AABB intersection. */
    protected final double[] t    = new double[ 2 ];

    /** Accelerator: attribute holding vertex coordinates. */
    protected final int attrVertex = geometry.getAttributeId( Brep.VERTEX, ctx, Brep.COORD );

    /** Initializes all constants depending on the ray geometry. */
    public void init ( double[] orig, double[] dir )
    {
      Geometry.toCartesian3D( orig, origin );
      Geometry.normalize3D( dir, direction );
      incx  = (int)Math.signum( dir[0] );
      incy  = (int)Math.signum( dir[1] );
      incz  = (int)Math.signum( dir[2] );
      inciy = incy * accel[1];
      inciz = incz * accel[2];
      dx    = Geometry.isZero( dir[0] ) ? Double.MAX_VALUE : Math.abs( cellSize[0] / dir[0] );
      dy    = Geometry.isZero( dir[1] ) ? Double.MAX_VALUE : Math.abs( cellSize[1] / dir[1] );
      dz    = Geometry.isZero( dir[2] ) ? Double.MAX_VALUE : Math.abs( cellSize[2] / dir[2] );
      reset();
    }

    /** Inserts all objects from the given cell into the heap. */
    protected void insertCell ( int index )
    {
      final int[] cell = data[index];
      if ( cell == null ) return;

      int id, i = 0;
      final int len = cell.length;
      double dist;

      while ( i < len && (id = cell[i]) != Brep.NULL )
      {
        if ( !used.get( id ) )
        {
          used.set( id );
          if ( geometry.getFaceVertices( id, vert ) == 3 )
          {
            geometry.getAttribute( attrVertex, vert[0], act0 );
            Geometry.toCartesian3D( act0, act0 );
            geometry.getAttribute( attrVertex, vert[1], act1 );
            Geometry.toCartesian3D( act1, act1 );
            geometry.getAttribute( attrVertex, vert[2], act2 );
            Geometry.toCartesian3D( act2, act2 );
            pointPointCounter++;
            dist = Geometry.rayTriangle3D( origin, direction, act0, act1, act2, null );
            if ( !Double.isInfinite( dist ) &&
                 dist >= 0.0 )
            {
              GridHeapItem hi = new GridHeapItem( dist );
              hi.id = id;
              heapCounter++;
              result.add( hi );
            }
          }
        }
        i++;
      }
    }

    /**
     * Initializes the pass (should be called at the beginning of each pass).
     *
     * @return {@code true} if the result set is empty.
     */
    protected boolean initPass ()
    {
      if ( data     == null ||
           geometry == null ||
           !Geometry.rayBox( origin, direction, leftBound, rightBound, t ) ) return true;

        // init 3DDDA pass:
      result = new TreeSet< GridHeapItem >();

      if ( t[0] > 0.0 )                   // origin lies outside of the grid hull
      {
        origin[0] += t[0] * direction[0];
        origin[1] += t[0] * direction[1];
        origin[2] += t[0] * direction[2];
      }

      final double x = (origin[0] - leftBound[0]) / cellSize[0];    // unit: grid cells
      final double y = (origin[1] - leftBound[1]) / cellSize[1];
      final double z = (origin[2] - leftBound[2]) / cellSize[2];
      xCell = Formula.clamp( (int)Math.floor( x ), 0, cells[0]-1 );
      yCell = Formula.clamp( (int)Math.floor( y ), 0, cells[1]-1 );
      zCell = Formula.clamp( (int)Math.floor( z ), 0, cells[2]-1 );
      iCell = xCell + yCell * accel[1] + zCell * accel[2];

        // nearest distance in x:
      lx = x - Math.floor( x );
      if ( incx > 0 )
        lx = 1.0 - lx;
      if ( incx == 0 )
        lx = Double.MAX_VALUE;
      else
        if ( lx < GRID_EPSILON )
          lx = dx;
        else
          lx *= dx;

        // nearest distance in y:
      ly = y - Math.floor( y );
      if ( incy > 0 )
        ly = 1.0 - ly;
      if ( incy == 0 )
        ly = Double.MAX_VALUE;
      else
        if ( ly < GRID_EPSILON )
          ly = dy;
        else
          ly *= dy;

        // nearest distance in z:
      lz = z - Math.floor( z );
      if ( incz > 0 )
        lz = 1.0 - lz;
      if ( incz == 0 )
        lz = Double.MAX_VALUE;
      else
        if ( lz < GRID_EPSILON )
          lz = dz;
        else
          lz *= dz;

      bound =
      minl  = Math.min( Math.min( lx, ly ), lz );

        // process the initial cell:
      insertCell( iCell );

      return false;
    }

    /** Reset the iterator. */
    public void reset ()
    {
      result = null;
      used.clear();
    }

    /** Returns the next object or {@link Brep#NULL NULL}. */
    public final int next ()
    {
      if ( result == null &&
           initPass() )
        return Brep.NULL;

      GridHeapItem hi = null;
      if ( !result.isEmpty() )
        hi = result.first();

      while ( hi == null ||
              hi.distance > bound )       // no element is ready
      {
          // 3DDDA step:
        if ( lx <= ly && lx <= lz )       // x-wall is the closest one
        {
          ly    -= lx;
          lz    -= lx;
          lx     = dx;
          xCell += incx;
          iCell += incx;
        }
        else
        if ( ly <= lx && ly <= lz )       // y-wall is the closest one
        {
          lx    -= ly;
          lz    -= ly;
          ly     = dy;
          yCell += incy;
          iCell += inciy;
        }
        else                              // z-wall is the closest one
        {
          lx    -= lz;
          ly    -= lz;
          lz     = dz;
          zCell += incz;
          iCell += inciz;
        }

        if ( xCell >= 0 && xCell < cells[0] &&
             yCell >= 0 && yCell < cells[1] &&
             zCell >= 0 && zCell < cells[2] )
        {
          bound +=
          minl   = Math.min( Math.min( lx, ly ), lz );
          insertCell( iCell );
          if ( !result.isEmpty() )
            hi = result.first();
        }
        else
        {
          if ( hi == null )
            return Brep.NULL;

          bound = Double.POSITIVE_INFINITY;
        }
      }

      if ( hi == null ) return Brep.NULL;

      heapCounter++;
      result.remove( hi );
      return hi.id;
    }
  }

  //==========================================================================
  //  Interface GeometrySearch:

  /**
   * Builds the whole data structure (directory) from scratch.
   *
   * @param  entity Type of stored entities ({@link Brep#VERTEX VERTEX}, {@link Brep#EDGE EDGE} or
   *                {@link Brep#FACE FACE} from interface {@link Brep}).
   * @param  ctx Handle to the used {@link Brep}-context.
   * @return {@code true} if this service is implemented.
   * @see    #resetDirectory
   * @see    #optimizeDirectory
   * @see    Brep#VERTEX
   * @see    Brep#EDGE
   * @see    Brep#SOLID
   */
  @Override
  public boolean buildDirectory ( int entity, int ctx )
  {
    geometry = (Brep)getInterface( PL_DATA, IFACE+"Brep" );
    if ( geometry == null ) return false;

    this.entity = entity;
    this.ctx    = ctx;

    // !!! TODO: build grid directory !!!
    // !!! {  My pilot implementation

      // grid bounds, cell size, accelerators..:
    computeBounds();
    scaleGrid();

      // init the data array:
    if ( data == null ||
         data.length != accel[0] )
      data = new int[ accel[0] ][];
    else
      Arrays.fill( data, null );
    deferred = null;

      // fill the data array:
    BrepIterator it;
    int id;
    double[] act   = new double[ dim ];
    int[][] icoord = null;
    int[] vert     = null;
    int[] ind      = null;

    switch ( entity )
    {
      case Brep.VERTEX:
        it = geometry.vertexIterator( null );

        while ( (id = it.next()) != Brep.NULL )  // insert one vertex
          insertVertex( id, act );
        break;

      case Brep.EDGE:
        it = geometry.edgeIterator( null );
        icoord = new int[ 2 ][ dim ];
        vert   = new int[ 2 ];
        ind    = new int[ dim ];

        while ( (id = it.next()) != Brep.NULL )  // insert one edge
          insertEdge( id, act, vert, icoord, ind );
        break;

      case Brep.FACE:
        it = geometry.faceIterator( ctx, null );
        icoord = new int[ 2 ][ dim ];
        vert   = new int[ 3 ];
        ind    = new int[ dim ];

        while ( (id = it.next()) != Brep.NULL )  // insert one face
          vert = insertFace( id, act, vert, icoord, ind );
        break;
    }

    // !!! }
    return true;
  }

  /**
   * (Re-)initializes the data structure (directory).
   *
   * @param  entity Type of stored entities ({@link Brep#VERTEX VERTEX}, {@link Brep#EDGE EDGE} or
   *                {@link Brep#FACE FACE} from interface {@link Brep}).
   * @param  ctx Handle to the used {@link Brep}-context.
   * @return {@code true} if this service is implemented.
   * @see    #buildDirectory
   * @see    #optimizeDirectory
   * @see    Brep#VERTEX
   * @see    Brep#EDGE
   * @see    Brep#SOLID
   */
  @Override
  public boolean resetDirectory ( int entity, int ctx )
  {
    geometry = (Brep)getInterface( PL_DATA, IFACE+"Brep" );
    if ( geometry == null ) return false;

    this.entity = entity;
    this.ctx    = ctx;

    // !!! TODO: reset data structure (make it empty) !!!
    // !!! {  My pilot implementation

    data     = null;
    deferred = null;

    // !!! }
    return true;
  }

  /**
   * Optimizes the data structure (directory). Can be used for <b>finishing</b> of the
   * update phase or for explicit data structure <b>optimization</b>. Non-mandatory.
   * (but application has to call it at the end of update phase).
   *
   * @return {@code true} if this service is implemented.
   * @see    #buildDirectory
   * @see    #resetDirectory
   */
  @Override
  public boolean optimizeDirectory ()
  {
    if ( geometry == null ) return false;

    // !!! TODO: build grid directory from deferred insertions !!!
    // !!! {  My pilot implementation

      // grid bounds, cell size, accelerators..:
    computeBoundsDeferred();
    scaleGrid();

      // init the new data array:
    int[][] tmp = data;
    data = new int[ accel[0] ][];

      // fill the new data array:
    int id, i, j, len;
    double[] act   = new double[ dim ];
    int[][] icoord = null;
    int[] vert     = null;
    int[] ind      = null;

    switch ( entity )
    {
      case Brep.VERTEX:
          // re-insert the old data set:
        if ( tmp != null )
          for ( i = 0; i < tmp.length; i++ )
            if ( tmp[i] != null )
            {
              len = tmp[i].length;
              for ( j = 0; j < len && (id = tmp[i][j]) != Brep.NULL; j++ )
                insertVertex( id, act );  // insert one old vertex
            }

          // deferred insertions:
        if ( deferred != null )
        {
          for ( i = 0; i < deferred.length && (id = deferred[i]) != Brep.NULL; i++ )
            insertVertex( id, act );
          deferred = null;
        }
        break;

      case Brep.EDGE:
        icoord = new int[ 2 ][ dim ];
        vert   = new int[ 2 ];
        ind    = new int[ dim ];

          // re-insert the old data set:
        if ( tmp != null )
          for ( i = 0; i < tmp.length; i++ )
            if ( tmp[i] != null )
            {
              len = tmp[i].length;
              for ( j = 0; j < len && (id = tmp[i][j]) != Brep.NULL; j++ )
                insertEdge( id, act, vert, icoord, ind );  // insert one old edge
            }

          // deferred insertions:
        if ( deferred != null )
        {
          for ( i = 0; i < deferred.length && (id = deferred[i]) != Brep.NULL; i++ )
            insertEdge( id, act, vert, icoord, ind );
          deferred = null;
        }
        break;

      case Brep.FACE:
        icoord = new int[ 2 ][ dim ];
        vert   = new int[ 3 ];
        ind    = new int[ dim ];

          // re-insert the old data set:
        if ( tmp != null )
          for ( i = 0; i < tmp.length; i++ )
            if ( tmp[i] != null )
            {
              len = tmp[i].length;
              for ( j = 0; j < len && (id = tmp[i][j]) != Brep.NULL; j++ )
                vert = insertFace( id, act, vert, icoord, ind ); // insert one old face
            }

          // deferred insertions:
        if ( deferred != null )
        {
          for ( i = 0; i < deferred.length && (id = deferred[i]) != Brep.NULL; i++ )
            vert = insertFace( id, act, vert, icoord, ind );
          deferred = null;
        }
        break;
    }

    // !!! }
    return true;
  }

  /**
   * Initializes a new <b>center pass</b>. Objects are enumerated in order defined by increasing
   * distance from the given center point.
   *
   * @param  center Center of a pass.
   * @param  radius Maximum radius to be searched.
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #getProgress
   */
  @Override
  public BrepIterator centerPass ( double[] center, double radius, BrepIterator pre )
  {
    if ( data == null ) return null;

    CenterIterator it;
    if ( pre == null ||
         !(pre instanceof CenterIterator) )
      pre = it = new CenterIterator();
    else
      it = (CenterIterator)pre;

    // !!! TODO: initialize the pass !!!
    // !!! {  My pilot implementation

    it.init( center, radius );

    // !!! }
    return pre;
  }

  /**
   * Initializes a new <b>ray-casting pass</b>.
   *
   * @param  start Starting point of the ray.
   * @param  direction Direction vector of the ray.
   * @param  pre Pre-allocated {@link BrepIterator} object (can be {@code null}).
   * @return The initialized iterator data.
   * @see    #getProgress
   */
  @Override
  public BrepIterator rayPass ( double[] start, double[] direction, BrepIterator pre )
  {
    if ( data == null ) return null;

    // !!! TODO: initialize the pass !!!
    // !!! {  My pilot implementation

    switch ( entity )
    {
      case Brep.EDGE:
        if ( alt )
        {
          RayIteratorOld it;
          if ( pre == null ||
               !(pre instanceof RayIteratorOld) )
            pre = it = new RayIteratorOld();
          else
            it = (RayIteratorOld)pre;

          it.init( start, direction );
        }
        else
        {
          RaySegmentIterator2D it;
          if ( pre == null ||
               !(pre instanceof RaySegmentIterator2D) )
            pre = it = new RaySegmentIterator2D();
          else
            it = (RaySegmentIterator2D)pre;

          it.init( start, direction );
        }
        break;

      case Brep.FACE:
      {
        RayTriangleIterator3D it;
        if ( pre == null ||
             !(pre instanceof RayTriangleIterator3D) )
          pre = it = new RayTriangleIterator3D();
        else
          it = (RayTriangleIterator3D)pre;

        it.init( start, direction );
        break;
      }
    }

    // !!! }
    return pre;
  }

  /**
   * Returns (and resets) counters (for statistics, profiling).
   * <p>counters[0] .. number of point-to-point distance calculations.
   * <p>counters[1] .. number of point-to-box distance calculations.
   * <p>counters[2] .. number of log(N) heap operations.
   *
   * @param  counters Array to hold result. If {@code null} functions returns required length.
   * @return Required array length.
   */
  @Override
  public int getStatistics ( long[] counters )
  {
    if ( counters != null &&
         counters.length > 0 )
    {
      counters[0] = pointPointCounter;
      if ( counters.length > 1 )
      {
        counters[1] = pointBoxCounter;
        if ( counters.length > 2 )
          counters[2] = heapCounter;
      }

      pointPointCounter =
      pointBoxCounter   =
      heapCounter       = 0L;
    }

    return 3;
  }

  //==========================================================================
  //  Interface Property:

  public final static String MAX_GRID_SIZE = "Max grid size";
  public final static String DIM           = "Dimension";
  public final static String ALTERNATIVE   = "Alternative";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // MAX_GRID_SIZE:
    if ( key.compareTo(MAX_GRID_SIZE) == 0 )
    {
      maxGrid = intProperty( value, maxGrid, 1, 1024 );
      return;
    }
      // DIM:
    if ( key.compareTo(DIM) == 0 )
    {
      dim = intProperty( value, dim, 1, 4 );
      return;
    }
      // ALTERNATIVE:
    if ( key.compareTo(ALTERNATIVE) == 0 )
    {
      alt = booleanProperty( value, alt );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(MAX_GRID_SIZE) == 0 )
      return maxGrid;
    if ( key.compareTo(DIM) == 0 )
      return dim;
    if ( key.compareTo(ALTERNATIVE) == 0 )
      return alt;
    if ( key.compareTo(TEXT_DESCRIPTION) == 0 )
      return String.format( dim == 2 ? "Uniform grid %d x %1$d" : "Uniform grid %d x %1$d x %1$d", maxGrid );

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "UniformGrid";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "GeometrySearchToBrep";

  /** Object description. */
  private final static String DESCRIPTION = "Uniform grid for fast searching in 2D, 3D..";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DATA;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT, IFACE+"GeometrySearch" );
      t.newOutputPlug( PL_DATA,  IFACE+"Brep" );
        // Property:
      t.propBegin( MAX_GRID_SIZE, TYPE_INTEGER, "Maximal grid size in 1 dimension", true );
        t.propDefault( 64 );
        t.propBounds(  1, 1024 );
      t.propEnd();
      t.propBegin( DIM, TYPE_INTEGER, "Data dimensionality", true );
        t.propDefault( 2 );
        t.propBounds(  1, 4 );
      t.propEnd();
      t.propBegin( ALTERNATIVE, TYPE_BOOLEAN, "Use alternative (slower) implementation?", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( TEXT_DESCRIPTION, TYPE_STRING, "Object description", false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
