package cz.cuni.jagrlib.testing;

import java.awt.Color;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Warping engine example - does naive forward mapping.
 *
 * @version 0.26 $Rev: 439 $ $Date: 2007-05-05 12:29:50 +0200 (so, 05 V 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/WarpEngine.java">WarpEngine.java</a>
 */
public class WarpEngine extends Piece implements Trigger
{
  //==========================================================================
  //  Data:

  /** Background (default) color. */
  protected Color background    = Color.BLACK;

  /** Transfer quincunx-interpolated pixels, too? */
  protected boolean interpolate = true;

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transformation.
   *
   * @param  type The action type (whatever it means..).
   * @return {@code true} if the action was successful.
   */
  public boolean fire ( int type )
  {
    final FunctionR2ToR2 func   = (FunctionR2ToR2)getInterface( PL_FUNCTION, IFACE+"FunctionR2ToR2" );
    final RasterGraphics input  = (RasterGraphics)getInterface( "image1",    IFACE+"RasterGraphics" );
    final RasterGraphics output = (RasterGraphics)getInterface( "image2",    IFACE+"RasterGraphics" );

    if ( func   == null ||
         input  == null ||
         output == null ) return false;

      // output image:
    int width    = input.getWidth();
    int height   = input.getHeight();
    double[] dom = new double[ 4 ];
    double[] rng = new double[ 4 ];

    if ( !func.getDomain( dom ) )
    {
      dom[0] = dom[2] = 0.0;
      dom[1] = dom[3] = 1.0;
    }
    if ( !func.getRange( rng ) )
      rng = dom;

      // source coordinates mapping (from [0,width-1] to [dom[0],dom[1]]..):
    double domKx = (dom[1] - dom[0]) / (width  - 1.0);
    double domAx = dom[0];
    double domKy = (dom[3] - dom[2]) / (height - 1.0);
    double domAy = dom[2];

      // destination coordinates mapping (from [rng[0],rng[1]] to [0,width-1]..):
    int destWidth  = (int)( (width  - 1.0) * (rng[1] - rng[0]) / (dom[1] - dom[0]) + 1.5 );
    int destHeight = (int)( (height - 1.0) * (rng[3] - rng[2]) / (dom[3] - dom[2]) + 1.5 );
    double rngKx   = (destWidth  - 1.0) / (rng[1] - rng[0]);
    double rngAx   = - rngKx * rng[0] + 0.5;
    double rngKy   = (destHeight - 1.0) / (rng[3] - rng[2]);
    double rngAy   = - rngKy * rng[2] + 0.5;

      // temporary values:
    int[] pixel = new int[ 4 ];           // I/O pixel
    double[] uv = new double[ 2 ];        // destination coordinates
    int i, j, u, v;

    output.init( destWidth, destHeight, RasterGraphics.MODE_RGB, 3 );
    pixel[0] = background.getRed();
    pixel[1] = background.getGreen();
    pixel[2] = background.getBlue();
    pixel[3] = background.getAlpha();
    output.setRectangle( 0, 0, destWidth, destHeight, pixel );

    if ( interpolate )                    // Quincunx-interpolated pixels go first:
    {
      int[] accum = new int[ 4 ];         // accumulator
      for ( j = 0; j < height - 1; j++ )
        for ( i = 0; i < width - 1; i++ ) // one interpolated pixel
        {
          func.f( (i + 0.5) * domKx + domAx, (j + 0.5) * domKy + domAy, uv );
          u = (int)( uv[0] * rngKx + rngAx );
          v = (int)( uv[1] * rngKy + rngAy );

          if ( u >= 0 && u < destWidth &&
               v >= 0 && v < destHeight )
          {
            input.getRGB( i,   j,   accum );
            input.getRGB( i+1, j,   pixel );
            accum[0] += pixel[0];
            accum[1] += pixel[1];
            accum[2] += pixel[2];
            accum[3] += pixel[3];
            input.getRGB( i,   j+1, pixel );
            accum[0] += pixel[0];
            accum[1] += pixel[1];
            accum[2] += pixel[2];
            accum[3] += pixel[3];
            input.getRGB( i+1, j+1, pixel );
              // average:
            accum[0] = (accum[0] + pixel[0] + 2) >> 2;
            accum[1] = (accum[1] + pixel[1] + 2) >> 2;
            accum[2] = (accum[2] + pixel[2] + 2) >> 2;
            accum[3] = (accum[3] + pixel[3] + 2) >> 2;
            output.putPixel( u, v, accum );
          }
        }
    }

    for ( j = 0; j < height; j++ )
      for ( i = 0; i < width; i++ )       // one source pixel
      {
        func.f( i * domKx + domAx, j * domKy + domAy, uv );
        u = (int)( uv[0] * rngKx + rngAx );
        v = (int)( uv[1] * rngKy + rngAy );

        if ( u >= 0 && u < destWidth &&
             v >= 0 && v < destHeight )
        {
          input.getRGB( i, j, pixel );
          output.putPixel( u, v, pixel );
        }
      }

    return true;
  }

  //==========================================================================
  //  Interface Property:

  public final static String BACKGROUND  = "Background color";
  public final static String INTERPOLATE = "Quincunx";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // BACKGROUND:
    if ( key.compareTo(BACKGROUND) == 0 )
    {
      background = colorProperty( value, background );
      return;
    }
      // INTERPOLATE:
    if ( key.compareTo(INTERPOLATE) == 0 )
    {
      interpolate = booleanProperty( value, interpolate );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(BACKGROUND) == 0 )
      return background;
    if ( key.compareTo(INTERPOLATE) == 0 )
      return interpolate;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "WarpEngine";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerToFunctionR2ToR2AndRasterGraphicsAndRasterGraphics";

  /** Object description. */
  private final static String DESCRIPTION = "Simple warping engine (implements naive forward mapping).";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,    IFACE+"Trigger" );
      t.newOutputPlug( PL_FUNCTION, IFACE+"FunctionR2ToR2" );
      t.newOutputPlug( "image1",    IFACE+"RasterGraphics" );
      t.newOutputPlug( "image2",    IFACE+"RasterGraphics" );
        // Property:
      t.propBegin( BACKGROUND, TYPE_OBJECT, "Background color", true );
        t.propDefault( "[0;0;0]" );
        t.propManipulator( "ParamColor" );
      t.propEnd();
      t.propBegin( INTERPOLATE, TYPE_BOOLEAN, "Quincunx", true );
        t.propDefault( true );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
