package cz.cuni.jagrlib.testing;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.RegPiece;

/**
 * Interactive image filter example - image warping specimen.
 *
 * @version 0.25 $Rev: 406 $ $Date: 2007-03-25 00:06:58 +0100 (ne, 25 III 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/WarpingSpecimen.java">WarpingSpecimen.java</a>
 */
public class WarpingSpecimen extends DefaultInputListener implements Trigger
{
  //==========================================================================
  //  Data:

  /** GUI actions/states. */
  protected static enum State
  {
    /** Nothing-to-do state. */
    NONE,
    /** Brand new arrow. */
    ARROW_NEW,
    /** Arrow dragging. */
    ARROW,
    /** Arrow finishing. */
    ARROW_FINISH,
    /** Clear the result image. */
    CLEAR,
    /** Loop break (will destroy the window). */
    BREAK,
    /** Loop quit (the window was destroyed before). */
    QUIT,
  }

    // Accelerators:

  /** Input raster image (for reset). */
  protected RasterGraphics input   = null;

  /** Output raster image. */
  protected RasterGraphics output  = null;

  /** Preview window. */
  protected GraphicsViewer preview = null;

  /** Image width in pixels. */
  protected int width;

  /** Image height in pixels. */
  protected int height;

  /** Actual action/state. */
  protected State state = State.NONE;

  /** Actual HUD line identifier. */
  protected int hudId = 0;

  /** Start-point of the actual arrow. */
  protected int x0, y0;

  /** End-point of the actual arrow. */
  protected int gx, gy;

  /** End-point marks' color. */
  protected final static int[] COLOR = { 0, 0, 0, 255 };

  /** End-point radius. */
  protected final static int   SIZE  = 2;

  //==========================================================================
  //  Support:

  /** Re-initializes the output image. */
  protected void resetImage ()
  {
    int[] pixel = new int[ 4 ];           // I/O pixel

    for ( int j = 0; j < height; j++ )
      for ( int i = 0; i < width; i++ )
      {
        input.getRGB( i, j, pixel );
        output.putPixel( i, j, pixel );
      }
  }

  //==========================================================================
  //  Interface Trigger:

  /**
   * Starts image transformation.
   *
   * @param  type The action type (0 .. initial rendering, 1 .. message loop).
   * @return <tt>true</tt> if the action was successful.
   */
  public boolean fire ( int type )
  {
    if ( type == 0 )
    {
        // initialization:
      input   = (RasterGraphics)getInterface( "image1",  IFACE+"RasterGraphics" );
      output  = (RasterGraphics)getInterface( "image2",  IFACE+"RasterGraphics" );
      preview = (GraphicsViewer)getInterface( PL_OUTPUT, IFACE+"GraphicsViewer" );

        // output image:
      width   = input.getWidth();
      height  = input.getHeight();
      output.init( width, height, RasterGraphics.MODE_RGB, 3 );

      resetImage();
    }
    else
        // interaction:
      do
      {
          // local copies of state variables (to minimize the critical section):
        State cState;
        int cx0, cy0, cgx, cgy;

        synchronized ( this )
        {
          try
          {
            wait();
          }
          catch ( InterruptedException e )
          {
            state = State.BREAK;
          }

            // copy state/action into cState and change the new state..
          cState = state;
          switch ( state )
          {
            case ARROW_NEW:
              state = State.ARROW;
              break;
            case ARROW_FINISH:
            case CLEAR:
              state = State.NONE;
              break;
          }

            // copy actual coordinates into c*:
          cx0 = x0; cy0 = y0;
          cgx = gx; cgy = gy;
        }

        switch ( cState )
        {
          case ARROW_NEW:
            output.setRectangle( cx0 - SIZE, cy0 - SIZE, cx0 + SIZE, cy0 + SIZE, COLOR );
            hudId = preview.hudGetFreeId();

          case ARROW:
            preview.hudSetLine( hudId, cx0, cy0, cgx, cgy );
            preview.invalidate( Math.min( cgx, cx0 ) - SIZE, Math.max( cgx, cx0 ) + SIZE,
                                Math.min( cgy, cy0 ) - SIZE, Math.max( cgy, cy0 ) + SIZE );
            preview.repaint();
            break;

          case ARROW_FINISH:
            output.setRectangle( cgx - SIZE, cgy - SIZE, cgx + SIZE, cgy + SIZE, COLOR );
            preview.invalidate(  cgx - SIZE, cgx + SIZE, cgy - SIZE, cgy + SIZE );
            preview.repaint();
            break;

          case CLEAR:
            resetImage();
            preview.hudClear();
            preview.repaintAll();
            break;

          case BREAK:
            preview.destroy();

          case QUIT:
            return true;
        }
      }
      while ( true );

    return true;
  }

  /**
   * Mouse button was pressed/released.
   *
   * @param sender Originator of the event
   * @param when System time in milliseconds.
   * @param x Horizontal screen position in pixels.
   * @param y Vertical screen position in pixels.
   * @param button Button number ({@link #BUTTON1}, etc.).
   * @param press Was it press or release?
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public void mouseButton ( Object sender, long when, int x, int y, int button, boolean press, int flags )
  {
    if ( button != BUTTON1 ) return;

    synchronized ( this )
    {
      if ( press )
      {
        x0 = gx = x;
        y0 = gy = y;
        state = State.ARROW_NEW;
        notify();
      }
      else
        if ( state == State.ARROW )
        {
          gx = x;
          gy = y;
          state = State.ARROW_FINISH;
          notify();
        }
        else
        if ( state == State.ARROW_NEW )
          state = State.NONE;
    }
  }

  /**
   * Keyboard key was pressed/released.
   *
   * @param sender Originator of the event
   * @param when System time in milliseconds.
   * @param press Was it press or release?
   * @param ch Character represented by the key.
   * @param keyCode Key code.
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public void keyboard ( Object sender, long when, boolean press, char ch, int keyCode, int flags )
  {
    if ( ch == 'q' )                         // break
    synchronized ( this )
    {
      state = State.BREAK;
      notify();
      return;
    }

    if ( ch == 'c' )                         // clear HUD
    synchronized ( this )
    {
      state = State.CLEAR;
      notify();
      return;
    }
  }

  /**
   * Mouse pointer was moved.
   *
   * @param sender Originator of the event
   * @param when System time in milliseconds.
   * @param x Horizontal screen position in pixels.
   * @param y Vertical screen position in pixels.
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public synchronized void mouseMove ( Object sender, long when, int x, int y, int flags )
  {
    if ( state == State.ARROW )
    {
      preview.invalidate( Math.min( gx, x0 ) - SIZE, Math.max( gx, x0 ) + SIZE,
                          Math.min( gy, y0 ) - SIZE, Math.max( gy, y0 ) + SIZE );
      gx = x;
      gy = y;
      notify();
    }
  }

  /**
   * Window was closed (user interaction..).
   *
   * @param sender Originator of the event
   */
  @Override
  public synchronized void windowClose ( Object sender )
  {
    state = State.QUIT;
    notify();
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;
      // CELL_SIZE:
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or <tt>null</tt>.
   */
  public Object get ( String key )
  {
    if ( key == null ) return null;
    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "WarpingSpecimen";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "TriggerAndInputListenerToRasterGraphics2AndGraphicsViewer";

  /** Object description. */
  private final static String DESCRIPTION = "Image warping specimen.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_RASTER + "." + C_FILTER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT,   IFACE+"Trigger" );
      t.newInputPlug( PL_DIRECT,  IFACE+"InputListener" );
      t.newOutputPlug( "image1",  IFACE+"RasterGraphics" );
      t.newOutputPlug( "image2",  IFACE+"RasterGraphics" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"GraphicsViewer" );
        // Property:
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
