package cz.cuni.jagrlib.worker;

import java.util.*;
import org.nfunk.jep.JEP;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Meta-worker for animation rendering.
 * Can be connected to any meta-worker-enabled {@link Worker} (see {@link Worker#REJECT}) and direct it
 * to create sequence of pictures.
 *
 * <p>Expressions are handled by the {@link org.nfunk.jep.JEP} class. Two special pre-defined variables
 * are provided to control the animation:</p>
 * <ul>
 * <li><b>{@code t}</b> - floating-point time variable varying from {@code 0.0} to {@code 1.0}</li>
 * <li><b>{@code frame}</b> - integer variable containing current frame number (from {@code 0} to {@link #frames}-1)</li>
 * </ul>
 *
 * <p>Template properties ({@link #fileNameTemplate}, {@link #variable1Template}, ..) are in format acceptable
 * by the {@link java.lang.String#format(String,Object...) String.format()} method (ala {@code sprintf()} in C).
 * If integer format-tag is used (<b>{@code h}</b>, <b>{@code d}</b>, <b>{@code o}</b>, <b>{@code x}</b>),
 * expression value is rounded to the nearest integer.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/AnimationWorker.java">AnimationWorker.java</a>
 */
public class AnimationWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Time variable identifier. */
  public final static String VAR_T = "t";

  /** Frame variable identifier. */
  public final static String VAR_FRAME = "frame";

  /** Number of frames to be rendered. */
  protected int frames = 200;

  /** Actual frame. */
  protected int frame = -1;

  /** File-name property name. */
  protected String fileNameProperty = BitStream.STREAM_NAME;

  /** File-name template. */
  protected String fileNameTemplate = "out%04d.png";

  /** Number of first generated frame. */
  protected int frameOffset = 0;

  /** Variable1: property name. */
  protected String variable1Property = "";

  /** Variable1: expression string. */
  protected String variable1Expression = VAR_T;

  /** Variable1: format template. */
  protected String variable1Template = "%.2f";

  /** Variable2: property name. */
  protected String variable2Property = "";

  /** Variable2: expression string. */
  protected String variable2Expression = VAR_T;

  /** Variable2: format template. */
  protected String variable2Template = "%.2f";

  /** Variable3: property name. */
  protected String variable3Property = "";

  /** Variable3: expression string. */
  protected String variable3Expression = VAR_T;

  /** Variable3: format template. */
  protected String variable3Template = "%.2f";

  //==========================================================================
  //  Interface Worker:

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final Property   module1 = (Property)getInterface( "module1", IFACE+"Property" );
    final Property   module2 = (Property)getInterface( "module2", IFACE+"Property" );
    final Property   module3 = (Property)getInterface( "module3", IFACE+"Property" );
    final Property   out     = (Property)getInterface( PL_STREAM, IFACE+"Property" );
    final Worker     worker  = (Worker)  getInterface( PL_OUTPUT, IFACE+"Worker" );

    if ( worker == null ||
         frames <= 0 ) return;

    if ( pl != null ) pl.progress( 0L, frames );

      // time-varying variables:
    double t  = 0.0;
    double dt = (frames > 1) ? 1.0 / (frames - 1) : 0.0;
    double value;
    String valString;

      // pre-parsed expressions:
    JEP v1 = null;
    JEP v2 = null;
    JEP v3 = null;
    boolean isVar1 = module1 != null &&
                     variable1Property.length() > 0 &&
                     variable1Expression != null &&
                     variable1Expression.length() > 0;
    boolean isVar2 = module2 != null &&
                     variable2Property.length() > 0 &&
                     variable2Expression != null &&
                     variable2Expression.length() > 0;
    boolean isVar3 = module3 != null &&
                     variable3Property.length() > 0 &&
                     variable3Expression != null &&
                     variable3Expression.length() > 0;
    boolean isInt1 = true;
    boolean isInt2 = true;
    boolean isInt3 = true;

    if ( isVar1 )
    {
      v1 = Formula.initExpression( variable1Expression );
      isInt1 = Formula.isIntFormat( variable1Template );
    }
    if ( isVar2 )
    {
      v2 = Formula.initExpression( variable2Expression );
      isInt2 = Formula.isIntFormat( variable2Template );
    }
    if ( isVar3 )
    {
      v3 = Formula.initExpression( variable3Expression );
      isInt3 = Formula.isIntFormat( variable3Template );
    }

      // rendering loop:
    for ( frame = 0;
          frame < frames && !userBreak;
          frame++, t += dt )              // render one frame
    {
        // set time-varying values:
      if ( isVar1 )
      {
        v1.addVariable( VAR_T,     t );
        v1.addVariable( VAR_FRAME, frame );
        value     = v1.getValue();
        valString = isInt1 ? String.format( Locale.US, variable1Template, Math.round( value ) ) :
                             String.format( Locale.US, variable1Template, value );
        LogFile.debug( "Setting " + variable1Property + " = '" + valString + "'" );
        module1.set( variable1Property, valString );
      }

      if ( isVar2 )
      {
        v2.addVariable( VAR_T,     t );
        v2.addVariable( VAR_FRAME, frame );
        value     = v2.getValue();
        valString = isInt2 ? String.format( Locale.US, variable2Template, Math.round( value ) ) :
                             String.format( Locale.US, variable2Template, value );
        LogFile.debug( "Setting " + variable2Property + " = '" + valString + "'" );
        module2.set( variable2Property, valString );
      }

      if ( isVar3 )
      {
        v3.addVariable( VAR_T,     t );
        v3.addVariable( VAR_FRAME, frame );
        value     = v3.getValue();
        valString = isInt3 ? String.format( Locale.US, variable3Template, Math.round( value ) ) :
                             String.format( Locale.US, variable3Template, value );
        LogFile.debug( "Setting " + variable3Property + " = '" + valString + "'" );
        module3.set( variable3Property, valString );
      }

      if ( out != null &&
           fileNameProperty.length() > 0 )
      {
        valString = String.format( Locale.US, fileNameTemplate, frame + frameOffset );
        LogFile.debug( "Setting " + fileNameProperty + " = '" + valString + "'" );
        out.set( fileNameProperty, valString );
      }

      worker.run();
      if ( pl != null ) pl.progress( frame + 1, frames );
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String FRAMES               = "Frames";
  public final static String FILE_NAME_PROPERTY   = "FileName property";
  public final static String FILE_NAME_TEMPLATE   = "FileName template";
  public final static String FRAME_OFFSET         = "Frame offset";
  public final static String VARIABLE1_PROPERTY   = "Variable1 property";
  public final static String VARIABLE1_TEMPLATE   = "Variable1 template";
  public final static String VARIABLE1_EXPRESSION = "Variable1 expression";
  public final static String VARIABLE2_PROPERTY   = "Variable2 property";
  public final static String VARIABLE2_TEMPLATE   = "Variable2 template";
  public final static String VARIABLE2_EXPRESSION = "Variable2 expression";
  public final static String VARIABLE3_PROPERTY   = "Variable3 property";
  public final static String VARIABLE3_TEMPLATE   = "Variable3 template";
  public final static String VARIABLE3_EXPRESSION = "Variable3 expression";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // FRAMES:
    if ( key.compareTo(FRAMES) == 0 )
    {
      frames = intProperty( value, frames );
      return;
    }
      // FILE_NAME_PROPERTY:
    if ( key.compareTo(FILE_NAME_PROPERTY) == 0 )
    {
      fileNameProperty = stringProperty( value, fileNameProperty );
      return;
    }
      // FILE_NAME_TEMPLATE:
    if ( key.compareTo(FILE_NAME_TEMPLATE) == 0 )
    {
      fileNameTemplate = stringProperty( value, fileNameTemplate );
      return;
    }
      // FRAME_OFFSET:
    if ( key.compareTo(FRAME_OFFSET) == 0 )
    {
      frameOffset = intProperty( value, frameOffset, 0, 1000000 );
      return;
    }
      // VARIABLE1_PROPERTY:
    if ( key.compareTo(VARIABLE1_PROPERTY) == 0 )
    {
      variable1Property = stringProperty( value, variable1Property );
      return;
    }
      // VARIABLE1_TEMPLATE:
    if ( key.compareTo(VARIABLE1_TEMPLATE) == 0 )
    {
      variable1Template = stringProperty( value, variable1Template );
      return;
    }
      // VARIABLE1_EXPRESSION:
    if ( key.compareTo(VARIABLE1_EXPRESSION) == 0 )
    {
      variable1Expression = stringProperty( value, variable1Expression );
      return;
    }
      // VARIABLE2_PROPERTY:
    if ( key.compareTo(VARIABLE2_PROPERTY) == 0 )
    {
      variable2Property = stringProperty( value, variable2Property );
      return;
    }
      // VARIABLE2_TEMPLATE:
    if ( key.compareTo(VARIABLE2_TEMPLATE) == 0 )
    {
      variable2Template = stringProperty( value, variable2Template );
      return;
    }
      // VARIABLE2_EXPRESSION:
    if ( key.compareTo(VARIABLE2_EXPRESSION) == 0 )
    {
      variable2Expression = stringProperty( value, variable2Expression );
      return;
    }
      // VARIABLE3_PROPERTY:
    if ( key.compareTo(VARIABLE3_PROPERTY) == 0 )
    {
      variable3Property = stringProperty( value, variable3Property );
      return;
    }
      // VARIABLE3_TEMPLATE:
    if ( key.compareTo(VARIABLE3_TEMPLATE) == 0 )
    {
      variable3Template = stringProperty( value, variable3Template );
      return;
    }
      // VARIABLE3_EXPRESSION:
    if ( key.compareTo(VARIABLE3_EXPRESSION) == 0 )
    {
      variable3Expression = stringProperty( value, variable3Expression );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(FRAMES) == 0 )
      return frames;
    if ( key.compareTo(FILE_NAME_PROPERTY) == 0 )
      return fileNameProperty;
    if ( key.compareTo(FILE_NAME_TEMPLATE) == 0 )
      return fileNameTemplate;
    if ( key.compareTo(FRAME_OFFSET) == 0 )
      return frameOffset;
    if ( key.compareTo(VARIABLE1_PROPERTY) == 0 )
      return variable1Property;
    if ( key.compareTo(VARIABLE1_TEMPLATE) == 0 )
      return variable1Template;
    if ( key.compareTo(VARIABLE1_EXPRESSION) == 0 )
      return variable1Expression;
    if ( key.compareTo(VARIABLE2_PROPERTY) == 0 )
      return variable2Property;
    if ( key.compareTo(VARIABLE2_TEMPLATE) == 0 )
      return variable2Template;
    if ( key.compareTo(VARIABLE2_EXPRESSION) == 0 )
      return variable2Expression;
    if ( key.compareTo(VARIABLE3_PROPERTY) == 0 )
      return variable3Property;
    if ( key.compareTo(VARIABLE3_TEMPLATE) == 0 )
      return variable3Template;
    if ( key.compareTo(VARIABLE3_EXPRESSION) == 0 )
      return variable3Expression;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "AnimationWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForWorkerAndProperty";

  /** Object description. */
  private final static String DESCRIPTION = "Meta-worker for animation rendering.";

  /** Object category. */
  protected final static String CATEGORY = "system." + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_OUTPUT, IFACE+"Worker" );
      t.newOptOutputPlug( "module1", IFACE+"Property" );
      t.newOptOutputPlug( "module2", IFACE+"Property" );
      t.newOptOutputPlug( "module3", IFACE+"Property" );
      t.newOptOutputPlug( PL_STREAM, IFACE+"Property" );
        // Property:
      t.propBegin( FRAMES, TYPE_INTEGER, "Number of frames", true );
        t.propDefault( 200 );
      t.propEnd();
      t.propBegin( FILE_NAME_PROPERTY, TYPE_STRING, "File-name property", true );
        t.propDefault( BitStream.STREAM_NAME );
      t.propEnd();
      t.propBegin( FILE_NAME_TEMPLATE, TYPE_STRING, "File-name template (to be used in String.format())", true );
        t.propDefault( "out%04d.png" );
      t.propEnd();
      t.propBegin( FRAME_OFFSET, TYPE_INTEGER, "Number of first generated frame", true );
        t.propDefault( 0 );
        t.propBounds(  0, 1000000 );
      t.propEnd();
      t.propBegin( VARIABLE1_PROPERTY, TYPE_STRING, "Variable1 property", true );
        t.propDefault( "" );
      t.propEnd();
      t.propBegin( VARIABLE1_TEMPLATE, TYPE_STRING, "Variable1 template (to be used in String.format())", true );
        t.propDefault( "%.2f" );
      t.propEnd();
      t.propBegin( VARIABLE1_EXPRESSION, TYPE_STRING, "Variable1 expression", true );
        t.propDefault( "t" );
      t.propEnd();
      t.propBegin( VARIABLE2_PROPERTY, TYPE_STRING, "Variable2 property", true );
        t.propDefault( "" );
      t.propEnd();
      t.propBegin( VARIABLE2_TEMPLATE, TYPE_STRING, "Variable2 template (to be used in String.format())", true );
        t.propDefault( "%.2f" );
      t.propEnd();
      t.propBegin( VARIABLE2_EXPRESSION, TYPE_STRING, "Variable2 expression", true );
        t.propDefault( "t" );
      t.propEnd();
      t.propBegin( VARIABLE3_PROPERTY, TYPE_STRING, "Variable3 property", true );
        t.propDefault( "" );
      t.propEnd();
      t.propBegin( VARIABLE3_TEMPLATE, TYPE_STRING, "Variable3 template (to be used in String.format())", true );
        t.propDefault( "%.2f" );
      t.propEnd();
      t.propBegin( VARIABLE3_EXPRESSION, TYPE_STRING, "Variable3 expression", true );
        t.propDefault( "t" );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
