package cz.cuni.jagrlib.worker;

import java.awt.event.KeyEvent;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: interactive ant-city rendering.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.24
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/AntCityWorker.java">AntCityWorker.java</a>
 */
public class AntCityWorker extends DefaultInputListener implements Worker
{
  //==========================================================================
  //  Data:

  /** Redraw period in milliseconds (or {@code 0} if no periodic redraw has to be done). */
  protected int period = 500;

  /** Continue flag for regular re-draw loop. */
  protected volatile boolean cont = true;

  /** Accelerator reference to the viewer window. */
  protected GraphicsViewer window;

  /** Redner object. */
  protected Render3D render;

  /** Timer thread used for automatic redraw. */
  protected Thread timerThread;

  /** Window close semaphore. */
  protected Semaphore sem = null;

  /** View direction as ordinal number 0 .. 3. */
  protected int viewOrd = 0;

  /** All possible view direction vectors. */
  protected static final double[][] views =
  {
    { -0.5, -0.5, -0.5 },
    {  0.5, -0.5, -0.5 },
    {  0.5,  0.5, -0.5 },
    { -0.5,  0.5, -0.5 },
  };

  /** Actual view direction. */
  protected double[] view = views[ viewOrd ];

  /** Constant 'up' vector. */
  protected static final double[] up = { 0.0, 0.0, 1.0 };

  /** Actual 'lookAt' point. */
  protected double[] lookAt = { 10.0, 10.0, 0.0 };

  /** Actual zoom factor (horizontal brick size). */
  protected double zoom = 30.0;

  /** Total rendering time of last 'frames' frames. */
  protected long totalTime = 0L;

  /** Number of recent frames to be measured. */
  protected int frames = 0;

  //==========================================================================
  //  Interface Worker:

  /**
   * Tries to stop the thread/computation..
   * The process should yield itself.
   * For {@link Runnable} implementators: {@link Thread#interrupt} will be called afterwards.
   */
  @Override
  public void stop ()
  {
    userBreak = true;
    cont      = false;
    if ( sem != null )
      sem.semSignal( 1 );
  }

  /** Executive routine of the working thread. */
  public void run ()
  {
                     window = (GraphicsViewer)getInterface( PL_WINDOW, IFACE+"GraphicsViewer" );
                     render = (Render3D)      getInterface( PL_OUTPUT, IFACE+"Render3D" );
    final SampleData data   = (SampleData)    getInterface( PL_DATA,   IFACE+"SampleData" );
    if ( window == null ||
         render == null ||
         data   == null ) return;

      // initialize data array:
    int[] dim = { 32, 100, 100 };
    data.resetData( SampleData.SAMPLE_RGBA, false, dim );
    int[] arr = new int[ 320000 ];
    RandomJames rnd = new RandomJames();
    int i, j;

    for ( i = 0; i < 10000; i++ )         // one cell in the floor projection
    {
      int col = (int)( rnd.uniformNumber() * 0x1000000 ) | 0xff000000;
      int hei = rnd.randomInteger( 1, 30 );
      for ( j = 0; j < hei; j++ )
        arr[ i + 10000 * j ] = col;
      for ( ; j < 32; j++ )
        arr[ i + 10000 * j ] = 0;
    }

    data.setDataArray( arr, 0, null, null );
    arr = null;

      // render window:
    sem = new Semaphore( 0 );
    window.addSemaphore( sem );
    window.set( GraphicsViewer.WINDOW_TITLE, "Ant City" );
    render.setProjection( view, up, lookAt, zoom );

      // viewer thread:
    GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( window );
    runIt.safeStart();

      // timer/render thread:
    timerThread = new Thread( "JaGrLib - AntCityWorker timer" )
    {
      @Override
      public void run ()
      {
        while ( cont )
        {
          long startTime = System.currentTimeMillis();
          render.render();
          totalTime += System.currentTimeMillis() - startTime;
          if ( ++frames == 10 )
          {
            window.set( GraphicsViewer.WINDOW_TITLE, "Ant City - " + ((totalTime + 5L) / 10L) + "ms" );
            totalTime = 0L;
            frames    = 0;
          }
          window.repaintAll();
          try
          {
            sleep( (period > 0L) ? period : 60000L );
          }
          catch ( InterruptedException e )
          {
            interrupted();
          }
        }
      }
    };
    try
    {
      timerThread.setPriority( Thread.MIN_PRIORITY );
    }
    catch ( Exception e )
    {
      LogFile.exception( "Cannot set priority of timer thread! " + e.getMessage(), e );
    }
    try
    {
      timerThread.start();
    }
    catch ( IllegalThreadStateException e )
    {
      LogFile.exception( "Cannot start timer thread! " + e.getMessage(), e );
    }

    sem.semWait( 3600000L );
    cont = false;
    timerThread.interrupt();
  }

  //==========================================================================
  //  Interface InputListener:

  /**
   * Keyboard key was pressed/released.
   *
   * @param sender Originator of the event
   * @param when System time in milliseconds.
   * @param press Was it press or release?
   * @param ch Character represented by the key.
   * @param keyCode Key code.
   * @param flags Input system status (key shifts/alts, ..).
   */
  @Override
  public void keyboard ( Object sender, long when, boolean press, char ch, int keyCode, int flags )
  {
    if ( press ) return;

    switch ( keyCode )
    {
      case KeyEvent.VK_PLUS:              // change view vector
        if ( ++viewOrd >= 4 ) viewOrd = 0;
        view = views[ viewOrd ];
        render.setProjection( view, up, lookAt, zoom );
        break;

      case KeyEvent.VK_MINUS:             // change view vector
        if ( --viewOrd < 0 ) viewOrd = 3;
        view = views[ viewOrd ];
        render.setProjection( view, up, lookAt, zoom );
        break;

      case KeyEvent.VK_PAGE_UP:           // zoom-in
        zoom *= 1.05;
        render.setProjection( view, up, lookAt, zoom );
        break;

      case KeyEvent.VK_PAGE_DOWN:         // zoom-out
        zoom /= 1.05;
        render.setProjection( view, up, lookAt, zoom );
        break;

      case KeyEvent.VK_LEFT:              // go to the left
        lookAt[0] += 1.0;
        render.setProjection( view, up, lookAt, zoom );
        break;

      case KeyEvent.VK_RIGHT:             // go to the right
        lookAt[0] -= 1.0;
        render.setProjection( view, up, lookAt, zoom );
        break;

      case KeyEvent.VK_UP:                // go forth
        lookAt[1] += 1.0;
        render.setProjection( view, up, lookAt, zoom );
        break;

      case KeyEvent.VK_DOWN:              // go back
        lookAt[1] -= 1.0;
        render.setProjection( view, up, lookAt, zoom );
        break;

      case KeyEvent.VK_HOME:              // reset initial view
        lookAt[0] = lookAt[1] = 10.0;
        render.setProjection( view, up, lookAt, zoom );
        break;

      default:
        return;
    }

    timerThread.interrupt();
  }

  //==========================================================================
  //  Interface Property:

  public final static String PERIOD  = "Redraw period";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // PERIOD:
    if ( key.compareTo(PERIOD) == 0 )
    {
      period = intProperty( value, period, 0, 10000 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(PERIOD) == 0 )
      return period;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "AntCityWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForAntCity";

  /** Object description. */
  private final static String DESCRIPTION = "Interactive ant-city rendering. " +
                                            "Uses keyboard to change projection parameters";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(  PL_INPUT,  IFACE+"InputListener" );
      t.newOutputPlug( PL_WINDOW, IFACE+"GraphicsViewer" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"Render3D" );
      t.newOutputPlug( PL_DATA,   IFACE+"SampleData" );
        // Property:
      t.propBegin( PERIOD, TYPE_INTEGER, "Auto-redraw period in milliseconds", true );
        t.propDefault( 500 );
        t.propBounds(  0, 10000 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
