package cz.cuni.jagrlib.worker;

import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;
import cz.cuni.jagrlib.piece.AudioReplay;

/**
 * Worker for audio replay and filtering.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/AudioFilterWorker.java">AudioFilterWorker.java</a>
 */
public class AudioFilterWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Audio filter (should be stopped in {@link #stop}). */
  protected Trigger filter = null;

  //==========================================================================
  //  Interface Worker:

  /**
   * Tries to stop the thread/computation..
   * The process should yield itself.
   * For {@link Runnable} implementators: {@link Thread#interrupt} will be called afterwards.
   */
  @Override
  public synchronized void stop ()
  {
    userBreak = true;
    if ( filter != null )
      filter.stop();
  }

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final DataFileFormat input   = (DataFileFormat)getInterface( PL_INPUT,  IFACE+"DataFileFormat" );
    final DataFileFormat output  = (DataFileFormat)getInterface( PL_OUTPUT, IFACE+"DataFileFormat" );
    final Trigger        replay1 = (Trigger)       getInterface( "replay1", IFACE+"Trigger" );
    final Trigger        replay2 = (Trigger)       getInterface( "replay2", IFACE+"Trigger" );

    if ( input == null ) return;

    synchronized ( this )
    {
      filter = (Trigger)getInterface( PL_FILTER, IFACE+"Trigger" );
      if ( filter != null &&
           output == null )
      {
        filter = null;
        return;
      }
    }

      // load source audio file:
    try
    {
      input.loadFile( (String)null, "r" ); // read-only mode
    }
    catch ( IOException e )
    {
      throw new JaGrLibException( "Error reading audio file!", e );
    }

      // replay source audio file:
    if ( replay1 != null )
    {
      replay1.set( AudioReplay.WINDOW_TITLE, "Input audio file" );
      replay1.fire( 0 );
    }

      // do the filtering job..
    try
    {
      if ( filter != null )
      {
        if ( output != null )
        {
          try
          {
            output.loadFile( (String)null, "rw" ); // read/write mode
          }
          catch ( IOException e )
          {
            throw new JaGrLibException( "Error creating audio file!", e );
          }
        }

        filter.fire( 0 );

          // replay target audio file:
        if ( replay2 != null )
        {
          replay2.set( AudioReplay.WINDOW_TITLE, "Output audio file" );
          replay2.fire( 0 );
        }
      }
    }
    finally
    {
      synchronized ( this )
      {
        filter = null;
      }
        // save the output audio file:
      try
      {
        input.closeFile();
        if ( output != null )
          output.closeFile();
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error closing audio file(s)!", e );
      }
    }
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "AudioilterWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForAudioFilter";

  /** Object description. */
  private final static String DESCRIPTION = "Tests audio files and filters.";

  /** Object category. */
  protected final static String CATEGORY = C_1D + '.' + C_RASTER + '.' + C_FILTER + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_INPUT,  IFACE+"DataFileFormat" );
      t.newOptOutputPlug( PL_OUTPUT, IFACE+"DataFileFormat" );
      t.newOptOutputPlug( "replay1", IFACE+"Trigger" );
      t.newOptOutputPlug( "replay2", IFACE+"Trigger" );
      t.newOptOutputPlug( PL_FILTER, IFACE+"Trigger" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
