package cz.cuni.jagrlib.worker;

import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: Cubic curve drawing test composition.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/BezierWorker.java">BezierWorker.java</a>
 */
public class BezierWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Number of lines to be drawn. */
  protected int repeat = 100;

  /** Set negative value for randomize. */
  protected int seed   = -1;

  //==========================================================================
  //  Run:

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final CurveRender    bezier1 = (CurveRender)   getInterface( "output1",  IFACE+"CurveRender" );
    final CurveRender    bezier2 = (CurveRender)   getInterface( "output2",  IFACE+"CurveRender" );
    final CurveRender    bezier3 = (CurveRender)   getInterface( "output3",  IFACE+"CurveRender" );
    final SolidBrush     brush   = (SolidBrush)    getInterface( PL_BITMASK, IFACE+"SolidBrush" );
    final GraphicsViewer preview = (GraphicsViewer)getInterface( PL_RASTER,  IFACE+"GraphicsViewer" );
    final DataFileFormat bak     = (DataFileFormat)getInterface( PL_TRIGGER, IFACE+"DataFileFormat" );

    if ( preview != null )
    {
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

    int[] colorOrig  = { 128, 255, 255, 255 };
    int[] colorOver2 = { 255,  50,  50, 255 };
    int[] colorOver3 = { 100, 255, 100, 255 };
    int[] cp = new int[ 8 ];
    int i, j;
    double width;

      // random generator:
    RandomJames rnd;
    if ( seed >= 0 )
      rnd = new RandomJames( seed >>> 15, seed & 0x7fff );
    else
    {
      rnd = new RandomJames();
      rnd.randomize();
    }

    for ( i = 0; i++ < repeat && !userBreak; )
    {
        // random cubic curve (with random line width):
      for ( j = 0; j < 8; j++ )
        cp[j] = rnd.randomInteger( 0, 511 );
      width = rnd.randomDouble( 0.5, 3.5 );

      if ( bezier1 != null )
      {
        bezier1.set( Render.LINE_WIDTH, width );
        bezier1.drawCurve( cp, 0, 4, 2 );
        brush.setColor( colorOrig );
        brush.fill();
      }

      if ( bezier2 != null )
      {
        bezier2.set( Render.LINE_WIDTH, width );
        bezier2.drawCurve( cp, 0, 4, 2 );
        brush.setColor( colorOver2 );
        brush.fill();
      }

      if ( bezier3 != null )
      {
        bezier3.set( Render.LINE_WIDTH, width );
        bezier3.drawCurve( cp, 0, 4, 2 );
        brush.setColor( colorOver3 );
        brush.fill();
      }

      if ( (i & 31) == 0 )
      {
        if ( pl != null )
          pl.progress( i, repeat );
        if ( preview != null )
          preview.repaintAll();
      }
    }

      // finish repainting:
    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - FINISHED" );
      preview.repaintAll();
      preview.stopRepaintLoop();
    }

      // save the result image:
    if ( bak != null )
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error writting image!", e );
      }
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // REPEAT:
    if ( key.compareTo(REPEAT) == 0 )
    {
      repeat = intProperty( value, repeat );
      return;
    }
      // SEED:
    if ( key.compareTo(SEED) == 0 )
    {
      seed = intProperty( value, seed );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(REPEAT) == 0 )
      return repeat;
    if ( key.compareTo(SEED) == 0 )
      return seed;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "BezierWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForCubicCurve";

  /** Object description. */
  private final static String DESCRIPTION = "Draws random cubic curves.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DRAW + "." + C_CURVE+ "." + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOptOutputPlug( "output1",  IFACE+"CurveRender" );
      t.newOptOutputPlug( "output2",  IFACE+"CurveRender" );
      t.newOptOutputPlug( "output3",  IFACE+"CurveRender" );
      t.newOutputPlug(    PL_BITMASK, IFACE+"SolidBrush" );
      t.newOptOutputPlug( PL_RASTER,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER, IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( REPEAT, TYPE_INTEGER, "Number of random curves", true );
        t.propDefault( 100 );
      t.propEnd();
      t.propBegin( SEED, TYPE_INTEGER, "Random generator seed (-1 for randomize)", true );
        t.propDefault( -1 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
