package cz.cuni.jagrlib.worker;

import java.io.*;
import java.util.Locale;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: color reduce test composition.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/ColorReduceWorker.java">ColorReduceWorker.java</a>
 */
public class ColorReduceWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Colormap generator (should be stopped in {@link #stop}). */
  protected Trigger colormap = null;

  /** Image convertor mapper (should be stopped in {@link #stop}). */
  protected Trigger mapper = null;

  //==========================================================================
  //  Interface Worker:

  /**
   * Tries to stop the thread/computation..
   * The process should yield itself.
   * For {@link Runnable} implementators: {@link Thread#interrupt} will be called afterwards.
   */
  @Override
  public synchronized void stop ()
  {
    userBreak = true;
    if ( colormap != null )
      colormap.stop();
    if ( mapper != null )
      mapper.stop();
  }

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final DataFileFormat output   = (DataFileFormat)getInterface( PL_OUTPUT,   IFACE+"DataFileFormat" );
    final GraphicsViewer preview2 = (GraphicsViewer)getInterface( "raster2",   IFACE+"GraphicsViewer" );
    final RealFunction   compare  = (RealFunction)  getInterface( PL_COMPARE,  IFACE+"RealFunction" );

    synchronized ( this )
    {
      mapper   = (Trigger)getInterface( PL_DIRECT,   IFACE+"Trigger" );
      if ( mapper == null ) return;
      colormap = (Trigger)getInterface( PL_COLORMAP, IFACE+"Trigger" );
    }

    long millis = System.currentTimeMillis();

      // optional colormap construction step:
    if ( colormap != null )
    {
      if ( pl != null )
        colormap.setProgressListener( pl );

        // generate (adaptive) colormap and store it as 0-th item:
      colormap.fire( 0 );

      synchronized ( this )
      {
        colormap = null;
      }
    }

      // do the mapping (using 0-th colormap):
    mapper.fire( 0 );
    synchronized ( this )
    {
      mapper = null;
    }
    millis = System.currentTimeMillis() - millis;

      // compare reconstructed image with original:
    double error = 0.0;
    if ( compare != null )
      error = compare.f( 0 );

      // attach preview window to the output image & render it:
    if ( preview2 != null )
    {
      final double elapsed = 0.001 * millis;
      preview2.set( GraphicsViewer.WINDOW_TITLE,
                    String.format( Locale.US, (compare == null) ? "FINISHED (%.1f sec)" :
                                                                  "FINISHED (%.1f sec), err = %.1f", elapsed, error ) );
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview2 );
      runIt.safeStart();
    }

      // save the output image:
    if ( output != null )
      try
      {
        output.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error writting image!", e );
      }
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "ColorReduceWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForColorReduce";

  /** Object description. */
  private final static String DESCRIPTION = "Tests color reduce filters.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_RASTER + '.' + C_FILTER + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOptOutputPlug( PL_OUTPUT,   IFACE+"DataFileFormat" );
      t.newOptOutputPlug( "raster2",   IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_COLORMAP, IFACE+"Trigger" );
      t.newOutputPlug(    PL_DIRECT,   IFACE+"Trigger" );
      t.newOptOutputPlug( PL_COMPARE,  IFACE+"RealFunction" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
