package cz.cuni.jagrlib.worker;

import java.io.*;
import java.util.Locale;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: general data compression.
 * <ol>
 *  <li>optional preprocessing</li>
 *  <li>compression (mandatory)</li>
 *  <li>decompression = reconstruction</li>
 *  <li>optional saving of reconstructed data (some ordinal file-format)</li>
 * </ol>
 * Another optional plugs: reconstruction error measure,
 * graphics viewer (if apropriate), binary code size (output {@link BitStream}).
 *
 * @version 0.26 $Rev: 551 $ $Date: 2007-10-10 19:58:19 +0200 (st, 10 X 2007) $ $Author: pepca $
 * @since   0.26
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/CompressionWorker.java">CompressionWorker.java</a>
 */
public class CompressionWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Log compression summary (log.txt)? */
  protected boolean logSummary = false;

  /** Compression module (should be stopped in {@link #stop}). */
  protected DataFileFormat comp1 = null;

  /** Decompression module (should be stopped in {@link #stop}). */
  protected DataFileFormat recon = null;

  /** Final save (should be stopped in {@link #stop}). */
  protected DataFileFormat comp2 = null;

  //==========================================================================
  //  Interface Worker:

  /**
   * Tries to stop the thread/computation..
   * The process should yield itself.
   * For {@link Runnable} implementators: {@link Thread#interrupt} will be called afterwards.
   */
  @Override
  public synchronized void stop ()
  {
    userBreak = true;
    if ( comp1 != null ) comp1.stop();
    if ( recon != null ) recon.stop();
    if ( comp2 != null ) comp2.stop();
  }

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final Trigger        preproc = (Trigger)       getInterface( "preproc",  IFACE+"Trigger" );
    final GraphicsViewer preview = (GraphicsViewer)getInterface( PL_RASTER,  IFACE+"GraphicsViewer" );
    final RealFunction   compare = (RealFunction)  getInterface( PL_COMPARE, IFACE+"RealFunction" );
    final Property       stat    = (Property)      getInterface( "stat",     IFACE+"Property" );
    final BitStream      ostream = (BitStream)     getInterface( PL_STREAM,  IFACE+"BitStream" );

    synchronized ( this )
    {
      comp1 = (DataFileFormat)getInterface( "comp1", IFACE+"DataFileFormat" );
      recon = (DataFileFormat)getInterface( "recon", IFACE+"DataFileFormat" );
      comp2 = (DataFileFormat)getInterface( "comp2", IFACE+"DataFileFormat" );
    }
    if ( comp1 == null ) return;

    if ( pl != null )
    {
      comp1.setProgressListener( pl );
      if ( recon != null ) recon.setProgressListener( pl );
      if ( comp2 != null ) comp2.setProgressListener( pl );
    }

      // codec/format statistics?
    if ( stat != null )
      stat.set( STATISTICS, true );

      // 0. preprocessing?
    if ( preproc != null )
      preproc.fire( 0 );

      // 1. compression:
    long millis = System.currentTimeMillis();
    try
    {
      comp1.saveFile( (String)null, null ); // file-name was specified in constructor
    }
    catch ( IOException e )
    {
      throw new JaGrLibException( "Error writting encoded file!", e );
    }
    finally
    {
      millis = System.currentTimeMillis() - millis;
      synchronized ( this )
      {
        comp1 = null;
      }
    }
    double elapsed1 = 0.001 * millis;     // compression time
    double elapsed2 = 0.0;                // reconstruction time
    double error    = 0.0;                // reconstruction error
    String msg      = null;
    long   outBits  = 0L;                 // compressed code size in bits
    if ( ostream != null )
      try
      {
        outBits = ostream.position();
      }
      catch ( IOException ex )
      {};

    if ( recon == null )
      msg = String.format( Locale.US, "enc %.1fs, %d", elapsed1, (int)(outBits >> 3) );

    else
    if ( !userBreak )
    {
        // 2. reconstruction:
      millis = System.currentTimeMillis();
      try
      {
        recon.loadFile( (String)null, null );
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error reading encoded file!", e );
      }
      finally
      {
        millis = System.currentTimeMillis() - millis;
        elapsed2 = 0.001 * millis;
        synchronized ( this )
        {
          recon = null;
        }
      }

        // 3. compare reconstructed file with original:
      if ( compare != null )
      {
        error = compare.f( 0 );
        msg   = String.format( Locale.US, "enc %.1fs, %d, dec %.1fs, err %.2f",
                               elapsed1, (int)(outBits >> 3), elapsed2, error );
      }
      else
        msg   = String.format( Locale.US, "enc %.1fs, %d, dec %.1fs",
                               elapsed1, (int)(outBits >> 3), elapsed2 );

        // attach preview window to the reconstructed image:
      if ( preview != null )
      {
        preview.set( GraphicsViewer.WINDOW_TITLE, "Reconstruction (" + msg + ')' );
        GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
        runIt.safeStart();
      }

        // 4. save reconstructed file (perhaps using some standard file-format):
      if ( comp2 != null )
        try
        {
          comp2.saveFile( (String)null, null ); // file-name was specified in constructor
        }
        catch ( IOException e )
        {
          throw new JaGrLibException( "Error saving reconstructed file!", e );
        }
        finally
        {
          synchronized ( this )
          {
            comp2 = null;
          }
        };
    }

    if ( logSummary )
      LogFile.log( "Compression summary: " + msg );
  }

  //==========================================================================
  //  Interface Property:

  public final static String LOG = "Write log";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

       // LOG:
    if ( key.compareTo(LOG) == 0 )
    {
      logSummary = booleanProperty( value, logSummary );
      return;
    }
 }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(LOG) == 0 )
      return logSummary;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "CompressionWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForDataCompression";

  /** Object description. */
  private final static String DESCRIPTION = "Data compression modules testing.";

  /** Object category. */
  protected final static String CATEGORY = C_DATA + '.' + C_COMPRESSION + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOptOutputPlug( "preproc",  IFACE+"Trigger" );
      t.newOutputPlug(    "comp1",    IFACE+"DataFileFormat" );
      t.newOptOutputPlug( "recon",    IFACE+"DataFileFormat" );
      t.newOptOutputPlug( "comp2",    IFACE+"DataFileFormat" );
      t.newOptOutputPlug( PL_RASTER,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_COMPARE, IFACE+"RealFunction" );
      t.newOptOutputPlug( "stat",     IFACE+"Property" );
      t.newOptOutputPlug( PL_STREAM,  IFACE+"BitStream" );
        // Property:
      t.propBegin( LOG, TYPE_BOOLEAN, "Log compression summary?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
