package cz.cuni.jagrlib.worker;

import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker for feature-based warping test composition.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/FeatureWarpingWorker.java">FeatureWarpingWorker.java</a>
 */
public class FeatureWarpingWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Number of features. */
  protected int    features     = 4;

  /** Amplitude of a random perturbation. */
  protected double rndAmplitude = 0.2;

  //==========================================================================
  //  Interface Worker:

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final DataFileFormat output   = (DataFileFormat)getInterface( PL_OUTPUT, IFACE+"DataFileFormat" );
    final GraphicsViewer preview2 = (GraphicsViewer)getInterface( "raster2", IFACE+"GraphicsViewer" );
    final Trigger        warping  = (Trigger)       getInterface( PL_DIRECT, IFACE+"Trigger" );
    final FeatureStore   data     = (FeatureStore)  getInterface( PL_DATA,   IFACE+"FeatureStore" );

    if ( output  == null ||
         warping == null ) return;

      // feature data:
    RandomJames rnd = new RandomJames();
    int i, j;

      // initialize the warping data = set of 1D features (arrows)
    if ( data != null )
    {
        // create feature set:
      double[] coords = new double[ 4 ];

      for ( i = 0; i++ < features; )
      {
        int id = data.insertFeature( 0, 1, FeatureStore.FEATURE_ARROW );
        if ( id < 0 ) break;

          // feature coordinates:
        coords[0] = rnd.randomDouble( 0.1, 0.9 );
        coords[1] = rnd.randomDouble( 0.1, 0.9 );
        do
        {
          coords[2] = rnd.randomDouble( 0.1, 0.9 );
          coords[3] = rnd.randomDouble( 0.1, 0.9 );
        } while ( Math.hypot( coords[0] - coords[2], coords[1] - coords[3] ) > 0.3 );

        data.setCoords( id, 0, coords );
        for ( j = 0; j < 4; j++ )         // jitter target coordinates
          coords[j] += (rnd.uniformNumber() - 0.5) * rndAmplitude;

        data.setCoords( id, 1, coords );

          // feature attributes:
        data.setParam( id, FeatureStore.NULL, "a", rnd.uniformNumber() );
        data.setParam( id, FeatureStore.NULL, "b", rnd.uniformNumber() * 2.0 );
      }
    }

      // transform source image into target bitmap..
    if ( pl != null )
      warping.setProgressListener( pl );
    warping.fire( 0 );

      // attach preview window to the output image & render it:
    if ( preview2 != null )
    {
      preview2.set( GraphicsViewer.WINDOW_TITLE, "Target Image" );
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview2 );
      runIt.safeStart();
    }

      // save the output image:
    try
    {
      output.saveFile( (String)null, null ); // file-name was specified in constructor
    }
    catch ( IOException e )
    {
      throw new JaGrLibException( "Error writting image!", e );
    }
  }

  //==========================================================================
  //  Interface Property:

  public final static String FEATURES      = "Features";
  public final static String RND_AMPLITUDE = "RND amplitude";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // FEATURES:
    if ( key.compareTo(FEATURES) == 0 )
    {
      features = intProperty( value, features, 1, 1000 );
      return;
    }
      // RND_AMPLITUDE:
    if ( key.compareTo(RND_AMPLITUDE) == 0 )
    {
      rndAmplitude = doubleProperty( value, rndAmplitude, 0.0, 1.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(FEATURES) == 0 )
      return features;
    if ( key.compareTo(RND_AMPLITUDE) == 0 )
      return rndAmplitude;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "FeatureWarpingWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForFeatureWarping";

  /** Object description. */
  private final static String DESCRIPTION = "Tests feature-based warping implementations.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_RASTER + '.' + C_FILTER + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_OUTPUT, IFACE+"DataFileFormat" );
      t.newOptOutputPlug( "raster2", IFACE+"GraphicsViewer" );
      t.newOutputPlug(    PL_DIRECT, IFACE+"Trigger" );
      t.newOptOutputPlug( PL_DATA,   IFACE+"FeatureStore" );
        // Property:
      t.propBegin( FEATURES, TYPE_INTEGER, "Feature number", true );
        t.propDefault( 4 );
        t.propBounds(  1, 1000 );
      t.propEnd();
      t.propBegin( RND_AMPLITUDE, TYPE_DOUBLE, "RND amplitude", true );
        t.propDefault( 0.2 );
        t.propBounds(  0.0, 1.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
