package cz.cuni.jagrlib.worker;

import java.io.*;
import java.awt.Point;
import java.util.Locale;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: flood-fill test composition.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/LineWorker.java">LineWorker.java</a>
 */
public class FloodFillWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Starting x-coordinate. */
  protected int xStart = 256;

  /** Starting y-coordinate. */
  protected int yStart = 256;

  //==========================================================================
  //  Interface Worker:

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final FloodFillRender fill    = (FloodFillRender)getInterface( PL_OUTPUT,    IFACE+"FloodFillRender" );
    final BitMask         cl      = (BitMask)        getInterface( "classifier", IFACE+"BitMask" );
    final BitMask         result  = (BitMask)        getInterface( "result",     IFACE+"BitMask" );
    final SolidPen        pen     = (SolidPen)       getInterface( PL_BITMASK,   IFACE+"SolidPen" );
    final DataFileFormat  border  = (DataFileFormat) getInterface( PL_TRIGGER,   IFACE+"DataFileFormat" );
    final GraphicsViewer  preview = (GraphicsViewer) getInterface( PL_RASTER,    IFACE+"GraphicsViewer" );

    if ( fill   == null ||
         cl     == null ||
         pen    == null ) return;

      // load initial border file:
    if ( border != null )
      try
      {
        border.loadFile( (String)null, null );
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error reading image!", e );
      }

      // attach preview window to the raster image:
    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE, "Input Image" );
      preview.repaintAll();
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

      // fill color:
    int[] color = { 255, 0, 0, 255 };

      // classifier initialization:
    cl.init();
    cl.setPixel( xStart, yStart );
    pen.setColor( color );

      // time measurement:
    long millis = System.currentTimeMillis();

      // flood-fill job:
    fill.floodFill( xStart, yStart );

      // and render the result:
    pen.stroke();

      // stop the timer..
    millis = System.currentTimeMillis() - millis;

      // maximal queue:
    int maxQueue = intProperty( fill, "MaxQueue", 0 );

      // draw maximal queue:
    if ( result != null )
    {
      Point[] queue = (Point[])(fill.get( "Queue" ));
      if ( queue != null )
      {
        for ( int i = 0; i < queue.length; i++ )
          result.setPixel( queue[i].x, queue[i].y );
        color[0] = 12;
        pen.setColor( color );
        pen.stroke();
      }
    }

      // refresh the output window:
    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE, "FINISHED - Queue size: " + maxQueue + ", time: " +
                   String.format( Locale.US, "%5.2f", 1.e-3 * millis ) );
      preview.repaintAll();
      preview.stopRepaintLoop();
    }
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // xStart:
    if ( key.compareTo("xStart") == 0 )
    {
      xStart = intProperty( value, xStart );
      return;
    }
      // yStart:
    if ( key.compareTo("yStart") == 0 )
    {
      yStart = intProperty( value, yStart );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo("xStart") == 0 )
      return xStart;
    if ( key.compareTo("yStart") == 0 )
      return yStart;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "FloodFillWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForFloodFill";

  /** Object description. */
  private final static String DESCRIPTION = "Tests flood-fill border, counts maximal FIFO size.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DRAW + '.' + C_FILL + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_OUTPUT,    IFACE+"FloodFillRender" );
      t.newOutputPlug(    PL_BITMASK,   IFACE+"SolidPen" );
      t.newOutputPlug(    "classifier", IFACE+"BitMask" );
      t.newOptOutputPlug( "result",     IFACE+"BitMask" );
      t.newOptOutputPlug( PL_RASTER,    IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER,   IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( "xStart", TYPE_INTEGER, "X-coordinate of starting point", true );
        t.propDefault( 256 );
      t.propEnd();
      t.propBegin( "yStart", TYPE_INTEGER, "Y-coordinate of starting point", true );
        t.propDefault( 256 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
