package cz.cuni.jagrlib.worker;

import java.io.*;
import java.util.Locale;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: computes illumination of scene using some global method and renders the result.
 *
 * @author  Tomas Svab
 * @version 0.26 $Rev: 760 $ $Date: 2008-05-19 04:05:39 +0200 (po, 19 V 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/testing/GlobalIlluminationWorker.java">GlobalIlluminationWorker.java</a>
 */
public class GlobalIlluminationWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** View vector: azimuth angle. */
  protected double azimuth    = 45.0;

  /** View vector: elevation angle. */
  protected double elevation  = 320.0;

  /** Camera distance (for perspective camera only). */
  protected double distance   = 520.0;

  /** Viewing volume (ortho: horizontal size, perspective: horizontal view angle in degrees). */
  protected double viewVolume = 80.0;

  /** LookAt point: X coordinate. */
  protected double lookX      = 278.0;

  /** LookAt point: Y coordinate. */
  protected double lookY      = 273.0;

  /** LookAt point: Z coordinate. */
  protected double lookZ      = 279.5;

  /** Maximal computation time in seconds (0.0 .. indefinite). */
  protected double maxTime    = 0.0;

  /** Maximal number of iterations (0 .. not limited). */
  protected int maxIter       = 0;

  /** Render object (should be stopped in case of user break). */
  protected Render3D render          = null;

  /** 3D scene initialization object (should be stopped in case of user break). */
  protected Trigger  init            = null;

  /** Preview-window redraw timer (for progress indication). */
  protected RedrawTimer timer        = null;

  /** Compute illumination of scene. */
  protected GlobalIllumination illum = null;

  /** Raster image = render target. */
  protected RasterGraphics image     = null;

  /** Preview window. */
  protected GraphicsViewer preview   = null;

  /** Raster file-format output. */
  protected DataFileFormat bak       = null;

  //==========================================================================
  //  Support - working stages:

  /** Initialize the job. */
  protected boolean initialize ()
  {
      // mandatory plugs:
    image  = (RasterGraphics)    getInterface( PL_RASTER, IFACE+"RasterGraphics" );
    render = (Render3D)          getInterface( PL_OUTPUT, IFACE+"Render3D" );
    illum  = (GlobalIllumination)getInterface( PL_DATA,   IFACE+"GlobalIllumination" );

    if ( image  == null ||
         render == null ||
         illum  == null )
      return false;

      // non-mandatory plugs:
    preview = (GraphicsViewer)getInterface( PL_WINDOW,  IFACE+"GraphicsViewer" );
    init    = (Trigger)       getInterface( PL_DIRECT,  IFACE+"Trigger" );
    bak     = (DataFileFormat)getInterface( PL_TRIGGER, IFACE+"DataFileFormat" );

      // 3D scene initialization:
    if ( init != null )
      init.fire( 0 );

    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE, "3D Scene" );
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

      // timer thread:
    timer = new RedrawTimer();
    timer.startTimer( preview, 0L );

    return true;
  }

  /** Set camera and projection. */
  protected void projection ()
  {
    final int width     = image.getWidth();
    final int height    = image.getHeight();
    final double aspect = width / (double)height;
    TrMatrix modelView  = new TrMatrix();
    TrMatrix proj       = new TrMatrix();
    final double az     = Math.toRadians( azimuth );
    final double el     = Math.toRadians( elevation );

    double[] look = { lookX, lookY, lookZ };
    double[] eye  = { look[0] + distance * Math.sin( az ) * Math.cos( el ),
                      look[1] + distance * Math.cos( az ) * Math.cos( el ),
                      look[2] + distance * Math.sin( el ) };
    double[] up   = { 0.0, 1.0, 0.0 };
    modelView.lookAt( eye, look, up );

    proj.perspective( viewVolume, aspect, 1.0, 50.0 );

      // set render module:
    render.setModelView( modelView );
    render.setProjection( proj );
    render.setViewport( 0, 0, width, height );
    render.setDepthComparison( Render3D.DEPTH_LEQUAL );

      // set illumination module:
    illum.setCamera( eye );
  }

  /** Finish the worker job. */
  protected void finish ( long millis )
  {
    synchronized ( this )
    {
      init   = null;
      render = null;
      illum  = null;
    }

      // finish repainting:
    if ( preview != null )
    {
      String tit = String.format( Locale.US, userBreak ? "User break after %.1f sec!" : "FINISHED (%.1f sec)", (0.001 * millis) );
      if ( timer != null )
        timer.setTitle( tit );
      preview.set( GraphicsViewer.WINDOW_TITLE, tit );
      preview.repaintAll();
      preview.stopRepaintLoop();
    }

    if ( timer != null )
      timer.stopTimer();

      // save the result image in raster format:
    if ( bak != null )
      try
      {
        bak.saveFile( (String)null, null );
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error writting image!", e );
      }
  }

  //==========================================================================
  //  Interface Worker:

  /**
   * Tries to stop the thread/computation..
   * The process should yield itself.
   * For {@link Runnable} implementators: {@link Thread#interrupt} will be called afterwards.
   */
  @Override
  public synchronized void stop ()
  {
    userBreak = true;

    if ( init != null )
      init.stop();
    if ( illum != null )
      illum.stop();
    if ( render != null )
      render.stop();
  }

  /**
   * Executive routine of the working thread.
   */
  @Override
  public void run ()
  {
    if ( !initialize() ) return;

      // time measurement:
    maxIter = (maxIter == 0)   ? Integer.MAX_VALUE : maxIter;
    maxTime = (maxTime == 0.0) ? Double.MAX_VALUE  : maxTime*1000.0;
    long startTime = System.currentTimeMillis();

      // prepare projection:
    projection();

      // illuminate:
    if ( pl != null ) illum.setProgressListener( pl );
    illum.illuminate( maxIter, Math.round( maxTime ) );

      // render results:
    if ( pl != null ) render.setProgressListener( pl );
    render.render();

    finish( System.currentTimeMillis() - startTime );
  }

  //==========================================================================
  //  Interface Property:

  public final static String AZIMUTH_ANGLE   = "Azimuth angle";
  public final static String ELEVATION_ANGLE = "Elevation angle";
  public final static String DISTANCE        = "Camera distance";
  public final static String VIEW_VOLUME     = "View volume";
  public final static String LOOK_X          = "Look X";
  public final static String LOOK_Y          = "Look Y";
  public final static String LOOK_Z          = "Look Z";
  public final static String MAX_TIME        = "Max time";
  public final static String MAX_ITER        = "Max iterations";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // AZIMUTH_ANGLE:
    if ( key.compareTo(AZIMUTH_ANGLE) == 0 )
    {
      azimuth = doubleProperty( value, azimuth );
      return;
    }
      // ELEVATION_ANGLE:
    if ( key.compareTo(ELEVATION_ANGLE) == 0 )
    {
      elevation = doubleProperty( value, elevation );
      return;
    }
      // DISTANCE:
    if ( key.compareTo(DISTANCE) == 0 )
    {
      distance = doubleProperty( value, distance, 0.0, 1.e+6 );
      return;
    }
      // VIEW_VOLUME:
    if ( key.compareTo(VIEW_VOLUME) == 0 )
    {
      viewVolume = doubleProperty( value, viewVolume, 0.01, 160.0 );
      return;
    }
      // LOOK_X:
    if ( key.compareTo(LOOK_X) == 0 )
    {
      lookX = doubleProperty( value, lookX );
      return;
    }
      // LOOK_Y:
    if ( key.compareTo(LOOK_Y) == 0 )
    {
      lookY = doubleProperty( value, lookY );
      return;
    }
      // LOOK_Z:
    if ( key.compareTo(LOOK_Z) == 0 )
    {
      lookZ = doubleProperty( value, lookZ );
      return;
    }
      // MAX_TIME:
    if ( key.compareTo(MAX_TIME) == 0 )
    {
      maxTime = doubleProperty( value, maxTime );
      return;
    }
      // MAX_ITER:
    if ( key.compareTo(MAX_ITER) == 0 )
    {
      maxIter = intProperty( value, maxIter );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(AZIMUTH_ANGLE) == 0 )
      return azimuth;
    if ( key.compareTo(ELEVATION_ANGLE) == 0 )
      return elevation;
    if ( key.compareTo(DISTANCE) == 0 )
      return distance;
    if ( key.compareTo(VIEW_VOLUME) == 0 )
      return viewVolume;
    if ( key.compareTo(LOOK_X) == 0 )
      return lookX;
    if ( key.compareTo(LOOK_Y) == 0 )
      return lookY;
    if ( key.compareTo(LOOK_Z) == 0 )
      return lookZ;
    if ( key.compareTo(REJECT) == 0 &&
         isConnected( PL_INPUT ) )
      return true;
    if ( key.compareTo(MAX_TIME) == 0 )
      return maxTime;
    if ( key.compareTo(MAX_ITER) == 0 )
      return maxIter;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Global Illumination Worker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForRender3D";

  /** Object description. */
  private final static String DESCRIPTION = "Computes illumination and renders 3D scene.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,   IFACE+"Worker" );
      t.newOutputPlug(    PL_RASTER,  IFACE+"RasterGraphics" );
      t.newOutputPlug(    PL_DATA,    IFACE+"GlobalIllumination" );
      t.newOptOutputPlug( PL_DIRECT,  IFACE+"Trigger" );
      t.newOptOutputPlug( PL_OUTPUT,  IFACE+"Render3D" );
      t.newOptOutputPlug( PL_WINDOW,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER, IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( AZIMUTH_ANGLE, TYPE_DOUBLE, "Azimuth angle in degrees", true );
        t.propDefault( 45.0 );
      t.propEnd();
      t.propBegin( ELEVATION_ANGLE, TYPE_DOUBLE, "Elevation angle in degrees", true );
        t.propDefault( 320.0 );
      t.propEnd();
      t.propBegin( DISTANCE, TYPE_DOUBLE, "Camera distance from the origin", true );
        t.propDefault( 520.0 );
        t.propBounds(  0.0, 1.e+6 );
      t.propEnd();
      t.propBegin( VIEW_VOLUME, TYPE_DOUBLE, "Viewing volume of the camera in world-space units / degrees", true );
        t.propDefault( 80.0 );
        t.propBounds(  0.01, 160.0 );
      t.propEnd();
      t.propBegin( LOOK_X, TYPE_DOUBLE, "Look X", true );
        t.propDefault( 278.0 );
      t.propEnd();
      t.propBegin( LOOK_Y, TYPE_DOUBLE, "Look Y", true );
        t.propDefault( 273.0 );
      t.propEnd();
      t.propBegin( LOOK_Z, TYPE_DOUBLE, "Look Z", true );
        t.propDefault( 279.5 );
      t.propEnd();
      t.propBegin( MAX_TIME, TYPE_DOUBLE, "Max time (in seconds) of computation", true );
        t.propDefault( 600 );
      t.propEnd();
      t.propBegin( MAX_ITER, TYPE_INTEGER, "Max iterations", true );
        t.propDefault( 500 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }
}
