package cz.cuni.jagrlib.worker;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker for raster-based imaging: {@link ImageSynthesizer} generates
 * raster image ({@link RasterGraphics}).
 *
 * <p>Calls {@link ImageSynthesizer#renderRectangle} multiple times:
 * in rectangular grid of the given size.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/GridRenderWorker.java">GridRenderWorker.java</a>
 */
public class GridRenderWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Number of grid rows. */
  protected int rows    = 4;

  /** Number of grid columns. */
  protected int columns = 4;

  /** Image synthesizer (should be stopped in {@link #stop}). */
  protected ImageSynthesizer synt = null;

  /** For progress indication.. */
  protected int stage;

  //==========================================================================
  //  Interface Worker:

  /**
   * Tries to stop the thread/computation..
   * The process should yield itself.
   * For {@link Runnable} implementators: {@link Thread#interrupt} will be called afterwards.
   */
  @Override
  public synchronized void stop ()
  {
    userBreak = true;
    if ( synt != null )
      synt.stop();
  }

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    stage = 0;
    final RasterGraphics image   = (RasterGraphics)  getInterface( PL_IMAGE,    IFACE+"RasterGraphics" );
    final GraphicsViewer preview = (GraphicsViewer)  getInterface( PL_RASTER,   IFACE+"GraphicsViewer" );
    final DataFileFormat bak     = (DataFileFormat)  getInterface( PL_TRIGGER,  IFACE+"DataFileFormat" );

    synchronized ( this )
    {
      synt = (ImageSynthesizer)getInterface( PL_OUTPUT, IFACE+"ImageSynthesizer" );
    }
    if ( synt == null ) return;

    if ( preview != null )
    {
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

      // timer thread:
    RedrawTimer timer = null;
    if ( preview != null )
    {
      timer = new RedrawTimer();
      timer.startTimer( preview, 0L );
    }

      // determine image size:
    int width, height;
    if ( image != null )
    {                                     // image has the highest priority
      width  = image.getWidth();
      height = image.getHeight();
    }
    else
    {                                     // at least I'll try the synthesizer..
      width  = intProperty( synt, ImageSynthesizer.WIDTH,  256 );
      height = intProperty( synt, ImageSynthesizer.HEIGHT, 256 );
    }
    if ( width  <= 0 ) width  = 256;
    if ( height <= 0 ) height = 256;
    synt.setBounds( width, height );

    int x, y, i, j;
    int dx = width  / columns;
    int dy = height / rows;
    int x1, y1;

    all:
    for ( j = 0, y = 0; j++ < rows; y += dy )
    {                                     // render one row
      y1 = (j < rows) ? y + dy : height;

      for ( i = 0, x = 0; i++ < columns; x += dx )
      {                                   // render one cell
        if ( userBreak ) break all;

        x1 = (i < columns) ? x + dx : width;

        synt.renderRectangle( x, y, x1, y1 );
        stage++;

        if ( preview != null )
        {
          preview.repaintAll();
          preview.set( GraphicsViewer.WINDOW_TITLE, "After the " + stage + "-th stage" );
        }
        if ( pl != null )
          pl.progress( stage, rows * columns );
      }
    }

    synchronized ( this )
    {
      synt = null;
    }

    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE, userBreak ? "User break!" : "FINISHED" );
      preview.repaintAll();
      preview.stopRepaintLoop();
    }

    if ( timer != null )
      timer.stopTimer();

      // save the result image:
    if ( bak != null )
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( Exception e )
      {
        throw new JaGrLibException( "Error writting image!", e );
      }
  }

  //==========================================================================
  //  Interface Property:

  public final static String ROWS    = "Rows";
  public final static String COLUMNS = "Columns";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // ROWS:
    if ( key.compareTo(ROWS) == 0 )
    {
      rows = intProperty( value, rows, 1, 1000 );
      return;
    }
      // COLUMNS:
    if ( key.compareTo(COLUMNS) == 0 )
    {
      columns = intProperty( value, columns, 1, 1000 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(ROWS) == 0 )
      return rows;
    if ( key.compareTo(COLUMNS) == 0 )
      return columns;
    if ( key.compareTo(REJECT) == 0 &&
         isConnected( PL_INPUT ) )
      return true;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "GridRenderWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForSampleRendering";

  /** Object description. */
  private final static String DESCRIPTION = "Worker for raster imaging: render grid";

  /** Object category. */
  protected final static String CATEGORY = C_3D + "." + C_RENDER + "." + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,       IFACE+"Worker" );
      t.newOutputPlug(    PL_OUTPUT,      IFACE+"ImageSynthesizer" );
      t.newOptOutputPlug( PL_IMAGE,       IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_RASTER,      IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER,     IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( ROWS, TYPE_INTEGER, "Number of grid rows", true );
        t.propDefault( 4 );
        t.propBounds(  1, 1000 );
      t.propEnd();
      t.propBegin( COLUMNS, TYPE_INTEGER, "Number of grid columns", true );
        t.propDefault( 4 );
        t.propBounds(  1, 1000 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
