package cz.cuni.jagrlib.worker;

import java.io.*;
import java.util.Locale;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: raster image composition.
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/ImageCompositionWorker.java">ImageCompositionWorker.java</a>
 */
public class ImageCompositionWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Maximum number of input raster images. */
  public final static int MAX_INPUT = 4;

  /** Image filter (should be stopped in {@link #stop}). */
  public Trigger filter = null;

  //==========================================================================
  //  Interface Worker:

  /**
   * Tries to stop the thread/computation..
   * The process should yield itself.
   * For {@link Runnable} implementators: {@link Thread#interrupt} will be called afterwards.
   */
  @Override
  public synchronized void stop ()
  {
    userBreak = true;
    if ( filter != null )
      filter.stop();
  }

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    int i;
    final DataFileFormat[] input = new DataFileFormat[ MAX_INPUT ];
    for ( i = 0; i < MAX_INPUT; i++ )
      input[i] = (DataFileFormat)getInterface( "input" + (i+1), IFACE+"DataFileFormat" );

    final DataFileFormat output   = (DataFileFormat)getInterface( PL_OUTPUT, IFACE+"DataFileFormat" );
    final GraphicsViewer previewo = (GraphicsViewer)getInterface( "rastero", IFACE+"GraphicsViewer" );

    if ( output == null ) return;

    synchronized ( this )
    {
      filter = (Trigger)getInterface( PL_DIRECT, IFACE+"Trigger" );
    }
    if ( filter == null ) return;
    if ( pl != null )
      filter.setProgressListener( pl );

      // load initial image files:
    for ( i = 0; i < MAX_INPUT; i++ )
      if ( input[i] != null )
        try
        {
          input[i].loadFile( (String)null, null );
        }
        catch ( IOException e )
        {
          synchronized ( this )
          {
            filter = null;
          }
          throw new JaGrLibException( "Error reading image #" + (i+1) + "!", e );
        }

      // do the job..
    if ( pl != null ) filter.setProgressListener( pl );
    long millis = System.currentTimeMillis();
    filter.fire( 0 );
    millis = System.currentTimeMillis() - millis;

    synchronized ( this )
    {
      filter = null;
    }

      // attach preview window to the output image & render it:
    if ( previewo != null )
    {
      final double elapsed = 0.001 * millis;
      previewo.set( GraphicsViewer.WINDOW_TITLE,
                    String.format( Locale.US, userBreak ? "User break after %.1f sec!" : "FINISHED (%.1f sec)", elapsed ) );
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( previewo );
      runIt.safeStart();
    }

      // save the output image:
    try
    {
      output.saveFile( (String)null, null ); // file-name was specified in constructor
    }
    catch ( IOException e )
    {
      throw new JaGrLibException( "Error writting image!", e );
    }
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(REJECT) == 0 &&
         isConnected( PL_INPUT ) )
      return true;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "ImageCompositionWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForImageComposition";

  /** Object description. */
  private final static String DESCRIPTION = "Raster image compositioning.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_RASTER + '.' + C_FILTER + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug( PL_INPUT, IFACE+"Worker" );
      int i;
      for ( i = 0; i < MAX_INPUT; i++ )
        t.newOptOutputPlug( "input" + (i+1), IFACE+"DataFileFormat" );
      t.newOutputPlug( PL_OUTPUT, IFACE+"DataFileFormat" );

      t.newOptOutputPlug( "rastero", IFACE+"GraphicsViewer" );

      t.newOutputPlug( PL_DIRECT, IFACE+"Trigger" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
