package cz.cuni.jagrlib.worker;

import java.io.*;
import java.util.Locale;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: image compression test composition.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/ImageCompressionWorker.java">ImageCompressionWorker.java</a>
 */
public class ImageCompressionWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Codec (should be stopped in {@link #stop}). */
  protected DataFileFormat output = null;

  //==========================================================================
  //  Interface Worker:

  /**
   * Tries to stop the thread/computation..
   * The process should yield itself.
   * For {@link Runnable} implementators: {@link Thread#interrupt} will be called afterwards.
   */
  @Override
  public synchronized void stop ()
  {
    userBreak = true;
    if ( output != null )
      output.stop();
  }

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final DataFileFormat recon    = (DataFileFormat)getInterface( "recon",    IFACE+"DataFileFormat" );
    final GraphicsViewer preview2 = (GraphicsViewer)getInterface( "raster2",  IFACE+"GraphicsViewer" );
    final RealFunction   compare  = (RealFunction)  getInterface( PL_COMPARE, IFACE+"RealFunction" );
    final Property       stat     = (Property)      getInterface( "stat",     IFACE+"Property" );

    if ( recon == null ) return;

    synchronized ( this )
    {
      output = (DataFileFormat)getInterface( PL_OUTPUT, IFACE+"DataFileFormat" );
    }
    if ( output == null ) return;
    if ( pl != null )
      output.setProgressListener( pl );

      // codec/format statistics?
    if ( stat != null )
      stat.set( STATISTICS, true );

      // save the output image:
    long millis = System.currentTimeMillis();
    try
    {
      output.saveFile( (String)null, null ); // file-name was specified in constructor
    }
    catch ( IOException e )
    {
      throw new JaGrLibException( "Error writting image!", e );
    }
    finally
    {
      millis = System.currentTimeMillis() - millis;
      synchronized ( this )
      {
        output = null;
      }
    }
    if ( userBreak ) return;

      // load reconstructed image:
    try
    {
      recon.loadFile( (String)null, null );
    }
    catch ( IOException e )
    {
      throw new JaGrLibException( "Error reading reconstructed image!", e );
    }

      // compare reconstructed image with original:
    final double elapsed = 0.001 * millis;
    double error = 0.0;
    if ( compare != null )
    {
      error = compare.f( 0 );
      if ( LogFile.debugging )
        LogFile.log( String.format( Locale.US, "Encoding in %.2f sec, reconstruction err = %.2f", elapsed, error ) );
    }

      // attach preview window to the reconstructed image:
    if ( preview2 != null )
    {
      preview2.set( GraphicsViewer.WINDOW_TITLE,
                    String.format( Locale.US, (compare == null) ? "Decoded Image (%.1f sec)" :
                                                                  "Decoded Image (%.1f sec), err = %.2f", elapsed, error ) );
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview2 );
      runIt.safeStart();
    }
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "ImageCompressionWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForImageCompression";

  /** Object description. */
  private final static String DESCRIPTION = "Tests image compression modules.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_RASTER + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"DataFileFormat" );
      t.newOutputPlug(    "recon",    IFACE+"DataFileFormat" );
      t.newOptOutputPlug( "raster2",  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_COMPARE, IFACE+"RealFunction" );
      t.newOptOutputPlug( "stat",     IFACE+"Property" );
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
