package cz.cuni.jagrlib.worker;

import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: line clipping & drawing test composition.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/LineClipWorker.java">LineClipWorker.java</a>
 */
public class LineClipWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Xmin coordinate for clipping. */
  protected int x0     = 100;

  /** Xmax coordinate for clipping. */
  protected int x1     = 200;

  /** Ymin coordinate for clipping. */
  protected int y0     = 100;

  /** Ymax coordinate for clipping. */
  protected int y1     = 200;

  /** Number of lines to be drawn. */
  protected int repeat = 1024;

  /** Set negative value for randomize. */
  protected int seed   = -1;

  //==========================================================================
  //  Interface Worker:

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final LineRender      line    = (LineRender)     getInterface( PL_OUTPUT,  IFACE+"LineRender" );
    final LineRender      direct  = (LineRender)     getInterface( PL_DIRECT,  IFACE+"LineRender" );
    final RectangleWindow window  = (RectangleWindow)getInterface( PL_WINDOW,  IFACE+"RectangleWindow" );
    final SolidPen        pen     = (SolidPen)       getInterface( PL_BITMASK, IFACE+"SolidPen" );
    final GraphicsViewer  preview = (GraphicsViewer) getInterface( PL_RASTER,  IFACE+"GraphicsViewer" );
    final DataFileFormat  bak     = (DataFileFormat) getInterface( PL_TRIGGER, IFACE+"DataFileFormat" );

    if ( line == null ||
         pen  == null ) return;

    if ( preview != null )
    {
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

    int[] color      = { 255, 255, 255, 255 };
    int[] colorOver  = {   0,   0, 192, 255 };
    int[] colorFrame = { 255, 255,   0, 255 };

      // draw clip frame first:
    direct.drawLine( x0, y0, x0, y1 );
    direct.drawLine( x0, y1, x1, y1 );
    direct.drawLine( x1, y1, x1, y0 );
    direct.drawLine( x1, y0, x0, y0 );
    pen.setColor( colorFrame );
    pen.stroke();

      // set rectangle window:
    window.rectangleWindow( x0, x1, y0, y1 );

      // random generator:
    RandomJames rnd;
    if ( seed >= 0 )
      rnd = new RandomJames( seed >>> 15, seed & 0x7fff );
    else
    {
      rnd = new RandomJames();
      rnd.randomize();
    }

      // draw random lines:
    for ( int i = 0; i++ < repeat && !userBreak; )
    {
        // random coordinates:
      int xx1 = rnd.randomInteger( 0, 511 );
      int xx2 = rnd.randomInteger( 0, 511 );
      int yy1 = rnd.randomInteger( 0, 511 );
      int yy2 = rnd.randomInteger( 0, 511 );

        // the whole line:
      direct.drawLine( xx1, yy1, xx2, yy2 );
      pen.setColor( color );
      pen.stroke();

        // clipped line:
      line.drawLine( xx1, yy1, xx2, yy2 );
      pen.setColor( colorOver );
      pen.stroke();

        // refresh the output window:
      if ( (i & 127) == 0 )
      {
        if ( pl != null )
          pl.progress( i, repeat );
        if ( preview != null )
        {
          preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - " + i + " lines" );
          preview.repaintAll();
        }
      }
    }

      // draw frame once more:
    direct.drawLine( x0, y0, x0, y1 );
    direct.drawLine( x0, y1, x1, y1 );
    direct.drawLine( x1, y1, x1, y0 );
    direct.drawLine( x1, y0, x0, y0 );
    pen.setColor( colorFrame );
    pen.stroke();

      // finish repainting:
    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - FINISHED" );
      preview.repaintAll();
      preview.stopRepaintLoop();
    }
      // save the result image:
    if ( bak != null )
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error writting image!", e );
      }
  }

  //==========================================================================
  //  Interface Property:

  public final static String X_MIN = "Xmin";
  public final static String X_MAX = "Xmax";
  public final static String Y_MIN = "Ymin";
  public final static String Y_MAX = "Ymax";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // REPEAT:
    if ( key.compareTo(REPEAT) == 0 )
    {
      repeat = intProperty( value, repeat );
      return;
    }
      // X_MIN:
    if ( key.compareTo(X_MIN) == 0 )
    {
      x0 = intProperty( value, x0 );
      return;
    }
      // X_MAX:
    if ( key.compareTo(X_MAX) == 0 )
    {
      x1 = intProperty( value, x1 );
      return;
    }
      // Y_MIN:
    if ( key.compareTo(Y_MIN) == 0 )
    {
      y0 = intProperty( value, y0 );
      return;
    }
      // Y_MAX:
    if ( key.compareTo(Y_MAX) == 0 )
    {
      y1 = intProperty( value, y1 );
      return;
    }
      // SEED:
    if ( key.compareTo(SEED) == 0 )
    {
      seed = intProperty( value, seed );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(REPEAT) == 0 )
      return repeat;
    if ( key.compareTo(X_MIN) == 0 )
      return x0;
    if ( key.compareTo(X_MAX) == 0 )
      return x1;
    if ( key.compareTo(Y_MIN) == 0 )
      return y0;
    if ( key.compareTo(Y_MAX) == 0 )
      return y1;
    if ( key.compareTo(SEED) == 0 )
      return seed;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "LineClipWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForLineClipping";

  /** Object description. */
  private final static String DESCRIPTION = "Draws and clips lines.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DRAW + '.' + C_LINE + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"LineRender" );
      t.newOutputPlug(    PL_DIRECT,  IFACE+"LineRender" );
      t.newOutputPlug(    PL_WINDOW,  IFACE+"RectangleWindow" );
      t.newOutputPlug(    PL_BITMASK, IFACE+"SolidPen" );
      t.newOptOutputPlug( PL_RASTER,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER, IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( REPEAT, TYPE_INTEGER, "Number of random test lines", true );
        t.propDefault( 512 );
      t.propEnd();
      t.propBegin( X_MIN, TYPE_INTEGER, "Xmin", true );
        t.propDefault( 100 );
      t.propEnd();
      t.propBegin( X_MAX, TYPE_INTEGER, "Xmax", true );
        t.propDefault( 200 );
      t.propEnd();
      t.propBegin( Y_MIN, TYPE_INTEGER, "Ymin", true );
        t.propDefault( 100 );
      t.propEnd();
      t.propBegin( Y_MAX, TYPE_INTEGER, "Ymax", true );
        t.propDefault( 200 );
      t.propEnd();
      t.propBegin( SEED, TYPE_INTEGER, "Random generator seed (-1 for randomize)", true );
        t.propDefault( -1 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
