package cz.cuni.jagrlib.worker;

import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: line drawing test composition.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/LineWorker.java">LineWorker.java</a>
 */
public class LineWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Number of lines to be drawn. */
  protected int repeat = 1024;

  /** Set negative value for randomize. */
  protected int seed   = -1;

  //==========================================================================
  //  Interface Worker:

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final LineRender     line    = (LineRender)    getInterface( PL_OUTPUT,  IFACE+"LineRender" );
    final SolidPen       pen     = (SolidPen)      getInterface( PL_BITMASK, IFACE+"SolidPen" );
    final GraphicsViewer preview = (GraphicsViewer)getInterface( PL_RASTER,  IFACE+"GraphicsViewer" );
    final DataFileFormat bak     = (DataFileFormat)getInterface( PL_TRIGGER, IFACE+"DataFileFormat" );

    if ( line == null ||
         pen  == null ) return;

    if ( preview != null )
    {
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

    double[] color = { 0.0, 0.0, 0.0, 1.0 };

      // random generator:
    RandomJames rnd;
    if ( seed >= 0 )
      rnd = new RandomJames( seed >>> 15, seed & 0x7fff );
    else
    {
      rnd = new RandomJames();
      rnd.randomize();
    }

      // draw random lines:
    for ( int i = 0; i++ < repeat && !userBreak; )
    {
        // random line:
      line.drawLine( rnd.uniformNumber() * 511.0, rnd.uniformNumber() * 511.0,
                     rnd.uniformNumber() * 511.0, rnd.uniformNumber() * 511.0 );

        // stroke it:
      if ( (i & 15) == 0 )
      {
          // set random color:
        color[0] = rnd.randomDouble( 0.3, 1.0 );
        color[1] = rnd.randomDouble( 0.3, 1.0 );
        color[2] = rnd.randomDouble( 0.3, 1.0 );
        pen.setColor( color );
        pen.stroke();
      }

        // refresh the output window:
      if ( (i & 255) == 0 )
      {
        if ( pl != null )
          pl.progress( i, repeat );
        if ( preview != null )
        {
          preview.repaintAll();
          preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - " + i + " lines" );
        }
      }
    }

      // set random color:
    color[0] = rnd.randomDouble( 0.3, 1.0 );
    color[1] = rnd.randomDouble( 0.3, 1.0 );
    color[2] = rnd.randomDouble( 0.3, 1.0 );
    pen.setColor( color );
    pen.stroke();

      // finish repainting:
    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - FINISHED" );
      preview.repaintAll();
      preview.stopRepaintLoop();
    }

      // save the result image in PNG format:
    if ( bak != null )
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error writting image!", e );
      }
  }

  //==========================================================================
  //  Interface Property:

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // REPEAT:
    if ( key.compareTo(REPEAT) == 0 )
    {
      repeat = intProperty( value, repeat );
      return;
    }
      // SEED:
    if ( key.compareTo(SEED) == 0 )
    {
      seed = intProperty( value, seed );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(REPEAT) == 0 )
      return repeat;
    if ( key.compareTo(SEED) == 0 )
      return seed;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "LineWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForLine";

  /** Object description. */
  private final static String DESCRIPTION = "Draws random lines.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DRAW + '.' + C_LINE + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"LineRender" );
      t.newOutputPlug(    PL_BITMASK, IFACE+"SolidPen" );
      t.newOptOutputPlug( PL_RASTER,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER, IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( REPEAT, TYPE_INTEGER, "Number of random test lines", true );
        t.propDefault( 512 );
      t.propEnd();
      t.propBegin( SEED, TYPE_INTEGER, "Random generator seed (-1 for randomize)", true );
        t.propDefault( -1 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
