package cz.cuni.jagrlib.worker;

import java.awt.Point;
import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: polygon clipping test composition.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/PolygonClipWorker.java">PolygonClipWorker.java</a>
 */
public class PolygonClipWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Maximum polygon size. */
  protected int maxPoly = 20;

  /** Number of polygon to draw. */
  protected int repeat  = 1024;

  /** Set negative value for randomize. */
  protected int seed    = -1;

  //==========================================================================
  //  Interface Worker:

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final PolygonFillRender polygon = (PolygonFillRender)getInterface( PL_OUTPUT,  IFACE+"PolygonFillRender" );
    final PolygonFillRender direct  = (PolygonFillRender)getInterface( PL_DIRECT,  IFACE+"PolygonFillRender" );
    final PolygonWindow     window  = (PolygonWindow)getInterface(     PL_WINDOW,  IFACE+"PolygonWindow" );
    final SolidPen          pen     = (SolidPen)         getInterface( PL_BITMASK, IFACE+"SolidPen" );
    final GraphicsViewer    preview = (GraphicsViewer)   getInterface( PL_RASTER,  IFACE+"GraphicsViewer" );
    final DataFileFormat    bak     = (DataFileFormat)   getInterface( PL_TRIGGER, IFACE+"DataFileFormat" );

    if ( polygon == null ||
         direct  == null ||
         window  == null ||
         pen     == null ) return;

    if ( preview != null )
    {
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

      // simple polygonal window:
    Point[] poly  = new Point[ 5 ];
    poly[0] = new Point( 100, 100 );
    poly[1] = new Point( 250, 100 );
    poly[2] = new Point( 300, 350 );
    poly[3] = new Point( 100, 300 );
    poly[4] = new Point( 200, 200 );
    window.polygonWindow( poly );

    int[]   color = new int[ 3 ];
    int i, j;

      // random generator:
    RandomJames rnd;
    if ( seed >= 0 )
      rnd = new RandomJames( seed >>> 15, seed & 0x7fff );
    else
    {
      rnd = new RandomJames();
      rnd.randomize();
    }

      // draw random polygons:
    poly  = new Point[ maxPoly ];
    for ( i = 0; i++ < repeat && !userBreak; )
    {
        // generate random polygon:
      int len = rnd.randomInteger( 2, maxPoly );
      for ( j = 0; j < len - 1; j++ )
        poly[j] = new Point( rnd.randomInteger( 0, 511 ),
                             rnd.randomInteger( 0, 511 ) );
      if ( rnd.uniformNumber() < 0.5 )
        poly[j] = poly[0];
      else
        poly[j] = new Point( rnd.randomInteger( 0, 511 ),
                             rnd.randomInteger( 0, 511 ) );
      if ( ++j < maxPoly )
        poly[j] = null;

        // fill the original:
      direct.fillPolygon( poly );

        // set random color:
      color[0] = rnd.randomInteger( 54, 255 );
      color[1] = rnd.randomInteger( 54, 255 );
      color[2] = rnd.randomInteger( 54, 255 );
      pen.setColor( color );
      pen.stroke();

        // fill the clipped polygon:
      polygon.fillPolygon( poly );

        // set random color:
      color[0] = rnd.randomInteger( 54, 255 );
      color[1] = rnd.randomInteger( 54, 255 );
      color[2] = rnd.randomInteger( 54, 255 );
      pen.setColor( color );
      pen.stroke();

        // refresh the output window:
      if ( (i & 31) == 0 )
      {
        if ( pl != null )
          pl.progress( i, repeat );
        if ( preview != null )
        {
          preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - " + i + " polygons" );
          preview.repaintAll();
        }
      }
    }

      // finish repainting:
    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - FINISHED" );
      preview.repaintAll();
      preview.stopRepaintLoop();
    }

      // save the result image in PNG format:
    if ( bak != null )
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error writting image!", e );
      }
  }

  //==========================================================================
  //  Interface Property:

  /** Key: maximal polygon size. */
  public final static String MAX_POLY = "MaxPoly";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // REPEAT:
    if ( key.compareTo(REPEAT) == 0 )
    {
      repeat = intProperty( value, repeat );
      return;
    }
      // MAX_POLY:
    if ( key.compareTo(MAX_POLY) == 0 )
    {
      maxPoly = intProperty( value, maxPoly, 3, 1000 );
      return;
    }
      // SEED:
    if ( key.compareTo(SEED) == 0 )
    {
      seed = intProperty( value, seed );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(REPEAT) == 0 )
      return repeat;
    if ( key.compareTo(MAX_POLY) == 0 )
      return maxPoly;
    if ( key.compareTo(SEED) == 0 )
      return seed;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "PolygonClipWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForPolygonFillRenderAndPolygonWindow";

  /** Object description. */
  private final static String DESCRIPTION = "Fills random polygons and clips them by PolygonWindow.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_CLIP + '.' + C_POLYGON + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"PolygonFillRender" );
      t.newOutputPlug(    PL_DIRECT,  IFACE+"PolygonFillRender" );
      t.newOutputPlug(    PL_WINDOW,  IFACE+"PolygonWindow" );
      t.newOutputPlug(    PL_BITMASK, IFACE+"SolidPen" );
      t.newOptOutputPlug( PL_RASTER,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER, IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( REPEAT, TYPE_INTEGER, "Number of random polygons", true );
        t.propBounds(  0, 10000 );
        t.propDefault( 128 );
      t.propEnd();
      t.propBegin( MAX_POLY, TYPE_INTEGER, "Maximal polygon size", true );
        t.propBounds(  3, 1000 );
        t.propDefault( 20 );
      t.propEnd();
      t.propBegin( SEED, TYPE_INTEGER, "Random generator seed (-1 for randomize)", true );
        t.propDefault( -1 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
