package cz.cuni.jagrlib.worker;

import java.awt.Point;
import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: anti-aliased thick poly-line drawing test composition.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/PolylineAntiWorker.java">PolylineAntiWorker.java</a>
 */
public class PolylineAntiWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Maximum polyline size. */
  protected int maxPoly = 20;

  /** Number of polylines to draw. */
  protected int repeat  = 1024;

  /** Set negative value for randomize. */
  protected int seed    = -1;

  //==========================================================================
  //  Interface Worker:

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final PolygonStrokeRender polyline = (PolygonStrokeRender)getInterface( PL_OUTPUT,  IFACE+"PolygonStrokeRender" );
    final SolidBrush          brush    = (SolidBrush)         getInterface( PL_BITMASK, IFACE+"SolidBrush" );
    final GraphicsViewer      preview  = (GraphicsViewer)     getInterface( PL_RASTER,  IFACE+"GraphicsViewer" );
    final DataFileFormat      bak      = (DataFileFormat)     getInterface( PL_TRIGGER, IFACE+"DataFileFormat" );

    if ( polyline == null ||
         brush    == null ) return;

    if ( preview != null )
    {
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

    int[] color  = new int[ 4 ];
    Point[] poly = new Point[ maxPoly ];
    int i, j;

      // random generator:
    RandomJames rnd;
    if ( seed >= 0 )
      rnd = new RandomJames( seed >>> 15, seed & 0x7fff );
    else
    {
      rnd = new RandomJames();
      rnd.randomize();
    }

      // draw random polylines:
    for ( i = 0; i++ < repeat && !userBreak; )
    {
        // generate random polyline:
      int len = rnd.randomInteger( 2, maxPoly );
      for ( j = 0; j < len - 1; j++ )
        poly[j] = new Point( rnd.randomInteger( 0, 511 ),
                             rnd.randomInteger( 0, 511 ) );
      if ( rnd.uniformNumber() < 0.5 )
        poly[j] = poly[0];
      else
        poly[j] = new Point( rnd.randomInteger( 0, 511 ),
                             rnd.randomInteger( 0, 511 ) );
      if ( ++j < maxPoly )
        poly[j] = null;

        // draw it:
      polyline.strokePolyline( poly );

        // and eventually stroke it:
      if ( (i & 7) == 0 )
      {
          // set random color:
        color[0] = rnd.randomInteger( 54, 255 );
        color[1] = rnd.randomInteger( 54, 255 );
        color[2] = rnd.randomInteger( 54, 255 );
        brush.setColor( color );
        brush.fill();
      }

        // refresh the output window:
      if ( (i & 31) == 0 )
      {
        if ( pl != null )
          pl.progress( i, repeat );
        if ( preview != null )
        {
          preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - " + i + " polylines" );
          preview.repaintAll();
        }
      }
    }

      // set random color:
    color[0] = rnd.randomInteger( 54, 255 );
    color[1] = rnd.randomInteger( 54, 255 );
    color[2] = rnd.randomInteger( 54, 255 );
    brush.setColor( color );
    brush.fill();

      // finish repainting:
    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - FINISHED" );
      preview.repaintAll();
      preview.stopRepaintLoop();
    }

      // save the result image in PNG format:
    if ( bak != null )
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error writting image!", e );
      }
  }

  //==========================================================================
  //  Interface Property:

  /** Key: maximal polygon size. */
  public final static String MAX_POLY = "MaxPoly";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // REPEAT:
    if ( key.compareTo(REPEAT) == 0 )
    {
      repeat = intProperty( value, repeat );
      return;
    }
      // MAX_POLY:
    if ( key.compareTo(MAX_POLY) == 0 )
    {
      maxPoly = intProperty( value, maxPoly );
      return;
    }
      // SEED:
    if ( key.compareTo(SEED) == 0 )
    {
      seed = intProperty( value, seed );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(REPEAT) == 0 )
      return repeat;
    if ( key.compareTo(MAX_POLY) == 0 )
      return maxPoly;
    if ( key.compareTo(SEED) == 0 )
      return seed;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "PolylineAntiWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForPolygonStrokeRender";

  /** Object description. */
  private final static String DESCRIPTION = "Draws thick anti-aliased poly lines.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DRAW + '.' + C_LINE + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"PolygonStrokeRender" );
      t.newOutputPlug(    PL_BITMASK, IFACE+"SolidBrush" );
      t.newOptOutputPlug( PL_RASTER,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER, IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( REPEAT, TYPE_INTEGER, "Number of random test polylines", true );
        t.propBounds( 0, 10000 );
        t.propDefault( 128 );
      t.propEnd();
      t.propBegin( MAX_POLY, TYPE_INTEGER, "Maximal polyline size", true );
        t.propBounds( 2, 1000 );
        t.propDefault( 20 );
      t.propEnd();
      t.propBegin( SEED, TYPE_INTEGER, "Random generator seed (-1 for randomize)", true );
        t.propDefault( -1 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
