package cz.cuni.jagrlib.worker;

import java.io.*;
import java.util.Locale;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: 3D rendering (3D scene, {@link Render3D} renderer, result raster image)).
 *
 * @version 0.26 $Rev: 701 $ $Date: 2008-01-24 00:19:06 +0100 (čt, 24 I 2008) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/Render3DWorker.java">Render3DWorker.java</a>
 */
public class Render3DWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Use perspective projection? */
  protected boolean perspective = false;

  /** View vector: azimuth angle. */
  protected double azimuth   = 30.0;

  /** View vector: elevation angle. */
  protected double elevation = 20.0;

  /** Camera distance (for perspective camera only). */
  protected double distance  = 10.0;

  /** Viewing volume (ortho: horizontal size, perspective: horizontal view angle in degrees). */
  protected double viewVolume = 60.0;

  /** Render object (should be stopped in case of user break). */
  protected Render3D render = null;

  /** 3D scene initialization object (should be stopped in case of user break). */
  protected Trigger  init   = null;

  //==========================================================================
  //  Interface Worker:

  /**
   * Tries to stop the thread/computation..
   * The process should yield itself.
   * For {@link Runnable} implementators: {@link Thread#interrupt} will be called afterwards.
   */
  @Override
  public synchronized void stop ()
  {
    userBreak = true;
    if ( init != null )
      init.stop();
    if ( render != null )
      render.stop();
  }

  /** Executive routine of the working thread. */
  public void run ()
  {
    final RasterGraphics image   = (RasterGraphics)getInterface( PL_RASTER,  IFACE+"RasterGraphics" );
    final GraphicsViewer preview = (GraphicsViewer)getInterface( PL_WINDOW,  IFACE+"GraphicsViewer" );
    final DataFileFormat bak     = (DataFileFormat)getInterface( PL_TRIGGER, IFACE+"DataFileFormat" );

    if ( image == null ) return;

    synchronized ( this )
    {
      render = (Render3D)getInterface( PL_OUTPUT, IFACE+"Render3D" );
      if ( render != null )
        init = (Trigger)getInterface( PL_DIRECT, IFACE+"Trigger" );
    }
    if ( render == null ) return;

      // 3D scene initialization:
    if ( init != null )
      init.fire( 0 );

    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE, "3D Scene" );
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

      // set projection matrix according to output canvas size:
    final int width     = image.getWidth();
    final int height    = image.getHeight();
    final double aspect = width / (double)height;
    TrMatrix modelView  = new TrMatrix();
    TrMatrix proj       = new TrMatrix();
    final double az     = Math.toRadians( azimuth );
    final double el     = Math.toRadians( elevation );

    double[] center     = { 0.0, 0.0, 0.0 };
    double[] eye        = { center[0] + distance * Math.sin( az ) * Math.cos( el ),
                            center[1] + distance * Math.cos( az ) * Math.cos( el ),
                            center[2] + distance * Math.sin( el ) };
    modelView.lookAt( eye, center, null );

    if ( perspective )
      proj.perspective( viewVolume, aspect,
                        1.0, 50.0 );
    else
    {
      final double vHalf  = 0.5 * viewVolume;
      proj.ortho( -vHalf,          vHalf,
                  -vHalf / aspect, vHalf / aspect,
                   1.0, 50.0 );
    }

    render.setModelView( modelView );
    render.setProjection( proj );
    render.setViewport( 0, 0, width, height );
    render.setDepthComparison( Render3D.DEPTH_LEQUAL );

      // render the image:
    if ( pl != null ) render.setProgressListener( pl );
    long millis = System.currentTimeMillis();
    render.render();
    millis = System.currentTimeMillis() - millis;

    synchronized ( this )
    {
      init   = null;
      render = null;
    }

      // finish repainting:
    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE,
                   String.format( Locale.US, userBreak ? "User break after %.1f sec!" :"3D Scene - FINISHED (%.1f sec)",
                                             (0.001 * millis) ) );
      preview.repaintAll();
      preview.stopRepaintLoop();
    }

      // save the result image in PNG format:
    if ( bak != null )
      try
      {
        bak.saveFile( (String)null, null );
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error writting image!", e );
      }
  }

  //==========================================================================
  //  Interface Property:

  public final static String PERSPECTIVE     = "Perspective";
  public final static String AZIMUTH_ANGLE   = "Azimuth angle";
  public final static String ELEVATION_ANGLE = "Elevation angle";
  public final static String DISTANCE        = "Camera distance";
  public final static String VIEW_VOLUME     = "View volume";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // PERSPECTIVE:
    if ( key.compareTo(PERSPECTIVE) == 0 )
    {
      perspective = booleanProperty( value, perspective );
      return;
    }
      // AZIMUTH_ANGLE:
    if ( key.compareTo(AZIMUTH_ANGLE) == 0 )
    {
      azimuth = doubleProperty( value, azimuth );
      return;
    }
      // ELEVATION_ANGLE:
    if ( key.compareTo(ELEVATION_ANGLE) == 0 )
    {
      elevation = doubleProperty( value, elevation );
      return;
    }
      // DISTANCE:
    if ( key.compareTo(DISTANCE) == 0 )
    {
      distance = doubleProperty( value, distance, 0.0, 1.e+6 );
      return;
    }
      // VIEW_VOLUME:
    if ( key.compareTo(VIEW_VOLUME) == 0 )
    {
      viewVolume = doubleProperty( value, viewVolume, 1.0, 150.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(PERSPECTIVE) == 0 )
      return perspective;
    if ( key.compareTo(AZIMUTH_ANGLE) == 0 )
      return azimuth;
    if ( key.compareTo(ELEVATION_ANGLE) == 0 )
      return elevation;
    if ( key.compareTo(DISTANCE) == 0 )
      return distance;
    if ( key.compareTo(VIEW_VOLUME) == 0 )
      return viewVolume;
    if ( key.compareTo(REJECT) == 0 &&
         isConnected( PL_INPUT ) )
      return true;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Render3DWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForRender3D";

  /** Object description. */
  private final static String DESCRIPTION = "Renders 3D scene into RasterGraphics.";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,   IFACE+"Worker" );
      t.newOptOutputPlug( PL_DIRECT,  IFACE+"Trigger" );
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"Render3D" );
      t.newOutputPlug(    PL_RASTER,  IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_WINDOW,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER, IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( PERSPECTIVE, TYPE_BOOLEAN, "Use perspective camera?", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( AZIMUTH_ANGLE, TYPE_DOUBLE, "Azimuth angle in degrees", true );
        t.propDefault( 30.0 );
      t.propEnd();
      t.propBegin( ELEVATION_ANGLE, TYPE_DOUBLE, "Elevation angle in degrees", true );
        t.propDefault( 20.0 );
      t.propEnd();
      t.propBegin( DISTANCE, TYPE_DOUBLE, "Camera distance from the origin", true );
        t.propDefault( 10.0 );
        t.propBounds(  0.0, 1.e+6 );
      t.propEnd();
      t.propBegin( VIEW_VOLUME, TYPE_DOUBLE, "Viewing volume of the camera in world-space units / degrees", true );
        t.propDefault( 60.0 );
        t.propBounds(  1.0, 150.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
