package cz.cuni.jagrlib.worker;

import java.util.Locale;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker for sample-based rendering (Ray-casting, ray-tracing).
 * {@link ImageFunction} or {@link RealFunction} should be provided,
 * {@link ImageSynthesizer} converts it to {@link RasterGraphics}.
 *
 * @version 0.26 $Rev: 470 $ $Date: 2007-05-27 07:43:43 +0200 (ne, 27 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/SampleRenderWorker.java">SampleRenderWorker.java</a>
 */
public class SampleRenderWorker extends Piece implements Worker, ProgressListener
{
  //==========================================================================
  //  Data:

  /** Use single pass to render the image? */
  protected boolean singlePass = false;

  /** Image synthesizer (should be stopped in {@link #stop}). */
  protected ImageSynthesizer synt = null;

  /** Preview-window redraw timer (for progress indication). */
  protected RedrawTimer timer = null;

  /** Progress-counter object (does progress-computation for me). */
  protected ProgressCounter pc = new ProgressCounter();

  //==========================================================================
  //  Interface ProgressListener:

  /** Sets progress state ({@code 0 &lt;= state &lt;= total}). */
  public void progress ( long state, long total )
  {
    pc.progress( state, total );
    if ( timer != null )
      timer.setTitle( pc.message() );
  }

  /** Repaint progress indicator (if applicable). */
  public void repaint ()
  {
    if ( timer != null )
      timer.interrupt();
  }

  //==========================================================================
  //  Interface Worker:

  /**
   * Tries to stop the thread/computation..
   * The process should yield itself.
   * For {@link Runnable} implementators: {@link Thread#interrupt} will be called afterwards.
   */
  @Override
  public synchronized void stop ()
  {
    userBreak = true;
    if ( synt != null )
      synt.stop();
  }

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final RasterGraphics image   = (RasterGraphics)getInterface( PL_IMAGE,   IFACE+"RasterGraphics" );
    final GraphicsViewer preview = (GraphicsViewer)getInterface( PL_RASTER,  IFACE+"GraphicsViewer" );
    final DataFileFormat bak     = (DataFileFormat)getInterface( PL_TRIGGER, IFACE+"DataFileFormat" );

    synchronized ( this )
    {
      synt = (ImageSynthesizer)getInterface( PL_OUTPUT, IFACE+"ImageSynthesizer" );
    }
    if ( synt == null ) return;

      // progress-indication:
    timer = null;
    if ( preview != null )
      synt.setProgressListener( this );
    else
    if ( pl != null )
      synt.setProgressListener( pl );

    if ( preview != null )
    {
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

      // timer thread:
    if ( preview != null )
    {
      timer = new RedrawTimer();
      timer.startTimer( preview, 0L );
    }

      // determine image size:
    int width, height;
    if ( image != null )
    {                                     // image has the highest priority
      width  = image.getWidth();
      height = image.getHeight();
      synt.setBounds( width, height );
    }
    else
    {                                     // at least I'll try the synthesizer..
      width  = intProperty( synt, ImageSynthesizer.WIDTH,  320 );
      height = intProperty( synt, ImageSynthesizer.HEIGHT, 240 );
    }

    if ( singlePass )
    {
      pc.setInterval( 1.0, 0.0 );
      pc.progress( 0, 1 );
      synt.renderRectangle( 0, 0, width, height );
    }

    else                                  // tri-part rendering
    {
        // render the 1st part:
      pc.setInterval( 0.25, 0.0 );
      pc.progress( 0, 1 );
      synt.renderRectangle( 0, 0, width/2, height/2 );
      if ( preview != null )
        preview.repaintAll();

      if ( !userBreak )
      {
          // render the 2nd part:
        pc.setInterval( 0.25, 0.25 );
        synt.renderRectangle( 0, height/2, width/2, height );
        if ( preview != null )
          preview.repaintAll();

          // finish rendering:
        if ( !userBreak )
        {
          pc.setInterval( 0.5, 0.5 );
          synt.renderRectangle( width/2, 0, width, height );
        }
      }
    }
    pc.progress( 1, 1 );

    synchronized ( this )
    {
      synt = null;
    }
    if ( preview != null )
    {
      String tit = String.format( Locale.US, userBreak ? "User break after %.1f sec!" :
                                                         "FINISHED (%.1f sec)", pc.totalTime() );
      if ( timer != null )
        timer.setTitle( tit );
      preview.set( GraphicsViewer.WINDOW_TITLE, tit );
      preview.repaintAll();
      preview.stopRepaintLoop();
    }

    if ( timer != null )
      timer.stopTimer();

      // save the result image in raster file-format:
    if ( bak != null )
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( Exception e )
      {
        throw new JaGrLibException( "Error writting image!", e );
      }
  }

  //==========================================================================
  //  Interface Property:

  public final static String SINGLE_PASS = "Single pass";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // SINGLE_PASS:
    if ( key.compareTo(SINGLE_PASS) == 0 )
    {
      singlePass = booleanProperty( value, singlePass );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(SINGLE_PASS) == 0 )
      return singlePass;
    if ( key.compareTo(REJECT) == 0 &&
         isConnected( PL_INPUT ) )
      return true;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "SampleRenderWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForSampleRendering";

  /** Object description. */
  private final static String DESCRIPTION = "Worker for sample-based rendering: Ray-Casting, Ray-tracing, ...";

  /** Object category. */
  protected final static String CATEGORY = C_3D + '.' + C_RENDER + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newInputPlug(     PL_INPUT,   IFACE+"Worker" );
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"ImageSynthesizer" );
      t.newOptOutputPlug( PL_IMAGE,   IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_RASTER,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER, IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( SINGLE_PASS, TYPE_BOOLEAN, "Single pass rendering", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
