package cz.cuni.jagrlib.worker;

import java.io.*;
import java.util.Locale;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: GeometrySearch tests (2D center-pass & ray-pass).
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/LineWorker.java">LineWorker.java</a>
 */
public class Search2DWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Number of objects in a set. */
  protected int setSize             = 4096;

  /** Number of point-queries. */
  protected int querySize           = 128;

  /** Number of nearest points to be looked for. */
  protected int nearest             = 8;

  /** Limit distance for searches. */
  protected double maxDistance      = 100.0;

  /** Use color based on point-center distance? */
  protected boolean colorByDistance = false;

  /** Use segments & ray-pass. */
  protected boolean segments        = false;

  /** Use visual presentation? */
  protected boolean visual          = true;

  /** Every point in the set will be painted using this color. */
  protected final static int[] BASE_COLOR  = {   0,   0, 128, 255 };

  /** Color of probe points. */
  protected final static int[] PROBE_COLOR = { 255, 255, 255, 255 };

  /** Minimal segment length in pixels. */
  protected final static double MIN_LEN    = 1.0;

  /** Maximal segment length in pixels. */
  protected final static double MAX_LEN    = 5.0;

  /** Implemented data-set types. */
  protected static enum DistributionType
  {
    /** Uniform random points. */
    DISTR_UNIFORM,
    /** [x,z] projection of Lorentz attractor. */
    DISTR_LORENTZ_XZ,
  }

  /** Actual function type. */
  protected DistributionType type = DistributionType.DISTR_LORENTZ_XZ;

  /** Constants for Lorenz attractor: dt */
  protected final static double DT    = 0.004;

  /** Constants for Lorenz attractor: delta */
  protected final static double DELTA = 10.0;

  /** Constants for Lorenz attractor: r */
  protected final static double R     = 28.0;

  /** Constants for Lorenz attractor: b */
  protected final static double B     = 8.0 / 3.0;

  //==========================================================================
  //  Interface Worker:

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final Brep           data    = (Brep)          getInterface( PL_DATA,    IFACE+"Brep" );
    final GeometrySearch geom    = (GeometrySearch)getInterface( PL_DIRECT,  IFACE+"GeometrySearch" );
    final RasterGraphics raster  = (RasterGraphics)getInterface( PL_OUTPUT,  IFACE+"RasterGraphics" );
    final GraphicsViewer preview = (GraphicsViewer)getInterface( PL_RASTER,  IFACE+"GraphicsViewer" );
    final DataFileFormat file    = (DataFileFormat)getInterface( PL_TRIGGER, IFACE+"DataFileFormat" );

    if ( data   == null ||
         geom   == null ||
         raster == null ) return;

      // initialize the preview window:
    if ( visual && preview != null )
    {
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

      // initialize the data set:
    long start, finish, tmp;
    LogFile.resetMemoryStatistics();

    RandomJames rnd = new RandomJames();
    data.setCoordTypes( Brep.ATTR_COORD_2D, Brep.ATTR_COORD_2I );
    data.init();
    int width  = raster.getWidth();
    int height = raster.getHeight();
    int i, j, v, v2, ed;
    double[] coord  = new double[ 2 ];
    double[] coord2 = new double[ 2 ];
    int ctx = Brep.CTX_DEFAULT;

    if ( visual )
      raster.setColor( BASE_COLOR );

      // variables for Lorenz:
    double lx = 0.1;
    double ly = 0.0;
    double lz = 0.0;
    double dt, dx, dy, dz;
    start     = System.currentTimeMillis();

    for ( i = 0; i++ < setSize && !userBreak; )
    {
      switch ( type )
      {
        case DISTR_LORENTZ_XZ:
            // Lorenz attractor (see e.g. http://astronomy.swin.edu.au/~pbourke/fractals/lorenz/):
          do
          {
            dt  = DT * ( 1.0 + rnd.uniformNumber() );
            dx  = dt * ( DELTA * (ly - lx) );
            dy  = dt * ( lx * (R - lz) - ly );
            dz  = dt * ( lx * ly - B * lz );
            lx += dx;
            ly += dy;
            lz += dz;
            coord[0] =  0.5 + 0.04 * lx;
            coord[1] = -0.2 + 0.03 * lz;
          } while ( coord[0] < 0.0 || coord[0] > 1.0 ||
                    coord[1] < 0.0 || coord[1] > 1.0 );
          break;

        default:
            // uniform random point:
          coord[0] = rnd.uniformNumber();
          coord[1] = rnd.uniformNumber();
      }

      coord[0] *= width;
      coord[1] *= height;
      v         = data.createVertex();
      data.setVertexCoords( ctx, v, coord );

      if ( segments )
      {
        final double len = rnd.randomDouble( MIN_LEN, MAX_LEN );
        final double a   = rnd.uniformNumber() * Math.PI;
        coord2[0] = coord[0] + len * Math.sin( a );
        coord2[1] = coord[1] + len * Math.cos( a );
        v2        = data.createVertex();
        data.setVertexCoords( ctx, v2, coord2 );
        ed        = data.createEdge();
        data.setEdgeVertices( ed, v, v2 );
          // draw the segment:
        if ( visual )
          DefaultRasterGraphics.drawLine( raster, coord[0], coord[1], coord2[0], coord2[1] );
      }
      else
          // draw the point:
        if ( visual )
          raster.setPixel( Formula.round( coord[0] ), Formula.round( coord[1] ) );
    }

      // refresh the output window:
    if ( visual && preview != null )
    {
      preview.repaintAll();
      preview.set( GraphicsViewer.WINDOW_TITLE, userBreak ? "User break!" : "Base set" );
    }

    if ( userBreak ) return;

      // prepare array for statistics:
    long[] stat = new long[ geom.getStatistics( null ) ];

      // init the search data structure:
    LogFile.log( LogFile.SEPARATOR3 );
    Object descr = geom.get( TEXT_DESCRIPTION );
    if ( descr != null )
      LogFile.log( "Data structure: " + descr.toString() + ", area: " + width + "x" + height );
    LogFile.memoryStatistics();

    tmp    = System.currentTimeMillis();
    geom.buildDirectory( segments ? Brep.EDGE : Brep.VERTEX, ctx );
    finish = System.currentTimeMillis();

    LogFile.log( "buildDirectory: " + ((type == DistributionType.DISTR_LORENTZ_XZ) ? "Lorentz [x,z]" : "uniform") +
                 ", size: " + setSize +
                 String.format( Locale.US, ", geometry generation (Brep): %.2f sec, building data structure: %.2f sec",
                                1.0e-3 * (tmp - start), 1.0e-3 * (finish - tmp) ) );
    if ( stat.length >= 3 )
    {
      geom.getStatistics( stat );
      LogFile.log( "buildDirectory operations: " + stat[0] + " point-point, " +
                   stat[1] + " point-box, " + stat[2] + " heap" );
    }
    LogFile.memoryStatistics();

      // refresh the output window:
    if ( visual && preview != null )
    {
      preview.repaintAll();
      preview.set( GraphicsViewer.WINDOW_TITLE, "Search data structure" );
    }

      // perform queries:
    int[] color     = { 0, 0, 0, 255 };
    double[] hit    = new double[ 2 ];
    BrepIterator it = null;
    double dist, last;
    double colorMul = 1.0 / nearest;
    long counter    = 0L;
    long hash       = 0L;
    long errors     = 0L;
    double[] dir    = new double[ 2 ];
    int[] edge      = new int[ 2 ];
    start           = System.currentTimeMillis();

    if ( segments )
      for ( i = 0; i < querySize && !userBreak; i++ )
      {
          // origin coordinates:
        coord[0] = rnd.randomDouble( 1.0, width  - 1.0 );
        coord[1] = rnd.randomDouble( 1.0, height - 1.0 );

          // random direction:
        final double angle = rnd.randomDouble( 0.0, Math.PI * 2.0 );
        dir[0] = Math.sin( angle );
        dir[1] = Math.cos( angle );

          // init center pass:
        it = geom.rayPass( coord, dir, it );
        if ( it == null ) continue;       // to be sure..

        //if ( visual )
        //{
        //  raster.setColor( PROBE_COLOR );
        //  DefaultRasterGraphics.drawLine( raster, coord[0], coord[1],
        //                                  coord[0] + width * 0.1 * dir[0],
        //                                  coord[1] + width * 0.1 * dir[1] );
        //}

          // result iteration:
        last = 0.0;
        for ( j = 0; j < nearest && (v = it.next()) != Brep.NULL; j++ )
        {                                 // process one result
          data.getEdgeVertices( v, edge );
          data.getVertexCoords( ctx, edge[0], hit );
          data.getVertexCoords( ctx, edge[1], coord2 );

          dist = Geometry.raySegment2D( coord, dir, hit, coord2 );
          if ( dist < last )
          {
            LogFile.debug( "Error in result ordering: query #" + i + ", result #" + (j-1) +
                           " was " + last + ", result #" + j + " is " + dist + "!" );
            errors++;
          }
          last = dist;

            // the segment is valid:
          counter++;
          hash = hash * 348937L + v * 8999L + 4831L;

          if ( visual )
          {
            Formula.colorRamp( 1.0 - colorMul * j, color );
            raster.setColor( color );
            DefaultRasterGraphics.drawLine( raster, hit[0], hit[1], coord2[0], coord2[1] );
          }
        }

          // refresh the output window, progress:
        if ( (i & 63) == 0 )
        {
          if ( pl != null )
            pl.progress( i, querySize );
          if ( visual && preview != null )
          {
            preview.repaintAll();
            preview.set( GraphicsViewer.WINDOW_TITLE, "Searching - " + i + " queries" );
          }
        }
      }

    else

      for ( i = 0; i < querySize && !userBreak; i++ )
      {
          // center coordinates:
        coord[0] = rnd.randomDouble( 1.0, width  - 1.0 );
        coord[1] = rnd.randomDouble( 1.0, height - 1.0 );

          // init center pass:
        it = geom.centerPass( coord, maxDistance, it );
        if ( it == null ) continue;       // to be sure..

          // result iteration:
        last = 0.0;
        for ( j = 0; j < nearest && (v = it.next()) != Brep.NULL; j++ )
        {                                 // process one result
          data.getVertexCoords( ctx, v, hit );
          dist = Math.hypot( hit[0] - coord[0], hit[1] - coord[1] );

          if ( dist > maxDistance ) break;

          if ( dist < last )
          {
            LogFile.debug( "Error in result ordering: query #" + i + ", result #" + (j-1) +
                           " was " + last + ", result #" + j + " is " + dist + "!" );
            errors++;
          }
          last = dist;

            // the point is valid:
          counter++;
          hash = hash * 348937L + v * 8999L + 4831L;

          if ( visual )
          {
            if ( colorByDistance )
              Formula.colorRamp( 1.0 - dist * 0.04, color );
            else
              Formula.colorRamp( 1.0 - colorMul * j, color );

            raster.putPixel( Formula.round( hit[0] ), Formula.round( hit[1] ), color );
          }
        }

        if ( visual )
        {
            // draw center:
          int x = Formula.round( coord[0] );
          int y = Formula.round( coord[1] );
          raster.putPixel( x,   y-1,    PROBE_COLOR );
          raster.setHLine( x-1, x+2, y, PROBE_COLOR );
          raster.putPixel( x,   y+1,    PROBE_COLOR );
        }

          // refresh the output window, progress:
        if ( (i & 63) == 0 )
        {
          if ( pl != null )
            pl.progress( i, querySize );
          if ( visual && preview != null )
          {
            preview.repaintAll();
            preview.set( GraphicsViewer.WINDOW_TITLE, "Searching - " + i + " queries" );
          }
        }
      }

    finish = System.currentTimeMillis();

    if ( userBreak )
      LogFile.log( String.format( Locale.US, "User break after %d operations, time: %.2f sec", i, 1.0e-3 * (finish - start) ) );

    else
    {
        // statistics about searching:
      if ( segments )
        LogFile.log( "Search: " + querySize + " rays, " + nearest + " nearest intersections, " +
                     counter + " objects returned" );
      else
        LogFile.log( "Search: " + querySize + " lookups, " + nearest + " nearest objects, " +
                     maxDistance + " max distance, " + counter + " objects returned" );

      LogFile.log( "Hash: " + hash + ", errors: " + errors +
                   String.format( Locale.US, ", time: %.2f sec", 1.0e-3 * (finish - start) ) );
      if ( stat.length >= 3 )
      {
        geom.getStatistics( stat );
        LogFile.log( "Search operations: " + stat[0] + " object, " + stat[1] + " box, " +
                     stat[2] + " heap" );
      }
    }
    LogFile.memoryStatistics();

    if ( visual )
    {
        // finish repainting:
      if ( preview != null )
      {
        preview.set( GraphicsViewer.WINDOW_TITLE, userBreak ? "User break!" : "FINISHED" );
        preview.repaintAll();
        preview.stopRepaintLoop();
      }

        // save the result image in PNG format:
      if ( file != null )
        try
        {
          file.saveFile( (String)null, null );
        }
        catch ( IOException e )
        {
          throw new JaGrLibException( "Error writting image!", e );
        }
      }
  }

  //==========================================================================
  //  Interface Property:

  public final static String SEGMENTS          = "Segments";
  public final static String VISUAL            = "Visual presentation";
  public final static String SET_TYPE          = "Set type";
  public final static String SET_SIZE          = "Set size";
  public final static String QUERY_SIZE        = "Query size";
  public final static String NEAREST           = "Nearest";
  public final static String MAX_DISTANCE      = "Max distance";
  public final static String COLOR_BY_DISTANCE = "Color by distance";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // SEGMENTS:
    if ( key.compareTo(SEGMENTS) == 0 )
    {
      segments = booleanProperty( value, segments );
      return;
    }
      // VISUAL:
    if ( key.compareTo(VISUAL) == 0 )
    {
      visual = booleanProperty( value, visual );
      return;
    }
      // SET_TYPE:
    if ( key.compareTo(SET_TYPE) == 0 )
    {
      type = enumProperty( value, type );
      return;
    }
      // SET_SIZE:
    if ( key.compareTo(SET_SIZE) == 0 )
    {
      setSize = intProperty( value, setSize, 100, 10000000 );
      return;
    }
      // QUERY_SIZE:
    if ( key.compareTo(QUERY_SIZE) == 0 )
    {
      querySize = intProperty( value, querySize, 1, 1000000 );
      return;
    }
      // NEAREST:
    if ( key.compareTo(NEAREST) == 0 )
    {
      nearest = intProperty( value, nearest, 1, 10000000 );
      return;
    }
      // MAX_DISTANCE:
    if ( key.compareTo(MAX_DISTANCE) == 0 )
    {
      maxDistance = doubleProperty( value, maxDistance, 0.0, 1.e6 );
      return;
    }
      // COLOR_BY_DISTANCE:
    if ( key.compareTo(COLOR_BY_DISTANCE) == 0 )
    {
      colorByDistance = booleanProperty( value, colorByDistance );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(SEGMENTS) == 0 )
      return segments;
    if ( key.compareTo(VISUAL) == 0 )
      return visual;
    if ( key.compareTo(SET_TYPE) == 0 )
      return type.ordinal();
    if ( key.compareTo(SET_SIZE) == 0 )
      return setSize;
    if ( key.compareTo(QUERY_SIZE) == 0 )
      return querySize;
    if ( key.compareTo(NEAREST) == 0 )
      return nearest;
    if ( key.compareTo(MAX_DISTANCE) == 0 )
      return maxDistance;
    if ( key.compareTo(COLOR_BY_DISTANCE) == 0 )
      return colorByDistance;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "Search2DWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForGeometrySearch";

  /** Object description. */
  private final static String DESCRIPTION = "Center-search test of the given GeometrySearch module.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + "." + C_DATA + "." + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_DATA,    IFACE+"Brep" );
      t.newOutputPlug(    PL_DIRECT,  IFACE+"GeometrySearch" );
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"RasterGraphics" );
      t.newOptOutputPlug( PL_RASTER,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER, IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( SEGMENTS, TYPE_BOOLEAN, "Generate line segments?", true );
        t.propDefault( false );
      t.propEnd();
      t.propBegin( VISUAL, TYPE_BOOLEAN, "Use visual presentation?", true );
        t.propDefault( true );
      t.propEnd();
      t.propBegin( SET_TYPE, TYPE_INTEGER, "Distribution type", true );
        t.propDefault( DistributionType.DISTR_LORENTZ_XZ.ordinal() );
        t.propManipulator( MANIPULATOR_COMBO );
        t.propEnum( "Uniform",       DistributionType.DISTR_UNIFORM.ordinal(),    "Uniform random distribution" );
        t.propEnum( "Lorentz [x,z]", DistributionType.DISTR_LORENTZ_XZ.ordinal(), "Lorentz attractor ([x,z] projection)" );
      t.propEnd();
      t.propBegin( SET_SIZE, TYPE_INTEGER, "Set size", true );
        t.propDefault( 4096 );
        t.propBounds(  100, 10000000 );
      t.propEnd();
      t.propBegin( QUERY_SIZE, TYPE_INTEGER, "Query size", true );
        t.propDefault( 128 );
        t.propBounds(  1, 1000000 );
      t.propEnd();
      t.propBegin( NEAREST, TYPE_INTEGER, "Number of points to be queried", true );
        t.propDefault( 8 );
        t.propBounds(  1, 10000000 );
      t.propEnd();
      t.propBegin( MAX_DISTANCE, TYPE_DOUBLE, "Maximal distance to be queried", true );
        t.propDefault( 100.0 );
        t.propBounds(  0.0, 1.e6 );
      t.propEnd();
      t.propBegin( COLOR_BY_DISTANCE, TYPE_BOOLEAN, "Use color based on distance?", true );
        t.propDefault( false );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
