package cz.cuni.jagrlib.worker;

import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker: text drawing test composition.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/TextWorker.java">TextWorker.java</a>
 */
public class TextWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Number of strings to be drawn. */
  protected int repeat = 1024;

  /** Set negative value for randomize. */
  protected int seed   = -1;

  /** Font family to use. */
  protected String family = "Arial";

  /** Font size in points. */
  protected int fontSize = 14;

  //==========================================================================
  //  Interface Worker:

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final TextRender     text    = (TextRender)    getInterface( PL_OUTPUT,  IFACE+"TextRender" );
    final SolidBrush     brush   = (SolidBrush)    getInterface( PL_BITMASK, IFACE+"SolidBrush" );
    final GraphicsViewer preview = (GraphicsViewer)getInterface( PL_RASTER,  IFACE+"GraphicsViewer" );
    final DataFileFormat bak     = (DataFileFormat)getInterface( PL_TRIGGER, IFACE+"DataFileFormat" );

    if ( text  == null ||
         brush == null ) return;

    if ( preview != null )
    {
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview );
      runIt.safeStart();
    }

    double[] color = { 0.0, 0.0, 0.0, 1.0 };

      // random generator:
    RandomJames rnd;
    if ( seed >= 0 )
      rnd = new RandomJames( seed >>> 15, seed & 0x7fff );
    else
    {
      rnd = new RandomJames();
      rnd.randomize();
    }

    StringBuffer buf = new StringBuffer();
    TextRender.TextExtent ext = new TextRender.TextExtent();
    int    len;
    double x, y, o;
    text.createFont( family, 0,
                     fontSize );
    text.putFont( 1 );
    text.createFont( family, TextRender.FONT_BOLD,
                     fontSize );
    text.putFont( 2 );
    text.createFont( family, TextRender.FONT_ITALIC,
                     fontSize );
    text.putFont( 3 );
    text.createFont( family, TextRender.FONT_BOLD | TextRender.FONT_ITALIC,
                     fontSize );
    text.putFont( 4 );

      // draw random strings:
    for ( int i = 0; i++ < repeat && !userBreak; )
    {
        // random position & orientation:
      x = rnd.randomDouble( 56.0, 456.0 );
      y = rnd.randomDouble( 56.0, 456.0 );
      o = rnd.uniformNumber() * 360.0;

        // random string:
      buf.setLength( 0 );
      len = rnd.randomInteger( 1, 12 );
      while ( len-- > 0 )
        buf.append( (char)rnd.randomInteger( 32, 127 ) );

        // draw the text:
      text.getFont( (i & 3) + 1 );
      text.fontOrientation( o );
      text.drawText( x, y,
                     buf.toString(), ext );
      text.drawText( x + ext.x1 - ext.x0, y + ext.y1 - ext.y0,
                     buf.toString(), null );
      //LogFile.log( String.format( "%6.2f, %6.2f, %5.1f, '%s'", x, y, o, buf.toString() ) );

        // stroke it:
      if ( (i & 7) == 0 )
      {
          // set random color:
        color[0] = rnd.randomDouble( 0.3, 1.0 );
        color[1] = rnd.randomDouble( 0.3, 1.0 );
        color[2] = rnd.randomDouble( 0.3, 1.0 );
        brush.setColor( color );
        brush.fill();
      }

        // refresh the output window:
      if ( (i & 63) == 0 )
      {
        if ( pl != null )
          pl.progress( i, repeat );
        if ( preview != null )
        {
          preview.repaintAll();
          preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - " + i + " strings" );
        }
      }
    }

      // set random color:
    color[0] = rnd.randomDouble( 0.3, 1.0 );
    color[1] = rnd.randomDouble( 0.3, 1.0 );
    color[2] = rnd.randomDouble( 0.3, 1.0 );
    brush.setColor( color );
    brush.fill();

      // finish repainting:
    if ( preview != null )
    {
      preview.set( GraphicsViewer.WINDOW_TITLE, "Output Image - FINISHED" );
      preview.repaintAll();
      preview.stopRepaintLoop();
    }

      // save the result image in PNG format:
    if ( bak != null )
      try
      {
        bak.saveFile( (String)null, null ); // file-name was specified in constructor
      }
      catch ( IOException e )
      {
        throw new JaGrLibException( "Error writting image!", e );
      }
  }

  //==========================================================================
  //  Interface Property:

  public final static String FAMILY    = "Font family";
  public final static String FONT_SIZE = "Font size";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // REPEAT:
    if ( key.compareTo(REPEAT) == 0 )
    {
      repeat = intProperty( value, repeat );
      return;
    }
      // FAMILY:
    if ( key.compareTo(FAMILY) == 0 )
    {
      family = stringProperty( value, family );
      return;
    }
      // FONT_SIZE:
    if ( key.compareTo(FONT_SIZE) == 0 )
    {
      fontSize = intProperty( value, fontSize, 1, 500 );
      return;
    }
      // SEED:
    if ( key.compareTo(SEED) == 0 )
    {
      seed = intProperty( value, seed );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(REPEAT) == 0 )
      return repeat;
    if ( key.compareTo(FAMILY) == 0 )
      return family;
    if ( key.compareTo(FONT_SIZE) == 0 )
      return fontSize;
    if ( key.compareTo(SEED) == 0 )
      return seed;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "TextWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForText";

  /** Object description. */
  private final static String DESCRIPTION = "Draws randomly placed text.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_DRAW + '.' + C_TEXT + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_OUTPUT,  IFACE+"TextRender" );
      t.newOutputPlug(    PL_BITMASK, IFACE+"SolidBrush" );
      t.newOptOutputPlug( PL_RASTER,  IFACE+"GraphicsViewer" );
      t.newOptOutputPlug( PL_TRIGGER, IFACE+"DataFileFormat" );
        // Property:
      t.propBegin( REPEAT, TYPE_INTEGER, "Number of random test strings", true );
        t.propDefault( 512 );
      t.propEnd();
      t.propBegin( FAMILY, TYPE_STRING, "Font family", true );
        t.propDefault( "Arial" );
      t.propEnd();
      t.propBegin( FONT_SIZE, TYPE_INTEGER, "Font size in points", true );
        t.propDefault( 14 );
        t.propBounds(  1, 500 );
      t.propEnd();
      t.propBegin( SEED, TYPE_INTEGER, "Random generator seed (-1 for randomize)", true );
        t.propDefault( -1 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
