package cz.cuni.jagrlib.worker;

import java.io.*;

import cz.cuni.jagrlib.*;
import cz.cuni.jagrlib.iface.*;
import cz.cuni.jagrlib.reg.*;

/**
 * Worker for triangle-mesh warping test composition.
 *
 * @version 0.26 $Rev: 466 $ $Date: 2007-05-26 07:12:19 +0200 (so, 26 V 2007) $ $Author: pepca $
 * @since   0.25
 * @see     <a href="../../../../../cz/cuni/jagrlib/worker/TriangleWarpingWorker.java">TriangleWarpingWorker.java</a>
 */
public class TriangleWarpingWorker extends Piece implements Worker
{
  //==========================================================================
  //  Data:

  /** Number of rows in a triangle mesh. */
  protected int    rows         = 8;

  /** Number of columns in a triangle mesh. */
  protected int    columns      = 8;

  /** Amplitude of a random perturbation. */
  protected double rndAmplitude = 0.4;

  //==========================================================================
  //  Interface Worker:

  /**
   * Executive routine of the working thread.
   */
  public void run ()
  {
    final DataFileFormat output   = (DataFileFormat)getInterface( PL_OUTPUT,   IFACE+"DataFileFormat" );
    final GraphicsViewer preview2 = (GraphicsViewer)getInterface( "raster2",   IFACE+"GraphicsViewer" );
    final Trigger        warping  = (Trigger)       getInterface( PL_DIRECT,   IFACE+"Trigger" );
    final Trigger        editor   = (Trigger)       getInterface( PL_EDITOR,   IFACE+"Trigger" );
    final Brep           data     = (Brep)          getInterface( PL_DATA,     IFACE+"Brep" );
    final Property       prop     = (Property)      getInterface( PL_PROPERTY, IFACE+"Property" );

    if ( output  == null ||
         warping == null ) return;

      // triangle mesh data:
    int ctx1        = Brep.CTX_DEFAULT;
    int ctx2        = Brep.NULL;
    RandomJames rnd = new RandomJames();
    int i, j;

      // initialize the warping data = two triangle meshes
    if ( data != null )
    {
        // create two contexts & coordinate attributes:
      data.setCoordTypes( Brep.ATTR_COORD_2D, Brep.ATTR_COORD_2I );
      data.init();
      ctx2 = data.createContext( Brep.NULL );
      data.createAttribute( Brep.VERTEX, ctx2, Brep.COORD, Brep.ATTR_COORD_2D );

        // triangle data: I. vertices
      int[][]  vert  = new int[ rows + 1 ][ columns + 1 ];
      double[] coord = new double[ 2 ];

      for ( j = 0; j <= rows; j++ )
        for ( i = 0; i <= columns; i++ )
        {                                 // init one vertex
          vert[j][i] = data.createVertex();

          coord[0] = i;
          coord[1] = j;
            // random perturbation:
          if ( i != 0 && i != columns )
            coord[0] += (rnd.uniformNumber() - 0.5) * rndAmplitude;
          if ( j != 0 && j != rows )
            coord[1] += (rnd.uniformNumber() - 0.5) * rndAmplitude;

            // source space:
          data.setVertexCoords( ctx1, vert[j][i], coord );

          coord[0] = i;
          coord[1] = j;
            // random perturbation:
          if ( i != 0 && i != columns )
            coord[0] += (rnd.uniformNumber() - 0.5) * rndAmplitude;
          if ( j != 0 && j != rows )
            coord[1] += (rnd.uniformNumber() - 0.5) * rndAmplitude;

            // target space:
          data.setVertexCoords( ctx2, vert[j][i], coord );
        }

        // triangle data: II. topology
      for ( j = 0; j < rows; j++ )
        for ( i = 0; i < columns; i++ )
        {                                 // one quad
          data.setTriangleVertices( data.createFace(), vert[j][i],   vert[j][i+1], vert[j+1][i]   );
          data.setTriangleVertices( data.createFace(), vert[j+1][i], vert[j][i+1], vert[j+1][i+1] );
        }
    }

      // set warping-mesh parameters to a module:
    if ( prop != null )
    {
      prop.set( CONTEXT2,     ctx2 );
      prop.set( MESH_ROWS,    rows );
      prop.set( MESH_COLUMNS, columns );
    }

      // transform source image into target bitmap for the 1st time..
    warping.fire( 0 );

      // attach preview window to the output image & render it:
    if ( preview2 != null )
    {
      preview2.set( GraphicsViewer.WINDOW_TITLE, "Target Image" );
      GraphicsViewer.PreviewThread runIt = new GraphicsViewer.PreviewThread( preview2 );
      runIt.safeStart();
    }

      // possibly do the interaction..
    if ( editor != null )
    {
      editor.set( CONTEXT2,     ctx2 );
      editor.set( MESH_ROWS,    rows );
      editor.set( MESH_COLUMNS, columns );
      editor.fire( 1 );
    }

      // save the output image:
    try
    {
      output.saveFile( (String)null, null ); // file-name was specified in constructor
    }
    catch ( IOException e )
    {
      throw new JaGrLibException( "Error writting image!", e );
    }
  }

  //==========================================================================
  //  Interface Property:

  /* Property to be set in a warping module. */
  public final static String CONTEXT2      = "Context2";

  public final static String MESH_ROWS     = "Mesh rows";
  public final static String MESH_COLUMNS  = "Mesh columns";
  public final static String RND_AMPLITUDE = "RND amplitude";

  /**
   * Sets the given property.
   *
   * @param  key Key string.
   * @param  value The new value.
   */
  @Override
  public void set ( String key, Object value )
  {
    if ( key == null || value == null ) return;

      // MESH_ROWS:
    if ( key.compareTo(MESH_ROWS) == 0 )
    {
      rows = intProperty( value, rows, 1, 1000 );
      return;
    }
      // MESH_COLUMNS:
    if ( key.compareTo(MESH_COLUMNS) == 0 )
    {
      columns = intProperty( value, columns, 1, 1000 );
      return;
    }
      // RND_AMPLITUDE:
    if ( key.compareTo(RND_AMPLITUDE) == 0 )
    {
      rndAmplitude = doubleProperty( value, rndAmplitude, 0.0, 1.0 );
      return;
    }
  }

  /**
   * Gets the given property.
   *
   * @param  key Key string.
   * @return The actual value or {@code null}.
   */
  @Override
  public Object get ( String key )
  {
    if ( key == null ) return null;

    if ( key.compareTo(MESH_ROWS) == 0 )
      return rows;
    if ( key.compareTo(MESH_COLUMNS) == 0 )
      return columns;
    if ( key.compareTo(RND_AMPLITUDE) == 0 )
      return rndAmplitude;

    return null;
  }

  //==========================================================================
  //  Module registration:

  /** Object name. */
  private final static String NAME = "TriangleWarpingWorker";

  /** Object template identifier. */
  protected final static String TEMPLATE_NAME = "WorkerForTriangleWarping";

  /** Object description. */
  private final static String DESCRIPTION = "Tests triangle-mesh warping implementations.";

  /** Object category. */
  protected final static String CATEGORY = C_2D + '.' + C_RASTER + '.' + C_FILTER + '.' + C_WORKER;

  /**
   * General-purpose registration routine.
   * Sets all plugs, strings, etc. to the given Template.
   */
  public static int setTemplate ( Template t, int ord )
  {
    if ( t != null && ord <= 0 )
    {
      t.setRegStrings( NAME, TEMPLATE_NAME, CATEGORY, DESCRIPTION );
        // Plugs:
      t.newOutputPlug(    PL_OUTPUT,   IFACE+"DataFileFormat" );
      t.newOptOutputPlug( "raster2",   IFACE+"GraphicsViewer" );
      t.newOutputPlug(    PL_DIRECT,   IFACE+"Trigger" );
      t.newOptOutputPlug( PL_EDITOR,   IFACE+"Trigger" );
      t.newOptOutputPlug( PL_DATA,     IFACE+"Brep" );
      t.newOptOutputPlug( PL_PROPERTY, IFACE+"Property" );
        // Property:
      t.propBegin( MESH_ROWS, TYPE_INTEGER, "Mesh rows", true );
        t.propDefault( 8 );
        t.propBounds(  1, 1000 );
      t.propEnd();
      t.propBegin( MESH_COLUMNS, TYPE_INTEGER, "Mesh columns", true );
        t.propDefault( 8 );
        t.propBounds(  1, 1000 );
      t.propEnd();
      t.propBegin( RND_AMPLITUDE, TYPE_DOUBLE, "RND amplitude", true );
        t.propDefault( 0.4 );
        t.propBounds(  0.0, 1.0 );
      t.propEnd();
    }
    return 1;
  }

  /**
   * Static registration instance for this class.
   * Automatically initialized in class-loading time.
   */
  public static final RegPiece reg;
  static
  {
    reg = new RegPiece();
    setTemplate( reg, 0 );
  }

}
