package cz.cuni.jagrlib.xml;

import java.io.*;
import java.util.List;
import java.util.*;
import java.lang.reflect.*;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import cz.cuni.jagrlib.*;

/**
 * Script read from XML file.
 *
 * @version 0.26 $Rev: 599 $ $Date: 2007-11-06 00:01:11 +0100 (út, 06 XI 2007) $ $Author: pepca $
 * @see     <a href="../../../../../cz/cuni/jagrlib/xml/JGLScript.java">JGLScript.java</a>
 */
public class JGLScript extends DefaultHandler
{
  //==========================================================================
  //  XML constants & states:

    // XML tags:
  public static final String SCRIPT        = "script";
  public static final String OBJECT        = "object";
  public static final String CLASS         = "class";
  public static final String CONSTRUCTOR   = "constructor";
  public static final String METHOD        = "method";
  public static final String MEMBER        = "member";
  public static final String NAME          = "name";
  public static final String ARG           = "arg";
  public static final String REF           = "ref";
  public static final String FUNCTION      = "function";
  public static final String RETURN        = "return";
  public static final String ARRAY         = "array";
  public static final String DATA          = "data";
  public static final String NULL          = "null";
  public static final String SET           = "set";
  public static final String KEY           = "key";
  public static final String VALUE         = "val";
  public static final String DEF           = "def";

    // XML attributes:
  public static final String A_ID          = "id";
  public static final String A_TYPE        = "type";
  public static final String A_KEY         = "key";

    // Sub-state types:
  public static final int STATE_ALL        = 0;
  public static final int STATE_CLASS      = 1;
  public static final int STATE_REF        = 2;
  public static final int STATE_NAME       = 3;

  /* Brief syntax definition:

   INSTANCE      <- <object> |
                    <ref>    |
                    <member> |
                    <function>

   OBJECT        <- INSTANCE |
                    <array>  |
                    <def>    |
                    <val>    |
                    <null>

   METHOD        <- <method> |
                    <set>

   SEQUENCE      <- OBJECT*

   <script>      <- { <object>   |
                      <array>    |
                      <def>      |
                      <function> |
                      <return> }*

   <object>      <- [ <class> [ <constructor> ] | <function> ] METHOD*

   <array>       <- { [ <class> ] SEQUENCE |
                      <data> }

   <def>         <- KEY_ATTRIBUTE [ TYPE_ATTRIBUTE ] LITERAL*

   <return>      <- ( <key> OBJECT )*

   <constructor> <- SEQUENCE

   <method>      <- <name> SEQUENCE

   <set>         <- { KEY_ATTRIBUTE | <key> | OBJECT } OBJECT

   <function>    <- INSTANCE <name> SEQUENCE

   <member>      <- { <class> | INSTANCE } <name>

   <val>         <- [ TYPE_ATTRIBUTE ] LITERAL*

   <data>        <- [ TYPE_ATTRIBUTE ] LITERAL*
  */

  /**
   * Preferred packages to be searched at object instantiation, member lookup, ..
   * Ordering of the packages does matter!
   */
  public static final String[] PREF_PACKAGES =
  {
    "",
    Template.JAGRLIB2,
    Template.JAGRLIB2 + "piece.",
    Template.JAGRLIB2 + "testing.",
    Template.JAGRLIB2 + "iface.",
    Template.JAGRLIB2 + "obscure.",
    Template.JAGRLIB2 + "worker.",
    "java.lang."
  };

  //==========================================================================
  //  Parser state variables:

  /** Object map. */
  protected Map< String, Object > objects;

  /** Result map. */
  protected Map< String, Object > results;

  /** Script file-name. */
  protected String scriptFileName;

  /** Actual line number (for error messages). */
  protected int lineNo;

  /** State stack (top of stack). */
  protected ParseState tos;

  /** For {@code <object>} section: anonymous id assigned if no explicit one was provided. */
  protected static int anonId = 1;

  /** Generic state object. */
  protected abstract class ParseState
  {
    /** Context identification (for debugging & error messages only, may be {@code null}). */
    public String context;

    /** Pointer to ancestor (FIFO). */
    protected ParseState parent;

    /** Pushes new parsing state on the stack. */
    public ParseState ()
    {
      parent = tos;
      tos    = this;
    }

    /** Swallows internal characters. */
    public void characters ( char[] ch, int start, int length )
    {
    }

    /** A new XML element is starting in my context.. */
    public void startElement ( String qName, Attributes atts )
    {
      error( "invalid tag <" + qName + '>' + (context != null ? "in the <" + context + "> context" : "") + '!' );
    }

    /** My element was finished. */
    public void endElement ( String qName )
    {
    }

    /** Text from one of descendants.. */
    public void setText ( int type, String value )
    {
    }

    /** Arbitrary value from one of descendants.. */
    public void setValue ( int type, Object value )
    {
    }

    /**
     * Checks whether the class instance was already created,
     * if not, creates it using the given constructor.
     */
    public void newInstance ( Object[] args )
    {
    }

    /**
     * Invokes the given method of the class instance.
     */
    public Object invokeMethod ( String name, Object[] args )
    {
      return null;
    }

    /** Pops this state from the stack. */
    public void pop ()
    throws SAXException
    {
      if ( parent == null )
        throw new SAXException( "Unrecoverable XML error" );

      tos = parent;
    }
  }

  //==========================================================================
  //  Support:

  /** XML format error. */
  protected void error ( String msg )
  {
    LogFile.error( "in script file '" + scriptFileName + "' on line " + lineNo + ": " + msg );
  }

  /** XML warning. */
  protected void warning ( String msg )
  {
    LogFile.log( "Script file '" + scriptFileName + "' warning on line " + lineNo + ": " + msg );
  }

  /** New object instance. */
  protected void newObject ( String name, Object instance )
  {
    objects.put( name, instance );
    LogFile.debug( "Object on line " + lineNo + ", key: " + name + ", val: " + instance.toString() );
  }

  /**
   * Invokes the given method.
   */
  protected Object globalInvoke ( Object instance, String name, Object[] args )
  {
    if ( instance == null ||
         name     == null ||
         name.length() == 0 ) return null;

    Object result = null;
    Method[] methods = instance.getClass().getMethods();
    for ( Method m : methods )
      if ( m.getName().equals( name ) )
      {
        try
        {
          result = m.invoke( instance, args );
        }
        catch ( IllegalArgumentException ex )
        {
          continue;
        }
        catch ( InvocationTargetException te )
        {
          final Throwable cause = te.getCause();
          error( "target exception in method '" + instance.getClass().getName() + '.' + name + "()'" );
          if ( cause != null )
            LogFile.log( " " + cause.getClass().getName() + ": " + cause.getMessage() );
          return null;
        }
        catch ( Exception e )
        {
          error( "invoking method '" + instance.getClass().getName() + '.' + name + "()'" );
          LogFile.log( " " + e.getClass().getName() + ": " + e.getMessage() );
          return null;
        }
        return result;
      }

    error( "cannot find '" + instance.getClass().getName() + '.' + name + "() method, ignoring.." );
    return null;
  }

  //==========================================================================
  //  Parser states:

  /**
   * Global parsing state encapsulating the whole XML file.
   * Only single {@code <script>} tag is expected.
   */
  protected class InitState extends ParseState
  {
    /** A new XML element is starting in my context.. */
    @Override
    public void startElement ( String qName, Attributes atts )
    {
      if ( SCRIPT.equals( qName ) )
        new ScriptState();
      else
        super.startElement( qName, atts );
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Parses the whole JGL script (inside of the {@code <script>} tag).
   */
  protected class ScriptState extends ParseState
  {
    /** A new XML element is starting in my context.. */
    @Override
    public void startElement ( String qName, Attributes atts )
    {
        // <object>
      if ( OBJECT.equals( qName ) )
        new ObjectState( atts );
      else
        // <array>
      if ( ARRAY.equals( qName ) )
        new ArrayState( atts );
      else
        // <def>
      if ( DEF.equals( qName ) )
        new DefState( atts );
      else
        // <function>
      if ( FUNCTION.equals( qName ) )
        new FunctionState();
      else
        // <return>
      if ( RETURN.equals( qName ) )
        new ReturnState();

      else
        super.startElement( qName, atts );
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Defines a new object instance.
   */
  protected class ObjectState extends ParseState
  {
    /** Object id (unique in the whole script, can be {@code null} for anonymous object). */
    public String id;

    /** Class object, if {@code null}, no class-name has been defined yet. */
    public Class<?> cl;

    /** [Result] object instance. */
    public Object instance;

    public ObjectState ( Attributes atts )
    {
      if ( atts != null )
        id = atts.getValue( A_ID );
    }

    /**
     * Checks whether the class instance was already created,
     * if not, creates it using the given constructor.
     */
    @Override
    public void newInstance ( Object[] args )
    {
      if ( instance != null ) return;

      if ( cl == null )
      {
        error( "class name was not defined, ignoring the whole <object> section" );
        return;
      }

      if ( args == null ||
           args.length == 0 )             // use default constructor
        try
        {
          instance = cl.newInstance();
        }
        catch ( Exception ex )
        {
        }
      else                                // constructor with arguments
      {
        Constructor<?>[] cons = cl.getConstructors();
        for ( Constructor<?> con : cons )
        {
          try
          {
            instance = con.newInstance( args );
          }
          catch ( Exception e )
          {
            continue;
          }
        }
      }

      if ( instance == null )
        error( "error instantiating the '" + cl.getName() + "' class" );
      else
        if ( id != null )
          newObject( id, instance );
    }

    /**
     * Invokes the given method of the class instance.
     */
    @Override
    public Object invokeMethod ( String name, Object[] args )
    {
      newInstance( null );
      return globalInvoke( instance, name, args );
    }

    /** A new XML element is starting in my context.. */
    @Override
    public void startElement ( String qName, Attributes atts )
    {
        // <class>
      if ( CLASS.equals( qName ) )
        new ClassState();
      else
        // <constructor>
      if ( CONSTRUCTOR.equals( qName ) )
        new ConstructorState();
      else
        // <function>
      if ( FUNCTION.equals( qName ) )
        new FunctionState();
      else
        // <method>
      if ( METHOD.equals( qName ) )
        new MethodState();
      else
        // <set>
      if ( SET.equals( qName ) )
        new SetState( atts );
      else

        super.startElement( qName, atts );
    }

    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      newInstance( null );
      parent.setValue( STATE_REF, instance );
    }

    /** Arbitrary value from one of descendants.. */
    @Override
    public void setValue ( int type, Object value )
    {
      switch ( type )
      {
        case STATE_CLASS:
          if ( value == null ||
               !(value instanceof Class) )
            return;

          if ( cl != null )
          {
            if ( instance != null )
            {
              error( "object instance was already created, ignoring redundant <class> definition" );
              return;
            }
            warning( "object class was already defined, ignoring the previous value.." );
          }

          cl = (Class<?>)value;
          return;

        case STATE_REF:                   // <function>: instead of a constructor
          if ( value != null )
          {
            instance = value;
            if ( id != null )
              newObject( id, instance );
          }
          return;
      }
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Fetches member variable from the given class/instance.
   */
  protected class MemberState extends ParseState
  {
    /** Class object. */
    protected Class<?> cl;

    /** Object instance. */
    protected Object instance;

    /** Member identifier. */
    protected String name;

    /** A new XML element is starting in my context.. */
    @Override
    public void startElement ( String qName, Attributes atts )
    {
        // <member>
      if ( MEMBER.equals( qName ) )
        new MemberState();
      else
        // <function>
      if ( FUNCTION.equals( qName ) )
        new FunctionState();
      else
        // <class>
      if ( CLASS.equals( qName ) )
        new ClassState();
      else
        // <ref>
      if ( REF.equals( qName ) )
        new RefState( STATE_REF );
      else
        // <object>
      if ( OBJECT.equals( qName ) )
        new ObjectState( atts );
      else
        // <name>
      if ( NAME.equals( qName ) )
        new TextParseState( STATE_NAME );
      else

        super.startElement( qName, atts );
    }

    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      if ( name == null )
      {
        error( "member identifier was not specified, ignoring.." );
        return;
      }

      Field fld = null;

      if ( instance != null )
        cl = instance.getClass();

      if ( cl == null )
      {
        error( "neither class name nor instance were specified, ignoring.." );
        return;
      }

      try
      {
        fld = cl.getField( name );
        parent.setValue( STATE_REF, fld.get( instance ) );
      }
      catch ( Exception e )
      {
        error( "cannot access the '" + cl.getName() + '.' + name + "' member field!" );
      }
    }

    /** Arbitrary value from one of descendants.. */
    @Override
    public void setValue ( int type, Object value )
    {
      switch ( type )
      {
        case STATE_CLASS:
          if ( value == null ||
               !(value instanceof Class) )
            return;

          cl = (Class<?>)value;
          return;

        case STATE_REF:
          instance = value;
          return;
      }
    }

    /** Text from one of descendants.. */
    @Override
    public void setText ( int type, String value )
    {
      switch ( type )
      {
        case STATE_NAME:
          if ( value == null ||
               (value = value.trim()).length() < 1 )
          {
            error( "invalid member identifier, ignoring.." );
            return;
          }

          name = value;
          break;

        default:
          super.setText( type, value );
      }
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Defines vector of the objects.
   */
  protected abstract class SequenceState extends ParseState
  {
    /** Array elements. */
    protected List< Object > list;

    /** A new XML element is starting in my context.. */
    @Override
    public void startElement ( String qName, Attributes atts )
    {
        // <val>
      if ( VALUE.equals( qName ) )
        new ArgParseState( STATE_REF, atts );
      else
        // <null>
      if ( NULL.equals( qName ) )
        new NullState();
      else
        // <ref>
      if ( REF.equals( qName ) )
        new RefState( STATE_REF );
      else
        // <object>
      if ( OBJECT.equals( qName ) )
        new ObjectState( atts );
      else
        // <array>
      if ( ARRAY.equals( qName ) )
        new ArrayState( atts );
      else
        // <def>
      if ( DEF.equals( qName ) )
        new DefState( atts );
      else
        // <member>
      if ( MEMBER.equals( qName ) )
        new MemberState();
      else
        // <function>
      if ( FUNCTION.equals( qName ) )
        new FunctionState();
      else

        super.startElement( qName, atts );
    }

    /** Arbitrary value from one of descendants.. */
    @Override
    public void setValue ( int type, Object value )
    {
      switch ( type )
      {
        case STATE_REF:
          if ( list == null )
            list = new ArrayList< Object > ();

          list.add( value );
          return;
      }
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Defines result object map.
   */
  protected class ReturnState extends SequenceState
  {
    /** Current key. */
    protected String key;

    /** A new XML element is starting in my context.. */
    @Override
    public void startElement ( String qName, Attributes atts )
    {
        // <key>
      if ( KEY.equals( qName ) )
        new TextParseState( STATE_NAME );
      else

        super.startElement( qName, atts );
    }

    /** Arbitrary value from one of descendants.. */
    @Override
    public void setValue ( int type, Object value )
    {
      switch ( type )
      {
        case STATE_REF:
          if ( key == null )
          {
            error( "return key was not defined, ignoring.." );
            return;
          }

          if ( results.get( key ) != null )
            warning( "redefinition of result, key = '" + key + "'" );

          results.put( key, value );
          key = null;
          return;
      }
    }

    /** Text from one of descendants.. */
    @Override
    public void setText ( int type, String value )
    {
      switch ( type )
      {
        case STATE_NAME:
          if ( value == null ||
               (value = value.trim()).length() < 1 )
          {
            error( "invalid key string, ignoring.." );
            return;
          }
          key = value;
          break;

        default:
          super.setText( type, value );
      }
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Defines general array object. Can be nested.
   */
  protected class ArrayState extends SequenceState
  {
    /** Object id (unique in the whole script). */
    public String id;

    /** Array element type, if {@code null}, no class-name has been defined yet. */
    public Class<?> cl;

    /** The whole array object (used in the {@code <data>} variant). */
    protected Object result;

    public ArrayState ( Attributes atts )
    {
      if ( atts != null )
        id = atts.getValue( A_ID );
    }

    /** A new XML element is starting in my context.. */
    @Override
    public void startElement ( String qName, Attributes atts )
    {
        // <data>
      if ( DATA.equals( qName ) )
        new ArgParseState( STATE_ALL, atts );
      else
        // <class>
      if ( CLASS.equals( qName ) )
        new ClassState();
      else

        super.startElement( qName, atts );
    }

    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      if ( result == null )
        if ( list == null )
          if ( cl == null )
            result = new Object[ 0 ];
          else
            result = Array.newInstance( cl, 0 );
        else
        {
          final int size = list.size();
          if ( cl == null )
            if ( size == 0 ||
                 list.get( 0 ) == null )
            {
              error( "array type was not specified and the array is empty, so it will be ignored" );
              return;
            }
            else
              cl = list.get( 0 ).getClass();

          try
          {
            result = Array.newInstance( cl, size );
            int i = 0;
            for ( Object o : list )
              Array.set( result, i++, o );
          }
          catch ( Exception e )
          {
            error( "incompatible array element types, the whole array will be ignored" );
            return;
          }
        }

      if ( id != null )
        newObject( id, result );
      parent.setValue( STATE_REF, result );
    }

    /** Arbitrary value from one of descendants.. */
    @Override
    public void setValue ( int type, Object value )
    {
      switch ( type )
      {
        case STATE_CLASS:
          if ( value == null ||
               !(value instanceof Class) )
            return;

          if ( cl != null )
          {
            warning( "array element class was already defined, ignoring the previous value.." );
          }

          if ( result != null )
          {
            warning( "both <data> and <class> sections are present, ignoring the <data> section.." );
            result = null;
          }

          cl = (Class<?>)value;
          if ( list == null )
            list = new ArrayList< Object > ();
          return;

        case STATE_ALL:
          if ( list != null )
          {
            warning( "both <data> and <class> sections are present, ignoring the <data> section.." );
            result = null;
            return;
          }

          if ( result != null )
          {
            warning( "duplicate <data> section, ignoring the old one.." );
          }

          result = value;
          return;

        default:
          super.setValue( type, value );
      }
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Constructs the object using explicitly defined constructor.
   * Implements argument-parsing schema.
   */
  protected class ConstructorState extends SequenceState
  {
    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      parent.newInstance( list == null ? (Object[])null : list.toArray() );
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Invokes the given method of the ancestor object.
   * Uses common parameter-parsing scheme from {@link JGLScript.ConstructorState}.
   */
  protected class MethodState extends SequenceState
  {
    /** Method's identifier. */
    protected String name;

    /** A new XML element is starting in my context.. */
    @Override
    public void startElement ( String qName, Attributes atts )
    {
        // <name>
      if ( NAME.equals( qName ) )
        new TextParseState( STATE_NAME );
      else

        super.startElement( qName, atts );
    }

    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      if ( name == null )
      {
        error( "missing method identifier, ignoring.." );
        return;
      }

      parent.invokeMethod( name, list == null ? (Object[])null : list.toArray() );
    }

    /** Text from one of descendants.. */
    @Override
    public void setText ( int type, String value )
    {
      switch ( type )
      {
        case STATE_NAME:
          if ( value == null ||
               (value = value.trim()).length() < 1 )
          {
            error( "invalid method identifier, ignoring.." );
            return;
          }
          name = value;
          break;

        default:
          super.setText( type, value );
      }
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Invokes the {@code set()} method of the ancestor object.
   * Uses common parameter-parsing scheme from {@link JGLScript.ConstructorState}.
   */
  protected class SetState extends SequenceState
  {
    /** Key string. */
    protected String key;

    /** Force ancestor to create object instance. */
    public SetState ( Attributes atts )
    {
      if ( atts != null )
        key = atts.getValue( A_KEY );

      parent.newInstance( null );
    }

    /** A new XML element is starting in my context.. */
    @Override
    public void startElement ( String qName, Attributes atts )
    {
        // <key>
      if ( KEY.equals( qName ) )
        new TextParseState( STATE_NAME );
      else

        super.startElement( qName, atts );
    }

    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      if ( key == null )
      {
        if ( list != null &&
             list.size() > 1 &&
             list.get( 0 ) != null )
        {
          key = list.get( 0 ).toString();
          list.remove( 0 );
        }

        if ( key == null )
        {
          error( "missing key string, ignoring.." );
          return;
        }
      }

      if ( list == null ||
           list.size() < 1 )
      {
        error( "set() - value argument is missing" );
        return;
      }

      parent.invokeMethod( "set", new Object[] { key, list.get( 0 ) } );
    }

    /** Text from one of descendants.. */
    @Override
    public void setText ( int type, String value )
    {
      switch ( type )
      {
        case STATE_NAME:
          if ( value == null ||
               (value = value.trim()).length() < 1 )
          {
            error( "invalid key string, ignoring.." );
            return;
          }
          key = value;
          break;

        default:
          super.setText( type, value );
      }
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Executes function from the other object.
   * Can be used as regular r-value or as substitution of constructor.
   */
  protected class FunctionState extends MethodState
  {
    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      if ( name == null )
      {
        error( "missing method identifier, ignoring.." );
        return;
      }
      if ( list == null ||
           list.size() < 1 )
      {
        error( "object instance was not defined, ignoring.." );
        return;
      }

      Object instance = list.get( 0 );
      list.remove( 0 );

      Object result = globalInvoke( instance, name, list.toArray() );
      if ( result != null )
        parent.setValue( STATE_REF, result );
    }
  }

  //--------------------------------------------------------------------------

  /**
   * General text-parsing state, accumulates text into {@link StringBuffer}.
   */
  protected class TextParseState extends ParseState
  {
    /** Node type (defined at constructing time by the parent node). */
    protected final int type;

    /** Accumulated text. */
    protected StringBuffer sb = new StringBuffer();

    public TextParseState ( int type )
    {
      this.type = type;
    }

    /** Swallows internal characters. */
    @Override
    public void characters ( char[] ch, int start, int length )
    {
      sb.append( ch, start, length );
    }

    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      parent.setText( type, sb.toString() );
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Reads full class specification and returns it using {@link JGLScript.ParseState#setValue setValue()}.
   */
  protected class ClassState extends TextParseState
  {
    public ClassState ()
    {
      super( STATE_CLASS );
    }

    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      final String className = sb.toString().trim();

      if ( className.length() < 1 )
      {
        error( "invalid class name will be ignored" );
        return;
      }

      Class<?> cl = null;
      for ( String packageName : PREF_PACKAGES )
      {
        try
        {
          if ( (cl = Class.forName( packageName + className )) != null ) break;
        }
        catch ( ClassNotFoundException ex )
        {}
      }

      if ( cl == null )
      {
        error( "cannot find class '" + className + "'" );
        return;
      }

      parent.setValue( type, cl );
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Reads reference to previously defined object.
   */
  protected class RefState extends TextParseState
  {
    public RefState ( int type )
    {
      super( type );
    }

    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      final String ident = sb.toString().trim();

      if ( ident.length() < 1 )
      {
        error( "invalid reference name will be ignored" );
        return;
      }

      Object reference = objects.get( ident );
      if ( reference == null )
        error( "invalid reference" + (ident != null ? " '" + ident + "'" : "") + ", null will be substituted" );
      else
        parent.setValue( type, reference );
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Null reference literal.
   */
  protected class NullState extends ParseState
  {
    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      parent.setValue( STATE_REF, null );
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Reads direct argument: atomic type, 1D array..
   */
  protected class ArgParseState extends ParseState
  {
    /** Node type (defined at constructing time by the parent node). */
    protected final int type;

    /** Argument type defined by tag attribute. */
    protected String typeAttr;

    /** Accumulated text. */
    protected StringBuffer sb = new StringBuffer();

    public ArgParseState ( int type, Attributes attr )
    {
      this.type = type;
      if ( attr != null )
        typeAttr = attr.getValue( A_TYPE );
    }

    /** Finishes element processing (overridable method). */
    protected void finish ( Object obj )
    {
      parent.setValue( type, obj );
    }

    /** Swallows internal characters. */
    @Override
    public void characters ( char[] ch, int start, int length )
    {
      sb.append( ch, start, length );
    }

    /** My element was finished. */
    @Override
    public void endElement ( String qName )
    {
      String  origStr = sb.toString();
      String  valStr  = origStr.trim();
      Byte    b       = null;
      Short   s       = null;
      Integer i       = null;
      Long    l       = null;
      Double  d       = null;
      Float   f       = null;

      if ( typeAttr == null ||
           typeAttr.length() < 1 )
      {
        try
        {
          i = Integer.decode( valStr );
        }
        catch ( NumberFormatException e )
        {
          i = null;
        }
        if ( i != null )
        {
          finish( i );
          return;
        }

        try
        {
          d = Double.valueOf( valStr );
        }
        catch ( NumberFormatException e )
        {
          d = null;
        }
        if ( d != null )
        {
          finish( d );
          return;
        }

        if ( valStr.equals( "false" ) )
        {
          finish( false );
          return;
        }
        if ( valStr.equals( "true" ) )
        {
          finish( true );
          return;
        }

        if ( valStr.equals( "null" ) )
        {
          finish( null );
          return;
        }

        finish( valStr );
        return;
      }

      switch ( Character.toLowerCase( typeAttr.charAt( 0 ) ) )
      {
        case 'b':
          try
          {
            b = Byte.decode( valStr );
          }
          catch ( NumberFormatException e )
          {
            error( "error in byte format, 0 will be substituted" );
            b = new Byte( (byte)0 );
          }
          finish( b );
          return;

        case 's':
          try
          {
            s = Short.decode( valStr );
          }
          catch ( NumberFormatException e )
          {
            error( "error in short format, 0 will be substituted" );
            s = new Short( (short)0 );
          }
          finish( s );
          return;

        case 'i':
          try
          {
            i = Integer.decode( valStr );
          }
          catch ( NumberFormatException e )
          {
            error( "error in integer format, 0 will be substituted" );
            i = new Integer( 0 );
          }
          finish( i );
          return;

        case 'j':
          try
          {
            l = Long.decode( valStr );
          }
          catch ( NumberFormatException e )
          {
            error( "error in long format, 0L will be substituted" );
            l = new Long( 0L );
          }
          finish( l );
          return;

        case 'd':
          try
          {
            d = Double.valueOf( valStr );
          }
          catch ( NumberFormatException e )
          {
            error( "error in double format, 0.0 will be substituted" );
            d = new Double( 0.0 );
          }
          finish( d );
          return;

        case 'f':
          try
          {
            f = Float.valueOf( valStr );
          }
          catch ( NumberFormatException e )
          {
            error( "error in float format, 0.0f will be substituted" );
            f = new Float( 0.0f );
          }
          finish( f );
          return;

        case 'c':
          if ( origStr.length() < 1 )
          {
            error( "error in char format, ' ' will be substituted" );
            finish( ' ' );
          }
          else
            finish( origStr.charAt( 0 ) );
          return;

        case 'z':
          finish( Boolean.valueOf( valStr ) );
          return;

        case 'l':
          if ( !typeAttr.equalsIgnoreCase( "Ljava.lang.String;" ) )
          {
            error( "cannot create literal of type '" + typeAttr + "', null will be substituted" );
            finish( null );
            return;
          }

        case 't':                         // String
          finish( origStr );
          return;

        case '[':
        {
          if ( typeAttr.length() < 2 )
          {
            error( "array item type is missing, null will be substituted" );
            finish( null );
            return;
          }

          char chType = Character.toLowerCase( typeAttr.charAt( 1 ) );
          if ( chType == 't' )
          {                               // array of String literals
            finish( Formula.parseStringList( origStr ) );
            return;
          }

          int beg = 0;
          while ( beg < valStr.length() &&
                  valStr.charAt( beg ) == '{' ) beg++;
          int end = valStr.length();
          while ( end > 0 &&
                  valStr.charAt( end - 1 ) == '}' ) end--;

          String[] split;
          if ( end > beg )
            split = valStr.substring( beg, end ).split( "," );
          else
            split = new String[ 0 ];

          final int len = split.length;
          int ii;
          for ( ii = 0; ii < len; ii++ )
            split[ii] = split[ii].trim();

          switch ( chType )
          {
            case 'b':
            case 's':
            case 'i':
            case 'j':
            {
              long[] rl = new long[ len ];
              for ( ii = 0; ii < len; ii++ )
              {
                try
                {
                  rl[ii] = Long.decode( split[ii] );
                }
                catch ( NumberFormatException e )
                {
                  error( "error in long format, 0L will be substituted" );
                  rl[ii] = 0L;
                }
              }
              switch ( chType )
              {
                case 'b':
                {
                  byte[] rb = new byte[ len ];
                  for ( ii = 0; ii < len; ii++ )
                    rb[ii] = (byte)rl[ii];
                  finish( rb );
                  return;
                }
                case 's':
                {
                  short[] rs = new short[ len ];
                  for ( ii = 0; ii < len; ii++ )
                    rs[ii] = (short)rl[ii];
                  finish( rs );
                  return;
                }
                case 'i':
                {
                  int[] ri = new int[ len ];
                  for ( ii = 0; ii < len; ii++ )
                    ri[ii] = (int)rl[ii];
                  finish( ri );
                  return;
                }
                default:
                  finish( rl );
              }
              return;
            }

            case 'f':
            case 'd':
            {
              double[] rd = new double[ len ];
              for ( ii = 0; ii < len; ii++ )
              {
                try
                {
                  rd[ii] = Double.valueOf( split[ii] );
                }
                catch ( NumberFormatException e )
                {
                  error( "error in double format, 0.0 will be substituted" );
                  rd[ii] = 0.0;
                }
              }
              if ( chType == 'f' )
              {
                float[] rf = new float[ len ];
                for ( ii = 0; ii < len; ii++ )
                  rf[ii] = (float)rd[ii];
                finish( rf );
                return;
              }
              finish( rd );
              return;
            }
          }
          return;
        }
      }
    }
  }

  //--------------------------------------------------------------------------

  /**
   * Defines literal (named) object: atomic type, 1D array..
   */
  protected class DefState extends ArgParseState
  {
    /** Object id (unique in the whole script, can be {@code null} for anonymous object). */
    public String id;

    public DefState ( Attributes attr )
    {
      super( STATE_REF, attr );

      if ( attr != null )
        id = attr.getValue( A_ID );
    }

    /** Finishes element processing (overridable method). */
    @Override
    protected void finish ( Object obj )
    {
      if ( id != null )
        newObject( id, obj );
      parent.setValue( STATE_REF, obj );
    }
  }

  //==========================================================================
  //  Script parsing:

  /**
   * Read and interprets XML script.
   *
   * @param  scriptFileName File-name to interpret.
   * @return Output (result) object map.
   */
  public Map< String, Object > interpret ( String scriptFileName )
  {
    return interpret( scriptFileName, null );
  }

  /**
   * Read and interprets XML script.
   *
   * @param  scriptFileName File-name to interpret.
   * @param  input Initial object-map (can be {@code null}, if not, it will be altered!).
   * @return Output (result) object map.
   */
  public synchronized Map< String, Object > interpret ( String scriptFileName, Map< String, Object > input )
  {
    if ( scriptFileName == null )
      return null;

    LogFile.debug( "Reading XML script: " + scriptFileName );

      // XML parser:
    SAXParser parser;
    try
    {
      parser = SAXParserFactory.newInstance().newSAXParser();
    }
    catch ( ParserConfigurationException e )
    {
      LogFile.exception( "SAX parser configuration error", e );
      return null;
    }
    catch ( SAXException e )
    {
      LogFile.exception( e );
      return null;
    }

      // init global parser data:
    this.scriptFileName = scriptFileName;
    objects = (input != null) ? input : new HashMap< String, Object > ();
    results = new HashMap< String, Object > ();
    lineNo  = 2;
    tos     = null;
    new InitState();

      // parse the script file:
    try
    {
      parser.parse( scriptFileName, this );
    }
    catch ( SAXException e )
    {
      LogFile.exception( "Error parsing script file '" + scriptFileName + "' on line " + lineNo + '!', e );
    }
    catch ( IOException e )
    {
      LogFile.exception( "Error reading the '" + scriptFileName + "' file on line " + lineNo + '!', e );
    }
    finally
    {
      LogFile.debug( "Finished XML script: " + scriptFileName );
    }

    return results;
  }

  //==========================================================================
  //  XML parsing handler:

  @Override
  public void characters ( char[] ch, int start, int length )
  throws SAXException
  {
    if ( ch  == null ) return;

    //LogFile.debug( "Chars at line " + lineNo + ": #" + new String( ch, start, length ) + '#' );

    tos.characters( ch, start, length );

    while ( length-- > 0 )
      if ( ch[start++] == '\n' ) lineNo++;
  }

  @Override
  public void startElement ( String uri, String localName, String qName, Attributes atts )
  throws SAXException
  {
    //LogFile.debug( "XML line: " + lineNo + ", start <" + qName + '>' );

    tos.startElement( qName, atts );
  }

  @Override
  public void endElement ( String uri, String localName, String qName )
  throws SAXException
  {
    //LogFile.debug( "XML line: " + lineNo + ", end <" + qName + '>' );

    tos.endElement( qName );
    tos.pop();
  }

  //==========================================================================
  //  Selftest:

  public static void main ( String[] args )
  {
    if ( args.length < 1 ) return;
    LogFile.debugging = true;

    JGLScript script = new JGLScript();
    Map< String, Object > results = script.interpret( args[0] );

    if ( results != null )
      for ( String key : results.keySet() )
      {
        Object o = results.get( key );
        LogFile.log( "Key: '" + key + "', class: " + o.getClass().getName() + ", value: '" + o.toString() + "'" );
      }
  }

}
